/*
 * BusyBox' main applet dispatcher.
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */
#ifdef APPLET_USAGE
# ifndef CONFIG_INSTALL_NO_USR
#  define D_USR_USAGE "(/usr)"
# else
#  define D_USR_USAGE ""
# endif
#define busybox_trivial_usage \
	"busybox [function] [arguments]..." \
	USE_FEATURE_INSTALLER("\n   or: --install [-s] [-l] [-r]")
#define busybox_full_usage \
	"\n\tBusyBox is a multi-call binary that combines many common Unix\n" \
	"\tutilities into a single executable.  Most people will create a\n"  \
	"\tlink to busybox for each function they wish to use and BusyBox\n" \
	"\twill act like whatever it was invoked as!" \
	USE_FEATURE_INSTALLER("\nInstall options:\n") \
	USE_FEATURE_INSTALLER("\t-s\tUse symbolic link vs hardlink\n") \
	USE_FEATURE_INSTALLER("\t-l\tMake link to current directory vs " D_USR_USAGE "/(s)bin\n") \
	USE_FEATURE_INSTALLER("\t-r\tRemove exting link")
#else

#include "busybox.h"
#include <string.h>
#include <unistd.h>
# if ENABLE_LOCALE_SUPPORT
#  include <locale.h>
# else
#  define setlocale(x,y)
# endif

# ifdef __BIONIC__
#  define CONFIG_INSTALL_NO_USR 1
# endif

#ifndef BB_EXTRA_VERSION
#define BANNER "BusyBox v" BB_VER " (" BB_BT ")"
#else
#define BANNER "BusyBox v" BB_VER " (" BB_EXTRA_VERSION ")"
#endif

const char BB_BANNER[]=BANNER;
const char bb_msg_full_version[] = BANNER " multi-call binary";

const char usage_messages[] =

#define MAKE_USAGE
#include "usage.h"

#include "applets.h"

;

#undef MAKE_USAGE
#undef APPLET
#undef APPLET_NOUSAGE
#undef PROTOTYPES

enum Location {
	_BB_DIR_ROOT = 0,
	_BB_DIR_BIN,
	_BB_DIR_SBIN,
#ifndef CONFIG_INSTALL_NO_USR
	_BB_DIR_USR_BIN,
	_BB_DIR_USR_SBIN
#endif
};

enum SUIDRoot {
	_BB_SUID_NEVER = 8,
	_BB_SUID_MAYBE = 16,
	_BB_SUID_ALWAYS = 32
};

#define  Location_mask (~(8|16|32))


#include "applets.h"


/* The -1 arises because of the {0,NULL,0,-1} entry above. */
static const size_t NUM_APPLETS =
			(sizeof (applets) / sizeof (struct BB_applet) - 1);


struct BB_COMMON_BSS {
	char bb_fake_common_bss[CALC_BSS+sizeof(char *)]; /* max and min common bss size */
} bb_common_bss;


#ifdef CONFIG_FEATURE_INSTALLER
#include "opt_gen.h"
OPT_GEN(
opt(s,BB_INST_FLAGS__USE_SYMBOLIC_LINKS),
opt(l,BB_INST_FLAGS__LINK_TO_CURRENT_DIR),
opt(r,BB_INST_FLAGS__REMOVE_OLD_LINK)
)
#include "opt_gen.h"
/*
 * directory table
 *              this should be consistent w/ the enum, busybox.h::Location,
 *              or else...
 */
#ifndef __BIONIC__
static const char usr_bin [] ="/usr/bin";
static const char usr_sbin[] ="/usr/sbin";

static const char* const install_dir[] = {
	&usr_bin [8], /* "", equivalent to "/" for concat_path_file() */
	&usr_bin [4], /* "/bin" */
	&usr_sbin[4], /* "/sbin" */
# ifndef CONFIG_INSTALL_NO_USR
	usr_bin,
	usr_sbin
# endif
};
#else
static const char bionic_bin[] = "/system/bin";
static const char* const install_dir[] = {
	&bionic_bin [11], /* "", equivalent to "/" for concat_path_file() */
	bionic_bin,
	bionic_bin        /* sbin = bin */
};
#endif

/* abstract link() */
typedef int (*__link_f)(const char *, const char *);


/* create (sym)links for each applet */
static void install_links(const char *busybox, int flags)
{
	__link_f Link =
		(flags & BB_INST_FLAGS__USE_SYMBOLIC_LINKS) ? symlink : link;
	char *fpc;
	int i;
	int rc;

	for (i = 0; applets[i].name != NULL; i++) {
		if((flags & BB_INST_FLAGS__LINK_TO_CURRENT_DIR) == 0) {
			fpc = concat_path_file(
				install_dir[applets[i].Loc_SUID & Location_mask],
				applets[i].name);
		} else {
			fpc = (char *)applets[i].name;
		}
		rc = Link(busybox, fpc);
		if(rc != 0 && errno == EEXIST &&
				(flags & BB_INST_FLAGS__REMOVE_OLD_LINK)) {
			if(strcmp(applets[i].name, "busybox")) {
				if(unlink(fpc))
					bb_perror_msg("remove %s", fpc);
				else
					rc = Link(busybox, fpc);
			}
		}
		if (rc != 0 && errno != EEXIST)
			bb_perror_msg("%s", fpc);
		if((flags & BB_INST_FLAGS__LINK_TO_CURRENT_DIR) == 0)
			free(fpc);
	}
}

#else
#define install_links(x,y)
#endif /* CONFIG_FEATURE_INSTALLER */

#ifdef CONFIG_FEATURE_SUID
static void check_suid (const struct BB_applet *applet)
{
  uid_t ruid = getuid ();               /* real [ug]id */
  uid_t rgid = getgid ();

  if ((applet->Loc_SUID & _BB_SUID_ALWAYS) == _BB_SUID_ALWAYS) {
	if (geteuid () != 0)
	  bb_error_msg_and_die ("This applet requires root privileges!");
  } else if ((applet->Loc_SUID & _BB_SUID_NEVER) == _BB_SUID_NEVER) {
	/* drop all privileges */
#ifndef __ANDROID_API__
	setgid (rgid);
	setuid (ruid);
#else
	/* android protect use setgid/setuid */
	if(geteuid () != ruid)
		setgid (rgid);
	if(getegid () != rgid)
		setuid (ruid);
#endif
  }
}
#else
#define check_suid(x)
#endif /* CONFIG_FEATURE_SUID */

#ifndef CONFIG_FEATURE_SH_STANDALONE_SHELL
static
#endif
void show_applets(void)
{
	const struct BB_applet *a;
	int col, output_width;

	if (ENABLE_FEATURE_AUTOWIDTH) {
		/* Obtain the terminal width.  */
		get_terminal_width_height(&output_width, NULL);
		/* leading tab and room to wrap */
		output_width -= 20;
	} else
		output_width = 60;

	col=0;
	for(a = applets; a->name;) {
		int t = col == 0 ? 7 : 0;
		col += printf("%s%s", (col ? ", " : "\t"), (a++)->name);
		col += t;
		if (col > output_width && a->name) {
			printf(",\n");
			col = 0;
		}
	}
	printf("\n\n");
}

static const struct BB_applet *applet_using;

static void bb_show_usage_int(int help) ATTRIBUTE_NORETURN;
static void bb_show_usage_int(int help)
{
	const char *format_string;
	const char *usage_string = usage_messages;
	int i;

	if(bb_default_error_retval == EXIT_FAILURE)
		bb_default_error_retval = help ? 0 : 2;
	for (i = applet_using - applets; i > 0;) {
		if (!*usage_string++) {
			--i;
		}
	}

	format_string = "%s\n\nUsage: %s %s\n\n";
	if(*usage_string == '-' && usage_string[1] >= 'A' && usage_string[1] <= 'z'
				&& usage_string[2] == '\n' && usage_string[3] == '\n' &&
				usage_string[4] == '\0') {
		char *h = xstrdup(usage_string);
		char *argv[3];

		argv[0] = Gargv[0];
		argv[1] = h;
		argv[2] = NULL;
		Gargv = argv;
		Gargc = 2;
		(*(applet_using->main)) ();
	} else {
		if (*usage_string == '\b')
			format_string = "%s\n\nNo help available.\n\n";
		fprintf ((help ? stdout : stderr),
			  format_string, bb_msg_full_version, applet_using->name, usage_string);
		if(strcmp(bb_applet_name, "busybox") == 0) {
			printf("\nCurrently defined functions:\n");
			show_applets();
		}
	}
	exit (bb_default_error_retval);
}

void bb_show_usage (void)
{
	bb_show_usage_int(0);
}

static int applet_name_compare (const void *x, const void *y)
{
  const char *name = x;
  const struct BB_applet *applet = y;

  return strcmp (name, applet->name);
}

struct BB_applet *find_applet_by_name (const char *name)
{
  return bsearch (name, applets, NUM_APPLETS, sizeof (struct BB_applet),
				  applet_name_compare);
}

static void run_applet_by_name (const char *name, int argc, char **argv) ATTRIBUTE_NORETURN;
static void run_applet_by_name (const char *name, int argc, char **argv)
{
	Gargc = argc;
	Gargv = argv;

	/* Do a binary search to find the applet entry given the name. */
	if(strncmp(name, "busybox", 7) == 0) {
		/* can call as busybox-new for example */
		if(argv[1] != NULL && argv[1][0] != '-') {
			/* call as busybox[any] applet ... */
			run_applet_by_name(argv[1], argc-1, argv+1);
		}
		/* after recursion we have "busybox [-options]" */
		name = "busybox";
	}
	applet_using = find_applet_by_name(name);
	if(applet_using) {
		bb_applet_name = applet_using->name;
		if(argc==2 && !strcmp(argv[1], "--help")) bb_show_usage_int(2);
		if(ENABLE_FEATURE_SUID) check_suid(applet_using);
		exit((*(applet_using->main)) ());
	}
	bb_applet_name = name;
	bb_error_msg_and_die("applet not found");
}


int main(int argc, char **argv)
{
	libbb_init(argc, argv);
	/* Set locale for everybody except `init' */
	if(ENABLE_LOCALE_SUPPORT && getpid() != 1)
		setlocale(LC_ALL, "");
	run_applet_by_name(bb_applet_name, argc, argv);
}

int busybox_main(void)
{
	int argc = Gargc;
	char **argv = Gargv;

	if (argc < 2)
		bb_show_usage_int(2);

#ifdef CONFIG_FEATURE_INSTALLER
	if (!strcmp(argv[1], "--install")) {
		int i;
		char *busybox;

		/* make opts for parse */
		for(i = 1; ; i++) {
			argv[i] = argv[i+1];
			if(argv[i] == NULL)
				break;
		}
		Gargc--;
		i = bb_getopt_ulflags(opt_str);
		/* link */
		busybox = xreadlink("/proc/self/exe");
		if (busybox) {
			install_links(busybox, i);
			free(busybox);
			return 0;
		} else {
			return 1;
		}
	}
#endif

	if ((argv[1][0] == '-' && strcmp(argv[1], "--help")) || argc == 2)
		bb_error_msg_and_die("unknown option %s", argv[1]);
	run_applet_by_name(argv[2], 2, argv);
}
#endif /* APPLET_USAGE */
