/*
 * Mini ar implementation for busybox
 *
 * Copyright (C) 2000 by Glenn McGrath
 * Written by Glenn McGrath <bug1@iinet.net.au> 1 June 2000
 *  Licensed under GPL v2, see file LICENSE in this tarball for details.
 *
 * Based in part on BusyBox tar, Debian dpkg-deb and GNU ar.
 *
 * There is no single standard to adhere to so ar may not portable
 * between different systems
 * http://www.unix-systems.org/single_unix_specification_v2/xcu/ar.html
 */
#ifdef APPLET_USAGE
#define ar_trivial_usage \
	"[-o] [-v] [-p] [-t] [-x] ARCHIVE FILES"
#define ar_full_usage \
	"Extract or list FILES from an ar archive.\n\n" \
	"Options:\n" \
	"\t-o\t\tpreserve original dates\n" \
	"\t-p\t\textract to stdout\n" \
	"\t-t\t\tlist\n" \
	"\t-x\t\textract\n" \
	"\t-v\t\tverbosely list files processed"
#else

#define COMPLEMENTARY "\0--:p:t:x:-1:?p--tx:t--px:x--pt"
#include "opt_gen.h"
OPT_GEN(
opt(p,AR_CTX_PRINT),
opt(t,AR_CTX_LIST),
opt(x,AR_CTX_EXTRACT),
opt(o,AR_OPT_PRESERVE_DATE),
opt(v,AR_OPT_VERBOSE),
opt(c,AR_OPT_CREATE),
opt(r,AR_OPT_INSERT)
)
#include "opt_gen.h"

#include "busybox.h"
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "unarchive.h"
#include "bb_xopen.h"

static void header_verbose_list_ar(const file_header_t *file_header)
{
	const char *mode;
	char *mtime;
	char buf[12];

	mode = bb_mode_string(file_header->mode, buf);
	mtime = ctime(&file_header->mtime);
	mtime[16] = ' ';
	memmove(&mtime[17], &mtime[20], 4);
	mtime[21] = '\0';
	printf("%s %d/%d%7d %s %s\n", &mode[1], file_header->uid, file_header->gid,
			(int) file_header->size, &mtime[4], file_header->name);
}


extern int ar_main(void)
{
	char **argv;
	archive_handle_t *archive_handle;
	unsigned opt;
	static const char msg_unsupported_err[] =
			"Archive %s not supported.  Install binutils 'ar'.";
	char magic[8];

	archive_handle = init_handle();

	/* Prepend '-' to the first argument if required */
	opt = bb_xgetopt_ulflags(opt_str);

	if (opt & AR_CTX_PRINT) {
		archive_handle->action_data = data_extract_to_stdout;
	}
	if (opt & AR_CTX_LIST) {
		archive_handle->action_header = header_list;
	}
	if (opt & AR_CTX_EXTRACT) {
		archive_handle->action_data = data_extract_all;
	}
	if (opt & AR_OPT_PRESERVE_DATE) {
		archive_handle->flags |= ARCHIVE_PRESERVE_DATE;
	}
	if (opt & AR_OPT_VERBOSE) {
		archive_handle->action_header = header_verbose_list_ar;
	}
	if (opt & AR_OPT_CREATE)
		bb_error_msg_and_die(msg_unsupported_err, "creation");
	if (opt & AR_OPT_INSERT)
		bb_error_msg_and_die(msg_unsupported_err, "insertion");
	argv = Gargv;
	archive_handle->src_fd = bb_xopen(*argv, O_RDONLY | BB_OPEN_ERR);
	argv++;

	while (*argv != NULL) {
		archive_handle->filter = filter_accept_list;
		archive_handle->accept = llist_add_to(archive_handle->accept, *argv);
		argv++;
	}

	archive_xread_all(archive_handle, magic, 7);
	if (strncmp(magic, "!<arch>", 7) != 0)
		bb_error_msg_and_die("invalid ar magic");
	archive_handle->offset += 7;

	while (get_header_ar(archive_handle) == EXIT_SUCCESS);

	return EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
