/*
 * Mini cpio implementation for busybox
 *
 * Copyright (C) 2001 by Glenn McGrath
 *  Licensed under GPL v2, see file LICENSE in this tarball for details.
 *
 * Limitations:
 *              Doesn't check CRC's
 *              Only supports new ASCII and CRC formats
 *
 */
#ifdef APPLET_USAGE
#define cpio_trivial_usage \
	"-[dimtuv][F cpiofile]"
#define cpio_full_usage \
	"Extract or list files from a cpio archive\n" \
	"Main operation mode:\n" \
	"\td\t\tmake leading directories\n" \
	"\ti\t\textract\n" \
	"\tm\t\tpreserve mtime\n" \
	"\tt\t\tlist\n" \
	"\tv\t\tverbose\n" \
	"\tu\t\tunconditional overwrite\n" \
	"\tF\t\tinput from file"
#else

#define COMPLEMENTARY "\0?i--t:t--i"
#include "opt_gen.h"
OPT_GEN(
opt(i,CPIO_OPT_EXTRACT),
opt(t,CPIO_OPT_TEST),
opt(u,CPIO_OPT_UNCONDITIONAL),
opt(v,CPIO_OPT_VERBOSE),
opt(F,CPIO_OPT_FILE, cpio_filename),
opt(d,CPIO_OPT_CREATE_LEADING_DIR),
opt(m,CPIO_OPT_PRESERVE_MTIME)
)
#include "opt_gen.h"

#include "busybox.h"
#include <string.h>
#include <unistd.h>
#include "unarchive.h"
#include "bb_xopen.h"


extern int cpio_main(void)
{
	char **argv;
	archive_handle_t *archive_handle;
	char *cpio_filename = NULL;
	int opt;

	/* Initialise */
	archive_handle = init_handle();
	archive_handle->src_fd = STDIN_FILENO;
	archive_handle->seek = seek_by_char;
	archive_handle->flags = ARCHIVE_EXTRACT_NEWER | ARCHIVE_PRESERVE_DATE;

	/* One of either extract or test options must be given */
	opt = bb_xgetopt_ulflags(opt_str, &cpio_filename);

	if (opt & CPIO_OPT_EXTRACT) {
		archive_handle->action_data = data_extract_all;
	}
	if (opt & CPIO_OPT_TEST) {
		archive_handle->action_header = header_list;
	}
	if (opt & CPIO_OPT_UNCONDITIONAL) {
		archive_handle->flags |= ARCHIVE_EXTRACT_UNCONDITIONAL;
		archive_handle->flags &= ~ARCHIVE_EXTRACT_NEWER;
	}
	if (opt & CPIO_OPT_VERBOSE) {
		if (archive_handle->action_header == header_list) {
			archive_handle->action_header = header_verbose_list;
		} else {
			archive_handle->action_header = header_list;
		}
	}
	if (cpio_filename) { /* CPIO_OPT_FILE */
		archive_handle->src_fd = bb_xopen(cpio_filename, O_RDONLY | BB_OPEN_ERR);
		archive_handle->seek = seek_by_jump;
	}
	if (opt & CPIO_OPT_CREATE_LEADING_DIR) {
		archive_handle->flags |= ARCHIVE_CREATE_LEADING_DIRS;
	}

	for (argv = Gargv; *argv != NULL; argv++) {
		archive_handle->filter = filter_accept_list;
		archive_handle->accept = llist_add_to(archive_handle->accept, *argv);
	}

	while (get_header_cpio(archive_handle) == EXIT_SUCCESS);

	return EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
