/*
 *  dpkg-deb packs, unpacks and provides information about Debian archives.
 *
 *  Licensed under GPL v2, see file LICENSE in this tarball for details.
 */
#ifdef APPLET_USAGE
#define dpkg_deb_trivial_usage \
	"[-cefxX] FILE [argument]"
#define dpkg_deb_full_usage \
	"Perform actions on Debian packages (.debs)\n\n" \
	"Options:\n" \
	"\t-c\tList contents of filesystem tree\n" \
	"\t-e\tExtract control files to [argument] directory\n" \
	"\t-f\tDisplay control field name starting with [argument]\n" \
	"\t-x\tExtract packages filesystem tree to directory\n" \
	"\t-X\tVerbose extract"
#else

#include "busybox.h"
#include <string.h>
#include <unistd.h>

#include "unarchive.h"
#include "bb_xopen.h"

#define DPKG_DEB_OPT_CONTENTS   1
#define DPKG_DEB_OPT_CONTROL    2
#define DPKG_DEB_OPT_FIELD      4
#define DPKG_DEB_OPT_EXTRACT    8
#define DPKG_DEB_OPT_EXTRACT_VERBOSE    16

extern int dpkg_deb_main(void)
{
	char **argv;
	archive_handle_t *ar_archive;
	archive_handle_t *tar_archive;
	llist_t *control_tar_llist = NULL;
	int opt;
	char *extract_dir = NULL;
	int need_dir = 0;

	/* Setup the tar archive handle */
	tar_archive = init_handle();

	/* Setup an ar archive handle that refers to the gzip sub archive */
	ar_archive = init_handle();
	ar_archive->sub_archive = tar_archive;
	ar_archive->filter = filter_accept_list_reassign;

#ifdef CONFIG_FEATURE_DEB_TAR_GZ
	ar_archive->accept = llist_add_to(NULL, "data.tar.gz");
	control_tar_llist = llist_add_to(NULL, "control.tar.gz");
#endif

#ifdef CONFIG_FEATURE_DEB_TAR_BZ2
	ar_archive->accept = llist_add_to(ar_archive->accept, "data.tar.bz2");
	control_tar_llist = llist_add_to(control_tar_llist, "control.tar.bz2");
#endif

	opt = bb_xgetopt_ulflags("cefXx\0-1:?c--efXx:e--cfXx:f--ceXx:X--cefx:x--cefX");
	argv = Gargv;

	if (opt & DPKG_DEB_OPT_CONTENTS) {
		tar_archive->action_header = header_verbose_list;
	}
	if (opt & DPKG_DEB_OPT_CONTROL) {
		ar_archive->accept = control_tar_llist;
		tar_archive->action_data = data_extract_all;
		if (argv[1] == NULL) {
			extract_dir = "./DEBIAN";
		} else {
			need_dir++;
		}
	}
	if (opt & DPKG_DEB_OPT_FIELD) {
		/* Print the entire control file
		 * it should accept a second argument which specifies a
		 * specific field to print */
		ar_archive->accept = control_tar_llist;
		tar_archive->accept = llist_add_to(NULL, "./control");
		tar_archive->filter = filter_accept_list;
		tar_archive->action_data = data_extract_to_stdout;
	}
	if (opt & DPKG_DEB_OPT_EXTRACT) {
		tar_archive->action_header = header_list;
	}
	if (opt & (DPKG_DEB_OPT_EXTRACT_VERBOSE | DPKG_DEB_OPT_EXTRACT)) {
		tar_archive->action_data = data_extract_all;
		need_dir++;
	}

	tar_archive->src_fd = ar_archive->src_fd = bb_xopen(*argv, O_RDONLY | BB_OPEN_ERR);
	argv++;

	/* Workout where to extract the files */
	/* 2nd argument is a dir name */
	if(*argv)
		extract_dir = *argv;
	else if(need_dir)
		bb_show_usage();

	if (extract_dir) {
		if(mkdir(extract_dir, 0777))
		     bb_perror_msg_and_die("mkdir %s", extract_dir);
		xchdir(extract_dir);
	}
	unpack_ar_archive(ar_archive);

	return EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
