/*
 * Gzip implementation for busybox
 *
 * Based on GNU gzip v1.2.4 Copyright (C) 1992-1993 Jean-loup Gailly.
 * Originally adjusted for busybox by Sven Rudolph <sr1@inf.tu-dresden.de>
 * based on gzip sources
 *
 * Adjusted further by Erik Andersen <andersen@codepoet.org> to support files as
 * well as stdin/stdout, and to generally behave itself wrt command line
 * handling.
 *
 * General cleanup to better adhere to the style guide and make use of standard
 * busybox functions by Glenn McGrath <bug1@iinet.net.au>
 *
 *  Licensed under GPL v2, see file LICENSE in this tarball for details.
 *
 * Common code for uncompress/bunzip2/unlzma
 * Copyright Vladimir Oleynik <dzo@simtreas.ru> (c) 2016
 */
#ifdef APPLET_USAGE
#define gunzip_trivial_usage \
	"[OPTION]... FILE"
#define gunzip_full_usage \
	"Uncompress FILE (or standard input if FILE is '-').\n\n" \
	"Options:\n" \
	"\t-c\tWrite output to standard output\n" \
	"\t-f\tForce read when source is a terminal\n" \
	"\t-t\tTest compressed file integrity"
#define zcat_trivial_usage \
	"FILE"
#define zcat_full_usage \
	"Uncompress to stdout."
#else

#include "busybox.h"
#include "unarchive.h"

static int chk_header_inflate_gunzip(int in, int out)
{
	if(check_header_gzip(in, 0))
		return -1;
	return inflate_gunzip(in, out);
}

extern int gunzip_main(void)
{
	return unx_main(NULL, bb_applet_name[0] == 'z', chk_header_inflate_gunzip, "\x1f\x8b");
}
#endif /* APPLET_USAGE */
