/*
 * Mini dumpkmap implementation for busybox
 *
 * Copyright (C) Arne Bernin <arne@matrix.loopback.org>
 *
 * Licensed under the GPL v2 or later, see the file LICENSE in this tarball.
 */
#ifdef APPLET_USAGE
#define dumpkmap_trivial_usage \
	"> keymap"
#define dumpkmap_full_usage \
	"Prints out a binary keyboard translation table to standard output."
#else

#include "busybox.h"
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>

/* From <linux/kd.h> */
struct kbentry {
	unsigned char kb_table;
	unsigned char kb_index;
	unsigned short kb_value;
};
#define KDGKBENT 0x4B46  /* gets one entry in translation table */

/* From <linux/keyboard.h> */
#define NR_KEYS 128
#define MAX_NR_KEYMAPS 256

CALC_BSS struct BB_COMMON_BSS {
	struct kbentry ke;
	char flags_d[MAX_NR_KEYMAPS];
} bb_common_bss;
#define flags_d bb_common_bss.flags_d
#define ke bb_common_bss.ke


int dumpkmap_main(void)
{
	int i, j, fd;
	static const char magic[] = "bkeymap";

	bb_xgetopt_ulflags("\0?0");

	fd = get_console_fd_and_die();

	write(1, magic, 7);

	/* Here we want to set everything to 0 except for indexes:
	 * [0-2] [4-6] [8-10] [12] */
	memset(flags_d, 0x00, MAX_NR_KEYMAPS);
	memset(flags_d, 0x01, 13);
	flags_d[3] = flags_d[7] = flags_d[11] = 0;

	/* dump flags_d */
	write(1, flags_d, MAX_NR_KEYMAPS);

	for (i = 0; i < MAX_NR_KEYMAPS; i++) {
		if (flags_d[i] == 1) {
			for (j = 0; j < NR_KEYS; j++) {
				ke.kb_index = j;
				ke.kb_table = i;
				if (xioctl(fd, KDGKBENT, &ke, "ioctl failed with %s, %s, %p",
						(char *)&ke.kb_index,
						(char *)&ke.kb_table,
						&ke.kb_value) >= 0)
					write(1, (void*)&ke.kb_value, 2);
			}
		}
	}
	/* Don't bother to close files.  Exit does that
	 * automagically, so we can save a few bytes */
	/* close(fd); */
	return EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
