/*
 * Mini loadkmap implementation for busybox
 *
 * Copyright (C) 1998 Enrique Zanardi <ezanardi@ull.es>
 *
 * Licensed under the GPL v2 or later, see the file LICENSE in this tarball.
 */
#ifdef APPLET_USAGE
#define loadkmap_trivial_usage \
	"< keymap"
#define loadkmap_full_usage \
	"Loads a binary keyboard translation table from standard input."
#else

#include "busybox.h"
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>

#define BINARY_KEYMAP_MAGIC "bkeymap"

/* From <linux/kd.h> */
struct kbentry {
	unsigned char kb_table;
	unsigned char kb_index;
	unsigned short kb_value;
};
/* sets one entry in translation table */
#define KDSKBENT        0x4B47

/* From <linux/keyboard.h> */
#define NR_KEYS         128
#define MAX_NR_KEYMAPS  256

CALC_BSS struct BB_COMMON_BSS
{
	struct kbentry ke;
	u_short ibuff[NR_KEYS];
	char flags_l[MAX_NR_KEYMAPS];
	char buff[7];
} bb_common_bss;
#define ibuff bb_common_bss.ibuff
#define flags_l bb_common_bss.flags_l
#define buff bb_common_bss.buff
#define ke bb_common_bss.ke


int loadkmap_main(void)
{
	int i, j, fd;

	bb_xgetopt_ulflags("\0?0");

	fd = get_console_fd_and_die();

	if ((bb_full_read(0, buff, 7) != 7) || (strncmp(buff, BINARY_KEYMAP_MAGIC, 7) != 0))
		bb_error_msg_and_die("This is not a valid binary keymap.");

	if (bb_full_read(0, flags_l, MAX_NR_KEYMAPS) != MAX_NR_KEYMAPS)
		bb_perror_msg_and_die("Error reading keymap flags_l");

	for (i = 0; i < MAX_NR_KEYMAPS; i++) {
		if (flags_l[i] == 1) {
			bb_full_read(0, ibuff, NR_KEYS * sizeof(u_short));
			for (j = 0; j < NR_KEYS; j++) {
				ke.kb_index = j;
				ke.kb_table = i;
				ke.kb_value = ibuff[j];
				ioctl(fd, KDSKBENT, &ke);
			}
		}
	}

	/* Don't bother to close files.  Exit does that
	 * automagically, so we can save a few bytes */
	/* close(fd); */
	return EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
