/*
 *  openvt.c - open a vt to run a command.
 *
 *  busyboxed by Quy Tonthat <quy@signal3.com>
 *  hacked by Tito <farmatito@tiscali.it>
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */
#ifdef APPLET_USAGE
#define openvt_trivial_usage \
	"<vtnum> <COMMAND> [ARGS...]"
#define openvt_full_usage \
	"Start a command on a new virtual terminal"
#else

/* getopt not needed */

#include "busybox.h"
#include <unistd.h>
#include <string.h>
#include "bb_xopen.h"

int openvt_main(void)
{
	char **argv = Gargv;
	int fd;
	char vtname[sizeof(VC_FORMAT) + 2];

	if (Gargc < 3)
		bb_show_usage();

	/* check for Illegal vt number: < 1 or > 12 */
	sprintf(vtname, VC_FORMAT, bb_xparse_i_number(10 | BB_PN_SET_LOWER1 | BB_PN_SET_UPPER, argv[1], 12));

	argv += 2;

	if(fork() == 0) {
		/* leave current vt */

		if (setsid() < 0)
			bb_perror_msg_and_die("Unable to set new session");
		close(0);                       /* so that new vt becomes stdin */

		/* and grab new one */
		fd = bb_xopen(vtname, O_RDWR | BB_OPEN_ERR);

		/* Reassign stdout and sterr */
		dup2(fd, STDOUT_FILENO);
		dup2(fd, STDERR_FILENO);

		execvp(argv[0], argv);
		_exit(1);
	}
	return EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
