/*
 * setkeycodes
 *
 * Copyright (C) 1994-1998 Andries E. Brouwer <aeb@cwi.nl>
 *
 * Adjusted for BusyBox by Erik Andersen <andersen@codepoet.org>
 *
 * Licensed under the GPL v2 or later, see the file LICENSE in this tarball.
 */
#ifdef APPLET_USAGE
#define setkeycodes_trivial_usage \
	"SCANCODE KEYCODE ..."
#define setkeycodes_full_usage \
	"Set entries into the kernel's scancode-to-keycode map,\n" \
	"allowing unusual keyboards to generate usable keycodes.\n\n" \
	"SCANCODE may be either xx or e0xx (hexadecimal),\n" \
	"and KEYCODE is given in decimal"
#else

#include "busybox.h"
#include <sys/ioctl.h>


/* From <linux/kd.h> */
struct kbkeycode {
	unsigned int scancode, keycode;
};
static const int KDSETKEYCODE = 0x4B4D;  /* write kernel keycode table entry */

extern int setkeycodes_main(void)
{
	int argc = Gargc;
	char **argv = Gargv;
	int fd;
	unsigned int sc;
	struct kbkeycode a;

	if (argc % 2 != 1 || argc < 2)
		bb_show_usage();

	fd = get_console_fd_and_die();

	while (argc > 2) {
		a.keycode = atoi(argv[2]);
		a.scancode = sc = bb_xparse_ui_number(16 | BB_PN_ERR_STR1,
			argv[1], "error reading SCANCODE: '%s'");
		if (a.scancode > 127) {
			a.scancode -= 0xe000;
			a.scancode += 128;
		}
		if (a.scancode > 255 || a.keycode > 127) {
			bb_error_msg_and_die("SCANCODE or KEYCODE outside bounds");
		}
		xioctl(fd,KDSETKEYCODE,&a, "\1failed to set SCANCODE %x to KEYCODE %d", sc, a.keycode);
		argc -= 2;
		argv += 2;
	}
	return EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
