/*
 * Mini chown implementation for busybox
 *
 * Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */

/* BB_AUDIT SUSv3 defects - unsupported options -h, -H, -L, and -P. */
/* BB_AUDIT GNU defects - unsupported options -h, -c, -f, -v, and long options. */
/* BB_AUDIT Note: gnu chown does not support -H, -L, or -P. */
/* http://www.opengroup.org/onlinepubs/007904975/utilities/chown.html */
#ifdef APPLET_USAGE
#define chown_trivial_usage \
	"[ -Rh ]... OWNER[<.|:>[GROUP]] FILE..."
#define chown_full_usage \
	"Change the owner and/or group of each FILE to OWNER and/or GROUP.\n" \
	"\nOptions:\n" \
	"\t-R\tChanges files and directories recursively\n" \
	"\t-h\tDo not dereference symbolic links"
#else

#define COMPLEMENTARY "\0-2" /* min 2 args */
#include "opt_gen.h"
OPT_GEN(opt_simple(R), opt_simple(h))
#include "opt_gen.h"

#include "busybox.h"
#include <unistd.h>
#include <string.h>


extern struct BB_COMMON_BSS
{
	int (*chown_func)(const char *, uid_t, gid_t);
	uid_t uid;
	gid_t gid;
} bb_common_bss;

static int fileAction(const char *fileName, struct stat *statbuf,
		void ATTRIBUTE_UNUSED *junk)
{
	if (!bb_common_bss.chown_func(fileName,
			(bb_common_bss.uid == (uid_t)-1) ? statbuf->st_uid : bb_common_bss.uid,
			(bb_common_bss.gid == (gid_t)-1) ? statbuf->st_gid : bb_common_bss.gid)) {
		return TRUE;
	}
	bb_perror_msg("%s", fileName);  /* A filename could have % in it... */
	return FALSE;
}


int chown_main(void)
{
	char **argv;
	int flags;
	int retval = EXIT_SUCCESS;
	char *groupName;


	flags = bb_xgetopt_ulflags(opt_str);

	bb_common_bss.chown_func = flags & OPT_h ? lchown : chown;

	argv = Gargv;

	/* First, check if there is a group name here */
	if ((groupName = strchr_set0(*argv, '.')) == NULL)
		groupName = strchr_set0(*argv, ':');

	/* Check for the username and groupname */
	if (groupName)
		bb_common_bss.gid = get_ug_id(groupName + 1, bb_xgetgrnam);
	else
		bb_common_bss.gid = -1;
	if (groupName != *argv)
		bb_common_bss.uid = get_ug_id(*argv, bb_xgetpwnam);
	else
		bb_common_bss.uid = -1;

	++argv;
	/* Ok, ready to do the deed now */
	do {
		if (! recursive_action (*argv, (flags & OPT_R), FALSE, FALSE,
					    fileAction, fileAction, NULL)) {
			retval = EXIT_FAILURE;
		}
	} while (*++argv);

	return retval;
}
#endif /* APPLET_USAGE */
