/*-
 * Copyright (c) 1990 The Regents of the University of California.
 * All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Michael Rendell of the Memorial University of Newfoundland.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Wed Jun 22 22:15:41 1994, faith@cs.unc.edu: Added internationalization
 *                           patches from Andries.Brouwer@cwi.nl
 * Wed Sep 14 22:31:17 1994: patches from Carl Christofferson
 *                           (cchris@connected.com)
 * 1999-02-22 Arkadiusz Miśkiewicz <misiek@pld.ORG.PL>
 *      added Native Language Support
 * 1999-09-19 Bruno Haible <haible@clisp.cons.org>
 *      modified to work correctly in multi-byte locales
 *
 */
#ifdef APPLET_USAGE
#define col_trivial_usage \
	"[-bfphxV] [-l NUM_LINES]"
#define col_full_usage \
	"Filter reverse line feeds from input.\n" \
	"\nOptions:\n" \
	"\t-b, --no-backspaces\tdo not output backspaces\n"  \
	"\t-f, --fine\t\tpermit forward half line feeds\n"  \
	"\t-p, --pass\t\tpass unknown control sequences\n" \
	"\t-h, --tabs\t\tconvert spaces to tabs\n"  \
	"\t-x, --spaces\t\tconvert tabs to spaces\n"  \
	"\t-l, --lines NUM\t\tbuffer at least NUM lines\n"  \
	"\t-V, --version\toutput version information and exit"  \

#else

#define COMPLEMENTARY "\0?0:h-x:x-h"
#include "opt_gen.h"
OPT_GEN(
opt(b, NO_BACKSPACES_OPT),
opt(f, FINE_OPT),
opt(h, COMPRESS_SPACES_OPT),
opt(p, PASS_UNKNOWN_SEQS_OPT),
opt(x, NO_COMPRESS_SPACES_OPT),
opt(V, VERSION_OPT),
opt(l, LINES_OPT,lines)
)
#include "opt_gen.h"


#include "busybox.h"
#include <string.h>
#include <unistd.h>

#ifdef CONFIG_FEATURE_UTF8_SUPPORT

# include <wchar.h>
# include <wctype.h>

#else /* !HAVE_WIDECHAR */

# include <ctype.h>
  /* Fallback for types */
# define wchar_t char
# define wint_t int
# define WEOF EOF
# define getwchar getchar
# define putwchar putchar
# define iswgraph isgraph
# define iswspace isspace
# define wcwidth(c) 1

#endif /* CONFIG_FEATURE_UTF8_SUPPORT */

#define BS      '\b'            /* backspace */
#define TAB     '\t'            /* tab */
#define SPACE   ' '             /* space */
#define NL      '\n'            /* newline */
#define CR      '\r'            /* carriage return */
#define ESC     '\033'          /* escape */
#define SI      '\017'          /* shift in to normal character set */
#define SO      '\016'          /* shift out to alternate character set */
#define VT      '\013'          /* vertical tab (aka reverse line feed) */
#define RLF     '\007'          /* ESC-07 reverse line feed */
#define RHLF    '\010'          /* ESC-010 reverse half-line feed */
#define FHLF    '\011'          /* ESC-011 forward half-line feed */

/* build up at least this many lines before flushing them out */
#define BUFFER_MARGIN           32

#define PACKAGE_STRING "util-linux 2.23"

typedef char CSET;

typedef struct char_str {
#define CS_NORMAL       1
#define CS_ALTERNATE    2
	short           c_column;       /* column character is in */
	CSET            c_set;          /* character set (currently only 2) */
	wchar_t         c_char;         /* character in question */
	int             c_width;        /* character width */
} CHAR;

typedef struct line_str LINE;

struct line_str {
	CHAR    *l_line;                /* characters on the line */
	LINE    *l_prev;                /* previous line */
	LINE    *l_next;                /* next line */
	int     l_lsize;                /* allocated sizeof l_line */
	int     l_line_len;             /* strlen(l_line) */
	int     l_needs_sort;           /* set if chars went in out of order */
	int     l_max_col;              /* max column in the line */
};

CALC_BSS struct BB_COMMON_BSS
{
	CSET last_set;                  /* char_set of last char printed */
	LINE *lines;
	int compress_spaces;            /* if doing space -> tab conversion */
	unsigned max_bufd_lines;        /* max # lines to keep in memory */
	int nblank_lines;               /* # blanks after last flushed line */
	LINE *line_freelist;
	int sorted_size;
	CHAR *sorted;
	int count_size;
	int *count;
} bb_common_bss;
#define last_set bb_common_bss.last_set
#define lines bb_common_bss.lines
#define compress_spaces bb_common_bss.compress_spaces
#define max_bufd_lines bb_common_bss.max_bufd_lines
#define nblank_lines bb_common_bss.nblank_lines
#define line_freelist bb_common_bss.line_freelist
#define sorted_size bb_common_bss.sorted_size
#define sorted bb_common_bss.sorted
#define count_size bb_common_bss.count_size
#define count bb_common_bss.count

#define PUTC(ch) \
	if (putwchar(ch) == WEOF) \
		bb_error_msg_and_die(bb_msg_write_error);

static void free_line(LINE *l)
{
	l->l_next = line_freelist;
	line_freelist = l;
}


/*
 * Write a line to stdout taking care of space to tab conversion (-h flag)
 * and character set shifts.
 */
static void flush_line(LINE *l)
{
	CHAR *c, *endc;
	int nchars, last_col, this_col, i, save, tot;

	last_col = 0;
	nchars = l->l_line_len;

	if (l->l_needs_sort) {
		/*
		 * Do an O(n) sort on l->l_line by column being careful to
		 * preserve the order of characters in the same column.
		 */
		if (l->l_lsize > sorted_size) {
			sorted_size = l->l_lsize;
			sorted = (CHAR *)xrealloc((void *)sorted,
						  (unsigned)sizeof(CHAR) * sorted_size);
		}
		if (l->l_max_col >= count_size) {
			count_size = l->l_max_col + 1;
			count = (int *)xrealloc((void *)count,
			    (unsigned)sizeof(int) * count_size);
		}
		memset(count, 0, sizeof(int) * l->l_max_col + 1);
		for (i = nchars, c = l->l_line; --i >= 0; c++)
			count[c->c_column]++;

		/*
		 * calculate running total (shifted down by 1) to use as
		 * indices into new line.
		 */
		for (tot = 0, i = 0; i <= l->l_max_col; i++) {
			save = count[i];
			count[i] = tot;
			tot += save;
		}

		for (i = nchars, c = l->l_line; --i >= 0; c++)
			sorted[count[c->c_column]++] = *c;
		c = sorted;
	} else
		c = l->l_line;
	while (nchars > 0) {
		this_col = c->c_column;
		endc = c;
		do {
			++endc;
		} while (--nchars > 0 && this_col == endc->c_column);

		/* if -b only print last character */
		if ((bss_flags & NO_BACKSPACES_OPT)) {
			c = endc - 1;
			if (nchars > 0 &&
			    this_col + c->c_width > endc->c_column)
				continue;
		}

		if (this_col > last_col) {
			int nspace = this_col - last_col;

			if (compress_spaces && nspace > 1) {
				int ntabs;

				ntabs = this_col / 8 - last_col / 8;
				if (ntabs > 0) {
					nspace = this_col & 7;
					while (--ntabs >= 0)
						PUTC('\t');
				}
			}
			while (--nspace >= 0)
				PUTC(' ');
			last_col = this_col;
		}

		for (;;) {
			if (c->c_set != last_set) {
				switch (c->c_set) {
				case CS_NORMAL:
					PUTC('\017');
					break;
				case CS_ALTERNATE:
					PUTC('\016');
				}
				last_set = c->c_set;
			}
			PUTC(c->c_char);
			if ((c + 1) < endc) {
				for (i=0; i < c->c_width; i++)
					PUTC('\b');
			}
			if (++c >= endc)
				break;
		}
		last_col += (c - 1)->c_width;
	}
}

/*
 * Print a number of newline/half newlines.  If fine flag is set, nblank_lines
 * is the number of half line feeds, otherwise it is the number of whole line
 * feeds.
 */
static void flush_blanks(void)
{
	int half, i, nb;

	half = 0;
	nb = nblank_lines;
	if (nb & 1) {
		if ((bss_flags & FINE_OPT))
			half = 1;
		else
			nb++;
	}
	nb /= 2;
	for (i = nb; --i >= 0;)
		PUTC('\n');
	if (half) {
		PUTC('\033');
		PUTC('9');
		if (!nb)
			PUTC('\r');
	}
	nblank_lines = 0;
}


static void flush_lines(int nflush)
{
	LINE *l;

	while (--nflush >= 0) {
		l = lines;
		lines = l->l_next;
		if (l->l_line) {
			flush_blanks();
			flush_line(l);
		}
		nblank_lines++;
		free((void *)l->l_line);
		free_line(l);
	}
	if (lines)
		lines->l_prev = NULL;
}


#define NALLOC 64


static LINE *alloc_line(void)
{
	LINE *l;
	int i;

	if (!line_freelist) {
		l = xmalloc(sizeof(LINE) * NALLOC);
		line_freelist = l;
		for (i = 1; i < NALLOC; i++, l++)
			l->l_next = l + 1;
		l->l_next = NULL;
	}
	l = line_freelist;
	line_freelist = l->l_next;

	memset(l, 0, sizeof(LINE));
	return l;
}


int col_main(void)
{
	wint_t ch;
	CHAR *c;
	CSET cur_set;                   /* current character set */
	LINE *l;                        /* current line */
	int extra_lines;                /* # of lines above first line */
	int cur_col;                    /* current column */
	int cur_line;                   /* line number of current position */
	int max_line;                   /* max value of cur_line */
	int this_line;                  /* line l points to */
	int nflushd_lines;              /* number of lines that were flushed */
	int adjust, opt, warned;
	int ret = EXIT_SUCCESS;
	char *sline = NULL;

#if ENABLE_FEATURE_LONG_OPTIONS
	static const struct option longopts[] = {
		{ "no-backspaces", no_argument,         0, 'b' },
		{ "fine",          no_argument,         0, 'f' },
		{ "pass",          no_argument,         0, 'p' },
		{ "tabs",          no_argument,         0, 'h' },
		{ "spaces",        no_argument,         0, 'x' },
		{ "lines",         required_argument,   0, 'l' },
		{ "version",       no_argument,         0, 'V' },
		{ NULL, 0, 0, 0 }
	};
	bb_applet_long_options = longopts;
#endif

	max_bufd_lines = 128 * 2;
	compress_spaces = 1;            /* compress spaces into tabs */

	opt = bb_xgetopt_ulflags(opt_str, &sline);
	if((opt & VERSION_OPT)) {
		printf("%s from %s\n", bb_applet_name, PACKAGE_STRING);
		return EXIT_SUCCESS;
	}
	if((opt & NO_COMPRESS_SPACES_OPT))
		compress_spaces = 0;    /* do not compress spaces into tabs */
	if(sline != NULL) {
		/*
		 * Buffered line count, which is a value in half
		 * lines e.g. twice the amount specified.
		 */
		max_bufd_lines = bb_xparse_ui_number(10 | BB_PN_BKM_SUFFIXES, sline);
	}

	adjust = cur_col = extra_lines = warned = 0;
	cur_line = max_line = nflushd_lines = this_line = 0;
	cur_set = last_set = CS_NORMAL;
	lines = l = alloc_line();

	while (feof(stdin) == 0) {
		errno = 0;
		if ((ch = getwchar()) == WEOF) {
			if (errno == EILSEQ) {
				bb_perror_nomsg();
				ret = EXIT_FAILURE;
			}
			break;
		}
		if (!iswgraph(ch)) {
			switch (ch) {
			case BS:                /* can't go back further */
				if (cur_col == 0)
					continue;
				--cur_col;
				continue;
			case CR:
				cur_col = 0;
				continue;
			case ESC:               /* just ignore EOF */
				switch(getwchar()) {
				case RLF:
					cur_line -= 2;
					break;
				case RHLF:
					cur_line--;
					break;
				case FHLF:
					cur_line++;
					if (cur_line > max_line)
						max_line = cur_line;
				}
				continue;
			case NL:
				cur_line += 2;
				if (cur_line > max_line)
					max_line = cur_line;
				cur_col = 0;
				continue;
			case SPACE:
				++cur_col;
				continue;
			case SI:
				cur_set = CS_NORMAL;
				continue;
			case SO:
				cur_set = CS_ALTERNATE;
				continue;
			case TAB:               /* adjust column */
				cur_col |= 7;
				++cur_col;
				continue;
			case VT:
				cur_line -= 2;
				continue;
			}
			if (iswspace(ch)) {
				if (wcwidth(ch) > 0)
					cur_col += wcwidth(ch);
				continue;
			}
			if (!(opt & PASS_UNKNOWN_SEQS_OPT))
				continue;
		}

		/* Must stuff ch in a line - are we at the right one? */
		if (cur_line != this_line - adjust) {
			LINE *lnew;
			int nmove;

			adjust = 0;
			nmove = cur_line - this_line;
			if (!(bss_flags & FINE_OPT)) {
				/* round up to next line */
				if (cur_line & 1) {
					adjust = 1;
					nmove++;
				}
			}
			if (nmove < 0) {
				for (; nmove < 0 && l->l_prev; nmove++)
					l = l->l_prev;
				if (nmove) {
					if (nflushd_lines == 0) {
						/*
						 * Allow backup past first
						 * line if nothing has been
						 * flushed yet.
						 */
						for (; nmove < 0; nmove++) {
							lnew = alloc_line();
							l->l_prev = lnew;
							lnew->l_next = l;
							l = lines = lnew;
							extra_lines++;
						}
					} else {
						if (!warned++)
							bb_error_msg(
			"warning: can't back up %s.", cur_line < 0 ?
			"past first line" : "-- line already flushed");
						cur_line -= nmove;
					}
				}
			} else {
				/* may need to allocate here */
				for (; nmove > 0 && l->l_next; nmove--)
					l = l->l_next;
				for (; nmove > 0; nmove--) {
					lnew = alloc_line();
					lnew->l_prev = l;
					l->l_next = lnew;
					l = lnew;
				}
			}
			this_line = cur_line + adjust;
			nmove = this_line - nflushd_lines;
			if (nmove > 0
			    && (unsigned) nmove >= max_bufd_lines + BUFFER_MARGIN) {
				nflushd_lines += nmove - max_bufd_lines;
				flush_lines(nmove - max_bufd_lines);
			}
		}
		/* grow line's buffer? */
		if (l->l_line_len + 1 >= l->l_lsize) {
			int need;

			need = l->l_lsize ? l->l_lsize * 2 : 90;
			l->l_line = (CHAR *)xrealloc((void *) l->l_line,
						    (unsigned) need * sizeof(CHAR));
			l->l_lsize = need;
		}
		c = &l->l_line[l->l_line_len++];
		c->c_char = ch;
		c->c_set = cur_set;
		c->c_column = cur_col;
		c->c_width = wcwidth(ch);
		/*
		 * If things are put in out of order, they will need sorting
		 * when it is flushed.
		 */
		if (cur_col < l->l_max_col)
			l->l_needs_sort = 1;
		else
			l->l_max_col = cur_col;
		if (c->c_width > 0)
			cur_col += c->c_width;
	}
	/* goto the last line that had a character on it */
	for (; l->l_next; l = l->l_next)
		this_line++;
	flush_lines(this_line - nflushd_lines + extra_lines + 1);

	/* make sure we leave things in a sane state */
	if (last_set != CS_NORMAL)
		PUTC('\017');

	/* flush out the last few blank lines */
	nblank_lines = max_line - this_line;
	if (max_line & 1)
		nblank_lines++;
	else if (!nblank_lines)
		/* missing a \n on the last line? */
		nblank_lines = 2;
	flush_blanks();
	bb_fflush_stdout_and_exit(ret);
}
#endif
