/*
 * Mini comm implementation for busybox
 *
 * Copyright (C) 2005 by Robert Sullivan <cogito.ergo.cogito@gmail.com>
 * Licensed under GPL v2, see file LICENSE in this tarball for details.
 */
#ifdef APPLET_USAGE
#define comm_trivial_usage \
	"[-123] FILE1 FILE2"
#define comm_full_usage \
	"Compares FILE1 to FILE2, or to stdin if = is specified.\n\n" \
	"Options:\n" \
	"\t-1\tSuppress lines unique to FILE1\n" \
	"\t-2\tSuppress lines unique to FILE2\n" \
	"\t-3\tSuppress lines common to both files"
#else

#define COMPLEMENTARY "\0=2" /* 2 opt only required */
#include "opt_gen.h"
OPT_GEN(opt_simple(1), opt_simple(2), opt_simple(3))
#include "opt_gen.h"

#include "busybox.h"
#include <string.h>
#include <unistd.h>

CALC_BSS struct BB_COMMON_BSS {
/* These three variables control behaviour if non-zero */
	int only_file_1;
	int only_file_2;
	int both;
} bb_common_bss;
#define only_file_1 bb_common_bss.only_file_1
#define only_file_2 bb_common_bss.only_file_2
#define both bb_common_bss.both


/* writeline outputs the input given, appropriately aligned according to class */
static void writeline(char *line, int class)
{
	if (class == 0) {
		if (!only_file_1)
			return;
	} else if (class == 1) {
		if (!only_file_2)
			return;
		if (only_file_1)
			putchar('\t');
	}
	else /*if (class == 2)*/ {
		if (!both)
			return;
		if (only_file_1)
			putchar('\t');
		if (only_file_2)
			putchar('\t');
	}
	fputs(line, stdout);
}

/* This is the real core of the program - lines are compared here */
static void cmp_files(void)
{
	char **infiles = Gargv;
#define LINE_LEN 100
#define BB_EOF_0 0x1
#define BB_EOF_1 0x2
	char thisline[2][LINE_LEN];
	char *r[2];
	FILE *streams[2];
	int i;

	for (i = 0; i < 2; ++i) {
		streams[i] = bb_xfopen(infiles[i], "A-r");
		r[i] = fgets(thisline[i], LINE_LEN, streams[i]);
	}

	while (r[0] || r[1]) {
		int order = 0;

		i = 0;
		if (feof(streams[0])) i |= BB_EOF_0;
		if (feof(streams[1])) i |= BB_EOF_1;

		if (!r[0])
			order = 1;
		else if (!r[1])
			order = -1;
		else {
			int tl0_len, tl1_len;
			tl0_len = strlen(thisline[0]);
			tl1_len = strlen(thisline[1]);
			order = memcmp(thisline[0], thisline[1], tl0_len < tl1_len ? tl0_len : tl1_len);
			if (!order)
				order = tl0_len < tl1_len ? -1 : tl0_len != tl1_len;
		}

		if (order == 0 && !i)
			writeline(thisline[1], 2);
		else if (order > 0 && !(i & BB_EOF_1))
			writeline(thisline[1], 1);
		else if (order < 0 && !(i & BB_EOF_0))
			writeline(thisline[0], 0);

		if (i & BB_EOF_0 & BB_EOF_1) {
			break;

		} else if (i) {
			i = (i & BB_EOF_0 ? 1 : 0);
			while (!feof(streams[i])) {
				if ((order < 0 && i) || (order > 0 && !i))
					writeline(thisline[i], i);
				fgets(thisline[i], LINE_LEN, streams[i]);
			}
			break;

		} else {
			if (order >= 0)
				r[1] = fgets(thisline[1], LINE_LEN, streams[1]);
			if (order <= 0)
				r[0] = fgets(thisline[0], LINE_LEN, streams[0]);
		}
	}

	fclose(streams[0]);
	fclose(streams[1]);
}

int comm_main(void)
{
	int flags;


	flags = bb_xgetopt_ulflags(opt_str);

	only_file_1 = !(flags & OPT_1);
	only_file_2 = !(flags & OPT_2);
	both = !(flags & OPT_3);

	cmp_files();
	return EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
