/*
 * head implementation for busybox
 *
 * Copyright (C) 2003  Manuel Novoa III  <mjn3@codepoet.org>
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */

/* BB_AUDIT SUSv3 compliant */
/* BB_AUDIT GNU compatible -c, -q, and -v options in 'fancy' configuration. */
/* http://www.opengroup.org/onlinepubs/007904975/utilities/head.html */
#ifdef APPLET_USAGE
#define head_trivial_usage \
	"[OPTION]... [FILE]..."
#define head_full_usage \
	"Print first 10 lines of each FILE to standard output.\n" \
	"With more than one FILE, precede each with a header giving the\n" \
	"file name. With no FILE, or when FILE is -, read standard input.\n\n" \
	"Options:\n" \
	"\t-n NUM\t\tPrint first NUM lines instead of first 10" \
	USE_FEATURE_FANCY_HEAD( \
	"\n\t-c NUM\t\toutput the first NUM bytes\n" \
	"\t-q\t\tnever output headers giving file names\n" \
	"\t-v\t\talways output headers giving file names" )
#else

#include "busybox.h"
#include <ctype.h>
#include <unistd.h>

#define COMPLEMENTARY "\0-0" USE_FEATURE_FANCY_HEAD(":cn:==n")
#include "opt_gen.h"
OPT_GEN(
opt(n, NUM_FLG, num),
opt_opt(ENABLE_FEATURE_FANCY_HEAD, c, NUM_BYTES_FLG, num),
opt_opt(ENABLE_FEATURE_FANCY_HEAD, q, Q_FLG),
opt_opt(ENABLE_FEATURE_FANCY_HEAD, v, V_FLG)
)
#include "opt_gen.h"

static const char header_fmt_str[] = "\n==> %s <==\n";

int head_main(void)
{
	char **argv;
	unsigned long count = 10;
	unsigned long i;
#if ENABLE_FEATURE_FANCY_HEAD
	int count_bytes;
	int header_threshhold = 1;
#else
# define header_threshhold   1
# define count_bytes         0
#endif

	FILE *fp;
	const char *fmt;
	char *p;
	int opt;
	int c;
	int retval = EXIT_SUCCESS;

	opt = bb_xgetopt_ulflags(opt_str, &p USE_FEATURE_FANCY_HEAD(,&p));

	if(opt & NUM_FLG)
	    count = bb_xparse_ul_number(10 | BB_PN_BKM_SUFFIXES, p);

#if ENABLE_FEATURE_FANCY_HEAD
	count_bytes = opt & NUM_BYTES_FLG;
	if(opt & Q_FLG)
		header_threshhold = INT_MAX;
	if(opt & V_FLG)
		header_threshhold = -1;
#endif

	fmt = header_fmt_str + 1;
#if ENABLE_FEATURE_FANCY_HEAD
	if (Gargc <= header_threshhold)
		header_threshhold = 0;
#else
	if (Gargc <= 1)
		fmt += 11;
#endif

	for (argv = Gargv; *argv; argv++) {
		if ((fp = bb_xfopen(*argv, "W-r")) != NULL) {
			if (fp == stdin)
				*argv = (char *) bb_msg_standard_input;
			if (header_threshhold)
				bb_printf(fmt, *argv);
			i = count;
			while (i && ((c = getc(fp)) != EOF)) {
				if (count_bytes || (c == '\n'))
					--i;
				putchar(c);
			}
			if (bb_fclose_nonstdin(fp)) {
				bb_perror_msg("%s", *argv);
				retval = EXIT_FAILURE;
			}
			bb_xferror_stdout();
			fmt = header_fmt_str;
		} else {
			retval = EXIT_FAILURE;
		}
	}

	bb_fflush_stdout_and_exit(retval);
}
#endif /* APPLET_USAGE */
