/*
 *  Copyright (C) 2003 by Glenn McGrath <bug1@iinet.net.au>
 * Licensed under GPL v2, see file LICENSE in this tarball for details.
 *
 * TODO: -d option, need a way of recursively making directories and changing
 *           owner/group, will probably modify bb_make_directory(...)
 */
#ifdef APPLET_USAGE
#define install_trivial_usage \
	"[-cgmops] [sources] <dest|directory>"
#define install_full_usage \
	"Copies files and set attributes\n\n" \
	"Options:\n" \
	"\t-c\tcopy the file, default\n" \
	"\t-d\tcreate directories\n" \
	"\t-g\tset group ownership\n" \
	"\t-m\tset permission modes\n" \
	"\t-o\tset ownership\n" \
	"\t-p\tpreserve date\n" \
	"\t-s\tstrip symbol tables"
#else

#include "busybox.h"
#include <string.h>
#include <unistd.h>
#include <libgen.h>

#include "libcoreutils/coreutils.h"

#define COMPLEMENTARY "\0?s--d:d--s"
#include "opt_gen.h"
OPT_GEN(
opt(c, INSTALL_OPT_CMD),
opt(d, INSTALL_OPT_DIRECTORY),
opt(p, INSTALL_OPT_PRESERVE_TIME),
opt(s, INSTALL_OPT_STRIP),
opt(g, INSTALL_OPT_GROUP, arg),
opt(m, INSTALL_OPT_MODE, arg),
opt(o, INSTALL_OPT_OWNER, arg)
)
#include "opt_gen.h"

#if ENABLE_FEATURE_LONG_OPTIONS
static const struct option install_long_options[] = {
	{ "directory",                  0,      NULL,   'd' },
	{ "preserve-timestamps",        0,      NULL,   'p' },
	{ "strip",                      0,      NULL,   's' },
	{ "group",                      0,      NULL,   'g' },
	{ "mode",                       0,      NULL,   'm' },
	{ "owner",                      0,      NULL,   'o' },
	{ 0,                            0,      0,      0 }
};
#endif

extern int install_main(void)
{
	char **argv;
	char *last;
	mode_t mode;
	uid_t uid;
	gid_t gid;
	char *gid_str = "-1";
	char *uid_str = "-1";
	char *mode_str = "0755";
	int copy_flags = FILEUTILS_DEREFERENCE | FILEUTILS_FORCE;
	int ret = EXIT_SUCCESS, flags, isdir;

#if ENABLE_FEATURE_LONG_OPTIONS
	bb_applet_long_options = install_long_options;
#endif
	/* -c exists for backwards compatibility, its needed */
	flags = bb_xgetopt_ulflags(opt_str, &gid_str, &mode_str, &uid_str);
	/* preserve access and modification time, this is GNU behaviour, BSD only preserves modification time */
	if (flags & INSTALL_OPT_PRESERVE_TIME) {
		copy_flags |= FILEUTILS_PRESERVE_STATUS;
	}
	bb_parse_mode(mode_str, &mode);
	gid = get_ug_id(gid_str, bb_xgetgrnam);
	uid = get_ug_id(uid_str, bb_xgetpwnam);
	umask(0);

	/* Create directories
	 * don't use bb_make_directory() as it can't change uid or gid
	 * perhaps bb_make_directory() should be improved.
	 */
	if (flags & INSTALL_OPT_DIRECTORY) {
		for (argv = Gargv; *argv; argv++) {
			char *old_argv_ptr = *argv + 1;
			char *argv_ptr;
			do {
				argv_ptr = strchr_set0(old_argv_ptr, '/');
				old_argv_ptr = argv_ptr;
				if (argv_ptr)
					old_argv_ptr++;
				if (mkdir(*argv, mode) == -1) {
					if (errno != EEXIST) {
						bb_perror_msg("coulnt create %s", *argv);
						ret = EXIT_FAILURE;
						break;
					}
				}
				else if (lchown(*argv, uid, gid) == -1) {
					bb_perror_msg("cannot change ownership of %s", *argv);
					ret = EXIT_FAILURE;
					break;
				}
				if (argv_ptr) {
					*argv_ptr = '/';
				}
			} while (old_argv_ptr);
		}
		return ret;
	}

	last = Gargv[Gargc - 1];
	isdir = is_directory(last, 0, NULL);
	for (argv = Gargv; *argv != last; argv++) {
		char *dest;

		dest = last;
		if (isdir > 0) dest = concat_path_file(last, basename(*argv));
		ret |= copy_file(*argv, dest, copy_flags);

		/* Set the file mode */
		if (chmod(dest, mode) == -1) {
			bb_perror_msg("cannot change permissions of %s", dest);
			ret = EXIT_FAILURE;
		}

		/* Set the user and group id */
		if (lchown(dest, uid, gid) == -1) {
			bb_perror_msg("cannot change ownership of %s", dest);
			ret = EXIT_FAILURE;
		}
		if (flags & INSTALL_OPT_STRIP) {
			if (execlp("strip", "strip", dest, NULL) == -1) {
				bb_error_msg("strip failed");
				ret = EXIT_FAILURE;
			}
		}
	}

	return ret;
}
#endif /* APPLET_USAGE */
