/*
 * Mini ln implementation for busybox
 *
 * Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 * Licensed under GPL v2, see file LICENSE in this tarball for details.
 */

/* BB_AUDIT SUSv3 compliant */
/* BB_AUDIT GNU options missing: -d, -F, -i, and -v. */
/* http://www.opengroup.org/onlinepubs/007904975/utilities/ln.html */
#ifdef APPLET_USAGE
#define ln_trivial_usage \
	"[OPTION] TARGET... LINK_NAME|DIRECTORY"
#define ln_full_usage \
	"Create a link named LINK_NAME or DIRECTORY to the specified TARGET\n" \
	"\nYou may use '--' to indicate that all following arguments are non-options.\n\n" \
	"Options:\n" \
	"\t-s\tmake symbolic links instead of hard links\n" \
	"\t-f\tremove existing destination files\n" \
	"\t-n\tno dereference symlinks - treat like normal file\n" \
	"\t-b\tmake a backup of the target (if exists) before link operation\n" \
	"\t-S suffix\tuse suffix instead of ~ when making backup files"
#else

#define COMPLEMENTARY "\0-1" /* min 1 args */
#include "opt_gen.h"
OPT_GEN(
opt(s, LN_SYMLINK),
opt(f, LN_FORCE),
opt(n, LN_NODEREFERENCE),
opt(b, LN_BACKUP),
opt(S, LN_SUFFIX, arg)
)
#include "opt_gen.h"

#include <unistd.h>
#include "busybox.h"

extern int ln_main(void)
{
	char **argv;
	int status = EXIT_SUCCESS;
	int flag;
	char *last;
	char *src_name;
	char *src;
	char *suffix = "~";
	int (*link_func)(const char *, const char *);

	flag = bb_xgetopt_ulflags(opt_str, &suffix);

	argv = Gargv;
	last = argv[Gargc - 1];

	if (argv[1] == NULL) {
		*--argv = last;
		last = bb_get_last_path_component(xstrdup(last));
	}

	do {
		src_name = NULL;
		src = last;

		if (is_directory(src, (flag & LN_NODEREFERENCE) ^ LN_NODEREFERENCE, NULL) > 0) {
			src_name = xstrdup(*argv);
			src = concat_path_file(src, bb_get_last_path_component(src_name));
			free(src_name);
			src_name = src;
		}
		if (!(flag & LN_SYMLINK) && is_directory(*argv, 1, NULL) < 0) {
			bb_perror_msg("%s", *argv);
			status = EXIT_FAILURE;
			free(src_name);
			continue;
		}

		if (flag & LN_BACKUP) {
				char *backup = xasprintf("%s%s", src, suffix);
				if (rename(src, backup) < 0 && errno != ENOENT) {
					bb_perror_msg("%s", src);
					status = EXIT_FAILURE;
					free(backup);
					continue;
				}
				free(backup);
				/*
				 * When the source and dest are both hard links to the same
				 * inode, a rename may succeed even though nothing happened.
				 * Therefore, always unlink().
				 */
				unlink(src);
		} else if (flag & LN_FORCE) {
			unlink(src);
		}

		link_func = link;
		if (flag & LN_SYMLINK) {
			link_func = symlink;
		}

		if (link_func(*argv, src) != 0) {
			bb_perror_msg("%s", src);
			status = EXIT_FAILURE;
		}

		free(src_name);

	} while ((++argv)[1]);

	return status;
}
#endif /* APPLET_USAGE */
