/*
 * Mini mkdir implementation for busybox
 *
 * Copyright (C) 2001 Matt Kraai <kraai@alumni.carnegiemellon.edu>
 * Licensed under GPL v2, see file LICENSE in this tarball for details.
 *
 * destroy problem: mode_t and (-1) to bb_make_directory(...long...) by Vladimir Oleynik
 */

/* BB_AUDIT SUSv3 compliant */
/* http://www.opengroup.org/onlinepubs/007904975/utilities/mkdir.html */

/* Mar 16, 2003      Manuel Novoa III   (mjn3@codepoet.org)
 *
 * Fixed broken permission setting when -p was used; especially in
 * conjunction with -m.
 */
#ifdef APPLET_USAGE
#define mkdir_trivial_usage \
	"[OPTION] DIRECTORY..."
#define mkdir_full_usage \
	"Create the DIRECTORY(ies) if they do not already exist\n\n" \
	"Options:\n" \
	"\t-m\tset permission mode (as in chmod), not rwxrwxrwx - umask\n" \
	"\t-p\tno error if existing, make parent directories as needed"
#else

#include <unistd.h>
#include "busybox.h"

#ifdef CONFIG_FEATURE_LONG_OPTIONS
static const struct option mkdir_long_options[] = {
	{ "mode", 1, NULL, 'm' },
	{ "parents", 0, NULL, 'p' },
	{ 0, 0, 0, 0 }
};
#endif

extern int mkdir_main(void)
{
	char **argv;
	long lmode = -1;
	int status = EXIT_SUCCESS;
	int flags = 0;
	unsigned long opt;
	char *smode;

#ifdef CONFIG_FEATURE_LONG_OPTIONS
	bb_applet_long_options = mkdir_long_options;
#endif
	/* min 1 arg */
	opt = bb_xgetopt_ulflags("m:p\0-1", &smode);
	if(opt & 1) {
		mode_t mode = 0777;
		if (!bb_parse_mode (smode, &mode)) {
			bb_error_msg_and_die ("invalid mode `%s'", smode);
		}
		lmode = mode;
	}
	if(opt & 2)
		flags |= FILEUTILS_RECUR;

	for(argv = Gargv; *argv; argv++) {
		if (bb_make_directory(*argv, lmode, flags)) {
			status = EXIT_FAILURE;
		}
	}

	return status;
}
#endif /* APPLET_USAGE */
