/*
 * mknod implementation for busybox
 *
 * Copyright (C) 2003  Manuel Novoa III  <mjn3@codepoet.org>
 * Licensed under GPL v2, see file LICENSE in this tarball for details.
 */

/* BB_AUDIT SUSv3 N/A -- Matches GNU behavior. */
#ifdef APPLET_USAGE
#define mknod_trivial_usage \
	"[OPTIONS] NAME TYPE MAJOR MINOR"
#define mknod_full_usage \
	"Create a special file (block, character, or pipe).\n\n" \
	"Options:\n" \
	"\t-m" USE_FEATURE_LONG_OPTIONS(", --mode") "\tcreate the special file using the specified mode (default a=rw)\n\n" \
	"TYPEs include:\n" \
	"\tb:\tMake a block (buffered) device\n" \
	"\tc or u:\tMake a character (un-buffered) device\n" \
	"\tp:\tMake a named pipe. MAJOR and MINOR are ignored for named pipes"
#else

#include "busybox.h"
#include <string.h>
#include <unistd.h>
#include "libcoreutils/coreutils.h"

static const char modes_chars[] = { 'p', 'c', 'u', 'b', 0, 1, 1, 2 };
static const mode_t modes_cubp[] = { S_IFIFO, S_IFCHR, S_IFBLK };

extern int mknod_main(void)
{
	int argc;
	char **argv;
	mode_t mode;
	dev_t dev;
	const char *name;

	mode = getopt_mk_fifo_nod(2);
	argv = Gargv;
	argc = Gargc;

	if ((name = strchr(modes_chars, argv[1][0])) != NULL) {
		mode |= modes_cubp[(int)(name[4])];

		dev = 0;
		if ((*name != 'p') && ((argc -= 2) == 2)) {
			dev = (bb_xparse_ui_number(10 | BB_PN_SET_UPPER | BB_PN_ERR_STR1,
							argv[2], 255, "invalid MAJOR `%s'") << 8)
				+ bb_xparse_ui_number(10 | BB_PN_SET_UPPER | BB_PN_ERR_STR1,
							argv[3], 255, "invalid MINOR `%s'");
		}

		if (argc == 2) {
			name = *argv;
			if (mknod(name, mode, dev) == 0)
				return EXIT_SUCCESS;
			bb_perror_msg_and_die("%s", name);
		}
	}
	bb_show_usage();
}
#endif /* APPLET_USAGE */
