/*
 * Mini mv implementation for busybox
 *
 * Copyright (C) 2000 by Matt Kraai <kraai@alumni.carnegiemellon.edu>
 * Licensed under GPL v2, see file LICENSE in this tarball for details.
 */

/* Mar 16, 2003      Manuel Novoa III   (mjn3@codepoet.org)
 *
 * Size reduction and improved error checking.
 */
#ifdef APPLET_USAGE
#define mv_trivial_usage \
	"[OPTION]... SOURCE DEST\n" \
	"or: mv [OPTION]... SOURCE... DIRECTORY"
#define mv_full_usage \
	"Rename SOURCE to DEST, or move SOURCE(s) to DIRECTORY.\n\n" \
	"Options:\n" \
	"\t-f\tdon't prompt before overwriting\n" \
	"\t-i\tinteractive, prompt before overwrite"
#else

#define COMPLEMENTARY "\0-2:f-i:i-f"
#include "opt_gen.h"
OPT_GEN(opt(i, OPT_FILEUTILS_FORCE), opt(f, OPT_FILEUTILS_INTERACTIVE))
#include "opt_gen.h"

#include "busybox.h"
#include <unistd.h>
#include <dirent.h>
#include "libcoreutils/coreutils.h"

extern int mv_main(void)
{
	int argc;
	char **argv;
	struct stat dest_stat;
	char *last;
	char *dest;
	unsigned long flags;
	int dest_exists;
	int status = 0;
	static const char fmt[] = "cannot overwrite %sdirectory with %sdirectory";

#if ENABLE_FEATURE_LONG_OPTIONS
	static const struct option mv_long_options[] = {
		{ "interactive", 0, NULL, 'i' },
		{ "force", 0, NULL, 'f' },
		{ 0, 0, 0, 0 }
	};
	bb_applet_long_options = mv_long_options;
#endif
	flags = bb_xgetopt_ulflags(opt_str);
	argc = Gargc;
	argv = Gargv;

	last = argv[argc - 1];

	if (2 == argc) {
		if ((dest_exists = cp_mv_stat(last, &dest_stat)) < 0)
			return 1;

		if (!(dest_exists & 2)) {
			dest = last;
			goto DO_MOVE;
		}
	}

	do {
		dest = concat_path_file(last, bb_get_last_path_component(*argv));

		if ((dest_exists = cp_mv_stat(dest, &dest_stat)) < 0)
			goto RET_1;

DO_MOVE:

		if (dest_exists && !(flags & OPT_FILEUTILS_FORCE) &&
				((access(dest, W_OK) < 0 && isatty(0)) ||
				(flags & OPT_FILEUTILS_INTERACTIVE))) {
			if (fprintf(stderr, "mv: overwrite `%s'? ", dest) < 0)
				goto RET_1;     /* Ouch! fprintf failed! */
			if (!bb_ask_confirmation())
				goto RET_0;
		}
		if (rename(*argv, dest) < 0) {
			struct stat source_stat;
			int source_exists;

			if (errno != EXDEV ||
				(source_exists = cp_mv_stat(*argv, &source_stat)) < 1) {
				bb_perror_msg("unable to rename `%s'", *argv);
			} else {
				if (dest_exists) {
					if (dest_exists == 3) {
						if (source_exists != 3) {
							bb_error_msg(fmt, "", "non-");
							goto RET_1;
						}
					} else {
						if (source_exists == 3) {
							bb_error_msg(fmt, "non-", "");
							goto RET_1;
						}
					}
					if (unlink(dest) < 0) {
						bb_perror_msg("cannot remove `%s'", dest);
						goto RET_1;
					}
				}
				if ((copy_file(*argv, dest,
					FILEUTILS_RECUR | FILEUTILS_PRESERVE_STATUS) >= 0) &&
					(remove_file(*argv, FILEUTILS_RECUR | FILEUTILS_FORCE) >= 0)) {
					goto RET_0;
				}
			}
RET_1:
			status = 1;
		}
RET_0:
		if (dest != last)
			free(dest);
	} while (*++argv != last);

	return (status);
}
#endif /* APPLET_USAGE */
