/*
 * nice implementation for busybox
 * Copyright (C) 1990-2005 Free Software Foundation, Inc.
 * David MacKenzie <djm@gnu.ai.mit.edu>
 *
 * Copyright (C) 2008 Vladimir Oleynik <dzo@simtreas.ru>
 *   implementation for busybox
 *
 */
#ifdef APPLET_USAGE
#define nice_trivial_usage \
	"[-n ADJUST COMMAND [ARG] ...]"
#define nice_full_usage \
	"Nice runs a program with modified scheduling priority.\n\n" \
	"Options:\n" \
	"\t-n ADJUST\tAdjust the scheduling priority by ADJUST"
#else

#include "busybox.h"
#include <string.h>
#include <unistd.h>
#include <sys/resource.h>

#define EXIT_CANNOT_INVOKE (126)
#define EXIT_ENOENT (127)

#ifndef NZERO
# define NZERO 20
#endif

/* This is required for Darwin Kernel Version 7.7.0.  */
#if NZERO == 0
# undef  NZERO
# define NZERO 20
#endif

#if ENABLE_FEATURE_LONG_OPTIONS
static struct option const longopts[] =
{
  {"adjustment", required_argument, NULL, 'n'},
  {NULL, 0, NULL, 0}
};
#endif

static inline int ISDIGIT(char c)
{
	int i = (int)c;
	return (unsigned)(i - '0') <= 9;
}

int
nice_main(void)
{
  int argc = Gargc;
  char **argv = Gargv;
  int adjustment = 10;
  char const *adjustment_given = NULL;
  int ok;
  int i;

  for (i = 1; i < argc; /* empty */) {
      char const *s = argv[i];

      if (s[0] == '-' && ISDIGIT(s[1 + (s[1] == '-' || s[1] == '+')])) {
	  adjustment_given = s + 1;
	  ++i;
      } else {
	  Gargc = argc - (i - 1);
	  Gargv = argv + (i - 1);

	  /* Ensure that any getopt diagnostics use the right name.  */
	  Gargv[0] = (char *)bb_applet_name;
#if ENABLE_FEATURE_LONG_OPTIONS
	  bb_applet_long_options = longopts;
#endif
	  bb_getopt_ulflags("+n:", &adjustment_given);
	  i += optind - 1;
	  break;
      }
  }

  if (adjustment_given) {
      /* If the requested adjustment is outside the valid range,
	 silently bring it to just within range; this mimics what
	 "setpriority" and "nice" do.  */
      enum { MIN_ADJUSTMENT = 1 - 2 * NZERO, MAX_ADJUSTMENT = 2 * NZERO - 1 };
      long int tmp = bb_xparse_l_number(10 | BB_PN_LOWER_MIN, adjustment_given);
      adjustment = MAX (MIN_ADJUSTMENT, MIN (tmp, MAX_ADJUSTMENT));
  }

  if (i == argc) {
      if (adjustment_given)
	  bb_show_usage();
      /* No command given; print the niceness.  */
      printf ("%d\n", nice(0));
      return EXIT_SUCCESS;
  }

  errno = 0;
  ok = (nice (adjustment) != -1 || errno == 0);
  if (!ok)
    bb_perror_msg("cannot set niceness");

  execvp (argv[i], &argv[i]);
  bb_default_error_retval = (errno == ENOENT ? EXIT_ENOENT : EXIT_CANNOT_INVOKE);
  bb_perror_msg_and_die("%s", *argv);
}
#endif /* APPLET_USAGE */
