/*
 * Licensed under GPL v2, see file LICENSE in this tarball for details.
 */

/* BB_AUDIT SUSv3 N/A -- Apparently a busybox extension. */

/* Mar 16, 2003      Manuel Novoa III   (mjn3@codepoet.org)
 *
 * Now does proper error checking on output and returns a failure exit code
 * if one or more paths can not be resolved.
 */

/* (C) 2022 Vladimir Oleynik <dzo@simtreas.ru>
 * support -e and -q
 */

#ifdef APPLET_USAGE
#define realpath_trivial_usage \
	"[-e] [-q] pathname..."
#define realpath_full_usage \
	"Returns the absolute pathnames of given argument.\n\n" \
	"Options:\n" \
	"\t-e\tAll components of the path must exist\n" \
	"\t-q\tSuppress most error messages"
#else

#include "busybox.h"
#include <string.h>

int realpath_main(void)
{
	char **pa, resolved_path[PATH_MAX];
	int o, retval = EXIT_SUCCESS;

	o = bb_xgetopt_ulflags("eq\0-1");
#define OPT_E 1
#define OPT_Q 2

	for (pa = Gargv; *pa != NULL; pa++) {
		char *p = *pa;

		if (realpath(p, resolved_path) != NULL) {
			puts(resolved_path);
		} else {
			if ((o & OPT_E) != 0) {
				retval = EXIT_FAILURE;
				if ((o & OPT_Q) == 0)
					bb_perror_msg("%s", p);
			} else {
				p = bb_simplify_path(p);
				puts(p);
				free(p);
			}
		}
	}

	bb_fflush_stdout_and_exit(retval);
}
#endif /* APPLET_USAGE */
