/*
 * Mini rm implementation for busybox
 *
 * Copyright (C) 2001 Matt Kraai <kraai@alumni.carnegiemellon.edu>
 * Licensed under GPL v2, see file LICENSE in this tarball for details.
 */

/* BB_AUDIT SUSv3 compliant */
/* http://www.opengroup.org/onlinepubs/007904975/utilities/rm.html */

/* Mar 16, 2003      Manuel Novoa III   (mjn3@codepoet.org)
 *
 * Size reduction.
 */
#ifdef APPLET_USAGE
#define rm_trivial_usage \
	"[OPTION]... FILE..."
#define rm_full_usage \
	"Remove (unlink) the FILE(s). You may use '--' to\n" \
	"indicate that all following arguments are non-options.\n\n" \
	"Options:\n" \
	"\t-i\t\talways prompt before removing each destination\n" \
	"\t-f\t\tremove existing destinations, never prompt\n" \
	"\t-r or -R\tremove the contents of directories recursively"
#else

#define COMPLEMENTARY "\0-2:f-i:i-f"
#include "opt_gen.h"
OPT_GEN(opt(i, OPT_FORCE), opt(f, OPT_INTERACTIVE), opt_simple(R), opt_simple(r))
#include "opt_gen.h"

#include <unistd.h>
#include "busybox.h"

extern int rm_main(void)
{
	char **argv;
	int status = 0;
	int flags = 0;
	int opt;

	opt = (int)bb_xgetopt_ulflags(opt_str);
	if(opt & OPT_FORCE)
		flags |= FILEUTILS_FORCE;
	if(opt & OPT_INTERACTIVE)
		flags |= FILEUTILS_INTERACTIVE;
	if(opt & (OPT_R|OPT_r))
		flags |= FILEUTILS_RECUR;
	argv = Gargv;
	if (*argv != NULL) {
		do {
			const char *base = bb_get_last_path_component(*argv);

			if ((base[0] == '.') && (!base[1] || ((base[1] == '.') && !base[2]))) {
				bb_error_msg("cannot remove `.' or `..'");
			} else if (remove_file(*argv, flags) >= 0) {
				continue;
			}
			status = 1;
		} while (*++argv);
	} else if (!(flags & FILEUTILS_FORCE)) {
		bb_show_usage();
	}

	return status;
}
#endif /* APPLET_USAGE */
