/*
 * rmdir implementation for busybox
 *
 * Copyright (C) 2003  Manuel Novoa III  <mjn3@codepoet.org>
 * Licensed under GPL v2, see file LICENSE in this tarball for details.
 */

/* BB_AUDIT SUSv3 compliant */
/* http://www.opengroup.org/onlinepubs/007904975/utilities/rmdir.html */
#ifdef APPLET_USAGE
#define rmdir_trivial_usage \
	"[OPTION]... DIRECTORY..."
#define rmdir_full_usage \
	"Remove the DIRECTORY(ies), if they are empty."
#else

#include <unistd.h>
#include <libgen.h>
#include "busybox.h"

extern int rmdir_main(void)
{
	char **argv;
	int status = EXIT_SUCCESS;
	int flags;
	int do_dot;
	char *path;

	/* min 1 arg */
	flags = bb_xgetopt_ulflags("p\0-1");

	for(argv = Gargv; *argv; argv++) {
		path = *argv;

		/* Record if the first char was a '.' so we can use dirname later. */
		do_dot = (*path == '.');

		do {
			if (rmdir(path) < 0) {
				bb_perror_msg("`%s'", path);    /* Match gnu rmdir msg. */
				status = EXIT_FAILURE;
			} else if (flags) {
				/* Note: path was not empty or null since rmdir succeeded. */
				path = dirname(path);
				/* Path is now just the parent component.  Note that dirname
				 * returns "." if there are no parents.  We must distinguish
				 * this from the case of the original path starting with '.'.
				 */
				if (do_dot || (*path != '.') || path[1]) {
					continue;
				}
			}
			break;
		} while (1);

	}

	return status;
}
#endif /* APPLET_USAGE */
