/*
 * sleep implementation for busybox
 *
 * Copyright (C) 2003  Manuel Novoa III  <mjn3@codepoet.org>
 * Licensed under GPL v2, see file LICENSE in this tarball for details.
 */

/* BB_AUDIT SUSv3 compliant */
/* BB_AUDIT GNU issues -- fancy version matches except args must be ints. */
/* http://www.opengroup.org/onlinepubs/007904975/utilities/sleep.html */

/* Mar 16, 2003      Manuel Novoa III   (mjn3@codepoet.org)
 *
 * Rewritten to do proper arg and error checking.
 * Also, added a 'fancy' configuration to accept multiple args with
 * time suffixes for seconds, minutes, hours, and days.
 */
#ifdef APPLET_USAGE
#define sleep_trivial_usage \
	USE_FEATURE_FANCY_SLEEP("[") "N" USE_FEATURE_FANCY_SLEEP("]...")
#define sleep_full_usage \
	SKIP_FEATURE_FANCY_SLEEP("Pause for N seconds.") \
	USE_FEATURE_FANCY_SLEEP( \
	"Pause for a time equal to the total of the args given, where each arg can\n" \
	"\t\thave an optional suffix of (s)econds, (m)inutes, (h)ours, or (d)ays.")
#else

#include "busybox.h"
#include <unistd.h>

#ifdef CONFIG_FEATURE_FANCY_SLEEP
static const struct suffix_mult sleep_suffixes[] = {
	{ "s", 1 },
	{ "m", 60 },
	{ "h", 60*60 },
	{ "d", 24*60*60 },
	{ NULL, 0 }
};
#endif

extern int sleep_main(void)
{
	unsigned int duration;

#ifdef CONFIG_FEATURE_FANCY_SLEEP
	char **argv;

	bb_xgetopt_ulflags("\0-1");
	argv = Gargv;
	duration = 0;
	do {
		duration += bb_xparse_ui_number(BB_PN_SET_UPPER | BB_PN_SUFFIX | BB_PN_ERR_STR1,
						*argv, (unsigned long)(UINT_MAX-duration),
						sleep_suffixes, "invalid time interval '%s'");
	} while (*++argv);

#else
	bb_xgetopt_ulflags("\0=1");
	duration = bb_xparse_ui_number(0, Gargv[0]);

#endif /* CONFIG_FEATURE_FANCY_SLEEP */

	if (sleep(duration))
		bb_perror_nomsg_and_die();

	return EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
