/*
 * Mini touch implementation for busybox
 *
 * Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 * Licensed under GPL v2, see file LICENSE in this tarball for details.
 */

/* BB_AUDIT SUSv3 _NOT_ compliant -- options -a, -m, -r, -t not supported. */
/* http://www.opengroup.org/onlinepubs/007904975/utilities/touch.html */

/* Mar 16, 2003      Manuel Novoa III   (mjn3@codepoet.org)
 *
 * Previous version called open() and then utimes().  While this will be
 * be necessary to implement -r and -t, it currently only makes things bigger.
 * Also, exiting on a failure was a bug.  All args should be processed.
 */
#ifdef APPLET_USAGE
#define touch_trivial_usage \
	"[-c] FILE [FILE ...]"
#define touch_full_usage \
	"Update the last-modified date on the given FILE[s].\n\n" \
	"Options:\n" \
	"\t-c\tDo not create any files"
#else

#include "busybox.h"
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>


extern int touch_main(void)
{
	char **argv;
	int fd;
	int flags;
	int status = EXIT_SUCCESS;

	/* args require */
	flags = bb_xgetopt_ulflags("c\0-1");

	for(argv = Gargv; *argv; argv++) {
		if (utimes(*argv, NULL)) {
			if (errno == ENOENT) {  /* no such file*/
				if (flags & 1) {        /* Creation is disabled, so ignore. */
					continue;
				}
				/* Try to create the file. */
				fd = open(*argv, O_RDWR | O_CREAT,
						  S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH
						  );
				if ((fd >= 0) && !close(fd)) {
					continue;
				}
			}
			status = EXIT_FAILURE;
			bb_perror_msg("%s", *argv);
		}
	}

	return status;
}
#endif /* APPLET_USAGE */
