/*
 * tty implementation for busybox
 *
 * Copyright (C) 2003  Manuel Novoa III  <mjn3@codepoet.org>
 *
 */

/* BB_AUDIT SUSv3 compliant */
/* http://www.opengroup.org/onlinepubs/007904975/utilities/tty.html */
#ifdef APPLET_USAGE
#define tty_trivial_usage \
	""
#define tty_full_usage \
	"Print the file name of the terminal connected to standard input.\n\n" \
	"Options:\n" \
	"\t-s\tprint nothing, only return an exit status"
#else

#include "busybox.h"
#include <unistd.h>

extern int tty_main(void)
{
	const char *s;
	int silent;             /* Note: No longer relevant in SUSv3. */
	int retval;

	bb_default_error_retval = 2;    /* SUSv3 requires > 1 for error. */

	silent = bb_xgetopt_ulflags("s\0??");

	retval = 0;

	if ((s = ttyname(0)) == NULL) {
	/* According to SUSv3, ttyname can on fail with EBADF or ENOTTY.
	 * We know the file descriptor is good, so failure means not a tty. */
		s = "not a tty";
		retval = 1;
	}

	if (!silent) {
		puts(s);
	}

	bb_fflush_stdout_and_exit(retval);
}
#endif /* APPLET_USAGE */
