/*
 * uniq implementation for busybox
 *
 * Copyright (C) 2005  Manuel Novoa III  <mjn3@codepoet.org>
 *
 * Licensed under the GPL v2 or later, see the file LICENSE in this tarball.
 */

/* BB_AUDIT SUSv3 compliant */
/* http://www.opengroup.org/onlinepubs/007904975/utilities/uniq.html */
#ifdef APPLET_USAGE
#define uniq_trivial_usage \
	"[-fscdu]... [INPUT [OUTPUT]]"
#define uniq_full_usage \
	"Discard all but one of successive identical lines from INPUT\n" \
	"(or standard input), writing to OUTPUT (or standard output).\n\n" \
	"Options:\n" \
	"\t-c\tprefix lines by the number of occurrences\n" \
	"\t-d\tonly print duplicate lines\n" \
	"\t-u\tonly print unique lines\n" \
	"\t-f N\tskip the first N fields\n" \
	"\t-s N\tskip the first N chars (after any skipped fields)"
#else

#include "busybox.h"
#include <string.h>
#include <ctype.h>
#include <unistd.h>


int uniq_main(void)
{
	char **argv;
	FILE *in, *out;
	unsigned long dups, skip_fields, skip_chars, i;
	int uniq_flags;
	const char *s0 = NULL, *e0, *s1 = NULL, *e1, *input_filename;

	/* "cdu" is special bit position */
	/* zero arg to "-", max 2 arg */
	uniq_flags = (int)bb_xgetopt_ulflags("cduf:s:\0-0?2", &s0, &s1)
			& (1|2|4);
	skip_fields = s0 != NULL ? bb_xparse_ul_number(10, s0) : 0;
	skip_chars = s1 != NULL ? bb_xparse_ul_number(10, s1) : 0;

	argv = Gargv;
	input_filename = *argv;

	in = bb_xfopen(*argv, "A-r");
	++argv;
	out = bb_xfopen(*argv, "A-w");

	s1 = e1 = NULL;                         /* prime the pump */

	do {
		s0 = s1;
		e0 = e1;
		dups = 0;

		/* gnu uniq ignores newlines */
		while ((s1 = bb_get_chomped_line_from_file(in)) != NULL) {
			e1 = s1;
			for (i = skip_fields; i; i--) {
				e1 = bb_skip_whitespace(e1);
				while (*e1 && !isspace(*e1)) {
					++e1;
				}
			}
			for (i = skip_chars; *e1 && i; i--) {
				++e1;
			}

			if (!s0 || strcmp(e0, e1)) {
				break;
			}

			++dups;          /* Note: Testing for overflow seems excessive. */
		}

		if (s0) {
			if (!(uniq_flags & (2 << !!dups))) {
				bb_fprintf(out, &("\0%d "[uniq_flags & 1]), dups + 1);
				bb_fprintf(out, "%s\n", s0);
			}
			free((void *)s0);
		}
	} while (s1);

	bb_xferror(in, input_filename);

	bb_fflush_stdout_and_exit(EXIT_SUCCESS);
}
#endif /* APPLET_USAGE */
