/*
 * Mini watch implementation for busybox
 *
 * Copyright (C) 2001 by Michael Habermann <mhabermann@gmx.de>
 * Licensed under GPL v2, see file LICENSE in this tarball for details.
 */

/* BB_AUDIT SUSv3 N/A */
/* BB_AUDIT GNU defects -- only option -n is supported. */

/* Mar 16, 2003      Manuel Novoa III   (mjn3@codepoet.org)
 *
 * Removed dependency on date_main(), added proper error checking, and
 * reduced size.
 */
#ifdef APPLET_USAGE
#define watch_trivial_usage \
	"[-n <seconds>] COMMAND..."
#define watch_full_usage \
	"Executes a program periodically.\n" \
	"Options:\n" \
	"\t-n\tLoop period in seconds - default is 2"
#else

#include "busybox.h"
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/wait.h>

extern int watch_main(void)
{
	char **argv;
	const int header_len = 40;
	time_t t;
	pid_t pid;
	unsigned period = 2;
	char *p;
	int old_stdout;
	int len, len2;
	char **watched_argv;
	char header[header_len + 1];

	/* '+' stop for first non-opt */
	/* require arg */
	if(bb_xgetopt_ulflags("+n:\0-1", &p))
		period = bb_xparse_ui_number(10 | BB_PN_LOWER_MIN | BB_PN_ERR_STR1,
						p, "invalid period `%s'");
	argv = watched_argv = Gargv;

	/* create header */

	len = snprintf(header, header_len, "Every %us:", period);
	do {
		len2 = strlen(*argv);
		if (len + len2 >= header_len-1) {
			break;
		}
		header[len] = ' ';
		memcpy(header+len+1, *argv, len2);
		len += len2+1;
	} while (*++argv);

	header[len] = 0;

	/* thanks to lye, who showed me how to redirect stdin/stdout */
	old_stdout = dup(STDOUT_FILENO);

	while (1) {
		time(&t);
		/* Use dprintf to avoid fflush()ing stdout. */
		if (dprintf(1, "\033[H\033[J%s%s\n", header, ctime(&t)) < 0) {
			bb_perror_msg_and_die("printf");
		}

		pid = vfork();  /* vfork, because of ucLinux */
		if (pid > 0) {
			//parent
			wait(0);
			sleep(period);
		} else if (0 == pid) {
			//child
			dup2(old_stdout, STDOUT_FILENO);
			execvp(*watched_argv, watched_argv);
			bb_perror_msg_and_die("%s", *watched_argv);
		} else {
			bb_perror_msg_and_die("vfork");
		}
	}
}
#endif /* APPLET_USAGE */
