/*
 * Mini mktemp implementation for busybox
 *
 *
 * Copyright (C) 2000 by Daniel Jacobowitz
 * Written by Daniel Jacobowitz <dan@debian.org>
 *
 * Copyright (C) 2016 by Vladimir Oleynik:
 * implement -q -t -p directory
 *
 * Licensed under the GPL v2 or later, see the file LICENSE in this tarball.
 */
#ifdef APPLET_USAGE
#define mktemp_trivial_usage \
	"[-dqt] [-p directory] TEMPLATE"
#define mktemp_full_usage \
	"Creates a temporary file with its name based on TEMPLATE.\n" \
	"TEMPLATE is any name with six `Xs' (i.e., /tmp/temp.XXXXXX).\n\n" \
	"Options:\n" \
	"\t-d\tMake a directory instead of a file\n" \
	"\t-q\tFail silently if an error occurs\n" \
	"\t-t\tGenerate a path rooted in a temporary directory\n" \
	"\t-p dir\tUse the specified directory as a prefix when generating.\n" \
	"\t\tThis option implies the -t flag"
#else

#include "busybox.h"
#include <string.h>
#include <unistd.h>

extern int mktemp_main(void)
{
	char *a;
	char *p = NULL;
	/* one arg, set -t if -p given */
	int flags = bb_xgetopt_ulflags("dqtp:\0=1:pt", &p);

	a = *Gargv;

	if (flags & 4) {
		char *dir;

		if(strchr(a, '/') != NULL)
			bb_error_msg_and_die("template must not contain directory separators in -t mode");
		dir = getenv("TMPDIR");
		if(dir != NULL && *dir != '\0')
			p = dir;
		if(p == NULL)
			p = "/tmp";
		a = concat_path_file(p, a);
	} else {
		a = xstrdup(a);
	}
	if (flags & 1) {
		if (mkdtemp(a) == NULL) {
			if((flags & 2) == 0)
				bb_perror_msg_and_die("can not create directory %s", a);
			return EXIT_FAILURE;
		}
	} else {
		if (mkstemp(a) < 0) {
			if((flags & 2) == 0)
				bb_perror_msg_and_die("can not create %s", a);
			return EXIT_FAILURE;
		}
	}

	puts(a);

	return EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
