#!/bin/sh

# ip of our external interface
LISTEN=xxx.xxx.xxx.xxx
PORT=12345
TIMEOUT=30
OPEN_TIME=15
# need absolute path
KNOCK_PASSWD=/etc/ssh_port_knock.txt
BUSYBOX="/bin/busybox"

which getpeername iptables md5sum od realpath > /dev/null || exit 1

usage() {
	echo "Usage: $0 [-l listen_ip] [-p listen_port] [-t open_time_seconds]" >&2
	exit 2
}

from_net=
addr=
while getopts "l:p:t:IA:" o; do
    case "$o" in
	l) LISTEN=$OPTARG ;;
	p) PORT=$OPTARG ;;
	t) OPEN_TIME=$OPTARG ;;
	I) from_net=1 ;;
	A) addr=$OPTARG ;;
	*) usage ;;
    esac
    done
shift $((OPTIND-1))

[ $# -ne 0 ] && usage

if [ -n "$addr" ]; then
	# open_time daemon

	echo "addr=$addr openned $OPEN_TIME seconds"
	iptables -I INPUT -s "$addr" -d "$LISTEN" -p tcp --syn --dport 22 -j ACCEPT 2>&1 || exit 1
	exec >&-
	sleep "$OPEN_TIME"
	iptables -D INPUT -s "$addr" -d "$LISTEN" -p tcp --syn --dport 22 -j ACCEPT
	exit $?
fi

prg="$(realpath -e "$0")"
[ $? -ne 0 ] && exit 1

if [ -z "$from_net" ]; then
	# start daemon

	# allow only established ssh connection
	iptables -A INPUT -s 0.0.0.0/0 -d $our_ip -m state --state ESTABLISHED -p tcp --dport 22 -j ACCEPT || exit 1
	iptables -A INPUT -s 0.0.0.0/0 -d "$LISTEN" -p tcp --syn --dport 22 -j DROP

	"$BUSYBOX" inetd -b "$LISTEN" "$PORT" "$prg" "-I" "-t$OPEN_TIME"
	exit $?
fi

# The inetd service

{ read port; read addr; } << EOF
$(getpeername)
EOF
# allow from network only
[ -z "$port" -o -z "$addr" ] && exit 1

# make RANDOM
read r1 r2 r3 r4 << EOF
$({ dd if=/dev/urandom bs=4 count=1 ; dd if=/dev/random bs=12 count=1 ; } 2> /dev/null | od -An -t x4)
EOF
r=$r1$r2$r3$r4

echo "$r"
# read hash from network
p=$("$BUSYBOX" eline -d '\n' -t 15 -n 34 -L -z 4 -Z ' ')
if [ $? -eq 125 ]; then
	echo -e "\nTimeout"
	exit 125
fi

while read pp; do
	m=${pp#?}
	[ "x${pp%"$m"}" = 'x#' ] && continue
	read pa m << EOF
$(echo "$pp $r" | md5sum)
EOF
	if [ "x$pa" = "x$p" ]; then
		# start open_time daemon
		env --block-signal=HUP -C / -- "$prg" "-A$addr" "-t$OPEN_TIME" < /dev/null 2> /dev/null &
		exit 0
	fi
done < "$KNOCK_PASSWD"

sleep 3
echo "Password incorrect"
exit 1
