//#include "platform.h"         This file must use after include libbb.h
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <net/if.h>

/* bb_gethostbyname flags */
#define BB_GHBN_IN_ADDR     1   /* next arg is (struct in_addr *) for set first found IPv4 address */
#define BB_GHBN_IN6_ADDR    2   /* next arg is (struct in6_addr *) for set first found IPv6 address */
#define BB_GHBN_SOCKADDR    4   /* next arg is (struct sockaddr *) for zeroed,
				   set family and set first found IPv4 or IPv6 address */
#define BB_GHBN_SOCKADDR_IN 8   /* next arg is (struct sockaddr_in *) for zeroed,
				   set family and set first found IPv4 address only */
#define BB_GHBN_SOCKADDR_IN6 16 /* next arg is (struct sockaddr_in6 *) for zeroed,
				   set family and set first found IPv6 address only */
#define BB_GHBN_NADDR       32  /* next arg is (int *) for set number of found IPv4 addresses */
#define BB_GHBN_NADDR6      64  /* next arg is (int *) for set number of found IPv6 addresses */
#define BB_GHBN_NSOCKADDR   128 /* next arg is (int *) for set number of found IPv4 or IPv6 addresses */
#define BB_GHBN_ADD_ERR_MSG 256 /* next arg is add messages for warning */
#define BB_GHBN_SILENT      512 /* do not show warning */
#define BB_GHBN_ERR_DIE     1024/* exit(default_error) if error */
#define BB_GHBN_CANON_NAME  2048/* next arg is (char **) for xstrdup(canonically name) */
#define BB_GHBN_NTOP        4096/* return dup(inet_ntop()) of found first IP address */
/* return:
   * (struct in_addr *) for BB_GHBN_IN_ADDR or BB_GHBN_SOCKADDR_IN or BB_GHBN_NADDR
   * (struct in6_addr *) for BB_GHBN_IN6_ADDR or BB_GHBN_SOCKADDR_IN6 or BB_GHBN_NADDR6
   * (struct sockaddr *) for BB_GHBN_SOCKADDR or BB_GHBN_NSOCKADDR
   * (char *) for BB_GHBN_NTOP only
   * (char *) for BB_GHBN_CANON_NAME only
   * NULL if getaddrinfo() return error and BB_GHBN_ERR_DIE is not set */
extern void *bb_gethostbyname(int flags, const char *name, ...);
extern int xconnect(const struct sockaddr *s_addr) FAST_FUNC;
extern unsigned short bb_lookup_tcp_port(const char *service, unsigned short default_port) FAST_FUNC;

/*
 * stolen from net-tools-1.59 and stripped down for busybox by
 *                      Erik Andersen <andersen@codepoet.org>
 *
 * Heavily modified by Manuel Novoa III       Mar 12, 2001
 *
 *
 */

/* hostfirst!=0 If we expect this to be a hostname,
   try hostname database first
 */
extern int INET_resolve(const char *name, struct sockaddr_in *s_in,
						int hostfirst) FAST_FUNC;


/* numeric: & 0x8000: default instead of *,
 *          & 0x4000: host instead of net,
 *          & 0x0fff: don't resolve
 */
extern int INET_rresolve(char *name, size_t len, struct sockaddr_in *s_in,
			 int numeric, unsigned int netmask) FAST_FUNC;

#if ENABLE_FEATURE_IPV6
extern int INET6_resolve(const char *name, struct sockaddr_in6 *sin6) FAST_FUNC;
extern int INET6_rresolve(char *name, size_t len, struct sockaddr_in6 *sin6,
							int numeric) FAST_FUNC;
#endif

extern void safe_ifname_cp(struct ifreq *ifr, const char *ifname) FAST_FUNC;
extern void safe_ifname_cp0(struct ifreq *pifr, const char *ifname) FAST_FUNC;

#define BB_XGETHOSTBYADDR_FLG_SHOW_ERROR 1
#define BB_XGETHOSTBYADDR_FLG_ERROR_DIE  2
#define BB_XGETHOSTBYADDR_FLG_IN_ADDR    4 /* a is (struct in_addr *) */
#define BB_XGETHOSTBYADDR_FLG_IN6_ADDR   8 /* a is (struct in6_addr *) */
#define BB_XGETHOSTBYADDR_FLG_NTOP       16/* return inet_ntop() if can not rresolve */
FAST_FUNC char *bb_xgethostbyaddr(const void *a, char hbuf[NI_MAXHOST], int flags);

#define BB_XSOCKET_FLG_AF_INET6   1     /* default AF_INET */
#define BB_XSOCKET_FLG_PF_PACKET  2
#define BB_XSOCKET_FLG_AF_UNIX    4
#define BB_XSOCKET_FLG_SOCK_DGRAM 8     /* default SOCK_STREAM */
#define BB_XSOCKET_FLG_SOCK_RAW   16
#define BB_XSOCKET_FLG_PROTO_SPEC 32
#define BB_XSOCKET_FLG_NO_WARN    64
#define BB_XSOCKET_FLG_ERR_DIE    128
#define BB_XSOCKET_FLG_AF_SPEC    256

extern int bb_xsocket(int flags, ...);

extern int setsockopt_1i(int sockfd, int level, int optname, int int_opt) FAST_FUNC;
extern void bb_xbind(int sockfd, struct sockaddr *addr, socklen_t len) FAST_FUNC;
extern unsigned short in_cksum(const void *addr, int len, unsigned short csum) FAST_FUNC;
