/*
 * Busybox main internal header file
 *
 * Licensed under the GPL v2 or later, see the file LICENSE in this tarball.
 *
 */
#ifndef __LIBBUSYBOX_H__
#define __LIBBUSYBOX_H__    1

#include <sys/types.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <stdint.h>
#include <features.h>

#if ENABLE_FEATURE_LONG_OPTIONS
#include <getopt.h> /* struct option */
#endif

#include "platform.h"
#include "bb_config.h"
#include "bb_ac.h"

/* Try to pull in PATH_MAX */
#include <limits.h>

/* for PATH_MAX on systems that don't have it in limits.h */
#include <sys/param.h>
#ifndef PATH_MAX
#define  PATH_MAX         4096
#endif

#ifdef DMALLOC
#include <dmalloc.h>
#endif

#include "pwd_.h"
#include "grp_.h"

#include <errno.h>
#if defined(__GLIBC__) || defined(__BIONIC__) || defined(__UCLIBC_HAS_THREADS__)
#undef errno
extern int *errnoptr;
#define errno (*errnoptr)
#endif

/* Some useful definitions */
#define FALSE   ((int) 0)
#define TRUE    ((int) 1)
#define SKIP    ((int) 2)

/* for mtab.c */
#define MTAB_GETMOUNTPT '1'
#define MTAB_GETDEVICE  '2'

/* Macros for min/max.  */
#ifndef MIN
#define MIN(a,b) (((a)<(b))?(a):(b))
#endif

#ifndef MAX
#define MAX(a,b) (((a)>(b))?(a):(b))
#endif

#ifndef BUFSIZ
#define BUFSIZ 4096
#endif

extern int Gargc;
extern char **Gargv;
extern char *f_l_a_proctitle[2];
extern const char *bb_applet_name;
extern int log_stderr;
extern int log_syslog;
extern int bb_default_error_retval;
USE_FEATURE_LONG_OPTIONS(extern const struct option *bb_applet_long_options;)

void libbb_init(int argc, char **argv);

extern void bb_setproctitle(const char *format, ...) __attribute__ ((format (printf, 1, 2)));


extern void bb_show_usage(void) ATTRIBUTE_NORETURN;
extern void bb_error_msg(const char *s, ...) __attribute__ ((format (printf, 1, 2)));
extern void bb_error_msg_and_die(const char *s, ...) __attribute__ ((noreturn, format (printf, 1, 2)));
extern void bb_perror_msg(const char *s, ...) __attribute__ ((format (printf, 1, 2)));
extern void bb_perror_msg_and_die(const char *s, ...) __attribute__ ((noreturn, format (printf, 1, 2)));

extern void bb_perror_nomsg_and_die(void) ATTRIBUTE_NORETURN;
extern void bb_perror_nomsg(void);

extern void xlog_warn(const char* fmt, ...) __attribute__ ((format (printf, 1, 2)));

/* These two are used internally -- you shouldn't need to use them */
extern void bb_verror_msg(const char *s, va_list p)
		__attribute__ ((format (printf, 1, 0))) FAST_FUNC;
extern void bb_vperror_msg(const char *s, va_list p)
		__attribute__ ((format (printf, 1, 0))) FAST_FUNC;

extern int xioctl(int d, int request, void *a, const char *msg, ...) FAST_FUNC;

extern int bb_echo(char** argv) FAST_FUNC;

extern const char *bb_mode_string(int mode, char buf[12]) FAST_FUNC;
extern int is_directory(const char *name, int followLinks, struct stat *statBuf) FAST_FUNC;

extern int remove_file(const char *path, int flags) FAST_FUNC;
extern int copy_file(const char *source, const char *dest, int flags) FAST_FUNC;
extern ssize_t safe_read(int fd, void *buf, size_t count) FAST_FUNC;
extern ssize_t bb_full_read(int fd, void *buf, size_t len) FAST_FUNC;
extern ssize_t safe_write(int fd, const void *buf, size_t count) FAST_FUNC;
extern ssize_t bb_full_write(int fd, const void *buf, size_t len) FAST_FUNC;
extern ssize_t write_all(int fd, const void *buf, size_t count) FAST_FUNC;
extern int recursive_action(const char *fileName, int recurse,
	  int followLinks, int depthFirst,
	  int (*fileAction) (const char *fileName, struct stat* statbuf, void* userData),
	  int (*dirAction) (const char *fileName, struct stat* statbuf, void* userData),
	  void* userData);


extern int get_kernel_revision(void) FAST_FUNC;

extern struct mntent *find_mount_point(const char *name, const char *table) FAST_FUNC;
extern long *find_pid_by_name(const char *pidName) FAST_FUNC;
extern char *find_block_device(char *path) FAST_FUNC;
extern int   get_console_fd_and_die(void) FAST_FUNC;
extern void  erase_mtab(const char * name) FAST_FUNC;

extern char *bb_get_line_from_file(FILE *file) FAST_FUNC;
extern char *bb_get_chomped_line_from_file(FILE *file) FAST_FUNC;
extern char *bb_get_chunk_from_file(FILE *file, int *end) FAST_FUNC;
extern int   bb_copyfd_size(int fd1, int fd2, const off_t size) FAST_FUNC;
extern int   bb_copyfd_eof(int fd1, int fd2) FAST_FUNC;
extern void  bb_xprint_and_close_file(FILE *file) FAST_FUNC;
extern char  bb_process_escape_sequence(const char **ptr) FAST_FUNC;
extern char *bb_get_last_path_component(char *path) FAST_FUNC;
extern FILE *bb_xfopen(const char *path, const char *mode) FAST_FUNC;
extern int   bb_fclose_nonstdin(FILE *f) FAST_FUNC;
extern void  bb_fflush_stdout_and_exit(int retval) ATTRIBUTE_NORETURN FAST_FUNC;
extern int   bb_parse_mode(const char *s, mode_t *theMode) FAST_FUNC;

extern unsigned long bss_flags;
extern unsigned long bb_getopt_ulflags(const char *applet_opts, ...);
extern unsigned long bb_xgetopt_ulflags(const char *applet_opts, ...);
extern char *xgetopt(const char *applet_opts) FAST_FUNC;

extern int bb_vfprintf(FILE * __restrict stream, const char * __restrict format,
					   va_list arg) __attribute__ ((format (printf, 2, 0)));
extern int bb_vprintf(const char * __restrict format, va_list arg)
	__attribute__ ((format (printf, 1, 0)));
extern int bb_fprintf(FILE * __restrict stream, const char * __restrict format, ...)
	__attribute__ ((format (printf, 2, 3)));
extern int bb_printf(const char * __restrict format, ...) __attribute__ ((format (printf, 1, 2)));

//#warning rename to xferror_filename?
extern void bb_xferror(FILE *fp, const char *fn) FAST_FUNC;
extern void bb_xferror_stdout(void) FAST_FUNC;
extern void bb_xfflush_stdout(void) FAST_FUNC;

extern void xchdir(const char *path) FAST_FUNC;
extern char *strchr_set0(char *s, int c) FAST_FUNC;

extern void chomp(char *s) FAST_FUNC;
extern void trim(char *s) FAST_FUNC;
extern char *bb_skip_whitespace(const char *) FAST_FUNC;

extern struct BB_applet *find_applet_by_name(const char *name);

/* dmalloc will redefine these to it's own implementation. It is safe
 * to have the prototypes here unconditionally.  */
extern void *xmalloc(size_t size) FAST_FUNC;
extern void *zmalloc(size_t size) FAST_FUNC;
extern void *xrealloc(void *old, size_t size) FAST_FUNC;

extern char *xstrdup(const char *s) FAST_FUNC;
extern char *xstrndup(const char *s, size_t n) FAST_FUNC;
extern char *safe_strncpy(char *dst, const char *src, size_t size) FAST_FUNC;
extern int safe_strtod(char *arg, double* value) FAST_FUNC;

struct suffix_mult {
	const char *suffix;
	unsigned int mult;
};

/* universal parse number */
#ifndef bb_xparse_i_number
extern int bb_xparse_i_number(int flags, const char *arg, ...);
#endif
extern long bb_xparse_l_number(int flags, const char *arg, ...);
#ifndef bb_xparse_im_number
extern intmax_t bb_xparse_im_number(int flags, const char *arg, ...);
#endif
#ifndef bb_xparse_ui_number
extern unsigned int bb_xparse_ui_number(int flags, const char *arg, ...);
#endif
extern unsigned long bb_xparse_ul_number(int flags, const char *arg, ...);
#ifndef bb_xparse_uim_number
extern uintmax_t bb_xparse_uim_number(int flags, const char *arg, ...);
#endif
#include "opt_gen.h"
BITS_GEN(BB_PN, BASE_1, BASE_2, BASE_4, BASE_8, BASE_16, BASE_HIGH_BIT,
SET_LOWER, SET_LOWER1, LOWER_MIN, SET_UPPER, END_PTR,
ALLOW_TAIL_SPACE, SUFFIX, BKM_SUFFIXES,
SHOW_USAGE, ERR_STR1, ERR_STRV, SILENT, RETURN_ERR)
#define BB_PN_SILENT_RE (BB_PN_SILENT | BB_PN_RETURN_ERR)

/* DO NOT CHANGE THESE VALUES!  cp.c depends on them. */
BITS_GEN(FILEUTILS, PRESERVE_STATUS, DEREFERENCE, RECUR, FORCE, INTERACTIVE)

BITS_GEN(XDAEMON, NO_CLOSE, NO_CHDIR, SET_DIR, PIDFILE, PF_ARG, NOFORK, NO_WRITE_PID)
#include "opt_gen.h"
extern char *xdaemon(int flags, ...) FAST_FUNC;


/* These parse entries in /etc/passwd and /etc/group.  This is desirable
 * for BusyBox since we want to avoid using the glibc NSS stuff, which
 * increases target size and is often not needed on embedded systems.  */
extern long  bb_xgetpwnam(const char *name);
extern long  bb_xgetgrnam(const char *name);
extern char *bb_getpwuid(char *name, long uid, int bufsize) FAST_FUNC;
extern char *bb_getgrgid(char *group, long gid, int bufsize) FAST_FUNC;
extern unsigned long get_ug_id(const char *s, long (*__bb_getxxnam)(const char *));

extern int bb_ask_confirmation(void);
#define PASSWORD_BUFFER_SIZE 256
extern char *bb_askpass(int timeout, const char * prompt,
				char pb[PASSWORD_BUFFER_SIZE]) FAST_FUNC;
#define ENCRYPT_BUFFER_SIZE 128
extern char *pw_encrypt(const char *clear, const char *salt,
				char enc[ENCRYPT_BUFFER_SIZE]) FAST_FUNC;

extern int device_open(const char *device, int mode) FAST_FUNC;

extern char *query_loop(const char *device) FAST_FUNC;
extern int   del_loop(const char *device) FAST_FUNC;
extern int   set_loop(char **device, const char *file, int offset) FAST_FUNC;

#if (__GLIBC__ < 2)
extern int vdprintf(int d, const char *format, va_list ap);
#endif

enum {
	KILOBYTE = 1024,
	MEGABYTE = (KILOBYTE*1024),
	GIGABYTE = (MEGABYTE*1024)
};
const char *make_human_readable_str(unsigned long long size,
	      unsigned long block_size, unsigned long display_unit, char buf[32]) FAST_FUNC;

int klogctl(int type, char * b, int len);

char *xgetcwd(char *cwd) FAST_FUNC;
char *xreadlink(const char *path) FAST_FUNC;
char *concat_path_file(const char *path, const char *filename) FAST_FUNC;
char *concat_subpath_file(const char *path, const char *filename) FAST_FUNC;
char *last_char_is(const char *s, int c) FAST_FUNC;

extern int bb_uncompress(int fd_in, int fd_out);

int bb_make_directory(char *path, long mode, int flags) FAST_FUNC;

const char *u_signal_num2name(int signo, int minsig, char *buf) FAST_FUNC;
const char *u_signal_name2num(const char *str_sig, int *signo, int minsig, char *buf) FAST_FUNC;

char *bb_simplify_path(const char *path) FAST_FUNC;


extern const char bb_msg_full_version[];
extern const char bb_msg_memory_exhausted[];
extern const char bb_msg_invalid_date[];
extern const char bb_msg_read_error[];
extern const char bb_msg_write_error[];
extern const char bb_msg_name_longer_than_foo[];
extern const char bb_msg_unknown[];
extern const char bb_msg_perm_denied_are_you_root[];
extern const char bb_msg_standard_input[];
extern const char bb_msg_standard_output[];

extern const char bb_path_nologin_file[];
extern const char bb_path_passwd_file[];
extern const char bb_path_shadow_file[];
extern const char bb_path_gshadow_file[];
extern const char bb_path_group_file[];
extern const char bb_path_securetty_file[];
extern const char bb_path_motd_file[];
extern const char bb_dev_null[];
extern const char bb_dev_tty[];
extern const char bb_default[];
extern const char bb_spacestok[];

/*
 * You can change LIBBB_DEFAULT_LOGIN_SHELL, but don`t use,
 * use bb_default_login_shell and next defines,
 * if you LIBBB_DEFAULT_LOGIN_SHELL change,
 * don`t lose change increment constant
 * and DEFAULT_LOGIN_SHELL_SHORT_NAME!
 */
#define LIBBB_DEFAULT_LOGIN_SHELL      "/bin/sh"

extern const char bb_default_login_shell[];
#define DEFAULT_SHELL     bb_default_login_shell
/* "sh" */
#define DEFAULT_SHELL_SHORT_NAME        (bb_default_login_shell+5)
/* "-sh" */
#define DEFAULT_LOGIN_SHELL_SHORT_NAME  "-sh"

extern const char bb_login_path[];
#define LOGIN_SHORT_NAME                (bb_login_path+5) /* offset /bin/ */

extern const char bb_path_mtab_file[];

#define VC_5 "/dev/tty5"
# if defined(__sh__) || defined(__H8300H__) || defined(__H8300S__)
#  define SC_0 "/dev/ttySC0"
#  define SC_1 "/dev/ttySC1"
#  define SC_FORMAT "/dev/ttySC%d"
# else
#  define SC_0 "/dev/ttyS0"
#  define SC_1 "/dev/ttyS1"
#  define SC_FORMAT "/dev/ttyS%d"
# endif
#define VC_FORMAT "/dev/tty%d"
#define LOOP_FORMAT "/dev/loop%d"
#define FB_0 "/dev/fb0"

//#warning put these in .o files

#define CONSOLE_DEV "/dev/console"

int is_in_ino_dev_hashtable(const struct stat *statbuf, char **name, void *ht) FAST_FUNC;
void add_to_ino_dev_hashtable(const struct stat *statbuf, const char *name, void *ht) FAST_FUNC;

/* Stupid gcc always includes its own builtin strlen()... */
extern size_t bb_strlen(const char *string);
#ifndef BB_STRLEN_IMPLEMENTATION
#define strlen(x)   bb_strlen(x)
#endif

char *xasprintf(const char *format, ...) __attribute__ ((format (printf, 1, 2)));

#define FAIL_DELAY    3
extern void bb_do_delay(int seconds) FAST_FUNC;

extern void change_identity(const struct passwd *pw ) FAST_FUNC;
extern const char *change_identity_e2str(const struct passwd *pw ) FAST_FUNC;
extern void run_shell(const char *shell, const char *command,
			const char **additional_args
			) ATTRIBUTE_NORETURN FAST_FUNC;
extern int bb_setenv1(const char *key, const char *value) FAST_FUNC;
extern int restricted_shell(const char *shell ) FAST_FUNC;
extern void setup_environment(const char *shell, int loginshell, int changeenv, const struct passwd *pw) FAST_FUNC;
extern int correct_password(const struct passwd *pw ) FAST_FUNC;
extern int obscure(const char *old, const char *newval, const struct passwd *pwdp) FAST_FUNC;
extern void print_login_issue(const char *issue_file, const char *tty) FAST_FUNC;
extern void print_login_prompt(void) FAST_FUNC;

extern ssize_t bb_xread(int fd, void *buf, size_t count) FAST_FUNC;
extern void bb_xread_all(int fd, void *buf, size_t count) FAST_FUNC;
extern unsigned char bb_xread_char(int fd) FAST_FUNC;

extern int compare_string_array(const char * const string_array[], const char *key) FAST_FUNC;
extern void *find_string_from_struct(const void *p, unsigned offset, const char *key) FAST_FUNC;

typedef struct llist_s {
	void *data;
	struct llist_s *link;
} llist_t;
extern llist_t *llist_add_to(llist_t *old_head, void *new_item) FAST_FUNC;
extern llist_t *llist_add_to_end(llist_t *list_head, void *data) FAST_FUNC;
extern llist_t *llist_free_one(llist_t *elm) FAST_FUNC;
extern void llist_free(llist_t *elm) FAST_FUNC;


extern int get_terminal_width_height(int *width, int *height) FAST_FUNC;

extern int bb_cloexec(int fd) FAST_FUNC;

extern void bb_openlog(int option, int facility) FAST_FUNC;

extern int waitpid_no_int(pid_t wpid, int *pstatus, int flags, int flg_wpid_only);

extern int bb_utime(const char *pathname, time_t actime, time_t modtime) FAST_FUNC;
extern const char *bb_basename(const char *name) FAST_FUNC;

extern void str_tolower(char *str) FAST_FUNC;


/* Options for size_to_human_string() */
enum
{
	SIZE_SUFFIX_1LETTER = 0,
	SIZE_SUFFIX_3LETTER = 1,
	SIZE_SUFFIX_SPACE   = 2
};
#define SIZE_TO_HUMAN_STRING_BUFSIZ 32
extern void size_to_human_string(int options, ubb64_t bytes,
				char buf[SIZE_TO_HUMAN_STRING_BUFSIZ]) FAST_FUNC;

extern void crc32_filltable(ubb32_t crc_table[256], int flg_big_endian) FAST_FUNC;

#ifdef CONFIG_FEATURE_UTF8_SUPPORT
FAST_FUNC int mbchar_len(const char *s);
#else
# define mbchar_len(s) bb_strlen(s)
#endif

extern void random_get_bytes(void *buf, size_t nbytes) FAST_FUNC;
extern int xusleep(unsigned int usec) FAST_FUNC;
extern int fsync_close(int fd) FAST_FUNC;
extern void bb_wutmp_logout(char *line) FAST_FUNC;
#endif /* __LIBBUSYBOX_H__ */
