/*
 * Ask for a password
 * I use a static buffer in this function.  Plan accordingly.
 *
 * Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */

#include "libbb.h"
#include "bb_sig.h"
#include <string.h>
#include <unistd.h>
#include <termios.h>
#include <sys/ioctl.h>

/* do nothing signal handler */
static void askpass_timeout(int ATTRIBUTE_UNUSED ignore)
{
}

#define size PASSWORD_BUFFER_SIZE
FAST_FUNC char *bb_askpass(int timeout, const char * prompt,
				char passwd[PASSWORD_BUFFER_SIZE])
{
	char *ret;
	int i;
	struct termios old, new;

	tcgetattr(STDIN_FILENO, &old);
	tcflush(STDIN_FILENO, TCIFLUSH);

	ret = passwd;
	memset(passwd, 0, size);

	fputs(prompt, stdout);
	fflush(stdout);

	tcgetattr(STDIN_FILENO, &new);
	new.c_iflag &= ~(IUCLC|IXON|IXOFF|IXANY);
	new.c_lflag &= ~(ECHO|ECHOE|ECHOK|ECHONL|TOSTOP);
	tcsetattr(STDIN_FILENO, TCSANOW, &new);

	if (timeout) {
		/* do not restart syscall */
		bb_signal( 1 | (1 << SIGALRM), askpass_timeout);
		alarm(timeout);
	}

	if (read(STDIN_FILENO, passwd, size-1) <= 0) {
		ret = NULL;
	} else {
		for(i = 0; i < size && passwd[i]; i++) {
			if (passwd[i] == '\r' || passwd[i] == '\n') {
				passwd[i] = 0;
				break;
			}
		}
	}

	if (timeout)
		alarm(0);

	tcsetattr(STDIN_FILENO, TCSANOW, &old);
	puts("");
	fflush(stdout);
	return ret;
}

#ifndef HAVE_GETPASS
char *getpass(const char *prompt)
{
	static char passwd[PASSWORD_BUFFER_SIZE];
	return bb_askpass(0, prompt, passwd);
}
#endif
