/*
 * Copyright (C) 2008 Vladimir Oleynik <dzo@simtreas.ru>
*/

#include "libbb.h"
#include <string.h>
#include "bb_sig.h"

FAST_FUNC sigset_t bb_sigmask(unsigned int sigs_bits)
{
	unsigned int b = 2;
	int signo = 1;
	sigset_t m;

	sigemptyset(&m);
	while(b) {
		if((sigs_bits & b) != 0)
			sigaddset(&m, signo);
		signo++;
		b <<= 1;
	}
	return m;
}

FAST_FUNC sigset_t bb_sigblock(unsigned int sigs_bits)
{
	sigset_t m;

	m = bb_sigmask(sigs_bits);
	sigprocmask(SIG_BLOCK, &m, NULL);
	return m;
}

FAST_FUNC void bb_sigunblock(sigset_t m)
{
	sigprocmask(SIG_UNBLOCK, &m, NULL);
}

FAST_FUNC int bb_sigaction2(int signum, const struct sigaction *act)
{
	return sigaction(signum, act, NULL);
}

FAST_FUNC sighandler_t bb_signal(unsigned int sigs_bits, sighandler_t h)
{
	unsigned int b = 2;
	int signo = 1;
	struct sigaction act, old;

	memset(&act, 0, sizeof(act));
	act.sa_handler = h;
	if((sigs_bits & 1) == 0)        /* use special signo=0 bit */
		act.sa_flags |= SA_RESTART;     /* bsd signal compatibility */
	while(b) {
		if((sigs_bits & b) != 0) {
			/* bsd signal compatibility */
			sigaddset(&act.sa_mask, signo);
			sigaction(signo, &act, &old);
		}
		signo++;
		b <<= 1;
	}
	return old.sa_handler;
}
