/*
 * special for BIONIC
 * code stolen from uClibc
 */

#include <sys/types.h>
#include <features.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include "bb_ac.h"

#include <errno.h>
#if defined(__GLIBC__) || defined(__BIONIC__) || defined(__UCLIBC_HAS_THREADS__)
#undef errno
extern int *errnoptr;
#define errno (*errnoptr)
#endif

#ifndef HAVE_ETHER_ATON
#include <netinet/ether.h>
#include <netinet/if_ether.h>
#endif
#ifndef HAVE_SYSINFO
#include <sys/syscall.h>
#include <sys/sysinfo.h>
#endif
#ifndef HAVE_SETHOSTNAME
#include <fcntl.h>
#endif
#ifndef HAVE_OPENPTY
#include <fcntl.h>
#endif
#ifndef HAVE_KILLPG
#include <signal.h>
#endif

#ifndef HAVE_STPCPY
char *stpcpy(char * s1, const char * s2)
{
	while ( (*s1++ = *s2++) != 0 );
	return s1 - 1;
}
#endif

#ifndef HAVE_MEMPCPY
void *mempcpy(void * s1, const void * s2, size_t n)
{
	char *r1 = s1;
	const char *r2 = s2;

	while (n) {
		*r1++ = *r2++;
		--n;
	}
	return r1;
}
#endif


#ifndef HAVE_ETHER_ATON
static struct ether_addr *ether_aton_r(const char *asc, struct ether_addr *addr)
{
	size_t cnt;

	for (cnt = 0; cnt < 6; ++cnt) {
		unsigned int number;
		char ch;

		ch = tolower(*asc++);
		if ((ch < '0' || ch > '9') && (ch < 'a' || ch > 'f'))
			return NULL;
		number = isdigit(ch) ? (ch - '0') : (ch - 'a' + 10);

		ch = tolower(*asc);
		if ((cnt < 5 && ch != ':')
			|| (cnt == 5 && ch != '\0' && !isspace(ch))) {
			++asc;
			if ((ch < '0' || ch > '9') && (ch < 'a' || ch > 'f'))
				return NULL;
			number <<= 4;
			number += isdigit(ch) ? (ch - '0') : (ch - 'a' + 10);

			ch = *asc;
			if (cnt < 5 && ch != ':')
				return NULL;
		}

		/* Store result.  */
		addr->ether_addr_octet[cnt] = (unsigned char) number;

		/* Skip ':'.  */
		++asc;
	}

	return addr;
}

struct ether_addr *ether_aton(const char *asc)
{
	static struct ether_addr result;

	return ether_aton_r(asc, &result);
}

char *ether_ntoa_r(const struct ether_addr *addr, char *buf)
{
	sprintf(buf, "%x:%x:%x:%x:%x:%x",
		addr->ether_addr_octet[0], addr->ether_addr_octet[1],
		addr->ether_addr_octet[2], addr->ether_addr_octet[3],
		addr->ether_addr_octet[4], addr->ether_addr_octet[5]);
	return buf;
}

char *ether_ntoa(const struct ether_addr *addr)
{
	static char asc[18];

	return ether_ntoa_r(addr, asc);
}

#endif

#ifndef HAVE_SYSINFO
int sysinfo(struct sysinfo * info) {
	return syscall(__NR_sysinfo, info);
}
#endif

#ifndef HAVE_SETHOSTNAME
/* stolen from Android core init/builtins.c */
static int write_file(const char *path, const char *value, size_t len)
{
    int fd, ret;

    fd = open(path, O_WRONLY|O_CREAT, 0622);

    if (fd < 0)
	return -errno;

    do {
	ret = write(fd, value, len);
    } while (ret < 0 && errno == EINTR);

    close(fd);
    if (ret < 0) {
	return -errno;
    } else {
	return 0;
    }
}

int sethostname(const char *name, size_t len)
{
    return write_file("/proc/sys/kernel/hostname", name, len);
}
#endif

#ifndef HAVE_GETPWENT
#include "pwd_.h"
static uid_t bb_bionic_getpwent_counter;

struct passwd *getpwent(void)
{
	struct passwd *p;

	for(;;) {
		if(bb_bionic_getpwent_counter > 9999) {
			/* last (nobody) Bionic user */
			bb_bionic_getpwent_counter = 0;
			return NULL;
		}
		p = getpwuid(bb_bionic_getpwent_counter);
		bb_bionic_getpwent_counter++;
		if(p != NULL)
			return p;
	}
}

#if defined(__BIONIC__)        /* Bionic error, <pwd.h> have errnoneus declare as int setpwent(void); */
int setpwent(void)
{
	bb_bionic_getpwent_counter = 0;
	return 0;
}
#else
void setpwent(void)
{
	bb_bionic_getpwent_counter = 0;
}
#endif


void endpwent(void)
{
	bb_bionic_getpwent_counter = 0;
}
#endif

#ifndef HAVE_GETGRENT
#include <grp.h>
static gid_t bb_bionic_getgrent_counter;

struct group *getgrent(void)
{
	struct group *g;

	for(;;) {
		if(bb_bionic_getgrent_counter > 9999) {
			/* last (nobody) Bionic user = group */
			bb_bionic_getgrent_counter = 0;
			return NULL;
		}
		g = getgrgid(bb_bionic_getgrent_counter);
		bb_bionic_getgrent_counter++;
		if(g != NULL)
			return g;
	}
}

void setgrent(void)
{
	bb_bionic_getgrent_counter = 0;
}

void endgrent(void)
{
	bb_bionic_getgrent_counter = 0;
}
#endif

#ifndef HAVE_CLEARENV
/* code stolen from Bionic version 8 */
/*
 * Copyright (C) 2010 The Android Open Source Project
 * All rights reserved.
 *
 */

extern char** environ;

int clearenv(void)
{
    char **P = environ;

    if (P != NULL) {
	for (; *P; ++P)
	    *P = NULL;
    }
    return 0;
}
#endif

#ifndef HAVE_KILLPG
/* code stolen from Bionic version 8 */
/*
 * Copyright (C) 2010 The Android Open Source Project
 * All rights reserved.
 *
 */

/* Backwards-compatible killpg(). */
int killpg(pid_t pgid, int sig)
{
	if (pgid == 1) {
		errno = ESRCH;
		return -1;
	}
	return (kill(-pgid, sig));
}
#endif


#ifndef HAVE_STRCHRNUL
char *strchrnul(const char *s, int ch)
{
	while (*s && *s != ch)
		++s;
	return (char *)s;
}
#endif

#ifndef HAVE_FFS
int ffs(int x)
{
  return __builtin_ffs(x);
}
#endif
