/*
 * Utility routines.
 *
 * Copyright (C) 1999-2005 by Erik Andersen <andersen@codepoet.org>
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */

#include "libbb.h"
#include <string.h>
#include <unistd.h>


static FAST_FUNC ssize_t
bb_full_fd_action(int src_fd, int dst_fd, size_t size)
{
	int status = -1;
	size_t total = 0;
	char buffer[BUFSIZ];

	if (src_fd < 0) goto out;
	while (!size || total < size)
	{
		ssize_t wrote, xread;

		xread = safe_read(src_fd, buffer,
				(!size || size - total > BUFSIZ) ? BUFSIZ : size - total);

		if (xread > 0) {
			/* A -1 dst_fd means we need to fake it... */
			wrote = (dst_fd < 0) ? xread : bb_full_write(dst_fd, buffer, xread);
			if (wrote < xread) {
				bb_perror_msg("%s", bb_msg_write_error);
				break;
			}
			total += wrote;
			if (total == size) status = 0;
		} else if (xread < 0) {
			bb_perror_msg("%s", bb_msg_read_error);
			break;
		} else if (xread == 0) {
			/* All done. */
			status = 0;
			break;
		}
	}

out:
	return status ? status : (ssize_t)total;
}


FAST_FUNC int bb_copyfd_size(int fd1, int fd2, const off_t size)
{
	if (size) {
		return(bb_full_fd_action(fd1, fd2, size));
	}
	return(0);
}

FAST_FUNC int bb_copyfd_eof(int fd1, int fd2)
{
	return(bb_full_fd_action(fd1, fd2, 0));
}
