/*
 * Utility routines.
 *
 * Copyright (C) many different people.  If you wrote this, please
 * acknowledge your work.
 * Licensed under the GPL v2 or later, see the file LICENSE in this tarball.
 *
 */

#include "libbb.h"
#include <unistd.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#define CURRENT_VC "/dev/tty0"

/* From <linux/kd.h> */
static const int KDGKBTYPE = 0x4B33;  /* get keyboard type */


static int open_a_console(const char *fnam)
{
	int fd;

	/* try read-write */
	fd = open(fnam, O_RDWR);

	/* if failed, try read-only */
	if (fd < 0 && errno == EACCES)
		fd = open(fnam, O_RDONLY);

	/* if failed, try write-only */
	if (fd < 0 && errno == EACCES)
		fd = open(fnam, O_WRONLY);

	return fd;
}

/*
 * Get an fd for use with kbd/console ioctls.
 * We try several things because opening /dev/console will fail
 * if someone else used X (which does a chown on /dev/console).
 */

FAST_FUNC int get_console_fd_and_die(void)
{
	int fd;

	static const char * const choise_console_names[] = {
		CONSOLE_DEV, CURRENT_VC, bb_dev_tty
	};

	for (fd = 2; fd >= 0; fd--) {
		int fd4name;
		int choise_fd;
		char arg;

		fd4name = open_a_console(choise_console_names[fd]);
	chk_std:
		choise_fd = fd4name >= 0 ? fd4name : fd;

		arg = 0;
		if (ioctl(choise_fd, KDGKBTYPE, &arg) == 0)
			return choise_fd;
		if(fd4name >= 0) {
			close(fd4name);
			fd4name = -1;
			goto chk_std;
		}
	}

	bb_error_msg_and_die("Couldn't get a file descriptor referring to the console");
	return fd;                      /* total failure */
}
