/*
 * Determine the width and height of the terminal.
 *
 * Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 * Copyright (C) 2008 by Vladimir Oleynik <dzo@simtreas.ru>
 *
 */

#include "libbb.h"
#include <unistd.h>
#include <termios.h>
#include <sys/ioctl.h>

/* It is perfectly ok to pass in a NULL for either width or for
 * height, in which case that value will not be set.  */
FAST_FUNC int get_terminal_width_height(int *width, int *height)
{
	struct winsize win = { 0, 0, 0, 0 };
	int ret = ioctl(STDOUT_FILENO, TIOCGWINSZ, &win);
	if (win.ws_row <= 1) win.ws_row = 25;
	if (win.ws_col <= 1) win.ws_col = 80;
	if (height) *height = (int) win.ws_row;
	if (width) *width = (int) win.ws_col;

	return ret;
}
