/*
 * Utility routines.
 *
 * Copyright (C) many different people.
 * If you wrote this, please acknowledge your work.
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 *
 */

#include "libbb.h"
#include <string.h>

#define HASH_SIZE       311             /* Should be prime */
#define hash_inode(i)   ((i) % HASH_SIZE)

typedef struct ino_dev_hash_bucket_struct {
  struct ino_dev_hash_bucket_struct *next;
  ino_t ino;
  dev_t dev;
  char name[1];
} ino_dev_hashtable_bucket_t;

/*
 * Return 1 if statbuf->st_ino && statbuf->st_dev are recorded in
 * `ino_dev_hashtable', else return 0
 *
 * If NAME is a non-NULL pointer to a character pointer, and there is
 * a match, then set *NAME to the value of the name slot in that
 * bucket.
 */
FAST_FUNC int is_in_ino_dev_hashtable(const struct stat *statbuf, char **name, void *ht)
{
	ino_dev_hashtable_bucket_t *bucket;
	ino_dev_hashtable_bucket_t **ino_dev_hashtable = *(ino_dev_hashtable_bucket_t ***)ht;

	if(ino_dev_hashtable != NULL) {
		bucket = ino_dev_hashtable[hash_inode(statbuf->st_ino)];
		while (bucket != NULL) {
			if ((bucket->ino == statbuf->st_ino) && (bucket->dev == statbuf->st_dev)) {
				if (name) *name = bucket->name;
				return 1;
			}
			bucket = bucket->next;
		}
	}
	return 0;
}

/* Add statbuf to statbuf hash table */
FAST_FUNC void add_to_ino_dev_hashtable(const struct stat *statbuf, const char *name, void *ht)
{
	int i;
	size_t s;
	ino_dev_hashtable_bucket_t **ino_dev_hashtable = *(ino_dev_hashtable_bucket_t ***)ht;
	ino_dev_hashtable_bucket_t *bucket;

	if(ino_dev_hashtable == NULL)
		ino_dev_hashtable = zmalloc(sizeof(ino_dev_hashtable_bucket_t *) * HASH_SIZE);
	if(name == NULL)
		name = "";
	s = strlen(name);
	bucket = xmalloc(sizeof(ino_dev_hashtable_bucket_t) + s);
	bucket->ino = statbuf->st_ino;
	bucket->dev = statbuf->st_dev;
	memcpy(bucket->name, name, s + 1);
	i = hash_inode(statbuf->st_ino);
	bucket->next = ino_dev_hashtable[i];
	ino_dev_hashtable[i] = bucket;
	*(ino_dev_hashtable_bucket_t ***)ht = ino_dev_hashtable;
}
