/*
 * Utility routines.
 *
 * Based in part on code from sash, Copyright (c) 1999 by David I. Bell
 * Permission has been granted to redistribute this code under the GPL.
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */

#include "libbb.h"

/*
 * Return stat(), and set errno ENOTDIR if a fileName is not a directory.
 * Nonexistent files return -1.
 * For check for directory stat_buffer may be NULL
 */
FAST_FUNC int
is_directory(const char *fileName, const int followLinks, struct stat *statBuf)
{
	int status;
	struct stat astatBuf;

	errno = 0;
	if (statBuf == NULL) {
	    /* set from auto stack buffer */
	    statBuf = &astatBuf;
	}

	if (followLinks)
		status = stat(fileName, statBuf);
	else
		status = lstat(fileName, statBuf);

	if (status >= 0) {
		status = S_ISDIR(statBuf->st_mode);
		if(!status)
			errno = ENOTDIR;
	}
	return status;
}
