#include "libbb.h"
#include <sys/ioctl.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include "pty_.h"

static int open_slave(int master, char *name, const char *s, int *aslave)
{
	int slave;

	if((slave = open(s, O_RDWR | O_NOCTTY)) < 0) {
		close(master);
	} else {
		if(name != NULL)
			strcpy(name, s);
		fchmod(slave, S_IRUSR | S_IWUSR);
	}
	*aslave = slave;
	return slave >= 0 ? 0 : slave;
}

int openpty(int *amaster, int *aslave, char *name, void *termp, void *winp)
{
	char buf[16];
	int i;
	static const char ptymajors[] = "pqrstuvwxyzabcdefghijklmnoABCDEFGHIJKLMNOPQRSTUVWXYZ";
	static const char ptyminors[] = "0123456789abcdef";
	int num_minors = strlen(ptyminors);
	int num_ptys = strlen(ptymajors) * num_minors;
	char namebuf[16];

	(void)termp;
	(void)winp;

#ifdef TIOCGPTN
	/* Try the unix98 way first. */
	if ((i = open("/dev/ptmx", O_RDWR)) >= 0) {
		int ptn;

		if (ioctl(i, TIOCGPTN, &ptn) >= 0) {
			snprintf(namebuf, sizeof(namebuf), "/dev/pts/%d", ptn);
			chmod(namebuf, S_IRUSR | S_IWUSR);
#ifdef TIOCSPTLCK
			ptn = 0;
			ioctl(i, TIOCSPTLCK, &ptn);
#endif
			*amaster = i;
			return open_slave(i, name, namebuf, aslave);
		}
	}
#endif /* TIOCGPTN */

	/* the old way - scan through the pty name space */
	for (i = 0; i < num_ptys; i++) {
		sprintf(buf, "/dev/pty%c%c", ptymajors[i / num_minors],
			 ptyminors[i % num_minors]);
		sprintf(namebuf, "/dev/tty%c%c",
		    ptymajors[i / num_minors], ptyminors[i % num_minors]);

		*amaster = open(buf, O_RDWR | O_NOCTTY);
		if (*amaster < 0)
			continue;

		return open_slave(*amaster, name, namebuf, aslave);
	}
	/* Failed to open any /dev/pty?? devices */
	return -1;
}
