#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include "bb_ac.h"
#ifndef HAVE_REGEX
/* regex from Uclibc Version: 0.9.30.2 */
/* Extended regular expression matching and search library,
   version 0.12.
   (Implements POSIX draft P1003.2/D11.2, except for some of the
   internationalization features.)
   Copyright (C) 1993-1999, 2000, 2001 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307 USA.  */

#include <string.h>
#include <string.h>
#include <strings.h>
#include <ctype.h>
#include <alloca.h>

#ifndef __USE_GNU
#define __USE_GNU
#endif
#include "regex.h"

static char re_syntax_table[256];

static void
init_syntax_once (void)
{
  int c;
  static int done = 0;
  if (done)
	return;
  memset (re_syntax_table, 0, sizeof re_syntax_table);
  for (c = 0; c < 256; ++c)
	if ((1 && isalnum (c)))
	  re_syntax_table[c] = 1;
  re_syntax_table['_'] = 1;
  done = 1;
}

typedef char boolean;

static int re_match_2_internal (struct re_pattern_buffer *bufp,
					const char *string1, int size1,
					const char *string2, int size2,
					int pos,
					struct re_registers *regs, int stop);

typedef enum
{
  no_op = 0,
  succeed,
  exactn,
  anychar,
  charset,
  charset_not,
  start_memory,
  stop_memory,
  duplicate,
  begline,
  endline,
  begbuf,
  endbuf,
  jump,
  jump_past_alt,
  on_failure_jump,
  on_failure_keep_string_jump,
  pop_failure_jump,
  maybe_pop_jump,
  dummy_failure_jump,
  push_dummy_failure,
  succeed_n,
  jump_n,
  set_number_at,
  wordchar,
  notwordchar,
  wordbeg,
  wordend,
  wordbound,
  notwordbound
} re_opcode_t;

int re_max_failures = 4000;

union fail_stack_elt
{
  unsigned char *pointer;
  int integer;
};

typedef union fail_stack_elt fail_stack_elt_t;

typedef struct
{
  fail_stack_elt_t *stack;
  unsigned size;
  unsigned avail;
} fail_stack_type;

typedef union
{
  fail_stack_elt_t word;
  struct
  {
	unsigned match_null_string_p:2;
	unsigned is_active:1;
	unsigned matched_something:1;
	unsigned ever_matched_something:1;
  } bits;
} register_info_type;

static char reg_unset_dummy;

static void store_op1 (re_opcode_t op, unsigned char *loc, int arg);
static void store_op2 (re_opcode_t op, unsigned char *loc, int arg1, int arg2);
static void insert_op1 (re_opcode_t op, unsigned char *loc, int arg, unsigned char *end);
static void insert_op2 (re_opcode_t op, unsigned char *loc, int arg1, int arg2, unsigned char *end);
static boolean at_begline_loc_p (const char *pattern, const char *p, reg_syntax_t syntax);
static boolean at_endline_loc_p (const char *p, const char *pend, reg_syntax_t syntax);
static reg_errcode_t
compile_range (unsigned int range_start, const char **p_ptr,
		const char *pend, char *translate,
		reg_syntax_t syntax, unsigned char *b);

typedef unsigned regnum_t;
typedef long pattern_offset_t;

typedef struct
{
  pattern_offset_t begalt_offset;
  pattern_offset_t fixup_alt_jump;
  pattern_offset_t inner_group_offset;
  pattern_offset_t laststart_offset;
  regnum_t regnum;
} compile_stack_elt_t;

typedef struct
{
  compile_stack_elt_t *stack;
  unsigned size;
  unsigned avail;
} compile_stack_type;

static boolean
group_in_compile_stack (compile_stack_type compile_stack, regnum_t regnum);

static reg_errcode_t
regex_compile (const char *pattern, size_t size,
			reg_syntax_t syntax, struct re_pattern_buffer *bufp)
{
  unsigned char c, c1;
  const char *p1;
  unsigned char *b;
  compile_stack_type compile_stack;
  const char *p = pattern;
  const char *pend = pattern + size;
  char *translate = (char *)bufp->translate;
  unsigned char *pending_exact = 0;
  unsigned char *laststart = 0;
  unsigned char *begalt;
  unsigned char *fixup_alt_jump = 0;
  regnum_t regnum = 0;

  compile_stack.stack =
	((compile_stack_elt_t *) malloc ((32) * sizeof (compile_stack_elt_t)));
  if (compile_stack.stack == NULL)
	return REG_ESPACE;
  compile_stack.size = 32;
  compile_stack.avail = 0;
  bufp->syntax = syntax;
  bufp->fastmap_accurate = 0;
  bufp->not_bol = bufp->not_eol = 0;
  bufp->used = 0;
  bufp->re_nsub = 0;
  init_syntax_once ();
  if (bufp->allocated == 0) {
	if (bufp->buffer) {
	  bufp->buffer = realloc (bufp->buffer,
					((32 * sizeof (unsigned char))) *
					sizeof (unsigned char));
	} else {
	  bufp->buffer = malloc (((32 * sizeof (unsigned char)) / sizeof (unsigned char)) *
				 sizeof (unsigned char));
	}
	if (!bufp->buffer)
	  return (free (compile_stack.stack), REG_ESPACE);
	bufp->allocated = (32 * sizeof (unsigned char));
  }
  begalt = b = bufp->buffer;
  while (p != pend) {
	do {
	  if (p == pend)
		return REG_EEND;
	  c = (unsigned char) *p++;
	  if (translate)
		c = (unsigned char) translate[c];
	} while (0);
	switch (c) {
	case '^':
	  {
		if (p == pattern + 1
			|| syntax & RE_CONTEXT_INDEP_ANCHORS
			|| at_begline_loc_p (pattern, p, syntax))
		  do {
			while ((unsigned long) (b - bufp->buffer + (1)) > bufp->allocated)
			  do {
				unsigned char *old_buffer = bufp->buffer;
				if (bufp->allocated == (1L << 16))
				  return REG_ESIZE;
				bufp->allocated <<= 1;
				if (bufp->allocated > (1L << 16))
				  bufp->allocated = (1L << 16);
				bufp->buffer =
				  (unsigned char *) realloc ((bufp->buffer),
											 (bufp->allocated));
				if (bufp->buffer == NULL)
				  return REG_ESPACE;
				if (old_buffer != bufp->buffer) {
				  int incr = bufp->buffer - old_buffer;
				  (b) += incr;
				  (begalt) += incr;
				  if (fixup_alt_jump)
					(fixup_alt_jump) += incr;
				  if (laststart)
					(laststart) += incr;
				  if (pending_exact)
					(pending_exact) += incr;
				}
			  } while (0);
			*b++ = (unsigned char) (begline);
		  } while (0);
		else
		  goto normal_char;
	  }
	  break;
	case '$':
	  {
		if (p == pend
			|| syntax & RE_CONTEXT_INDEP_ANCHORS
			|| at_endline_loc_p (p, pend, syntax))
		  do {
			while ((unsigned long) (b - bufp->buffer + (1)) > bufp->allocated)
			  do {
				unsigned char *old_buffer = bufp->buffer;
				if (bufp->allocated == (1L << 16))
				  return REG_ESIZE;
				bufp->allocated <<= 1;
				if (bufp->allocated > (1L << 16))
				  bufp->allocated = (1L << 16);
				bufp->buffer =
				  (unsigned char *) realloc ((bufp->buffer),
											 (bufp->allocated));
				if (bufp->buffer == NULL)
				  return REG_ESPACE;
				if (old_buffer != bufp->buffer) {
				  int incr = bufp->buffer - old_buffer;
				  (b) += incr;
				  (begalt) += incr;
				  if (fixup_alt_jump)
					(fixup_alt_jump) += incr;
				  if (laststart)
					(laststart) += incr;
				  if (pending_exact)
					(pending_exact) += incr;
				}
			  } while (0);
			*b++ = (unsigned char) (endline);
		  } while (0);
		else
		  goto normal_char;
	  }
	  break;
	case '+':
	case '?':
	  if ((syntax & RE_BK_PLUS_QM)
		  || (syntax & RE_LIMITED_OPS))
		goto normal_char;
	handle_plus:
	case '*':
	  if (!laststart) {
		if (syntax & RE_CONTEXT_INVALID_OPS)
		  return (free (compile_stack.stack), REG_BADRPT);
		else if (!(syntax & RE_CONTEXT_INDEP_OPS))
		  goto normal_char;
	  }
	  {
		boolean keep_string_p = 0;
		char zero_times_ok = 0, many_times_ok = 0;
		for (;;) {
		  zero_times_ok |= c != '+';
		  many_times_ok |= c != '?';
		  if (p == pend)
			break;
		  do {
			if (p == pend)
			  return REG_EEND;
			c = (unsigned char) *p++;
			if (translate)
			  c = (unsigned char) translate[c];
		  } while (0);
		  if (c == '*'
			  || (!(syntax & RE_BK_PLUS_QM) && (c == '+' || c == '?')));
		  else if (syntax & RE_BK_PLUS_QM && c == '\\') {
			if (p == pend)
			  return (free (compile_stack.stack), REG_EESCAPE);
			do {
			  if (p == pend)
				return REG_EEND;
			  c1 = (unsigned char) *p++;
			  if (translate)
				c1 = (unsigned char) translate[c1];
			} while (0);
			if (!(c1 == '+' || c1 == '?')) {
			  p--;
			  p--;
			  break;
			}
			c = c1;
		  } else {
			p--;
			break;
		  }
		}
		if (!laststart)
		  break;
		if (many_times_ok) {
		  while ((unsigned long) (b - bufp->buffer + (1 + 2)) >
				 bufp->allocated)
			do {
			  unsigned char *old_buffer = bufp->buffer;
			  if (bufp->allocated == (1L << 16))
				return REG_ESIZE;
			  bufp->allocated <<= 1;
			  if (bufp->allocated > (1L << 16))
				bufp->allocated = (1L << 16);
			  bufp->buffer =
				(unsigned char *) realloc ((bufp->buffer), (bufp->allocated));
			  if (bufp->buffer == NULL)
				return REG_ESPACE;
			  if (old_buffer != bufp->buffer) {
				int incr = bufp->buffer - old_buffer;
				(b) += incr;
				(begalt) += incr;
				if (fixup_alt_jump)
				  (fixup_alt_jump) += incr;
				if (laststart)
				  (laststart) += incr;
				if (pending_exact)
				  (pending_exact) += incr;
			  }
			} while (0);
		  if ((translate ? (char) translate[(unsigned char) (*(p - 2))]
			   : (*(p - 2))) ==
			  (translate ? (char) translate[(unsigned char) ('.')] : ('.'))
			  && zero_times_ok && p < pend
			  && (translate ? (char) translate[(unsigned char) (*p)] : (*p))
			  ==
			  (translate ? (char) translate[(unsigned char) ('\n')] : ('\n'))
			  && !(syntax & RE_DOT_NEWLINE)) {
			store_op1 (jump, b, (int) ((laststart) - (b) - (1 + 2)));
			keep_string_p = 1;
		  } else
			store_op1 (maybe_pop_jump, b,
							(int) ((laststart - (1 + 2)) - (b) - (1 + 2)));
		  b += 1 + 2;
		}
		while ((unsigned long) (b - bufp->buffer + (1 + 2)) > bufp->allocated)
		  do {
			unsigned char *old_buffer = bufp->buffer;
			if (bufp->allocated == (1L << 16))
			  return REG_ESIZE;
			bufp->allocated <<= 1;
			if (bufp->allocated > (1L << 16))
			  bufp->allocated = (1L << 16);
			bufp->buffer =
			  (unsigned char *) realloc ((bufp->buffer), (bufp->allocated));
			if (bufp->buffer == NULL)
			  return REG_ESPACE;
			if (old_buffer != bufp->buffer) {
			  int incr = bufp->buffer - old_buffer;
			  (b) += incr;
			  (begalt) += incr;
			  if (fixup_alt_jump)
				(fixup_alt_jump) += incr;
			  if (laststart)
				(laststart) += incr;
			  if (pending_exact)
				(pending_exact) += incr;
			}
		  } while (0);
		insert_op1 (keep_string_p ? on_failure_keep_string_jump :
						 on_failure_jump, laststart,
						 (int) ((b + 1 + 2) - (laststart) - (1 + 2)), b);
		pending_exact = 0;
		b += 1 + 2;
		if (!zero_times_ok) {
		  while ((unsigned long) (b - bufp->buffer + (1 + 2)) >
				 bufp->allocated)
			do {
			  unsigned char *old_buffer = bufp->buffer;
			  if (bufp->allocated == (1L << 16))
				return REG_ESIZE;
			  bufp->allocated <<= 1;
			  if (bufp->allocated > (1L << 16))
				bufp->allocated = (1L << 16);
			  bufp->buffer =
				(unsigned char *) realloc ((bufp->buffer), (bufp->allocated));
			  if (bufp->buffer == NULL)
				return REG_ESPACE;
			  if (old_buffer != bufp->buffer) {
				int incr = bufp->buffer - old_buffer;
				(b) += incr;
				(begalt) += incr;
				if (fixup_alt_jump)
				  (fixup_alt_jump) += incr;
				if (laststart)
				  (laststart) += incr;
				if (pending_exact)
				  (pending_exact) += incr;
			  }
			} while (0);
		  insert_op1 (dummy_failure_jump, laststart,
						   (int) ((laststart + 2 + 2 * 2) - (laststart) -
								  (1 + 2)), b);
		  b += 1 + 2;
		}
	  }
	  break;
	case '.':
	  laststart = b;
	  do {
		while ((unsigned long) (b - bufp->buffer + (1)) > bufp->allocated)
		  do {
			unsigned char *old_buffer = bufp->buffer;
			if (bufp->allocated == (1L << 16))
			  return REG_ESIZE;
			bufp->allocated <<= 1;
			if (bufp->allocated > (1L << 16))
			  bufp->allocated = (1L << 16);
			bufp->buffer =
			  (unsigned char *) realloc ((bufp->buffer), (bufp->allocated));
			if (bufp->buffer == NULL)
			  return REG_ESPACE;
			if (old_buffer != bufp->buffer) {
			  int incr = bufp->buffer - old_buffer;
			  (b) += incr;
			  (begalt) += incr;
			  if (fixup_alt_jump)
				(fixup_alt_jump) += incr;
			  if (laststart)
				(laststart) += incr;
			  if (pending_exact)
				(pending_exact) += incr;
			}
		  } while (0);
		*b++ = (unsigned char) (anychar);
	  } while (0);
	  break;
	case '[':
	  {
		boolean had_char_class = 0;
		unsigned int range_start = 0xffffffff;
		if (p == pend)
		  return (free (compile_stack.stack), REG_EBRACK);
		while ((unsigned long) (b - bufp->buffer + (34)) > bufp->allocated)
		  do {
			unsigned char *old_buffer = bufp->buffer;
			if (bufp->allocated == (1L << 16))
			  return REG_ESIZE;
			bufp->allocated <<= 1;
			if (bufp->allocated > (1L << 16))
			  bufp->allocated = (1L << 16);
			bufp->buffer =
			  (unsigned char *) realloc ((bufp->buffer), (bufp->allocated));
			if (bufp->buffer == NULL)
			  return REG_ESPACE;
			if (old_buffer != bufp->buffer) {
			  int incr = bufp->buffer - old_buffer;
			  (b) += incr;
			  (begalt) += incr;
			  if (fixup_alt_jump)
				(fixup_alt_jump) += incr;
			  if (laststart)
				(laststart) += incr;
			  if (pending_exact)
				(pending_exact) += incr;
			}
		  } while (0);
		laststart = b;
		do {
		  while ((unsigned long) (b - bufp->buffer + (1)) > bufp->allocated)
			do {
			  unsigned char *old_buffer = bufp->buffer;
			  if (bufp->allocated == (1L << 16))
				return REG_ESIZE;
			  bufp->allocated <<= 1;
			  if (bufp->allocated > (1L << 16))
				bufp->allocated = (1L << 16);
			  bufp->buffer =
				(unsigned char *) realloc ((bufp->buffer), (bufp->allocated));
			  if (bufp->buffer == NULL)
				return REG_ESPACE;
			  if (old_buffer != bufp->buffer) {
				int incr = bufp->buffer - old_buffer;
				(b) += incr;
				(begalt) += incr;
				if (fixup_alt_jump)
				  (fixup_alt_jump) += incr;
				if (laststart)
				  (laststart) += incr;
				if (pending_exact)
				  (pending_exact) += incr;
			  }
			} while (0);
		  *b++ = (unsigned char) (*p == '^' ? charset_not : charset);
		} while (0);
		if (*p == '^')
		  p++;
		p1 = p;
		do {
		  while ((unsigned long) (b - bufp->buffer + (1)) > bufp->allocated)
			do {
			  unsigned char *old_buffer = bufp->buffer;
			  if (bufp->allocated == (1L << 16))
				return REG_ESIZE;
			  bufp->allocated <<= 1;
			  if (bufp->allocated > (1L << 16))
				bufp->allocated = (1L << 16);
			  bufp->buffer =
				(unsigned char *) realloc ((bufp->buffer), (bufp->allocated));
			  if (bufp->buffer == NULL)
				return REG_ESPACE;
			  if (old_buffer != bufp->buffer) {
				int incr = bufp->buffer - old_buffer;
				(b) += incr;
				(begalt) += incr;
				if (fixup_alt_jump)
				  (fixup_alt_jump) += incr;
				if (laststart)
				  (laststart) += incr;
				if (pending_exact)
				  (pending_exact) += incr;
			  }
			} while (0);
		  *b++ = (unsigned char) ((1 << 8) / 8);
		} while (0);
		memset (b, 0, (1 << 8) / 8);
		if ((re_opcode_t) b[-2] == charset_not
			&& (syntax & RE_HAT_LISTS_NOT_NEWLINE))
		  (b[((unsigned char) ('\n')) / 8] |=
		   1 << (((unsigned char) '\n') % 8));
		for (;;) {
		  if (p == pend)
			return (free (compile_stack.stack), REG_EBRACK);
		  do {
			if (p == pend)
			  return REG_EEND;
			c = (unsigned char) *p++;
			if (translate)
			  c = (unsigned char) translate[c];
		  } while (0);
		  if ((syntax & RE_BACKSLASH_ESCAPE_IN_LISTS) && c == '\\') {
			if (p == pend)
			  return (free (compile_stack.stack), REG_EESCAPE);
			do {
			  if (p == pend)
				return REG_EEND;
			  c1 = (unsigned char) *p++;
			  if (translate)
				c1 = (unsigned char) translate[c1];
			} while (0);
			(b[((unsigned char) (c1)) / 8] |=
			 1 << (((unsigned char) c1) % 8));
			range_start = c1;
			continue;
		  }
		  if (c == ']' && p != p1 + 1)
			break;
		  if (had_char_class && c == '-' && *p != ']')
			return (free (compile_stack.stack), REG_ERANGE);
		  if (c == '-' && !(p - 2 >= pattern && p[-2] == '[')
			  && !(p - 3 >= pattern && p[-3] == '[' && p[-2] == '^')
			  && *p != ']') {
			reg_errcode_t ret
			  = compile_range (range_start, &p, pend, translate,
									syntax, b);
			if (ret != REG_NOERROR)
			  return (free (compile_stack.stack), ret);
			range_start = 0xffffffff;
		  }
		  else if (p[0] == '-' && p[1] != ']') {
			reg_errcode_t ret;
			do {
			  if (p == pend)
				return REG_EEND;
			  c1 = (unsigned char) *p++;
			  if (translate)
				c1 = (unsigned char) translate[c1];
			} while (0);
			ret = compile_range (c, &p, pend, translate, syntax, b);
			if (ret != REG_NOERROR)
			  return (free (compile_stack.stack), ret);
			range_start = 0xffffffff;
		  }
		  else if (syntax & RE_CHAR_CLASSES && c == '[' && *p == ':') {
			char str[6 + 1];
			do {
			  if (p == pend)
				return REG_EEND;
			  c = (unsigned char) *p++;
			  if (translate)
				c = (unsigned char) translate[c];
			} while (0);
			c1 = 0;
			if (p == pend)
			  return (free (compile_stack.stack), REG_EBRACK);
			for (;;) {
			  do {
				if (p == pend)
				  return REG_EEND;
				c = (unsigned char) *p++;
				if (translate)
				  c = (unsigned char) translate[c];
			  } while (0);
			  if ((c == ':' && *p == ']') || p == pend)
				break;
			  if (c1 < 6)
				str[c1++] = c;
			  else
				str[0] = '\0';
			}
			str[c1] = '\0';
			if (c == ':' && *p == ']') {
			  int ch;
			  boolean is_alnum = ((strcmp (str, "alnum") == 0));
			  boolean is_alpha = ((strcmp (str, "alpha") == 0));
			  boolean is_blank = ((strcmp (str, "blank") == 0));
			  boolean is_cntrl = ((strcmp (str, "cntrl") == 0));
			  boolean is_digit = ((strcmp (str, "digit") == 0));
			  boolean is_graph = ((strcmp (str, "graph") == 0));
			  boolean is_lower = ((strcmp (str, "lower") == 0));
			  boolean is_print = ((strcmp (str, "print") == 0));
			  boolean is_punct = ((strcmp (str, "punct") == 0));
			  boolean is_space = ((strcmp (str, "space") == 0));
			  boolean is_upper = ((strcmp (str, "upper") == 0));
			  boolean is_xdigit = ((strcmp (str, "xdigit") == 0));
			  if (!
				  (((strcmp (str, "alpha") == 0))
				   || ((strcmp (str, "upper") == 0))
				   || ((strcmp (str, "lower") == 0))
				   || ((strcmp (str, "digit") == 0))
				   || ((strcmp (str, "alnum") == 0))
				   || ((strcmp (str, "xdigit") == 0))
				   || ((strcmp (str, "space") == 0))
				   || ((strcmp (str, "print") == 0))
				   || ((strcmp (str, "punct") == 0))
				   || ((strcmp (str, "graph") == 0))
				   || ((strcmp (str, "cntrl") == 0))
				   || ((strcmp (str, "blank") == 0))))
				return (free (compile_stack.stack), REG_ECTYPE);
			  do {
				if (p == pend)
				  return REG_EEND;
				c = (unsigned char) *p++;
				if (translate)
				  c = (unsigned char) translate[c];
			  } while (0);
			  if (p == pend)
				return (free (compile_stack.stack), REG_EBRACK);
			  for (ch = 0; ch < 1 << 8; ch++) {
				if ((is_alnum && (1 && isalnum (ch)))
					|| (is_alpha && (1 && isalpha (ch)))
					|| (is_blank && ((ch) == ' ' || (ch) == '\t'))
					|| (is_cntrl && (1 && iscntrl (ch))))
				  (b[((unsigned char) (ch)) / 8] |=
				   1 << (((unsigned char) ch) % 8));
				if ((is_digit && (1 && isdigit (ch)))
					|| (is_graph && (1 && isprint (ch) && !isspace (ch)))
					|| (is_lower && (1 && islower (ch)))
					|| (is_print && (1 && isprint (ch))))
				  (b[((unsigned char) (ch)) / 8] |=
				   1 << (((unsigned char) ch) % 8));
				if ((is_punct && (1 && ispunct (ch)))
					|| (is_space && (1 && isspace (ch)))
					|| (is_upper && (1 && isupper (ch)))
					|| (is_xdigit && (1 && isxdigit (ch))))
				  (b[((unsigned char) (ch)) / 8] |=
				   1 << (((unsigned char) ch) % 8));
				if (translate && (is_upper || is_lower)
					&& ((1 && isupper (ch)) || (1 && islower (ch))))
				  (b[((unsigned char) (ch)) / 8] |=
				   1 << (((unsigned char) ch) % 8));
			  }
			  had_char_class = 1;
			} else {
			  c1++;
			  while (c1--)
				p--;
			  (b[((unsigned char) ('[')) / 8] |=
			   1 << (((unsigned char) '[') % 8));
			  (b[((unsigned char) (':')) / 8] |=
			   1 << (((unsigned char) ':') % 8));
			  range_start = ':';
			  had_char_class = 0;
			}
		  } else if (syntax & RE_CHAR_CLASSES && c == '[' && *p == '=') {
			unsigned char str[1 + 1];
			do {
			  if (p == pend)
				return REG_EEND;
			  c = (unsigned char) *p++;
			  if (translate)
				c = (unsigned char) translate[c];
			} while (0);
			c1 = 0;
			if (p == pend)
			  return (free (compile_stack.stack), REG_EBRACK);
			for (;;) {
			  do {
				if (p == pend)
				  return REG_EEND;
				c = (unsigned char) *p++;
				if (translate)
				  c = (unsigned char) translate[c];
			  } while (0);
			  if ((c == '=' && *p == ']') || p == pend)
				break;
			  if (c1 < 1)
				str[c1++] = c;
			  else
				str[0] = '\0';
			}
			str[c1] = '\0';
			if (c == '=' && *p == ']' && str[0] != '\0') {
			  {
				if (c1 != 1)
				  return (free (compile_stack.stack), REG_ECOLLATE);
				do {
				  if (p == pend)
					return REG_EEND;
				  c = (unsigned char) *p++;
				  if (translate)
					c = (unsigned char) translate[c];
				} while (0);
				(b[((unsigned char) (str[0])) / 8] |=
				 1 << (((unsigned char) str[0]) % 8));
			  }
			  had_char_class = 1;
			} else {
			  c1++;
			  while (c1--)
				p--;
			  (b[((unsigned char) ('[')) / 8] |=
			   1 << (((unsigned char) '[') % 8));
			  (b[((unsigned char) ('=')) / 8] |=
			   1 << (((unsigned char) '=') % 8));
			  range_start = '=';
			  had_char_class = 0;
			}
		  } else if (syntax & RE_CHAR_CLASSES && c == '[' && *p == '.') {
			unsigned char str[128];
			do {
			  if (p == pend)
				return REG_EEND;
			  c = (unsigned char) *p++;
			  if (translate)
				c = (unsigned char) translate[c];
			} while (0);
			c1 = 0;
			if (p == pend)
			  return (free (compile_stack.stack), REG_EBRACK);
			for (;;) {
			  do {
				if (p == pend)
				  return REG_EEND;
				c = (unsigned char) *p++;
				if (translate)
				  c = (unsigned char) translate[c];
			  } while (0);
			  if ((c == '.' && *p == ']') || p == pend)
				break;
			  if (c1 < sizeof (str))
				str[c1++] = c;
			  else
				str[0] = '\0';
			}
			str[c1] = '\0';
			if (c == '.' && *p == ']' && str[0] != '\0') {
			  {
				if (c1 != 1)
				  return (free (compile_stack.stack), REG_ECOLLATE);
				do {
				  if (p == pend)
					return REG_EEND;
				  c = (unsigned char) *p++;
				  if (translate)
					c = (unsigned char) translate[c];
				} while (0);
				(b[((unsigned char) (str[0])) / 8] |=
				 1 << (((unsigned char) str[0]) % 8));
				range_start = ((const unsigned char *) str)[0];
			  }
			  had_char_class = 0;
			} else {
			  c1++;
			  while (c1--)
				p--;
			  (b[((unsigned char) ('[')) / 8] |=
			   1 << (((unsigned char) '[') % 8));
			  (b[((unsigned char) ('.')) / 8] |=
			   1 << (((unsigned char) '.') % 8));
			  range_start = '.';
			  had_char_class = 0;
			}
		  } else {
			had_char_class = 0;
			(b[((unsigned char) (c)) / 8] |= 1 << (((unsigned char) c) % 8));
			range_start = c;
		  }
		}
		while ((int) b[-1] > 0 && b[b[-1] - 1] == 0)
		  b[-1]--;
		b += b[-1];
	  }
	  break;
	case '(':
	  if (syntax & RE_NO_BK_PARENS)
		goto handle_open;
	  else
		goto normal_char;
	case ')':
	  if (syntax & RE_NO_BK_PARENS)
		goto handle_close;
	  else
		goto normal_char;
	case '\n':
	  if (syntax & RE_NEWLINE_ALT)
		goto handle_alt;
	  else
		goto normal_char;
	case '|':
	  if (syntax & RE_NO_BK_VBAR)
		goto handle_alt;
	  else
		goto normal_char;
	case '{':
	  if (syntax & RE_INTERVALS && syntax & RE_NO_BK_BRACES)
		goto handle_interval;
	  else
		goto normal_char;
	case '\\':
	  if (p == pend)
		return (free (compile_stack.stack), REG_EESCAPE);
	  do {
		if (p == pend)
		  return REG_EEND;
		c = (unsigned char) *p++;
	  } while (0);
	  switch (c) {
	  case '(':
		if (syntax & RE_NO_BK_PARENS)
		  goto normal_backslash;
	  handle_open:
		bufp->re_nsub++;
		regnum++;
		if (compile_stack.avail == compile_stack.size) {
		  compile_stack.stack =
		   (compile_stack_elt_t *) realloc (compile_stack.stack,
							(compile_stack.size << 1) *
							sizeof (compile_stack_elt_t));
		  if (compile_stack.stack == NULL)
			return REG_ESPACE;
		  compile_stack.size <<= 1;
		}
		(compile_stack.stack[compile_stack.avail]).begalt_offset =
		  begalt - bufp->buffer;
		(compile_stack.stack[compile_stack.avail]).fixup_alt_jump =
		  fixup_alt_jump ? fixup_alt_jump - bufp->buffer + 1 : 0;
		(compile_stack.stack[compile_stack.avail]).laststart_offset =
		  b - bufp->buffer;
		(compile_stack.stack[compile_stack.avail]).regnum = regnum;
		if (regnum <= 255) {
		  (compile_stack.stack[compile_stack.avail]).inner_group_offset = b
			- bufp->buffer + 2;
		  do {
			while ((unsigned long) (b - bufp->buffer + (3)) > bufp->allocated)
			  do {
				unsigned char *old_buffer = bufp->buffer;
				if (bufp->allocated == (1L << 16))
				  return REG_ESIZE;
				bufp->allocated <<= 1;
				if (bufp->allocated > (1L << 16))
				  bufp->allocated = (1L << 16);
				bufp->buffer =
				  (unsigned char *) realloc ((bufp->buffer),
											 (bufp->allocated));
				if (bufp->buffer == NULL)
				  return REG_ESPACE;
				if (old_buffer != bufp->buffer) {
				  int incr = bufp->buffer - old_buffer;
				  (b) += incr;
				  (begalt) += incr;
				  if (fixup_alt_jump)
					(fixup_alt_jump) += incr;
				  if (laststart)
					(laststart) += incr;
				  if (pending_exact)
					(pending_exact) += incr;
				}
			  } while (0);
			*b++ = (unsigned char) (start_memory);
			*b++ = (unsigned char) (regnum);
			*b++ = (unsigned char) (0);
		  } while (0);
		}
		compile_stack.avail++;
		fixup_alt_jump = 0;
		laststart = 0;
		begalt = b;
		pending_exact = 0;
		break;
	  case ')':
		if (syntax & RE_NO_BK_PARENS)
		  goto normal_backslash;
		if (compile_stack.avail == 0) {
		  if (syntax & RE_UNMATCHED_RIGHT_PAREN_ORD)
			goto normal_backslash;
		  else
			return (free (compile_stack.stack), REG_ERPAREN);
		}
	  handle_close:
		if (fixup_alt_jump) {
		  do {
			while ((unsigned long) (b - bufp->buffer + (1)) > bufp->allocated)
			  do {
				unsigned char *old_buffer = bufp->buffer;
				if (bufp->allocated == (1L << 16))
				  return REG_ESIZE;
				bufp->allocated <<= 1;
				if (bufp->allocated > (1L << 16))
				  bufp->allocated = (1L << 16);
				bufp->buffer =
				  (unsigned char *) realloc ((bufp->buffer),
											 (bufp->allocated));
				if (bufp->buffer == NULL)
				  return REG_ESPACE;
				if (old_buffer != bufp->buffer) {
				  int incr = bufp->buffer - old_buffer;
				  (b) += incr;
				  (begalt) += incr;
				  if (fixup_alt_jump)
					(fixup_alt_jump) += incr;
				  if (laststart)
					(laststart) += incr;
				  if (pending_exact)
					(pending_exact) += incr;
				}
			  } while (0);
			*b++ = (unsigned char) (push_dummy_failure);
		  } while (0);
		  store_op1 (jump_past_alt, fixup_alt_jump,
						  (int) ((b - 1) - (fixup_alt_jump) - (1 + 2)));
		}
		if (compile_stack.avail == 0) {
		  if (syntax & RE_UNMATCHED_RIGHT_PAREN_ORD)
			goto normal_char;
		  else
			return (free (compile_stack.stack), REG_ERPAREN);
		}

		{
		  regnum_t this_group_regnum;
		  compile_stack.avail--;
		  begalt =
			bufp->buffer +
			(compile_stack.stack[compile_stack.avail]).begalt_offset;
		  fixup_alt_jump =
			(compile_stack.stack[compile_stack.avail]).fixup_alt_jump ? bufp->
			buffer +
			(compile_stack.stack[compile_stack.avail]).fixup_alt_jump - 1 : 0;
		  laststart =
			bufp->buffer +
			(compile_stack.stack[compile_stack.avail]).laststart_offset;
		  this_group_regnum =
			(compile_stack.stack[compile_stack.avail]).regnum;
		  pending_exact = 0;
		  if (this_group_regnum <= 255) {
			unsigned char *inner_group_loc
			  =
			  bufp->buffer +
			  (compile_stack.stack[compile_stack.avail]).inner_group_offset;
			*inner_group_loc = regnum - this_group_regnum;
			do {
			  while ((unsigned long) (b - bufp->buffer + (3)) >
					 bufp->allocated)
				do {
				  unsigned char *old_buffer = bufp->buffer;
				  if (bufp->allocated == (1L << 16))
					return REG_ESIZE;
				  bufp->allocated <<= 1;
				  if (bufp->allocated > (1L << 16))
					bufp->allocated = (1L << 16);
				  bufp->buffer =
					(unsigned char *) realloc ((bufp->buffer),
											   (bufp->allocated));
				  if (bufp->buffer == NULL)
					return REG_ESPACE;
				  if (old_buffer != bufp->buffer) {
					int incr = bufp->buffer - old_buffer;
					(b) += incr;
					(begalt) += incr;
					if (fixup_alt_jump)
					  (fixup_alt_jump) += incr;
					if (laststart)
					  (laststart) += incr;
					if (pending_exact)
					  (pending_exact) += incr;
				  }
				} while (0);
			  *b++ = (unsigned char) (stop_memory);
			  *b++ = (unsigned char) (this_group_regnum);
			  *b++ = (unsigned char) (regnum - this_group_regnum);
			} while (0);
		  }
		}
		break;
	  case '|':
		if (syntax & RE_LIMITED_OPS || syntax & RE_NO_BK_VBAR)
		  goto normal_backslash;
	  handle_alt:
		if (syntax & RE_LIMITED_OPS)
		  goto normal_char;
		while ((unsigned long) (b - bufp->buffer + (1 + 2)) > bufp->allocated)
		  do {
			unsigned char *old_buffer = bufp->buffer;
			if (bufp->allocated == (1L << 16))
			  return REG_ESIZE;
			bufp->allocated <<= 1;
			if (bufp->allocated > (1L << 16))
			  bufp->allocated = (1L << 16);
			bufp->buffer =
			  (unsigned char *) realloc ((bufp->buffer), (bufp->allocated));
			if (bufp->buffer == NULL)
			  return REG_ESPACE;
			if (old_buffer != bufp->buffer) {
			  int incr = bufp->buffer - old_buffer;
			  (b) += incr;
			  (begalt) += incr;
			  if (fixup_alt_jump)
				(fixup_alt_jump) += incr;
			  if (laststart)
				(laststart) += incr;
			  if (pending_exact)
				(pending_exact) += incr;
			}
		  } while (0);
		insert_op1 (on_failure_jump, begalt,
						 (int) ((b + 2 + 2 * 2) - (begalt) - (1 + 2)), b);
		pending_exact = 0;
		b += 1 + 2;
		if (fixup_alt_jump)
		  store_op1 (jump_past_alt, fixup_alt_jump,
						  (int) ((b) - (fixup_alt_jump) - (1 + 2)));
		fixup_alt_jump = b;
		while ((unsigned long) (b - bufp->buffer + (1 + 2)) > bufp->allocated)
		  do {
			unsigned char *old_buffer = bufp->buffer;
			if (bufp->allocated == (1L << 16))
			  return REG_ESIZE;
			bufp->allocated <<= 1;
			if (bufp->allocated > (1L << 16))
			  bufp->allocated = (1L << 16);
			bufp->buffer =
			  (unsigned char *) realloc ((bufp->buffer), (bufp->allocated));
			if (bufp->buffer == NULL)
			  return REG_ESPACE;
			if (old_buffer != bufp->buffer) {
			  int incr = bufp->buffer - old_buffer;
			  (b) += incr;
			  (begalt) += incr;
			  if (fixup_alt_jump)
				(fixup_alt_jump) += incr;
			  if (laststart)
				(laststart) += incr;
			  if (pending_exact)
				(pending_exact) += incr;
			}
		  } while (0);
		b += 1 + 2;
		laststart = 0;
		begalt = b;
		break;
	  case '{':
		if (!(syntax & RE_INTERVALS)
			|| (syntax & RE_NO_BK_BRACES))
		  goto normal_backslash;
	  handle_interval:
		{
		  int lower_bound = -1, upper_bound = -1;
		  const char *beg_interval = p;
		  if (p == pend)
			goto invalid_interval;
		  {
			while (p != pend) {
			  do {
				if (p == pend)
				  return REG_EEND;
				c = (unsigned char) *p++;
				if (translate)
				  c = (unsigned char) translate[c];
			  } while (0);
			  if (c < '0' || c > '9')
				break;
			  if (lower_bound <= RE_DUP_MAX) {
				if (lower_bound < 0)
				  lower_bound = 0;
				lower_bound = lower_bound * 10 + c - '0';
			  }
			}
		  };
		  if (c == ',') {
			{
			  while (p != pend) {
				do {
				  if (p == pend)
					return REG_EEND;
				  c = (unsigned char) *p++;
				  if (translate)
					c = (unsigned char) translate[c];
				} while (0);
				if (c < '0' || c > '9')
				  break;
				if (upper_bound <= RE_DUP_MAX) {
				  if (upper_bound < 0)
					upper_bound = 0;
				  upper_bound = upper_bound * 10 + c - '0';
				}
			  }
			};
			if (upper_bound < 0)
			  upper_bound = RE_DUP_MAX;
		  } else
			upper_bound = lower_bound;
		  if (!(0 <= lower_bound && lower_bound <= upper_bound))
			goto invalid_interval;
		  if (!(syntax & RE_NO_BK_BRACES)) {
			if (c != '\\' || p == pend)
			  goto invalid_interval;
			do {
			  if (p == pend)
				return REG_EEND;
			  c = (unsigned char) *p++;
			  if (translate)
				c = (unsigned char) translate[c];
			} while (0);
		  }
		  if (c != '}')
			goto invalid_interval;
		  if (!laststart) {
			if (syntax & RE_CONTEXT_INVALID_OPS
				&& !(syntax & RE_INVALID_INTERVAL_ORD))
			  return (free (compile_stack.stack), REG_BADRPT);
			else if (syntax & RE_CONTEXT_INDEP_OPS)
			  laststart = b;
			else
			  goto unfetch_interval;
		  }
		  if (RE_DUP_MAX < upper_bound)
			return (free (compile_stack.stack), REG_BADBR);
		  if (upper_bound == 0) {
			while ((unsigned long) (b - bufp->buffer + (1 + 2)) >
				   bufp->allocated)
			  do {
				unsigned char *old_buffer = bufp->buffer;
				if (bufp->allocated == (1L << 16))
				  return REG_ESIZE;
				bufp->allocated <<= 1;
				if (bufp->allocated > (1L << 16))
				  bufp->allocated = (1L << 16);
				bufp->buffer =
				  (unsigned char *) realloc ((bufp->buffer),
											 (bufp->allocated));
				if (bufp->buffer == NULL)
				  return REG_ESPACE;
				if (old_buffer != bufp->buffer) {
				  int incr = bufp->buffer - old_buffer;
				  (b) += incr;
				  (begalt) += incr;
				  if (fixup_alt_jump)
					(fixup_alt_jump) += incr;
				  if (laststart)
					(laststart) += incr;
				  if (pending_exact)
					(pending_exact) += incr;
				}
			  } while (0);
			insert_op1 (jump, laststart,
							 (int) ((b + 1 + 2) - (laststart) - (1 + 2)), b);
			b += 1 + 2;
		  } else {
			unsigned nbytes = 2 + 4 * 2 + (upper_bound > 1) * (2 + 4 * 2);
			while ((unsigned long) (b - bufp->buffer + (nbytes)) >
				   bufp->allocated)
			  do {
				unsigned char *old_buffer = bufp->buffer;
				if (bufp->allocated == (1L << 16))
				  return REG_ESIZE;
				bufp->allocated <<= 1;
				if (bufp->allocated > (1L << 16))
				  bufp->allocated = (1L << 16);
				bufp->buffer =
				  (unsigned char *) realloc ((bufp->buffer),
											 (bufp->allocated));
				if (bufp->buffer == NULL)
				  return REG_ESPACE;
				if (old_buffer != bufp->buffer) {
				  int incr = bufp->buffer - old_buffer;
				  (b) += incr;
				  (begalt) += incr;
				  if (fixup_alt_jump)
					(fixup_alt_jump) += incr;
				  if (laststart)
					(laststart) += incr;
				  if (pending_exact)
					(pending_exact) += incr;
				}
			  } while (0);
			insert_op2 (succeed_n, laststart,
							 (int) ((b + 1 + 2 * 2 +
									 (upper_bound >
									  1) * (1 + 2 * 2)) - (laststart) - (1 +
																		 2)),
							 lower_bound, b);
			b += 1 + 2 * 2;
			insert_op2 (set_number_at, laststart, 1
							 + 2 * 2, lower_bound, b);
			b += 1 + 2 * 2;
			if (upper_bound > 1) {
			  store_op2 (jump_n, b,
							  (int) ((laststart + 2 * 2 + 1) - (b) - (1 + 2)),
							  upper_bound - 1);
			  b += 1 + 2 * 2;
			  insert_op2 (set_number_at, laststart,
							   b - laststart, upper_bound - 1, b);
			  b += 1 + 2 * 2;
			}
		  }
		  pending_exact = 0;
		  break;
		invalid_interval:
		  if (!(syntax & RE_INVALID_INTERVAL_ORD))
			return (free (compile_stack.stack),
					p == pend ? REG_EBRACE : REG_BADBR);
		unfetch_interval:
		  p = beg_interval;
		  c = '{';
		  if (syntax & RE_NO_BK_BRACES)
			goto normal_char;
		  else
			goto normal_backslash;
		}
	  case 'w':
		if (syntax & RE_NO_GNU_OPS)
		  goto normal_char;
		laststart = b;
		do {
		  while ((unsigned long) (b - bufp->buffer + (1)) > bufp->allocated)
			do {
			  unsigned char *old_buffer = bufp->buffer;
			  if (bufp->allocated == (1L << 16))
				return REG_ESIZE;
			  bufp->allocated <<= 1;
			  if (bufp->allocated > (1L << 16))
				bufp->allocated = (1L << 16);
			  bufp->buffer =
				(unsigned char *) realloc ((bufp->buffer), (bufp->allocated));
			  if (bufp->buffer == NULL)
				return REG_ESPACE;
			  if (old_buffer != bufp->buffer) {
				int incr = bufp->buffer - old_buffer;
				(b) += incr;
				(begalt) += incr;
				if (fixup_alt_jump)
				  (fixup_alt_jump) += incr;
				if (laststart)
				  (laststart) += incr;
				if (pending_exact)
				  (pending_exact) += incr;
			  }
			} while (0);
		  *b++ = (unsigned char) (wordchar);
		} while (0);
		break;
	  case 'W':
		if (syntax & RE_NO_GNU_OPS)
		  goto normal_char;
		laststart = b;
		do {
		  while ((unsigned long) (b - bufp->buffer + (1)) > bufp->allocated)
			do {
			  unsigned char *old_buffer = bufp->buffer;
			  if (bufp->allocated == (1L << 16))
				return REG_ESIZE;
			  bufp->allocated <<= 1;
			  if (bufp->allocated > (1L << 16))
				bufp->allocated = (1L << 16);
			  bufp->buffer =
				(unsigned char *) realloc ((bufp->buffer), (bufp->allocated));
			  if (bufp->buffer == NULL)
				return REG_ESPACE;
			  if (old_buffer != bufp->buffer) {
				int incr = bufp->buffer - old_buffer;
				(b) += incr;
				(begalt) += incr;
				if (fixup_alt_jump)
				  (fixup_alt_jump) += incr;
				if (laststart)
				  (laststart) += incr;
				if (pending_exact)
				  (pending_exact) += incr;
			  }
			} while (0);
		  *b++ = (unsigned char) (notwordchar);
		} while (0);
		break;
	  case '<':
		if (syntax & RE_NO_GNU_OPS)
		  goto normal_char;
		do {
		  while ((unsigned long) (b - bufp->buffer + (1)) > bufp->allocated)
			do {
			  unsigned char *old_buffer = bufp->buffer;
			  if (bufp->allocated == (1L << 16))
				return REG_ESIZE;
			  bufp->allocated <<= 1;
			  if (bufp->allocated > (1L << 16))
				bufp->allocated = (1L << 16);
			  bufp->buffer =
				(unsigned char *) realloc ((bufp->buffer), (bufp->allocated));
			  if (bufp->buffer == NULL)
				return REG_ESPACE;
			  if (old_buffer != bufp->buffer) {
				int incr = bufp->buffer - old_buffer;
				(b) += incr;
				(begalt) += incr;
				if (fixup_alt_jump)
				  (fixup_alt_jump) += incr;
				if (laststart)
				  (laststart) += incr;
				if (pending_exact)
				  (pending_exact) += incr;
			  }
			} while (0);
		  *b++ = (unsigned char) (wordbeg);
		} while (0);
		break;
	  case '>':
		if (syntax & RE_NO_GNU_OPS)
		  goto normal_char;
		do {
		  while ((unsigned long) (b - bufp->buffer + (1)) > bufp->allocated)
			do {
			  unsigned char *old_buffer = bufp->buffer;
			  if (bufp->allocated == (1L << 16))
				return REG_ESIZE;
			  bufp->allocated <<= 1;
			  if (bufp->allocated > (1L << 16))
				bufp->allocated = (1L << 16);
			  bufp->buffer =
				(unsigned char *) realloc ((bufp->buffer), (bufp->allocated));
			  if (bufp->buffer == NULL)
				return REG_ESPACE;
			  if (old_buffer != bufp->buffer) {
				int incr = bufp->buffer - old_buffer;
				(b) += incr;
				(begalt) += incr;
				if (fixup_alt_jump)
				  (fixup_alt_jump) += incr;
				if (laststart)
				  (laststart) += incr;
				if (pending_exact)
				  (pending_exact) += incr;
			  }
			} while (0);
		  *b++ = (unsigned char) (wordend);
		} while (0);
		break;
	  case 'b':
		if (syntax & RE_NO_GNU_OPS)
		  goto normal_char;
		do {
		  while ((unsigned long) (b - bufp->buffer + (1)) > bufp->allocated)
			do {
			  unsigned char *old_buffer = bufp->buffer;
			  if (bufp->allocated == (1L << 16))
				return REG_ESIZE;
			  bufp->allocated <<= 1;
			  if (bufp->allocated > (1L << 16))
				bufp->allocated = (1L << 16);
			  bufp->buffer =
				(unsigned char *) realloc ((bufp->buffer), (bufp->allocated));
			  if (bufp->buffer == NULL)
				return REG_ESPACE;
			  if (old_buffer != bufp->buffer) {
				int incr = bufp->buffer - old_buffer;
				(b) += incr;
				(begalt) += incr;
				if (fixup_alt_jump)
				  (fixup_alt_jump) += incr;
				if (laststart)
				  (laststart) += incr;
				if (pending_exact)
				  (pending_exact) += incr;
			  }
			} while (0);
		  *b++ = (unsigned char) (wordbound);
		} while (0);
		break;
	  case 'B':
		if (syntax & RE_NO_GNU_OPS)
		  goto normal_char;
		do {
		  while ((unsigned long) (b - bufp->buffer + (1)) > bufp->allocated)
			do {
			  unsigned char *old_buffer = bufp->buffer;
			  if (bufp->allocated == (1L << 16))
				return REG_ESIZE;
			  bufp->allocated <<= 1;
			  if (bufp->allocated > (1L << 16))
				bufp->allocated = (1L << 16);
			  bufp->buffer =
				(unsigned char *) realloc ((bufp->buffer), (bufp->allocated));
			  if (bufp->buffer == NULL)
				return REG_ESPACE;
			  if (old_buffer != bufp->buffer) {
				int incr = bufp->buffer - old_buffer;
				(b) += incr;
				(begalt) += incr;
				if (fixup_alt_jump)
				  (fixup_alt_jump) += incr;
				if (laststart)
				  (laststart) += incr;
				if (pending_exact)
				  (pending_exact) += incr;
			  }
			} while (0);
		  *b++ = (unsigned char) (notwordbound);
		} while (0);
		break;
	  case '`':
		if (syntax & RE_NO_GNU_OPS)
		  goto normal_char;
		do {
		  while ((unsigned long) (b - bufp->buffer + (1)) > bufp->allocated)
			do {
			  unsigned char *old_buffer = bufp->buffer;
			  if (bufp->allocated == (1L << 16))
				return REG_ESIZE;
			  bufp->allocated <<= 1;
			  if (bufp->allocated > (1L << 16))
				bufp->allocated = (1L << 16);
			  bufp->buffer =
				(unsigned char *) realloc ((bufp->buffer), (bufp->allocated));
			  if (bufp->buffer == NULL)
				return REG_ESPACE;
			  if (old_buffer != bufp->buffer) {
				int incr = bufp->buffer - old_buffer;
				(b) += incr;
				(begalt) += incr;
				if (fixup_alt_jump)
				  (fixup_alt_jump) += incr;
				if (laststart)
				  (laststart) += incr;
				if (pending_exact)
				  (pending_exact) += incr;
			  }
			} while (0);
		  *b++ = (unsigned char) (begbuf);
		} while (0);
		break;
	  case '\'':
		if (syntax & RE_NO_GNU_OPS)
		  goto normal_char;
		do {
		  while ((unsigned long) (b - bufp->buffer + (1)) > bufp->allocated)
			do {
			  unsigned char *old_buffer = bufp->buffer;
			  if (bufp->allocated == (1L << 16))
				return REG_ESIZE;
			  bufp->allocated <<= 1;
			  if (bufp->allocated > (1L << 16))
				bufp->allocated = (1L << 16);
			  bufp->buffer =
				(unsigned char *) realloc ((bufp->buffer), (bufp->allocated));
			  if (bufp->buffer == NULL)
				return REG_ESPACE;
			  if (old_buffer != bufp->buffer) {
				int incr = bufp->buffer - old_buffer;
				(b) += incr;
				(begalt) += incr;
				if (fixup_alt_jump)
				  (fixup_alt_jump) += incr;
				if (laststart)
				  (laststart) += incr;
				if (pending_exact)
				  (pending_exact) += incr;
			  }
			} while (0);
		  *b++ = (unsigned char) (endbuf);
		} while (0);
		break;
	  case '1':
	  case '2':
	  case '3':
	  case '4':
	  case '5':
	  case '6':
	  case '7':
	  case '8':
	  case '9':
		if (syntax & RE_NO_BK_REFS)
		  goto normal_char;
		c1 = c - '0';
		if (c1 > regnum)
		  return (free (compile_stack.stack), REG_ESUBREG);
		if (group_in_compile_stack (compile_stack, (regnum_t) c1))
		  goto normal_char;
		laststart = b;
		do {
		  while ((unsigned long) (b - bufp->buffer + (2)) > bufp->allocated)
			do {
			  unsigned char *old_buffer = bufp->buffer;
			  if (bufp->allocated == (1L << 16))
				return REG_ESIZE;
			  bufp->allocated <<= 1;
			  if (bufp->allocated > (1L << 16))
				bufp->allocated = (1L << 16);
			  bufp->buffer =
				(unsigned char *) realloc ((bufp->buffer), (bufp->allocated));
			  if (bufp->buffer == NULL)
				return REG_ESPACE;
			  if (old_buffer != bufp->buffer) {
				int incr = bufp->buffer - old_buffer;
				(b) += incr;
				(begalt) += incr;
				if (fixup_alt_jump)
				  (fixup_alt_jump) += incr;
				if (laststart)
				  (laststart) += incr;
				if (pending_exact)
				  (pending_exact) += incr;
			  }
			} while (0);
		  *b++ = (unsigned char) (duplicate);
		  *b++ = (unsigned char) (c1);
		} while (0);
		break;
	  case '+':
	  case '?':
		if (syntax & RE_BK_PLUS_QM)
		  goto handle_plus;
		else
		  goto normal_backslash;
	  default:
	  normal_backslash:
		c = (translate ? (char) translate[(unsigned char) (c)] : (c));
		goto normal_char;
	  }
	  break;
	default:
	normal_char:
	  if (!pending_exact
		  || pending_exact + *pending_exact + 1 != b
		  || *pending_exact == (1 << 8) - 1
		  || *p == '*' || *p == '^' || ((syntax & RE_BK_PLUS_QM)
										? *p == '\\' && (p[1] == '+'
														 || p[1] == '?')
										: (*p == '+' || *p == '?'))
		  || ((syntax & RE_INTERVALS)
			  && ((syntax & RE_NO_BK_BRACES)
				  ? *p == '{' : (p[0] == '\\' && p[1] == '{')))) {
		laststart = b;
		do {
		  while ((unsigned long) (b - bufp->buffer + (2)) > bufp->allocated)
			do {
			  unsigned char *old_buffer = bufp->buffer;
			  if (bufp->allocated == (1L << 16))
				return REG_ESIZE;
			  bufp->allocated <<= 1;
			  if (bufp->allocated > (1L << 16))
				bufp->allocated = (1L << 16);
			  bufp->buffer =
				(unsigned char *) realloc ((bufp->buffer), (bufp->allocated));
			  if (bufp->buffer == NULL)
				return REG_ESPACE;
			  if (old_buffer != bufp->buffer) {
				int incr = bufp->buffer - old_buffer;
				(b) += incr;
				(begalt) += incr;
				if (fixup_alt_jump)
				  (fixup_alt_jump) += incr;
				if (laststart)
				  (laststart) += incr;
				if (pending_exact)
				  (pending_exact) += incr;
			  }
			} while (0);
		  *b++ = (unsigned char) (exactn);
		  *b++ = (unsigned char) (0);
		} while (0);
		pending_exact = b - 1;
	  }
	  do {
		while ((unsigned long) (b - bufp->buffer + (1)) > bufp->allocated)
		  do {
			unsigned char *old_buffer = bufp->buffer;
			if (bufp->allocated == (1L << 16))
			  return REG_ESIZE;
			bufp->allocated <<= 1;
			if (bufp->allocated > (1L << 16))
			  bufp->allocated = (1L << 16);
			bufp->buffer =
			  (unsigned char *) realloc ((bufp->buffer), (bufp->allocated));
			if (bufp->buffer == NULL)
			  return REG_ESPACE;
			if (old_buffer != bufp->buffer) {
			  int incr = bufp->buffer - old_buffer;
			  (b) += incr;
			  (begalt) += incr;
			  if (fixup_alt_jump)
				(fixup_alt_jump) += incr;
			  if (laststart)
				(laststart) += incr;
			  if (pending_exact)
				(pending_exact) += incr;
			}
		  } while (0);
		*b++ = (unsigned char) (c);
	  } while (0);
	  (*pending_exact)++;
	  break;
	}
  }
  if (fixup_alt_jump)
	store_op1 (jump_past_alt, fixup_alt_jump,
					(int) ((b) - (fixup_alt_jump) - (1 + 2)));
  if (!(compile_stack.avail == 0))
	return (free (compile_stack.stack), REG_EPAREN);
  if (syntax & RE_NO_POSIX_BACKTRACKING)
	do {
	  while ((unsigned long) (b - bufp->buffer + (1)) > bufp->allocated)
		do {
		  unsigned char *old_buffer = bufp->buffer;
		  if (bufp->allocated == (1L << 16))
			return REG_ESIZE;
		  bufp->allocated <<= 1;
		  if (bufp->allocated > (1L << 16))
			bufp->allocated = (1L << 16);
		  bufp->buffer =
			(unsigned char *) realloc ((bufp->buffer), (bufp->allocated));
		  if (bufp->buffer == NULL)
			return REG_ESPACE;
		  if (old_buffer != bufp->buffer) {
			int incr = bufp->buffer - old_buffer;
			(b) += incr;
			(begalt) += incr;
			if (fixup_alt_jump)
			  (fixup_alt_jump) += incr;
			if (laststart)
			  (laststart) += incr;
			if (pending_exact)
			  (pending_exact) += incr;
		  }
		} while (0);
	  *b++ = (unsigned char) (succeed);
	} while (0);
  free (compile_stack.stack);
  bufp->used = b - bufp->buffer;
  return REG_NOERROR;
}

static void
store_op1 (re_opcode_t op, unsigned char *loc, int arg)
{
  *loc = (unsigned char) op;
  do {
	(loc + 1)[0] = (arg) & 0377;
	(loc + 1)[1] = (arg) >> 8;
  } while (0);
}

static void
store_op2 (re_opcode_t op, unsigned char *loc, int arg1, int arg2)
{
  *loc = (unsigned char) op;
  do {
	(loc + 1)[0] = (arg1) & 0377;
	(loc + 1)[1] = (arg1) >> 8;
  } while (0);
  do {
	(loc + 1 + 2)[0] = (arg2) & 0377;
	(loc + 1 + 2)[1] = (arg2) >> 8;
  } while (0);
}

static void
insert_op1 (re_opcode_t op,
				 unsigned char *loc, int arg, unsigned char *end)
{
  unsigned char *pfrom = end;
  unsigned char *pto = end + 1 + 2;
  while (pfrom != loc)
	*--pto = *--pfrom;
  store_op1 (op, loc, arg);
}

static void
insert_op2 (re_opcode_t op,
				 unsigned char *loc, int arg1, int arg2, unsigned char *end)
{
  unsigned char *pfrom = end;
  unsigned char *pto = end + 1 + 2 * 2;
  while (pfrom != loc)
	*--pto = *--pfrom;
  store_op2 (op, loc, arg1, arg2);
}

static boolean
at_begline_loc_p (const char *pattern, const char *p,
					   reg_syntax_t syntax)
{
  const char *prev = p - 2;
  boolean prev_prev_backslash = prev > pattern && prev[-1] == '\\';
  return (*prev == '(' && (syntax & RE_NO_BK_PARENS || prev_prev_backslash))
	|| (*prev == '|' && (syntax & RE_NO_BK_VBAR || prev_prev_backslash));
}

static boolean
at_endline_loc_p (const char *p, const char *pend, reg_syntax_t syntax)
{
  const char *next = p;
  boolean next_backslash = *next == '\\';
  const char *next_next = p + 1 < pend ? p + 1 : 0;
  return
	(syntax & RE_NO_BK_PARENS ? *next == ')'
	 : next_backslash && next_next && *next_next == ')')
	|| (syntax & RE_NO_BK_VBAR ? *next == '|'
		: next_backslash && next_next && *next_next == '|');
}

static reg_errcode_t
compile_range (unsigned int range_start_char,
					const char **p_ptr, const char *pend,
					char *translate, reg_syntax_t syntax, unsigned char *b)
{
  unsigned this_char;
  const char *p = *p_ptr;
  reg_errcode_t ret;
  unsigned end_char;
  if (p == pend)
	return REG_ERANGE;
  (*p_ptr)++;
  ret = syntax & RE_NO_EMPTY_RANGES ? REG_ERANGE : REG_NOERROR;
  range_start_char =
	(translate ? (char) translate[(unsigned char) (range_start_char)]
	 : (range_start_char));
  end_char =
	((unsigned)
	 (translate ? (char) translate[(unsigned char) (p[0])] : (p[0])) &
	 ((1 << 8) - 1));
  for (this_char = range_start_char; this_char <= end_char; ++this_char) {
	(b
	 [((unsigned
		char) ((translate ? (char) translate[(unsigned char) (this_char)]
				: (this_char)))) / 8] |=
	 1 <<
	 (((unsigned char) (translate ? (char)
						translate[(unsigned char) (this_char)] : (this_char)))
	  % 8));
	ret = REG_NOERROR;
  }
  return ret;
}

int
re_compile_fastmap (struct re_pattern_buffer *bufp)
{
  int j, k;
  fail_stack_type fail_stack;
  char *destination;
  char *fastmap = bufp->fastmap;
  unsigned char *pattern = bufp->buffer;
  unsigned char *pend = pattern + bufp->used;
  unsigned char *p = pattern;
  boolean path_can_be_null = 1;
  boolean succeed_n_p = 0;

  do {
	fail_stack.stack =
	  (fail_stack_elt_t *) alloca (5 * sizeof (fail_stack_elt_t));
	if (fail_stack.stack == NULL)
	  return -2;
	fail_stack.size = 5;
	fail_stack.avail = 0;
  } while (0);
  memset (fastmap, 0, 1 << 8);
  bufp->fastmap_accurate = 1;
  bufp->can_be_null = 0;
  while (1) {
	if (p == pend || *p == succeed) {
	  if (!(fail_stack.avail == 0)) {
		bufp->can_be_null |= path_can_be_null;
		path_can_be_null = 1;
		p = fail_stack.stack[--fail_stack.avail].pointer;
		continue;
	  } else
		break;
	}

	switch (((re_opcode_t) * p++)) {
	case duplicate:
	  bufp->can_be_null = 1;
	  goto done;
	case exactn:
	  fastmap[p[1]] = 1;
	  break;
	case charset:
	  for (j = *p++ * 8 - 1; j >= 0; j--)
		if (p[j / 8] & (1 << (j % 8)))
		  fastmap[j] = 1;
	  break;
	case charset_not:
	  for (j = *p * 8; j < (1 << 8); j++)
		fastmap[j] = 1;
	  for (j = *p++ * 8 - 1; j >= 0; j--)
		if (!(p[j / 8] & (1 << (j % 8))))
		  fastmap[j] = 1;
	  break;
	case wordchar:
	  for (j = 0; j < (1 << 8); j++)
		if (re_syntax_table[(unsigned char) (j)] == 1)
		  fastmap[j] = 1;
	  break;
	case notwordchar:
	  for (j = 0; j < (1 << 8); j++)
		if (re_syntax_table[(unsigned char) (j)] != 1)
		  fastmap[j] = 1;
	  break;
	case anychar:
	  {
		int fastmap_newline = fastmap['\n'];
		for (j = 0; j < (1 << 8); j++)
		  fastmap[j] = 1;
		if (!(bufp->syntax & RE_DOT_NEWLINE))
		  fastmap['\n'] = fastmap_newline;
		else if (bufp->can_be_null)
		  goto done;
		break;
	  }
	case no_op:
	case begline:
	case endline:
	case begbuf:
	case endbuf:
	case wordbound:
	case notwordbound:
	case wordbeg:
	case wordend:
	case push_dummy_failure:
	  continue;
	case jump_n:
	case pop_failure_jump:
	case maybe_pop_jump:
	case jump:
	case jump_past_alt:
	case dummy_failure_jump:
	  do {
		do {
		  (j) = *(p) & 0377;
		  (j) += ((signed char) (*((p) + 1))) << 8;
		} while (0);
		(p) += 2;
	  } while (0);
	  p += j;
	  if (j > 0)
		continue;
	  if ((re_opcode_t) * p != on_failure_jump
		  && (re_opcode_t) * p != succeed_n)
		continue;
	  p++;
	  do {
		do {
		  (j) = *(p) & 0377;
		  (j) += ((signed char) (*((p) + 1))) << 8;
		} while (0);
		(p) += 2;
	  } while (0);
	  p += j;
	  if (!(fail_stack.avail == 0)
		  && fail_stack.stack[fail_stack.avail - 1].pointer == p)
		fail_stack.avail--;
	  continue;
	case on_failure_jump:
	case on_failure_keep_string_jump:
	handle_on_failure_jump:
	  do {
		do {
		  (j) = *(p) & 0377;
		  (j) += ((signed char) (*((p) + 1))) << 8;
		} while (0);
		(p) += 2;
	  } while (0);
	  if (p + j < pend) {
		if (!
			(((fail_stack.avail == fail_stack.size)
			  && !((fail_stack).size >
				   (unsigned) (re_max_failures *
							   (5 * 3 + 4)) ? 0 : ((fail_stack).stack =
												   (fail_stack_elt_t
													*) (destination =
														(char *)
														alloca (((fail_stack).
																 size << 1) *
																sizeof
																(fail_stack_elt_t)),
														memcpy (destination,
																(fail_stack).
																stack,
																(fail_stack).
																size *
																sizeof
																(fail_stack_elt_t))),
												   (fail_stack).stack ==
												   NULL ? 0 : ((fail_stack).
															   size <<= 1,
															   1)))) ? 0
			 : ((fail_stack).stack[(fail_stack).avail++].pointer =
				p + j, 1))) {
		  return -2;
		}
	  } else
		bufp->can_be_null = 1;
	  if (succeed_n_p) {
		do {
		  do {
			(k) = *(p) & 0377;
			(k) += ((signed char) (*((p) + 1))) << 8;
		  } while (0);
		  (p) += 2;
		} while (0);
		succeed_n_p = 0;
	  }
	  continue;
	case succeed_n:
	  p += 2;
	  do {
		do {
		  (k) = *(p) & 0377;
		  (k) += ((signed char) (*((p) + 1))) << 8;
		} while (0);
		(p) += 2;
	  } while (0);
	  if (k == 0) {
		p -= 2 * 2;
		succeed_n_p = 1;
		goto handle_on_failure_jump;
	  }
	  continue;
	case set_number_at:
	  p += 2 * 2;
	  continue;
	case start_memory:
	case stop_memory:
	  p += 2;
	  continue;
	default:
	  abort ();
	}
	path_can_be_null = 0;
	p = pend;
  }
  bufp->can_be_null |= path_can_be_null;
done:
  return 0;
}

int
re_search_2 (struct re_pattern_buffer *bufp,
				  const char *string1, int size1,
				  const char *string2, int size2,
				  int startpos,
				  int range, struct re_registers *regs, int stop)
{
  int val;
  char *fastmap = bufp->fastmap;
  char *translate = (char *)bufp->translate;
  int total_size = size1 + size2;
  int endpos = startpos + range;
  if (startpos < 0 || startpos > total_size)
	return -1;
  if (endpos < 0)
	range = 0 - startpos;
  else if (endpos > total_size)
	range = total_size - startpos;
  if (bufp->used > 0 && range > 0
	  && ((re_opcode_t) bufp->buffer[0] == begbuf
		  || ((re_opcode_t) bufp->buffer[0] == begline
			  && !bufp->newline_anchor))) {
	if (startpos > 0)
	  return -1;
	else
	  range = 1;
  }
  if (fastmap && !bufp->fastmap_accurate)
	if (re_compile_fastmap (bufp) == -2)
	  return -2;
  for (;;) {
	if (fastmap && startpos < total_size && !bufp->can_be_null) {
	  if (range > 0) {
		const char *d;
		int lim = 0;
		int irange = range;
		if (startpos < size1 && startpos + range >= size1)
		  lim = range - (size1 - startpos);
		d = (startpos >= size1 ? string2 - size1 : string1) + startpos;
		if (translate)
		  while (range > lim && !fastmap[(unsigned char)
										 translate[(unsigned char) *d++]])
			range--;
		else
		  while (range > lim && !fastmap[(unsigned char) *d++])
			range--;
		startpos += irange - range;
	  } else {
		char c = (size1 == 0 || startpos >= size1
						   ? string2[startpos - size1]
						   : string1[startpos]);
		if (!fastmap
			[(unsigned char) (translate ? (char)
							  translate[(unsigned char) (c)] : (c))])
		  goto advance;
	  }
	}
	if (range >= 0 && startpos == total_size && fastmap && !bufp->can_be_null) {
	  return -1;
	}
	val = re_match_2_internal (bufp, string1, size1, string2,
									size2, startpos, regs, stop);
	if (val >= 0) {
	  return startpos;
	}
	if (val == -2) {
	  return -2;
	}
  advance:
	if (!range)
	  break;
	else if (range > 0) {
	  range--;
	  startpos++;
	} else {
	  range++;
	  startpos--;
	}
  }
  return -1;
}

static boolean
group_match_null_string_p (unsigned char **p, unsigned char *end,
				register_info_type *reg_info);
static boolean
alt_match_null_string_p (unsigned char *p, unsigned char *end,
				register_info_type *reg_info);
static boolean
common_op_match_null_string_p (unsigned char **p, unsigned char *end,
					register_info_type *reg_info);
static int
bcmp_translate (const char *s1, const char *s2, int len, char *translate);

static int
re_match_2_internal (struct re_pattern_buffer *bufp,
						  const char *string1, int size1,
						  const char *string2, int size2,
						  int pos, struct re_registers *regs, int stop)
{
  int mcnt;
  unsigned char *p1;
  const char *end1, *end2;
  const char *end_match_1, *end_match_2;
  const char *d, *dend;
  unsigned char *p = bufp->buffer;
  unsigned char *pend = p + bufp->used;
  unsigned char *just_past_start_mem = 0;
  char *translate = (char *)bufp->translate;
  fail_stack_type fail_stack;
  size_t num_regs = bufp->re_nsub + 1;
  active_reg_t lowest_active_reg = ((1 << 8) + 1);
  active_reg_t highest_active_reg = (1 << 8);
  const char **regstart, **regend;
  const char **old_regstart, **old_regend;
  register_info_type *reg_info;
  unsigned best_regs_set = 0;
  const char **best_regstart, **best_regend;
  const char *match_end = NULL;
  int set_regs_matched_done = 0;
  const char **reg_dummy;
  register_info_type *reg_info_dummy;

  do {
	fail_stack.stack =
	  (fail_stack_elt_t *) alloca (5 * sizeof (fail_stack_elt_t));
	if (fail_stack.stack == NULL)
	  return -2;
	fail_stack.size = 5;
	fail_stack.avail = 0;
  } while (0);
  if (bufp->re_nsub) {
	regstart = ((const char **) alloca ((num_regs) * sizeof (const char *)));
	regend = ((const char **) alloca ((num_regs) * sizeof (const char *)));
	old_regstart =
	  ((const char **) alloca ((num_regs) * sizeof (const char *)));
	old_regend =
	  ((const char **) alloca ((num_regs) * sizeof (const char *)));
	best_regstart =
	  ((const char **) alloca ((num_regs) * sizeof (const char *)));
	best_regend =
	  ((const char **) alloca ((num_regs) * sizeof (const char *)));
	reg_info =
	  ((register_info_type *)
	   alloca ((num_regs) * sizeof (register_info_type)));
	reg_dummy = ((const char **) alloca ((num_regs) * sizeof (const char *)));
	reg_info_dummy =
	  ((register_info_type *)
	   alloca ((num_regs) * sizeof (register_info_type)));
	if (!(regstart && regend && old_regstart && old_regend && reg_info
		  && best_regstart && best_regend && reg_dummy && reg_info_dummy)) {
	  do {;
		if (regstart)
		  ((void) 0);
		regstart = NULL;
		if (regend)
		  ((void) 0);
		regend = NULL;
		if (old_regstart)
		  ((void) 0);
		old_regstart = NULL;
		if (old_regend)
		  ((void) 0);
		old_regend = NULL;
		if (best_regstart)
		  ((void) 0);
		best_regstart = NULL;
		if (best_regend)
		  ((void) 0);
		best_regend = NULL;
		if (reg_info)
		  ((void) 0);
		reg_info = NULL;
		if (reg_dummy)
		  ((void) 0);
		reg_dummy = NULL;
		if (reg_info_dummy)
		  ((void) 0);
		reg_info_dummy = NULL;
	  } while (0);
	  return -2;
	}
  } else {
	regstart = regend = old_regstart = old_regend = best_regstart
	  = best_regend = reg_dummy = NULL;
	reg_info = reg_info_dummy = (register_info_type *) NULL;
  }
  if (pos < 0 || pos > size1 + size2)
  {
	do {;
	  if (regstart)
		((void) 0);
	  regstart = NULL;
	  if (regend)
		((void) 0);
	  regend = NULL;
	  if (old_regstart)
		((void) 0);
	  old_regstart = NULL;
	  if (old_regend)
		((void) 0);
	  old_regend = NULL;
	  if (best_regstart)
		((void) 0);
	  best_regstart = NULL;
	  if (best_regend)
		((void) 0);
	  best_regend = NULL;
	  if (reg_info)
		((void) 0);
	  reg_info = NULL;
	  if (reg_dummy)
		((void) 0);
	  reg_dummy = NULL;
	  if (reg_info_dummy)
		((void) 0);
	  reg_info_dummy = NULL;
	} while (0);
	return -1;
  }
  for (mcnt = 1; (unsigned) mcnt < num_regs; mcnt++) {
	regstart[mcnt] = regend[mcnt]
	  = old_regstart[mcnt] = old_regend[mcnt] = (&reg_unset_dummy);
	((reg_info[mcnt]).bits.match_null_string_p) = 3;
	((reg_info[mcnt]).bits.is_active) = 0;
	((reg_info[mcnt]).bits.matched_something) = 0;
	((reg_info[mcnt]).bits.ever_matched_something) = 0;
  }
  if (size2 == 0 && string1 != NULL) {
	string2 = string1;
	size2 = size1;
	string1 = 0;
	size1 = 0;
  }
  end1 = string1 + size1;
  end2 = string2 + size2;
  if (stop <= size1) {
	end_match_1 = string1 + stop;
	end_match_2 = string2;
  } else {
	end_match_1 = end1;
	end_match_2 = string2 + stop - size1;
  }
  if (size1 > 0 && pos <= size1) {
	d = string1 + pos;
	dend = end_match_1;
  } else {
	d = string2 + pos - size1;
	dend = end_match_2;
  }

  for (;;) {
	if (p == pend) {
	  if (d != end_match_2) {
		boolean same_str_p =
		  ((size1 && string1 <= (match_end) && (match_end) <= string1 + size1)
		   == (dend == end_match_1));
		boolean best_match_p;
		if (same_str_p)
		  best_match_p = d > match_end;
		else
		  best_match_p = !(dend == end_match_1);

		if (!(fail_stack.avail == 0)) {
		  if (!best_regs_set || best_match_p) {
			best_regs_set = 1;
			match_end = d;

			for (mcnt = 1; (unsigned) mcnt < num_regs; mcnt++) {
			  best_regstart[mcnt] = regstart[mcnt];
			  best_regend[mcnt] = regend[mcnt];
			}
		  }
		  goto fail;
		}
		else if (best_regs_set && !best_match_p) {
		restore_best_regs:

		  d = match_end;
		  dend = ((d >= string1 && d <= end1)
				  ? end_match_1 : end_match_2);
		  for (mcnt = 1; (unsigned) mcnt < num_regs; mcnt++) {
			regstart[mcnt] = best_regstart[mcnt];
			regend[mcnt] = best_regend[mcnt];
		  }
		}
	  }
	succeed_label:

	  if (regs && !bufp->no_sub) {
		if (bufp->regs_allocated == REGS_UNALLOCATED) {
		  regs->num_regs =
			((RE_NREGS) > (num_regs + 1) ? (RE_NREGS) : (num_regs + 1));
		  regs->start =
			((regoff_t *) malloc ((regs->num_regs) * sizeof (regoff_t)));
		  regs->end =
			((regoff_t *) malloc ((regs->num_regs) * sizeof (regoff_t)));
		  if (regs->start == NULL || regs->end == NULL) {
			do {;
			  if (regstart)
				((void) 0);
			  regstart = NULL;
			  if (regend)
				((void) 0);
			  regend = NULL;
			  if (old_regstart)
				((void) 0);
			  old_regstart = NULL;
			  if (old_regend)
				((void) 0);
			  old_regend = NULL;
			  if (best_regstart)
				((void) 0);
			  best_regstart = NULL;
			  if (best_regend)
				((void) 0);
			  best_regend = NULL;
			  if (reg_info)
				((void) 0);
			  reg_info = NULL;
			  if (reg_dummy)
				((void) 0);
			  reg_dummy = NULL;
			  if (reg_info_dummy)
				((void) 0);
			  reg_info_dummy = NULL;
			} while (0);
			return -2;
		  }
		  bufp->regs_allocated = REGS_REALLOCATE;
		} else if (bufp->regs_allocated == REGS_REALLOCATE) {
		  if (regs->num_regs < num_regs + 1) {
			regs->num_regs = num_regs + 1;
			((regs->start) =
			 (regoff_t *) realloc (regs->start,
								   (regs->num_regs) * sizeof (regoff_t)));
			((regs->end) =
			 (regoff_t *) realloc (regs->end,
								   (regs->num_regs) * sizeof (regoff_t)));
			if (regs->start == NULL || regs->end == NULL) {
			  do {;
				if (regstart)
				  ((void) 0);
				regstart = NULL;
				if (regend)
				  ((void) 0);
				regend = NULL;
				if (old_regstart)
				  ((void) 0);
				old_regstart = NULL;
				if (old_regend)
				  ((void) 0);
				old_regend = NULL;
				if (best_regstart)
				  ((void) 0);
				best_regstart = NULL;
				if (best_regend)
				  ((void) 0);
				best_regend = NULL;
				if (reg_info)
				  ((void) 0);
				reg_info = NULL;
				if (reg_dummy)
				  ((void) 0);
				reg_dummy = NULL;
				if (reg_info_dummy)
				  ((void) 0);
				reg_info_dummy = NULL;
			  } while (0);
			  return -2;
			}
		  }
		} else {
		  ;
		}
		if (regs->num_regs > 0) {
		  regs->start[0] = pos;
		  regs->end[0] = ((dend == end_match_1)
						  ? ((regoff_t) (d - string1))
						  : ((regoff_t) (d - string2 + size1)));
		}
		for (mcnt = 1;
			 (unsigned) mcnt < ((num_regs) <
								(regs->num_regs) ? (num_regs) : (regs->
																 num_regs));
			 mcnt++) {
		  if (((regstart[mcnt]) == (&reg_unset_dummy))
			  || ((regend[mcnt]) == (&reg_unset_dummy)))
			regs->start[mcnt] = regs->end[mcnt] = -1;
		  else {
			regs->start[mcnt]
			  =
			  (regoff_t) ((size1 && string1 <= (regstart[mcnt])
						   && (regstart[mcnt]) <=
						   string1 + size1) ? ((regoff_t) ((regstart[mcnt]) -
														   string1))
						  : ((regoff_t)
							 ((regstart[mcnt]) - string2 + size1)));
			regs->end[mcnt]
			  =
			  (regoff_t) ((size1 && string1 <= (regend[mcnt])
						   && (regend[mcnt]) <=
						   string1 + size1) ? ((regoff_t) ((regend[mcnt]) -
														   string1))
						  : ((regoff_t) ((regend[mcnt]) - string2 + size1)));
		  }
		}
		for (mcnt = num_regs; (unsigned) mcnt < regs->num_regs; mcnt++)
		  regs->start[mcnt] = regs->end[mcnt] = -1;
	  }

	  mcnt = d - pos - ((dend == end_match_1)
						? string1 : string2 - size1);
	  do {
		if (regstart)
		  ((void) 0);
		regstart = NULL;
		if (regend)
		  ((void) 0);
		regend = NULL;
		if (old_regstart)
		  ((void) 0);
		old_regstart = NULL;
		if (old_regend)
		  ((void) 0);
		old_regend = NULL;
		if (best_regstart)
		  ((void) 0);
		best_regstart = NULL;
		if (best_regend)
		  ((void) 0);
		best_regend = NULL;
		if (reg_info)
		  ((void) 0);
		reg_info = NULL;
		if (reg_dummy)
		  ((void) 0);
		reg_dummy = NULL;
		if (reg_info_dummy)
		  ((void) 0);
		reg_info_dummy = NULL;
	  } while (0);
	  return mcnt;
	}
	switch (((re_opcode_t) * p++)) {
	case no_op:
	  break;
	case succeed:
	  goto succeed_label;
	case exactn:
	  mcnt = *p++;
	  if (translate) {
		do {
		  while (d == dend) {
			if (dend == end_match_2)
			  goto fail;
			d = string2;
			dend = end_match_2;
		  };
		  if ((unsigned char) translate[(unsigned char) *d++]
			  != (unsigned char) *p++)
			goto fail;
		}
		while (--mcnt);
	  } else {
		do {
		  while (d == dend) {
			if (dend == end_match_2)
			  goto fail;
			d = string2;
			dend = end_match_2;
		  };
		  if (*d++ != (char) *p++)
			goto fail;
		}
		while (--mcnt);
	  }
	  do {
		if (!set_regs_matched_done) {
		  active_reg_t r;
		  set_regs_matched_done = 1;
		  for (r = lowest_active_reg; r <= highest_active_reg; r++) {
			((reg_info[r]).bits.matched_something) =
			  ((reg_info[r]).bits.ever_matched_something) = 1;
		  }
		}
	  } while (0);
	  break;
	case anychar:
	  while (d == dend) {
		if (dend == end_match_2)
		  goto fail;
		d = string2;
		dend = end_match_2;
	  };
	  if ((!(bufp->syntax & RE_DOT_NEWLINE)
		   && (translate ? (char) translate[(unsigned char) (*d)] : (*d)) ==
		   '\n')
		  || (bufp->syntax & RE_DOT_NOT_NULL
			  && (translate ? (char) translate[(unsigned char) (*d)] : (*d))
			  == '\000'))
		goto fail;
	  do {
		if (!set_regs_matched_done) {
		  active_reg_t r;
		  set_regs_matched_done = 1;
		  for (r = lowest_active_reg; r <= highest_active_reg; r++) {
			((reg_info[r]).bits.matched_something) =
			  ((reg_info[r]).bits.ever_matched_something) = 1;
		  }
		}
	  } while (0);
	  d++;
	  break;
	case charset:
	case charset_not:
	  {
		unsigned char c;
		boolean not = (re_opcode_t) * (p - 1) == charset_not;

		while (d == dend) {
		  if (dend == end_match_2)
			goto fail;
		  d = string2;
		  dend = end_match_2;
		};
		c = (translate ? (char) translate[(unsigned char) (*d)] : (*d));
		if (c < (unsigned) (*p * 8)
			&& p[1 + c / 8] & (1 << (c % 8)))
		  not = !not;
		p += 1 + *p;
		if (!not)
		  goto fail;
		do {
		  if (!set_regs_matched_done) {
			active_reg_t r;
			set_regs_matched_done = 1;
			for (r = lowest_active_reg; r <= highest_active_reg; r++) {
			  ((reg_info[r]).bits.matched_something) =
				((reg_info[r]).bits.ever_matched_something) = 1;
			}
		  }
		} while (0);
		d++;
		break;
	  }
	case start_memory:
	  p1 = p;
	  if (((reg_info[*p]).bits.match_null_string_p) == 3)
		((reg_info[*p]).bits.match_null_string_p)
		  = group_match_null_string_p (&p1, pend, reg_info);
	  old_regstart[*p] = ((reg_info[*p]).bits.match_null_string_p)
		? ((regstart[*p]) == (&reg_unset_dummy)) ? d : regstart[*p]
		: regstart[*p];
	  regstart[*p] = d;
	  ((reg_info[*p]).bits.is_active) = 1;
	  ((reg_info[*p]).bits.matched_something) = 0;
	  set_regs_matched_done = 0;
	  highest_active_reg = *p;
	  if (lowest_active_reg == ((1 << 8) + 1))
		lowest_active_reg = *p;
	  p += 2;
	  just_past_start_mem = p;
	  break;
	case stop_memory:
	  old_regend[*p] = ((reg_info[*p]).bits.match_null_string_p)
		? ((regend[*p]) == (&reg_unset_dummy)) ? d : regend[*p]
		: regend[*p];
	  regend[*p] = d;
	  ((reg_info[*p]).bits.is_active) = 0;
	  set_regs_matched_done = 0;
	  if (lowest_active_reg == highest_active_reg) {
		lowest_active_reg = ((1 << 8) + 1);
		highest_active_reg = (1 << 8);
	  } else {
		unsigned char r = *p - 1;
		while (r > 0 && !((reg_info[r]).bits.is_active))
		  r--;
		if (r == 0) {
		  lowest_active_reg = ((1 << 8) + 1);
		  highest_active_reg = (1 << 8);
		} else
		  highest_active_reg = r;
	  }
	  if ((!((reg_info[*p]).bits.matched_something)
		   || just_past_start_mem == p - 1)
		  && (p + 2) < pend) {
		boolean is_a_jump_n = 0;
		p1 = p + 2;
		mcnt = 0;
		switch ((re_opcode_t) * p1++) {
		case jump_n:
		  is_a_jump_n = 1;
		case pop_failure_jump:
		case maybe_pop_jump:
		case jump:
		case dummy_failure_jump:
		  do {
			do {
			  (mcnt) = *(p1) & 0377;
			  (mcnt) += ((signed char) (*((p1) + 1))) << 8;
			} while (0);
			(p1) += 2;
		  } while (0);
		  if (is_a_jump_n)
			p1 += 2;
		  break;
		default:
		  ;
		}
		p1 += mcnt;
		if (mcnt < 0 && (re_opcode_t) * p1 == on_failure_jump
			&& (re_opcode_t) p1[1 + 2] == start_memory && p1[2 + 2] == *p) {
		  if (((reg_info[*p]).bits.ever_matched_something)) {
			unsigned r;
			((reg_info[*p]).bits.ever_matched_something) = 0;
			for (r = *p; r < (unsigned) *p + (unsigned) *(p + 1); r++) {
			  regstart[r] = old_regstart[r];
			  if (old_regend[r] >= regstart[r])
				regend[r] = old_regend[r];
			}
		  }
		  p1++;
		  do {
			do {
			  (mcnt) = *(p1) & 0377;
			  (mcnt) += ((signed char) (*((p1) + 1))) << 8;
			} while (0);
			(p1) += 2;
		  } while (0);
		  do {
			char *destination;
			active_reg_t this_reg;;;;;;;;
			while (((fail_stack).size - (fail_stack).avail) <
				   (((0 ? 0 : highest_active_reg - lowest_active_reg +
					  1) * 3) + 4)) {
			  if (!
				  ((fail_stack).size >
				   (unsigned) (re_max_failures *
							   (5 * 3 + 4)) ? 0 : ((fail_stack).stack =
												   (fail_stack_elt_t
													*) (destination =
														(char *)
														alloca (((fail_stack).
																 size << 1) *
																sizeof
																(fail_stack_elt_t)),
														memcpy (destination,
																(fail_stack).
																stack,
																(fail_stack).
																size *
																sizeof
																(fail_stack_elt_t))),
												   (fail_stack).stack ==
												   NULL ? 0 : ((fail_stack).
															   size <<= 1,
															   1))))
				return -2;;;
			};
			if (1)
			  for (this_reg = lowest_active_reg;
				   this_reg <= highest_active_reg; this_reg++) {;;;
				fail_stack.stack[fail_stack.avail++].pointer =
				  (unsigned char *) (regstart[this_reg]);;
				fail_stack.stack[fail_stack.avail++].pointer =
				  (unsigned char *) (regend[this_reg]);;;;;;;
				fail_stack.stack[fail_stack.avail++] =
				  (reg_info[this_reg].word);
			  };
			fail_stack.stack[fail_stack.avail++].integer =
			  (lowest_active_reg);;
			fail_stack.stack[fail_stack.avail++].integer =
			  (highest_active_reg);;;
			fail_stack.stack[fail_stack.avail++].pointer =
			  (unsigned char *) (p1 + mcnt);;;;
			fail_stack.stack[fail_stack.avail++].pointer =
			  (unsigned char *) d;;;
		  } while (0);
		  goto fail;
		}
	  }
	  p += 2;
	  break;
	case duplicate:
	  {
		const char *d2, *dend2;
		int regno = *p++;

		if (((regstart[regno]) == (&reg_unset_dummy))
			|| ((regend[regno]) == (&reg_unset_dummy)))
		  goto fail;
		d2 = regstart[regno];
		dend2 =
		  (((size1 && string1 <= (regstart[regno])
			 && (regstart[regno]) <= string1 + size1)
			== (size1 && string1 <= (regend[regno])
				&& (regend[regno]) <= string1 + size1))
		   ? regend[regno] : end_match_1);
		for (;;) {
		  while (d2 == dend2) {
			if (dend2 == end_match_2)
			  break;
			if (dend2 == regend[regno])
			  break;
			d2 = string2;
			dend2 = regend[regno];
		  }
		  if (d2 == dend2)
			break;
		  while (d == dend) {
			if (dend == end_match_2)
			  goto fail;
			d = string2;
			dend = end_match_2;
		  };
		  mcnt = dend - d;
		  if (mcnt > dend2 - d2)
			mcnt = dend2 - d2;
		  if (translate ? bcmp_translate (d, d2, mcnt, translate)
			  : memcmp (d, d2, mcnt * sizeof (unsigned char)))
			goto fail;
		  d += mcnt, d2 += mcnt;
		  do {
			if (!set_regs_matched_done) {
			  active_reg_t r;
			  set_regs_matched_done = 1;
			  for (r = lowest_active_reg; r <= highest_active_reg; r++) {
				((reg_info[r]).bits.matched_something) =
				  ((reg_info[r]).bits.ever_matched_something) = 1;
			  }
			}
		  } while (0);
		}
	  }
	  break;
	case begline:
	  if (d == (size1 ? string1 : string2) || !size2) {
		if (!bufp->not_bol)
		  break;
	  } else if (d[-1] == '\n' && bufp->newline_anchor) {
		break;
	  }
	  goto fail;
	case endline:
	  if (d == end2) {
		if (!bufp->not_eol)
		  break;
	  }
	  else if ((d == end1 ? *string2 : *d) == '\n' && bufp->newline_anchor) {
		break;
	  }
	  goto fail;
	case begbuf:
	  if (d == (size1 ? string1 : string2) || !size2)
		break;
	  goto fail;
	case endbuf:
	  if (d == end2)
		break;
	  goto fail;
	case on_failure_keep_string_jump:
	  do {
		do {
		  (mcnt) = *(p) & 0377;
		  (mcnt) += ((signed char) (*((p) + 1))) << 8;
		} while (0);
		(p) += 2;
	  } while (0);

	  do {
		char *destination;
		active_reg_t this_reg;;;;;;;;
		while (((fail_stack).size - (fail_stack).avail) <
			   (((0 ? 0 : highest_active_reg - lowest_active_reg + 1) * 3) +
				4)) {
		  if (!
			  ((fail_stack).size >
			   (unsigned) (re_max_failures *
						   (5 * 3 + 4)) ? 0 : ((fail_stack).stack =
											   (fail_stack_elt_t
												*) (destination =
													(char *)
													alloca (((fail_stack).
															 size << 1) *
															sizeof
															(fail_stack_elt_t)),
													memcpy (destination,
															(fail_stack).
															stack,
															(fail_stack).
															size *
															sizeof
															(fail_stack_elt_t))),
											   (fail_stack).stack ==
											   NULL ? 0 : ((fail_stack).
														   size <<= 1, 1))))
			return -2;;;
		};
		if (1)
		  for (this_reg = lowest_active_reg; this_reg <= highest_active_reg;
			   this_reg++) {;;;
			fail_stack.stack[fail_stack.avail++].pointer =
			  (unsigned char *) (regstart[this_reg]);;
			fail_stack.stack[fail_stack.avail++].pointer =
			  (unsigned char *) (regend[this_reg]);;;;;;;
			fail_stack.stack[fail_stack.avail++] = (reg_info[this_reg].word);
		  };
		fail_stack.stack[fail_stack.avail++].integer = (lowest_active_reg);;
		fail_stack.stack[fail_stack.avail++].integer = (highest_active_reg);;;
		fail_stack.stack[fail_stack.avail++].pointer =
		  (unsigned char *) (p + mcnt);;;;
		fail_stack.stack[fail_stack.avail++].pointer =
		  (unsigned char *) (NULL);;;
	  } while (0);
	  break;
	case on_failure_jump:
	on_failure:

	  do {
		do {
		  (mcnt) = *(p) & 0377;
		  (mcnt) += ((signed char) (*((p) + 1))) << 8;
		} while (0);
		(p) += 2;
	  } while (0);

	  p1 = p;
	  while (p1 < pend && (re_opcode_t) * p1 == no_op)
		p1++;
	  if (p1 < pend && (re_opcode_t) * p1 == start_memory) {
		highest_active_reg = *(p1 + 1) + *(p1 + 2);
		if (lowest_active_reg == ((1 << 8) + 1))
		  lowest_active_reg = *(p1 + 1);
	  }

	  do {
		char *destination;
		active_reg_t this_reg;;;;;;;;
		while (((fail_stack).size - (fail_stack).avail) <
			   (((0 ? 0 : highest_active_reg - lowest_active_reg + 1) * 3) +
				4)) {
		  if (!
			  ((fail_stack).size >
			   (unsigned) (re_max_failures *
						   (5 * 3 + 4)) ? 0 : ((fail_stack).stack =
											   (fail_stack_elt_t
												*) (destination =
													(char *)
													alloca (((fail_stack).
															 size << 1) *
															sizeof
															(fail_stack_elt_t)),
													memcpy (destination,
															(fail_stack).
															stack,
															(fail_stack).
															size *
															sizeof
															(fail_stack_elt_t))),
											   (fail_stack).stack ==
											   NULL ? 0 : ((fail_stack).
														   size <<= 1, 1))))
			return -2;;;
		};
		if (1)
		  for (this_reg = lowest_active_reg; this_reg <= highest_active_reg;
			   this_reg++) {;;;
			fail_stack.stack[fail_stack.avail++].pointer =
			  (unsigned char *) (regstart[this_reg]);;
			fail_stack.stack[fail_stack.avail++].pointer =
			  (unsigned char *) (regend[this_reg]);;;;;;;
			fail_stack.stack[fail_stack.avail++] = (reg_info[this_reg].word);
		  };
		fail_stack.stack[fail_stack.avail++].integer = (lowest_active_reg);;
		fail_stack.stack[fail_stack.avail++].integer = (highest_active_reg);;;
		fail_stack.stack[fail_stack.avail++].pointer =
		  (unsigned char *) (p + mcnt);;;;
		fail_stack.stack[fail_stack.avail++].pointer =
		  (unsigned char *) d;;;
	  } while (0);
	  break;
	case maybe_pop_jump:
	  do {
		do {
		  (mcnt) = *(p) & 0377;
		  (mcnt) += ((signed char) (*((p) + 1))) << 8;
		} while (0);
		(p) += 2;
	  } while (0);

	  {
		unsigned char *p2 = p;
		while (1) {
		  if (p2 + 2 < pend
			  && ((re_opcode_t) * p2 == stop_memory
				  || (re_opcode_t) * p2 == start_memory))
			p2 += 3;
		  else if (p2 + 2 + 2 * 2 < pend
				   && (re_opcode_t) * p2 == dummy_failure_jump)
			p2 += 2 + 2 * 2;
		  else
			break;
		}
		p1 = p + mcnt;
		if (p2 == pend) {
		  p[-(1 + 2)] = (unsigned char) pop_failure_jump;
		}
		else if ((re_opcode_t) * p2 == exactn
				 || (bufp->newline_anchor && (re_opcode_t) * p2 == endline)) {
		  unsigned char c
			= *p2 == (unsigned char) endline ? '\n' : p2[2];
		  if (((re_opcode_t) p1[1 + 2] == exactn) && p1[3 + 2] != c) {
			p[-(1 + 2)] = (unsigned char) pop_failure_jump;
		  }
		  else if ((re_opcode_t) p1[3] == charset
				   || (re_opcode_t) p1[3] == charset_not) {
			int not = (re_opcode_t) p1[3] == charset_not;
			if (c < (unsigned) (p1[4] * 8)
				&& p1[5 + c / 8] & (1 << (c % 8)))
			  not = !not;
			if (!not) {
			  p[-3] = (unsigned char) pop_failure_jump;
			}
		  }
		}
		else if ((re_opcode_t) * p2 == charset) {
		  if ((re_opcode_t) p1[3] == exactn && !((int) p2[1] * 8 > (int) p1[5]
												 && (p2[2 + p1[5] / 8]
													 & (1 << (p1[5] % 8))))) {
			p[-3] = (unsigned char) pop_failure_jump;
		  }
		  else if ((re_opcode_t) p1[3] == charset_not) {
			int idx;
			for (idx = 0; idx < (int) p2[1]; idx++)
			  if (!(p2[2 + idx] == 0 || (idx < (int) p1[4]
										 && ((p2[2 + idx] & ~p1[5 + idx]) ==
											 0))))
				break;
			if (idx == p2[1]) {
			  p[-3] = (unsigned char) pop_failure_jump;
			}
		  } else if ((re_opcode_t) p1[3] == charset) {
			int idx;
			for (idx = 0; idx < (int) p2[1] && idx < (int) p1[4]; idx++)
			  if ((p2[2 + idx] & p1[5 + idx]) != 0)
				break;
			if (idx == p2[1] || idx == p1[4]) {
			  p[-3] = (unsigned char) pop_failure_jump;
			}
		  }
		}
	  }
	  p -= 2;
	  if ((re_opcode_t) p[-1] != pop_failure_jump) {
		p[-1] = (unsigned char) jump;
		goto unconditional_jump;
	  }
	case pop_failure_jump:
	  {
		active_reg_t dummy_low_reg, dummy_high_reg;

		{
		  active_reg_t this_reg;
		  /* dummy */
		  --fail_stack.avail;
		  --fail_stack.avail;
		  dummy_high_reg =
			(active_reg_t) fail_stack.stack[--fail_stack.avail].integer;;
		  dummy_low_reg =
			(active_reg_t) fail_stack.stack[--fail_stack.avail].integer;;
		  if (1)
			for (this_reg = dummy_high_reg; this_reg >= dummy_low_reg;
				 this_reg--) {;
			  reg_info_dummy[this_reg].word =
				fail_stack.stack[--fail_stack.avail];;
			  reg_dummy[this_reg] =
				(const char *) fail_stack.stack[--fail_stack.avail].pointer;;
			  reg_dummy[this_reg] =
				(const char *) fail_stack.stack[--fail_stack.avail].pointer;;
		  } else {
			for (this_reg = highest_active_reg; this_reg > dummy_high_reg;
				 this_reg--) {
			  reg_info_dummy[this_reg].word.integer = 0;
			  reg_dummy[this_reg] = 0;
			  reg_dummy[this_reg] = 0;
			}
			highest_active_reg = dummy_high_reg;
		  }
		  set_regs_matched_done = 0;;
		};
	  }
	unconditional_jump:
	  ;
	case jump:
	  do {
		do {
		  (mcnt) = *(p) & 0377;
		  (mcnt) += ((signed char) (*((p) + 1))) << 8;
		} while (0);
		(p) += 2;
	  } while (0);

	  p += mcnt;
	  break;
	case jump_past_alt:
	  goto unconditional_jump;
	case dummy_failure_jump:
	  do {
		char *destination;
		active_reg_t this_reg;;;;;;;;
		while (((fail_stack).size - (fail_stack).avail) <
			   (((0 ? 0 : highest_active_reg - lowest_active_reg + 1) * 3) +
				4)) {
		  if (!
			  ((fail_stack).size >
			   (unsigned) (re_max_failures *
						   (5 * 3 + 4)) ? 0 : ((fail_stack).stack =
											   (fail_stack_elt_t
												*) (destination =
													(char *)
													alloca (((fail_stack).
															 size << 1) *
															sizeof
															(fail_stack_elt_t)),
													memcpy (destination,
															(fail_stack).
															stack,
															(fail_stack).
															size *
															sizeof
															(fail_stack_elt_t))),
											   (fail_stack).stack ==
											   NULL ? 0 : ((fail_stack).
														   size <<= 1, 1))))
			return -2;;;
		};
		if (1)
		  for (this_reg = lowest_active_reg; this_reg <= highest_active_reg;
			   this_reg++) {;;;
			fail_stack.stack[fail_stack.avail++].pointer =
			  (unsigned char *) (regstart[this_reg]);;
			fail_stack.stack[fail_stack.avail++].pointer =
			  (unsigned char *) (regend[this_reg]);;;;;;;
			fail_stack.stack[fail_stack.avail++] = (reg_info[this_reg].word);
		  };
		fail_stack.stack[fail_stack.avail++].integer = (lowest_active_reg);;
		fail_stack.stack[fail_stack.avail++].integer = (highest_active_reg);;;
		fail_stack.stack[fail_stack.avail++].pointer =
		  (unsigned char *) (NULL);;;;
		fail_stack.stack[fail_stack.avail++].pointer =
		  (unsigned char *) (NULL);;;
	  } while (0);
	  goto unconditional_jump;
	case push_dummy_failure:
	  do {
		char *destination;
		active_reg_t this_reg;;;;;;;;
		while (((fail_stack).size - (fail_stack).avail) <
			   (((0 ? 0 : highest_active_reg - lowest_active_reg + 1) * 3) +
				4)) {
		  if (!
			  ((fail_stack).size >
			   (unsigned) (re_max_failures *
						   (5 * 3 + 4)) ? 0 : ((fail_stack).stack =
											   (fail_stack_elt_t
												*) (destination =
													(char *)
													alloca (((fail_stack).
															 size << 1) *
															sizeof
															(fail_stack_elt_t)),
													memcpy (destination,
															(fail_stack).
															stack,
															(fail_stack).
															size *
															sizeof
															(fail_stack_elt_t))),
											   (fail_stack).stack ==
											   NULL ? 0 : ((fail_stack).
														   size <<= 1, 1))))
			return -2;;;
		};
		if (1)
		  for (this_reg = lowest_active_reg; this_reg <= highest_active_reg;
			   this_reg++) {;;;
			fail_stack.stack[fail_stack.avail++].pointer =
			  (unsigned char *) (regstart[this_reg]);;
			fail_stack.stack[fail_stack.avail++].pointer =
			  (unsigned char *) (regend[this_reg]);;;;;;;
			fail_stack.stack[fail_stack.avail++] = (reg_info[this_reg].word);
		  };
		fail_stack.stack[fail_stack.avail++].integer = (lowest_active_reg);;
		fail_stack.stack[fail_stack.avail++].integer = (highest_active_reg);;;
		fail_stack.stack[fail_stack.avail++].pointer =
		  (unsigned char *) (NULL);;;;
		fail_stack.stack[fail_stack.avail++].pointer =
		  (unsigned char *) (NULL);;;
	  } while (0);
	  break;
	case succeed_n:
	  do {
		(mcnt) = *(p + 2) & 0377;
		(mcnt) += ((signed char) (*((p + 2) + 1))) << 8;
	  } while (0);

	  if (mcnt > 0) {
		mcnt--;
		p += 2;
		do {
		  do {
			(p)[0] = (mcnt) & 0377;
			(p)[1] = (mcnt) >> 8;
		  } while (0);
		  (p) += 2;
		} while (0);
	  } else if (mcnt == 0) {
		p[2] = (unsigned char) no_op;
		p[3] = (unsigned char) no_op;
		goto on_failure;
	  }
	  break;
	case jump_n:
	  do {
		(mcnt) = *(p + 2) & 0377;
		(mcnt) += ((signed char) (*((p + 2) + 1))) << 8;
	  } while (0);

	  if (mcnt) {
		mcnt--;
		do {
		  (p + 2)[0] = (mcnt) & 0377;
		  (p + 2)[1] = (mcnt) >> 8;
		} while (0);

		goto unconditional_jump;
	  }
	  else
		p += 2 * 2;
	  break;
	case set_number_at:
	  {
		do {
		  do {
			(mcnt) = *(p) & 0377;
			(mcnt) += ((signed char) (*((p) + 1))) << 8;
		  } while (0);
		  (p) += 2;
		} while (0);
		p1 = p + mcnt;
		do {
		  do {
			(mcnt) = *(p) & 0377;
			(mcnt) += ((signed char) (*((p) + 1))) << 8;
		  } while (0);
		  (p) += 2;
		} while (0);

		do {
		  (p1)[0] = (mcnt) & 0377;
		  (p1)[1] = (mcnt) >> 8;
		} while (0);
		break;
	  }
	case wordbound:
	  {
		boolean prevchar, thischar;

		if (d == (size1 ? string1 : string2) || !size2 || d == end2)
		  break;
		prevchar =
		  (re_syntax_table
		   [(unsigned char) ((d - 1) == end1 ? *string2 : (d - 1) ==
							 string2 - 1 ? *(end1 - 1) : *(d - 1))] == 1);
		thischar =
		  (re_syntax_table
		   [(unsigned char) (d == end1 ? *string2 : d ==
							 string2 - 1 ? *(end1 - 1) : *d)] == 1);
		if (prevchar != thischar)
		  break;
		goto fail;
	  }
	case notwordbound:
	  {
		boolean prevchar, thischar;

		if (d == (size1 ? string1 : string2) || !size2 || d == end2)
		  goto fail;
		prevchar =
		  (re_syntax_table
		   [(unsigned char) ((d - 1) == end1 ? *string2 : (d - 1) ==
							 string2 - 1 ? *(end1 - 1) : *(d - 1))] == 1);
		thischar =
		  (re_syntax_table
		   [(unsigned char) (d == end1 ? *string2 : d ==
							 string2 - 1 ? *(end1 - 1) : *d)] == 1);
		if (prevchar != thischar)
		  goto fail;
		break;
	  }
	case wordbeg:
	  if (!(d == end2)
		  &&
		  (re_syntax_table
		   [(unsigned char) (d == end1 ? *string2 : d ==
							 string2 - 1 ? *(end1 - 1) : *d)] == 1)
		  && ((d == (size1 ? string1 : string2) || !size2)
			  ||
			  !(re_syntax_table
				[(unsigned char) ((d - 1) == end1 ? *string2 : (d - 1) ==
								  string2 - 1 ? *(end1 - 1) : *(d - 1))] ==
				1)))
		break;
	  goto fail;
	case wordend:
	  if (!(d == (size1 ? string1 : string2) || !size2)
		  &&
		  (re_syntax_table
		   [(unsigned char) ((d - 1) == end1 ? *string2 : (d - 1) ==
							 string2 - 1 ? *(end1 - 1) : *(d - 1))] == 1)
		  && ((d == end2)
			  ||
			  !(re_syntax_table
				[(unsigned char) (d == end1 ? *string2 : d ==
								  string2 - 1 ? *(end1 - 1) : *d)] == 1)))
		break;
	  goto fail;
	case wordchar:
	  while (d == dend) {
		if (dend == end_match_2)
		  goto fail;
		d = string2;
		dend = end_match_2;
	  };
	  if (!
		  (re_syntax_table
		   [(unsigned char) (d == end1 ? *string2 : d ==
							 string2 - 1 ? *(end1 - 1) : *d)] == 1))
		goto fail;
	  do {
		if (!set_regs_matched_done) {
		  active_reg_t r;
		  set_regs_matched_done = 1;
		  for (r = lowest_active_reg; r <= highest_active_reg; r++) {
			((reg_info[r]).bits.matched_something) =
			  ((reg_info[r]).bits.ever_matched_something) = 1;
		  }
		}
	  } while (0);
	  d++;
	  break;
	case notwordchar:
	  while (d == dend) {
		if (dend == end_match_2)
		  goto fail;
		d = string2;
		dend = end_match_2;
	  };
	  if (re_syntax_table
		   [(unsigned char) (d == end1 ? *string2 : d ==
							 string2 - 1 ? *(end1 - 1) : *d)] == 1)
		goto fail;
	  do {
		if (!set_regs_matched_done) {
		  active_reg_t r;
		  set_regs_matched_done = 1;
		  for (r = lowest_active_reg; r <= highest_active_reg; r++) {
			((reg_info[r]).bits.matched_something) =
			  ((reg_info[r]).bits.ever_matched_something) = 1;
		  }
		}
	  } while (0);
	  d++;
	  break;
	default:
	  abort ();
	}
	continue;
  fail:
	if (!(fail_stack.avail == 0)) {
	  ;
	  {
		active_reg_t this_reg;
		const unsigned char *string_temp;;;;;;;;
		string_temp = fail_stack.stack[--fail_stack.avail].pointer;
		if (string_temp != NULL)
		  d = (const char *) string_temp;;;;
		p = (unsigned char *) fail_stack.stack[--fail_stack.avail].pointer;;;
		highest_active_reg =
		  (active_reg_t) fail_stack.stack[--fail_stack.avail].integer;;
		lowest_active_reg =
		  (active_reg_t) fail_stack.stack[--fail_stack.avail].integer;;
		if (1)
		  for (this_reg = highest_active_reg; this_reg >= lowest_active_reg;
			   this_reg--) {;
			reg_info[this_reg].word = fail_stack.stack[--fail_stack.avail];;
			regend[this_reg] =
			  (const char *) fail_stack.stack[--fail_stack.avail].pointer;;
			regstart[this_reg] =
			  (const char *) fail_stack.stack[--fail_stack.avail].pointer;;
		} else {
		  for (this_reg = highest_active_reg; this_reg > highest_active_reg;
			   this_reg--) {
			reg_info[this_reg].word.integer = 0;
			regend[this_reg] = 0;
			regstart[this_reg] = 0;
		  }
		  highest_active_reg = highest_active_reg;
		}
		set_regs_matched_done = 0;;
	  };
	  if (!p)
		goto fail;
	  if (p < pend) {
		boolean is_a_jump_n = 0;
		switch ((re_opcode_t) * p) {
		case jump_n:
		  is_a_jump_n = 1;
		case maybe_pop_jump:
		case pop_failure_jump:
		case jump:
		  p1 = p + 1;
		  do {
			do {
			  (mcnt) = *(p1) & 0377;
			  (mcnt) += ((signed char) (*((p1) + 1))) << 8;
			} while (0);
			(p1) += 2;
		  } while (0);
		  p1 += mcnt;
		  if ((is_a_jump_n && (re_opcode_t) * p1 == succeed_n)
			  || (!is_a_jump_n && (re_opcode_t) * p1 == on_failure_jump))
			goto fail;
		  break;
		default:
		  ;
		}
	  }
	  if (d >= string1 && d <= end1)
		dend = end_match_1;
	} else
	  break;
  }
  if (best_regs_set)
	goto restore_best_regs;
  do {;
	if (regstart)
	  ((void) 0);
	regstart = NULL;
	if (regend)
	  ((void) 0);
	regend = NULL;
	if (old_regstart)
	  ((void) 0);
	old_regstart = NULL;
	if (old_regend)
	  ((void) 0);
	old_regend = NULL;
	if (best_regstart)
	  ((void) 0);
	best_regstart = NULL;
	if (best_regend)
	  ((void) 0);
	best_regend = NULL;
	if (reg_info)
	  ((void) 0);
	reg_info = NULL;
	if (reg_dummy)
	  ((void) 0);
	reg_dummy = NULL;
	if (reg_info_dummy)
	  ((void) 0);
	reg_info_dummy = NULL;
  } while (0);
  return -1;
}

static boolean
group_match_null_string_p (unsigned char **p, unsigned char *end,
								register_info_type * reg_info)
{
  int mcnt;
  unsigned char *p1 = *p + 2;
  while (p1 < end) {
	switch ((re_opcode_t) * p1) {
	case on_failure_jump:
	  p1++;
	  do {
		do {
		  (mcnt) = *(p1) & 0377;
		  (mcnt) += ((signed char) (*((p1) + 1))) << 8;
		} while (0);
		(p1) += 2;
	  } while (0);
	  if (mcnt >= 0) {
		while ((re_opcode_t) p1[mcnt - (1 + 2)] == jump_past_alt) {
		  if (!alt_match_null_string_p (p1, p1 + mcnt -
											 (1 + 2), reg_info))
			return 0;
		  p1 += mcnt;
		  if ((re_opcode_t) * p1 != on_failure_jump)
			break;
		  p1++;
		  do {
			do {
			  (mcnt) = *(p1) & 0377;
			  (mcnt) += ((signed char) (*((p1) + 1))) << 8;
			} while (0);
			(p1) += 2;
		  } while (0);
		  if ((re_opcode_t) p1[mcnt - (1 + 2)] != jump_past_alt) {
			p1 -= 1 + 2;
			break;
		  }
		}
		do {
		  (mcnt) = *(p1 - 2) & 0377;
		  (mcnt) += ((signed char) (*((p1 - 2) + 1))) << 8;
		} while (0);
		if (!alt_match_null_string_p (p1, p1 + mcnt, reg_info))
		  return 0;
		p1 += mcnt;
	  }
	  break;
	case stop_memory:
	  *p = p1 + 2;
	  return 1;
	default:
	  if (!common_op_match_null_string_p (&p1, end, reg_info))
		return 0;
	}
  }
  return 0;
}

static boolean
alt_match_null_string_p (unsigned char *p, unsigned char *end,
							  register_info_type * reg_info)
{
  int mcnt;
  unsigned char *p1 = p;
  while (p1 < end) {
	switch ((re_opcode_t) * p1) {
	case on_failure_jump:
	  p1++;
	  do {
		do {
		  (mcnt) = *(p1) & 0377;
		  (mcnt) += ((signed char) (*((p1) + 1))) << 8;
		} while (0);
		(p1) += 2;
	  } while (0);
	  p1 += mcnt;
	  break;
	default:
	  if (!common_op_match_null_string_p (&p1, end, reg_info))
		return 0;
	}
  }
  return 1;
}
static boolean
common_op_match_null_string_p (unsigned char **p, unsigned char *end,
									register_info_type * reg_info)
{
  int mcnt;
  boolean ret;
  int reg_no;
  unsigned char *p1 = *p;
  switch ((re_opcode_t) * p1++) {
  case no_op:
  case begline:
  case endline:
  case begbuf:
  case endbuf:
  case wordbeg:
  case wordend:
  case wordbound:
  case notwordbound:
	break;
  case start_memory:
	reg_no = *p1;
	ret = group_match_null_string_p (&p1, end, reg_info);
	if (((reg_info[reg_no]).bits.match_null_string_p) == 3)
	  ((reg_info[reg_no]).bits.match_null_string_p) = ret;
	if (!ret)
	  return 0;
	break;
  case jump:
	do {
	  do {
		(mcnt) = *(p1) & 0377;
		(mcnt) += ((signed char) (*((p1) + 1))) << 8;
	  } while (0);
	  (p1) += 2;
	} while (0);
	if (mcnt >= 0)
	  p1 += mcnt;
	else
	  return 0;
	break;
  case succeed_n:
	p1 += 2;
	do {
	  do {
		(mcnt) = *(p1) & 0377;
		(mcnt) += ((signed char) (*((p1) + 1))) << 8;
	  } while (0);
	  (p1) += 2;
	} while (0);
	if (mcnt == 0) {
	  p1 -= 2 * 2;
	  do {
		do {
		  (mcnt) = *(p1) & 0377;
		  (mcnt) += ((signed char) (*((p1) + 1))) << 8;
		} while (0);
		(p1) += 2;
	  } while (0);
	  p1 += mcnt;
	} else
	  return 0;
	break;
  case duplicate:
	if (!((reg_info[*p1]).bits.match_null_string_p))
	  return 0;
	break;
  case set_number_at:
	p1 += 2 * 2;
  default:
	return 0;
  }
  *p = p1;
  return 1;
}

static int
bcmp_translate (const char *s1, const char *s2,
					 int len, char *translate)
{
  const unsigned char *p1 = (const unsigned char *) s1;
  const unsigned char *p2 = (const unsigned char *) s2;
  while (len) {
	if (translate[*p1++] != translate[*p2++])
	  return 1;
	len--;
  }
  return 0;
}

reg_syntax_t re_syntax_options;
reg_syntax_t
re_set_syntax (reg_syntax_t syntax)
{
  reg_syntax_t ret = re_syntax_options;
  re_syntax_options = syntax;
  return ret;
}

/* This table gives an error message for each of the error codes listed
   in regex.h.  Obviously the order here has to be same as there.
   POSIX doesn't require that we do anything for REG_NOERROR,
   but why not be nice?  */

static const char re_error_msgid[] =
  {
# define REG_NOERROR_IDX        0
    "Success" "\0"   /* REG_NOERROR */
# define REG_NOMATCH_IDX (REG_NOERROR_IDX + sizeof "Success")
    "No match" "\0"  /* REG_NOMATCH */
# define REG_BADPAT_IDX (REG_NOMATCH_IDX + sizeof "No match")
    "Invalid regular expression" "\0"/* REG_BADPAT */
# define REG_ECOLLATE_IDX (REG_BADPAT_IDX + sizeof "Invalid regular expression")
    "Invalid collation character" "\0"/* REG_ECOLLATE */
# define REG_ECTYPE_IDX (REG_ECOLLATE_IDX + sizeof "Invalid collation character")
    "Invalid character class name" "\0"/* REG_ECTYPE */
# define REG_EESCAPE_IDX        (REG_ECTYPE_IDX + sizeof "Invalid character class name")
    "Trailing backslash" "\0"/* REG_EESCAPE */
# define REG_ESUBREG_IDX        (REG_EESCAPE_IDX + sizeof "Trailing backslash")
    "Invalid back reference" "\0"/* REG_ESUBREG */
# define REG_EBRACK_IDX (REG_ESUBREG_IDX + sizeof "Invalid back reference")
    "Unmatched [ or [^" "\0" /* REG_EBRACK */
# define REG_EPAREN_IDX (REG_EBRACK_IDX + sizeof "Unmatched [ or [^")
    "Unmatched ( or \\(" "\0"/* REG_EPAREN */
# define REG_EBRACE_IDX (REG_EPAREN_IDX + sizeof "Unmatched ( or \\(")
    "Unmatched \\{" "\0"/* REG_EBRACE */
# define REG_BADBR_IDX  (REG_EBRACE_IDX + sizeof "Unmatched \\{")
    "Invalid content of \\{\\}" "\0"/* REG_BADBR */
# define REG_ERANGE_IDX (REG_BADBR_IDX + sizeof "Invalid content of \\{\\}")
    "Invalid range end" "\0" /* REG_ERANGE */
# define REG_ESPACE_IDX (REG_ERANGE_IDX + sizeof "Invalid range end")
    "Memory exhausted" "\0"/* REG_ESPACE */
# define REG_BADRPT_IDX (REG_ESPACE_IDX + sizeof "Memory exhausted")
    "Invalid preceding regular expression" "\0"/* REG_BADRPT */
# define REG_EEND_IDX   (REG_BADRPT_IDX + sizeof "Invalid preceding regular expression")
    "Premature end of regular expression" "\0"/* REG_EEND */
# define REG_ESIZE_IDX  (REG_EEND_IDX + sizeof "Premature end of regular expression")
    "Regular expression too big" "\0"/* REG_ESIZE */
# define REG_ERPAREN_IDX        (REG_ESIZE_IDX + sizeof "Regular expression too big")
    "Unmatched ) or \\)" /* REG_ERPAREN */
};

static const size_t re_error_msgid_idx[] = {
    REG_NOERROR_IDX,
    REG_NOMATCH_IDX,
    REG_BADPAT_IDX,
    REG_ECOLLATE_IDX,
    REG_ECTYPE_IDX,
    REG_EESCAPE_IDX,
    REG_ESUBREG_IDX,
    REG_EBRACK_IDX,
    REG_EPAREN_IDX,
    REG_EBRACE_IDX,
    REG_BADBR_IDX,
    REG_ERANGE_IDX,
    REG_ESPACE_IDX,
    REG_BADRPT_IDX,
    REG_EEND_IDX,
    REG_ESIZE_IDX,
    REG_ERPAREN_IDX
};

static boolean
group_in_compile_stack (compile_stack_type compile_stack, regnum_t regnum)
{
  int this_element;

  for (this_element = compile_stack.avail - 1;
	   this_element >= 0; this_element--)
	if (compile_stack.stack[this_element].regnum == regnum)
	  return 1;
  return 0;
}

void
re_set_registers (struct re_pattern_buffer *bufp, struct re_registers *regs,
					unsigned num_regs, regoff_t * starts, regoff_t * ends)
{
  if (num_regs) {
	bufp->regs_allocated = REGS_REALLOCATE;
	regs->num_regs = num_regs;
	regs->start = starts;
	regs->end = ends;
  } else {
	bufp->regs_allocated = REGS_UNALLOCATED;
	regs->num_regs = 0;
	regs->start = regs->end = (regoff_t *) 0;
  }
}

int
re_search (struct re_pattern_buffer *bufp, const char *string, int size,
			 int startpos, int range, struct re_registers *regs)
{
  return re_search_2 (bufp, NULL, 0, string, size, startpos, range, regs,
						size);
}

int
re_match (struct re_pattern_buffer *bufp, const char *string, int size,
			int pos, struct re_registers *regs)
{
  int result;
  result = re_match_2_internal (bufp, NULL, 0, string, size, pos, regs, size);
  return result;
}

int
re_match_2 (struct re_pattern_buffer *bufp, const char *string1, int size1,
			  const char *string2, int size2, int pos,
			  struct re_registers *regs, int stop)
{
  int result;
  result = re_match_2_internal (bufp, string1, size1, string2, size2, pos, regs, stop);
  return result;
}

const char *
re_compile_pattern (const char *pattern, size_t length,
					  struct re_pattern_buffer *bufp)
{
  reg_errcode_t ret;
  bufp->regs_allocated = REGS_UNALLOCATED;
  bufp->no_sub = 0;
  bufp->newline_anchor = 1;
  ret = regex_compile (pattern, length, re_syntax_options, bufp);
  if (!ret)
	return NULL;
  return (re_error_msgid + re_error_msgid_idx[(int) ret]);
}

static struct re_pattern_buffer re_comp_buf;

char *
re_comp (const char *s)
{
  reg_errcode_t ret;

  if (!s) {
	if (!re_comp_buf.buffer)
	  return ("No previous regular expression");
	return 0;
  }
  if (!re_comp_buf.buffer) {
	re_comp_buf.buffer = (unsigned char *) malloc (200);
	if (re_comp_buf.buffer == NULL)
	  return (char *) (re_error_msgid + re_error_msgid_idx[(int) REG_ESPACE]);
	re_comp_buf.allocated = 200;
	re_comp_buf.fastmap = (char *) malloc (1 << 8);
	if (re_comp_buf.fastmap == NULL)
	  return (char *) (re_error_msgid + re_error_msgid_idx[(int) REG_ESPACE]);
  }
  re_comp_buf.newline_anchor = 1;
  ret = regex_compile (s, strlen (s), re_syntax_options, &re_comp_buf);
  if (!ret)
	return NULL;
  return (char *) (re_error_msgid + re_error_msgid_idx[(int) ret]);
}

int
re_exec (const char *s)
{
  const int len = strlen (s);
  return 0 <= re_search (&re_comp_buf, s, len, 0, len, NULL);
}

int
regcomp (regex_t * preg, const char *pattern, int cflags)
{
  reg_errcode_t ret;
  reg_syntax_t syntax = (cflags & REG_EXTENDED) ?
			RE_SYNTAX_POSIX_EXTENDED : RE_SYNTAX_POSIX_BASIC;

  preg->buffer = 0;
  preg->allocated = 0;
  preg->used = 0;
  preg->fastmap = (char *) malloc (1 << 8);
  if (cflags & REG_ICASE) {
	unsigned i;

	preg->translate = malloc (256 * sizeof (*(char *)0));
	if (preg->translate == NULL)
	  return (int) REG_ESPACE;
	for (i = 0; i < 256; i++)
	  preg->translate[i] = (1 && isupper (i)) ? tolower (i) : i;
  } else
	preg->translate = NULL;
  if (cflags & REG_NEWLINE) {
	syntax &= ~RE_DOT_NEWLINE;
	syntax |= RE_HAT_LISTS_NOT_NEWLINE;
	preg->newline_anchor = 1;
  } else
	preg->newline_anchor = 0;
  preg->no_sub = !!(cflags & REG_NOSUB);
  ret = regex_compile (pattern, strlen (pattern), syntax, preg);
  if (ret == REG_ERPAREN)
	ret = REG_EPAREN;
  if (ret == REG_NOERROR && preg->fastmap) {
	if (re_compile_fastmap (preg) == -2) {
	  free (preg->fastmap);
	  preg->fastmap = NULL;
	}
  }
  return (int) ret;
}

int
regexec (const regex_t * preg, const char *string, size_t nmatch,
		   regmatch_t pmatch[], int eflags)
{
  int ret;
  struct re_registers regs;
  regex_t private_preg;
  int len = strlen (string);
  boolean want_reg_info = !preg->no_sub && nmatch > 0;

  memcpy (&private_preg, preg, sizeof (*preg));
  private_preg.not_bol = !!(eflags & REG_NOTBOL);
  private_preg.not_eol = !!(eflags & REG_NOTEOL);
  private_preg.regs_allocated = REGS_FIXED;
  if (want_reg_info) {
	regs.num_regs = nmatch;
	regs.start = ((regoff_t *) malloc ((nmatch * 2) * sizeof (regoff_t)));
	if (regs.start == NULL)
	  return (int) REG_NOMATCH;
	regs.end = regs.start + nmatch;
  }
  ret = re_search (&private_preg, string, len, 0, len, want_reg_info ? &regs : NULL);
  if (want_reg_info) {
	if (ret >= 0) {
	  unsigned r;

	  for (r = 0; r < nmatch; r++) {
		pmatch[r].rm_so = regs.start[r];
		pmatch[r].rm_eo = regs.end[r];
	  }
	}
	free (regs.start);
  }
  return ret >= 0 ? (int) REG_NOERROR : (int) REG_NOMATCH;
}

size_t
regerror (int errcode, const regex_t * preg, char *errbuf, size_t errbuf_size)
{
  const char *msg;
  size_t msg_size;

  if (errcode < 0 || errcode >= (int) (sizeof (re_error_msgid_idx) / sizeof (re_error_msgid_idx[0])))
	abort ();
  msg = (re_error_msgid + re_error_msgid_idx[errcode]);
  msg_size = strlen (msg) + 1;
  if (errbuf_size != 0) {
	if (msg_size > errbuf_size) {
	  memcpy (errbuf, msg, errbuf_size - 1);
	  errbuf[errbuf_size - 1] = 0;
	} else
	  memcpy (errbuf, msg, msg_size);
  }
  return msg_size;
}

void
regfree (regex_t * preg)
{
  free (preg->buffer);
  preg->buffer = NULL;
  preg->allocated = 0;
  preg->used = 0;
  free (preg->fastmap);
  preg->fastmap = NULL;
  preg->fastmap_accurate = 0;
  free (preg->translate);
  preg->translate = NULL;
}
#endif /* HAVE_REGEX */
