/*
 * Copyright 1989 - 1991, Julianne Frances Haugh <jockgrrl@austin.rr.com>
 * All rights reserved.
 *
 * add if(shell==NULL) shell=geten("SHELL"); if(shell==NULL) shell=DEFAULT_SHELL;
 * auto login shell if arg==NULL
 * by Vladimir Oleynik (C) 2016 <dzo@simtreas.ru>
 *
 */

#include "libbb.h"
#include <unistd.h>
#include <string.h>
#include <syslog.h>
#include <ctype.h>

/* Run SHELL, or DEFAULT_SHELL if SHELL is empty.
   If COMMAND is nonzero, pass it to the shell with the -c option.
   If ADDITIONAL_ARGS is nonzero, pass it to the shell as more
   arguments.  */

FAST_FUNC void
run_shell(const char *shell, const char *command, const char **additional_args)
{
	const char **args;
	int argno = 1;
	int additional_args_cnt = 0;
	int loginshell = command == NULL;

	for (args = additional_args; args && *args; args++)
		additional_args_cnt++;

	args = xmalloc(sizeof (char *) * (4  + additional_args_cnt));

	if(shell == NULL) {
		shell = getenv("SHELL");
		if (shell == NULL)
			shell = DEFAULT_SHELL;
	}
	if(shell == DEFAULT_SHELL || !strcmp(shell, DEFAULT_SHELL)) {
		args [0] = loginshell ? DEFAULT_LOGIN_SHELL_SHORT_NAME : DEFAULT_SHELL_SHORT_NAME;
	} else {
		args [0] = bb_basename ( shell );

		if ( loginshell ) {
			size_t l = strlen(args[0]) + 1;
			char *minus_argv0 = alloca(l + 1);
			minus_argv0[0] = '-';
			memcpy(minus_argv0 + 1, args [0], l);
			args[0] = minus_argv0;
		}
	}

	if (command) {
		args[argno++] = "-c";
		args[argno++] = command;
	}
	if (additional_args) {
		for ( ; *additional_args; ++additional_args )
			args[argno++] = *additional_args;
	}
	args [argno] = 0;
	execv (shell, (char **)args);
	bb_perror_msg_and_die ( "cannot run %s", shell );
}
