/*
 * Copyright 1989 - 1991, Julianne Frances Haugh <jockgrrl@austin.rr.com>
 * All rights reserved.
 *
 * Copyright (C) 2008 Vladimir N. Oleynik <dzo@simtreas.ru>
 */

#include "libbb.h"
#include <unistd.h>
#include <string.h>
#include <syslog.h>
#include <ctype.h>


#define DEFAULT_LOGIN_PATH      "/bin:/usr/bin"
#define DEFAULT_ROOT_LOGIN_PATH "/usr/sbin:/bin:/usr/bin:/sbin"

extern FAST_FUNC int bb_setenv1 ( const char *key, const char *value )
{
	    return setenv (key, value, 1);
}

static void exsetenv ( const char *key, const char *value )
{
	    if (bb_setenv1(key, value))
		    bb_error_msg_and_die("%s", bb_msg_memory_exhausted);
}

void FAST_FUNC
setup_environment (const char *shell, int loginshell, int changeenv, const struct passwd *pw)
{
	if ( loginshell ) {
		const char *term;

		/* Change the current working directory to be the home directory
		 * of the user.  It is a fatal error for this process to be unable
		 * to change to that directory.  There is no "default" home
		 * directory.
		 * Some systems default to HOME=/
		 */
		if ( chdir ( pw-> pw_dir )) {
			if ( chdir ( "/" )) {
				xlog_warn("unable to cd to %s' for user %s'", pw-> pw_dir, pw-> pw_name );
				bb_error_msg_and_die ( "cannot cd to home directory or /" );
			}
			fputs ( "warning: cannot change to home directory\n", stderr );
		}

		/* Leave TERM unchanged.  Set HOME, SHELL, USER, LOGNAME, PATH.
		   Unset all other environment variables.  */
		term = getenv ("TERM");
		clearenv ( );
		if ( term )
			exsetenv ( "TERM", term );
		exsetenv ( "HOME",    pw-> pw_dir );
		exsetenv ( "SHELL",   shell );
		exsetenv ( "USER",    pw-> pw_name );
		exsetenv ( "LOGNAME", pw-> pw_name );
		exsetenv ( "PATH",    ( pw-> pw_uid ? DEFAULT_LOGIN_PATH : DEFAULT_ROOT_LOGIN_PATH ));
	}
	else if ( changeenv ) {
		/* Set HOME, SHELL, and if not becoming a super-user,
		   USER and LOGNAME.  */
		exsetenv ( "HOME",  pw-> pw_dir );
		exsetenv ( "SHELL", shell );
		if  ( pw-> pw_uid ) {
			exsetenv ( "USER",    pw-> pw_name );
			exsetenv ( "LOGNAME", pw-> pw_name );
		}
	}
}
