/*
 * Copyright (C) 2010 Karel Zak <kzak@redhat.com>
 * Copyright (C) 2010 Davidlohr Bueso <dave@gnu.org>
 *
 * Copyright (C) 2016 Vladimir Oleynik <dzo@simtreas.ru>
 * convert malloc buffer to stack buffer
 */

#include "libbb.h"

/*
 * returns exponent (2^x=n) in range KiB..PiB
 */
static int get_exp(ubb64_t n)
{
	int shft;

	for (shft = 10; shft <= 60; shft += 10) {
		if (n < (1ULL << shft))
			break;
	}
	return shft - 10;
}

FAST_FUNC void size_to_human_string(int options, ubb64_t bytes, char buf[SIZE_TO_HUMAN_STRING_BUFSIZ])
{
	int dec, exp;
	ubb64_t frac;
	static const char letters[] = "BKMGTPE";
	char suffix[sizeof(" KiB")], *psuf = suffix;
	char c;

	if (options & SIZE_SUFFIX_SPACE)
		*psuf++ = ' ';

	exp  = get_exp(bytes);
	c    = letters[(exp ? exp / 10 : 0)];
	dec  = exp ? bytes / (1ULL << exp) : bytes;
	frac = exp ? bytes % (1ULL << exp) : 0;

	*psuf++ = c;

	if ((options & SIZE_SUFFIX_3LETTER) && (c != 'B')) {
		*psuf++ = 'i';
		*psuf++ = 'B';
	}

	*psuf = '\0';

	/* fprintf(stderr, "exp: %d, unit: %c, dec: %d, frac: %jd\n",
	 *                 exp, suffix[0], dec, frac);
	 */

	if (frac) {
		/* round */
		frac = (frac / (1ULL << (exp - 10)) + 50) / 100;
		if (frac == 10)
			dec++, frac = 0;
	}

	if (frac)
		snprintf(buf, SIZE_TO_HUMAN_STRING_BUFSIZ, "%d.%"P_BB64T"u%s", dec, frac, suffix);
	 else
		snprintf(buf, SIZE_TO_HUMAN_STRING_BUFSIZ, "%d%s", dec, suffix);
}
