/*
 * Utility routines.
 *
 * Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 * Copyright (C) 2005-2019 Vladimir Oleynik <dzo@simtreas.ru>
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */

#include "libbb.h"
#include <string.h>
#include <syslog.h>

extern FAST_FUNC void bb_vperror_msg(const char *s, va_list p)
{
	int e = errno;
	size_t l;
	const char *sep;

	if(s == NULL) {
		sep = s = "";
		l = 3;  /* "%m\0" */
	} else {
		sep = ": ";
		l = strlen(s) + 5;  /* + strlen(sep) */
	}
	if(log_syslog) {
		va_list copy;
		char *se = alloca(l);

		sprintf(se, "%s%s%%m", s, sep);
		va_copy(copy, p);
		errno = e;
		vsyslog(log_syslog, se, p);
		va_end(copy);
	}
	if(log_stderr) {
		char *se = alloca(strlen(bb_applet_name) + l + 3); /* + ": \n" */
		sprintf(se, "%s%s%s: %%m\n", bb_applet_name, sep, s);
		errno = e;
		bb_vfprintf(stderr, se, p);
	}
}
