/*
 * Utility routines.
 *
 *  * Written by Karel Zak <kzak@redhat.com>
 *            Petr Uzel <petr.uzel@suse.cz>
 *            Vladimir Oleynik <dzo@simtreas.ru> 2019
 *               Busybox port
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */

#include "libbb.h"
#include <unistd.h>


FAST_FUNC ssize_t write_all(int fd, const void *buf, size_t count)
{
	while (count) {
		ssize_t tmp;

		tmp = safe_write(fd, buf, count);
		if(tmp < 0) {
			if (errno == EAGAIN)    /* Try later, *sigh* */
				xusleep(250000);
			else
				return -1;      /* EINTR process from safe_write */
		} else {
			count -= tmp;
			buf = (const void *) ((const char *) buf + tmp);
		}
	}
	return 0;
}
