/*
 * chdir to working directory
 * close stdin and stdout, stderr (optional)
 * detach from controlling terminal
 * create pidfile if foreground (now)
 *
 * Copyright (C) 2004-2017 Vladimir N. Oleynik <dzo@simtreas.ru>
 */

#include "libbb.h"
#include <unistd.h>
#include <paths.h>
#include <fcntl.h>
#include <syslog.h>

FAST_FUNC char *xdaemon(int flags, ...)
{
	int noclose = flags & XDAEMON_NO_CLOSE;
	char *pf = NULL;
	va_list p;
	FILE *fp = NULL;

	va_start(p, flags);
	if((flags & XDAEMON_SET_DIR) != 0) {
		const char *wp = va_arg(p, const char *);
		xchdir(wp);
		flags |= XDAEMON_NO_CHDIR;
	}
	if((flags & XDAEMON_PF_ARG) != 0)
		pf = va_arg(p, char *);
	va_end(p);
	if((flags & XDAEMON_PIDFILE) != 0 && pf == NULL)
		pf = xasprintf("/var/run/%s.pid", bb_applet_name);
	if(pf != NULL && (flags & XDAEMON_NO_WRITE_PID) == 0)
		fp = bb_xfopen(pf, "Ww");
	if((flags & XDAEMON_NOFORK) == 0) {
		if(daemon((flags & XDAEMON_NO_CHDIR) != 0, noclose) < 0) {
			log_syslog = LOG_ERR;   /* may be error after close, put to syslog too */
			bb_perror_msg_and_die("can not stay daemon");
		}
		if(!noclose) {
			log_stderr = 0;
			log_syslog = LOG_ERR;
		}
	}
	if(fp != NULL) {
		fprintf (fp, "%lu\n", (unsigned long)getpid());
		fclose(fp);
	}
	return pf;
}
