/*
 * bb_xfopen implementation for busybox
 *
 * Copyright (C) 2022  Vladimir Oleynik  <dzo@simtreas.ru>
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */

#include "libbb.h"
#include <string.h>

/* xmode : [-][W][E][A]w/r/a/etc fopen mode
 * '-': return stdin/stdout if filename is NULL or "-"
 * 'W': warning if error open file
 * 'E': exit if error open file
 * 'A': both "W" and "E"
 */
FAST_FUNC FILE *bb_xfopen(const char *filename, const char *xmode)
{
	FILE *fp;
	int warn = 0, ee = 0, minus = 0;

	while (1) {
		switch(*xmode++) {
			case '-':
				minus++;
				continue;
			case 'E':
				ee = 1;
				continue;
			case 'A':
				ee = 1;
				/* FALLTHROUGH */
			case 'W':
				warn = 1;
				continue;
			default:
				xmode--;
				break;
		}
		break;
	}

	if (minus && (filename == bb_msg_standard_input || filename == NULL ||
			(filename[0] == '-' && filename[1] == '\0'))) {
		return *xmode == 'r' ? stdin : stdout;
	}
	fp = fopen(filename, xmode);
	if (fp == NULL) {
		if (warn)
			bb_perror_msg("%s", filename);
		if (ee)
			exit(bb_default_error_retval);
	}
	return fp;
}
