/*
 * addgroup - add users to /etc/passwd and /etc/shadow
 *
 * Copyright (C) 1999 by Lineo, inc. and John Beppu
 * Copyright (C) 1999,2000,2001 by John Beppu <beppu@codepoet.org>
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */

#ifdef APPLET_USAGE
#define addgroup_trivial_usage \
	"[-g GID] group_name [user_name]"
#define addgroup_full_usage \
	"Adds a group to the system\n\n" \
	"Options:\n" \
	"\t-g GID\t\tspecify gid"
#else

#include "busybox.h"
#include <string.h>
#include <unistd.h>
#include "bb_shadow.h"

#if defined(__BIONIC__)
#error "BIONIC have not support add user/group. Reconfigure for disable the 'addgroup' applet."
#endif

#if !defined(__MUSL__) || defined(CONFIG_USE_BB_PWD_GRP)
CALC_BSS struct BB_COMMON_BSS {
	char buffer[512];
} bb_common_bss;
#endif


/* defaults : should this be in an external file? */
static const char default_passwd[] = "x";


/* make sure gr_name isn't taken, make sure gid is kosher
 * return 1 on failure */
static int group_study(struct group *g)
{
	FILE *etc_group;
	gid_t desired;
#if !defined(__MUSL__) || defined(CONFIG_USE_BB_PWD_GRP)
	struct group resultbuf;
#endif
	struct group *grp;
	const int max = 65000;

	etc_group = bb_xfopen(bb_path_group_file, "Ar");

	/* make sure gr_name isn't taken, make sure gid is kosher */
	desired = g->gr_gid;
	while (1) {
#if defined(__MUSL__) && !defined(CONFIG_USE_BB_PWD_GRP) /* musl have not fgetgrent_r */
		grp = fgetgrent(etc_group);
#else
		fgetgrent_r(etc_group, &resultbuf, bb_common_bss.buffer, sizeof(bb_common_bss.buffer), &grp);
#endif
		if(grp == NULL)
			break;
		if ((strcmp(grp->gr_name, g->gr_name)) == 0)
			bb_error_msg_and_die("%s: group already in use", g->gr_name);
		if ((desired) && grp->gr_gid == desired)
			bb_error_msg_and_die("%d: gid has already been allocated", desired);
		if ((grp->gr_gid > g->gr_gid) && (grp->gr_gid < max))
			g->gr_gid = grp->gr_gid;
	}
	fclose(etc_group);

	/* gid */
	if (desired) {
		g->gr_gid = desired;
	} else {
		g->gr_gid++;
	}
	/* return 1; */
	return 0;
}

/* append a new user to the passwd file */
static int addgroup(char *group, gid_t gid, const char *user)
{
	FILE *etc_group;

#ifdef CONFIG_FEATURE_SHADOWPASSWDS
	FILE *etc_gshadow;
#endif

	struct group gr;

	/* group:passwd:gid:userlist */
	static const char entryfmt[] = "%s:%s:%d:%s\n";

	/* make sure gid and group haven't already been allocated */
	gr.gr_gid = gid;
	gr.gr_name = group;
	if (group_study(&gr))
		return 1;

	/* add entry to group */
	etc_group = bb_xfopen(bb_path_group_file, "Aa");

	fprintf(etc_group, entryfmt, group, default_passwd, gr.gr_gid, user);
	fclose(etc_group);


#ifdef CONFIG_FEATURE_SHADOWPASSWDS
	/* add entry to gshadow if necessary */
	if (access(bb_path_gshadow_file, F_OK|W_OK) == 0) {
		etc_gshadow = bb_xfopen(bb_path_gshadow_file, "Aa");
		fprintf(etc_gshadow, "%s:!::\n", group);
		fclose(etc_gshadow);
	}
#endif

	/* return 1; */
	return 0;
}

#ifndef CONFIG_ADDUSER
static inline void if_i_am_not_root(void)
{
	if (geteuid()) {
		bb_error_msg_and_die( "Only root may add a user or group to the system.");
	}
}
#else
extern void if_i_am_not_root(void);
#endif

/*
 * addgroup will take a login_name as its first parameter.
 *
 * gid
 *
 * can be customized via command-line parameters.
 * ________________________________________________________________________ */
int addgroup_main(void)
{
	char **argv;
	char *group;
	char *user;
	gid_t gid = 0;

	/* get remaining args */
	bb_xgetopt_ulflags("g:\0:g:-1", &gid);
	argv = Gargv;
	group = *argv;
	argv++;

	if (*argv != NULL) {
		user = *argv;
	} else {
		user = "";
	}

	if_i_am_not_root();

	/* werk */
	return addgroup(group, gid, user);
}
#endif /* APPLET_USAGE */
