#ifdef APPLET_USAGE
#define sulogin_trivial_usage \
	"[-t timeout] [tty-device]"
#define sulogin_full_usage \
	"Single user login\n" \
	"Options:\n" \
	"\t-t timeout\tfor ask password"
#else

#include "busybox.h"
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <utmp.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <ctype.h>
#include <time.h>
#include <fcntl.h>
#include "bb_shadow.h"
#include "bb_sig.h"


#define SULOGIN_PROMPT "\nGive root password for system maintenance\n" \
	"(or type Control-D for normal startup): "

static const char * const forbid[] = {
	"ENV",
	"BASH_ENV",
	"HOME",
	"IFS",
	"PATH",
	"SHELL",
	"LD_LIBRARY_PATH",
	"LD_PRELOAD",
	"LD_TRACE_LOADED_OBJECTS",
	"LD_BIND_NOW",
	"LD_AOUT_LIBRARY_PATH",
	"LD_AOUT_PRELOAD",
	"LD_NOWARN",
	"LD_KEEPDIR",
	(char *) 0
};



static void catchalarm(int ATTRIBUTE_UNUSED junk)
{
	exit(EXIT_FAILURE);
}


#define root_name "root"

extern int sulogin_main(void)
{
	char *cp;
	int timeout = 0;

	struct passwd pwent;
	struct passwd *pwd;
	const char * const *p;
#if ENABLE_FEATURE_SHADOWPASSWDS
	struct spwd *spwd = NULL;
#endif

	bb_openlog(LOG_PID | LOG_CONS | LOG_NOWAIT, LOG_AUTH);
	bb_xgetopt_ulflags("t:\0:-t", &timeout);
	if(*Gargv != NULL) {
		close(0);
		close(1);
		close(2);
		if (open(*Gargv, O_RDWR) >= 0) {
			dup(0);
			dup(0);
		} else {
			log_syslog = 0;
			log_syslog = LOG_WARNING;
			bb_perror_msg_and_die("cannot open %s", *Gargv);
		}
	}

	if (access(bb_path_passwd_file, 0) == -1) {
		log_syslog = log_stderr = LOG_WARNING;
		bb_error_msg_and_die("No password file");
	}
	if (!isatty(0) || !isatty(1) || !isatty(2))
		return EXIT_FAILURE;

	/* Clear out anything dangerous from the environment */
	for (p = forbid; *p; p++)
		unsetenv(*p);

	bb_signal(1 | (1 << SIGALRM), catchalarm);
	if (!(pwd = getpwnam(root_name))) {
		log_syslog = log_stderr = LOG_WARNING;
		bb_error_msg_and_die("No password entry for `root'");
	}
	pwent = *pwd;
#if ENABLE_FEATURE_SHADOWPASSWDS
	spwd = NULL;
	if (pwd && ((strcmp(pwd->pw_passwd, "x") == 0)
				|| (strcmp(pwd->pw_passwd, "*") == 0))) {
		endspent();
		spwd = getspnam(root_name);
		if (spwd) {
			pwent.pw_passwd = spwd->sp_pwdp;
		}
	}
#endif
	while (1) {
		char enc[ENCRYPT_BUFFER_SIZE];
		char pass[PASSWORD_BUFFER_SIZE];

		cp = bb_askpass(timeout, SULOGIN_PROMPT, pass);
		if (!cp || !*cp) {
			puts("\n");
			log_syslog = LOG_INFO;
			log_stderr = 0;
			bb_default_error_retval = EXIT_SUCCESS;
			bb_error_msg_and_die("Normal startup");
		}
		if (strcmp(pw_encrypt(pass, pwent.pw_passwd, enc), pwent.pw_passwd) == 0)
			break;
		bb_do_delay(FAIL_DELAY);
		puts("Login incorrect");
		fflush(stdout);
		xlog_warn("Incorrect root password");
	}
	bb_signal((1 << SIGALRM), SIG_DFL);
	puts("Entering System Maintenance Mode\n");
	fflush(stdout);
	syslog(LOG_INFO, "System Maintenance Mode");

#if ENABLE_SELINUX
	renew_current_security_context();
#endif

	run_shell(pwent.pw_shell, 0, 0);

	return 0;
}
#endif /* APPLET_USAGE */
