/*
 * vlock implementation for busybox
 *
 * Copyright (C) 2000 by spoon <spoon@ix.netcom.com>
 * Written by spoon <spon@ix.netcom.com>
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */

/* Shoutz to Michael K. Johnson <johnsonm@redhat.com>, author of the
 * original vlock.  I snagged a bunch of his code to write this
 * minimalistic vlock.
 */
/* Fixed by Erik Andersen to do passwords the tinylogin way...
 * It now works with md5, sha1, etc passwords. */
#ifdef APPLET_USAGE
#define vlock_trivial_usage \
	"[OPTIONS]"
#define vlock_full_usage \
	"Lock a virtual terminal. A password is required to unlock\n" \
	"Options:\n" \
	"\t-a\tLock all VTs"
#else

#include "busybox.h"
#include <sys/vt.h>
#include <sys/ioctl.h>
#include <string.h>
#include <unistd.h>
#include <termios.h>

#include "bb_xopen.h"
#include "bb_sig.h"

CALC_BSS struct BB_COMMON_BSS
{
	struct passwd *pw;
	struct vt_mode ovtm;
	struct termios oterm;
	int vfd;
} bb_common_bss;
#define o_lock_all bss_flags
#define pw bb_common_bss.pw
#define ovtm bb_common_bss.ovtm
#define oterm bb_common_bss.oterm
#define vfd bb_common_bss.vfd


static void release_vt(int signo)
{
	if (!o_lock_all)
		ioctl(vfd, VT_RELDISP, 1);
	else
		ioctl(vfd, VT_RELDISP, 0);
}

static void acquire_vt(int signo)
{
	ioctl(vfd, VT_RELDISP, VT_ACKACQ);
}

static void restore_terminal(void)
{
	ioctl(vfd, VT_SETMODE, &ovtm);
	tcsetattr(STDIN_FILENO, TCSANOW, &oterm);
}

extern int vlock_main(void)
{
	struct vt_mode vtm;
	struct termios term;

	bb_xgetopt_ulflags("a\0?1");

	if((pw = getpwuid(getuid())) == NULL)
		bb_error_msg_and_die("Unknown uid %d", getuid());

	vfd = bb_xopen(bb_dev_tty, O_RDWR | BB_OPEN_ERR);

	if (ioctl(vfd, VT_GETMODE, &vtm) < 0)
		bb_perror_msg_and_die("VT_GETMODE");

	bb_signal((1 << SIGUSR1), release_vt);
	bb_signal((1 << SIGUSR2), acquire_vt);

	/* need to handle some signals so that we don't get killed by them */
	bb_signal( (1 << SIGHUP) | (1 << SIGQUIT)
			| (1 << SIGINT) | (1 << SIGTSTP), SIG_IGN);

	ovtm = vtm;
	vtm.mode = VT_PROCESS;
	vtm.relsig = SIGUSR1;
	vtm.acqsig = SIGUSR2;
	ioctl(vfd, VT_SETMODE, &vtm);

	tcgetattr(STDIN_FILENO, &oterm);
	term = oterm;
	term.c_iflag &= ~BRKINT;
	term.c_iflag |= IGNBRK;
	term.c_lflag &= ~ISIG;
	term.c_lflag &= ~(ECHO | ECHOCTL);
	tcsetattr(STDIN_FILENO, TCSANOW, &term);

	do {
		printf("Virtual Console%s locked.\n%s's ",
				o_lock_all ? "s" : "", pw->pw_name);
		fflush(stdout);
		if (correct_password (pw))
			break;
		bb_do_delay(FAIL_DELAY);
		puts("Password incorrect.");
	} while (1);
	restore_terminal();
	return 0;
}
#endif /* APPLET_USAGE */
