#ifdef APPLET_USAGE
#define mt_trivial_usage \
	"[-f device] opcode value"
#define mt_full_usage \
	"Control magnetic tape drive operation\n" \
	"\nAvailable Opcodes:\n\n" \
	"bsf bsfm bsr bss datacompression drvbuffer eof eom erase\n" \
	"fsf fsfm fsr fss load lock mkpart nop offline ras1 ras2\n" \
	"ras3 reset retension rewind rewoffline seek setblk setdensity\n" \
	"setpart tell unload unlock weof wset"
#else

#include "busybox.h"
#include <string.h>
#include "bb_xopen.h"
#if defined(__BIONIC__)
#error "BIONIC have not support MTIO stuff. Reconfigure for disable the 'mt' applet."
#endif
#include <sys/mtio.h>

struct mt_opcodes {
	const char *name;
	short value;
};

/* missing: eod/seod, stoptions, stwrthreshold, densities */
static const struct mt_opcodes opcodes[] = {
	{"bsf", MTBSF},
	{"bsfm", MTBSFM},
	{"bsr", MTBSR},
	{"bss", MTBSS},
	{"datacompression", MTCOMPRESSION},
	{"eom", MTEOM},
	{"erase", MTERASE},
	{"fsf", MTFSF},
	{"fsfm", MTFSFM},
	{"fsr", MTFSR},
	{"fss", MTFSS},
	{"load", MTLOAD},
	{"lock", MTLOCK},
	{"mkpart", MTMKPART},
	{"nop", MTNOP},
	{"offline", MTOFFL},
	{"rewoffline", MTOFFL},
	{"ras1", MTRAS1},
	{"ras2", MTRAS2},
	{"ras3", MTRAS3},
	{"reset", MTRESET},
	{"retension", MTRETEN},
	{"rewind", MTREW},
	{"seek", MTSEEK},
	{"setblk", MTSETBLK},
	{"setdensity", MTSETDENSITY},
	{"drvbuffer", MTSETDRVBUFFER},
	{"setpart", MTSETPART},
	{"tell", MTTELL},
	{"wset", MTWSM},
	{"unload", MTUNLOAD},
	{"unlock", MTUNLOCK},
	{"eof", MTWEOF},
	{"weof", MTWEOF},
	{0, 0}
};

extern int mt_main(void)
{
	char **argv;
	const char *file = "/dev/tape";
	const struct mt_opcodes *code = opcodes;
	struct mtop op;
	struct mtpos position;
	int fd, mode;

	bb_xgetopt_ulflags("f:\0-1?2", &file);
	argv = Gargv;

	while (code->name != 0) {
		if (strcmp(code->name, *argv) == 0)
			break;
		code++;
	}

	if (code->name == 0)
		bb_error_msg_and_die("unrecognized opcode %s.", *argv);

	op.mt_op = code->value;
	if (argv[1] != NULL)
		op.mt_count = atoi(argv[1]);
	else
		op.mt_count = 1;                /* One, not zero, right? */

	switch (code->value) {
		case MTWEOF:
		case MTERASE:
		case MTWSM:
		case MTSETDRVBUFFER:
			mode = O_WRONLY;
			break;

		default:
			mode = O_RDONLY;
			break;
	}

	fd = bb_xopen(file, mode | BB_OPEN_ERR);

	switch (code->value) {
		case MTTELL:
			xioctl(fd, MTIOCPOS, &position, "\1%s", file);
			printf ("At block %d.\n", (int) position.mt_blkno);
			break;

		default:
			xioctl(fd, MTIOCTOP, &op, "\1%s", file);
			break;
	}

	return EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
