/*
 * setsid.c -- execute a command in a new session
 * Rick Sladkey <jrs@world.std.com>
 * In the public domain.
 *
 * 1999-02-22 Arkadiusz Mikiewicz <misiek@pld.ORG.PL>
 * - added Native Language Support
 *
 * 2001-01-18 John Fremlin <vii@penguinpowered.com>
 * - fork in case we are process group leader
 *
 * 2004-11-12 Paul Fox
 * - busyboxed
 */
#ifdef APPLET_USAGE
#define setsid_trivial_usage \
	"program [arg ...]"
#define setsid_full_usage \
	"Runs any program in a new session by calling setsid() before\n" \
	"exec'ing the rest of its arguments. See setsid(2) for details."
#else

#include <unistd.h>
#include "busybox.h"

int setsid_main(void)
{
	char **argv = Gargv + 1;

	if (Gargc < 2)
		bb_show_usage();

	if (getpgrp() == getpid()) {
		switch(fork()){
		case -1:
			bb_perror_msg_and_die("fork");
		case 0:
			break;
		default:        /* parent */
			exit(0);
		}
		/* child falls through */
	}

	setsid();  /* no error possible */

	execvp(*argv, argv);

	bb_perror_msg_and_die("%s", *argv);
}
#endif /* APPLET_USAGE */
