/*
 * Mini watchdog implementation for busybox
 *
 * Copyright (C) 2003  Paul Mundt <lethal@linux-sh.org>
 *
 * Licensed under the GPL v2 or later, see the file LICENSE in this tarball.
 */
#ifdef APPLET_USAGE
#define watchdog_trivial_usage \
	"[-t <seconds>] DEV"
#define watchdog_full_usage \
	"Periodically write to watchdog device DEV.\n" \
	"Options:\n" \
	"\t-t\tTimer period in seconds - default is 30"
#else

#include "busybox.h"
#include <unistd.h>
#include "bb_xopen.h"
#include "bb_sig.h"

CALC_BSS struct BB_COMMON_BSS {
	/* Watchdog file descriptor */
	int fd;
} bb_common_bss;
#define fd bb_common_bss.fd

static void watchdog_shutdown(int ATTRIBUTE_UNUSED unused)
{
	write(fd, "V", 1);      /* Magic, see watchdog-api.txt in kernel */
//        close(fd);
	exit(0);
}

int watchdog_main(void)
{
	/* Userspace timer duration, in seconds */
	long td = 30;
	unsigned int timer_duration;

	/* We're only interested in the watchdog device .. */
	bb_xgetopt_ulflags("t:\0:t:-1", &td);
	timer_duration = td;

	xdaemon(XDAEMON_PIDFILE);

	bb_signal( (1 << SIGINT) | (1 << SIGHUP), watchdog_shutdown);

	fd = bb_xopen(*Gargv, O_WRONLY | BB_OPEN_ERR | BB_OPEN_CREAT_0777);

	while (1) {
		/*
		 * Make sure we clear the counter before sleeping, as the counter value
		 * is undefined at this point -- PFM
		 */
		write(fd, "\0", 1);
		sleep(timer_duration);
	}

	watchdog_shutdown(0);

	return EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
