/*
 * Mini lsmod implementation for busybox
 *
 * Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 *
 * Modified by Alcove, Julien Gaulmin <julien.gaulmin@alcove.fr> and
 * Nicolas Ferre <nicolas.ferre@alcove.fr> to support pre 2.1 kernels
 * (which lack the query_module() interface).
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */
#ifdef APPLET_USAGE
#define lsmod_trivial_usage \
	""
#define lsmod_full_usage \
	"List the currently loaded kernel modules."
#else

#include "busybox.h"
#include <string.h>
#include <stddef.h>
#include <unistd.h>
#include <dirent.h>
#include <ctype.h>
#include <assert.h>
#include <sys/file.h>


#ifndef CONFIG_FEATURE_CHECK_TAINTED_MODULE
static inline void check_tainted(void) { printf("\n"); }
#else
#define TAINT_FILENAME                  "/proc/sys/kernel/tainted"
#define TAINT_PROPRIETORY_MODULE        (1<<0)
#define TAINT_FORCED_MODULE             (1<<1)
#define TAINT_UNSAFE_SMP                (1<<2)

static void check_tainted(void)
{
	int tainted;
	FILE *f;

	tainted = 0;
	if ((f = fopen(TAINT_FILENAME, "r"))) {
		fscanf(f, "%d", &tainted);
		fclose(f);
	}
	if (f && tainted) {
		printf("    Tainted: %c%c%c\n",
				tainted & TAINT_PROPRIETORY_MODULE      ? 'P' : 'G',
				tainted & TAINT_FORCED_MODULE           ? 'F' : ' ',
				tainted & TAINT_UNSAFE_SMP              ? 'S' : ' ');
	}
	else {
		printf("    Not tainted\n");
	}
}
#endif

extern int lsmod_main(void)
{
	FILE *file;
	char line[4096];

	printf("Module                  Size  Used by");
	check_tainted();

	file = bb_xfopen("/proc/modules", "Ar");

	while (fgets(line, sizeof(line), file)) {
	  char *tok;

	  tok = strtok(line, bb_spacestok);
	  printf("%-19s", tok);
	  tok = strtok(NULL, " \t\n");
	  printf(" %8s", tok);
	  tok = strtok(NULL, " \t\n");
	  /* Null if no module unloading support. */
	  if (tok) {
	    printf("  %s", tok);
	    tok = strtok(NULL, "\n");
	    if (!tok)
	      tok = "";
	    /* New-style has commas, or -.  If so,
	       truncate (other fields might follow). */
	    else if (strchr(tok, ',')) {
	      tok = strtok(tok, bb_spacestok);
	      /* Strip trailing comma. */
	      if (tok[strlen(tok)-1] == ',')
		tok[strlen(tok)-1] = '\0';
	    } else if (tok[0] == '-'
		       && (tok[1] == '\0' || isspace(tok[1])))
	      tok = "";
	    printf(" %s", tok);
	  }
	  printf("\n");
	}
	fclose(file);

	return 0;  /* Success  */
}
#endif /* APPLET_USAGE */
