/* vi: set sw=4 ts=4: */
/*
 * Mini rmmod implementation for busybox
 *
 * Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 * Copyright (C) 2008 Timo Teras <timo.teras@iki.fi>
 *
 * Licensed under GPLv2 or later, see file LICENSE in this source tree.
 */
#ifdef APPLET_USAGE
#define rmmod_trivial_usage \
	"[-wfa] [MODULE]..."
#define rmmod_full_usage \
	"Unloads the specified kernel modules from the kernel.\n\n" \
	"Options:\n" \
	"\t-a\tRemove all unused modules (recursively)\n" \
	"\t-w\tWait until the module is no longer used\n" \
	"\t-f\tForce unload"
#else

#include "busybox.h"
#include <fcntl.h>
#include "modutils.h"

int rmmod_main(void)
{
	int n;
	unsigned flags = O_NONBLOCK | O_EXCL;
	char **argv;

	/* Parse command line. */
	n = bb_getopt_ulflags("wfas"); // -s ignored
	argv = Gargv;
	if (n & 1)  // --wait
		flags &= ~O_NONBLOCK;
	if (n & 2)  // --force
		flags |= O_TRUNC;
	if (n & 4) {
		/* Unload _all_ unused modules via NULL delete_module() call */
		if (bb_delete_module(NULL, flags) != 0 && errno != EFAULT)
			bb_perror_msg_and_die("rmmod");
		return EXIT_SUCCESS;
	}

	if (*argv == NULL)
		bb_show_usage();

	while (*argv != NULL) {
		char modname[MODULE_NAME_LEN];
		const char *bname;

		bname = bb_basename(*argv++);
		filename2modname(bname, modname);
		if (bb_delete_module(modname, flags))
			bb_error_msg_and_die("can't unload '%s': %s",
					modname, moderror(errno));
	}

	return EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
