#define XT_FUNCTION_MAXNAMELEN 30
#define IPT_FUNCTION_MAXNAMELEN XT_FUNCTION_MAXNAMELEN

/* only for userspace compatibility */
/* IP Cache bits. */
/* Src IP address. */
#define NFC_IP_SRC              0x0001
/* Dest IP address. */
#define NFC_IP_DST              0x0002
/* Input device. */
#define NFC_IP_IF_IN            0x0004
/* Output device. */
#define NFC_IP_IF_OUT           0x0008
/* TOS. */
#define NFC_IP_TOS              0x0010
/* Protocol. */
#define NFC_IP_PROTO            0x0020
/* IP options. */
#define NFC_IP_OPTIONS          0x0040
/* Frag & flags. */
#define NFC_IP_FRAG             0x0080

/* Per-protocol information: only matters if proto match. */
/* TCP flags. */
#define NFC_IP_TCPFLAGS         0x0100
/* Source port. */
#define NFC_IP_SRC_PT           0x0200
/* Dest port. */
#define NFC_IP_DST_PT           0x0400
/* Something else about the proto */
#define NFC_IP_PROTO_UNKNOWN    0x2000

/* Generic cache responses from hook functions.
   <= 0x2000 is used for protocol-flags. */
#define NFC_UNKNOWN 0x4000
#define NFC_ALTERED 0x8000


#define ipt_udp xt_udp
#define ipt_tcp xt_tcp
#define ipt_entry_target xt_entry_target
#define ipt_entry_match xt_entry_match

/* TCP matching stuff */
struct xt_tcp {
	__u16 spts[2];                  /* Source port range. */
	__u16 dpts[2];                  /* Destination port range. */
	__u8 option;                    /* TCP Option iff non-zero*/
	__u8 flg_mask;                  /* TCP flags mask byte */
	__u8 flg_cmp;                   /* TCP flags compare byte */
	__u8 invflags;                  /* Inverse flags */
};

/* UDP matching stuff */
struct xt_udp {
	__u16 spts[2];                  /* Source port range. */
	__u16 dpts[2];                  /* Destination port range. */
	__u8 invflags;                  /* Inverse flags */
};

/* Yes, Virginia, you have to zero the padding. */
struct ipt_ip {
	/* Source and destination IP addr */
	struct in_addr src, dst;
	/* Mask for src and dest IP addr */
	struct in_addr smsk, dmsk;
	char iniface[IFNAMSIZ], outiface[IFNAMSIZ];
	unsigned char iniface_mask[IFNAMSIZ], outiface_mask[IFNAMSIZ];

	/* Protocol, 0 = ANY */
	u_int16_t proto;

	/* Flags word */
	u_int8_t flags;
	/* Inverse flags */
	u_int8_t invflags;
};

struct xt_counters {
	__u64 pcnt, bcnt;                       /* Packet and byte counters */
};

/* This structure defines each of the firewall rules.  Consists of 3
   parts which are 1) general IP header stuff 2) match specific
   stuff 3) the target to perform if the rule matches */
struct ipt_entry {
	struct ipt_ip ip;

	/* Mark with fields that we care about. */
	unsigned int nfcache;

	/* Size of ipt_entry + matches */
	u_int16_t target_offset;
	/* Size of ipt_entry + matches + target */
	u_int16_t next_offset;

	/* Back pointer */
	unsigned int comefrom;

	/* Packet and byte counters. */
	struct xt_counters counters;

	/* The matches (if any), then the target. */
	unsigned char elems[0];
};

#define XT_EXTENSION_MAXNAMELEN 29

struct xt_entry_target {
	union {
		struct {
			__u16 target_size;

			/* Used by userspace */
			char name[XT_EXTENSION_MAXNAMELEN];
			__u8 revision;
		} user;
		struct {
			__u16 target_size;

			/* Used inside the kernel */
			struct xt_target *target;
		} kernel;

		/* Total length */
		__u16 target_size;
	} u;

	unsigned char data[0];
};

struct xt_entry_match {
	union {
		struct {
			__u16 match_size;

			/* Used by userspace */
			char name[XT_EXTENSION_MAXNAMELEN];
			__u8 revision;
		} user;
		struct {
			__u16 match_size;

			/* Used inside the kernel */
			struct xt_match *match;
		} kernel;

		/* Total length */
		__u16 match_size;
	} u;

	unsigned char data[0];
};

/* this is a dummy structure to find out the alignment requirement for a struct
 * containing all the fundamental data types that are used in ipt_entry,
 * ip6t_entry and arpt_entry.  This sucks, and it is a hack.  It will be my
 * personal pleasure to remove it -HW
 */
struct _xt_align {
	__u8 u8;
	__u16 u16;
	__u32 u32;
	__u64 u64;
};

#define __ALIGN_KERNEL_MASK(x, mask)    (((x) + (mask)) & ~(mask))
#define __ALIGN_KERNEL(x, a)            __ALIGN_KERNEL_MASK(x, (typeof(x))(a) - 1)
#define XT_ALIGN(s) __ALIGN_KERNEL((s), __alignof__(struct _xt_align))

typedef char xt_chainlabel[32];

FOR_MINIUPNPD struct xtc_handle *iptc_init(const char *tablename);
FOR_MINIUPNPD const char * iptc_strerror(int err);
FOR_MINIUPNPD void iptc_free(struct xtc_handle *h);
/* Does this chain exist? */
FOR_MINIUPNPD int iptc_is_chain(const char *chain, struct xtc_handle *const handle);
FOR_MINIUPNPD const struct ipt_entry *iptc_first_rule(const char *chain, struct xtc_handle *handle);
FOR_MINIUPNPD const struct ipt_entry *iptc_next_rule(const struct ipt_entry *prev, struct xtc_handle *handle);
FOR_MINIUPNPD int iptc_delete_num_entry(const xt_chainlabel chain, unsigned int rulenum, struct xtc_handle *handle);
FOR_MINIUPNPD int iptc_commit(struct xtc_handle *handle);
FOR_MINIUPNPD int iptc_append_entry(const xt_chainlabel chain, const struct ipt_entry *e, struct xtc_handle *handle);
FOR_MINIUPNPD int iptc_replace_entry(const xt_chainlabel chain, const struct ipt_entry *e, unsigned int rulenum, struct xtc_handle *handle);

#if ENABLE_FEATURE_IPV6
/* Yes, Virginia, you have to zero the padding. */
struct ip6t_ip6 {
	/* Source and destination IP6 addr */
	struct in6_addr src, dst;
	/* Mask for src and dest IP6 addr */
	struct in6_addr smsk, dmsk;
	char iniface[IFNAMSIZ], outiface[IFNAMSIZ];
	unsigned char iniface_mask[IFNAMSIZ], outiface_mask[IFNAMSIZ];

	/* Upper protocol number
	 * - The allowed value is 0 (any) or protocol number of last parsable
	 *   header, which is 50 (ESP), 59 (No Next Header), 135 (MH), or
	 *   the non IPv6 extension headers.
	 * - The protocol numbers of IPv6 extension headers except of ESP and
	 *   MH do not match any packets.
	 * - You also need to set IP6T_FLAGS_PROTO to "flags" to check protocol.
	 */
	u_int16_t proto;
	/* TOS to match iff flags & IP6T_F_TOS */
	u_int8_t tos;

	/* Flags word */
	u_int8_t flags;
	/* Inverse flags */
	u_int8_t invflags;
};

/* This structure defines each of the firewall rules.  Consists of 3
   parts which are 1) general IP header stuff 2) match specific
   stuff 3) the target to perform if the rule matches */
struct ip6t_entry {
	struct ip6t_ip6 ipv6;

	/* Mark with fields that we care about. */
	unsigned int nfcache;

	/* Size of ipt_entry + matches */
	u_int16_t target_offset;
	/* Size of ipt_entry + matches + target */
	u_int16_t next_offset;

	/* Back pointer */
	unsigned int comefrom;

	/* Packet and byte counters. */
	struct xt_counters counters;

	/* The matches (if any), then the target. */
	unsigned char elems[0];
};

FOR_MINIUPNPD struct xtc_handle6 *ip6tc_init(const char *tablename);
FOR_MINIUPNPD const char *ip6tc_strerror(int err);
FOR_MINIUPNPD int ip6tc_is_chain(const char *chain, struct xtc_handle6 *const handle);
FOR_MINIUPNPD const struct ip6t_entry *ip6tc_first_rule(const char *chain, struct xtc_handle6 *handle);
FOR_MINIUPNPD const struct ip6t_entry *ip6tc_next_rule(const struct ip6t_entry *prev, struct xtc_handle6 *handle);
FOR_MINIUPNPD int ip6tc_delete_num_entry(const xt_chainlabel chain, unsigned int rulenum, struct xtc_handle6 *handle);
FOR_MINIUPNPD int ip6tc_commit(struct xtc_handle6 *handle);
FOR_MINIUPNPD int ip6tc_append_entry(const xt_chainlabel chain, const struct ip6t_entry *e, struct xtc_handle6 *handle);
FOR_MINIUPNPD void ip6tc_free(struct xtc_handle6 *h);
#endif
