/*
 * Copyright (C) 1998 Brad M. Garcia <garsh@home.com>
 *
 * query.c:
 * This is a complete rewrite of Brad garcias query.c
 * This file contains the data definitions, function definitions, and
 * variables used to implement our DNS query list.
 * Assumptions: No multithreading.
 * Copyright (C) Natanael Copa <ncopa@users.sourceforge.net>
 *
 * tcprequest.c - handle TCP request by transparent proxying.
 * Copyright (C) 1999 Wolfgang Zekoll <wzk@quietsche-entchen.de>
 *
 * Copyright (C) 2017 Vladimir N. Oleynik <dzo@simtreas.ru>
 * busybox port, size reduction, many signed/unsigned/int/long corrections
 * -p port, usage for nonroot user, reduce data segment to zero
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#ifdef APPLET_USAGE
#define dnrd_trivial_usage \
	"[-a addr] [-p port] [-bfhikv] [-M max-sock] [-s IPADDR[:domain]] [-R root] [-u uid]"
#define dnrd_full_usage \
	"Cache and forward DNS server\n\n" \
	"Options:\n" \
	USE_FEATURE_LONG_OPTIONS("\t-a, --address=LOCALADDRESS\n") \
	USE_FEATURE_LONG_OPTIONS("\t\t\t\tOnly bind to the port on the given address,\n") \
	USE_FEATURE_LONG_OPTIONS("\t\t\t\trather than all local addresses.\n") \
	USE_FEATURE_LONG_OPTIONS("\t-p, --port=port\t\tSet port #, default 53.\n\n") \
	USE_FEATURE_LONG_OPTIONS("\t-b, --load-balance\tRound-Robin load balance forwarding servers.\n") \
	USE_FEATURE_LONG_OPTIONS("\t-f, --foreground\tRun program in foreground rather than fork.\n") \
	USE_FEATURE_LONG_OPTIONS("\t-h, --help\t\tPrint this message, then exit.\n") \
	USE_FEATURE_LONG_OPTIONS("\t-i, --ignore\t\tIgnore cache for disabled servers.\n") \
	USE_FEATURE_LONG_OPTIONS("\t-k, --kill\t\tKill a running daemon.\n") \
	USE_FEATURE_LONG_OPTIONS("\t-M, --max-sock=N\tSet maximum number of open sockets to N.\n") \
	USE_FEATURE_LONG_OPTIONS("\t-s, --server=IPADDR[:domain]\n") \
	USE_FEATURE_LONG_OPTIONS("\t\t\t\tSet the DNS server.  You can specify an\n") \
	USE_FEATURE_LONG_OPTIONS("\t\t\t\toptional domain name, in which case a DNS\n") \
	USE_FEATURE_LONG_OPTIONS("\t\t\t\trequest will only be sent to that server for\n") \
	USE_FEATURE_LONG_OPTIONS("\t\t\t\tnames in that domain.\n") \
	USE_FEATURE_LONG_OPTIONS("\t\t\t\t(Can be used more than once for multiple or\n") \
	USE_FEATURE_LONG_OPTIONS("\t\t\t\tbackup servers).\n") \
	USE_FEATURE_LONG_OPTIONS("\t-u, --uid=UID\t\tUsername or numeric id to switch to.\n") \
	USE_FEATURE_LONG_OPTIONS("\t-R, --dnrd-root=DIR\tThe dnrd root directory. dnrd will chroot to\n") \
	USE_FEATURE_LONG_OPTIONS("\t\t\t\tthis dir.\n") \
	USE_FEATURE_LONG_OPTIONS("\t-v, --version\t\tPrint out the version number and exit.") \
	SKIP_FEATURE_LONG_OPTIONS("\t-a IPADDR Only bind to the port on the given address, rather than all\n") \
	SKIP_FEATURE_LONG_OPTIONS("\t\tlocal addresses\n") \
	SKIP_FEATURE_LONG_OPTIONS("\t-p\t\tSet port #, default 53.\n\n") \
	SKIP_FEATURE_LONG_OPTIONS("\t-b\t\tRound-Robin load balance forwarding servers\n") \
	SKIP_FEATURE_LONG_OPTIONS("\t-f\t\tRun program in foreground rather than fork.\n") \
	SKIP_FEATURE_LONG_OPTIONS("\t-h\t\tPrint this message, then exit.\n") \
	SKIP_FEATURE_LONG_OPTIONS("\t-i\t\tIgnore cache for disabled servers\n") \
	SKIP_FEATURE_LONG_OPTIONS("\t-k\t\tKill a running daemon.\n") \
	SKIP_FEATURE_LONG_OPTIONS("\t-M N\tSet maximum number of open sockets to N\n") \
	SKIP_FEATURE_LONG_OPTIONS("\t-s IPADDR(:domain)\n") \
	SKIP_FEATURE_LONG_OPTIONS("\t\t\tSet the DNS server.  You can specify an optional domain name,\n") \
	SKIP_FEATURE_LONG_OPTIONS("\t\t\tin which case a DNS request will only be sent to that server\n") \
	SKIP_FEATURE_LONG_OPTIONS("\t\t\tfor names in that domain. (Can be used more than once for\n") \
	SKIP_FEATURE_LONG_OPTIONS("\t\t\tmultiple or backup servers)\n") \
	SKIP_FEATURE_LONG_OPTIONS("\t-u UID\tUsername or numeric id to switch to\n") \
	SKIP_FEATURE_LONG_OPTIONS("\t-R DIR\tThe dnrd root directory. dnrd will chroot to this dir.\n") \
	SKIP_FEATURE_LONG_OPTIONS("\t-v\t\tPrint out the version number and exit).")
#else

#define COMPLEMENTARY "\0=0"
#include "opt_gen.h"
OPT_GEN(
opt(a, OPT_LOCAL_ADDRESS, arg),
opt(p, OPT_PORT, arg),
opt(b, OPT_LOAD_BALANCE),
opt(f, OPT_FOREGROUND),
opt(h, OPT_HELP),
opt(i, OPT_IGNORE_CACHE),
opt(k, OPT_KILL),
opt(M, OPT_MAX_SOCK, arg),
opt(s, OPT_SERVER, arg),
opt(R, OPT_DNRD_ROOT, arg),
opt(u, OPT_UID, arg),
opt(v, OPT_VERSION)
)
#include "opt_gen.h"

/* Define to the version of this package. */
#define PACKAGE_VERSION "2.20.4"

/* Enable random source ports */
#define RANDOM_SRC

#include "busybox.h"
#include <sys/select.h>
#include <sys/time.h>
#include <stddef.h>
#include <time.h>
#include <unistd.h>
#include <assert.h>
#include <string.h>
#include <syslog.h>
#include <netinet/in.h>
#include <dirent.h>
#include <fcntl.h>
#include <ctype.h>
#include "bb_sig.h"
#include "bb_inet.h"

#if ENABLE_FEATURE_LONG_OPTIONS
static const struct option long_options[] = {
  {"address", 1, 0, 'a'},
  {"port", 1, 0, 'p'},
  {"load-balance", 0, 0, 'b'},
  {"foreground", 0, 0, 'f'},
  {"help", 0, 0, 'h'},
  {"ignore", 0, 0, 'i'},
  {"kill", 0, 0, 'k'},
  {"max-sock", 1, 0, 'M'},
  {"server", 1, 0, 's'},
  {"uid", 1, 0, 'u'},
  {"version", 0, 0, 'v'},
  {"dnrd-root", 1, 0, 'R'},
  {0, 0, 0, 0}
};
#endif

typedef unsigned long int ub4;  /* unsigned 4-byte quantities */
typedef unsigned char ub1;
typedef int word;               /* fastest type available */

#ifndef max
# define max(a,b) (((a)<(b)) ? (b) : (a))
#endif /* max */

typedef struct _srvnode
{
  struct sockaddr_in addr;      /* IP address of server */
  time_t inactive;              /* is this server active? */
  unsigned int send_count;
  int send_time;
  int tcp;
  struct _query *newquery;      /* new opened socket, prepared for a new query */
  struct _srvnode *next;        /* ptr to next server */
} srvnode_t;

typedef struct dnsheader_struct
{
  unsigned short int id;
  unsigned short u;

  short int qdcount;
  short int ancount;
  short int nscount;
  short int arcount;    /* Till here it would fit perfectly to a real
			 * DNS packet if we had big endian. */

  char *here;           /* For packet parsing. */
  unsigned char *packet;/* The actual data packet ... */
  int len;              /* ... with this size in bytes. */

  char *rdata;          /* For packet assembly. */
} dnsheader_t;

typedef struct _cache
{
  unsigned int code;
  char *name;                                   /* Objectname */
  /* Query type and class. */
  unsigned int type;
  unsigned int class;

  int positive;                                 /* Positive or error response? */
  time_t created;
  time_t lastused;
  time_t expires;

  dnsheader_t *p;                               /* The DNS packet with decoded header */

  srvnode_t *server;                    /* the server that gave this answer */
  struct _cache *next, *prev;
} cache_t;

#define CACHE_HIGHWATER         1000
#define CACHE_LOWWATER           800

/* According to RFC 1035 (2.3.4) */
#define UDP_MAXSIZE 512

/* default chroot path. this might be redefined in compile time */
#ifndef DNRD_ROOT
#define DNRD_ROOT "/var/dnrd"
#endif

/* Set the default timeout value for select in seconds */
#ifndef SELECT_TIMEOUT
#define SELECT_TIMEOUT 1
#endif

/* Set the default timeout value for forward DNS. If we get no
 * response from a DNS server within forward_timeout, deactivate the
 * server.  note that if SELECT_TIMEOUT is greater than this, the
 * forward timeout *might* increase to select_timeout. This value
 * should be >= SELECT_TIMEOUT
 */
/* 12 seems to be a good value under heavy load... */
#ifndef FORWARD_TIMEOUT
#define FORWARD_TIMEOUT 12
#endif

/* not used yet */
#ifndef FORWARD_RETRIES
#define FORWARD_RETRIES 5
#endif

/* only check if any server are to be reactivated every
 * REACTIVATE_INTERVAL seconds
 */
#define REACTIVATE_INTERVAL 10

/* Internet DNS is the only class we support. */
#define DNS_CLASS_INET                  1

/* Here's where the packet's first field starts. */
#define PACKET_DATABEGIN                12


/* Macros for manipulating the flags field */
#define MASK_RCODE  0x000f
#define MASK_Z      0x0040
#define MASK_RA     0x0080
#define MASK_RD     0x0100
#define MASK_TC     0x0200
#define MASK_AA     0x0400
#define MASK_OPCODE 0xe800
#define MASK_QR     0x8000

#define GET_RCODE(x)    ((x) & MASK_RCODE)
#define GET_Z(x)        (((x) & MASK_Z) >> 4)
#define GET_RA(x)       (((x) & MASK_RA) >> 7)
#define GET_RD(x)       (((x) & MASK_RD) >> 8)
#define GET_TC(x)       (((x) & MASK_TC) >> 9)
#define GET_AA(x)       (((x) & MASK_AA) >> 10)
#define GET_OPCODE(x)   (((x) & MASK_OPCODE) >> 11)
#define GET_QR(x)       (((x) & MASK_QR) >> 15)

#define SET_RCODE(x, y)     ((x) = ((x) & ~MASK_RCODE) | ((y) & MASK_RCODE))
#define SET_RA(x, y)        ((x) = ((x) & ~MASK_RA) | (((y) << 7) & MASK_RA))
#define SET_RD(x, y)        ((x) = ((x) & ~MASK_RD) | (((y) << 8) & MASK_RD))
#define SET_TC(x, y)        ((x) = ((x) & ~MASK_TC) | (((y) << 9) & MASK_TC))
#define SET_AA(x, y)        ((x) = ((x) & ~MASK_AA) | (((y) << 10) & MASK_AA))
#define SET_OPCODE(x, y)    ((x) = ((x) & ~MASK_OPCODE) | \
			     (((y) << 11) & MASK_OPCODE))
#define SET_QR(x, y)        ((x) = ((x) & ~MASK_QR) | (((y) << 15) & MASK_QR))

#define RR_LABELMAXLEN 63
#define RR_LABELSIZE (RR_LABELMAXLEN + 1)
#define RR_NAMEMAXLEN 255
#define RR_NAMESIZE (RR_NAMEMAXLEN + 1)

struct dnssrv_t
{
  int sock;                     /* for communication with server */
  struct sockaddr_in addr;      /* IP address of server */
  char *domain; /* optional domain to match.  Set to zero for a default server */
};

typedef struct _domnode
{
  char *domain;                 /* the domain */
  srvnode_t *srvlist;           /* linked list of servers */
  srvnode_t *current;
  int roundrobin;               /* load balance the servers */
  int retrydelay;               /* delay before reactivating the servers */
  struct _domnode *next;        /* ptr to next server */
} domnode_t;

typedef struct _rr
{
  unsigned short flags;
  char name[RR_NAMESIZE];

  unsigned int type;
  unsigned int class;

  unsigned int ttl;
  int len;
  char data[RR_NAMESIZE];
} rr_t;

typedef struct _query
{
  int sock;                                             /* the communication socket */
  srvnode_t *srv;                               /* the upstream server */
  domnode_t *domain;                    /* the domain the query belongs to */

  unsigned short my_qid;                /* the local qid */
  unsigned short client_qid;    /* the qid from the client */
  struct sockaddr_in client;    /* */

  /*  int send_count; * number of retries */
  /*  time_t send_time; * time of last sent packet */
  time_t client_time;                   /* last time we got this query from client */
  int client_count;                             /* number of times we got this same request */

  time_t ttl;                           /* time to live for this query */

  struct _query *next;                  /* ptr to next query */
} query_t;

#define RANDSIZL   (4)                  /* I recommend 8 for crypto, 4 for simulations */
#define RANDSIZ    (1<<RANDSIZL)

/* context of random number generator */
struct randctx
{
  ub4 randcnt;
  ub4 randrsl[RANDSIZ];
  ub4 randmem[RANDSIZ];
  ub4 randa;
  ub4 randb;
  ub4 randc;
};
typedef struct randctx randctx;

CALC_BSS struct BB_COMMON_BSS {
	char *dnsbuf_dyn;
	cache_t *cachelist;
	cache_t *lastcache;
	int cache_hits;
	int cache_misses;
	const char *pid_file;
	int isock;
	int tcpsock;
	uid_t daemonuid;
	char *dnrd_user;
	int foreground;  /* 0 if deattached to a terminal */
	/* The path where we chroot. All config files are relative this path */
	char *dnrd_root;
	int port;
	domnode_t *domain_list;
	/* turn this on to skip cache hits from responses of inactive dns servers */
	int ignore_inactive_cache_hits;
	int maxsock;    /* highest socket number */
	/* maximum number of open sockets. If we have this amount of
	   concurrent queries, we start dropping new ones */
	int max_sockets;
	fd_set fdmaster; /* the fd set. query modifies this so we make it global */
	struct sockaddr_in recv_addr; /* init recv_addr in main.c instead of here */
	query_t qlist;                /* the active query list */
	query_t *qlist_tail;
	unsigned long total_timeouts;
	int upstream_sockets;    /* number of upstream sockets */
	int dropping;            /* dropping new packets */
	randctx isaac_ctx;
	time_t last_stats;
	time_t last_q;
	int last_try_ds;
	int last_try_rs;
	time_t lastexpire;
	char udp_msg[UDP_MAXSIZE + 4];
	char reply_msg[UDP_MAXSIZE + 4];
#define QID_POOL_SIZE 65536
	unsigned short int qid_pool[QID_POOL_SIZE];
	int pool_ptr;
} bb_common_bss;
#define dnsbuf_dyn bb_common_bss.dnsbuf_dyn
#define cachelist bb_common_bss.cachelist
#define lastcache bb_common_bss.lastcache
#define cache_hits bb_common_bss.cache_hits
#define cache_misses bb_common_bss.cache_misses
#define pid_file bb_common_bss.pid_file
#define isock bb_common_bss.isock
#define tcpsock bb_common_bss.tcpsock
#define daemonuid bb_common_bss.daemonuid
#define dnrd_user bb_common_bss.dnrd_user
#define foreground bb_common_bss.foreground
#define dnrd_root bb_common_bss.dnrd_root
#define port bb_common_bss.port
#define domain_list bb_common_bss.domain_list
#define ignore_inactive_cache_hits bb_common_bss.ignore_inactive_cache_hits
#define maxsock bb_common_bss.maxsock
#define max_sockets bb_common_bss.max_sockets
#define fdmaster bb_common_bss.fdmaster
#define recv_addr bb_common_bss.recv_addr
#define qlist bb_common_bss.qlist
#define qlist_tail bb_common_bss.qlist_tail
#define total_timeouts bb_common_bss.total_timeouts
#define upstream_sockets bb_common_bss.upstream_sockets
#define dropping bb_common_bss.dropping
#define isaac_ctx bb_common_bss.isaac_ctx
#define last_stats bb_common_bss.last_stats
#define last_q bb_common_bss.last_q
#define last_try_ds bb_common_bss.last_try_ds
#define last_try_rs bb_common_bss.last_try_rs
#define lastexpire bb_common_bss.lastexpire
#define udp_msg bb_common_bss.udp_msg
#define reply_msg bb_common_bss.reply_msg
#define qid_pool bb_common_bss.qid_pool
#define pool_ptr bb_common_bss.pool_ptr

/*
definitions for a random number generator
By Bob Jenkins, 1996, Public Domain
MODIFIED:
  960327: Creation (addition of randinit, really)
  970719: use context, not global variables, for internal state
  980324: renamed seed to flag
  980605: recommend RANDSIZL=4 for noncryptography.
  010626: note this is public domain
*/


/* If (flag==TRUE), then use the contents of randrsl[0..RANDSIZ-1] as the seed. */
static void randinit (randctx *ctx, word flag);
static void isaac (randctx *ctx);


/* Call rand(/o_ randctx *r _o/) to retrieve a single 32-bit random value */
#define rand(r) \
   (!(r)->randcnt-- ? \
     (isaac(r), (r)->randcnt=RANDSIZ-1, (r)->randrsl[(r)->randcnt]) : \
     (r)->randrsl[(r)->randcnt])


static const char *get_typestr(int type)
{
  static const char WARNING[] = "Warning";
  static const char NOTICE[] = "Notice";
  static const char INFO[] = "Info";

  switch (type) {
  case LOG_WARNING:
	return WARNING;
  case LOG_NOTICE:
	return NOTICE;
  default:
	return INFO;
  }
}

/* log_msg()
 *
 * In:      type - a syslog priority
 *          fmt  - a formatting string, ala printf.
 *          ...  - other printf-style arguments.
 *
 * Sends a message to stdout or stderr if attached to a terminal, otherwise
 * it sends a message to syslog. */
static void __attribute__ ((format (printf, 2, 3))) log_msg(int type, const char *fmt, ...)
{
  va_list ap;

  va_start (ap, fmt);

  if (foreground) {
	fprintf(stderr, "%s: %s: ", bb_applet_name, get_typestr (type));
	vfprintf(stderr, fmt, ap);
	fprintf(stderr, "\n");
  } else {
	vsyslog(type, fmt, ap);
  }
  va_end (ap);
}

/* make_cname()
 *
 * In:       text - human readable domain name string
 *
 * Returns:  Pointer to the allocated, filled in character string on success,
 *           NULL on failure.
 *
 * Abstract: converts the human-readable domain name to the DNS CNAME
 *           form, where each node has a length byte followed by the
 *           text characters, and ends in a null byte.  The space for
 *           this new representation is allocated by this function. */
static char *make_cname(const char *text, const int maxlen)
{
  /* this kind of code can easily contain buffer overflow.
     I have checked it and double checked it so I believe it does not.
     Natanael */
  const char *tptr = text;
  const char *end = text;
  char *cname = (char *) zmalloc (strnlen (text, maxlen) + 2);
  char *cptr = cname;

  while (*end != 0) {
	size_t diff;
	end = strchr (tptr, '.');
	if (end == NULL)
	  end = text + strnlen (text, maxlen);
	if (end <= tptr) {
	  free (cname);
	  return NULL;
	}
	diff = end - tptr;
	*cptr++ = diff;
	memcpy (cptr, tptr, diff);
	cptr += diff;
	tptr = end + 1;
  }
  *cptr = 0;
  assert ((unsigned) (cptr - cname) == strnlen (text, maxlen) + 1);
  return cname;
}


static void free_packet (dnsheader_t * x)
{
  free (x->packet);
  free (x);
}

/*
 * Cache time calculations are done in seconds.  CACHE_TIMEUNIT
 * defines a minute.  You might want to set it to 1 for
 * debugging.
 */

#define CACHE_TIMEUNIT          60

/*
 * DNS queries that have been answered positively are stored for
 * CACHE_TIME, errors for only CACHE_NEGTIME minutes.  Reusing
 * a positive answer resets the entry's expire time.  After
 * CACHE_MAXTIME the item is removed anyway.
 */

#define CACHE_NEGTIME           (     5 * CACHE_TIMEUNIT)
#define CACHE_TIME              (    60 * CACHE_TIMEUNIT)
#define CACHE_MAXTIME           (6 * 60 * CACHE_TIMEUNIT)

/* The expire function can be called as often as wanted.
 * It will however wait CACHE_MINCYCLE between two
 * expire runs. */

#define CACHE_MINCYCLE          (5 * CACHE_TIMEUNIT)

/* If after an expire the cache holds more than CACHE_HIGHWATER
 * items the oldest items are removed until there are only
 * CACHE_LOWWATER left. */

static void free_cx(cache_t * cx)
{
  free_packet (cx->p);
  free (cx->name);
  free (cx);
}

static unsigned int get_stringcode(char *string)
{
  unsigned int c, code;
  int i;

  code = 0;
  for (i = 0; (c = (unsigned char) string[i]) != 0; i++) {
	if (isupper (c)) {
	  c = tolower (c);
	}

	code = code + c;
  }

  code = (code & 0xFF) + (strlen (string) << 8);
  return (code);
}

static cache_t *create_cx(dnsheader_t * x, rr_t * query, srvnode_t * server)
{
  cache_t *cx;

  cx = zmalloc (sizeof (cache_t));

  cx->name = strdup (query->name);
  cx->code = get_stringcode (cx->name);

  cx->positive = x->ancount;
  cx->type = query->type;
  cx->class = query->class;
  cx->p = x;
  cx->lastused = time (NULL);
  cx->server = server;

  return (cx);
}

static cache_t *append_cx(cache_t * cx)
{
  if (lastcache == NULL) {
	cachelist = cx;
	lastcache = cx;
  } else {
	lastcache->next = cx;
	cx->prev = lastcache;
	lastcache = cx;
  }

  cx->created = time (NULL);
  return cx;
}

static cache_t *remove_cx(cache_t * cx)
{
  if (cx->next != NULL) {
	cx->next->prev = cx->prev;
  } else {
	lastcache = cx->prev;
  }

  if (cx->prev != NULL) {
	cx->prev->next = cx->next;
  } else {
	cachelist = cx->next;
  }

  return (cx);
}

/* skip_labels()
 *
 * In: packet - pointer to the current position within the packet
 *     term   - beyond the end of the packet. Passing this is bad
 *
 * Returns:
 *     a pointer to the byte immediately beyond the name
 *     OR null, if we walk off the end of the packet */
static unsigned char *skip_labels(unsigned char *packet, unsigned char *term)
{
  while (1) {
	if (*packet > 63)
	  return packet + 2;
	if (*packet == 0)
	  return packet + 1;
	packet += *packet + 1;
	if (packet >= term)
	  return NULL;
  }
}

/* cache_dns_ttl()
 *
 * In: packet   - a pointer to the packet data
 *     len      - the size of the packet
 *     hdr      - Preprocessed header data
 *
 * Returns: the minimum TTL within the packet
 *    OR 0 if we can't work it out */
static unsigned int cache_dns_ttl(void *packet, int len, dnsheader_t * hdr)
{
  unsigned char *p = packet;
  unsigned char *term = p + len;
  unsigned long long ttl = 0x100000000; /* unit32 + 1 */
  unsigned int t;
  unsigned short i;
  unsigned int buf[1];
  unsigned short sbuf[1];

  // skip header
  p += PACKET_DATABEGIN;

  // skip questions
  i = hdr->qdcount;
  while (i) {
	p = skip_labels (p, term);
	if (p == NULL)
	  return 0;
	p += 4;
	i--;
	if (p >= term)
	  return 0;                                     //walked off the end
  }

  i = hdr->ancount + hdr->nscount + hdr->arcount;
  while (i) {
	if (p >= term || p < (unsigned char *) packet)
	  return 0;                                     //walked off the end
	p = skip_labels (p, term);
	if (p == NULL)
	  return 0;
	p += 4;
	if (p >= term)
	  return 0;                                     //walked off the end

	memcpy (buf, p, sizeof (buf));
	t = ntohl (buf[0]);
	if ((unsigned long long)t < ttl)
	  ttl = t;
	p += 4;
	if (p >= term)
	  return 0;                                     //walked off the end

	memcpy (sbuf, p, sizeof (sbuf));
	p += ntohs (sbuf[0]);
	p += 2;

	i--;
  }

  if (ttl == 0x1FFFFFFFF)
	return 0;
  return (unsigned int)ttl;
}

static dnsheader_t *alloc_packet(void *packet, int len)
{
  dnsheader_t *x;

  x = zmalloc (sizeof (dnsheader_t));

  x->packet = zmalloc (len + 2);
  x->len = len;
  memcpy (x->packet, packet, len);

  return (x);
}

static dnsheader_t *decode_header(void *packet, int len)
{
  unsigned short int *p;
  dnsheader_t *x;

  x = alloc_packet (packet, len);
  p = (unsigned short int *) x->packet;

  x->id = ntohs (p[0]);
  x->u = ntohs (p[1]);
  x->qdcount = ntohs (p[2]);
  x->ancount = ntohs (p[3]);
  x->nscount = ntohs (p[4]);
  x->arcount = ntohs (p[5]);

  x->here = (char *) &x->packet[12];
  return (x);
}

static dnsheader_t *parse_packet(unsigned char *packet, int len)
{
  dnsheader_t *x;

  x = decode_header (packet, len);
  return (x);
}


/* cmp_cachemru() - Compare the "ages" of two cache entries.
 *
 * In:      cx - pointer to first cache entry.
 *          cy - pointer to second cache entry.
 *
 * Returns: >0 if cx is older than cy
 *          <0 if cx is younger than cy
 *           0 if cx is equal to cy */
static int cmp_cachemru(const void *x, const void *y)
{
  const cache_t **cx = (const cache_t **) x;
  const cache_t **cy = (const cache_t **) y;
  return ((*cy)->lastused - (*cx)->lastused);
}

/* expire_oldest() - expire cache entries until lowwater is reached
 *
 * In:      itemcount - number of entries in cache.
 *
 * Returns: 0, always.
 *
 * Sorts all the entries in order of age, oldest first.
 * Then removes the first (itemcount - CACHE_LOWWATER) entries. */
static int expire_oldest(int itemcount)
{
  int i, n;
  cache_t *cx, **cv;

  if ((n = itemcount - CACHE_LOWWATER) < 0)
	return 0;

  cv = zmalloc (itemcount * sizeof (cache_t *));
  for (i = 0, cx = cachelist; i < itemcount; i++, cx = cx->next) {
	cv[i] = cx;
  }

  qsort (cv, itemcount, sizeof (cache_t *), cmp_cachemru);
  for (i = 0; i < n; i++) {
	cx = cv[i];
	remove_cx (cx);
	free_cx (cx);
  }

  log_msg (LOG_NOTICE, "cache: %d of %d mru expired, %d remaining\n",
		   n, itemcount, itemcount - n);

  return 0;
}

/* cache_expire() - Expire old entries from the cache.
 *
 * Returns: 0, always.
 *
 * When called, this function walks the entire cachelist, looking for entries
 * that should be expired. */
static int cache_expire(void)
{
  int total, expired;
  time_t now;
  cache_t *cx, *next;

  /* Initialize lastexpire to the current time */
  if (lastexpire == 0)
	lastexpire = time (NULL);

  now = time (NULL);
  if (now - lastexpire < CACHE_MINCYCLE)
	return 0;

  total = 0;
  expired = 0;

  cx = cachelist;
  while (cx != NULL) {
	total += 1;
	next = cx->next;

	if (cx->expires < now || (now - cx->created) > CACHE_MAXTIME) {
	  remove_cx (cx);
	  free_cx (cx);
	  expired++;
	}

	cx = next;
  }

  total = total - expired;
  if (total > CACHE_HIGHWATER)
	expire_oldest (total);

  lastexpire = now;
  return 0;
}

/* check_query()
 *
 * This functions does some test to verify that the message is sane
 * returns:
 *          -1 if the query is insane and should be ignored
 *          0 if the query is ok and should be processed
 *          1 if the query is not ok and a format error should be sent */
static int check_query(void *msg, int len)
{
  short int flags = ntohs (((short int *) msg)[1]);     /* flags */

  /* first check the size */
  if (len < 12) {
	/* Query packet is to small. Ignoring */
	return -1;
  }

  if (len > UDP_MAXSIZE) {
	/* Query packet is too big. Ignoring */
	return -1;
  }

  /* Flags check. If Z flag or QR is set, just ignore
   * the request. According to rfc1035 4.1.1 Z flag must be zero in
   * all queries and responses.
   * BIND set the RCODE on recursive lookups so we dont check that.
   */

  /* check if Z is set. It should never been set */
  if (flags & MASK_Z) {
	/* Z was set. Ignoring query */
	return -1;
  }

  /* Check if it is a query, if QR is set */
  if (flags & MASK_QR) {
	/* QR was set. Ignoring query */
	return -1;
  }

  /* ok this is valid */
  return 0;
}

/* check_reply()
 *
 * This functions does some test to verify that the message is sane
 * returns:
 *          -1 if the reply is insane and should be ignored
 *          0 if the query is ok and should be processed */
static int check_reply(srvnode_t * s, void *msg, int len)
{
  short int flags = ntohs (((short int *) msg)[1]);     /* flags */

  /* bad server replies are pretty bad. Lets log them to syslog */
  if (len < 12) {
	log_msg (LOG_WARNING, "Reply packet was to small. Ignoring reply from %s",
			 inet_ntoa (s->addr.sin_addr));
	return -1;
  }
  /* we have already checkd for oversized packets */

  /* check if Z is set. It should never been set */
  if (flags & MASK_Z) {
	log_msg (LOG_WARNING, "Z was set. Ignoring reply from %s",
			 inet_ntoa (s->addr.sin_addr));
	return -1;
  }

  /* Check if it is a query, if QR is set */
  if (!(flags & MASK_QR)) {
	log_msg (LOG_WARNING, "QR was not set. Ignoring reply from %s",
			 inet_ntoa (s->addr.sin_addr));
	return -1;
  }

  /* check for possible cache poisoning attempts etc here.... */

  /* ok this packet is ok */
  return 0;
}


static int
get_objname(const unsigned char *msg, const int msgsize, int index,
			 char *dest, const int destsize)
{
  int i = index;
  int j = 0;
  int depth = 0;
  int compressed = 0;
  int c;

  /* we refuse to try to decode anything in the header or after the packet */
  if (index < PACKET_DATABEGIN || i >= msgsize)
	return -1;

  while ((c = msg[i++]) != '\0') {
	if (i >= msgsize)
	  return -1;

	/* check if it is a pointer */
	if ((c & 0xc0) == 0xc0) {
	  /* check that next byte is readable */
	  if (i >= msgsize)
		return -1;

	  /* Avoid circular pointers. */
	  depth++;
	  if ((depth >= msgsize) || (depth > UDP_MAXSIZE))
		return -1;

	  /* we store the first location */
	  if (!compressed)
		compressed = i + 1;

	  /* check that the pointer points within limits */
	  if ((i = ((c & 0x3f) << 8) + msg[i]) >= msgsize)
		return -1;
	  continue;
	} else if (c < 64) {
	  /* check that label length don't cross any borders */
	  if (((j + c + 1) >= destsize) || ((i + c) >= msgsize))
		return -1;

	  while (c--) {
		/* currently we accept everything... */
		dest[j++] = msg[i++];
	  }
	  dest[j++] = '.';
	} else
	  return -1;                          /* a label cannot be longer than 63 */
  }

  if (--j >= (destsize))
	return -1;                            /* we need space for '\0' */
  dest[j] = '\0';
  /* if we used compression, return the location from the first ptr */
  return (compressed ? compressed : i);
}

static int
snprintf_cname(char *msg, const int msgsize,   /* the dns packet */
				int index,              /* where in the DNS packet the name is */
				char *dest, int destsize)
{                                                               /* where to store the cname */
  static const char malformatted[] = "(malformatted)";

  if (get_objname ((unsigned char *)msg, msgsize, index, dest, destsize) < 0) {
	strncpy (dest, malformatted, destsize);
	return -1;
  }
  return 0;
}

static srvnode_t *alloc_srvnode(void)
{
  srvnode_t *p = zmalloc (sizeof (srvnode_t));
  p->inactive = -1;
  p->send_time = 0;
  p->send_count = 0;
  /* actually we return a new emty list... */
  return p->next = p;
}

/* Allocate a domain node */
static domnode_t *alloc_domnode(void)
{
  domnode_t *p = zmalloc (sizeof (domnode_t));
  p->domain = NULL;
  p->srvlist = alloc_srvnode ();
  /* actually we return a new emty list... */
  return p->next = p;
}

/* add a domain */
/* note: cname must be allocated! */
static domnode_t *add_domain(domnode_t * list, const int load_balance, char *cname)
{
  domnode_t *p;

  p = alloc_domnode ();
  p->domain = cname;
  p->roundrobin = load_balance;
  p->next = list->next;
  list->next = p;
  return p;
}

/* search for a domain. returns the node if found or NULL if not */
static domnode_t *search_domnode (domnode_t * head, const char *name)
{
  domnode_t *d = head;
  /* the list head is pointing to the default domain */
  while ((d = d->next) != head) {
	if (strcmp (d->domain, name) == 0)
	  return d;
  }
  return NULL;
}

/* search for the domnode that has the domain. Returns domnode if
   domain is a subdomain under domnode */
static domnode_t *
search_subdomnode(domnode_t * head, const char *name, const int maxlen)
{
  domnode_t *d = head, *curr = head;
  int h, n, maxfound = 0;
  const char *p;
  assert ((head != NULL) && (name != NULL));
  /* the list head is pointing to the default domain */
  if ((name == NULL) || (d == NULL))
	return head;
  while ((d = d->next) != head) {
	if ((n = strnlen (name, maxlen)) > (h = strnlen (d->domain, maxlen))) {
	  p = name + n - h;
	} else
	  p = name;

	/* this works because the domain names are in cname format so
	   hayes.org will nor appear as a subdomain under yes.org yes.org
	   will be encoded as "\3yes\3org" while hayes.org will be encoded
	   as "\5hayes\3org"
	 */
	if ((strncmp (d->domain, p, maxlen - (p - name)) == 0) && (h > maxfound)) {
	  maxfound = h;                         /* max length found */
	  curr = d;
	}
  }
  return curr;
}

/* convert cname to ascii and return a static buffer */
/* this func must *never* be called with an incomming DNS
   packet as input. Never. */
static char *cname2asc(const char *cname, char buf[256])
{
  /* Note: we don't really check the size of the incomming cname. but
     according to RFC 1035 a name must not be bigger than 255 octets.
   */
  if (cname)
	snprintf_cname ((char *) cname, strlen (cname), 0, buf, 256); /* sizeof(buf) */
  else
	strcpy (buf, "(default)");
  return buf;
}

/* wrapper for setting current server */
static srvnode_t *set_current(domnode_t * d, srvnode_t * s)
{
  char buf[256];
  assert (d != NULL);
  if (s) {
	if (d->roundrobin == 0)
	  log_msg (LOG_NOTICE, "Setting server %s for domain %s", inet_ntoa (s->addr.sin_addr), cname2asc (d->domain, buf));
  } else
	log_msg (LOG_WARNING, "No active servers for domain %s", cname2asc (d->domain, buf));
  return (d->current = s);
}

/* next active server
   Returns: NULL - if there are no active servers in list
	    the next active server otherwise */
static srvnode_t *next_active(domnode_t * d)
{
  srvnode_t *s, *start;
  assert (d != NULL);
  if (d->current) {
	start = d->current;
  } else {                           /* previously deactivated everything, lets check from start */
	start = d->srvlist;
  }
  for (s = start->next; s->inactive && s != start; s = s->next);
  if (s->inactive)
	s = NULL;
  return (s);
}

/* deactivate current server
   Returns: next active server, NULL if there are none */
static srvnode_t *deactivate_current(domnode_t * d)
{
  assert (d != NULL);
  if (d->current) {
	d->current->inactive = time (NULL);
	log_msg (LOG_NOTICE, "Deactivating DNS server %s",
			 inet_ntoa (d->current->addr.sin_addr));
  }
  return set_current (d, next_active (d));
}


static char *strnlwr(char *string, const int maxlen)
{
  int i;

  if (!string)
	return NULL;
  for (i = 0; (i <= maxlen) && (string[i] != 0); i++) {
	string[i] = tolower (string[i]);
  }

  return (string);
}


static int myrand(int max)
{
  return (int) ((float) max * (unsigned int) rand (&isaac_ctx) /
				(0xffffffff + 1.0));
}

static void qid_init_pool(void)
{
  int i;
  struct timeval tv;

  for (i = 0; i < QID_POOL_SIZE; i++)
	qid_pool[i] = i;

  /* get random seed from time */
  for (i = 0; i < RANDSIZ; i++) {
	gettimeofday (&tv, 0);
	isaac_ctx.randrsl[i] = tv.tv_sec + tv.tv_usec;
  }
  /* init prng */
  randinit (&isaac_ctx, TRUE);
}

static unsigned short int qid_get(void)
{
  unsigned short int t;
  int i = myrand (pool_ptr) % QID_POOL_SIZE;

  t = qid_pool[i];
  /* shrink the pool */
  qid_pool[i] = qid_pool[pool_ptr-- % QID_POOL_SIZE];
  return (t);
}

static unsigned short int qid_return(unsigned short int qid)
{
  return (qid_pool[++pool_ptr % QID_POOL_SIZE] = qid);
}


/* init the query list */
static void query_init(void)
{
  qlist_tail = (qlist.next = &qlist);
}

/* create a new query, and open a socket to the server */
static query_t *query_create(domnode_t * d, srvnode_t * s)
{
  query_t *q;
#ifdef RANDOM_SRC
  struct sockaddr_in my_addr;
#endif

  /* should never be called with no server */
  assert (s != NULL);

  /* check if we have reached maximum of sockets */
  if (upstream_sockets >= max_sockets) {
	if (!dropping) {
	  dropping = 1;
	  log_msg (LOG_WARNING, "Socket limit reached. Dropping new queries");
	}
	return NULL;
  }

  dropping = 0;
  /* allocate */
  if ((q = (query_t *) zmalloc (sizeof (query_t))) == NULL)
	return NULL;

  /* return an emtpy circular list */
  q->next = (struct _query *) q;

  /* we specify both domain and server */
  /* the dummy queries will use server but no domain is attatched */
  q->domain = d;
  q->srv = s;

  /* set the default time to live value */
  q->ttl = FORWARD_TIMEOUT;

  /* open a new socket */
  if ((q->sock = socket (AF_INET, SOCK_DGRAM, 0)) < 0) {
	bb_perror_msg("query_create: Couldn't open socket");
	free (q);
	return NULL;
  } else
	upstream_sockets++;

  /* bind to random source port */
#ifdef RANDOM_SRC
  memset (&my_addr, 0, sizeof (my_addr));
  my_addr.sin_family = AF_INET;
  my_addr.sin_addr.s_addr = INADDR_ANY;

  my_addr.sin_port = htons (myrand (65536 - 1026) + 1025);
  while ((bind (q->sock, (struct sockaddr *) &my_addr, sizeof (struct sockaddr))) == -1) {
	if (errno != EADDRINUSE) {
	  log_msg (LOG_WARNING, "bind: %s", strerror (errno));
	  break;
	}

	/* try another port */
	my_addr.sin_port = htons (myrand (65535 - 1026) + 1025);
  }
#endif

  /* add the socket to the master FD set */
  FD_SET (q->sock, &fdmaster);
  if (q->sock > maxsock)
	maxsock = q->sock;

  /* get an unused QID */
  q->my_qid = qid_get ();
  return q;
}

static query_t *query_destroy(query_t * q)
{
  /* close the socket and return mem */
  qid_return (q->my_qid);

  /* unset the socket */
  FD_CLR (q->sock, &fdmaster);
  close (q->sock);

  upstream_sockets--;
  free (q);
  return NULL;
}

/* Get a new query */
static query_t *query_get_new(domnode_t * dom, srvnode_t * srv)
{
  query_t *q;
  assert (srv != NULL);
  /* if there are no prepared queries waiting for us, lets create one */
  if ((q = srv->newquery) == NULL) {
	if ((q = query_create (dom, srv)) == NULL)
	  return NULL;
  }
  srv->newquery = NULL;
  q->domain = dom;
  q->srv = srv;
  return q;
}

/* get qid, rewrite and add to list. Retruns the query before the added  */
static query_t *
query_add(domnode_t * dom, srvnode_t * srv,
		   const struct sockaddr_in * client, char *msg, unsigned len)
{

  query_t *q, *p, *oldtail;
  unsigned short client_qid = *((unsigned short *) msg);
  time_t now = time (NULL);

  /* look if the query are in the list if it is, don't add it again. */
  for (p = &qlist; p->next != &qlist; p = p->next) {
	if ((p->next->client_qid == client_qid) &&
		((p->next->client.sin_addr.s_addr == client->sin_addr.s_addr) &&
		 (p->next->client.sin_port == client->sin_port))) {

	  /* we found the qid in the list */
	  *((unsigned short *) msg) = p->next->my_qid;
	  p->next->client_time = now;
	  return p;
	}
  }

  if ((q = query_get_new (dom, srv)) == NULL) {
	/* if we could not allocate any new query, return with NULL */
	return NULL;
  }

  q->client_qid = client_qid;
  memcpy (&(q->client), client, sizeof (struct sockaddr_in));
  q->client_time = now;
  q->client_count = 1;

  /* set new qid from random generator */
  client_qid = htons (q->my_qid);
  memcpy(msg, &client_qid, sizeof(client_qid));

  /* add the query to the list */
  q->next = qlist_tail->next;
  qlist_tail->next = q;

  /* new query is new tail */
  oldtail = qlist_tail;
  qlist_tail = q;
  return oldtail;
}

/* remove query after */
static query_t *query_delete_next(query_t * q)
{
  query_t *tmp = q->next;

  /* unlink tmp */
  q->next = q->next->next;

  /* if this was the last query in the list, we need to update the tail */
  if (qlist_tail == tmp)
	qlist_tail = q;

  /* destroy query */
  query_destroy (tmp);
  return q;
}


/* remove old unanswered queries */
static void query_timeout(void)
{
  int count = 0;
  time_t now = time (NULL);
  query_t *q;

  for (q = &qlist; q->next != &qlist; q = q->next) {
	if (q->next->client_time < (now - q->next->ttl)) {
	  count++;
	  query_delete_next (q);
	}
  }
  /* count ? query_timeout: removed count entries */
  total_timeouts += count;
}

/*
rand.c: By Bob Jenkins.  My random number generator, ISAAC.  Public Domain.
MODIFIED:
  960327: Creation (addition of randinit, really)
  970719: use context, not global variables, for internal state
  980324: added main (ifdef'ed out), also rearranged randinit()
  010626: Note that this is public domain
*/

#define ind(mm,x)  (*(ub4 *)((ub1 *)(mm) + ((x) & ((RANDSIZ-1)<<2))))
#define rngstep(mix,a,b,mm,m,m2,r,x) \
{ \
  x = *m;  \
  a = (a^(mix)) + *(m2++); \
  *(m++) = y = ind(mm,x) + a + b; \
  *(r++) = b = ind(mm,y>>RANDSIZL) + x; \
}

void
isaac (randctx *ctx)
{
  ub4 a, b, x, y, *m, *mm, *m2, *r, *mend;

  mm = ctx->randmem;
  r = ctx->randrsl;
  a = ctx->randa;
  b = ctx->randb + (++ctx->randc);
  for (m = mm, mend = m2 = m + (RANDSIZ / 2); m < mend;) {
	rngstep (a << 13, a, b, mm, m, m2, r, x);
	rngstep (a >> 6, a, b, mm, m, m2, r, x);
	rngstep (a << 2, a, b, mm, m, m2, r, x);
	rngstep (a >> 16, a, b, mm, m, m2, r, x);
  }
  for (m2 = mm; m2 < mend;) {
	rngstep (a << 13, a, b, mm, m, m2, r, x);
	rngstep (a >> 6, a, b, mm, m, m2, r, x);
	rngstep (a << 2, a, b, mm, m, m2, r, x);
	rngstep (a >> 16, a, b, mm, m, m2, r, x);
  }
  ctx->randb = b;
  ctx->randa = a;
}


#define mix(a,b,c,d,e,f,g,h) \
{ \
   a^=b<<11; d+=a; b+=c; \
   b^=c>>2;  e+=b; c+=d; \
   c^=d<<8;  f+=c; d+=e; \
   d^=e>>16; g+=d; e+=f; \
   e^=f<<10; h+=e; f+=g; \
   f^=g>>4;  a+=f; g+=h; \
   g^=h<<8;  b+=g; h+=a; \
   h^=a>>9;  c+=h; a+=b; \
}

/* if (flag==TRUE), then use the contents of randrsl[] to initialize mm[]. */
void
randinit (randctx *ctx, word flag)
{
  word i;
  ub4 a, b, c, d, e, f, g, h;
  ub4 *m, *r;
  ctx->randa = ctx->randb = ctx->randc = 0;
  m = ctx->randmem;
  r = ctx->randrsl;
  a = b = c = d = e = f = g = h = 0x9e3779b9;   /* the golden ratio */

  for (i = 0; i < 4; ++i) {             /* scramble it */
	mix (a, b, c, d, e, f, g, h);
  }

  if (flag) {
	/* initialize using the contents of r[] as the seed */
	for (i = 0; i < RANDSIZ; i += 8) {
	  a += r[i];
	  b += r[i + 1];
	  c += r[i + 2];
	  d += r[i + 3];
	  e += r[i + 4];
	  f += r[i + 5];
	  g += r[i + 6];
	  h += r[i + 7];
	  mix (a, b, c, d, e, f, g, h);
	  m[i] = a;
	  m[i + 1] = b;
	  m[i + 2] = c;
	  m[i + 3] = d;
	  m[i + 4] = e;
	  m[i + 5] = f;
	  m[i + 6] = g;
	  m[i + 7] = h;
	}
	/* do a second pass to make all of the seed affect all of m */
	for (i = 0; i < RANDSIZ; i += 8) {
	  a += m[i];
	  b += m[i + 1];
	  c += m[i + 2];
	  d += m[i + 3];
	  e += m[i + 4];
	  f += m[i + 5];
	  g += m[i + 6];
	  h += m[i + 7];
	  mix (a, b, c, d, e, f, g, h);
	  m[i] = a;
	  m[i + 1] = b;
	  m[i + 2] = c;
	  m[i + 3] = d;
	  m[i + 4] = e;
	  m[i + 5] = f;
	  m[i + 6] = g;
	  m[i + 7] = h;
	}
  } else {
	/* fill in m[] with messy stuff */
	for (i = 0; i < RANDSIZ; i += 8) {
	  mix (a, b, c, d, e, f, g, h);
	  m[i] = a;
	  m[i + 1] = b;
	  m[i + 2] = c;
	  m[i + 3] = d;
	  m[i + 4] = e;
	  m[i + 5] = f;
	  m[i + 6] = g;
	  m[i + 7] = h;
	}
  }

  isaac (ctx);                                  /* fill in the first set of results */
  ctx->randcnt = RANDSIZ;               /* prepare to use the first set of results */
}


/* prepare the dns packet for a Server Failure reply */
static char *set_srvfail(char *msg, const int len)
{
  if (len < 4)
	return NULL;
  /* FIXME: host to network should be called here */
  /* Set flags QR and AA */
  msg[2] |= 0x84;
  /* Set flags RA and RCODE=3 */
  msg[3] = 0x82;
  return msg;
}

/* parse_query()
 *
 * The function get_dnsquery() returns us the query part of an
 * DNS packet.  For this we must already have a dnsheader_t
 * packet which is additional work.  To speed things a little
 * bit up (we use it often) parse_query() gets the query direct.
 *
 * returns non-zero on error and zero on success. */
static int parse_query(rr_t * y, unsigned char *msg, int len)
{
  int k;
  int i;

  /* If QDCOUNT, the number of entries in the question section,
   * is zero, we just give up */

  if (ntohs (((dnsheader_t *) msg)->qdcount) == 0) {
	/*      QDCOUNT will always be zero on responses. */
	return -1;
  }

  y->flags = ntohs (((short int *) msg)[1]);

  if ((i = get_objname (msg, len, PACKET_DATABEGIN, y->name, sizeof (y->name))) < 0)
	return -1;

  /* check that there is type and class */
  if (i + 4 > len)
	return -1;

  y->type = ntohs (*(unsigned short *) (&msg[i]));
  i += 2;

  y->class = ntohs (*(unsigned short *) (&msg[i]));
  i += 2;

  /* those strings should have been checked in get_objectname */
  if ((k = strlen (y->name))) {
	if (y->name[--k] == '.')
	  y->name[k] = '\0';
  }

  /* should we really convert the name to lowercase?
   * rfc1035 2.3.3
   */
  strnlwr (y->name, sizeof (y->name));

  return 0;
}

/* cache_lookup()
 *
 * In/Out:  packet - the query packet on input, response on output.
 *          len    - length of the query packet.
 *
 * Returns: length of the response packet "cached",
 *          0 if no cached response is available.
 *
 * Look into the query packet too see if it is a lookup request that
 * might already have been cached.  If so, copy the data to the
 * cached pointer returning the length of the cached packet
 * as return value.
 *
 * The function assumes that the area cached points to is
 * large enough. */
static int cache_lookup(void *packet, int len)
{
  unsigned int code;
  /*    dnsheader_t *x; */
  rr_t query;
  cache_t *cx = NULL;

  if (parse_query (&query, packet, len) ||
	  (GET_QR (query.flags) == 1) ||
	  (*query.name == 0) || (query.class != DNS_CLASS_INET)) {
	return 0;
  }

  /* The query could be in the cache.  Let's take the packet ...
   * ... and search our cache for this request. */
  code = get_stringcode (query.name);
  for (cx = cachelist; cx != NULL; cx = cx->next) {
	if (cx->code == code &&
		cx->type == query.type &&
		cx->class == query.class && strcasecmp (cx->name, query.name) == 0) {

	  if (cx->positive > 0) {
		cx->lastused = time (NULL);
	  }

	  if (cx->expires <= cx->lastused) {
		// this packet has expired, get rid of it!
		remove_cx (cx);
		free_cx (cx);
		/* found cached entry is expired, deleting */
		return 0;
	  }

	  memcpy (packet + 2, cx->p->packet + 2, cx->p->len - 2);
	  cache_hits++;

	  /* lets check if the server is active */
	  if (ignore_inactive_cache_hits && cx->server->inactive) {
		/* server is inactive. Skipping cache entry */
		return 0;
	  }

	  return (cx->p->len);
	}
  }

  cache_misses++;
  return 0;
}

/* check if there are a list or not
   retruns 1 if its empty or NULL
   returns 0 if there are servers in the list */
static int no_srvlist(srvnode_t * head)
{
  if (!head)
	return 1;
  return (head->next == head);
}

/* handle_query()
 *
 * In/Out:  msg       - the query on input, the reply on output.
 *          len       - length of the query/reply
 *
 * Out:     dptr      - dptr->current contains the server to which to forward the query
 *
 * Returns:  -1 if the query is bogus
 *           1  if the query should be forwarded to the srvidx server
 *           0  if msg now contains the reply
 *
 * Takes a single DNS query and determines what to do with it.
 * This is common code used for both TCP and UDP.
 *
 * Assumptions: There is only one request per message. */
static int handle_query(char *msg, int *len, domnode_t ** dptr)
{
  int replylen;
  domnode_t *d;

  /* Next, see if we have the answer cached */
  if ((replylen = cache_lookup (msg, *len)) > 0) {
	/* Replying to query with cached answer */
	*len = replylen;
	return 0;
  } else if (replylen < 0)
	return -1;

  /* get the server list for this domain */
  d = search_subdomnode (domain_list, &msg[12], *len);

  if (no_srvlist (d->srvlist)) {
	/* there is no servers for this domain, reply with "Server failure" */
	if (!set_srvfail (msg, *len))
	  return -1;
	return 0;
  }

  if (d->roundrobin) {
	set_current (d, next_active (d));
  } else {
	/* find the first active server */
	d->current = NULL;
	d->current = next_active (d);
  }

  /* Send to a server until it "times out". */
  if (d->current) {
	time_t now = time (NULL);
	if ((d->current->send_time != 0)
		&& (FORWARD_TIMEOUT != 0)
		&& (REACTIVATE_INTERVAL != 0)
		&& (now - d->current->send_time > FORWARD_TIMEOUT)) {
	  deactivate_current (d);
	}
  }

  if (d->current) {
	/* Forwarding the query to DNS server inet_ntoa (d->current->addr.sin_addr) */ ;
  } else {
	/* All servers deactivated. Replying with "Server failure" */
	if (!set_srvfail (msg, *len))
	  return -1;
	return 0;
  }

  *dptr = d;
  return 1;
}

/* Check if any server are timing out and should be deactivated */
static void deactivate_servers (int interval)
{
  time_t now = time (NULL);
  domnode_t *d = domain_list;
  srvnode_t *s;

  if (!last_try_ds)
	last_try_ds = now;
  if (now - last_try_ds < interval)
	return;
  last_try_ds = now;

  do {
	if ((s = d->srvlist))
	  while ((s = s->next) != d->srvlist) {
		int current_disabled = 0;
		if (s->inactive)
		  continue;
		if (s->send_time && (difftime (now, s->send_time) > FORWARD_TIMEOUT)) {
		  s->inactive = now;
		  if (s == d->current)  /* we need to update d->current */
			current_disabled = 1;
		}
		if (current_disabled)
		  deactivate_current (d);
	  }
  } while ((d = d->next) != domain_list);
}

static void srv_stats(time_t interval)
{
  srvnode_t *s;
  domnode_t *d = domain_list;
  time_t now = time (NULL);

  if (last_stats + interval > now) {
	last_stats = now;
	do {
	  if ((s = d->srvlist))
		while ((s = s->next) != d->srvlist)
		  /* stats for %s: send count=%i", inet_ntoa (s->addr.sin_addr), s->send_count */ ;
	} while ((d = d->next) != domain_list);
  }
}


/* print statics about the query list and open sockets */
static void query_stats(time_t interval)
{
  time_t now = time (NULL);

  if (interval == 0)
	return;
  if (last_q + interval < now) {
	last_q = now;
	log_msg (LOG_INFO, "Hits: %i, Misses: %i, Total: %i, Timeouts: %li",
			 cache_hits, cache_misses, cache_hits + cache_misses,
			 total_timeouts);
	cache_hits = cache_misses = total_timeouts = 0;
  }
}

/* Setup signal handlers.
 * See: select_tut manual page
 *
 * basicly, we mask out the signals that is not supposed to be
 * recieved except within the pselect call. */
static void init_sig_handler(sigset_t * orig_sigmask)
{
  sigset_t sigmask;

  sigmask = bb_sigmask((1 << SIGINT) | (1 << SIGQUIT) | (1 << SIGTERM) | (1 << SIGUSR1) | (1 << SIGCHLD));

  sigprocmask (SIG_BLOCK, &sigmask, orig_sigmask);

  /* Handling TCP requests is done by forking a child.  When they terminate
   * they send SIGCHLDs to the parent.  This will eventually interrupt
   * some system calls.  Because I don't know if this is handled it's better
   * to ignore them -- 14OCT99wzk */
  signal (SIGCHLD, SIG_IGN);
}

/* add a server.*/
static void add_srv(srvnode_t * head, const char *ipaddr, int rport)
{
  srvnode_t *p;
  struct sockaddr_in addr;

  if (!inet_aton (ipaddr, &addr.sin_addr))
	bb_error_msg_and_die ("bad ip address \"%s\"", ipaddr);
  addr.sin_family = AF_INET;
  addr.sin_port = htons (rport);
  p = alloc_srvnode ();
  memcpy (&p->addr, &addr, sizeof (addr));
  p->inactive = 0;
  p->next = head->next;
  head->next = p;
}

/* returns the last srvnode in the list */
static srvnode_t *last_srvnode(srvnode_t * head)
{
  srvnode_t *p = head;
  /* head should always be != NULL */
  assert (p != NULL);
  while (p->next != head)
	p = p->next;
  return p;
}


#define TCP_EXIT(x) _exit((x))

typedef struct tcp_handle_info
{
  int connect;
  struct sockaddr_in client;
  unsigned len;
} tcp_handle_t;

/*
 * alarm_handler() - Timeout handler for the connect() function in ip_open().
 *              It's purpose is just to interrupt the connect() system call.
 */
static void
alarm_handler(int dummy)
{
  (void)&dummy;
  return;
}

/*
 * ip_open() - Open a tcp socket to the given DNS server
 *
 * In:      server - The address to which to open a connection
 *
 * Returns: the opened socket on success, -1 on failure.
 */
static int ip_open (struct sockaddr_in server)
{
  int sock;

  sock = socket (PF_INET, SOCK_STREAM, IPPROTO_IP);
  if (sock < 0)
	return -1;

  signal (SIGALRM, alarm_handler);
  alarm (10);
  if (connect (sock, (struct sockaddr *) &server, sizeof (server)) < 0)
	return -1;

  alarm (0);
  signal (SIGALRM, SIG_DFL);

  return (sock);
}

/* cache_dnspacket()
 *
 * In:      packet - the response packet to cache.
 *          len    - length of the response packet.
 *
 * Take the response packet and look if it meets some basic
 * conditions for caching.  If so put the entire response into
 * our cache. */
static void cache_dnspacket(void *packet, int len, srvnode_t * server)
{
  dnsheader_t *x;
  rr_t query;
  cache_t *cx = NULL;
  unsigned int dns_ttl;

  if (parse_query (&query, packet, len) ||
	  (GET_QR (query.flags) == 0) || (*query.name == 0)) {
	return;
  }

  x = parse_packet (packet, len);

  /* Ok, the packet is interesting for us.  Let's put it into our
   * cache list. */
  cx = create_cx (x, &query, server);
  append_cx (cx);

  /* Set the expire time of the cached object. */
  cx->lastused = time (NULL);
  dns_ttl = cache_dns_ttl (packet, len, x);
  if (dns_ttl == 0) {
	/* Using static cache timeouts */
	cx->expires = cx->lastused +
	  ((cx->p->ancount > 0) ? CACHE_TIME : CACHE_NEGTIME);
  } else {
	/* Using dynamic cache timeouts -> dns_ttl seconds */
	cx->expires = cx->lastused + dns_ttl;
  }
}

/* tcp_handler() - This function accepts an incoming connection on
 *                 tcpsock and handles the client request by transparent
 *                 proxying.  After the request is served the thread
 *                 terminates. */
static void *tcp_handler(tcp_handle_t * dummy)
{
  tcp_handle_t *arg = (tcp_handle_t *) dummy;
  int rc, bytes;
  int max_sock;
  char buffer[4096];                    /* FIXME: do we want this on the stack? */
  struct timeval tov;
  fd_set connection, available;
  unsigned short tcpsize;
  int connect = arg->connect;
  domnode_t *d = domain_list;
  srvnode_t *s;

  do {
	if ((s = d->srvlist))
	  while ((s = s->next) != d->srvlist)
		s->tcp = -1;
  } while ((d = d->next) != domain_list);

  max_sock = connect + 1;
  FD_ZERO (&connection);
  FD_SET (connect, &connection);

  while (1) {
	int child_die = 0;

	memcpy (&available, &connection, sizeof (fd_set));
	tov.tv_sec = 120;                       /* Give the DNS 2min to answer. */
	tov.tv_usec = 0;

	rc = select (max_sock, &available, NULL, NULL, &tov);
	if (rc < 0) {
	  bb_perror_msg_and_die("[%d] select() error", getpid());
	} else if (rc == 0) {
	  log_msg (LOG_NOTICE, "[%d] connection timed out", getpid ());
	  TCP_EXIT (0);
	}

	/* Forward replies from DNS server to client */
	d = domain_list;
	do {
	  if ((s = d->srvlist))
		while ((s = s->next) != d->srvlist) {
		  if (s->tcp == -1)
			continue;
		  if (FD_ISSET (s->tcp, &available)) {
			/* [getpid ()] Received tcp reply.  Forwarding... */
			if ((bytes = read (s->tcp, buffer, sizeof (buffer))) <= 0) {
			  child_die = 1;
			  break;
			}
			cache_dnspacket (buffer + 2, bytes - 2, s);
			if (write (connect, buffer, bytes) != bytes) {
			  child_die = 1;
			  break;
			}
		  }
		}
	  if (child_die)
		break;
	} while ((d = d->next) != domain_list);

	if (child_die)
	  break;

	/* Forward requests from client to DNS server */
	if (FD_ISSET (connect, &available)) {
	  int retn;
	  domnode_t *dptr;


	  bytes = read (connect, &tcpsize, sizeof (tcpsize));
	  /* check for connection close */
	  if (bytes == 0)
		break;
	  /* [getpid ()] Received tcp DNS query...", getpid () */

	  if (bytes < 0) {
		/* tcp read error */
		break;
	  }
	  memcpy (buffer, &tcpsize, sizeof (tcpsize));
	  tcpsize = ntohs (tcpsize);
	  if (tcpsize > (sizeof (buffer) + 2)) {
		log_msg (LOG_WARNING, "[%d] Received tcp message is too big to process", getpid ());
		break;
	  }
	  bytes = read (connect, buffer + 2, (size_t) tcpsize);
	  if (bytes == 0) {
		bb_error_msg("[%d] tcp DNS query is mangled", getpid());
		break;
	  }
	  if (bytes < 0) {
		/* tcp read error */
		break;
	  }
	  if ((retn = handle_query (buffer + 2, &bytes, &dptr)) < 0) {
		/* bogus query */
		break;
	  }
	  /* If we can reply locally (master, cache, no servers), do so. */
	  else if (retn == 0) {
		unsigned short len = htons ((unsigned short) bytes);
		memcpy (buffer, &len, sizeof (len));
		if (write (connect, buffer, bytes + 2) != (bytes + 2))
		  break;
	  }

	  /* Forward DNS request to the appropriate server.
	   * Open a socket if one doesn't already exist. */
	  else if (retn == 1) {
		s = dptr->current;
		if (s->tcp == -1) {
		  s->tcp = ip_open (s->addr);
		  if (s->tcp < 0) {
			bb_error_msg("[%d] Can't connect to server", getpid());
			break;
		  }
		  if (s->tcp > (max_sock - 1))
			max_sock = s->tcp + 1;
		  FD_SET (s->tcp, &connection);
		}
		if (write (s->tcp, buffer, bytes + 2) != (bytes + 2))
		  break;
	  }
	}

  }
  /* The child process is done.  It can now die */
  /* "[getpid ()] Exiting child */
  TCP_EXIT (0);
  /* NOTREACHED */
  return (void *) 0;
}

/* handle_tcprequest()
 *
 * This function spawns a thread to actually handle the tcp request.
 * A thread is used instead of a process so that any DNS replies can
 * be placed in cache of the parent process. */
static void tcp_handle_request(void)
{
  tcp_handle_t *arg;
  pid_t pid;

  arg = (tcp_handle_t *) zmalloc (sizeof (tcp_handle_t));
  arg->len = sizeof (arg->client);

  arg->connect = accept (tcpsock, (struct sockaddr *) &(arg->client),
						 &(arg->len));
  if (arg->connect < 0) {
	bb_perror_msg("accept error for TCP connection");
	free (arg);
	return;
  }

  if ((pid = fork ()) == 0) {
	/* we are the child */
	tcp_handler (arg);
  } else if (pid < 0) {
	bb_perror_msg("fork for TCP connection failed");
  }
  close (arg->connect);
  free (arg);
}

/* dnssend()                                            22OCT99wzk
 *
 * Abstract: A small wrapper for send()/sendto().  If an error occurs a
 *           message is written to syslog.
 *
 * Returns:  The return code from sendto(). */
static int udp_send(int sock, srvnode_t * srv, void *msg, int len)
{
  int rc;
  time_t now = time (NULL);
  rc = sendto (sock, msg, len, 0, (const struct sockaddr *) &srv->addr, sizeof (struct sockaddr_in));

  if (rc != len) {
	bb_perror_msg("sendto %s error", inet_ntoa (srv->addr.sin_addr));
	return rc;
  }
  if (srv->send_time == 0)
	srv->send_time = now;
  srv->send_count++;
  return rc;
}

static int
send2current (query_t * q, void *msg, const int len)
{
  /* If we have domains associated with our servers, send it to the
     appropriate server as determined by srvr */
  domnode_t *d;
  assert (q != NULL);
  assert (q->domain != NULL);

  d = q->domain;
  while ((d->current != NULL)
		 && (udp_send (q->sock, d->current, msg, len) != len)) {
	if (REACTIVATE_INTERVAL)
	  deactivate_current (d);
  }
  if (d->current != NULL) {
	return len;
  } else
	return 0;
}


/* handle_udprequest()
 *
 * This function handles udp DNS requests by either replying to them (if we
 * know the correct reply via master, caching, etc.), or forwarding them to
 * an appropriate DNS server. */
static query_t *udp_handle_request(void)
{
  unsigned addr_len;
  int len;
  const int maxsize = UDP_MAXSIZE;
  struct sockaddr_in from_addr;
  int fwd;
  domnode_t *dptr;
  query_t *q, *prev;

  /* Read in the message */
  addr_len = sizeof (struct sockaddr_in);
  len = recvfrom (isock, udp_msg, maxsize, 0, (struct sockaddr *) &from_addr, &addr_len);
  if (len < 0) {
	/* recvfrom error */
	return NULL;
  }

  /* do some basic checking */
  if (check_query (udp_msg, len) < 0)
	return NULL;

  /* Determine how query should be handled */
  if ((fwd = handle_query (udp_msg, &len, &dptr)) < 0)
	return NULL;                            /* if its bogus, just ignore it */

  /* If we already know the answer, send it and we're done */
  if (fwd == 0) {
	if (sendto (isock, udp_msg, len, 0, (const struct sockaddr *) &from_addr, addr_len) != len) {
	  /* sendto error */ ;
	}
	return NULL;
  }

  /* rewrite udp_msg, get id and add to list */
  if ((prev = query_add (dptr, dptr->current, &from_addr, udp_msg, len)) == NULL) {
	/* of some reason we could not get any new queries. we have to
	   drop this packet */
	return NULL;
  }
  q = prev->next;


  if (send2current (q, udp_msg, len) > 0) {
	/* add to query list etc etc */
	return q;
  }
  return q;
}

/* dnsrecv()                                                    22OCT99wzk
 *
 * Abstract: A small wrapper for recv()/recvfrom() with output of an
 *           error message if needed.
 *
 * Returns:  A positove number indicating of the bytes received, -1 on a
 *           recvfrom error and 0 if the received message is too large. */
static int reply_recv(query_t * q, void *msg, int len)
{
  int rc;
  struct sockaddr_in from;
  socklen_t fromlen = sizeof (struct sockaddr_in);

  rc = recvfrom (q->sock, msg, len, 0, (struct sockaddr *) &from, &fromlen);

  if (rc == -1) {
	bb_perror_msg("recvfrom error: %s", inet_ntoa (q->srv->addr.sin_addr));
	return -1;
  } else if (rc > len) {
	log_msg (LOG_NOTICE, "packet too large: %s", inet_ntoa (q->srv->addr.sin_addr));
	return 0;
  } else if (memcmp (&from.sin_addr, &q->srv->addr.sin_addr, sizeof (from.sin_addr)) != 0) {
	log_msg (LOG_WARNING, "unexpected server: %s", inet_ntoa (from.sin_addr));
	return 0;
  }

  return (rc);
}

/* handle_udpreply()
 *
 * This function handles udp DNS requests by either replying to them (if we
 * know the correct reply via master, caching, etc.), or forwarding them to
 * an appropriate DNS server.
 *
 * Note that the mached query is prev->next and not prev. */
static void udp_handle_reply(query_t * prev)
{
  int len;
  unsigned addr_len;
  query_t *q = prev->next;

  /* handling socket q->sock */
  if ((len = reply_recv (q, reply_msg, UDP_MAXSIZE)) < 0) {
	/* dnsrecv failed: len */
	query_delete_next (prev);
	return;                                         /* recv error */
  }

  /* do basic checking */
  if (check_reply (q->srv, reply_msg, len) < 0) {
	/* check_reply failed */
	query_delete_next (prev);
	return;
  }

  addr_len = sizeof (struct sockaddr_in);

  /* was this a dummy reactivate query? */
  if (q->domain != NULL) {
	/* no, lets cache the reply and send to client */
	cache_dnspacket (reply_msg, len, q->srv);

	/* set the client qid */
	memcpy(reply_msg, &q->client_qid, sizeof(q->client_qid));
	/* Forwarding the reply to the host inet_ntoa (q->client.sin_addr) */
	if (sendto (isock, reply_msg, len, 0,
				(const struct sockaddr *) &q->client, addr_len) != len) {
	  /* sendto error */ ;
	}
  } else {
	/* We got a reactivation dummy reply. Cool! */ ;
  }

  /* this server is obviously alive, we reset the counters */
  q->srv->send_time = 0;
  if (q->srv->inactive)
	/* Reactivating server inet_ntoa (q->srv->addr.sin_addr */ ;
  q->srv->inactive = 0;
  /* remove query from list and destroy it */
  query_delete_next (prev);
}


/* send a dummy packet to a deactivated server to check if its back*/
static int udp_send_dummy(srvnode_t *s)
{
  static const unsigned char dnsbuf[] = {
	/* HEADER */
	/* we send a lookup for localhost */
	/* will this work on a big endian system? */
	0x00, 0x00,                                     /* ID */
	0x01, 0x00,                                     /* QR|OC|AA|TC|RD -  RA|Z|RCODE  */
	0x00, 0x01,                                     /* QDCOUNT */
	0x00, 0x00,                                     /* ANCOUNT */
	0x00, 0x00,                                     /* NSCOUNT */
	0x00, 0x00,                                     /* ARCOUNT */

	9, 'l', 'o', 'c', 'a', 'l', 'h', 'o', 's', 't', 0,      /* QNAME */
	0x00, 0x01,                                     /* QTYPE A record */
	0x00, 0x01                                      /* QCLASS: IN */
	  /* in case you want to lookup root servers instead, use this: */
	  /*    0x00,       *//* QNAME:  empty */
	  /*    0x00, 0x02, *//* QTYPE:  a authorative name server */
	  /*    0x00, 0x01  *//* QCLASS: IN */
  };
  query_t *q;
  struct sockaddr_in srcaddr;

  /* should not happen */
  assert (s != NULL);
  if(dnsbuf_dyn == NULL) {
	dnsbuf_dyn = xmalloc(sizeof(dnsbuf));
	memcpy(dnsbuf_dyn, dnsbuf, sizeof(dnsbuf));
  }
  if ((q = query_add (NULL, s, &srcaddr, dnsbuf_dyn, sizeof (dnsbuf))) != NULL) {
	int rc;
	q = q->next;                            /* query add returned the query 1 before in list */
	/* don't let those queries live too long */
	q->ttl = REACTIVATE_INTERVAL;
	memset (&srcaddr, 0, sizeof (srcaddr));
	/* Sending dummy id=%i to %s", ((unsigned short *) dnsbuf_dyn)[0], inet_ntoa (s->addr.sin_addr) */
	rc = udp_send (q->sock, s, dnsbuf_dyn, sizeof (dnsbuf));
	((unsigned short *) dnsbuf_dyn)[0]++;
	return rc;
  }
  return -1;
}

static void init_dnrd_uid(void)
{
  char *ep;

  /* find the uid */
  daemonuid = (uid_t) strtoul (dnrd_user, &ep, 10);
  if (dnrd_user[0] == '\0')            /* should never happen */
	bb_error_msg_and_die ("please specify a valid user with -u");

  if (*ep) {                                    /* dnrd_user is not numeric */
	struct passwd *pwent;
	if ((pwent = getpwnam (dnrd_user))) {
	  if ((daemonuid = pwent->pw_uid) == 0)
		bb_error_msg_and_die ("please specify a non-root user (uid != 0) with -u");
	  if (pwent->pw_gid == 0)
		bb_error_msg_and_die ("please specify a non-root user group (gid != 0)");
	} else {
	  bb_error_msg_and_die ("Could not become \"%s\" user. Please create the user "
			    "account or specify a valid user with  the -u option.", dnrd_user);
	}
  }

}


static void dnrd_root_sanity_check(void)
{
  DIR *dirp;
  int rslt;
  struct dirent *direntry;
  struct stat st;

  if (chdir (dnrd_root))
	bb_perror_msg_and_die ("could not chdir to %s", dnrd_root);

  dirp = opendir (dnrd_root);
  if (!dirp)
	bb_perror_msg_and_die ("the directory %s must be created before dnrd will run", dnrd_root);

  rslt = stat (dnrd_root, &st);
  if (st.st_uid != 0)
	bb_error_msg_and_die ("the %s directory must be owned by root", dnrd_root);
  if ((st.st_mode & (S_IWGRP | S_IWOTH)) != 0)
	bb_perror_msg_and_die ("the %s directory should only be user writable", dnrd_root);
  while ((direntry = readdir (dirp)) != NULL) {
	if (!strcmp (direntry->d_name, ".") || !strcmp (direntry->d_name, ".."))
	  continue;

	rslt = stat (direntry->d_name, &st);

	if (rslt)
	  continue;
	if (S_ISDIR (st.st_mode))
	  bb_error_msg_and_die ("The %s directory must not contain subdirectories", dnrd_root);
	if ((st.st_mode & (S_IXUSR | S_IXGRP | S_IXOTH | S_IWGRP | S_IWOTH)) != 0) {
	  bb_error_msg_and_die ("A file in %s has either execute "
					"permissions or non-user write permission.  Please do a "
					"\"chmod a-x,go-w\" on all files in this directory",
					dnrd_root);
	}
	if (st.st_uid != 0)
	  bb_error_msg_and_die ("All files in %s must be owned by root", dnrd_root);
  }
  closedir (dirp);
}

static void init_socket(void)
{
  recv_addr.sin_port = htons (port);

  /* Setup our DNS query reception socket. */
  isock = bb_xsocket(BB_XSOCKET_FLG_SOCK_DGRAM | BB_XSOCKET_FLG_ERR_DIE);
  setsockopt_1i(isock, SOL_SOCKET, SO_REUSEADDR, 1);

  bb_xbind(isock, (struct sockaddr *) &recv_addr, sizeof (recv_addr));

  /* Setup our DNS tcp proxy socket. */
  tcpsock = bb_xsocket(BB_XSOCKET_FLG_ERR_DIE);
  setsockopt_1i(tcpsock, SOL_SOCKET, SO_REUSEADDR, 1);
  bb_xbind(tcpsock, (struct sockaddr *) &recv_addr, sizeof (recv_addr));
  if (listen (tcpsock, 5) != 0)
	bb_perror_msg_and_die ("can't listen");
}

/* check if a pid is running
 * from the unix faq
 * http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC18 */
static int isrunning (int pid)
{
  if (kill (pid, 0)) {
	if (errno == EPERM)
	  return 1;
	 else
	  return 0;
  } else {
	return 1;
  }
}

/* wait_for_exit()
 *
 * In: pid     - the process id to wait for
 *     timeout - maximum time to wait in 1/100 secs
 *
 * Returns: 1 if it died in before timeout
 *
 * Abstract: Check if a process is running and wait til it does not
 */
static int wait_for_exit(int pid, int timeout)
{
  while (timeout--) {
	if (!isrunning (pid))
	  return 1;
	xusleep (10000);
  }
  /* ouch... we timed out */
  return 0;
}

/* kill_current()
 *
 * Returns: 1 if a currently running dnrd was found & killed, 0 otherwise.
 *
 * Abstract: This function sees if pid_file already exists and, if it does,
 *           will kill the current dnrd process and remove the file. */
static int kill_current(void)
{
  int pid;
  int retn;
  FILE *filep;

  filep = bb_xfopen(pid_file, "Ar");
  if ((retn = (fscanf (filep, "%i%*s", &pid) == 1))) {
	kill (pid, SIGTERM);
	/* dnrd gets 4 seconds to die or we give up */
	if (!wait_for_exit (pid, 400))
	  bb_error_msg("The dnrd process didn't die within 4 seconds");
  }
  fclose (filep);
  unlink (pid_file);
  return retn;
}


static void parse_args(void)
{
  int load_balance = 0;
  int c;
  char *a, *p, *mm, *ss, *u;

  USE_FEATURE_LONG_OPTIONS(bb_applet_long_options = long_options);
  c = bb_xgetopt_ulflags(opt_str, &a, &p, &mm, &ss, &dnrd_root, &u);
  if(c & OPT_LOCAL_ADDRESS) {
	if (!inet_aton (a, &recv_addr.sin_addr))
		bb_error_msg_and_die ("bad ip address \"%s\"", a);
  }
  if(c & OPT_PORT)
	port = bb_xparse_i_number(10 | BB_PN_ERR_STR1 | BB_PN_SET_LOWER1 | BB_PN_SET_UPPER,
					p, 65535, "bad port \"%s\"");
  load_balance = c & OPT_LOAD_BALANCE;
  foreground = c & OPT_FOREGROUND;
  ignore_inactive_cache_hits = c & OPT_IGNORE_CACHE;
  if(c & OPT_MAX_SOCK) /* Setting maximum number of open sockets to max_sockets */
	max_sockets = bb_xparse_i_number(10 | BB_PN_SET_LOWER1, mm);
  if(c & OPT_SERVER) {
	domnode_t *dp;
	char *s, *sep = strchr (ss, ':'), *sep2 = strrchr (ss, ':');
	int rport = 53;

	if(sep2 != NULL) {
		char *e;
		rport = strtol(sep2 + 1, &e, 10);
		if(e != (sep2 + 1) && *e == '\0') {
			if(rport <= 0 || rport > 65535)
				bb_error_msg_and_die ("bad port number \"%s\"", sep2 + 1);
			*sep2 = 0;
			if(sep == sep2)
				sep = NULL;
		} else {
			if(sep != sep2)
				bb_error_msg_and_die ("invalid format -s server:domain:port \"%s\"", ss);
			sep2 = NULL;
			rport = 53;
		}
	}
	if (sep) {                              /* is a domain specified? */
	  s = make_cname (strnlwr (sep + 1, 200), 200);
	  *sep = 0;
	  if ((dp = search_domnode (domain_list, s)) == NULL) {
		/* Added domain load balancing, load_balance ? "with" : "without" */
		dp = add_domain (domain_list, load_balance, s);
	  } else {
		free (s);
	  }
	} else
	  dp = domain_list;

	add_srv (last_srvnode (dp->srvlist), ss, rport);
	/* Server %s#%d added to domain %s", ss, rport, sep ? sep + 1 : "(default)" */
	if (dp->roundrobin != load_balance) {
	  dp->roundrobin = load_balance;
	  /* Turned on load balancing for domain %s", cname2asc (dp->domain, buf) */
	}
	if (sep2)
	  *sep2 = ':';
	if (sep)
	  *sep = ':';
  }
  if(c & OPT_KILL) {
	if (!kill_current())
		  printf ("No %s daemon found.  Exiting.\n", bb_applet_name);
	exit (0);
  }
  if(c & OPT_VERSION) {
	printf ("dnrd version %s\n\n", PACKAGE_VERSION);
	exit(0);
  }
  if(c & OPT_HELP) {
	bb_default_error_retval = 0;
	bb_show_usage();
  }
}

/* reactivate servers that have been inactive for delay seconds */
static void retry_srvlist(domnode_t * d, const int delay)
{
  time_t now = time (NULL);
  srvnode_t *s;
  assert (d != NULL);                   /* should never happen */
  s = d->srvlist;
  while ((s = s->next) && (s != d->srvlist))
	if (s->inactive && (now - s->inactive) >= delay) {
	  s->inactive = now;
	  udp_send_dummy (s);
	}
}

/* Check if any deactivated server are back online again */
static void reactivate_servers(int interval)
{
  time_t now = time (NULL);
  domnode_t *d = domain_list;

  if (!last_try_rs)
	last_try_rs = now;
  /* check for reactivate servers */
  if ((now - last_try_rs < interval))
	return;

  last_try_rs = now;
  do {
	if (!no_srvlist (d->srvlist))
	  retry_srvlist (d, interval);
  } while ((d = d->next) != domain_list);
}

/* run()
 *
 * Abstract: This function runs continuously, waiting for packets to arrive
 *           and processing them accordingly. */
static void ATTRIBUTE_NORETURN run(void)
{
  struct timespec tout;
  fd_set fdread;
  int retn;
  sigset_t orig_sigmask;

  FD_ZERO (&fdmaster);
  FD_SET (isock, &fdmaster);
  FD_SET (tcpsock, &fdmaster);
  maxsock = (tcpsock > isock) ? tcpsock : isock;

  init_sig_handler (&orig_sigmask);

  while (1) {
	query_t *q;
	tout.tv_sec = SELECT_TIMEOUT;
	tout.tv_nsec = 0;
	fdread = fdmaster;

	/* Wait for input or timeout */
	retn = pselect (maxsock + 1, &fdread, 0, 0, &tout, &orig_sigmask);

	/* reactivate servers */
	if (REACTIVATE_INTERVAL != 0) {
	  reactivate_servers (REACTIVATE_INTERVAL);
	  /* check if any server should be timed out */
	  if (FORWARD_TIMEOUT != 0)
		deactivate_servers (1);
	}

	/* Handle errors */
	if (retn < 0) {
	  if (errno != EINTR) {
		bb_perror_msg("select returned");
	  }
	  continue;
	} else if (retn != 0) {
	  for (q = &qlist; q->next != &qlist; q = q->next) {
		if (FD_ISSET (q->next->sock, &fdread)) {
		  udp_handle_reply (q);
		}
	  }

	  /* Check for incoming TCP requests */
	  if (FD_ISSET (tcpsock, &fdread))
		tcp_handle_request ();
	  /* Check for new DNS queries */
	  if (FD_ISSET (isock, &fdread)) {
		q = udp_handle_request ();
	  }
	}

	/* ok, we are done with replies and queries, lets do some
	   maintenance work */

	/* Expire lookups from the cache */
	cache_expire ();
	/* Remove old unanswered queries */
	query_timeout();

	/* print som query statestics */
	query_stats (0);
	srv_stats (10);
  }
}

/* main() - startup the program. */
int dnrd_main(void)
{
  domnode_t *p;

  dnrd_root = DNRD_ROOT;
  dnrd_user = "nobody";
  foreground = -1;
  max_sockets = 200;
  pool_ptr = QID_POOL_SIZE - 1;
  port = 53;
  /*
   * Initialization in common.h of recv_addr is broken, causing at
   * least the '-a' and -p switchs not to work.  Instead of assuming
   * positions of fields in the struct across platforms I thought it
   * safer to do a standard initialization in main().
   */
  recv_addr.sin_family = AF_INET;

  bb_openlog (LOG_PID, LOG_DAEMON);

  /* create the domain list */
  domain_list = alloc_domnode ();

  parse_args ();

  /* create pidfile name only */
  pid_file = xdaemon(XDAEMON_PIDFILE | XDAEMON_NOFORK | XDAEMON_NO_WRITE_PID);

  init_socket ();

  /* init the qid pool */
  qid_init_pool ();

  /* init query list */
  query_init ();

  /* Now it's time to become a daemon. */
  if (foreground < 0) {
	xdaemon(XDAEMON_PF_ARG, pid_file);
	umask (077);
	foreground = 0;
  }

  if(port < 1024) {
	/* we need to find the uid and gid from /etc/passwd before we chroot. */
	init_dnrd_uid();
	/* Change our root and current working directories to
	 * /usr/local/etc/dnrd.  Also, so some sanity checking on that
	 * directory first. */
	dnrd_root_sanity_check ();
	if (chroot (dnrd_root))
		bb_perror_msg_and_die ("couldn't chroot to %s", dnrd_root);

	/* drop supplementary groups */
	if (setgroups (0, NULL) < 0)
		bb_perror_msg_and_die ("can't drop supplementary groups");
	/* Switch uid/gid to something safer than root. */
	/* setting uid to daemonuid */
	if (setuid (daemonuid) < 0)
		bb_perror_msg_and_die ("could not switch to uid %i", daemonuid);
  }

  /* Setup our DNS query forwarding socket. */
  p = domain_list;
  do {
	/* set the first server as current */
	p->current = p->srvlist->next;
  } while ((p = p->next) != domain_list);

  /* Run forever. */
  run();
}
#endif /* APPLET_USAGE */
