/*
 * FTP User Program -- Command Interface.
 */

/*
 * busybox port by Vladimir Oleynik <dzo@simtreas.ru> (C) 2001-2003
 * - size optimization, memory usage reducind
 * - converting IAC to IAC+IAC and IAC+IAC to IAC, IAC+EOF correction
 * - remove potential overflow from makeargv()
 */

/*
 * Copyright (c) 1985, 1989 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */
#ifdef APPLET_USAGE
#define ftp_trivial_usage \
	"[-vdinp" USE_FEATURE_COMMAND_EDITING("e") "g] [host [port]]"
#define ftp_full_usage \
	"A FTP user program\n\n" \
	"Options:\n" \
	"\t-v\tVerbose mode\n" \
	"\t-d\tDebug mode\n" \
	"\t-i\tDisable interactive mode\n" \
	"\t-n\tDisable autologin mode\n" \
	"\t-p\tPassive mode\n" \
	USE_FEATURE_COMMAND_EDITING("\t-e\tDisable command line editing\n") \
	"\t-g\tDisable global metacharecters for local file names"
#else

#include "busybox.h"
#include "bb_inet.h"
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <paths.h>
#include <dirent.h>
#include <setjmp.h>

#include <sys/wait.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/file.h>

#include <glob.h>
#ifndef GLOB_BRACE
#define GLOB_BRACE 0
#endif
#ifndef GLOB_TILDE
#define GLOB_TILDE 0
#endif

#include "bb_sig.h"

#include <netinet/in.h>
#include <netinet/ip.h>

#ifdef HAVE_ARPA_FTP
#include <arpa/ftp.h>
#else
#include "ftp.h"
#endif
#ifdef HAVE_ARPA_TELNET
#include <arpa/telnet.h>
#else
#include "telnet.h"
#endif

#ifdef CONFIG_FEATURE_COMMAND_EDITING
#include "cmdedit.h"
#endif


#undef _PATH_TMP
#ifndef __BIONIC__
#define _PATH_TMP       "/tmp/ftpXXXXXX"
#else
/* Android do not have /tmp */
#define _PATH_TMP       "ftpXXXXXX"
#endif

#define NT_SIZE 17
struct comvars {
	int connected;                  /* connected to server */
	char hostname[MAXHOSTNAMELEN];  /* name of host connected to */
	struct sockaddr_in myctladdr;
	struct sockaddr_in hisctladdr;
	FILE *cin;
	int cin_fileno;
	FILE *cout;
	int cout_fileno;
	int type;                       /* requested file transfer type */
	int curtype;                    /* current file transfer type */
	char ntin[NT_SIZE];             /* input translation table */
	char ntout[NT_SIZE];            /* output translation table */
	char mapin[MAXPATHLEN];         /* input map template */
	char mapout[MAXPATHLEN];        /* output map template */
	char dotrans_domap_buf[PATH_MAX];
	char cpend;     /* flag: if != 0, then pending server reply */
	char sunique;   /* store files on server with unique name */
	char runique;   /* store local files with unique name */
	char mcase;     /* map upper to lower case for mget names */
	char ntflag;    /* use ntin ntout tables for name xlation */
	char mapflag;   /* use mapin mapout templates on file names */
};

/* We only support "file" structure, sorry. */
#define structname "file"
/* We only support non-print format, sorry. */
#define formname "non-print"
/* We only support stream mode, sorry. */
#define modename "stream"

struct macel {
	char mac_name[9];       /* macro name */
	char *mac_start;        /* start of macro in macbuf */
	char *mac_end;          /* end of macro in macbuf */
};


#define MACBUF_SIZE 4096

CALC_BSS struct BB_COMMON_BSS {
/*
 * Options and other state info.
 */
	int      traceflag;      /* trace packets exchanged */
	int      hash;           /* print # for each buffer transferred */
	int      tick;           /* print byte counter during transfers */
	int      verbose;        /* print messages coming back from server */
	int      fromatty;       /* input is from a terminal */
	int      debug;          /* debugging level */
	int      bell;           /* ring bell on cmd completion */
	int      proxy;          /* proxy server connection active */
	int      proxflag;       /* proxy connection exists */
	int      code;           /* return/reply code for ftp command */
	char     pasv[64];       /* passive port for proxy data connection */
	int      passivemode;    /* passive mode enabled */
	char     *altarg;        /* argv[1] with no shell-like preprocessing  */
	char     typename[32];           /* name of file transfer type */
	int      stru;                   /* file transfer structure */
	int      form;                   /* file transfer format */
	int      mode;                   /* file transfer mode */
	char     bytename[32];           /* local byte size in ascii */
	int      unix_server;    /* server is unix, can use binary for ascii */
	int      unix_proxy;     /* proxy is unix, can use binary for ascii */
	sigjmp_buf toplevel;     /* non-local goto stuff for cmd scanner */
	char     line[BUFSIZ];   /* input line buffer */
	char     *stringbase;    /* current scan point in line buffer */
	char     argbuf[200];    /* argument storage buffer */
	char     *argbase;       /* current storage point in arg buffer */
	int      mflag;          /* flag: if != 0, then active multi command */
	int      options;        /* used during socket creation */
	glob_t globbuf;
	struct comvars proxstruct;
	struct comvars o_proxstruct;
	struct comvars *cur;
	int macnum;                      /* number of defined macros */
	struct macel macros[16];
	char macbuf[MACBUF_SIZE];
	off_t restart_point;
	char reply_string[BUFSIZ];       /* last line of previous reply */

	const char *mname;
	sigjmp_buf jabort;
	sigjmp_buf abortprox;
	const char *globerr;
	int slrflag;
	char **rargv;
	int n_rargv;
	FILE *ftemp;
	char **remglob_args;
	struct sockaddr_in data_addr;
	int ptflag;
	sigjmp_buf recvabort;
	sigjmp_buf sendabort;
	sigjmp_buf ptabort;
	int ptabflg;
	int abrtflag;
	unsigned bufsize;
	char *recrq_buf;
	unsigned char  oline[BUFSIZ+16];     /* output line buffer + cmd */
	unsigned char  ailine[BUFSIZ*2+16];  /* output line buffer with coding IAC+IAC */
	/* Options and other state info. */
	int sendport;   /* use PORT cmd for each data connection */
	int interactive;/* interactively prompt on m* cmds */
	int doglob;     /* glob local file names */
	int crflag;     /* if 1, strip car. rets. on ascii gets */
	int data;
} bb_common_bss;
#define traceflag bb_common_bss.traceflag
#define hash bb_common_bss.hash
#define tick bb_common_bss.tick
#define verbose bb_common_bss.verbose
#define fromatty bb_common_bss.fromatty
#define debug bb_common_bss.debug
#define bell bb_common_bss.bell
#define proxy bb_common_bss.proxy
#define proxflag bb_common_bss.proxflag
#define code bb_common_bss.code
#define pasv bb_common_bss.pasv
#define passivemode bb_common_bss.passivemode
#define altarg bb_common_bss.altarg
#define typename bb_common_bss.typename
#define stru bb_common_bss.stru
#define form bb_common_bss.form
#define mode bb_common_bss.mode
#define bytename bb_common_bss.bytename
#define unix_server bb_common_bss.unix_server
#define unix_proxy bb_common_bss.unix_proxy
#define toplevel bb_common_bss.toplevel
#define line bb_common_bss.line
#define stringbase bb_common_bss.stringbase
#define argbuf bb_common_bss.argbuf
#define argbase bb_common_bss.argbase
#define mflag bb_common_bss.mflag
#define options bb_common_bss.options
#define globbuf bb_common_bss.globbuf
#define proxstruct bb_common_bss.proxstruct
#define o_proxstruct bb_common_bss.o_proxstruct
#define cur bb_common_bss.cur
#define macnum bb_common_bss.macnum
#define macros bb_common_bss.macros
#define macbuf bb_common_bss.macbuf
#define restart_point bb_common_bss.restart_point
#define reply_string bb_common_bss.reply_string
#define mname bb_common_bss.mname
#define jabort bb_common_bss.jabort
#define abortprox bb_common_bss.abortprox
#define globerr bb_common_bss.globerr
#define slrflag bb_common_bss.slrflag
#define rargv bb_common_bss.rargv
#define n_rargv bb_common_bss.n_rargv
#define ftemp bb_common_bss.ftemp
#define remglob_args bb_common_bss.remglob_args
#define data_addr bb_common_bss.data_addr
#define ptflag bb_common_bss.ptflag
#define recvabort bb_common_bss.recvabort
#define sendabort bb_common_bss.sendabort
#define ptabort bb_common_bss.ptabort
#define ptabflg bb_common_bss.ptabflg
#define abrtflag bb_common_bss.abrtflag
#define bufsize bb_common_bss.bufsize
#define recrq_buf bb_common_bss.recrq_buf
#define oline bb_common_bss.oline
#define ailine bb_common_bss.ailine
#define sendport bb_common_bss.sendport
#define interactive bb_common_bss.interactive
#define doglob bb_common_bss.doglob
#define crflag bb_common_bss.crflag
#define data bb_common_bss.data

/*
 * Format of command table.
 */
struct cmd {
	const char *c_str;
#define c_name(cmd_ptr) ((cmd_ptr->c_str)+4)
#ifdef CONFIG_FEATURE_FTP_HELP_MSG_SUPPORT
	const char *c_help;     /* help string */
#endif
#define c_bell          c_str[0]
#define c_conn          c_str[1]
#define c_proxy         c_str[2]
#define c_handler_type  c_str[3]  /* function to call type */

	void *c_handler;          /* function to call */
};


static char *hookup(const char *host, int port);
static const struct cmd *getcmd(const char *);
static char **makeargv(int *pargc, char **parg);
static int dologin(const char *host);
static int command(const char *fmt, ...);
static void sendrequest(const char *cmd, char *local, char *remote, int printnames);
static void recvrequest(const char *cmd, char *local, char *remote,
		 const char *lmode, int printnames);
static int another(int *pargc, char ***pargv, const char *prompt);
static int getreply(int expecteof);
static void domacro(int argc, char *argv[]);
static void pswitch(int flag);
static int xruserpass(const char *host, char **aname, char **apass, char **aacct);
static void setpeer(int argc, char *argv[]);
static void quit(void);
static void changetype(int newtype, int show);

static void settick(void);
static void disconnect(void);

static void setascii(void);
static void setbell(void);
static void setbinary(void);
static void setdebug(int argc, char *argv[]);
static void setform(void);
static void setglob(void);
static void sethash(void);
static void setmode(void);
static void setport(void);
static void setprompt(void);
static void setstruct(void);
static void settenex(void);
static void settrace(void);
static void settype(int argc, char *argv[]);
static void setverbose(void);
static void restart(int argc, char *argv[]);
static void reget(int argc, char *argv[]);
static void syst(void);
static void cd(int argc, char *argv[]);
static void lcd(int argc, char *argv[]);
static void delete_cmd(int argc, char *argv[]);
static void mdelete(int argc, char *argv[]);
static void user(int argc, char *argv[]);
static void ls(int argc, char *argv[]);
static void mls(int argc, char *argv[]);
static void get(int argc, char *argv[]);
static void mget(int argc, char *argv[]);
static void help(int argc, char *argv[]);
static void put(int argc, char *argv[]);
static void mput(int argc, char *argv[]);
static void renamefile(int argc, char *argv[]);
static void status(void);
static void quote(int argc, char *argv[]);
static void rmthelp(int argc, char *argv[]);
static void shell(const char *arg);
static void site(int argc, char *argv[]);
static void pwd(void);
static void makedir(int argc, char *argv[]);
static void removedir(int argc, char *argv[]);
static void setcr(void);
static void account(int argc, char *argv[]);
static void doproxy(int argc, char *argv[]);
static void reset(void);
static void setcase(void);
static void setntrans(int argc, char *argv[]);
static void setnmap(int argc, char *argv[]);
static void setsunique(void);
static void setrunique(void);
static void cdup(void);
static void macdef(int argc, char *argv[]);
static void sizecmd(int argc, char *argv[]);
static void modtime(int argc, char *argv[]);
static void newer(int argc, char *argv[]);
static void rmtstatus(int argc, char *argv[]);
static void do_chmod(int argc, char *argv[]);
static void do_umask(int argc, char *argv[]);
static void idle_cmd(int argc, char *argv[]);
static void setpassive(void);

static void intr(int);

static char *remglob(char *argv[], int doswitch, char buf[PATH_MAX]);
static int checkglob(int fd, const char *pattern);
static char *dotrans(char *name);
static char *domap(char *name);
static char *globulize(char *str);
static int confirm(const char *cmd, const char *file);
static int getit(int argc, char *argv[], int restartit, const char *modestr);
static void quote1(const char *initial, int argc, char **argv);

/*
 * User FTP -- Command Tables.
 */

#ifdef CONFIG_FEATURE_FTP_HELP_MSG_SUPPORT
static const char accounthelp[] = "send account command to remote server";
static const char appendhelp [] = "append to a file";
static const char asciihelp  [] = "set ascii transfer type";
static const char beephelp   [] = "beep when command completed";
static const char binaryhelp [] = "set binary transfer type";
static const char casehelp   [] = "toggle mget upper/lower case id mapping";
static const char cdhelp     [] = "change remote working directory";
static const char cduphelp   [] = "change remote working directory to parent directory";
static const char chmodhelp  [] = "change file permissions of remote file";
static const char connecthelp[] = "connect to remote ftp";
static const char crhelp     [] = "toggle carriage return stripping on ascii gets";
static const char deletehelp [] = "delete remote file";
static const char debughelp  [] = "toggle/set debugging mode";
static const char dirhelp    [] = "list contents of remote directory";
static const char disconhelp [] = "terminate ftp session";
static const char domachelp  [] = "execute macro";
static const char formhelp   [] = "set file transfer format";
static const char globhelp   [] = "toggle metacharacter expansion of local file names";
static const char hashhelp   [] = "toggle printing `#' for each buffer transferred";
static const char helphelp   [] = "print local help information";
static const char idlehelp   [] = "get (set) idle timer on remote side";
static const char lcdhelp    [] = "change local working directory";
static const char lshelp     [] = "list contents of remote directory";
static const char macdefhelp [] = "define a macro";
static const char mdeletehelp[] = "delete multiple files";
static const char mdirhelp   [] = "list contents of multiple remote directories";
static const char mgethelp   [] = "get multiple files";
static const char mkdirhelp  [] = "make directory on the remote machine";
static const char mlshelp    [] = "list contents of multiple remote directories";
static const char modtimehelp[] = "show last modification time of remote file";
static const char modehelp   [] = "set file transfer mode";
static const char mputhelp   [] = "send multiple files";
static const char newerhelp  [] = "get file if remote file is newer than local file ";
static const char nlisthelp  [] = "nlist contents of remote directory";
static const char nmaphelp   [] = "set templates for default file name mapping";
static const char ntranshelp [] = "set translation table for default file name mapping";
static const char passivehelp[] = "enter passive transfer mode";
static const char porthelp   [] = "toggle use of PORT cmd for each data connection";
static const char prompthelp [] = "force interactive prompting on multiple commands";
static const char proxyhelp  [] = "issue command on alternate connection";
static const char pwdhelp    [] = "print working directory on remote machine";
static const char quithelp   [] = "terminate ftp session and exit";
static const char quotehelp  [] = "send arbitrary ftp command";
static const char receivehelp[] = "receive file";
static const char regethelp  [] = "get file restarting at end of local file";
static const char remotehelp [] = "get help from remote server";
static const char renamehelp [] = "rename file";
static const char restarthelp[] = "restart file transfer at bytecount";
static const char rmdirhelp  [] = "remove directory on the remote machine";
static const char rmtstatushelp[]="show status of remote machine";
static const char runiquehelp[] = "toggle store unique for local files";
static const char resethelp  [] = "clear queued command replies";
static const char sendhelp   [] = "send one file";
static const char sitehelp   [] = "send site specific command to remote server\n"
	     "\t\tTry \"rhelp site\" or \"site help\" for more information";
static const char shellhelp  [] = "escape to the shell";
static const char sizecmdhelp[] = "show size of remote file";
static const char statushelp [] = "show current status";
static const char structhelp [] = "set file transfer structure";
static const char suniquehelp[] = "toggle store unique on remote machine";
static const char systemhelp [] = "show remote system type";
static const char tenexhelp  [] = "set tenex file transfer type";
static const char tickhelp   [] = "toggle printing byte counter during transfers";
static const char tracehelp  [] = "toggle packet tracing";
static const char typehelp   [] = "set file transfer type";
static const char umaskhelp  [] = "get (set) umask on remote side";
static const char userhelp   [] = "send new user information";
static const char verbosehelp[] = "toggle verbose mode";
#define H(help) help,
#else
#define H(help)
#endif

static const struct cmd cmdtab[] = {
	{ "\0\0\0\2!",          H(shellhelp)       shell },
	{ "\1\0\0\0$",          H(domachelp)       domacro },
	{ "\0\1\1\0account",    H(accounthelp)     account },
	{ "\1\1\1\0append",     H(appendhelp)      put },
	{ "\0\1\1\1ascii",      H(asciihelp)       setascii },
	{ "\0\0\0\1bell",       H(beephelp)        setbell },
	{ "\0\1\1\1binary",     H(binaryhelp)      setbinary, },
	{ "\0\0\0\1bye",        H(quithelp)        quit },
	{ "\0\0\1\1case",       H(casehelp)        setcase },
	{ "\0\1\1\0cd",         H(cdhelp)          cd },
	{ "\0\1\1\1cdup",       H(cduphelp)        cdup },
	{ "\0\1\1\0chmod",      H(chmodhelp)       do_chmod },
	{ "\0\1\1\1close",      H(disconhelp)      disconnect },
	{ "\0\0\0\1cr",         H(crhelp)          setcr },
	{ "\0\1\1\0delete",     H(deletehelp)      delete_cmd },
	{ "\0\0\0\0debug",      H(debughelp)       setdebug },
	{ "\1\1\1\0dir",        H(dirhelp)         ls },
	{ "\0\1\1\1disconnect", H(disconhelp)      disconnect },
	{ "\0\0\0\1exit",       H(quithelp)        quit },
	{ "\0\1\1\1form",       H(formhelp)        setform },
	{ "\1\1\1\0get",        H(receivehelp)     get },
	{ "\0\0\0\1glob",       H(globhelp)        setglob },
	{ "\0\0\0\1hash",       H(hashhelp)        sethash },
	{ "\0\0\1\0help",       H(helphelp)        help },
	{ "\0\1\1\0idle",       H(idlehelp)        idle_cmd },
	{ "\0\1\1\1image",      H(binaryhelp)      setbinary },
	{ "\0\0\0\0lcd",        H(lcdhelp)         lcd },
	{ "\1\1\1\0ls",         H(lshelp)          ls },
	{ "\0\0\0\0macdef",     H(macdefhelp)      macdef },
	{ "\1\1\1\0mdelete",    H(mdeletehelp)     mdelete },
	{ "\1\1\1\0mdir",       H(mdirhelp)        mls },
	{ "\1\1\1\0mget",       H(mgethelp)        mget },
	{ "\0\1\1\0mkdir",      H(mkdirhelp)       makedir },
	{ "\1\1\1\0mls",        H(mlshelp)         mls },
	{ "\0\1\1\1mode",       H(modehelp)        setmode },
	{ "\0\1\1\0modtime",    H(modtimehelp)     modtime },
	{ "\1\1\1\0mput",       H(mputhelp)        mput },
	{ "\1\1\1\0newer",      H(newerhelp)       newer },
	{ "\0\0\1\0nmap",       H(nmaphelp)        setnmap },
	{ "\1\1\1\0nlist",      H(nlisthelp)       ls },
	{ "\0\0\1\0ntrans",     H(ntranshelp)      setntrans },
	{ "\0\0\1\0open",       H(connecthelp)     setpeer },
	{ "\0\0\0\1prompt",     H(prompthelp)      setprompt },
	{ "\0\0\0\1passive",    H(passivehelp)     setpassive },
	{ "\0\0\1\0proxy",      H(proxyhelp)       doproxy },
	{ "\0\0\0\1sendport",   H(porthelp)        setport },
	{ "\1\1\1\0put",        H(sendhelp)        put },
	{ "\0\1\1\1pwd",        H(pwdhelp)         pwd },
	{ "\0\0\0\1quit",       H(quithelp)        quit },
	{ "\1\1\1\0quote",      H(quotehelp)       quote },
	{ "\1\1\1\0recv",       H(receivehelp)     get },
	{ "\1\1\1\0reget",      H(regethelp)       reget },
	{ "\0\1\1\0rstatus",    H(rmtstatushelp)   rmtstatus },
	{ "\0\1\1\0rhelp",      H(remotehelp)      rmthelp },
	{ "\0\1\1\0rename",     H(renamehelp)      renamefile },
	{ "\0\1\1\1reset",      H(resethelp)       reset },
	{ "\1\1\1\0restart",    H(restarthelp)     restart },
	{ "\0\1\1\0rmdir",      H(rmdirhelp)       removedir },
	{ "\0\0\1\1runique",    H(runiquehelp)     setrunique },
	{ "\1\1\1\0send",       H(sendhelp)        put },
	{ "\0\1\1\0site",       H(sitehelp)        site },
	{ "\1\1\1\0size",       H(sizecmdhelp)     sizecmd },
	{ "\0\0\1\1status",     H(statushelp)      status },
	{ "\0\1\1\1struct",     H(structhelp)      setstruct },
	{ "\0\1\1\1system",     H(systemhelp)      syst },
	{ "\0\0\1\1sunique",    H(suniquehelp)     setsunique },
	{ "\0\1\1\1tenex",      H(tenexhelp)       settenex },
	{ "\0\0\0\1tick",       H(tickhelp)        settick },
	{ "\0\0\0\1trace",      H(tracehelp)       settrace },
	{ "\0\1\1\0type",       H(typehelp)        settype },
	{ "\0\1\1\0user",       H(userhelp)        user },
	{ "\0\1\1\0umask",      H(umaskhelp)       do_umask },
	{ "\0\0\0\1verbose",    H(verbosehelp)     setverbose },
	{ "\0\0\1\0?",          H(helphelp)        help }
};

static const int  NCMDS = (sizeof (cmdtab) / sizeof (cmdtab[0])) - 1;


static char **ftpglob(const char *v);


static void lostpeer(int);

static void cmdscanner(int top);

/*
   printf("%?", bool) as
   printf("%s", (bool ? "on" : "off"))
*/

/* avoid gcc warning for format printf("%?", ...) */
static void printf_on_off(const char *s, ...)
{
	va_list p;
	int on_off;
	const char *str;

	va_start(p, s);
	while(*s != '\0') {
		if(*s != '%') {
			putchar(*s);
		} else {
			s++;
			if(*s == '?') {
				on_off = va_arg(p, int);
				str = on_off ? "on" : "off";
			} else /* 's' */
				str = va_arg(p, const char *);
			fputs(str, stdout);
		}
		s++;
	}
	va_end(p);
}

int
ftp_main(void)
{
	int top;
	unsigned long ch;

	cur = &proxstruct;
	sendport = -1;
	interactive = 1;
	doglob = 1;
	crflag = 1;
	data = -1;

	ch = bb_getopt_ulflags("vdinpge");
	verbose = ch & 1;
	if((ch & 2) != 0)
	    options |= SO_DEBUG;
	if((ch & 4) != 0)
	    interactive = 0;

/* if 0 establish user account on connection */
#define noautologin (bss_flags & 8)

	if((ch & 16) != 0)
	    passivemode = 1;
	if((ch & 32) != 0)
	    doglob = 0;

#ifdef CONFIG_FEATURE_COMMAND_EDITING
	/* disable readline support */
#define rl_inhibit (bss_flags & 64)
#endif

	fromatty = isatty(fileno(stdin));
	if (fromatty != 0)
		verbose++;

	if (*Gargv != NULL) {
		if (sigsetjmp(toplevel, 1))
			exit(0);
		bb_signal((1 << SIGINT), intr);
		bb_signal((1 << SIGPIPE), lostpeer);
		setpeer(Gargc + 1, Gargv - 1);
	}
	top = sigsetjmp(toplevel, 1) == 0;
	if (top) {
		bb_signal((1 << SIGINT), intr);
		bb_signal((1 << SIGPIPE), lostpeer);
	}
	for (;;) {
		cmdscanner(top);
		top = 1;
	}
}

static void
intr(int ignore)
{
	(void)ignore;
	siglongjmp(toplevel, 1);
}

static void
lostpeer(int ignore)
{
	(void)ignore;

	if (cur->connected) {
		if (cur->cout != NULL) {
			shutdown(cur->cout_fileno, 1+1);
			fclose(cur->cout);
			cur->cout = NULL;
		}
		if (data >= 0) {
			shutdown(data, 1+1);
			close(data);
			data = -1;
		}
		cur->connected = 0;
	}
	pswitch(1);
	if (cur->connected) {
		if (cur->cout != NULL) {
			shutdown(cur->cout_fileno, 1+1);
			fclose(cur->cout);
			cur->cout = NULL;
		}
		cur->connected = 0;
	}
	proxflag = 0;
	pswitch(0);
}

static char *get_input_line(char *buf, char *prompt)
{
#ifdef CONFIG_FEATURE_COMMAND_EDITING
	if (fromatty && !rl_inhibit) {
		if(cmdedit_read_input(prompt, buf)==0)
			intr(0);
		return buf;
	}
#endif
	if (fromatty) {
		printf("%s", prompt);
		fflush(stdout);
	}
	return fgets(buf, BUFSIZ, stdin);
}


static void c_handler(const struct cmd *c, int argc, char **argv, char *arg)
{
	if(c->c_handler_type == 0) {
		void (*c_handler_v)(int, char **);

		c_handler_v = c->c_handler;
		c_handler_v(argc, argv);
	}
	else if(c->c_handler_type == 1) {
		void (*c_handler_0)(void);

		c_handler_0 = c->c_handler;
		c_handler_0();
	}
	else if(c->c_handler_type == 2) {
		void (*c_handler_1)(const char *);

		c_handler_1 = c->c_handler;
		c_handler_1(arg);
	}
}

/*
 * Command parser.
 */
static void
cmdscanner(int top)
{
	int margc;
	char *marg;
	char **margv;
	const struct cmd *c;
	int l;

	if (!top)
		putchar('\n');
	for (;;) {
		if (!get_input_line(line, "ftp> ")) {
			quit();
		}
		l = strlen(line);
		if (l == 0)
			break;
		if (line[--l] == '\n') {
			if (l == 0)
				break;
			line[l] = '\0';
		}
		else if (l == sizeof(line) - 2) {
			puts("sorry, input line too long");
			while ((l = getchar()) != '\n' && l != EOF)
				/* void */;
			break;
		} /* else it was a line without a newline */
		margv = makeargv(&margc, &marg);
		if (margc == 0) {
			continue;
		}
		c = getcmd(margv[0]);
		if (c == (struct cmd *)-1) {
			puts("?Ambiguous command");
			continue;
		}
		if (c == NULL) {
			puts("?Invalid command");
			continue;
		}
		if (c->c_conn && !cur->connected) {
			puts("Not connected.");
			continue;
		}
		c_handler(c, margc, margv, marg);
		if (bell && c->c_bell) putchar('\007');
		if (c->c_handler != help)
			break;
	}
	bb_signal((1 << SIGINT), intr);
	bb_signal((1 << SIGPIPE), lostpeer);
}

static const struct cmd *
getcmd(const char *name)
{
	const char *p, *q;
	const struct cmd *c, *found;
	int nmatches, longest;

	longest = 0;
	nmatches = 0;
	found = 0;
	for (c = cmdtab; c < &cmdtab[NCMDS]; c++) {
		p = c_name(c);
		for (q = name; *q == *p++; q++)
			if (*q == 0)            /* exact match? */
				return (c);
		if (!*q) {                      /* the name was a prefix */
			if (q - name > longest) {
				longest = q - name;
				nmatches = 1;
				found = c;
			} else if (q - name == longest)
				nmatches++;
		}
	}
	if (nmatches > 1)
		return ((struct cmd *)-1);
	return (found);
}

/*
 * Parse string into argbuf;
 * implemented with FSM to
 * handle quoting and strings
 */
static char *
slurpstring(void)
{
	int got_one = 0;
	char *sb = stringbase;
	char *ap = argbase;
	char *tmp = argbase;            /* will return this if token found */

	if (*sb == '!' || *sb == '$') { /* recognize ! as a token for shell */
		switch (slrflag) {      /* and $ as token for macro invoke */
			case 0:
				slrflag++;
				stringbase++;
				return ((*sb == '!') ? "!" : "$");
				/* NOTREACHED */
			case 1:
				slrflag++;
				altarg = stringbase;
				break;
			default:
				break;
		}
	}

S0:
	switch (*sb) {

	case '\0':
		goto OUT;

	case ' ':
	case '\t':
		sb++; goto S0;

	default:
		switch (slrflag) {
			case 0:
				slrflag++;
				break;
			case 1:
				slrflag++;
				altarg = sb;
				break;
			default:
				break;
		}
		goto S1;
	}

S1:
	switch (*sb) {

	case ' ':
	case '\t':
	case '\0':
		goto OUT;       /* end of token */

	case '\\':
		sb++; goto S2;  /* slurp next character */

	case '"':
		sb++; goto S3;  /* slurp quoted string */

	default:
		*ap++ = *sb++;  /* add character to token */
		got_one = 1;
		goto S1;
	}

S2:
	switch (*sb) {

	case '\0':
		goto OUT;

	default:
		*ap++ = *sb++;
		got_one = 1;
		goto S1;
	}

S3:
	switch (*sb) {

	case '\0':
		goto OUT;

	case '"':
		sb++; goto S1;

	default:
		*ap++ = *sb++;
		got_one = 1;
		goto S3;
	}

OUT:
	if (got_one)
		*ap++ = '\0';
	argbase = ap;                   /* update storage pointer */
	stringbase = sb;                /* update scan pointer */
	if (got_one) {
		return(tmp);
	}
	switch (slrflag) {
		case 0:
			slrflag++;
			break;
		case 1:
			slrflag++;
			altarg = NULL;
			break;
		default:
			break;
	}
	return NULL;
}

/*
 * Slice a string up into argc/argv.
 */

static char **
makeargv(int *pargc, char **parg)
{
	int rargc = 0;
	char **argp;

	argp = rargv;
	stringbase = line;              /* scan from first of buffer */
	argbase = argbuf;               /* store from first of buffer */
	slrflag = 0;

	if(argp == NULL) {
l_realloc:
		n_rargv += 20;
		rargv = xrealloc(rargv, sizeof(char *) * n_rargv);
		argp = rargv+rargc;
	}

	while ((*argp++ = slurpstring())!=NULL) {
		rargc++;
		if(rargc >= n_rargv)
			goto l_realloc;
	}

	*pargc = rargc;
	if (parg) *parg = altarg;
	return rargv;
}


#define HELPINDENT ((int) sizeof ("directory"))

/*
 * Help command.
 * Call each command handler with argc == 0 and argv[0] == name.
 */
static void
help(int argc, char *argv[])
{
	const struct cmd *c;

	if (argc == 1) {
		int i, j;
		int columns, width = 0, lines;

		puts("Commands may be abbreviated.  Commands are:\n");
		for (c = cmdtab; c < &cmdtab[NCMDS]; c++) {
			int len = strlen(c_name(c));

			if (len > width)
				width = len;
		}
		width++;
		columns = 80 / width;
		if (columns == 0)
			columns = 1;
		lines = (NCMDS + columns - 1) / columns;
		for (i = 0; i < lines; i++) {
			for (j = 0; j < columns; j++) {
				c = cmdtab + j * lines + i;
				if (c >= &cmdtab[NCMDS]) {
					break;
				}
				printf("%-*s", width, ((!proxy || c->c_proxy) ?
							c_name(c) : " "));
			}
			puts("");
		}
		return;
	}
	while (--argc > 0) {
		char *arg;
		arg = *++argv;
		c = getcmd(arg);
		if (c == (struct cmd *)-1)
			printf("?Ambiguous help command %s\n", arg);
#ifdef CONFIG_FEATURE_FTP_HELP_MSG_SUPPORT
		else if (c == NULL)
			printf("?Invalid help command %s\n", arg);
		else
			printf("%-*s\t%s\n", HELPINDENT,
				c_name(c), c->c_help);
#else
		else
			printf("?This version do not support help messages\n");
#endif
	}
}
static int token(char *tokval, FILE *cfile);

#define DEFAULT 1
#define LOGIN   2
#define PASSWD  3
#define ACCOUNT 4
#define MACDEF  5
#define ID      10
#define MACH    11

static const struct toktab {
	const char *tokstr;
	/* int tval; */
} toktab[]= {
	{ "\1default"  /* ,    DEFAULT */ },
	{ "\2login"    /* ,    LOGIN   */ },
	{ "\3password" /* ,    PASSWD  */ },
	{ "\3passwd"   /* ,    PASSWD  */ },
	{ "\4account"  /* ,    ACCOUNT */ },
	{ "\13machine" /* ,    MACH    */ },
	{ "\5macdef"   /* ,    MACDEF  */ },
	{ NULL         /* ,    0       */  }
};

static int
xruserpass(const char *host, char **aname, char **apass, char **aacct)
{
	const char *hdir;
	char buf[BUFSIZ], *tmp;
	char myname[MAXHOSTNAMELEN];
	char tokval[100];
	const char *mydomain;
	int t, i, c, usedefault = 0;
	struct stat stb;
	FILE *cfile;

	hdir = getenv("HOME");
	if (hdir == NULL)
		hdir = ".";
	snprintf(buf, sizeof(buf), "%s/.netrc", hdir);
	cfile = fopen(buf, "r");
	if (cfile == NULL) {
		if (errno != ENOENT)
			bb_perror_msg("%s", buf);
		return(0);
	}
	if (gethostname(myname, sizeof(myname)) < 0)
		myname[0] = '\0';
	if ((mydomain = strchr(myname, '.')) == NULL)
		mydomain = "";
next:
	while ((t = token(tokval, cfile))) switch(t) {

	case DEFAULT:
		usedefault = 1;
		/* FALL THROUGH */

	case MACH:
		if (!usedefault) {
			if (token(tokval, cfile) != ID)
				continue;
			/*
			 * Allow match either for user's input host name
			 * or official hostname.  Also allow match of
			 * incompletely-specified host in local domain.
			 */
			if (strcasecmp(host, tokval) == 0)
				goto match;
			if (strcasecmp(cur->hostname, tokval) == 0)
				goto match;
			if ((tmp = strchr(cur->hostname, '.')) != NULL &&
			    strcasecmp(tmp, mydomain) == 0 &&
			    strncasecmp(cur->hostname, tokval, tmp-cur->hostname) == 0 &&
			    tokval[tmp - cur->hostname] == '\0')
				goto match;
			if ((tmp = strchr(host, '.')) != NULL &&
			    strcasecmp(tmp, mydomain) == 0 &&
			    strncasecmp(host, tokval, tmp - host) == 0 &&
			    tokval[tmp - host] == '\0')
				goto match;
			continue;
		}
	match:
		while ((t = token(tokval, cfile)) && t != MACH && t != DEFAULT)
		switch(t) {

		case LOGIN:
			if (token(tokval, cfile)) {
				if (*aname == 0) {
					*aname = xstrdup(tokval);
				} else {
					if (strcmp(*aname, tokval))
						goto next;
				}
			}
			break;
		case PASSWD:
			if (*aname==NULL) {
	fprintf(stderr, "Error: `password' must follow `login' in .netrc\n");
				goto bad;
			}
			if (strcmp(*aname, "anonymous") &&
			    fstat(fileno(cfile), &stb) >= 0 &&
			    (stb.st_mode & 077) != 0) {
	fprintf(stderr, "Error - .netrc file not correct permissions.\n"
			"Remove password or correct mode (should be 600).\n");
				goto bad;
			}
			if (token(tokval, cfile) && *apass == 0) {
				*apass = xstrdup(tokval);
			}
			break;
		case ACCOUNT:
			if (fstat(fileno(cfile), &stb) >= 0
			    && (stb.st_mode & 077) != 0) {
	fprintf(stderr, "Error - .netrc file not correct permissions.\n"
			"Remove account or correct mode (should be 600).\n");
				goto bad;
			}
			if (token(tokval, cfile) && *aacct == 0) {
				*aacct = xstrdup(tokval);
			}
			break;
		case MACDEF:
			if (proxy) {
				fclose(cfile);
				return(0);
			}
			while ((c=getc(cfile)) != EOF && (c == ' ' || c == '\t'));
			if (c == EOF || c == '\n') {
				puts("Missing macdef name argument.");
				goto bad;
			}
			if (macnum == 16) {
				puts("Limit of 16 macros have already been defined");
				goto bad;
			}
			tmp = macros[macnum].mac_name;
			*tmp++ = c;
			for (i=0; i < 8 && (c=getc(cfile)) != EOF &&
			    !isspace(c); ++i) {
				*tmp++ = c;
			}
			if (c == EOF) {
				puts("Macro definition missing null line terminator.");
				goto bad;
			}
			*tmp = '\0';
			if (c != '\n') {
				while ((c=getc(cfile)) != EOF && c != '\n');
			}
			if (c == EOF) {
				puts("Macro definition missing null line terminator.");
				goto bad;
			}
			if (macnum == 0) {
				macros[macnum].mac_start = macbuf;
			}
			else {
				macros[macnum].mac_start = macros[macnum-1].mac_end + 1;
			}
			tmp = macros[macnum].mac_start;
			while (tmp != macbuf + MACBUF_SIZE) {
				if ((c=getc(cfile)) == EOF) {
				puts("Macro definition missing null line terminator.");
					goto bad;
				}
				*tmp = c;
				if (*tmp == '\n') {
					if (*(tmp-1) == '\0') {
					   macros[macnum++].mac_end = tmp - 1;
					   break;
					}
					*tmp = '\0';
				}
				tmp++;
			}
			if (tmp == macbuf + MACBUF_SIZE) {
				puts("4K macro buffer exceeded");
				goto bad;
			}
			break;
		default:
	fprintf(stderr, "Unknown .netrc keyword %s\n", tokval);
			break;
		}
		goto done;
	}
done:
	fclose(cfile);
	return(0);
bad:
	fclose(cfile);
	return(-1);
}

static int
token(char *tokval, FILE *cfile)
{
	char *cp;
	int c;
	const struct toktab *t;

	if (feof(cfile))
		return 0;
	while ((c = getc(cfile)) != EOF &&
	    (c == '\n' || c == '\t' || c == ' ' || c == ','))
		continue;
	if (c == EOF)
		return 0;
	cp = tokval;
	if (c == '"') {
		while ((c = getc(cfile)) != EOF && c != '"') {
			if (c == '\\')
				c = getc(cfile);
			*cp++ = c;
		}
	} else {
		*cp++ = c;
		while ((c = getc(cfile)) != EOF
		    && c != '\n' && c != '\t' && c != ' ' && c != ',') {
			if (c == '\\')
				c = getc(cfile);
			*cp++ = c;
		}
	}
	*cp = 0;
	if (tokval[0] == 0)
		return 0;
	for (t = toktab; t->tokstr; t++)
		if (!strcmp(t->tokstr + 1, tokval))
			return (t->tokstr[0]);
	return (ID);
}

/*
 * FTP User Program -- Command Routines.
 */


/*
 * Tick counter step size.
 */
#define TICKBYTES     10240

/*
 * pipeprotect: protect against "special" local filenames by prepending
 * "./". Special local filenames are "-" and "|..." AND "/...".
 */
static char *pipeprotect(char *name)
{
	char *nu;
	if (strcmp(name, "-") && *name!='|' && *name!='/') {
		return name;
	}
	while(*name == '/')
		name++;

	/* We're going to leak this memory. XXX. */
	nu = xasprintf("./%s", name);
	return nu;
}

/*
 * Look for embedded ".." in a pathname and change it to "!!", printing
 * a warning.
 */
static char *pathprotect(char *name)
{
	int gotdots=0, i, len;

	/* Convert null terminator to trailing / to catch a trailing ".." */
	len = strlen(name)+1;
	name[len-1] = '/';

	/*
	 * State machine loop. gotdots is < 0 if not looking at dots,
	 * 0 if we just saw a / and thus might start getting dots,
	 * and the count of dots seen so far if we have seen some.
	 */
	for (i=0; i<len; i++) {
		if (name[i]=='.' && gotdots>=0) gotdots++;
		else if (name[i]=='/' && gotdots<0) gotdots=0;
		else if (name[i]=='/' && gotdots==2) {
		    printf("Warning: embedded .. in %.*s (changing to !!)\n",
			   len-1, name);
		    name[i-1] = '!';
		    name[i-2] = '!';
		    gotdots = 0;
		}
		else if (name[i]=='/') gotdots = 0;
		else gotdots = -1;
	}
	name[len-1] = 0;
	return name;
}


/*
 * `Another' gets another argument, and stores the new argc and argv.
 * It reverts to the top level (via main.c's intr()) on EOF/error.
 *
 * Returns false if no new arguments have been added.
 */
static int
another(int *pargc, char ***pargv, const char *prompt)
{
	int margc;
	char **margv;

	unsigned len = strlen(line);
	int ret;
	char prmtbuf[BUFSIZ];
	char addbufline[BUFSIZ];

	if (len >= sizeof(line) - 3) {
		puts("sorry, arguments too long");
		intr(0);
	}
	snprintf(prmtbuf, BUFSIZ, "(%s) ", prompt);
	line[len++] = ' ';
	line[len]=0;
	if(get_input_line(addbufline, prmtbuf)==0)
		intr(0);
	ret = strlen(addbufline);
	if(ret > 0 && addbufline[ret - 1] == '\n')
		addbufline[ret - 1] = '\0';
	for(ret=0; addbufline[ret]; ret++) {
		if(len >= (BUFSIZ-1))
			break;
		line[len++] = addbufline[ret];
	}
	line[len] = '\0';
	margv = makeargv(&margc, NULL);
	ret = margc > *pargc;
	*pargc = margc;
	*pargv = margv;
	return ret;
}

/*
 * Connect to peer server and
 * auto-login, if possible.
 */
static void
setpeer(int argc, char *argv[])
{
	char *host;
	unsigned short port;

	if (cur->connected) {
		printf("Already connected to %s, use close first.\n",
			cur->hostname);
		code = -1;
		return;
	}
	if (argc < 2)
		another(&argc, &argv, "to");
	if (argc < 2 || argc > 3) {
		printf("usage: %s host-name [port]\n", argv[0]);
		code = -1;
		return;
	}
	port = bb_lookup_tcp_port(argc > 2 ? argv[2] : "ftp", 21);
	if (argc > 2 && errno != 0) {
		printf("%s: bad port number-- %s\n", argv[1], argv[2]);
		printf ("usage: %s host-name [port]\n", argv[0]);
		code = -1;
		return;
	}
	host = hookup(argv[1], port);
	if (host) {
		int overbose;

		cur->connected = 1;
		/*
		 * Set up defaults for FTP.
		 */
		strcpy(typename, "ascii"), cur->type = TYPE_A;
		cur->curtype = TYPE_A;
		/* strcpy(formname, "non-print"),*/ form = FORM_N;
		/* strcpy(modename, "stream"), */   mode = MODE_S;
		/* strcpy(structname, "file"), */ stru = STRU_F;
		strcpy(bytename, "8");
		if (!noautologin)
			dologin(argv[1]);

/*
 * this ifdef is to keep someone form "porting" this to an incompatible
 * system and not checking this out. This way they have to think about it.
 */
		overbose = verbose;
		if (debug == 0)
			verbose = -1;
		if (command("SYST") == COMPLETE && overbose) {
			char *cp, c = 0;
			cp = strchr(reply_string+4, ' ');
			if (cp == NULL)
				cp = strchr(reply_string+4, '\r');
			if (cp) {
				if (cp[-1] == '.')
					cp--;
				c = *cp;
				*cp = '\0';
			}

			printf("Remote system type is %s.\n",
				reply_string+4);
			if (cp)
				*cp = c;
		}
		if (!strncmp(reply_string, "215 UNIX Type: L8", 17)) {
			if (proxy)
				unix_proxy = 1;
			else
				unix_server = 1;
			/*
			 * Set type to 0 (not specified by user),
			 * meaning binary by default, but don't bother
			 * telling server.  We can use binary
			 * for text files unless changed by the user.
			 */
			cur->type = 0;
			strcpy(typename, "binary");
			if (overbose)
			    printf("Using %s mode to transfer files.\n",
				typename);
		} else {
			if (proxy)
				unix_proxy = 0;
			else
				unix_server = 0;
			if (overbose &&
			    !strncmp(reply_string, "215 TOPS20", 10))
				puts(
"Remember to set tenex mode when transfering binary files from this machine.");
		}
		verbose = overbose;
	}
}

struct  types {
	const char *t_name;
	int t_type;
};

static const struct types types[] = {
	{ "Aascii",      TYPE_A },
	{ "Ibinary",     TYPE_I },
	{ "Iimage",      TYPE_I },
	{ "Eebcdic",     TYPE_E },
	{ "Ltenex",      TYPE_L },
	{ NULL,          0     }
};

/*
 * Set transfer type.
 */
static void
do_settype(const char *thetype)
{
	const struct types *p;
	int comret;

	for (p = types; p->t_name; p++)
		if (strcmp(thetype, p->t_name+1) == 0)
			break;
	if (p->t_name == 0) {
		printf("%s: unknown mode\n", thetype);
		code = -1;
		return;
	}
	if ((p->t_type == TYPE_L) && (bytename[0] != '\0'))
		comret = command("TYPE %c %s", p->t_name[0], bytename);
	else
		comret = command("TYPE %c", p->t_name[0]);
	if (comret == COMPLETE) {
		strcpy(typename, p->t_name+1);
		cur->curtype = cur->type = p->t_type;
	}
}

static void
settype(int argc, char *argv[])
{
	const struct types *p;
	if (argc > 2) {
		const char *sep;

		printf("usage: %s [", argv[0]);
		sep = " ";
		for (p = types; p->t_name; p++) {
			printf("%s%s", sep, p->t_name+1);
			sep = " | ";
		}
		puts(" ]");
		code = -1;
		return;
	}
	if (argc < 2) {
		printf("Using %s mode to transfer files.\n", typename);
		code = 0;
		return;
	}
	do_settype(argv[1]);
}

/*
 * Internal form of settype; changes current type in use with server
 * without changing our notion of the type for data transfers.
 * Used to change to and from ascii for listings.
 */
static void
changetype(int newtype, int show)
{
	const struct types *p;
	int comret, oldverbose = verbose;
	int oldtick = tick;

	if (newtype == 0)
		newtype = TYPE_I;
	if (newtype == cur->curtype)
		return;
	if (debug == 0 && show == 0)
		verbose = 0;
	tick = 0;
	for (p = types; p->t_name; p++)
		if (newtype == p->t_type)
			break;
	if (p->t_name == 0) {
		bb_error_msg("internal error: unknown type %d", newtype);
		return;
	}
	if (newtype == TYPE_L && bytename[0] != '\0')
		comret = command("TYPE %c %s", p->t_name[0], bytename);
	else
		comret = command("TYPE %c", p->t_name[0]);
	if (comret == COMPLETE)
		cur->curtype = newtype;
	verbose = oldverbose;
	tick = oldtick;
}

/*
 * Set binary transfer type.
 */
/*VARARGS*/
static void
setbinary(void)
{
	do_settype("binary");
}

/*
 * Set ascii transfer type.
 */
/*VARARGS*/
static void
setascii(void)
{
	do_settype("ascii");
}

/*
 * Set tenex transfer type.
 */
/*VARARGS*/
static void
settenex(void)
{
	do_settype("tenex");
}

/*
 * Set file transfer mode.
 */
/*ARGSUSED*/
static void
setmode(void)
{
	printf("We only support %s mode, sorry.\n", modename);
	code = -1;
}

/*
 * Set file transfer format.
 */
/*ARGSUSED*/
static void
setform(void)
{
	printf("We only support %s format, sorry.\n", formname);
	code = -1;
}

/*
 * Set file transfer structure.
 */
static void
setstruct(void)
{
	printf("We only support %s structure, sorry.\n", structname);
	code = -1;
}

/*
 * Send a single file.
 */
static void
put(int argc, char *argv[])
{
	const char *cmd;
	int loc = 0;
	char *oldargv1, *oldargv2;

	if (argc == 2) {
		argc++;
		argv[2] = argv[1];
		loc++;
	}
	if (argc < 2 && !another(&argc, &argv, "local-file"))
		goto usage;
	if (argc < 3 && !another(&argc, &argv, "remote-file")) {
usage:
		printf("usage: %s local-file remote-file\n", argv[0]);
		code = -1;
		return;
	}
	oldargv1 = argv[1];
	oldargv2 = argv[2];
	argv[1] = globulize(argv[1]);
	if (!argv[1]) {
		code = -1;
		return;
	}
	/*
	 * If "globulize" modifies argv[1], and argv[2] is a copy of
	 * the old argv[1], make it a copy of the new argv[1].
	 */
	if (argv[1] != oldargv1 && argv[2] == oldargv1) {
		argv[2] = argv[1];
	}
	cmd = (argv[0][0] == 'a') ? "APPE" : ((cur->sunique) ? "STOU" : "STOR");
	if (loc && cur->ntflag) {
		argv[2] = dotrans(argv[2]);
	}
	if (loc && cur->mapflag) {
		argv[2] = domap(argv[2]);
	}
	sendrequest(cmd, argv[1], argv[2],
	    argv[1] != oldargv1 || argv[2] != oldargv2);
}

static void mabort(int);

/*
 * Send multiple files.
 */
static void
mput(int argc, char *argv[])
{
	int i;
	void (*oldintr)(int);
	int ointer;
	char *tp;

	if (argc < 2 && !another(&argc, &argv, "local-files")) {
		printf("usage: %s local-files\n", argv[0]);
		code = -1;
		return;
	}
	mname = argv[0];
	mflag = 1;
	oldintr = bb_signal((1 << SIGINT), mabort);
	sigsetjmp(jabort, 1);
	if (proxy) {
		char *cp, *tp2, tmpbuf[PATH_MAX];
		char buf[PATH_MAX];

		while ((cp = remglob(argv, 0, buf)) != NULL) {
			if (*cp == 0) {
				mflag = 0;
				continue;
			}
			if (mflag && confirm(argv[0], cp)) {
				tp = cp;
				if (cur->mcase) {
					while (*tp && !islower(*tp)) {
						tp++;
					}
					if (!*tp) {
						tp = cp;
						tp2 = tmpbuf;
						while ((*tp2 = *tp) != '\0') {
						     if (isupper(*tp2)) {
							*tp2 = 'a' + *tp2 - 'A';
						     }
						     tp++;
						     tp2++;
						}
					}
					tp = tmpbuf;
				}
				if (cur->ntflag) {
					tp = dotrans(tp);
				}
				if (cur->mapflag) {
					tp = domap(tp);
				}
				sendrequest((cur->sunique) ? "STOU" : "STOR",
				    cp, tp, cp != tp || !interactive);
				if (!mflag && fromatty) {
					ointer = interactive;
					interactive = 1;
					if (confirm("Continue with", "mput")) {
						mflag++;
					}
					interactive = ointer;
				}
			}
		}
		bb_signal((1 << SIGINT), oldintr);
		mflag = 0;
		return;
	}
	for (i = 1; i < argc; i++) {
		char **cpp, **gargs;

		if (!doglob) {
			if (mflag && confirm(argv[0], argv[i])) {
				tp = (cur->ntflag) ? dotrans(argv[i]) : argv[i];
				tp = (cur->mapflag) ? domap(tp) : tp;
				sendrequest((cur->sunique) ? "STOU" : "STOR",
				    argv[i], tp, tp != argv[i] || !interactive);
				if (!mflag && fromatty) {
					ointer = interactive;
					interactive = 1;
					if (confirm("Continue with", "mput")) {
						mflag++;
					}
					interactive = ointer;
				}
			}
			continue;
		}
		gargs = ftpglob(argv[i]);
		if (globerr != NULL) {
			puts(globerr);
			continue;
		}
		for (cpp = gargs; cpp && *cpp != NULL; cpp++) {
			if (mflag && confirm(argv[0], *cpp)) {
				tp = (cur->ntflag) ? dotrans(*cpp) : *cpp;
				tp = (cur->mapflag) ? domap(tp) : tp;
				sendrequest((cur->sunique) ? "STOU" : "STOR",
				    *cpp, tp, *cpp != tp || !interactive);
				if (!mflag && fromatty) {
					ointer = interactive;
					interactive = 1;
					if (confirm("Continue with", "mput")) {
						mflag++;
					}
					interactive = ointer;
				}
			}
		}
		globfree(&globbuf);
	}
	bb_signal((1 << SIGINT), oldintr);
	mflag = 0;
}

static void
reget(int argc, char *argv[])
{
	getit(argc, argv, 1, "r+w");
}

static void
get(int argc, char *argv[])
{
	getit(argc, argv, 0, restart_point ? "r+w" : "w" );
}

/*
 * Receive one file.
 */

static void print_local_err(const char *local_act)
{
	fprintf(stderr, "local: %s: %m\n", local_act);
}

static int
getit(int argc, char *argv[], int restartit, const char *modestr)
{
	int loc = 0;
	char *oldargv1, *oldargv2;

	if (argc == 2) {
		argc++;
		/*
		 * Protect the user from accidentally retrieving special
		 * local names.
		 */
		argv[2] = pipeprotect(argv[1]);
		if (!argv[2]) {
			code = -1;
			return 0;
		}
		loc++;
	}
	if (argc < 2 && !another(&argc, &argv, "remote-file"))
		goto usage;
	if (argc < 3 && !another(&argc, &argv, "local-file")) {
usage:
		printf("usage: %s remote-file [ local-file ]\n", argv[0]);
		code = -1;
		return 0;
	}
	oldargv1 = argv[1];
	oldargv2 = argv[2];
	argv[2] = globulize(argv[2]);
	if (!argv[2]) {
		code = -1;
		return 0;
	}
	if (loc && cur->mcase) {
		char *tp = argv[1], *tp2, tmpbuf[PATH_MAX];

		while (*tp && !islower(*tp)) {
			tp++;
		}
		if (!*tp) {
			tp = argv[2];
			tp2 = tmpbuf;
			while ((*tp2 = *tp) != '\0') {
				if (isupper(*tp2)) {
					*tp2 = 'a' + *tp2 - 'A';
				}
				tp++;
				tp2++;
			}
			argv[2] = tmpbuf;
		}
	}
	if (loc && cur->ntflag)
		argv[2] = dotrans(argv[2]);
	if (loc && cur->mapflag)
		argv[2] = domap(argv[2]);
	if (restartit) {
		struct stat stbuf;
		int ret;

		ret = stat(argv[2], &stbuf);
		if (restartit == 1) {
			if (ret < 0) {
				print_local_err(argv[2]);
				return 0;
			}
			restart_point = stbuf.st_size;
		} else {
			if (ret == 0) {
				int overbose;

				overbose = verbose;
				if (debug == 0)
					verbose = -1;
				if (command("MDTM %s", argv[1]) == COMPLETE) {
					int yy, mo, day, hour, min, sec;
					struct tm *tm;
					time_t timep = stbuf.st_mtime;

					verbose = overbose;
					sscanf(reply_string,
					    "%*s %04d%02d%02d%02d%02d%02d",
					    &yy, &mo, &day, &hour, &min, &sec);
					tm = gmtime(&timep);
					tm->tm_mon++;
/* Indentation is misleading, but changes keep small. */
/*
 * I think the indentation and braces are now correct. Whoever put this
 * in the way it was originally should be prohibited by law.
 */
					if (tm->tm_year+1900 > yy)
						return 1;
					if (tm->tm_year+1900 == yy) {
					   if (tm->tm_mon > mo)
					      return 1;
					   if (tm->tm_mon == mo) {
					      if (tm->tm_mday > day)
						 return 1;
					      if (tm->tm_mday == day) {
						 if (tm->tm_hour > hour)
							return 1;
						 if (tm->tm_hour == hour) {
						    if (tm->tm_min > min)
						       return 1;
						    if (tm->tm_min == min) {
						       if (tm->tm_sec > sec)
							  return 1;
						    }
						 }
					      }
					   }
					}
				} else {
					puts(reply_string);
					verbose = overbose;
					return 0;
				}
			}
		}
	}

	recvrequest("RETR", argv[2], argv[1], modestr,
		    argv[1] != oldargv1 || argv[2] != oldargv2);
	restart_point = 0;
	return 0;
}

static void
mabort(int ignore)
{
	int ointer;

	(void)ignore;

	puts("");
	fflush(stdout);
	if (mflag && fromatty) {
		ointer = interactive;
		interactive = 1;
		if (confirm("Continue with", mname)) {
			interactive = ointer;
			siglongjmp(jabort,0);
		}
		interactive = ointer;
	}
	mflag = 0;
	siglongjmp(jabort,0);
}

/*
 * Get multiple files.
 */
static void
mget(int argc, char **argv)
{
	void (*oldintr)(int);
	int ointer;
	char *cp, *tp, *tp2, tmpbuf[PATH_MAX];
	char buf[PATH_MAX];

	if (argc < 2 && !another(&argc, &argv, "remote-files")) {
		printf("usage: %s remote-files\n", argv[0]);
		code = -1;
		return;
	}
	mname = argv[0];
	mflag = 1;
	oldintr = bb_signal((1 << SIGINT), mabort);
	sigsetjmp(jabort, 1);
	while ((cp = remglob(argv, proxy, buf)) != NULL) {
		if (*cp == '\0') {
			mflag = 0;
			continue;
		}
		if (mflag && confirm(argv[0], cp)) {
			tp = cp;
			if (cur->mcase) {
				while (*tp && !islower(*tp)) {
					tp++;
				}
				if (!*tp) {
					tp = cp;
					tp2 = tmpbuf;
					while ((*tp2 = *tp) != '\0') {
						if (isupper(*tp2)) {
							*tp2 = 'a' + *tp2 - 'A';
						}
						tp++;
						tp2++;
					}
				}
				tp = tmpbuf;
			}
			if (cur->ntflag) {
				tp = dotrans(tp);
			}
			if (cur->mapflag) {
				tp = domap(tp);
			}
			/* Reject embedded ".." */
			tp = pathprotect(tp);

			/* Prepend ./ to "-" or "!*" or leading "/" */
			tp = pipeprotect(tp);
			if (tp == NULL) {
				/* hmm... how best to handle this? */
				mflag = 0;
			}
			else {
				recvrequest("RETR", tp, cp, "w",
					    tp != cp || !interactive);
			}
			if (!mflag && fromatty) {
				ointer = interactive;
				interactive = 1;
				if (confirm("Continue with", "mget")) {
					mflag++;
				}
				interactive = ointer;
			}
		}
	}
	bb_signal((1 << SIGINT), oldintr);
	mflag = 0;
}

static char *
remglob(char *argv[], int doswitch, char buf[PATH_MAX])
{
	char temp[16];
	int oldverbose, oldhash, badglob = 0;
	char *cp;

	if (!mflag) {
		if (!doglob) {
			remglob_args = NULL;
		} else {
			if (ftemp) {
				fclose(ftemp);
				ftemp = NULL;
			}
		}
		return NULL;
	}
	if (!doglob) {
		if (remglob_args == NULL)
			remglob_args = argv;
		if ((cp = *++remglob_args) == NULL)
			remglob_args = NULL;
		return (cp);
	}
	if (ftemp == NULL) {
		int oldumask, fd;
		strcpy(temp, _PATH_TMP);

		/* libc 5.2.18 creates with mode 0666, which is dumb */
		oldumask = umask(077);
		fd = mkstemp(temp);
		umask(oldumask);

		if (fd<0) {
			printf("Error creating temporary file %s\n", temp);
			return NULL;
		}

		oldverbose = verbose, verbose = 0;
		oldhash = hash, hash = 0;
		if (doswitch) {
			pswitch(!proxy);
		}
		while (*++argv != NULL) {
			int     dupfd = dup(fd);

			recvrequest ("NLST", temp, *argv, "a", 0);
			if (!checkglob(dupfd, *argv)) {
				badglob = 1;
				break;
			}
		}
		unlink(temp);

		if (doswitch) {
			pswitch(!proxy);
		}
		verbose = oldverbose; hash = oldhash;
		if (badglob) {
			puts("Refusing to handle insecure file list");
			close(fd);
			return NULL;
		}
		ftemp = fdopen(fd, "r");
		if (ftemp == NULL) {
			puts("fdopen failed, oops");
			return NULL;
		}
		rewind(ftemp);
	}
	if (fgets(buf, PATH_MAX, ftemp) == NULL) {
		fclose(ftemp), ftemp = NULL;
		return (NULL);
	}
	strchr_set0(buf, '\n');
	return (buf);
}

/*
 * Check whether given pattern matches `..'
 * We assume only a glob pattern starting with a dot will match
 * dot entries on the server.
 */
static int
isdotdotglob(const char *pattern)
{
	int     havedot = 0;
	char    c;

	if (*pattern++ != '.')
		return 0;
	while ((c = *pattern++) != '\0' && c != '/') {
		if (c == '*' || c == '?')
			continue;
		if (c == '.' && havedot++)
			return 0;
	}
	return 1;
}

/*
 * This function makes sure the list of globbed files returned from
 * the server doesn't contain anything dangerous such as
 * /home/<yourname>/.forward, or ../.forward,
 * or |mail foe@doe </etc/passwd, etc.
 * Covered areas:
 *  -   returned name starts with / but glob pattern doesn't
 *  -   glob pattern starts with / but returned name doesn't
 *  -   returned name starts with |
 *  -   returned name contains .. in a position where glob
 *      pattern doesn't match ..
 *      I.e. foo/.* allows foo/../bar but not foo/.bar/../fly
 *
 * Note that globbed names starting with / should really be stored
 * under the current working directory; this is handled in mget above.
 *                                              --okir
 */
static int
checkglob(int fd, const char *pattern)
{
	const char      *sp;
	char            buffer[MAXPATHLEN], dotdot[MAXPATHLEN];
	int             okay = 1, nrslash, initial, nr;
	FILE            *fp;

	/* Find slashes in glob pattern, and verify whether component
	 * matches `..'
	 */
	initial = (pattern[0] == '/');
	for (sp = pattern, nrslash = 0; sp != 0; sp = strchr(sp, '/')) {
		while (*sp == '/')
			sp++;
		if (nrslash >= MAXPATHLEN) {
			printf("Incredible pattern: %s\n", pattern);
			return 0;
		}
		dotdot[nrslash++] = isdotdotglob(sp);
	}

	fp = fdopen(fd, "r");
	while (okay && fgets(buffer, sizeof(buffer), fp) != NULL) {
		char    *spt;

		if ((spt = strchr_set0(buffer, '\n')) == NULL) {
			printf("Extremely long filename from server: %s",
				buffer);
			okay = 0;
			break;
		}
		if (buffer[0] == '|'
		 || (buffer[0] != '/' && initial)
		 || (buffer[0] == '/' && !initial))
			okay = 0;
		for (spt = buffer, nr = 0; spt; spt = strchr(spt, '/'), nr++) {
			while (*spt == '/')
				spt++;
			if (spt[0] == '.' && !strncmp(spt, "../", 3)
			 && (nr >= nrslash || !dotdot[nr]))
				okay = 0;
		}
	}

	if (!okay)
		printf("Filename provided by server "
		       "doesn't match pattern `%s': %s\n", pattern, buffer);

	fclose(fp);
	return okay;
}


/*
 * Show status.
 */
static void
status(void)
{
	int i;

	if (cur->connected)
		printf("Connected to %s.\n", cur->hostname);
	else
		printf("Not connected.\n");
	if (!proxy) {
		pswitch(1);
		if (cur->connected) {
			printf("Connected for proxy commands to %s.\n", cur->hostname);
		}
		else {
			printf("No proxy connection.\n");
		}
		pswitch(0);
	}
	printf_on_off("Mode: %s; Type: %s; Form: %s; Structure: %s\n"
			"Verbose: %?; Bell: %?; Prompting: %?; Globbing: %?\n"
			"Store unique: %?; Receive unique: %?\n"
			"Case: %?; CR stripping: %?\n",
		modename, typename, formname, structname,
		verbose, bell, interactive, doglob,
		cur->sunique, cur->runique,
		cur->mcase, crflag);
	if (cur->ntflag) {
		printf("Ntrans: (in) %s (out) %s\n", cur->ntin, cur->ntout);
	} else {
		printf("Ntrans: off\n");
	}
	if (cur->mapflag) {
		printf("Nmap: (in) %s (out) %s\n", cur->mapin, cur->mapout);
	} else {
		printf("Nmap: off\n");
	}
	printf_on_off("Hash mark printing: %?; Use of PORT cmds: %?\n"
		"Tick counter printing: %?\n",
		hash, sendport, tick);
	if (macnum > 0) {
		puts("Macros:");
		for (i=0; i<macnum; i++) {
			printf("\t%s\n",macros[i].mac_name);
		}
	}
	code = 0;
}

/*
 * Set beep on cmd completed mode.
 */
static void
setbell(void)
{

	bell = !bell;
	printf_on_off("Bell mode %?.\n", bell);
	code = bell;
}

/*
 * Turn on packet tracing.
 */
static void
settrace(void)
{
	traceflag = !traceflag;
	printf_on_off("Packet tracing %?.\n", traceflag);
	code = traceflag;
}

/*
 * Toggle hash mark printing during transfers.
 */
static void
sethash(void)
{
	hash = !hash;
	if (hash && tick)
		settick();

	printf_on_off("Hash mark printing %?", hash);
	code = hash;
	if (hash)
		printf(" (%d bytes/hash mark)", 1024);
	printf(".\n");
}

/*
 * Toggle tick counter printing during transfers.
 */
static void
settick(void)
{
	tick = !tick;
	if (hash && tick)
		sethash();
	printf_on_off("Tick counter printing %?", tick);
	code = tick;
	if (tick)
		printf(" (%d bytes/tick increment)", TICKBYTES);
	printf(".\n");
}

/*
 * Turn on printing of server echos.
 */
static void
setverbose(void)
{
	verbose = !verbose;
	printf_on_off("Verbose mode %?.\n", verbose);
	code = verbose;
}

/*
 * Toggle PORT cmd use before each data connection.
 */
static void
setport(void)
{
	sendport = !sendport;
	printf_on_off("Use of PORT cmds %?.\n", sendport);
	code = sendport;
}

/*
 * Turn on interactive prompting
 * during mget, mput, and mdelete.
 */
static void
setprompt(void)
{
	interactive = !interactive;
	printf_on_off("Interactive mode %?.\n", interactive);
	code = interactive;
}

/*
 * Toggle metacharacter interpretation
 * on local file names.
 */
static void
setglob(void)
{
	doglob = !doglob;
	printf_on_off("Globbing %?.\n", doglob);
	code = doglob;
}

/*
 * Set debugging mode on/off and/or
 * set level of debugging.
 */
static void
setdebug(int argc, char *argv[])
{
	int val;

	if (argc > 1) {
		val = atoi(argv[1]);
		if (val < 0) {
			printf("%s: bad debugging value.\n", argv[1]);
			code = -1;
			return;
		}
	} else
		val = !debug;
	debug = val;
	if (debug)
		options |= SO_DEBUG;
	else
		options &= ~SO_DEBUG;
	printf_on_off("Debugging %?.\n", debug);
	code = debug > 0;
}

/*
 * Set current working directory
 * on remote machine.
 */
static void
cd(int argc, char *argv[])
{

	if (argc < 2 && !another(&argc, &argv, "remote-directory")) {
		printf("usage: %s remote-directory\n", argv[0]);
		code = -1;
		return;
	}
	if (command("CWD %s", argv[1]) == ERROR && code == 500) {
		if (verbose)
			puts("CWD command not recognized, trying XCWD");
		command("XCWD %s", argv[1]);
	}
}

/*
 * Set current working directory
 * on local machine.
 */
static void
lcd(int argc, char *argv[])
{
	char buf[PATH_MAX];
	const char *dir;

	if (argc == 1) {
	    dir = ".";
	}
	else if (argc != 2) {
		printf("usage: %s local-directory\n", argv[0]);
		code = -1;
		return;
	}
	else {
	    dir = globulize(argv[1]);
	}
	if (!dir) {
		code = -1;
		return;
	}
	if (chdir(dir) < 0) {
		print_local_err(dir);
		code = -1;
		return;
	}
	dir = buf;
	if (!getcwd(buf, sizeof(buf))) {
	    if (errno==ERANGE) dir = "<too long>";
	    else dir = "???";
	}
	printf("Local directory now %s\n", dir);
	code = 0;
}

/*
 * Delete a single file.
 */
static void
delete_cmd(int argc, char *argv[])
{

	if (argc < 2 && !another(&argc, &argv, "remote-file")) {
		printf("usage: %s remote-file\n", argv[0]);
		code = -1;
		return;
	}
	command("DELE %s", argv[1]);
}

/*
 * Delete multiple files.
 */
static void
mdelete(int argc, char *argv[])
{
	void (*oldintr)(int);
	int ointer;
	char *cp;
	char buf[PATH_MAX];

	if (argc < 2 && !another(&argc, &argv, "remote-files")) {
		printf("usage: %s remote-files\n", argv[0]);
		code = -1;
		return;
	}
	mname = argv[0];
	mflag = 1;
	oldintr = bb_signal((1 << SIGINT), mabort);
	sigsetjmp(jabort, 1);
	while ((cp = remglob(argv, 0, buf)) != NULL) {
		if (*cp == '\0') {
			mflag = 0;
			continue;
		}
		if (mflag && confirm(argv[0], cp)) {
			command("DELE %s", cp);
			if (!mflag && fromatty) {
				ointer = interactive;
				interactive = 1;
				if (confirm("Continue with", "mdelete")) {
					mflag++;
				}
				interactive = ointer;
			}
		}
	}
	bb_signal((1 << SIGINT), oldintr);
	mflag = 0;
}

/*
 * Rename a remote file.
 */
static void
renamefile(int argc, char *argv[])
{

	if (argc < 2 && !another(&argc, &argv, "from-name"))
		goto usage;
	if (argc < 3 && !another(&argc, &argv, "to-name")) {
usage:
		printf("%s from-name to-name\n", argv[0]);
		code = -1;
		return;
	}
	if (command("RNFR %s", argv[1]) == CONTINUE)
		command("RNTO %s", argv[2]);
}

/*
 * Get a directory listing
 * of remote files.
 */
static void
ls(int argc, char *argv[])
{
	const char *cmd;

	if (argc < 2) {
		argc++, argv[1] = NULL;
	}
	if (argc < 3) {
		argc++, argv[2] = "-";
	}
	if (argc > 3) {
		printf("usage: %s remote-directory local-file\n", argv[0]);
		code = -1;
		return;
	}
	cmd = argv[0][0] == 'n' ? "NLST" : "LIST";
	if (strcmp(argv[2], "-") && (argv[2] = globulize(argv[2]))==NULL) {
		code = -1;
		return;
	}
	if (strcmp(argv[2], "-") && *argv[2] != '|')
		if ((argv[2] = globulize(argv[2]))==NULL ||
		    !confirm("output to local-file:", argv[2])) {
			code = -1;
			return;
	}
	recvrequest(cmd, argv[2], argv[1], "w", 0);
}

/*
 * Get a directory listing
 * of multiple remote files.
 */
static void
mls(int argc, char *argv[])
{
	void (*oldintr)(int);
	int ointer, i;
	const char *volatile cmd;
	char *volatile dest;
	const char *modestr;

	if (argc < 2 && !another(&argc, &argv, "remote-files"))
		goto usage;
	if (argc < 3 && !another(&argc, &argv, "local-file")) {
usage:
		printf("usage: %s remote-files local-file\n", argv[0]);
		code = -1;
		return;
	}
	dest = argv[argc - 1];
	argv[argc - 1] = NULL;
	if (strcmp(dest, "-") && *dest != '|')
		if ((dest = globulize(dest))==NULL ||
		    !confirm("output to local-file:", dest)) {
			code = -1;
			return;
	}
	cmd = argv[0][1] == 'l' ? "NLST" : "LIST";
	mname = argv[0];
	mflag = 1;
	oldintr = bb_signal((1 << SIGINT), mabort);

	/*
	 * This just plain seems wrong.
	 */
	sigsetjmp(jabort, 1);

	for (i = 1; mflag && i < argc-1; ++i) {
		modestr = (i == 1) ? "w" : "a";
		recvrequest(cmd, dest, argv[i], modestr, 0);
		if (!mflag && fromatty) {
			ointer = interactive;
			interactive = 1;
			if (confirm("Continue with", argv[0])) {
				mflag ++;
			}
			interactive = ointer;
		}
	}
	bb_signal((1 << SIGINT), oldintr);
	mflag = 0;
}

/*
 * Do a shell escape
 */
static void
shell(const char *arg)
{
	int pid;
	void (*old1)(int);
	void (*old2)(int);

	old1 = bb_signal((1 << SIGINT), SIG_IGN);
	old2 = bb_signal((1 << SIGQUIT), SIG_IGN);
	if ((pid = fork()) == 0) {
		for (pid = 3; pid < 20; pid++)
			close(pid);
		bb_signal((1 << SIGINT) | (1 << SIGQUIT), SIG_DFL);
		run_shell(NULL, arg, NULL);
	}
	if (pid > 0) while (wait(NULL) != pid);

	bb_signal((1 << SIGINT), old1);
	bb_signal((1 << SIGQUIT), old2);
	if (pid == -1) {
		bb_perror_msg("Try again later");
		code = -1;
	}
	else {
		code = 0;
	}
}

/*
 * Send new user information (re-login)
 */
static void
user(int argc, char *argv[])
{
	char theacct[BUFSIZ];
	int n, aflag = 0;

	if (argc < 2)
		another(&argc, &argv, "username");
	if (argc < 2 || argc > 4) {
		printf("usage: %s username [password] [account]\n", argv[0]);
		code = -1;
		return;
	}
	n = command("USER %s", argv[1]);
	if (n == CONTINUE) {
		if (argc < 3 )
			argv[2] = getpass("Password: "), argc++;
		n = command("PASS %s", argv[2]);
	}
	if (n == CONTINUE) {
		if (argc < 4) {
			get_input_line(theacct, "Account: ");
			argv[3] = theacct; argc++;
		}
		n = command("ACCT %s", argv[3]);
		aflag++;
	}
	if (n != COMPLETE) {
		printf("Login failed.\n");
		return;
	}
	if (!aflag && argc == 4) {
		command("ACCT %s", argv[3]);
	}
}

/*
 * Print working directory.
 */
static void
pwd(void)
{
	int oldverbose = verbose;

	/*
	 * If we aren't verbose, this doesn't do anything!
	 */
	verbose = 1;
	if (command("PWD") == ERROR && code == 500) {
		puts("PWD command not recognized, trying XPWD");
		command("XPWD");
	}
	verbose = oldverbose;
}

/*
 * Make a directory.
 */
static void
makedir(int argc, char *argv[])
{

	if (argc < 2 && !another(&argc, &argv, "directory-name")) {
		printf("usage: %s directory-name\n", argv[0]);
		code = -1;
		return;
	}
	if (command("MKD %s", argv[1]) == ERROR && code == 500) {
		if (verbose)
			puts("MKD command not recognized, trying XMKD");
		command("XMKD %s", argv[1]);
	}
}

/*
 * Remove a directory.
 */
static void
removedir(int argc, char *argv[])
{

	if (argc < 2 && !another(&argc, &argv, "directory-name")) {
		printf("usage: %s directory-name\n", argv[0]);
		code = -1;
		return;
	}
	if (command("RMD %s", argv[1]) == ERROR && code == 500) {
		if (verbose)
			puts("RMD command not recognized, trying XRMD");
		command("XRMD %s", argv[1]);
	}
}

/*
 * Send a line, verbatim, to the remote machine.
 */
static void
quote(int argc, char *argv[])
{
	if (argc < 2 && !another(&argc, &argv, "command line to send")) {
		printf("usage: %s line-to-send\n", argv[0]);
		code = -1;
		return;
	}
	quote1("", argc, argv);
}

/*
 * Send a SITE command to the remote machine.  The line
 * is sent verbatim to the remote machine, except that the
 * word "SITE" is added at the front.
 */
static void
site(int argc, char *argv[])
{
	if (argc < 2 && !another(&argc, &argv, "arguments to SITE command")) {
		printf("usage: %s line-to-send\n", argv[0]);
		code = -1;
		return;
	}
	quote1("SITE ", argc, argv);
}

/*
 * Turn argv[1..argc) into a space-separated string, then prepend initial text.
 * Send the result as a one-line command and get response.
 */
static void
quote1(const char *initial, int argc, char **argv)
{
	int i, len;
	char buf[BUFSIZ];               /* must be >= sizeof(line) */

	strcpy(buf, initial);
	if (argc > 1) {
		len = strlen(buf);
		len += strlen(strcpy(&buf[len], argv[1]));
		for (i = 2; i < argc; i++) {
			buf[len++] = ' ';
			len += strlen(strcpy(&buf[len], argv[i]));
		}
	}
	if (command(buf) == PRELIM) {
		while (getreply(0) == PRELIM);
	}
}

static void
do_chmod(int argc, char *argv[])
{

	if (argc < 2 && !another(&argc, &argv, "mode"))
		goto usage;
	if (argc < 3 && !another(&argc, &argv, "file-name")) {
usage:
		printf("usage: %s mode file-name\n", argv[0]);
		code = -1;
		return;
	}
	command("SITE CHMOD %s %s", argv[1], argv[2]);
}

static void
do_umask(int argc, char *argv[])
{
	int oldverbose = verbose;

	verbose = 1;
	command(argc == 1 ? "SITE UMASK" : "SITE UMASK %s", argv[1]);
	verbose = oldverbose;
}

static void
idle_cmd(int argc, char *argv[])
{
	int oldverbose = verbose;

	verbose = 1;
	command(argc == 1 ? "SITE IDLE" : "SITE IDLE %s", argv[1]);
	verbose = oldverbose;
}

/*
 * Ask the other side for help.
 */
static void
rmthelp(int argc, char *argv[])
{
	int oldverbose = verbose;

	verbose = 1;
	command(argc == 1 ? "HELP" : "HELP %s", argv[1]);
	verbose = oldverbose;
}

/*
 * Terminate session and exit.
 */
static void
quit(void)
{

	if (cur->connected)
		disconnect();
	pswitch(1);
	if (cur->connected) {
		disconnect();
	}
	exit(0);
}

/*
 * Terminate session, but don't exit.
 */
static void
disconnect(void)
{
	if (!cur->connected)
		return;
	command("QUIT");
	if (cur->cout) {
		fclose(cur->cout);
	}
	cur->cout = NULL;
	cur->connected = 0;
	data = -1;
	if (!proxy) {
		macnum = 0;
	}
}

static int
confirm(const char *cmd, const char *file)
{
	char lyne[BUFSIZ];
	char lineread[BUFSIZ];

	if (!interactive)
		return 1;
	snprintf(lyne, BUFSIZ, "%s %s? ", cmd, file);
	if(get_input_line(lineread, lyne)==0)
		return 0;
	return (*lineread != 'n' && *lineread != 'N');
}

/*
 * Glob a local file name specification with
 * the expectation of a single return value.
 * Can't control multiple values being expanded
 * from the expression, we return only the first.
 */
static char *
globulize(char *cpp)
{
	char **globbed;
	char *rv = cpp;

	if (!doglob) return cpp;

	globbed = ftpglob(cpp);
	if (globerr != NULL) {
		printf("%s: %s\n", cpp, globerr);
		return NULL;
	}
	if (globbed) {
		rv = xstrdup(globbed[0]);
		globfree(&globbuf);
	}
	return rv;
}

static void
account(int argc, char *argv[])
{
	char buf[128], *ap;

	if (argc > 1) {
		*buf = 0;
		while (argc > 1) {
			--argc;
			++argv;
			strncat(buf, *argv, sizeof(buf)-strlen(buf));
			buf[sizeof(buf)-1] = 0;
		}
		ap = buf;
	}
	else {
		ap = getpass("Account:");
	}
	command("ACCT %s", ap);
}

static void
proxabort(int ignore)
{
	(void)ignore;

	if (!proxy) {
		pswitch(1);
	}
	if (cur->connected) {
		proxflag = 1;
	}
	else {
		proxflag = 0;
	}
	pswitch(0);
	siglongjmp(abortprox,1);
}

static void
doproxy(int argc, char *argv[])
{
	const struct cmd *c;
	void (*oldintr)(int);

	if (argc < 2 && !another(&argc, &argv, "command")) {
		printf("usage: %s command\n", argv[0]);
		code = -1;
		return;
	}
	c = getcmd(argv[1]);
	if (c == (struct cmd *) -1) {
		puts("?Ambiguous command");
		fflush(stdout);
		code = -1;
		return;
	}
	if (c == 0) {
		puts("?Invalid command");
		fflush(stdout);
		code = -1;
		return;
	}
	if (!c->c_proxy) {
		puts("?Invalid proxy command");
		fflush(stdout);
		code = -1;
		return;
	}
	if (sigsetjmp(abortprox, 1)) {
		code = -1;
		return;
	}
	oldintr = bb_signal((1 << SIGINT), proxabort);
	pswitch(1);
	if (c->c_conn && !cur->connected) {
		puts("Not connected");
		fflush(stdout);
		pswitch(0);
		bb_signal((1 << SIGINT), oldintr);
		code = -1;
		return;
	}
	c_handler(c, argc-1, argv+1, NULL);  /* should not reach this */
	if (cur->connected) {
		proxflag = 1;
	}
	else {
		proxflag = 0;
	}
	pswitch(0);
	bb_signal((1 << SIGINT), oldintr);
}

static void
setcase(void)
{
	cur->mcase = !cur->mcase;
	printf_on_off("Case mapping %?.\n", cur->mcase);
	code = cur->mcase;
}

static void
setcr(void)
{
	crflag = !crflag;
	printf_on_off("Carriage Return stripping %?.\n", crflag);
	code = crflag;
}

static void
setntrans(int argc, char *argv[])
{
	if (argc == 1) {
		code = cur->ntflag = 0;
		puts("Ntrans off.");
		return;
	}
	code = cur->ntflag = 1;
	safe_strncpy(cur->ntin, argv[1], NT_SIZE);
	if (argc == 2) {
		cur->ntout[0] = '\0';
		return;
	}
	safe_strncpy(cur->ntout, argv[2], NT_SIZE);
}

static char *
dotrans(char *name)
{
	int i, ostop, found;
	char *dotrans_new = proxstruct.dotrans_domap_buf;
	char *cp1, *cp2 = dotrans_new;

	for (ostop = 0; cur->ntout[ostop]; ostop++)
		;
	for (cp1 = name; *cp1; cp1++) {
		found = 0;
		for (i = 0; cur->ntin[i]; i++) {
			if (*cp1 == cur->ntin[i]) {
				found++;
				if (i < ostop) {
					*cp2++ = cur->ntout[i];
				}
				break;
			}
		}
		if (!found) {
			*cp2++ = *cp1;
		}
	}
	*cp2 = '\0';
	return(dotrans_new);
}

static void
setnmap(int argc, char *argv[])
{
	char *cp;

	if (argc == 1) {
		code = cur->mapflag = 0;
		puts("Nmap off.");
		return;
	}
	if (argc < 3 && !another(&argc, &argv, "mapout")) {
		printf("Usage: %s [mapin mapout]\n", argv[0]);
		code = -1;
		return;
	}
	code = cur->mapflag = 1;
	cp = strchr(altarg, ' ');
	if (proxy) {
		while(*++cp == ' ');
		altarg = cp;
		cp = strchr(altarg, ' ');
	}
	*cp = '\0';
	safe_strncpy(cur->mapin, altarg, MAXPATHLEN);
	while (*++cp == ' ');
	safe_strncpy(cur->mapout, cp, MAXPATHLEN);
}

static char *
domap(char *name)
{
	char *domap_new = o_proxstruct.dotrans_domap_buf;
	char *cp1 = name, *cp2 = cur->mapin;
	char *tp[9], *te[9];
	int i, toks[9], toknum = 0, match = 1;

	for (i=0; i < 9; ++i) {
		toks[i] = 0;
	}
	while (match && *cp1 && *cp2) {
		switch (*cp2) {
			case '\\':
				if (*++cp2 != *cp1) {
					match = 0;
				}
				break;
			case '$':
				if (*(cp2+1) >= '1' && *(cp2+1) <= '9') {
					if (*cp1 != *(++cp2+1)) {
						toknum = *cp2 - '1';
						toks[toknum]++;
						tp[toknum] = cp1;
						while (*++cp1 && *(cp2+1)
							!= *cp1);
						te[toknum] = cp1;
					}
					cp2++;
					break;
				}
				/* FALLTHROUGH */
			default:
				if (*cp2 != *cp1) {
					match = 0;
				}
				break;
		}
		if (match && *cp1) {
			cp1++;
		}
		if (match && *cp2) {
			cp2++;
		}
	}
	if (!match && *cp1) /* last token mismatch */
	{
		toks[toknum] = 0;
	}
	cp1 = domap_new;
	*cp1 = '\0';
	cp2 = cur->mapout;
	while (*cp2) {
		match = 0;
		switch (*cp2) {
			case '\\':
				if (*(cp2 + 1)) {
					*cp1++ = *++cp2;
				}
				break;
			case '[':
LOOP:
				if (*++cp2 == '$' && isdigit(*(cp2+1))) {
					if (*++cp2 == '0') {
						char *cp3 = name;

						while (*cp3) {
							*cp1++ = *cp3++;
						}
						match = 1;
					}
					else if (toks[toknum = *cp2 - '1']) {
						char *cp3 = tp[toknum];

						while (cp3 != te[toknum]) {
							*cp1++ = *cp3++;
						}
						match = 1;
					}
				}
				else {
					while (*cp2 && *cp2 != ',' &&
					    *cp2 != ']') {
						if (*cp2 == '\\') {
							cp2++;
						}
						else if (*cp2 == '$' &&
							isdigit(*(cp2+1))) {
							if (*++cp2 == '0') {
							   char *cp3 = name;

							   while (*cp3) {
								*cp1++ = *cp3++;
							   }
							}
							else if (toks[toknum =
							    *cp2 - '1']) {
							   char *cp3=tp[toknum];

							   while (cp3 !=
								  te[toknum]) {
								*cp1++ = *cp3++;
							   }
							}
						}
						else if (*cp2) {
							*cp1++ = *cp2++;
						}
					}
					if (!*cp2) {
						puts("nmap: unbalanced brackets");
						return(name);
					}
					match = 1;
					cp2--;
				}
				if (match) {
					while (*++cp2 && *cp2 != ']') {
					      if (*cp2 == '\\' && *(cp2 + 1)) {
							cp2++;
					      }
					}
					if (!*cp2) {
						puts("nmap: unbalanced brackets");
						return(name);
					}
					break;
				}
				switch (*++cp2) {
					case ',':
						goto LOOP;
					case ']':
						break;
					default:
						cp2--;
						goto LOOP;
				}
				break;
			case '$':
				if (isdigit(*(cp2 + 1))) {
					if (*++cp2 == '0') {
						char *cp3 = name;

						while (*cp3) {
							*cp1++ = *cp3++;
						}
					}
					else if (toks[toknum = *cp2 - '1']) {
						char *cp3 = tp[toknum];

						while (cp3 != te[toknum]) {
							*cp1++ = *cp3++;
						}
					}
					break;
				}
				/* intentional drop through */
			default:
				*cp1++ = *cp2;
				break;
		}
		cp2++;
	}
	*cp1 = '\0';
	if (!*domap_new) {
		return(name);
	}
	return(domap_new);
}

static void
setsunique(void)
{
	cur->sunique = !cur->sunique;
	printf_on_off("Store unique %?.\n", cur->sunique);
	code = cur->sunique;
}

static void
setrunique(void)
{
	cur->runique = !cur->runique;
	printf_on_off("Receive unique %?.\n", cur->runique);
	code = cur->runique;
}

/* change directory to parent directory */
static void
cdup(void)
{
	if (command("CDUP") == ERROR && code == 500) {
		if (verbose)
			puts("CDUP command not recognized, trying XCUP");
		command("XCUP");
	}
}

/* restart transfer at specific point */
static void
restart(int argc, char *argv[])
{
	if (argc != 2)
		puts("restart: offset not specified");
	else {
#if defined(NATIVE_64) || !defined(CONFIG_LFS)
		restart_point = atol(argv[1]);
#else
		restart_point = atoll(argv[1]);
#endif
		printf("restarting at %" P_OFFT "d. %s\n", restart_point,
		    "execute get, put or append to initiate transfer");
	}
}

/* show remote system type */
static void
syst(void)
{
	command("SYST");
}

static void
macdef(int argc, char *argv[])
{
	char *tmp;
	int c;

	if (macnum == 16) {
		puts("Limit of 16 macros have already been defined");
		code = -1;
		return;
	}
	if (argc < 2 && !another(&argc, &argv, "macro name")) {
		printf("Usage: %s macro_name\n",argv[0]);
		code = -1;
		return;
	}
	if (interactive) {
		puts("Enter macro line by line, terminating it with a null line");
	}
	safe_strncpy(macros[macnum].mac_name, argv[1],
						sizeof(macros[0].mac_name));
	if (macnum == 0) {
		macros[macnum].mac_start = macbuf;
	}
	else {
		macros[macnum].mac_start = macros[macnum - 1].mac_end + 1;
	}
	tmp = macros[macnum].mac_start;
	/* stepping over the end of the array, remember to take away 1! */
	while (tmp != macbuf+MACBUF_SIZE) {
		if ((c = getchar()) == EOF) {
			puts("macdef: end of file encountered");
			code = -1;
			return;
		}
		if ((*tmp = c) == '\n') {
			if (tmp == macros[macnum].mac_start) {
				macros[macnum++].mac_end = tmp;
				code = 0;
				return;
			}
			if (*(tmp-1) == '\0') {
				macros[macnum++].mac_end = tmp - 1;
				code = 0;
				return;
			}
			*tmp = '\0';
		}
		tmp++;
	}
	while (1) {
		while ((c = getchar()) != '\n' && c != EOF)
			/* LOOP */;
		if (c == EOF || getchar() == '\n') {
			puts("Macro not defined - 4k buffer exceeded");
			code = -1;
			return;
		}
	}
}

/*
 * Start up passive mode interaction
 */
static void
setpassive(void)
{
	passivemode = !passivemode;
	printf_on_off("Passive mode %?.\n", passivemode);
	code = passivemode;
}

/*
 * get size of file on remote machine
 */
static void
sizecmd(int argc, char *argv[])
{

	if (argc < 2 && !another(&argc, &argv, "filename")) {
		printf("usage: %s filename\n", argv[0]);
		code = -1;
		return;
	}
	command("SIZE %s", argv[1]);
}

/*
 * get last modification time of file on remote machine
 */
static void
modtime(int argc, char *argv[])
{
	int overbose;

	if (argc < 2 && !another(&argc, &argv, "filename")) {
		printf("usage: %s filename\n", argv[0]);
		code = -1;
		return;
	}
	overbose = verbose;
	if (debug == 0)
		verbose = -1;
	if (command("MDTM %s", argv[1]) == COMPLETE) {
		int yy, mo, day, hour, min, sec;
		sscanf(reply_string, "%*s %04d%02d%02d%02d%02d%02d", &yy, &mo,
			&day, &hour, &min, &sec);
		/* might want to print this in local time */
		printf("%s\t%02d/%02d/%04d %02d:%02d:%02d GMT\n", argv[1],
			mo, day, yy, hour, min, sec);
	} else
		puts(reply_string);
	verbose = overbose;
}

/*
 * show status on remote machine
 */
static void
rmtstatus(int argc, char *argv[])
{
	command(argc > 1 ? "STAT %s" : "STAT" , argv[1]);
}

/*
 * get file if modtime is more recent than current file
 */
static void
newer(int argc, char *argv[])
{
	if (getit(argc, argv, -1, "w")) {
		/* This should be controlled by some verbose flag */
		printf("Local file \"%s\" is newer than remote file \"%s\"\n",
			argv[2], argv[1]);
	}
}


static char *gunique(char *, char gunique_buf[MAXPATHLEN]);
static void proxtrans(const char *cmd, char *local, char *remote);
static int initconn(void);
static void ptransfer(const char *direction, long bytes,
		      const struct timeval *t0,
		      const struct timeval *t1);
static void abort_remote(FILE *din);

static char *
hookup(const char *host, int port)
{
	struct in_addr *addr_buf;
	char *canon_name;
	int naddr = 0, na = 0;

	int s;
	socklen_t len;
	struct sockaddr_in hi;  /* sizeof only */

	memset(&(cur->hisctladdr), 0, sizeof(hi));
	cur->hisctladdr.sin_family = AF_INET;

	addr_buf = bb_gethostbyname(BB_GHBN_NADDR | BB_GHBN_CANON_NAME, host, &naddr, &canon_name);
	if (addr_buf == NULL) {
		code = -1;
		return((char *) 0);
	}
	memcpy(&(cur->hisctladdr.sin_addr), addr_buf, sizeof(struct in_addr));
	safe_strncpy(cur->hostname, canon_name, MAXHOSTNAMELEN);
	free(canon_name);

	s = bb_xsocket(0);
	if (s < 0) {
		code = -1;
		return 0;
	}
	cur->hisctladdr.sin_port = port;
	while (connect(s, (struct sockaddr *)&(cur->hisctladdr), sizeof (hi)) < 0) {
		if (naddr > 1) {
			bb_perror_msg("connect to address %s",
				inet_ntoa(cur->hisctladdr.sin_addr));
			naddr--;
			na++;
			memcpy(&(cur->hisctladdr.sin_addr), addr_buf + na, sizeof(struct in_addr));
			printf("Trying %s...\n",
				inet_ntoa(cur->hisctladdr.sin_addr));
		}
		bb_perror_msg("connect");
		code = -1;
		goto bad;
	}
	len = sizeof (hi);
	if (getsockname(s, (struct sockaddr *)&(cur->myctladdr), &len) < 0) {
		bb_perror_msg("getsockname");
		code = -1;
		goto bad;
	}
#ifdef IP_TOS
	if (setsockopt_1i(s, IPPROTO_IP, IP_TOS, IPTOS_LOWDELAY) < 0)
		bb_perror_msg("setsockopt TOS (ignored)");
#endif
	cur->cin = fdopen(s, "r");
	cur->cout = fdopen(s, "w");
	if (cur->cin == NULL || cur->cout == NULL) {
		bb_error_msg("fdopen failed.");
		if (cur->cin)
			fclose(cur->cin);
		if (cur->cout)
			fclose(cur->cout);
		code = -1;
		goto bad;
	}
	cur->cin_fileno = fileno(cur->cin);
	cur->cout_fileno = fileno(cur->cout);
	if (verbose)
		printf("Connected to %s.\n", cur->hostname);
	if (getreply(0) > 2) {  /* read startup message from server */
		if (cur->cin)
			fclose(cur->cin);
		if (cur->cout)
			fclose(cur->cout);
		code = -1;
		goto bad;
	}
#ifdef SO_OOBINLINE
	if (setsockopt_1i(s, SOL_SOCKET, SO_OOBINLINE, 1) < 0 && debug)
		bb_perror_msg("setsockopt");
#endif /* SO_OOBINLINE */

	return (cur->hostname);
bad:
	close(s);
	return ((char *)0);
}

static int
dologin(const char *host)
{
	char tmp[BUFSIZ];
	char prompt[BUFSIZ];
	char *luser, *pass, *zacct;
	int n, aflag = 0;

	luser = pass = zacct = 0;
	if (xruserpass(host, &luser, &pass, &zacct) < 0) {
		code = -1;
		return(0);
	}
	while (luser == NULL) {
		char *myname = getlogin();

		if (myname == NULL)
			myname = bb_getpwuid(NULL, getuid(), 0);
		if (myname)
			snprintf(prompt, BUFSIZ, "Name (%s:%s): ", host, myname);
		else
			snprintf(prompt, BUFSIZ, "Name (%s): ", host);
		if(get_input_line(tmp, prompt)==0) {
			bb_error_msg("\nLogin failed.");
			return 0;
		}
		tmp[strlen(tmp) - 1] = '\0';
		if (*tmp == '\0')
			luser = myname;
		else
			luser = tmp;
	}
	n = command("USER %s", luser);
	if (n == CONTINUE) {
		if (pass == NULL) {
			/* fflush(stdout); */
			pass = getpass("Password:");
		}
		n = command("PASS %s", pass);
	}
	if (n == CONTINUE) {
		aflag++;
		/* fflush(stdout); */
		zacct = getpass("Account:");
		n = command("ACCT %s", zacct);
	}
	if (n != COMPLETE) {
		fprintf(stderr, "Login failed.\n");
		return 0;
	}
	if (!aflag && zacct != NULL)
		command("ACCT %s", zacct);
	if (proxy)
		return(1);
	for (n = 0; n < macnum; ++n) {
		if (!strcmp("init", macros[n].mac_name)) {
			int margc;
			char **margv;
			strcpy(line, "$init");
			margv = makeargv(&margc, NULL);
			domacro(margc, margv);
			break;
		}
	}
	return 1;
}


static void
cmdabort(int ignore)
{
	(void)ignore;

	puts("");
	fflush(stdout);
	abrtflag++;
	if (ptflag) siglongjmp(ptabort,1);
}

static int
command(const char *fmt, ...)
{
	va_list ap;
	int r;
	void (*oldintr)(int);
	unsigned char *optr, *chkiac;

	abrtflag = 0;
	va_start(ap, fmt);
	vsnprintf((char *)oline, sizeof(oline), fmt, ap);
	va_end(ap);
	optr = oline;
	if (debug) {
		if (strncmp("PASS ", fmt, 5) == 0)
			chkiac = (unsigned char *)"PASS XXXX";
		else
			chkiac = oline;
		printf("---> %s\n", chkiac);
		fflush(stdout);
	}
	if (cur->cout == NULL) {
		bb_perror_msg ("No control connection for command");
		code = -1;
		return 0;
	}
	oldintr = bb_signal((1 << SIGINT), cmdabort);
	chkiac = ailine;
	while(*optr) {
		if(*optr == IAC)
			*chkiac++ = IAC;
		*chkiac++ = *optr++;
	}
	*chkiac = 0;
	fprintf(cur->cout, "%s\r\n", ailine);
	fflush(cur->cout);
	cur->cpend = 1;
	r = getreply(!strcmp(fmt, "QUIT"));
	if (abrtflag && oldintr != SIG_IGN)
		(*oldintr)(SIGINT);
	bb_signal((1 << SIGINT), oldintr);
	return(r);
}


static int
getreply(int expecteof)
{
	int c, n;
	int dig;
	char *cp;
	int originalcode = 0, continuation = 0;
	void (*oldintr)(int);
	int pflag = 0;
	size_t px = 0;
	size_t psize = sizeof(pasv);

	oldintr = bb_signal((1 << SIGINT), cmdabort);
	for (;;) {
		dig = n = code = 0;
		cp = reply_string;
		while ((c = getc(cur->cin)) != '\n') {
			if (c == IAC) {     /* handle telnet commands */
				switch (c = getc(cur->cin)) {
				case WILL:
				case WONT:
					c = getc(cur->cin);
					fprintf(cur->cout, "%c%c%c", IAC, DONT, c);
					fflush(cur->cout);
					continue;
				case DO:
				case DONT:
					c = getc(cur->cin);
					fprintf(cur->cout, "%c%c%c", IAC, WONT, c);
					fflush(cur->cout);
					continue;
				case IAC:
				case EOF:
					break;
				default:
					continue;
				}
			}
			dig++;
			if (c == EOF) {
				if (expecteof) {
					bb_signal((1 << SIGINT), oldintr);
					code = 221;
					return 0;
				}
				lostpeer(0);
				if (verbose) {
					puts("421 Service not available, remote server has closed connection");
					fflush(stdout);
				}
				code = 421;
				return(4);
			}
			if (c != '\r' && (verbose > 0 ||
			    (verbose > -1 && n == '5' && dig > 4))) {
				if (proxflag &&
				   (dig == 1 || (dig == 5 && verbose == 0)))
					printf("%s:", cur->hostname);
				putchar(c);
			}
			if (dig < 4 && isdigit(c))
				code = code * 10 + (c - '0');
			if (!pflag && code == 227)
				pflag = 1;
			if (dig > 4 && pflag == 1 && isdigit(c))
				pflag = 2;
			if (pflag == 2) {
				if (c != '\r' && c != ')') {
					if (px < psize-1) pasv[px++] = c;
				}
				else {
					pasv[px] = '\0';
					pflag = 3;
				}
			}
			if (dig == 4 && c == '-') {
				if (continuation)
					code = 0;
				continuation++;
			}
			if (n == 0)
				n = c;
			if (cp < &reply_string[sizeof(reply_string) - 1])
				*cp++ = c;
		}
		if (verbose > 0 || (verbose > -1 && n == '5')) {
			putchar(c);
			fflush (stdout);
		}
		if (continuation && code != originalcode) {
			if (originalcode == 0)
				originalcode = code;
			continue;
		}
		*cp = '\0';
		if (n != '1')
			cur->cpend = 0;
		bb_signal((1 << SIGINT), oldintr);
		if (code == 421 || originalcode == 421)
			lostpeer(0);
		if (abrtflag && oldintr != cmdabort && oldintr != SIG_IGN)
			(*oldintr)(SIGINT);
		return (n - '0');
	}
}

static int
empty(fd_set *mask, int hifd, int sec)
{
	struct timeval t;

	t.tv_sec = (long) sec;
	t.tv_usec = 0;
	return(select(hifd+1, mask, (fd_set *) 0, (fd_set *) 0, &t));
}

static void
abortsend(int ignore)
{
	(void)ignore;

	mflag = 0;
	abrtflag = 0;
	puts("\nsend aborted\nwaiting for remote to finish abort");
	fflush(stdout);
	siglongjmp(sendabort, 1);
}

static FILE *dataconn(const char *lmode)
{
	struct sockaddr_in from;
	int s;
	socklen_t fromlen = sizeof(from);

	if (passivemode)
	    return (fdopen(data, lmode));

	s = accept(data, (struct sockaddr *) &from, &fromlen);
	if (s < 0) {
		bb_perror_msg("accept");
		close(data), data = -1;
		return (NULL);
	}
	close(data);
	data = s;
#ifdef IP_TOS
	if (setsockopt_1i(s, IPPROTO_IP, IP_TOS, IPTOS_THROUGHPUT) < 0)
		bb_perror_msg("setsockopt TOS (ignored)");
#endif
	return (fdopen(data, lmode));
}


#define HASHBYTES 1024

static void
sendrequest(const char *cmd, char *local, char *remote, int printnames)
{
	struct stat st;
	struct timeval start, stop;
	int c, d;
	FILE *volatile fin, *volatile dout = 0;
	int (*volatile closefunc)(FILE *);
	void (*volatile oldintr)(int);
	void (*volatile oldintp)(int);
	volatile long bytes = 0, hashbytes = HASHBYTES;
	char buf[BUFSIZ], *bufp;
	const char *volatile lmode;

	if (verbose && printnames) {
		if (local && *local != '-')
			printf("local: %s ", local);
		if (remote)
			printf("remote: %s\n", remote);
	}
	if (proxy) {
		proxtrans(cmd, local, remote);
		return;
	}
	if (cur->curtype != cur->type)
		changetype(cur->type, 0);
	closefunc = NULL;
	oldintr = NULL;
	oldintp = NULL;
	lmode = "w";
	if (sigsetjmp(sendabort, 1)) {
		while (cur->cpend) {
			getreply(0);
		}
		if (data >= 0) {
			close(data);
			data = -1;
		}
		if (oldintr)
			bb_signal((1 << SIGINT), oldintr);
		if (oldintp)
			bb_signal((1 << SIGPIPE), oldintp);
		code = -1;
		return;
	}
	oldintr = bb_signal((1 << SIGINT), abortsend);
	if (strcmp(local, "-") == 0)
		fin = stdin;
	else if (*local == '|') {
		oldintp = bb_signal((1 << SIGPIPE), SIG_IGN);
		fin = popen(local + 1, "r");
		if (fin == NULL) {
			bb_perror_msg("%s", local + 1);
			bb_signal((1 << SIGINT), oldintr);
			bb_signal((1 << SIGPIPE), oldintp);
			code = -1;
			return;
		}
		closefunc = pclose;
	} else {
		fin = fopen(local, "r");
		if (fin == NULL) {
			print_local_err(local);
			bb_signal((1 << SIGINT), oldintr);
			code = -1;
			return;
		}
		closefunc = fclose;
		if (fstat(fileno(fin), &st) < 0 ||
		    (st.st_mode&S_IFMT) != S_IFREG) {
			printf("%s: not a plain file.\n", local);
			bb_signal((1 << SIGINT), oldintr);
			fclose(fin);
			code = -1;
			return;
		}
	}
	if (initconn()) {
		bb_signal((1 << SIGINT), oldintr);
		if (oldintp)
			bb_signal((1 << SIGPIPE), oldintp);
		code = -1;
		if (closefunc != NULL)
			(*closefunc)(fin);
		return;
	}
	if (sigsetjmp(sendabort, 1))
		goto abort;

	if (restart_point &&
	    (strcmp(cmd, "STOR") == 0 || strcmp(cmd, "APPE") == 0)) {
		if (BB_fseeko(fin, restart_point, 0) < 0) {
			print_local_err(local);
			restart_point = 0;
			if (closefunc != NULL)
				(*closefunc)(fin);
			return;
		}
		if (command("REST %" P_OFFT "d", restart_point) != CONTINUE) {
			restart_point = 0;
			if (closefunc != NULL)
				(*closefunc)(fin);
			return;
		}
		restart_point = 0;
		lmode = "r+w";
	}
	if (remote) {
		if (command("%s %s", cmd, remote) != PRELIM) {
			bb_signal((1 << SIGINT), oldintr);
			if (oldintp)
				bb_signal((1 << SIGPIPE), oldintp);
			if (closefunc != NULL)
				(*closefunc)(fin);
			return;
		}
	} else
		if (command("%s", cmd) != PRELIM) {
			bb_signal((1 << SIGINT), oldintr);
			if (oldintp)
				bb_signal((1 << SIGPIPE), oldintp);
			if (closefunc != NULL)
				(*closefunc)(fin);
			return;
		}
	dout = dataconn(lmode);
	if (dout == NULL)
		goto abort;
	gettimeofday(&start, (struct timezone *)0);
	oldintp = bb_signal((1 << SIGPIPE), SIG_IGN);
	switch (cur->curtype) {

	case TYPE_I:
	case TYPE_L:
		errno = d = 0;
		while ((c = read(fileno(fin), buf, sizeof (buf))) > 0) {
			bytes += c;
			for (bufp = buf; c > 0; c -= d, bufp += d)
				if ((d = write(fileno(dout), bufp, c)) <= 0)
					break;
			if (hash) {
				while (bytes >= hashbytes) {
					putchar('#');
					hashbytes += HASHBYTES;
				}
				fflush(stdout);
			}
			if (tick && (bytes >= hashbytes)) {
				printf("\rBytes transferred: %ld", bytes);
				fflush(stdout);
				while (bytes >= hashbytes)
					hashbytes += TICKBYTES;
			}
		}
		if (hash && (bytes > 0)) {
			if (bytes < HASHBYTES)
				putchar('#');
			putchar('\n');
			fflush(stdout);
		}
		if (tick) {
			printf("\rBytes transferred: %ld\n", bytes);
			fflush(stdout);
		}
		if (c < 0)
			print_local_err(local);
		if (d < 0) {
			if (errno != EPIPE)
				bb_perror_msg("netout");
			bytes = -1;
		}
		break;

	case TYPE_A:
		while ((c = getc(fin)) != EOF) {
			if (c == '\n') {
				while (hash && (bytes >= hashbytes)) {
					putchar('#');
					fflush(stdout);
					hashbytes += HASHBYTES;
				}
				if (tick && (bytes >= hashbytes)) {
					printf("\rBytes transferred: %ld",
						bytes);
					fflush(stdout);
					while (bytes >= hashbytes)
						hashbytes += TICKBYTES;
				}
				if (ferror(dout))
					break;
				putc('\r', dout);
				bytes++;
			}
			putc(c, dout);
			bytes++;
	/*              if (c == '\r') {                                */
	/*               putc('\0', dout);  (* this violates rfc */
	/*                      bytes++;                                */
	/*              }                                               */
		}
		if (hash) {
			if (bytes < hashbytes)
				putchar('#');
			putchar('\n');
			fflush(stdout);
		}
		if (tick) {
			printf("\rBytes transferred: %ld\n", bytes);
			fflush(stdout);
		}
		if (ferror(fin))
			print_local_err(local);
		if (ferror(dout)) {
			if (errno != EPIPE)
				bb_perror_msg("netout");
			bytes = -1;
		}
		break;
	}
	gettimeofday(&stop, (struct timezone *)0);
	if (closefunc != NULL)
		(*closefunc)(fin);
	fclose(dout);
	/* closes data as well, so discard it */
	data = -1;
	getreply(0);
	bb_signal((1 << SIGINT), oldintr);
	if (oldintp)
		bb_signal((1 << SIGPIPE), oldintp);
	if (bytes > 0)
		ptransfer("sent", bytes, &start, &stop);
	return;
abort:
	gettimeofday(&stop, (struct timezone *)0);
	bb_signal((1 << SIGINT), oldintr);
	if (oldintp)
		bb_signal((1 << SIGPIPE), oldintp);
	if (!cur->cpend) {
		code = -1;
		return;
	}
	if (dout) {
		fclose(dout);
	}
	if (data >= 0) {
		/* if it just got closed with dout, again won't hurt */
		close(data);
		data = -1;
	}
	getreply(0);
	code = -1;
	if (closefunc != NULL && fin != NULL)
		(*closefunc)(fin);
	if (bytes > 0)
		ptransfer("sent", bytes, &start, &stop);
}

static void
abortrecv(int ignore)
{
	(void)ignore;

	mflag = 0;
	abrtflag = 0;
	puts("\nreceive aborted\nwaiting for remote to finish abort");
	fflush(stdout);
	siglongjmp(recvabort, 1);
}

static void
recvrequest(const char *cmd,
	    char *volatile local, char *remote,
	    const char *lmode, int printnames)
{
	FILE *volatile fout, *volatile din = 0;
	int (*volatile closefunc)(FILE *);
	void (*volatile oldintp)(int);
	void (*volatile oldintr)(int);
	volatile int is_retr, tcrflag, bare_lfs = 0;
	volatile long bytes = 0, hashbytes = HASHBYTES;
	int c, d;
	struct timeval start, stop;
	struct stat st;
	char gunique_buf[MAXPATHLEN];

	is_retr = strcmp(cmd, "RETR") == 0;
	if (is_retr && verbose && printnames) {
		if (local && *local != '-')
			printf("local: %s ", local);
		if (remote)
			printf("remote: %s\n", remote);
	}
	if (proxy && is_retr) {
		proxtrans(cmd, local, remote);
		return;
	}
	closefunc = NULL;
	oldintr = NULL;
	oldintp = NULL;
	tcrflag = !crflag && is_retr;
	if (sigsetjmp(recvabort, 1)) {
		while (cur->cpend) {
			getreply(0);
		}
		if (data >= 0) {
			close(data);
			data = -1;
		}
		if (oldintr)
			bb_signal((1 << SIGINT), oldintr);
		code = -1;
		return;
	}
	oldintr = bb_signal((1 << SIGINT), abortrecv);
	if (strcmp(local, "-") && *local != '|') {
		if (access(local, W_OK) < 0) {
			char *dir = strrchr(local, '/');

			if (errno != ENOENT && errno != EACCES) {
				print_local_err(local);
				bb_signal((1 << SIGINT), oldintr);
				code = -1;
				return;
			}
			if (dir != NULL)
				*dir = 0;
			d = access(dir ? local : ".", W_OK);
			if (dir != NULL)
				*dir = '/';
			if (d < 0) {
				print_local_err(local);
				bb_signal((1 << SIGINT), oldintr);
				code = -1;
				return;
			}
			if (!cur->runique && errno == EACCES &&
			    chmod(local, 0600) < 0) {
				print_local_err(local);
				/*
				 * Believe it or not, this was actually
				 * repeated in the original source.
				 */
				bb_signal((1 << SIGINT), oldintr);
				code = -1;
				return;
			}
			if (cur->runique && errno == EACCES &&
			   (local = gunique(local, gunique_buf)) == NULL) {
				bb_signal((1 << SIGINT), oldintr);
				code = -1;
				return;
			}
		}
		else if (cur->runique && (local = gunique(local, gunique_buf)) == NULL) {
			bb_signal((1 << SIGINT), oldintr);
			code = -1;
			return;
		}
	}
	if (!is_retr) {
		if (cur->curtype != TYPE_A)
			changetype(TYPE_A, 0);
	}
	else if (cur->curtype != cur->type) {
		changetype(cur->type, 0);
	}
	if (initconn()) {
		bb_signal((1 << SIGINT), oldintr);
		code = -1;
		return;
	}
	if (sigsetjmp(recvabort, 1))
		goto abort;
	if (is_retr && restart_point &&
	    command("REST %" P_OFFT "d", restart_point) != CONTINUE)
		return;
	if (remote) {
		if (command("%s %s", cmd, remote) != PRELIM) {
			bb_signal((1 << SIGINT), oldintr);
			return;
		}
	}
	else {
		if (command("%s", cmd) != PRELIM) {
			bb_signal((1 << SIGINT), oldintr);
			return;
		}
	}
	din = dataconn("r");
	if (din == NULL)
		goto abort;
	if (strcmp(local, "-") == 0)
		fout = stdout;
	else if (*local == '|') {
		oldintp = bb_signal((1 << SIGPIPE), SIG_IGN);
		fout = popen(local + 1, "w");
		if (fout == NULL) {
			bb_perror_msg("%s", local+1);
			goto abort;
		}
		closefunc = pclose;
	}
	else {
		fout = fopen(local, lmode);
		if (fout == NULL) {
			print_local_err(local);
			goto abort;
		}
		closefunc = fclose;
	}
	if (fstat(fileno(fout), &st) < 0 || st.st_blksize == 0)
		st.st_blksize = BUFSIZ;
	if (st.st_blksize > bufsize) {
		if (recrq_buf)
			free(recrq_buf);
		recrq_buf = malloc((unsigned)st.st_blksize);
		if (recrq_buf == NULL) {
			bb_perror_msg("%s", bb_msg_memory_exhausted);
			bufsize = 0;
			goto abort;
		}
		bufsize = st.st_blksize;
	}
	gettimeofday(&start, (struct timezone *)0);
	switch (cur->curtype) {

	case TYPE_I:
	case TYPE_L:
		if (restart_point &&
		    BB_fseeko(fout, restart_point, SEEK_SET) < 0) {
			print_local_err(local);
			if (closefunc != NULL)
				(*closefunc)(fout);
			return;
		}
		errno = d = 0;
		while ((c = read(fileno(din), recrq_buf, bufsize)) > 0) {
			if ((d = write(fileno(fout), recrq_buf, c)) != c)
				break;
			bytes += c;
			if (hash && is_retr) {
				while (bytes >= hashbytes) {
					putchar('#');
					hashbytes += HASHBYTES;
				}
				fflush(stdout);
			}
			if (tick && (bytes >= hashbytes) && is_retr) {
				printf("\rBytes transferred: %ld",
					bytes);
				fflush(stdout);
				while (bytes >= hashbytes)
					hashbytes += TICKBYTES;
			}
		}
		if (hash && bytes > 0) {
			if (bytes < HASHBYTES)
				putchar('#');
			putchar('\n');
			fflush(stdout);
		}
		if (tick && is_retr) {
			printf("\rBytes transferred: %ld\n", bytes);
			fflush(stdout);
		}
		if (c < 0) {
			if (errno != EPIPE)
				bb_perror_msg("netin");
			bytes = -1;
		}
		if (d < c) {
			if (d < 0)
				print_local_err(local);
			else
				fprintf(stderr, "%s: short write\n", local);
		}
		break;

	case TYPE_A:
		if (restart_point) {
			off_t i, n;
			int ch;

			if (fseek(fout, 0L, SEEK_SET) < 0)
				goto done;
			n = restart_point;
			for (i = 0; i++ < n;) {
				if ((ch = getc(fout)) == EOF)
					goto done;
				if (ch == '\n')
					i++;
			}
			if (fseek(fout, 0L, SEEK_CUR) < 0) {
done:
				print_local_err(local);
				if (closefunc != NULL)
					(*closefunc)(fout);
				return;
			}
		}
		while ((c = getc(din)) != EOF) {
			if (c == '\n')
				bare_lfs++;
			while (c == '\r') {
				while (hash && (bytes >= hashbytes)
					&& is_retr) {
					putchar('#');
					fflush(stdout);
					hashbytes += HASHBYTES;
				}
				if (tick && (bytes >= hashbytes) && is_retr) {
					printf("\rBytes transferred: %ld",
						bytes);
					fflush(stdout);
					while (bytes >= hashbytes)
						hashbytes += TICKBYTES;
				}
				bytes++;
				if ((c = getc(din)) != '\n' || tcrflag) {
					if (ferror(fout))
						goto break2;
					putc('\r', fout);
					if (c == '\0') {
						bytes++;
						goto contin2;
					}
					if (c == EOF)
						goto contin2;
				}
			}
			putc(c, fout);
			bytes++;
	contin2:        ;
		}
break2:
		if (hash && is_retr) {
			if (bytes < hashbytes)
				putchar('#');
			putchar('\n');
			fflush(stdout);
		}
		if (tick && is_retr) {
			printf("\rBytes transferred: %ld\n", bytes);
			fflush(stdout);
		}
		if (bare_lfs) {
			printf("WARNING! %d bare linefeeds received in ASCII mode\n"
	"File may not have transferred correctly.\n", bare_lfs);
		}
		if (ferror(din)) {
			if (errno != EPIPE)
				bb_perror_msg("netin");
			bytes = -1;
		}
		if (ferror(fout))
			print_local_err(local);
		break;
	}
	if (closefunc != NULL)
		(*closefunc)(fout);
	bb_signal((1 << SIGINT), oldintr);
	if (oldintp)
		bb_signal((1 << SIGPIPE), oldintp);
	gettimeofday(&stop, (struct timezone *)0);
	fclose(din);
	/* closes data as well, so discard it */
	data = -1;
	getreply(0);
	if (bytes > 0 && is_retr)
		ptransfer("received", bytes, &start, &stop);
	return;
abort:

/* abort using RFC959 recommended IP,SYNC sequence  */

	gettimeofday(&stop, (struct timezone *)0);
	if (oldintp)
		bb_signal((1 << SIGPIPE), oldintp);
	bb_signal((1 << SIGINT), SIG_IGN);
	if (!cur->cpend) {
		code = -1;
		bb_signal((1 << SIGINT), oldintr);
		return;
	}

	abort_remote(din);
	code = -1;
	if (closefunc != NULL && fout != NULL)
		(*closefunc)(fout);
	if (din) {
		fclose(din);
	}
	if (data >= 0) {
		/* if it just got closed with din, again won't hurt */
		close(data);
		data = -1;
	}
	if (bytes > 0)
		ptransfer("received", bytes, &start, &stop);
	bb_signal((1 << SIGINT), oldintr);
}

/*
 * Need to start a listen on the data channel before we send the command,
 * otherwise the server's connect may fail.
 */
static int
initconn(void)
{
	char *p, *a;
	int result, tmpno = 0;
	socklen_t len;
	u_long a1,a2,a3,a4,p1,p2;

	if (passivemode) {
		data = bb_xsocket(0);
		if (data < 0)
			return(1);
		if (options & SO_DEBUG &&
		    setsockopt_1i(data, SOL_SOCKET, SO_DEBUG, 1) < 0)
			bb_perror_msg("setsockopt (ignored)");
		if (command("PASV") != COMPLETE) {
			puts("Passive mode refused.");
			return(1);
		}

		/*
		 * What we've got at this point is a string of comma separated
		 * one-byte unsigned integer values, separated by commas.
		 * The first four are the an IP address. The fifth is the MSB
		 * of the port number, the sixth is the LSB. From that we'll
		 * prepare a sockaddr_in.
		 */

		if (sscanf(pasv,"%ld,%ld,%ld,%ld,%ld,%ld",
			   &a1,&a2,&a3,&a4,&p1,&p2)
		    != 6)
		{
			puts("Passive mode address scan failure. Shouldn't happen!");
			return(1);
		}

		data_addr.sin_family = AF_INET;
		data_addr.sin_addr.s_addr = htonl((a1 << 24) | (a2 << 16) |
						  (a3 << 8) | a4);
		data_addr.sin_port = htons((p1 << 8) | p2);

		if (connect(data, (struct sockaddr *) &data_addr,
		    sizeof(data_addr))<0) {
			bb_perror_msg("connect");
			return(1);
		}
#ifdef IP_TOS
		if (setsockopt_1i(data, IPPROTO_IP, IP_TOS, IPTOS_THROUGHPUT) < 0)
			bb_perror_msg("setsockopt TOS (ignored)");
#endif
		return(0);
	}
noport:
	data_addr = cur->myctladdr;
	if (sendport)
		data_addr.sin_port = 0; /* let system pick one */
	if (data != -1)
		close(data);
	data = bb_xsocket(0);
	if (data < 0) {
		if (tmpno)
			sendport = 1;
		return 1;
	}
	if (!sendport)
		if (setsockopt_1i(data, SOL_SOCKET, SO_REUSEADDR, 1) < 0) {
			bb_perror_msg("setsockopt (reuse address)");
			goto bad;
		}
	if (bind(data, (struct sockaddr *)&data_addr, sizeof (data_addr)) < 0) {
		bb_perror_msg("bind");
		goto bad;
	}
	if (options & SO_DEBUG &&
	    setsockopt_1i(data, SOL_SOCKET, SO_DEBUG, 1) < 0)
		bb_perror_msg("setsockopt (ignored)");
	len = sizeof (data_addr);
	if (getsockname(data, (struct sockaddr *)&data_addr, &len) < 0) {
		bb_perror_msg("getsockname");
		goto bad;
	}
	if (listen(data, 1) < 0)
		bb_perror_msg("listen");
	if (sendport) {
		a = (char *)&data_addr.sin_addr;
		p = (char *)&data_addr.sin_port;
#define UC(b)   (((int)b)&0xff)
		result =
		    command("PORT %d,%d,%d,%d,%d,%d",
		      UC(a[0]), UC(a[1]), UC(a[2]), UC(a[3]),
		      UC(p[0]), UC(p[1]));
		if (result == ERROR && sendport == -1) {
			sendport = 0;
			tmpno = 1;
			goto noport;
		}
		return (result != COMPLETE);
	}
	if (tmpno)
		sendport = 1;
#ifdef IP_TOS
	if (setsockopt_1i(data, IPPROTO_IP, IP_TOS, IPTOS_THROUGHPUT) < 0)
		bb_perror_msg("setsockopt TOS (ignored)");
#endif
	return 0;
bad:
	close(data), data = -1;
	if (tmpno)
		sendport = 1;
	return 1;
}


static void
ptransfer(const char *direction, long bytes,
	  const struct timeval *t0,
	  const struct timeval *t1)
{
	struct timeval td;
	float s, bs;

	if (verbose) {
		td.tv_sec = t1->tv_sec - t0->tv_sec;
		td.tv_usec = t1->tv_usec - t0->tv_usec;
		if (td.tv_usec < 0) {
			td.tv_sec--;
			td.tv_usec += 1000000;
		}
		s = td.tv_sec + (td.tv_usec / 1000000.);
#define nz(x)   ((x) == 0 ? 1 : (x))
		bs = bytes / nz(s);
		printf("%ld bytes %s in %.3g secs (%.2g Kbytes/sec)\n",
		    bytes, direction, s, bs / 1024.0);
	}
}

static void
psabort(int ignore)
{
	(void)ignore;
	abrtflag++;
}

static void
pswitch(int flag)
{
	void (*oldintr)(int);

	abrtflag = 0;
	oldintr = bb_signal((1 << SIGINT), psabort);
	if (flag) {
		if (proxy)
			return;
		cur = &o_proxstruct;
		proxy++;
	}
	else {
		if (!proxy)
			return;
		cur = &proxstruct;
		proxy = 0;
	}
	if (abrtflag) {
		abrtflag = 0;
		(*oldintr)(SIGINT);
	}
}

static void
abortpt(int ignore)
{
	(void)ignore;
	puts("");
	fflush(stdout);
	ptabflg++;
	mflag = 0;
	abrtflag = 0;
	siglongjmp(ptabort, 1);
}

static void
proxtrans(const char *cmd, char *local, char *remote)
{
	void (*volatile oldintr)(int);
	volatile int secndflag = 0, prox_type, nfnd;
	const char *volatile cmd2;
	fd_set mask;

	if (strcmp(cmd, "RETR"))
		cmd2 = "RETR";
	else
		cmd2 = cur->runique ? "STOU" : "STOR";
	if ((prox_type = cur->type) == 0) {
		if (unix_server && unix_proxy)
			prox_type = TYPE_I;
		else
			prox_type = TYPE_A;
	}
	if (cur->curtype != prox_type)
		changetype(prox_type, 1);
	if (command("PASV") != COMPLETE) {
		puts("proxy server does not support third party transfers.");
		return;
	}
	pswitch(0);
	if (!cur->connected) {
		puts("No primary connection");
		pswitch(1);
		code = -1;
		return;
	}
	if (cur->curtype != prox_type)
		changetype(prox_type, 1);
	if (command("PORT %s", pasv) != COMPLETE) {
		pswitch(1);
		return;
	}
	if (sigsetjmp(ptabort, 1))
		goto abort;
	oldintr = bb_signal((1 << SIGINT), abortpt);
	if (command("%s %s", cmd, remote) != PRELIM) {
		bb_signal((1 << SIGINT), oldintr);
		pswitch(1);
		return;
	}
	sleep(2);
	pswitch(1);
	secndflag++;
	if (command("%s %s", cmd2, local) != PRELIM)
		goto abort;
	ptflag++;
	getreply(0);
	pswitch(0);
	getreply(0);
	bb_signal((1 << SIGINT), oldintr);
	pswitch(1);
	ptflag = 0;
	printf("local: %s remote: %s\n", local, remote);
	return;
abort:
	bb_signal((1 << SIGINT), SIG_IGN);
	ptflag = 0;
	if (strcmp(cmd, "RETR") && !proxy)
		pswitch(1);
	else if (!strcmp(cmd, "RETR") && proxy)
		pswitch(0);
	if (!cur->cpend && !secndflag) {  /* only here if cmd = "STOR" (proxy=1) */
		if (command("%s %s", cmd2, local) != PRELIM) {
			pswitch(0);
			if (cur->cpend)
				abort_remote((FILE *) NULL);
		}
		pswitch(1);
		if (ptabflg)
			code = -1;
		bb_signal((1 << SIGINT), oldintr);
		return;
	}
	if (cur->cpend)
		abort_remote((FILE *) NULL);
	pswitch(!proxy);
	if (!cur->cpend && !secndflag) {  /* only if cmd = "RETR" (proxy=1) */
		if (command("%s %s", cmd2, local) != PRELIM) {
			pswitch(0);
			if (cur->cpend)
				abort_remote((FILE *) NULL);
			pswitch(1);
			if (ptabflg)
				code = -1;
			bb_signal((1 << SIGINT), oldintr);
			return;
		}
	}
	if (cur->cpend)
		abort_remote((FILE *) NULL);
	pswitch(!proxy);
	if (cur->cpend) {
		FD_ZERO(&mask);
		FD_SET(cur->cin_fileno, &mask);
		if ((nfnd = empty(&mask, cur->cin_fileno, 10)) <= 0) {
			if (nfnd < 0)
				bb_perror_msg("abort");
			if (ptabflg)
				code = -1;
			lostpeer(0);
		}
		getreply(0);
		getreply(0);
	}
	if (proxy)
		pswitch(0);
	pswitch(1);
	if (ptabflg)
		code = -1;
	bb_signal((1 << SIGINT), oldintr);
}

static void
reset(void)
{
	fd_set mask;
	int nfnd = 1;

	FD_ZERO(&mask);
	while (nfnd > 0) {
		FD_SET(cur->cin_fileno, &mask);
		if ((nfnd = empty(&mask, cur->cin_fileno, 0)) < 0) {
			bb_perror_msg("reset");
			code = -1;
			lostpeer(0);
		}
		else if (nfnd) {
			getreply(0);
		}
	}
}

static char *
gunique(char *local, char gunique_buf[MAXPATHLEN])
{
	char *cp = strrchr(local, '/');
	int d, count = 0;
	char ext = '1';

	if (cp)
		*cp = '\0';
	d = access(cp ? local : ".", W_OK);
	if (cp)
		*cp = '/';
	if (d < 0) {
		print_local_err(local);
		return((char *) 0);
	}
	strcpy(gunique_buf, local);
	cp = gunique_buf + strlen(gunique_buf);
	*cp++ = '.';
	while (!d) {
		if (++count == 100) {
			puts("runique: can't find unique file name.");
			return((char *) 0);
		}
		*cp++ = ext;
		*cp = '\0';
		if (ext == '9')
			ext = '0';
		else
			ext++;
		if ((d = access(gunique_buf, F_OK)) < 0)
			break;
		if (ext != '0')
			cp--;
		else if (*(cp - 2) == '.')
			*(cp - 1) = '1';
		else {
			*(cp - 2) = *(cp - 2) + 1;
			cp--;
		}
	}
	return(gunique_buf);
}

static void
abort_remote(FILE *din)
{
	char buf[BUFSIZ];
	int nfnd, hifd;
	fd_set mask;

	/*
	 * send IAC in urgent mode instead of DM because 4.3BSD places oob mark
	 * after urgent byte rather than before as is protocol now
	 */
	snprintf(buf, sizeof(buf), "%c%c%c", IAC, IP, IAC);
	if (send(cur->cout_fileno, buf, 3, MSG_OOB) != 3)
		bb_perror_msg("abort");
	fprintf(cur->cout,"%cABOR\r\n", DM);
	fflush(cur->cout);
	FD_ZERO(&mask);
	FD_SET(cur->cin_fileno, &mask);
	hifd = cur->cin_fileno;
	if (din) {
		FD_SET(fileno(din), &mask);
		if (hifd < fileno(din)) hifd = fileno(din);
	}
	if ((nfnd = empty(&mask, hifd, 10)) <= 0) {
		if (nfnd < 0)
			bb_perror_msg("abort");
		if (ptabflg)
			code = -1;
		lostpeer(0);
	}
	if (din && FD_ISSET(fileno(din), &mask)) {
		while (read(fileno(din), buf, BUFSIZ) > 0)
			/* LOOP */;
	}
	if (getreply(0) == ERROR && code == 552) {
		/* 552 needed for nic style abort */
		getreply(0);
	}
	getreply(0);
}

static void
domacro(int argc, char *argv[])
{
	int margc;
	char *marg;
	char **margv;
	int i, j;
	char *cp1, *cp2;
	int count = 2, loopflg = 0;
	char line2[BUFSIZ];
	const struct cmd *c;

	if (argc < 2 && !another(&argc, &argv, "macro name")) {
		printf("Usage: %s macro_name.\n", argv[0]);
		code = -1;
		return;
	}
	for (i = 0; i < macnum; ++i) {
		if (!strncmp(argv[1], macros[i].mac_name, 9)) {
			break;
		}
	}
	if (i == macnum) {
		printf("'%s' macro not found.\n", argv[1]);
		code = -1;
		return;
	}
	strcpy(line2, line);
TOP:
	cp1 = macros[i].mac_start;
	while (cp1 != macros[i].mac_end) {
		while (isspace(*cp1)) {
			cp1++;
		}
		cp2 = line;
		while (*cp1 != '\0') {
		      switch(*cp1) {
			    case '\\':
				 *cp2++ = *++cp1;
				 break;
			    case '$':
				 if (isdigit(*(cp1+1))) {
				    j = 0;
				    while (isdigit(*++cp1)) {
					  j = 10*j +  *cp1 - '0';
				    }
				    cp1--;
				    if (argc - 2 >= j) {
					strcpy(cp2, argv[j+1]);
					cp2 += strlen(argv[j+1]);
				    }
				    break;
				 }
				 if (*(cp1+1) == 'i') {
					loopflg = 1;
					cp1++;
					if (count < argc) {
					   strcpy(cp2, argv[count]);
					   cp2 += strlen(argv[count]);
					}
					break;
				}
				/* intentional drop through */
			    default:
				*cp2++ = *cp1;
				break;
		      }
		      if (*cp1 != '\0') {
			 cp1++;
		      }
		}
		*cp2 = '\0';
		margv = makeargv(&margc, &marg);
		c = getcmd(margv[0]);
		if (c == (struct cmd *)-1) {
			puts("?Ambiguous command");
			code = -1;
		}
		else if (c == NULL) {
			puts("?Invalid command");
			code = -1;
		}
		else if (c->c_conn && !cur->connected) {
			puts("Not connected.");
			code = -1;
		}
		else {
			if (verbose) {
				puts(line);
			}
			c_handler(c, margc, margv, marg);

			if (bell && c->c_bell) {
				putchar('\007');
			}
			strcpy(line, line2);
			margv = makeargv(&margc, &marg);
			argc = margc;
			argv = margv;
		}
		if (cp1 != macros[i].mac_end) {
			cp1++;
		}
	}
	if (loopflg && ++count < argc) {
		goto TOP;
	}
}

static char **
ftpglob(const char *v)
{

	int retglob;

	globbuf.gl_pathv = 0;
	retglob = glob(v, GLOB_BRACE|GLOB_TILDE, NULL, &globbuf);
	switch(retglob) {
		case GLOB_NOSPACE:
			globerr = "Not have space";
			break;
		case GLOB_ABORTED:
			globerr = "Aborted";
			break;
		case GLOB_NOMATCH:
			globerr = "No matched";
			break;
		default:
			globerr = 0;
			break;
	}
	return globbuf.gl_pathv;
}
#endif /* APPLET_USAGE */
