/* ftpd.c 1.81 1999/06/11 05:18:53 */
/* ls.c 1.11 1999/03/09 20:16:17 */
/* daemons.c 1.1 1998/10/28 23:22:22 */

/*
 * Copyright 1995-2000 Trolltech AS. Copyright 2001 Arnt Gulbrandsen.
 * Use, modification and distribution is allowed without limitation,
 * warranty, or liability of any kind.
*/

/*
 * Busybox port and other small perfects by
 * (C) 2003,2007 Vladimir Oleynik <dzo@simtreas.ru>
*/

#ifdef APPLET_USAGE
#define ftpd_trivial_usage \
	"[-M]\nor\n[-ab] [-c clients] [-m load] [-p first:last] [-u uid_limit] [-s] [-x group]"
#define ftpd_full_usage \
	"The Trolltech AS public domain ftp server daemon.\n\n" \
	"Options:\n" \
	"\t-M\tCreate users and groups list and exit.\n" \
	"\t\tThis files used after chroot.\n" \
	"\t-a\tAllow active-mode FTP. This option makes troll-ftpd more\n" \
	"\t\tcompatible with some old clients\n" \
	"\t-b\tBe broken. Turns on some compatibility hacks for shoddy\n" \
	"\t\tclients.\n" \
	"\t-c clients\tAllow a maximum of clients to be connected.\n" \
	"\t-m load\tDo not allow anonymous users to download files if the\n" \
	"\t\tload is above 'load' when the user connects. Uploads and file\n" \
	"\t\tlistings are still allowed, as are downloads by real users.\n" \
	"\t\tThe user is not told about this until he/she tries to download\n" \
	"\t\ta file.\n" \
	"\t-p first:last\tUse only ports in the range inclusive for\n" \
	"\t\tpassive-mode downloads.\n" \
	"\t-u uid\tDo not allow uids below uid to log in. Default -u 1\n" \
	"\t\tfor denies access only to root accounts.\n" \
	"\t-s\tDon't allow anonymous users to retrieve files owned by 'ftp'\n" \
	"\t-x group\tChroot most users to their home directories, but\n" \
	"\t\tgive root and all users in 'group' access to the entire\n" \
	"\t\tfile system."
#else

static const char ftpd_users[]=  "/var/adm/ftp/users";
static const char ftpd_groups[]= "/var/adm/ftp/groups";

#ifdef DEBUG
static int debug;  /* 0 - don't give debug output */
#endif

#include "busybox.h"
#include <ctype.h>
#include <dirent.h>
#include <glob.h>       /* linux rules! */
#include <netinet/in.h>
#include <netinet/ip.h>
#include <setjmp.h>
#include "bb_shadow.h"
#include <string.h>
#include <sys/fsuid.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <sys/vfs.h>
#include <time.h>
#include <unistd.h>
#include <syslog.h>
#include <crypt.h>
#include "bb_xopen.h"
#include "bb_inet.h"

#ifndef HAVE_SETFSUID
#include <sys/syscall.h>
int setfsuid(uid_t uid) {
	return syscall(__NR_setfsuid, uid);
}
#endif

#ifndef HAVE_GETPWENT
struct passwd *getpwent(void);
#endif
#ifndef HAVE_GETGRENT
struct group *getgrent(void);
#endif

#if PATH_MAX < 1024
#error "troll-ftpd requres that PATH_MAX be at least 1024"
#endif

#ifdef DEBUG
#include <signal.h>
#endif

#define MAX_GROUPS 256

struct reply {
    struct reply *prev;
    char line[1];
};

struct filename {
    struct filename *down;
    struct filename *right;
    int top;
    char line[1];
};

CALC_BSS struct BB_COMMON_BSS
{
	int opt_a, opt_C, opt_d, opt_F, opt_l, opt_R, opt_r, opt_t, opt_S;
	bb64_t downloaded;               /* bytes downloaded */
	bb64_t uploaded;                 /* bytes uploaded */
	int passive;
	int datafd;                      /* data connection file descriptor */
	struct sockaddr_in ctrlconn;     /* stdin/stdout, for using the same ip number */

	char cmd[PATH_MAX + 32]; /* command line - about 30 chars for command */
	char wd[PATH_MAX + 1];  /* current working directory */
	char *rd;               /* root directory, for chroot'd environments */
	char * cpwd;            /* crypted password of the account in use */
	int loggedin;
	char account[9];
	char * renamefrom;
	int epsvall;            /* nonzero means we've seen an EPSV ALL */

	int userchroot;         /* don't chroot() by default for regular users */
	uid_t chroot_trustedgid;
	int broken;             /* don'be be broken by default */
	int keeproot;           /* don't keep root after login (or do keep it) */
	uid_t warez;            /* don't guard against warez */
	int guest;              /* if non-zero, it's a guest user */
	uid_t useruid;          /* smallest uid that can ftp */
	int candownload;        /* if non-zero, don't let the user download */
	double load;            /* for reporting what the load was */

	int firstport;          /* first and last ports to use, if we're */
	int lastport;           /* packet filter friendly. */
	bbst_size_t restartat;

	struct reply *lastreply;
	int replycode;

	int colwidth;
	int filenames;

	struct filename *head;
	struct filename *tail;

	struct sockaddr_in peer;
	unsigned short int peerdataport;

	char *users;
	char *groups;

	int matches;
	char outbuf[1024];
	int outptr;
	char u_number[9];
	char g_number[9];

	unsigned int idletime;
	int type;            /* type - 0 = error, 1 = ascii, 2 = binary */
} bb_common_bss;
#define opt_a bb_common_bss.opt_a
#define opt_C bb_common_bss.opt_C
#define opt_d bb_common_bss.opt_d
#define opt_F bb_common_bss.opt_F
#define opt_l bb_common_bss.opt_l
#define opt_R bb_common_bss.opt_R
#define opt_r bb_common_bss.opt_r
#define opt_t bb_common_bss.opt_t
#define opt_S bb_common_bss.opt_S
#define downloaded bb_common_bss.downloaded
#define uploaded bb_common_bss.uploaded
#define passive bb_common_bss.passive
#define datafd bb_common_bss.datafd
#define ctrlconn bb_common_bss.ctrlconn
#define cmd bb_common_bss.cmd
#define wd bb_common_bss.wd
#define rd bb_common_bss.rd
#define cpwd bb_common_bss.cpwd
#define loggedin bb_common_bss.loggedin
#define account bb_common_bss.account
#define renamefrom bb_common_bss.renamefrom
#define epsvall bb_common_bss.epsvall
#define userchroot bb_common_bss.userchroot
#define chroot_trustedgid bb_common_bss.chroot_trustedgid
#define broken bb_common_bss.broken
#define keeproot bb_common_bss.keeproot
#define warez bb_common_bss.warez
#define guest bb_common_bss.guest
#define useruid bb_common_bss.useruid
#define candownload bb_common_bss.candownload
#define load bb_common_bss.load
#define firstport bb_common_bss.firstport
#define lastport bb_common_bss.lastport
#define restartat bb_common_bss.restartat
#define lastreply bb_common_bss.lastreply
#define replycode bb_common_bss.replycode
#define colwidth bb_common_bss.colwidth
#define filenames bb_common_bss.filenames
#define head bb_common_bss.head
#define tail bb_common_bss.tail
#define peer bb_common_bss.peer
#define peerdataport bb_common_bss.peerdataport
#define users bb_common_bss.users
#define groups bb_common_bss.groups
#define matches bb_common_bss.matches
#define outbuf bb_common_bss.outbuf
#define outptr bb_common_bss.outptr
#define u_number bb_common_bss.u_number
#define g_number bb_common_bss.g_number
#define idletime bb_common_bss.idletime
#define type bb_common_bss.type

static const int window = 51200;        /* window size */
static const int cmdsize = PATH_MAX + 32;


static void
w_setsockopt(int sockfd, int level, int optname, const void *optval)
{
	if(setsockopt(sockfd, level, optname, optval, sizeof(int)) < 0) {
		log_syslog = LOG_WARNING;
		bb_perror_msg("setsockopt");
		log_syslog = LOG_ERR;
	}
}

#ifdef DEBUG
static volatile int logging;

static void sighandler( int sig )
{
    if ( sig == SIGUSR1 && logging < 2 )
	logging++;
    else if ( sig == SIGUSR2 && logging > 0 )
	logging--;
}
#endif

static ATTRIBUTE_NORETURN
void quit421( int code, const char *fmt, ... )
{
#define strlen_421_ 4   /* strlen("421 "); */
#define strlen_r_n  3   /* strlen("\r\n\0"); */
    char buf[PATH_MAX + 128];
    int e = errno;

    va_list ap;
    int offs;

    va_start( ap, fmt );
    strcpy(buf, "421 ");
    errno = e;
    vsnprintf( buf + strlen_421_, sizeof(buf) - strlen_421_ - strlen_r_n,
								fmt, ap );
    va_end( ap );
    offs = strlen(buf);
    strcat(buf, "\r\n");
    printf( "%s", buf );
    fflush(stdout);
    buf[offs] = 0;      /* remove "\r\n" */
    bb_default_error_retval = code;
    bb_error_msg_and_die("%s", buf + strlen_421_);
}

/* rewroted by Vladimir Oleynik */
static void addreply( int code, const char *line,... )
{
    struct reply *p;
    struct reply *q;
    char buf[PATH_MAX + 50];
    va_list ap;
    char *s, *e;

    if ( code )
	replycode = code;
    q = NULL;

    va_start( ap, line );
    vsnprintf( buf, sizeof(buf), line, ap );
    va_end( ap );

    /* reduce all '\r' chars */
    s = e = buf;
    while(*s) {
	if(*s != '\r')
		*e++ = *s;
	s++;
    }
    *e = 0;

    s = NULL;
    while( (e = strrchr(buf, '\n')) != NULL || s == NULL) {
	if(e == NULL) {
	    /* without '\n' */
	    s = e = buf;
	} else {
	    /* found '\n' */
	    s = e - 1;  /* before '\n' */
	    *e++ = 0;   /* buf[last('\n')] = 0; e - ptr to after '\n' */
	    /* reduce [:space:] before '\n' */
	    while(s >= buf && isspace(*s) && *s != '\n')
		*s-- = 0;
	    s = NULL;
	}

	p = malloc(sizeof(struct reply) + strlen(e));
	if ( !p )
	    quit421(1, bb_msg_memory_exhausted);
	strcpy( p->line, e );
	if ( q ) {
	    p->prev = q->prev;
	    q->prev = p;
	} else {
	    p->prev = lastreply;
	    lastreply = p;
	}
	q = p;
    }
}

static void replycont( struct reply *p )               /* ick */
{
    if ( p ) {
	replycont( p->prev );
	printf( "%3d-%s\r\n", replycode, p->line );
#ifdef DEBUG
	if ( logging > 1 )
	    syslog( LOG_DEBUG,  "%3d-%s\n", replycode, p->line );
#endif
	free( (char * ) p );
    }
}

static void doreply( void )
{
    if ( lastreply ) {
	replycont( lastreply->prev );
	printf( "%3d %s\r\n", replycode, lastreply->line );
#ifdef DEBUG
	if ( logging > 1 )
	    syslog( LOG_DEBUG,  "%3d %s\n", replycode, lastreply->line );
#endif
	free( (char * ) lastreply );
	lastreply = NULL;
    }
    fflush( stdout );
}


/* this is dog-slow.  but now it's 100% in compliance with the protocol. */

static int sfgets( void )
{
    int l;
    int nread;
    fd_set rs;
    struct timeval tv;

    FD_ZERO( &rs );
    tv.tv_sec = idletime;
    tv.tv_usec = 0;
    nread = 0;
    l = 0;
    do {
	if ( nread == 0 ) {
	    FD_SET( 0, &rs );
	    select( 1, &rs, NULL, NULL, &tv );
	    if ( ioctl( 0, FIONREAD, &nread ) < 0 )
		nread = 0;
	}
	if ( FD_ISSET( 0, &rs ) ) {
	    int i;
	    i = read( 0, cmd+l, 1 );
	    if ( i == 0 ) {
		syslog( LOG_INFO, "exiting due to EOF from client" );
		exit( 21 );
	    } else if ( i < 0 ) {
		syslog(LOG_INFO, "exiting due to read error from client: %m");
		exit( 22 );
	    }
	    if ( nread )
		nread--;
	    if ( cmd[l] == '\n' ) {
		/* if you need to debug ftpd, you can (e.g.) connect
		   to it using an ordinary client, connect to the
		   process using gdb or another ptrace()-aware
		   debugger, and set a breakpoint on the next
		   line. ftpd will hit the breakpoint immediately
		   after the next command or whever the timeout hits. */
		cmd[l+1] = '\0';
		return 1;
	    }
	    if ( l < cmdsize - 2 )
		l++;
	} else {
	    return 0;
	}
    } while( 1 );
}

static void error( int n, const char *msg )
{
    int e = errno;

    bb_perror_msg("%s", msg );
    errno = e;
    addreply( n, "%s: %m", msg);
}

static int close_datafd(void)
{
    close( datafd );
    return datafd = 0;
}

static int opendata( void )
{
    struct sockaddr_in dataconn;        /* his data connection endpoint */
    int fd;
    unsigned int fodder;
    socklen_t addrlen;

    if ( !datafd ) {
	error( 425, "No data connection" );
	return 0;
    }

    if ( passive ) {
	fd_set rs;
	struct timeval tv;

	FD_ZERO( &rs );
	FD_SET( datafd, &rs );
	tv.tv_sec = idletime;
	tv.tv_usec = 0;
	/* I suppose it would be better to listen for ABRT too... */
	if ( !select( datafd + 1, &rs, NULL, NULL, &tv ) ) {
	    addreply( 421,
		      "timeout (no connection for %d seconds)", idletime );
	    doreply();
	    exit( 23 );
	}
	addrlen = sizeof( dataconn );
	fd = accept( datafd, (struct sockaddr *)&dataconn, &addrlen );
	if ( fd < 0 ) {
	    error( 421, "accept failed" );
	    return close_datafd();
	}
	if ( !guest && dataconn.sin_addr.s_addr != peer.sin_addr.s_addr ) {
	    addreply( 425, "Connection must originate at %s (not %s)",
		      inet_ntoa( peer.sin_addr ),
		      inet_ntoa( dataconn.sin_addr ) );
	    return close_datafd();
	}
	addreply( 150, "Accepted data connection from %s:%d",
		  inet_ntoa( dataconn.sin_addr ),
		  ntohs((unsigned short int)dataconn.sin_port) );
    } else {
	dataconn.sin_addr.s_addr = peer.sin_addr.s_addr;
	dataconn.sin_port = htons( peerdataport );
	dataconn.sin_family = AF_INET;

	if (connect(datafd, (struct sockaddr *)&dataconn, sizeof(dataconn))) {
	    addreply( 425, "Could not open data connection to %s port %d: %m",
		      inet_ntoa( dataconn.sin_addr ), peerdataport);
	    return close_datafd();
	}
	fd = datafd;
	datafd = 0;
	addreply( 150, "Connecting to %s:%d",
		  inet_ntoa( dataconn.sin_addr ), peerdataport );
    }

    fodder = IPTOS_THROUGHPUT;
    w_setsockopt(fd, SOL_IP, IP_TOS, (char *)&fodder);
    fodder = window;
    w_setsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *)&fodder);
    fodder = window;    /* not that important, but... */
    w_setsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char *)&fodder);

    return fd;
}

static void wrstr( int f, const char * s, ... )
{
    int l;
    va_list ap;

    va_start( ap, s );
    do {
	if ( !s ) {
	    if ( outptr )
		write( f, outbuf, outptr );
	    outptr = 0;
	    break;
	}

	l = strlen( s );
	if ( l + outptr > 1024 ) {
	    if ( outptr )
		write( f, outbuf, outptr );
	    outptr = 0;
	}

	if ( l > 1024 ) {
	    write( f, s, l );
	} else {
	    memcpy( outbuf+outptr, s, l );
	    outptr += l;
	}
	s = va_arg(ap, char *);
    } while(s);
    va_end( ap );
}

static void outputfiles( int f )
{
    int n;
    struct filename *p;
    struct filename *q;

    if ( !head )
	return;

    tail->down = NULL;
    tail = NULL;
    colwidth = ( colwidth | 7 ) + 1;
    if ( opt_l || !opt_C )
	colwidth = 75;

    /* set up first column */
    p = head;
    p->top = 1;
    n = (filenames + (75 / colwidth)-1) / (75 / colwidth);
    while ( n && p ) {
	p = p->down;
	if ( p )
	    p->top = 0;
	n--;
    }

    /* while there's a neighbour to the right, point at it */
    q = head;
    while ( p ) {
	p->top = q->top;
	q->right = p;
	q = q->down;
	p = p->down;
    }

    /* some are at the right end */
    while ( q ) {
	q->right = NULL;
	q = q->down;
    }

    /* don't want wraparound, do we? */
    p = head;
    while ( p && p->down && !p->down->top )
	p = p->down;
    if ( p && p->down )
	p->down = NULL;

    /* print each line, which consists of each column */
    p = head;
    while ( p ) {
	q = p;
	p = p->down;
	while ( q ) {
	    char pad[6];
	    char *tmp = ( char * ) q;

	    if ( q->right ) {
		strcpy( pad, "\t\t\t\t\t" );
		pad[( colwidth + 7 - strlen( q->line ) ) / 8] = '\0';
	    } else {
		strcpy( pad, "\r\n" );
	    }
	    wrstr( f, q->line, pad, NULL );
	    q = q->right;
	    free( tmp );
	}
    }

    /* reset variables for next time */
    head = tail = NULL;
    colwidth = 0;
    filenames = 0;
}

static char *getname( const uid_t uid )
{
    if ( users && users[9 * uid] ) {
	return ( users + ( 9 * uid ) );
    } else {
	snprintf( u_number, 9, "%-8d", uid );
	return ( u_number );
    }
}

static char *getgroup( const gid_t gid )
{
    if ( groups && groups[9 * gid] ) {
	return ( groups + ( 9 * gid ) );
    } else {
	snprintf( g_number, 9, "%-8d", gid );
	return ( g_number );
    }
}

static void addfile( const char *name, const char *suffix )
{
    struct filename *p;
    int l;

    if ( !name || !suffix )
	return;

    matches++;

    l = strlen( name ) + strlen( suffix );
    if ( l > colwidth )
	colwidth = l;
    l = l + sizeof( struct filename );

    p = malloc(l);

    if ( !p ) {
	quit421(7, bb_msg_memory_exhausted);
    }
    snprintf( p->line, l, "%s%s", name, suffix );
    if ( tail )
	tail->down = p;
    else
	head = p;
    tail = p;
    filenames++;
}

/* listfile returns non-zero if the file is a directory */
static int listfile( const char *name )
{
    int rval = 0;
    char m[1024];
    struct stat st;
    char months[12][4] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
			   "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
    struct tm *t;
    char suffix[2];

    if ( lstat( name, &st ) == 0 ) {
	t = localtime( (time_t * ) & st.st_mtime );
	if ( !t ) {
	    quit421(2, "Bailing out, localtime() is insane");
	}
	suffix[0] = suffix[1] = '\0';
	if ( opt_F ) {
	    if ( S_ISLNK( st.st_mode ) )
		suffix[0] = '@';
	    else if ( S_ISDIR( st.st_mode ) ) {
		suffix[0] = '/';
		rval = 1;
	    } else if ( st.st_mode & 010101 )
		suffix[0] = '*';
	}
	if ( opt_l ) {
	    strcpy( m, " ---------" );
	    switch ( st.st_mode & S_IFMT ) {
		case S_IFREG:
		m[0] = '-';
		break;
		case S_IFLNK:
		m[0] = 'l';
		break;  /* readlink() here? */
		case S_IFDIR:
		m[0] = 'd';
		rval = 1;
		break;
	    }
	    if ( m[0] != ' ' ) {
		char nameline[PATH_MAX + PATH_MAX + 128];
		char timeline[6];

		if ( st.st_mode & 256 )
		    m[1] = 'r';
		if ( st.st_mode & 128 )
		    m[2] = 'w';
		if ( st.st_mode & 64 )
		    m[3] = 'x';
		if ( st.st_mode & 32 )
		    m[4] = 'r';
		if ( st.st_mode & 16 )
		    m[5] = 'w';
		if ( st.st_mode & 8 )
		    m[6] = 'x';
		if ( st.st_mode & 4 )
		    m[7] = 'r';
		if ( st.st_mode & 2 )
		    m[8] = 'w';
		if ( st.st_mode & 1 )
		    m[9] = 'x';

		if ( time( NULL ) - st.st_mtime > 180 * 24 * 60 * 60 )
		    snprintf( timeline, 6, "%5d", t->tm_year + 1900 );
		else
		    snprintf( timeline, 6, "%02d:%02d",
			      t->tm_hour, t->tm_min );

		snprintf( nameline, PATH_MAX + 128 - 1,
			  "%s %3d %s %s %7" P_ST_SIZE "d %s %2d %s %s", m,
			  (int)st.st_nlink, getname( st.st_uid ),
			  getgroup( st.st_gid ),
			  st.st_size, months[t->tm_mon],
			  t->tm_mday, timeline, name );
		if ( S_ISLNK( st.st_mode ) ) {
		    char *p = nameline + strlen( nameline );

		    m[readlink( name, m, 1023 )] = '\0';
		    suffix[0] = '\0';
		    if ( opt_F && stat( name, &st ) == 0 ) {
			if ( S_ISLNK( st.st_mode ) )
			    suffix[0] = '@';
			else if ( S_ISDIR( st.st_mode ) )
			    suffix[0] = '/';
			else if ( st.st_mode & 010101 )
			    suffix[0] = '*';
		    }
		    snprintf( p, PATH_MAX, " -> %s", m );
		}
		addfile( nameline, suffix );
	    }   /* hide non-downloadable files */
	} else {
	    if ( S_ISREG( st.st_mode ) ||
		S_ISDIR( st.st_mode ) ||
		S_ISLNK( st.st_mode ) ) {
		addfile( name, suffix );
	    }
	}
    }
    return rval;
}

/* functions to to sort for qsort() */
static int cmp( const void *a, const void *b ) {
    return strcmp( *( const char ** ) a, *( const char ** ) b );
}
static int cmp_r ( const void *a, const void *b ) {
    return strcmp( *( const char ** ) b, *( const char ** ) a );
}
static int cmp_t ( const void *a, const void *b ) {
    return *(*(const int **)a -2) - *(*(const int **)b -2);
}
static int cmp_rt ( const void *a, const void *b ) {
    return *(*(const int **)b -2) - *(*(const int **)a -2);
}
static int cmp_S ( const void *a, const void *b ) {
    return *(*(const int **)b -1) - *(*(const int **)a -1);
}
static int cmp_rS ( const void *a, const void *b ) {
    return *(*(const int **)a -1) - *(*(const int **)b -1);
}


static char **sreaddir( const char *r_dirname )
{
    DIR *d;
    struct dirent *de;
    struct stat st;
    int i = is_directory( r_dirname, 1, &st );
    char **p;
    unsigned int s;
    int dsize;

    if ( i <= 0 )
	return NULL;

    if ( (d = opendir( r_dirname ) ) == NULL )
	return NULL;

    /* st_size is enough for any sane fs, but procfs is insane */
    dsize = st.st_size + 100;   /* okay okay, a little margin is cool */

 berkeley:
    dsize = dsize * 2;
    p = malloc(dsize);

    if ( !p ) {
	closedir( d );
	errno = ENOMEM;
	return NULL;
    }
    s = dsize;
    i = 0;

    while ( (de = readdir( d ) ) != NULL ) {
	if ( s < i*sizeof(int) + 3*sizeof(int) + strlen( de->d_name ) + 1 ) {
	    /* should leak some memory too, make it perfect : ) */
	    free( p );
	    rewinddir( d );
	    goto berkeley;
	}
	s -= strlen( de->d_name ) + 1;
	strcpy( ((char *)p)+s, de->d_name );
	p[i++] = ((char *)p)+s;
	if (!lstat(de->d_name,&st)) {
	    s -= sizeof(int);
	    *((int *)(((char *)p) + s))=st.st_size;
	    s -= sizeof(int);
	    *((int *)(((char *)p) + s))=st.st_size;
	} else {
	    s -= sizeof(int);
	    *((int *)(((char *)p) + s))=0;
	    s -= sizeof(int);
	    *((int *)(((char *)p) + s))=0;
	}
    }
    closedir( d );
    p[i] = NULL;

    if (opt_t)
	if (opt_r)
	    qsort( p, i, sizeof( char * ), cmp_rt );
	else
	    qsort( p, i, sizeof( char * ), cmp_t );
    else if (opt_S)
	if (opt_r)
	    qsort( p, i, sizeof( char * ), cmp_rS );
	else
	    qsort( p, i, sizeof( char * ), cmp_S );
    else
	if (opt_r)
	    qsort( p, i, sizeof( char * ), cmp_r );
	else
	    qsort( p, i, sizeof( char * ), cmp );

    return p;
}

/* have to change to the directory first ( speed hack for -R ) */
static void listdir( int f, const char *name )
{
    char **dir;

    dir = sreaddir( "." );
    if ( dir ) {
	char **s;
	char **r;
	int d;

	wrstr( f, "total 1\r\n", NULL );      /* so what is total anyway */
	s = dir;
	while ( *s ) {
	    if ( **s != '.' ) {
		d = listfile( *s );
	    } else if ( opt_a ) {
		d = listfile( *s );
		if ( ( (*s)[1] == '\0' ) ||
		     ( ( (*s)[1] == '.' ) &&
		       ( (*s)[2] == '\0' ) ) )
		    d = 0;
	    } else {
		d = 0;
	    }
	    if ( !d )
		*s = NULL;
	    s++;
	}
	outputfiles( f );
	r = dir;
	while ( opt_R && r != s ) {
	    if ( *r && !chdir( *r ) ) {
		char subdir[PATH_MAX];

		snprintf( subdir, PATH_MAX, "%s/%s", name, *r );
		wrstr( f, "\r\n", subdir, ":\r\n", NULL);
		listdir( f, subdir );
		if ( chdir( ".." ) ) {  /* defensive in the extreme... */
		    chdir( wd );
		    if ( chdir( name ) ) {      /* someone rmdir()'d it? */
			quit421(13, "Unrecoverable file system error: %m");
		    }
		}
	    }
	    r++;
	}
	free( dir );
    } else {
	addreply( 226, "Error during reading of %s", name );
    }
}

static void donlist( char *arg )
{
    int c;

    matches = 0;

    opt_l = opt_a = opt_C = opt_d = opt_F = opt_R = opt_r = opt_t = opt_S = 0;

    while ( isspace( *arg ) )
	arg++;

    while ( arg && *arg == '-' ) {
	while ( arg++ && isalnum( *arg ) ) {
	    switch ( *arg ) {
	    case 'a':
		opt_a = 1;
		break;
	    case 'l':
		opt_l = 1;
		opt_C = 0;
		break;
	    case '1':
		opt_l = opt_C = 0;
		break;
	    case 'C':
		opt_l = 0;
		opt_C = 1;
		break;
	    case 'F':
		opt_F = 1;
		break;
	    case 'R':
		opt_R = 1;
		break;
	    case 'd':
		opt_d = 1;
		break;
	    case 'r':
		opt_r = 1;
		break;
	    case 't':
		opt_t = 1;
		opt_S = 0;
		break;
	    case 'S':
		opt_S = 1;
		opt_t = 0;
		break;
	    default:
		;
	    }
	}
	while ( isspace( *arg ) )
	    arg++;
    }

    c = opendata();
    if ( !c )
	return;

    doreply();

    if ( type == 2 )
	addreply( 0, "Binary mode requested, but A (ASCII) used." );

    if ( arg && *arg ) {
	int justone;
	char * dos;

	justone = 1;    /* just one argument, so don't print dir name */

	while ( arg ) {
	    glob_t g;
	    int a;
	    char buffer[PATH_MAX+1];

	    char *endarg = strchr_set0( arg, ' ' );

	    if ( endarg ) {
		endarg++;
		justone = 0;
	    }

#ifdef DEBUG
	    if ( debug )
		addreply( 226, "Glob argument: %s", arg );
#endif
	    strcpy( buffer, arg );
	    if ( loggedin && !guest && *arg == '~' ) {
		struct passwd *pw;
		int i;
		const char *p;

		i = 0;
		p = arg;
		p++;
		while ( *p && *p != '/' && i < PATH_MAX )
		    buffer[i++] = *p++;
		buffer[i] = '\0';

		if ( (pw = getpwnam( i ? buffer : account ) ) )
		    snprintf( buffer, PATH_MAX, "%s%s", pw->pw_dir, p );
		else
		    strncpy( buffer, arg, PATH_MAX );
	    }

	    /* detect and block /../ DoS attacks */
	    dos = buffer;
	    while( dos && *dos ) {
		while( dos && *dos && *dos != '/' )
		    dos++;
		/* ok, we're at a slash. now, this DoS should always
		   match one of these two cases:
		   1. /../blah where blah contains wildcards.
		   2. /.blah/ where blah contains wildcards.
		   we detect both and refuse to expand the pattern. */
		if ( *++dos == '.' ) {
		    if ( dos[1] == '.' && dos[2] == '/' )
			dos += 3;
		    while( dos && *dos && *dos != '/' ) {
			if( ( *dos == '{' || *dos == '?' || *dos == '*' ) &&
			    strchr( dos, '/' ) ) {
			    dos = NULL;
			    buffer[0] = '\0';
			} else {
			    dos++;
			}
		    }
		}
	    }

	    if ( !buffer[0] ) {
		addreply( 0, "%s not expanded: May be too complex",
			  arg );
	    } else if ( (a=glob( buffer,
				 opt_a ? GLOB_PERIOD : 0, NULL, &g )) == 0 ) {
		char **path;

		path = g.gl_pathv;
		if ( path && path[0] && path[1] )
		    justone = 0;

		while ( path && *path ) {
		    int lst = is_directory(*path, 0, NULL);

		    if ( lst >= 0 ) {
			if ( opt_d || lst == 0 ) {
			    listfile( *path );
			    **path = '\0';
			}
		    } else {
			**path = '\0';
		    }
		    path++;
		}
		outputfiles( c );               /* in case of opt_C */
		path = g.gl_pathv;
		while ( path && *path ) {
		    if ( **path ) {
			if ( !justone ) {
			    wrstr( c, "\r\r", *path, ":\r\n", NULL );
			}
			if ( !chdir( *path ) ) {
			    listdir( c, *path );
			    chdir( wd );
			}
		    }
		    path++;
		}
		globfree( &g );
	    } else {
		if ( a != GLOB_NOMATCH )
		    addreply( 226,
			      "Unknown error during globbing of %s", arg );
		globfree( &g );
	    }
	    arg = endarg;
	}
    } else {
	if ( opt_d )
	    listfile( "." );
	else
	    listdir( c, "." );
	outputfiles( c );
    }
    wrstr( c, NULL );
    close( c );
    if ( opt_a || opt_C || opt_d || opt_F || opt_l || opt_r || opt_R ||
	 opt_t || opt_S )
	addreply( 0, "Options: %s%s%s%s%s%s%s%s%s",
		  opt_a ? "-a " : "",
		  opt_C ? "-C " : "",
		  opt_d ? "-d " : "",
		  opt_F ? "-F " : "",
		  opt_l ? "-l " : "",
		  opt_r ? "-r " : "",
		  opt_R ? "-R " : "",
		  opt_S ? "-S " : "",
		  opt_t ? "-t" : "" );
    addreply( 226, "%d matches total", matches );
}

/* small help routine to display a banner */
static void dobanner( void )
{
    FILE *msg;

    if ( (msg = fopen( ".banner", "r" ) ) != NULL ) {
	/* if you think this is too small, send me mail.  NOTE: It should
	   not be unlimited.  Take a look at e.g. ftp.3com.com to see why:
	   at the time of writing the logon banner is 250-odd lines */
	char buffer[1025];
	int len = fread( (void * ) buffer, 1, 1024, msg );

	fclose( msg );
	if ( len > 0 && len < 1024 ) {
	    buffer[len] = '\0';
	    addreply( 0, "%s", buffer );
	}
    }
}

static void douser( const char *username )
{
    struct passwd *pw;
#ifdef CONFIG_FEATURE_SHADOWPASSWDS
    struct spwd *spw;
#endif

    if ( loggedin ) {
	if ( username ) {
	    if ( !guest )
		addreply( 530, "You're already logged in." );
	    else if ( broken )
		addreply( 331, "Any password will work." );
	    else
		addreply( 230, "Anonymous user logged in." );
	}
	return;
    }
    if ( username &&
	 strcasecmp( username, "ftp" ) &&
	 strcasecmp( username, "anonymous" ) ) {

	free( (void*) cpwd );

	pw = getpwnam( username );

	if (pw == NULL || pw->pw_shell == NULL ||
		restricted_shell(pw->pw_shell) || pw->pw_uid < useruid) {
	    cpwd = strdup("*");
	} else {
	    rd = "/";
	    if ( chdir( pw->pw_dir ) ) {
		/* non-chrooted users can see everything, so let them
		   in anyway */
		addreply(0, "%s does not exist.  Starting in /", pw->pw_dir);
		chdir( "/" );
	    }
#ifdef CONFIG_FEATURE_SHADOWPASSWDS
	    if ( !strcmp(pw->pw_passwd, "x") && (spw = getspnam( username ) ) )
		cpwd = strdup( spw->sp_pwdp[0] == '@' ? "*" : spw->sp_pwdp );
	    else
#endif
		{
			if(pw->pw_passwd == NULL)       /* Bionic */
				pw->pw_passwd = "*";
			cpwd = strdup( pw->pw_passwd );
		}
	    setregid( pw->pw_gid, pw->pw_gid );
	    initgroups( pw->pw_name, pw->pw_gid );
	    /* never chroot our beloved sysadmin */
	    if ( pw->pw_uid == (uid_t) 0 ) {
		userchroot = 0;
	    } else if (userchroot != 0) {
		/* check if user belongs to the trusted group id */
		if ( pw->pw_gid == chroot_trustedgid ) {
		    userchroot = 0;
		} else {
		    gid_t *extragroups;
		    int n;

		    n = getgroups( 0, NULL );
		    if ( n > 0 ) {
			extragroups = malloc(n * sizeof(gid_t));
			if ( extragroups == NULL )
			    quit421(1, bb_msg_memory_exhausted);
			n = getgroups( n, extragroups );
			while ( userchroot && n-- >= 0 )
			    if ( extragroups[n] == chroot_trustedgid )
				userchroot = 0;
			free( extragroups );
		    }
		}
	    }
	    if ( keeproot )
		setfsuid( pw->pw_uid );
	    else
		setreuid( pw->pw_uid, pw->pw_uid );
	}
	strncpy( account, username, 8 );
	account[8] = '\0';
	addreply( 331, "User %s OK.  Password required.", account );
	loggedin = 0;
    } else {
	if ( (pw = getpwnam( "ftp" ) ) == NULL ||
	     pw->pw_uid == 0 || pw->pw_gid == 0 ||
	     setregid( pw->pw_gid, pw->pw_gid ) ||
	     chroot( pw->pw_dir ) ||
	     !( rd = strdup( pw->pw_dir ) ) ||
	     chdir( "/" ) )
	    quit421(1, "unable to set up secure anonymous FTP");
	if ( keeproot )
	    setfsuid( pw->pw_uid );
	else
	    setreuid( pw->pw_uid, pw->pw_uid );
	dobanner();
	/* the 230 will be overwritten if this is an implicit login */
	addreply( 230, "Anonymous user logged in." );
	strcpy( account, "ftp" );
	rd = strdup( pw->pw_dir );
	loggedin = guest = 1;
	syslog( LOG_INFO, "guest logged in" );
    }
    (void)getcwd( wd, PATH_MAX );
}

static void dopass( const char *password )
{
#ifdef HAVE_CRYPT
    /* accept non anonymous */
    if ( loggedin ||
	 ( cpwd && !strcmp( cpwd, (const char *)crypt( password, cpwd ) ) ) ) {
	gid_t g[NGROUPS_MAX];
	int ngroups;

	/* note that the password stays around for a short while, so
	   root is able to find out what the user's password is */

	if ( !loggedin )
	    candownload = 1;    /* real users can always download */

	ngroups = getgroups( NGROUPS_MAX, g );
	if ( !guest && ngroups > 1 ) {
	    char reply[80];
	    int p;
	    char *q;

	    snprintf( reply, 79, "User %s has group access to:", account );
	    p = strlen( reply );
	    do {
		ngroups--;
		q = getgroup( g[ngroups] );
		if ( p + strlen(q) > 75 ) {
		    reply[p] = '\0';
		    addreply( 0, "%s", reply );
		    *reply = '\0';
		    p = 0;
		}
		reply[p++] = ' ';
		while ( q && *q && !isspace( *q ) )
		    reply[p++] = *q++;
	    } while ( ngroups > 0 );
	    reply[p] = '\0';
	    addreply( 0, "%s", reply );
	}
	if ( userchroot ) {
	    if ( chroot( wd ) < 0 )
		quit421(1, "File system failure");
	    strcpy( wd, "/" );
	    if ( chdir( wd ) < 0 )
		quit421(1, "File system failure");
	} else if ( chroot_trustedgid ) {
	    addreply( 0, "Admin login; no chroot" );
	}
	addreply( 230, "OK.  Current directory is %s", wd );
	syslog( LOG_INFO, "%s logged in", account );
	loggedin = 1;
    } else
#endif
	/* do not accept non anonymous */
	addreply( 530, "Sorry" );
}

static void docwd( const char *dir )
{
    char buffer[PATH_MAX + 256];        /* let's hope... */

    if ( loggedin && !guest && dir && *dir == '~' ) {
	struct passwd *pw;
	int i;
	const char *p;

	i = 0;
	p = dir;
	p++;
	while ( *p && *p != '/' )
	    buffer[i++] = *p++;
	buffer[i] = '\0';

	if ( (pw = getpwnam( i ? buffer : account ) ) )
	    snprintf( buffer, PATH_MAX+255, "%s%s", pw->pw_dir, p );
	else
	    strncpy( buffer, dir, PATH_MAX );
    } else {
	strncpy( buffer, dir, PATH_MAX );
    }

    if ( chdir( *buffer ? buffer :  "/" ) ) {
	snprintf( buffer, PATH_MAX+255, "Can't change directory to %s: %m",
		  dir);
	syslog( LOG_INFO, "%s", buffer );
	addreply( 530, "%s", buffer );
    } else {
	FILE *msg;

	if ( (msg = fopen( ".message", "r" ) ) != NULL ) {
	    int len = fread( (void *)buffer, 1, 1024, msg );
	    fclose( msg );
	    if ( len > 0 && len < 1024 ) {
		buffer[len] = '\0';
		addreply( 0, "%s", buffer );
	    }
	}
	if ( !getcwd( wd, PATH_MAX ) ) {
	    if ( *dir == '/' ) {
		snprintf( wd, PATH_MAX, "%s", dir );    /* already checked */
	    } else {
		int lpm = snprintf( buffer, PATH_MAX, "%s/%s", wd, dir );
		if ( lpm < 0 || lpm >= PATH_MAX )
		    quit421(1, "Path too long");
		strcpy(wd, buffer);
	    }
	}
	addreply( 250, "Changed to %s", wd );
    }
}

static void doretr( const char *name )
{
    int c, f, o, s, skip;
    struct stat st;
    char *p, *buf;
    int left;
    struct timeval started, ended;
    double t;
    double speed;
    char speedstring[30];

    if ( !candownload ) {
	addreply( 550, "The load was %3.2f when you connected.  We do not "
		  "allow downloads\nby anonymous users when the load is "
		  "that high.  Uploads are always\nallowed.", load );
	return;
    }

    f = open( name, O_RDONLY );
    if ( f < 0 ) {
	char buffer[PATH_MAX + 40];
	snprintf( buffer, PATH_MAX+39, "Can't open %s", name );
	error( 550, buffer );
	return;
    }
    if ( fstat( f, &st ) ) {
	close( f );
	error( 451, "can't find file size" );
	return;
    }
    if ( restartat && ( restartat > st.st_size ) ) {
	addreply( 451, "Restart offset %" P_ST_SIZE "d is too large for file size %" P_ST_SIZE "d.\n"
		  "Restart offset reset to 0.",
		  restartat, st.st_size );
	return;
    }
    if ( !S_ISREG( st.st_mode ) ) {
	close( f );
	addreply( 450, "Not a regular file" );
	return;
    }
    if ( warez && ( st.st_uid == warez ) && guest ) {
	close( f );
	addreply( 550, "This file has been uploaded by an anonymous user.  "
		  "It has not\nyet been approved for downloading by "
		  "the site administrators.\n" );
	return;
    }

    c = opendata();
    if ( !c ) {
	close( f );
	return;
    }

    if ( restartat == st.st_size ) {
	/* some clients insist on doing this.  I can't imagine why. */
	addreply(226, "Nothing left to download.  Restart offset reset to 0.");
	close( f );
	close( c );
	return;
    }

    if ( (s = fcntl( c, F_GETFL, 0 ) ) < 0 ) {
	error( 451, "fcntl failed" );
	close( f );
	close( c );
	return;
    }
    s |= O_NDELAY;
    fcntl( c, F_SETFL, s );

    if ( type == 1 ) {
	addreply( 0,
		  "NOTE: ASCII mode requested, but binary mode used" );
	if ( (time( 0 ) % 1000000 ) == 0 )
	    addreply( 0, "The computer is your friend.  Trust the computer" );
    }
    if ( st.st_size - restartat > 4096 )
	addreply( 0, "%.1f kbytes to download",
		  (st.st_size - restartat) / 1024.0 );

    doreply();

    (void)gettimeofday( &started, NULL );

    o = restartat & ~262143;
    skip = restartat - o;
    while ( o < st.st_size ) {
	left = st.st_size - o;
	if ( left > 262144 )
	    left = 262144;
	buf = mmap( 0, left, PROT_READ, MAP_FILE | MAP_SHARED, f, o );
	if ( buf == (char *)-1 ) {
	    error( 451, "mmap of file failed" );
	    close( f );
	    close( c );
	    return;
	}
	p = buf;
	o += left;
	s = left;
	while ( left > skip ) {
	    size_t w;

	    w = write( c, p+skip, (size_t) (left - skip) );
	    if ( (int ) w < 0 ) {
		if ( errno == EAGAIN ) {
		    /* wait idletime seconds for progress */
		    fd_set rs;
		    fd_set ws;
		    struct timeval tv;

		    FD_ZERO( &rs );
		    FD_ZERO( &ws );
		    FD_SET( 0, &rs );
		    FD_SET( c, &ws );
		    tv.tv_sec = idletime;
		    tv.tv_usec = 0;
		    select( c + 1, &rs, &ws, NULL, &tv );
		    if ( FD_ISSET( 0, &rs ) ) {
			/* we assume is is ABRT since nothing else is legal */
			addreply( 426, "Transfer aborted" );
			munmap( buf, s );
			close( f );
			close( c );
			return;
		    } else if ( !( FD_ISSET( c, &ws ) ) ) {
			/* client presumably gone away */
			syslog( LOG_INFO,
				"died: %d seconds without download progress",
				idletime );
			exit( 11 );
		    }
		    w = 0;
		} else {
		    error( 450, "Error during write to data connection" );
		    close( f );
		    close( c );
		    return;
		}
	    }
	    left -= w;
	    p += w;
	}
	skip = 0;
	munmap( buf, s );
    }

    (void)gettimeofday( &ended, NULL );
    t = ( ended.tv_sec + ended.tv_usec / 1000000.0 ) -
	( started.tv_sec + started.tv_usec / 1000000.0 );
    addreply( 226, "File written successfully" );
    if ( t && ( st.st_size - restartat ) > 2*window )
	speed = ( st.st_size - restartat - window ) / t;
    else
	speed = 0.0;
    addreply( 0, "%.3f seconds (measured by the server), %.2f %sb/s", t,
	      speed > 524288 ? speed / 1048576 : speed / 1024,
	      speed > 524288 ? "M" : "K" );
    close( f );
    close( c );

    downloaded = downloaded + st.st_size - restartat;

    if ( restartat ) {
	restartat = 0;
	addreply( 0, "Restart offset reset to 0." );
    }

    snprintf( speedstring, 29, " (%.2fKB/sec)", speed/1024 );
    syslog( LOG_INFO, "%s%s%s%s downloaded%s",
	    rd ? rd :  "",
	    *name == '/' ? "" : wd,
	    ( *name != '/' && ( !*wd || wd[strlen( wd ) - 1] != '/' ) )
	    ? "/" : "",
	    name,
	    speed > 0.1 ? speedstring : "" );
}

static void dorest (const char *name)
{
    char *endptr;
    restartat = BB_st_size_strtol( name, &endptr, 10 );
    if ( *endptr ) {
	restartat = 0;
	addreply( 501, "RESTART needs numeric parameter.\n"
		  "Restart offset set to 0." );
    } else {
	syslog( LOG_NOTICE, "info: restart %" P_ST_SIZE "d", restartat );
	addreply( 350,
		  "Restarting at %ld. Send STOR or RETR to initiate transfer.",
		  restartat );
    }
}

static void dodele( const char *name )
{
    if ( guest )
	addreply( 550, "Anonymous users can not delete files." );
    else if ( unlink( name ) )
	addreply( 550, "Could not delete %s: %m", name);
    else
	addreply( 250, "Deleted %s", name );
}

static void dostor( const char *name )
{
    int c, f;
    char *p;
    char buf[16384];
    int r;
    bbst_size_t filesize;
    struct statfs statfsbuf;
    struct stat st;
    // Added for ascii upload
    uint i,j;
    char cpy[16384];
    char *q;

    filesize = 0;

    if ( type < 1 ) {
	addreply( 503, "Only ASCII and binary modes are supported" );
	return;
    }

    if ( !stat( name, &st ) )  {
	if ( guest ) {
	    addreply( 553,
		      "Anonymous users may not overwrite existing files" );
	    return;
	}
    } else if ( errno != ENOENT ) {
	error( 553, "Can't check for file presence" );
	return;
    }
    f = open( name, O_CREAT | O_TRUNC | O_WRONLY, 0600 );
    if ( f < 0 ) {
	error( 553, "Can't open file" );
	return;
    }
    if ( restartat && lseek(f, restartat, SEEK_SET) < 0) {
	error (451, "can't seek" );
	return;
    }

    c = opendata();
    if ( !c ) {
	close( f );
	return;
    }
    doreply();

    do {
	/* wait idletime seconds for data to be available */
	fd_set rs;
	struct timeval tv;

	FD_ZERO( &rs );
	FD_SET( 0, &rs );
	FD_SET( c, &rs );
	tv.tv_sec = idletime;
	tv.tv_usec = 0;
	select( c + 1, &rs, NULL, NULL, &tv );
	if ( FD_ISSET( 0, &rs ) ) {
	    addreply( 0, "ABRT is the only legal command while uploading" );
	    addreply( 426, "Transfer aborted" );
	    close( f );
	    close( c );
	    addreply( 0, "%s %s", name,
		      unlink( name ) ? "partially uploaded" : "removed" );
	    return;
	} else if ( !( FD_ISSET( c, &rs ) ) ) {
	    /* client presumably gone away */
	    unlink( name );
	    syslog( LOG_INFO, "died: %d seconds without upload progress",
		    idletime );
	    exit( 20 );
	}
	r = read( c, &buf, 16384 );
	if ( r > 0 ) {
	    p = buf;

	    filesize += r;
	    while ( r ) {
		size_t w;

		if ( type == 1 ) {
		    int k = 0;
		    i = 0;
		    j = 0;
		    while ( i < (size_t) r ) {
			if ( p[i] == '\r' ) {
			    i++;
			    k++;
			}
			cpy[j++] = buf[i++];
		    }
		    q = cpy;
		    r -= k;
		    w = write( f, q, (size_t) r );
		} else {
		    w = write( f, p, (size_t) r );
		}

		if ( (signed int)w < 0 ) {
		    error( -450, "Error during write to file" );
		    close( f );
		    close( c );
		    addreply( 450, "%s %s", name,
			      unlink(name)?"partially uploaded":"removed" );
		    return;
		}
		r -= w;
		p += w;
	    }
	    r = 1;
	} else if ( r < 0 ) {
	    error( -451, "Error during read from data connection" );
	    close( f );
	    close( c );
	    addreply( 451, "%s %s", name,
		      unlink( name ) ? "partially uploaded" : "removed" );
	    return;
	}
    } while ( r > 0 );

    fchmod( f, 0644 );
    addreply( 226, "File written successfully" );

    if ( fstatfs( f, &statfsbuf ) == 0 ) {
	double space;

	space = (double)statfsbuf.f_bsize * (double)statfsbuf.f_bavail;
	if ( space > 524288 )
	    addreply( 0, "%.1f M%s", "bytes free disk space", space / 1048576 );
	else
	    addreply( 0, "%f K%s", "bytes free disk space", space / 1024 );
    }
    close( f );
    close( c );
    uploaded += filesize;
    syslog( LOG_INFO, "%s%s%s%s uploaded",
	    rd ? rd :  "",
	    *name == '/' ? "" : wd,
	    ( *name != '/' && ( !*wd || wd[strlen( wd ) - 1] != '/' ) )
	    ? "/" : "",
	    name );
    if ( restartat ) {
	restartat = 0;
	addreply( 0, "Restart offset reset to 0." );
    }
}

/* next two functions contributed by Patrick Michael Kane <modus@asimov.net> */
static void domkd( const char *name )
{
    if ( guest )
	addreply( 550, "Sorry, anonymous users are not allowed to %s",
		  "make directories." );
    else if ( (mkdir( name, 0755 ) ) < 0 )
	error( 550, "Can't create directory" );
    else
	addreply( 257, "MKD command successful." );
}


static void dormd( const char *name )
{
    if ( guest )
	addreply( 550, "Sorry, anonymous users are not allowed to %s",
		  "remove directories." );
    else if ( (rmdir( name ) ) < 0 )
	error( 550, "Can't remove directory" );
    else
	addreply( 250, "RMD command successful." );
}

static void domdtm( const char *name )
{
    struct stat st;
    struct tm *t;

    if ( !name || !*name ) {
	addreply( 500, "Command not understood" );
    } else if ( lstat( name, &st ) ) {
#ifdef DEBUG
	if ( debug )
	    addreply( 0, "arg is %s, wd is %s", name, wd );
#endif
	addreply( 550, "Unable to stat()" );
    } else if ( !S_ISREG( st.st_mode ) ) {
	addreply( 550, "Not a regular file" );
    } else {
	t = gmtime( (time_t * ) & st.st_mtime );
	if ( !t ) {
	    addreply( 550, "gmtime() returned NULL" );
	} else {
	    addreply( 213, "%04d%02d%02d%02d%02d%02d",
		     t->tm_year + 1900, t->tm_mon + 1, t->tm_mday,
		     t->tm_hour, t->tm_min, t->tm_sec );
	}
    }
}

static void dopasv( int useepsv )
{
    socklen_t socklen;
    unsigned int a;
    int p;
    unsigned int on;
    struct sockaddr_in dataconn;        /* my data connection endpoint */
    int firstporttried;

    if ( datafd ) {                     /* for buggy clients */
	close_datafd();
    }

    datafd = socket( AF_INET, SOCK_STREAM, 0 );
    if ( datafd < 0 ) {
	error( 425, "Can't open passive connection" );
	datafd = 0;
	return;
    }
    on = 1;
    if ( setsockopt( datafd, SOL_SOCKET, SO_REUSEADDR,
		     &on, sizeof(on) ) < 0 ) {
	error( 421, "setsockopt" );
	return;
    }

    dataconn = ctrlconn;
    if ( firstport && lastport )
	firstporttried = firstport + ( getpid() % (lastport-firstport+1) );
    else
	firstporttried = 0;

    p = firstporttried;
    dataconn.sin_port = ntohs( p );
    while ( bind( datafd, (struct sockaddr *)&dataconn, sizeof(dataconn) ) < 0 ) {
	if ( firstporttried ) {
	    p--;
	    if ( p < firstport )
		p = lastport;
	}
	if ( firstporttried == 0 || firstporttried == p ) {
	    if ( firstporttried )
		addreply( 0, "TCP ports %d to %d inclusive are all busy",
			  firstport, lastport );
	    error( 425, "Can't bind to socket" );
	    close_datafd();
	    return;
	}
	dataconn.sin_port = ntohs( p );
    }
    (void)listen( datafd, 1 );  /* maybe 0 on some unices? */

    socklen = sizeof( dataconn );
    if ( getsockname( datafd, (struct sockaddr *)&dataconn, &socklen ) < 0 ) {
	error( 425, "Can't getsockname( dataconn )" );
	close_datafd();
	return;
    }
    a = ntohl( dataconn.sin_addr.s_addr );
    p = ntohs( (unsigned short int ) ( dataconn.sin_port ) );
    if ( useepsv )
	addreply( 229, "Extended Passive mode OK (|||%d|)", p );
    else
	addreply( 227, "Passive mode OK (%d,%d,%d,%d,%d,%d)",
		  (a >> 24) & 255, (a >> 16) & 255, (a >> 8) & 255, a & 255,
		  (p >> 8) & 255, p & 255 );
    passive = 1;
}


static void doport( unsigned int a, unsigned int p )
{
    struct sockaddr_in dataconn;        /* his endpoint */
    int on;

    if ( datafd ) { /* for buggy clients saying PORT over and over */
	close_datafd();
    }

    datafd = socket( AF_INET, SOCK_STREAM, 0 );
    if ( datafd < 0 ) {
	error( 425, "Can't make data socket" );
	datafd = 0;
	return;
    }
    on = 1;
    if ( setsockopt( datafd, SOL_SOCKET, SO_REUSEADDR,
		     &on, sizeof(on) ) < 0 ) {
	error( 421, "setsockopt" );
	return;
    }
    dataconn = ctrlconn;
    dataconn.sin_port = htons( (short ) 20 ); /* FTP data connection port */
    if ( bind( datafd, (struct sockaddr *)&dataconn, sizeof(dataconn) ) < 0 ) {
	error( -220, "bind" );
	close_datafd();
	return;
    }

#ifdef DEBUG
    if ( debug )
	addreply( 0, "My data connection endpoint is %s:%d",
		  inet_ntoa( *(struct in_addr *)&dataconn.sin_addr.s_addr ),
		  ntohs( dataconn.sin_port ) );
#endif
    peerdataport = p;

    if ( htonl( a ) != peer.sin_addr.s_addr ) {
	addreply( 425, "Will not open connection to %d.%d.%d.%d (only to %s)",
		  (a >> 24) & 255, (a >> 16) & 255, (a >> 8) & 255, a & 255,
		  inet_ntoa( peer.sin_addr ) );
	close_datafd();
	return;
    }

    passive = 0;

    addreply( 200, "PORT command successful" );
}

static void dosize( const char *name )
{
    struct stat st;

    if ( !name || !*name ) {
	addreply( 500, "Command not understood" );
    } else if ( lstat( name, &st ) ) {
#ifdef DEBUG
	if ( debug )
	    addreply( 0, "arg is %s, wd is %s", name, wd );
#endif
	addreply( 550, "Unable to stat()" );
    } else if ( !S_ISREG( st.st_mode ) ) {
	addreply( 550, "Not a regular file" );
    } else {
	addreply( 213, "%" P_ST_SIZE "d", st.st_size );
    }
}


static void dotype( const char *arg )
{
    replycode = 200;    /* bloody awful hack */

    if ( !arg || !*arg ) {
	addreply( 501, "TYPE needs an argument\n"
		 "Only A(scii), I(mage) and L(ocal) are supported" );
    } else if ( tolower(*arg) == 'a' )
	type = 1;
    else if ( tolower(*arg) == 'i' )
	type = 2;
    else if ( tolower(*arg) == 'l' ) {
	if ( arg[1] == '8' ) {
	    type = 2;
	} else if ( isdigit( arg[1] ) ) {
	    addreply( 504, "Only 8-bit bytes are supported" );
	} else {
	    addreply( 0, "Byte size not specified" );
	    type = 2;
	}
    } else {
	addreply( 504, "Unknown TYPE: %s", arg );
    }

    addreply( 0, "TYPE is now %s", ( type > 1 ) ? "8-bit binary" : "ASCII" );
}


static void dostru( const char *arg )
{
    if ( !arg || !*arg )
	addreply( 500, "No arguments\n"
		 "Not that it matters, only STRU F is supported" );
    else if ( strcasecmp( arg, "F" ) )
	addreply( 504, "STRU %s is not supported\nOnly F(ile) is supported",
		 arg );
    else
	addreply( 200, "F OK" );
}


static void domode( const char *arg )
{
    if ( !arg || !*arg )
	addreply( 500, "No arguments\n"
		 "Not that it matters, only MODE S is supported" );
    else if ( strcasecmp( arg, "S" ) )
	addreply( 504, "MODE %s is not supported\n"
		 "Only S(tream) is supported", arg );
    else
	addreply( 200, "S OK" );
}


static void dornfr( const char *name )
{
    struct stat st;

    if ( guest ) {
	addreply( 550, "Sorry, anonymous users are not allowed to %s",
							"rename files." );
    } else if ( ( stat( name, &st ) ) == 0 ) {
	if ( renamefrom ) {
	    addreply( 0, "Aborting previous rename operation." );
	    (void) free( renamefrom );
	}
	renamefrom = strdup( name );
	addreply( 350, "RNFR accepted - file exists, ready for destination." );
    } else {
	addreply( 550, "File does not exist!" );
    }
}


/* rnto can return 550, which is not strictly allowed */

static void dornto( const char *name )
{
    struct stat st;

    if ( guest )
	addreply( 550, "Sorry, anonymous users are not allowed to %s",
							"rename files." );
    else if ( ( stat( name, &st ) ) == 0 )
	addreply( 550, "RENAME Failed - destination file already exists." );
    else if ( !renamefrom )
	addreply( 503, "Need RNFR before RNTO" );
    else if ( rename( renamefrom, name ) < 0 )
	addreply( 550, "Rename failed: %m");
    else
	addreply( 250, "File renamed." );

    (void) free( renamefrom );
    renamefrom = 0;
}


static char *u_g_getnames( const char *fn )
{
    char *u_g_mmap = NULL;
    int f;

    /* uid_t can have 65536 values on linux */

    f = open(fn, O_RDONLY );
    if ( f >= 0 ) {
	u_g_mmap = mmap( 0, (size_t) 9 * 65536,
		      PROT_READ, MAP_FILE | MAP_SHARED, f, 0 );
	if ( u_g_mmap == (void *) -1 ) {
	    u_g_mmap = NULL;
	    close( f );
	}
    }
    return u_g_mmap;
}

static int daemons( void )
{
    char buf[2049];
    int f;
    int r;

    int b, e;
    int c;
    int d = 0;

    f = open( "/proc/net/tcp", O_RDONLY );
    if ( f < 0 )
	return d;

    buf[2048] = '\0';

    b = 0;
    e = 0;
    do {
	r = read( f, buf+e, 2048-e );
	if ( r <= 0 ) /* ignore errors.  0 is okay, in fact common. */
	    break;
	e += r;

	/* b now is offset of the start of the first line to be parsed
	   and e the end of the available data */
	c = b;
	while( c < e && buf[c] != '\n' )
	    c++;
	while ( c < e ) {
	    buf[c++] = '\0';
	    while( b < c && buf[b] != ':' && buf[b] != '\n' )
		b++;
	    if ( b < c && buf[b] == ':' ) {
		b++;
		while( b < e && buf[b] != ':' )
		    b++;
		b++;
		if ( strtoul( buf+b, NULL, 16 ) == 21 )
		    d++;
	    }
	    b = c;
	    while( c < e && buf[c] != '\n' )
		c++;
	}
	if ( e > b )
	    (void)memmove( buf, buf+b, e-b );
	e = e-b;
	b = 0;
    } while( 1 );
    close( f );
    if ( d > 0 )
	d--; /* don't count inetd, but also don't report -1 */
    return d;
}

static void parser( void )
{
    char *arg;
    int n;

    while ( 1 ) {
	doreply();

	if ( !sfgets() ) {
	    addreply( 421, "Timeout (%d seconds).", idletime );
	    syslog( LOG_INFO, "exiting due to timeout (idle time %d)",
		    idletime );
	    return;
	}
#ifdef DEBUG
	if ( debug )
	    addreply( 0, "%s", cmd );
#endif
	n = 0;
	while ( isalpha( cmd[n] ) && n < cmdsize ) {
	    cmd[n] = tolower( cmd[n] );
	    n++;
	}

	if ( !n ) {
	    addreply( 221,
		      "Goodbye.  You uploaded %d and downloaded %d kbytes.",
		      (uploaded+1023)/1024, (downloaded+1023)/1024 );
	    continue;
	}
	while ( isspace( cmd[n] ) && n < cmdsize )
	    cmd[n++] = '\0';

	arg = cmd + n;

	while ( cmd[n] && n < cmdsize )
	    n++;
	n--;

	while ( isspace( cmd[n] ) )
	    cmd[n--] = '\0';

#ifdef DEBUG
	if ( logging )
	    syslog( LOG_DEBUG, "command %s %s",
		    cmd, strcasecmp( cmd, "pass" ) ? arg : "<password>" );
#endif
	if ( strlen( cmd ) > 10 ) {
	    addreply( 500, "Unknown command." );
	} else if ( strlen( arg ) >= PATH_MAX ) { /* ">=" on purpose. */
	    addreply( 501, "Cannot handle %d-character file names" );
	} else if ( !strcasecmp( cmd, "user" ) ) {
	    douser( arg );
	} else if ( !strcasecmp( cmd, "pass" ) ) {
	    dopass( arg );
	} else if ( !strcasecmp( cmd, "quit" ) ) {
	    addreply( 221,
		      "Goodbye.  You uploaded %d and downloaded %d kbytes.",
		      (uploaded+1023)/1024, (downloaded+1023)/1024 );
	    return;
	} else if ( !strcasecmp( cmd, "noop" ) ) {
	    addreply( 200, "NOOP command successful" );
	} else if ( !strcasecmp( cmd, "syst" ) ) {
	    addreply( 215, "UNIX Type: L8" );
	} else if ( !strcasecmp( cmd, "port" ) ||
		    !strcasecmp( cmd, "eprt" ) ) {
	    /* don't auto-login for PORT or PASV, but do auto-login
	       for the command which _uses_ the data connection */
	    unsigned int a1, a2, a3, a4, p1, p2;
	    if ( epsvall ) {
		addreply( 501, "Cannot use PORT/EPRT after EPSV ALL" );
	    } else if ( cmd[0] == 'e' && strncasecmp( arg, "|2|", 3 ) == 0 ) {
		addreply( 522, "IPv6 not supported, use IPv4 (1)" );
	    } else if ( !keeproot ) {
		addreply( 502, "PORT/EPRT is disabled for security reasons." );
		addreply( 0, "Please use PASV/EPSV instead (\"passive on\")" );
	    } else if ( cmd[0] == 'e' &&
			5 == sscanf( arg, "|1|%u.%u.%u.%u|%u|",
			      &a1, &a2, &a3, &a4, &p1 ) &&
		 a1 < 256 && a2 < 256 && a3 < 256 && a4 < 256 &&
		 p1 < 65536 ) {
		doport( (a1 << 24) + (a2 << 16) + (a3 << 8) + a4, p1 );
	    } else if ( cmd[0] == 'p' &&
			6 == sscanf( arg, "%u,%u,%u,%u,%u,%u",
				     &a1, &a2, &a3, &a4, &p1, &p2 ) &&
			a1 < 256 && a2 < 256 && a3 < 256 && a4 < 256 &&
			p1 < 256 && p2 < 256 ) {
		doport( (a1 << 24) + (a2 << 16) + (a3 << 8) + a4,
			( (p1 << 8 ) + p2 ) );
	    } else {
		addreply( 501, "Syntax error." );
	    }
	} else if ( !strcasecmp( cmd, "pasv" ) ) {
	    dopasv( 0 );
	} else if ( !strcasecmp( cmd, "epsv" ) ) {
	    if ( !strcasecmp( arg, "all" ) ) {
		addreply( 220, "OK; will reject non-EPSV data connections" );
		epsvall++;
	    } else if ( !strcasecmp( arg, "2" ) ) {
		addreply( 522, "IPv6 not supported, use IPv4 (1)" );
	    } else if ( strlen( arg ) == 0 || !strcasecmp( arg, "1" ) ) {
		dopasv( 1 );
	    } else {
	    }
	} else if ( !strcasecmp( cmd, "pwd" ) ||
		    !strcasecmp( cmd, "xpwd" ) ) {
	    if ( loggedin )
		addreply( 257, "\"%s\"", wd );
	    else
		addreply( 550, "Not logged in" );
	} else if( !strcasecmp( cmd, "auth" ) ) {
	    /* RFC 2228 Page 5 Authentication/Security mechanism (AUTH) */
	    addreply( 502, "Security extensions not implemented" );
	} else {
	    /* from this point, all commands trigger an automatic login */
	    douser( NULL );

	    if ( !strcasecmp( cmd, "cwd" ) ) {
		docwd( arg );
	    } else if ( !strcasecmp( cmd, "cdup" ) ) {
		docwd( ".." );
	    } else if ( !strcasecmp( cmd, "retr" ) ) {
		if ( arg && *arg )
		    doretr( arg );
		else
		    addreply( 501, "No file name given." );
	    } else if ( !strcasecmp( cmd, "rest" ) ) {
		if (arg && *arg)
		    dorest(arg);
		else
		    addreply (501, "No restart point");
	    } else if ( !strcasecmp( cmd, "dele" ) ) {
		if ( arg && *arg )
		    dodele( arg );
		else
		    addreply( 501, "No file name given." );
	    } else if ( !strcasecmp( cmd, "stor" ) ) {
		if ( arg && *arg )
		    dostor( arg );
		else
		    addreply( 501, "No file name given." );
	    } else if ( !strcasecmp( cmd, "mkd" ) ||
			!strcasecmp( cmd, "xmkd" ) ) {
		if ( arg && *arg )
		    domkd( arg );
		else
		    addreply( 501, "No directory name." );
	    } else if ( !strcasecmp( cmd, "rmd" ) ||
			!strcasecmp( cmd, "xrmd" ) ) {
		if ( arg && *arg )
		    dormd( arg );
		else
		    addreply( 550, "No directory name." );
	    } else if ( !strcasecmp( cmd, "list" ) ||
			!strcasecmp( cmd, "nlst" ) ) {
		donlist( (arg && *arg ) ? arg :  "-l" );
	    } else if ( !strcasecmp( cmd, "type" ) ) {
		dotype( arg );
	    } else if ( !strcasecmp( cmd, "mode" ) ) {
		domode( arg );
	    } else if ( !strcasecmp( cmd, "stru" ) ) {
		dostru( arg );
	    } else if ( !strcasecmp( cmd, "abor" ) ) {
		addreply( 226, "ABOR succeeded." );
	    } else if ( !strcasecmp( cmd, "site" ) ) {
		char *sitearg;

		sitearg = arg;
		while ( sitearg && *sitearg && !isspace( *sitearg ) )
		    sitearg++;
		if ( sitearg )
		    *sitearg++ = '\0';

		if ( !strcasecmp( arg, "idle" ) ) {
		    if ( !*sitearg ) {
			addreply( 501, "SITE IDLE: need argument" );
		    } else {
			unsigned long int i;

			i = strtoul( sitearg, &sitearg, 10 );
			if ( sitearg && *sitearg ) {
			    addreply( 501, "Garbage (%s) after value (%u)",
				      sitearg, i );
			} else {
			    if ( i > 7200 )
				i = 7200;
			    if ( i < 10 )
				i = 10;
			    idletime = i;
			    addreply( 200, "Idle time set to %u seconds", i );
			}
		    }
		} else if ( arg && *arg ) {
		    addreply( 500, "SITE %s unknown", arg );
		} else {
		    addreply( 500, "SITE: argument needed" );
		}
#ifdef DEBUG
	    } else if ( !strcasecmp( cmd, "xdbg" ) ) {
		debug++;
		addreply( 200,
			  "XDBG command succeeded, debug level is now %d.",
			  debug );
#endif
	    } else if ( !strcasecmp( cmd, "mdtm" ) ) {
		domdtm( (arg && *arg ) ? arg :  "" );
	    } else if ( !strcasecmp( cmd, "size" ) ) {
		dosize( (arg && *arg ) ? arg :  "" );
	    } else if ( !strcasecmp( cmd, "rnfr" ) ) {
		if ( arg && *arg )
		    dornfr( arg );
		else
		    addreply( 550, "No file name given." );
	    } else if ( !strcasecmp( cmd, "rnto" ) ) {
		if ( arg && *arg )
		    dornto( arg );
		else
		    addreply( 550, "No file name given." );
	    } else {
		addreply( 500, "Unknown command." );
	    }

	    if ( strcasecmp( cmd, "rest" ) )
		restartat = 0;
	}
    }
}

static int mkdir_subpath_open(const char *dir)
{
    int f;
    char *s0, *s;

    s0 = bb_simplify_path(dir);
    for(s = s0 + 1; (s = strchr(s, '/')) != NULL; s++) {
	*s = '\0';
	if(mkdir(s0, 0755) != 0 && errno != EEXIST)
		bb_perror_msg_and_die("mkdir(%s)", s0);
	*s = '/';
    }
    f = bb_xopen(s0, O_WRONLY|O_CREAT|O_TRUNC| BB_OPEN_ERR, 0644);
    free(s0);
    return f;
}

static void write_ftpd_pw(int f, long id, const char *name)
{
	char dummy[16];

	lseek(f, 9 * id, SEEK_SET);
	sprintf(dummy, "%-.8s", name);
	write(f, dummy, 9);
}

static int mkusers_main(void)
{
    int f;
    struct passwd * pw;
    struct group * gr;

    f = mkdir_subpath_open(ftpd_users);
    while((pw = getpwent()) != NULL)
	write_ftpd_pw(f, pw->pw_uid, pw->pw_name);
    close(f);

    f = mkdir_subpath_open(ftpd_groups);
    while((gr = getgrent()) != NULL)
	write_ftpd_pw(f, gr->gr_gid, gr->gr_name);
    close(f);
    return 0;
}

int ftpd_main(void)
{
    int fodder;
    char *he;
    char hbuf[NI_MAXHOST];
    struct passwd * pw;
    FILE *f;
    double maxload = 0.0;
    time_t tmt;
    struct tm *t;
    struct rusage ru;
    int maxusers = 0;
    socklen_t sl;
    char *xopt, *user_limit, *load_limit, *port_range, *uid_limit;

#ifdef DEBUG
    int d_opt_count = 0;
    struct sigaction sa;

    /* logging stuff.  use sigaction so they're not reset */
    bb_signal((1 << SIGUSR1) | (1 << SIGUSR2), sighandler);
    /* we pretend urgent data doesn't exist */
    bb_signal((1 << SIGURG), SIG_IGN);
#endif

    type = 1;

    /* the first openlog will probably be overwritten by -f ten lines
       below, so don't put too much code between them. */
    bb_openlog(LOG_CONS | LOG_PID, LOG_LOCAL2);
    log_stderr = 0;
    log_syslog = LOG_ERR;

#ifdef DEBUG
    fodder = bb_xgetopt_ulflags("aMsx:bc:m:p:u:d\0dd",
		&xopt, &user_limit, &load_limit, &port_range, &uid_limit
		, &d_opt_count);
#else
    fodder = bb_xgetopt_ulflags("aMsx:bc:m:p:u:",
		&xopt, &user_limit, &load_limit, &port_range, &uid_limit);
#endif
    keeproot = fodder & (1<<0); /* -a */
    if(fodder & (1<<1)) {
	/* -M */
	log_stderr = 1;
	log_syslog = 0;
	return mkusers_main();
    }
    if(fodder & (1<<2)) {
	/* -s */
	pw = getpwnam( "ftp" );
	if ( pw != NULL )
		warez = pw->pw_uid;
	else
		bb_error_msg("Configuration error: can't find \"ftp\" uid");
    }
    if(fodder & (1<<3)) {
	/* -x group */
	struct group * gr = getgrnam( xopt );
	if ( gr != NULL )
		chroot_trustedgid = gr->gr_gid;
	else
		bb_error_msg("Configuration error: "
			"can't find group \"%s\" (for -x)", xopt );
	/* but the ftpd can run in this case */
	userchroot = 1;
    }
    broken = fodder & (1<<4);   /* -b */
    if(fodder & (1<<5)) {
	/* -c user_limit */
	maxusers = bb_xparse_i_number(BB_PN_SET_LOWER1 | BB_PN_RETURN_ERR | BB_PN_ERR_STR1,
					user_limit, "Illegal user limit: %s");
    }
    if(fodder & (1<<6)) {
	/* -m load_limit */
	if ( safe_strtod(load_limit, &maxload) || maxload <= 0.0 )
		bb_error_msg("Illegal load limit: %s", load_limit);
    }
    if(fodder & (1<<7)) {
	/* -p port:range */
	int ret;

	ret = sscanf(port_range, "%d:%d", &firstport, &lastport);
	if ( ret != 2 || firstport < 1024 || lastport > 65535 ||
						lastport <= firstport )
		bb_error_msg("Illegal port range: %s", port_range);
    }
    if(fodder & (1<<8)) {
	/* -u uid_limit */
	int tmp = bb_xparse_i_number(BB_PN_SET_UPPER | BB_PN_SILENT_RE, uid_limit, 65535);
	if ( tmp < 0 ) {
		/* have to quit in this case - anything else might be
		   a security problem */
		quit421(16, "Configuration error: Illegal uid limit: %s", uid_limit);
	}
	useruid = (uid_t) tmp;
    }
#ifdef DEBUG
    if(d_opt_count) {
	/* -d */
	logging = d_opt_count == 1 ? 1 : 2;
    }
#endif
    /* mmap in ftpd_users and ftpd_groups, for use by
       ls. do it before the chroot if ls is supposed to show the same
       stuff. */
    users = u_g_getnames(ftpd_users);
    groups = u_g_getnames(ftpd_groups);

    if ( firstport || maxusers ) {
	int u;

	if ( firstport ) {
	    int portmax = (lastport-firstport+1)/2;
	    if ( !maxusers || maxusers > portmax )
		maxusers = portmax; /* ... so we don't run out of ports */
	}

	u = daemons();
	if ( u > maxusers )
	    quit421(5, "%lu users (the maximum) are already logged in", maxusers );
	addreply( 0, "You are user number %d of %d allowed.", u, maxusers );
    }

    load = -1.0;
    if ( (f = fopen( "/proc/loadavg", "r" ) ) != NULL) {
	 fscanf( f, "%lf", &load );
	 fclose( f );
    }

    sl = sizeof(struct sockaddr_in);

    if ( getsockname( 0, (struct sockaddr *)&ctrlconn, &sl ) )
	quit421(3, "Cannot getsockname: %m");

    loggedin = 0;

    if ( getpeername( 0, (struct sockaddr *)&peer, &sl ) )
	quit421(9, "Cannot getpeername: %m");

    syslog( LOG_INFO, "connection from %s", inet_ntoa( peer.sin_addr ) );

    he = bb_xgethostbyaddr(&ctrlconn.sin_addr.s_addr, hbuf, BB_XGETHOSTBYADDR_FLG_IN_ADDR);

    /* if you want authenticated sites that don't use /etc/passed,
       this is a nice place to hack.  the below is an example of how
       it can be done, using SITEDIR/etc/passwd */

    if ( he ) {
	char *name;

	if ( asprintf( &name, "/var/adm/ftp/%s/.", he ) < 0 )
	    quit421(1, bb_msg_memory_exhausted);
	if ( !chdir( name ) ) {
	    int l = strlen( he ) + 3;

	    rd = malloc(l);
	    /* it'd be much better with a separate name for "ftp" here */
	    pw = getpwnam( "ftp" );
	    if ( !rd || !pw ||
		 setregid( pw->pw_gid, pw->pw_gid ) ||
		 initgroups( pw->pw_name, pw->pw_gid ) ||
		 chroot( name ) ||
		 chdir( "/" ) )
		quit421(1, "Unable to cage in anonymous user");
	    snprintf( rd, l, "%s:", he );
	    if ( keeproot )
		setfsuid( pw->pw_uid );
	    else
		setreuid( pw->pw_uid, pw->pw_uid );
	    dobanner();
	    loggedin = guest = 1;
	    syslog( LOG_INFO, "logged in guest of %s", he );
	}
	free(name);
	cpwd = strdup( "*" );
    }

    chdir( "/" );
    strcpy( wd, "/" );

    fodder = IPTOS_LOWDELAY;
    w_setsockopt(0, SOL_IP, IP_TOS, (char *)&fodder);
    fodder = 1;
    w_setsockopt(0, SOL_SOCKET, SO_OOBINLINE, (char *)&fodder);

    tmt = time( NULL );
    t = localtime( (time_t *) &tmt );
    if ( t != 0 && load >= 0.0 )
	addreply( 220, "Local time is now %02d:%02d and the load is %3.2f.",
		  t->tm_hour, t->tm_min, load );
    else
	bb_error_msg("unable to read load average or current time" );

    if ( loggedin )
	addreply( 220, "Only anonymous FTP is allowed at %s.", he );

    addreply( 220, "You will be disconnected after %d seconds of inactivity.",
	      idletime );

    candownload = maxload <= 0.0 || load < maxload;

    parser();

    if ( getrusage( RUSAGE_SELF, &ru ) == 0 ) {
	unsigned long s, u;

	u = ( ru.ru_utime.tv_usec + ru.ru_stime.tv_usec + 500 ) / 1000;
	s = ru.ru_utime.tv_sec + ru.ru_stime.tv_sec;
	if ( u > 999 ) {
	    u -= 1000;
	    s++;
	}
	addreply( 0, "CPU time spent on you: %ld.%03ld seconds.", s, u );
    }
    doreply();
    sleep( 1 );
    return 0;
}
#endif /* APPLET_USAGE */
