/*

    getpeername.c, part of
    faucet and hose: network pipe utilities
    Copyright (C) 1995-98 Robert Forsman

    busyboxing an IPv6:
    Copyright (C) 2022 Vladimir Oleynik <dzo@simtreas.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    */
#ifdef APPLET_USAGE
#define getpeername_trivial_usage \
	"[ -verbose ] [ fd# ]"
#define getpeername_full_usage \
	"Get information about this or that end of the socket's connection.\n" \
	"fd default is 0"
#else

#include "busybox.h"
#include "bb_inet.h"
#include <string.h>


int getpeername_main(void)
{
	int i;
	int fd = 0;                 /* assume socket is attached to stdin.
				  (if it's attached to stdout, our
				  output will go over the socket) */
	int verbose=0;

	union {
		struct sockaddr     base;
		struct sockaddr_in  in;
#if ENABLE_FEATURE_IPV6
		struct sockaddr_in6 in6;
#endif
	} saddr;

	int port;
	socklen_t saddrlen = sizeof(saddr);
	const char *name;
	char buf[NI_MAXHOST];

	for (i=1; Gargv[i] != NULL; i++) {
		if (0==strncmp(Gargv[i], "-verbose", strlen(Gargv[i]))) {
			verbose++;
		} else
			break;
	}
	if (Gargv[i]) {
		fd = bb_xparse_i_number(BB_PN_SHOW_USAGE, Gargv[i]);
		i++;
	}
	if (Gargv[i])
		bb_show_usage();

	if (verbose > 1)
		bb_error_msg("netpipes version 4.2" USE_FEATURE_IPV6("+IPv6"));

	if (bb_applet_name[3] == 'p') {
		i = getpeername(fd, &saddr.base, &saddrlen);
		name = "peer";
	} else {
		i = getsockname(fd, &saddr.base, &saddrlen);
		name = "sock";
	}
	if (i != 0)
		bb_perror_msg_and_die("get%sname failed on descriptor %d", name, fd);

	if (saddr.base.sa_family == AF_INET) {
		struct sockaddr_in *sinp = &saddr.in;
		port = ntohs(sinp->sin_port);
	}
#if ENABLE_FEATURE_IPV6
	else if (saddr.base.sa_family == AF_INET6) {
		struct sockaddr_in6 *sinp = &saddr.in6;
		port = ntohs(sinp->sin_port);
	}
#endif
	else {
	    bb_perror_msg_and_die("unknown address family (%d) returned by get%sname", saddr.base.sa_family, name);
	}
	if (getnameinfo(&saddr.base, saddrlen, buf, NI_MAXHOST, NULL, 0, NI_NUMERICHOST) != 0)
		bb_perror_msg_and_die("get%sname failed on descriptor %d", name, fd);
	if (verbose)
		puts("Internet\nPort");
	printf("%d\n", port);
	if (verbose) puts("Host");
	printf("%s\n", buf);
	if (verbose) {
		if (getnameinfo(&saddr.base, saddrlen, buf, NI_MAXHOST, NULL, 0, NI_NAMEREQD) == 0)
			puts(buf);
		else
			puts(" (no name for host)");
	}
	return 0;
}
#endif /* APPLET_USAGE */
