/*
 * portmap, Implements the program,version to port number mapping for rpc.
 *
 * Busybox port by Vladimir Oleynik 2005 (c) <dzo@simtreas.ru>
 */

/*-
 * Copyright (c) 1990 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
2.3 88/08/11 4.0 RPCSRC
static char sccsid[] = "@(#)portmap.c 1.32 87/08/06 Copyr 1984 Sun Micro";
*/


/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 *
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 *
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 *
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 *
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 *
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */

  /*
  * pmap_check - additional portmap security.
  *
  * Always reject non-local requests to update the portmapper tables.
  *
  * Refuse to forward mount requests to the nfs mount daemon. Otherwise, the
  * requests would appear to come from the local system, and nfs export
  * restrictions could be bypassed.
  *
  * Refuse to forward requests to the nfsd process.
  *
  * Refuse to forward requests to NIS (YP) daemons; The only exception is the
  * YPPROC_DOMAIN_NONACK broadcast rpc call that is used to establish initial
  * contact with the NIS server.
  *
  * Always allocate an unprivileged port when forwarding a request.
  *
  * If compiled with -DCHECK_PORT, require that requests to register or
  * unregister a privileged port come from a privileged port. This makes it
  * more difficult to replace a critical service by a trojan. Also, require
  * that requests to set/unset the NFSD port come form a privileged port.
  *
  * If compiled with -DHOSTS_ACCESS, reject requests from hosts that are not
  * authorized by the /etc/hosts.{allow,deny} files. The local system is
  * always treated as an authorized host. The access control tables are never
  * consulted for requests from the local system, and are always consulted
  * for requests from other hosts. Access control is based on IP addresses
  * only; attempts to map an address to a host name might cause the
  * portmapper to hang.
  *
  * Author: Wietse Venema (wietse@wzv.win.tue.nl), dept. of Mathematics and
  * Computing Science, Eindhoven University of Technology, The Netherlands.
  */

 /*
  * Check if an address belongs to the local system. Adapted from:
  *
  * 91/03/11 Copyright 1984,1990 Sun Microsystems, Inc.
  * 2.1 88/07/29 4.0 RPCSRC.
  */
#ifdef APPLET_USAGE
#define portmap_trivial_usage \
	"[-dv]"
#define portmap_full_usage \
	"Implements the program, version to port number mapping for RPC.\n" \
	"Options:\n" \
	"\t-d\tdebugging mode\n" \
	"\t-v\tverbose logging"
#else

#undef HOSTS_ACCESS
#define CHECK_PORT
#define FACILITY LOG_DAEMON
#undef DEBUGING


#include "busybox.h"
#if defined(__MUSL__)
#error "MUSL libc have not RPC support. We can not compile portmap applet. :("
#endif
/* This is just a warning of a common mistake.  Possibly this should be a
 * uclibc faq entry rather than in busybox... */
#if defined(__UCLIBC__) && ! defined(__UCLIBC_HAS_RPC__)
#error "You need to build uClibc with UCLIBC_HAS_RPC for portmap build."
#endif
#include <rpc/rpc.h>
#include <rpc/pmap_prot.h>
#include <syslog.h>
#include "bb_inet.h"
#include <netinet/in.h>
#ifdef SYSV40
#include <rpc/rpcent.h>
#endif
#include <unistd.h>
#include <stdio.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <string.h>

#ifdef HOSTS_ACCESS
#include <tcpd.h>
#endif
#if defined(__GLIBC__)
#define _BSD_SOURCE 1   /* for daemon(3) */
#include <rpc/xdr.h>
#endif /* __GLIBC__ */
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "bb_sig.h"


#ifdef LOOPBACK_SETUNSET
#define CHECK_SETUNSET  check_setunset
#else
#define CHECK_SETUNSET(xprt,ludp,ltcp,proc,prog,port) \
	check_setunset(svc_getcaller(xprt),proc,prog,port)
#endif

/* Explicit #defines in case the include files are not available. */

#define NFSPROG         ((unsigned long) 100003)
#define MOUNTPROG       ((unsigned long) 100005)
#define YPXPROG         ((unsigned long) 100069)
#define YPPROG          ((unsigned long) 100004)
#define YPPROC_DOMAIN_NONACK ((unsigned long) 2)
#define MOUNTPROC_MNT   ((unsigned long) 1)
#define NFS_PORT        2049

#define allow_severity LOG_INFO
#define deny_severity LOG_WARNING

/* A handful of macros for "readability". */

#define good_client(a) hosts_ctl("portmap", "", inet_ntoa(a->sin_addr), "")

#define reserved_port(p) (IPPORT_RESERVED/2 < (p) && (p) < IPPORT_RESERVED)

#define unreserved_port(p) (IPPORT_RESERVED <= (p) && (p) != NFS_PORT)

#define legal_port(a,p) \
  (reserved_port(ntohs((a)->sin_port)) || unreserved_port(p))

#define log_bad_port(addr, proc, prog) \
  logit(deny_severity, addr, proc, prog, ": request from unprivileged port")

#define log_bad_host(addr, proc, prog) \
  logit(deny_severity, addr, proc, prog, ": request from unauthorized host")

#define log_bad_owner(addr, proc, prog) \
  logit(deny_severity, addr, proc, prog, ": request from non-local host")

#define log_no_forward(addr, proc, prog) \
  logit(deny_severity, addr, proc, prog, ": request not forwarded")

#define log_client(addr, proc, prog) \
  logit(allow_severity, addr, proc, prog, "")

CALC_BSS struct BB_COMMON_BSS {
	int     verboselog;
 /*
  * With virtual hosting, each hardware network interface can have multiple
  * network addresses. On such machines the number of machine addresses can
  * be surprisingly large.
  */
	int num_local;
	int num_addrs;
	struct in_addr *addrs;
	struct pmaplist *g_pmaplist;
#ifdef DEBUGING
	int debugging;
#define debugging bb_common_bss.debugging
#else
#define debugging 0
#endif
#ifdef LOOPBACK_SETUNSET
	SVCXPRT *ludpxprt, *ltcpxprt;
#endif
} bb_common_bss;
#define verboselog bb_common_bss.verboselog
#define num_local bb_common_bss.num_local
#define num_addrs bb_common_bss.num_addrs
#define addrs bb_common_bss.addrs
#define g_pmaplist bb_common_bss.g_pmaplist
#define ludpxprt bb_common_bss.ludpxprt
#define ltcpxprt bb_common_bss.ltcpxprt


/* toggle_verboselog - toggle verbose logging flag */

static void toggle_verboselog(int sig)
{
    bb_signal((1 << sig), toggle_verboselog);
    verboselog = !verboselog;
}

/* grow_addrs - extend list of local interface addresses */

static int grow_addrs(void)
{
    struct in_addr *new_addrs;
    int     new_num;

    /*
     * Keep the previous result if we run out of memory. The system would
     * really get hosed if we simply give up.
     */
    new_num = (addrs == 0) ? 1 : num_addrs + num_addrs;
    new_addrs = malloc(sizeof(*addrs) * new_num);
    if (new_addrs == 0) {
	bb_perror_msg("out of memory");
	return 0;
    } else {
	if (addrs != 0) {
	    memcpy(new_addrs, addrs, sizeof(*addrs) * num_addrs);
	    free(addrs);
	}
	num_addrs = new_num;
	addrs = new_addrs;
	return 1;
    }
}

/* find_local - find all IP addresses for this host */

static int find_local(void)
{
    struct ifconf ifc;
    struct ifreq ifreq;
    struct ifreq *ifr;
    struct ifreq *the_end;
    int     sock;
    char    buf[BUFSIZ];

    /*
     * Get list of network interfaces. We use a huge buffer to allow for the
     * presence of non-IP interfaces.
     */

    if ((sock = bb_xsocket(BB_XSOCKET_FLG_SOCK_DGRAM)) < 0)
	return 0;
    ifc.ifc_len = sizeof(buf);
    ifc.ifc_buf = buf;
    if (xioctl(sock, SIOCGIFCONF, &ifc, "SIOCGIFCONF") < 0) {
	(void) close(sock);
	return 0;
    }
    /* Get IP address of each active IP network interface. */

    the_end = (struct ifreq *) (ifc.ifc_buf + ifc.ifc_len);
    num_local = 0;
    for (ifr = ifc.ifc_req; ifr < the_end; ifr++) {
	if (ifr->ifr_addr.sa_family == AF_INET) {       /* IP net interface */
	    ifreq = *ifr;
	    if (xioctl(sock, SIOCGIFFLAGS, &ifreq, "SIOCGIFFLAGS") < 0) {
		;
	    } else if (ifreq.ifr_flags & IFF_UP) {      /* active interface */
		if (xioctl(sock, SIOCGIFADDR, &ifreq, "SIOCGIFADDR") < 0) {
		    ;
		} else {
		    if (num_local >= num_addrs)
			if (grow_addrs() == 0)
			    break;
		    addrs[num_local++] = ((struct sockaddr_in *)
					  & ifreq.ifr_addr)->sin_addr;
		}
	    }
	}
	/* Support for variable-length addresses. */
#ifdef HAS_SA_LEN
	ifr = (struct ifreq *) ((caddr_t) ifr
		      + ifr->ifr_addr.sa_len - sizeof(struct sockaddr));
#endif
    }
    (void) close(sock);
    return num_local;
}


/* from_local - determine whether request comes from the local system */

static int from_local(struct sockaddr_in *addr)
{
    int     i;

    if (addrs == 0 && find_local() == 0)
	bb_error_msg("cannot find any active local network interfaces");

    for (i = 0; i < num_local; i++) {
	if (memcmp((char *) &(addr->sin_addr), (char *) &(addrs[i]),
		   sizeof(struct in_addr)) == 0)
	    return 1;
    }
    return 0;
}

/* logit - report events of interest via the syslog daemon */

static void logit(int severity, struct sockaddr_in *addr,
    unsigned long procnum, unsigned long prognum, char *text)
{
    const char   *procname;
    char    procbuf[4 * sizeof(unsigned long)];
    char   *progname;
    char    progbuf[4 * sizeof(unsigned long)];
    struct rpcent *rpc;
    struct proc_map {
	unsigned long  code;
	const char   *proc;
    };
    struct proc_map const *procp;
    static const struct proc_map procmap[] = {
	{ PMAPPROC_CALLIT, "callit" },
	{ PMAPPROC_DUMP, "dump" },
	{ PMAPPROC_GETPORT, "getport" },
	{ PMAPPROC_NULL, "null" },
	{ PMAPPROC_SET, "set" },
	{ PMAPPROC_UNSET, "unset" },
	{ 0, 0 }
    };

    /*
     * Fork off a process or the portmap daemon might hang while
     * getrpcbynumber() or syslog() does its thing.
     */

    if (fork() == 0) {

	/* Try to map program number to name. */

	if (prognum == 0) {
	    progname = "";
	} else if ((rpc = getrpcbynumber((int) prognum))) {
	    progname = rpc->r_name;
	} else {
	    sprintf(progname = progbuf, "%lu", prognum);
	}

	/* Try to map procedure number to name. */

	for (procp = procmap; procp->proc && procp->code != procnum; procp++)
	     /* void */ ;
	if ((procname = procp->proc) == 0) {
	    sprintf(procbuf, "%lu", (unsigned long) procnum);
	    procname = procbuf;
	}

	/* Write syslog record. */

	syslog(severity, "connect from %s to %s(%s)%s",
	       inet_ntoa(addr->sin_addr), procname, progname, text);
	exit(0);
    }
}

/* check_default - additional checks for NULL, DUMP, GETPORT and unknown */

static int check_default(struct sockaddr_in *addr,
				unsigned long proc, unsigned long prog)
{
#ifdef HOSTS_ACCESS
    if (!(from_local(addr) || good_client(addr))) {
	log_bad_host(addr, proc, prog);
	return 0;
    }
#endif
    if (verboselog)
	log_client(addr, proc, prog);
    return 1;
}

/* check_privileged_port - additional checks for privileged-port updates */

static int check_privileged_port(struct sockaddr_in *addr,
		unsigned long proc, unsigned long prog, unsigned long port)
{
#ifdef CHECK_PORT
    if (!legal_port(addr, port)) {
	log_bad_port(addr, proc, prog);
	return 0;
    }
#endif
    return 1;
}

/* check_setunset - additional checks for update requests */

#ifdef LOOPBACK_SETUNSET

static int check_setunset(SVCXPRT *xprt, SVCXPRT *ludp_xprt,
				SVCXPRT *ltcp_xprt,
		unsigned long proc, unsigned long prog, unsigned long port)
{
    struct sockaddr_in *addr = svc_getcaller(xprt);

    if (xprt != ludp_xprt && xprt != ltcp_xprt) {
#ifdef HOSTS_ACCESS
	(void) good_client(addr);               /* because of side effects */
#endif
	log_bad_owner(addr, proc, prog);
	return 0;
    }
    if (port && !check_privileged_port(addr, proc, prog, port))
	return 0;
    if (verboselog)
	log_client(addr, proc, prog);
    return 1;
}

#else

static int
check_setunset(struct sockaddr_in *addr,
		unsigned long proc, unsigned long prog, unsigned long port)
{
    if (!from_local(addr)) {
#ifdef HOSTS_ACCESS
	(void) good_client(addr);               /* because of side effects */
#endif
	log_bad_owner(addr, proc, prog);
	return 0;
    }
    if (port && !check_privileged_port(addr, proc, prog, port))
	return 0;
    if (verboselog)
	log_client(addr, proc, prog);
    return 1;
}

#endif

/* check_callit - additional checks for forwarded requests */

static int
check_callit(struct sockaddr_in *addr,
		unsigned long proc, unsigned long prog, unsigned long aproc)
{
#ifdef HOSTS_ACCESS
    if (!(from_local(addr) || good_client(addr))) {
	log_bad_host(addr, proc, prog);
	return 0;
    }
#endif
    if (prog == PMAPPROG || prog == NFSPROG || prog == YPXPROG ||
	(prog == MOUNTPROG && aproc == MOUNTPROC_MNT) ||
	(prog == YPPROG && aproc != YPPROC_DOMAIN_NONACK)) {
	log_no_forward(addr, proc, prog);
	return 0;
    }
    if (verboselog)
	log_client(addr, proc, prog);
    return 1;
}


#ifndef LOG_PERROR
#define LOG_PERROR 0
#endif

#ifndef LOG_DAEMON
#define LOG_DAEMON 0
#endif

/* Older SYSV. */
#if !defined(SIGCHLD) && defined(SIGCLD)
#define SIGCHLD      SIGCLD
#endif

#ifndef svc_getcaller           /* SYSV4 */
#  define svc_getcaller svc_getrpccaller
#endif

#ifdef USE_SETPGRP00
#define setsid() setpgrp(0,0)
#endif


/*
 * Stuff for the rmtcall service
 */

static struct pmaplist *
find_service(unsigned long prog, unsigned long vers, unsigned long prot)
{
	struct pmaplist *hit = NULL;
	struct pmaplist *pml;

	for (pml = g_pmaplist; pml != NULL; pml = pml->pml_next) {
		if ((pml->pml_map.pm_prog != prog) ||
			(pml->pml_map.pm_prot != prot))
			continue;
		hit = pml;
		if (pml->pml_map.pm_vers == vers)
		    break;
	}
	return hit;
}

struct encap_parms {
	unsigned int arglen;
	char *args;
};

struct rmtcallargs {
	unsigned long  rmt_prog;
	unsigned long  rmt_vers;
	unsigned long  rmt_port;
	unsigned long  rmt_proc;
	struct encap_parms rmt_args;
};

#define ARGSIZE 9000

static bool_t
xdr_encap_parms(XDR *xdrs, struct encap_parms *epp)
{
	return xdr_bytes(xdrs, &(epp->args), &(epp->arglen), ARGSIZE);
}


static bool_t xdr_rmtcall_args(XDR *xdrs, struct rmtcallargs *cap)
{

	/* does not get a port number */
	if (xdr_u_long(xdrs, &(cap->rmt_prog)) &&
	    xdr_u_long(xdrs, &(cap->rmt_vers)) &&
	    xdr_u_long(xdrs, &(cap->rmt_proc))) {
		return xdr_encap_parms(xdrs, &(cap->rmt_args));
	}
	return 0;
}

static bool_t xdr_rmtcall_result(XDR *xdrs, struct rmtcallargs *cap)
{
	if (xdr_u_long(xdrs, &(cap->rmt_port)))
		return xdr_encap_parms(xdrs, &(cap->rmt_args));
	return 0;
}

/* only worries about the struct encap_parms part of struct rmtcallargs.
 * The arglen must already be set!! */
static bool_t xdr_opaque_parms(XDR *xdrs, struct rmtcallargs *cap)
{
	return xdr_opaque(xdrs, cap->rmt_args.args, cap->rmt_args.arglen);
}

/* This routine finds and sets the length of incoming opaque paraters
 * and then calls xdr_opaque_parms. */
static bool_t xdr_len_opaque_parms(XDR *xdrs, struct rmtcallargs *cap)
{
	unsigned int beginpos, lowpos, highpos, currpos, pos;

	beginpos = lowpos = pos = xdr_getpos(xdrs);
	highpos = lowpos + ARGSIZE;
	while ((int)(highpos - lowpos) >= 0) {
		currpos = (lowpos + highpos) / 2;
		if (xdr_setpos(xdrs, currpos)) {
			pos = currpos;
			lowpos = currpos + 1;
		} else {
			highpos = currpos - 1;
		}
	}
	xdr_setpos(xdrs, beginpos);
	cap->rmt_args.arglen = pos - beginpos;
	return xdr_opaque_parms(xdrs, cap);
}

/* Call a remote procedure service
 * This procedure is very quiet when things go wrong.
 * The proc is written to support broadcast rpc.  In the broadcast case,
 * a machine should shut-up instead of complain, less the requestor be
 * overrun with complaints at the expense of not hearing a valid reply ...
 *
 * This now forks so that the program & process that it calls can call
 * back to the portmapper. */
static void callit(struct svc_req *rqstp, SVCXPRT *xprt)
{
	struct rmtcallargs a;
	struct pmaplist *pml;
	u_short port;
	struct sockaddr_in me;
	int pid, so = -1;
	CLIENT *client;
	struct authunix_parms *au = (struct authunix_parms *)rqstp->rq_clntcred;
	struct timeval timeout;
	char buf[ARGSIZE];

	timeout.tv_sec = 5;
	timeout.tv_usec = 0;
	a.rmt_args.args = buf;
	if (!svc_getargs(xprt, (xdrproc_t) xdr_rmtcall_args, (caddr_t) &a))
		return;
	/* host and service access control */
	if (!check_callit(svc_getcaller(xprt),
	    rqstp->rq_proc, a.rmt_prog, a.rmt_proc))
		return;
	if ((pml = find_service(a.rmt_prog, a.rmt_vers,
	    (unsigned long)IPPROTO_UDP)) == NULL)
		return;
	/*
	 * fork a child to do the work.  Parent immediately returns.
	 * Child exits upon completion.
	 */
	if ((pid = fork()) != 0) {
		if (pid < 0)
			bb_perror_msg("CALLIT (prog %lu): fork", a.rmt_prog);
		return;
	}
	port = pml->pml_map.pm_port;
	get_myaddress(&me);
	me.sin_port = htons(port);
	client = clntudp_create(&me, a.rmt_prog, a.rmt_vers, timeout, &so);
	if (client != (CLIENT *)NULL) {
		if (rqstp->rq_cred.oa_flavor == AUTH_UNIX) {
			client->cl_auth = authunix_create(au->aup_machname,
			   au->aup_uid, au->aup_gid, au->aup_len, au->aup_gids);
		}
		a.rmt_port = (unsigned long)port;
		if (clnt_call(client, a.rmt_proc, (xdrproc_t) xdr_opaque_parms, (char*) &a,
		    (xdrproc_t) xdr_len_opaque_parms, (char*) &a, timeout) == RPC_SUCCESS) {
			svc_sendreply(xprt, (xdrproc_t) xdr_rmtcall_result, (caddr_t)&a);
		}
		AUTH_DESTROY(client->cl_auth);
		clnt_destroy(client);
	}
	(void)close(so);
	exit(0);
}

/* 1 OK, 0 not */
static void reg_service(struct svc_req *rqstp, SVCXPRT *xprt)
{
	struct pmap reg;
	struct pmaplist *pml, *prevpml, *fnd;
	int ans, port;
	caddr_t t;

	if (debugging)
		(void) fprintf(stderr, "server: about do a switch\n");
	switch (rqstp->rq_proc) {

	case PMAPPROC_NULL:
		/*
		 * Null proc call
		 */
		/* remote host authorization check */
		check_default(svc_getcaller(xprt), rqstp->rq_proc, (unsigned long) 0);
		if (!svc_sendreply(xprt, (xdrproc_t) xdr_void, (caddr_t)0) && debugging) {
			abort();
		}
		break;

	case PMAPPROC_SET:
		/*
		 * Set a program,version to port mapping
		 */
		if (!svc_getargs(xprt, (xdrproc_t) xdr_pmap, (caddr_t) &reg))
			svcerr_decode(xprt);
		else {
			/* reject non-local requests, protect priv. ports */
			if (!CHECK_SETUNSET(xprt, ludpxprt, ltcpxprt,
			    rqstp->rq_proc, reg.pm_prog, reg.pm_port)) {
				ans = 0;
				goto done;
			}
			/*
			 * check to see if already used
			 * find_service returns a hit even if
			 * the versions don't match, so check for it
			 */
			fnd = find_service(reg.pm_prog, reg.pm_vers, reg.pm_prot);
			if (fnd && fnd->pml_map.pm_vers == reg.pm_vers) {
				if (fnd->pml_map.pm_port == reg.pm_port) {
					ans = 1;
					goto done;
				}
				else {
					ans = 0;
					goto done;
				}
			} else {
				/*
				 * add to END of list
				 */
				pml = xmalloc(sizeof(struct pmaplist));
				pml->pml_map = reg;
				pml->pml_next = 0;
				if (g_pmaplist == 0) {
					g_pmaplist = pml;
				} else {
					for (fnd= g_pmaplist; fnd->pml_next != 0;
					    fnd = fnd->pml_next);
					fnd->pml_next = pml;
				}
				ans = 1;
			}
		done:
			if ((!svc_sendreply(xprt, (xdrproc_t) xdr_int, (caddr_t)&ans)) &&
			    debugging) {
				(void) fprintf(stderr, "svc_sendreply\n");
				abort();
			}
		}
		break;

	case PMAPPROC_UNSET:
		/*
		 * Remove a program,version to port mapping.
		 */
		if (!svc_getargs(xprt, (xdrproc_t) xdr_pmap, (caddr_t) &reg))
			svcerr_decode(xprt);
		else {
			ans = 0;
			/* reject non-local requests */
			if (!CHECK_SETUNSET(xprt, ludpxprt, ltcpxprt,
			    rqstp->rq_proc, reg.pm_prog, (unsigned long) 0))
				goto done;
			for (prevpml = NULL, pml = g_pmaplist; pml != NULL; ) {
				if ((pml->pml_map.pm_prog != reg.pm_prog) ||
					(pml->pml_map.pm_vers != reg.pm_vers)) {
					/* both pml & prevpml move forwards */
					prevpml = pml;
					pml = pml->pml_next;
					continue;
				}
				/* found it; pml moves forward, prevpml stays */
				/* privileged port check */
				if (!check_privileged_port(svc_getcaller(xprt),
				    rqstp->rq_proc,
				    reg.pm_prog,
				    pml->pml_map.pm_port)) {
					ans = 0;
					break;
				}
				ans = 1;
				t = (caddr_t)pml;
				pml = pml->pml_next;
				if (prevpml == NULL)
					g_pmaplist = pml;
				else
					prevpml->pml_next = pml;
				free(t);
			}
			if ((!svc_sendreply(xprt, (xdrproc_t) xdr_int, (caddr_t)&ans)) &&
			    debugging) {
				(void) fprintf(stderr, "svc_sendreply\n");
				abort();
			}
		}
		break;

	case PMAPPROC_GETPORT:
		/*
		 * Lookup the mapping for a program,version and return its port
		 */
		if (!svc_getargs(xprt, (xdrproc_t) xdr_pmap, (caddr_t) &reg))
			svcerr_decode(xprt);
		else {
			/* remote host authorization check */
			if (!check_default(svc_getcaller(xprt),
			    rqstp->rq_proc,
			    reg.pm_prog)) {
				ans = 0;
				goto done;
			}
			fnd = find_service(reg.pm_prog, reg.pm_vers, reg.pm_prot);
			if (fnd)
				port = fnd->pml_map.pm_port;
			else
				port = 0;
			if ((!svc_sendreply(xprt, (xdrproc_t) xdr_int, (caddr_t)&port)) &&
			    debugging) {
				(void) fprintf(stderr, "svc_sendreply\n");
				abort();
			}
		}
		break;

	case PMAPPROC_DUMP:
		/*
		 * Return the current set of mapped program,version
		 */
		if (!svc_getargs(xprt, (xdrproc_t) xdr_void, (caddr_t) NULL))
			svcerr_decode(xprt);
		else {
			/* remote host authorization check */
			struct pmaplist *p;
			if (!check_default(svc_getcaller(xprt),
			    rqstp->rq_proc, (unsigned long) 0)) {
				p = 0;  /* send empty list */
			} else {
				p = g_pmaplist;
			}
			if ((!svc_sendreply(xprt, (xdrproc_t) xdr_pmaplist,
			    (caddr_t)&p)) && debugging) {
				(void) fprintf(stderr, "svc_sendreply\n");
				abort();
			}
		}
		break;

	case PMAPPROC_CALLIT:
		/*
		 * Calls a procedure on the local machine.  If the requested
		 * procedure is not registered this procedure does not return
		 * error information!!
		 * This procedure is only supported on rpc/udp and calls via
		 * rpc/udp.  It passes null authentication parameters.
		 */
		callit(rqstp, xprt);
		break;

	default:
		/* remote host authorization check */
		check_default(svc_getcaller(xprt), rqstp->rq_proc, (unsigned long) 0);
		svcerr_noproc(xprt);
		break;
	}
}

 /*
  * How desperate can one be. It is possible to prevent an attacker from
  * manipulating your portmapper tables from outside with requests that
  * contain spoofed source address information. The countermeasure is to
  * force all rpc servers to register and unregister with the portmapper via
  * the loopback network interface, instead of via the primary network
  * interface that every host can talk to. For this countermeasure to work it
  * is necessary to #define LOOPBACK_SETUNSET, to disable source routing in
  * the kernel, and to modify libc so that get_myaddress() chooses the
  * loopback interface address.
  */

#ifdef LOOPBACK_SETUNSET
#ifndef INADDR_LOOPBACK
#define INADDR_LOOPBACK ntohl(inet_addr("127.0.0.1"))
#endif
#endif

int portmap_main(void)
{
	SVCXPRT *xprt;
	int sock;
	struct sockaddr_in addr;
	struct pmaplist *pml;
	unsigned long c;

	c = bb_getopt_ulflags("vdf");
	verboselog = c & 1;
#ifdef DEBUGING
	debugging = c & 2;
#endif
	if(!debugging) {
#ifdef LOG_MAIL
		bb_openlog(debugging ? LOG_PID | LOG_PERROR : LOG_PID, FACILITY);
#else
		bb_openlog(debugging ? LOG_PID | LOG_PERROR : LOG_PID);
#endif
		if ((c & 4) == 0)
			xdaemon(XDAEMON_PIDFILE);
	}


	sock = bb_xsocket(BB_XSOCKET_FLG_SOCK_DGRAM | BB_XSOCKET_FLG_ERR_DIE);
#ifdef LOOPBACK_SETUNSET
	setsockopt_1i(sock, SOL_SOCKET, SO_REUSEADDR, 1);
#endif

	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(PMAPPORT);
	bb_xbind(sock, (struct sockaddr *)&addr, sizeof(struct sockaddr_in));

	if ((xprt = svcudp_create(sock)) == (SVCXPRT *)NULL)
		bb_perror_msg_and_die("couldn't do udp_create");
	/* make an entry for ourself */
	pml = xmalloc(sizeof(struct pmaplist));
	pml->pml_next = 0;
	pml->pml_map.pm_prog = PMAPPROG;
	pml->pml_map.pm_vers = PMAPVERS;
	pml->pml_map.pm_prot = IPPROTO_UDP;
	pml->pml_map.pm_port = PMAPPORT;
	g_pmaplist = pml;

	sock = bb_xsocket(BB_XSOCKET_FLG_ERR_DIE);
#ifdef LOOPBACK_SETUNSET
	setsockopt_1i(sock, SOL_SOCKET, SO_REUSEADDR, 1);
#endif
	bb_xbind(sock, (struct sockaddr *)&addr, sizeof(struct sockaddr_in));
	if ((xprt = svctcp_create(sock, RPCSMALLMSGSIZE, RPCSMALLMSGSIZE)) == (SVCXPRT *)NULL)
		bb_perror_msg_and_die("couldn't do tcp_create");
	/* make an entry for ourself */
	pml = xmalloc(sizeof(struct pmaplist));
	pml->pml_map.pm_prog = PMAPPROG;
	pml->pml_map.pm_vers = PMAPVERS;
	pml->pml_map.pm_prot = IPPROTO_TCP;
	pml->pml_map.pm_port = PMAPPORT;
	pml->pml_next = g_pmaplist;
	g_pmaplist = pml;

#ifdef LOOPBACK_SETUNSET
	sock = bb_xsocket(BB_XSOCKET_FLG_SOCK_DGRAM | BB_XSOCKET_FLG_ERR_DIE);
	setsockopt_1i(sock, SOL_SOCKET, SO_REUSEADDR, 1);

	addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	bb_xbind(sock, (struct sockaddr *)&addr, sizeof(struct sockaddr_in));

	if ((ludpxprt = svcudp_create(sock)) == (SVCXPRT *)NULL)
		bb_perror_msg_and_die("couldn't do udp_create");
	sock = bb_xsocket(BB_XSOCKET_FLG_ERR_DIE);
	setsockopt_1i(sock, SOL_SOCKET, SO_REUSEADDR, 1);
	bb_xbind(sock, (struct sockaddr *)&addr, sizeof(struct sockaddr_in));
	if ((ltcpxprt = svctcp_create(sock, RPCSMALLMSGSIZE, RPCSMALLMSGSIZE)) == (SVCXPRT *)NULL)
		bb_perror_msg_and_die("couldn't do tcp_create");
#endif

	(void)svc_register(xprt, PMAPPROG, PMAPVERS, reg_service, 0);

	/*
	* Give up root privileges so that we can never allocate a privileged
	* port when forwarding an rpc request.
	*/
	if (setgid(1) == -1)
	    bb_perror_msg_and_die("setgid(1) failed");
	if (setuid(1) == -1)
	    bb_perror_msg_and_die("setuid(1) failed");
	initgroups("daemon", 1);

	bb_signal((1 << SIGINT), toggle_verboselog);
	bb_signal((1 << SIGCHLD), SIG_IGN);
	svc_run();
	bb_perror_msg_and_die("run_svc returned unexpectedly");
	return 1;       /* fake */
}
#endif /* APPLET_USAGE */
