/*
Poptop -- The PPTP Server
-------------------------

Current Maintainer:
James Cameron <james.cameron at hp dot com>

Previous Maintainer:
Richard de Vroede <r.devroede at linvision dot com>

Original Authors:
Matthew Ramsay <matthewr at lineo dot com>
Kevin Thayer <tmk at netmagic.net>
David Luyer <luyer at ucs.uwa.edu.au>
Peter Galbavy <Peter.Galbavy at knowledge dot com>

Other Contributors:
Ron O'Hara <rono at sentuny dot com.au>
Allan Clark <allanc at sco dot com>
Harald Vogt <vogt at serc.nl>
Bruno Lopes F. Cabral <bruno at openline dot com.br> [!3runo]
Peter Stamfest <poptop at stamfest dot at>
Sergio M. Ammirata      Norbert van Bolhuis
Visarion Mandzgaladze   Hernan Otero
Jan Dubiec              Pawel Guraj
Chris Wilson            Anton Gorlov
Charlie Brady

The Linux PPTP Server takes advantage of some Linux PPTP client code
written by C. Scott Ananian <cananian at alumni.princeton.edu>

Other Credits:
Moreton Bay <http://www.moretonbay.com>
Lineo       <http://www.lineo.com>
Snapgear    <http://www.snapgear.com>
Linvision   <http://www.linvision.com>

 Copyright (C) 2003-2006 Vladimir Oleynik <dzo@simtreas.ru>
	seq is u_int32 now
	send_pptp_packet is ssize_t now
	add exit(1) to bail()   (forgotten?)
	busybox port
	size optimization
*/
#ifdef APPLET_USAGE
#define pptpd_trivial_usage \
	"[options]"
#define pptpd_full_usage \
	"PPTP server daemon.\n" \
	"Call pptpd -h for more options"

#define pptpctrl_trivial_usage \
	"7 args options"
#define pptpctrl_full_usage \
	"PPTP server control.\n" \
	"See man pptpctrl"
#else

#define VERSION "1.3.4"

/* Where is my pppd? */
#define PPP_BINARY "/usr/sbin/pppd"

/* Syslog Facility to use?  See openlog(3). */
#define PPTP_FACILITY LOG_DAEMON

/* Have libwrap? */
/* #undef HAVE_LIBWRAP */

/* Define to 1 if you have the `fork' function. */
#define HAVE_FORK 1

#include "busybox.h"
#include "bb_inet.h"
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/un.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <string.h>
#include <assert.h>
#include <syslog.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <ctype.h>
#include <dirent.h>
#include "pty_.h"
#include <termios.h>
// placing net/if.h here fixes build on Solaris
#include <net/if.h>
#if HAVE_LIBWRAP
/* re-include */
#include <tcpd.h>
#endif

#include "bb_sig.h"
#include "config_read.h"

/* String sizes for the config file */

#define MAX_CONFIG_FILENAME_SIZE        256
#define MAX_CONFIG_STRING_SIZE          512

/* For IP parser */

#define LOCAL 0
#define REMOTE 1

/* Default configuration values, mostly configurable */

#define CONNECTIONS_DEFAULT             100
#define DEFAULT_LOCAL_IP_LIST           "192.168.0.1-100"
#define DEFAULT_REMOTE_IP_LIST          "192.168.1.1-100"

#define MAX_CALLS_PER_TCP_LINK          128

#define PPP_SPEED_DEFAULT               "115200"
#if EMBED
#define PPTPD_CONFIG_FILE_DEFAULT       "/etc/config/pptpd.conf"
#else
#define PPTPD_CONFIG_FILE_DEFAULT       "/etc/pptpd.conf"
#endif

#define STIMEOUT_DEFAULT                10 /* seconds */

/* Location of binaries */

#define PPTP_CTRL_BIN                   "/usr/sbin/pptpctrl"

/* Parameters permitted in the config file */

#define CONNECTIONS_KEYWORD             "connections"
#define SPEED_KEYWORD                   "speed"
#define PPPD_OPTION_KEYWORD             "option"
#define DEBUG_KEYWORD                   "debug"
#define LOCALIP_KEYWORD                 "localip"
#define REMOTEIP_KEYWORD                "remoteip"
#define LISTEN_KEYWORD                  "listen"
#define PIDFILE_KEYWORD                 "pidfile"
#define STIMEOUT_KEYWORD                "stimeout"
#define NOIPPARAM_KEYWORD               "noipparam"
#define PPP_BINARY_KEYWORD              "ppp"
#define LOGWTMP_KEYWORD                 "logwtmp"
#define DELEGATE_KEYWORD                "delegate"

/* assume packet is bad/spoofed if it's more than this many seqs ahead */
#define MISSING_WINDOW 300

/* Packet queue structure: linked list of packets received out-of-order */
typedef struct pqueue {
  struct pqueue *next;
  struct pqueue *prev;
  u_int32_t seq;
  struct timeval expires;
  unsigned char *packet;
  int packlen;
  int capacity;
} pqueue_t;

/* data about connection slots */
struct slot {
  pid_t pid;
  char *local;
  char *remote;
};

/* PPTP ctrl message port */
#define PPTP_PORT                       1723

/* PPTP gre prototype */
#define PPTP_PROTO                      47

/* PPTP version */
#define PPTP_VERSION                    0x0100
#define PPTP_FIRMWARE_VERSION           0x0001

/* Hostname and Vendor */
#define PPTP_HOSTNAME                   "local"
#define PPTP_VENDOR                     "linux"

#define MAX_HOSTNAME_SIZE               64
#define MAX_VENDOR_SIZE                 64

/* Magic Cookie */
#define PPTP_MAGIC_COOKIE               0x1a2b3c4d

/* Message types */
#define PPTP_CTRL_MESSAGE               1

/* Maximum size of any PPTP control packet we will get */
#define PPTP_MAX_CTRL_PCKT_SIZE         220

/* Control Connection Management */
#define START_CTRL_CONN_RQST            1
#define START_CTRL_CONN_RPLY            2
#define STOP_CTRL_CONN_RQST             3
#define STOP_CTRL_CONN_RPLY             4
#define ECHO_RQST                       5
#define ECHO_RPLY                       6

/* Call Management */
#define OUT_CALL_RQST                   7
#define OUT_CALL_RPLY                   8
#define IN_CALL_RQST                    9
#define IN_CALL_RPLY                    10
#define IN_CALL_CONN                    11
#define CALL_CLR_RQST                   12
#define CALL_DISCONN_NTFY               13

/* Error Reporting */
#define WAN_ERR_NTFY                    14

/* PPP Session Control */
#define SET_LINK_INFO                   15

/* how long before a link is idle? (seconds) */
#define IDLE_WAIT                       60

/* how long should we wait for an echo reply? (seconds) */
#define MAX_ECHO_WAIT                   60

#define RESERVED                        0x0000

/* Start Control Connection Reply */
#define ASYNCHRONOUS_FRAMING            0x00000001
#define SYNCHRONOUS_FRAMING             0x00000002
#define ANALOG_ACCESS                   0x00000001
#define DIGITAL_ACCESS                  0x00000002

/* Our properties - we don't actually have any physical serial i/f's and only want
 * one call per client!
 */
#define OUR_FRAMING                     0x00000000
#define OUR_BEARER                      0x00000000
#define MAX_CHANNELS                    0x0001

/* Out Call Reply Defines */
#define PCKT_RECV_WINDOW_SIZE           0x0001
#define PCKT_PROCESS_DELAY              0x0000
#define CHANNEL_ID                      0x00000000

/* ERROR CODES */
#define NO_ERROR                        0x00

/* CALL_CLEAR RESULT CODES */
#define LOST_CARRIER                    0x01
#define ADMIN_SHUTDOWN                  0x03
#define CALL_CLEAR_REQUEST              0x04

/* RESULT CODES */
#define CONNECTED                       0x01
#define DISCONNECTED                    0x01
#define GENERAL_ERROR                   0x02    /* also for ERROR CODES, CALL CLEAR */
#define NO_CARRIER                      0x03
#define BUSY                            0x04
#define NO_DIAL_TONE                    0x05
#define TIME_OUT                        0x06
#define DO_NOT_ACCEPT                   0x07

/* CTRL CLOSE CODES */
#define GENERAL_STOP_CTRL               0x01
#define STOP_PROTOCOL                   0x02
#define STOP_LOCAL_SHUTDOWN             0x03

/* PPTP CTRL structs */

struct pptp_header {
	u_int16_t length;               /* pptp message length incl header */
	u_int16_t pptp_type;            /* pptp message type */
	u_int32_t magic;                /* magic cookie */
	u_int16_t ctrl_type;            /* control message type */
	u_int16_t reserved0;            /* reserved */
};

struct pptp_start_ctrl_conn_rqst {
	struct pptp_header header;      /* pptp header */
	u_int16_t version;              /* pptp protocol version */
	u_int16_t reserved1;            /* reserved */
	u_int32_t framing_cap;          /* framing capabilities */
	u_int32_t bearer_cap;           /* bearer capabilities */
	u_int16_t max_channels;         /* maximum channels */
	u_int16_t firmware_rev;         /* firmware revision */
	u_int8_t hostname[MAX_HOSTNAME_SIZE];   /* hostname */
	u_int8_t vendor[MAX_VENDOR_SIZE];       /* vendor */
};

struct pptp_start_ctrl_conn_rply {
	struct pptp_header header;      /* pptp header */
	u_int16_t version;              /* pptp protocol version */
	u_int8_t result_code;           /* result code */
	u_int8_t error_code;            /* error code */
	u_int32_t framing_cap;          /* framing capabilities */
	u_int32_t bearer_cap;           /* bearer capabilities */
	u_int16_t max_channels;         /* maximum channels */
	u_int16_t firmware_rev;         /* firmware revision */
	u_int8_t hostname[MAX_HOSTNAME_SIZE];   /* hostname */
	u_int8_t vendor[MAX_VENDOR_SIZE];       /* vendor */
};

struct pptp_stop_ctrl_conn_rqst {
	struct pptp_header header;      /* header */
	u_int8_t reason;                /* reason for closing */
	u_int8_t reserved1;             /* reserved */
	u_int16_t reserved2;            /* reserved */
};

struct pptp_stop_ctrl_conn_rply {
	struct pptp_header header;      /* header */
	u_int8_t result_code;           /* result code */
	u_int8_t error_code;            /* error code */
	u_int16_t reserved1;            /* reserved */
};

struct pptp_echo_rqst {
	struct pptp_header header;      /* header */
	u_int32_t identifier;           /* value to match rply with rqst */
};

struct pptp_echo_rply {
	struct pptp_header header;      /* header */
	u_int32_t identifier;           /* identifier of echo rqst */
	u_int8_t result_code;           /* result code */
	u_int8_t error_code;            /* error code */
	u_int16_t reserved1;            /* reserved */
};

struct pptp_out_call_rqst {
	struct pptp_header header;      /* header */
	u_int16_t call_id;              /* unique identifier to PAC-PNS pair */
	u_int16_t call_serial;          /* session identifier */
	u_int32_t min_bps;              /* minimum line speed */
	u_int32_t max_bps;              /* maximum line speed */
	u_int32_t bearer_type;          /* bearer type */
	u_int32_t framing_type;         /* framing type */
	u_int16_t pckt_recv_size;       /* packet recv window size */
	u_int16_t pckt_delay;           /* packet processing delay */
	u_int16_t phone_len;            /* phone number length */
	u_int16_t reserved1;            /* reserved */
	u_int8_t phone_num[64];         /* phone number */
	u_int8_t subaddress[64];        /* additional dialing info */
};

struct pptp_out_call_rply {
	struct pptp_header header;      /* header */
	u_int16_t call_id;              /* unique identifier to PAC-PNS pair */
	u_int16_t call_id_peer;         /* set to call_id of the call rqst */
	u_int8_t result_code;           /* result code */
	u_int8_t error_code;            /* error code */
	u_int16_t cause_code;           /* additional failure information */
	u_int32_t speed;                /* actual connection speed */
	u_int16_t pckt_recv_size;       /* packet recv window size */
	u_int16_t pckt_delay;           /* packet processing delay */
	u_int32_t channel_id;           /* physical channel ID */
};

struct pptp_in_call_rqst {
	struct pptp_header header;      /* header */
	u_int16_t call_id;              /* unique identifier for tunnel */
	u_int16_t call_serial;          /* session identifier */
	u_int32_t bearer_type;          /* bearer capability */
	u_int32_t channel_id;           /* channel ID */
	u_int16_t dialed_len;           /* dialed length */
	u_int16_t dialing_len;          /* dialing length */
	u_int8_t dialed_num[64];        /* number that was dialed by the caller */
	u_int8_t dialing_num[64];       /* the number from which the call was placed */
	u_int8_t subaddress[64];        /* additional dialing information */
};

struct pptp_in_call_rply {
	struct pptp_header header;      /* header */
	u_int16_t call_id;              /* unique identifier for the tunnel */
	u_int16_t peers_call_id;        /* set to rcvd call ID */
	u_int8_t result_code;           /* result code */
	u_int8_t error_code;            /* error code */
	u_int16_t pckt_recv_size;       /* packet recv window size */
	u_int16_t pckt_delay;           /* packet transmit delay */
	u_int16_t reserved1;            /* reserved */
};

struct pptp_in_call_connect {
	struct pptp_header header;      /* header */
	u_int16_t peers_call_id;        /* set to rcvd call ID */
	u_int16_t reserved1;            /* reserved */
	u_int32_t speed;                /* connect speed */
	u_int16_t pckt_recv_size;       /* packet rcvd window size */
	u_int16_t pckt_delay;           /* packet transmit delay */
	u_int32_t framing_type;         /* framing type */
};

struct pptp_call_clr_rqst {
	struct pptp_header header;      /* header */
	u_int16_t call_id;              /* call ID assigned by the PNS */
	u_int16_t reserved1;            /* reserved */
};

struct pptp_call_disconn_ntfy {
	struct pptp_header header;      /* header */
	u_int16_t call_id;              /* call ID assigned by the PAC */
	u_int8_t result_code;           /* result code */
	u_int8_t error_code;            /* error code */
	u_int16_t cause_code;           /* additional disconnect info */
	u_int16_t reserved1;            /* reserved */
	u_int8_t call_stats[128];       /* vendor specific call stats */
};

struct pptp_wan_err_ntfy {
	struct pptp_header header;      /* header */
	u_int16_t peers_call_id;        /* call ID assigned by PNS */
	u_int16_t reserved1;            /* reserved */
	u_int32_t crc_errors;           /* # of PPP frames rcvd with CRC errors */
	u_int32_t framing_errors;       /* # of improperly framed PPP pckts */
	u_int32_t hardware_overruns;    /* # of receive buffer overruns */
	u_int32_t buff_overruns;        /* # of buffer overruns */
	u_int32_t timeout_errors;       /* # of timeouts */
	u_int32_t align_errors;         /* # of alignment errors */
};

struct pptp_set_link_info {
	struct pptp_header header;
	u_int16_t peers_call_id;        /* call ID assigned by PAC */
	u_int16_t reserved1;            /* reserved */
	u_int32_t send_accm;            /* send ACCM value the client should use */
	u_int32_t recv_accm;            /* recv ACCM value the client should use */
};

/* GRE and PPP structs */

/* Copied from C. S. Ananian */

#define HDLC_FLAG               0x7E
#define HDLC_ESCAPE             0x7D

#define PPTP_GRE_PROTO          0x880B
#define PPTP_GRE_VER            0x1

#define PPTP_GRE_FLAG_C         0x80
#define PPTP_GRE_FLAG_R         0x40
#define PPTP_GRE_FLAG_K         0x20
#define PPTP_GRE_FLAG_S         0x10
#define PPTP_GRE_FLAG_A         0x80

#define PPTP_GRE_IS_C(f)        ((f)&PPTP_GRE_FLAG_C)
#define PPTP_GRE_IS_R(f)        ((f)&PPTP_GRE_FLAG_R)
#define PPTP_GRE_IS_K(f)        ((f)&PPTP_GRE_FLAG_K)
#define PPTP_GRE_IS_S(f)        ((f)&PPTP_GRE_FLAG_S)
#define PPTP_GRE_IS_A(f)        ((f)&PPTP_GRE_FLAG_A)

struct pptp_gre_header {
	u_int8_t flags;         /* bitfield */
	u_int8_t ver;           /* should be PPTP_GRE_VER (enhanced GRE) */
	u_int16_t protocol;     /* should be PPTP_GRE_PROTO (ppp-encaps) */
	u_int16_t payload_len;  /* size of ppp payload, not inc. gre header */
	u_int16_t call_id;      /* peer's call_id for this session */
	u_int32_t seq;          /* sequence number.  Present if S==1 */
	u_int32_t ack;          /* seq number of highest packet recieved by */
	/* sender in this session */
};

struct gre_state {
	u_int32_t ack_sent, ack_recv;
	u_int32_t seq_sent, seq_recv;
	u_int32_t call_id_pair;
};

typedef struct pack_track {
  uint32_t seq;       // seq no of this tracked packet
  uint64_t time;      // time when this tracked packet was sent (in usecs)
} pack_track_t;

typedef struct gre_stats {
  /* statistics for GRE receive */

  uint32_t rx_accepted;  // data packet was passed to pppd
  uint32_t rx_lost;      // data packet did not arrive before timeout
  uint32_t rx_underwin;  // data packet was under window (arrived too late
			 // or duplicate packet)
  uint32_t rx_overwin;   // data packet was over window
			 // (too many packets lost?)
  uint32_t rx_buffered;  // data packet arrived earlier than expected,
			 // packet(s) before it were lost or reordered
  uint32_t rx_errors;    // OS error on receive
  uint32_t rx_truncated; // truncated packet
  uint32_t rx_invalid;   // wrong protocol or invalid flags
  uint32_t rx_acks;      // acknowledgement only

  /* statistics for GRE transmit */

  uint32_t tx_sent;      // data packet write() to GRE socket succeeded
  uint32_t tx_failed;    // data packet write() to GRE socket returned error
  uint32_t tx_short;     // data packet write() to GRE socket underflowed
  uint32_t tx_acks;      // sent packet with just ACK
  uint32_t tx_oversize;  // data packet dropped because it was too large

  /* statistics for packet tracking, for RTT calculation */

  pack_track_t pt;       // last data packet seq/time
  int rtt;               // estimated round-trip time in us

} gre_stats_t;

#define PPPINITFCS16    0xffff  /* Initial FCS value */
#define PPPGOODFCS16    0xf0b8  /* Good final FCS value */

/* For our call ID pairs */
#define PNS_VALUE 0
#define PAC_VALUE 1

#define PACKET_MAX 8196

CALC_BSS struct BB_COMMON_BSS
{
	pqueue_t *pq_head, *pq_tail;
	/* contains a list of free queue elements.*/
	pqueue_t *pq_freelist_head;

	char *pppdoptstr;
	char *speedstr;
	char *bindaddr;
	int pptp_debug;
	int pptp_noipparam;
	int pptp_logwtmp;
	int pptp_delegate;
	/* this end's call identifier */
	uint16_t unique_call_id;

	struct slot *slots;
	/* number of connection slots allocated */
	int slot_count;
	/* specifies if debugging is on or off */
	int pptpctrl_debug;
	/* pipe private to process */
	int sigpipe[2];
	gre_stats_t stats;
	int noipparam;  /* if true, don't send ipparam to ppp */
	char Gspeed[32];
	char pppdxfig[256];
	pid_t pppfork;  /* so we can kill it after disconnect */

	u_int32_t Gcall_id_pair;    /* call id (to terminate call) */

	int _pac_call_id;
	u_int16_t _pac_init;
	struct gre_state gre;

	unsigned char dg_buf[PACKET_MAX + 64 /*ip header */ ];
	union {
		struct pptp_gre_header header;
		unsigned char buffer[PACKET_MAX + sizeof(struct pptp_gre_header)];
	} u;
	char *rh_buf;           /* buffer between calls */
	int rh_bufed;           /* size of buffer */
	unsigned char dh_buf[PACKET_MAX], dh_buf_copy[PACKET_MAX];
	unsigned escape;
	unsigned char dest[2 * PACKET_MAX + 6];
	unsigned char err;
	unsigned dh_start, dh_end;
	unsigned dh_len;
	u_int16_t gci;
	/* option for timeout on starting ctrl connection */
	long pptp_stimeout;
	int pptp_connections;
	char *ppp_binary;
	u_int16_t decaps_hdlc_fcs;
	char need_strdup[1];
} bb_common_bss;
#define pq_head bb_common_bss.pq_head
#define pq_tail bb_common_bss.pq_tail
#define pq_freelist_head bb_common_bss.pq_freelist_head
#define pppdoptstr bb_common_bss.pppdoptstr
#define speedstr bb_common_bss.speedstr
#define bindaddr bb_common_bss.bindaddr
#define pptp_debug bb_common_bss.pptp_debug
#define pptp_noipparam bb_common_bss.pptp_noipparam
#define pptp_logwtmp bb_common_bss.pptp_logwtmp
#define pptp_delegate bb_common_bss.pptp_delegate
#define unique_call_id bb_common_bss.unique_call_id
#define slots bb_common_bss.slots
#define slot_count bb_common_bss.slot_count
#define pptpctrl_debug bb_common_bss.pptpctrl_debug
#define sigpipe bb_common_bss.sigpipe
#define stats bb_common_bss.stats
#define noipparam bb_common_bss.noipparam
#define Gspeed bb_common_bss.Gspeed
#define pppdxfig bb_common_bss.pppdxfig
#define pppfork bb_common_bss.pppfork
#define Gcall_id_pair bb_common_bss.Gcall_id_pair
#define _pac_call_id bb_common_bss._pac_call_id
#define _pac_init bb_common_bss._pac_init
#define gre bb_common_bss.gre
#define dg_buf bb_common_bss.dg_buf
#define u bb_common_bss.u
#define rh_buf bb_common_bss.rh_buf
#define rh_bufed bb_common_bss.rh_bufed
#define dh_buf bb_common_bss.dh_buf
#define dh_buf_copy bb_common_bss.dh_buf_copy
#define escape bb_common_bss.escape
#define dest bb_common_bss.dest
#define err bb_common_bss.err
#define dh_start bb_common_bss.dh_start
#define dh_end bb_common_bss.dh_end
#define dh_len bb_common_bss.dh_len
#define gci bb_common_bss.gci
#define pptp_stimeout bb_common_bss.pptp_stimeout
#define pptp_connections bb_common_bss.pptp_connections
#define ppp_binary bb_common_bss.ppp_binary
#define decaps_hdlc_fcs bb_common_bss.decaps_hdlc_fcs
#define need_strdup bb_common_bss.need_strdup

#define MIN_CAPACITY 128 /* min allocated buffer for a packet */

/* wait this many seconds for missing packets before forgetting about them */
#define DEFAULT_PACKET_TIMEOUT 0.3
/* DEFAULT_PACKET_TIMEOUT * 1000000 */
#define packet_timeout_usecs 300000l

static int pqueue_alloc(u_int32_t seq, unsigned char *packet, int packlen, pqueue_t **new)
{
  pqueue_t *newent;

  /* search the freelist for one that has sufficient space  */
  if (pq_freelist_head) {

    for (newent = pq_freelist_head; newent; newent = newent->next) {

      if (newent->capacity >= packlen) {

	/* unlink from freelist */
	if (pq_freelist_head == newent)
	  pq_freelist_head = newent->next;

	if (newent->prev)
	  newent->prev->next = newent->next;

	if (newent->next)
	  newent->next->prev = newent->prev;

	if (pq_freelist_head)
	  pq_freelist_head->prev = NULL;

	break;
      } /* end if capacity >= packlen */
    } /* end for */

    /* nothing found? Take first and reallocate it */
    if (NULL == newent) {

      newent = pq_freelist_head;
      pq_freelist_head = pq_freelist_head->next;

      if (pq_freelist_head)
	pq_freelist_head->prev = NULL;

      newent->packet = (unsigned char *)realloc(newent->packet, packlen);

      if (!newent->packet) {
	xlog_warn("error reallocating packet: %m");
	return -1;
      }
      newent->capacity = packlen;
    }

  } else {

    /* allocate a new one */
    newent = (pqueue_t *)malloc( sizeof(pqueue_t) );
    if (!newent) {
      xlog_warn("error allocating newent: %m");
      return -1;
    }
    newent->capacity = 0;

  }

  if ( ! newent->capacity  ) {
    /* a new queue entry was allocated. Allocate the packet buffer */
    int size = packlen < MIN_CAPACITY ? MIN_CAPACITY : packlen;
    /* Allocate at least MIN_CAPACITY */
    newent->packet = (unsigned char *)malloc(size);
    if (!newent->packet) {
      xlog_warn("error allocating packet: %m");
      return -1;
    }
    newent->capacity = size;
  } /* endif ! capacity */
  assert( newent->capacity >= packlen );
  /* store the contents into the buffer */
  memcpy(newent->packet, packet, packlen);

  newent->next = newent->prev = NULL;
  newent->seq     = seq;
  newent->packlen = packlen;

  gettimeofday(&newent->expires, NULL);
  newent->expires.tv_usec += packet_timeout_usecs;
  newent->expires.tv_sec  += (newent->expires.tv_usec / 1000000);
  newent->expires.tv_usec %= 1000000;

  *new = newent;
  return 0;
}


static int pqueue_add (u_int32_t seq, unsigned char *packet, int packlen)
{
  pqueue_t *newent, *point;

  /* get a new entry */
  if ( 0 != pqueue_alloc(seq, packet, packlen, &newent) ) {
    return -1;
  }

  for (point = pq_head; point != NULL; point = point->next) {
    if (point->seq == seq) {
      // queue already contains this packet
      xlog_warn("discarding duplicate packet %u", seq);
      return -1;
    }
    if (point->seq > seq) {
      // gone too far: point->seq > seq and point->prev->seq < seq
      if (point->prev) {
	// insert between point->prev and point
	point->prev->next = newent;
      } else {
	// insert at head of queue, before point
	pq_head = newent;
      }
      newent->prev = point->prev; // will be NULL, at head of queue
      newent->next = point;
      point->prev = newent;
      return 0;
    }
  }

  /* We didn't find anywhere to insert the packet,
   * so there are no packets in the queue with higher sequences than this one,
   * so all the packets in the queue have lower sequences,
   * so this packet belongs at the end of the queue (which might be empty)
   */

  if (pq_head == NULL) {
    pq_head = newent;
  } else {
    pq_tail->next = newent;
  }
  newent->prev = pq_tail;
  pq_tail = newent;

  return 0;
}


static int pqueue_del (pqueue_t *point)
{
  /* unlink from pq */
  if (pq_head == point) pq_head = point->next;
  if (pq_tail == point) pq_tail = point->prev;
  if (point->prev) point->prev->next = point->next;
  if (point->next) point->next->prev = point->prev;

  /* add point to the freelist */
  point->next = pq_freelist_head;
  point->prev = NULL;

  if (point->next)
    point->next->prev = point;
  pq_freelist_head = point;

  return 0;
}


static int pqueue_expiry_time (pqueue_t *entry)
{
  struct timeval tv;
  int expiry_time;

  gettimeofday(&tv, NULL);
  expiry_time  = (entry->expires.tv_sec  - tv.tv_sec) * 1000000;
  expiry_time += (entry->expires.tv_usec - tv.tv_usec);
  return expiry_time;
}

/*
 * pptpd.c
 *
 * Grabs any command line argument and processes any further options in
 * the pptpd config file, before throwing over to pptpmanager.c.
 *
 */


#define DEBUG_IP_PARSER 0

/* Return the address or NULL if not valid */
static char *validip(char *hostname)
{
	/* Try to parse IP directly */
	if (inet_addr(hostname) != INADDR_NONE)
		return hostname;
	else
		return NULL;
}

/* Check if it's a valid IP range */
static int isIpRange(char *str)
{
	int dashes = 0;
	int dots = 0;

#if DEBUG_IP_PARSER
	syslog(LOG_DEBUG, "MGR: Checking if %s is a valid IP range", str);
#endif
	do {
		if (*str == '-')
			dashes++;
		else if (*str == '.')
			dots++;
		else if (!strchr("0123456789", *str)) {
#if DEBUG_IP_PARSER
			syslog(LOG_DEBUG, "MGR: Not an IP range: character %c is not valid", *str);
#endif
			return 0;
		}
	} while (*++str);
#if DEBUG_IP_PARSER
	syslog(LOG_DEBUG, "MGR: Dashes = %d (wanted: 1), Dots = %d (wanted: 4)", dashes, dots);
#endif
	return (dashes == 1 && dots == 3);
}

/* added for hostname/address lookup */
static char *lookup(char *hostname, const char *add_err_msg)
{
	return bb_gethostbyname(BB_GHBN_NTOP | BB_GHBN_ADD_ERR_MSG | BB_GHBN_ERR_DIE,
				hostname, add_err_msg);
}

static void slot_set_local(int i, char *ip)
{
  struct slot *slot = &slots[i];
  if (slot->local) free(slot->local);
  slot->local = strdup(ip);
}

static void slot_set_remote(int i, char *ip)
{
  struct slot *slot = &slots[i];
  if (slot->remote) free(slot->remote);
  slot->remote = strdup(ip);
}

static char *slot_get_local(int i)
{
  struct slot *slot = &slots[i];
  return slot->local;
}

/* process a type 0 (LOCAL) or type 1 (REMOTE) IP string */
static void processIPStr(int type, char *ipstr)
{
	int pos;

	char *tmpstr;
	/* char tmpstr2[20]; xxx.xxx.xxx.xxx-xxx (largest we can get) */
	char tmpstr2[128];      /* allow hostnames */
	char *tmpstr3;
	char tmpstr5[16];
	char *tmpstr6;
	char *tmpstr7;
	int num;

	char ipa[8];            /* xxx-xxx (largest we can get) */
	char ipb[8];
	char ipc[8];
	char ipd[8];

	char ip_pre[13];        /* xxx.xxx.xxx. (largest we can get) */
	char ip_post[13];

	char ipl[4];
	char ipu[4];

	int bail = FALSE;       /* so we know when to stop formatting the ip line */

	int lower, upper, n;

	num = 0;

	while (!bail) {
		if ((tmpstr = strchr(ipstr, ',')) == NULL) {
			/* last (or only) entry reached */
			safe_strncpy(tmpstr2, ipstr, sizeof(tmpstr2));
			bail = TRUE;
		} else {
			pos = tmpstr - ipstr;
			ipstr[pos] = '\0';
			safe_strncpy(tmpstr2, ipstr, sizeof(tmpstr2));
			ipstr = tmpstr + 1;
		}

#if DEBUG_IP_PARSER
		syslog(LOG_DEBUG, "MGR: Parsing segment: %s", tmpstr2);
#endif

		if (!isIpRange(tmpstr2)) {
			/* We got a normal IP
			 * Check if the IP address is valid, use it if so
			 */
			tmpstr7 = lookup(tmpstr2, "MGR: Bad IPv4 address in config file!");
			if (num == pptp_connections) {
				xlog_warn("MGR: connections limit (%d) reached, extra IP addresses ignored", pptp_connections);
				return;
			}
#if DEBUG_IP_PARSER
			syslog(LOG_DEBUG, "MGR: Setting IP %d = %s", num, tmpstr7);
#endif
			if (type == LOCAL)
				slot_set_local(num, tmpstr7);
			else
				slot_set_remote(num, tmpstr7);
			free(tmpstr7);
			num++;
		} else {
			/* Got a range;
			 * eg. 192.168.0.234-238
			 * or (thanx Kev! :-).. i thought i was finished :-)
			 * 192.168-178.1.231
			 */

			/* lose the "."'s */
			while ((tmpstr3 = strchr(tmpstr2, '.')) != NULL) {
				pos = tmpstr3 - tmpstr2;
				tmpstr2[pos] = ' ';
			}

			if ((tmpstr3 = strchr(tmpstr2, '-')) == NULL ||
			    strchr(tmpstr3 + 1, '-') != NULL) {
				bb_error_msg("MGR: Confused in IP parse routines (multiple hyphens)");
				continue;
			}
			/* should be left with "192 168 0 234-238"
			 * or 192 168-178 1 231
			 */

			sscanf(tmpstr2, "%7s %7s %7s %7s", ipa, ipb, ipc, ipd);

			if ((tmpstr6 = strchr(ipd, '-')) != NULL) {
				pos = tmpstr6 - ipd;
				ipd[pos] = ' ';
				sscanf(ipd, "%3s %3s", ipl, ipu);
#if DEBUG_IP_PARSER
				syslog(LOG_DEBUG, "MGR: (lower upper) = (%s %s)", ipl, ipu);
#endif
				lower = atoi(ipl);
				upper = atoi(ipu);
#if DEBUG_IP_PARSER
				syslog(LOG_DEBUG, "MGR: Range = %d to %d on 4th segment", lower, upper);
#endif
				sprintf(ip_pre, "%.3s.%.3s.%.3s.", ipa, ipb, ipc);
				ip_post[0] = '\0';
#if DEBUG_IP_PARSER
				syslog(LOG_DEBUG, "MGR: Pre = %s Post = %s", ip_pre, ip_post);
#endif
			} else if ((tmpstr6 = strchr(ipc, '-')) != NULL) {
				pos = tmpstr6 - ipc;
				ipc[pos] = ' ';
				sscanf(ipc, "%3s %3s", ipl, ipu);
#if DEBUG_IP_PARSER
				syslog(LOG_DEBUG, "MGR: (lower upper) = (%s %s)", ipl, ipu);
#endif
				lower = atoi(ipl);
				upper = atoi(ipu);
#if DEBUG_IP_PARSER
				syslog(LOG_DEBUG, "MGR: Range = %d to %d on 3rd segment", lower, upper);
#endif
				sprintf(ip_pre, "%.3s.%.3s.", ipa, ipb);
				sprintf(ip_post, ".%.3s", ipd);
#if DEBUG_IP_PARSER
				syslog(LOG_DEBUG, "MGR: Pre = %s Post = %s", ip_pre, ip_post);
#endif
			} else if ((tmpstr6 = strchr(ipb, '-')) != NULL) {
				pos = tmpstr6 - ipb;
				ipb[pos] = ' ';
				sscanf(ipb, "%3s %3s", ipl, ipu);
#if DEBUG_IP_PARSER
				syslog(LOG_DEBUG, "MGR: (lower upper) = (%s %s)", ipl, ipu);
#endif
				lower = atoi(ipl);
				upper = atoi(ipu);
#if DEBUG_IP_PARSER
				syslog(LOG_DEBUG, "MGR: Range = %d to %d on 2nd segment", lower, upper);
#endif
				sprintf(ip_pre, "%.3s.", ipa);
				sprintf(ip_post, ".%.3s.%.3s", ipc, ipd);
#if DEBUG_IP_PARSER
				syslog(LOG_DEBUG, "MGR: Pre = %s Post = %s", ip_pre, ip_post);
#endif
			} else if ((tmpstr6 = strchr(ipa, '-')) != NULL) {
				pos = tmpstr6 - ipa;
				ipa[pos] = ' ';
				sscanf(ipa, "%3s %3s", ipl, ipu);
#if DEBUG_IP_PARSER
				syslog(LOG_DEBUG, "MGR: (lower upper) = (%s %s)", ipl, ipu);
#endif
				lower = atoi(ipl);
				upper = atoi(ipu);
#if DEBUG_IP_PARSER
				syslog(LOG_DEBUG, "MGR: Range = %d to %d on 1st segment", lower, upper);
#endif
				ip_pre[0] = '\0';
				sprintf(ip_post, ".%.3s.%.3s.%.3s", ipb, ipc, ipd);
#if DEBUG_IP_PARSER
				syslog(LOG_DEBUG, "MGR: Pre = %s Post = %s", ip_pre, ip_post);
#endif
			} else {
				bb_error_msg("MGR: Confused in IP parse routines (lost hyphen)");
				continue;
			}

			for (n = lower; n <= upper; n++) {
				sprintf(tmpstr5, "%s%d%s", ip_pre, n, ip_post);
				/* Check if the ip address is valid */
				if ((tmpstr7 = validip(tmpstr5)) == NULL)
					bb_error_msg_and_die("MGR: Bad IP address (%s) in config file!", tmpstr5);
				if (num == pptp_connections) {
					xlog_warn("MGR: connections limit (%d) reached, extra IP addresses ignored", pptp_connections);
					return;
				}
#if DEBUG_IP_PARSER
				syslog(LOG_DEBUG, "MGR: Setting IP %d = %s", num, tmpstr7);
#endif
				if (type == LOCAL)
					slot_set_local(num, tmpstr7);
				else
					slot_set_remote(num, tmpstr7);
				num++;
			}
		}
	}
	if (num == 1 && type == LOCAL && pptp_connections > 1) {
#if DEBUG_IP_PARSER
		syslog(LOG_DEBUG, "MGR: Setting all %d local IPs to %s", pptp_connections, slot_get_local(0));
#endif
		for (n = 1; n < pptp_connections; n++)
			slot_set_local(n, slot_get_local(0));
	} else if (pptp_connections > num) {
		syslog(LOG_INFO, "MGR: Maximum of %d connections reduced to %d, not enough IP addresses given",
		       pptp_connections, num);
		pptp_connections = num;
	}
}

static void showusage(void)
{
	printf("\npptpd v%s\n", VERSION);
	printf("Usage: pptpd [options], where options are:\n\n");
	printf(" [-c] [--conf file]        Specifies the config file to read default\n");
	printf("                           settings from (default is %s).\n", PPTPD_CONFIG_FILE_DEFAULT);
	printf(" [-d] [--debug]            Turns on debugging (to syslog).\n");
	printf(" [-e] [--ppp file]         Use alternate pppd binary, default %s.\n", PPP_BINARY);
	printf(" [-f] [--fg]               Run in foreground.\n");
	printf(" [-h] [--help]             Displays this help message.\n");
	printf(" [-i] [--noipparam]        Suppress the passing of the client's IP address\n");
	printf("                           to PPP, which is done by default otherwise.\n");
	printf(" [-l] [--listen x.x.x.x]   Specifies IP of local interface to listen to.\n");
	printf(" [-o] [--option file]      Specifies the PPP options file to use\n");
	printf("                           (default is /etc/ppp/options).\n");
	printf(" [-p] [--pidfile file]     Specifies the file to write the process ID to\n");
	printf("                           (default is /var/run/pptpd.pid).\n");
	printf(" [-s] [--speed baud]       Specifies the baud speed for the PPP daemon\n");
	printf("                           (default is 115200).\n");
	printf(" [-t] [--stimeout seconds] Specifies the timeout for the first packet. This is a DOS protection\n");
	printf("                           (default is 10).\n");
	printf(" [-v] [--version]          Displays the pptpd version number.\n");
	printf(" [-w] [--logwtmp]          Update wtmp as users login.\n");
	printf(" [-C] [--connections n]    Limit on number of connections.\n");
	printf(" [-D] [--delegate]         Delegate IP allocation to pppd.\n");

	printf("\n\nLogs and debugging go to syslog as DAEMON.");

	printf("\n\nCommand line options will override any default settings and any settings\n");
	printf("specified in the config file (default config file: %s).\n\n", PPTPD_CONFIG_FILE_DEFAULT);
}

/*
 * configfile.c
 *
 * Methods for accessing the PPTPD config file and searching for
 * PPTPD keywords.
 */

/*
 * read_config_file
 *
 * This method opens up the file specified by 'filename' and searches
 * through the file for 'keyword'. If 'keyword' is found any string
 * following it is stored in 'value'.
 *
 * args: filename (IN) - config filename
 *       keyword (IN) - word to search for in config file
 *       value (OUT) - value of keyword
 *
 * retn: -1 on error, 0 if keyword not found, 1 on value success
 */

static char *read_config_file(config_read_t *crt, char *keyword, char *value)
{
	size_t len, keyword_len;

	if(crt == NULL)
		return NULL;
	if(value)
		*value = '\0';
	rewind(crt->in);
	keyword_len = strlen(keyword);
	while( (crt = config_read(crt, "#\1\2\0\0")) != NULL) {
		if(crt->ntokens == 0)/* EOF, to rewind for next search next key */
			return NULL;
		len = strlen(crt->tokens[0]);
		if (len == keyword_len && !strncmp(crt->tokens[0], keyword, len)) {
			char *out = crt->ntokens > 1 ? crt->tokens[1] : "";

			if(value) {
				if(value != need_strdup)
					out = safe_strncpy(value, out, MAX_CONFIG_STRING_SIZE);
				else
					out = xstrdup(out);
			}
			return out;
		}
	}
	return NULL;
}


static void
slot_iterate(int count, void (*callback) (struct slot *slot))
{
  int i;
  for(i=0; i<count; i++)
    (*callback)(&slots[i]);
}

#if KEEP_UNUSE
static void slot_slot_free(struct slot *slot)
{
  slot->pid = 0;
  if (slot->local) free(slot->local);
  slot->local = NULL;
  if (slot->remote) free(slot->remote);
  slot->remote = NULL;
}

static void slot_free(void)
{
  slot_iterate(slot_count, slot_slot_free);
  free(slots);
  slots = NULL;
  slot_count = 0;
}
#endif

static void slot_slot_init(struct slot *slot)
{
  slot->pid = 0;
  slot->local = NULL;
  slot->remote = NULL;
}

static void slot_init(int count)
{
  slot_count = count;
  slots = (struct slot *) calloc(slot_count, sizeof(struct slot));
  slot_iterate(slot_count, slot_slot_init);
}

static char *slot_get_remote(int i)
{
  struct slot *slot = &slots[i];
  return slot->remote;
}

/*
 * Author: Kevin Thayer
 *
 * this routine sets up the arguments for the call handler and calls it.
 */
static void connectCall(int clientSocket, int clientNumber) ATTRIBUTE_NORETURN;
static void connectCall(int clientSocket, int clientNumber)
{

#define NUM2ARRAY(array, num) snprintf(array, sizeof(array), "%d", num)

	char *ctrl_argv[16];    /* arguments for launching 'pptpctrl' binary */

	int pptpctrl_argc = 0;  /* count the number of arguments sent to pptpctrl */

	/* lame strings to hold passed args. */
	char ctrl_debug[2];
	char ctrl_noipparam[2];
	char pppdoptfile_argv[2];
	char speedgiven_argv[2];
	extern char **environ;

	char callid_argv[16];

	/*
	 * Launch the CTRL manager binary; we send it some information such as
	 * speed and option file on the command line.
	 */

	ctrl_argv[pptpctrl_argc++] = PPTP_CTRL_BIN "                                ";

	/* Pass socket as stdin */
	if (clientSocket != 0) {
		dup2(clientSocket, 0);
		close(clientSocket);
	}

	/* get argv set up */
	NUM2ARRAY(ctrl_debug, pptp_debug ? 1 : 0);
	ctrl_debug[1] = '\0';
	ctrl_argv[pptpctrl_argc++] = ctrl_debug;

	NUM2ARRAY(ctrl_noipparam, pptp_noipparam ? 1 : 0);
	ctrl_noipparam[1] = '\0';
	ctrl_argv[pptpctrl_argc++] = ctrl_noipparam;

	/* optionfile = TRUE or FALSE; so the CTRL manager knows whether to load a non-standard options file */
	NUM2ARRAY(pppdoptfile_argv, pppdoptstr ? 1 : 0);
	pppdoptfile_argv[1] = '\0';
	ctrl_argv[pptpctrl_argc++] = pppdoptfile_argv;
	if (pppdoptstr) {
		/* send the option filename so the CTRL manager can launch pppd with this alternate file */
		ctrl_argv[pptpctrl_argc++] = pppdoptstr;
	}
	/* tell the ctrl manager whether we were given a speed */
	NUM2ARRAY(speedgiven_argv, speedstr ? 1 : 0);
	speedgiven_argv[1] = '\0';
	ctrl_argv[pptpctrl_argc++] = speedgiven_argv;
	if (speedstr) {
		/* send the CTRL manager the speed of the connection so it can fire pppd at that speed */
		ctrl_argv[pptpctrl_argc++] = speedstr;
	}
	if (pptp_delegate) {
		/* no local or remote address to specify */
		ctrl_argv[pptpctrl_argc++] = "0";
		ctrl_argv[pptpctrl_argc++] = "0";
	} else {
		/* specify local & remote addresses for this call */
		ctrl_argv[pptpctrl_argc++] = "1";
		ctrl_argv[pptpctrl_argc++] = slot_get_local(clientNumber);
		ctrl_argv[pptpctrl_argc++] = "1";
		ctrl_argv[pptpctrl_argc++] = slot_get_remote(clientNumber);
	}

	/* our call id to be included in GRE packets the client
	 * will send to us */
	NUM2ARRAY(callid_argv, unique_call_id);
	ctrl_argv[pptpctrl_argc++] = callid_argv;

	/* pass path to ppp binary */
	ctrl_argv[pptpctrl_argc++] = ppp_binary;

	/* pass logwtmp flag */
	ctrl_argv[pptpctrl_argc++] = pptp_logwtmp ? "1" : "0";

	/* note: update pptpctrl.8 if the argument list format is changed */

	/* terminate argv array with a NULL */
	ctrl_argv[pptpctrl_argc] = NULL;
	pptpctrl_argc++;

	/* ok, args are setup: invoke the call handler */
	execve(PPTP_CTRL_BIN, ctrl_argv, environ);
	bb_perror_msg_and_die("MGR: Failed to exec " PPTP_CTRL_BIN "!");
}


/*
 * Author: Kevin Thayer
 *
 * This creates a socket to listen on, sets the max # of pending connections and
 * various other options.
 *
 * Returns the fd of the host socket.
 *
 * The function return values are:
 * 0 for sucessful
 * -1 for bad socket creation
 * -2 for bad socket options
 * -3 for bad bind
 * -4 for bad listen
 */
static int createHostSocket(int *hostSocket)
{
	struct sockaddr_in address;

	/* create the master socket and check it worked */
	*hostSocket = bb_xsocket(BB_XSOCKET_FLG_ERR_DIE);

	/* set master socket to allow daemon to be restarted with connections active  */
	if (setsockopt_1i(*hostSocket, SOL_SOCKET, SO_REUSEADDR, 1) < 0)
		return -2;

	/* set up socket */
	memset(&address, 0, sizeof(address));
	address.sin_family = AF_INET;
	if(bindaddr)
		address.sin_addr.s_addr = inet_addr(bindaddr);
	else
		address.sin_addr.s_addr = INADDR_ANY;
	address.sin_port = bb_lookup_tcp_port("pptp", PPTP_PORT);

	/* bind the socket to the pptp port */
	bb_xbind(*hostSocket, (struct sockaddr *) &address, sizeof(address));

	/* minimal backlog to avoid DoS */
	if (listen(*hostSocket, 3) < 0)
		return -4;

	return 0;
}

#define GET_VALUE(which, where) ((which ## _VALUE) ? ((where) & 0xffff) : ((where) >> 16))

#define NOTE_VALUE(which, where, what) ((which ## _VALUE) \
					  ? ((where) = ((where) & 0xffff0000) | (what)) \
					  : ((where) = ((where) & 0xffff) | ((what) << 16)))

static void slot_set_pid(int i, pid_t pid)
{
  struct slot *slot = &slots[i];
  slot->pid = pid;
}

static int slot_find_by_pid(pid_t pid)
{
  int i;
  for(i=0; i < slot_count; i++) {
    struct slot *slot = &slots[i];
    if (slot->pid == pid) return i;
  }
  return -1;
}

static int slot_find_empty(void)
{
  return slot_find_by_pid(0);
}


static void sigchld_responder(int sig)
{
  int child, status;

  while ((child = waitpid_no_int(-1, &status, WNOHANG, 0)) > 0) {
    if (pptp_delegate) {
      if (pptp_debug) syslog(LOG_DEBUG, "MGR: Reaped child %d", child);
    } else {
      int i;
      i = slot_find_by_pid(child);
      if (i != -1) {
	slot_set_pid(i, 0);
	if (pptp_debug) syslog(LOG_DEBUG, "MGR: Reaped child %d", child);
      } else {
	syslog(LOG_INFO, "MGR: Reaped unknown child %d", child);
      }
    }
  }
}

/* signal to pipe delivery implementation */


/* return the signal pipe read file descriptor for select(2) */
#define sigpipe_fd() sigpipe[0]

/* create a signal pipe, returns 0 for success, -1 with errno for failure */
static int sigpipe_create(void)
{
  int rc;

  rc = pipe(sigpipe);
  if (rc < 0) return rc;

  bb_cloexec(sigpipe[0]);
  bb_cloexec(sigpipe[1]);

#ifdef O_NONBLOCK
#define FLAG_TO_SET O_NONBLOCK
#else
#ifdef SYSV
#define FLAG_TO_SET O_NDELAY
#else /* BSD */
#define FLAG_TO_SET FNDELAY
#endif
#endif

  rc = fcntl(sigpipe[1], F_GETFL);
  if (rc != -1)
    rc = fcntl(sigpipe[1], F_SETFL, rc | FLAG_TO_SET);
  if (rc < 0) return rc;
  return 0;
#undef FLAG_TO_SET
}

/* generic handler for signals, writes signal number to pipe */
static void sigpipe_handler(int signum);

/* assign a signal number to the pipe */
static void sigpipe_assign(int signum)
{
  bb_signal((1 << signum), sigpipe_handler);
}

static void sigpipe_handler(int signum)
{
  write(sigpipe[1], &signum, sizeof(signum));
  sigpipe_assign(signum);
}

/* read and return the pending signal from the pipe */
static int sigpipe_read(void)
{
  int signum;
  read(sigpipe[0], &signum, sizeof(signum));
  return signum;
}


static int pptp_manager(void)
{
	int firstOpen = -1;
	int ctrl_pid;
	socklen_t addrsize;

	int hostSocket;
	fd_set connSet;

	int rc, sig_fd;

	rc = sigpipe_create();
	if (rc < 0)
		bb_perror_msg_and_die("MGR: unable to setup sigchld pipe!");

	sigpipe_assign(SIGCHLD);
	sigpipe_assign(SIGTERM);
	sig_fd = sigpipe_fd();

	/* openlog() not required, done in pptpd.c */

	syslog(LOG_INFO, "MGR: Manager process started");

	if (!pptp_delegate) {
		syslog(LOG_INFO, "MGR: Maximum of %d connections available",
		       pptp_connections);
	}

	/* Connect the host socket and activate it for listening */
	if (createHostSocket(&hostSocket) < 0)
		bb_perror_msg_and_die("MGR: Couldn't create host socket");

	while (1) {
		int max_fd;
		FD_ZERO(&connSet);
		if (pptp_delegate) {
			FD_SET(hostSocket, &connSet);
		} else {
			firstOpen = slot_find_empty();
			if (firstOpen == -1) {
				bb_error_msg("MGR: No free connection slots or IPs - no more clients can connect!");
			} else {
				FD_SET(hostSocket, &connSet);
			}
		}
		max_fd = hostSocket;

		FD_SET(sig_fd, &connSet);
		if (max_fd < sig_fd) max_fd = sig_fd;

		while (1) {
			if (select(max_fd + 1, &connSet, NULL, NULL, NULL) != -1) break;
			if (errno == EINTR) continue;
			bb_perror_msg_and_die("MGR: Error with manager select()");
		}

		if (FD_ISSET(sig_fd, &connSet)) {       /* SIGCHLD */
			int signum = sigpipe_read();
			if (signum == SIGCHLD)
				sigchld_responder(signum);
			else if (signum == SIGTERM)
				return signum;
		}

		if (FD_ISSET(hostSocket, &connSet)) {   /* A call came! */
			int clientSocket;
			struct sockaddr_in client_addr;

			/* Accept call and launch PPTPCTRL */
			addrsize = sizeof(client_addr);
			clientSocket = accept(hostSocket, (struct sockaddr *) &client_addr, &addrsize);

#if HAVE_LIBWRAP
			if (clientSocket != -1) {
				struct request_info r;
				request_init(&r, RQ_DAEMON, "pptpd", RQ_FILE, clientSocket, NULL);
				fromhost(&r);
				if (!hosts_access(&r)) {
					/* send a permission denied message? this is a tcp wrapper
					 * type deny so probably best to just drop it immediately like
					 * this, as tcp wrappers usually do.
					 */
					close(clientSocket);
					/* this would never be file descriptor 0, so use it as a error
					 * value
					 */
					clientSocket = 0;
				}
			}
#endif
			if (clientSocket == -1) {
				/* accept failed, but life goes on... */
				bb_perror_msg("MGR: accept() failed");
			} else if (clientSocket != 0) {
				fd_set rfds;
				struct timeval tv;
				struct pptp_header ph;

				/* TODO: this select below prevents
				   other connections from being
				   processed during the wait for the
				   first data packet from the
				   client. */

				/*
				 * DOS protection: get a peek at the first packet
				 * and do some checks on it before we continue.
				 * A 10 second timeout on the first packet seems reasonable
				 * to me,  if anything looks sus,  throw it away.
				 */

				FD_ZERO(&rfds);
				FD_SET(clientSocket, &rfds);
				tv.tv_sec = pptp_stimeout;
				tv.tv_usec = 0;
				if (select(clientSocket + 1, &rfds, NULL, NULL, &tv) <= 0) {
					bb_perror_msg("MGR: dropped slow initial connection");
					close(clientSocket);
					continue;
				}

				if (recv(clientSocket, &ph, sizeof(ph), MSG_PEEK) !=
						sizeof(ph)) {
					bb_error_msg("MGR: dropped small initial connection");
					close(clientSocket);
					continue;
				}

				ph.length = ntohs(ph.length);
				ph.pptp_type = ntohs(ph.pptp_type);
				ph.magic = ntohl(ph.magic);
				ph.ctrl_type = ntohs(ph.ctrl_type);

				if (ph.length <= 0 || ph.length > PPTP_MAX_CTRL_PCKT_SIZE) {
					xlog_warn("MGR: initial packet length %d outside "
							"(0 - %d)",  ph.length, PPTP_MAX_CTRL_PCKT_SIZE);
					goto dos_exit;
				}

				if (ph.magic != PPTP_MAGIC_COOKIE) {
					xlog_warn("MGR: initial packet bad magic");
					goto dos_exit;
				}

				if (ph.pptp_type != PPTP_CTRL_MESSAGE) {
					xlog_warn("MGR: initial packet has bad type");
					goto dos_exit;
				}

				if (ph.ctrl_type != START_CTRL_CONN_RQST) {
					xlog_warn("MGR: initial packet has bad ctrl type "
							"0x%x", ph.ctrl_type);
		dos_exit:
					close(clientSocket);
					continue;
				}

#ifndef HAVE_FORK
				switch (ctrl_pid = vfork())
#else
				switch (ctrl_pid = fork())
#endif
				{
				case -1:        /* error */
					bb_perror_msg("MGR: fork() failed launching " PPTP_CTRL_BIN);
					close(clientSocket);
					break;

				case 0: /* child */
					close(hostSocket);
					if (pptp_debug)
						syslog(LOG_DEBUG, "MGR: Launching " PPTP_CTRL_BIN " to handle client");
					connectCall(clientSocket, !pptp_delegate ? firstOpen : 0);
					/* NORETURN */
				default:        /* parent */
					close(clientSocket);
					unique_call_id += MAX_CALLS_PER_TCP_LINK;
					if (!pptp_delegate)
						slot_set_pid(firstOpen, ctrl_pid);
					break;
				}
			}
		}               /* FD_ISSET(hostSocket, &connSet) */
	}                       /* while (1) */
}                               /* pptp_manager() */

#if ENABLE_FEATURE_LONG_OPTIONS
static const struct option long_options[] =
{
	{"conf", 1, 0, 'c'},
	{"debug", 0, 0, 'd'},
	{"ppp", 1, 0, 'e'},
	{"fg", 0, 0, 'f'},
	{"help", 0, 0, 'h'},
	{"noipparam", 0, 0, 'i'},
	{"listen", 1, 0, 'l'},
	{"option", 1, 0, 'o'},
	{"pidfile", 1, 0, 'p'},
	{"speed", 1, 0, 's'},
	{"stimeout", 1, 0, 't'},
	{"version", 0, 0, 'v'},
	{"logwtmp", 0, 0, 'w'},
	{"connections", 1, 0, 'C'},
	{"delegate", 0, 0, 'D'},
	{0, 0, 0, 0}
};
#endif

int pptpd_main(void)
{
	/* command line options */
	int c;

	/* function-local options */
	int foreground;
	char *pid_file = NULL;

	/* config file */
	char *configFile = NULL;
	config_read_t *crt;

	/* config file parsing temp strings */
	char tmp[MAX_CONFIG_STRING_SIZE];

	decaps_hdlc_fcs = PPPINITFCS16;
	/* open a connection to the syslog daemon */
	bb_openlog(LOG_PID, PPTP_FACILITY);

	/* Start value for our call IDs on this TCP link */
	unique_call_id = 0xFFFF;
	/* process command line options */
#if ENABLE_FEATURE_LONG_OPTIONS
	bb_applet_long_options = long_options;
#endif
	c = bb_xgetopt_ulflags("c:de:fhil:o:p:s:t:vwC:D\0:t-C", &configFile,
	    &ppp_binary, &bindaddr, &pppdoptstr, &pid_file, &speedstr,
	    &pptp_stimeout, &pptp_connections);

	if(configFile) {
		/* --conf */
		crt = config_read_prepare(configFile, 'e');
	} else {
		configFile = PPTPD_CONFIG_FILE_DEFAULT;
		crt = config_read_prepare(configFile, 'w');
	}
	if(crt != NULL)
		crt->flg_no_close = 1;

	pptp_debug = c & 2; /* --debug */
	/* --ppp 4 */
	foreground = c & 8; /* --fg */
	if((c & 16)) {
	    /* --help */
	    showusage();
	    return 0;
	}
	pptp_noipparam = c & 32; /* --noipparam */
	/* --listen 64 */
	/* --option 128 */
	/* --pidfile 256 */
	/* --speed 512 */
	/* --stimeout 1024 */
	if((c & 2048)) {
		/* --version */
		printf("pptpd v%s\n", VERSION);
		return 0;
	}
	pptp_logwtmp = c & 4096;    /* --logwtmp */
	/* --connections 8192 */
	pptp_delegate = c & 16384;  /* --delegate */

	/* Now that we have all the command line args.. lets open the
	 * conf file and add anything else (remembering not to override
	 * anything since the command line has more privilages :-)
	 */

	if (pptp_connections <= 0 && read_config_file(crt, CONNECTIONS_KEYWORD, tmp) != NULL) {
		pptp_connections = atoi(tmp);
		if (pptp_connections <= 0)
			pptp_connections = CONNECTIONS_DEFAULT;
	}

	slot_init(pptp_connections);

	if (!pptp_debug && read_config_file(crt, DEBUG_KEYWORD, NULL) != NULL)
		pptp_debug = TRUE;

	if (pptp_stimeout <= 0 && read_config_file(crt, STIMEOUT_KEYWORD, tmp) != NULL) {
		pptp_stimeout = atol(tmp);
		if (pptp_stimeout <= 0)
			pptp_stimeout = STIMEOUT_DEFAULT;
	}

	if (!pptp_noipparam && read_config_file(crt, NOIPPARAM_KEYWORD, NULL) != NULL)
		pptp_noipparam = TRUE;

	if (!bindaddr)
		bindaddr = read_config_file(crt, LISTEN_KEYWORD, tmp);

	if (bindaddr)
		bindaddr = lookup(bindaddr, "MGR: Invalid listening address");

	if (!speedstr)
		speedstr = read_config_file(crt, SPEED_KEYWORD, need_strdup);

	if (!pppdoptstr)
		pppdoptstr = read_config_file(crt, PPPD_OPTION_KEYWORD, need_strdup);

	if (!ppp_binary)
		ppp_binary = read_config_file(crt, PPP_BINARY_KEYWORD, need_strdup);

	if (!pptp_logwtmp && read_config_file(crt, LOGWTMP_KEYWORD, NULL) != NULL)
		pptp_logwtmp = TRUE;

	if (!pptp_delegate && read_config_file(crt, DELEGATE_KEYWORD, NULL) != NULL)
		pptp_delegate = TRUE;

	if (!pid_file)
		pid_file = read_config_file(crt, PIDFILE_KEYWORD, need_strdup);

	if (!pptp_delegate) {
		/* NOTE: remote then local, reason can be seen at the end of processIPStr */

		/* grab the remoteip string from the config file */
		if (read_config_file(crt, REMOTEIP_KEYWORD, tmp) == NULL) {
			/* use "smart" defaults */
			safe_strncpy(tmp, DEFAULT_REMOTE_IP_LIST, sizeof(tmp));
		}
		processIPStr(REMOTE, tmp);

		/* grab the localip string from the config file */
		if (read_config_file(crt, LOCALIP_KEYWORD, tmp) == NULL) {
			/* use "smart" defaults */
			safe_strncpy(tmp, DEFAULT_LOCAL_IP_LIST, sizeof(tmp));
		}
		processIPStr(LOCAL, tmp);
	}
	config_close(crt);

	/* if not yet set, adopt default PPP binary path */
	if (!ppp_binary) ppp_binary = PPP_BINARY;
	/* check that the PPP binary is executable */
	if (access(ppp_binary, X_OK) < 0)
		bb_error_msg_and_die("MGR: PPP binary %s not executable", ppp_binary);
	/* check that the PPP options file is readable */
	if (pppdoptstr && access(pppdoptstr, R_OK) < 0)
		bb_error_msg_and_die("MGR: PPP options file %s not readable", pppdoptstr);

	if(!foreground)
		xdaemon(XDAEMON_PF_ARG | XDAEMON_PIDFILE, pid_file);

	/* manage connections until SIGTERM */
	pptp_manager();
#if KEEP_UNUSE
	slot_free();
#endif
	return 0;
}


/*
 * pptpctrl.c
 *
 * PPTP control connection between PAC-PNS pair
 *
 */

#define seq_greater(A,B) ((A)>(B) || \
			 (((u_int32_t)(A)<0xff) && ((~((u_int32_t)(B)))<0xff)))

/* Macro used in encaps_hdlc().  add "val" to "dest" at position "pos",
 * incrementing "pos" to point after the added value.  set "tmp" to "val"
 * as a side-effect.
 */
#define ADD_CHAR(pos, val, tmp) \
	tmp = (val); \
	if ((tmp<0x20) || (tmp==HDLC_FLAG) || (tmp==HDLC_ESCAPE)) { \
		dest[pos++]=HDLC_ESCAPE; \
		dest[pos++]=tmp^0x20; \
	} else \
		dest[pos++]=tmp

/* Fast Frame Check Sequence (FCS) Implementation, for HDLC-like framing of
 * PPP.  Adapted by C. Scott Ananian <cananian@alumni.princeton.edu>
 * from RFC1662:
 *
 * C.2.  16-bit FCS Computation Method
 *
 *  The following code provides a table lookup computation for
 *  calculating the Frame Check Sequence as data arrives at the
 *  interface.  This implementation is based on [7], [8], and [9].
 *
 *  [7]   Perez, "Byte-wise CRC Calculations", IEEE Micro, June 1983.
 *
 *  [8]   Morse, G., "Calculating CRC's by Bits and Bytes", Byte,
 *        September 1986.
 *
 *  [9]   LeVan, J., "A Fast CRC", Byte, November 1987.
 */

/*
 * FCS lookup table as calculated by the table generator.
 */
static const u_int16_t fcstab[256] =
{
	0x0000, 0x1189, 0x2312, 0x329b, 0x4624, 0x57ad, 0x6536, 0x74bf,
	0x8c48, 0x9dc1, 0xaf5a, 0xbed3, 0xca6c, 0xdbe5, 0xe97e, 0xf8f7,
	0x1081, 0x0108, 0x3393, 0x221a, 0x56a5, 0x472c, 0x75b7, 0x643e,
	0x9cc9, 0x8d40, 0xbfdb, 0xae52, 0xdaed, 0xcb64, 0xf9ff, 0xe876,
	0x2102, 0x308b, 0x0210, 0x1399, 0x6726, 0x76af, 0x4434, 0x55bd,
	0xad4a, 0xbcc3, 0x8e58, 0x9fd1, 0xeb6e, 0xfae7, 0xc87c, 0xd9f5,
	0x3183, 0x200a, 0x1291, 0x0318, 0x77a7, 0x662e, 0x54b5, 0x453c,
	0xbdcb, 0xac42, 0x9ed9, 0x8f50, 0xfbef, 0xea66, 0xd8fd, 0xc974,
	0x4204, 0x538d, 0x6116, 0x709f, 0x0420, 0x15a9, 0x2732, 0x36bb,
	0xce4c, 0xdfc5, 0xed5e, 0xfcd7, 0x8868, 0x99e1, 0xab7a, 0xbaf3,
	0x5285, 0x430c, 0x7197, 0x601e, 0x14a1, 0x0528, 0x37b3, 0x263a,
	0xdecd, 0xcf44, 0xfddf, 0xec56, 0x98e9, 0x8960, 0xbbfb, 0xaa72,
	0x6306, 0x728f, 0x4014, 0x519d, 0x2522, 0x34ab, 0x0630, 0x17b9,
	0xef4e, 0xfec7, 0xcc5c, 0xddd5, 0xa96a, 0xb8e3, 0x8a78, 0x9bf1,
	0x7387, 0x620e, 0x5095, 0x411c, 0x35a3, 0x242a, 0x16b1, 0x0738,
	0xffcf, 0xee46, 0xdcdd, 0xcd54, 0xb9eb, 0xa862, 0x9af9, 0x8b70,
	0x8408, 0x9581, 0xa71a, 0xb693, 0xc22c, 0xd3a5, 0xe13e, 0xf0b7,
	0x0840, 0x19c9, 0x2b52, 0x3adb, 0x4e64, 0x5fed, 0x6d76, 0x7cff,
	0x9489, 0x8500, 0xb79b, 0xa612, 0xd2ad, 0xc324, 0xf1bf, 0xe036,
	0x18c1, 0x0948, 0x3bd3, 0x2a5a, 0x5ee5, 0x4f6c, 0x7df7, 0x6c7e,
	0xa50a, 0xb483, 0x8618, 0x9791, 0xe32e, 0xf2a7, 0xc03c, 0xd1b5,
	0x2942, 0x38cb, 0x0a50, 0x1bd9, 0x6f66, 0x7eef, 0x4c74, 0x5dfd,
	0xb58b, 0xa402, 0x9699, 0x8710, 0xf3af, 0xe226, 0xd0bd, 0xc134,
	0x39c3, 0x284a, 0x1ad1, 0x0b58, 0x7fe7, 0x6e6e, 0x5cf5, 0x4d7c,
	0xc60c, 0xd785, 0xe51e, 0xf497, 0x8028, 0x91a1, 0xa33a, 0xb2b3,
	0x4a44, 0x5bcd, 0x6956, 0x78df, 0x0c60, 0x1de9, 0x2f72, 0x3efb,
	0xd68d, 0xc704, 0xf59f, 0xe416, 0x90a9, 0x8120, 0xb3bb, 0xa232,
	0x5ac5, 0x4b4c, 0x79d7, 0x685e, 0x1ce1, 0x0d68, 0x3ff3, 0x2e7a,
	0xe70e, 0xf687, 0xc41c, 0xd595, 0xa12a, 0xb0a3, 0x8238, 0x93b1,
	0x6b46, 0x7acf, 0x4854, 0x59dd, 0x2d62, 0x3ceb, 0x0e70, 0x1ff9,
	0xf78f, 0xe606, 0xd49d, 0xc514, 0xb1ab, 0xa022, 0x92b9, 0x8330,
	0x7bc7, 0x6a4e, 0x58d5, 0x495c, 0x3de3, 0x2c6a, 0x1ef1, 0x0f78
};


typedef int (*callback_t)(int cl, void *pack, unsigned int len);

/* test for a 32 bit counter overflow */
#define WRAPPED( curseq, lastseq) \
    ((((curseq) & 0xffffff00) == 0) && \
     (((lastseq) & 0xffffff00 ) == 0xffffff00))



/* Make stripped packet into HDLC packet */
static int encaps_hdlc(int fd, void *pack, unsigned len)
{
	unsigned char *source = (unsigned char *) pack;
	/* largest expansion possible - double all + double fcs + 2 flags */
	unsigned pos = 1, i;
	u_int16_t fcs;
	unsigned char c;

	fcs = PPPINITFCS16;

	/* make sure overflow is impossible so we don't have to bounds check
	 * in loop.  drop large packets.
	 */
	if (len > PACKET_MAX) {
		bb_error_msg("GRE: Asked to encapsulate too large packet (len = %d)", len);
		return -1;
	}
	/* start character */
	dest[0] = HDLC_FLAG;

	/* escape the payload */
	for (i = 0; i < len; i++) {
		ADD_CHAR(pos, source[i], c);
		fcs = (fcs >> 8) ^ fcstab[(fcs ^ c) & 0xff];
	}

	fcs ^= 0xFFFF;

	ADD_CHAR(pos, fcs & 0xFF, c);
	ADD_CHAR(pos, fcs >> 8, c);

	/* tack on the end-flag */
	dest[pos++] = HDLC_FLAG;

	/* now write this packet */
	return write(fd, dest, pos);
}

#undef ADD_CHAR


static int dequeue_gre (callback_t callback, int cl)
{
	pqueue_t *head;
	int status;
	/* process packets in the queue that either are expected or
	   have timed out. */
	head = pq_head;
	while ( head != NULL &&
		( (head->seq == gre.seq_recv + 1) ||
			/* wrap-around safe */
		  (pqueue_expiry_time(head) <= 0)
		  )
		) {
		/* if it is timed out... */
		if (head->seq != gre.seq_recv + 1 ) {  /* wrap-around safe */
			stats.rx_lost += head->seq - gre.seq_recv - 1;
			if (pptpctrl_debug)
				syslog(LOG_DEBUG,
				       "GRE: timeout waiting for %u packets",
				       head->seq - gre.seq_recv - 1);
		}
		if (pptpctrl_debug)
			syslog(LOG_DEBUG, "GRE: accepting #%u from queue",
			       head->seq);
		gre.seq_recv = head->seq;
		status = callback(cl, head->packet, head->packlen);
		pqueue_del(head);
		if (status < 0) return status;
		head = pq_head;
	}
	return 0;
}

static uint64_t time_now_usecs(void)
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return (tv.tv_sec * 1000000) + tv.tv_usec;
}

static int
decaps_gre(int fd, int (*cb) (int cl, void *pack, unsigned len), int cl)
{
	struct pptp_gre_header *header;
	int status, ip_len = 0;

	dequeue_gre(cb, cl);
	if ((status = read(fd, dg_buf, sizeof(dg_buf))) <= 0) {
		bb_error_msg("GRE: read(fd=%d,len=%d) from network failed: status = %d error = %s",
		       fd, (int)sizeof(dg_buf), status, status ? strerror(errno) : "No error");
		stats.rx_errors++;
		return -1;
	}
	/* strip off IP header, if present */
	if ((dg_buf[0] & 0xF0) == 0x40)
		ip_len = (dg_buf[0] & 0xF) * 4;
	header = (struct pptp_gre_header *) (dg_buf + ip_len);

	/* verify packet (else discard) */
	if (((header->ver & 0x7F) != PPTP_GRE_VER) ||    /* version should be 1   */
	    (ntohs(header->protocol) != PPTP_GRE_PROTO) ||     /* GRE protocol for PPTP */
	    PPTP_GRE_IS_C(header->flags) ||      /* flag C should be clear   */
	    PPTP_GRE_IS_R(header->flags) ||      /* flag R should be clear   */
	    (!PPTP_GRE_IS_K(header->flags)) ||   /* flag K should be set     */
	    ((header->flags & 0xF) != 0)) {      /* routing and recursion ctrl = 0  */
		/* if invalid, discard this packet */
		bb_error_msg("GRE: Discarding packet by header check");
		stats.rx_invalid++;
		return 0;
	}
	if (header->call_id != GET_VALUE(PAC, gre.call_id_pair)) {
		/*
		 * Discard silently to allow more than one GRE tunnel from
		 * the same IP address in case clients are behind the
		 * firewall.
		 *
		 * syslog(LOG_ERR, "GRE: Discarding for incorrect call");
		 */
		return 0;
	}
	if (PPTP_GRE_IS_A(header->ver)) {        /* acknowledgement present */
		u_int32_t ack = (PPTP_GRE_IS_S(header->flags)) ?
			ntohl(header->ack) : ntohl(header->seq);
			/* ack in different place if S=0 */

		if (seq_greater(ack, gre.ack_recv))
			gre.ack_recv = ack;

		/* also handle sequence number wrap-around  */
		if (WRAPPED(ack,gre.ack_recv)) gre.ack_recv = ack;
		if (gre.ack_recv == stats.pt.seq) {
			int rtt = time_now_usecs() - stats.pt.time;
			stats.rtt = (stats.rtt + rtt) / 2;
		}
	}
	if (PPTP_GRE_IS_S(header->flags)) {      /* payload present */
		unsigned headersize = sizeof(*header);
		unsigned payload_len = ntohs(header->payload_len);
		u_int32_t seq = ntohl(header->seq);

		if (!PPTP_GRE_IS_A(header->ver))
			headersize -= sizeof(header->ack);
		/* check for incomplete packet (length smaller than expected) */
		if (status - headersize < payload_len) {
			stats.rx_truncated++;
			return 0;
		}
		/* check for out-of-order sequence number
		 * N.B.: some client implementations violate RFC 2637
		 * and start their sequence numbers at 1 instead of 0,
		 * so we have to introduce a kludge to deal with it.
		 * on wrap we may allow an out of order packet to pass
		*/
		if (seq == gre.seq_recv + 1 || seq == 1) {
			if (pptpctrl_debug)
				syslog(LOG_DEBUG, "GRE: accepting packet #%u",
				       seq);
			stats.rx_accepted++;
			gre.seq_recv = seq;
			return cb(cl, dg_buf + ip_len + headersize, payload_len);
		} else if (!seq_greater(seq, gre.seq_recv)) {
			if (pptpctrl_debug)
				syslog(LOG_DEBUG,
				       "GRE: discarding duplicate or old packet #%d (expecting #%u)",
				       seq, gre.seq_recv + 1);
			return 0;       /* discard duplicate packets */
		} else {
			stats.rx_buffered++;
			if (pptpctrl_debug)
				syslog(LOG_DEBUG,
				       "GRE: buffering packet #%d (expecting #%d, lost or reordered)",
				       seq, gre.seq_recv + 1);
			pqueue_add(seq, dg_buf + ip_len + headersize, payload_len);
			return 0;       /* discard out-of-order packets */
		}
	}
	return 0;               /* ack, but no payload */
}


static int encaps_gre(int fd, void *pack, unsigned len)
{
	unsigned header_len;
	struct iovec iovec[2];

	if(fd == -1)
		/* peek mode */
		return (gre.ack_sent == gre.seq_recv) ? 0 : -1;

	/* package this up in a GRE shell. */
	u.header.flags = PPTP_GRE_FLAG_K;
	u.header.ver = PPTP_GRE_VER;
	u.header.protocol = htons(PPTP_GRE_PROTO);
	u.header.payload_len = htons(len);
	u.header.call_id = GET_VALUE(PNS, gre.call_id_pair);

	/* special case ACK with no payload */
	if (pack == NULL) {
		if (gre.ack_sent != gre.seq_recv) {
			u.header.ver |= PPTP_GRE_FLAG_A;
			u.header.payload_len = htons(0);
			u.header.seq = htonl(gre.seq_recv);    /* ack is in odd place because S=0 */
			gre.ack_sent = gre.seq_recv;
			/* don't sent ACK field, ACK is in SYN field */
			return write(fd, u.buffer, sizeof(u.header) - sizeof(u.header.ack));
		} else
			return 0;       /* we don't need to send ACK */
	}
	/* send packet with payload */
	u.header.flags |= PPTP_GRE_FLAG_S;
	u.header.seq = htonl(gre.seq_sent);
	gre.seq_sent++;
	if (gre.ack_sent != gre.seq_recv) {     /* send ack with this message */
		u.header.ver |= PPTP_GRE_FLAG_A;
		u.header.ack = htonl(gre.seq_recv);
		gre.ack_sent = gre.seq_recv;
		header_len = sizeof(u.header);
	} else {        /* don't send ack */
		header_len = sizeof(u.header) - sizeof(u.header.ack);
	}
	if (len > PACKET_MAX) {
		bb_error_msg("GRE: packet is too large %d", len);
		stats.tx_oversize++;
		return 0;       /* drop this, it's too big */
	}
	/* write header and buffer without copying. */
	iovec[0].iov_base = u.buffer;
	iovec[0].iov_len = header_len;
	iovec[1].iov_base = pack;
	iovec[1].iov_len = len;
	return writev(fd, iovec, 2);
}

#define ctrl_clientSocket 0 /* in case it changes back to a variable */

/*
 * freecall
 *
 * Notes termination of current call
 *
 * retn: -1 on failure, PAC call ID on success
 */


static u_int16_t freecall(void)
{
	u_int16_t ret;

	if(!_pac_init) {
		_pac_call_id = htons(-1);
		_pac_init = 1;
	}
	ret = _pac_call_id;
	if(_pac_call_id == htons(-1))
		bb_error_msg("CTRL: Asked to free call when no call open, not handled well");
	_pac_call_id = htons(-1);
	return ret;
}

/* Macros to use in making response packets */

#define MAKE_CTRL_HEADER(where, what) \
	where.header.length = htons(sizeof(where)); \
	where.header.pptp_type = htons(PPTP_CTRL_MESSAGE); \
	where.header.magic = htonl(PPTP_MAGIC_COOKIE); \
	where.header.ctrl_type = htons(what); \
	where.header.reserved0 = htons(RESERVED)

#define COPY_CTRL_PACKET(from, to, size) \
	memcpy(to, &from, ((*size) = sizeof(from)))

#define DEBUG_PACKET(what) \
	if(pptpctrl_debug) \
		syslog(LOG_DEBUG, "CTRL: Made a " what " packet")



static void
make_call_admin_shutdown(unsigned char *rply_packet, ssize_t * rply_size)
{
	struct pptp_call_disconn_ntfy call_disconn_ntfy;
	u_int16_t pac_call_id;

	/* Form a reply
	 * The reply packet is a CALL-DISCONECT-NOTIFY
	 * In single call mode we don't care what peer's call ID is, so don't even bother looking
	 */
	if ((pac_call_id = freecall()) == htons(-1)) {
		/* XXX should return an error */
		bb_error_msg("CTRL: Could not free Call ID [admin shutdown]!");
	}
	MAKE_CTRL_HEADER(call_disconn_ntfy, CALL_DISCONN_NTFY);
	call_disconn_ntfy.call_id = pac_call_id;
	call_disconn_ntfy.result_code = ADMIN_SHUTDOWN;         /* disconnected by admin shutdown */
	call_disconn_ntfy.error_code = NO_ERROR;
	call_disconn_ntfy.cause_code = htons(NO_ERROR);
	call_disconn_ntfy.reserved1 = htons(RESERVED);
	memset(call_disconn_ntfy.call_stats, 0, 128);
	COPY_CTRL_PACKET(call_disconn_ntfy, rply_packet, rply_size);
	DEBUG_PACKET("CALL DISCONNECT RPLY");
}


/*
 * send_pptp_packet
 *
 * Sends a PPTP packet to a file descriptor.
 *
 * args:        clientFd (IN) - file descriptor to write the packet to.
 *              packet (IN) - the packet data to write.
 *              packet_size (IN) - the packet size.
 *
 * retn:        Number of bytes written on success.
 *              -1 on write failure.
 */
static ssize_t
send_pptp_packet(int clientFd, unsigned char *packet, size_t packet_size)
{
	ssize_t bytes_written;

	bytes_written = write(clientFd, packet, packet_size);
	if (bytes_written == -1) {
		/* write failed */
		bb_perror_msg("CTRL: Couldn't write packet to client");
	} else {
		/* debugging */
		if (pptpctrl_debug) {
			syslog(LOG_DEBUG, "CTRL: I wrote %d bytes to the client.", (int)packet_size);
			syslog(LOG_DEBUG, "CTRL: Sent packet to client");
		}
	}
	return bytes_written;
}

static void
make_stop_ctrl_req(unsigned char *rply_packet, ssize_t * rply_size)
{
	struct pptp_stop_ctrl_conn_rqst stop_ctrl;

	MAKE_CTRL_HEADER(stop_ctrl, STOP_CTRL_CONN_RQST);
	stop_ctrl.reason = GENERAL_STOP_CTRL;
	stop_ctrl.reserved1 = RESERVED;
	stop_ctrl.reserved2 = htons(RESERVED);
	COPY_CTRL_PACKET(stop_ctrl, rply_packet, rply_size);
	DEBUG_PACKET("STOP CTRL REQ");
}

/*
 * ignoreErrno
 *
 * Check if an errno represents a read error which should be ignored, and
 * put back to be select()ed on again later.
 *
 * Very similar to the function in Squid by Duane Wessels (under GPL).
 *
 * args: an errno value
 *
 * retn: 1 if the error is unimportant
 *       0 if the error is important
 */
static int ignoreErrno(int ierrno)
{
	switch (ierrno) {
	case EAGAIN:            /* nothing to read */
	case EINTR:             /* signal received */
#ifdef ERESTART
#if ERESTART != EINTR
	case ERESTART:          /* signal received, should restart syscall */
#endif
#endif
#if EWOULDBLOCK != EAGAIN
	case EWOULDBLOCK:       /* shouldn't get this one but anyway, just in case */
#endif
		return 1;
	default:
		return 0;
	}
}

/*
 * read_pptp_header
 *
 * Reads a packet from a file descriptor and determines whether it is a
 * valid PPTP Control Message. If a valid PPTP Control Message is detected
 * it extracts the Control Message type from the packet header.
 *
 * args:        clientFd (IN) - Clients file descriptor.
 *              packet (OUT) - Packet we read from the client.
 *              pptp_ctrl_type (OUT) - PPTP Control Message type of the packet.
 *
 * retn:        Number of bytes read on success.
 *              -1 on retryable error.
 *              0 on error to exit on.
 */
static ssize_t
read_pptp_header(int clientFd, unsigned char *packet, int *pptp_ctrl_type)
{
	ssize_t bytes_ttl, bytes_this;  /* quantities read (total and this read) */
	u_int16_t length;               /* length of this packet */
	struct pptp_header *header;     /* the received header */


	*pptp_ctrl_type = 0;            /* initialise return arg        */

	/* read any previously rh_bufed data */
	if (rh_bufed) {
		memcpy(packet, rh_buf, rh_bufed);
		free(rh_buf);
		rh_buf = NULL;
		bytes_ttl = rh_bufed;
		rh_bufed = 0;
		if (pptpctrl_debug)
			syslog(LOG_DEBUG, "CTRL: Read in previous incomplete ctrl packet");
	} else
		bytes_ttl = 0;

	/* try and get the length in */
	if (bytes_ttl < 2) {
		bytes_this = read(clientFd, packet + bytes_ttl, 2 - bytes_ttl);
		switch (bytes_this) {
		case -1:
			if (ignoreErrno(errno)) {
				/* re-tryable error, re-buffer and return */
				if (bytes_ttl) {
					rh_bufed = bytes_ttl;
					rh_buf = malloc(bytes_ttl);
					if (!rh_buf)
						return(0);
					memcpy(rh_buf, packet, bytes_ttl);
				}
				bb_error_msg("CTRL: Error reading ctrl packet length (bytes_ttl=%d): %m", (int)bytes_ttl);
				return -1;
			}
			/* FALLTHRU */
		case 0:
			bb_error_msg("CTRL: EOF or bad error reading ctrl packet length.");
			return 0;
		default:
			bytes_ttl += bytes_this;
			/* Not enough data to proceed */
			if (bytes_ttl == 1) {
				rh_bufed = bytes_ttl;
				rh_buf = malloc(bytes_ttl);
				if (!rh_buf)
					return(0);
				memcpy(rh_buf, packet, bytes_ttl);
				if (pptpctrl_debug)
					syslog(LOG_DEBUG, "CTRL: Incomplete ctrl packet length, retry later");
				return -1;
			}
		}
	}
	/* OK, we have (at least) the first 2 bytes, and there is data waiting
	 *
	 * length includes the header,  so a length less than 2 is someone
	 * trying to hack into us or a badly corrupted packet.
	 * Why not require length to be at least 10? Since we later cast
	 * packet to struct pptp_header and use at least the 10 first bytes..
	 * Thanks to Timo Sirainen for mentioning this.
	 */
	length = htons(*(u_int16_t *) packet);
	if (length <= 10 || length > PPTP_MAX_CTRL_PCKT_SIZE) {
		bb_error_msg("CTRL: 11 < Control packet (length=%d) < "
				"PPTP_MAX_CTRL_PCKT_SIZE (%d)",
				length, PPTP_MAX_CTRL_PCKT_SIZE);
		/* we loose sync (unless we malloc something big, which isn't a good
		 * idea - potential DoS) so we must close connection (draft states that
		 * if you loose sync you must close the control connection immediately)
		 */
		return 0;
	}
	/* Now read the actual control packet */
	bytes_this = read(clientFd, packet + bytes_ttl, length - bytes_ttl);
	switch (bytes_this) {
	case -1:
		if(ignoreErrno(errno)) {
			/* re-tryable error, re-buffer and return */
			if (bytes_ttl) {
				rh_bufed = bytes_ttl;
				rh_buf = malloc(bytes_ttl);
				if (!rh_buf)
					return(0);
				memcpy(rh_buf, packet, bytes_ttl);
			}
			bb_error_msg("CTRL: Error reading ctrl packet (bytes_ttl=%d,length=%d): %m", (int)bytes_ttl, length);
			return -1;
		}
		/* FALLTHRU */
	case 0:
		bb_error_msg("CTRL: EOF or bad error reading ctrl packet.");
		return 0;
	default:
		bytes_ttl += bytes_this;
		/* not enough data to proceed */
		if (bytes_ttl != length) {
			rh_bufed = bytes_ttl;
			rh_buf = malloc(bytes_ttl);
			if (!rh_buf)
				return(0);
			memcpy(rh_buf, packet, bytes_ttl);
			if (pptpctrl_debug)
				syslog(LOG_DEBUG, "CTRL: Incomplete ctrl packet, retry later");
			return -1;
		}
	}

	/* We got one :-) */

	/* Cast the packet into the PPTP Control Message format */
	header = (struct pptp_header *) packet;

	/* Packet sanity check on magic cookie */
	if (ntohl(header->magic) != PPTP_MAGIC_COOKIE) {
		/* Oops! Not a valid Control Message */
		bb_error_msg("CTRL: Bad magic cookie - lost syncronization, closing control connection.");
		/* draft states loss of syncronization must result in
		 * immediate closing of the control connection
		 */
		return 0;
	}
	/* Woohoo! Looks like we got a valid PPTP packet */
	*pptp_ctrl_type = (int) (ntohs(header->ctrl_type));
	if (pptpctrl_debug)
		syslog(LOG_DEBUG, "CTRL: Received PPTP Control Message (type: %d)", *pptp_ctrl_type);
	return bytes_ttl;
}

/*
 * deal_start_ctrl_conn
 *
 * This method 'deals' with a START-CONTROL-CONNECTION-REQUEST. After
 * stripping down the connection request a suitable reply is formed and
 * stored in 'rply_packet' ready for sending.
 *
 * args: packet (IN) - the packet that we have to deal with (should be a
 *                     START-CONTROL-CONNECTION-REQUEST packet)
 *       rply_packet (OUT) - suitable reply to the 'packet' we got.
 *       rply_size (OUT) - size of the reply packet
 */
static void
deal_start_ctrl_conn(unsigned char *packet, unsigned char *rply_packet, ssize_t * rply_size)
{
	struct pptp_start_ctrl_conn_rply start_ctrl_conn_rply;

	MAKE_CTRL_HEADER(start_ctrl_conn_rply, START_CTRL_CONN_RPLY);
	start_ctrl_conn_rply.version = htons(PPTP_VERSION);
	start_ctrl_conn_rply.result_code = CONNECTED;
	start_ctrl_conn_rply.error_code = NO_ERROR;
	start_ctrl_conn_rply.framing_cap = htons(OUR_FRAMING);
	start_ctrl_conn_rply.bearer_cap = htons(OUR_BEARER);
	start_ctrl_conn_rply.max_channels = htons(MAX_CHANNELS);
	start_ctrl_conn_rply.firmware_rev = htons(PPTP_FIRMWARE_VERSION);
	memset(start_ctrl_conn_rply.hostname, 0, MAX_HOSTNAME_SIZE);
	strncpy((char *)start_ctrl_conn_rply.hostname, PPTP_HOSTNAME, MAX_HOSTNAME_SIZE);
	memset(start_ctrl_conn_rply.vendor, 0, MAX_VENDOR_SIZE);
	strncpy((char *)start_ctrl_conn_rply.vendor, PPTP_VENDOR, MAX_VENDOR_SIZE);
	COPY_CTRL_PACKET(start_ctrl_conn_rply, rply_packet, rply_size);
	DEBUG_PACKET("START CTRL CONN RPLY");
}


/*
 * deal_stop_ctrl_conn
 *
 * This method response to a STOP-CONTROL-CONNECTION-REQUEST with a
 * STOP-CONTROL-CONNECTION-REPLY.
 */
static void
deal_stop_ctrl_conn(unsigned char *packet, unsigned char *rply_packet, ssize_t * rply_size)
{
	struct pptp_stop_ctrl_conn_rply stop_ctrl_conn_rply;

	MAKE_CTRL_HEADER(stop_ctrl_conn_rply, STOP_CTRL_CONN_RPLY);
	stop_ctrl_conn_rply.result_code = DISCONNECTED;
	stop_ctrl_conn_rply.error_code = NO_ERROR;
	stop_ctrl_conn_rply.reserved1 = htons(RESERVED);
	COPY_CTRL_PACKET(stop_ctrl_conn_rply, rply_packet, rply_size);
	DEBUG_PACKET("STOP CTRL CONN RPLY");
}

/*
 * getcall
 *
 * Assigns a call ID to the session and stores/returns it
 *
 * we only permit one call at a time, so the chance of wrapping 65k on one
 * control connection is zero to none...
 */
static u_int16_t getcall(void)
{

	/* Start with a random Call ID.  This is to allocate unique
	 * Call ID's across multiple TCP PPTP connections.  In this
	 * way remote clients masqueraded by a firewall will put
	 * unique peer call ID's into GRE packets that will have the
	 * same source IP address of the firewall. */

	if (!gci) {
		if (unique_call_id == 0xFFFF) {
			struct timeval tv;
			if (gettimeofday(&tv, NULL) == 0) {
				gci = ((tv.tv_sec & 0x0FFF) << 4) +
				    (tv.tv_usec >> 16);
			}
		} else {
			gci = unique_call_id;
		}
	}

	if(!_pac_init) {
		_pac_call_id = htons(-1);
		_pac_init = 1;
	}
	if(_pac_call_id != htons(-1))
		bb_error_msg("CTRL: Asked to allocate call id when call open, not handled well");
	_pac_call_id = htons(gci);
	gci++;
	return _pac_call_id;
}

/*
 * deal_out_call
 *
 * This method 'deals' with a OUT-GOING-CALL-REQUEST. After
 * stripping down the request a suitable reply is formed and stored in
 * 'rply_packet' ready for sending.
 *
 * args: packet (IN) - the packet that we have to deal with (should be a
 *                      OUT-GOING-CALL-REQUEST packet)
 *       rply_packet (OUT) - suitable reply to the 'packet' we got.
 *       rply_size (OUT) - size of the reply packet
 *
 */
static void
deal_out_call(unsigned char *packet, unsigned char *rply_packet, ssize_t * rply_size)
{
	u_int16_t pac_call_id;
	struct pptp_out_call_rqst *out_call_rqst;
	struct pptp_out_call_rply out_call_rply;

	out_call_rqst = (struct pptp_out_call_rqst *) packet;

	if ((pac_call_id = getcall()) == htons(-1)) {
		/* XXX should reject call */
		bb_error_msg("CTRL: No free Call IDs!");
		pac_call_id = 0;
	}
	MAKE_CTRL_HEADER(out_call_rply, OUT_CALL_RPLY);
	/* call_id is used for ctrl, call_id_peer is used for GRE
	 * call_id_peer is what we were sent by the other end in ctrl initilization
	 */
	out_call_rply.call_id = pac_call_id;
	out_call_rply.call_id_peer = out_call_rqst->call_id;
	out_call_rply.result_code = CONNECTED;
	out_call_rply.error_code = NO_ERROR;
	out_call_rply.cause_code = NO_ERROR;
	/* maybe limit to pppd speed? but pppd doesn't accept 10Mbps as a speed and yet
	 * still performs at over 115200, eg, 60kbyte/sec and higher observed.
	 */
	out_call_rply.speed = out_call_rqst->max_bps;
	/* lets match their window size for now... was htons(PCKT_RECV_WINDOW_SIZE)
	 */
	out_call_rply.pckt_recv_size = out_call_rqst->pckt_recv_size;
	if(pptpctrl_debug)
		syslog(LOG_DEBUG, "CTRL: Set parameters to %d maxbps, %d window size",
			ntohl(out_call_rply.speed), ntohs(out_call_rply.pckt_recv_size));
	out_call_rply.pckt_delay = htons(PCKT_PROCESS_DELAY);
	out_call_rply.channel_id = htonl(CHANNEL_ID);
	COPY_CTRL_PACKET(out_call_rply, rply_packet, rply_size);
	DEBUG_PACKET("OUT CALL RPLY");
}

/*
 * deal_echo
 *
 * This method 'deals' with a ECHO-REQUEST. After stripping down the
 * connection request a suitable reply is formed and stored in
 * 'rply_packet' ready for sending.
 *
 * args: packet (IN) - the packet that we have to deal with (should be a
 *                      ECHO-REQUEST packet)
 *       rply_packet (OUT) - suitable reply to the 'packet' we got.
 *       rply_size (OUT) - size of the reply packet
 *
 */
static void
deal_echo(unsigned char *packet, unsigned char *rply_packet, ssize_t * rply_size)
{
	struct pptp_echo_rqst *echo_rqst;
	struct pptp_echo_rply echo_rply;

	echo_rqst = (struct pptp_echo_rqst *) packet;

	MAKE_CTRL_HEADER(echo_rply, ECHO_RPLY);
	echo_rply.identifier = echo_rqst->identifier;
	echo_rply.result_code = CONNECTED;
	echo_rply.error_code = NO_ERROR;
	echo_rply.reserved1 = htons(RESERVED);
	COPY_CTRL_PACKET(echo_rply, rply_packet, rply_size);
	DEBUG_PACKET("ECHO RPLY");
}

/*
 * deal_call_clr
 *
 * This method 'deals' with a CALL-CLEAR-REQUEST. After stripping down the
 * connection request a suitable reply is formed and stored in
 * 'rply_packet' ready for sending.
 *
 * args: packet (IN) - the packet that we have to deal with (should be a
 *                      CALL-CLEAR-REQUEST packet)
 *       rply_packet (OUT) - suitable reply to the 'packet' we got.
 *       rply_size (OUT) - size of the reply packet
 *
 */
static void
deal_call_clr(unsigned char *packet, unsigned char *rply_packet, ssize_t *rply_size)
{
	struct pptp_call_disconn_ntfy call_disconn_ntfy;
	u_int16_t pac_call_id;

	/* Form a reply
	 * The reply packet is a CALL-DISCONECT-NOTIFY
	 * In single call mode we don't care what peer's call ID is, so don't even bother looking
	 */
	if ((pac_call_id = freecall()) == htons(-1)) {
		/* XXX should return an error */
		bb_error_msg("CTRL: Could not free Call ID [call clear]!");
	}
	MAKE_CTRL_HEADER(call_disconn_ntfy, CALL_DISCONN_NTFY);
	call_disconn_ntfy.call_id = pac_call_id;
	call_disconn_ntfy.result_code = CALL_CLEAR_REQUEST;     /* disconnected by call_clr_rqst */
	call_disconn_ntfy.error_code = NO_ERROR;
	call_disconn_ntfy.cause_code = htons(NO_ERROR);
	call_disconn_ntfy.reserved1 = htons(RESERVED);
	memset(call_disconn_ntfy.call_stats, 0, 128);
	COPY_CTRL_PACKET(call_disconn_ntfy, rply_packet, rply_size);
	DEBUG_PACKET("CALL DISCONNECT RPLY");
}

/*
 * deal_set_link_info
 *
 * @FIXME This function is *not* completed
 *
 * This method 'deals' with a SET-LINK-INFO. After stripping down the
 * connection request a suitable reply is formed and stored in
 * 'rply_packet' ready for sending.
 *
 * args: packet (IN) - the packet that we have to deal with (should be a
 *                      SET-LINK-INFO packet)
 *       rply_packet (OUT) - suitable reply to the 'packet' we got.
 *       rply_size (OUT) - size of the reply packet
 *
 */
static void deal_set_link_info(unsigned char *packet)
{
	struct pptp_set_link_info *set_link_info;

	set_link_info = (struct pptp_set_link_info *) packet;
	if(set_link_info->send_accm != 0xffffffff || set_link_info->recv_accm != 0xffffffff)
		bb_error_msg("CTRL: Ignored a SET LINK INFO packet with real ACCMs!");
	else if(pptpctrl_debug)
		syslog(LOG_DEBUG, "CTRL: Got a SET LINK INFO packet with standard ACCMs");
}


/*
 * read_pptp_packet
 *
 * Sees if a packet can be read and if so what type of packet it is. The
 * method then calls the appropriate function to examine the details of the
 * packet and form a suitable reply packet.
 *
 * args:        clientFd (IN) - Client socket to read from.
 *              packet (OUT) - Packet read from the client.
 *              rply_packet (OUT) - Reply packet for the client.
 *              rply_size (OUT) - Size of the reply packet.
 *
 * retn:        PPTP control message type of the packet on success.
 *              -1 on retryable error.
 *              0 on error to abort on.
 */
static int
read_pptp_packet(int clientFd, unsigned char *packet, unsigned char *rply_packet, ssize_t * rply_size)
{

	size_t bytes_read;
	int pptp_ctrl_type = 0; /* Control Message Type */

	/* read a packet and parse header */
	if ((bytes_read = read_pptp_header(clientFd, packet, &pptp_ctrl_type)) <= 0) {
		/* error reading packet */
		bb_error_msg("CTRL: couldn't read packet header (%s)", bytes_read ? "retry" : "exit");
		return bytes_read;
	}

	/* launch appropriate method to form suitable reply to the packet */
	switch (pptp_ctrl_type) {
	case START_CTRL_CONN_RQST:      /* Start Control Connection Request */
		deal_start_ctrl_conn(packet, rply_packet, rply_size);
		break;

	case STOP_CTRL_CONN_RQST:
		deal_stop_ctrl_conn(packet, rply_packet, rply_size);
		break;

	case OUT_CALL_RQST:             /* Outgoing Call Request */
		deal_out_call(packet, rply_packet, rply_size);
		break;

	case ECHO_RQST:                 /* Echo Request */
		deal_echo(packet, rply_packet, rply_size);
		break;

	case CALL_CLR_RQST:             /* Call Clear Request (Disconnect Request) */
		deal_call_clr(packet, rply_packet, rply_size);
		break;

	case SET_LINK_INFO:             /* Set Link Info */
		/* no reply packet but process it */
		deal_set_link_info(packet);
		break;

	case ECHO_RPLY:                 /* Echo Reply */
	case STOP_CTRL_CONN_RPLY:       /* Stop Control Connection Reply */
	case CALL_DISCONN_NTFY:         /* Call Disconnect Notify */
		/* no reply packet */
		break;

	default:
		bb_error_msg("CTRL: PPTP Control Message type %d not supported.", pptp_ctrl_type);
		pptp_ctrl_type = -1;
	}

	return pptp_ctrl_type;
}

/*
 * This is the custom exit() for this program.
 *
 * Updated to also be the default SIGTERM handler, and if
 * the link is going down for unnatural reasons, we will close it
 * right now, it's only been tested for win98, other tests would be nice
 * -tmk
 */
static void bail(int sigraised) ATTRIBUTE_NORETURN;

static void bail(int sigraised)
{
	if (sigraised)
		syslog(LOG_INFO, "CTRL: Exiting on signal %d", sigraised);

	/* send a disconnect to the other end */
	/* ignore any errors */
	if (GET_VALUE(PAC, Gcall_id_pair) != htons(-1)) {
		fd_set connSet;         /* fd_set for select() */
		struct timeval tv;      /* time to wait for reply */
		unsigned char packet[PPTP_MAX_CTRL_PCKT_SIZE];
		unsigned char rply_packet[PPTP_MAX_CTRL_PCKT_SIZE];
		ssize_t rply_size;      /* reply packet size */
		int pkt;
		int retry = 0;

		if (pptpctrl_debug)
			syslog(LOG_DEBUG, "CTRL: Exiting with active call");

		make_call_admin_shutdown(rply_packet, &rply_size);
		if(send_pptp_packet(ctrl_clientSocket, rply_packet, rply_size) < 0)
			goto skip;

		make_stop_ctrl_req(rply_packet, &rply_size);
		if(send_pptp_packet(ctrl_clientSocket, rply_packet, rply_size) < 0)
			goto skip;

		FD_ZERO(&connSet);
		FD_SET(ctrl_clientSocket, &connSet);
		tv.tv_sec = 5;  /* wait 5 secs for a reply then quit */
		tv.tv_usec = 0;

		/* Wait for STOP CTRL CONN RQST or RPLY */
		while (select(ctrl_clientSocket + 1, &connSet, NULL, NULL, &tv) == 1) {
			switch((pkt = read_pptp_packet(ctrl_clientSocket, packet, rply_packet, &rply_size))) {
			case STOP_CTRL_CONN_RQST:
				send_pptp_packet(ctrl_clientSocket, rply_packet, rply_size);
				goto skip;
			case CALL_CLR_RQST:
				xlog_warn("CTRL: Got call clear request after call manually shutdown - buggy client");
				break;
			case STOP_CTRL_CONN_RPLY:
				goto skip;
			case -1:
				xlog_warn("CTRL: Retryable error in disconnect sequence");
				retry++;
				break;
			case 0:
				xlog_warn("CTRL: Fatal error reading control message in disconnect sequence");
				goto skip;
			default:
				xlog_warn("CTRL: Unexpected control message %d in disconnect sequence", pkt);
				retry++;
				break;
			}
			tv.tv_sec = 5;  /* wait 5 secs for another reply then quit */
			tv.tv_usec = 0;
			if (retry > 100) {
				xlog_warn("CTRL: Too many retries (%d) - giving up", retry);
				break;
			}
		}

	skip:
		close(ctrl_clientSocket);
	}

	if (pptpctrl_debug)
		syslog(LOG_DEBUG, "CTRL: Exiting now");
	exit(1);
}

/*
 * launch_pppd
 *
 * Launches the PPP daemon. The PPP daemon is responsible for assigning the
 * PPTP client its IP address.. These values are assigned via the command
 * line.
 *
 * Add return of connected ppp interface
 *
 * retn: 0 on success, -1 on failure.
 *
 */
static void launch_pppd(char **pppaddrs, struct in_addr *inetaddrs)
{
	char *pppd_argv[14];
	int an = 0;
	sigset_t sigs;

	pppd_argv[an++] = ppp_binary;

	if (pptpctrl_debug) {
		syslog(LOG_DEBUG,
		       "CTRL (PPPD Launcher): program binary = %s",
		       pppd_argv[an - 1]);
	}

	/* options for pppd */

	pppd_argv[an++] = "local";

	/* If a pppd option file is specified, use it
	 * if not, pppd will default to /etc/ppp/options
	 */
	if (*pppdxfig) {
		pppd_argv[an++] = "file";
		pppd_argv[an++] = pppdxfig;
	}

	/* If a speed has been specified, use it
	 * if not, use "smart" default
	 */
	if (*Gspeed) {
		pppd_argv[an++] = Gspeed;
	} else {
		pppd_argv[an++] = PPP_SPEED_DEFAULT;
	}

	if (pptpctrl_debug) {
		if (*pppaddrs[0])
			syslog(LOG_DEBUG, "CTRL (PPPD Launcher): local address = %s", pppaddrs[0]);
		if (*pppaddrs[1])
			syslog(LOG_DEBUG, "CTRL (PPPD Launcher): remote address = %s", pppaddrs[1]);
	}

	if (*pppaddrs[0] || *pppaddrs[1]) {
		char pppInterfaceIPs[33];
		sprintf(pppInterfaceIPs, "%s:%s", pppaddrs[0], pppaddrs[1]);
		pppd_argv[an++] = pppInterfaceIPs;
	}

	if (!noipparam) {
		 pppd_argv[an++] = "ipparam";
		 pppd_argv[an++] = inet_ntoa(inetaddrs[1]);
	}

	if (pptp_logwtmp) {
		 pppd_argv[an++] = "plugin";
		 pppd_argv[an++] = "/usr/lib/pptpd/pptpd-logwtmp.so";
		 pppd_argv[an++] = "pptpd-original-ip";
		 pppd_argv[an++] = inet_ntoa(inetaddrs[1]);
	}

	/* argv arrays must always be NULL terminated */
	pppd_argv[an++] = NULL;
	/* make sure SIGCHLD is unblocked, pppd does not expect it */
	sigfillset(&sigs);
	sigprocmask(SIG_UNBLOCK, &sigs, NULL);
	/* run pppd now */
	execvp(pppd_argv[0], pppd_argv);
	/* execvp() failed */
	bb_perror_msg("CTRL (PPPD Launcher): Failed to launch PPP daemon");
}


/*
 * startCall
 *
 * Launches PPPD for the call.
 *
 * args:        pppaddrs - local/remote IPs or "" for either/both if none
 * retn:        pty file descriptor
 *
 */
static int startCall(char **pppaddrs, struct in_addr *inetaddrs)
{
	/* PTY/TTY pair for talking to PPPd */
	int pty_fd, tty_fd;
	/* register pids of children */
	/* Finds an open pty/tty pair */
	if (openpty(&pty_fd, &tty_fd, NULL, NULL, NULL) != 0) {
		bb_perror_msg_and_die("CTRL: openpty() error");
	} else {
		struct termios tios;

		/* Turn off echo in the slave - to prevent loopback.
		   pppd will do this, but might not do it before we
		   try to send data. */
		if (tcgetattr(tty_fd, &tios) < 0)
			bb_perror_msg_and_die("CTRL: tcgetattr() error");
		tios.c_lflag &= ~(ECHO | ECHONL);
		if (tcsetattr(tty_fd, TCSAFLUSH, &tios) < 0)
			bb_perror_msg_and_die("CTRL: tcsetattr() error");
	}
	if (pptpctrl_debug) {
		syslog(LOG_DEBUG, "CTRL: pty_fd = %d", pty_fd);
		syslog(LOG_DEBUG, "CTRL: tty_fd = %d", tty_fd);
	}
	/* Launch the PPPD  */
#ifndef HAVE_FORK
	switch(pppfork=vfork())
#else
	switch(pppfork=fork())
#endif
	{
	case -1:        /* fork() error */
		bb_perror_msg_and_die("CTRL: Error forking to exec pppd");

	case 0:         /* child */
		if (dup2(tty_fd, 0) == -1)
		  bb_perror_msg_and_die("CTRL: child tty_fd dup2 to stdin");
		if (dup2(tty_fd, 1) == -1)
		  bb_perror_msg_and_die("CTRL: child tty_fd dup2 to stdout");
		if (tty_fd > 1)
			close(tty_fd);
		if (pty_fd > 1)
			close(pty_fd);
		launch_pppd(pppaddrs, inetaddrs);
		bb_perror_msg_and_die("CTRL: PPPD launch failed! (launch_pppd did not fork)");
	}

	close(tty_fd);
	return pty_fd;
}


/* ONE blocking read per call; dispatches all packets possible */
/* returns 0 on success, or <0 on read failure                 */
static int
decaps_hdlc(int fd, int (*cb) (int cl, void *pack, unsigned len), int cl)
{
	unsigned char c;
	int status;

	/* we do one read only, since it may block.  and only if the
	 * dh_buf is empty (dh_start == dh_end)
	 */

	if (fd == -1) {
		if(cb == NULL) {
			/* peek mode */
			return err ? -1 : 0;
		} else if (!err) {
			/* re-xmit and nothing queued */
			bb_error_msg("GRE: Re-xmit called with nothing queued");
			return -1;
		}
	}

	if (!err) {
		/* All known data is processed.  This true unless the last
		 * network write failed.
		 */
		if ((status = read(fd, dh_buf, sizeof(dh_buf))) <= 0) {
			bb_error_msg("GRE: read(fd=%d,len=%d) from PTY failed: status = %d error = %s%s",
			       fd, (int)sizeof(dh_buf),
			       status, status ? strerror(errno) : "No error",
			       errno != EIO ? "" : ", usually caused by unexpected termination of pppd, check option syntax and pppd logs");
			/* FAQ: mistakes in pppd option spelling in
			 * /etc/ppp/options.pptpd often cause EIO,
			 * with pppd not reporting the problem to any
			 * logs.  Termination of pppd by signal can
			 * *also* cause this situation. -- James Cameron
			 */
			return -1;
		}
		dh_end = status;
		dh_start = 0;
	} else {
		/* We're here because of a network write failure.  Try again.
		 * Then do what we would do normally and enter the loop as if
		 * just continuing the while(1).  Not sure that this ever
		 * really happens, but since we error-check status then we
		 * should have the code to handle an error :-)
		 */
		err = 0;
		if ((status = cb(cl, dh_buf_copy, dh_len)) < 0) {
			bb_perror_msg("GRE: re-xmit failed from decaps_hdlc");
			err = 1;
			return status;  /* return error */
		}
		/* Great!  Let's do more! */
		decaps_hdlc_fcs = PPPINITFCS16;
		dh_len = 0;
		escape = 0;
	}

	while (1) {
		/* Infinite loop, we return when we're out of data */

		/* Check if out of data */
		if (dh_start == dh_end)
			return 0;

		/* Add to the packet up till the next HDLC_FLAG (dh_start/dh_end of
		 * packet marker).  Copy to 'dh_buf_copy', un-escape and checksum as we go.
		 */
		while (dh_buf[dh_start] != HDLC_FLAG) {

			/* Dispose of 'too long' packets */
			if (dh_len >= PACKET_MAX) {
				bb_error_msg("GRE: Received too long packet from pppd.");
				while (dh_buf[dh_start] != HDLC_FLAG && dh_start < dh_end)
					dh_start++;
				if (dh_start < dh_end) {
					goto newpacket;
				} else
					return 0;
			}
			/* Read a character, un-escaping if needed */
			if (dh_buf[dh_start] == HDLC_ESCAPE && !escape)
				escape = 1;
			else {
				if (escape) {
					dh_buf_copy[dh_len] = c = dh_buf[dh_start] ^ 0x20;
					escape = 0;
				} else
					dh_buf_copy[dh_len] = c = dh_buf[dh_start];
				decaps_hdlc_fcs = (decaps_hdlc_fcs >> 8) ^ fcstab[(decaps_hdlc_fcs ^ c) & 0xff];
				dh_len++;
			}
			dh_start++;

			/* Check if out of data */
			if (dh_start == dh_end)
				return 0;
		}

		/* Found flag.  Skip past it */
		dh_start++;

		/* Check for over-short packets and silently discard, as per RFC1662 */
		if ((dh_len < 4) || (escape == 1)) {
			/* dh_len == 0 is possible, we generate it :-) [using HDLC_ESCAPE at
			 * dh_start and dh_end of packet].  Others are worth recording.
			 */
			if (dh_len && dh_len < 4)
				bb_error_msg("GRE: Received too short packet from pppd.");
			if (escape)
				bb_error_msg("GRE: Received bad packet from pppd.");
			goto newpacket;
		}
		/* Check, then remove the 16-bit FCS checksum field */
		if (decaps_hdlc_fcs != PPPGOODFCS16) {
			bb_error_msg("GRE: Bad checksum from pppd.");
			goto newpacket;
		}
		dh_len -= sizeof(u_int16_t);

		/* So now we have a packet of length 'dh_len' in 'dh_buf_copy' */
		if ((status = cb(cl, dh_buf_copy, dh_len)) < 0) {
			bb_perror_msg("GRE: xmit failed from decaps_hdlc");
			err = 1;
			return status;  /* return error */
		}
	      newpacket:
		/* Great!  Let's do more! */
		decaps_hdlc_fcs = PPPINITFCS16;
		dh_len = 0;
		escape = 0;
	}
}


static void
make_echo_req_packet(unsigned char *rply_packet, ssize_t * rply_size, u_int32_t echo_id)
{
	struct pptp_echo_rqst echo_packet;

	MAKE_CTRL_HEADER(echo_packet, ECHO_RQST);
	echo_packet.identifier = echo_id;
	COPY_CTRL_PACKET(echo_packet, rply_packet, rply_size);
	DEBUG_PACKET("ECHO REQ");
}


static int pptp_gre_init(struct in_addr *inetaddrs)
{
	struct sockaddr_in addr;
	int gre_fd;

	/* Open IP protocol socket */
	gre_fd = bb_xsocket(BB_XSOCKET_FLG_SOCK_RAW | BB_XSOCKET_FLG_ERR_DIE | BB_XSOCKET_FLG_PROTO_SPEC, PPTP_PROTO);

	memset(&addr, 0, sizeof(addr));

	addr.sin_family = AF_INET;
	addr.sin_addr = inetaddrs[0];
	addr.sin_port = 0;
	if (bind(gre_fd, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
		bb_perror_msg("GRE: bind() failed");
		bb_error_msg("GRE: continuing, but may not work if multi-homed");
	}

	addr.sin_family = AF_INET;
	addr.sin_addr = inetaddrs[1];
	addr.sin_port = 0;
	if (connect(gre_fd, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
		bb_perror_msg("GRE: connect() failed");
		return -1;
	}

	gre.seq_sent = 0;
	gre.ack_sent = gre.ack_recv = gre.seq_recv = 0xFFFFFFFF;
	/* seq_recv is -1, therefore next packet expected is seq 0,
	   to comply with RFC 2637: 'The sequence number for each
	   user session is set to zero at session startup.' */

	gre.call_id_pair = Gcall_id_pair;   /* network byte order */
	return gre_fd;
}

/*
 * pptp_handle_ctrl_connection
 *
 * 1. read a packet (should be start_ctrl_conn_rqst)
 * 2. reply to packet (send a start_ctrl_conn_rply)
 * 3. proceed with GRE and CTRL connections
 *
 * args: pppaddrs - ppp local and remote addresses (strings)
 *       inetaddrs - local and client socket address
 * retn: 0 success, -1 failure
 */
static void pptp_handle_ctrl_connection(char **pppaddrs, struct in_addr *inetaddrs)
{
	/* For echo requests used to check link is alive */
	int echo_wait = FALSE;          /* Waiting for echo? */
	u_int32_t echo_count = 0;       /* Sequence # of echo */
	time_t echo_time = 0;           /* Time last echo req sent */
	struct timeval idleTime;        /* How long to select() */

	/* General local variables */
	ssize_t rply_size;              /* Reply packet size */
	fd_set fds;                     /* For select() */
	int maxfd = ctrl_clientSocket;  /* For select() */
	int send_packet;                /* Send a packet this time? */
	int init = 0;                   /* Has pppd initialized the pty? */
	int pty_fd = -1;                /* File descriptor of pty */
	int gre_fd = -1;                /* Network file descriptor */
	int sig_fd = sigpipe_fd();      /* Signal pipe descriptor       */

	unsigned char packet[PPTP_MAX_CTRL_PCKT_SIZE];
	unsigned char rply_packet[PPTP_MAX_CTRL_PCKT_SIZE];

	for (;;) {

		FD_ZERO(&fds);
		FD_SET(sig_fd, &fds);
		FD_SET(ctrl_clientSocket, &fds);
		if (pty_fd != -1)
			FD_SET(pty_fd, &fds);
		if (gre_fd != -1 && init)
			FD_SET(gre_fd, &fds);

		/* set timeout */
		if (encaps_gre(-1, NULL, 0) || decaps_hdlc(-1, NULL, 0)) {
			idleTime.tv_sec = 0;
			idleTime.tv_usec = 50000; /* don't ack immediately */
		} else {
			idleTime.tv_sec = IDLE_WAIT;
			idleTime.tv_usec = 0;
		}

		/* default: do nothing */
		send_packet = FALSE;

		switch (select(maxfd + 1, &fds, NULL, NULL, &idleTime)) {
		case -1:        /* Error with select() */
			if (errno != EINTR)
				bb_error_msg("CTRL: Error with select(), quitting");
			goto leave_clear_call;

		case 0:
			if (decaps_hdlc(-1, NULL, 0)) {
				if(decaps_hdlc(-1, encaps_gre, gre_fd))
					bb_error_msg("CTRL: GRE re-xmit failed");
			} else if (encaps_gre(-1, NULL, 0))
				/* Pending ack and nothing else to do */
				encaps_gre(gre_fd, NULL, 0);    /* send ack with no payload */
			else if (echo_wait != TRUE) {
				/* Timeout. Start idle link detection. */
				echo_count++;
				if (pptpctrl_debug)
					syslog(LOG_DEBUG, "CTRL: Sending ECHO REQ id %d", echo_count);
				time(&echo_time);
				make_echo_req_packet(rply_packet, &rply_size, echo_count);
				echo_wait = TRUE;
				send_packet = TRUE;
			}
			break;

		default:
			break;
		}

		/* check for pending SIGTERM delivery */
		if (FD_ISSET(sig_fd, &fds)) {
			if (sigpipe_read() == SIGTERM)
				bail(SIGTERM);
		}

		/* detect startup of pppd */
		if (!init && pty_fd != -1 && FD_ISSET(pty_fd, &fds))
			init = 1;

		/* handle actual packets */

		/* send from pty off via GRE */
		if (pty_fd != -1 && FD_ISSET(pty_fd, &fds) && decaps_hdlc(pty_fd, encaps_gre, gre_fd) < 0) {
			bb_error_msg("CTRL: PTY read or GRE write failed (pty,gre)=(%d,%d)", pty_fd, gre_fd);
			break;
		}
		/* send from GRE off to pty */
		if (gre_fd != -1 && FD_ISSET(gre_fd, &fds) && decaps_gre(gre_fd, encaps_hdlc, pty_fd) < 0) {
			if (gre_fd == 6 && pty_fd == 5) {
				bb_error_msg("CTRL: GRE-tunnel has collapsed (GRE read or PTY write failed (gre,pty)=(%d,%d))", gre_fd, pty_fd);
			} else {
				bb_error_msg("CTRL: GRE read or PTY write failed (gre,pty)=(%d,%d)", gre_fd, pty_fd);
			}
			break;
		}
		/* handle control messages */

		if (FD_ISSET(ctrl_clientSocket, &fds)) {
			send_packet = TRUE;
			switch (read_pptp_packet(ctrl_clientSocket, packet, rply_packet, &rply_size)) {
			case 0:
				bb_error_msg("CTRL: CTRL read failed");
				goto leave_drop_call;

			case -1:
				send_packet = FALSE;
				break;

			case STOP_CTRL_CONN_RQST:
				if (pptpctrl_debug)
					syslog(LOG_DEBUG, "CTRL: Received STOP CTRL CONN request (disconnecting)");
				if (gre_fd != -1 || pty_fd != -1)
					xlog_warn("CTRL: Request to close control connection when call is open, closing");
				send_pptp_packet(ctrl_clientSocket, rply_packet, rply_size);
				goto leave_drop_call;

			case CALL_CLR_RQST:
				if(pptpctrl_debug)
					syslog(LOG_DEBUG, "CTRL: Received CALL CLR request (closing call)");
				if (gre_fd == -1 || pty_fd == -1)
					xlog_warn("CTRL: Request to close call but call not open");
				if (gre_fd != -1) {
					FD_CLR(gre_fd, &fds);
					close(gre_fd);
					gre_fd = -1;
				}
				if (pty_fd != -1) {
					FD_CLR(pty_fd, &fds);
					close(pty_fd);
					pty_fd = -1;
				}
				/* violating RFC */
				goto leave_drop_call;

			case OUT_CALL_RQST:
				/* for killing off the link later (ugly) */
				NOTE_VALUE(PAC, Gcall_id_pair, ((struct pptp_out_call_rply *) (rply_packet))->call_id);
				NOTE_VALUE(PNS, Gcall_id_pair, ((struct pptp_out_call_rply *) (rply_packet))->call_id_peer);
				if (gre_fd != -1 || pty_fd != -1) {
					xlog_warn("CTRL: Request to open call when call is already open, closing");
					if (gre_fd != -1) {
						FD_CLR(gre_fd, &fds);
						close(gre_fd);
						gre_fd = -1;
					}
					if (pty_fd != -1) {
						FD_CLR(pty_fd, &fds);
						close(pty_fd);
						pty_fd = -1;
					}
				}
				/* change process title for accounting and status scripts */
				bb_setproctitle("pptpd [%s:%04X - %04X]",
				      inet_ntoa(inetaddrs[1]),
				      ntohs(((struct pptp_out_call_rply *) (rply_packet))->call_id_peer),
				      ntohs(((struct pptp_out_call_rply *) (rply_packet))->call_id));
				/* start the call, by launching pppd */
				syslog(LOG_INFO, "CTRL: Starting call (launching pppd, opening GRE)");
				pty_fd = startCall(pppaddrs, inetaddrs);
				if (pty_fd > maxfd) maxfd = pty_fd;
				if ((gre_fd = pptp_gre_init(inetaddrs)) > maxfd)
					maxfd = gre_fd;
				break;

			case ECHO_RPLY:
				if (echo_wait == TRUE && ((struct pptp_echo_rply *) (packet))->identifier == echo_count)
					echo_wait = FALSE;
				else
					xlog_warn("CTRL: Unexpected ECHO REPLY packet");
				/* FALLTHRU */
			case SET_LINK_INFO:
				send_packet = FALSE;
				break;

			case CALL_DISCONN_NTFY:
			case STOP_CTRL_CONN_RPLY:
				/* These don't generate replies.  Also they come from things we don't send in this section. */
				xlog_warn("CTRL: Got a reply to a packet we didn't send");
				send_packet = FALSE;
				break;

			/* Otherwise, the already-formed reply will do fine, so send it */
			}
		}

		/* send reply packet - this may block, but it should be very rare */
		if (send_packet == TRUE && send_pptp_packet(ctrl_clientSocket, rply_packet, rply_size) < 0) {
			bb_error_msg("CTRL: Error sending GRE, aborting call");
			goto leave_clear_call;
		}

		/* waiting for echo reply and curtime - echo_time > max wait */
		if (echo_wait == TRUE && (time(NULL) - echo_time) > MAX_ECHO_WAIT) {
			syslog(LOG_INFO, "CTRL: Session timed out, ending call");
			goto leave_clear_call;
		}
	}
	/* Finished! :-) */
leave_drop_call:
	NOTE_VALUE(PAC, Gcall_id_pair, htons(-1));
	NOTE_VALUE(PNS, Gcall_id_pair, htons(-1));
	close(ctrl_clientSocket);
leave_clear_call:
	/* leave ctrl_clientSocket and Gcall_id_pair alone for bail() */
	if (gre_fd != -1)
		close(gre_fd);
	gre_fd = -1;
	if (pty_fd != -1)
		close(pty_fd);
	pty_fd = -1;
	return;
}

/* Oh the horror.. lets hope this covers all the ones we have to handle */
#if defined(O_NONBLOCK) && !defined(__sun__) && !defined(__sun)
#define OUR_NB_MODE O_NONBLOCK
#else
#define OUR_NB_MODE O_NDELAY
#endif

/* read a command line argument, a flag alone */
#define GETARG_INT(X) \
	X = atoi(argv[arg++])

/* read a command line argument, a string alone */
#define GETARG_STRING(X) \
	X = strdup(argv[arg++])

/* read a command line argument, a presence flag followed by string */
#define GETARG_VALUE(X) \
	if(atoi(argv[arg++]) != 0) \
		safe_strncpy(X, argv[arg++], sizeof(X)); \
	else \
		*X = '\0'

int pptpctrl_main(void)
{
	int argc = Gargc;
	char **argv = Gargv;
	char pppLocal[16];              /* local IP to pass to pppd */
	char pppRemote[16];             /* remote IP address to pass to pppd */
	struct sockaddr_in addr;        /* client address */
	socklen_t addrlen;
	int arg = 1;
	int flags;
	struct in_addr inetaddrs[2];
	char *pppaddrs[2] = { pppLocal, pppRemote };

	/* fail if argument count invalid */
	if (argc < 7) {
		bb_default_error_retval = 2;
		bb_error_msg_and_die("insufficient arguments, see man pptpctrl");
	}

	/* open a connection to the syslog daemon */
	bb_openlog(LOG_PID, PPTP_FACILITY);
	log_syslog = LOG_ERR;
	log_stderr = 0;

	/* autoreap if supported */
	bb_signal((1 << SIGCHLD), SIG_IGN);

	/* note: update pptpctrl.8 if the argument list format is changed */
	GETARG_INT(pptpctrl_debug);
	GETARG_INT(noipparam);
	GETARG_VALUE(pppdxfig);
	GETARG_VALUE(Gspeed);
	GETARG_VALUE(pppLocal);
	GETARG_VALUE(pppRemote);
	if (arg < argc) GETARG_INT(unique_call_id);
	if (arg < argc) GETARG_STRING(ppp_binary);
	if (arg < argc) GETARG_INT(pptp_logwtmp);

	if (pptpctrl_debug) {
		if (*pppLocal)
			syslog(LOG_DEBUG, "CTRL: local address = %s", pppLocal);
		if (*pppRemote)
			syslog(LOG_DEBUG, "CTRL: remote address = %s", pppRemote);
		if (*Gspeed)
			syslog(LOG_DEBUG, "CTRL: pppd speed = %s", Gspeed);
		if (*pppdxfig)
			syslog(LOG_DEBUG, "CTRL: pppd options file = %s", pppdxfig);
	}

	addrlen = sizeof(addr);
	if (getsockname(ctrl_clientSocket, (struct sockaddr *) &addr, &addrlen) != 0) {
		bb_perror_msg("CTRL: getsockname() failed");
		close(ctrl_clientSocket);
		bail(0);
	}
	inetaddrs[0] = addr.sin_addr;

	addrlen = sizeof(addr);
	if (getpeername(ctrl_clientSocket, (struct sockaddr *) &addr, &addrlen) != 0) {
		bb_perror_msg("CTRL: getpeername() failed");
		close(ctrl_clientSocket);
		bail(0);
	}
	inetaddrs[1] = addr.sin_addr;

	/* Set non-blocking */
	if ((flags = fcntl(ctrl_clientSocket, F_GETFL, arg /* ignored */)) == -1 ||
	    fcntl(ctrl_clientSocket, F_SETFL, flags|OUR_NB_MODE) == -1) {
		bb_perror_msg("CTRL: Failed to set client socket non-blocking");
		close(ctrl_clientSocket);
		bail(0);
	}

	/* Fiddle with argv */
	bb_setproctitle("pptpd [%s]", inet_ntoa(addr.sin_addr));

	/* be ready for a grisly death */
	sigpipe_create();
	sigpipe_assign(SIGTERM);
	NOTE_VALUE(PAC, Gcall_id_pair, htons(-1));
	NOTE_VALUE(PNS, Gcall_id_pair, htons(-1));

	syslog(LOG_INFO, "CTRL: Client %s control connection started", inet_ntoa(addr.sin_addr));
	pptp_handle_ctrl_connection(pppaddrs, inetaddrs);
	syslog(LOG_DEBUG, "CTRL: Reaping child PPP[%i]", pppfork);
	if (pppfork > 0)
		waitpid_no_int(pppfork, NULL, 0, 1);
	syslog(LOG_INFO, "CTRL: Client %s control connection finished", inet_ntoa(addr.sin_addr));

	bail(0);
	return 1;               /* make gcc happy */
}
#endif /* APPLET_USAGE */
