/*-
 * Copyright (c) 1983, 1990 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * rsh.c      5.24 (Berkeley) 7/1/91
 * char pkg[]="$NetKit: netkit-rsh-0.17 $";
 */
#ifdef APPLET_USAGE
#define rsh_trivial_usage \
	"[-nds] [-l login] host [command]"
#define rsh_full_usage \
	"Executes command on host\n\n" \
	"Options:\n" \
	"\t-n Redirect standart input from /dev/null\n" \
	"\t-d Turns on socket debugging\n" \
	"\t-s Redirect standart stderr to /dev/null, use it into firewall"
#else

#include "busybox.h"
#include "bb_inet.h"
#include "bb_sig.h"
#include <sys/ioctl.h>
#include <sys/file.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <unistd.h>
#include <string.h>

/* This is just a warning of a common mistake.  Possibly this should be a
 * uclibc faq entry rather than in busybox... */
#if defined(__UCLIBC__) && ! defined(__UCLIBC_HAS_RPC__)
#error "You need to build uClibc with UCLIBC_HAS_RPC for portmap build."
#endif
#if defined(__MUSL__)
#error "MUSL libc have not RPC support. We can not compile rsh applet. :("
#endif

#define _PATH_RLOGIN    "/usr/bin/rlogin"

/*
 * rsh - remote shell
 */

CALC_BSS struct BB_COMMON_BSS {
	int rfd2, sflag;
} bb_common_bss;
#define rfd2 bb_common_bss.rfd2
#define sflag bb_common_bss.sflag

static void
talk(int nflag, sigset_t omask, int pid, int rem)
{
	int cc, wc;
	char *bp;
	fd_set readfrom, rembits;
	int rfd2_ok, rem_ok;
	char buf[BUFSIZ];

	FD_ZERO(&rembits);

	if (!nflag && pid == 0) {
		if (!sflag)
			close(rfd2);

reread:         errno = 0;
		if ((cc = read(0, buf, sizeof buf)) <= 0)
			goto done;
		bp = buf;

rewrite:        FD_ZERO(&rembits);
		FD_SET(rem, &rembits);
		if (select(rem+1, 0, &rembits, 0, 0) < 0) {
			if (errno != EINTR)
				bb_perror_msg_and_die("select");
			goto rewrite;
		}
		if (! FD_ISSET(rem, &rembits))
			goto rewrite;
			wc = write(rem, bp, cc);
		if (wc < 0) {
			if (errno == EWOULDBLOCK)
				goto rewrite;
			goto done;
		}
		bp += wc;
		cc -= wc;
		if (cc == 0)
			goto reread;
		goto rewrite;
done:
		shutdown(rem, 1);
		exit(0);
	}

	rfd2_ok = !sflag;
	rem_ok = 1;
	bb_sigunblock(omask);
	while (rfd2_ok || rem_ok) {
		FD_ZERO(&readfrom);
		if (!sflag && rfd2_ok)
			FD_SET(rfd2, &readfrom);
		if (rem_ok)
			FD_SET(rem, &readfrom);
		if (select(!sflag && (rfd2 > rem) ? rfd2+1 : rem+1,
			   &readfrom, 0, 0, 0) < 0) {
			if (errno != EINTR)
				bb_perror_msg_and_die("select");
			continue;
		}
		if (!sflag && FD_ISSET(rfd2, &readfrom)) {
			errno = 0;
				cc = read(rfd2, buf, sizeof buf);
			if (cc > 0)
				write(2, buf, cc);
			else if (cc == 0 || errno != EWOULDBLOCK)
				rfd2_ok = 0;
		}
		if (FD_ISSET(rem, &readfrom)) {
			errno = 0;
				cc = read(rem, buf, sizeof buf);
			if (cc > 0)
				write(1, buf, cc);
			else if (cc == 0 || errno != EWOULDBLOCK)
				rem_ok = 0;
		}
	}
}

static void
sendsig(int signo)
{
	char x = (char) signo;

	if (!sflag)
		write(rfd2, &x, 1);
}

static char *
copyargs(char **argv)
{
	int cc;
	char **ap, *p;
	char *args;

	cc = 0;
	for (ap = argv; *ap; ++ap)
		cc += strlen(*ap) + 1;
	args = xmalloc(cc);
	for (p = args, ap = argv; *ap; ++ap) {
		/*strcpy(p, *ap);*/
		for (p = strcpy(p, *ap); *p; ++p)
			;
		if (ap[1])
			*p++ = ' ';
	}
	return(args);
}

int rsh_main(void)
{
	char *pw;
	char **gargv0 = Gargv;
	sigset_t omask;
	int argoff, ch, dflag, nflag, one, pid=0, rem, uid;
	char *args, *host, *user;
	char *null = NULL;
	unsigned short s_port;

	argoff = 0;
	one = 1;
	host = user = NULL;

	/* handle "rsh host flags" */
	if (Gargv[1] != NULL && Gargv[1][0] != '-') {
		host = Gargv[1];
		ch = 1;
		do {
			ch++;
			Gargv[ch - 1] = Gargv[ch];
		} while(Gargv[ch] != NULL);
		Gargc--;
		argoff = 1;
	}
	ch = bb_getopt_ulflags("+dsn8KLel:w", &user);
	dflag = ch & 1;
	sflag = ch & 2;
	nflag = ch & 4;
	/* -8Lew are ignored to allow rlogin aliases */

	/* if haven't gotten a host yet, do so */
	if (host == NULL && (host = *Gargv++) == NULL)
		bb_show_usage();

	/* if no further arguments, must have been called as rlogin. */
	if (*Gargv == NULL) {
		if (setuid(getuid()))
			bb_perror_msg_and_die("setuid");
		gargv0[0] = (char *)"rlogin";
		if(argoff) {
			*Gargv++ = host;
			*Gargv = NULL;
		}
		execv(_PATH_RLOGIN, gargv0);
		bb_perror_msg_and_die("can't exec %s", _PATH_RLOGIN);
	}

	pw = bb_getpwuid(NULL, uid = getuid(), -1);
	if (!user)
		user = pw;

	__environ = &null;
	args = copyargs(Gargv);

	s_port = bb_lookup_tcp_port("shell", 514);

	if (sflag)
		rem = rcmd(&host, s_port, pw, user, args, NULL);
	else
		rem = rcmd(&host, s_port, pw, user, args, &rfd2);

	if (rem < 0)
		exit(1);

	if (!sflag && rfd2 < 0)
		bb_error_msg_and_die("can't establish stderr.");

	if (setuid(uid))
		bb_perror_msg_and_die("setuid");

	if (dflag) {
		if (setsockopt(rem, SOL_SOCKET, SO_DEBUG, &one,
		    sizeof(one)) < 0)
			bb_perror_msg("setsockopt");
		if (!sflag && (setsockopt(rfd2, SOL_SOCKET, SO_DEBUG, &one,
		    sizeof(one)) < 0))
			bb_perror_msg("setsockopt");
	}

	omask = bb_sigblock((1<<SIGINT)|(1<<SIGQUIT)|(1<<SIGTERM));
	if (bb_signal(1 | (1 << SIGINT), SIG_IGN) != SIG_IGN)
		bb_signal(1 | (1 << SIGINT), sendsig);
	if (bb_signal(1 | (1 << SIGQUIT), SIG_IGN) != SIG_IGN)
		bb_signal(1 | (1 << SIGQUIT), sendsig);
	if (bb_signal(1 | (1 << SIGTERM), SIG_IGN) != SIG_IGN)
		bb_signal(1 | (1 << SIGTERM), sendsig);

	if (!nflag) {
		pid = fork();
		if (pid < 0)
			bb_perror_msg_and_die("fork");
	}

	if (!sflag)
		ioctl(rfd2, FIONBIO, &one);
	ioctl(rem, FIONBIO, &one);

	talk(nflag, omask, pid, rem);

	if (!nflag)
		kill(pid, SIGKILL);
	return 0;
}
#endif /* APPLET_USAGE */
