/*
 * mini-snmpd 1.4-rc1
 *  bugreport: https://github.com/troglobit/mini-snmpd/issues
 *
 * Copyright (C) 2008-2010 Robert Ernst <robert.ernst@linux-solutions.at>
 * Copyright (C) 2011       Javier Palacios <javiplx@gmail.com>
 *
 * Busybox port by Vladimir N. Oleynik (c) 2010-2016 <dzo@simtreas.ru>
 */
#ifdef APPLET_USAGE
#define snmpd_trivial_usage \
	"[-avl][-p n][-P n][-c n][-D n][-V n][-L n][-C n][-d n][-i n][-t t]"
#define snmpd_full_usage \
	USE_FEATURE_LONG_OPTIONS( \
	"\t-p, --udp-port nnn\tset the UDP port to bind to (161)\n" \
	"\t-P, --tcp-port nnn\tset the TCP port to bind to (161)\n" \
	"\t-c, --community STR\tset the community string (public)\n" \
	"\t-D, --description STR\tset the system description (empty)\n" \
	"\t-V, --vendor OID\tset the system vendor (empty)\n" \
	"\t-L, --location STR\tset the system location (empty)\n" \
	"\t-C, --contact STR\tset the system contact (empty)\n" \
	"\t-d, --disks PATH\tset the disks to monitor (/)\n" \
	"\t-i, --interfaces IFACE\tset the network interfaces to monitor (lo)\n" \
	"\t-I, --listen IFACE\tset the network interface to listen (all)\n" \
	"\t-t, --timeout SEC\tset the timeout for MIB updates (1 second)\n" \
	"\t-a, --auth\t\trequire authentication (thus SNMP version 2c)\n" \
	"\t-v, --verbose\tverbose syslog messages\n" \
	"\t-n, --foreground\t\trun in foreground, do not detach\n" \
	"\t-s, --syslog\t\tuse syslog for logging") \
	SKIP_FEATURE_LONG_OPTIONS( \
	"\t-p nnn\tset the UDP port to bind to (161)\n" \
	"\t-P nnn\tset the TCP port to bind to (161)\n" \
	"\t-c STR\tset the community string (public)\n" \
	"\t-D STR\tset the system description (empty)\n" \
	"\t-V OID\tset the system vendor (empty)\n" \
	"\t-L STR\tset the system location (empty)\n" \
	"\t-C STR\tset the system contact (empty)\n" \
	"\t-d PATH\tset the disks to monitor (/)\n" \
	"\t-i IFACE\tset the network interfaces to monitor (lo)\n" \
	"\t-I FACE\tset the network interface to listen (all)\n" \
	"\t-t SEC\tset the timeout for MIB updates (1 second)\n" \
	"\t-a\t\trequire authentication (thus SNMP version 2c)\n" \
	"\t-v\tverbose syslog messages\n" \
	"\t-n\trun in foreground, do not detach\n" \
	"\t-s\tuse syslog for logging")
#else

/* Vendor OID */
#define VENDOR ".1.3.6.1.4.1"

/* Version number of package */
#define VERSION "1.4-rc1"

#include "busybox.h"

#include <sys/time.h>
#include <netinet/in.h>
#include <limits.h>
#include <syslog.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <unistd.h>
#include <sys/sysinfo.h>
#include <sys/ioctl.h>
#include <sys/vfs.h>
#include <math.h>
#include "bb_inet.h"
#include "bb_sig.h"


/* Project dependent defines */

#define EXIT_OK                                         0
#define EXIT_ARGS                                       1
#define EXIT_SYSCALL                                    2

#define MAX_NR_CLIENTS                                  16
#define MAX_NR_OIDS                                     16
#define MAX_NR_SUBIDS                                   16
#define MAX_NR_DISKS                                    4
#define MAX_NR_INTERFACES                               8
#define MAX_NR_VALUES                                   192

#define MAX_PACKET_SIZE                                 2048
#define MAX_STRING_SIZE                                 64

/* SNMP dependent defines */

#define BER_TYPE_BOOLEAN                                0x01
#define BER_TYPE_INTEGER                                0x02
#define BER_TYPE_BIT_STRING                             0x03
#define BER_TYPE_OCTET_STRING                           0x04
#define BER_TYPE_NULL                                   0x05
#define BER_TYPE_OID                                    0x06
#define BER_TYPE_SEQUENCE                               0x30
#define BER_TYPE_COUNTER                                0x41
#define BER_TYPE_GAUGE                                  0x42
#define BER_TYPE_TIME_TICKS                             0x43
#define BER_TYPE_NO_SUCH_OBJECT                         0x80
#define BER_TYPE_NO_SUCH_INSTANCE                       0x81
#define BER_TYPE_END_OF_MIB_VIEW                        0x82
#define BER_TYPE_SNMP_GET                               0xA0
#define BER_TYPE_SNMP_GETNEXT                           0xA1
#define BER_TYPE_SNMP_RESPONSE                          0xA2
#define BER_TYPE_SNMP_SET                               0xA3
#define BER_TYPE_SNMP_GETBULK                           0xA5
#define BER_TYPE_SNMP_INFORM                            0xA6
#define BER_TYPE_SNMP_TRAP                              0xA7
#define BER_TYPE_SNMP_REPORT                            0xA8

#define SNMP_VERSION_1                                  0
#define SNMP_VERSION_2C                                 1
#define SNMP_VERSION_3                                  3

#define SNMP_STATUS_OK                                  0
#define SNMP_STATUS_TOO_BIG                             1
#define SNMP_STATUS_NO_SUCH_NAME                        2
#define SNMP_STATUS_BAD_VALUE                           3
#define SNMP_STATUS_READ_ONLY                           4
#define SNMP_STATUS_GEN_ERR                             5
#define SNMP_STATUS_NO_ACCESS                           6
#define SNMP_STATUS_WRONG_TYPE                          7
#define SNMP_STATUS_WRONG_LENGTH                        8
#define SNMP_STATUS_WRONG_ENCODING                      9
#define SNMP_STATUS_WRONG_VALUE                         10
#define SNMP_STATUS_NO_CREATION                         11
#define SNMP_STATUS_INCONSISTENT_VALUE                  12
#define SNMP_STATUS_RESOURCE_UNAVAILABLE                13
#define SNMP_STATUS_COMMIT_FAILED                       14
#define SNMP_STATUS_UNDO_FAILED                         15
#define SNMP_STATUS_AUTHORIZATION_ERROR                 16
#define SNMP_STATUS_NOT_WRITABLE                        17
#define SNMP_STATUS_INCONSISTENT_NAME                   18

#if defined(__BIONIC__)
typedef uint16_t in_port_t;
#endif

#if !ENABLE_FEATURE_IPV6
#define my_sockaddr_t           sockaddr_in
#define my_sin_addr             sin_addr
#define my_sin_port             sin_port
#define my_sin_family           sin_family
#define my_af_inet              AF_INET
#define my_pf_inet              PF_INET
#define my_in_addr_t            in_addr
#define my_inet_addrstrlen      INET_ADDRSTRLEN

#else /* IPv6 */

#define my_sockaddr_t           sockaddr_in6
#define my_sin_addr             sin6_addr
#define my_sin_port             sin6_port
#define my_sin_family           sin6_family
#define my_af_inet              AF_INET6
#define my_pf_inet              PF_INET6
#define my_in_addr_t            in6_addr
#define my_inet_addrstrlen      INET6_ADDRSTRLEN
#endif/* IPV6 */


/* Data types */

typedef struct client_s {
	time_t              timestamp;
	int                 sockfd;
	struct my_in_addr_t addr;
	in_port_t           port;
	unsigned char       packet[MAX_PACKET_SIZE];
	size_t              size;
	int                 outgoing;
} client_t;

typedef struct oid_s {
	unsigned int subid_list[MAX_NR_SUBIDS];
	size_t       subid_list_length;
	short        encoded_length;
} oid_t;

typedef struct data_s {
	unsigned char *buffer;
	size_t         max_length;
	short          encoded_length;
} data_t;

typedef struct value_s {
	oid_t  oid;
	data_t data;
} value_t;

typedef struct field_s {
	char         *prefix;

	size_t        len;
	unsigned int *value[12];
} field_t;

typedef struct request_s {
	char      community[MAX_STRING_SIZE];
	int       type;
	int       version;
	int       id;
	uint32_t  non_repeaters;
	uint32_t  max_repetitions;
	oid_t     oid_list[MAX_NR_OIDS];
	size_t    oid_list_length;
} request_t;

typedef struct response_s {
	int     error_status;
	int     error_index;
	value_t value_list[MAX_NR_VALUES];
	size_t  value_list_length;
} response_t;

typedef struct loadinfo_s {
	unsigned int avg[3];
} loadinfo_t;

typedef struct meminfo_s {
	unsigned int total;
	unsigned int free;
	unsigned int shared;
	unsigned int buffers;
	unsigned int cached;
} meminfo_t;

typedef struct cpuinfo_s {
	unsigned int user;
	unsigned int nice;
	unsigned int system;
	unsigned int idle;
	unsigned int irqs;
	unsigned int cntxts;
} cpuinfo_t;

typedef struct diskinfo_s {
	unsigned int total[MAX_NR_DISKS];
	unsigned int free[MAX_NR_DISKS];
	unsigned int used[MAX_NR_DISKS];
	unsigned int blocks_used_percent[MAX_NR_DISKS];
	unsigned int inodes_used_percent[MAX_NR_DISKS];
} diskinfo_t;

typedef struct netinfo_s {
	unsigned int status[MAX_NR_INTERFACES];
	unsigned int rx_bytes[MAX_NR_INTERFACES];
	unsigned int rx_packets[MAX_NR_INTERFACES];
	unsigned int rx_errors[MAX_NR_INTERFACES];
	unsigned int rx_drops[MAX_NR_INTERFACES];
	unsigned int tx_bytes[MAX_NR_INTERFACES];
	unsigned int tx_packets[MAX_NR_INTERFACES];
	unsigned int tx_errors[MAX_NR_INTERFACES];
	unsigned int tx_drops[MAX_NR_INTERFACES];
} netinfo_t;


static const struct in_addr inaddr_any = { INADDR_ANY };

CALC_BSS struct BB_COMMON_BSS {
#if ENABLE_FEATURE_IPV6
	int       g_family;
#endif
	int       g_timeout;
	char     *g_community;
	char     *g_vendor;
	char     *g_disk_list[MAX_NR_DISKS];
	size_t    g_disk_list_length;
	in_port_t g_udp_port;
	in_port_t g_tcp_port;
	int       g_udp_sockfd;
	int       g_tcp_sockfd;
	int       g_auth;
	int       g_no_daemon;
	int       g_syslog;
	int       g_verbose;
	int       g_quit;

	char     *g_description;
	char     *g_location;
	char     *g_contact;
	char     *g_bind_to_device;

	char     *g_interface_list[MAX_NR_INTERFACES];
	size_t    g_interface_list_length;

	client_t  g_udp_client;
	client_t *g_tcp_client_list[MAX_NR_CLIENTS];
	size_t    g_tcp_client_list_length;

	value_t   g_mib[MAX_NR_VALUES];
	size_t    g_mib_length;
	char      oid_ntoa_buf[MAX_NR_SUBIDS * 10 + 2];
	unsigned int uptime_start;
	oid_t     oid_aton_oid;
} bb_common_bss;
#if ENABLE_FEATURE_IPV6
#define g_family bb_common_bss.g_family
#else
#define g_family AF_INET
#endif
#define g_timeout bb_common_bss.g_timeout
#define g_community bb_common_bss.g_community
#define g_vendor bb_common_bss.g_vendor
#define g_disk_list bb_common_bss.g_disk_list
#define g_disk_list_length bb_common_bss.g_disk_list_length
#define g_udp_port bb_common_bss.g_udp_port
#define g_tcp_port bb_common_bss.g_tcp_port
#define g_udp_sockfd bb_common_bss.g_udp_sockfd
#define g_tcp_sockfd bb_common_bss.g_tcp_sockfd
#define g_auth bb_common_bss.g_auth
#define g_no_daemon bb_common_bss.g_no_daemon
#define g_syslog bb_common_bss.g_syslog
#define g_verbose bb_common_bss.g_verbose
#define g_quit bb_common_bss.g_quit
#define g_description bb_common_bss.g_description
#define g_location bb_common_bss.g_location
#define g_contact bb_common_bss.g_contact
#define g_bind_to_device bb_common_bss.g_bind_to_device
#define g_interface_list bb_common_bss.g_interface_list
#define g_interface_list_length bb_common_bss.g_interface_list_length
#define g_udp_client bb_common_bss.g_udp_client
#define g_tcp_client_list bb_common_bss.g_tcp_client_list
#define g_tcp_client_list_length bb_common_bss.g_tcp_client_list_length
#define g_mib bb_common_bss.g_mib
#define g_mib_length bb_common_bss.g_mib_length
#define oid_ntoa_buf bb_common_bss.oid_ntoa_buf
#define uptime_start bb_common_bss.uptime_start
#define oid_aton_oid bb_common_bss.oid_aton_oid

static void lprintf(int level, const char *format, ...)
{
	if (g_verbose || (level != LOG_DEBUG)) {
		va_list arg;
		va_start(arg, format);

		if (g_syslog)
			vsyslog(level, format, arg);
		else {
			bb_vfprintf(stderr, format, arg);
			putc('\n', stderr);
		}
		va_end(arg);
	}
}

static inline int parse_lineint(char *buf, field_t *f)
{
	char *ptr, *prefixptr;
	size_t i;

	ptr = buf;
	while (isspace(*ptr))
		ptr++;
	if (!*ptr)
		return 0;

	/* Check if buffer begins with prefix */
	prefixptr = f->prefix;
	while (*prefixptr) {
		if (*prefixptr++ != *ptr++)
			return 0;
	}
	if (*ptr == ':')        /* Prefix may have a ':', skip it too! */
		ptr++;
	else if (!isspace(*ptr))/* If there is NO ':' after prefix there must be a space, otherwise we got a partial match */
		return 0;

	for (i = 0; i < f->len; i++) {
		while (isspace(*ptr))
			ptr++;

		if (f->value[i]) {
			*(f->value[i]) = strtoull(ptr, NULL, 0);
		}

		while (!isspace(*ptr))
			ptr++;
	}

	return 1;
}

static int parse_file(char *file, field_t fields[])
{
	char buf[512];
	FILE *fp;

	if (!file || !fields)
		return -1;

	fp = fopen(file, "r");
	if (!fp)
		return -1;

	while (fgets(buf, sizeof(buf), fp)) {
		int i;

		for (i = 0; fields[i].prefix; i++) {
			if (parse_lineint(buf, &fields[i]))
				break;
		}
	}

	return fclose(fp);
}

static int read_file(const char *filename, char *buf, size_t size)
{
	int ret;
	FILE *fp;
	size_t len;

	fp = fopen(filename, "r");
	if (!fp) {
		lprintf(LOG_WARNING, "Failed opening %s: %m", filename);
		return -1;
	}

	len = fread(buf, 1, size - 1, fp);
	ret = fclose(fp);
	if (len == 0 || ret == -1) {
		lprintf(LOG_WARNING, "Failed reading %s: %m", filename);
		return -1;
	}

	buf[len] = '\0';

	return 0;
}

static int ticks_since(const struct timeval *tv_last, struct timeval *tv_now)
{
	float ticks;

	if (gettimeofday(tv_now, NULL) == -1) {
		lprintf(LOG_WARNING, "could not get ticks: %m");
		return -1;
	}

	if (tv_now->tv_sec < tv_last->tv_sec || (tv_now->tv_sec == tv_last->tv_sec && tv_now->tv_usec < tv_last->tv_usec)) {
		lprintf(LOG_WARNING, "could not get ticks: time running backwards");
		return -1;
	}

	ticks = (float)(tv_now->tv_sec - 1 - tv_last->tv_sec) * 100.0 + (float)((tv_now->tv_usec + 1000000 - tv_last->tv_usec) / 10000);
	if (ticks < INT_MIN)
		return INT_MIN;
	if (ticks > INT_MAX)
		return INT_MAX;

	return ticks;
}

static char *oid_ntoa(const oid_t *oid)
{
	size_t i, len = 0;

	oid_ntoa_buf[0] = '\0';
	for (i = 0; i < oid->subid_list_length; i++) {
		len += snprintf(oid_ntoa_buf + len, sizeof(oid_ntoa_buf) - len, ".%u", oid->subid_list[i]);
		if (len >= sizeof(oid_ntoa_buf))
			break;
	}

	return oid_ntoa_buf;
}

static oid_t *oid_aton(const char *str)
{
	char *ptr = (char *)str;

	if (!str)
		return NULL;

	oid_aton_oid.subid_list_length = 0;
	while (*ptr != 0) {
		if (oid_aton_oid.subid_list_length >= MAX_NR_SUBIDS)
			return NULL;

		if (*ptr != '.')
			return NULL;

		ptr++;
		if (*ptr == 0)
			return NULL;

		oid_aton_oid.subid_list[oid_aton_oid.subid_list_length++] = strtoul(ptr, &ptr, 0);
	}

	if (oid_aton_oid.subid_list_length < 2 || (oid_aton_oid.subid_list[0] * 40 + oid_aton_oid.subid_list[1]) > 0xFF)
		return NULL;

	return &oid_aton_oid;
}

static int oid_cmp(const oid_t *oid1, const oid_t *oid2)
{
	int subid1, subid2;
	size_t i;

	for (i = 0; i < MAX_NR_OIDS; i++) {
		subid1 = (oid1->subid_list_length > i) ? (int)oid1->subid_list[i] : -1;
		subid2 = (oid2->subid_list_length > i) ? (int)oid2->subid_list[i] : -1;

		if (subid1 == -1 && subid2 == -1)
			return 0;
		if (subid1 > subid2)
			return 1;
		if (subid1 < subid2)
			return -1;
	}

	return 0;
}

static int split(const char *str, char *delim, char **list, int max_list_length)
{
	int len = 0;
	char *ptr;
	char *buf = strdup(str);

	if (!buf)
		return 0;

	for (ptr = strtok(buf, delim); ptr; ptr = strtok(NULL, delim)) {
		if (len < max_list_length)
			list[len++] = strdup(ptr);
	}

	free(buf);

	return len;
}

static client_t *find_oldest_client(void)
{
	size_t i, found = 0, pos = 0;
	time_t timestamp = (time_t)LONG_MAX;

	for (i = 0; i < g_tcp_client_list_length; i++) {
		if (timestamp > g_tcp_client_list[i]->timestamp) {
			timestamp = g_tcp_client_list[i]->timestamp;
			found = 1;
			pos = i;
		}
	}

	return found ? g_tcp_client_list[pos] : NULL;
}

#define SNMP_VERSION_1_ERROR(resp, code, index) {                       \
	(resp)->error_status = code;                                    \
	(resp)->error_index  = index;                                   \
	return 0;                                                       \
}

#define SNMP_VERSION_2_ERROR(resp, req, index, err) {                   \
	size_t len = (resp)->value_list_length;                         \
	memcpy(&(resp)->value_list[len].oid, &(req)->oid_list[index],   \
	       sizeof((req)->oid_list[index]));                         \
	memcpy(&(resp)->value_list[len].data, &err, sizeof(err));       \
	(resp)->value_list_length++;                                    \
	continue;                                                       \
}

#define SNMP_GET_ERROR(resp, req, index, code, err, msg) {              \
	if ((req)->version == SNMP_VERSION_1)                           \
		SNMP_VERSION_1_ERROR((resp), (code), (index));          \
									\
	if ((resp)->value_list_length < MAX_NR_VALUES)                  \
		SNMP_VERSION_2_ERROR((resp), (req), (index), err);      \
									\
	lprintf(LOG_ERR, "%s", msg);                                    \
	return -1;                                                      \
}

static const data_t m_null              = { (unsigned char *)"\x05\x00", 2, 2 };
static const data_t m_no_such_object    = { (unsigned char *)"\x80\x00", 2, 2 };
static const data_t m_no_such_instance  = { (unsigned char *)"\x81\x00", 2, 2 };
static const data_t m_end_of_mib_view   = { (unsigned char *)"\x82\x00", 2, 2 };


static int decode_len(const unsigned char *packet, size_t size, size_t *pos, int *type, size_t *len)
{
	size_t length_of_len;

	if (*pos >= size) {
		lprintf(LOG_DEBUG, "underflow for element type");
		errno = EINVAL;
		return -1;
	}

	/* Fetch the ASN.1 element type (only subset of universal tags supported) */
	switch (packet[*pos]) {
		case BER_TYPE_BOOLEAN:
		case BER_TYPE_INTEGER:
		case BER_TYPE_BIT_STRING:
		case BER_TYPE_OCTET_STRING:
		case BER_TYPE_NULL:
		case BER_TYPE_OID:
		case BER_TYPE_SEQUENCE:
		case BER_TYPE_COUNTER:
		case BER_TYPE_GAUGE:
		case BER_TYPE_TIME_TICKS:
		case BER_TYPE_NO_SUCH_OBJECT:
		case BER_TYPE_NO_SUCH_INSTANCE:
		case BER_TYPE_END_OF_MIB_VIEW:
		case BER_TYPE_SNMP_GET:
		case BER_TYPE_SNMP_GETNEXT:
		case BER_TYPE_SNMP_RESPONSE:
		case BER_TYPE_SNMP_SET:
		case BER_TYPE_SNMP_GETBULK:
		case BER_TYPE_SNMP_INFORM:
		case BER_TYPE_SNMP_TRAP:
			*type = packet[*pos];
			*pos = *pos + 1;
			break;

		default:
			lprintf(LOG_DEBUG, "unsupported element type %02X", packet[*pos]);
			errno = EINVAL;
			return -1;
	}

	if (*pos >= size) {
		lprintf(LOG_DEBUG, "underflow for element length");
		errno = EINVAL;
		return -1;
	}

	/* Fetch the ASN.1 element length (only lengths up to 16 bit supported) */
	if (!(packet[*pos] & 0x80)) {
		*len = packet[*pos];
		*pos = *pos + 1;
	} else {
		length_of_len = packet[*pos] & 0x7F;
		if (length_of_len > 2) {
			lprintf(LOG_DEBUG, "overflow for element length");
			errno = EINVAL;
			return -1;
		}

		*pos = *pos + 1;
		*len = 0;
		while (length_of_len--) {
			if (*pos >= size) {
				lprintf(LOG_DEBUG, "underflow for element length");
				errno = EINVAL;
				return -1;
			}

			*len = (*len << 8) + packet[*pos];
			*pos = *pos + 1;
		}
	}

	return 0;
}

/* Fetch the value as unsigned integer (copy sign bit into all bytes first) */
static int decode_int(const unsigned char *packet, size_t size, size_t *pos, size_t len, int *value)
{
	unsigned int tmp;

	if (*pos >= (size - len + 1)) {
		lprintf(LOG_DEBUG, "underflow for integer");
		errno = EINVAL;
		return -1;
	}

	memset(&tmp, (packet[*pos] & 0x80) ? 0xFF : 0x00, sizeof(tmp));
	while (len--) {
		tmp = (tmp << 8) | packet[*pos];
		*pos = *pos + 1;
	}
	*(int *)value = tmp;

	return 0;
}

/* Fetch the value as unsigned integer (copy sign bit into all bytes first) */
static int decode_cnt(const unsigned char *packet, size_t size, size_t *pos, size_t len, uint32_t *value)
{
	if (*pos >= (size - len + 1)) {
		lprintf(LOG_DEBUG, "underflow for unsigned");
		errno = EINVAL;
		return -1;
	}

	*value = 0;
	while (len--) {
		*value = (*value << 8) | packet[*pos];
		*pos = *pos + 1;
	}

	return 0;
}

/* Fetch the value as C string (user must have made sure the length is ok) */
static int decode_str(const unsigned char *packet, size_t size, size_t *pos, size_t len, char *str, size_t str_len)
{
	if (*pos >= (size - len + 1)) {
		lprintf(LOG_DEBUG, "underflow for string");
		errno = EINVAL;
		return -1;
	}

	snprintf(str, str_len, "%.*s", (int)len, &packet[*pos]);
	*pos = *pos + len;

	return 0;
}

/* Fetch the value as C string (user must have made sure the length is ok) */
static int decode_oid(const unsigned char *packet, size_t size, size_t *pos, size_t len, oid_t *value)
{
	if (*pos >= (size - len + 1)) {
		lprintf(LOG_DEBUG, "underflow for oid");
		errno = EINVAL;
		return -1;
	}

	value->encoded_length = len;
	if (len > 0xFFFF) {
		lprintf(LOG_ERR, "could not decode: internal error");
		return -1;
	}

	if (len > 0xFF)
		value->encoded_length += 4;
	else if (len > 0x7F)
		value->encoded_length += 3;
	else
		value->encoded_length += 2;

	value->subid_list_length = 0;
	if (!len) {
		lprintf(LOG_DEBUG, "underflow for OID startbyte");
		errno = EINVAL;
		return -1;
	}

	if (packet[*pos] & 0x80) {
		lprintf(LOG_DEBUG, "unsupported OID startbyte %02X", packet[*pos]);
		errno = EINVAL;
		return -1;
	}

	value->subid_list[value->subid_list_length++] = packet[*pos] / 40;
	value->subid_list[value->subid_list_length++] = packet[*pos] % 40;
	*pos = *pos + 1;
	len--;

	while (len) {
		if (value->subid_list_length >= MAX_NR_SUBIDS) {
			lprintf(LOG_DEBUG, "overflow for OID byte");
			errno = EFAULT;
			return -1;
		}

		value->subid_list[value->subid_list_length] = 0;
		while (len--) {
			value->subid_list[value->subid_list_length]
				= (value->subid_list[value->subid_list_length] << 7) + (packet[*pos] & 0x7F);
			if (packet[*pos] & 0x80) {
				if (!len) {
					lprintf(LOG_DEBUG, "underflow for OID byte");
					errno = EINVAL;
					return -1;
				}
				*pos = *pos + 1;
			} else {
				*pos = *pos + 1;
				break;
			}
		}
		value->subid_list_length++;
	}

	return 0;
}

/* Fetch the value as pointer (user must make sure not to overwrite packet) */
static int decode_ptr(size_t size, size_t *pos, int len)
{
	if (*pos >= (size - len + 1)) {
		lprintf(LOG_DEBUG, "underflow for ptr");
		errno = EINVAL;
		return -1;
	}

	*pos = *pos + len;

	return 0;
}

static int decode_snmp_request(request_t *request, client_t *client)
{
	int type;
	size_t pos = 0, len = 0;
	static const char header_msg[]  = "Unexpected SNMP header";
	static const char error_msg[]   = "Unexpected SNMP error";
	static const char request_msg[] = "Unexpected SNMP request";
	static const char varbind_msg[] = "Unexpected SNMP varbindings";
	static const char commun_msg[]  = "SNMP community";
	static const char version_msg[] = "SNMP version";

	/* The SNMP message is enclosed in a sequence */
	if (decode_len(client->packet, client->size, &pos, &type, &len) == -1)
		return -1;

	if (type != BER_TYPE_SEQUENCE || len != (client->size - pos)) {
		lprintf(LOG_DEBUG, "%s type %02X length %zu", header_msg, type, len);
		errno = EINVAL;
		return -1;
	}

	/* The first element of the sequence is the version */
	if (decode_len(client->packet, client->size, &pos, &type, &len) == -1)
		return -1;

	if (type != BER_TYPE_INTEGER || len != 1) {
		lprintf(LOG_DEBUG, "Unexpected %s type %02X length %zu", version_msg, type, len);
		errno = EINVAL;
		return -1;
	}

	if (decode_int(client->packet, client->size, &pos, len, &request->version) == -1)
		return -1;

	if (request->version != SNMP_VERSION_1 && request->version != SNMP_VERSION_2C) {
		lprintf(LOG_DEBUG, "Unsupported %s %d", version_msg, request->version);
		errno = EINVAL;
		return -1;
	}

	/* The second element of the sequence is the community string */
	if (decode_len(client->packet, client->size, &pos, &type, &len) == -1)
		return -1;

	if (type != BER_TYPE_OCTET_STRING || len >= sizeof(request->community)) {
		lprintf(LOG_DEBUG, "Unexpected %s type %02X length %zu", commun_msg, type, len);
		errno = EINVAL;
		return -1;
	}

	if (decode_str(client->packet, client->size, &pos, len, request->community, sizeof(request->community)) == -1)
		return -1;

	if (strlen(request->community) < 1) {
		lprintf(LOG_DEBUG, "unsupported %s '%s'", commun_msg, request->community);
		errno = EINVAL;
		return -1;
	}

	/* The third element of the sequence is the SNMP request */
	if (decode_len(client->packet, client->size, &pos, &type, &len) == -1)
		return -1;

	if (len != (client->size - pos)) {
		lprintf(LOG_DEBUG, "%s type type %02X length %zu", request_msg, type, len);
		errno = EINVAL;
		return -1;
	}
	request->type = type;

	/* The first element of the SNMP request is the request ID */
	if (decode_len(client->packet, client->size, &pos, &type, &len) == -1)
		return -1;

	if (type != BER_TYPE_INTEGER || len < 1) {
		lprintf(LOG_DEBUG, "%s id type %02X length %zu", request_msg, type, len);
		errno = EINVAL;
		return -1;
	}

	if (decode_int(client->packet, client->size, &pos, len, &request->id) == -1)
		return -1;

	/* The second element of the SNMP request is the error state / non repeaters (0..2147483647) */
	if (decode_len(client->packet, client->size, &pos, &type, &len) == -1)
		return -1;

	if (type != BER_TYPE_INTEGER || len < 1) {
		lprintf(LOG_DEBUG, "%s state type %02X length %zu", error_msg, type, len);
		errno = EINVAL;
		return -1;
	}

	if (decode_cnt(client->packet, client->size, &pos, len, &request->non_repeaters) == -1)
		return -1;

	/* The third element of the SNMP request is the error index / max repetitions (0..2147483647) */
	if (decode_len(client->packet, client->size, &pos, &type, &len) == -1)
		return -1;

	if (type != BER_TYPE_INTEGER || len < 1) {
		lprintf(LOG_DEBUG, "%s index type %02X length %zu", error_msg, type, len);
		errno = EINVAL;
		return -1;
	}

	if (decode_cnt(client->packet, client->size, &pos, len, &request->max_repetitions) == -1)
		return -1;

	/* The fourth element of the SNMP request are the variable bindings */
	if (decode_len(client->packet, client->size, &pos, &type, &len) == -1)
		return -1;

	if (type != BER_TYPE_SEQUENCE || len != (client->size - pos)) {
		lprintf(LOG_DEBUG, "%s type %02X length %zu", varbind_msg, type, len);
		errno = EINVAL;
		return -1;
	}

	/* Loop through the variable bindings */
	request->oid_list_length = 0;
	while (pos < client->size) {
		/* If there is not enough room in the OID list, bail out now */
		if (request->oid_list_length >= MAX_NR_OIDS) {
			lprintf(LOG_DEBUG, "Overflow in OID list");
			errno = EFAULT;
			return -1;
		}

		/* Each variable binding is a sequence describing the variable */
		if (decode_len(client->packet, client->size, &pos, &type, &len) == -1)
			return -1;

		if (type != BER_TYPE_SEQUENCE || len < 1) {
			lprintf(LOG_DEBUG, "%s type %02X length %zu", varbind_msg, type, len);
			errno = EINVAL;
			return -1;
		}

		/* The first element of the variable binding is the OID */
		if (decode_len(client->packet, client->size, &pos, &type, &len) == -1)
			return -1;

		if (type != BER_TYPE_OID || len < 1) {
			lprintf(LOG_DEBUG, "%s OID type %02X length %zu", varbind_msg, type, len);
			errno = EINVAL;
			return -1;
		}

		if (decode_oid(client->packet, client->size, &pos, len, &request->oid_list[request->oid_list_length]) == -1)
			return -1;

		/* The second element of the variable binding is the new type and value */
		if (decode_len(client->packet, client->size, &pos, &type, &len) == -1)
			return -1;

		if ((type == BER_TYPE_NULL && len) || (type != BER_TYPE_NULL && !len)) {
			lprintf(LOG_DEBUG, "%s value type %02X length %zu", varbind_msg, type, len);
			errno = EINVAL;
			return -1;
		}

		if (decode_ptr(client->size, &pos, len) == -1)
			return -1;

		/* Now the OID list has one more entry */
		request->oid_list_length++;
	}

	return 0;
}


static size_t get_intlen(int val)
{
	if (val < -8388608 || val > 8388607)
		return 6;
	if (val < -32768 || val > 32767)
		return 5;
	if (val < -128 || val > 127)
		return 4;

	return 3;
}

static size_t get_strlen(const char *str)
{
	size_t len = strlen(str);

	if (len > 0xFFFF)
		return MAX_PACKET_SIZE;
	if (len > 0xFF)
		return len + 4;
	if (len > 0x7F)
		return len + 3;

	return len + 2;
}

static size_t get_hdrlen(size_t len)
{
	if (len > 0xFFFF)
		return MAX_PACKET_SIZE;
	if (len > 0xFF)
		return 4;
	if (len > 0x7F)
		return 3;

	return 2;
}

static int encode_snmp_integer(unsigned char *buf, int val)
{
	size_t len;

	if (val < -8388608 || val > 8388607)
		len = 4;
	else if (val < -32768 || val > 32767)
		len = 3;
	else if (val < -128 || val > 127)
		len = 2;
	else
		len = 1;

	*buf++ = BER_TYPE_INTEGER;
	*buf++ = len;
	while (len--)
		*buf++ = ((unsigned int)val >> (8 * len)) & 0xFF;

	return 0;
}

static int encode_snmp_string(unsigned char *buf, const char *str)
{
	size_t len;

	len = strlen(str);
	if (len > 0xFFFF)
		return -1;

	*buf++ = BER_TYPE_OCTET_STRING;
	if (len > 0xFF) {
		*buf++ = 0x82;
		*buf++ = (len >> 8) & 0xFF;
		*buf++ = len & 0xFF;
	} else if (len > 0x7F) {
		*buf++= 0x81;
		*buf++ = len & 0xFF;
	} else {
		*buf++ = len & 0x7F;
	}
	memcpy(buf, str, len);

	return 0;
}

static int encode_snmp_sequence_header(unsigned char *buf, size_t len, int type)
{
	if (len > 0xFFFF)
		return -1;

	*buf++ = type;
	if (len > 0xFF) {
		*buf++ = 0x82;
		*buf++ = (len >> 8) & 0xFF;
		*buf++ = len & 0xFF;
	} else if (len > 0x7F) {
		*buf++= 0x81;
		*buf++ = len & 0xFF;
	} else {
		*buf++ = len & 0x7F;
	}

	return 0;
}

static int encode_snmp_oid(unsigned char *buf, const oid_t *oid)
{
	size_t i, len;

	len = 1;
	for (i = 2; i < oid->subid_list_length; i++) {
		if (oid->subid_list[i] >= (1 << 28))
			len += 5;
		else if (oid->subid_list[i] >= (1 << 21))
			len += 4;
		else if (oid->subid_list[i] >= (1 << 14))
			len += 3;
		else if (oid->subid_list[i] >= (1 << 7))
			len += 2;
		else
			len += 1;
	}

	*buf++ = BER_TYPE_OID;
	if (len > 0xFFFF) {
		lprintf(LOG_ERR, "could not encode '%s': OID overflow", oid_ntoa(oid));
		return -1;
	}

	if (len > 0xFF) {
		*buf++ = 0x82;
		*buf++ = (len >> 8) & 0xFF;
		*buf++ = len & 0xFF;
	} else if (len > 0x7F) {
		*buf++ = 0x81;
		*buf++ = len & 0xFF;
	} else {
		*buf++ = len & 0x7F;
	}

	*buf++ = oid->subid_list[0] * 40 + oid->subid_list[1];
	for (i = 2; i < oid->subid_list_length; i++) {
		if (oid->subid_list[i] >= (1 << 28))
			len = 5;
		else if (oid->subid_list[i] >= (1 << 21))
			len = 4;
		else if (oid->subid_list[i] >= (1 << 14))
			len = 3;
		else if (oid->subid_list[i] >= (1 << 7))
			len = 2;
		else
			len = 1;

		while (len--) {
			if (len)
				*buf++ = ((oid->subid_list[i] >> (7 * len)) & 0x7F) | 0x80;
			else
				*buf++ = (oid->subid_list[i] >> (7 * len)) & 0x7F;
		}
	}

	return 0;
}

static int log_encoding_error(const char *what, const char *why)
{
	lprintf(LOG_ERR, "Failed encoding %s: %s", what, why);
	return -1;
}

static int encode_snmp_varbind(unsigned char *buf, size_t *pos, const value_t *value)
{
	size_t len;

	/* The value of the variable binding (NULL for error responses) */
	len = value->data.encoded_length;
	if (*pos < len)
		return log_encoding_error(oid_ntoa(&value->oid), "DATA overflow");

	memcpy(&buf[*pos - len], value->data.buffer, len);
	*pos = *pos - len;

	/* The OID of the variable binding */
	len = value->oid.encoded_length;
	if (*pos < len)
		return log_encoding_error(oid_ntoa(&value->oid), "OID overflow");

	encode_snmp_oid(&buf[*pos - len], &value->oid);
	*pos = *pos - len;

	/* The sequence header (type and length) of the variable binding */
	len = get_hdrlen(value->oid.encoded_length + value->data.encoded_length);
	if (*pos < len)
		return log_encoding_error(oid_ntoa(&value->oid), "VARBIND overflow");

	encode_snmp_sequence_header(&buf[*pos - len], value->oid.encoded_length + value->data.encoded_length, BER_TYPE_SEQUENCE);
	*pos = *pos - len;

	return 0;
}

static int encode_snmp_response(request_t *request, response_t *response, client_t *client)
{
	size_t i, len, pos;

	/* If there was an error, we have to encode the original varbind list, but
	 * omit any varbind values (replace them with NULL values)
	 */
	if (response->error_status != SNMP_STATUS_OK) {
		if (request->oid_list_length > MAX_NR_VALUES)
			return log_encoding_error("SNMP response", "value list overflow");

		for (i = 0; i < request->oid_list_length; i++) {
			memcpy(&response->value_list[i].oid, &request->oid_list[i], sizeof(request->oid_list[i]));
			memcpy(&response->value_list[i].data, &m_null, sizeof(m_null));
		}
		response->value_list_length = request->oid_list_length;
	}

	/* To make the code more compact and save processing time, we are encoding the
	 * data beginning at the last byte of the buffer backwards. Thus, the encoded
	 * packet will not be positioned at offset 0..(size-1) of the client's packet
	 * buffer, but at offset (bufsize-size..bufsize-1)!
	 */
	pos = MAX_PACKET_SIZE;
	for (i = response->value_list_length; i > 0; i--) {
		if (encode_snmp_varbind(client->packet, &pos, &response->value_list[i-1]) == -1)
			return -1;
	}

	len = get_hdrlen(MAX_PACKET_SIZE - pos);
	if (pos < len)
		return log_encoding_error("SNMP response", "VARBINDS overflow");

	encode_snmp_sequence_header(&client->packet[pos - len], MAX_PACKET_SIZE - pos, BER_TYPE_SEQUENCE);
	pos = pos - len;

	len = get_intlen(response->error_index);
	if (pos < len)
		return log_encoding_error("SNMP response", "ERROR INDEX overflow");

	encode_snmp_integer(&client->packet[pos - len], response->error_index);
	pos = pos - len;

	len = get_intlen(response->error_status);
	if (pos < len)
		return log_encoding_error("SNMP response", "ERROR STATUS overflow");

	encode_snmp_integer(&client->packet[pos - len], response->error_status);
	pos = pos - len;

	len = get_intlen(request->id);
	if (pos < len)
		return log_encoding_error("SNMP response", "ID overflow");

	encode_snmp_integer(&client->packet[pos - len], request->id);
	pos = pos - len;

	len = get_hdrlen(MAX_PACKET_SIZE - pos);
	if (pos < len)
		return log_encoding_error("SNMP response", "PDU overflow");

	encode_snmp_sequence_header(&client->packet[pos - len], MAX_PACKET_SIZE - pos, BER_TYPE_SNMP_RESPONSE);
	pos = pos - len;

	len = get_strlen(request->community);
	if (pos < len)
		return log_encoding_error("SNMP response", "COMMUNITY overflow");

	encode_snmp_string(&client->packet[pos - len], request->community);
	pos = pos - len;

	len = get_intlen(request->version);
	if (pos < len)
		return log_encoding_error("SNMP response", "VERSION overflow");

	encode_snmp_integer(&client->packet[pos - len], request->version);
	pos = pos - len;

	len = get_hdrlen(MAX_PACKET_SIZE - pos);
	if (pos < len)
		return log_encoding_error("SNMP response", "RESPONSE overflow");

	encode_snmp_sequence_header(&client->packet[pos - len], MAX_PACKET_SIZE - pos, BER_TYPE_SEQUENCE);
	pos = pos - len;

	/*
	 * Now move the packet to the start of the buffer so that the caller does not have
	 * to deal with this messy detail (the CPU cycles needed are worth their money!)
	 * and set up the packet size.
	 */
	if (pos > 0)
		memmove(&client->packet[0], &client->packet[pos], MAX_PACKET_SIZE - pos);
	client->size = MAX_PACKET_SIZE - pos;

	return 0;
}

/* Find the OID in the MIB that is exactly the given one or a subid */
static value_t *mib_find(const oid_t *oid, size_t *pos)
{
	while (*pos < g_mib_length) {
		value_t *curr = &g_mib[*pos];
		size_t len = oid->subid_list_length * sizeof(oid->subid_list[0]);

		if (curr->oid.subid_list_length >= oid->subid_list_length &&
		    !memcmp(curr->oid.subid_list, oid->subid_list, len))
			return curr;
		*pos = *pos + 1;
	}

	return NULL;
}

static int handle_snmp_get(request_t *request, response_t *response)
{
	size_t i, pos;
	value_t *value;
	static const char msg[] = "Failed handling SNMP GET: value list overflow";

	/*
	 * Search each varbinding of the request and append the value to the
	 * response. Note that if the length does not match, we might have found a
	 * subid of the requested one (table cell of table column)!
	 */
	for (i = 0; i < request->oid_list_length; i++) {
		pos = 0;
		value = mib_find(&request->oid_list[i], &pos);
		if (!value)
			SNMP_GET_ERROR(response, request, i, SNMP_STATUS_NO_SUCH_NAME, m_no_such_object, msg);

		if (pos >= g_mib_length)
			SNMP_GET_ERROR(response, request, i, SNMP_STATUS_NO_SUCH_NAME, m_no_such_object, msg);

		if (value->oid.subid_list_length == (request->oid_list[i].subid_list_length + 1))
			SNMP_GET_ERROR(response, request, i, SNMP_STATUS_NO_SUCH_NAME, m_no_such_instance, msg);

		if (value->oid.subid_list_length != request->oid_list[i].subid_list_length)
			SNMP_GET_ERROR(response, request, i, SNMP_STATUS_NO_SUCH_NAME, m_no_such_object, msg);

		if (response->value_list_length < MAX_NR_VALUES) {
			memcpy(&response->value_list[response->value_list_length], value, sizeof(*value));
			response->value_list_length++;
			continue;
		}

		lprintf(LOG_ERR, "%s", msg);
		return -1;
	}

	return 0;
}

/* Find the OID in the MIB that is the one after the given one */
static value_t *mib_findnext(const oid_t *oid)
{
	size_t pos;

	for (pos = 0; pos < g_mib_length; pos++) {
		if (oid_cmp(&g_mib[pos].oid, oid) > 0)
			return &g_mib[pos];
	}

	return NULL;
}

static int handle_snmp_getnext(request_t *request, response_t *response)
{
	size_t i;
	value_t *value;
	static const char msg[] = "Failed handling SNMP GETNEXT: value list overflow";

	/*
	 * Search each varbinding of the request and append the value to the
	 * response. Note that if the length does not match, we might have found a
	 * subid of the requested one (table cell of table column)!
	 */
	for (i = 0; i < request->oid_list_length; i++) {
		value = mib_findnext(&request->oid_list[i]);
		if (!value)
			SNMP_GET_ERROR(response, request, i, SNMP_STATUS_NO_SUCH_NAME, m_end_of_mib_view, msg);

		if (response->value_list_length < MAX_NR_VALUES) {
			memcpy(&response->value_list[response->value_list_length], value, sizeof(*value));
			response->value_list_length++;
			continue;
		}

		lprintf(LOG_ERR, "%s", msg);
		return -1;
	}

	return 0;
}

static int handle_snmp_set(request_t *request, response_t *response)
{
	SNMP_VERSION_1_ERROR(response, (request->version == SNMP_VERSION_1)
			     ? SNMP_STATUS_NO_SUCH_NAME : SNMP_STATUS_NO_ACCESS, 0);
}

static int handle_snmp_getbulk(request_t *request, response_t *response)
{
	size_t i, j;
	oid_t oid_list[MAX_NR_OIDS];
	value_t *value;
	static const char msg[] = "Failed handling SNMP GETBULK: value list overflow";

	/* Make a local copy of the OID list since we are going to modify it */
	memcpy(oid_list, request->oid_list, sizeof(request->oid_list));

	/* The non-repeaters are handled like with the GETNEXT request */
	for (i = 0; i < request->oid_list_length; i++) {
		if (i >= request->non_repeaters)
			break;

		value = mib_findnext(&oid_list[i]);
		if (!value)
			SNMP_GET_ERROR(response, request, i, SNMP_STATUS_NO_SUCH_NAME, m_end_of_mib_view, msg);

		if (response->value_list_length < MAX_NR_VALUES) {
			memcpy(&response->value_list[response->value_list_length], value, sizeof(*value));
			response->value_list_length++;
			continue;
		}

		lprintf(LOG_ERR, "%s", msg);
		return -1;
	}

	/*
	 * The repeaters are handled like with the GETNEXT request, except that:
	 *
	 * - the access is interleaved (i.e. first repetition of all varbinds,
	 *   then second repetition of all varbinds, then third,...)
	 * - the repetitions are aborted as soon as there is no successor found
	 *   for all of the varbinds
	 * - other than with getnext, the last variable in the MIB is named if
	 *   the variable queried is not after the end of the MIB
	 */
	for (j = 0; j < request->max_repetitions; j++) {
		int found_repeater = 0;

		for (i = request->non_repeaters; i < request->oid_list_length; i++) {
			value = mib_findnext(&oid_list[i]);
			if (!value)
				SNMP_GET_ERROR(response, request, i, SNMP_STATUS_NO_SUCH_NAME, m_end_of_mib_view, msg);

			if (response->value_list_length < MAX_NR_VALUES) {
				memcpy(&response->value_list[response->value_list_length], value, sizeof(*value));
				response->value_list_length++;
				memcpy(&oid_list[i], &value->oid, sizeof(value->oid));
				found_repeater++;
				continue;
			}

			lprintf(LOG_ERR, "%s", msg);
			return -1;
		}

		if (found_repeater == 0)
			break;
	}

	return 0;
}


static int snmp_packet_complete(const client_t *client)
{
	int type;
	size_t pos = 0, len = 0;

	/*
	 * The SNMP message must be at least have a header containing sequence,
	 * version, community, sequence, request id, 2 integers, sequence, oid
	 * and null value.
	 */
	if (client->size < 25)
		return 0;

	/* The SNMP message is enclosed in a sequence */
	if (decode_len(client->packet, client->size, &pos, &type, &len) == -1)
		return -1;

	if (type != BER_TYPE_SEQUENCE || len < 1 || len > (client->size - pos)) {
		lprintf(LOG_DEBUG, "Unexpected SNMP header type %02X length %zu", type, len);
		errno = EINVAL;
		return -1;
	}

	/* Return whether we received the whole packet */
	return ((client->size - pos) == len) ? 1 : 0;
}

static int snmp(client_t *client)
{
	response_t response;
	request_t request;

	/* Setup request and response (other code only changes non-defaults) */
	memset(&request, 0, sizeof(request));
	memset(&response, 0, sizeof(response));

	/* Decode the request (only checks for syntax of the packet) */
	if (decode_snmp_request(&request, client) == -1)
		return -1;

	/*
	 * If we are using SNMP v2c or require authentication, check the community
	 * string for length and validity.
	 */
	if (request.version == SNMP_VERSION_2C) {
		if (strcmp(g_community, request.community)) {
			response.error_status = (request.version == SNMP_VERSION_2C) ? SNMP_STATUS_NO_ACCESS : SNMP_STATUS_GEN_ERR;
			response.error_index = 0;
			goto done;
		}
	} else if (g_auth) {
		response.error_status = SNMP_STATUS_GEN_ERR;
		response.error_index = 0;
		goto done;
	}

	/* Now handle the SNMP requests depending on their type */
	switch (request.type) {
		case BER_TYPE_SNMP_GET:
			if (handle_snmp_get(&request, &response) == -1)
				return -1;
			break;

		case BER_TYPE_SNMP_GETNEXT:
			if (handle_snmp_getnext(&request, &response) == -1)
				return -1;
			break;

		case BER_TYPE_SNMP_SET:
			if (handle_snmp_set(&request, &response) == -1)
				return -1;
			break;

		case BER_TYPE_SNMP_GETBULK:
			if (handle_snmp_getbulk(&request, &response) == -1)
				return -1;
			break;

		default:
			lprintf(LOG_ERR, "UNHANDLED REQUEST TYPE %d", request.type);
			client->size = 0;
			return 0;
	}

done:
	/* Encode the request (depending on error status and encode flags) */
	if (encode_snmp_response(&request, &response, client) == -1)
		return -1;

	return 0;
}

/*
 * Module variables
 *
 * To extend the MIB, add the definition of the SNMP table here. Note that the
 * variables use OIDs that have two subids more, which both are specified in the
 * mib_build_entry() and mib_build_entries() function calls. For example, the
 * system table uses the OID .1.3.6.1.2.1.1, the first system table variable,
 * system.sysDescr.0 (using OID .1.3.6.1.2.1.1.1.0) is appended to the MIB using
 * the function call mib_build_entry(&m_system_oid, 1, 0, ...).
 *
 * The first parameter is the array containing the list of subids (up to 14 here),
 * the next is the number of subids. The last parameter is the length that this
 * OID will need encoded in SNMP packets (including the BER type and length fields).
 */

static const oid_t m_system_oid         = { { 1, 3, 6, 1, 2, 1, 1               }, 7, 8  };
static const oid_t m_if_1_oid           = { { 1, 3, 6, 1, 2, 1, 2               }, 7, 8  };
static const oid_t m_if_2_oid           = { { 1, 3, 6, 1, 2, 1, 2, 2, 1         }, 9, 10 };
static const oid_t m_host_oid           = { { 1, 3, 6, 1, 2, 1, 25, 1           }, 8, 9  };
static const oid_t m_memory_oid         = { { 1, 3, 6, 1, 4, 1, 2021, 4,        }, 8, 10 };
static const oid_t m_disk_oid           = { { 1, 3, 6, 1, 4, 1, 2021, 9, 1      }, 9, 11 };
static const oid_t m_load_oid           = { { 1, 3, 6, 1, 4, 1, 2021, 10, 1     }, 9, 11 };
static const oid_t m_cpu_oid            = { { 1, 3, 6, 1, 4, 1, 2021, 11        }, 8, 10 };

static const int m_load_avg_times[3] = { 1, 5, 15 };

static int encode_integer(data_t *data, int integer_value)
{
	unsigned char *buffer;
	int length;

	buffer = data->buffer;
	if (integer_value < -8388608 || integer_value > 8388607)
		length = 4;
	else if (integer_value < -32768 || integer_value > 32767)
		length = 3;
	else if (integer_value < -128 || integer_value > 127)
		length = 2;
	else
		length = 1;

	*buffer++ = BER_TYPE_INTEGER;
	*buffer++ = length;
	while (length--)
		*buffer++ = ((unsigned int)integer_value >> (8 * length)) & 0xFF;

	data->encoded_length = buffer - data->buffer;

	return 0;
}

static int encode_string(data_t *data, const char *string)
{
	size_t len;
	unsigned char *buffer;

	if (!string)
		return 2;

	len = strlen(string);
	if ((len + 4) > data->max_length) {
		data->max_length = len + 4;
		data->buffer = realloc(data->buffer, data->max_length);
		if (!data->buffer)
			return 2;
	}

	if (len > 0xFFFF) {
		lprintf(LOG_ERR, "Failed encoding '%s': string overflow", string);
		return -1;
	}

	buffer    = data->buffer;
	*buffer++ = BER_TYPE_OCTET_STRING;
	if (len > 255) {
		*buffer++ = 0x82;
		*buffer++ = (len >> 8) & 0xFF;
		*buffer++ = len & 0xFF;
	} else if (len > 127) {
		*buffer++ = 0x81;
		*buffer++ = len & 0xFF;
	} else {
		*buffer++ = len & 0x7F;
	}

	while (*string)
		*buffer++ = (unsigned char)(*string++);

	data->encoded_length = buffer - data->buffer;

	return 0;
}

static int encode_oid(data_t *data, const oid_t *oid)
{
	size_t i, len = 1;
	unsigned char *buffer = data->buffer;

	if (!oid)
		return 2;

	for (i = 2; i < oid->subid_list_length; i++) {
		if (oid->subid_list[i] >= (1 << 28))
			len += 5;
		else if (oid->subid_list[i] >= (1 << 21))
			len += 4;
		else if (oid->subid_list[i] >= (1 << 14))
			len += 3;
		else if (oid->subid_list[i] >= (1 << 7))
			len += 2;
		else
			len += 1;
	}

	if (len > 0xFFFF) {
		lprintf(LOG_ERR, "Failed encoding '%s': OID overflow", oid_ntoa(oid));
		return -1;
	}

	*buffer++ = BER_TYPE_OID;
	if (len > 0xFF) {
		*buffer++ = 0x82;
		*buffer++ = (len >> 8) & 0xFF;
		*buffer++ = len & 0xFF;
	} else if (len > 0x7F) {
		*buffer++ = 0x81;
		*buffer++ = len & 0xFF;
	} else {
		*buffer++ = len & 0x7F;
	}

	*buffer++ = oid->subid_list[0] * 40 + oid->subid_list[1];
	for (i = 2; i < oid->subid_list_length; i++) {
		if (oid->subid_list[i] >= (1 << 28))
			len = 5;
		else if (oid->subid_list[i] >= (1 << 21))
			len = 4;
		else if (oid->subid_list[i] >= (1 << 14))
			len = 3;
		else if (oid->subid_list[i] >= (1 << 7))
			len = 2;
		else
			len = 1;

		while (len--) {
			if (len)
				*buffer++ = ((oid->subid_list[i] >> (7 * len)) & 0x7F) | 0x80;
			else
				*buffer++ = (oid->subid_list[i] >> (7 * len)) & 0x7F;
		}
	}

	data->encoded_length = buffer - data->buffer;

	return 0;
}

static int encode_unsigned(data_t *data, int type, unsigned int ticks_value)
{
	unsigned char *buffer;
	int length;

	buffer = data->buffer;
	if (ticks_value & 0xFF000000)
		length = 4;
	else if (ticks_value & 0x00FF0000)
		length = 3;
	else if (ticks_value & 0x0000FF00)
		length = 2;
	else
		length = 1;

	*buffer++ = type;
	*buffer++ = length;
	while (length--)
		*buffer++ = (ticks_value >> (8 * length)) & 0xFF;

	data->encoded_length = buffer - data->buffer;

	return 0;
}

/* Create OID from the given prefix, column, and row */
static int oid_build(oid_t *oid, const oid_t *prefix, int column, int row)
{
	memcpy(oid, prefix, sizeof(*oid));

	if (oid->subid_list_length >= MAX_NR_SUBIDS)
		return -1;

	oid->subid_list[oid->subid_list_length++] = column;

	if (oid->subid_list_length >= MAX_NR_SUBIDS)
		return -1;

	oid->subid_list[oid->subid_list_length++] = row;

	return 0;
}

/* Create a data buffer for the value depending on the type:
 *
 * - strings and oids are assumed to be static or have the maximum allowed length
 * - integers are assumed to be dynamic and don't have more than 32 bits
 */
static int data_alloc(data_t *data, int type)
{
	switch (type) {
		case BER_TYPE_INTEGER:
			data->max_length = sizeof(int) + 2;
			data->encoded_length = 0;
			data->buffer = xmalloc(data->max_length);
			break;

		case BER_TYPE_OCTET_STRING:
			data->max_length = 4;
			data->encoded_length = 0;
			data->buffer = xmalloc(data->max_length);
			break;

		case BER_TYPE_OID:
			data->max_length = MAX_NR_SUBIDS * 5 + 4;
			data->encoded_length = 0;
			data->buffer = xmalloc(data->max_length);
			break;

		case BER_TYPE_COUNTER:
		case BER_TYPE_GAUGE:
		case BER_TYPE_TIME_TICKS:
			data->max_length = sizeof(unsigned int) + 2;
			data->encoded_length = 0;
			data->buffer = xmalloc(data->max_length);
			break;

		default:
			return -1;
	}

	data->buffer[0] = type;
	data->buffer[1] = 0;
	data->buffer[2] = 0;
	data->encoded_length = 3;

	return 0;
}

/*
 * Calculate the encoded length of the created OID (note: first the length
 * of the subid list, then the length of the length/type header!)
 */
static int encode_oid_len(oid_t *oid)
{
	uint32_t len = 1;
	size_t i;

	for (i = 2; i < oid->subid_list_length; i++) {
		if (oid->subid_list[i] >= (1 << 28))
			len += 5;
		else if (oid->subid_list[i] >= (1 << 21))
			len += 4;
		else if (oid->subid_list[i] >= (1 << 14))
			len += 3;
		else if (oid->subid_list[i] >= (1 << 7))
			len += 2;
		else
			len += 1;
	}

	if (len > 0xFFFF) {
		lprintf(LOG_ERR, "Failed encoding '%s': OID overflow", oid_ntoa(oid));
		oid->encoded_length = -1;
		return -1;
	}

	if (len > 0xFF)
		len += 4;
	else if (len > 0x7F)
		len += 3;
	else
		len += 2;

	oid->encoded_length = (short)len;

	return 0;
}

static value_t *mib_alloc_entry(const oid_t *prefix, int column, int row, int type)
{
	int ret;
	value_t *value;
	static const char msg[] = "Failed creating MIB entry";

	/* Create a new entry in the MIB table */
	if (g_mib_length >= MAX_NR_VALUES) {
		lprintf(LOG_ERR, "%s '%s.%d.%d': table overflow", msg, oid_ntoa(prefix), column, row);
		return NULL;
	}

	value = &g_mib[g_mib_length++];
	memcpy(&value->oid, prefix, sizeof(value->oid));

	/* Create the OID from the prefix, the column and the row */
	if (oid_build(&value->oid, prefix, column, row)) {
		lprintf(LOG_ERR, "%s '%s.%d.%d': oid overflow", msg, oid_ntoa(prefix), column, row);
		return NULL;
	}

	ret  = encode_oid_len(&value->oid);
	ret += data_alloc(&value->data, type);
	if (ret) {
		lprintf(LOG_ERR, "%s '%s.%d.%d': unsupported type %d", msg,
			oid_ntoa(&value->oid), column, row, type);
		return NULL;
	}

	return value;
}



/*
 * Set data buffer to its new value, depending on the type.
 *
 * Note: we assume the buffer was allocated to hold the maximum possible
 *       value when the MIB was built.
 */
static int data_set(data_t *data, int type, const void *arg)
{
	/* Make sure to always initialize the buffer, in case of error below. */
	memset(data->buffer, 0, data->max_length);

	switch (type) {
		case BER_TYPE_INTEGER:
			return encode_integer(data, (intptr_t)arg);

		case BER_TYPE_OCTET_STRING:
			return encode_string(data, (const char *)arg);

		case BER_TYPE_OID:
			return encode_oid(data, oid_aton((const char *)arg));

		case BER_TYPE_COUNTER:
		case BER_TYPE_GAUGE:
		case BER_TYPE_TIME_TICKS:
			return encode_unsigned(data, type, (uintptr_t)arg);

		default:
			break;  /* Fall through */
	}

	return 1;
}


static int mib_build_entries(const oid_t *prefix, int column, int row_from, int row_to, int type)
{
	int row;

	for (row = row_from; row <= row_to; row++) {
		if (!mib_alloc_entry(prefix, column, row, type))
			return -1;
	}

	return 0;
}


static int mib_data_set(const oid_t *oid, data_t *data, int column, int row, int type, const void *arg)
{
	int ret;
	static const char msg[] = "Failed assigning value to OID";

	ret = data_set(data, type, arg);
	if (ret) {
		if (ret == 1)
			lprintf(LOG_ERR, "%s '%s.%d.%d': unsupported type %d", msg, oid_ntoa(oid), column, row, type);
		else if (ret == 2)
			lprintf(LOG_ERR, "%s '%s.%d.%d': invalid default value", msg, oid_ntoa(oid), column, row);

		return -1;
	}

	return 0;
}

static int mib_build_entry(const oid_t *prefix, int column, int row, int type, const void *arg)
{
	value_t *value;

	value = mib_alloc_entry(prefix, column, row, type);
	if (!value)
		return -1;

	return mib_data_set(&value->oid, &value->data, column, row, type, arg);
}

/* -----------------------------------------------------------------------------
 * Interface functions
 *
 * To extend the MIB, add the relevant mib_build_entry() calls (to add one MIB
 * variable) or mib_build_entries() calls (to add a column of a MIB table) in
 * the mib_build() function. Note that building the MIB must be done strictly in
 * ascending OID order or the SNMP getnext/getbulk functions will not work as
 * expected!
 *
 * To extend the MIB, add the relevant mib_update_entry() calls (to update one
 * MIB variable or one cell in a MIB table) in the mib_update() function. Note
 * that the MIB variables must be added in the correct order (i.e. ascending).
 * How to get the value for that variable is up to you, but bear in mind that
 * the mib_update() function is called between receiving the request from the
 * client and sending back the response; thus you should avoid time-consuming
 * actions!
 *
 * The variable types supported up to now are OCTET_STRING, INTEGER (32 bit
 * signed), COUNTER (32 bit unsigned), TIME_TICKS (32 bit unsigned, in 1/10s)
 * and OID.
 *
 * Note that the maximum number of MIB variables is restricted by the length of
 * the MIB array, (see mini_snmpd.h for the value of MAX_NR_VALUES).
 */

static int mib_build(void)
{
	char hostname[MAX_STRING_SIZE];
	char name[16];
	size_t i;

	/* Determine some static values that are not known at compile-time */
	if (gethostname(hostname, sizeof(hostname)) == -1)
		hostname[0] = '\0';
	else if (hostname[sizeof(hostname) - 1] != '\0')
		hostname[sizeof(hostname) - 1] = '\0';

	/*
	 * The system MIB: basic info about the host (SNMPv2-MIB.txt)
	 * Caution: on changes, adapt the corresponding mib_update() section too!
	 */
	if (mib_build_entry(&m_system_oid, 1, 0, BER_TYPE_OCTET_STRING, g_description ?: "") == -1 ||
	    mib_build_entry(&m_system_oid, 2, 0, BER_TYPE_OID,          g_vendor)            == -1 ||
	   !mib_alloc_entry(&m_system_oid, 3, 0, BER_TYPE_TIME_TICKS)                              ||
	    mib_build_entry(&m_system_oid, 4, 0, BER_TYPE_OCTET_STRING, g_contact ?: "")     == -1 ||
	    mib_build_entry(&m_system_oid, 5, 0, BER_TYPE_OCTET_STRING, hostname)            == -1 ||
	    mib_build_entry(&m_system_oid, 6, 0, BER_TYPE_OCTET_STRING, g_location ?: "")    == -1)
		return -1;

	/*
	 * The interface MIB: network interfaces (IF-MIB.txt)
	 * Caution: on changes, adapt the corresponding mib_update() section too!
	 */
	if (g_interface_list_length > 0) {
		if (mib_build_entry(&m_if_1_oid, 1, 0, BER_TYPE_INTEGER, (const void *)(intptr_t)g_interface_list_length) == -1)
			return -1;

		/* ifIndex -- XXX: Should be system ifindex! */
		for (i = 0; i < g_interface_list_length; i++) {
			if (mib_build_entry(&m_if_2_oid, 1, i + 1, BER_TYPE_INTEGER, (const void *)(intptr_t)(i + 1)) == -1)
				return -1;
		}

		/* ifDescription */
		for (i = 0; i < g_interface_list_length; i++) {
			if (mib_build_entry(&m_if_2_oid, 2, i + 1, BER_TYPE_OCTET_STRING, g_interface_list[i]) == -1)
				return -1;
		}

		/* ifType: ENUM, ethernetCsmacd(6) <-- recommended for all types of Ethernets */
		for (i = 0; i < g_interface_list_length; i++) {
			if (mib_build_entry(&m_if_2_oid, 3, i + 1, BER_TYPE_INTEGER, (const void *)(intptr_t)6) == -1)
				return -1;
		}

		/* ifMtu */
		for (i = 0; i < g_interface_list_length; i++) {
			if (mib_build_entry(&m_if_2_oid, 4, i + 1, BER_TYPE_INTEGER, (const void *)(intptr_t)1500) == -1)
				return -1;
		}

		/* ifSpeed (in bps) */
		for (i = 0; i < g_interface_list_length; i++) {
			if (mib_build_entry(&m_if_2_oid, 5, i + 1, BER_TYPE_GAUGE, (const void *)(intptr_t)1000000000) == -1)
				return -1;
		}

		/* ifPhysAddress */
		for (i = 0; i < g_interface_list_length; i++) {
			unsigned char mac[] = { 0xc0, 0xff, 0xee, 0xde, 0xad, i + 1, 0x00 };

			if (mib_build_entry(&m_if_2_oid, 6, i + 1, BER_TYPE_OCTET_STRING, mac) == -1)
				return -1;
		}

		/* ifAdminStatus: up(1), down(2), testing(3) */
		for (i = 0; i < g_interface_list_length; i++) {
			if (mib_build_entry(&m_if_2_oid, 7, i + 1, BER_TYPE_INTEGER, (const void *)(intptr_t)1) == -1)
				return -1;
		}

		/* ifOperStatus: up(1), down(2), testing(3), unknown(4), dormant(5), notPresent(6), lowerLayerDown(7) */
		for (i = 0; i < g_interface_list_length; i++) {
			if (mib_build_entry(&m_if_2_oid, 8, i + 1, BER_TYPE_INTEGER, (const void *)(intptr_t)1) == -1)
				return -1;
		}

		/* ifLastChange */
		for (i = 0; i < g_interface_list_length; i++) {
			if (mib_build_entry(&m_if_2_oid, 9, i + 1, BER_TYPE_TIME_TICKS, (const void *)(intptr_t)0) == -1)
				return -1;
		}

		if (mib_build_entries(&m_if_2_oid, 10, 1, g_interface_list_length, BER_TYPE_COUNTER) == -1 ||
		    mib_build_entries(&m_if_2_oid, 11, 1, g_interface_list_length, BER_TYPE_COUNTER) == -1 ||
		    mib_build_entries(&m_if_2_oid, 13, 1, g_interface_list_length, BER_TYPE_COUNTER) == -1 ||
		    mib_build_entries(&m_if_2_oid, 14, 1, g_interface_list_length, BER_TYPE_COUNTER) == -1 ||
		    mib_build_entries(&m_if_2_oid, 16, 1, g_interface_list_length, BER_TYPE_COUNTER) == -1 ||
		    mib_build_entries(&m_if_2_oid, 17, 1, g_interface_list_length, BER_TYPE_COUNTER) == -1 ||
		    mib_build_entries(&m_if_2_oid, 19, 1, g_interface_list_length, BER_TYPE_COUNTER) == -1 ||
		    mib_build_entries(&m_if_2_oid, 20, 1, g_interface_list_length, BER_TYPE_COUNTER) == -1)
			return -1;
	}

	/*
	 * The host MIB: additional host info (HOST-RESOURCES-MIB.txt)
	 * Caution: on changes, adapt the corresponding mib_update() section too!
	 */
	if (!mib_alloc_entry(&m_host_oid, 1, 0, BER_TYPE_TIME_TICKS))
		return -1;

	/*
	 * The memory MIB: total/free memory (UCD-SNMP-MIB.txt)
	 * Caution: on changes, adapt the corresponding mib_update() section too!
	 */
	if (!mib_alloc_entry(&m_memory_oid,  5, 0, BER_TYPE_INTEGER) ||
	    !mib_alloc_entry(&m_memory_oid,  6, 0, BER_TYPE_INTEGER) ||
	    !mib_alloc_entry(&m_memory_oid, 13, 0, BER_TYPE_INTEGER) ||
	    !mib_alloc_entry(&m_memory_oid, 14, 0, BER_TYPE_INTEGER) ||
	    !mib_alloc_entry(&m_memory_oid, 15, 0, BER_TYPE_INTEGER))
		return -1;

	/*
	 * The disk MIB: mounted partitions (UCD-SNMP-MIB.txt)
	 * Caution: on changes, adapt the corresponding mib_update() section too!
	 */
	if (g_disk_list_length > 0) {
		for (i = 0; i < g_disk_list_length; i++) {
			if (mib_build_entry(&m_disk_oid, 1, i + 1, BER_TYPE_INTEGER, (const void *)(intptr_t)(i + 1)) == -1)
				return -1;
		}

		for (i = 0; i < g_disk_list_length; i++) {
			if (mib_build_entry(&m_disk_oid, 2, i + 1, BER_TYPE_OCTET_STRING, g_disk_list[i]) == -1)
				return -1;
		}

		if (mib_build_entries(&m_disk_oid,  6, 1, g_disk_list_length, BER_TYPE_INTEGER) == -1 ||
		    mib_build_entries(&m_disk_oid,  7, 1, g_disk_list_length, BER_TYPE_INTEGER) == -1 ||
		    mib_build_entries(&m_disk_oid,  8, 1, g_disk_list_length, BER_TYPE_INTEGER) == -1 ||
		    mib_build_entries(&m_disk_oid,  9, 1, g_disk_list_length, BER_TYPE_INTEGER) == -1 ||
		    mib_build_entries(&m_disk_oid, 10, 1, g_disk_list_length, BER_TYPE_INTEGER) == -1)
			return -1;
	}

	/*
	 * The load MIB: CPU load averages (UCD-SNMP-MIB.txt)
	 * Caution: on changes, adapt the corresponding mib_update() section too!
	 */
	for (i = 0; i < 3; i++) {
		if (mib_build_entry(&m_load_oid, 1, i + 1, BER_TYPE_INTEGER, (const void *)(intptr_t)(i + 1)) == -1)
			return -1;
	}

	for (i = 0; i < 3; i++) {
		snprintf(name, sizeof(name), "Load-%d", m_load_avg_times[i]);
		if (mib_build_entry(&m_load_oid, 2, i + 1, BER_TYPE_OCTET_STRING, name) == -1)
			return -1;
	}

	if (mib_build_entries(&m_load_oid, 3, 1, 3, BER_TYPE_OCTET_STRING) == -1)
		return -1;

	for (i = 0; i < 3; i++) {
		snprintf(name, sizeof(name), "%d", m_load_avg_times[i]);
		if (mib_build_entry(&m_load_oid, 4, i + 1, BER_TYPE_OCTET_STRING, name) == -1)
			return -1;
	}

	if (mib_build_entries(&m_load_oid, 5, 1, 3, BER_TYPE_INTEGER) == -1)
		return -1;

	/* The CPU MIB: CPU statistics (UCD-SNMP-MIB.txt)
	 * Caution: on changes, adapt the corresponding mib_update() section too!
	 */
	if (!mib_alloc_entry(&m_cpu_oid, 50, 0, BER_TYPE_COUNTER) ||
	    !mib_alloc_entry(&m_cpu_oid, 51, 0, BER_TYPE_COUNTER) ||
	    !mib_alloc_entry(&m_cpu_oid, 52, 0, BER_TYPE_COUNTER) ||
	    !mib_alloc_entry(&m_cpu_oid, 53, 0, BER_TYPE_COUNTER) ||
	    !mib_alloc_entry(&m_cpu_oid, 59, 0, BER_TYPE_COUNTER) ||
	    !mib_alloc_entry(&m_cpu_oid, 60, 0, BER_TYPE_COUNTER))
		return -1;

	return 0;
}

/* We need the uptime in 1/100 seconds, so we can't use sysinfo() */
static unsigned int get_system_uptime(void)
{
	char buf[128];

	if (read_file("/proc/uptime", buf, sizeof(buf)) == -1)
		return -1;

	return (unsigned int)(atof(buf) * 100);
}

/* We need the uptime in 1/100 seconds, so we can't use sysinfo() */
static unsigned int get_process_uptime(void)
{
	unsigned int uptime_now = get_system_uptime();

	if (uptime_start == 0)
		uptime_start = uptime_now;

	return uptime_now - uptime_start;
}

static void get_loadinfo(loadinfo_t *loadinfo)
{
	int i;
	char buf[128];
	char *ptr;

	if (read_file("/proc/loadavg", buf, sizeof(buf)) == -1) {
		memset(loadinfo, 0, sizeof(loadinfo_t));
		return;
	}

	ptr = buf;
	for (i = 0; i < 3; i++) {
		while (isspace(*ptr))
			ptr++;

		if (*ptr != 0)
			loadinfo->avg[i] = strtod(ptr, &ptr) * 100;
	}
}

static void get_meminfo(meminfo_t *meminfo)
{
	field_t fields[] = {
		{ "MemTotal",  1, { &meminfo->total   }},
		{ "MemFree",   1, { &meminfo->free    }},
		{ "MemShared", 1, { &meminfo->shared  }},
		{ "Buffers",   1, { &meminfo->buffers }},
		{ "Cached",    1, { &meminfo->cached  }},
		{ NULL }
	};

	if (parse_file("/proc/meminfo", fields))
		memset(meminfo, 0, sizeof(meminfo_t));
}

static void get_cpuinfo(cpuinfo_t *cpuinfo)
{
	field_t fields[] = {
		{ "cpu ",  4, { &cpuinfo->user, &cpuinfo->nice, &cpuinfo->system, &cpuinfo->idle }},
		{ "intr ", 1, { &cpuinfo->irqs   }},
		{ "ctxt ", 1, { &cpuinfo->cntxts }},
		{ NULL }
	};

	if (parse_file("/proc/stat", fields))
		memset(cpuinfo, 0, sizeof(cpuinfo_t));
}

static void get_diskinfo(diskinfo_t *diskinfo)
{
	size_t i;
	struct statfs fs;

	for (i = 0; i < g_disk_list_length; i++) {
		if (statfs(g_disk_list[i], &fs) == -1) {
			diskinfo->total[i]               = 0;
			diskinfo->free[i]                = 0;
			diskinfo->used[i]                = 0;
			diskinfo->blocks_used_percent[i] = 0;
			diskinfo->inodes_used_percent[i] = 0;
			continue;
		}

		diskinfo->total[i] = ((float)fs.f_blocks * fs.f_bsize) / 1024;
		diskinfo->free[i]  = ((float)fs.f_bfree  * fs.f_bsize) / 1024;
		diskinfo->used[i]  = ((float)(fs.f_blocks - fs.f_bfree) * fs.f_bsize) / 1024;
		diskinfo->blocks_used_percent[i] =
			((float)(fs.f_blocks - fs.f_bfree) * 100 + fs.f_blocks - 1) / fs.f_blocks;
		if (fs.f_files <= 0)
			diskinfo->inodes_used_percent[i] = 0;
		else
			diskinfo->inodes_used_percent[i] =
				((float)(fs.f_files - fs.f_ffree) * 100 + fs.f_files - 1) / fs.f_files;
	}
}

static void get_netinfo(netinfo_t *netinfo)
{
	int fd = bb_xsocket(BB_XSOCKET_FLG_SOCK_DGRAM | BB_XSOCKET_FLG_NO_WARN);
	size_t i;
	struct ifreq ifreq;
	field_t fields[MAX_NR_INTERFACES + 1];

	memset(fields, 0, (MAX_NR_INTERFACES + 1) * sizeof(field_t));
	for (i = 0; i < g_interface_list_length; i++) {
		fields[i].prefix    = g_interface_list[i];
		fields[i].len       = 12;
		fields[i].value[0]  = &netinfo->rx_bytes[i];
		fields[i].value[1]  = &netinfo->rx_packets[i];
		fields[i].value[2]  = &netinfo->rx_errors[i];
		fields[i].value[3]  = &netinfo->rx_drops[i];
		fields[i].value[8]  = &netinfo->tx_bytes[i];
		fields[i].value[9]  = &netinfo->tx_packets[i];
		fields[i].value[10] = &netinfo->tx_errors[i];
		fields[i].value[11] = &netinfo->tx_drops[i];

		safe_ifname_cp(&ifreq, g_interface_list[i]);
		if (fd == -1 || ioctl(fd, SIOCGIFFLAGS, &ifreq) == -1) {
			netinfo->status[i] = 4;
			continue;
		}

		if (ifreq.ifr_flags & IFF_UP)
			netinfo->status[i] = (ifreq.ifr_flags & IFF_RUNNING) ? 1 : 7;
		else
			netinfo->status[i] = 2;
	}
	if (fd != -1)
		close(fd);

	if (parse_file("/proc/net/dev", fields))
		memset(netinfo, 0, sizeof(*netinfo));
}

static int mib_update_entry(const oid_t *prefix, int column, int row, size_t *pos, int type, const void *arg)
{
	oid_t oid;
	value_t *value;
	static const char msg[] = "Failed updating OID";

	memcpy(&oid, prefix, sizeof(oid));

	/* Create the OID from the prefix, the column and the row */
	if (oid_build(&oid, prefix, column, row)) {
		lprintf(LOG_ERR, "%s '%s.%d.%d': OID overflow", msg, oid_ntoa(prefix), column, row);
		return -1;
	}

	/* Search the MIB for the given OID beginning at the given position */
	value = mib_find(&oid, pos);
	if (!value) {
		lprintf(LOG_ERR, "%s '%s.%d.%d': OID not found", msg, oid_ntoa(prefix), column, row);
		return -1;
	}

	return mib_data_set(prefix, &value->data, column, row, type, arg);
}

static int mib_update(int full)
{
	char nr[16];
	size_t i, pos;
	union {
		diskinfo_t diskinfo;
		loadinfo_t loadinfo;
		meminfo_t meminfo;
		cpuinfo_t cpuinfo;
		netinfo_t netinfo;
	} u;

	/* Begin searching at the first MIB entry */
	pos = 0;

	/*
	 * The system MIB: basic info about the host (SNMPv2-MIB.txt)
	 * Caution: on changes, adapt the corresponding mib_build() section too!
	 */
	if (mib_update_entry(&m_system_oid, 3, 0, &pos, BER_TYPE_TIME_TICKS, (const void *)(uintptr_t)get_process_uptime()) == -1)
		return -1;

	/*
	 * The interface MIB: network interfaces (IF-MIB.txt)
	 * Caution: on changes, adapt the corresponding mib_build() section too!
	 */
	if (full) {
		if (g_interface_list_length > 0) {
			get_netinfo(&u.netinfo);
			for (i = 0; i < g_interface_list_length; i++) {
				if (mib_update_entry(&m_if_2_oid, 8, i + 1, &pos, BER_TYPE_INTEGER, (const void *)(intptr_t)u.netinfo.status[i]) == -1)
					return -1;
			}

			for (i = 0; i < g_interface_list_length; i++) {
				if (mib_update_entry(&m_if_2_oid, 10, i + 1, &pos, BER_TYPE_COUNTER, (const void *)(uintptr_t)u.netinfo.rx_bytes[i]) == -1)
					return -1;
			}

			for (i = 0; i < g_interface_list_length; i++) {
				if (mib_update_entry(&m_if_2_oid, 11, i + 1, &pos, BER_TYPE_COUNTER, (const void *)(uintptr_t)u.netinfo.rx_packets[i]) == -1)
					return -1;
			}

			for (i = 0; i < g_interface_list_length; i++) {
				if (mib_update_entry(&m_if_2_oid, 13, i + 1, &pos, BER_TYPE_COUNTER, (const void *)(uintptr_t)u.netinfo.rx_drops[i]) == -1)
					return -1;
			}

			for (i = 0; i < g_interface_list_length; i++) {
				if (mib_update_entry(&m_if_2_oid, 14, i + 1, &pos, BER_TYPE_COUNTER, (const void *)(uintptr_t)u.netinfo.rx_errors[i]) == -1)
					return -1;
			}

			for (i = 0; i < g_interface_list_length; i++) {
				if (mib_update_entry(&m_if_2_oid, 16, i + 1, &pos, BER_TYPE_COUNTER, (const void *)(uintptr_t)u.netinfo.tx_bytes[i]) == -1)
					return -1;
			}

			for (i = 0; i < g_interface_list_length; i++) {
				if (mib_update_entry(&m_if_2_oid, 17, i + 1, &pos, BER_TYPE_COUNTER, (const void *)(uintptr_t)u.netinfo.tx_packets[i]) == -1)
					return -1;
			}

			for (i = 0; i < g_interface_list_length; i++) {
				if (mib_update_entry(&m_if_2_oid, 19, i + 1, &pos, BER_TYPE_COUNTER, (const void *)(uintptr_t)u.netinfo.tx_drops[i]) == -1)
					return -1;
			}

			for (i = 0; i < g_interface_list_length; i++) {
				if (mib_update_entry(&m_if_2_oid, 20, i + 1, &pos, BER_TYPE_COUNTER, (const void *)(uintptr_t)u.netinfo.tx_errors[i]) == -1)
					return -1;
			}
		}
	}

	/*
	 * The host MIB: additional host info (HOST-RESOURCES-MIB.txt)
	 * Caution: on changes, adapt the corresponding mib_build() section too!
	 */
	if (mib_update_entry(&m_host_oid, 1, 0, &pos, BER_TYPE_TIME_TICKS, (const void *)(uintptr_t)get_system_uptime()) == -1)
		return -1;

	/*
	 * The memory MIB: total/free memory (UCD-SNMP-MIB.txt)
	 * Caution: on changes, adapt the corresponding mib_build() section too!
	 */
	if (full) {
		get_meminfo(&u.meminfo);
		if (mib_update_entry(&m_memory_oid,  5, 0, &pos, BER_TYPE_INTEGER, (const void *)(intptr_t)u.meminfo.total)   == -1 ||
		    mib_update_entry(&m_memory_oid,  6, 0, &pos, BER_TYPE_INTEGER, (const void *)(intptr_t)u.meminfo.free)    == -1 ||
		    mib_update_entry(&m_memory_oid, 13, 0, &pos, BER_TYPE_INTEGER, (const void *)(intptr_t)u.meminfo.shared)  == -1 ||
		    mib_update_entry(&m_memory_oid, 14, 0, &pos, BER_TYPE_INTEGER, (const void *)(intptr_t)u.meminfo.buffers) == -1 ||
		    mib_update_entry(&m_memory_oid, 15, 0, &pos, BER_TYPE_INTEGER, (const void *)(intptr_t)u.meminfo.cached)  == -1)
			return -1;
	}

	/*
	 * The disk MIB: mounted partitions (UCD-SNMP-MIB.txt)
	 * Caution: on changes, adapt the corresponding mib_build() section too!
	 */
	if (full) {
		if (g_disk_list_length > 0) {
			get_diskinfo(&u.diskinfo);
			for (i = 0; i < g_disk_list_length; i++) {
				if (mib_update_entry(&m_disk_oid, 6, i + 1, &pos, BER_TYPE_INTEGER, (const void *)(intptr_t)u.diskinfo.total[i]) == -1)
					return -1;
			}

			for (i = 0; i < g_disk_list_length; i++) {
				if (mib_update_entry(&m_disk_oid, 7, i + 1, &pos, BER_TYPE_INTEGER, (const void *)(intptr_t)u.diskinfo.free[i]) == -1)
					return -1;
			}

			for (i = 0; i < g_disk_list_length; i++) {
				if (mib_update_entry(&m_disk_oid, 8, i + 1, &pos, BER_TYPE_INTEGER, (const void *)(intptr_t)u.diskinfo.used[i]) == -1)
					return -1;
			}

			for (i = 0; i < g_disk_list_length; i++) {
				if (mib_update_entry(&m_disk_oid, 9, i + 1, &pos, BER_TYPE_INTEGER, (const void *)(intptr_t)u.diskinfo.blocks_used_percent[i]) == -1)
					return -1;
			}

			for (i = 0; i < g_disk_list_length; i++) {
				if (mib_update_entry(&m_disk_oid, 10, i + 1, &pos, BER_TYPE_INTEGER, (const void *)(intptr_t)u.diskinfo.inodes_used_percent[i]) == -1)
					return -1;
			}
		}
	}

	/*
	 * The load MIB: CPU load averages (UCD-SNMP-MIB.txt)
	 * Caution: on changes, adapt the corresponding mib_build() section too!
	 */
	if (full) {
		get_loadinfo(&u.loadinfo);
		for (i = 0; i < 3; i++) {
			snprintf(nr, sizeof(nr), "%d.%02d", u.loadinfo.avg[i] / 100, u.loadinfo.avg[i] % 100);
			if (mib_update_entry(&m_load_oid, 3, i + 1, &pos, BER_TYPE_OCTET_STRING, nr) == -1)
				return -1;
		}

		for (i = 0; i < 3; i++) {
			if (mib_update_entry(&m_load_oid, 5, i + 1, &pos, BER_TYPE_INTEGER, (const void *)(intptr_t)u.loadinfo.avg[i]) == -1)
				return -1;
		}
	}

	/*
	 * The cpu MIB: CPU statistics (UCD-SNMP-MIB.txt)
	 * Caution: on changes, adapt the corresponding mib_build() section too!
	 */
	if (full) {
		get_cpuinfo(&u.cpuinfo);
		if (mib_update_entry(&m_cpu_oid, 50, 0, &pos, BER_TYPE_COUNTER, (const void *)(uintptr_t)u.cpuinfo.user)   == -1 ||
		    mib_update_entry(&m_cpu_oid, 51, 0, &pos, BER_TYPE_COUNTER, (const void *)(uintptr_t)u.cpuinfo.nice)   == -1 ||
		    mib_update_entry(&m_cpu_oid, 52, 0, &pos, BER_TYPE_COUNTER, (const void *)(uintptr_t)u.cpuinfo.system) == -1 ||
		    mib_update_entry(&m_cpu_oid, 53, 0, &pos, BER_TYPE_COUNTER, (const void *)(uintptr_t)u.cpuinfo.idle)   == -1 ||
		    mib_update_entry(&m_cpu_oid, 59, 0, &pos, BER_TYPE_COUNTER, (const void *)(uintptr_t)u.cpuinfo.irqs)   == -1 ||
		    mib_update_entry(&m_cpu_oid, 60, 0, &pos, BER_TYPE_COUNTER, (const void *)(uintptr_t)u.cpuinfo.cntxts) == -1)
			return -1;
	}
	return 0;
}


static void handle_signal(int signo)
{
	g_quit = signo;
}

static void handle_udp_client(void)
{
	static const char req_msg[] = "Failed UDP request from";
	static const char snd_msg[] = "Failed UDP response to";
	ssize_t rv;
	char straddr[my_inet_addrstrlen];
	socklen_t socklen;
	struct my_sockaddr_t sockaddr;

	/* Read the whole UDP packet from the socket at once */
	socklen = sizeof(sockaddr);
	rv = recvfrom(g_udp_sockfd, g_udp_client.packet, sizeof(g_udp_client.packet),
		      0, (struct sockaddr *)&sockaddr, &socklen);
	if (rv == -1) {
		lprintf(LOG_WARNING, "Failed receiving UDP request on port %d: %m", g_udp_port);
		return;
	}

	g_udp_client.timestamp = time(NULL);
	g_udp_client.sockfd = g_udp_sockfd;
	g_udp_client.addr = sockaddr.my_sin_addr;
	g_udp_client.port = sockaddr.my_sin_port;
	g_udp_client.size = rv;
	g_udp_client.outgoing = 0;

	/* Call the protocol handler which will prepare the response packet */
	inet_ntop(my_af_inet, &sockaddr.my_sin_addr, straddr, sizeof(straddr));
	if (snmp(&g_udp_client) == -1) {
		lprintf(LOG_WARNING, "%s %s:%d: %m", req_msg, straddr, sockaddr.my_sin_port);
		return;
	}
	if (g_udp_client.size == 0) {
		lprintf(LOG_WARNING, "%s %s:%d: ignored", req_msg, straddr, sockaddr.my_sin_port);
		return;
	}
	g_udp_client.outgoing = 1;

	/* Send the whole UDP packet to the socket at once */
	rv = sendto(g_udp_sockfd, g_udp_client.packet, g_udp_client.size,
		MSG_DONTWAIT, (struct sockaddr *)&sockaddr, socklen);
	inet_ntop(my_af_inet, &sockaddr.my_sin_addr, straddr, sizeof(straddr));
	if (rv == -1)
		lprintf(LOG_WARNING, "%s %s:%d: %m", snd_msg, straddr, sockaddr.my_sin_port);
	else if ((size_t)rv != g_udp_client.size)
		lprintf(LOG_WARNING, "%s %s:%d: only %zd of %zu bytes sent", snd_msg, straddr, sockaddr.my_sin_port, rv, g_udp_client.size);

}

static void handle_tcp_connect(void)
{
	int rv;
	static const char msg[] = "Could not accept TCP connection";
	char straddr[my_inet_addrstrlen];
	client_t *client;
	socklen_t socklen;
	struct my_sockaddr_t tmp_sockaddr;
	struct my_sockaddr_t sockaddr;

	/* Accept the new connection (remember the client's IP address and port) */
	socklen = sizeof(sockaddr);
	rv = accept(g_tcp_sockfd, (struct sockaddr *)&sockaddr, &socklen);
	if (rv == -1) {
		lprintf(LOG_ERR, "%s: %m", msg);
		return;
	}
	if (rv >= FD_SETSIZE) {
		lprintf(LOG_ERR, "%s: FD set overflow", msg);
		close(rv);
		return;
	}

	/* Create a new client control structure or overwrite the oldest one */
	if (g_tcp_client_list_length >= MAX_NR_CLIENTS) {
		client = find_oldest_client();
		if (!client) {
			lprintf(LOG_ERR, "%s: internal error", msg);
			exit(EXIT_SYSCALL);
		}

		tmp_sockaddr.my_sin_addr = client->addr;
		tmp_sockaddr.my_sin_port = client->port;
		inet_ntop(my_af_inet, &tmp_sockaddr.my_sin_addr, straddr, sizeof(straddr));
		lprintf(LOG_WARNING, "Maximum number of %d clients reached, kicking out %s:%d",
			MAX_NR_CLIENTS, straddr, tmp_sockaddr.my_sin_port);
		close(client->sockfd);
	} else {
		client = xmalloc(sizeof(client_t));
		g_tcp_client_list[g_tcp_client_list_length++] = client;
	}

	/* Now fill out the client control structure values */
	inet_ntop(my_af_inet, &sockaddr.my_sin_addr, straddr, sizeof(straddr));
	lprintf(LOG_DEBUG, "Connected TCP client %s:%d", straddr, sockaddr.my_sin_port);
	client->timestamp = time(NULL);
	client->sockfd = rv;
	client->addr = sockaddr.my_sin_addr;
	client->port = sockaddr.my_sin_port;
	client->size = 0;
	client->outgoing = 0;
}

static void handle_tcp_client_write(client_t *client)
{
	static const char msg[] = "Failed TCP response to";
	ssize_t rv;
	char straddr[my_inet_addrstrlen];
	struct my_sockaddr_t sockaddr;

	/* Send the packet atomically and close socket if that did not work */
	sockaddr.my_sin_addr = client->addr;
	sockaddr.my_sin_port = client->port;
	rv = send(client->sockfd, client->packet, client->size, 0);
	inet_ntop(my_af_inet, &sockaddr.my_sin_addr, straddr, sizeof(straddr));
	if (rv == -1) {
		lprintf(LOG_WARNING, "%s %s:%d: %m", msg, straddr, sockaddr.my_sin_port);
		close(client->sockfd);
		client->sockfd = -1;
		return;
	}
	if ((size_t)rv != client->size) {
		lprintf(LOG_WARNING, "%s %s:%d: only %zd of %zu bytes written",
			msg, straddr, sockaddr.my_sin_port, rv, client->size);
		close(client->sockfd);
		client->sockfd = -1;
		return;
	}

	/* Put the client into listening mode again */
	client->size = 0;
	client->outgoing = 0;
}

static void handle_tcp_client_read(client_t *client)
{
	static const char req_msg[] = "Failed TCP request from";
	int rv;
	char straddr[my_inet_addrstrlen];
	struct my_sockaddr_t sockaddr;

	/* Read from the socket what arrived and put it into the buffer */
	sockaddr.my_sin_addr = client->addr;
	sockaddr.my_sin_port = client->port;
	rv = read(client->sockfd, client->packet + client->size, sizeof(client->packet) - client->size);
	inet_ntop(my_af_inet, &sockaddr.my_sin_addr, straddr, sizeof(straddr));
	if (rv == -1) {
		lprintf(LOG_WARNING, "%s %s:%d: %m", req_msg, straddr, sockaddr.my_sin_port);
		close(client->sockfd);
		client->sockfd = -1;
		return;
	}
	if (rv == 0) {
		lprintf(LOG_DEBUG, "TCP client %s:%d disconnected", straddr, sockaddr.my_sin_port);
		close(client->sockfd);
		client->sockfd = -1;
		return;
	}
	client->timestamp = time(NULL);
	client->size += rv;

	/* Check whether the packet was fully received and handle packet if yes */
	rv = snmp_packet_complete(client);
	if (rv == -1) {
		lprintf(LOG_WARNING, "%s %s:%d: %m", req_msg, straddr, sockaddr.my_sin_port);
		close(client->sockfd);
		client->sockfd = -1;
		return;
	}
	if (rv == 0) {
		return;
	}
	client->outgoing = 0;

	/* Call the protocol handler which will prepare the response packet */
	if (snmp(client) == -1) {
		lprintf(LOG_WARNING, "%s %s:%d: %m", req_msg, straddr, sockaddr.my_sin_port);
		close(client->sockfd);
		client->sockfd = -1;
		return;
	}
	if (client->size == 0) {
		lprintf(LOG_WARNING, "%s %s:%d: ignored", req_msg, straddr, sockaddr.my_sin_port);
		close(client->sockfd);
		client->sockfd = -1;
		return;
	}

	client->outgoing = 1;
}

/* bb_xgetopt_ulflags option list helper function */

static void split_disklist_opt(const char *o)
{
  g_disk_list_length = split(o, ",:;", g_disk_list, MAX_NR_DISKS);
}

static void split_iflist_opt(const char *o)
{
  g_interface_list_length = split(o, ",;", g_interface_list, MAX_NR_INTERFACES);
}


int snmpd_main(void)
{
	USE_FEATURE_LONG_OPTIONS(static const struct option long_options[] = {
#if ENABLE_FEATURE_IPV6
		{ "use-ipv4", 0, 0, '4' },
		{ "use-ipv6", 0, 0, '6' },
#endif
		{ "udp-port", 1, 0, 'p' },
		{ "tcp-port", 1, 0, 'P' },
		{ "community", 1, 0, 'c' },
		{ "description", 1, 0, 'D' },
		{ "vendor", 1, 0, 'V' },
		{ "location", 1, 0, 'L' },
		{ "contact", 1, 0, 'C' },
		{ "disks", 1, 0, 'd' },
		{ "interfaces", 1, 0, 'i' },
		{ "listen", 1, 0, 'I' },
		{ "timeout", 1, 0, 't' },
		{ "auth", 0, 0, 'a' },
		{ "foreground", 0, 0, 'n' },
		{ "verbose", 0, 0, 'v' },
		{ "syslog", 0, 0, 's' },
		{ NULL, 0, 0, 0 }
	};)
	int ticks, nfds, c;
	size_t i;
	fd_set rfds, wfds;
	struct ifreq ifreq;
	struct timeval tv_last;
	struct timeval tv_now;
	struct timeval tv_sleep;
	socklen_t socklen;
	union {
		struct sockaddr_in sa;
#if ENABLE_FEATURE_IPV6
		struct sockaddr_in6 sa6;
#endif
	} sockaddr;
	char *t = NULL;

	g_timeout = 100;
	g_community = "public";
	g_vendor = VENDOR;
	g_disk_list[0] = "/";
	g_disk_list_length = 1;
	g_udp_port = 161;
	g_tcp_port = 161;
	/* Prevent TERM and HUP signals from interrupting system calls */
	/* 1 << 0 is flag SA_RESTART */
	bb_signal ((1<<0) | (1 << SIGHUP) | (1 << SIGTERM), handle_signal);

	/* Parse commandline options */
	USE_FEATURE_LONG_OPTIONS(bb_applet_long_options = long_options);
	c = bb_xgetopt_ulflags("ansv46p:P:t:c:D:V:L:C:I:d:i:h\0:-pP:h-:4-6:6-4:=di",
			&g_udp_port, &g_tcp_port, &t, &g_community, &g_description,
			&g_vendor, &g_location, &g_contact, &g_bind_to_device,
			split_disklist_opt, split_iflist_opt);
	g_auth = c & 1;
	g_no_daemon = c & 2;
	g_syslog = c & 4;
	g_verbose = c & 8;
#if ENABLE_FEATURE_IPV6
	g_family = ((c & (16|32)) != 32) ? AF_INET : AF_INET6;
#endif
	if(t != NULL)
		g_timeout = atoi(t) * 100;

	if (!g_no_daemon) {
		lprintf(LOG_DEBUG, "Daemonizing ...");
		if (-1 == daemon(0, 0)) {
			log_syslog = LOG_ERR;
			bb_error_msg_and_die("Failed daemonizing: %m");
		}
		g_syslog = 1;
	}
	if(g_syslog) {
		bb_openlog(LOG_PID, LOG_DAEMON);
		log_stderr = 0;
		log_syslog = LOG_ERR;
	}

	/* Store the starting time since we need it for MIB updates */
	if (gettimeofday(&tv_last, NULL) == -1) {
		memset(&tv_last, 0, sizeof(tv_last));
		memset(&tv_sleep, 0, sizeof(tv_sleep));
	} else {
		tv_sleep.tv_sec = g_timeout / 100;
		tv_sleep.tv_usec = (g_timeout % 100) * 10000;
	}

	/* Build the MIB and execute the first MIB update to get actual values */
	if (mib_build() == -1)
		return EXIT_SYSCALL;
	if (mib_update(1) == -1)
		return EXIT_SYSCALL;

	/* Open the server's UDP port and prepare it for listening */
	g_udp_sockfd = bb_xsocket(BB_XSOCKET_FLG_SOCK_DGRAM | BB_XSOCKET_FLG_ERR_DIE
#if ENABLE_FEATURE_IPV6
				| ((g_family == AF_INET) ? 0 : BB_XSOCKET_FLG_AF_INET6)
#endif
				);
#if ENABLE_FEATURE_IPV6
	if (g_family == AF_INET) {
#endif
		sockaddr.sa.sin_family = g_family;
		sockaddr.sa.sin_port = htons(g_udp_port);
		sockaddr.sa.sin_addr = inaddr_any;
		socklen = sizeof(sockaddr.sa);
#if ENABLE_FEATURE_IPV6
	} else {
		sockaddr.sa6.sin6_family = g_family;
		sockaddr.sa6.sin6_port = htons(g_udp_port);
		sockaddr.sa6.sin6_addr = in6addr_any;
		socklen = sizeof(sockaddr.sa6);
	}
#endif
	bb_xbind(g_udp_sockfd, (struct sockaddr *)&sockaddr, socklen);

	if (g_bind_to_device) {
		safe_ifname_cp(&ifreq, g_bind_to_device);
		if (setsockopt(g_udp_sockfd, SOL_SOCKET, SO_BINDTODEVICE, (char *)&ifreq, sizeof(ifreq)) == -1) {
			lprintf(LOG_WARNING, "could not bind UDP socket to device %s: %m", g_bind_to_device);
			return EXIT_SYSCALL;
		}
	}
	/* Open the server's TCP port and prepare it for listening */
	g_tcp_sockfd = bb_xsocket(BB_XSOCKET_FLG_ERR_DIE
#if ENABLE_FEATURE_IPV6
				| ((g_family == AF_INET) ? 0 : BB_XSOCKET_FLG_AF_INET6)
#endif
				);

	if (g_bind_to_device) {
		safe_ifname_cp(&ifreq, g_bind_to_device);
		if (setsockopt(g_tcp_sockfd, SOL_SOCKET, SO_BINDTODEVICE, (char *)&ifreq, sizeof(ifreq)) == -1) {
			lprintf(LOG_WARNING, "could not bind TCP socket to device %s: %m", g_bind_to_device);
			return EXIT_SYSCALL;
		}
	}

	if (setsockopt_1i(g_tcp_sockfd, SOL_SOCKET, SO_REUSEADDR, 1) == -1) {
		lprintf(LOG_WARNING, "could not set SO_REUSEADDR on TCP socket: %m");
		return EXIT_SYSCALL;
	}

#if ENABLE_FEATURE_IPV6
	if (g_family == AF_INET) {
#endif
		sockaddr.sa.sin_family = g_family;
		sockaddr.sa.sin_port = htons(g_udp_port);
		sockaddr.sa.sin_addr = inaddr_any;
		socklen = sizeof(sockaddr.sa);
#if ENABLE_FEATURE_IPV6
	} else {
		sockaddr.sa6.sin6_family = g_family;
		sockaddr.sa6.sin6_port = htons(g_udp_port);
		sockaddr.sa6.sin6_addr = in6addr_any;
		socklen = sizeof(sockaddr.sa6);
	}
#endif
	bb_xbind(g_tcp_sockfd, (struct sockaddr *)&sockaddr, socklen);

	if (listen(g_tcp_sockfd, 128) == -1) {
		lprintf(LOG_ERR, "could not prepare TCP socket for listening: %m");
		return EXIT_SYSCALL;
	}

	/* Print a starting message (so the user knows the args were ok) */
	if (g_bind_to_device) {
		lprintf(LOG_INFO, "Listening on port %d/udp and %d/tcp on interface %s",
			g_udp_port, g_tcp_port, g_bind_to_device);
	} else {
		lprintf(LOG_INFO, "Listening on port %d/udp and %d/tcp", g_udp_port, g_tcp_port);
	}

	/* Handle incoming connect requests and incoming data */
	while (!g_quit) {
		/* Sleep until we get a request or the timeout is over */
		FD_ZERO(&rfds);
		FD_ZERO(&wfds);
		FD_SET(g_udp_sockfd, &rfds);
		FD_SET(g_tcp_sockfd, &rfds);
		nfds = (g_udp_sockfd > g_tcp_sockfd) ? g_udp_sockfd : g_tcp_sockfd;

		for (i = 0; i < g_tcp_client_list_length; i++) {
			if (g_tcp_client_list[i]->outgoing)
				FD_SET(g_tcp_client_list[i]->sockfd, &wfds);
			else
				FD_SET(g_tcp_client_list[i]->sockfd, &rfds);

			if (nfds < g_tcp_client_list[i]->sockfd)
				nfds = g_tcp_client_list[i]->sockfd;
		}

		if (select(nfds + 1, &rfds, &wfds, NULL, &tv_sleep) == -1) {
			if (g_quit)
				break;

			lprintf(LOG_ERR, "could not select from sockets: %m");
			return EXIT_SYSCALL;
		}

		/* Determine whether to update the MIB and the next ticks to sleep */
		ticks = ticks_since(&tv_last, &tv_now);
		if (ticks < 0 || ticks >= g_timeout) {
			lprintf(LOG_DEBUG, "updating the MIB (full)");
			if (mib_update(1) == -1)
				return EXIT_SYSCALL;

			memcpy(&tv_last, &tv_now, sizeof(tv_now));
			tv_sleep.tv_sec = g_timeout / 100;
			tv_sleep.tv_usec = (g_timeout % 100) * 10000;
		} else {
			lprintf(LOG_DEBUG, "updating the MIB (partial)");
			if (mib_update(0) == -1)
				return EXIT_SYSCALL;

			tv_sleep.tv_sec = (g_timeout - ticks) / 100;
			tv_sleep.tv_usec = ((g_timeout - ticks) % 100) * 10000;
		}

		/* Handle UDP packets, TCP packets and TCP connection connects */
		if (FD_ISSET(g_udp_sockfd, &rfds))
			handle_udp_client();

		if (FD_ISSET(g_tcp_sockfd, &rfds))
			handle_tcp_connect();

		for (i = 0; i < g_tcp_client_list_length; i++) {
			if (g_tcp_client_list[i]->outgoing) {
				if (FD_ISSET(g_tcp_client_list[i]->sockfd, &wfds))
					handle_tcp_client_write(g_tcp_client_list[i]);
			} else {
				if (FD_ISSET(g_tcp_client_list[i]->sockfd, &rfds))
					handle_tcp_client_read(g_tcp_client_list[i]);
			}
		}

		/* If there was a TCP disconnect, remove the client from the list */
		for (i = 0; i < g_tcp_client_list_length; i++) {
			if (g_tcp_client_list[i]->sockfd == -1) {
				g_tcp_client_list_length--;
				if (i < g_tcp_client_list_length) {
					size_t len = (g_tcp_client_list_length - i) * sizeof(g_tcp_client_list[i]);

					free(g_tcp_client_list[i]);
					memmove(&g_tcp_client_list[i], &g_tcp_client_list[i + 1], len);
				}
			}
		}
	}

	/* We were killed, print a message and exit */
	lprintf(LOG_INFO, "stopped");

	return EXIT_OK;
}
#endif /* APPLET_USAGE */
