/*-
 * Copyright (c) 1991, 1993
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Busybox port by Vladimir Oleynik (C) 2004 <dzo@simtreas.ru>
 */

/* OpenBSD telnet_locl.h 1.2 1999/12/11 09:08:09 itojun  */
#ifdef APPLET_USAGE
#define telnet_trivial_usage \
	"[-a] [-e escape_char] [-l USER] HOST [PORT]"
#define telnet_full_usage \
	"Telnet is used to establish interactive communication with another\n" \
	"computer over a network using the TELNET protocol.\n\n" \
	"Options:\n" \
	"\t-a\t\tAttempt an automatic login with the USER variable\n" \
	"\t-e escape_char\tSets the initial escape character\n" \
	"\t-l USER\t\tAttempt an automatic login with the USER argument\n" \
	"\tHOST\t\tThe official name, alias or the IP address of the\n" \
	"\t\t\tremote host.\n" \
	"\tPORT\t\tThe remote port number to connect to. If it is not\n" \
	"\t\t\tspecified, the default telnet (23) port is used."
#else

#include "busybox.h"
#include <sys/resource.h>

#include <string.h>
#include <ctype.h>
#include <setjmp.h>
#include <unistd.h>

#include <termios.h>
#include <fcntl.h>
#include <sys/select.h>
#include <sys/time.h>
#include <time.h>

#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/file.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>

#define TELOPTS
//#define TELCMDS
//#define SLC_NAMES


#include "speed_table.h"
#include "bb_sig.h"
#include "bb_inet.h"

#ifdef HAVE_ARPA_TELNET
#include <arpa/telnet.h>
#else
#include "telnet.h"
#endif

#ifndef howmany
# define howmany(x, y)    (((x) + ((y) - 1)) / (y))
#endif
#ifdef __BIONIC__
typedef unsigned long fd_mask;
#endif

//#define SUPPORT_DEBUGING

#ifndef _POSIX_VDISABLE
# ifdef VDISABLE
#  define _POSIX_VDISABLE VDISABLE
# else
#  define _POSIX_VDISABLE ((cc_t)'\377')
# endif
#endif


#define LINEMODE

/*
 * This defines a structure for a ring buffer.
 *
 * The circular buffer has two parts:
 *(((
 *      full:   [consume, supply)
 *      empty:  [supply, consume)
 *]]]
 *
 */
typedef struct {
    unsigned char       *consume,       /* where data comes out of */
			*supply,        /* where data comes in to */
			*bottom,        /* lowest address in buffer */
			*top,           /* highest address+1 in buffer */
			*mark;          /* marker (user defined) */
    int         size;           /* size in bytes of buffer */
    u_long      consumetime,    /* help us keep straight full, empty, etc. */
		supplytime;
} Ring;

/* Here are some functions and macros to deal with the ring buffer */

/* Initialization routine */
static int ring_init(Ring *ring, unsigned char *buffer, int count);

/* Buffer state transition routines */
static void
	ring_supplied(Ring *ring, int count),
	ring_consumed(Ring *ring, int count);

/* Buffer state query routines */
static int
	ring_empty_count(Ring *ring),
	ring_full_count(Ring *ring),
	ring_full_consecutive(Ring *ring);


/*
 * Clear any mark set on the ring.
 */

static inline void
ring_clear_mark(Ring *ring)
{
    ring->mark = 0;
}

static void ring_mark(Ring *);


static int ring_at_mark(Ring *);

typedef struct {
    int
	system,                 /* what the current time is */
	echotoggle,             /* last time user entered echo character */
	modenegotiated,         /* last time operating mode negotiated */
	didnetreceive,          /* last time we read data from network */
	gotDM;                  /* when did we last see a data mark */
} Clocks;

struct env_lst {
	struct env_lst *next;   /* pointer to next structure */
	struct env_lst *prev;   /* pointer to previous structure */
	unsigned char *var;     /* pointer to variable name */
	unsigned char *value;   /* pointer to variable value */
	int export;             /* 1 -> export with default list of variables */
	int welldefined;        /* A well defined variable */
};

/*
 * slc()
 * Handle special character suboption of LINEMODE.
 */

struct spc {
	cc_t val;
	cc_t *valp;
	char flags;     /* Current flags & level */
	char mylevel;   /* Maximum level & flags */
};


#define SUBBUFSIZE      256

#define SLC_IMPORT      0
#define SLC_EXPORT      1
#define SLC_RVALUE      2

CALC_BSS struct BB_COMMON_BSS {
	int binary;
	int autologin;      /* Autologin anyone? Autologin enabled */
	int skiprc;         /* Don't process the ~/.telnetrc file */
	int connected;      /* Are we connected to the other side? */
	int ISend;          /* trying to send network data in */
	int crmod;
	int crlf;           /* Should '\r' be mapped to <CR><LF> (or <CR><NUL>)? */
	int telnetport;     /* Are we connected to the telnet port? */
	int SYNCHing;       /* we are in TELNET SYNCH mode */
	int flushout;       /* flush output */
	int autoflush;      /* flush output when interrupting? */
	int autosynch;      /* send interrupt characters with SYNCH? */
	int localflow;      /* we handle flow control locally */
	int restartany;     /* if flow control enabled, restart on any character */
	int localchars;     /* we recognize interrupt/quit */
	int donelclchars;   /* the user has set "localchars" */
	int donebinarytoggle;       /* the user has put us in binary */
	int dontlecho;      /* do we suppress local echoing right now? */
	int globalmode;     /* Mode tty should be in */
	int clienteof;
	int tout;           /* Output file descriptor */
	int tin;            /* Input file descriptor */
	int net;            /* Network file descriptor */
	cc_t escape;     /* Escape to command mode */
	cc_t rlogin;     /* Rlogin mode escape character */
	cc_t echoc;      /* Toggle local echoing */
	char options[256]; /* The combined all the little options */
	jmp_buf peerdied;
	struct termios new_tc;
# ifndef VDISCARD
	cc_t termFlushChar;
# endif
# ifndef VLNEXT
	cc_t termLiteralNextChar;
# endif
# ifndef VSUSP
	cc_t termSuspChar;
# endif
# ifndef VWERASE
	cc_t termWerasChar;
# endif
# ifndef VREPRINT
	cc_t termRprntChar;
# endif
# ifndef VSTART
	cc_t termStartChar;
# endif
# ifndef VSTOP
	cc_t termStopChar;
# endif
# ifndef VEOL
	cc_t termForw1Char;
# endif
# ifndef VEOL2
	cc_t termForw2Char;
# endif
# ifndef VSTATUS
	cc_t termAytChar;
# endif
/* Ring buffer structures which are shared */

	Ring            ttyoring, ttyiring;
	unsigned char   ttyobuf[2*BUFSIZ], ttyibuf[BUFSIZ];

	Ring netoring;
	Ring netiring;
	unsigned char netobuf[2*BUFSIZ], netibuf[BUFSIZ];
/*
 * The following are some clocks used to decide how to interpret
 * the relationship between various variables.
 */
	Clocks clocks;
	char *ambiguous;         /* special return value for command routines */
	char    *hostname;  /* Who are we connected to? */
	char line[256];
	char saveline[256];
	int  margc;
	char *margv[20];
	char control_buf[5];
	int binmode;
	int linemode;
	struct env_lst envlisthead;
	struct env_lst *nep;
	char rcname[128];
	const char *prompt;    /* Prompt for command. */
	char lsr[44];
#ifdef  sysV88
	IOPTN ipopt;
#endif
	struct timeval timeout;
/*
 * The following is a clock, used to determine full, empty, etc.
 *
 * There is some trickiness here.  Since the ring buffers are initialized
 * to ZERO on allocation, we need to make sure, when interpreting the
 * clock, that when the times are EQUAL, then the buffer is FULL.
 */
	u_long ring_clock;
	struct  termios old_tc;
	fd_set *ibitsp, *obitsp, *xbitsp;
	int fdsn;
	int prevmode;
	struct timeval TimeValue;
	unsigned char    subbuffer[SUBBUFSIZE],
			*subpointer, *subend;    /* buffer for sub-options */
	char    do_dont_resp[256];
	char    will_wont_resp[256];
	int      telrcv_state;
	int     flushline;
	char *unknown[2];
	struct spc spc_data[NSLC+1];
	unsigned char slc_reply[128];
	unsigned char *slc_replyp;
	unsigned char *opt_reply;
	unsigned char *opt_replyp;
	unsigned char *opt_replyend;
	int local;
	int want_status_response;
	int bol;
	int eight;      /* use eight bit mode (binary in and/or out */
	int kludgelinemode;
	int slc_mode;
	unsigned char str_lm[7];
	unsigned char str_lm_mode[7];
	int tos;
} bb_common_bss;
#define binary bb_common_bss.binary
#define autologin bb_common_bss.autologin
#define skiprc bb_common_bss.skiprc
#define connected bb_common_bss.connected
#define ISend bb_common_bss.ISend
#define crmod bb_common_bss.crmod
#define crlf bb_common_bss.crlf
#define telnetport bb_common_bss.telnetport
#define SYNCHing bb_common_bss.SYNCHing
#define flushout bb_common_bss.flushout
#define autoflush bb_common_bss.autoflush
#define autosynch bb_common_bss.autosynch
#define localflow bb_common_bss.localflow
#define restartany bb_common_bss.restartany
#define localchars bb_common_bss.localchars
#define donelclchars bb_common_bss.donelclchars
#define donebinarytoggle bb_common_bss.donebinarytoggle
#define dontlecho bb_common_bss.dontlecho
#define globalmode bb_common_bss.globalmode
#define clienteof bb_common_bss.clienteof
#define tout bb_common_bss.tout
#define tin bb_common_bss.tin
#define net bb_common_bss.net
#define escape bb_common_bss.escape
#define rlogin bb_common_bss.rlogin
#define echoc bb_common_bss.echoc
#define options bb_common_bss.options
#define peerdied bb_common_bss.peerdied
#define new_tc bb_common_bss.new_tc
# ifndef VDISCARD
#define termFlushChar bb_common_bss.termFlushChar
# endif
# ifndef VLNEXT
#define termLiteralNextChar bb_common_bss.termLiteralNextChar
# endif
# ifndef VSUSP
#define termSuspChar bb_common_bss.termSuspChar
# endif
# ifndef VWERASE
#define termWerasChar bb_common_bss.termWerasChar
# endif
# ifndef VREPRINT
#define termRprntChar bb_common_bss.termRprntChar
# endif
# ifndef VSTART
#define termStartChar bb_common_bss.termStartChar
# endif
# ifndef VSTOP
#define termStopChar bb_common_bss.termStopChar
# endif
# ifndef VEOL
#define termForw1Char bb_common_bss.termForw1Char
# endif
# ifndef VEOL2
#define termForw2Char bb_common_bss.termForw2Char
# endif
# ifndef VSTATUS
#define termAytChar bb_common_bss.termAytChar
# endif
#define ttyoring bb_common_bss.ttyoring
#define ttyiring bb_common_bss.ttyiring
#define ttyobuf bb_common_bss.ttyobuf
#define ttyibuf bb_common_bss.ttyibuf
#define netoring bb_common_bss.netoring
#define netiring bb_common_bss.netiring
#define netobuf bb_common_bss.netobuf
#define netibuf bb_common_bss.netibuf
#define clocks bb_common_bss.clocks
#define ambiguous bb_common_bss.ambiguous
#define hostname bb_common_bss.hostname
#define line bb_common_bss.line
#define saveline bb_common_bss.saveline
#define margc bb_common_bss.margc
#define margv bb_common_bss.margv
#define control_buf bb_common_bss.control_buf
#define binmode bb_common_bss.binmode
#define linemode bb_common_bss.linemode
#define envlisthead bb_common_bss.envlisthead
#define nep bb_common_bss.nep
#define rcname bb_common_bss.rcname
#define prompt bb_common_bss.prompt
#define lsr bb_common_bss.lsr
#ifdef  sysV88
#define ipopt bb_common_bss.ipopt
#endif
#define timeout bb_common_bss.timeout
#define ring_clock bb_common_bss.ring_clock
#define old_tc bb_common_bss.old_tc
#define ibitsp bb_common_bss.ibitsp
#define obitsp bb_common_bss.obitsp
#define xbitsp bb_common_bss.xbitsp
#define fdsn bb_common_bss.fdsn
#define prevmode bb_common_bss.prevmode
#define TimeValue bb_common_bss.TimeValue
#define subbuffer bb_common_bss.subbuffer
#define subpointer bb_common_bss.subpointer
#define subend bb_common_bss.subend
#define do_dont_resp bb_common_bss.do_dont_resp
#define will_wont_resp bb_common_bss.will_wont_resp
#define telrcv_state bb_common_bss.telrcv_state
#define flushline bb_common_bss.flushline
#define unknown bb_common_bss.unknown
#define spc_data bb_common_bss.spc_data
#define slc_reply bb_common_bss.slc_reply
#define slc_replyp bb_common_bss.slc_replyp
#define opt_reply bb_common_bss.opt_reply
#define opt_replyp bb_common_bss.opt_replyp
#define opt_replyend bb_common_bss.opt_replyend
#define local bb_common_bss.local
#define want_status_response bb_common_bss.want_status_response
#define bol bb_common_bss.bol
#define eight bb_common_bss.eight
#define kludgelinemode bb_common_bss.kludgelinemode
#define slc_mode bb_common_bss.slc_mode
#define str_lm bb_common_bss.str_lm
#define str_lm_mode bb_common_bss.str_lm_mode
#define tos bb_common_bss.tos

#ifdef SUPPORT_DEBUGING
static int
	netdata,        /* Print out network data flow */
	showoptions,
	prettydump,     /* Print "netdata" output in user readable format */
	termdata,       /* Debugging flag, print out terminal data flow */
	debug;          /* Debug level */
#else
#define netdata 0
#define termdata 0
#endif


/*
 * We keep track of each side of the option negotiation.
 */

#define MY_STATE_WILL           0x01
#define MY_WANT_STATE_WILL      0x02
#define MY_STATE_DO             0x04
#define MY_WANT_STATE_DO        0x08

/*
 * Macros to check the current state of things
 */

#define my_state_is_do(opt)             (options[opt]&MY_STATE_DO)
#define my_state_is_will(opt)           (options[opt]&MY_STATE_WILL)
#define my_want_state_is_do(opt)        (options[opt]&MY_WANT_STATE_DO)
#define my_want_state_is_will(opt)      (options[opt]&MY_WANT_STATE_WILL)

#define my_state_is_dont(opt)           (!my_state_is_do(opt))
#define my_state_is_wont(opt)           (!my_state_is_will(opt))
#define my_want_state_is_dont(opt)      (!my_want_state_is_do(opt))
#define my_want_state_is_wont(opt)      (!my_want_state_is_will(opt))

#define set_my_state_do(opt)            {options[opt] |= MY_STATE_DO;}
#define set_my_state_will(opt)          {options[opt] |= MY_STATE_WILL;}
#define set_my_want_state_do(opt)       {options[opt] |= MY_WANT_STATE_DO;}
#define set_my_want_state_will(opt)     {options[opt] |= MY_WANT_STATE_WILL;}

#define set_my_state_dont(opt)          {options[opt] &= ~MY_STATE_DO;}
#define set_my_state_wont(opt)          {options[opt] &= ~MY_STATE_WILL;}
#define set_my_want_state_dont(opt)     {options[opt] &= ~MY_WANT_STATE_DO;}
#define set_my_want_state_wont(opt)     {options[opt] &= ~MY_WANT_STATE_WILL;}

/*
 * Make everything symmetrical
 */

#define HIS_STATE_WILL                  MY_STATE_DO
#define HIS_WANT_STATE_WILL             MY_WANT_STATE_DO
#define HIS_STATE_DO                    MY_STATE_WILL
#define HIS_WANT_STATE_DO               MY_WANT_STATE_WILL

#define his_state_is_do                 my_state_is_will
#define his_state_is_will               my_state_is_do
#define his_want_state_is_do            my_want_state_is_will
#define his_want_state_is_will          my_want_state_is_do

#define his_state_is_dont               my_state_is_wont
#define his_state_is_wont               my_state_is_dont
#define his_want_state_is_dont          my_want_state_is_wont
#define his_want_state_is_wont          my_want_state_is_dont

#define set_his_state_do                set_my_state_will
#define set_his_state_will              set_my_state_do
#define set_his_want_state_do           set_my_want_state_will
#define set_his_want_state_will         set_my_want_state_do

#define set_his_state_dont              set_my_state_wont
#define set_his_state_wont              set_my_state_dont
#define set_his_want_state_dont         set_my_want_state_wont
#define set_his_want_state_wont         set_my_want_state_dont


#ifdef SUPPORT_DEBUGING
static FILE    *NetTrace;          /* Where debugging output goes */

/* Name of file where debugging output goes */
static unsigned char NetTraceFile[MAXPATHLEN];

static void SetNetTrace (const char *);       /* Function to change where debugging goes */
#else
#define SetNetTrace(f)
#endif

static void init_network(void);
static void env_undefine (const unsigned char *, const unsigned char *);
static void env_export (const unsigned char *);
static void env_unexport (const unsigned char *, const unsigned char *);
static void env_send (const unsigned char *, const unsigned char *);
static void env_list (const unsigned char *, const unsigned char *);
static int stilloob(void);
static void setneturg(void);
static int netflush(void);
static void init_sys(void);
static int  TerminalWrite(char *buf, int n);
static int  TerminalRead(unsigned char *buf, int n);
static int  TerminalAutoFlush(void);
static int  TerminalSpecialChars(int c);
static void TerminalFlushOutput(void);
static void TerminalSaveState(void);
static void TerminalDefaultChars(void);
static void TerminalNewMode(int f);
static cc_t *tcval(int func);
static void TerminalSpeeds(long *input_speed, long *output_speed);
static void init_telnet(void);
static void tel_leave_binary(int rw);
static void tel_enter_binary(int rw);
static int  opt_welldefined(char *ep);
static int  telrcv(void);
static int  rlogin_susp(void);
static void intp(void);
static void sendbrk(void);
static void sendabort(void);
static void sendsusp(void);
static void sendeof(void);

#ifdef  SIGINFO
static void sendayt(void);
#endif

static void xmitAO(void);
static void xmitEL(void);
static void xmitEC(void);

static void sendnaws (void);
static void setconnmode (int);
static void setcommandmode (void);
static void setneturg (void);
static void telnet (void);
static void tel_enter_binary (int);
static void TerminalFlushOutput (void);
static void TerminalNewMode (int);
static void willoption (int);
static void wontoption (int);

static void send_do (int, int);
static void send_dont (int, int);
static void send_will (int, int);
static void send_wont (int, int);

static void lm_will (unsigned char *, int);
static void lm_wont (unsigned char *, int);
static void lm_do (unsigned char *, int);
static void lm_dont (unsigned char *, int);
static void lm_mode (unsigned char *, int, int);

static void slc_init (void);
static void slcstate (void);
static void slc_mode_export (void);
static void slc_mode_import (int);
static void slc_import (int);
static void slc_export (void);
static void slc (unsigned char *, int);
static void slc_check (void);
static void slc_start_reply (void);
static void slc_add_reply (unsigned char, unsigned char, cc_t);
static void slc_end_reply (void);
static int  slc_update (void);

static void env_opt (unsigned char *, int);
static void env_opt_start (void);
static void env_opt_start_info (void);
static void env_opt_add (unsigned char *);
static void env_opt_end (int);

static int  get_status (const char *);
static int  dosynch (const char *);

static void quit (void) __attribute__ ((noreturn));

/* terminal.c */

static void init_terminal(void);
static int  ttyflush(int drop);
static int  getconnmode(void);

/* utilities.c */

static void EmptyTerminal(void);
static void Exit(void) __attribute__ ((noreturn));
static void ExitString(char *string) __attribute__ ((noreturn));


#ifdef SUPPORT_DEBUGING
static void Dump(char direction, unsigned char *buffer, int length);
static void optionstatus(void);
static void printsub(int direction, unsigned char *pointer, int length);
static void printoption(char *direction, int cmd, int option);
#else
#define Dump(direction, buffer, length)
#define optionstatus()
#define printsub(direction, pointer, length)
#define printoption(direction, cmd, option)
#endif

# define termEofChar            new_tc.c_cc[VEOF]
# define termEraseChar          new_tc.c_cc[VERASE]
# define termIntChar            new_tc.c_cc[VINTR]
# define termKillChar           new_tc.c_cc[VKILL]
# define termQuitChar           new_tc.c_cc[VQUIT]


# ifdef        VSUSP
#  define termSuspChar          new_tc.c_cc[VSUSP]
# endif
# if    defined(VFLUSHO) && !defined(VDISCARD)
#  define VDISCARD VFLUSHO
# endif
# ifdef        VDISCARD
#  define termFlushChar         new_tc.c_cc[VDISCARD]
# endif
# ifdef VWERASE
#  define termWerasChar         new_tc.c_cc[VWERASE]
# endif
# ifdef        VREPRINT
#  define termRprntChar         new_tc.c_cc[VREPRINT]
# endif
# ifdef        VLNEXT
#  define termLiteralNextChar   new_tc.c_cc[VLNEXT]
# endif
# ifdef        VSTART
#  define termStartChar         new_tc.c_cc[VSTART]
# endif
# ifdef        VSTOP
#  define termStopChar          new_tc.c_cc[VSTOP]
# endif
# ifdef        VEOL
#  define termForw1Char         new_tc.c_cc[VEOL]
# endif
# ifdef        VEOL2
#  define termForw2Char         new_tc.c_cc[VEOL]
# endif
# ifdef        VSTATUS
#  define termAytChar           new_tc.c_cc[VSTATUS]
#endif


#define settimer(x)     clocks.x = clocks.system++


#define NETADD(c)       { *netoring.supply = c; ring_supplied(&netoring, 1); }
#define NET2ADD(c1,c2)  { NETADD(c1); NETADD(c2); }
#define NETBYTES()      (ring_full_count(&netoring))
#define NETROOM()       (ring_empty_count(&netoring))

#define TTYADD(c)       if (!(SYNCHing||flushout)) { \
				*ttyoring.supply = c; \
				ring_supplied(&ttyoring, 1); \
			}
#define TTYBYTES()      (ring_full_count(&ttyoring))
#define TTYROOM()       (ring_empty_count(&ttyoring))

/*      Various modes */
#define MODE_LOCAL_CHARS(m)     ((m)&(MODE_EDIT|MODE_TRAPSIG))
#define MODE_LOCAL_ECHO(m)      ((m)&MODE_ECHO)
#define MODE_COMMAND_LINE(m)    ((m)==-1)

#define CONTROL(x)      ((x)&0x1f)              /* CTRL(x) is not portable */

#define MODE_OUT8      0x8000 /* binary mode sans -opost */


#define LOWER(x) (isupper((int)x) ? tolower((int)x) : (x))
/*
 * The prefix function returns 0 if *s1 is not a prefix
 * of *s2.  If *s1 exactly matches *s2, the negative of
 * the length is returned.  If *s1 is a prefix of *s2,
 * the length of *s1 is returned.
 */
static int
isprefix(const char *s1, const char *s2)
{
    const char *os1;
    char c1, c2;

    if (*s1 == '\0')
	return(-1);
    os1 = s1;
    c1 = *s1;
    c2 = *s2;
    while (LOWER(c1) == LOWER(c2)) {
	if (c1 == '\0')
	    break;
	c1 = *++s1;
	c2 = *++s2;
    }
    return(*s1 ? 0 : (*s2 ? (s1 - os1) : (os1 - s1)));
}


static char **
genget(const char *name, char **table, int stlen)
     /* name to match */
     /* name entry in table */

{
    char **c, **found;
    int n;

    if (name == 0)
	return 0;

    found = 0;
    for (c = table; *c != 0; c = (char **)((char *)c + stlen)) {
	if ((n = isprefix(name, *c)) == 0)
	    continue;
	if (n < 0)              /* exact match */
	    return(c);
	if (found)
	    return(&ambiguous);
	found = c;
    }
    return(found);
}

/* Function call version of Ambiguous() */
static int Ambiguous(const void *s)
{
    return((char **)s == &ambiguous);
}


typedef int (*intrtn_t)(int, char**);
static int call(intrtn_t, ...);

typedef struct {
	const char *name;       /* command name */
	const char *help;       /* help string (NULL for no help) */
	intrtn_t handler;       /* routine which executes command */
	int     needconnect;    /* Do we need to be connected to execute? */
} Command;


static void
makeargv(void)
{
    char *cp, *cp2, c;
    char **argp = margv;

    margc = 0;
    cp = line;
    if (*cp == '!') {           /* Special case shell escape */
	strcpy(saveline, line); /* save for shell command */
	*argp++ = "!";          /* No room in string to get this */
	margc++;
	cp++;
    }
    while ((c = *cp)) {
	int inquote = 0;
	while (isspace(c))
	    c = *++cp;
	if (c == '\0')
	    break;
	*argp++ = cp;
	margc += 1;
	for (cp2 = cp; c != '\0'; c = *++cp) {
	    if (inquote) {
		if (c == inquote) {
		    inquote = 0;
		    continue;
		}
	    } else {
		if (c == '\\') {
		    if ((c = *++cp) == '\0')
			break;
		} else if (c == '"') {
		    inquote = '"';
		    continue;
		} else if (c == '\'') {
		    inquote = '\'';
		    continue;
		} else if (isspace(c))
		    break;
	    }
	    *cp2++ = c;
	}
	*cp2 = '\0';
	if (c == '\0')
	    break;
	cp++;
    }
    *argp++ = 0;
}

/*
 * Make a character string into a number.
 *
 * Todo:  1.  Could take random integers (12, 0x12, 012, 0b1).
 */

static char
special(const char *s)
{
	char c;
	char b;

	switch (*s) {
	case '^':
		b = *++s;
		if (b == '?') {
		    c = b | 0x40;               /* DEL */
		} else {
		    c = b & 0x1f;
		}
		break;
	default:
		c = *s;
		break;
	}
	return c;
}

/*
 * Construct a control character sequence
 * for a special character.
 */
static const char *
control(cc_t c)
{
	/*
	 * The only way I could get the Sun 3.5 compiler
	 * to shut up about
	 *      if ((unsigned int)c >= 0x80)
	 * was to assign "c" to an unsigned int variable...
	 * Arggg....
	 */
	unsigned int uic = (unsigned int)c;

	if (uic == 0x7f)
		return ("^?");
	if (c == (cc_t)_POSIX_VDISABLE) {
		return "off";
	}
	if (uic >= 0x80) {
		control_buf[0] = '\\';
		control_buf[1] = ((c>>6)&07) + '0';
		control_buf[2] = ((c>>3)&07) + '0';
		control_buf[3] = (c&07) + '0';
		control_buf[4] = 0;
	} else if (uic >= 0x20) {
		control_buf[0] = c;
		control_buf[1] = 0;
	} else {
		control_buf[0] = '^';
		control_buf[1] = '@'+c;
		control_buf[2] = 0;
	}
	return (const char *)control_buf;
}



/*
 *      The following are data structures and routines for
 *      the "send" command.
 *
 */

struct sendlist {
    const char  *name;          /* How user refers to it (case independent) */
    const char  *help;          /* Help information (0 ==> no help) */
    int         needconnect;    /* Need to be connected */
    int         narg;           /* Number of arguments */
    int         (*handler)(const char *);   /* Routine to perform (for special ops) */
    int         nbyte;          /* Number of bytes to send this command */
    int         what;           /* Character to be sent (<0 ==> special) */
};



static int
	send_esc(const char *),
	send_help(const char *),
	send_docmd(const char *),
	send_dontcmd(const char *),
	send_willcmd(const char *),
	send_wontcmd(const char *);

static const struct sendlist Sendlist[] = {
    { "ao",     "Send Telnet Abort output",             1, 0, 0, 2, AO },
    { "ayt",    "Send Telnet 'Are You There'",          1, 0, 0, 2, AYT },
    { "brk",    "Send Telnet Break",                    1, 0, 0, 2, BREAK },
    { "break",  0,                                      1, 0, 0, 2, BREAK },
    { "ec",     "Send Telnet Erase Character",          1, 0, 0, 2, EC },
    { "el",     "Send Telnet Erase Line",               1, 0, 0, 2, EL },
    { "escape", "Send current escape character",        1, 0, send_esc, 1, 0 },
    { "ga",     "Send Telnet 'Go Ahead' sequence",      1, 0, 0, 2, GA },
    { "ip",     "Send Telnet Interrupt Process",        1, 0, 0, 2, IP },
    { "intp",   0,                                      1, 0, 0, 2, IP },
    { "interrupt", 0,                                   1, 0, 0, 2, IP },
    { "intr",   0,                                      1, 0, 0, 2, IP },
    { "nop",    "Send Telnet 'No operation'",           1, 0, 0, 2, NOP },
    { "eor",    "Send Telnet 'End of Record'",          1, 0, 0, 2, EOR },
    { "abort",  "Send Telnet 'Abort Process'",          1, 0, 0, 2, ABORT },
    { "susp",   "Send Telnet 'Suspend Process'",        1, 0, 0, 2, SUSP },
    { "eof",    "Send Telnet End of File Character",    1, 0, 0, 2, xEOF },
    { "synch",  "Perform Telnet 'Synch operation'",     1, 0, dosynch, 2, 0 },
    { "getstatus", "Send request for STATUS",           1, 0, get_status, 6, 0 },
    { "?",      "Display send options",                 0, 0, send_help, 0, 0 },
    { "help",   0,                                      0, 0, send_help, 0, 0 },
    { "do",     0,                                      0, 1, send_docmd, 3, 0 },
    { "dont",   0,                                      0, 1, send_dontcmd, 3, 0 },
    { "will",   0,                                      0, 1, send_willcmd, 3, 0 },
    { "wont",   0,                                      0, 1, send_wontcmd, 3, 0 },
    { 0,        0,                                      0, 0, 0, 0, 0 }
};

#define GETSEND(name) ((struct sendlist *) genget(name, (char **) Sendlist, \
				sizeof(struct sendlist)))

static int
send_help(const char ATTRIBUTE_UNUSED *u)
{
    const struct sendlist *s; /* pointer to current command */

    for (s = Sendlist; s->name; s++) {
	if (s->help)
	    printf("%-15s %s\r\n", s->name, s->help);
    }
    return(0);
}

static int
sendcmd(int argc, char **argv)
{
    int count;          /* how many bytes we are going to need to send */
    int i;
    struct sendlist *s; /* pointer to current command */
    int success = 0;
    int needconnect = 0;

    if (argc < 2) {
	printf("need at least one argument for 'send' command\r\n");
	printf("'send ?' for help\r\n");
	return 0;
    }
    /*
     * First, validate all the send arguments.
     * In addition, we see how much space we are going to need, and
     * whether or not we will be doing a "SYNCH" operation (which
     * flushes the network queue).
     */
    count = 0;
    for (i = 1; i < argc; i++) {
	s = GETSEND(argv[i]);
	if (s == 0) {
	    printf("Unknown send argument '%s'\r\n'send ?' for help.\r\n",
			argv[i]);
	    return 0;
	} else if (Ambiguous(s)) {
	    printf("Ambiguous send argument '%s'\r\n'send ?' for help.\r\n",
			argv[i]);
	    return 0;
	}
	if (i + s->narg >= argc) {
	    fprintf(stderr,
	    "Need %d argument%s to 'send %s' command.  'send %s ?' for help.\r\n",
		s->narg, s->narg == 1 ? "" : "s", s->name, s->name);
	    return 0;
	}
	count += s->nbyte;
	if (s->handler == send_help) {
	    send_help(NULL);
	    return 0;
	}

	i += s->narg;
	needconnect += s->needconnect;
    }
    if (!connected && needconnect) {
	printf("?Need to be connected first.\r\n'send ?' for help\r\n");
	return 0;
    }
    /* Now, do we have enough room? */
    if (NETROOM() < count) {
	printf("There is %s TO the network\r\n"
	"to process your request.  Nothing will be done.\r\n"
	"('send synch' will throw away most data in the network\r\n"
	"buffer, if this might help.)\r\n", "Not enough room in the buffer");
	return 0;
    }
    /* OK, they are all OK, now go through again and actually send */
    count = 0;
    for (i = 1; i < argc; i++) {
	if ((s = GETSEND(argv[i])) == 0) {
	    fprintf(stderr, "Telnet 'send' error - argument disappeared!\r\n");
	    quit();
	    /*NOTREACHED*/
	}
	if (s->handler) {
	    count++;
	    success += (*s->handler)((s->narg > 0) ? argv[i+1] : NULL);
	    i += s->narg;
	} else {
	    NET2ADD(IAC, s->what);
	    printoption("SENT", IAC, s->what);
	}
    }
    return (count == success);
}


static int
send_tncmd(void (*func)(int, int), const char *cmd, const char *name)
{
    char **cpp;
//    extern char *telopts[];
    int val = 0;

    if (isprefix(name, "help") || isprefix(name, "?")) {
	int col, len;

	printf("Usage: send %s <value|option>\r\n"
	"\"value\" must be from 0 to 255\r\n"
	"Valid options are:\r\n\t", cmd);

	col = 8;
	for (cpp = telopts; *cpp; cpp++) {
	    len = strlen(*cpp) + 3;
	    if (col + len > 65) {
		printf("\r\n\t");
		col = 8;
	    }
	    printf(" \"%s\"", *cpp);
	    col += len;
	}
	printf("\r\n");
	return 0;
    }
    cpp = (char **)genget(name, (char **)telopts, sizeof(char *));
    if (Ambiguous(cpp)) {
	fprintf(stderr,"'%s': ambiguous argument ('send %s ?' for help).\r\n",
					name, cmd);
	return 0;
    }
    if (cpp) {
	val = cpp - (char **)telopts;
    } else {
	const char *cp = name;

	while (*cp >= '0' && *cp <= '9') {
	    val *= 10;
	    val += *cp - '0';
	    cp++;
	}
	if (*cp != 0) {
	    fprintf(stderr, "'%s': unknown argument ('send %s ?' for help).\r\n",
					name, cmd);
	    return 0;
	} else if (val < 0 || val > 255) {
	    fprintf(stderr, "'%s': bad value ('send %s ?' for help).\r\n",
					name, cmd);
	    return 0;
	}
    }
    if (!connected) {
	printf("?Need to be connected first.\r\n");
	return 0;
    }
    (*func)(val, 1);
    return 1;
}

static int
send_esc(const char ATTRIBUTE_UNUSED *u)
{
    NETADD(escape);
    return 1;
}

static int
send_docmd(const char *name)
{
    return(send_tncmd(send_do, "do", name));
}

static int
send_dontcmd(const char *name)
{
    return(send_tncmd(send_dont, "dont", name));
}

static int
send_willcmd(const char *name)
{
    return(send_tncmd(send_will, "will", name));
}

static int
send_wontcmd(const char *name)
{
    return(send_tncmd(send_wont, "wont", name));
}


/*
 * The following are the routines and data structures referred
 * to by the arguments to the "toggle" command.
 */

static int
lclchars(void)
{
    donelclchars = 1;
    return 1;
}

#ifdef SUPPORT_DEBUGING
static int
togdebug(void)
{
    if (net > 0 &&
	(setsockopt_1i(net, SOL_SOCKET, SO_DEBUG, debug)) < 0) {
	    bb_perror_msg("setsockopt (SO_DEBUG)");
    }
    return 1;
}
#endif /* SUPPORT_DEBUGING */

static int
togcrlf(void)
{
    printf("Will send carriage returns as telnet <CR><%s>.\r\n",
		(crlf ? "LF" : "NUL"));
    return 1;
}


static int
togbinary(int val)
{
    donebinarytoggle = 1;

    if (val >= 0) {
	binmode = val;
    } else {
	if (my_want_state_is_will(TELOPT_BINARY) &&
				my_want_state_is_do(TELOPT_BINARY)) {
	    binmode = 1;
	} else if (my_want_state_is_wont(TELOPT_BINARY) &&
				my_want_state_is_dont(TELOPT_BINARY)) {
	    binmode = 0;
	}
	val = binmode ? 0 : 1;
    }

    if (val == 1) {
	if (my_want_state_is_will(TELOPT_BINARY) &&
					my_want_state_is_do(TELOPT_BINARY)) {
	    printf("Already operating in binary mode with remote host.\r\n");
	} else {
	    printf("Negotiating binary mode with remote host.\r\n");
	    tel_enter_binary(3);
	}
    } else {
	if (my_want_state_is_wont(TELOPT_BINARY) &&
					my_want_state_is_dont(TELOPT_BINARY)) {
	    printf("Already in network ascii mode with remote host.\r\n");
	} else {
	    printf("Negotiating network ascii mode with remote host.\r\n");
	    tel_leave_binary(3);
	}
    }
    return 1;
}

static int
togrbinary(int val)
{
    donebinarytoggle = 1;

    if (val == -1)
	val = my_want_state_is_do(TELOPT_BINARY) ? 0 : 1;

    if (val == 1) {
	if (my_want_state_is_do(TELOPT_BINARY)) {
	    printf("Already receiving in binary mode.\r\n");
	} else {
	    printf("Negotiating binary mode on input.\r\n");
	    tel_enter_binary(1);
	}
    } else {
	if (my_want_state_is_dont(TELOPT_BINARY)) {
	    printf("Already receiving in network ascii mode.\r\n");
	} else {
	    printf("Negotiating network ascii mode on input.\r\n");
	    tel_leave_binary(1);
	}
    }
    return 1;
}

static int
togxbinary(int val)
{
    donebinarytoggle = 1;

    if (val == -1)
	val = my_want_state_is_will(TELOPT_BINARY) ? 0 : 1;

    if (val == 1) {
	if (my_want_state_is_will(TELOPT_BINARY)) {
	    printf("Already transmitting in binary mode.\r\n");
	} else {
	    printf("Negotiating binary mode on output.\r\n");
	    tel_enter_binary(2);
	}
    } else {
	if (my_want_state_is_wont(TELOPT_BINARY)) {
	    printf("Already transmitting in network ascii mode.\r\n");
	} else {
	    printf("Negotiating network ascii mode on output.\r\n");
	    tel_leave_binary(2);
	}
    }
    return 1;
}


static int togglehelp(void);

struct togglelist {
    const char  *name;                  /* name of toggle */
    const char  *help;                  /* help message */
    int         (*handler)(int);        /* routine to do actual setting */
    int         * const variable;
    const char  *actionexplanation;
    int         needconnect;            /* Need to be connected */
};

static const struct togglelist Togglelist[] = {
    { "autoflush",
	"flushing of output when sending interrupt characters",
	    0,
		&autoflush,
		    "flush output when sending interrupt characters" },
    { "autosynch",
	"automatic sending of interrupt characters in urgent mode",
	    0,
		&autosynch,
		    "send interrupt characters in urgent mode" },
    { "skiprc",
	"don't read ~/.telnetrc file",
	    0,
		&skiprc,
		    "skip reading of ~/.telnetrc file" },
    { "binary",
	"sending and receiving of binary data",
	    togbinary,
		0,
		    0 },
    { "inbinary",
	"receiving of binary data",
	    togrbinary,
		0,
		    0 },
    { "outbinary",
	"sending of binary data",
	    togxbinary,
		0,
		    0 },
    { "crlf",
	"sending carriage returns as telnet <CR><LF>",
	    (int (*)(int))togcrlf,
		&crlf,
		    0 },
    { "crmod",
	"mapping of received carriage returns",
	    0,
		&crmod,
		    "map carriage return on output" },
    { "localchars",
	"local recognition of certain control characters",
	    (int (*)(int))lclchars,
		&localchars,
		    "recognize certain control characters" },
    { " ", "", 0, 0 },          /* empty line */
#ifdef SUPPORT_DEBUGING
    { "debug",
	"debugging",
	    (int (*)(int))togdebug,
		&debug,
		    "turn on socket level debugging" },

    { "netdata",
	"printing of hexadecimal network data (debugging)",
	    0,
		&netdata,
		    "print hexadecimal representation of network traffic" },
    { "prettydump",
	"output of \"netdata\" to user readable format (debugging)",
	    0,
		&prettydump,
		    "print user readable output for \"netdata\"" },
    { "options",
	"viewing of options processing (debugging)",
	    0,
		&showoptions,
		    "show option processing" },
    { "termdata",
	"(debugging) toggle printing of hexadecimal terminal data",
	    0,
		&termdata,
		    "print hexadecimal representation of terminal traffic" },
#endif  /* SUPPORT_DEBUGING */
    { "?",
	0,
	    (int (*)(int))togglehelp },
    { "help",
	0,
	    (int (*)(int))togglehelp },
    { 0 }
};

static int
togglehelp(void)
{
    const struct togglelist *c;

    for (c = Togglelist; c->name; c++) {
	if (c->help) {
	    if (*c->help)
		printf("%-15s toggle %s\r\n", c->name, c->help);
	    else
		printf("\r\n");
	}
    }
    printf("\r\n");
    printf("%-15s %s\r\n", "?", "display help information");
    return 0;
}

static void
settogglehelp(int set)
{
    const struct togglelist *c;

    for (c = Togglelist; c->name; c++) {
	if (c->help) {
	    if (*c->help)
		printf("%-15s %s %s\r\n", c->name, set ? "enable" : "disable",
						c->help);
	    else
		printf("\r\n");
	}
    }
}

#define GETTOGGLE(name) (struct togglelist *) \
		genget(name, (char **) Togglelist, sizeof(struct togglelist))

static int
toggle(int argc, char *argv[])
{
    int retval = 1;
    char *name;
    const struct togglelist *c;

    if (argc < 2) {
	fprintf(stderr,
	    "Need an argument to 'toggle' command.  'toggle ?' for help.\r\n");
	return 0;
    }
    argc--;
    argv++;
    while (argc--) {
	name = *argv++;
	c = GETTOGGLE(name);
	if (Ambiguous(c)) {
	    fprintf(stderr, "'%s': ambiguous argument ('toggle ?' for help).\r\n",
					name);
	    return 0;
	} else if (c == 0) {
	    fprintf(stderr, "'%s': unknown argument ('toggle ?' for help).\r\n",
					name);
	    return 0;
	} else if (!connected && c->needconnect) {
	    printf("?Need to be connected first.\r\n");
	    printf("'send ?' for help\r\n");
	    return 0;
	} else {
	    if (c->variable) {
		*c->variable = !*c->variable;           /* invert it */
		if (c->actionexplanation) {
		    printf("%s %s.\r\n", *c->variable? "Will" : "Won't",
							c->actionexplanation);
		}
	    }
	    if (c->handler) {
		retval &= (*c->handler)(-1);
	    }
	}
    }
    return retval;
}

/*
 * The following perform the "set" command.
 */


struct setlist {
    const char *name;                   /* name */
    const char *help;                   /* help information */
    void (*handler)(const char *);
    cc_t *charp;                        /* where it is located at */
};

static const struct setlist Setlist[] = {
    { "echo",   "character to toggle local echoing on/off", 0, &echoc },
    { "escape", "character to escape back to telnet command mode", 0, &escape },
    { "rlogin", "rlogin escape character", 0, &rlogin },
#ifdef SUPPORT_DEBUGING
    { "tracefile", "file to write trace information to", SetNetTrace, (cc_t *)NetTraceFile},
#endif
    { " ", "" },
    { " ", "The following need 'localchars' to be toggled true", 0, 0 },
    { "flushoutput", "character to cause an Abort Output", 0, &termFlushChar },
    { "interrupt", "character to cause an Interrupt Process", 0, &termIntChar },
    { "quit",   "character to cause an Abort process", 0, &termQuitChar },
    { "eof",    "character to cause an EOF ", 0, &termEofChar },
    { " ", "" },
    { " ", "The following are for local editing in linemode", 0, 0 },
    { "erase",  "character to use to erase a character", 0, &termEraseChar },
    { "kill",   "character to use to erase a line", 0, &termKillChar },
    { "lnext",  "character to use for literal next", 0, &termLiteralNextChar },
    { "susp",   "character to cause a Suspend Process", 0, &termSuspChar },
    { "reprint", "character to use for line reprint", 0, &termRprntChar },
    { "worderase", "character to use to erase a word", 0, &termWerasChar },
    { "start",  "character to use for XON", 0, &termStartChar },
    { "stop",   "character to use for XOFF", 0, &termStopChar },
    { "forw1",  "alternate end of line character", 0, &termForw1Char },
    { "forw2",  "alternate end of line character", 0, &termForw2Char },
    { "ayt",    "alternate AYT character", 0, &termAytChar },
    { 0 }
};

static struct setlist *
getset(char *name)
{
    return (struct setlist *)
		genget(name, (char **) Setlist, sizeof(struct setlist));
}

static void
set_escape_char(const char *s)
{
	if (rlogin != _POSIX_VDISABLE) {
		rlogin = (s && *s) ? special(s) : _POSIX_VDISABLE;
		printf("Telnet rlogin escape character is '%s'.\r\n",
					control(rlogin));
	} else {
		escape = (s && *s) ? special(s) : _POSIX_VDISABLE;
		printf("Telnet escape character is '%s'.\r\n", control(escape));
	}
}

static int
setcmd(int argc, char *argv[])
{
    int value;
    const struct setlist *ct;
    const struct togglelist *c;

    if (argc < 2 || argc > 3) {
	printf("Format is 'set Name Value'\r\n'set ?' for help.\r\n");
	return 0;
    }
    if ((argc == 2) && (isprefix(argv[1], "?") || isprefix(argv[1], "help"))) {
	for (ct = Setlist; ct->name; ct++)
	    printf("%-15s %s\r\n", ct->name, ct->help);
	printf("\r\n");
	settogglehelp(1);
	printf("%-15s %s\r\n", "?", "display help information");
	return 0;
    }

    ct = getset(argv[1]);
    if (ct == 0) {
	c = GETTOGGLE(argv[1]);
	if (c == 0) {
	    fprintf(stderr, "'%s': unknown argument ('set ?' for help).\r\n",
			argv[1]);
	    return 0;
	} else if (Ambiguous(c)) {
	    fprintf(stderr, "'%s': ambiguous argument ('set ?' for help).\r\n",
			argv[1]);
	    return 0;
	} else if (!connected && c->needconnect) {
	    printf("?Need to be connected first.\r\n");
	    printf("'send ?' for help\r\n");
	    return 0;
	}

	if (c->variable) {
	    if ((argc == 2) || (strcmp("on", argv[2]) == 0))
		*c->variable = 1;
	    else if (strcmp("off", argv[2]) == 0)
		*c->variable = 0;
	    else {
		printf("Format is 'set togglename [on|off]'\r\n'set ?' for help.\r\n");
		return 0;
	    }
	    if (c->actionexplanation) {
		printf("%s %s.\r\n", *c->variable? "Will" : "Won't",
							c->actionexplanation);
	    }
	}
	if (c->handler)
	    (*c->handler)(1);
    } else if (argc != 3) {
	printf("Format is 'set Name Value'\r\n'set ?' for help.\r\n");
	return 0;
    } else if (Ambiguous(ct)) {
	fprintf(stderr, "'%s': ambiguous argument ('set ?' for help).\r\n",
			argv[1]);
	return 0;
    } else if (ct->handler) {
	(*ct->handler)(argv[2]);
	printf("%s set to \"%s\".\r\n", ct->name, (char *)ct->charp);
    } else {
	if (strcmp("off", argv[2])) {
	    value = special(argv[2]);
	} else {
	    value = _POSIX_VDISABLE;
	}
	*(ct->charp) = (cc_t)value;
	printf("%s character is '%s'.\r\n", ct->name, control(*(ct->charp)));
    }
    slc_check();
    return 1;
}

static int
unsetcmd(int argc, char *argv[])
{
    const struct setlist *ct;
    const struct togglelist *c;
    char *name;

    if (argc < 2) {
	fprintf(stderr,
	    "Need an argument to 'unset' command.  'unset ?' for help.\r\n");
	return 0;
    }
    if (isprefix(argv[1], "?") || isprefix(argv[1], "help")) {
	for (ct = Setlist; ct->name; ct++)
	    printf("%-15s %s\r\n", ct->name, ct->help);
	printf("\r\n");
	settogglehelp(0);
	printf("%-15s %s\r\n", "?", "display help information");
	return 0;
    }

    argc--;
    argv++;
    while (argc--) {
	name = *argv++;
	ct = getset(name);
	if (ct == 0) {
	    c = GETTOGGLE(name);
	    if (c == 0) {
		fprintf(stderr, "'%s': unknown argument ('unset ?' for help).\r\n",
			name);
		return 0;
	    } else if (Ambiguous(c)) {
		fprintf(stderr, "'%s': ambiguous argument ('unset ?' for help).\r\n",
			name);
		return 0;
	    }
	    if (c->variable) {
		*c->variable = 0;
		if (c->actionexplanation) {
		    printf("%s %s.\r\n", *c->variable? "Will" : "Won't",
							c->actionexplanation);
		}
	    }
	    if (c->handler)
		(*c->handler)(0);
	} else if (Ambiguous(ct)) {
	    fprintf(stderr, "'%s': ambiguous argument ('unset ?' for help).\r\n",
			name);
	    return 0;
	} else if (ct->handler) {
	    (*ct->handler)(0);
	    printf("%s reset to \"%s\".\r\n", ct->name, (char *)ct->charp);
	} else {
	    *(ct->charp) = _POSIX_VDISABLE;
	    printf("%s character is '%s'.\r\n", ct->name, control(*(ct->charp)));
	}
    }
    return 1;
}

/* The following are the data structures and routines for the 'mode' command. */
static int
dokludgemode(void)
{
    kludgelinemode = 1;
    send_wont(TELOPT_LINEMODE, 1);
    send_dont(TELOPT_SGA, 1);
    send_dont(TELOPT_ECHO, 1);
    return 1;
}

static int
dolinemode(void)
{
    if (kludgelinemode)
	send_dont(TELOPT_SGA, 1);
    send_will(TELOPT_LINEMODE, 1);
    send_dont(TELOPT_ECHO, 1);
    return 1;
}

static int
docharmode(void)
{
    if (kludgelinemode)
	send_do(TELOPT_SGA, 1);
    else
	send_wont(TELOPT_LINEMODE, 1);
    send_do(TELOPT_ECHO, 1);
    return 1;
}

static int
dolmmode(int bit, int on)
{
    unsigned char c;

    if (my_want_state_is_wont(TELOPT_LINEMODE)) {
	printf("?Need to have LINEMODE option enabled first.\r\n%s.\r\n", "'mode ?' for help");
	return 0;
    }

    if (on)
	c = (linemode | bit);
    else
	c = (linemode & ~bit);
    lm_mode(&c, 1, 1);
    return 1;
}

static int
tn_setmode(int bit)
{
    return dolmmode(bit, 1);
}

static int
tn_clearmode(int bit)
{
    return dolmmode(bit, 0);
}

struct modelist {
	const char *name;       /* command name */
	const char *help;       /* help string */
	int     (*handler)(int);/* routine which executes command */
	int     needconnect;    /* Do we need to be connected to execute? */
	int     arg1;
};

static int modehelp(void);

static const struct modelist ModeList[] = {
    { "character", "Disable LINEMODE option",   (int (*)(int))docharmode, 1 },
    { "",       "(or disable obsolete line-by-line mode)", 0 },
    { "line",   "Enable LINEMODE option",       (int (*)(int))dolinemode, 1 },
    { "",       "(or enable obsolete line-by-line mode)", 0 },
    { "", "", 0 },
    { "",       "These require the LINEMODE option to be enabled", 0 },
    { "isig",   "Enable signal trapping",       tn_setmode, 1, MODE_TRAPSIG },
    { "+isig",  0,                              tn_setmode, 1, MODE_TRAPSIG },
    { "-isig",  "Disable signal trapping",      tn_clearmode, 1, MODE_TRAPSIG },
    { "edit",   "Enable character editing",     tn_setmode, 1, MODE_EDIT },
    { "+edit",  0,                              tn_setmode, 1, MODE_EDIT },
    { "-edit",  "Disable character editing",    tn_clearmode, 1, MODE_EDIT },
    { "softtabs", "Enable tab expansion",       tn_setmode, 1, MODE_SOFT_TAB },
    { "+softtabs", 0,                           tn_setmode, 1, MODE_SOFT_TAB },
    { "-softtabs", "Disable character editing", tn_clearmode, 1, MODE_SOFT_TAB },
    { "litecho", "Enable literal character echo", tn_setmode, 1, MODE_LIT_ECHO },
    { "+litecho", 0,                            tn_setmode, 1, MODE_LIT_ECHO },
    { "-litecho", "Disable literal character echo", tn_clearmode, 1, MODE_LIT_ECHO },
    { "help",   0,                              (int (*)(int))modehelp, 0 },
    { "kludgeline", 0,                          (int (*)(int))dokludgemode, 1 },
    { "", "", 0 },
    { "?",      "Print help information",       (int (*)(int))modehelp, 0 },
    { 0 },
};


static int
modehelp(void)
{
    const struct modelist *mt;

    printf("format is:  'mode Mode', where 'Mode' is one of:\r\n\r\n");
    for (mt = ModeList; mt->name; mt++) {
	if (mt->help) {
	    if (*mt->help)
		printf("%-15s %s\r\n", mt->name, mt->help);
	    else
		printf("\r\n");
	}
    }
    return 0;
}

#define GETMODECMD(name) (struct modelist *) \
		genget(name, (char **) ModeList, sizeof(struct modelist))

static int
modecmd(int argc, char *argv[])
{
    const struct modelist *mt;

    if (argc != 2) {
	printf("'mode' command requires an argument\r\n%s.\r\n", "'mode ?' for help");
    } else if ((mt = GETMODECMD(argv[1])) == 0) {
	fprintf(stderr, "Unknown mode '%s' (%s).\r\n", argv[1], "'mode ?' for help");
    } else if (Ambiguous(mt)) {
	fprintf(stderr, "Ambiguous mode '%s' (%s).\r\n", argv[1], "'mode ?' for help");
    } else if (mt->needconnect && !connected) {
	printf("?Need to be connected first.\r\n%s.\r\n", "'mode ?' for help");
    } else if (mt->handler) {
	return (*mt->handler)(mt->arg1);
    }
    return 0;
}

/*
 * The following data structures and routines implement the
 * "display" command.
 */

static int
display(int argc, char *argv[])
{
    const struct togglelist *tl;
    const struct setlist *sl;

#define dotog(tl)       if (tl->variable && tl->actionexplanation) { \
			    if (*tl->variable) { \
				printf("will"); \
			    } else { \
				printf("won't"); \
			    } \
			    printf(" %s.\r\n", tl->actionexplanation); \
			}

#define doset(sl)   if (sl->name && *sl->name != ' ') { \
			if (sl->handler == 0) \
			    printf("%-15s [%s]\r\n", sl->name, control(*sl->charp)); \
			else \
			    printf("%-15s \"%s\"\r\n", sl->name, (char *)sl->charp); \
		    }

    if (argc == 1) {
	for (tl = Togglelist; tl->name; tl++) {
	    dotog(tl);
	}
	printf("\r\n");
	for (sl = Setlist; sl->name; sl++) {
	    doset(sl);
	}
    } else {
	int i;

	for (i = 1; i < argc; i++) {
	    sl = getset(argv[i]);
	    tl = GETTOGGLE(argv[i]);
	    if (Ambiguous(sl) || Ambiguous(tl)) {
		printf("?Ambiguous argument '%s'.\r\n", argv[i]);
		return 0;
	    } else if (!sl && !tl) {
		printf("?Unknown argument '%s'.\r\n", argv[i]);
		return 0;
	    } else {
		if (tl) {
		    dotog(tl);
		}
		if (sl) {
		    doset(sl);
		}
	    }
	}
    }
/*@*/optionstatus();
    return 1;
#undef  doset
#undef  dotog
}

/*
 * The following are the data structures, and many of the routines,
 * relating to command processing.
 */

/*
 * Set the escape character.
 */
static int
setescape(int argc, char *argv[])
{
	char *arg;
	char setescape_buf[50];

	printf(
	    "Deprecated usage - please use 'set escape%s%s' in the future.\r\n",
				(argc > 2)? " ":"", (argc > 2)? argv[1]: "");
	if (argc > 2)
		arg = argv[1];
	else {
		printf("new escape character: ");
		(void) fgets(setescape_buf, sizeof(setescape_buf), stdin);
		arg = setescape_buf;
	}
	if (arg[0] != '\0')
		escape = arg[0];
	printf("Escape character is '%s'.\r\n", control(escape));
	(void) fflush(stdout);
	return 1;
}

    /*VARARGS*/
static int
togcrmod(void)
{
    crmod = !crmod;
    printf("Deprecated usage - please use 'toggle crmod' in the future.\r\n"
	   "%s map carriage return on output.\r\n", crmod ? "Will" : "Won't");
    (void) fflush(stdout);
    return 1;
}

static inline int
TerminalWindowSize(int *rows, int *cols)
{
    if (get_terminal_width_height(cols, rows) >= 0)
	return 1;
    return 0;
}

    /*VARARGS*/
static int
telnetsuspend(void)
{
#ifdef  SIGTSTP
    setcommandmode();
    {
	int oldrows, oldcols, newrows, newcols, err;

	err = (TerminalWindowSize(&oldrows, &oldcols) == 0) ? 1 : 0;
	(void) kill(0, SIGTSTP);
	/*
	 * If we didn't get the window size before the SUSPEND, but we
	 * can get them now (?), then send the NAWS to make sure that
	 * we are set up for the right window size.
	 */
	if (TerminalWindowSize(&newrows, &newcols) && connected &&
	    (err || ((oldrows != newrows) || (oldcols != newcols)))) {
		sendnaws();
	}
    }
    /* reget parameters in case they were changed */
    TerminalSaveState();
    setconnmode(0);
#else
    printf("Suspend is not supported.  Try the '!' command instead\r\n");
#endif
    return 1;
}

    /*ARGSUSED*/
static int
shell(int argc)
{
	int oldrows, oldcols, newrows, newcols, err;

	setcommandmode();

	err = (TerminalWindowSize(&oldrows, &oldcols) == 0) ? 1 : 0;
	switch(vfork()) {
	case -1:
		bb_perror_msg("Fork failed");
		break;
	case 0:
		/* Fire up the shell in the child. */
		run_shell(NULL, argc > 1 ? &saveline[1] : NULL, NULL);
	default:
		(void)wait((int *)0);       /* Wait for the shell to complete */

		if (TerminalWindowSize(&newrows, &newcols) && connected &&
				(err || ((oldrows != newrows) ||
				(oldcols != newcols)))) {
			sendnaws();
		}
		break;
	}
	return 1;
}


static inline int
NetClose(int fd)
{
    return close(fd);
}

/*
 * Initialize variables.
 */
static void tninit(void)
{
    init_terminal();
    init_network();
    init_telnet();
    init_sys();
}

    /*VARARGS*/
static int
bye(int argc, char *argv[])
{

    if (connected) {
	(void) shutdown(net, 2);
	printf("Connection closed.\r\n");
	(void) NetClose(net);
	connected = 0;
	/* reset options */
	tninit();
    }
    if ((argc != 2) || (strcmp(argv[1], "fromquit") != 0)) {
	Exit();
	/* NOTREACHED */
    }
    return 0; /* NOTREACHED */
}

static void quit(void)
{
	(void) call(bye, "bye", "fromquit", 0);
	Exit();
}

static int logout(void)
{
	send_do(TELOPT_LOGOUT, 1);
	(void) netflush();
	return 1;
}


/*
 * The SLC command.
 */

struct slclist {
	const char *name;
	const char *help;
	void    (*handler)(int);
	int     arg;
};

static void slc_help(void);

static const struct slclist SlcList[] = {
    { "export", "Use local special character definitions",
						(void (*)(int))slc_mode_export,        0 },
    { "import", "Use remote special character definitions",
						slc_mode_import,        1 },
    { "check",  "Verify remote special character definitions",
						slc_mode_import,        0 },
    { "help",   0,                              (void (*)(int))slc_help,               0 },
    { "?",      "Print help information",       (void (*)(int))slc_help,               0 },
    { 0 },
};

static void
slc_help(void)
{
    const struct slclist *c;

    for (c = SlcList; c->name; c++) {
	if (c->help) {
	    if (*c->help)
		printf("%-15s %s\r\n", c->name, c->help);
	    else
		printf("\r\n");
	}
    }
}

static struct slclist *
getslc(const char *name)
{
    return (struct slclist *)
		genget(name, (char **) SlcList, sizeof(struct slclist));
}

static int
slccmd(int argc, char *argv[])
{
    struct slclist *c;

    if (argc != 2) {
	fprintf(stderr,
	    "Need an argument to 'slc' command.  'slc ?' for help.\r\n");
	return 0;
    }
    c = getslc(argv[1]);
    if (c == 0) {
	fprintf(stderr, "'%s': unknown argument ('slc ?' for help).\r\n",
				argv[1]);
	return 0;
    }
    if (Ambiguous(c)) {
	fprintf(stderr, "'%s': ambiguous argument ('slc ?' for help).\r\n",
				argv[1]);
	return 0;
    }
    (*c->handler)(c->arg);
    slcstate();
    return 1;
}

/*
 * The ENVIRON command.
 */

struct envlist {
	const char *name;
	const char *help;
	void  (*handler)(const unsigned char *, const unsigned char *);
	int   narg;
};


static struct env_lst *
env_find(const unsigned char *var)
{
	struct env_lst *ep;

	for (ep = envlisthead.next; ep; ep = ep->next) {
		if (strcmp((char *)ep->var, (char *)var) == 0)
			return(ep);
	}
	return NULL;
}


static struct env_lst *
env_define(const unsigned char *var, const unsigned char *value)
{
	struct env_lst *ep;

	if ((ep = env_find(var))) {
		if (ep->var)
			free(ep->var);
		if (ep->value)
			free(ep->value);
	} else {
		ep = xmalloc(sizeof(struct env_lst));
		ep->next = envlisthead.next;
		envlisthead.next = ep;
		ep->prev = &envlisthead;
		if (ep->next)
			ep->next->prev = ep;
	}
	ep->welldefined = opt_welldefined((char *)var);
	ep->export = 1;
	ep->var = (unsigned char *)xstrdup((char *)var);
	ep->value = (unsigned char *)xstrdup((char *)value);
	return(ep);
}

static void     env_help(const unsigned char *unuse1, const unsigned char *unuse2);

static void
env_export_2_envlist(const unsigned char *var, const unsigned char *unuse2)
{
	(void)&unuse2;
	env_export(var);
}

static void
env_define_2_envlist(const unsigned char *var, const unsigned char *value)
{
	env_define_2_envlist(var, value);
}


static const struct envlist EnvList[] = {
    { "define", "Define an environment variable", env_define_2_envlist, 2 },
    { "undefine", "Undefine an environment variable", env_undefine, 1 },
    { "export", "Mark an environment variable for automatic export", env_export_2_envlist, 1 },
    { "unexport", "Don't mark an environment variable for automatic export", env_unexport, 1 },
    { "send",   "Send an environment variable", env_send, 1 },
    { "list",   "List the current environment variables", env_list, 0 },
    { "help",   0,                              env_help, 0 },
    { "?",      "Print help information",       env_help, 0 },
    { 0 },
};

static void
env_help(const unsigned char *unuse1, const unsigned char *unuse2)
{
    const struct envlist *c;

    (void)&unuse1;
    (void)&unuse2;
    for (c = EnvList; c->name; c++) {
	if (c->help) {
	    if (*c->help)
		printf("%-15s %s\r\n", c->name, c->help);
	    else
		printf("\r\n");
	}
    }
}

static struct envlist *
getenvcmd(const char *name)
{
    return (struct envlist *)
		genget(name, (char **) EnvList, sizeof(struct envlist));
}

static int
env_cmd(int argc, char *argv[])
{
    struct envlist *c;

    if (argc < 2) {
	fprintf(stderr,
	    "Need an argument to 'environ' command.  'environ ?' for help.\r\n");
	return 0;
    }
    c = getenvcmd(argv[1]);
    if (c == 0) {
	fprintf(stderr, "'%s': unknown argument ('environ ?' for help).\r\n",
				argv[1]);
	return 0;
    }
    if (Ambiguous(c)) {
	fprintf(stderr, "'%s': ambiguous argument ('environ ?' for help).\r\n",
				argv[1]);
	return 0;
    }
    if (c->narg + 2 != argc) {
	fprintf(stderr,
	    "Need %s%d argument%s to 'environ %s' command.  'environ ?' for help.\r\n",
		c->narg < argc + 2 ? "only " : "",
		c->narg, c->narg == 1 ? "" : "s", c->name);
	return 0;
    }
    (*c->handler)((const unsigned char *)argv[2], (const unsigned char *)argv[3]);
    return 1;
}


static void
env_init(void)
{
	extern char **environ;
	char **epp, *cp;
	struct env_lst *ep;

	for (epp = environ; *epp; epp++) {
		if ((cp = strchr_set0(*epp, '='))) {
			ep = env_define((unsigned char *)*epp,
					(unsigned char *)cp+1);
			ep->export = 0;
			*cp = '=';
		}
	}
	/*
	 * Special case for DISPLAY variable.  If it is ":0.0" or
	 * "unix:0.0", we have to get rid of "unix" and insert our
	 * hostname.
	 */
	if ((ep = env_find((const unsigned char *)"DISPLAY"))
	    && ((*ep->value == ':')
		|| (strncmp((char *)ep->value, "unix:", 5) == 0))) {
		char hbuf[MAXHOSTNAMELEN];
		char *cp2 = strchr((char *)ep->value, ':');

		gethostname(hbuf, sizeof hbuf);

		/* If this is not the full name, try to get it via DNS */
		if (strchr(hbuf, '.') == 0) {
			char *out = NULL;

			bb_gethostbyname(BB_GHBN_CANON_NAME | BB_GHBN_SILENT, hbuf, &out);
			if(out != NULL) {
				safe_strncpy(hbuf, out, sizeof hbuf);
				free(out);
			}
		}

		cp = xasprintf ("%s%s", hbuf, cp2);

		free(ep->value);
		ep->value = (unsigned char *)cp;
	}
	/*
	 * If USER is not defined, but LOGNAME is, then add
	 * USER with the value from LOGNAME.  By default, we
	 * don't export the USER variable.
	 */
	if ((env_find((const unsigned char *)"USER") == NULL) &&
			(ep = env_find((const unsigned char *)"LOGNAME"))) {
		env_define((unsigned char *)"USER", ep->value);
		env_unexport((unsigned char *)"USER", NULL);
	}
	env_export((unsigned char *)"DISPLAY");
	env_export((unsigned char *)"PRINTER");
	env_export((unsigned char *)"XAUTHORITY");
}


static void
env_undefine(const unsigned char *var, const unsigned char *unuse)
{
	struct env_lst *ep;

	(void)&unuse;
	if ((ep = env_find(var))) {
		ep->prev->next = ep->next;
		if (ep->next)
			ep->next->prev = ep->prev;
		if (ep->var)
			free(ep->var);
		if (ep->value)
			free(ep->value);
		free(ep);
	}
}

static void
env_export(const unsigned char *var)
{
	struct env_lst *ep;

	if ((ep = env_find(var)))
		ep->export = 1;
}

static void
env_unexport(const unsigned char *var, const unsigned char *unuse)
{
	struct env_lst *ep;

	(void)&unuse;
	if ((ep = env_find(var)) != NULL)
		ep->export = 0;
}

static void
env_send(const unsigned char *var, const unsigned char *unuse)
{
	struct env_lst *ep;

	(void)&unuse;
	if (my_state_is_wont(TELOPT_NEW_ENVIRON)) {
		fprintf(stderr,
		    "Cannot send '%s': Telnet ENVIRON option not enabled\r\n",
									var);
		return;
	}
	ep = env_find(var);
	if (ep == 0) {
		fprintf(stderr, "Cannot send '%s': variable not defined\r\n",
									var);
		return;
	}
	env_opt_start_info();
	env_opt_add(ep->var);
	env_opt_end(0);
}

static void
env_list(const unsigned char *unuse1, const unsigned char *unuse2)
{
	struct env_lst *ep;

	(void)&unuse1;
	(void)&unuse2;

	for (ep = envlisthead.next; ep; ep = ep->next) {
		printf("%c %-20s %s\r\n", ep->export ? '*' : ' ',
					ep->var, ep->value);
	}
}

static unsigned char *
env_default(int init, int welldefined)
{

	if (init) {
		nep = &envlisthead;
		return NULL;
	}
	if (nep) {
		while ((nep = nep->next)) {
			if (nep->export && (nep->welldefined == welldefined))
				return(nep->var);
		}
	}
	return NULL;
}

static unsigned char *
env_getvalue(const unsigned char *var)
{
	struct env_lst *ep;

	if ((ep = env_find(var)))
		return(ep->value);
	return NULL;
}

/*
 * Print status about the connection.
 */
    /*ARGSUSED*/
static int
status(int  argc, char *argv[])
{
    if (connected) {
	printf("Connected to %s.\r\n", hostname);
	if ((argc < 2) || strcmp(argv[1], "notmuch")) {
	    int mode = getconnmode();

	    if (my_want_state_is_will(TELOPT_LINEMODE)) {
		printf("Operating with LINEMODE option\r\n");
		printf("%s line editing\r\n", (mode&MODE_EDIT) ? "Local" : "No");
		printf("%s catching of signals\r\n",
					(mode&MODE_TRAPSIG) ? "Local" : "No");
		slcstate();
	    } else if (kludgelinemode && my_want_state_is_dont(TELOPT_SGA)) {
		printf("Operating in obsolete linemode\r\n");
	    } else {
		printf("Operating in single character mode\r\n");
		if (localchars)
		    printf("Catching signals locally\r\n");
	    }
	    printf("%s character echo\r\n", (mode&MODE_ECHO) ? "Local" : "Remote");
	    if (my_want_state_is_will(TELOPT_LFLOW))
		printf("%s flow control\r\n", (mode&MODE_FLOW) ? "Local" : "No");
	}
    } else {
	printf("No connection.\r\n");
    }
    printf("Escape character is '%s'.\r\n", control(escape));
    (void) fflush(stdout);
    return 1;
}

#ifdef  SIGINFO
/*
 * Function that gets called when SIGINFO is received.
 */
static void
ayt_status(int sig)
{
    (void) call(status, "status", "notmuch", 0);
}
#endif

static const Command *getcmd(const char *name);

static void
cmdrc(char *m1, char *m2)
{
    const Command *c;
    FILE *rcfile;
    int gotmachine = 0;
    int l1 = strlen(m1);
    int l2 = strlen(m2);
    char m1save[MAXHOSTNAMELEN];

    if (skiprc)
	return;

    strncpy(m1save, m1, sizeof(m1save));
    m1 = m1save;

    if (rcname[0] == 0) {
	char *home = getenv("HOME");

	if (home == NULL || *home == '\0')
	    return;
	snprintf (rcname, sizeof(rcname), "%s/.telnetrc",
		  home ? home : "");
    }

    if ((rcfile = fopen(rcname, "r")) == 0) {
	return;
    }

    for (;;) {
	if (fgets(line, sizeof(line), rcfile) == NULL)
	    break;
	if (line[0] == 0)
	    break;
	if (line[0] == '#')
	    continue;
	if (gotmachine) {
	    if (!isspace(line[0]))
		gotmachine = 0;
	}
	if (gotmachine == 0) {
	    if (isspace(line[0]))
		continue;
	    if (strncasecmp(line, m1, l1) == 0)
		strncpy(line, &line[l1], sizeof(line) - l1);
	    else if (strncasecmp(line, m2, l2) == 0)
		strncpy(line, &line[l2], sizeof(line) - l2);
	    else if (strncasecmp(line, "DEFAULT", 7) == 0)
		strncpy(line, &line[7], sizeof(line) - 7);
	    else
		continue;
	    if (line[0] != ' ' && line[0] != '\t' && line[0] != '\n')
		continue;
	    gotmachine = 1;
	}
	makeargv();
	if (margv[0] == 0)
	    continue;
	c = getcmd(margv[0]);
	if (Ambiguous(c)) {
	    printf("?Ambiguous command: %s\r\n", margv[0]);
	    continue;
	}
	if (c == 0) {
	    printf("?Invalid command: %s\r\n", margv[0]);
	    continue;
	}
	/*
	 * This should never happen...
	 */
	if (c->needconnect && !connected) {
	    printf("?Need to be connected first for %s.\r\n", margv[0]);
	    continue;
	}
	(*c->handler)(margc, margv);
    }
    fclose(rcfile);
}


static int tn(int argc, char *argv[])
{
    struct addrinfo hints, *res, *res0;
    int error;
    char *cmd, *hostp = 0, *portp = 0, *user = 0, *aliasp = 0;
    int retry;
#ifdef NI_WITHSCOPEID
    const int niflags = NI_NUMERICHOST | NI_WITHSCOPEID;
#else
    const int niflags = NI_NUMERICHOST;
#endif

    if (connected) {
	printf("?Already connected to %s\r\n", hostname);
	return 0;
    }
    if (argc < 2) {
	(void) strcpy(line, "open ");
	printf("(to) ");
	(void) fgets(&line[strlen(line)], sizeof(line) - strlen(line), stdin);
	makeargv();
	argc = margc;
	argv = margv;
    }
    cmd = *argv;
    --argc; ++argv;
    while (argc) {
	if (strcmp(*argv, "help") == 0 || isprefix(*argv, "?"))
	    goto usage;
	if (strcmp(*argv, "-l") == 0) {
	    --argc; ++argv;
	    if (argc == 0)
		goto usage;
	    user = xstrdup(*argv++);
	    --argc;
	    continue;
	}
	if (strcmp(*argv, "-b") == 0) {
	    --argc; ++argv;
	    if (argc == 0)
		goto usage;
	    aliasp = *argv++;
	    --argc;
	    continue;
	}
	if (strcmp(*argv, "-a") == 0) {
	    --argc; ++argv;
	    autologin = 1;
	    continue;
	}
	if (hostp == 0) {
	    hostp = *argv++;
	    --argc;
	    continue;
	}
	if (portp == 0) {
	    portp = *argv++;
	    --argc;
	    continue;
	}
    usage:
	printf("usage: %s [-l user] [-a] host-name [port]\r\n", cmd);
	return 0;
    }
    if (hostp == 0)
	goto usage;

    hostname = hostp;
    memset(&hints, 0, sizeof(hints));
    hints.ai_family = PF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_CANONNAME;
    if (portp == NULL) {
	portp = "telnet";
	telnetport = 1;
    } else if (*portp == '-') {
	portp++;
	telnetport = 1;
    } else
	telnetport = 0;
    error = getaddrinfo(hostp, portp, &hints, &res0);
    if (error) {
	if (error == EAI_SERVICE)
	    bb_error_msg("%s: bad port", portp);
	else
	    bb_error_msg("%s: %s", hostp, gai_strerror(error));
	return 0;
    }

    net = -1;
    retry = 0;
    for (res = res0; res; res = res->ai_next) {
	if (1 /* retry */) {
	    char hbuf[NI_MAXHOST];

	    if (getnameinfo(res->ai_addr, res->ai_addrlen, hbuf, sizeof(hbuf),
		    NULL, 0, niflags) != 0) {
		strcpy(hbuf, "(invalid)");
	    }
	    printf("Trying %s...\r\n", hbuf);
	}
	net = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
	if (net < 0)
	    continue;

	if (aliasp) {
	    struct addrinfo ahints, *ares;

	    memset(&ahints, 0, sizeof(ahints));
	    ahints.ai_family = PF_UNSPEC;
	    ahints.ai_socktype = SOCK_STREAM;
	    ahints.ai_flags = AI_PASSIVE;
	    error = getaddrinfo(aliasp, "0", &ahints, &ares);
	    if (error) {
		bb_error_msg("%s: %s", aliasp, gai_strerror(error));
		close(net);
		continue;
	    }
	    if (bind(net, ares->ai_addr, ares->ai_addrlen) < 0) {
		bb_perror_msg("%s", aliasp);
		(void) close(net);   /* dump descriptor */
		freeaddrinfo(ares);
		continue;
	    }
	    freeaddrinfo(ares);
	}
	if (res->ai_family == AF_INET) {
	    tos = IPTOS_LOWDELAY;   /* Low Delay bit */
	    if (tos &&
		   (setsockopt(net, IPPROTO_IP, IP_TOS, (void *)&tos, sizeof(int)) < 0)
		&& (errno != ENOPROTOOPT))
		    bb_perror_msg("setsockopt (IP_TOS) (ignored)");
	}

#ifdef SUPPORT_DEBUGING
	if (debug && setsockopt_1i(net, SOL_SOCKET, SO_DEBUG, 1) < 0) {
		bb_perror_msg("setsockopt (SO_DEBUG)");
	}
#endif

	if (connect(net, res->ai_addr, res->ai_addrlen) < 0) {
	    char hbuf[NI_MAXHOST];

	    if (getnameinfo(res->ai_addr, res->ai_addrlen, hbuf, sizeof(hbuf),
		    NULL, 0, niflags) != 0) {
		strcpy(hbuf, "(invalid)");
	    }
	    bb_perror_msg("connect to address %s", hbuf);

	    close(net);
	    net = -1;
	    retry++;
	    continue;
	}

	connected++;
	break;
    }
    freeaddrinfo(res0);
    if (net < 0) {
	return 0;
    }
    cmdrc(hostp, hostname);
    if (autologin && user == NULL) {
	struct passwd *pw;

	user = getenv("USER");
	if (user == NULL ||
	    ((pw = getpwnam(user)) && pw->pw_uid != getuid())) {
		user = bb_getpwuid(NULL, getuid(), 0);
	}
    }
    if (user) {
	env_define((unsigned char *)"USER", (unsigned char *)user);
	env_export((unsigned char *)"USER");
    }
    (void) call(status, "status", "notmuch", 0);
    if (setjmp(peerdied) == 0)
	telnet();
    (void) NetClose(net);
    ExitString("Connection closed by foreign host");
    /*NOTREACHED*/
}

#define HELPINDENT (sizeof ("connect"))

static const char
	openhelp[] =    "connect to a site",
	closehelp[] =   "close current connection",
	logouthelp[] =  "forcibly logout remote user and close the connection",
	quithelp[] =    "exit telnet",
	statushelp[] =  "print status information",
	helphelp[] =    "print help information",
	sendhelp[] =    "transmit special characters ('send ?' for more)",
	sethelp[] =     "set operating parameters ('set ?' for more)",
	unsethelp[] =   "unset operating parameters ('unset ?' for more)",
	togglestring[] ="toggle operating parameters ('toggle ?' for more)",
	slchelp[] =     "change state of special charaters ('slc ?' for more)",
	displayhelp[] = "display operating parameters",
	zhelp[] =       "suspend telnet",
	shellhelp[] =   "invoke a subshell",
	envhelp[] =     "change environment variables ('environ ?' for more)",
	modestring[] = "try to enter line or character mode ('mode ?' for more)";

static int      help(int, char**);

static const Command cmdtab[] = {
	{ "close",      closehelp,      bye,            1 },
	{ "logout",     logouthelp,     (intrtn_t)logout,         1 },
	{ "display",    displayhelp,    display,        0 },
	{ "mode",       modestring,     modecmd,        0 },
	{ "open",       openhelp,       tn,             0 },
	{ "quit",       quithelp,       (intrtn_t)quit,           0 },
	{ "send",       sendhelp,       sendcmd,        0 },
	{ "set",        sethelp,        setcmd,         0 },
	{ "unset",      unsethelp,      unsetcmd,       0 },
	{ "status",     statushelp,     status,         0 },
	{ "toggle",     togglestring,   toggle,         0 },
	{ "slc",        slchelp,        slccmd,         0 },
	{ "z",          zhelp,          (intrtn_t)telnetsuspend,  0 },
	{ "!",          shellhelp,      (intrtn_t)shell,          0 },
	{ "environ",    envhelp,        env_cmd,        0 },
	{ "?",          helphelp,       help,           0 },
	{ 0,            0,              0,              0 }
};

static const char crmodhelp[] =   "deprecated command -- use 'toggle crmod' instead";
static const char escapehelp[] =  "deprecated command -- use 'set escape' instead";

static const Command cmdtab2[] = {
	{ "help",       0,              help,               0 },
	{ "escape",     escapehelp,     setescape,          0 },
	{ "crmod",      crmodhelp,      (intrtn_t)togcrmod, 0 },
	{ 0,            0,              0,                  0 }
};


/*
 * Call routine with argc, argv set from args (terminated by 0).
 */

    /*VARARGS1*/
static int
call(intrtn_t routine, ...)
{
    va_list ap;
    char *args[100];
    int argno = 0;

    va_start(ap, routine);
    while ((args[argno++] = va_arg(ap, char *)) != 0);
    va_end(ap);
    return (*routine)(argno-1, args);
}


static const Command *
getcmd(const char *name)
{
    const Command *cm;

    if ((cm = (const Command *) genget(name, (char **) cmdtab, sizeof(Command))))
	return cm;
    return (const Command *) genget(name, (char **) cmdtab2, sizeof(Command));
}

static void
command(int top, char *tbuf, int cnt)
{
    const Command *c;

    setcommandmode();
    if (!top) {
	putchar('\n');
#if     defined(unix)
    } else {
	bb_signal((1 << SIGINT) | (1 << SIGQUIT), SIG_DFL);
#endif  /* defined(unix) */
    }
    for (;;) {
	if (rlogin == _POSIX_VDISABLE)
		printf("%s> ", prompt);
	if (tbuf) {
	    char *cp;
	    cp = line;
	    while (cnt > 0 && (*cp++ = *tbuf++) != '\n')
		cnt--;
	    tbuf = 0;
	    if (cp == line || *--cp != '\n' || cp == line)
		goto getline;
	    *cp = '\0';
	    if (rlogin == _POSIX_VDISABLE)
		printf("%s\r\n", line);
	} else {
	getline:
	    if (rlogin != _POSIX_VDISABLE)
		printf("%s> ", prompt);
	    if (fgets(line, sizeof(line), stdin) == NULL) {
		if (feof(stdin) || ferror(stdin)) {
		    quit();
		    /*NOTREACHED*/
		}
		break;
	    }
	}
	if (line[0] == 0)
	    break;
	makeargv();
	if (margv[0] == 0) {
	    break;
	}
	c = getcmd(margv[0]);
	if (Ambiguous(c)) {
	    printf("?Ambiguous command\r\n");
	    continue;
	}
	if (c == 0) {
	    printf("?Invalid command\r\n");
	    continue;
	}
	if (c->needconnect && !connected) {
	    printf("?Need to be connected first.\r\n");
	    continue;
	}
	if ((*c->handler)(margc, margv)) {
	    break;
	}
    }
    if (!top) {
	if (!connected) {
	    Exit();
	    /*NOTREACHED*/
	}
	setconnmode(0);
    }
}

/*
 * Help command.
 */
static int
help(int argc, char *argv[])
{
	const Command *c;

	if (argc == 1) {
		printf("Commands may be abbreviated.  Commands are:\r\n\r\n");
		for (c = cmdtab; c->name; c++)
			if (c->help) {
				printf("%-*s\t%s\r\n", (int)HELPINDENT, c->name,
								    c->help);
			}
		return 0;
	}
	while (--argc > 0) {
		char *arg;
		arg = *++argv;
		c = getcmd(arg);
		if (Ambiguous(c))
			printf("?Ambiguous help command %s\r\n", arg);
		else if (c == (Command *)0)
			printf("?Invalid help command %s\r\n", arg);
		else
			printf("%s\r\n", c->help);
	}
	return 0;
}

/* Initialize internal network data structures. */
static void
init_network(void)
{
    if (ring_init(&netoring, netobuf, sizeof netobuf) != 1) {
	exit(1);
    }
    if (ring_init(&netiring, netibuf, sizeof netibuf) != 1) {
	exit(1);
    }
    SetNetTrace(0);
}


/*
 * Check to see if any out-of-band data exists on a socket (for
 * Telnet "synch" processing).
 */

static int
stilloob(void)
{
    fd_set *fdsp;
    int fd_sn;
    int value;

    fd_sn = howmany(net+1, NFDBITS) * sizeof(fd_mask);
    fdsp = (fd_set *)xmalloc(fd_sn);

    do {
	memset(fdsp, 0, fd_sn);
	FD_SET(net, fdsp);
	value = select(net+1, (fd_set *)0, (fd_set *)0, fdsp, &timeout);
    } while ((value == -1) && (errno == EINTR));

    if (value < 0) {
	bb_perror_msg("select");
	free(fdsp);
	quit();
	/* NOTREACHED */
    }
    if (FD_ISSET(net, fdsp)) {
	free(fdsp);
	return 1;
    } else {
	free(fdsp);
	return 0;
    }
}


/*
 *  setneturg()
 *
 *      Sets "neturg" to the current location.
 */

static void
setneturg(void)
{
    ring_mark(&netoring);
}


/*
 *  netflush
 *              Send as much data as possible to the network,
 *      handling requests for urgent data.
 *
 *              The return value indicates whether we did any
 *      useful work.
 */

static int
netflush(void)
{
    int n, n1;

    if ((n1 = n = ring_full_consecutive(&netoring)) > 0) {
	if (!ring_at_mark(&netoring)) {
	    n = send(net, (char *)netoring.consume, n, 0); /* normal write */
	} else {
	    /*
	     * In 4.2 (and 4.3) systems, there is some question about
	     * what byte in a sendOOB operation is the "OOB" data.
	     * To make ourselves compatible, we only send ONE byte
	     * out of band, the one WE THINK should be OOB (though
	     * we really have more the TCP philosophy of urgent data
	     * rather than the Unix philosophy of OOB data).
	     */
	    n = send(net, (char *)netoring.consume, 1, MSG_OOB);/* URGENT data */
	}
    }
    if (n < 0) {
	if (errno != ENOBUFS && errno != EWOULDBLOCK) {
	    setcommandmode();
	    bb_perror_msg("%s", hostname);
	    (void)NetClose(net);
	    ring_clear_mark(&netoring);
	    longjmp(peerdied, -1);
	    /*NOTREACHED*/
	}
	n = 0;
    }
    if (netdata && n) {
	Dump('>', netoring.consume, n);
    }
    if (n) {
	ring_consumed(&netoring, n);
	/*
	 * If we sent all, and more to send, then recurse to pick
	 * up the other half.
	 */
	if ((n1 == n) && ring_full_consecutive(&netoring)) {
	    (void) netflush();
	}
	return 1;
    } else {
	return 0;
    }
}

/*
 * This defines a structure for a ring buffer.
 *
 * The circular buffer has two parts:
 *(((
 *      full:   [consume, supply)
 *      empty:  [supply, consume)
 *]]]
 *
 */

/* Internal macros */

#define ring_subtract(d,a,b)    (((a)-(b) >= 0)? \
					(a)-(b): (((a)-(b))+(d)->size))

#define ring_increment(d,a,c)   (((a)+(c) < (d)->top)? \
					(a)+(c) : (((a)+(c))-(d)->size))

#define ring_decrement(d,a,c)   (((a)-(c) >= (d)->bottom)? \
					(a)-(c) : (((a)-(c))-(d)->size))


#define ring_empty(d) (((d)->consume == (d)->supply) && \
				((d)->consumetime >= (d)->supplytime))
#define ring_full(d) (((d)->supply == (d)->consume) && \
				((d)->supplytime > (d)->consumetime))

/* Buffer state transition routines */

static int
ring_init(Ring *ring, unsigned char *buffer, int count)
{
    memset((char *)ring, 0, sizeof *ring);

    ring->size = count;

    ring->supply = ring->consume = ring->bottom = buffer;

    ring->top = ring->bottom+ring->size;

    return 1;
}

/* Mark routines */

/*
 * Mark the most recently supplied byte.
 */

static void
ring_mark(Ring *ring)
{
    ring->mark = ring_decrement(ring, ring->supply, 1);
}

/*
 * Is the ring pointing to the mark?
 */

static int
ring_at_mark(Ring *ring)
{
    if (ring->mark == ring->consume) {
	return 1;
    } else {
	return 0;
    }
}


/*
 * Add characters from current segment to ring buffer.
 */
static void
ring_supplied(Ring *ring, int count)
{
    ring->supply = ring_increment(ring, ring->supply, count);
    ring->supplytime = ++ring_clock;
}

/*
 * We have just consumed "c" bytes.
 */
static void
ring_consumed(Ring *ring, int count)
{
    if (count == 0)     /* don't update anything */
	return;

    if (ring->mark &&
		(ring_subtract(ring, ring->mark, ring->consume) < count)) {
	ring->mark = 0;
    }
    ring->consume = ring_increment(ring, ring->consume, count);
    ring->consumetime = ++ring_clock;
    /*
     * Try to encourage "ring_empty_consecutive()" to be large.
     */
    if (ring_empty(ring)) {
	ring->consume = ring->supply = ring->bottom;
    }
}


/* Buffer state query routines */


/* Number of bytes that may be supplied */
static int
ring_empty_count(Ring *ring)
{
    if (ring_empty(ring)) {     /* if empty */
	    return ring->size;
    } else {
	return ring_subtract(ring, ring->consume, ring->supply);
    }
}

/* number of CONSECUTIVE bytes that may be supplied */
static int
ring_empty_consecutive(Ring *ring)
{
    if ((ring->consume < ring->supply) || ring_empty(ring)) {
			    /*
			     * if consume is "below" supply, or empty, then
			     * return distance to the top
			     */
	return ring_subtract(ring, ring->top, ring->supply);
    } else {
				    /*
				     * else, return what we may.
				     */
	return ring_subtract(ring, ring->consume, ring->supply);
    }
}

/* Return the number of bytes that are available for consuming
 * (but don't give more than enough to get to cross over set mark)
 */

static int
ring_full_count(Ring *ring)
{
    if ((ring->mark == 0) || (ring->mark == ring->consume)) {
	if (ring_full(ring)) {
	    return ring->size;  /* nothing consumed, but full */
	} else {
	    return ring_subtract(ring, ring->supply, ring->consume);
	}
    } else {
	return ring_subtract(ring, ring->mark, ring->consume);
    }
}

/*
 * Return the number of CONSECUTIVE bytes available for consuming.
 * However, don't return more than enough to cross over set mark.
 */
static int
ring_full_consecutive(Ring *ring)
{
    if ((ring->mark == 0) || (ring->mark == ring->consume)) {
	if ((ring->supply < ring->consume) || ring_full(ring)) {
	    return ring_subtract(ring, ring->top, ring->consume);
	} else {
	    return ring_subtract(ring, ring->supply, ring->consume);
	}
    } else {
	if (ring->mark < ring->consume) {
	    return ring_subtract(ring, ring->top, ring->consume);
	} else {        /* Else, distance to mark */
	    return ring_subtract(ring, ring->mark, ring->consume);
	}
    }
}

/* Move data into the "supply" portion of of the ring buffer. */
static void
ring_supply_data(Ring *ring, const unsigned char *buffer, int count)
{
    int i;

    while (count) {
	i = MIN(count, ring_empty_consecutive(ring));
	memmove(ring->supply, buffer, i);
	ring_supplied(ring, i);
	count -= i;
	buffer += i;
    }
}

/*
 * The following routines try to encapsulate what is system dependent
 * (at least between 4.x and dos) which is used in telnet.c.
 */


# ifndef        TCSANOW
#  ifdef TCSETS
#   define      TCSANOW         TCSETS
#   define      TCSADRAIN       TCSETSW
#   define      tcgetattr(f, t) ioctl(f, TCGETS, (char *)t)
#  else
#   ifdef TCSETA
#    define     TCSANOW         TCSETA
#    define     TCSADRAIN       TCSETAW
#    define     tcgetattr(f, t) ioctl(f, TCGETA, (char *)t)
#   else
#    define     TCSANOW         TIOCSETA
#    define     TCSADRAIN       TIOCSETAW
#    define     tcgetattr(f, t) ioctl(f, TIOCGETA, (char *)t)
#   endif
#  endif
#  define       tcsetattr(f, a, t) ioctl(f, a, (char *)t)
#  define       cfgetospeed(ptr)        ((ptr)->c_cflag&CBAUD)
#  ifdef CIBAUD
#   define      cfgetispeed(ptr)        (((ptr)->c_cflag&CIBAUD) >> IBSHIFT)
#  else
#   define      cfgetispeed(ptr)        cfgetospeed(ptr)
#  endif
# endif /* TCSANOW */
# ifdef sysV88
# define TIOCFLUSH TC_PX_DRAIN
# endif


static void
init_sys(void)
{
    tout = fileno(stdout);
    tin = fileno(stdin);

    errno = 0;
}


static int
TerminalWrite(char *buf, int n)
{
    return write(tout, buf, n);
}

static int
TerminalRead(unsigned char *buf, int n)
{
    return read(tin, buf, n);
}

/*
 *
 */

static int
TerminalAutoFlush(void)
{
#if     defined(LNOFLSH)
    int flush;

    ioctl(0, TIOCLGET, (char *)&flush);
    return !(flush&LNOFLSH);    /* if LNOFLSH, no autoflush */
#else   /* LNOFLSH */
    return 1;
#endif  /* LNOFLSH */
}

/*
 * TerminalSpecialChars()
 *
 * Look at an input character to see if it is a special character
 * and decide what to do.
 *
 * Output:
 *
 *      0       Don't add this character.
 *      1       Do add this character
 */

static int
TerminalSpecialChars(int c)
{
    if (c == termIntChar) {
	intp();
	return 0;
    } else if (c == termQuitChar) {
	if (kludgelinemode)
	    sendbrk();
	else
	    sendabort();
	return 0;
    } else if (c == termEofChar) {
	if (my_want_state_is_will(TELOPT_LINEMODE)) {
	    sendeof();
	    return 0;
	}
	return 1;
    } else if (c == termSuspChar) {
	sendsusp();
	return(0);
    } else if (c == termFlushChar) {
	xmitAO();               /* Transmit Abort Output */
	return 0;
    } else if (!MODE_LOCAL_CHARS(globalmode)) {
	if (c == termKillChar) {
	    xmitEL();
	    return 0;
	} else if (c == termEraseChar) {
	    xmitEC();           /* Transmit Erase Character */
	    return 0;
	}
    }
    return 1;
}


/*
 * Flush output to the terminal
 */

static void
TerminalFlushOutput(void)
{
#ifdef  TIOCFLUSH
    int com = FWRITE;
    (void) ioctl(fileno(stdout), TIOCFLUSH, (int *) &com);
#else
    (void) ioctl(fileno(stdout), TCFLSH, (int *) 0);
#endif
}

static void
TerminalSaveState(void)
{
    tcgetattr(0, &old_tc);

    new_tc = old_tc;

#ifndef VDISCARD
    termFlushChar = CONTROL('O');
#endif
#ifndef VWERASE
    termWerasChar = CONTROL('W');
#endif
#ifndef VREPRINT
    termRprntChar = CONTROL('R');
#endif
#ifndef VLNEXT
    termLiteralNextChar = CONTROL('V');
#endif
#ifndef VSTART
    termStartChar = CONTROL('Q');
#endif
#ifndef VSTOP
    termStopChar = CONTROL('S');
#endif
#ifndef VSTATUS
    termAytChar = CONTROL('T');
#endif
}

static cc_t *
tcval(int func)
{
    switch(func) {
    case SLC_IP:        return(&termIntChar);
    case SLC_ABORT:     return(&termQuitChar);
    case SLC_EOF:       return(&termEofChar);
    case SLC_EC:        return(&termEraseChar);
    case SLC_EL:        return(&termKillChar);
    case SLC_XON:       return(&termStartChar);
    case SLC_XOFF:      return(&termStopChar);
    case SLC_FORW1:     return(&termForw1Char);
    case SLC_FORW2:     return(&termForw2Char);
# ifdef VDISCARD
    case SLC_AO:        return(&termFlushChar);
# endif
# ifdef VSUSP
    case SLC_SUSP:      return(&termSuspChar);
# endif
# ifdef VWERASE
    case SLC_EW:        return(&termWerasChar);
# endif
# ifdef VREPRINT
    case SLC_RP:        return(&termRprntChar);
# endif
# ifdef VLNEXT
    case SLC_LNEXT:     return(&termLiteralNextChar);
# endif
# ifdef VSTATUS
    case SLC_AYT:       return(&termAytChar);
# endif

    case SLC_SYNCH:
    case SLC_BRK:
    case SLC_EOR:
    default:
	return((cc_t *)0);
    }
}

static void
TerminalDefaultChars(void)
{
    memmove(new_tc.c_cc, old_tc.c_cc, sizeof(old_tc.c_cc));
# ifndef        VDISCARD
    termFlushChar = CONTROL('O');
# endif
# ifndef        VWERASE
    termWerasChar = CONTROL('W');
# endif
# ifndef        VREPRINT
    termRprntChar = CONTROL('R');
# endif
# ifndef        VLNEXT
    termLiteralNextChar = CONTROL('V');
# endif
# ifndef        VSTART
    termStartChar = CONTROL('Q');
# endif
# ifndef        VSTOP
    termStopChar = CONTROL('S');
# endif
# ifndef        VSTATUS
    termAytChar = CONTROL('T');
# endif
}

/*
 * TerminalNewMode - set up terminal to a specific mode.
 *      MODE_ECHO: do local terminal echo
 *      MODE_FLOW: do local flow control
 *      MODE_TRAPSIG: do local mapping to TELNET IAC sequences
 *      MODE_EDIT: do local line editing
 *
 *      Command mode:
 *              MODE_ECHO|MODE_EDIT|MODE_FLOW|MODE_TRAPSIG
 *              local echo
 *              local editing
 *              local xon/xoff
 *              local signal mapping
 *
 *      Linemode:
 *              local/no editing
 *      Both Linemode and Single Character mode:
 *              local/remote echo
 *              local/no xon/xoff
 *              local/no signal mapping
 */

#ifdef SIGTSTP
static void susp(int);
#endif /* SIGTSTP */
#ifdef SIGINFO
static void ayt(int);
#endif

static void
TerminalNewMode(int f)
{
    struct termios tmp_tc;
    int onoff;
    int old;
    cc_t esc;

    globalmode = f&~MODE_FORCE;
    if (prevmode == f)
	return;

    /*
     * Write any outstanding data before switching modes
     * ttyflush() returns 0 only when there is no more data
     * left to write out, it returns -1 if it couldn't do
     * anything at all, otherwise it returns 1 + the number
     * of characters left to write.
     */
    old = ttyflush(SYNCHing|flushout);
    if (old < 0 || old > 1) {
	tcgetattr(tin, &tmp_tc);
	do {
	    /*
	     * Wait for data to drain, then flush again.
	     */
	    tcsetattr(tin, TCSADRAIN, &tmp_tc);
	    old = ttyflush(SYNCHing|flushout);
	} while (old < 0 || old > 1);
    }

    old = prevmode;
    prevmode = f&~MODE_FORCE;
    tmp_tc = new_tc;

    if (f&MODE_ECHO) {
	tmp_tc.c_lflag |= ECHO;
	tmp_tc.c_oflag |= ONLCR;
	if (crlf)
		tmp_tc.c_iflag |= ICRNL;
    } else {
	tmp_tc.c_lflag &= ~ECHO;
	tmp_tc.c_oflag &= ~ONLCR;
    }

    if ((f&MODE_FLOW) == 0) {
	tmp_tc.c_iflag &= ~(IXOFF|IXON);        /* Leave the IXANY bit alone */
    } else {
	if (restartany < 0) {
		tmp_tc.c_iflag |= IXOFF|IXON;   /* Leave the IXANY bit alone */
	} else if (restartany > 0) {
		tmp_tc.c_iflag |= IXOFF|IXON|IXANY;
	} else {
		tmp_tc.c_iflag |= IXOFF|IXON;
		tmp_tc.c_iflag &= ~IXANY;
	}
    }

    if ((f&MODE_TRAPSIG) == 0) {
	tmp_tc.c_lflag &= ~ISIG;
	localchars = 0;
    } else {
	tmp_tc.c_lflag |= ISIG;
	localchars = 1;
    }

    if (f&MODE_EDIT) {
	tmp_tc.c_lflag |= ICANON;
    } else {
	tmp_tc.c_lflag &= ~ICANON;
	tmp_tc.c_iflag &= ~ICRNL;
	tmp_tc.c_cc[VMIN] = 1;
	tmp_tc.c_cc[VTIME] = 0;
    }

    if ((f&(MODE_EDIT|MODE_TRAPSIG)) == 0) {
	tmp_tc.c_lflag &= ~IEXTEN;
    }

    if (f&MODE_SOFT_TAB) {
# ifdef OXTABS
	tmp_tc.c_oflag |= OXTABS;
# endif
# ifdef TABDLY
	tmp_tc.c_oflag &= ~TABDLY;
	tmp_tc.c_oflag |= TAB3;
# endif
    } else {
# ifdef OXTABS
	tmp_tc.c_oflag &= ~OXTABS;
# endif
# ifdef TABDLY
	tmp_tc.c_oflag &= ~TABDLY;
# endif
    }

    if (f&MODE_LIT_ECHO) {
# ifdef ECHOCTL
	tmp_tc.c_lflag &= ~ECHOCTL;
# endif
    } else {
# ifdef ECHOCTL
	tmp_tc.c_lflag |= ECHOCTL;
# endif
    }

    if (f == -1) {
	onoff = 0;
    } else {
	if (f & MODE_INBIN)
		tmp_tc.c_iflag &= ~ISTRIP;
	else
		tmp_tc.c_iflag |= ISTRIP;
	if ((f & MODE_OUTBIN) || (f & MODE_OUT8)) {
		tmp_tc.c_cflag &= ~(CSIZE|PARENB);
		tmp_tc.c_cflag |= CS8;
		if(f & MODE_OUTBIN)
			tmp_tc.c_oflag &= ~OPOST;
		else
			tmp_tc.c_oflag |= OPOST;

	} else {
		tmp_tc.c_cflag &= ~(CSIZE|PARENB);
		tmp_tc.c_cflag |= old_tc.c_cflag & (CSIZE|PARENB);
		tmp_tc.c_oflag |= OPOST;
	}
	onoff = 1;
    }

    if (f != -1) {
#ifdef  SIGTSTP
	bb_signal((1 << SIGTSTP), susp);
#endif  /* SIGTSTP */
#ifdef  SIGINFO
	bb_signal((1 << SIGINFO), ayt);
#endif
#if     defined(NOKERNINFO)
	tmp_tc.c_lflag |= NOKERNINFO;
#endif
	/*
	 * We don't want to process ^Y here.  It's just another
	 * character that we'll pass on to the back end.  It has
	 * to process it because it will be processed when the
	 * user attempts to read it, not when we send it.
	 */
# ifdef VDSUSP
	tmp_tc.c_cc[VDSUSP] = (cc_t)(_POSIX_VDISABLE);
# endif
	/*
	 * If the VEOL character is already set, then use VEOL2,
	 * otherwise use VEOL.
	 */
	esc = (rlogin != _POSIX_VDISABLE) ? rlogin : escape;
	if ((tmp_tc.c_cc[VEOL] != esc)
# ifdef VEOL2
	    && (tmp_tc.c_cc[VEOL2] != esc)
# endif
	    ) {
		if (tmp_tc.c_cc[VEOL] == (cc_t)(_POSIX_VDISABLE))
		    tmp_tc.c_cc[VEOL] = esc;
# ifdef VEOL2
		else if (tmp_tc.c_cc[VEOL2] == (cc_t)(_POSIX_VDISABLE))
		    tmp_tc.c_cc[VEOL2] = esc;
# endif
	}
    } else {
#ifdef  SIGTSTP
	sigset_t mask;
#endif  /* SIGTSTP */
#ifdef  SIGINFO
	bb_signal((1 << SIGINFO), ayt_status);
#endif
#ifdef  SIGTSTP
	bb_signal((1 << SIGTSTP), SIG_DFL);
	mask = bb_sigmask((1 << SIGTSTP));
	bb_sigunblock(mask);
#endif  /* SIGTSTP */
	tmp_tc = old_tc;
    }
    if (tcsetattr(tin, TCSADRAIN, &tmp_tc) < 0)
	tcsetattr(tin, TCSANOW, &tmp_tc);

# if    !defined(sysV88)
    ioctl(tin, FIONBIO, (char *)&onoff);
    ioctl(tout, FIONBIO, (char *)&onoff);
# endif
}

static void
TerminalSpeeds(long *ispeed, long *ospeed)
{
    speed_t in, out;

    out = cfgetospeed(&old_tc);
    in = cfgetispeed(&old_tc);
    if (in == 0)
	in = out;
    *ispeed = bb_baud_to_value(in);
    *ospeed = bb_baud_to_value(out);
}



/*
 * Various signal handling routines.
 */

static void
deadpeer(int sig)
{
	setcommandmode();
	longjmp(peerdied, sig);
}

static void
intr(int sig)
{
#if __GNUC__
	/* Avoid ARGSUSED */
	(void)&sig;
#endif
    if (localchars) {
	intp();
	return;
    }
    setcommandmode();
    Exit();
}

    /* ARGSUSED */
static void
intr2(int sig)
{
#if __GNUC__
	/* Avoid ARGSUSED */
	(void)&sig;
#endif

    if (localchars) {
	if (kludgelinemode)
	    sendbrk();
	else
	    sendabort();
    }
}

#ifdef  SIGTSTP
    /* ARGSUSED */
static void
susp(int sig)
{
#if __GNUC__
	/* Avoid ARGSUSED */
	(void)&sig;
#endif

    if ((rlogin != _POSIX_VDISABLE) && rlogin_susp())
	return;
    if (localchars)
	sendsusp();
}
#endif

#ifdef  SIGWINCH
    /* ARGSUSED */
static void
sendwin(int sig)
{
#if __GNUC__
	/* Avoid ARGSUSED */
	(void)&sig;
#endif

    if (connected) {
	sendnaws();
    }
}
#endif

#ifdef  SIGINFO
static void
ayt(int sig)
{
    if (connected)
	sendayt();
    else
	ayt_status(sig);
}
#endif


static void
sys_telnet_init(void)
{
    bb_signal((1 << SIGINT), intr);
    bb_signal((1 << SIGQUIT), intr2);
    bb_signal((1 << SIGPIPE), deadpeer);
#ifdef  SIGWINCH
    bb_signal((1 << SIGWINCH), sendwin);
#endif
#ifdef  SIGTSTP
    bb_signal((1 << SIGTSTP), susp);
#endif
#ifdef  SIGINFO
    bb_signal((1 << SIGINFO), ayt);
#endif

    setconnmode(0);

    {
	int onoff = 1;
	ioctl(net, FIONBIO, (char *)&onoff);  /* set NetNonblockingIO */
    }

    if(setsockopt_1i(net, SOL_SOCKET, SO_OOBINLINE, 1) == -1)
	bb_perror_msg("setsockopt");
}

/*
 * Process rings -
 *
 *      This routine tries to fill up/empty our various rings.
 *
 *      The parameter specifies whether this is a poll operation,
 *      or a block-until-something-happens operation.
 *
 *      The return value is 1 if something happened, 0 if not.
 */

/* If poll==0, then block until something to do */
static int
process_rings(int netin, int netout, int netex, int ttyin, int ttyout, int poll)
{
    int returnValue = 0;
    int maxfd = -1;
    int tmp;
    int c;

    if ((netout || netin || netex) && net > maxfd)
	maxfd = net;
    if (ttyout && tout > maxfd)
	maxfd = tout;
    if (ttyin && tin > maxfd)
	maxfd = tin;
    tmp = howmany(maxfd+1, NFDBITS) * sizeof(fd_mask);
    if (tmp > fdsn) {
	if (ibitsp)
	    free(ibitsp);
	if (obitsp)
	    free(obitsp);
	if (xbitsp)
	    free(xbitsp);
	fdsn = tmp;
	ibitsp = (fd_set *)xmalloc(fdsn);
	obitsp = (fd_set *)xmalloc(fdsn);
	xbitsp = (fd_set *)xmalloc(fdsn);
	memset(ibitsp, 0, fdsn);
	memset(obitsp, 0, fdsn);
	memset(xbitsp, 0, fdsn);
    }

    if (netout)
	FD_SET(net, obitsp);
    if (ttyout)
	FD_SET(tout, obitsp);
    if (ttyin)
	FD_SET(tin, ibitsp);
    if (netin)
	FD_SET(net, ibitsp);
    if (netex)
	FD_SET(net, xbitsp);

    if ((c = select(maxfd+1, ibitsp, obitsp, xbitsp,
      (poll == 0)? (struct timeval *)0 : &TimeValue)) < 0) {
	if (c == -1) {
		    /*
		     * we can get EINTR if we are in line mode,
		     * and the user does an escape (TSTP), or
		     * some other signal generator.
		     */
	    if (errno == EINTR) {
		return 0;
	    }
		    /* I don't like this, does it ever happen? */
	    printf("sleep(5) from telnet, after select\r\n");
	    sleep(5);
	}
	return 0;
    }

    /*
     * Any urgent data?
     */
    if (FD_ISSET(net, xbitsp)) {
	FD_CLR(net, xbitsp);
	SYNCHing = 1;
	(void) ttyflush(1);     /* flush already enqueued data */
    }

    /*
     * Something to read from the network...
     */
    if (FD_ISSET(net, ibitsp)) {
	int canread;

	FD_CLR(net, ibitsp);
	canread = ring_empty_consecutive(&netiring);
	c = recv(net, (char *)netiring.supply, canread, 0);
	if (c < 0 && errno == EWOULDBLOCK) {
	    c = 0;
	} else if (c <= 0) {
	    return -1;
	}
	if (netdata) {
	    Dump('<', netiring.supply, c);
	}
	if (c)
	    ring_supplied(&netiring, c);
	returnValue = 1;
    }

    /*
     * Something to read from the tty...
     */
    if (FD_ISSET(tin, ibitsp)) {
	FD_CLR(tin, ibitsp);
	c = TerminalRead(ttyiring.supply, ring_empty_consecutive(&ttyiring));
	if (c < 0 && errno == EIO)
	    c = 0;
	if (c < 0 && errno == EWOULDBLOCK) {
	    c = 0;
	} else {
	    /* EOF detection for line mode!!!! */
	    if ((c == 0) && MODE_LOCAL_CHARS(globalmode) && isatty(tin)) {
		/* must be an EOF... */
		*ttyiring.supply = termEofChar;
		c = 1;
	    }
	    if (c <= 0) {
		return -1;
	    }
	    if (termdata) {
		Dump('<', ttyiring.supply, c);
	    }
	    ring_supplied(&ttyiring, c);
	}
	returnValue = 1;                /* did something useful */
    }

    if (FD_ISSET(net, obitsp)) {
	FD_CLR(net, obitsp);
	returnValue |= netflush();
    }
    if (FD_ISSET(tout, obitsp)) {
	FD_CLR(tout, obitsp);
	returnValue |= (ttyflush(SYNCHing|flushout) > 0);
    }

    return returnValue;
}

#define        strip(x) (eight ? (x) : ((x) & 0x7f))

#define SB_CLEAR()      subpointer = subbuffer;
#define SB_TERM()       { subend = subpointer; SB_CLEAR(); }
#define SB_ACCUM(c)     if (subpointer < (subbuffer+sizeof subbuffer)) { \
				*subpointer++ = (c); \
			}

#define SB_GET()        ((*subpointer++)&0xff)
#define SB_PEEK()       ((*subpointer)&0xff)
#define SB_EOF()        (subpointer >= subend)
#define SB_LEN()        (subend - subpointer)


/*
 * Telnet receiver states for fsm
 */
#define TS_DATA         0
#define TS_IAC          1
#define TS_WILL         2
#define TS_WONT         3
#define TS_DO           4
#define TS_DONT         5
#define TS_CR           6
#define TS_SB           7               /* sub-option collection */
#define TS_SE           8               /* looking for sub-option end */

#define telopt_environ TELOPT_NEW_ENVIRON

/*
 * Initialize telnet environment.
 */

static void
init_telnet(void)
{
    env_init();

    SB_CLEAR();
    memset((char *)options, 0, sizeof options);

    connected = ISend = localflow = donebinarytoggle = 0;
    restartany = -1;

    SYNCHing = 0;

    /* Don't change NetTrace */

    escape = CONTROL(']');
    rlogin = _POSIX_VDISABLE;
    echoc = CONTROL('E');

    flushline = 1;
    telrcv_state = TS_DATA;
}


/*
 * These routines are in charge of sending option negotiations
 * to the other side.
 *
 * The basic idea is that we send the negotiation if either side
 * is in disagreement as to what the current state should be.
 */

static void
send_do(int c, int init)
{
    if (init) {
	if (((do_dont_resp[c] == 0) && my_state_is_do(c)) ||
				my_want_state_is_do(c))
	    return;
	set_my_want_state_do(c);
	do_dont_resp[c]++;
    }
    NET2ADD(IAC, DO);
    NETADD(c);
    printoption("SENT",DO, c);
}

static void
send_dont(int c, int init)
{
    if (init) {
	if (((do_dont_resp[c] == 0) && my_state_is_dont(c)) ||
				my_want_state_is_dont(c))
	    return;
	set_my_want_state_dont(c);
	do_dont_resp[c]++;
    }
    NET2ADD(IAC, DONT);
    NETADD(c);
    printoption("SENT", DONT, c);
}

static void
send_will(int c, int init)
{
    if (init) {
	if (((will_wont_resp[c] == 0) && my_state_is_will(c)) ||
				my_want_state_is_will(c))
	    return;
	set_my_want_state_will(c);
	will_wont_resp[c]++;
    }
    NET2ADD(IAC, WILL);
    NETADD(c);
    printoption("SENT", WILL, c);
}

static void
send_wont(int c, int init)
{
    if (init) {
	if (((will_wont_resp[c] == 0) && my_state_is_wont(c)) ||
				my_want_state_is_wont(c))
	    return;
	set_my_want_state_wont(c);
	will_wont_resp[c]++;
    }
    NET2ADD(IAC, WONT);
    NETADD(c);
    printoption("SENT", WONT, c);
}


static void
willoption(int option)
{
	int new_state_ok = 0;

	if (do_dont_resp[option]) {
	    --do_dont_resp[option];
	    if (do_dont_resp[option] && my_state_is_do(option))
		--do_dont_resp[option];
	}

	if ((do_dont_resp[option] == 0) && my_want_state_is_dont(option)) {

	    switch (option) {

	    case TELOPT_ECHO:
	    case TELOPT_BINARY:
	    case TELOPT_SGA:
		settimer(modenegotiated);
		/* FALL THROUGH */
	    case TELOPT_STATUS:
		new_state_ok = 1;
		break;

	    case TELOPT_TM:
		if (flushout)
		    flushout = 0;
		/*
		 * Special case for TM.  If we get back a WILL,
		 * pretend we got back a WONT.
		 */
		set_my_want_state_dont(option);
		set_my_state_dont(option);
		return;                 /* Never reply to TM will's/wont's */

	    case TELOPT_LINEMODE:
	    default:
		break;
	    }

	    if (new_state_ok) {
		set_my_want_state_do(option);
		send_do(option, 0);
		setconnmode(0);         /* possibly set new tty mode */
	    } else {
		do_dont_resp[option]++;
		send_dont(option, 0);
	    }
	}
	set_my_state_do(option);
}

static void
wontoption(int option)
{
	if (do_dont_resp[option]) {
	    --do_dont_resp[option];
	    if (do_dont_resp[option] && my_state_is_dont(option))
		--do_dont_resp[option];
	}

	if ((do_dont_resp[option] == 0) && my_want_state_is_do(option)) {

	    switch (option) {

	    case TELOPT_SGA:
		if (!kludgelinemode)
		    break;
		/* FALL THROUGH */
	    case TELOPT_ECHO:
		settimer(modenegotiated);
		break;

	    case TELOPT_TM:
		if (flushout)
		    flushout = 0;
		set_my_want_state_dont(option);
		set_my_state_dont(option);
		return;         /* Never reply to TM will's/wont's */

	    default:
		break;
	    }
	    set_my_want_state_dont(option);
	    if (my_state_is_do(option))
		send_dont(option, 0);
	    setconnmode(0);                     /* Set new tty mode */
	} else if (option == TELOPT_TM) {
	    /*
	     * Special case for TM.
	     */
	    if (flushout)
		flushout = 0;
	    set_my_want_state_dont(option);
	}
	set_my_state_dont(option);
}

static void
dooption(int option)
{
	int new_state_ok = 0;

	if (will_wont_resp[option]) {
	    --will_wont_resp[option];
	    if (will_wont_resp[option] && my_state_is_will(option))
		--will_wont_resp[option];
	}

	if (will_wont_resp[option] == 0) {
	  if (my_want_state_is_wont(option)) {

	    switch (option) {

	    case TELOPT_TM:
		/*
		 * Special case for TM.  We send a WILL, but pretend
		 * we sent WONT.
		 */
		send_will(option, 0);
		set_my_want_state_wont(TELOPT_TM);
		set_my_state_wont(TELOPT_TM);
		return;

	    case TELOPT_BINARY:         /* binary mode */
	    case TELOPT_NAWS:           /* window size */
	    case TELOPT_TSPEED:         /* terminal speed */
	    case TELOPT_LFLOW:          /* local flow control */
	    case TELOPT_TTYPE:          /* terminal type option */
	    case TELOPT_SGA:            /* no big deal */
		new_state_ok = 1;
		break;

	    case TELOPT_NEW_ENVIRON:    /* New environment variable option */
		new_state_ok = 1;
		break;

	    case TELOPT_XDISPLOC:       /* X Display location */
		if (env_getvalue((unsigned char *)"DISPLAY"))
		    new_state_ok = 1;
		break;

	    case TELOPT_LINEMODE:
		kludgelinemode = 0;
		send_do(TELOPT_SGA, 1);
		set_my_want_state_will(TELOPT_LINEMODE);
		send_will(option, 0);
		set_my_state_will(TELOPT_LINEMODE);
		slc_init();
		return;

	    case TELOPT_ECHO:           /* We're never going to echo... */
	    default:
		break;
	    }

	    if (new_state_ok) {
		set_my_want_state_will(option);
		send_will(option, 0);
		setconnmode(0);                 /* Set new tty mode */
	    } else {
		will_wont_resp[option]++;
		send_wont(option, 0);
	    }
	  } else {
	    /*
	     * Handle options that need more things done after the
	     * other side has acknowledged the option.
	     */
	    switch (option) {
	    case TELOPT_LINEMODE:
		kludgelinemode = 0;
		send_do(TELOPT_SGA, 1);
		set_my_state_will(option);
		slc_init();
		send_do(TELOPT_SGA, 0);
		return;
	    }
	  }
	}
	set_my_state_will(option);
}

static void
dontoption(int option)
{

	if (will_wont_resp[option]) {
	    --will_wont_resp[option];
	    if (will_wont_resp[option] && my_state_is_wont(option))
		--will_wont_resp[option];
	}

	if ((will_wont_resp[option] == 0) && my_want_state_is_will(option)) {
	    switch (option) {
	    case TELOPT_LINEMODE:
		linemode = 0;   /* put us back to the default state */
		break;
	    }
	    /* we always accept a DONT */
	    set_my_want_state_wont(option);
	    if (my_state_is_will(option))
		send_wont(option, 0);
	    setconnmode(0);                     /* Set new tty mode */
	}
	set_my_state_wont(option);
}

/*
 * This routine will turn a pipe separated list of names in the buffer
 * into an array of pointers to NUL terminated names.  We toss out any
 * bad, duplicate, or verbose names (names with spaces).
 */

static const char *
gettermname(void)
{
	static const char name_unknown[] = "UNKNOWN";
	char *tname;

	if ((tname = (char *)env_getvalue((unsigned char *)"TERM"))) {
		return tname;
	} else {
		return name_unknown;
	}
}

/*
 * suboption()
 *
 *      Look at the sub-option buffer, and try to be helpful to the other
 * side.
 *
 *      Currently we recognize:
 *
 *              Terminal type, send request.
 *              Terminal speed (send request).
 *              Local flow control (is request).
 *              Linemode
 */

static void
suboption(void)
{
    unsigned char subchar;

    printsub('<', subbuffer, SB_LEN()+2);
    switch (subchar = SB_GET()) {
    case TELOPT_TTYPE:
	if (my_want_state_is_wont(TELOPT_TTYPE))
	    return;
	if (SB_EOF() || SB_GET() != TELQUAL_SEND) {
	    return;
	} else {
	    const char *name;
	    unsigned char temp[50];
	    int len;

	    name = gettermname();
	    len = strlen(name) + 4 + 2;
	    if (len < NETROOM()) {
		snprintf((char *)temp, sizeof(temp),
			 "%c%c%c%c%s%c%c", IAC, SB, TELOPT_TTYPE,
			 TELQUAL_IS, name, IAC, SE);
		ring_supply_data(&netoring, temp, len);
		printsub('>', &temp[2], len-2);
	    } else {
		ExitString("Not enough room in the buffer");
		/*NOTREACHED*/
	    }
	}
	break;
    case TELOPT_TSPEED:
	if (my_want_state_is_wont(TELOPT_TSPEED))
	    return;
	if (SB_EOF())
	    return;
	if (SB_GET() == TELQUAL_SEND) {
	    long ospeed, ispeed;
	    unsigned char temp[50];
	    int len;

	    TerminalSpeeds(&ispeed, &ospeed);

	    snprintf((char *)temp, sizeof(temp),
		     "%c%c%c%c%ld,%ld%c%c", IAC, SB, TELOPT_TSPEED,
		     TELQUAL_IS, ospeed, ispeed, IAC, SE);
	    len = strlen((char *)temp+4) + 4;   /* temp[3] is 0 ... */

	    if (len < NETROOM()) {
		ring_supply_data(&netoring, temp, len);
		printsub('>', temp+2, len - 2);
	    }
/*@*/       else printf("Not enough room in the buffer");
	}
	break;
    case TELOPT_LFLOW:
	if (my_want_state_is_wont(TELOPT_LFLOW))
	    return;
	if (SB_EOF())
	    return;
	switch(SB_GET()) {
	case LFLOW_RESTART_ANY:
	    restartany = 1;
	    break;
	case LFLOW_RESTART_XON:
	    restartany = 0;
	    break;
	case LFLOW_ON:
	    localflow = 1;
	    break;
	case LFLOW_OFF:
	    localflow = 0;
	    break;
	default:
	    return;
	}
	setcommandmode();
	setconnmode(0);
	break;

    case TELOPT_LINEMODE:
	if (my_want_state_is_wont(TELOPT_LINEMODE))
	    return;
	if (SB_EOF())
	    return;
	switch (SB_GET()) {
	case WILL:
	    lm_will(subpointer, SB_LEN());
	    break;
	case WONT:
	    lm_wont(subpointer, SB_LEN());
	    break;
	case DO:
	    lm_do(subpointer, SB_LEN());
	    break;
	case DONT:
	    lm_dont(subpointer, SB_LEN());
	    break;
	case LM_SLC:
	    slc(subpointer, SB_LEN());
	    break;
	case LM_MODE:
	    lm_mode(subpointer, SB_LEN(), 0);
	    break;
	default:
	    break;
	}
	break;

    case TELOPT_NEW_ENVIRON:
	if (SB_EOF())
	    return;
	switch(SB_PEEK()) {
	case TELQUAL_IS:
	case TELQUAL_INFO:
	    if (my_want_state_is_dont(subchar))
		return;
	    break;
	case TELQUAL_SEND:
	    if (my_want_state_is_wont(subchar)) {
		return;
	    }
	    break;
	default:
	    return;
	}
	env_opt(subpointer, SB_LEN());
	break;

    case TELOPT_XDISPLOC:
	if (my_want_state_is_wont(TELOPT_XDISPLOC))
	    return;
	if (SB_EOF())
	    return;
	if (SB_GET() == TELQUAL_SEND) {
	    unsigned char temp[50], *dp;
	    int len;

	    if ((dp = env_getvalue((unsigned char *)"DISPLAY")) == NULL) {
		/*
		 * Something happened, we no longer have a DISPLAY
		 * variable.  So, turn off the option.
		 */
		send_wont(TELOPT_XDISPLOC, 1);
		break;
	    }
	    snprintf((char *)temp, sizeof(temp),
		    "%c%c%c%c%s%c%c", IAC, SB, TELOPT_XDISPLOC,
		    TELQUAL_IS, dp, IAC, SE);
	    len = strlen((char *)temp+4) + 4;   /* temp[3] is 0 ... */

	    if (len < NETROOM()) {
		ring_supply_data(&netoring, temp, len);
		printsub('>', temp+2, len - 2);
	    }
/*@*/       else printf("Not enough room in the buffer");
	}
	break;

    default:
	break;
    }
}

static void
lm_will(unsigned char *cmd, int len)
{
    if (len < 1) {
/*@*/   printf("lm_will: no command!!!\n");     /* Should not happen... */
	return;
    }
    switch(cmd[0]) {
    case LM_FORWARDMASK:        /* We shouldn't ever get this... */
    default:
	str_lm[3] = DONT;
	str_lm[4] = cmd[0];
	if (NETROOM() > (int)(sizeof(str_lm))) {
	    ring_supply_data(&netoring, str_lm, sizeof(str_lm));
	    printsub('>', &str_lm[2], sizeof(str_lm)-2);
	}
/*@*/   else printf("Not enough room in the buffer");
	break;
    }
}

static void
lm_wont(unsigned char *cmd, int len)
{
    if (len < 1) {
/*@*/   printf("lm_wont: no command!!!\n");     /* Should not happen... */
	return;
    }
    switch(cmd[0]) {
    case LM_FORWARDMASK:        /* We shouldn't ever get this... */
    default:
	/* We are always DONT, so don't respond */
	return;
    }
}

static void
lm_do(unsigned char *cmd, int len)
{
    if (len < 1) {
/*@*/   printf("lm_do: no command!!!\n");       /* Should not happen... */
	return;
    }
    switch(cmd[0]) {
    case LM_FORWARDMASK:
    default:
	str_lm[3] = WONT;
	str_lm[4] = cmd[0];
	if (NETROOM() > (int)(sizeof(str_lm))) {
	    ring_supply_data(&netoring, str_lm, sizeof(str_lm));
	    printsub('>', &str_lm[2], sizeof(str_lm)-2);
	}
/*@*/   else printf("Not enough room in the buffer");
	break;
    }
}

static void
lm_dont(unsigned char *cmd, int len)
{
    if (len < 1) {
/*@*/   printf("lm_dont: no command!!!\n");     /* Should not happen... */
	return;
    }
    switch(cmd[0]) {
    case LM_FORWARDMASK:
    default:
	/* we are always WONT, so don't respond */
	break;
    }
}


static void
lm_mode(unsigned char *cmd, int len, int init)
{
	if (len != 1)
		return;
	if ((linemode&MODE_MASK&~MODE_ACK) == *cmd)
		return;
	if (*cmd&MODE_ACK)
		return;
	linemode = *cmd&(MODE_MASK&~MODE_ACK);
	str_lm_mode[4] = linemode;
	if (!init)
	    str_lm_mode[4] |= MODE_ACK;
	if (NETROOM() > (int)(sizeof(str_lm_mode))) {
	    ring_supply_data(&netoring, str_lm_mode, sizeof(str_lm_mode));
	    printsub('>', &str_lm_mode[2], sizeof(str_lm_mode)-2);
	}
/*@*/   else printf("Not enough room in the buffer");
	setconnmode(0); /* set changed mode */
}



static void
slc_init(void)
{
	struct spc *spcp;

	localchars = 1;
	for (spcp = spc_data; spcp < &spc_data[NSLC+1]; spcp++) {
		spcp->val = 0;
		spcp->valp = 0;
		spcp->flags = spcp->mylevel = SLC_NOSUPPORT;
	}

#define initfunc(func, flags) { \
					spcp = &spc_data[func]; \
					if ((spcp->valp = tcval(func))) { \
					    spcp->val = *spcp->valp; \
					    spcp->mylevel = SLC_VARIABLE|flags; \
					} else { \
					    spcp->val = 0; \
					    spcp->mylevel = SLC_DEFAULT; \
					} \
				    }

	initfunc(SLC_SYNCH, 0);
	/* No BRK */
	initfunc(SLC_AO, 0);
	initfunc(SLC_AYT, 0);
	/* No EOR */
	initfunc(SLC_ABORT, SLC_FLUSHIN|SLC_FLUSHOUT);
	initfunc(SLC_EOF, 0);
#ifndef SYSV_TERMIO
	initfunc(SLC_SUSP, SLC_FLUSHIN);
#endif
	initfunc(SLC_EC, 0);
	initfunc(SLC_EL, 0);
#ifndef SYSV_TERMIO
	initfunc(SLC_EW, 0);
	initfunc(SLC_RP, 0);
	initfunc(SLC_LNEXT, 0);
#endif
	initfunc(SLC_XON, 0);
	initfunc(SLC_XOFF, 0);
#ifdef  SYSV_TERMIO
	spc_data[SLC_XON].mylevel = SLC_CANTCHANGE;
	spc_data[SLC_XOFF].mylevel = SLC_CANTCHANGE;
#endif
	initfunc(SLC_FORW1, 0);
	initfunc(SLC_FORW2, 0);
	/* No FORW2 */

	initfunc(SLC_IP, SLC_FLUSHIN|SLC_FLUSHOUT);
#undef  initfunc

	if (slc_mode == SLC_EXPORT)
		slc_export();
	else
		slc_import(1);

}

static void
slcstate(void)
{
    printf("Special characters are %s values\n",
		slc_mode == SLC_IMPORT ? "remote default" :
		slc_mode == SLC_EXPORT ? "local" :
					 "remote");
}

static void
slc_mode_export(void)
{
    slc_mode = SLC_EXPORT;
    if (my_state_is_will(TELOPT_LINEMODE))
	slc_export();
}

static void
slc_mode_import(int def)
{
    slc_mode = def ? SLC_IMPORT : SLC_RVALUE;
    if (my_state_is_will(TELOPT_LINEMODE))
	slc_import(def);
}

static const unsigned char slc_import_val[] = {
	IAC, SB, TELOPT_LINEMODE, LM_SLC, 0, SLC_VARIABLE, 0, IAC, SE
};

static const unsigned char slc_import_def[] = {
	IAC, SB, TELOPT_LINEMODE, LM_SLC, 0, SLC_DEFAULT, 0, IAC, SE
};

static void
slc_import(int def)
{
    if (NETROOM() > (int)(sizeof(slc_import_val))) {
	if (def) {
	    ring_supply_data(&netoring, slc_import_def, sizeof(slc_import_def));
	    printsub('>', &slc_import_def[2], sizeof(slc_import_def)-2);
	} else {
	    ring_supply_data(&netoring, slc_import_val, sizeof(slc_import_val));
	    printsub('>', &slc_import_val[2], sizeof(slc_import_val)-2);
	}
    }
/*@*/ else printf("Not enough room in the buffer");
}

static void
slc_export(void)
{
    struct spc *spcp;

    TerminalDefaultChars();

    slc_start_reply();
    for (spcp = &spc_data[1]; spcp < &spc_data[NSLC+1]; spcp++) {
	if (spcp->mylevel != SLC_NOSUPPORT) {
	    if (spcp->val == (cc_t)(_POSIX_VDISABLE))
		spcp->flags = SLC_NOSUPPORT;
	    else
		spcp->flags = spcp->mylevel;
	    if (spcp->valp)
		spcp->val = *spcp->valp;
	    slc_add_reply(spcp - spc_data, spcp->flags, spcp->val);
	}
    }
    slc_end_reply();
    (void)slc_update();
    setconnmode(1);     /* Make sure the character values are set */
}

static void
slc(unsigned char *cp, int len)
{
	struct spc *spcp;
	int func, level;

	slc_start_reply();

	for (; len >= 3; len -=3, cp +=3) {

		func = cp[SLC_FUNC];

		if (func == 0) {
			/*
			 * Client side: always ignore 0 function.
			 */
			continue;
		}
		if (func > NSLC) {
			if ((cp[SLC_FLAGS] & SLC_LEVELBITS) != SLC_NOSUPPORT)
				slc_add_reply(func, SLC_NOSUPPORT, 0);
			continue;
		}

		spcp = &spc_data[func];

		level = cp[SLC_FLAGS]&(SLC_LEVELBITS|SLC_ACK);

		if ((cp[SLC_VALUE] == (unsigned char)spcp->val) &&
		    ((level&SLC_LEVELBITS) == (spcp->flags&SLC_LEVELBITS))) {
			continue;
		}

		if (level == (SLC_DEFAULT|SLC_ACK)) {
			/*
			 * This is an error condition, the SLC_ACK
			 * bit should never be set for the SLC_DEFAULT
			 * level.  Our best guess to recover is to
			 * ignore the SLC_ACK bit.
			 */
			cp[SLC_FLAGS] &= ~SLC_ACK;
		}

		if (level == ((spcp->flags&SLC_LEVELBITS)|SLC_ACK)) {
			spcp->val = (cc_t)cp[SLC_VALUE];
			spcp->flags = cp[SLC_FLAGS];    /* include SLC_ACK */
			continue;
		}

		level &= ~SLC_ACK;

		if (level <= (spcp->mylevel&SLC_LEVELBITS)) {
			spcp->flags = cp[SLC_FLAGS]|SLC_ACK;
			spcp->val = (cc_t)cp[SLC_VALUE];
		}
		if (level == SLC_DEFAULT) {
			if ((spcp->mylevel&SLC_LEVELBITS) != SLC_DEFAULT)
				spcp->flags = spcp->mylevel;
			else
				spcp->flags = SLC_NOSUPPORT;
		}
		slc_add_reply(func, spcp->flags, spcp->val);
	}
	slc_end_reply();
	if (slc_update())
		setconnmode(1); /* set the  new character values */
}

static void
slc_check(void)
{
    struct spc *spcp;

    slc_start_reply();
    for (spcp = &spc_data[1]; spcp < &spc_data[NSLC+1]; spcp++) {
	if (spcp->valp && spcp->val != *spcp->valp) {
	    spcp->val = *spcp->valp;
	    if (spcp->val == (cc_t)(_POSIX_VDISABLE))
		spcp->flags = SLC_NOSUPPORT;
	    else
		spcp->flags = spcp->mylevel;
	    slc_add_reply(spcp - spc_data, spcp->flags, spcp->val);
	}
    }
    slc_end_reply();
    setconnmode(1);
}


static void
slc_start_reply(void)
{
	slc_replyp = slc_reply;
	*slc_replyp++ = IAC;
	*slc_replyp++ = SB;
	*slc_replyp++ = TELOPT_LINEMODE;
	*slc_replyp++ = LM_SLC;
}

static void
slc_add_reply(unsigned char func, unsigned char flags, cc_t value)
{
	if ((*slc_replyp++ = func) == IAC)
		*slc_replyp++ = IAC;
	if ((*slc_replyp++ = flags) == IAC)
		*slc_replyp++ = IAC;
	if ((*slc_replyp++ = (unsigned char)value) == IAC)
		*slc_replyp++ = IAC;
}

static void
slc_end_reply(void)
{
    int len;

    *slc_replyp++ = IAC;
    *slc_replyp++ = SE;
    len = slc_replyp - slc_reply;
    if (len <= 6)
	return;
    if (NETROOM() > len) {
	ring_supply_data(&netoring, slc_reply, slc_replyp - slc_reply);
	printsub('>', &slc_reply[2], slc_replyp - slc_reply - 2);
    }
/*@*/else printf("Not enough room in the buffer");
}

static int
slc_update(void)
{
	struct spc *spcp;
	int need_update = 0;

	for (spcp = &spc_data[1]; spcp < &spc_data[NSLC+1]; spcp++) {
		if (!(spcp->flags&SLC_ACK))
			continue;
		spcp->flags &= ~SLC_ACK;
		if (spcp->valp && (*spcp->valp != spcp->val)) {
			*spcp->valp = spcp->val;
			need_update = 1;
		}
	}
	return(need_update);
}

static void
env_opt(unsigned char *buf, int len)
{
	unsigned char *ep = 0, *epc = 0;
	int i;

	switch(buf[0]&0xff) {
	case TELQUAL_SEND:
		env_opt_start();
		if (len == 1) {
			env_opt_add(NULL);
		} else for (i = 1; i < len; i++) {
			switch (buf[i]&0xff) {
			case NEW_ENV_VAR:
			case ENV_USERVAR:
				if (ep) {
					*epc = 0;
					env_opt_add(ep);
				}
				ep = epc = &buf[i+1];
				break;
			case ENV_ESC:
				i++;
				/*FALL THROUGH*/
			default:
				if (epc)
					*epc++ = buf[i];
				break;
			}
		}
		if (ep) {
			*epc = 0;
			env_opt_add(ep);
		}
		env_opt_end(1);
		break;

	case TELQUAL_IS:
	case TELQUAL_INFO:
		/* Ignore for now.  We shouldn't get it anyway. */
		break;

	default:
		break;
	}
}

#define OPT_REPLY_SIZE  256

static void
env_opt_start(void)
{
	unsigned char *p;

	if (opt_reply) {
		p = (unsigned char *)realloc(opt_reply, OPT_REPLY_SIZE);
		if (p == NULL)
			free(opt_reply);
	} else
		p = (unsigned char *)malloc(OPT_REPLY_SIZE);
	opt_reply = p;
	if (opt_reply == NULL) {
/*@*/           printf("env_opt_start: malloc()/realloc() failed!!!\n");
		opt_reply = opt_replyp = opt_replyend = NULL;
		return;
	}
	opt_replyp = opt_reply;
	opt_replyend = opt_reply + OPT_REPLY_SIZE;
	*opt_replyp++ = IAC;
	*opt_replyp++ = SB;
	*opt_replyp++ = telopt_environ;
	*opt_replyp++ = TELQUAL_IS;
}

static void
env_opt_start_info(void)
{
	env_opt_start();
	if (opt_replyp)
	    opt_replyp[-1] = TELQUAL_INFO;
}

static void
env_opt_add(unsigned char *ep)
{
	unsigned char *vp, c;

	if (opt_reply == NULL)          /*XXX*/
		return;                 /*XXX*/

	if (ep == NULL || *ep == '\0') {
		/* Send user defined variables first. */
		env_default(1, 0);
		while ((ep = env_default(0, 0)))
			env_opt_add(ep);

		/* Now add the list of well know variables.  */
		env_default(1, 1);
		while ((ep = env_default(0, 1)))
			env_opt_add(ep);
		return;
	}
	vp = env_getvalue(ep);
	if (opt_replyp + (vp ? strlen((char *)vp) : 0) +
				strlen((char *)ep) + 6 > opt_replyend)
	{
		int len;
		unsigned char *p;
		opt_replyend += OPT_REPLY_SIZE;
		len = opt_replyend - opt_reply;
		p = (unsigned char *)realloc(opt_reply, len);
		if (p == NULL)
			free(opt_reply);
		opt_reply = p;
		if (opt_reply == NULL) {
/*@*/                   printf("env_opt_add: realloc() failed!!!\n");
			opt_reply = opt_replyp = opt_replyend = NULL;
			return;
		}
		opt_replyp = opt_reply + len - (opt_replyend - opt_replyp);
		opt_replyend = opt_reply + len;
	}
	if (opt_welldefined((char *)ep))
		*opt_replyp++ = NEW_ENV_VAR;
	else
		*opt_replyp++ = ENV_USERVAR;
	for (;;) {
		while ((c = *ep++)) {
			switch(c&0xff) {
			case IAC:
				*opt_replyp++ = IAC;
				break;
			case NEW_ENV_VAR:
			case NEW_ENV_VALUE:
			case ENV_ESC:
			case ENV_USERVAR:
				*opt_replyp++ = ENV_ESC;
				break;
			}
			*opt_replyp++ = c;
		}
		if ((ep = vp)) {
			*opt_replyp++ = NEW_ENV_VALUE;
			vp = NULL;
		} else
			break;
	}
}

static int
opt_welldefined(char *ep)
{
	if ((strcmp(ep, "USER") == 0) ||
	    (strcmp(ep, "DISPLAY") == 0) ||
	    (strcmp(ep, "PRINTER") == 0) ||
	    (strcmp(ep, "SYSTEMTYPE") == 0) ||
	    (strcmp(ep, "JOB") == 0) ||
	    (strcmp(ep, "ACCT") == 0))
		return(1);
	return(0);
}

static void
env_opt_end(int emptyok)
{
	int len;

	len = opt_replyp - opt_reply + 2;
	if (emptyok || len > 6) {
		*opt_replyp++ = IAC;
		*opt_replyp++ = SE;
		if (NETROOM() > len) {
			ring_supply_data(&netoring, opt_reply, len);
			printsub('>', &opt_reply[2], len - 2);
		}
/*@*/           else printf("Not enough room in the buffer");
	}
	if (opt_reply) {
		free(opt_reply);
		opt_reply = opt_replyp = opt_replyend = NULL;
	}
}



static int
telrcv(void)
{
    int c;
    int scc;
    unsigned char *sbp = 0;
    int count;
    int returnValue = 0;

    scc = 0;
    count = 0;
    while (TTYROOM() > 2) {
	if (scc == 0) {
	    if (count) {
		ring_consumed(&netiring, count);
		returnValue = 1;
		count = 0;
	    }
	    sbp = netiring.consume;
	    scc = ring_full_consecutive(&netiring);
	    if (scc == 0) {
		/* No more data coming in */
		break;
	    }
	}

	c = *sbp++ & 0xff, scc--; count++;

	switch (telrcv_state) {

	case TS_CR:
	    telrcv_state = TS_DATA;
	    if (c == '\0') {
		break;  /* Ignore \0 after CR */
	    }
	    else if ((c == '\n') && my_want_state_is_dont(TELOPT_ECHO) && !crmod) {
		TTYADD(c);
		break;
	    }
	    /* Else, fall through */

	case TS_DATA:
	    if (c == IAC) {
		telrcv_state = TS_IAC;
		break;
	    }
		    /*
		     * The 'crmod' hack (see following) is needed
		     * since we can't set CRMOD on output only.
		     * Machines like MULTICS like to send \r without
		     * \n; since we must turn off CRMOD to get proper
		     * input, the mapping is done here (sigh).
		     */
	    if ((c == '\r') && my_want_state_is_dont(TELOPT_BINARY)) {
		if (scc > 0) {
		    c = *sbp&0xff;
		    if (c == 0) {
			sbp++, scc--; count++;
			/* a "true" CR */
			TTYADD('\r');
		    } else if (my_want_state_is_dont(TELOPT_ECHO) &&
					(c == '\n')) {
			sbp++, scc--; count++;
			TTYADD('\n');
		    } else {
			TTYADD('\r');
			if (crmod) {
				TTYADD('\n');
			}
		    }
		} else {
		    telrcv_state = TS_CR;
		    TTYADD('\r');
		    if (crmod) {
			    TTYADD('\n');
		    }
		}
	    } else {
		TTYADD(c);
	    }
	    continue;

	case TS_IAC:
process_iac:
	    switch (c) {

	    case WILL:
		telrcv_state = TS_WILL;
		continue;

	    case WONT:
		telrcv_state = TS_WONT;
		continue;

	    case DO:
		telrcv_state = TS_DO;
		continue;

	    case DONT:
		telrcv_state = TS_DONT;
		continue;

	    case DM:
		    /*
		     * We may have missed an urgent notification,
		     * so make sure we flush whatever is in the
		     * buffer currently.
		     */
		printoption("RCVD", IAC, DM);
		SYNCHing = 1;
		(void) ttyflush(1);
		SYNCHing = stilloob();
		settimer(gotDM);
		break;

	    case SB:
		SB_CLEAR();
		telrcv_state = TS_SB;
		continue;

	    case IAC:
		TTYADD(IAC);
		break;

	    case NOP:
	    case GA:
	    default:
		printoption("RCVD", IAC, c);
		break;
	    }
	    telrcv_state = TS_DATA;
	    continue;

	case TS_WILL:
	    printoption("RCVD", WILL, c);
	    willoption(c);
	    telrcv_state = TS_DATA;
	    continue;

	case TS_WONT:
	    printoption("RCVD", WONT, c);
	    wontoption(c);
	    telrcv_state = TS_DATA;
	    continue;

	case TS_DO:
	    printoption("RCVD", DO, c);
	    dooption(c);
	    if (c == TELOPT_NAWS) {
		sendnaws();
	    } else if (c == TELOPT_LFLOW) {
		localflow = 1;
		setcommandmode();
		setconnmode(0);
	    }
	    telrcv_state = TS_DATA;
	    continue;

	case TS_DONT:
	    printoption("RCVD", DONT, c);
	    dontoption(c);
	    flushline = 1;
	    setconnmode(0);     /* set new tty mode (maybe) */
	    telrcv_state = TS_DATA;
	    continue;

	case TS_SB:
	    if (c == IAC) {
		telrcv_state = TS_SE;
	    } else {
		SB_ACCUM(c);
	    }
	    continue;

	case TS_SE:
	    if (c != SE) {
		if (c != IAC) {
		    /*
		     * This is an error.  We only expect to get
		     * "IAC IAC" or "IAC SE".  Several things may
		     * have happend.  An IAC was not doubled, the
		     * IAC SE was left off, or another option got
		     * inserted into the suboption are all possibilities.
		     * If we assume that the IAC was not doubled,
		     * and really the IAC SE was left off, we could
		     * get into an infinate loop here.  So, instead,
		     * we terminate the suboption, and process the
		     * partial suboption if we can.
		     */
		    SB_ACCUM(IAC);
		    SB_ACCUM(c);
		    subpointer -= 2;
		    SB_TERM();

		    printoption("In SUBOPTION processing, RCVD", IAC, c);
		    suboption();        /* handle sub-option */
		    telrcv_state = TS_IAC;
		    goto process_iac;
		}
		SB_ACCUM(c);
		telrcv_state = TS_SB;
	    } else {
		SB_ACCUM(IAC);
		SB_ACCUM(SE);
		subpointer -= 2;
		SB_TERM();
		suboption();    /* handle sub-option */
		telrcv_state = TS_DATA;
	    }
	}
    }
    if (count)
	ring_consumed(&netiring, count);
    return returnValue||count;
}

static int
rlogin_susp(void)
{
    if (local) {
	local = 0;
	bol = 1;
	command(0, "z\n", 2);
	return(1);
    }
    return(0);
}

static int
telsnd(void)
{
    int tcc;
    int count;
    int returnValue = 0;
    unsigned char *tbp = 0;

    tcc = 0;
    count = 0;
    while (NETROOM() > 2) {
	int sc;
	int c;

	if (tcc == 0) {
	    if (count) {
		ring_consumed(&ttyiring, count);
		returnValue = 1;
		count = 0;
	    }
	    tbp = ttyiring.consume;
	    tcc = ring_full_consecutive(&ttyiring);
	    if (tcc == 0) {
		break;
	    }
	}
	c = *tbp++ & 0xff, sc = strip(c), tcc--; count++;
	if (rlogin != _POSIX_VDISABLE) {
		if (bol) {
			bol = 0;
			if (sc == rlogin) {
				local = 1;
				continue;
			}
		} else if (local) {
			local = 0;
			if (sc == '.' || c == termEofChar) {
				bol = 1;
				command(0, "close\n", 6);
				continue;
			}
			if (sc == termSuspChar) {
				bol = 1;
				command(0, "z\n", 2);
				continue;
			}
			if (sc == escape) {
				command(0, (char *)tbp, tcc);
				bol = 1;
				count += tcc;
				tcc = 0;
				flushline = 1;
				break;
			}
			if (sc != rlogin) {
				++tcc;
				--tbp;
				--count;
				c = sc = rlogin;
			}
		}
		if ((sc == '\n') || (sc == '\r'))
			bol = 1;
	} else if (escape != _POSIX_VDISABLE && sc == escape) {
	    /*
	     * Double escape is a pass through of a single escape character.
	     */
	    if (tcc && strip(*tbp) == escape) {
		tbp++;
		tcc--;
		count++;
		bol = 0;
	    } else {
		command(0, (char *)tbp, tcc);
		bol = 1;
		count += tcc;
		tcc = 0;
		flushline = 1;
		break;
	    }
	} else
	    bol = 0;
	if (kludgelinemode && (globalmode&MODE_EDIT) && (sc == echoc)) {
	    if (tcc > 0 && strip(*tbp) == echoc) {
		tcc--; tbp++; count++;
	    } else {
		dontlecho = !dontlecho;
		settimer(echotoggle);
		setconnmode(0);
		flushline = 1;
		break;
	    }
	}
	if (sc != _POSIX_VDISABLE && MODE_LOCAL_CHARS(globalmode)) {
	    if (TerminalSpecialChars(sc) == 0) {
		bol = 1;
		break;
	    }
	}
	if (my_want_state_is_wont(TELOPT_BINARY)) {
	    switch (c) {
	    case '\n':
		    /*
		     * If we are in CRMOD mode (\r ==> \n)
		     * on our local machine, then probably
		     * a newline (unix) is CRLF (TELNET).
		     */
		if (MODE_LOCAL_CHARS(globalmode)) {
		    NETADD('\r');
		}
		NETADD('\n');
		bol = flushline = 1;
		break;
	    case '\r':
		if (!crlf) {
		    NET2ADD('\r', '\0');
		} else {
		    NET2ADD('\r', '\n');
		}
		bol = flushline = 1;
		break;
	    case IAC:
		NET2ADD(IAC, IAC);
		break;
	    default:
		NETADD(c);
		break;
	    }
	} else if (c == IAC) {
	    NET2ADD(IAC, IAC);
	} else {
	    NETADD(c);
	}
    }
    if (count)
	ring_consumed(&ttyiring, count);
    return returnValue||count;          /* Non-zero if we did anything */
}

/*
 * Scheduler()
 *
 * Try to do something.
 *
 * If we do something useful, return 1; else return 0.
 *
 */


/* should we block in the select ? */
static int
Scheduler(int block)
{
    int returnValue;
    int netin, netout, netex, ttyin, ttyout;

    /* Decide which rings should be processed */

    netout = ring_full_count(&netoring) &&
	    (flushline ||
		(my_want_state_is_wont(TELOPT_LINEMODE)
			&& (!kludgelinemode || my_want_state_is_do(TELOPT_SGA))
		) ||
			my_want_state_is_will(TELOPT_BINARY));
    ttyout = ring_full_count(&ttyoring);

    ttyin = ring_empty_count(&ttyiring) && (clienteof == 0);

    netin = !ISend && ring_empty_count(&netiring);

    netex = !SYNCHing;

    /* Call to system code to process rings */

    returnValue = process_rings(netin, netout, netex, ttyin, ttyout, !block);

    /* Now, look at the input rings, looking for work to do. */

    if (ring_full_count(&ttyiring)) {
	    returnValue |= telsnd();
    }

    if (ring_full_count(&netiring)) {
	returnValue |= telrcv();
    }
    return returnValue;
}

/*
 * Select from tty and network...
 */
static void
telnet(void)
{
    sys_telnet_init();

    if (telnetport) {
	send_do(TELOPT_SGA, 1);
	send_will(TELOPT_TTYPE, 1);
	send_will(TELOPT_NAWS, 1);
	send_will(TELOPT_TSPEED, 1);
	send_will(TELOPT_LFLOW, 1);
	send_will(TELOPT_LINEMODE, 1);
	send_will(TELOPT_NEW_ENVIRON, 1);
	send_do(TELOPT_STATUS, 1);
	if (env_getvalue((unsigned char *)"DISPLAY"))
	    send_will(TELOPT_XDISPLOC, 1);
	if (binary)
	    tel_enter_binary(binary);
    }

    for (;;) {
	int schedValue;

	while ((schedValue = Scheduler(0)) != 0) {
	    if (schedValue == -1) {
		setcommandmode();
		return;
	    }
	}

	if (Scheduler(1) == -1) {
	    setcommandmode();
	    return;
	}
    }
}

#if     0       /* XXX - this not being in is a bug */
/*
 * nextitem()
 *
 *      Return the address of the next "item" in the TELNET data
 * stream.  This will be the address of the next character if
 * the current address is a user data character, or it will
 * be the address of the character following the TELNET command
 * if the current address is a TELNET IAC ("I Am a Command")
 * character.
 */

static char *
nextitem(char *current)
{
    if ((*current&0xff) != IAC) {
	return current+1;
    }
    switch (*(current+1)&0xff) {
    case DO:
    case DONT:
    case WILL:
    case WONT:
	return current+3;
    case SB:            /* loop forever looking for the SE */
	{
	    char *look = current+2;

	    for (;;) {
		if ((*look++&0xff) == IAC) {
		    if ((*look++&0xff) == SE) {
			return look;
		    }
		}
	    }
	}
    default:
	return current+2;
    }
}
#endif  /* 0 */

/*
 * netclear()
 *
 *      We are about to do a TELNET SYNCH operation.  Clear
 * the path to the network.
 *
 *      Things are a bit tricky since we may have sent the first
 * byte or so of a previous TELNET command into the network.
 * So, we have to scan the network buffer from the beginning
 * until we are up to where we want to be.
 *
 *      A side effect of what we do, just to keep things
 * simple, is to clear the urgent data pointer.  The principal
 * caller should be setting the urgent data pointer AFTER calling
 * us in any case.
 */

static void
netclear(void)
{
#if     0       /* XXX */
    char *thisitem, *next;
    char *good;
#define wewant(p)       ((nfrontp > p) && ((*p&0xff) == IAC) && \
				((*(p+1)&0xff) != EC) && ((*(p+1)&0xff) != EL))

    thisitem = netobuf;

    while ((next = nextitem(thisitem)) <= netobuf.send) {
	thisitem = next;
    }

    /* Now, thisitem is first before/at boundary. */

    good = netobuf;     /* where the good bytes go */

    while (netoring.add > thisitem) {
	if (wewant(thisitem)) {
	    int length;

	    next = thisitem;
	    do {
		next = nextitem(next);
	    } while (wewant(next) && (nfrontp > next));
	    length = next-thisitem;
	    memmove(good, thisitem, length);
	    good += length;
	    thisitem = next;
	} else {
	    thisitem = nextitem(thisitem);
	}
    }

#endif  /* 0 */
}

/*
 * These routines add various telnet commands to the data stream.
 */

static void
doflush(void)
{
    NET2ADD(IAC, DO);
    NETADD(TELOPT_TM);
    flushline = 1;
    flushout = 1;
    (void) ttyflush(1);                 /* Flush/drop output */
    /* do printoption AFTER flush, otherwise the output gets tossed... */
    printoption("SENT", DO, TELOPT_TM);
}

static void
xmitAO(void)
{
    NET2ADD(IAC, AO);
    printoption("SENT", IAC, AO);
    if (autoflush) {
	doflush();
    }
}


static void
xmitEL(void)
{
    NET2ADD(IAC, EL);
    printoption("SENT", IAC, EL);
}

static void
xmitEC(void)
{
    NET2ADD(IAC, EC);
    printoption("SENT", IAC, EC);
}


static int
dosynch(const char ATTRIBUTE_UNUSED *u)
{
    netclear();                 /* clear the path to the network */
    NETADD(IAC);
    setneturg();
    NETADD(DM);
    printoption("SENT", IAC, DM);
    return 1;
}

static int
get_status(const char ATTRIBUTE_UNUSED *u)
{
    unsigned char tmp[16];
    unsigned char *cp;

    if (my_want_state_is_dont(TELOPT_STATUS)) {
	printf("Remote side does not support STATUS option\n");
	return 0;
    }
    cp = tmp;

    *cp++ = IAC;
    *cp++ = SB;
    *cp++ = TELOPT_STATUS;
    *cp++ = TELQUAL_SEND;
    *cp++ = IAC;
    *cp++ = SE;
    if (NETROOM() >= (int)((cp - tmp))) {
	ring_supply_data(&netoring, tmp, cp-tmp);
	printsub('>', tmp+2, cp - tmp - 2);
    }
    ++want_status_response;
    return 1;
}

static void
intp(void)
{
    NET2ADD(IAC, IP);
    printoption("SENT", IAC, IP);
    flushline = 1;
    if (autoflush)
	doflush();
    if (autosynch)
	dosynch(NULL);
}

static void
sendbrk(void)
{
    NET2ADD(IAC, BREAK);
    printoption("SENT", IAC, BREAK);
    flushline = 1;
    if (autoflush)
	doflush();
    if (autosynch)
	dosynch(NULL);
}

static void
sendabort(void)
{
    NET2ADD(IAC, ABORT);
    printoption("SENT", IAC, ABORT);
    flushline = 1;
    if (autoflush)
	doflush();
    if (autosynch)
	dosynch(NULL);
}

static void
sendsusp(void)
{
    NET2ADD(IAC, SUSP);
    printoption("SENT", IAC, SUSP);
    flushline = 1;
    if (autoflush)
	doflush();
    if (autosynch)
	dosynch(NULL);
}

static void
sendeof(void)
{
    NET2ADD(IAC, xEOF);
    printoption("SENT", IAC, xEOF);
}

#ifdef  SIGINFO
static void
sendayt(void)
{
    NET2ADD(IAC, AYT);
    printoption("SENT", IAC, AYT);
}
#endif

/*
 * Send a window size update to the remote system.
 */

static void
sendnaws(void)
{
    int rows, cols;
    unsigned char tmp[16];
    unsigned char *cp;

    if (my_state_is_wont(TELOPT_NAWS))
	return;

#define PUTSHORT(cp, x) { if ((*cp++ = ((x)>>8)&0xff) == IAC) *cp++ = IAC; \
			    if ((*cp++ = ((x))&0xff) == IAC) *cp++ = IAC; }

    if (TerminalWindowSize(&rows, &cols) == 0) {        /* Failed */
	return;
    }

    cp = tmp;

    *cp++ = IAC;
    *cp++ = SB;
    *cp++ = TELOPT_NAWS;
    PUTSHORT(cp, cols);
    PUTSHORT(cp, rows);
    *cp++ = IAC;
    *cp++ = SE;
    if (NETROOM() >= (int)(cp - tmp)) {
	ring_supply_data(&netoring, tmp, cp-tmp);
	printsub('>', tmp+2, cp - tmp - 2);
    }
}

static void
tel_enter_binary(int rw)
{
    if (rw&1)
	send_do(TELOPT_BINARY, 1);
    if (rw&2)
	send_will(TELOPT_BINARY, 1);
}

static void
tel_leave_binary(int rw)
{
    if (rw&1)
	send_dont(TELOPT_BINARY, 1);
    if (rw&2)
	send_wont(TELOPT_BINARY, 1);
}

/*
 * initialize the terminal data structures.
 */

static void
init_terminal(void)
{
    if (ring_init(&ttyoring, ttyobuf, sizeof ttyobuf) != 1) {
	exit(1);
    }
    if (ring_init(&ttyiring, ttyibuf, sizeof ttyibuf) != 1) {
	exit(1);
    }
    autoflush = TerminalAutoFlush();
}


/*
 *              Send as much data as possible to the terminal.
 *
 *              Return value:
 *                      -1: No useful work done, data waiting to go out.
 *                       0: No data was waiting, so nothing was done.
 *                       1: All waiting data was written out.
 *                       n: All data - n was written out.
 */


static int
ttyflush(int drop)
{
    int n, n0, n1;

    n0 = ring_full_count(&ttyoring);
    if ((n1 = n = ring_full_consecutive(&ttyoring)) > 0) {
	if (drop) {
	    TerminalFlushOutput();
	    /* we leave 'n' alone! */
	} else {
	    n = TerminalWrite((char *)ttyoring.consume, n);
	}
    }
    if (n > 0) {
	if (termdata && n) {
	    Dump('>', ttyoring.consume, n);
	}
	/*
	 * If we wrote everything, and the full count is
	 * larger than what we wrote, then write the
	 * rest of the buffer.
	 */
	if (n1 == n && n0 > n) {
		n1 = n0 - n;
		if (!drop)
			n1 = TerminalWrite((char *)ttyoring.bottom, n1);
		if (n1 > 0)
			n += n1;
	}
	ring_consumed(&ttyoring, n);
    }
    if (n < 0)
	return -1;
    if (n == n0) {
	if (n0)
	    return -1;
	return 0;
    }
    return n0 - n + 1;
}


/*
 * These routines decides on what the mode should be (based on the values
 * of various global variables).
 */


static int
getconnmode(void)
{
    int mode = 0;

    if (my_want_state_is_dont(TELOPT_ECHO))
	mode |= MODE_ECHO;

    if (localflow)
	mode |= MODE_FLOW;

    if ((eight & 1) || my_want_state_is_will(TELOPT_BINARY))
	mode |= MODE_INBIN;

    if (eight & 2)
	mode |= MODE_OUT8;
    if (his_want_state_is_will(TELOPT_BINARY))
	mode |= MODE_OUTBIN;

    if (kludgelinemode) {
	if (my_want_state_is_dont(TELOPT_SGA)) {
	    mode |= (MODE_TRAPSIG|MODE_EDIT);
	    if (dontlecho && (clocks.echotoggle > clocks.modenegotiated)) {
		mode &= ~MODE_ECHO;
	    }
	}
	return(mode);
    }
    if (my_want_state_is_will(TELOPT_LINEMODE))
	mode |= linemode;
    return(mode);
}

static void
setconnmode(int force)
{
    int newmode;

    newmode = getconnmode()|(force?MODE_FORCE:0);

    TerminalNewMode(newmode);

}


static void
setcommandmode(void)
{
    TerminalNewMode(-1);
}


# ifdef SUPPORT_DEBUGING
/*
 * The following are routines used to print out debugging information.
 */

static void
SetNetTrace(const char *file)
{
    if (NetTrace && NetTrace != stdout)
	fclose(NetTrace);
    if (file  && (strcmp(file, "-") != 0)) {
	NetTrace = fopen(file, "w");
	if (NetTrace) {
	    strncpy((char *)NetTraceFile, file, sizeof(NetTraceFile));
	    return;
	}
	bb_perror_msg("cannot open %s", file);
    }
    NetTrace = stdout;
    strcpy((char *)NetTraceFile, "(standard output)");
}

static void
Dump(char direction, unsigned char *buffer, int length)
{
#   define BYTES_PER_LINE       32
#   define min(x,y)     ((x<y)? x:y)
    unsigned char *pThis;
    int offset;

    offset = 0;

    while (length) {
	/* print one line */
	fprintf(NetTrace, "%c 0x%x\t", direction, offset);
	pThis = buffer;
	if (prettydump) {
	    buffer = buffer + min(length, BYTES_PER_LINE/2);
	    while (pThis < buffer) {
		fprintf(NetTrace, "%c%.2x",
		    (((*pThis)&0xff) == 0xff) ? '*' : ' ',
		    (*pThis)&0xff);
		pThis++;
	    }
	    length -= BYTES_PER_LINE/2;
	    offset += BYTES_PER_LINE/2;
	} else {
	    buffer = buffer + min(length, BYTES_PER_LINE);
	    while (pThis < buffer) {
		fprintf(NetTrace, "%.2x", (*pThis)&0xff);
		pThis++;
	    }
	    length -= BYTES_PER_LINE;
	    offset += BYTES_PER_LINE;
	}
	if (NetTrace == stdout) {
	    fprintf(NetTrace, "\r\n");
	} else {
	    fprintf(NetTrace, "\n");
	}
	if (length < 0) {
	    fflush(NetTrace);
	    return;
	}
	/* find next unique line */
    }
    fflush(NetTrace);
}


static void
printoption(char *direction, int cmd, int option)
{
	if (!showoptions)
		return;
	if (cmd == IAC) {
		if (TELCMD_OK(option))
		    fprintf(NetTrace, "%s IAC %s", direction, TELCMD(option));
		else
		    fprintf(NetTrace, "%s IAC %d", direction, option);
	} else {
		char *fmt;
		fmt = (cmd == WILL) ? "WILL" : (cmd == WONT) ? "WONT" :
			(cmd == DO) ? "DO" : (cmd == DONT) ? "DONT" : 0;
		if (fmt) {
		    fprintf(NetTrace, "%s %s ", direction, fmt);
		    if (TELOPT_OK(option))
			fprintf(NetTrace, "%s", TELOPT(option));
		    else if (option == TELOPT_EXOPL)
			fprintf(NetTrace, "EXOPL");
		    else
			fprintf(NetTrace, "%d", option);
		} else
		    fprintf(NetTrace, "%s %d %d", direction, cmd, option);
	}
	if (NetTrace == stdout) {
	    fprintf(NetTrace, "\r\n");
	    fflush(NetTrace);
	} else {
	    fprintf(NetTrace, "\n");
	}
	return;
}

static void
optionstatus(void)
{
    int i;

    for (i = 0; i < 256; i++) {
	if (do_dont_resp[i]) {
	    if (TELOPT_OK(i))
		printf("resp DO_DONT %s: %d\n", TELOPT(i), do_dont_resp[i]);
	    else if (TELCMD_OK(i))
		printf("resp DO_DONT %s: %d\n", TELCMD(i), do_dont_resp[i]);
	    else
		printf("resp DO_DONT %d: %d\n", i,
				do_dont_resp[i]);
	    if (my_want_state_is_do(i)) {
		if (TELOPT_OK(i))
		    printf("want DO   %s\n", TELOPT(i));
		else if (TELCMD_OK(i))
		    printf("want DO   %s\n", TELCMD(i));
		else
		    printf("want DO   %d\n", i);
	    } else {
		if (TELOPT_OK(i))
		    printf("want DONT %s\n", TELOPT(i));
		else if (TELCMD_OK(i))
		    printf("want DONT %s\n", TELCMD(i));
		else
		    printf("want DONT %d\n", i);
	    }
	} else {
	    if (my_state_is_do(i)) {
		if (TELOPT_OK(i))
		    printf("     DO   %s\n", TELOPT(i));
		else if (TELCMD_OK(i))
		    printf("     DO   %s\n", TELCMD(i));
		else
		    printf("     DO   %d\n", i);
	    }
	}
	if (will_wont_resp[i]) {
	    if (TELOPT_OK(i))
		printf("resp WILL_WONT %s: %d\n", TELOPT(i), will_wont_resp[i]);
	    else if (TELCMD_OK(i))
		printf("resp WILL_WONT %s: %d\n", TELCMD(i), will_wont_resp[i]);
	    else
		printf("resp WILL_WONT %d: %d\n",
				i, will_wont_resp[i]);
	    if (my_want_state_is_will(i)) {
		if (TELOPT_OK(i))
		    printf("want WILL %s\n", TELOPT(i));
		else if (TELCMD_OK(i))
		    printf("want WILL %s\n", TELCMD(i));
		else
		    printf("want WILL %d\n", i);
	    } else {
		if (TELOPT_OK(i))
		    printf("want WONT %s\n", TELOPT(i));
		else if (TELCMD_OK(i))
		    printf("want WONT %s\n", TELCMD(i));
		else
		    printf("want WONT %d\n", i);
	    }
	} else {
	    if (my_state_is_will(i)) {
		if (TELOPT_OK(i))
		    printf("     WILL %s\n", TELOPT(i));
		else if (TELCMD_OK(i))
		    printf("     WILL %s\n", TELCMD(i));
		else
		    printf("     WILL %d\n", i);
	    }
	}
    }

}

/* direction '<' or '>' */
/* pointer where suboption data sits */
/* length of suboption data */
static void
printsub(int direction, unsigned char *pointer, int length)
{
    int i;

    if (showoptions || direction == 0 ||
	(want_status_response && (pointer[0] == TELOPT_STATUS))) {
	if (direction) {
	    fprintf(NetTrace, "%s IAC SB ",
				(direction == '<')? "RCVD":"SENT");
	    if (length >= 3) {
		int j;

		i = pointer[length-2];
		j = pointer[length-1];

		if (i != IAC || j != SE) {
		    fprintf(NetTrace, "(terminated by ");
		    if (TELOPT_OK(i))
			fprintf(NetTrace, "%s ", TELOPT(i));
		    else if (TELCMD_OK(i))
			fprintf(NetTrace, "%s ", TELCMD(i));
		    else
			fprintf(NetTrace, "%d ", i);
		    if (TELOPT_OK(j))
			fprintf(NetTrace, "%s", TELOPT(j));
		    else if (TELCMD_OK(j))
			fprintf(NetTrace, "%s", TELCMD(j));
		    else
			fprintf(NetTrace, "%d", j);
		    fprintf(NetTrace, ", not IAC SE!) ");
		}
	    }
	    length -= 2;
	}
	if (length < 1) {
	    fprintf(NetTrace, "(Empty suboption??\?)");
	    if (NetTrace == stdout)
		fflush(NetTrace);
	    return;
	}
	switch (pointer[0]) {
	case TELOPT_TTYPE:
	    fprintf(NetTrace, "TERMINAL-TYPE ");
	    switch (pointer[1]) {
	    case TELQUAL_IS:
		fprintf(NetTrace, "IS \"%.*s\"", length-2, (char *)pointer+2);
		break;
	    case TELQUAL_SEND:
		fprintf(NetTrace, "SEND");
		break;
	    default:
		fprintf(NetTrace,
				"- unknown qualifier %d (0x%x).",
				pointer[1], pointer[1]);
	    }
	    break;
	case TELOPT_TSPEED:
	    fprintf(NetTrace, "TERMINAL-SPEED");
	    if (length < 2) {
		fprintf(NetTrace, " (empty suboption??\?)");
		break;
	    }
	    switch (pointer[1]) {
	    case TELQUAL_IS:
		fprintf(NetTrace, " IS ");
		fprintf(NetTrace, "%.*s", length-2, (char *)pointer+2);
		break;
	    default:
		if (pointer[1] == 1)
		    fprintf(NetTrace, " SEND");
		else
		    fprintf(NetTrace, " %d (unknown)", pointer[1]);
		for (i = 2; i < length; i++)
		    fprintf(NetTrace, " ?%d?", pointer[i]);
		break;
	    }
	    break;

	case TELOPT_LFLOW:
	    fprintf(NetTrace, "TOGGLE-FLOW-CONTROL");
	    if (length < 2) {
		fprintf(NetTrace, " (empty suboption??\?)");
		break;
	    }
	    switch (pointer[1]) {
	    case LFLOW_OFF:
		fprintf(NetTrace, " OFF"); break;
	    case LFLOW_ON:
		fprintf(NetTrace, " ON"); break;
	    case LFLOW_RESTART_ANY:
		fprintf(NetTrace, " RESTART-ANY"); break;
	    case LFLOW_RESTART_XON:
		fprintf(NetTrace, " RESTART-XON"); break;
	    default:
		fprintf(NetTrace, " %d (unknown)", pointer[1]);
	    }
	    for (i = 2; i < length; i++)
		fprintf(NetTrace, " ?%d?", pointer[i]);
	    break;

	case TELOPT_NAWS:
	    fprintf(NetTrace, "NAWS");
	    if (length < 2) {
		fprintf(NetTrace, " (empty suboption??\?)");
		break;
	    }
	    if (length == 2) {
		fprintf(NetTrace, " ?%d?", pointer[1]);
		break;
	    }
	    fprintf(NetTrace, " %d %d (%d)",
		pointer[1], pointer[2],
		(int)((((unsigned int)pointer[1])<<8)|((unsigned int)pointer[2])));
	    if (length == 4) {
		fprintf(NetTrace, " ?%d?", pointer[3]);
		break;
	    }
	    fprintf(NetTrace, " %d %d (%d)",
		pointer[3], pointer[4],
		(int)((((unsigned int)pointer[3])<<8)|((unsigned int)pointer[4])));
	    for (i = 5; i < length; i++)
		fprintf(NetTrace, " ?%d?", pointer[i]);
	    break;

	case TELOPT_LINEMODE:
	    fprintf(NetTrace, "LINEMODE ");
	    if (length < 2) {
		fprintf(NetTrace, " (empty suboption??\?)");
		break;
	    }
	    switch (pointer[1]) {
	    case WILL:
		fprintf(NetTrace, "WILL ");
		goto common;
	    case WONT:
		fprintf(NetTrace, "WONT ");
		goto common;
	    case DO:
		fprintf(NetTrace, "DO ");
		goto common;
	    case DONT:
		fprintf(NetTrace, "DONT ");
	    common:
		if (length < 3) {
		    fprintf(NetTrace, "(no option??\?)");
		    break;
		}
		switch (pointer[2]) {
		case LM_FORWARDMASK:
		    fprintf(NetTrace, "Forward Mask");
		    for (i = 3; i < length; i++)
			fprintf(NetTrace, " %x", pointer[i]);
		    break;
		default:
		    fprintf(NetTrace, "%d (unknown)", pointer[2]);
		    for (i = 3; i < length; i++)
			fprintf(NetTrace, " %d", pointer[i]);
		    break;
		}
		break;

	    case LM_SLC:
		fprintf(NetTrace, "SLC");
		for (i = 2; i < length - 2; i += 3) {
		    if (SLC_NAME_OK(pointer[i+SLC_FUNC]))
			fprintf(NetTrace, " %s", SLC_NAME(pointer[i+SLC_FUNC]));
		    else
			fprintf(NetTrace, " %d", pointer[i+SLC_FUNC]);
		    switch (pointer[i+SLC_FLAGS]&SLC_LEVELBITS) {
		    case SLC_NOSUPPORT:
			fprintf(NetTrace, " NOSUPPORT"); break;
		    case SLC_CANTCHANGE:
			fprintf(NetTrace, " CANTCHANGE"); break;
		    case SLC_VARIABLE:
			fprintf(NetTrace, " VARIABLE"); break;
		    case SLC_DEFAULT:
			fprintf(NetTrace, " DEFAULT"); break;
		    }
		    fprintf(NetTrace, "%s%s%s",
			pointer[i+SLC_FLAGS]&SLC_ACK ? "|ACK" : "",
			pointer[i+SLC_FLAGS]&SLC_FLUSHIN ? "|FLUSHIN" : "",
			pointer[i+SLC_FLAGS]&SLC_FLUSHOUT ? "|FLUSHOUT" : "");
		    if (pointer[i+SLC_FLAGS]& ~(SLC_ACK|SLC_FLUSHIN|
						SLC_FLUSHOUT| SLC_LEVELBITS))
			fprintf(NetTrace, "(0x%x)", pointer[i+SLC_FLAGS]);
		    fprintf(NetTrace, " %d;", pointer[i+SLC_VALUE]);
		    if ((pointer[i+SLC_VALUE] == IAC) &&
			(pointer[i+SLC_VALUE+1] == IAC))
				i++;
		}
		for (; i < length; i++)
		    fprintf(NetTrace, " ?%d?", pointer[i]);
		break;

	    case LM_MODE:
		fprintf(NetTrace, "MODE ");
		if (length < 3) {
		    fprintf(NetTrace, "(no mode??\?)");
		    break;
		}
		{
		    char tbuf[64];
		    snprintf(tbuf, sizeof(tbuf),
			     "%s%s%s%s%s",
			     pointer[2]&MODE_EDIT ? "|EDIT" : "",
			     pointer[2]&MODE_TRAPSIG ? "|TRAPSIG" : "",
			     pointer[2]&MODE_SOFT_TAB ? "|SOFT_TAB" : "",
			     pointer[2]&MODE_LIT_ECHO ? "|LIT_ECHO" : "",
			     pointer[2]&MODE_ACK ? "|ACK" : "");
		    fprintf(NetTrace, "%s", tbuf[1] ? &tbuf[1] : "0");
		}
		if (pointer[2]&~(MODE_MASK))
		    fprintf(NetTrace, " (0x%x)", pointer[2]);
		for (i = 3; i < length; i++)
		    fprintf(NetTrace, " ?0x%x?", pointer[i]);
		break;
	    default:
		fprintf(NetTrace, "%d (unknown)", pointer[1]);
		for (i = 2; i < length; i++)
		    fprintf(NetTrace, " %d", pointer[i]);
	    }
	    break;

	case TELOPT_STATUS: {
	    char *cp;
	    int j, k;

	    fprintf(NetTrace, "STATUS");

	    switch (pointer[1]) {
	    default:
		if (pointer[1] == TELQUAL_SEND)
		    fprintf(NetTrace, " SEND");
		else
		    fprintf(NetTrace, " %d (unknown)", pointer[1]);
		for (i = 2; i < length; i++)
		    fprintf(NetTrace, " ?%d?", pointer[i]);
		break;
	    case TELQUAL_IS:
		if (--want_status_response < 0)
		    want_status_response = 0;
		if (NetTrace == stdout)
		    fprintf(NetTrace, " IS\r\n");
		else
		    fprintf(NetTrace, " IS\n");

		for (i = 2; i < length; i++) {
		    switch(pointer[i]) {
		    case DO:    cp = "DO"; goto common2;
		    case DONT:  cp = "DONT"; goto common2;
		    case WILL:  cp = "WILL"; goto common2;
		    case WONT:  cp = "WONT"; goto common2;
		    common2:
			i++;
			if (TELOPT_OK((int)pointer[i]))
			    fprintf(NetTrace, " %s %s", cp, TELOPT(pointer[i]));
			else
			    fprintf(NetTrace, " %s %d", cp, pointer[i]);

			if (NetTrace == stdout)
			    fprintf(NetTrace, "\r\n");
			else
			    fprintf(NetTrace, "\n");
			break;

		    case SB:
			fprintf(NetTrace, " SB ");
			i++;
			j = k = i;
			while (j < length) {
			    if (pointer[j] == SE) {
				if (j+1 == length)
				    break;
				if (pointer[j+1] == SE)
				    j++;
				else
				    break;
			    }
			    pointer[k++] = pointer[j++];
			}
			printsub(0, &pointer[i], k - i);
			if (i < length) {
			    fprintf(NetTrace, " SE");
			    i = j;
			} else
			    i = j - 1;

			if (NetTrace == stdout)
			    fprintf(NetTrace, "\r\n");
			else
			    fprintf(NetTrace, "\n");

			break;

		    default:
			fprintf(NetTrace, " %d", pointer[i]);
			break;
		    }
		}
		break;
	    }
	    break;
	  }

	case TELOPT_XDISPLOC:
	    fprintf(NetTrace, "X-DISPLAY-LOCATION ");
	    switch (pointer[1]) {
	    case TELQUAL_IS:
		fprintf(NetTrace, "IS \"%.*s\"", length-2, (char *)pointer+2);
		break;
	    case TELQUAL_SEND:
		fprintf(NetTrace, "SEND");
		break;
	    default:
		fprintf(NetTrace, "- unknown qualifier %d (0x%x).",
				pointer[1], pointer[1]);
	    }
	    break;

	case TELOPT_NEW_ENVIRON:
	    fprintf(NetTrace, "NEW-ENVIRON ");
	    switch (pointer[1]) {
	    case TELQUAL_IS:
		fprintf(NetTrace, "IS ");
		goto env_common;
	    case TELQUAL_SEND:
		fprintf(NetTrace, "SEND ");
		goto env_common;
	    case TELQUAL_INFO:
		fprintf(NetTrace, "INFO ");
	    env_common:
		{
		    int noquote = 2;
		    for (i = 2; i < length; i++ ) {
			switch (pointer[i]) {
			case NEW_ENV_VALUE:
			    fprintf(NetTrace, "\" VALUE " + noquote);
			    noquote = 2;
			    break;

			case NEW_ENV_VAR:
			    fprintf(NetTrace, "\" VAR " + noquote);
			    noquote = 2;
			    break;

			case ENV_ESC:
			    fprintf(NetTrace, "\" ESC " + noquote);
			    noquote = 2;
			    break;

			case ENV_USERVAR:
			    fprintf(NetTrace, "\" USERVAR " + noquote);
			    noquote = 2;
			    break;

			default:
			    if (isprint(pointer[i]) && pointer[i] != '"') {
				if (noquote) {
				    putc('"', NetTrace);
				    noquote = 0;
				}
				putc(pointer[i], NetTrace);
			    } else {
				fprintf(NetTrace, "\" %03o " + noquote,
							pointer[i]);
				noquote = 2;
			    }
			    break;
			}
		    }
		    if (!noquote)
			putc('"', NetTrace);
		    break;
		}
	    }
	    break;

	default:
	    if (TELOPT_OK(pointer[0]))
		fprintf(NetTrace, "%s (unknown)", TELOPT(pointer[0]));
	    else
		fprintf(NetTrace, "%d (unknown)", pointer[0]);
	    for (i = 1; i < length; i++)
		fprintf(NetTrace, " %d", pointer[i]);
	    break;
	}
	if (direction) {
	    if (NetTrace == stdout)
		fprintf(NetTrace, "\r\n");
	    else
		fprintf(NetTrace, "\n");
	}
	if (NetTrace == stdout)
	    fflush(NetTrace);
    }
}
#endif /* SUPPORT_DEBUGING */

/* EmptyTerminal - called to make sure that the terminal buffer is empty.
 *                      Note that we consider the buffer to run all the
 *                      way to the kernel (thus the select).
 */

static void
EmptyTerminal(void)
{
#if     defined(unix)
    fd_set      outs;

    FD_ZERO(&outs);
#endif  /* defined(unix) */

    if (TTYBYTES() == 0) {
#if     defined(unix)
	FD_SET(tout, &outs);
	(void) select(tout+1, (fd_set *) 0, &outs, (fd_set *) 0,
			(struct timeval *) 0);  /* wait for TTLOWAT */
#endif  /* defined(unix) */
    } else {
	while (TTYBYTES()) {
	    (void) ttyflush(0);
#if     defined(unix)
	    ttyflush(0);
	    FD_SET(tout, &outs);
	    (void) select(tout+1, (fd_set *) 0, &outs, (fd_set *) 0,
				(struct timeval *) 0);  /* wait for TTLOWAT */
#endif  /* defined(unix) */
	}
    }
}

static void
SetForExit(void)
{
    setconnmode(0);
    do {
	(void)telrcv();                 /* Process any incoming data */
	EmptyTerminal();
    } while (ring_full_count(&netiring));       /* While there is any */
    setcommandmode();
    fflush(stdout);
    fflush(stderr);
    setconnmode(0);
    EmptyTerminal();                    /* Flush the path to the tty */
    setcommandmode();
}

static void
Exit(void)
{
    SetForExit();
    exit(0);
}

static void
ExitString(char *string)
{
    SetForExit();
    bb_error_msg_and_die("%s.\r", string);
}


/*
 * main.  Parse arguments, invoke the protocol or command parser.
 */
int
telnet_main(void)
{
	int argc;
	char **argv;
	unsigned long ch;
	char *user, *alias, *esc_opt;

#ifdef SUPPORT_DEBUGING
	char *netrace_f;
#endif

	bol = 1;
	eight = 3;
	kludgelinemode = 1;
	slc_mode = SLC_EXPORT;
	str_lm[0] = IAC;
	str_lm[1] = SB;
	str_lm[2] = TELOPT_LINEMODE;
	str_lm[3] = 0;
	str_lm[4] = 0;
	str_lm[5] = IAC;
	str_lm[6] = SE;
	str_lm_mode[0] = IAC;
	str_lm_mode[1] = SB;
	str_lm_mode[2] = TELOPT_LINEMODE;
	str_lm_mode[3] = LM_MODE;
	str_lm_mode[4] = 0;
	str_lm_mode[5] = IAC;
	str_lm_mode[6] = SE;
	tninit();               /* Clear out things */

	TerminalSaveState();

	prompt = bb_applet_name;

	user = alias = NULL;

	rlogin = (strncmp(prompt, "rlog", 4) == 0) ? '~' : _POSIX_VDISABLE;

	autologin = -1;

	ch = bb_xgetopt_ulflags("78DELab:ce:l:r"
#ifdef SUPPORT_DEBUGING
				"dn:"
#endif
					"\07-8:8-7:a-l:l-a"
				, &alias, &esc_opt, &user
#ifdef SUPPORT_DEBUGING
				, &netrace_f
#endif
				);
	if(ch & 1)
		eight = 3;      /* binary output and input */
	if(ch & 2)
		eight = 0;
	if(ch & 4) {
		/* sometimes we don't want a mangled display */
		char *p;

		if((p = getenv("DISPLAY")))
			env_define((unsigned char*)"DISPLAY", (unsigned char*)p);
	}
	if(ch & 8)
		rlogin = escape = _POSIX_VDISABLE;
	if(ch & 16)
		eight |= 2;     /* binary output only */
	if(ch & 32)
		autologin = 1;
	/* if(ch & 64) */
	if(ch & 128)
		skiprc = 1;
	if(ch & 256)
		set_escape_char(esc_opt);
	/* if(ch & 512)
		autologin = -1; */
	if(ch & 1024)
		rlogin = '~';

#ifdef SUPPORT_DEBUGING
	if(ch & 2048)
		debug = 1;
	if(ch & 4096)
		SetNetTrace(netrace_f);
#endif

	if (autologin == -1)
		autologin = (rlogin == _POSIX_VDISABLE) ? 0 : 1;

	argc = Gargc;
	argv = Gargv;

	if (argc) {
		char *args[7];
		char **argp = args;

		if (argc > 2)
			bb_show_usage();
		*argp++ = (char *)prompt;
		if (user) {
			*argp++ = "-l";
			*argp++ = user;
		}
		if (alias) {
			*argp++ = "-b";
			*argp++ = alias;
		}
		*argp++ = argv[0];              /* host */
		if (argc > 1)
			*argp++ = argv[1];      /* port */
		*argp = 0;

		if (tn(argp - args, args) == 1)
			return 0;
		else
			return 1;
	}
	for (;;) {
		command(1, 0, 0);
	}
	return 0;
}
#endif /* APPLET_USAGE */
