/* Telnet server */

/* char pkg[] = "Package: netkit-telnet-0.14"; */

/* Philip Copeland, David A. Holland, Al Howat
 *
 * netbug@ftp.uk.linux.org
 *
 * Original BSD license see bottom. */

/* Busybox style by Vladimir Oleynik (C) 2003, 2004 <dzo@simtreas.ru> */
#ifdef APPLET_USAGE
#define telnetd_trivial_usage \
	"[-hn] [-L loginprg]"
#define telnetd_full_usage \
	"Telnetd uses incoming TELNET connections via inetd.\n" \
	"Options:\n" \
	"\t-L LOGIN\texec LOGIN on connect (default /bin/login)\n" \
	"\t-h\tDisable print login banner\n" \
	"\t-n\tDisable TCP keep-alive"
#else


#define ISSUE_FILE "/etc/issue.net"

/*
const char telnetd_usage[] =
	" [-h]"
	" [-L login_program]"
	" [-n]";
*/

#include "busybox.h"
#include <sys/wait.h>
#include <fcntl.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <syslog.h>
#include <time.h>
#include <utmp.h>
#include <netdb.h>
#include <netinet/ip.h>
#include <assert.h>
#include <string.h>
#include <unistd.h>
#include <sys/utsname.h>
#include <ctype.h>
#include <termios.h>
#include "pty_.h"
#include <paths.h>

#include "speed_table.h"
#include "bb_sig.h"
#include "bb_inet.h"

#ifdef HAVE_ARPA_TELNET
#include <arpa/telnet.h>
#else
#include "telnet.h"
#endif


#ifndef _POSIX_VDISABLE
#ifdef VDISABLE
#define _POSIX_VDISABLE VDISABLE
#else
#define _POSIX_VDISABLE ((unsigned char)'\377')
#endif
#endif

#ifndef LOG_ODELAY
#define LOG_ODELAY      0
#endif


#define ENV_VAR         NEW_ENV_VAR
#define ENV_VALUE       NEW_ENV_VALUE
#define TELOPT_ENVIRON  TELOPT_NEW_ENVIRON

/*
 * I/O data buffers defines
 */
#define NETSLOP 64

#define NIACCUM(c)      {   *netip++ = c; \
			    ncc++; \
			}

/* clock manipulations */
#define settimer(x)     (telnetd_clocks.x = ++telnetd_clocks.system)
#define sequenceIs(x,y) (telnetd_clocks.x < telnetd_clocks.y)

/*
 * Structures of information for each special character function.
 */
typedef struct {
	unsigned char flag;     /* the flags for this function */
	cc_t val;               /* the value of the special character */
} slcent;

typedef struct {
	slcent          defset;         /* the default settings */
	slcent          current;        /* the current settings */
	cc_t            *sptr;          /* a pointer to the char in */
					/* system data structures */
} slcfun;

/*
 * We keep track of each side of the option negotiation.
 */

#define MY_STATE_WILL           0x01
#define MY_WANT_STATE_WILL      0x02
#define MY_STATE_DO             0x04
#define MY_WANT_STATE_DO        0x08

/*
 * Macros to check the current state of things
 */

#define my_state_is_do(opt)             (telnetd_options[opt]&MY_STATE_DO)
#define my_state_is_will(opt)           (telnetd_options[opt]&MY_STATE_WILL)
#define my_want_state_is_do(opt)        (telnetd_options[opt]&MY_WANT_STATE_DO)
#define my_want_state_is_will(opt)      (telnetd_options[opt]&MY_WANT_STATE_WILL)

#define my_state_is_dont(opt)           (!my_state_is_do(opt))
#define my_state_is_wont(opt)           (!my_state_is_will(opt))
#define my_want_state_is_dont(opt)      (!my_want_state_is_do(opt))
#define my_want_state_is_wont(opt)      (!my_want_state_is_will(opt))

#define set_my_state_do(opt)            (telnetd_options[opt] |= MY_STATE_DO)
#define set_my_state_will(opt)          (telnetd_options[opt] |= MY_STATE_WILL)
#define set_my_want_state_do(opt)       (telnetd_options[opt] |= MY_WANT_STATE_DO)
#define set_my_want_state_will(opt)     (telnetd_options[opt] |= MY_WANT_STATE_WILL)

#define set_my_state_dont(opt)          (telnetd_options[opt] &= ~MY_STATE_DO)
#define set_my_state_wont(opt)          (telnetd_options[opt] &= ~MY_STATE_WILL)
#define set_my_want_state_dont(opt)     (telnetd_options[opt] &= ~MY_WANT_STATE_DO)
#define set_my_want_state_wont(opt)     (telnetd_options[opt] &= ~MY_WANT_STATE_WILL)

/*
 * Tricky code here.  What we want to know is if the MY_STATE_WILL
 * and MY_WANT_STATE_WILL bits have the same value.  Since the two
 * bits are adjacent, a little arithmatic will show that by adding
 * in the lower bit, the upper bit will be set if the two bits were
 * different, and clear if they were the same.
 */
#define my_will_wont_is_changing(opt) \
			((telnetd_options[opt]+MY_STATE_WILL) & MY_WANT_STATE_WILL)

#define my_do_dont_is_changing(opt) \
			((telnetd_options[opt]+MY_STATE_DO) & MY_WANT_STATE_DO)

/*
 * Make everything symetrical
 */

#define HIS_STATE_WILL                  MY_STATE_DO
#define HIS_WANT_STATE_WILL             MY_WANT_STATE_DO
#define HIS_STATE_DO                    MY_STATE_WILL
#define HIS_WANT_STATE_DO               MY_WANT_STATE_WILL

#define his_state_is_do                 my_state_is_will
#define his_state_is_will               my_state_is_do
#define his_want_state_is_do            my_want_state_is_will
#define his_want_state_is_will          my_want_state_is_do

#define his_state_is_dont               my_state_is_wont
#define his_state_is_wont               my_state_is_dont
#define his_want_state_is_dont          my_want_state_is_wont
#define his_want_state_is_wont          my_want_state_is_dont

#define set_his_state_do                set_my_state_will
#define set_his_state_will              set_my_state_do
#define set_his_want_state_do           set_my_want_state_will
#define set_his_want_state_will         set_my_want_state_do

#define set_his_state_dont              set_my_state_wont
#define set_his_state_wont              set_my_state_dont
#define set_his_want_state_dont         set_my_want_state_wont
#define set_his_want_state_wont         set_my_want_state_dont

#define his_will_wont_is_changing       my_do_dont_is_changing
#define his_do_dont_is_changing         my_will_wont_is_changing

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 64
#endif  /* MAXHOSTNAMELEN */

/*
 * The following are some clocks used to decide how to interpret
 * the relationship between various variables.
 */

struct _clocks {
    int system;                 /* what the current time is */
    int echotoggle;             /* last time user entered echo character */
    int modenegotiated;         /* last time operating mode negotiated */
    int didnetreceive;          /* last time we read data from network */
    int ttypesubopt;            /* ttype subopt is received */
    int tspeedsubopt;           /* tspeed subopt is received */
    int environsubopt;          /* environ subopt is received */
    int xdisplocsubopt;         /* xdisploc subopt is received */
    int baseline;               /* time started to do timed action */
    int gotDM;                  /* when did we last see a data mark */
};

CALC_BSS struct BB_COMMON_BSS
{
	char telnetd_options[256];
	char do_dont_resp[256];
	char will_wont_resp[256];
	int flowmode;    /* current flow control state */
	slcfun slctab[NSLC + 1]; /* slc mapping table */
	char *terminaltype;
	struct _clocks telnetd_clocks;
/*
 * I/O data buffers, pointers, and counters.
 */
	char ptyobuf[BUFSIZ+NETSLOP], *pfrontp, *pbackp;
	char netibuf[BUFSIZ], *netip;
	char netobuf[2*BUFSIZ];
	char *nfrontp, *nbackp;
	char *neturg;            /* one past last byte of urgent data */
	int pcc, ncc;

	int pty;
	int SYNCHing;            /* we are in TELNET SYNCH mode */

	struct termios termbuf, termbuf2;        /* pty control structure */

	char host_name[MAXHOSTNAMELEN];
/*
 * I/O data buffers,
 * pointers, and counters.
 */
	char    ptyibuf[BUFSIZ];
	char    ptyibuf2[BUFSIZ];
	unsigned char subbuffer[512];
	int turn_on_sga;
	char linedata[PATH_MAX];
#ifdef  TIOCGWINSZ
	int def_row, def_col;
#endif
	char terminalname[41];
	struct timeval timeout;
	char *putlocation;
	char editedhost[32];
	struct utsname kerninfo;
	char *envinit[2];
	int dis_hostinfo;       /* do we print login banner? */
	int state /* = TS_DATA */;
	int def_tspeed, def_rspeed;
	const char *loginprg;
	char  *ptyip;
	char *Tline;
	int not42;
	int ptyslavefd;
	unsigned char *subpointer;
	const unsigned char *subend;
} bb_common_bss;
#define telnetd_options bb_common_bss.telnetd_options
#define do_dont_resp bb_common_bss.do_dont_resp
#define will_wont_resp bb_common_bss.will_wont_resp
#define flowmode bb_common_bss.flowmode
#define slctab bb_common_bss.slctab
#define terminaltype bb_common_bss.terminaltype
#define telnetd_clocks bb_common_bss.telnetd_clocks
#define ptyobuf bb_common_bss.ptyobuf
#define pfrontp bb_common_bss.pfrontp
#define pbackp bb_common_bss.pbackp
#define netibuf bb_common_bss.netibuf
#define netip bb_common_bss.netip
#define netobuf bb_common_bss.netobuf
#define nfrontp bb_common_bss.nfrontp
#define nbackp bb_common_bss.nbackp
#define neturg bb_common_bss.neturg
#define pcc bb_common_bss.pcc
#define ncc bb_common_bss.ncc
#define pty bb_common_bss.pty
#define SYNCHing bb_common_bss.SYNCHing
#define termbuf bb_common_bss.termbuf
#define termbuf2 bb_common_bss.termbuf2
#define host_name bb_common_bss.host_name
#define ptyibuf bb_common_bss.ptyibuf
#define ptyibuf2 bb_common_bss.ptyibuf2
#define subbuffer bb_common_bss.subbuffer
#define turn_on_sga bb_common_bss.turn_on_sga
#define linedata bb_common_bss.linedata
#define def_row bb_common_bss.def_row
#define def_col bb_common_bss.def_col
#define terminalname bb_common_bss.terminalname
#define timeout bb_common_bss.timeout
#define putlocation bb_common_bss.putlocation
#define editedhost bb_common_bss.editedhost
#define kerninfo bb_common_bss.kerninfo
#define envinit bb_common_bss.envinit
#define dis_hostinfo bb_common_bss.dis_hostinfo
#define state bb_common_bss.state
#define def_tspeed bb_common_bss.def_tspeed
#define def_rspeed bb_common_bss.def_rspeed
#define loginprg bb_common_bss.loginprg
#define ptyip bb_common_bss.ptyip
#define Tline bb_common_bss.Tline
#define not42 bb_common_bss.not42
#define ptyslavefd bb_common_bss.ptyslavefd
#define subpointer bb_common_bss.subpointer
#define subend bb_common_bss.subend

static const char ttytype_sbbuf[] =
			{ IAC, SB, TELOPT_TTYPE, TELQUAL_SEND, IAC, SE };


#define DEFAULT_IM      "%i\r\n%s %r (%h) (%t)\r\n\r\n"

/* other external variables */
extern  char **environ;

static void __attribute__ ((noreturn))
fatalperror(int f, const char *msg)
{
	char buf[BUFSIZ];
	int n;

#if defined(__ANDROID_API__) && (__ANDROID_API__ < 29)
	n = sprintf(buf, "%s: %s: %s.\r\n", bb_applet_name, msg, strerror(errno));
#else
	n = sprintf(buf, "%s: %s: %m.\r\n", bb_applet_name, msg);
#endif
	(void) write(f, buf, n);
	sleep(1);       /*XXX*/
	exit(1);
}


/* init_termbuf()
 * set_termbuf()
 *
 * These three routines are used to get and set the "termbuf" structure
 * to and from the kernel.  init_termbuf() gets the current settings.
 * set_termbuf() writes the structure into the kernel. */

static void init_termbuf(void)
{
    tcgetattr(pty, &termbuf);
    termbuf2 = termbuf;
}

static void set_termbuf(void)
{
    if (memcmp(&termbuf, &termbuf2, sizeof(termbuf))) {
	tcsetattr(pty, TCSANOW, &termbuf);
    }
}

/* spcset(func, valp, valpp)
 *
 * This function takes various special characters (func), and
 * sets *valp to the current value of that character, and
 * *valpp to point to where in the "termbuf" structure that
 * value is kept.
 *
 * It returns the SLC_ level of support for this function. */

static int spcset(int func, cc_t *valp, cc_t **valpp)
{
#define setval(a, b)    *valp = termbuf.c_cc[a]; \
			*valpp = &termbuf.c_cc[a]; \
			return(b);
#define defval(a) *valp = ((cc_t)a); *valpp = (cc_t *)0; return(SLC_DEFAULT);

    switch(func) {
    case SLC_EOF:
	setval(VEOF, SLC_VARIABLE);
    case SLC_EC:
	setval(VERASE, SLC_VARIABLE);
    case SLC_EL:
	setval(VKILL, SLC_VARIABLE);
    case SLC_IP:
	setval(VINTR, SLC_VARIABLE|SLC_FLUSHIN|SLC_FLUSHOUT);
    case SLC_ABORT:
	setval(VQUIT, SLC_VARIABLE|SLC_FLUSHIN|SLC_FLUSHOUT);
    case SLC_XON:
#ifdef VSTART
	setval(VSTART, SLC_VARIABLE);
#else
	defval(0x13);
#endif
    case SLC_XOFF:
#ifdef  VSTOP
	setval(VSTOP, SLC_VARIABLE);
#else
	defval(0x11);
#endif
    case SLC_EW:
#ifdef  VWERASE
	setval(VWERASE, SLC_VARIABLE);
#else
	defval(0);
#endif
    case SLC_RP:
#ifdef  VREPRINT
	setval(VREPRINT, SLC_VARIABLE);
#else
	defval(0);
#endif
    case SLC_LNEXT:
#ifdef  VLNEXT
	setval(VLNEXT, SLC_VARIABLE);
#else
	defval(0);
#endif
    case SLC_AO:
#if     !defined(VDISCARD) && defined(VFLUSHO)
# define VDISCARD VFLUSHO
#endif
#ifdef  VDISCARD
	setval(VDISCARD, SLC_VARIABLE|SLC_FLUSHOUT);
#else
	defval(0);
#endif
    case SLC_SUSP:
#ifdef  VSUSP
	setval(VSUSP, SLC_VARIABLE|SLC_FLUSHIN);
#else
	defval(0);
#endif
#ifdef  VEOL
    case SLC_FORW1:
	setval(VEOL, SLC_VARIABLE);
#endif
#ifdef  VEOL2
    case SLC_FORW2:
	setval(VEOL2, SLC_VARIABLE);
#endif
    case SLC_AYT:
#ifdef  VSTATUS
	setval(VSTATUS, SLC_VARIABLE);
#else
	defval(0);
#endif

    case SLC_BRK:
    case SLC_SYNCH:
    case SLC_EOR:
	defval(0);

    default:
	*valp = 0;
	*valpp = 0;
	return(SLC_NOSUPPORT);
    }
}

/* get_slc_defaults
 * Initialize the slc mapping table. */
static void get_slc_defaults(void)
{
    int i;

    init_termbuf();
    for (i = 1; i <= NSLC; i++) {
	slctab[i].defset.flag = spcset(i, &slctab[i].defset.val,
				       &slctab[i].sptr);
	slctab[i].current.flag = SLC_NOSUPPORT;
	slctab[i].current.val = 0;
    }
}


static int envvarok(const char *varp)
{
    /*
     * Allow only these variables.
     */
    if (!strcmp(varp, "TERM")) return 1;
    if (!strcmp(varp, "DISPLAY")) return 1;
    if (!strcmp(varp, "USER")) return 1;
    if (!strcmp(varp, "LOGNAME")) return 1;
    if (!strcmp(varp, "POSIXLY_CORRECT")) return 1;

    /* optionally syslog(LOG_INFO) here */
    return 0;
}

static const char doopt[] = { IAC, DO,   '%', 'c', 0 };
static const char dont[] =  { IAC, DONT, '%', 'c', 0 };
static const char will[] =  { IAC, WILL, '%', 'c', 0 };
static const char wont[] =  { IAC, WONT, '%', 'c', 0 };

/*
 * Buffer for sub-options, and macros
 * for suboptions buffer manipulations
 */
#define SB_CLEAR()      subpointer = subbuffer;
#define SB_TERM()       { subend = subpointer; SB_CLEAR(); }
#define SB_ACCUM(c)     if (subpointer < (subbuffer+sizeof subbuffer)) { \
				*subpointer++ = (c); \
			}
#define SB_GET()        ((*subpointer++)&0xff)
#define SB_EOF()        (subpointer >= subend)
#define SB_LEN()        (subend - subpointer)

/*
 * State for recv fsm
 */
#define TS_DATA         0       /* base state */
#define TS_IAC          1       /* look for double IAC's */
#define TS_CR           2       /* CR-LF ->'s CR */
#define TS_SB           3       /* throw away begin's... */
#define TS_SE           4       /* ...end's (suboption negotiation) */
#define TS_WILL         5       /* will option negotiation */
#define TS_WONT         6       /* wont " */
#define TS_DO           7       /* do " */
#define TS_DONT         8       /* dont " */

static void doeof(void)
{
    init_termbuf();

    *pfrontp++ = slctab[SLC_EOF].sptr ?
	     (unsigned char)*slctab[SLC_EOF].sptr : '\004';
}

static void tty_setecho(int on)
{
    if (on) termbuf.c_lflag |= ECHO;
    else termbuf.c_lflag &= ~ECHO;
}

static void tty_binaryout(int on)
{
    if (on) {
	termbuf.c_cflag &= ~(CSIZE|PARENB);
	termbuf.c_cflag |= CS8;
	termbuf.c_oflag &= ~OPOST;
    }
    else {
	termbuf.c_cflag &= ~CSIZE;
	termbuf.c_cflag |= CS7|PARENB;
	termbuf.c_oflag |= OPOST;
    }
}

static void telnetd_send_will(int option, int init)
{
    if (init) {
	if ((will_wont_resp[option] == 0 && my_state_is_will(option))||
	    my_want_state_is_will(option))
	    return;
	set_my_want_state_will(option);
	will_wont_resp[option]++;
    }
    (void) sprintf(nfrontp, will, option);
    nfrontp += sizeof (doopt) - 2;

}

static void send_wont_telnetd(int option, int init)
{
    if (init) {
	if ((will_wont_resp[option] == 0 && my_state_is_wont(option)) ||
	    my_want_state_is_wont(option))
	    return;
	set_my_want_state_wont(option);
	will_wont_resp[option]++;
    }
    (void) sprintf(nfrontp, wont, option);
    nfrontp += sizeof (wont) - 2;

}

/*
 * When we get a DONT SGA, we will try once to turn it
 * back on.  If the other side responds DONT SGA, we
 * leave it at that.  This is so that when we talk to
 * clients that understand KLUDGELINEMODE but not LINEMODE,
 * we'll keep them in char-at-a-time mode.
 */

static void dontoption(int option)
{
    /*
     * Process client input.
     */

    if (will_wont_resp[option]) {
	will_wont_resp[option]--;
	if (will_wont_resp[option] && my_state_is_wont(option))
	    will_wont_resp[option]--;
    }
    if ((will_wont_resp[option] == 0) && (my_want_state_is_will(option))) {
	switch (option) {
	case TELOPT_BINARY:
	    init_termbuf();
	    tty_binaryout(0);
	    set_termbuf();
	    break;

	case TELOPT_ECHO:       /* we should stop echoing */
	    {
		init_termbuf();
		tty_setecho(0);
		set_termbuf();
	    }
	    break;

	case TELOPT_SGA:
	    set_my_want_state_wont(option);
	    if (my_state_is_will(option))
		send_wont_telnetd(option, 0);
	    set_my_state_wont(option);
	    if (turn_on_sga ^= 1) telnetd_send_will(option,1);
	    return;
	 default:
	    break;
	}

	set_my_want_state_wont(option);
	if (my_state_is_will(option))
	    send_wont_telnetd(option, 0);
    }
    set_my_state_wont(option);
}

/*
 * cleanup()
 *
 * This is the routine to call when we are all through, to
 * clean up anything that needs to be cleaned up.
 */
static void __attribute__ ((noreturn)) cleanup(int sig)
{
    (void)sig;

    bb_wutmp_logout(Tline);
    shutdown(0, 2);
    exit(1);
}

/*
 *  netflush
 *              Send as much data as possible to the network,
 *      handling requests for urgent data.
 */
static void
telnetd_netflush(void)
{
    int n;

    if ((n = nfrontp - nbackp) > 0) {
	/*
	 * if no urgent data, or if the other side appears to be an
	 * old 4.2 client (and thus unable to survive TCP urgent data),
	 * write the entire buffer in non-OOB mode.
	 */
	if ((neturg == 0) || (not42 == 0)) {
	    n = write(1, nbackp, n);  /* normal write */
	} else {
	    n = neturg - nbackp;
	    /*
	     * In 4.2 (and 4.3) systems, there is some question about
	     * what byte in a sendOOB operation is the "OOB" data.
	     * To make ourselves compatible, we only send ONE byte
	     * out of band, the one WE THINK should be OOB (though
	     * we really have more the TCP philosophy of urgent data
	     * rather than the Unix philosophy of OOB data).
	     */
	    if (n > 1) {
		n = send(1, nbackp, n-1, 0);  /* send URGENT all by itself */
	    } else {
		n = send(1, nbackp, n, MSG_OOB);      /* URGENT data */
	    }
	}
    }
    if (n < 0) {
	if (errno == EWOULDBLOCK || errno == EINTR)
		return;
	cleanup(0);
    }
    nbackp += n;
    if (nbackp >= neturg) {
	neturg = 0;
    }
    if (nbackp == nfrontp) {
	nbackp = nfrontp = netobuf;
    }
}  /* end of netflush */


static void dooption(int option)
{
    int changeok = 0;

    /*
     * Process client input.
     */

    if (will_wont_resp[option]) {
	will_wont_resp[option]--;
	if (will_wont_resp[option] && my_state_is_will(option))
	    will_wont_resp[option]--;
    }
    if ((will_wont_resp[option] == 0) && (my_want_state_is_wont(option))) {
	switch (option) {
	case TELOPT_ECHO:
	    {
		init_termbuf();
		tty_setecho(1);
		set_termbuf();
	    }
	    changeok++;
	    break;

	case TELOPT_BINARY:
	    init_termbuf();
	    tty_binaryout(1);
	    set_termbuf();
	    changeok++;
	    break;

	case TELOPT_SGA:
	    turn_on_sga = 0;
	    changeok++;
	    break;

	case TELOPT_STATUS:
	    changeok++;
	    break;

	case TELOPT_TM:
	    /*
	     * Special case for TM.  We send a WILL, but
	     * pretend we sent a WONT.
	     */
	    telnetd_send_will(option, 0);
	    set_my_want_state_wont(option);
	    set_my_state_wont(option);
	    return;

	case TELOPT_LOGOUT:
	    /*
	     * When we get a LOGOUT option, respond
	     * with a WILL LOGOUT, make sure that
	     * it gets written out to the network,
	     * and then just go away...
	     */
	    set_my_want_state_will(TELOPT_LOGOUT);
	    telnetd_send_will(TELOPT_LOGOUT, 0);
	    set_my_state_will(TELOPT_LOGOUT);
	    telnetd_netflush();
	    cleanup(0);
	    /* NOT REACHED */
	    break;

	case TELOPT_LINEMODE:
	case TELOPT_TTYPE:
	case TELOPT_NAWS:
	case TELOPT_TSPEED:
	case TELOPT_LFLOW:
	case TELOPT_XDISPLOC:
	case TELOPT_ENVIRON:
	default:
	    break;
	}
	if (changeok) {
	    set_my_want_state_will(option);
	    telnetd_send_will(option, 0);
	}
	else {
	    will_wont_resp[option]++;
	    send_wont_telnetd(option, 0);
	}
    }
    set_my_state_will(option);
}

static void tty_r_t_speed(int valr, int valt)
{
    speed_t s;

    s = bb_value_to_baud(valr);
    if(s == (speed_t) - 1)
	s = bb_value_to_baud(9600);
    cfsetispeed(&termbuf, s);
    s = bb_value_to_baud(valt);
    if(s == (speed_t) - 1)
	s = bb_value_to_baud(9600);
    cfsetospeed(&termbuf, s);
}

/*
 * clientstat
 *
 * Process linemode related requests from the client.
 * Client can request a change to only one of linemode, editmode or slc's
 * at a time, and if using kludge linemode, then only linemode may be
 * affected.
 */
static void
clientstat(int code, int parm1, int parm2)
{
	/*
	 * Get a copy of terminal characteristics.
	 */
	init_termbuf();

	/*
	 * Process request from client. code tells what it is.
	 */
	switch (code) {
	case TELOPT_NAWS:
#ifdef  TIOCSWINSZ
	    {
		struct winsize ws;

		def_col = parm1;
		def_row = parm2;

		/*
		 * Change window size as requested by client.
		 */

		ws.ws_col = parm1;
		ws.ws_row = parm2;
		(void) ioctl(pty, TIOCSWINSZ, (char *)&ws);
	    }
#endif  /* TIOCSWINSZ */

		break;

	case TELOPT_TSPEED:
	    {
		def_tspeed = parm1;
		def_rspeed = parm2;
		/*
		 * Change terminal speed as requested by client.
		 * We set the receive speed first, so that if we can't
		 * store seperate receive and transmit speeds, the transmit
		 * speed will take precedence.
		 */
		tty_r_t_speed(parm2, parm1);
		set_termbuf();

		break;

	    }  /* end of case TELOPT_TSPEED */

	default:
		/* What? */
		break;
	}  /* end of switch */

	telnetd_netflush();

}  /* end of clientstat */


/*
 * writenet
 *
 * Just a handy little function to write a bit of raw data to the net.
 * It will force a transmit of the buffer if necessary
 *
 * arguments
 *    ptr - A pointer to a character string to write
 *    len - How many bytes to write
 */
static void
writenet(unsigned char *ptr, int len)
{
	/* flush buffer if no room for new data) */
	if ((&netobuf[BUFSIZ] - nfrontp) < len) {
		/* if this fails, don't worry, buffer is a little big */
		telnetd_netflush();
	}

	memcpy(nfrontp, ptr, len);
	nfrontp += len;

}  /* end of writenet */


static void send_status(void)
{
#define ADD(c)   *ncp++ = c;
#define ADD_DATA(c) { *ncp++ = c; if (c == SE) *ncp++ = c; }

    unsigned char statusbuf[256];
    unsigned char *ncp;
    unsigned char i;

    ncp = statusbuf;

    telnetd_netflush(); /* get rid of anything waiting to go out */

    ADD(IAC);
    ADD(SB);
    ADD(TELOPT_STATUS);
    ADD(TELQUAL_IS);

    /*
     * We check the want_state rather than the current state,
     * because if we received a DO/WILL for an option that we
     * don't support, and the other side didn't send a DONT/WONT
     * in response to our WONT/DONT, then the "state" will be
     * WILL/DO, and the "want_state" will be WONT/DONT.  We
     * need to go by the latter.
     */
    for (i = 0; i < NTELOPTS; i++) {
	if (my_want_state_is_will(i)) {
	    ADD(WILL);
	    ADD_DATA(i);
	    if (i == IAC) ADD(IAC);
	}
	if (his_want_state_is_will(i)) {
	    ADD(DO);
	    ADD_DATA(i);
	    if (i == IAC) ADD(IAC);
	}
    }

    if (his_want_state_is_will(TELOPT_LFLOW)) {
	ADD(SB);
	ADD(TELOPT_LFLOW);
	ADD(flowmode);
	ADD(SE);
    }

    ADD(IAC);
    ADD(SE);

    writenet(statusbuf, ncp - statusbuf);
    telnetd_netflush(); /* Send it on its way */

#undef ADD
#undef ADD_DATA
}

/*
 * suboption()
 *
 *      Look at the sub-option buffer, and try to be helpful to the other
 * side.
 *
 *      Currently we recognize:
 *
 *      Terminal type is
 *      Linemode
 *      Window size
 *      Terminal speed
 */
static void suboption(void)
{
    int subchar;

    subchar = SB_GET();
    switch (subchar) {
     case TELOPT_TSPEED: {
	int xspeed, rspeed;
	if (his_state_is_wont(TELOPT_TSPEED))   /* Ignore if option disabled */
	    break;

	settimer(tspeedsubopt);
	if (SB_EOF() || SB_GET() != TELQUAL_IS) return;
	xspeed = atoi((char *)subpointer);

	while (SB_GET() != ',' && !SB_EOF());
	if (SB_EOF()) return;

	rspeed = atoi((char *)subpointer);
	clientstat(TELOPT_TSPEED, xspeed, rspeed);
	break;
    }

    case TELOPT_TTYPE: {                /* Yaaaay! */
	if (his_state_is_wont(TELOPT_TTYPE))    /* Ignore if option disabled */
	    break;
	settimer(ttypesubopt);

	if (SB_EOF() || SB_GET() != TELQUAL_IS) {
	    return;             /* ??? XXX but, this is the most robust */
	}

	terminaltype = terminalname;

	while ((terminaltype < (terminalname + sizeof terminalname-1)) &&
	       !SB_EOF())
	{
	    int c;
	    c = SB_GET();
	    if (isupper(c)) {
		c = tolower(c);
	    }
	    *terminaltype++ = c;    /* accumulate name */
	}
	*terminaltype = 0;
	terminaltype = terminalname;
	break;
    }

    case TELOPT_NAWS: {
	int xwinsize, ywinsize;
	if (his_state_is_wont(TELOPT_NAWS))     /* Ignore if option disabled */
	    break;

	if (SB_EOF()) return;
	xwinsize = SB_GET() << 8;
	if (SB_EOF()) return;
	xwinsize |= SB_GET();
	if (SB_EOF()) return;
	ywinsize = SB_GET() << 8;
	if (SB_EOF()) return;
	ywinsize |= SB_GET();
	clientstat(TELOPT_NAWS, xwinsize, ywinsize);
	break;
    }

    case TELOPT_STATUS: {
	int mode;

	if (SB_EOF())
	    break;
	mode = SB_GET();
	switch (mode) {
	case TELQUAL_SEND:
	    if (my_state_is_will(TELOPT_STATUS))
		send_status();
	    break;

	case TELQUAL_IS:
	    break;

	default:
	    break;
	}
	break;
    }  /* end of case TELOPT_STATUS */

    case TELOPT_XDISPLOC: {
	if (SB_EOF() || SB_GET() != TELQUAL_IS)
		return;
	settimer(xdisplocsubopt);
	subpointer[SB_LEN()] = '\0';
	bb_setenv1("DISPLAY", (char *)subpointer);
	break;
    }  /* end of case TELOPT_XDISPLOC */

    case TELOPT_ENVIRON: {
	int c;
	char *cp, *varp, *valp;

	if (SB_EOF())
		return;
	c = SB_GET();
	if (c == TELQUAL_IS)
		settimer(environsubopt);
	else if (c != TELQUAL_INFO)
		return;

	while (!SB_EOF() && SB_GET() != ENV_VAR)
		;

	if (SB_EOF())
		return;

	cp = varp = (char *)subpointer;
	valp = 0;

	while (!SB_EOF()) {
	    switch (c = SB_GET()) {
	    case ENV_VALUE:
		*cp = '\0';
		cp = valp = (char *)subpointer;
		break;

	    case ENV_VAR:
		*cp = '\0';
		if (envvarok(varp)) {
		    if (valp)
			bb_setenv1(varp, valp);
		    else
			unsetenv(varp);
		}
		cp = varp = (char *)subpointer;
		valp = 0;
		break;

	    case ENV_ESC:
		if (SB_EOF())
		    break;
		c = SB_GET();
		/* FALL THROUGH */
	    default:
		*cp++ = c;
		break;
	    }
	}
	*cp = '\0';
	if (envvarok(varp)) {
	    if (valp)
		bb_setenv1(varp, valp);
	    else
		unsetenv(varp);
	}
	break;
    }  /* end of case TELOPT_ENVIRON */
    default:
	break;
    }  /* end of switch */

}  /* end of suboption */

#ifndef TCSIG
# ifdef TIOCSIG
#  define TCSIG TIOCSIG
# endif
#endif

static void
ptyflush(void)
{
	int n;

	if ((n = pfrontp - pbackp) > 0) {
		n = write(pty, pbackp, n);
	}
	if (n < 0) {
		if (errno == EWOULDBLOCK || errno == EINTR)
			return;
		cleanup(0);
	}
	pbackp += n;
	if (pbackp == pfrontp)
		pbackp = pfrontp = ptyobuf;
}


/*
 * Send interrupt to process on other side of pty.
 * If it is in raw mode, just write NULL;
 * otherwise, write intr char.
 */
static void interrupt(void)
{
    ptyflush(); /* half-hearted */

#ifdef  TCSIG
    (void) ioctl(pty, TCSIG, (char *)SIGINT);
#else   /* TCSIG */
    init_termbuf();
    *pfrontp++ = slctab[SLC_IP].sptr ?
	 (unsigned char)*slctab[SLC_IP].sptr : '\177';
#endif  /* TCSIG */
}

/*
 * nextitem()
 *
 *      Return the address of the next "item" in the TELNET data
 * stream.  This will be the address of the next character if
 * the current address is a user data character, or it will
 * be the address of the character following the TELNET command
 * if the current address is a TELNET IAC ("I Am a Command")
 * character.
 */
static char *
nextitem(char *current)
{
    if ((*current&0xff) != IAC) {
	return current+1;
    }
    switch (*(current+1)&0xff) {
    case DO:
    case DONT:
    case WILL:
    case WONT:
	return current+3;
    case SB:            /* loop forever looking for the SE */
	{
	    char *look = current+2;

	    for (;;) {
		if ((*look++&0xff) == IAC) {
		    if ((*look++&0xff) == SE) {
			return look;
		    }
		}
	    }
	}
    default:
	return current+2;
    }
}  /* end of nextitem */

/*
 * netclear()
 *
 *      We are about to do a TELNET SYNCH operation.  Clear
 * the path to the network.
 *
 *      Things are a bit tricky since we may have sent the first
 * byte or so of a previous TELNET command into the network.
 * So, we have to scan the network buffer from the beginning
 * until we are up to where we want to be.
 *
 *      A side effect of what we do, just to keep things
 * simple, is to clear the urgent data pointer.  The principal
 * caller should be setting the urgent data pointer AFTER calling
 * us in any case.
 */
static void
netclear(void)
{
    char *thisitem, *next;
    char *good;
#define wewant(p)       ((nfrontp > p) && ((*p&0xff) == IAC) && \
				((*(p+1)&0xff) != EC) && ((*(p+1)&0xff) != EL))

    thisitem = netobuf;

    while ((next = nextitem(thisitem)) <= nbackp) {
	thisitem = next;
    }

    /* Now, thisitem is first before/at boundary. */

    good = netobuf;     /* where the good bytes go */

    while (nfrontp > thisitem) {
	if (wewant(thisitem)) {
	    int length;

	    next = thisitem;
	    do {
		next = nextitem(next);
	    } while (wewant(next) && (nfrontp > next));
	    length = next-thisitem;
	    memcpy(good, thisitem, length);
	    good += length;
	    thisitem = next;
	} else {
	    thisitem = nextitem(thisitem);
	}
    }

    nbackp = netobuf;
    nfrontp = good;             /* next byte to be sent */
    neturg = 0;
}  /* end of netclear */

/*
 * When we get an AYT, if ^T is enabled, use that.  Otherwise,
 * just send back "[Yes]".
 */
static void recv_ayt(void)
{
#if     defined(SIGINFO) && defined(TCSIG)
    if (slctab[SLC_AYT].sptr && *slctab[SLC_AYT].sptr != _POSIX_VDISABLE) {
	(void) ioctl(pty, TCSIG, (char *)SIGINFO);
	return;
    }
#endif
    sprintf(nfrontp, "\r\n[%s : yes]\r\n", host_name);
    nfrontp += strlen(nfrontp);
}

/*
 * Send quit to process on other side of pty.
 * If it is in raw mode, just write NULL;
 * otherwise, write quit char.
 */
static void telnetd_sendbrk(void)
{
    ptyflush(); /* half-hearted */
#ifdef  TCSIG
    (void) ioctl(pty, TCSIG, (char *)SIGQUIT);
#else   /* TCSIG */
    init_termbuf();
    *pfrontp++ = slctab[SLC_ABORT].sptr ?
	 (unsigned char)*slctab[SLC_ABORT].sptr : '\034';
#endif  /* TCSIG */
}


static void telnetd_sendsusp(void)
{
#ifdef  SIGTSTP
    ptyflush(); /* half-hearted */
# ifdef TCSIG
    (void) ioctl(pty, TCSIG, (char *)SIGTSTP);
# else  /* TCSIG */
    *pfrontp++ = slctab[SLC_SUSP].sptr ?
	(unsigned char)*slctab[SLC_SUSP].sptr : '\032';
# endif /* TCSIG */
#endif  /* SIGTSTP */
}

/*
 * Check a descriptor to see if out of band data exists on it.
 */
static int
telnetd_stilloob(void) /* socket number */
{
    fd_set      excepts;
    int value;
    int s = 0;

    do {
	FD_ZERO(&excepts);
	FD_SET(s, &excepts);
	value = select(s+1, (fd_set *)0, (fd_set *)0, &excepts, &timeout);
    } while ((value == -1) && (errno == EINTR));

    if (value < 0) {
	fatalperror(pty, "select");
    }
    if (FD_ISSET(s, &excepts)) {
	return 1;
    } else {
	return 0;
    }
}

static void tty_binaryin(int on)
{
    if (on) {
	termbuf.c_iflag &= ~ISTRIP;
    }
    else {
	termbuf.c_iflag |= ISTRIP;
    }
}

/*
 * The will/wont/do/dont state machines are based on Dave Borman's
 * Telnet option processing state machine.
 *
 * These correspond to the following states:
 *      my_state = the last negotiated state
 *      want_state = what I want the state to go to
 *      want_resp = how many requests I have sent
 * All state defaults are negative, and resp defaults to 0.
 *
 * When initiating a request to change state to new_state:
 *
 * if ((want_resp == 0 && new_state == my_state) || want_state == new_state) {
 *      do nothing;
 * } else {
 *      want_state = new_state;
 *      send new_state;
 *      want_resp++;
 * }
 *
 * When receiving new_state:
 *
 * if (want_resp) {
 *      want_resp--;
 *      if (want_resp && (new_state == my_state))
 *              want_resp--;
 * }
 * if ((want_resp == 0) && (new_state != want_state)) {
 *      if (ok_to_switch_to new_state)
 *              want_state = new_state;
 *      else
 *              want_resp++;
 *      send want_state;
 * }
 * my_state = new_state;
 *
 * Note that new_state is implied in these functions by the function itself.
 * will and do imply positive new_state, wont and dont imply negative.
 *
 * Finally, there is one catch.  If we send a negative response to a
 * positive request, my_state will be the positive while want_state will
 * remain negative.  my_state will revert to negative when the negative
 * acknowlegment arrives from the peer.  Thus, my_state generally tells
 * us not only the last negotiated state, but also tells us what the peer
 * wants to be doing as well.  It is important to understand this difference
 * as we may wish to be processing data streams based on our desired state
 * (want_state) or based on what the peer thinks the state is (my_state).
 *
 * This all works fine because if the peer sends a positive request, the data
 * that we receive prior to negative acknowlegment will probably be affected
 * by the positive state, and we can process it as such (if we can; if we
 * can't then it really doesn't matter).  If it is that important, then the
 * peer probably should be buffering until this option state negotiation
 * is complete.
 *
 */
static void telnetd_send_do(int option, int init)
{
    if (init) {
	if ((do_dont_resp[option] == 0 && his_state_is_will(option)) ||
	    his_want_state_is_will(option))
	    return;
	/*
	 * Special case for TELOPT_TM:  We send a DO, but pretend
	 * that we sent a DONT, so that we can send more DOs if
	 * we want to.
	 */
	if (option == TELOPT_TM)
	    set_his_want_state_wont(option);
	else
	    set_his_want_state_will(option);
	do_dont_resp[option]++;
    }
    sprintf(nfrontp, doopt, option);
    nfrontp += sizeof (dont) - 2;

}

static void send_dont_telnetd(int option, int init)
{
    if (init) {
	if ((do_dont_resp[option] == 0 && his_state_is_wont(option)) ||
	    his_want_state_is_wont(option))
	    return;
	set_his_want_state_wont(option);
	do_dont_resp[option]++;
    }
    sprintf(nfrontp, dont, option);
    nfrontp += sizeof (doopt) - 2;

}

static void telnetd_willoption(int option)
{
    int changeok = 0;
    void (*func)(void) = 0;

    /*
     * process input from peer.
     */

    if (do_dont_resp[option]) {
	do_dont_resp[option]--;
	if (do_dont_resp[option] && his_state_is_will(option))
	    do_dont_resp[option]--;
    }
    if (do_dont_resp[option] == 0) {
	if (his_want_state_is_wont(option)) {
	    switch (option) {

	    case TELOPT_BINARY:
		init_termbuf();
		tty_binaryin(1);
		set_termbuf();
		changeok++;
		break;

	    case TELOPT_ECHO:
		/*
		 * See comments below for more info.
		 */
		not42 = 0;      /* looks like a 4.2 system */
		break;

	    case TELOPT_TM:
		/*
		 * We never respond to a WILL TM, and
		 * we leave the state WONT.
		 */
		return;

	    case TELOPT_LFLOW:
		 /*
		  * If we are going to support flow control
		  * option, then don't worry peer that we can't
		  * change the flow control characters.
		  */
		slctab[SLC_XON].defset.flag &= ~SLC_LEVELBITS;
		slctab[SLC_XON].defset.flag |= SLC_DEFAULT;
		slctab[SLC_XOFF].defset.flag &= ~SLC_LEVELBITS;
		slctab[SLC_XOFF].defset.flag |= SLC_DEFAULT;
	    case TELOPT_TTYPE:
	    case TELOPT_SGA:
	    case TELOPT_NAWS:
	    case TELOPT_TSPEED:
	    case TELOPT_XDISPLOC:
	    case TELOPT_ENVIRON:
		changeok++;
		break;

	    default:
		break;
	    }
	    if (changeok) {
		set_his_want_state_will(option);
		telnetd_send_do(option, 0);
	    }
	    else {
		do_dont_resp[option]++;
		send_dont_telnetd(option, 0);
	    }
	}
	else {
	    /*
	     * Option processing that should happen when
	     * we receive conformation of a change in
	     * state that we had requested.
	     */
	    switch (option) {
	     case TELOPT_ECHO:
		not42 = 0;      /* looks like a 4.2 system */
		/*
		 * Egads, he responded "WILL ECHO".  Turn
		 * it off right now!
		 */
		send_dont_telnetd(option, 1);
		/*
		 * "WILL ECHO".  Kludge upon kludge!
		 * A 4.2 client is now echoing user input at
		 * the tty.  This is probably undesireable and
		 * it should be stopped.  The client will
		 * respond WONT TM to the DO TM that we send to
		 * check for kludge linemode.  When the WONT TM
		 * arrives, linemode will be turned off and a
		 * change propogated to the pty.  This change
		 * will cause us to process the new pty state
		 * in localstat(), which will notice that
		 * linemode is off and send a WILL ECHO
		 * so that we are properly in character mode and
		 * all is well.
		 */
		break;
	    }
	}
    }
    set_his_state_will(option);
    if (func) (*func)();
}

static void wontoption(int option)
{
    /*
     * Process client input.
     */

    if (do_dont_resp[option]) {
	do_dont_resp[option]--;
	if (do_dont_resp[option] && his_state_is_wont(option))
	    do_dont_resp[option]--;
    }
    if (do_dont_resp[option] == 0) {
	if (his_want_state_is_will(option)) {
	    /* it is always ok to change to negative state */
	    switch (option) {
	    case TELOPT_ECHO:
		not42 = 1; /* doesn't seem to be a 4.2 system */
		break;

	    case TELOPT_BINARY:
		init_termbuf();
		tty_binaryin(0);
		set_termbuf();
		break;

	    case TELOPT_TM:
		/*
		 * If we get a WONT TM, and had sent a DO TM,
		 * don't respond with a DONT TM, just leave it
		 * as is.  Short circut the state machine to
		 * achive this.
		 */
		set_his_want_state_wont(TELOPT_TM);
		return;

	    case TELOPT_LFLOW:
		/*
		 * If we are not going to support flow control
		 * option, then let peer know that we can't
		 * change the flow control characters.
		 */
		slctab[SLC_XON].defset.flag &= ~SLC_LEVELBITS;
		slctab[SLC_XON].defset.flag |= SLC_CANTCHANGE;
		slctab[SLC_XOFF].defset.flag &= ~SLC_LEVELBITS;
		slctab[SLC_XOFF].defset.flag |= SLC_CANTCHANGE;
		break;

		/*
		 * For options that we might spin waiting for
		 * sub-negotiation, if the client turns off the
		 * option rather than responding to the request,
		 * we have to treat it here as if we got a response
		 * to the sub-negotiation, (by updating the timers)
		 * so that we'll break out of the loop.
		 */
	    case TELOPT_TTYPE:
		settimer(ttypesubopt);
		break;

	    case TELOPT_TSPEED:
		settimer(tspeedsubopt);
		break;

	    case TELOPT_XDISPLOC:
		settimer(xdisplocsubopt);
		break;

	    case TELOPT_ENVIRON:
		settimer(environsubopt);
		break;

	    default:
		break;
	    }
	    set_his_want_state_wont(option);
	    if (his_state_is_will(option)) send_dont_telnetd(option, 0);
	}
	else {
	    switch (option) {
	     case TELOPT_TM:
		 break;

	    default:
		break;
	    }
	}
    }
}  /* end of wontoption */

static void telnetd_telrcv(void)
{
    int c;

    while (ncc > 0) {
	if ((&ptyobuf[BUFSIZ] - pfrontp) < 2) break;
	c = *netip++ & 0377;
	ncc--;

	switch (state) {
	 case TS_CR:
	     state = TS_DATA;
	     /* Strip off \n or \0 after a \r */
	     if ((c == 0) || (c == '\n')) {
		 break;
	     }
	     /* FALL THROUGH */

	 case TS_DATA:
	     if (c == IAC) {
		 state = TS_IAC;
		 break;
	     }
	     /*
	      * We now map \r\n ==> \r for pragmatic reasons.
	      * Many client implementations send \r\n when
	      * the user hits the CarriageReturn key.
	      *
	      * We USED to map \r\n ==> \n, since \r\n says
	      * that we want to be in column 1 of the next
	      * printable line, and \n is the standard
	      * unix way of saying that (\r is only good
	      * if CRMOD is set, which it normally is).
	      */
	     if ((c == '\r') && his_state_is_wont(TELOPT_BINARY)) {
		     state = TS_CR;
	     }
	     *pfrontp++ = c;
	     break;

	 case TS_IAC:
	 gotiac:
	     switch (c) {

		 /*
		  * Send the process on the pty side an
		  * interrupt.  Do this with a NULL or
		  * interrupt char; depending on the tty mode.
		  */
	      case IP:
		  interrupt();
		  break;
	      case BREAK:
		  telnetd_sendbrk();
		  break;

		  /*
		   * Are You There?
		   */
	      case AYT:
		  recv_ayt();
		  break;

		  /*
		   * Abort Output
		   */
	      case AO:
		  {
		      ptyflush();       /* half-hearted */
		      init_termbuf();

		      if (slctab[SLC_AO].sptr &&
			  *slctab[SLC_AO].sptr != (cc_t)(_POSIX_VDISABLE))
		      {
			  *pfrontp++ =
			      (unsigned char)*slctab[SLC_AO].sptr;
		      }

		      netclear();       /* clear buffer back */
		      *nfrontp++ = IAC;
		      *nfrontp++ = DM;
		      neturg = nfrontp-1; /* off by one XXX */
		      break;
		  }

		  /*
		   * Erase Character and
		   * Erase Line
		   */
	      case EC:
	      case EL:
		 {
		     cc_t ch;
		     ptyflush();        /* half-hearted */
		     init_termbuf();
		     if (c == EC) ch = *slctab[SLC_EC].sptr;
		     else ch = *slctab[SLC_EL].sptr;
		     if (ch != (cc_t)(_POSIX_VDISABLE))
			 *pfrontp++ = (unsigned char)ch;
		     break;
		 }

		  /*
		   * Check for urgent data...
		   */
	      case DM:
		  SYNCHing = telnetd_stilloob();
		  settimer(gotDM);
		  break;

		  /*
		   * Begin option subnegotiation...
		   */
	      case SB:
		  state = TS_SB;
		  SB_CLEAR();
		  continue;

	      case WILL:
		  state = TS_WILL;
		  continue;

	      case WONT:
		  state = TS_WONT;
		  continue;

	      case DO:
		  state = TS_DO;
		  continue;

	      case DONT:
		  state = TS_DONT;
		  continue;

	      case EOR:
		  if (his_state_is_will(TELOPT_EOR)) doeof();
		  break;

		  /*
		   * Handle RFC 10xx Telnet linemode option additions
		   * to command stream (EOF, SUSP, ABORT).
		   */
	      case xEOF:
		  doeof();
		  break;

	      case SUSP:
		  telnetd_sendsusp();
		  break;

	      case ABORT:
		  telnetd_sendbrk();
		  break;

	      case IAC:
		 *pfrontp++ = c;
		  break;
	     }
	     state = TS_DATA;
	     break;

	 case TS_SB:
	     if (c == IAC) {
		 state = TS_SE;
	     }
	     else {
		 SB_ACCUM(c);
	     }
	     break;

	 case TS_SE:
	     if (c != SE) {
		 if (c != IAC) {
				/*
				 * bad form of suboption negotiation.
				 * handle it in such a way as to avoid
				 * damage to local state.  Parse
				 * suboption buffer found so far,
				 * then treat remaining stream as
				 * another command sequence.
				 */

				/* for DIAGNOSTICS */
		     SB_ACCUM(IAC);
		     SB_ACCUM(c);
		     subpointer -= 2;

		     SB_TERM();
		     suboption();
		     state = TS_IAC;
		     goto gotiac;
		 }
		 SB_ACCUM(c);
		 state = TS_SB;
	     }
	     else {
		 /* for DIAGNOSTICS */
		 SB_ACCUM(IAC);
		 SB_ACCUM(SE);
		 subpointer -= 2;

		 SB_TERM();
		 suboption();   /* handle sub-option */
		 state = TS_DATA;
	     }
	     break;

	 case TS_WILL:
	     telnetd_willoption(c);
	     state = TS_DATA;
	     continue;

	 case TS_WONT:
	     wontoption(c);
	     state = TS_DATA;
	     continue;

	 case TS_DO:
	     dooption(c);
	     state = TS_DATA;
	     continue;

	 case TS_DONT:
	     dontoption(c);
	     state = TS_DATA;
	     continue;

	 default:
	     printf("panic state=%d\n", state);
	     bb_error_msg_and_die("panic state=%d\n", state);
	}
    }
}


/* getptyslave()
 *
 * Open the slave side of the pty, and do any initialization
 * that is necessary. */

static void getptyslave(void)
{
    struct winsize ws;
    int t = ptyslavefd;

    /*
     * set up the tty modes as we like them to be.
     */
    init_termbuf();
# ifdef TIOCGWINSZ
    if (def_row || def_col) {
	memset(&ws, 0, sizeof(ws));
	ws.ws_col = def_col;
	ws.ws_row = def_row;
	ioctl(t, TIOCSWINSZ, (char *)&ws);
    }
# endif

    /*
     * Settings for all other termios/termio based
     * systems, other than 4.4BSD.  In 4.4BSD the
     * kernel does the initial terminal setup.
     *
     * XXX what about linux?
     */
#  ifndef       OXTABS
#   define OXTABS       0
#  endif
    termbuf.c_lflag |= ECHO;
    termbuf.c_oflag |= OPOST|ONLCR|OXTABS;
    termbuf.c_iflag |= ICRNL;
    termbuf.c_iflag &= ~IXOFF;

    tty_r_t_speed((def_rspeed > 0 ? def_rspeed : 9600),
		  (def_tspeed > 0 ? def_tspeed : 9600));
    /*
     * Set the tty modes, and make this our controlling tty.
     */
    set_termbuf();
    if (setsid() < 0) fatalperror(2, "setsid()");
    if (ioctl(t, TIOCSCTTY, (char *)0) < 0) {
	fatalperror(2, "ioctl(sctty)");
    }
    if (t != 0) dup2(t, 0);
    if (t != 1) dup2(t, 1);
    if (t != 2) dup2(t, 2);
    if (t > 2) close(t);

    if (pty > 2) close(pty);
}


/*
 * start_login(host)
 *
 * Assuming that we are now running as a child processes, this
 * function will turn us into the login process.
 */


static void __attribute__ ((noreturn)) start_login(const char *host)
{
    const char *argv[6];
    const char *u;

    u = strrchr(loginprg, '/');
    if(loginprg == bb_login_path) {
	if(u != NULL)
		argv[0] = u + 1;
	else
		argv[0] = loginprg;
	/* -h : pass on name of host.
	 *      WARNING:  -h is accepted by login if and only if getuid() == 0.
	 * -p : don't clobber the environment (so terminal type stays set). */
	argv[1] = "-h";
	argv[2] = host;
	argv[3] = "-p";
	u = getenv("USER");

	if (u) {
		argv[4] = u;
		if (*u == '-') {
			write(1,"I don't hear you!\r\n",19);
			bb_error_msg_and_die("Attempt to login with an option!");
		}
		argv[5] = NULL;
	} else {
		argv[4] = NULL;
	}
    } else {
	if(u != NULL && strlen(u + 1) >= 2 && !strcmp("sh", u + 1 + strlen(u + 1) - 2))
		argv[0] = "-sh";
	else
		argv[0] = loginprg;
	argv[1] = NULL;
    }
    closelog();
    execv(loginprg, (char **)argv);

    bb_perror_msg("%s", loginprg);
    fatalperror(2, loginprg);
}


/*
 * Get a pty, scan input lines.
 */

static int
terminaltypeok(char *s)
{
    if (terminaltype == NULL)
	return(1);

#if 0
  {
    char buf[2048];

    /*
     * Fix from Chris Evans: if it has a / in it, termcap will
     * treat it as a filename. Oops.
     */
    if (strchr(s, '/')) {
	return 0;
    }

    /*
     * tgetent() will return 1 if the type is known, and
     * 0 if it is not known.  If it returns -1, it couldn't
     * open the database.  But if we can't open the database,
     * it won't help to say we failed, because we won't be
     * able to verify anything else.  So, we treat -1 like 1.
     */
    if (tgetent(buf, s) == 0)
	return(0);
  }
#endif
    return(1);
}

/*
 * ttloop
 *
 *      A small subroutine to flush the network output buffer, get some data
 * from the network, and pass it through the telnet state machine.  We
 * also flush the pty input buffer (by dropping its data) if it becomes
 * too full.
 */

static void ttloop(void)
{
    if (nfrontp-nbackp) {
	telnetd_netflush();
    }
    ncc = read(0, netibuf, sizeof netibuf);
    if (ncc < 0) {
	syslog(LOG_INFO, "ttloop: read: %m");
	exit(1);
    } else if (ncc == 0) {
	syslog(LOG_INFO, "ttloop: peer died: EOF");
	exit(1);
    }
    netip = netibuf;
    telnetd_telrcv();                   /* state machine */
    if (ncc > 0) {
	pfrontp = pbackp = ptyobuf;
	telnetd_telrcv();
    }
}  /* end of ttloop */


static void
telnetd_gettermname(void)
{
    /*
     * If the client turned off the option,
     * we can't send another request, so we
     * just return.
     */
    if (his_state_is_wont(TELOPT_TTYPE))
	return;
    settimer(baseline);
    memcpy(nfrontp, ttytype_sbbuf, sizeof ttytype_sbbuf);
    nfrontp += sizeof ttytype_sbbuf;
    while (sequenceIs(ttypesubopt, baseline))
	ttloop();
}


/*
 * getterminaltype
 *
 *      Ask the other end to send along its terminal type and speed.
 * Output is the variable terminaltype filled in.
 */
static void
getterminaltype(void)
{
    settimer(baseline);

    telnetd_send_do(TELOPT_TTYPE, 1);
    telnetd_send_do(TELOPT_TSPEED, 1);
    telnetd_send_do(TELOPT_XDISPLOC, 1);
    telnetd_send_do(TELOPT_ENVIRON, 1);
    while (
	   his_will_wont_is_changing(TELOPT_TTYPE) ||
	   his_will_wont_is_changing(TELOPT_TSPEED) ||
	   his_will_wont_is_changing(TELOPT_XDISPLOC) ||
	   his_will_wont_is_changing(TELOPT_ENVIRON)) {
	ttloop();
    }
    if (his_state_is_will(TELOPT_TSPEED)) {
	static const char sbbuf[] = { IAC, SB, TELOPT_TSPEED, TELQUAL_SEND, IAC, SE };

	memcpy(nfrontp, sbbuf, sizeof sbbuf);
	nfrontp += sizeof sbbuf;
    }
    if (his_state_is_will(TELOPT_XDISPLOC)) {
	static const char sbbuf[] = { IAC, SB, TELOPT_XDISPLOC, TELQUAL_SEND, IAC, SE };

	memcpy(nfrontp, sbbuf, sizeof sbbuf);
	nfrontp += sizeof sbbuf;
    }
    if (his_state_is_will(TELOPT_ENVIRON)) {
	static const char sbbuf[] = { IAC, SB, TELOPT_ENVIRON, TELQUAL_SEND, IAC, SE };

	memcpy(nfrontp, sbbuf, sizeof sbbuf);
	nfrontp += sizeof sbbuf;
    }
    if (his_state_is_will(TELOPT_TTYPE)) {

	memcpy(nfrontp, ttytype_sbbuf, sizeof ttytype_sbbuf);
	nfrontp += sizeof ttytype_sbbuf;
    }
    if (his_state_is_will(TELOPT_TSPEED)) {
	while (sequenceIs(tspeedsubopt, baseline))
	    ttloop();
    }
    if (his_state_is_will(TELOPT_XDISPLOC)) {
	while (sequenceIs(xdisplocsubopt, baseline))
	    ttloop();
    }
    if (his_state_is_will(TELOPT_ENVIRON)) {
	while (sequenceIs(environsubopt, baseline))
	    ttloop();
    }
    if (his_state_is_will(TELOPT_TTYPE)) {
	char first[256], last[256];

	while (sequenceIs(ttypesubopt, baseline))
	    ttloop();

	/*
	 * If the other side has already disabled the option, then
	 * we have to just go with what we (might) have already gotten.
	 */
	if (his_state_is_will(TELOPT_TTYPE) && !terminaltypeok(terminaltype)) {
	    /*
	     * Due to state.c, terminaltype points to a static char[41].
	     * Therefore, this assert cannot fail, and therefore, strings
	     * arising from "terminaltype" can be safely strcpy'd into
	     * first[] or last[].
	     */
	    assert(strlen(terminaltype) < sizeof(first));

	    strcpy(first, terminaltype);

	    for(;;) {
		/*
		 * Save the unknown name, and request the next name.
		 */
		strcpy(last, terminaltype);

		telnetd_gettermname();
		assert(strlen(terminaltype) < sizeof(first));

		if (terminaltypeok(terminaltype))
		    break;

		if (!strcmp(last, terminaltype) ||
		    his_state_is_wont(TELOPT_TTYPE)) {
		    /*
		     * We've hit the end.  If this is the same as
		     * the first name, just go with it.
		     */
		    if (!strcmp(first, terminaltype))
			break;
		    /*
		     * Get the terminal name one more time, so that
		     * RFC1091 compliant telnets will cycle back to
		     * the start of the list.
		     */
		    telnetd_gettermname();
		    assert(strlen(terminaltype) < sizeof(first));

		    if (strcmp(first, terminaltype)) {
			/*
			 * first[] came from terminaltype, so it must fit
			 * back in.
			 */
			strcpy(terminaltype, first);
		    }
		    break;
		}
	    }
	}
    }
}  /* end of getterminaltype */


static void
putchr(int cc)
{
	*putlocation++ = cc;
}

static void
putstr(const char *s)
{
    while (*s) putchr(*s++);
}


static void
edithost(const char *pat, const char *host)
{
	char *res = editedhost;

	uname(&kerninfo);

	if (!pat)
		pat = "";
	while (*pat) {
		switch (*pat) {

		case '#':
			if (*host)
				host++;
			break;

		case '@':
			if (*host)
				*res++ = *host++;
			break;

		default:
			*res++ = *pat;
			break;
		}
		if (res == &editedhost[sizeof editedhost - 1]) {
			*res = '\0';
			return;
		}
		pat++;
	}
	if (*host)
		(void) strncpy(res, host,
				sizeof editedhost - (res - editedhost) -1);
	else
		*res = '\0';
	editedhost[sizeof editedhost - 1] = '\0';
}

static const char fmtstr[] = { "%H:%M on %A, %d %B %Y" };

static void putf(const char *cp, char *where)
{
	time_t t;
	char db[100];

	if (where)
		putlocation = where;

	while (*cp) {
		if (*cp != '%') {
			putchr(*cp++);
			continue;
		}
		switch (*++cp) {

		case 't':
			putstr(bb_basename(Tline));
			break;

		case 'h':
			putstr(editedhost);
			break;

		case 'd':
			(void)time(&t);
			(void)strftime(db, sizeof(db), fmtstr, localtime(&t));
			putstr(db);
			break;

		case '%':
			putchr('%');
			break;
#ifndef __BIONIC__
		case 'D':
			{
				char    buff[128];

				if (getdomainname(buff,sizeof(buff)) < 0
					|| buff[0] == '\0'
					|| strcmp(buff, "(none)") == 0)
					break;
				putstr(buff);
			}
			break;
#endif
		case 'i':
			{
				char buff[3];
				FILE *fp;
				int p, c;

				if ((fp = fopen(ISSUE_FILE, "r")) == NULL)
					break;
				p = '\n';
				while ((c = fgetc(fp)) != EOF) {
					if (p == '\n' && c == '#') {
						do {
							c = fgetc(fp);
						} while (c != EOF && c != '\n');
						continue;
					} else if (c == '%') {
						buff[0] = c;
						c = fgetc(fp);
						if (c == EOF) break;
						buff[1] = c;
						buff[2] = '\0';
						putf(buff, NULL);
					} else {
						if (c == '\n') putchr('\r');
						putchr(c);
						p = c;
					}
				};
				(void) fclose(fp);
			}
			return; /* ignore remainder of the banner string */
			/*NOTREACHED*/

		case 's':
			putstr(kerninfo.sysname);
			break;

		case 'm':
			putstr(kerninfo.machine);
			break;

		case 'r':
			putstr(kerninfo.release);
			break;

		case 'v':
			putstr(kerninfo.version);
			break;
		}
		cp++;
	}
}

/*
 * Main loop.  Select from pty and network, and
 * hand data to telnet receiver finite state machine.
 */
static void __attribute__ ((noreturn)) telnet(void)
{
    int on = 1;
    char *HE;
    const char *IM;
    int p = pty;

    /*
     * Initialize the slc mapping table.
     */
    get_slc_defaults();

    /*
     * Do some tests where it is desireable to wait for a response.
     * Rather than doing them slowly, one at a time, do them all
     * at once.
     */
    if (my_state_is_wont(TELOPT_SGA))
	telnetd_send_will(TELOPT_SGA, 1);
    /*
     * Is the client side a 4.2 (NOT 4.3) system?  We need to know this
     * because 4.2 clients are unable to deal with TCP urgent data.
     *
     * To find out, we send out a "DO ECHO".  If the remote system
     * answers "WILL ECHO" it is probably a 4.2 client, and we note
     * that fact ("WILL ECHO" ==> that the client will echo what
     * WE, the server, sends it; it does NOT mean that the client will
     * echo the terminal input).
     */
    telnetd_send_do(TELOPT_ECHO, 1);

    /*
     * Send along a couple of other options that we wish to negotiate.
     */
    telnetd_send_do(TELOPT_NAWS, 1);
    telnetd_send_will(TELOPT_STATUS, 1);
    flowmode = 1;  /* default flow control state */
    telnetd_send_do(TELOPT_LFLOW, 1);

    /*
     * Spin, waiting for a response from the DO ECHO.  However,
     * some REALLY DUMB telnets out there might not respond
     * to the DO ECHO.  So, we spin looking for NAWS, (most dumb
     * telnets so far seem to respond with WONT for a DO that
     * they don't understand...) because by the time we get the
     * response, it will already have processed the DO ECHO.
     * Kludge upon kludge.
     */
    while (his_will_wont_is_changing(TELOPT_NAWS)) {
	ttloop();
    }

    /*
     * But...
     * The client might have sent a WILL NAWS as part of its
     * startup code; if so, we'll be here before we get the
     * response to the DO ECHO.  We'll make the assumption
     * that any implementation that understands about NAWS
     * is a modern enough implementation that it will respond
     * to our DO ECHO request; hence we'll do another spin
     * waiting for the ECHO option to settle down, which is
     * what we wanted to do in the first place...
     */
    if (his_want_state_is_will(TELOPT_ECHO) &&
	his_state_is_will(TELOPT_NAWS)) {
	while (his_will_wont_is_changing(TELOPT_ECHO))
	    ttloop();
    }
    /*
     * On the off chance that the telnet client is broken and does not
     * respond to the DO ECHO we sent, (after all, we did send the
     * DO NAWS negotiation after the DO ECHO, and we won't get here
     * until a response to the DO NAWS comes back) simulate the
     * receipt of a will echo.  This will also send a WONT ECHO
     * to the client, since we assume that the client failed to
     * respond because it believes that it is already in DO ECHO
     * mode, which we do not want.
     */
    if (his_want_state_is_will(TELOPT_ECHO)) {
	telnetd_willoption(TELOPT_ECHO);
    }

    /*
     * Finally, to clean things up, we turn on our echo.  This
     * will break stupid 4.2 telnets out of local terminal echo.
     */

    if (my_state_is_wont(TELOPT_ECHO))
	telnetd_send_will(TELOPT_ECHO, 1);

    /*
     * Turn on packet mode
     */
    ioctl(p, TIOCPKT, (char *)&on);
    /*
     * Call telrcv() once to pick up anything received during
     * terminal type negotiation, 4.2/4.3 determination, and
     * linemode negotiation.
     */
    telnetd_telrcv();

    ioctl(0, FIONBIO, (char *)&on);
    ioctl(p, FIONBIO, (char *)&on);

#if defined(SO_OOBINLINE)
    setsockopt(0, SOL_SOCKET, SO_OOBINLINE, &on, sizeof on);
#endif  /* defined(SO_OOBINLINE) */

#ifdef  SIGTSTP
    bb_signal((1 << SIGTSTP), SIG_IGN);
#endif
#ifdef  SIGTTOU
    /*
     * Ignoring SIGTTOU keeps the kernel from blocking us
     * in ttioct() in /sys/tty.c.
     */
    bb_signal((1 << SIGTTOU), SIG_IGN);
#endif

    bb_signal((1 << SIGCHLD), cleanup);

#ifdef TIOCNOTTY
    {
	int t;

	t = open(bb_dev_tty, O_RDWR);
	if (t >= 0) {
	    (void) ioctl(t, TIOCNOTTY, (char *)0);
	    (void) close(t);
	}
    }
#endif

    /*
     * Show banner that getty never gave.
     *
     * We put the banner in the pty input buffer.  This way, it
     * gets carriage return null processing, etc., just like all
     * other pty --> client data.
     */

    if (getenv("USER"))
	dis_hostinfo = 1;

    IM = DEFAULT_IM;
    HE = 0;

    edithost(HE, host_name);
    if (dis_hostinfo == 0 && *IM)
	putf(IM, ptyibuf2);

    if (pcc) strncat(ptyibuf2, ptyip, pcc+1);
    ptyip = ptyibuf2;
    pcc = strlen(ptyip);
    for (;;) {
	fd_set ibits, obits, xbits;
	int c, hifd;

	if (ncc < 0 && pcc < 0)
	    break;

	FD_ZERO(&ibits);
	FD_ZERO(&obits);
	FD_ZERO(&xbits);
	/*
	 * Never look for input if there's still
	 * stuff in the corresponding output buffer
	 */
	if (nfrontp - nbackp || pcc > 0) {
	    FD_SET(1, &obits);
	    hifd = 1+1;
	}
	else {
	    FD_SET(p, &ibits);
	    hifd = p+1;
	}
	if (pfrontp - pbackp || ncc > 0) {
	    FD_SET(p, &obits);
	    hifd = p+1;
	}
	else {
	    FD_SET(0, &ibits);
	}
	if (!SYNCHing) {
	    FD_SET(0, &xbits);
	}
	if ((c = select(hifd, &ibits, &obits, &xbits,
			(struct timeval *)0)) < 1) {
	    if (c == -1) {
		if (errno == EINTR) {
		    continue;
		}
	    }
	    sleep(5);
	    continue;
	}

	/*
	 * Any urgent data?
	 */
	if (FD_ISSET(0, &xbits)) {
	    SYNCHing = 1;
	}

	/*
	 * Something to read from the network...
	 */
	if (FD_ISSET(0, &ibits)) {
#if !defined(SO_OOBINLINE)
	    /*
	     * In 4.2 (and 4.3 beta) systems, the
	     * OOB indication and data handling in the kernel
	     * is such that if two separate TCP Urgent requests
	     * come in, one byte of TCP data will be overlaid.
	     * This is fatal for Telnet, but we try to live
	     * with it.
	     *
	     * In addition, in 4.2 (and...), a special protocol
	     * is needed to pick up the TCP Urgent data in
	     * the correct sequence.
	     *
	     * What we do is:  if we think we are in urgent
	     * mode, we look to see if we are "at the mark".
	     * If we are, we do an OOB receive.  If we run
	     * this twice, we will do the OOB receive twice,
	     * but the second will fail, since the second
	     * time we were "at the mark", but there wasn't
	     * any data there (the kernel doesn't reset
	     * "at the mark" until we do a normal read).
	     * Once we've read the OOB data, we go ahead
	     * and do normal reads.
	     *
	     * There is also another problem, which is that
	     * since the OOB byte we read doesn't put us
	     * out of OOB state, and since that byte is most
	     * likely the TELNET DM (data mark), we would
	     * stay in the TELNET SYNCH (SYNCHing) state.
	     * So, clocks to the rescue.  If we've "just"
	     * received a DM, then we test for the
	     * presence of OOB data when the receive OOB
	     * fails (and AFTER we did the normal mode read
	     * to clear "at the mark").
	     */
	    if (SYNCHing) {
		int atmark;

		ioctl(0, SIOCATMARK, (char *)&atmark);
		if (atmark) {
		    ncc = recv(0, netibuf, sizeof (netibuf), MSG_OOB);
		    if ((ncc == -1) && (errno == EINVAL)) {
			ncc = read(0, netibuf, sizeof (netibuf));
			if (sequenceIs(didnetreceive, gotDM)) {
			    SYNCHing = telnetd_stilloob();
			}
		    }
		}
		else {
		    ncc = read(0, netibuf, sizeof (netibuf));
		}
	    }
	    else {
		ncc = read(0, netibuf, sizeof (netibuf));
	    }
	    settimer(didnetreceive);
#else   /* !defined(SO_OOBINLINE)) */
	    ncc = read(0, netibuf, sizeof (netibuf));
#endif  /* !defined(SO_OOBINLINE)) */
	    if (ncc < 0 && errno == EWOULDBLOCK)
		ncc = 0;
	    else {
		if (ncc <= 0) {
		    break;
		}
		netip = netibuf;
	    }
	}

	/*
	 * Something to read from the pty...
	 */
	if (FD_ISSET(p, &ibits)) {
	    pcc = read(p, ptyibuf, BUFSIZ);
	    /*
	     * On some systems, if we try to read something
	     * off the master side before the slave side is
	     * opened, we get EIO.
	     */
	    if (pcc < 0 && (errno == EWOULDBLOCK || errno == EIO)) {
		pcc = 0;
	    }
	    else {
		if (pcc <= 0)
		    break;
		if (ptyibuf[0] & TIOCPKT_FLUSHWRITE) {
		    netclear(); /* clear buffer back */
#ifndef NO_URGENT
		    /*
		     * There are client telnets on some
		     * operating systems get screwed up
		     * royally if we send them urgent
		     * mode data.
		     */
		    *nfrontp++ = IAC;
		    *nfrontp++ = DM;
		    neturg = nfrontp-1; /* off by one XXX */
#endif
		}
		if (his_state_is_will(TELOPT_LFLOW) &&
		    (ptyibuf[0] &
		     (TIOCPKT_NOSTOP|TIOCPKT_DOSTOP))) {
		    (void) sprintf(nfrontp, "%c%c%c%c%c%c",
				   IAC, SB, TELOPT_LFLOW,
				   ptyibuf[0] & TIOCPKT_DOSTOP ? 1 : 0,
				   IAC, SE);
		    nfrontp += 6;
		}
		pcc--;
		ptyip = ptyibuf+1;
	    }
	}

	while (pcc > 0) {
	    if ((&netobuf[BUFSIZ] - nfrontp) < 2)
		break;
	    c = *ptyip++ & 0377, pcc--;
	    if (c == IAC)
		*nfrontp++ = c;
	    *nfrontp++ = c;
	    if ((c == '\r'  ) && (my_state_is_wont(TELOPT_BINARY))) {
		if (pcc > 0 && ((*ptyip & 0377) == '\n')) {
		    *nfrontp++ = *ptyip++ & 0377;
		    pcc--;
		}
		else *nfrontp++ = '\0';
	    }
	}

	if (FD_ISSET(1, &obits) && (nfrontp - nbackp) > 0)
	    telnetd_netflush();
	if (ncc > 0)
	    telnetd_telrcv();
	if (FD_ISSET(p, &obits) && (pfrontp - pbackp) > 0)
	    ptyflush();
    }
    cleanup(0);
}  /* end of telnet */


int
telnetd_main(void)
{
#if     defined(IPPROTO_IP) && defined(IP_TOS)
	int tos = -1;
#endif
	struct sockaddr from;
	const char *host = NULL;
	char hbuf[NI_MAXHOST];
	char *p;
	int i;

	Tline = linedata;
	def_tspeed = -1;
	def_rspeed = -1;
	ptyip = ptyibuf;

	not42 = 1;
	subpointer = subbuffer;
	subend = subbuffer;

	bb_openlog(LOG_PID | LOG_ODELAY, LOG_DAEMON);
	log_syslog = LOG_ERR;
	log_stderr = 0;

	{
		socklen_t fromlen = sizeof (from);

		if (getpeername(0, &from, &fromlen) < 0)
			bb_perror_msg_and_die("getpeername");
	}

	i = bb_xgetopt_ulflags("hnL:\0?0", &loginprg);
	if(loginprg != NULL)
		loginprg = xstrdup(loginprg);   /* need for bb_setproctitle */
	else
		loginprg = bb_login_path;
	dis_hostinfo = i & 1;
	if ((i & 2) == 0) {
	    int on = 1;

	    if(setsockopt(0, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof (on)) < 0)
		xlog_warn("setsockopt (SO_KEEPALIVE): %m");
	}

#if     defined(IPPROTO_IP) && defined(IP_TOS)
	if (tos < 0)
		tos = 020;      /* Low Delay bit */
	if (tos && (setsockopt(0, IPPROTO_IP, IP_TOS, &tos, sizeof(tos)) < 0)
		   && (errno != ENOPROTOOPT) )
		xlog_warn("setsockopt (IP_TOS): %m");
#endif  /* defined(IPPROTO_IP) && defined(IP_TOS) */

#ifndef __ANDROID_API__
	p = getenv("TZ");
	if(p) {
		p = strdup(p-3);
		envinit[0] = p;
	}

	clearenv();
	environ = envinit;
#endif

	/* Find an available pty to use. */
	if (openpty(&pty, &ptyslavefd, Tline, NULL, NULL) < 0)
		fatalperror(2, "All network ports in use");

	/* Get local host name */
	gethostname(host_name, sizeof(host_name));

	p = NULL;
	bb_gethostbyname(BB_GHBN_SILENT | BB_GHBN_CANON_NAME, host_name, &p);
	if(p != NULL) {
		safe_strncpy(host_name, p, sizeof(host_name));
		free(p);
	}

	/* get name of connected client */
	host = bb_xgethostbyaddr(&(((struct sockaddr_in *)&from)->sin_addr), hbuf,
#if ENABLE_FEATURE_IPV6
		from.sa_family == AF_INET ?
			BB_XGETHOSTBYADDR_FLG_IN_ADDR | BB_XGETHOSTBYADDR_FLG_NTOP :
			BB_XGETHOSTBYADDR_FLG_IN6_ADDR | BB_XGETHOSTBYADDR_FLG_NTOP
#else
		BB_XGETHOSTBYADDR_FLG_IN_ADDR | BB_XGETHOSTBYADDR_FLG_NTOP
#endif
		);

	/* Set REMOTEHOST to proctitle and environment variable */
	bb_setproctitle("%s: %s", bb_applet_name, host);

	setenv("REMOTEHOST", host, 0);

	/* get terminal type. */
	pfrontp = pbackp = ptyobuf;
	netip = netibuf;
	nfrontp = nbackp = netobuf;

	getterminaltype();
	bb_setenv1("TERM", terminaltype ? terminaltype : "network");

	/*
	 * Start up the login process on the slave side of the terminal
	 */

	i = fork();
	if (i < 0) fatalperror(2, "fork");
	if (i) {
		/* parent */
		bb_signal((1 << SIGHUP), SIG_IGN);
		close(ptyslavefd);
	}
	else {
		/* child */
		bb_signal((1 << SIGHUP), SIG_IGN);
		getptyslave();
		start_login(host);
		/*NOTREACHED*/
	}

	telnet();       /* begin server processing */
	/*NOTREACHED*/
	return 0;
}  /* end of main */

/*
 * Copyright (c) 1989 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#endif /* APPLET_USAGE */
