/* tftp common header */

/* Standard includes */

#include <string.h>
#include <strings.h>
#include <inttypes.h>
#include <unistd.h>
#include <setjmp.h>
#include <sys/time.h>
#include <time.h>
#include <fcntl.h>
#include <signal.h>

#include "bb_inet.h"

#ifndef HAVE_ARPA_TFTP
#include "tftp.h"
#else
#include <arpa/tftp.h>
#endif

/* Test for EAGAIN/EWOULDBLOCK */
#ifdef EAGAIN
#if defined(EWOULDBLOCK) && (EWOULDBLOCK != EAGAIN)
#define E_WOULD_BLOCK(x) ((x) == EAGAIN || (x) == EWOULDBLOCK)
#else
#define E_WOULD_BLOCK(x) ((x) == EAGAIN)
#endif
#else
#define E_WOULD_BLOCK(x) ((x) == EWOULDBLOCK)
#endif

#ifndef O_TEXT
#define O_TEXT          0
#endif
#ifndef O_BINARY
#define O_BINARY        0
#endif

/* sysexits.h */

#ifndef __BIONIC__
/* Define to 1 if you have the <sysexits.h> header file. */
#define HAVE_SYSEXITS_H 1
#endif
#ifdef HAVE_SYSEXITS_H
#include <sysexits.h>
#else
#define EX_USAGE        64      /* command line usage error */
#define EX_DATAERR      65      /* data format error */
#define EX_NOINPUT      66      /* cannot open input */
#define EX_NOUSER       67      /* addressee unknown */
#define EX_NOHOST       68      /* host name unknown */
#define EX_UNAVAILABLE  69      /* service unavailable */
#define EX_SOFTWARE     70      /* internal software error */
#define EX_OSERR        71      /* system error (e.g., can't fork) */
#define EX_OSFILE       72      /* critical OS file missing */
#define EX_CANTCREAT    73      /* can't create (user) output file */
#define EX_IOERR        74      /* input/output error */
#define EX_TEMPFAIL     75      /* temp failure; user is invited to retry */
#define EX_PROTOCOL     76      /* remote error in protocol */
#define EX_NOPERM       77      /* permission denied */
#define EX_CONFIG       78      /* configuration error */
#endif

/* netinet/in.h, and possible missing pieces */

#include <netinet/in.h>
#ifndef IPPORT_TFTP
#define IPPORT_TFTP 69
#endif

/* arpa/{inet,tftp}.h, and possible missing pieces */

#include <arpa/inet.h>

#ifndef OACK
#define OACK 6
#endif
#ifndef EOPTNEG
#define EOPTNEG 8
#endif

/* Prototypes for libxtra functions */

#ifndef HAVE_BSD_SIGNAL
void (*bsd_signal(int, void (*)(int)))(int);
#endif

/* tftp-hpa version and configuration strings */

#define VERSION "tftp-hpa 0.48"

struct tftphdr;

extern struct tftphdr *tftp_r_init(void);
extern struct tftphdr *tftp_w_init(void);

extern void    tftp_read_ahead(FILE *, int);
extern int     tftp_readit(FILE *, struct tftphdr **, int);

extern int     tftp_synchnet(int);

extern int     tftp_write_behind(FILE *, int);
extern int     tftp_writeit(FILE *, struct tftphdr **, int, int);

extern int tftp_pick_port_bind(int sockfd, struct sockaddr_in *myaddr,
					unsigned int from, unsigned int to);

#define MAX_SEGSIZE     65464
#define PKTSIZE MAX_SEGSIZE+4       /* should be moved to tftp.h */

struct bf {
	int counter;            /* size of data in buffer, or flag */
	char buf[PKTSIZE];      /* room for data packet */
};
