/*
    Copyright (c)  2000-2007           Dmitry Butskoy
					<buc@citadel.stu.neva.ru>
    mod-dccp.c:
    Copyright (c)  2012         Samuel Jero <sj323707@ohio.edu>

 * Busybox port by Vladimir Oleynik (C) 2020 <dzo@simtreas.ru>
   * rewrote from CLIF_lib to getopt

    License:  GPL v2 or any later

*/

#ifdef APPLET_USAGE

#define traceroute_trivial_usage \
	"[-"\
	USE_FEATURE_IPV6("46") \
	USE_FEATURE_TRACEROUTE_DEBUG("d") \
	USE_FEATURE_TRACEROUTE_USE_ICMP("I") \
	USE_FEATURE_TRACEROUTE_USE_TCP("T") \
	USE_FEATURE_TRACEROUTE_USE_UDP("U") \
	USE_FEATURE_TRACEROUTE_USE_UDPL("u") \
	USE_FEATURE_TRACEROUTE_USE_DCCP("D") \
	USE_FEATURE_TRACEROUTE_SHOW_ICMP_EXT("e") \
	"FnrV] [-f 1st_ttl] [-m max_ttl] [-N squeries] [-p port#]\n" \
	"\t[-q nqueries] " USE_FEATURE_IPV6("[-l flow_label] ") "[-P proto]\n" \
	"\t[-s src_addr] [-t tos] [-w wait] [-g gate,...] [-i iface]\n" \
	"\t[-z pausemsecs] " \
	USE_FEATURE_LONG_OPTIONS("[--sport=port] ") \
	SKIP_FEATURE_LONG_OPTIONS("[-S port] ") \
	USE_FEATURE_LONG_OPTIONS("[--mtu] ") \
	SKIP_FEATURE_LONG_OPTIONS("[-c] ") \
	USE_FEATURE_LONG_OPTIONS("[--back]\n") \
	SKIP_FEATURE_LONG_OPTIONS("[-b]\n") \
	"\thost [data size]"
#define traceroute_full_usage \
	"trace the route ip packets follow going to \"host\"\n" \
	"Options:\n" \
	USE_FEATURE_IPV6("\t-4\tExplicitly force IPv4 tracerouting.\n") \
	USE_FEATURE_IPV6("\t-6\tExplicitly force IPv6 tracerouting.\n") \
	USE_FEATURE_TRACEROUTE_DEBUG("\t-d\tSet SO_DEBUG options to socket\n") \
	USE_FEATURE_TRACEROUTE_USE_ICMP("\t-I\tUse ICMP ECHO instead of UDP datagrams\n") \
	USE_FEATURE_TRACEROUTE_USE_TCP("\t-T\tUse TCP SYN for probes\n") \
	USE_FEATURE_TRACEROUTE_USE_UDP("\t-U\tUse UDP to particular destination port\n" \
	"for tracerouting (instead of increasing the port per each probe).\n" \
	"\t\tDefault port is 53 (dns).\n") \
	USE_FEATURE_TRACEROUTE_USE_UDPL("\t-u\tUse UDPLITE for tracerouting (default port is 53).\n" \
	"Warning, the original programm used the -UL option for this.\n") \
	USE_FEATURE_TRACEROUTE_USE_DCCP("\t-D\tUse DCCP Requests for probes.\n") \
	USE_FEATURE_TRACEROUTE_SHOW_ICMP_EXT("\t-e\tShow  ICMP extensions (rfc4884).\n") \
	"\t-F\tSet the don't fragment bit\n" \
	USE_FEATURE_IPV6("\t-l\tUse specified flow label for IPv6 packets.\n") \
	"\t-n\tPrint hop addresses numerically rather than symbolically\n" \
	"\t-r\tBypass the normal routing tables and send directly to a host\n" \
	"\t-V\tPrint the version and exit\n" \
	"\t-f ttl\tSpecifies with what TTL to start. Defaults to 1\n" \
	"\t-N sq\tSpecifies the number of probe packets sent out simultaneously\n\t\t(default 16).\n" \
	"\t-m mttl\tSet the max time-to-live (max number of hops)\n" \
	"\t-p p#\tSet the base UDP/TCP port number used in probes\n" \
	"\t\t(default is 33434, for UDP will be incremented by each probe)\n\t\tor the initial ICMP sequence value\n" \
	"\t-q nq\tSet the number of probes per ``ttl'' to nqueries. Default is 3\n" \
	"\t-P pr\tUse raw packet of specified protocol for tracerouting. Default 253\n" \
	"\t-s sa\tUse the following IP address as the source address\n" \
	"\t-t tos\tSet the type-of-service in probe packets to the following value\n" \
	"\t\t(default 0)\n" \
	"\t-w wait\tSet the time (in seconds) to wait for a response to a probe\n" \
	"\t\t(default 3 sec)\n" \
	"\t-g\tSpecify a loose source route gateway (8 maximum)\n" \
	"\t-M method\tUse specified method for traceroute operations\n" \
	USE_FEATURE_LONG_OPTIONS("\t--mtu\tDiscover  MTU along the path being traced. Implies -F -N 1\n") \
	SKIP_FEATURE_LONG_OPTIONS("\t-c\tDiscover  MTU along the path being traced. Implies -F -N 1\n") \
	USE_FEATURE_LONG_OPTIONS("\t--back\tPrint the number of backward hops when it seems different\n\t\twith the forward direction\n") \
	SKIP_FEATURE_LONG_OPTIONS("\t-b\tPrint the number of backward hops when it seems different\n\t\twith the forward direction\n") \
	USE_FEATURE_TRACEROUTE_FWMARK(USE_FEATURE_LONG_OPTIONS("\t--fwmark=mark\tSet the firewall mark for outgoing packets\n")) \
	USE_FEATURE_TRACEROUTE_FWMARK(SKIP_FEATURE_LONG_OPTIONS("\t-a mark\tSet the firewall mark for outgoing packets\n")) \
	USE_FEATURE_LONG_OPTIONS("\t--sport=port\tChooses the source port to use. Implies -N 1") \
	SKIP_FEATURE_LONG_OPTIONS("\t-S port\tChooses the source port to use. Implies -N 1") \

#else


#define VERSION 2.1.0

#include "busybox.h"
#include <string.h>
#include <unistd.h>
#include <sys/times.h>
#include <sys/time.h>
#include <poll.h>
#include <math.h>
#include <ctype.h>
#include <locale.h>

#include "bb_xopen.h"
#include "bb_inet.h"

#include <netinet/ip_icmp.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <linux/dccp.h>

#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <sys/utsname.h>
#include <linux/types.h>
#include <linux/errqueue.h>

#if ENABLE_FEATURE_IPV6
#define INET_4_6_ADDRSTRLEN INET6_ADDRSTRLEN
#include <netinet/ip6.h>
#include <netinet/icmp6.h>
#else
#define INET_4_6_ADDRSTRLEN INET_ADDRSTRLEN
#endif

#define COMPLEMENTARY "\0=g:I-UTuD:U-ITuD:T-IUuD:u-IUTD:D-IUTu:V-h:h-V:O::"
#include "opt_gen.h"
OPT_GEN(
opt_simple(4),
opt_opt(ENABLE_FEATURE_IPV6, 6, OPT_6),
opt_opt(ENABLE_FEATURE_TRACEROUTE_DEBUG, d, OPT_DEBUG),
opt(F, OPT_DONTFRAG),
opt_simple(f, arg),
opt_opt(ENABLE_FEATURE_TRACEROUTE_SPEC_GATEWAYS, g, OPT_g, arg),
opt_opt(ENABLE_FEATURE_TRACEROUTE_USE_ICMP, I, OPT_ICMP),
opt_opt(ENABLE_FEATURE_TRACEROUTE_USE_TCP, T, OPT_TCP),
opt_opt(ENABLE_FEATURE_TRACEROUTE_USE_UDP, U, OPT_UDP),
opt_opt(ENABLE_FEATURE_TRACEROUTE_USE_UDPL, u, OPT_UDPL),
opt_opt(ENABLE_FEATURE_TRACEROUTE_USE_DCCP, D, OPT_DCCP),
opt_simple(i, arg),
opt_simple(m, arg),
opt_simple(N, arg),
opt_simple(n),
opt_simple(p, arg),
opt_simple(t, arg),
opt_opt(ENABLE_FEATURE_IPV6, l, OPT_l, arg),
opt_simple(w, arg),
opt_simple(q, arg),
opt_simple(r),
opt_simple(s, arg),
opt_simple(z, arg),
opt_simple(e),
opt_simple(M, arg),
opt_simple(S, arg),
opt_simple(O, arg),
opt_simple(P, arg),
#ifdef SO_MARK
opt_opt(ENABLE_FEATURE_TRACEROUTE_FWMARK, a, OPT_a, arg),
#endif
opt_simple(c),
opt_simple(b),
opt_simple(h),
opt_simple(V),
)
#include "opt_gen.h"

/*
   It is just a stripped copy of the kernel header "linux/in6.h"

   "Flow label" things are still not defined in "netinet/in*.h" headers,
   but we cannot use "linux/in6.h" immediately because it currently
   conflicts with "netinet/in.h" .
*/
#if ENABLE_FEATURE_IPV6
# ifndef _UAPI_LINUX_IN6_H
struct in6_flowlabel_req
{
	struct in6_addr flr_dst;
	__u32   flr_label;
	__u8    flr_action;
	__u8    flr_share;
	__u16   flr_flags;
	__u16   flr_expires;
	__u16   flr_linger;
	__u32   __flr_pad;
	/* Options in format of IPV6_PKTOPTIONS */
};
# endif

#define IPV6_FL_A_GET   0
#define IPV6_FL_A_PUT   1
#define IPV6_FL_A_RENEW 2

#define IPV6_FL_F_CREATE        1
#define IPV6_FL_F_EXCL          2

#define IPV6_FL_S_NONE          0
#define IPV6_FL_S_EXCL          1
#define IPV6_FL_S_PROCESS       2
#define IPV6_FL_S_USER          3
#define IPV6_FL_S_ANY           255

#define IPV6_FLOWINFO_FLOWLABEL         0x000fffff
#define IPV6_FLOWINFO_PRIORITY          0x0ff00000

#define IPV6_FLOWLABEL_MGR      32
#define IPV6_FLOWINFO_SEND      33
#endif

union common_sockaddr {
	struct sockaddr sa;
	struct sockaddr_in sin;
#if ENABLE_FEATURE_IPV6
	struct sockaddr_in6 sin6;
#endif
};
typedef union common_sockaddr sockaddr_any;

struct probe_struct {
	int done;
	int final;
	sockaddr_any res;
	double send_time;
	double recv_time;
	int recv_ttl;
	int sk;
	int seq;
	char *ext;
	char err_str[16];       /*  assume enough   */
};
typedef struct probe_struct probe;


struct tr_module_struct {
	const char *name;
	int (*init) (const sockaddr_any *dest,
				unsigned int port_seq, size_t *packet_len);
	void (*send_probe) (probe *pb, int ttl);
	void (*recv_probe) (int fd, int revents);
	int one_per_time;       /*  no simultaneous probes   */
	size_t header_len;      /*  additional header length (aka for udp)   */
};
typedef struct tr_module_struct tr_module;


#define __TEXT(X)       #X
#define _TEXT(X)        __TEXT(X)

#define DEF_START_PORT  33434   /*  start for traditional udp method   */
#define DEF_UDP_PORT    53      /*  dns   */
#define DEF_TCP_PORT    80      /*  web   */
#define DEF_DCCP_PORT   DEF_START_PORT  /*  is it a good choice?...  */
#define DEF_RAW_PROT    253     /*  for experimentation and testing, rfc3692  */


typedef probe *(*check_reply_t) (int sk, int err, sockaddr_any *from,
							char *buf, size_t len);
#ifndef ICMP6_DST_UNREACH_BEYONDSCOPE
#ifdef ICMP6_DST_UNREACH_NOTNEIGHBOR
#define ICMP6_DST_UNREACH_BEYONDSCOPE ICMP6_DST_UNREACH_NOTNEIGHBOR
#else
#define ICMP6_DST_UNREACH_BEYONDSCOPE 2
#endif
#endif

#ifndef IPV6_RECVHOPLIMIT
#define IPV6_RECVHOPLIMIT IPV6_HOPLIMIT
#endif

#ifndef IP_PMTUDISC_PROBE
#define IP_PMTUDISC_PROBE 3
#endif

#ifndef IPV6_PMTUDISC_PROBE
#define IPV6_PMTUDISC_PROBE 3
#endif

#ifndef AI_IDN
#define AI_IDN  0
#endif

#ifndef NI_IDN
#define NI_IDN  0
#endif


#define MAX_HOPS        255
#define MAX_PROBES      10
#define MAX_GATEWAYS_4  8
#define MAX_GATEWAYS_6  127
#define DEF_HOPS        30
#define DEF_SIM_PROBES  16      /*  including several hops   */
#define DEF_NUM_PROBES  3
#define DEF_WAIT_SECS   5.0
#define DEF_HERE_FACTOR 3
#define DEF_NEAR_FACTOR 10
#ifndef DEF_WAIT_PREC
#define DEF_WAIT_PREC   0.001   /*  +1 ms  to avoid precision issues   */
#endif
#define DEF_SEND_SECS   0
#define DEF_DATA_LEN    40      /*  all but IP header...  */
#define MAX_PACKET_LEN  65000
#ifndef DEF_AF
#define DEF_AF          AF_INET
#endif

#define ttl2hops(X)     (((X) <= 64 ? 65 : ((X) <= 128 ? 129 : 256)) - (X))

struct icmp_ext_header {
#if __BYTE_ORDER == __BIG_ENDIAN
	unsigned int version:4;
	unsigned int reserved:4;
#else
	unsigned int reserved:4;
	unsigned int version:4;
#endif
	uint8_t reserved1;
	uint16_t checksum;
} __attribute__ ((packed));


struct icmp_ext_object {
	uint16_t length;
	uint8_t class;
	uint8_t c_type;
	uint8_t pk_data[0];
};

#define MPLS_CLASS 1
#define MPLS_C_TYPE 1

static const char version_string[] = "Modern traceroute for Linux, "
				"version " _TEXT(VERSION)
				"\nCopyright (c) 2016  Dmitry Butskoy, "
				"  License: GPL v2 or any later";

CALC_BSS struct BB_COMMON_BSS {
#ifdef CONFIG_FEATURE_TRACEROUTE_DEBUG
	int debug;
#endif
	int first_hop;
	unsigned int max_hops;
	unsigned int sim_probes;
	unsigned int probes_per_hop;

#ifdef CONFIG_FEATURE_TRACEROUTE_SPEC_GATEWAYS
	char **gateways;
	int num_gateways;
#endif
	unsigned char *rtbuf;
	size_t rtbuf_len;
#if ENABLE_FEATURE_IPV6
	unsigned int ipv6_rthdr_type;
#endif
	size_t g_header_len;
	size_t data_len;
	int dontfrag;
	unsigned int dst_port_seq;
	unsigned int tos;
#if ENABLE_FEATURE_IPV6
	unsigned int flow_label;
#endif
	int noroute;
	unsigned int fwmark;
	int packet_len;
	double wait_secs;
	double here_factor;
	double near_factor;
	double send_secs;
	int mtudisc;
	int backward;
	sockaddr_any dst_addr;
	char *dst_name;
	unsigned int src_port;
	const char *module;
	const tr_module *ops;
	probe *probes;
	llist_t *O;
	struct pollfd *pfd;
	int num_polls;
	char *device;
	sockaddr_any src_addr;
	sockaddr_any dest_addr;
	int af;
	int raw_sk;
	char *fill_data;
	size_t *length_p;
#if defined(CONFIG_FEATURE_TRACEROUTE_USE_TCP) || defined(CONFIG_FEATURE_TRACEROUTE_USE_DCCP)
	size_t csum_len;
#define TCP_BUBSZ 1024          /*  enough, enough...  */
	uint8_t buf[TCP_BUBSZ];
	unsigned int dest_port;
#endif
#ifdef CONFIG_FEATURE_TRACEROUTE_USE_DCCP
	int last_ttl_dccp;
	struct dccp_hdr *dh;
	struct dccp_hdr_ext *dhe;
#endif
	unsigned int coverage;
	unsigned int num_probes;
/*  There is a bug in the kernel before 2.6.25, which prevents icmp errors
  to be obtained by MSG_ERRQUEUE for ipv6 connected raw sockets.
*/
	int can_connect;
#if defined(CONFIG_FEATURE_TRACEROUTE_USE_TCPCONN) || defined(CONFIG_FEATURE_TRACEROUTE_USE_ICMP)
	int icmp_sk;
#endif
#ifdef CONFIG_FEATURE_TRACEROUTE_USE_ICMP
	uint16_t seq_icmp;
	uint16_t ident;
	int last_ttl_icmp;
#endif
	unsigned int curr_port;
	unsigned int protocol_udp;
	int protocol_raw;
	int last_ttl_raw;
	int seq_raw;
#ifdef CONFIG_FEATURE_TRACEROUTE_USE_TCP
	int last_ttl_tcp;
	struct tcphdr *th;
	int flags;       /*  & 0xff == tcp_flags ...  */
	int sysctl;
	int reuse;
	unsigned int mss;
	int tcp_info;
#endif
} bb_common_bss;
#define debug bb_common_bss.debug
#define first_hop bb_common_bss.first_hop
#define max_hops bb_common_bss.max_hops
#define sim_probes bb_common_bss.sim_probes
#define probes_per_hop bb_common_bss.probes_per_hop
#define gateways bb_common_bss.gateways
#define num_gateways bb_common_bss.num_gateways
#define rtbuf bb_common_bss.rtbuf
#define rtbuf_len bb_common_bss.rtbuf_len
#define ipv6_rthdr_type bb_common_bss.ipv6_rthdr_type
#define g_header_len bb_common_bss.g_header_len
#define data_len bb_common_bss.data_len
#define dontfrag bb_common_bss.dontfrag
#define dst_port_seq bb_common_bss.dst_port_seq
#define tos bb_common_bss.tos
#define flow_label bb_common_bss.flow_label
#define noroute bb_common_bss.noroute
#define fwmark bb_common_bss.fwmark
#define packet_len bb_common_bss.packet_len
#define wait_secs bb_common_bss.wait_secs
#define here_factor bb_common_bss.here_factor
#define near_factor bb_common_bss.near_factor
#define send_secs bb_common_bss.send_secs
#define mtudisc bb_common_bss.mtudisc
#define backward bb_common_bss.backward
#define dst_addr bb_common_bss.dst_addr
#define dst_name bb_common_bss.dst_name
#define src_port bb_common_bss.src_port
#define module bb_common_bss.module
#define ops bb_common_bss.ops
#define probes bb_common_bss.probes
#define O bb_common_bss.O
#define pfd bb_common_bss.pfd
#define num_polls bb_common_bss.num_polls
#define device bb_common_bss.device
#define src_addr bb_common_bss.src_addr
#define dest_addr bb_common_bss.dest_addr
#define af bb_common_bss.af
#define raw_sk bb_common_bss.raw_sk
#define fill_data bb_common_bss.fill_data
#define length_p bb_common_bss.length_p
#define dest_port bb_common_bss.dest_port
#define coverage bb_common_bss.coverage
#define num_probes bb_common_bss.num_probes
#define curr_port bb_common_bss.curr_port
#define protocol_udp bb_common_bss.protocol_udp
#define dh bb_common_bss.dh
#define dhe bb_common_bss.dhe
#define mss bb_common_bss.mss
#define csum_len bb_common_bss.csum_len
#define buf bb_common_bss.buf
#define last_ttl_dccp bb_common_bss.last_ttl_dccp
#define can_connect bb_common_bss.can_connect
#define icmp_sk bb_common_bss.icmp_sk
#define last_ttl_icmp bb_common_bss.last_ttl_icmp
#define protocol_raw bb_common_bss.protocol_raw
#define last_ttl_raw bb_common_bss.last_ttl_raw
#define seq_raw bb_common_bss.seq_raw
#define last_ttl_tcp bb_common_bss.last_ttl_tcp
#define flags bb_common_bss.flags
#define sysctl bb_common_bss.sysctl
#define reuse bb_common_bss.reuse
#define tcp_info bb_common_bss.tcp_info
#define seq_icmp bb_common_bss.seq_icmp
#define ident bb_common_bss.ident
#define th bb_common_bss.th


static void ex_error(const char *s, ...) __attribute__ ((noreturn, format (printf, 1, 2)));
static void ex_error (const char *format, ...) {
	va_list ap;

	va_start (ap, format);
	bb_verror_msg(format, ap);
	va_end (ap);

	exit (2);
}

static void set_send_secs (char *arg)
{
	char *q;

	if (arg == NULL)  return;
	send_secs = strtod (arg, &q);
	if (q == arg || *q || send_secs < 0)
		ex_error ("bad sendtime `%s' specified", arg);
}

#define do_snprintf(CURR, END, FMT, ARGS...)    \
	do {                                    \
	    CURR += snprintf (CURR, END - CURR, (FMT), ## ARGS);\
	    if (CURR > END)  CURR = END;                        \
	} while (0)


static int try_extension (probe *pb, char *buf_try_extension, size_t len) {
	struct icmp_ext_header *iext = (struct icmp_ext_header *) buf_try_extension;
	char str[1024];
	char *curr_ext = str;
	char *end = str + sizeof (str) / sizeof (*str);

	/*  a check for len >= 8 already done for all cases   */

	if (iext->version != 2)  return -1;

	if (iext->checksum &&
	    in_cksum (iext, len, 0) != (uint16_t) ~0
	)  return -1;

	buf_try_extension += sizeof (*iext);
	len -= sizeof (*iext);

	while (len >= sizeof (struct icmp_ext_object)) {
	    struct icmp_ext_object *obj = (struct icmp_ext_object *) buf_try_extension;
	    size_t objlen = ntohs (obj->length);
	    size_t data_len_ext;
	    uint32_t *ui = (uint32_t *) obj->pk_data;
	    int i, n;

	    if (objlen < sizeof (*obj) ||
		objlen > len
	    )  return -1;

	    data_len_ext = objlen - sizeof (*obj);
	    if (data_len_ext % sizeof (uint32_t))
		    return -1;  /*  must be 32bit rounded...  */

	    n = data_len_ext / sizeof (*ui);


	    if (curr_ext > (char *) str && curr_ext < end)
		    *curr_ext++ = ';';      /*  a separator   */

	    if (obj->class == MPLS_CLASS &&
		obj->c_type == MPLS_C_TYPE &&
		n >= 1
	    ) {    /*  people prefer MPLS to be parsed...  */

		do_snprintf (curr_ext, end, "MPLS:");

		for (i = 0; i < n; i++, ui++) {
		    uint32_t mpls = ntohl (*ui);

		    do_snprintf (curr_ext, end, "%sL=%u,E=%u,S=%u,T=%u",
					i ? "/" : "",
					mpls >> 12,
					(mpls >> 9) & 0x7,
					(mpls >> 8) & 0x1,
					mpls & 0xff);
		}

	    }
	    else {      /*  common case...  */

		do_snprintf (curr_ext, end, "%u/%u:", obj->class, obj->c_type);

		for (i = 0; i < n && curr_ext < end; i++, ui++)
		    do_snprintf (curr_ext, end, "%s%08x", i ? "," : "", ntohl(*ui));
	    }

	    buf_try_extension += objlen;
	    len -= objlen;
	}

	if (len)  return -1;

	pb->ext = strdup (str);

	return 0;
}

static void handle_extensions (probe *pb, char *buf_handle_extension, int len, int step) {

	if (!step)
	    try_extension (pb, buf_handle_extension, len);
	else {
	    for ( ; len >= 8; buf_handle_extension += step, len -= step)
		if (try_extension (pb, buf_handle_extension, len) == 0)
			break;
	}
}

#if defined(CONFIG_FEATURE_TRACEROUTE_USE_TCP) || defined(CONFIG_FEATURE_TRACEROUTE_USE_DCCP)
static void init_random_seq (void) {

	srand (times (NULL) + getpid ());
}

static unsigned int random_seq (void) {

	/*  To not worry about RANDOM_MAX and precision...  */
	return  (rand () << 16) ^ (rand () << 8) ^ rand () ^ (rand () >> 8);
}
#endif

/*  Just returns current time as double, with most possible precision...  */

static double get_time (void) {
	struct timeval tv;
	double d;

	gettimeofday (&tv, NULL);

	d = ((double) tv.tv_usec) / 1000000. + (unsigned long) tv.tv_sec;

	return d;
}

static void error_or_perm (const char *str) __attribute__((noreturn));

static void error_or_perm (const char *str) {

	if (errno == EPERM)
		fprintf (stderr, "You do not have enough privileges to use "
				"this traceroute method. ");
	bb_perror_msg_and_die ("%s", str);
}

static void add_poll (int fd, int events) {
	int i;

	for (i = 0; i < num_polls && pfd[i].fd > 0; i++) ;

	if (i == num_polls)
	    pfd = xrealloc (pfd, ++num_polls * sizeof (*pfd));

	pfd[i].fd = fd;
	pfd[i].events = events;
}

static void del_poll (int fd) {
	int i;

	for (i = 0; i < num_polls && pfd[i].fd != fd; i++) ;

	if (i < num_polls)  pfd[i].fd = -1;    /*  or just zero it...  */
}

static int cleanup_polls (void) {
	int i;

	for (i = 0; i < num_polls && pfd[i].fd > 0; i++) ;

	if (i < num_polls) {    /*  a hole have found   */
	    int j;

	    for (j = i + 1; j < num_polls; j++) {
		if (pfd[j].fd > 0) {
		    pfd[i++] = pfd[j];
		    pfd[j].fd = -1;
		}
	    }
	}

	return i;
}

static void do_poll (double timeout, void (*callback) (int fd, int revents)) {
	int nfds;
	int msecs = ceil (timeout * 1000);

	while ((nfds = cleanup_polls ()) > 0) {
	    int i, n;

	    n = poll (pfd, nfds, msecs);

	    if (n <= 0) {
		if (n == 0 || errno == EINTR)
			return;
		bb_perror_msg_and_die ("poll");
	    }

	    for (i = 0; n && i < num_polls; i++) {
		if (pfd[i].revents) {
		    callback (pfd[i].fd, pfd[i].revents);
		    n--;
		}
	    }

	    msecs = 0;      /*  no more wait, just eat all the pending   */
	}
}

static void bind_socket (int sk) {
	sockaddr_any *addr, tmp;

	if (device) {
	    if (setsockopt (sk, SOL_SOCKET, SO_BINDTODEVICE,
					device, strlen (device) + 1) < 0
	    )  bb_perror_msg_and_die ("setsockopt SO_BINDTODEVICE");
	}

	if (!src_addr.sa.sa_family) {
	    memset (&tmp, 0, sizeof (tmp));
	    tmp.sa.sa_family = af;
	    addr = &tmp;
	} else
	    addr = &src_addr;

	bb_xbind (sk, &addr->sa, sizeof (*addr));
}

static int equal_addr (const sockaddr_any *a, const sockaddr_any *b) {

	if (!a->sa.sa_family)
		return 0;

	if (a->sa.sa_family != b->sa.sa_family)
		return 0;

#if ENABLE_FEATURE_IPV6
	if (a->sa.sa_family == AF_INET6)
	    return  !memcmp (&a->sin6.sin6_addr, &b->sin6.sin6_addr,
						sizeof (a->sin6.sin6_addr));
	else
#endif
	    return  !memcmp (&a->sin.sin_addr, &b->sin.sin_addr,
						sizeof (a->sin.sin_addr));
}


#define DCCP_HEADER_LEN         (sizeof (struct dccp_hdr) + \
				 sizeof (struct dccp_hdr_ext) \
				 + sizeof (struct dccp_hdr_request))


static probe *probe_by_seq (int seq) {
	unsigned int n;

	if (seq <= 0)  return NULL;

	for (n = 0; n < num_probes; n++) {
	    if (probes[n].seq == seq)
		    return &probes[n];
	}

	return NULL;
}

static probe *probe_by_sk (int sk) {
	unsigned int n;

	if (sk <= 0)  return NULL;

	for (n = 0; n < num_probes; n++) {
	    if (probes[n].sk == sk)
		    return &probes[n];
	}

	return NULL;
}

static void probe_done (probe *pb) {

	if (pb->sk) {
	    del_poll (pb->sk);
	    close (pb->sk);
	    pb->sk = 0;
	}

	pb->seq = 0;

	pb->done = 1;
}

static void use_recv_ttl (int sk) {
	if (af == AF_INET)
		setsockopt_1i (sk, SOL_IP, IP_RECVTTL, 1);
#if ENABLE_FEATURE_IPV6
	else if (af == AF_INET6)
		setsockopt_1i (sk, SOL_IPV6, IPV6_RECVHOPLIMIT, 1);
#endif
	/*  foo on errors   */
}


#define VER(A,B,C,D)    (((((((A) << 8) | (B)) << 8) | (C)) << 8) | (D))

static int raw_can_connect (void) {

	if (can_connect < 0) {

	    if (af == AF_INET)
		    can_connect = 1;
#if ENABLE_FEATURE_IPV6
	    else {      /*  AF_INET6   */
		struct utsname uts;
		int n;
		unsigned int a, b, c, d = 0;

		if (uname (&uts) < 0)
			return 0;

		n = sscanf (uts.release, "%u.%u.%u.%u", &a, &b, &c, &d);
		can_connect = (n >= 3 && VER (a, b, c, d) >= VER (2, 6, 25, 0));
	    }
#endif
	}

	return can_connect;
}

static void use_timestamp (int sk) {
	setsockopt_1i (sk, SOL_SOCKET, SO_TIMESTAMP, 1);
	/*  foo on errors...  */
}

static void tune_socket (int sk) {
	int i = 0;

#ifdef CONFIG_FEATURE_TRACEROUTE_DEBUG
	if (debug) {
	    if (setsockopt_1i (sk, SOL_SOCKET, SO_DEBUG, 1) < 0)
		    bb_perror_msg_and_die ("setsockopt SO_DEBUG");
	}
#endif

#ifdef SO_MARK
	if (fwmark) {
	    if (setsockopt (sk, SOL_SOCKET, SO_MARK,
					&fwmark, sizeof (fwmark)) < 0
	    )  bb_perror_msg_and_die ("setsockopt SO_MARK");
	}
#endif

	if (rtbuf && rtbuf_len) {
	    if (af == AF_INET) {
		if (setsockopt (sk, IPPROTO_IP, IP_OPTIONS,
						rtbuf, rtbuf_len) < 0
		)  bb_perror_msg_and_die ("setsockopt IP_OPTIONS");
	    }
#if ENABLE_FEATURE_IPV6
	    else if (af == AF_INET6) {
		if (setsockopt (sk, IPPROTO_IPV6, IPV6_RTHDR,
						rtbuf, rtbuf_len) < 0
		)  bb_perror_msg_and_die ("setsockopt IPV6_RTHDR");
	    }
#endif
	}

	bind_socket (sk);

	if (af == AF_INET) {

	    i = dontfrag ? IP_PMTUDISC_PROBE : IP_PMTUDISC_DONT;
	    if (setsockopt_1i (sk, SOL_IP, IP_MTU_DISCOVER, i) < 0 &&
		(!dontfrag || (setsockopt_1i (sk, SOL_IP, IP_MTU_DISCOVER, IP_PMTUDISC_DO) < 0))
	    )  bb_perror_msg_and_die ("setsockopt IP_MTU_DISCOVER");

	    if (tos) {
		if (setsockopt_1i (sk, SOL_IP, IP_TOS, tos) < 0)
			bb_perror_msg_and_die ("setsockopt IP_TOS");
	    }

	}
#if ENABLE_FEATURE_IPV6
	else if (af == AF_INET6) {

	    i = dontfrag ? IPV6_PMTUDISC_PROBE : IPV6_PMTUDISC_DONT;
	    if (setsockopt_1i (sk, SOL_IPV6, IPV6_MTU_DISCOVER,i) < 0 &&
		(!dontfrag || (setsockopt_1i (sk, SOL_IPV6, IPV6_MTU_DISCOVER,IPV6_PMTUDISC_DO) < 0))
	    )  bb_perror_msg_and_die ("setsockopt IPV6_MTU_DISCOVER");


	    if (flow_label) {
		struct in6_flowlabel_req flr;

		memset (&flr, 0, sizeof (flr));
		flr.flr_label = htonl (flow_label & 0x000fffff);
		flr.flr_action = IPV6_FL_A_GET;
		flr.flr_flags = IPV6_FL_F_CREATE;
		flr.flr_share = IPV6_FL_S_ANY;
		memcpy (&flr.flr_dst, &dst_addr.sin6.sin6_addr,
						    sizeof (flr.flr_dst));

		if (setsockopt (sk, IPPROTO_IPV6, IPV6_FLOWLABEL_MGR,
						    &flr, sizeof (flr)) < 0
		)  bb_perror_msg_and_die ("setsockopt IPV6_FLOWLABEL_MGR");
	    }

	    if (tos) {
		if (setsockopt_1i (sk, IPPROTO_IPV6, IPV6_TCLASS, tos) < 0
		)  bb_perror_msg_and_die ("setsockopt IPV6_TCLASS");
	    }

	    if (tos || flow_label) {
		if (setsockopt_1i (sk, IPPROTO_IPV6, IPV6_FLOWINFO_SEND, 1) < 0
		)  bb_perror_msg_and_die ("setsockopt IPV6_FLOWINFO_SEND");
	    }
	}
#endif
	if (noroute) {
	    if (setsockopt_1i (sk, SOL_SOCKET, SO_DONTROUTE, noroute) < 0)
		    bb_perror_msg_and_die ("setsockopt SO_DONTROUTE");
	}

	use_timestamp (sk);

	use_recv_ttl (sk);

	fcntl (sk, F_SETFL, O_NONBLOCK);
}

static void set_ttl (int sk, int ttl) {

	if (af == AF_INET) {
	    if (setsockopt_1i (sk, SOL_IP, IP_TTL, ttl) < 0)
		    bb_perror_msg_and_die ("setsockopt IP_TTL");
	}
#if ENABLE_FEATURE_IPV6
	else if (af == AF_INET6) {
	    if (setsockopt_1i (sk, SOL_IPV6, IPV6_UNICAST_HOPS, ttl) < 0
	    )  bb_perror_msg_and_die ("setsockopt IPV6_UNICAST_HOPS");
	}
#endif
}

static void use_recverr (int sk) {
	if (af == AF_INET) {
	    if (setsockopt_1i (sk, SOL_IP, IP_RECVERR, 1) < 0)
		    bb_perror_msg_and_die ("setsockopt IP_RECVERR");
	}
#if ENABLE_FEATURE_IPV6
	else if (af == AF_INET6) {
	    if (setsockopt_1i (sk, SOL_IPV6, IPV6_RECVERR, 1) < 0)
		    bb_perror_msg_and_die ("setsockopt IPV6_RECVERR");
	}
#endif
}

static void parse_icmp_res (probe *pb, int type, int code, int info) {
	char *str = NULL;
	char buf_icmp_res[sizeof (pb->err_str)];

	if (af == AF_INET) {

	    if (type == ICMP_TIME_EXCEEDED) {
		if (code == ICMP_EXC_TTL)
			return;
	    }
	    else if (type == ICMP_DEST_UNREACH) {

		switch (code) {
		    case ICMP_UNREACH_NET:
		    case ICMP_UNREACH_NET_UNKNOWN:
		    case ICMP_UNREACH_ISOLATED:
		    case ICMP_UNREACH_TOSNET:
			    str = "!N";
			    break;

		    case ICMP_UNREACH_HOST:
		    case ICMP_UNREACH_HOST_UNKNOWN:
		    case ICMP_UNREACH_TOSHOST:
			    str = "!H";
			    break;

		    case ICMP_UNREACH_NET_PROHIB:
		    case ICMP_UNREACH_HOST_PROHIB:
		    case ICMP_UNREACH_FILTER_PROHIB:
			    str = "!X";
			    break;

		    case ICMP_UNREACH_PORT:
			    /*  dest host is reached   */
			    str = "";
			    break;

		    case ICMP_UNREACH_PROTOCOL:
			    str = "!P";
			    break;

		    case ICMP_UNREACH_NEEDFRAG:
			    snprintf (buf_icmp_res, sizeof (buf_icmp_res), "!F-%d", info);
			    str = buf_icmp_res;
			    break;

		    case ICMP_UNREACH_SRCFAIL:
			    str = "!S";
			    break;

		    case ICMP_UNREACH_HOST_PRECEDENCE:
			    str = "!V";
			    break;

		    case ICMP_UNREACH_PRECEDENCE_CUTOFF:
			    str = "!C";
			    break;

		    default:
			    snprintf (buf_icmp_res, sizeof (buf_icmp_res), "!<%u>", code);
			    str = buf_icmp_res;
			    break;
		}
	    }

	}
#if ENABLE_FEATURE_IPV6
	else if (af == AF_INET6) {

	    if (type == ICMP6_TIME_EXCEEDED) {
		if (code == ICMP6_TIME_EXCEED_TRANSIT)
			return;
	    }
	    else if (type == ICMP6_DST_UNREACH) {

		switch (code) {

		    case ICMP6_DST_UNREACH_NOROUTE:
			    str = "!N";
			    break;

		    case ICMP6_DST_UNREACH_BEYONDSCOPE:
		    case ICMP6_DST_UNREACH_ADDR:
			    str = "!H";
			    break;

		    case ICMP6_DST_UNREACH_ADMIN:
			    str = "!X";
			    break;

		    case ICMP6_DST_UNREACH_NOPORT:
			    /*  dest host is reached   */
			    str = "";
			    break;

		    default:
			    snprintf (buf_icmp_res, sizeof (buf_icmp_res), "!<%u>", code);
			    str = buf_icmp_res;
			    break;
		}
	    }
	    else if (type == ICMP6_PACKET_TOO_BIG) {
		snprintf (buf_icmp_res, sizeof (buf_icmp_res), "!F-%d", info);
		str = buf_icmp_res;
	    }
	}
#endif
	if (!str) {
	    snprintf (buf_icmp_res, sizeof (buf_icmp_res), "!<%u-%u>", type, code);
	    str = buf_icmp_res;
	}

	if (*str)
	    safe_strncpy (pb->err_str, str, sizeof (pb->err_str));

	pb->final = 1;
}

static void parse_local_res (probe *pb, int ee_errno, int info) {

	if (ee_errno == EMSGSIZE && info != 0) {
	    snprintf (pb->err_str, sizeof(pb->err_str)-1, "!F-%d", info);
	    pb->final = 1;
	    return;
	}

	errno = ee_errno;
	bb_perror_msg_and_die ("local recverr");
}

static void recv_reply (int sk, int err, check_reply_t check_reply) {
	struct msghdr msg;
	sockaddr_any from;
	struct iovec iov;
	int n;
	probe *pb;
	char buf_recv_reply[1280];         /*  min mtu for ipv6 ( >= 576 for ipv4)  */
	char *bufp = buf_recv_reply;
	char control[1024];
	struct cmsghdr *cm;
	double recv_time = 0;
	int recv_ttl = 0;
	struct sock_extended_err *ee = NULL;

	memset (&msg, 0, sizeof (msg));
	msg.msg_name = &from;
	msg.msg_namelen = sizeof (from);
	msg.msg_control = control;
	msg.msg_controllen = sizeof (control);
	iov.iov_base = buf_recv_reply;
	iov.iov_len = sizeof (buf_recv_reply);
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;

	n = recvmsg (sk, &msg, err ? MSG_ERRQUEUE : 0);
	if (n < 0)  return;

	/*  when not MSG_ERRQUEUE, AF_INET returns full ipv4 header
	    on raw sockets...
	*/

	if (!err &&
	    af == AF_INET &&
	    /*  XXX: Assume that the presence of an extra header means
		that it is not a raw socket...
	    */
	    ops->header_len == 0
	) {
	    struct iphdr *ip = (struct iphdr *) bufp;
	    int hlen;

	    if (n < (int)sizeof (struct iphdr))  return;

	    hlen = ip->ihl << 2;
	    if (n < hlen)  return;

	    bufp += hlen;
	    n -= hlen;
	}

	pb = check_reply (sk, err, &from, bufp, n);
	if (!pb) {

	    /*  for `frag needed' case at the local host,
		kernel >= 3.13 sends local error (no more icmp)
	    */
	    if (!n && err && dontfrag) {
		pb = &probes[(first_hop - 1) * probes_per_hop];
		if (pb->done)  return;
	    } else
		return;
	}

	/*  Parse CMSG stuff   */

	for (cm = CMSG_FIRSTHDR (&msg); cm; cm = CMSG_NXTHDR (&msg, cm)) {
	    void *ptr = CMSG_DATA (cm);

	    if (cm->cmsg_level == SOL_SOCKET) {

		if (cm->cmsg_type == SO_TIMESTAMP) {
		    struct timeval *tv = (struct timeval *) ptr;

		    recv_time = tv->tv_sec + tv->tv_usec / 1000000.;
		}
	    }
	    else if (cm->cmsg_level == SOL_IP) {

		if (cm->cmsg_type == IP_TTL)
			recv_ttl = *((int *) ptr);
		else if (cm->cmsg_type == IP_RECVERR) {

		    ee = (struct sock_extended_err *) ptr;

		    if (ee->ee_origin != SO_EE_ORIGIN_ICMP &&
			ee->ee_origin != SO_EE_ORIGIN_LOCAL
		    )  return;

		    /*  dgram icmp sockets might return extra things...  */
		    if (ee->ee_origin == SO_EE_ORIGIN_ICMP &&
			(ee->ee_type == ICMP_SOURCE_QUENCH ||
			 ee->ee_type == ICMP_REDIRECT)
		    )  return;
		}
	    }
#if ENABLE_FEATURE_IPV6
	    else if (cm->cmsg_level == SOL_IPV6) {

		if (cm->cmsg_type == IPV6_HOPLIMIT)
			recv_ttl = *((int *) ptr);
		else if (cm->cmsg_type == IPV6_RECVERR) {

		    ee = (struct sock_extended_err *) ptr;

		    if (ee->ee_origin != SO_EE_ORIGIN_ICMP6 &&
			ee->ee_origin != SO_EE_ORIGIN_LOCAL
		    )  return;
		}
	    }
#endif
	}

	if (!recv_time)
		recv_time = get_time ();

	if (!err)
	    memcpy (&pb->res, &from, sizeof (pb->res));

	pb->recv_time = recv_time;

	pb->recv_ttl = recv_ttl;

	if (ee && ee->ee_origin != SO_EE_ORIGIN_LOCAL) {    /*  icmp or icmp6   */
	    memcpy (&pb->res, SO_EE_OFFENDER (ee), sizeof(pb->res));
	    parse_icmp_res (pb, ee->ee_type, ee->ee_code, ee->ee_info);
	}

	if (ee && ee->ee_origin == SO_EE_ORIGIN_LOCAL)
		parse_local_res (pb, ee->ee_errno, ee->ee_info);

	if (ee &&
	    mtudisc &&
	    ee->ee_info >= g_header_len &&
	    ee->ee_info < g_header_len + data_len
	) {
	    data_len = ee->ee_info - g_header_len;

	    probe_done (pb);

	    /*  clear this probe (as actually the previous hop answers here)
	      but fill its `err_str' by the info obtained. Ugly, but easy...
	    */
	    memset (pb, 0, sizeof (*pb));
	    snprintf (pb->err_str, sizeof(pb->err_str)-1, "F=%d", ee->ee_info);

	    return;
	}

	if (ee &&
	    (bss_flags & OPT_e) &&
	    g_header_len + n >= (128 + 8) &&      /*  at least... (rfc4884)  */
	    g_header_len <= 128 &&        /*  paranoia   */
	    ((af == AF_INET && (ee->ee_type == ICMP_TIME_EXCEEDED ||
				ee->ee_type == ICMP_DEST_UNREACH ||
				ee->ee_type == ICMP_PARAMETERPROB))
#if ENABLE_FEATURE_IPV6
	    ||
	     (af == AF_INET6 && (ee->ee_type == ICMP6_TIME_EXCEEDED ||
				 ee->ee_type == ICMP6_DST_UNREACH))
#endif
	    )
	) {
	    int step;
	    int offs = 128 - g_header_len;

	    if (n > (int)data_len)  step = 0;        /*  guaranteed at 128 ...  */
	    else
		step = af == AF_INET ? 4 : 8;

	    handle_extensions (pb, bufp + offs, n - offs, step);
	}

	probe_done (pb);
}

static int do_send (int sk, const void *s_data, size_t len, const sockaddr_any *addr) {
	int res;

	if (!addr || raw_can_connect ())
		res = send (sk, s_data, len, 0);
	else
	    res = sendto (sk, s_data, len, 0, &addr->sa, sizeof (*addr));

	if (res < 0) {
	    if (errno == ENOBUFS || errno == EAGAIN)
		    return res;
	    if (errno == EMSGSIZE)
		    return 0;   /*  recverr will say more...  */
	    bb_perror_msg_and_die ("send");     /*  not recoverable   */
	}

	return res;
}

#ifdef CONFIG_FEATURE_TRACEROUTE_USE_DCCP
#define DEF_SERVICE_CODE        1885957735

static int dccp_init (const sockaddr_any *dest,
				unsigned int port_seq, size_t *packet_len_p) {
	int af_dccp_init = dest->sa.sa_family;
	sockaddr_any src;
	socklen_t len;
	uint8_t *ptr;
	uint16_t *lenp;
	llist_t *o;
	unsigned int service_code = DEF_SERVICE_CODE;
	struct dccp_hdr_request *dhr;

	dest_addr = *dest;
	dest_addr.sin.sin_port = 0;     /*  raw sockets can be confused   */

	if (!port_seq)  port_seq = DEF_DCCP_PORT;
	dest_port = htons (port_seq);

	/*  Create raw socket for DCCP   */
	raw_sk = socket (af_dccp_init, SOCK_RAW, IPPROTO_DCCP);
	if (raw_sk < 0)
		error_or_perm ("socket");

	tune_socket (raw_sk);       /*  including bind, if any   */

	if (connect (raw_sk, &dest_addr.sa, sizeof (dest_addr)) < 0)
		bb_perror_msg_and_die ("connect");

	len = sizeof (src);
	if (getsockname (raw_sk, &src.sa, &len) < 0)
		bb_perror_msg_and_die ("getsockname");

	if (!raw_can_connect ()) {      /*  work-around for buggy kernels  */
		close (raw_sk);
		raw_sk = bb_xsocket (BB_XSOCKET_FLG_AF_SPEC|BB_XSOCKET_FLG_SOCK_RAW|BB_XSOCKET_FLG_PROTO_SPEC,
				af_dccp_init, IPPROTO_DCCP);
		tune_socket (raw_sk);
		/*  but do not connect it...  */
	}

	use_recverr (raw_sk);

	add_poll (raw_sk, POLLIN | POLLERR);

	/*  Now create the sample packet.  */

	/*  For easy checksum computing:
		saddr
		daddr
		length
		protocol
		dccphdr
	*/

	ptr = buf;

	if (af_dccp_init == AF_INET) {
		len = sizeof (src.sin.sin_addr);
		memcpy (ptr, &src.sin.sin_addr, len);
		ptr += len;
		memcpy (ptr, &dest_addr.sin.sin_addr, len);
		ptr += len;
	}
#if ENABLE_FEATURE_IPV6
	  else {
		len = sizeof (src.sin6.sin6_addr);
		memcpy (ptr, &src.sin6.sin6_addr, len);
		ptr += len;
		memcpy (ptr, &dest_addr.sin6.sin6_addr, len);
		ptr += len;
	}
#endif

	lenp = (uint16_t *) ptr;
	ptr += sizeof (uint16_t);
	*((uint16_t *) ptr) = htons ((uint16_t) IPPROTO_DCCP);
	ptr += sizeof (uint16_t);

	/*  Construct DCCP header   */
	dh = (struct dccp_hdr *) ptr;

	dh->dccph_ccval = 0;
	dh->dccph_checksum = 0;
	dh->dccph_cscov = 0;
	dh->dccph_dport = dest_port;
	dh->dccph_reserved = 0;
	dh->dccph_sport = 0;    /*  temporary   */
	dh->dccph_x = 1;
	dh->dccph_type = DCCP_PKT_REQUEST;
	dh->dccph_seq2 = 0;     /*  reserved if using 48 bit sequence numbers  */
	/*  high 16 bits of sequence number. Always make 0 for simplicity.  */
	dh->dccph_seq = 0;
	ptr += sizeof (struct dccp_hdr);

	dhe = (struct dccp_hdr_ext *) ptr;
	dhe->dccph_seq_low = 0;         /*  temporary   */
	ptr += sizeof (struct dccp_hdr_ext);

	for(o = O; o != NULL; o = o->link) {
		char *sopt = o->data;

		if(strncmp(sopt, "service=", sizeof("service=") - 1))
			ex_error ("unknown option -O `%s' specified", sopt);
		sopt += sizeof("service=") - 1;
		service_code = bb_xparse_ui_number(BB_PN_ERR_STR1, sopt, "invalid service value '%s'");
	}

	dhr = (struct dccp_hdr_request *) ptr;
	dhr->dccph_req_service = htonl (service_code);
	ptr += sizeof (struct dccp_hdr_request);

	csum_len = ptr - buf;

	if (csum_len > sizeof (buf))
		bb_perror_msg_and_die ("impossible");   /*  paranoia   */

	len = ptr - (uint8_t *) dh;
	if (len & 0x03)  bb_perror_msg_and_die ("impossible");  /*  as >>2 ...  */

	*lenp = htons (len);
	dh->dccph_doff = len >> 2;

	*packet_len_p = len;

	return 0;
}

static void dccp_send_probe (probe *pb, int ttl) {
	int sk;
	int af_dccp_send_probe = dest_addr.sa.sa_family;
	sockaddr_any addr;
	socklen_t len = sizeof (addr);

	/*  To make sure we have chosen a free unused "source port",
	   just create, (auto)bind and hold a socket while the port is needed.
	*/

	sk = socket (af_dccp_send_probe, SOCK_DCCP, IPPROTO_DCCP);
	if (sk < 0)  bb_perror_msg_and_die ("socket");

	bind_socket (sk);

	if (getsockname (sk, &addr.sa, &len) < 0)
		bb_perror_msg_and_die ("getsockname");

	/*  When we reach the target host, it can send us either Reset or Response.
	  For Reset all is OK (we and kernel just answer nothing), but
	  for Response we should reply with our Close.
	    It is well-known "half-open technique", used by port scanners etc.
	  This way we do not touch remote applications at all, unlike
	  the ordinary connect(2) call.
	    As the port-holding socket neither connect() nor listen(),
	  it means "no such port yet" for remote ends, and kernel always
	  send Reset in such a situation automatically (we have to do nothing).
	*/

	dh->dccph_sport = addr.sin.sin_port;

	dhe->dccph_seq_low = random_seq ();

	dh->dccph_checksum = 0;
	dh->dccph_checksum = in_cksum (buf, csum_len, 0);

	if (ttl != last_ttl_dccp) {
		set_ttl (raw_sk, ttl);
		last_ttl_dccp = ttl;
	}

	pb->send_time = get_time ();

	if (do_send (raw_sk, dh, dh->dccph_doff << 2, &dest_addr) < 0) {
		close (sk);
		pb->send_time = 0;
		return;
	}

	pb->seq = dh->dccph_sport;

	pb->sk = sk;
}

static probe *dccp_check_reply (int sk, int err, sockaddr_any *from,
				char *buf_dccp_check_reply, size_t len) {
	(void)sk;
	probe *pb;
	struct dccp_hdr *ndh = (struct dccp_hdr *) buf_dccp_check_reply;
	uint16_t sport, dport;

	if (len < 8)  return NULL;          /*  too short   */

	if (err) {
	    sport = ndh->dccph_sport;
	    dport = ndh->dccph_dport;
	} else {
	    sport = ndh->dccph_dport;
	    dport = ndh->dccph_sport;
	}

	if (dport != dest_port)
		return NULL;

	if (!equal_addr (&dest_addr, from))
		return NULL;

	pb = probe_by_seq (sport);
	if (!pb)  return NULL;

	if (!err)  pb->final = 1;

	return pb;
}

static void dccp_recv_probe (int sk, int revents) {

	if (!(revents & (POLLIN | POLLERR)))
		return;

	recv_reply (sk, !!(revents & POLLERR), dccp_check_reply);
}

static const tr_module dccp_ops = {
	.name = "dccp",
	.init = dccp_init,
	.send_probe = dccp_send_probe,
	.recv_probe = dccp_recv_probe,
};

#endif /* CONFIG_FEATURE_TRACEROUTE_USE_DCCP */

#ifdef CONFIG_FEATURE_TRACEROUTE_USE_ICMP
static int icmp_init (const sockaddr_any *dest,
			    unsigned int port_seq, size_t *packet_len_p) {
	size_t i;
	int af_icmp_init = dest->sa.sa_family;
	int protocol;
	int dgram = 0, raw = 0;
	llist_t *o;

	dest_addr = *dest;
	dest_addr.sin.sin_port = 0;

	if (port_seq)  seq_icmp = port_seq;

	length_p = packet_len_p;
	if (*length_p < sizeof (struct icmphdr))
		*length_p = sizeof (struct icmphdr);

	fill_data = xmalloc (*length_p);

	for (i = sizeof (struct icmphdr); i < *length_p; i++)
		fill_data[i] = 0x40 + (i & 0x3f);

#if ENABLE_FEATURE_IPV6
	protocol = (af_icmp_init == AF_INET) ? IPPROTO_ICMP : IPPROTO_ICMPV6;
#else
	protocol = IPPROTO_ICMP;
#endif

	for(o = O; o != NULL; o = o->link) {
		char *sopt = o->data;

		if(!strcmp(sopt, "raw")) {
			raw = 1;
			continue;
		}
		if(!strcmp(sopt, "dgram")) {
			dgram = 1;
			continue;
		}
		ex_error ("unknown option -O `%s' specified", sopt);
	}
	if(raw && dgram)
		ex_error ("-O raw and -O dgram are mutually exclusive");

	if (!raw) {
	    icmp_sk = socket (af_icmp_init, SOCK_DGRAM, protocol);
	    if (icmp_sk < 0 && dgram)
		    bb_perror_msg_and_die ("socket");
	}

	if (!dgram) {
	    int tst_raw_sk = socket (af_icmp_init, SOCK_RAW, protocol);
	    if (tst_raw_sk < 0) {
		if (raw || icmp_sk < 0)
			error_or_perm ("socket");
		dgram = 1;
	    } else {
		/*  prefer the traditional "raw" way when possible   */
		close (icmp_sk);
		icmp_sk = tst_raw_sk;
	    }
	}

	tune_socket (icmp_sk);

	/*  Don't want to catch packets from another hosts   */
	if (raw_can_connect () &&
	    connect (icmp_sk, &dest_addr.sa, sizeof (dest_addr)) < 0
	)  bb_perror_msg_and_die ("connect");

	use_recverr (icmp_sk);

	if (dgram) {
	    sockaddr_any addr;
	    socklen_t len = sizeof (addr);

	    if (getsockname (icmp_sk, &addr.sa, &len) < 0)
		    bb_perror_msg_and_die ("getsockname");
	    ident = ntohs (addr.sin.sin_port);  /*  both IPv4 and IPv6   */

	} else
	    ident = getpid () & 0xffff;

	add_poll (icmp_sk, POLLIN | POLLERR);

	return 0;
}

static void icmp_send_probe (probe *pb, int ttl) {
	int af_icmp_send_probe = dest_addr.sa.sa_family;

	if (ttl != last_ttl_icmp) {

	    set_ttl (icmp_sk, ttl);

	    last_ttl_icmp = ttl;
	}

	if (af_icmp_send_probe == AF_INET) {
	    struct icmp *icmp = (struct icmp *) fill_data;

	    icmp->icmp_type = ICMP_ECHO;
	    icmp->icmp_code = 0;
	    icmp->icmp_cksum = 0;
	    icmp->icmp_id = htons (ident);
	    icmp->icmp_seq = htons (seq_icmp);

	    icmp->icmp_cksum = in_cksum (fill_data, *length_p, 0);
	}
#if ENABLE_FEATURE_IPV6
	else if (af_icmp_send_probe == AF_INET6) {
	    struct icmp6_hdr *icmp6 = (struct icmp6_hdr *) fill_data;

	    icmp6->icmp6_type = ICMP6_ECHO_REQUEST;
	    icmp6->icmp6_code = 0;
	    icmp6->icmp6_cksum = 0;
	    icmp6->icmp6_id = htons (ident);
	    icmp6->icmp6_seq = htons(seq_icmp);

	    /*  icmp6->icmp6_cksum always computed by kernel internally   */
	}
#endif
	pb->send_time = get_time ();

	if (do_send (icmp_sk, fill_data, *length_p, &dest_addr) < 0) {
	    pb->send_time = 0;
	    return;
	}

	pb->seq = seq_icmp;

	seq_icmp++;
}

static probe *icmp_check_reply (int sk, int err, sockaddr_any *from,
				char *buf_icmp_check_reply, size_t len) {

	(void)sk;
	(void)from;
	int af_check_reply = dest_addr.sa.sa_family;
	int type;
	uint16_t recv_id, recv_seq;
	probe *pb;

	if (len < sizeof (struct icmphdr))
		return NULL;

#if ENABLE_FEATURE_IPV6
	if (af_check_reply == AF_INET)
#endif
	{
	    struct icmp *icmp = (struct icmp *) buf_icmp_check_reply;

	    type = icmp->icmp_type;

	    recv_id = ntohs (icmp->icmp_id);
	    recv_seq = ntohs (icmp->icmp_seq);

	}
#if ENABLE_FEATURE_IPV6
	else {      /*  AF_INET6   */
	    struct icmp6_hdr *icmp6 = (struct icmp6_hdr *) buf_icmp_check_reply;

	    type = icmp6->icmp6_type;

	    recv_id = ntohs (icmp6->icmp6_id);
	    recv_seq = ntohs (icmp6->icmp6_seq);
	}
#endif

	if (recv_id != ident)
		return NULL;

	pb = probe_by_seq (recv_seq);
	if (!pb)  return NULL;

	if (!err) {

	    if (!(af_check_reply == AF_INET && type == ICMP_ECHOREPLY)
#if ENABLE_FEATURE_IPV6
		&&
		!(af_check_reply == AF_INET6 && type == ICMP6_ECHO_REPLY)
#endif
	    )  return NULL;

	    pb->final = 1;
	}

	return pb;
}

static void icmp_recv_probe (int sk, int revents) {

	if (!(revents & (POLLIN | POLLERR)))
		return;

	recv_reply (sk, !!(revents & POLLERR), icmp_check_reply);
}

static const tr_module icmp_ops = {
	.name = "icmp",
	.init = icmp_init,
	.send_probe = icmp_send_probe,
	.recv_probe = icmp_recv_probe,
};
#endif /* CONFIG_FEATURE_TRACEROUTE_USE_ICMP */


static void set_protocol (char *arg) {
	char *q;

	protocol_raw = strtoul (arg, &q, 0);
	if (q == arg) {
	    struct protoent *p = getprotobyname (arg);

	    if (!p)  ex_error ("unknown protocol specified '%s'", arg);
	    protocol_raw = p->p_proto;
	}
}

static int raw_init (const sockaddr_any *dest,
			    unsigned int port_seq, size_t *packet_len_p) {
	size_t i;
	int af_raw_init = dest->sa.sa_family;
	llist_t *o;
	char *sopt = NULL;

	dest_addr = *dest;
	dest_addr.sin.sin_port = 0;

	for(o = O; o != NULL; o = o->link) {
		sopt = o->data;

		if(strncmp(sopt, "protocol=", sizeof("protocol=") - 1))
			ex_error ("unknown option -O `%s' specified", sopt);
	}
	if(sopt != NULL)
		set_protocol(sopt + sizeof("protocol=") - 1);

	if (port_seq)  protocol_raw = port_seq;

	length_p = packet_len_p;

	fill_data = xmalloc (*length_p);

	for (i = 0; i < *length_p; i++)
		fill_data[i] = 0x40 + (i & 0x3f);

	raw_sk = socket (af_raw_init, SOCK_RAW, protocol_raw);
	if (raw_sk < 0)
		error_or_perm ("socket");

	tune_socket (raw_sk);

	/*  Don't want to catch packets from another hosts   */
	if (raw_can_connect () &&
	    connect (raw_sk, &dest_addr.sa, sizeof (dest_addr)) < 0
	)  bb_perror_msg_and_die ("connect");

	use_recverr (raw_sk);

	add_poll (raw_sk, POLLIN | POLLERR);

	return 0;
}

static void raw_send_probe (probe *pb, int ttl) {

	if (ttl != last_ttl_raw) {

	    set_ttl (raw_sk, ttl);

	    last_ttl_raw = ttl;
	}

	pb->send_time = get_time ();

	if (do_send (raw_sk, fill_data, *length_p, &dest_addr) < 0) {
	    pb->send_time = 0;
	    return;
	}

	pb->seq = ++seq_raw;
}

static probe *raw_check_reply (int sk, int err, sockaddr_any *from,
			       char *buf_raw_check_reply, size_t len) {
	probe *pb;

	(void)sk;
	(void)buf_raw_check_reply;
	(void)len;

	if (!equal_addr (&dest_addr, from))
		return NULL;

	pb = probe_by_seq (seq_raw);
	if (!pb)  return NULL;

	if (!err)  pb->final = 1;

	return pb;
}

static void raw_recv_probe (int sk, int revents) {

	if (!(revents & (POLLIN | POLLERR)))
		return;

	recv_reply (sk, !!(revents & POLLERR), raw_check_reply);
}

#ifdef CONFIG_FEATURE_TRACEROUTE_USE_TCP
#ifndef IP_MTU
#define IP_MTU  14
#endif

#define TH_FLAGS(TH)    (((uint8_t *) (TH))[13])
#define TH_FIN  0x01
#define TH_SYN  0x02
#define TH_RST  0x04
#define TH_PSH  0x08
#define TH_ACK  0x10
#define TH_URG  0x20
#define TH_ECE  0x40
#define TH_CWR  0x80

#define FL_FLAGS        0x0100
#define FL_ECN          0x0200
#define FL_SACK         0x0400
#define FL_TSTAMP       0x0800
#define FL_WSCALE       0x1000


static const struct {
	const char *name;
	unsigned int flag;
} tcp_flags[] = {
	{ "fin", TH_FIN },
	{ "syn", TH_SYN },
	{ "rst", TH_RST },
	{ "psh", TH_PSH },
	{ "ack", TH_ACK },
	{ "urg", TH_URG },
	{ "ece", TH_ECE },
	{ "cwr", TH_CWR },
};

static char *names_by_flags (unsigned int flags_nf) {
	size_t i;
	char str[64];   /*  enough...  */
	char *curr_nf = str;
	char *end = str + sizeof (str) / sizeof (*str);

	for (i = 0; i < sizeof (tcp_flags) / sizeof (*tcp_flags); i++) {
	    const char *p;

	    if (!(flags_nf & tcp_flags[i].flag))  continue;

	    if (curr_nf > str && curr_nf < end)  *curr_nf++ = ',';
	    for (p = tcp_flags[i].name; *p && curr_nf < end; *curr_nf++ = *p++) ;
	}

	*curr_nf = '\0';

	return  strdup (str);
}

static int set_tcp_flags (char *arg) {
	unsigned int value;

	value = bb_xparse_ui_number(BB_PN_SET_UPPER|BB_PN_ERR_STR1, arg, 0xff, "invalid flags value '%s'");
	flags = (flags & ~0xff) | (value & 0xff) | FL_FLAGS;
	return 0;
}

#define SYSCTL_PREFIX   "/proc/sys/net/ipv4/tcp_"
static int check_sysctl (const char *name) {
	int fd, res;
	char buf_check_sysctl[sizeof (SYSCTL_PREFIX) + strlen (name) + 1];
	uint8_t ch;

	strcpy (buf_check_sysctl, SYSCTL_PREFIX);
	strcat (buf_check_sysctl, name);

	fd = open (buf_check_sysctl, O_RDONLY, 0);
	if (fd < 0)  return 0;

	res = read (fd, &ch, sizeof (ch));
	close (fd);

	if (res != sizeof (ch))
		return 0;

	/*  since kernel 2.6.31 "tcp_ecn" can have value of '2'...  */
	if (ch == '1')  return 1;

	return 0;
}

static int tcp_init (const sockaddr_any *dest,
			    unsigned int port_seq, size_t *packet_len_p) {
	int af_tcp_init = dest->sa.sa_family;
	sockaddr_any src;
	int mtu;
	socklen_t len;
	uint8_t *ptr;
	uint16_t *lenp;
	llist_t *o;

	dest_addr = *dest;
	dest_addr.sin.sin_port = 0;     /*  raw sockets can be confused   */

	if (!port_seq)  port_seq = DEF_TCP_PORT;
	dest_port = htons (port_seq);

	/*  Create raw socket for tcp   */

	raw_sk = socket (af_tcp_init, SOCK_RAW, IPPROTO_TCP);
	if (raw_sk < 0)
		error_or_perm ("socket");

	tune_socket (raw_sk);       /*  including bind, if any   */

	if (connect (raw_sk, &dest_addr.sa, sizeof (dest_addr)) < 0)
		bb_perror_msg_and_die ("connect");

	len = sizeof (src);
	if (getsockname (raw_sk, &src.sa, &len) < 0)
		bb_perror_msg_and_die ("getsockname");

	len = sizeof (mtu);
#if ENABLE_FEATURE_IPV6
	if (getsockopt (raw_sk, af_tcp_init == AF_INET ? SOL_IP : SOL_IPV6,
				af_tcp_init == AF_INET ? IP_MTU : IPV6_MTU,
				&mtu, &len) < 0 || mtu < 576
	)  mtu = 576;
#else
	if (getsockopt (raw_sk, SOL_IP, IP_MTU, &mtu, &len) < 0 || mtu < 576)
		mtu = 576;
#endif

	/*  mss = mtu - headers   */
#if ENABLE_FEATURE_IPV6
	mtu -= af_tcp_init == AF_INET ? sizeof (struct iphdr) : sizeof (struct ip6_hdr);
#else
	mtu -= sizeof (struct iphdr);
#endif
	mtu -= sizeof (struct tcphdr);

	if (!raw_can_connect ()) {      /*  work-around for buggy kernels  */
	    close (raw_sk);
	    raw_sk = bb_xsocket (BB_XSOCKET_FLG_AF_SPEC|BB_XSOCKET_FLG_SOCK_RAW|BB_XSOCKET_FLG_PROTO_SPEC,
				af_tcp_init, IPPROTO_TCP);
	    tune_socket (raw_sk);
	    /*  but do not connect it...  */
	}

	use_recverr (raw_sk);

	add_poll (raw_sk, POLLIN | POLLERR);

	for(o = O; o != NULL; o = o->link) {
		char *sopt = o->data;
		size_t i;

		for (i = 0; i < sizeof (tcp_flags) / sizeof (*tcp_flags); i++) {
			if (!strcmp (sopt, tcp_flags[i].name)) {
				flags |= tcp_flags[i].flag;
				i = sizeof (tcp_flags);
				break;
			}
		}
		if(i == sizeof (tcp_flags)) continue;

		if(!strncmp(sopt, "flags=", sizeof("flags=") - 1)) {
			set_tcp_flags(sopt + sizeof("flags=") - 1);
			continue;
		}
		if (!strcmp (sopt, "ecn"))              { flags |= FL_ECN;    continue; }
		if (!strcmp (sopt, "sack"))             { flags |= FL_SACK;   continue; }
		if (!strcmp (sopt, "timestamps"))       { flags |= FL_TSTAMP; continue; }
		if (!strcmp (sopt, "window_scaling"))   { flags |= FL_WSCALE; continue; }
		if (!strcmp (sopt, "sysctl"))           { sysctl = 1;         continue; }
		if (!strcmp (sopt, "reuse"))            { reuse = 1;          continue; }
		if (!strcmp (sopt, "info"))             { tcp_info = 1;           continue; }
		if(!strncmp(sopt, "mss=", sizeof("mss=") - 1)) {
			sopt += sizeof("mss=") - 1;
			mss = bb_xparse_ui_number(BB_PN_ERR_STR1, sopt, "invalid mss value '%s'");
			continue;
		}
		ex_error ("unknown option -O `%s' specified", sopt);
	}
	/*  Now create the sample packet.  */

	if (!flags)  sysctl = 1;

	if (sysctl) {
	    if (check_sysctl ("ecn"))  flags |= FL_ECN;
	    if (check_sysctl ("sack"))  flags |= FL_SACK;
	    if (check_sysctl ("timestamps"))  flags |= FL_TSTAMP;
	    if (check_sysctl ("window_scaling"))  flags |= FL_WSCALE;
	}

	if (!(flags & (FL_FLAGS | 0xff))) {     /*  no any tcp flag set   */
	    flags |= TH_SYN;
	    if (flags & FL_ECN)
		    flags |= TH_ECE | TH_CWR;
	}

	/*  For easy checksum computing:
	    saddr
	    daddr
	    length
	    protocol
	    tcphdr
	    tcpoptions
	*/

	ptr = buf;

#if ENABLE_FEATURE_IPV6
	if (af_tcp_init == AF_INET)
#endif
	{
	    len = sizeof (src.sin.sin_addr);
	    memcpy (ptr, &src.sin.sin_addr, len);
	    ptr += len;
	    memcpy (ptr, &dest_addr.sin.sin_addr, len);
	    ptr += len;
	}
#if ENABLE_FEATURE_IPV6
	  else {
	    len = sizeof (src.sin6.sin6_addr);
	    memcpy (ptr, &src.sin6.sin6_addr, len);
	    ptr += len;
	    memcpy (ptr, &dest_addr.sin6.sin6_addr, len);
	    ptr += len;
	}
#endif

	lenp = (uint16_t *) ptr;
	ptr += sizeof (uint16_t);
	*((uint16_t *) ptr) = htons ((uint16_t) IPPROTO_TCP);
	ptr += sizeof (uint16_t);

	/*  Construct TCP header   */

	th = (struct tcphdr *) ptr;

	th->source = 0;     /*  temporary   */
	th->dest = dest_port;
	th->seq = 0;        /*  temporary   */
	th->ack_seq = 0;
	th->doff = 0;       /*  later...  */
	TH_FLAGS(th) = flags & 0xff;
	th->window = htons (4 * mtu);
	th->check = 0;
	th->urg_ptr = 0;

	/*  Build TCP options   */

	ptr = (uint8_t *) (th + 1);

	if (flags & TH_SYN) {
	    *ptr++ = TCPOPT_MAXSEG;     /*  2   */
	    *ptr++ = TCPOLEN_MAXSEG;    /*  4   */
	    *((uint16_t *) ptr) = htons (mss ? mss : mtu);
	    ptr += sizeof (uint16_t);
	}

	if (flags & FL_TSTAMP) {

	    if (flags & FL_SACK) {
		*ptr++ = TCPOPT_SACK_PERMITTED; /*  4   */
		*ptr++ = TCPOLEN_SACK_PERMITTED;/*  2   */
	    } else {
		*ptr++ = TCPOPT_NOP;    /*  1   */
		*ptr++ = TCPOPT_NOP;    /*  1   */
	    }
	    *ptr++ = TCPOPT_TIMESTAMP;  /*  8   */
	    *ptr++ = TCPOLEN_TIMESTAMP; /*  10  */

	    *((uint32_t *) ptr) = random_seq ();        /*  really!  */
	    ptr += sizeof (uint32_t);
	    *((uint32_t *) ptr) = (flags & TH_ACK) ? random_seq () : 0;
	    ptr += sizeof (uint32_t);
	}
	else if (flags & FL_SACK) {
	    *ptr++ = TCPOPT_NOP;        /*  1   */
	    *ptr++ = TCPOPT_NOP;        /*  1   */
	    *ptr++ = TCPOPT_SACK_PERMITTED;     /*  4   */
	    *ptr++ = TCPOLEN_SACK_PERMITTED;    /*  2   */
	}

	if (flags & FL_WSCALE) {
	    *ptr++ = TCPOPT_NOP;        /*  1   */
	    *ptr++ = TCPOPT_WINDOW;     /*  3   */
	    *ptr++ = TCPOLEN_WINDOW;    /*  3   */
	    *ptr++ = 2; /*  assume some corect value...  */
	}

	csum_len = ptr - buf;

	if (csum_len > TCP_BUBSZ)
		bb_perror_msg_and_die ("impossible");   /*  paranoia   */

	len = ptr - (uint8_t *) th;
	if (len & 0x03)  bb_perror_msg_and_die ("impossible");  /*  as >>2 ...  */

	*lenp = htons (len);
	th->doff = len >> 2;

	*packet_len_p = len;

	return 0;
}

static void tcp_send_probe (probe *pb, int ttl) {
	int sk;
	sockaddr_any addr;
	socklen_t len = sizeof (addr);

	/*  To make sure we have chosen a free unused "source port",
	   just create, (auto)bind and hold a socket while the port is needed.
	*/

	sk = bb_xsocket (BB_XSOCKET_FLG_AF_SPEC, dest_addr.sa.sa_family);

	if (reuse && setsockopt_1i (sk, SOL_SOCKET, SO_REUSEADDR, reuse) < 0)
		bb_perror_msg_and_die ("setsockopt SO_REUSEADDR");

	bind_socket (sk);

	if (getsockname (sk, &addr.sa, &len) < 0)
		bb_perror_msg_and_die ("getsockname");

	/*  When we reach the target host, it can send us either RST or SYN+ACK.
	  For RST all is OK (we and kernel just answer nothing), but
	  for SYN+ACK we should reply with our RST.
	    It is well-known "half-open technique", used by port scanners etc.
	  This way we do not touch remote applications at all, unlike
	  the ordinary connect(2) call.
	    As the port-holding socket neither connect() nor listen(),
	  it means "no such port yet" for remote ends, and kernel always
	  send RST in such a situation automatically (we have to do nothing).
	*/

	th->source = addr.sin.sin_port;

	th->seq = random_seq ();

	th->check = 0;
	th->check = in_cksum (buf, csum_len, 0);

	if (ttl != last_ttl_tcp) {

	    set_ttl (raw_sk, ttl);

	    last_ttl_tcp = ttl;
	}

	pb->send_time = get_time ();

	if (do_send (raw_sk, th, th->doff << 2, &dest_addr) < 0) {
	    close (sk);
	    pb->send_time = 0;
	    return;
	}

	pb->seq = th->source;

	pb->sk = sk;
}

static probe *tcp_check_reply (int sk, int err, sockaddr_any *from,
				char *buf_tcp_check_reply, size_t len)
{
	probe *pb;
	struct tcphdr *tcp = (struct tcphdr *) buf_tcp_check_reply;
	uint16_t sport, dport;

	(void)sk;
	if (len < 8)  return NULL;          /*  too short   */

	if (err) {
	    sport = tcp->source;
	    dport = tcp->dest;
	} else {
	    sport = tcp->dest;
	    dport = tcp->source;
	}

	if (dport != dest_port)
		return NULL;

	if (!equal_addr (&dest_addr, from))
		return NULL;

	pb = probe_by_seq (sport);
	if (!pb)  return NULL;

	if (!err) {

	    pb->final = 1;

	    if (tcp_info)
		pb->ext = names_by_flags (TH_FLAGS(tcp));
	}

	return pb;
}

static void tcp_recv_probe (int sk, int revents) {

	if (!(revents & (POLLIN | POLLERR)))
		return;

	recv_reply (sk, !!(revents & POLLERR), tcp_check_reply);
}

static const tr_module tcp_ops = {
	.name = "tcp",
	.init = tcp_init,
	.send_probe = tcp_send_probe,
	.recv_probe = tcp_recv_probe,
};
#endif /* CONFIG_FEATURE_TRACEROUTE_USE_TCP */

#ifdef CONFIG_FEATURE_TRACEROUTE_USE_TCPCONN
static int tcp_init_tcpconn (const sockaddr_any *dest,
			    unsigned int port_seq, size_t *packet_len_p) {
	int af_tcp_init_tcpconn = dest->sa.sa_family;

	(void)packet_len_p;
	dest_addr = *dest;
	dest_addr.sin.sin_port = htons (DEF_TCP_PORT);

	if (port_seq)
	    dest_addr.sin.sin_port = htons (port_seq);

	/*  Currently an ICMP socket is the only way
	  to obtain the needed info...
	*/
#if ENABLE_FEATURE_IPV6
	icmp_sk = socket (af_tcp_init_tcpconn, SOCK_RAW, (af_tcp_init_tcpconn == AF_INET) ? IPPROTO_ICMP
							: IPPROTO_ICMPV6);
#else
	icmp_sk = socket (af_tcp_init_tcpconn, SOCK_RAW, IPPROTO_ICMP);
#endif

	if (icmp_sk < 0)
		error_or_perm ("socket");

	/*  icmp_sk not need full tune_socket() here, just a receiving one  */
	bind_socket (icmp_sk);
	use_timestamp (icmp_sk);
	use_recv_ttl (icmp_sk);

	add_poll (icmp_sk, POLLIN);

	return 0;
}

static void tcp_send_probe_tcpconn (probe *pb, int ttl) {
	int sk;
	sockaddr_any addr;
	socklen_t length = sizeof (addr);

	sk = bb_xsocket (BB_XSOCKET_FLG_AF_SPEC, dest_addr.sa.sa_family);

	tune_socket (sk);       /*  common stuff   */

	set_ttl (sk, ttl);

	pb->send_time = get_time ();

	if (connect (sk, &dest_addr.sa, sizeof (dest_addr)) < 0) {
	    if (errno != EINPROGRESS)
		    bb_perror_msg_and_die ("connect");
	}


	if (getsockname (sk, &addr.sa, &length) < 0)
		bb_perror_msg_and_die ("getsockname");

	pb->seq = addr.sin.sin_port;    /*  both ipv4/ipv6  */

	pb->sk = sk;

	add_poll (sk, POLLERR | POLLHUP | POLLOUT);
}

static probe *tcp_check_reply_tcpconn (int sk, int err, sockaddr_any *from,
				       char *buf_tcp_check_reply_tcpconn, size_t len) {
#if ENABLE_FEATURE_IPV6
	int af_tcp_check_reply_tcpconn = dest_addr.sa.sa_family;
#endif
	int type, code, info_tcp_check_reply_tcpconn;
	probe *pb;
	struct tcphdr *tcp;

	(void)err;
	(void)from;
	(void)sk;
	if (len < sizeof (struct icmphdr))
		return NULL;

#if ENABLE_FEATURE_IPV6
	if (af_tcp_check_reply_tcpconn == AF_INET)
#endif
	{
	    struct icmp *icmp = (struct icmp *) buf_tcp_check_reply_tcpconn;
	    struct iphdr *ip;
	    int hlen;

	    type = icmp->icmp_type;
	    code = icmp->icmp_code;
	    info_tcp_check_reply_tcpconn = icmp->icmp_void;

	    if (type != ICMP_TIME_EXCEEDED && type != ICMP_DEST_UNREACH)
		    return NULL;

	    if (len < sizeof (struct icmphdr) + sizeof (struct iphdr) + 8)
		    /* `8' - rfc1122: 3.2.2  */
		    return NULL;

	    ip = (struct iphdr *) (((char *)icmp) + sizeof(struct icmphdr));
	    hlen = ip->ihl << 2;

	    if (len < sizeof (struct icmphdr) + hlen + 8)
		    return NULL;
	    if (ip->protocol != IPPROTO_TCP)
		    return NULL;

	    tcp = (struct tcphdr *) (((char *) ip) + hlen);

	}
#if ENABLE_FEATURE_IPV6
	else {      /*  AF_INET6   */
	    struct icmp6_hdr *icmp6 = (struct icmp6_hdr *) buf_tcp_check_reply_tcpconn;
	    struct ip6_hdr *ip6;

	    type = icmp6->icmp6_type;
	    code = icmp6->icmp6_code;
	    info_tcp_check_reply_tcpconn = icmp6->icmp6_mtu;

	    if (type != ICMP6_TIME_EXCEEDED &&
		type != ICMP6_DST_UNREACH &&
		type != ICMP6_PACKET_TOO_BIG
	    )  return NULL;

	    if (len < sizeof(struct icmp6_hdr) + sizeof(struct ip6_hdr) + 8)
		    return NULL;

	    ip6 = (struct ip6_hdr *) (icmp6 + 1);
	    if (ip6->ip6_nxt != IPPROTO_TCP)
		    return NULL;

	    tcp = (struct tcphdr *) (ip6 + 1);

	}
#endif
	if (tcp->dest != dest_addr.sin.sin_port)
		return NULL;

	pb = probe_by_seq (tcp->source);
	if (!pb)  return NULL;

	/*  here only, high level has no data to do this   */
	parse_icmp_res (pb, type, code, info_tcp_check_reply_tcpconn);

	return pb;
}

static void tcp_recv_probe_tcpconn (int sk, int revents)
{
	if (sk != icmp_sk) {    /*  a tcp socket   */
	    probe *pb;

	    pb = probe_by_sk (sk);
	    if (!pb) {
		del_poll (sk);
		return;
	    }


	    /*  do connect() again and check errno, regardless of revents  */
	    if (connect (sk, &dest_addr.sa, sizeof (dest_addr)) < 0) {
		if (errno != EISCONN && errno != ECONNREFUSED)
			return; /*  ICMP say more   */
	    }

	    /*  we have reached the dest host (either connected or refused)  */

	    memcpy (&pb->res, &dest_addr, sizeof (pb->res));

	    pb->final = 1;

	    pb->recv_time = get_time ();

	    probe_done (pb);

	    return;
	}

	/*  ICMP stuff   */

	if (!(revents & POLLIN))
		return;

	recv_reply (icmp_sk, 0, tcp_check_reply_tcpconn);
}

static const tr_module tcp_ops_tcpconn = {
	.name = "tcpconn",
	.init = tcp_init_tcpconn,
	.send_probe = tcp_send_probe_tcpconn,
	.recv_probe = tcp_recv_probe_tcpconn,
};
#endif /* CONFIG_FEATURE_TRACEROUTE_USE_TCPCONN */

#ifndef IPPROTO_UDPLITE
#define IPPROTO_UDPLITE 136
#endif

#ifndef UDPLITE_SEND_CSCOV
#define UDPLITE_SEND_CSCOV      10
#define UDPLITE_RECV_CSCOV      11
#endif


static void do_fill_data (size_t *packet_len_p) {
	size_t i;

	length_p = packet_len_p;

	fill_data = xmalloc (*length_p);

	for (i = 0; i < *length_p; i++)
		fill_data[i] = 0x40 + (i & 0x3f);
}

static int udp_default_init (const sockaddr_any *dest,
				unsigned int port_seq, size_t *packet_len_p)
{
	curr_port = port_seq ? port_seq : DEF_START_PORT;

	dest_addr = *dest;
	dest_addr.sin.sin_port = htons (curr_port);

	do_fill_data (packet_len_p);

	return 0;
}

#define MIN_COVERAGE    (sizeof (struct udphdr))

static void set_coverage (int sk) {
	if (setsockopt_1i (sk, IPPROTO_UDPLITE, UDPLITE_SEND_CSCOV, coverage) < 0
	)  bb_perror_msg_and_die ("UDPLITE_SEND_CSCOV");

	if (setsockopt_1i (sk, IPPROTO_UDPLITE, UDPLITE_RECV_CSCOV, MIN_COVERAGE) < 0
	)  bb_perror_msg_and_die ("UDPLITE_RECV_CSCOV");
}

static void udp_send_probe (probe *pb, int ttl) {
	int sk;

	sk = bb_xsocket (BB_XSOCKET_FLG_AF_SPEC|BB_XSOCKET_FLG_SOCK_DGRAM|BB_XSOCKET_FLG_PROTO_SPEC,
				dest_addr.sa.sa_family, protocol_udp);
	tune_socket (sk);       /*  common stuff   */

	if (coverage)  set_coverage (sk);       /*  udplite case   */

	set_ttl (sk, ttl);

	if (connect (sk, &dest_addr.sa, sizeof (dest_addr)) < 0)
		bb_perror_msg_and_die ("connect");

	use_recverr (sk);

	pb->send_time = get_time ();

	if (do_send (sk, fill_data, *length_p, NULL) < 0) {
	    close (sk);
	    pb->send_time = 0;
	    return;
	}

	pb->sk = sk;

	add_poll (sk, POLLIN | POLLERR);

	pb->seq = dest_addr.sin.sin_port;

	if (curr_port) {        /*  traditional udp method   */
	    curr_port++;
	    dest_addr.sin.sin_port = htons (curr_port); /* both ipv4 and ipv6 */
	}
}

static probe *udp_check_reply (int sk, int err, sockaddr_any *from,
			       char *buf_udp_check_reply, size_t len) {
	probe *pb;

	(void)buf_udp_check_reply;
	(void)len;
	pb = probe_by_sk (sk);
	if (!pb)  return NULL;

	if (pb->seq != from->sin.sin_port)
		return NULL;

	if (!err)  pb->final = 1;

	return pb;
}

static void udp_recv_probe (int sk, int revents)
{
	if (!(revents & (POLLIN | POLLERR)))
		return;

	recv_reply (sk, !!(revents & POLLERR), udp_check_reply);
}

#ifdef CONFIG_FEATURE_TRACEROUTE_USE_UDP
/*  All three modules share the same methods except the init...  */
static int udp_init (const sockaddr_any *dest,
			    unsigned int port_seq, size_t *packet_len_p) {

	dest_addr = *dest;

	if (!port_seq)  port_seq = DEF_UDP_PORT;
	dest_addr.sin.sin_port = htons ((uint16_t) port_seq);

	do_fill_data (packet_len_p);

	return 0;
}

static const tr_module udp_ops = {
	.name = "udp",
	.init = udp_init,
	.send_probe = udp_send_probe,
	.recv_probe = udp_recv_probe,
	.header_len = sizeof (struct udphdr),
};
#endif /* CONFIG_FEATURE_TRACEROUTE_USE_UDP */

#ifdef CONFIG_FEATURE_TRACEROUTE_USE_UDPL
static int udplite_init (const sockaddr_any *dest,
			    unsigned int port_seq, size_t *packet_len_p) {

	llist_t *o;
	char *sopt = NULL;

	coverage = MIN_COVERAGE;
	for(o = O; o != NULL; o = o->link) {
		sopt = o->data;

		if(strncmp(sopt, "coverage=", sizeof("coverage=") - 1))
			ex_error ("unknown option -O `%s' specified", sopt);
	}
	if(sopt != NULL) {
		sopt += sizeof("coverage=") - 1;
		coverage = bb_xparse_ui_number(BB_PN_ERR_STR1, sopt, "invalid coverage value '%s'");
	}

	dest_addr = *dest;

	if (!port_seq)  port_seq = DEF_UDP_PORT;    /*  XXX: Hmmm...   */
	dest_addr.sin.sin_port = htons ((uint16_t) port_seq);

	protocol_udp = IPPROTO_UDPLITE;

	do_fill_data (packet_len_p);

	return 0;
}

static const tr_module udplite_ops = {
	.name = "udplite",
	.init = udplite_init,
	.send_probe = udp_send_probe,
	.recv_probe = udp_recv_probe,
	.header_len = sizeof (struct udphdr),
};
#endif /* CONFIG_FEATURE_TRACEROUTE_USE_UDPL */

/*  Set initial parameters according to how we was called   */
static void check_progname (void) {
	const char *p = bb_applet_name;
	int l;

	l = strlen (p);
	if (l <= 0)  return;
	l--;

#if ENABLE_FEATURE_IPV6
	if (p[l] == '6')  af = AF_INET6;
	else
#else
		if (p[l] == '4')  af = AF_INET;
#endif

	if (!strncmp (p, "tcp", 3))
		module = "tcp";
	if (!strncmp (p, "tracert", 7))
		module = "icmp";
}

static int getaddr (const char *name, sockaddr_any *addr) {
	int ret;
	struct addrinfo hints, *ai, *res = NULL;

	memset (&hints, 0, sizeof (hints));
	hints.ai_family = af;
	hints.ai_flags = AI_IDN;

	ret = getaddrinfo (name, NULL, &hints, &res);
	if (ret) {
		fprintf (stderr, "%s: %s\n", name, gai_strerror (ret));
		return -1;
	}

	for (ai = res; ai; ai = ai->ai_next) {
	    if (ai->ai_family == af)  break;
	    /*  when af not specified, choose DEF_AF if present   */
	    if (!af && ai->ai_family == DEF_AF)
		    break;
	}
	if (!ai)  ai = res;     /*  anything...  */

	if (ai->ai_addrlen > sizeof (*addr)) /*  paranoia   */
		bb_error_msg_and_die ("ai->ai_addrlen > sizeof (*addr)");
	memcpy (addr, ai->ai_addr, ai->ai_addrlen);

	freeaddrinfo (res);

	return 0;
}

static void make_fd_used (int fd) {
	int nfd;

	if (fcntl (fd, F_GETFL) != -1)
		return;

	if (errno != EBADF)
		bb_perror_msg_and_die ("fcntl F_GETFL");

	nfd = bb_xopen (bb_dev_null, O_RDONLY | BB_OPEN_ERR);

	if (nfd != fd) {
	    dup2 (nfd, fd);
	    close (nfd);
	}
}

static const char *addr2str (const sockaddr_any *addr, char *addr2str_buf) {

	getnameinfo (&addr->sa, sizeof (*addr),
		addr2str_buf, INET_4_6_ADDRSTRLEN, 0, 0, NI_NUMERICHOST);

	return addr2str_buf;
}

/*      IP  options  stuff          */
static void init_ip_options (void) {
#ifdef CONFIG_FEATURE_TRACEROUTE_SPEC_GATEWAYS
	sockaddr_any *gates;
	int i, max;

	if (!num_gateways)
		return;

#if ENABLE_FEATURE_IPV6
	/*  check for TYPE,ADDR,ADDR... form   */
	if (af == AF_INET6 && num_gateways > 1 && gateways[0]) {
	    char *q;
	    unsigned int value = strtoul (gateways[0], &q, 0);

	    if (!*q) {
		ipv6_rthdr_type = value;
		num_gateways--;
		for (i = 0; i < num_gateways; i++)
			gateways[i] = gateways[i + 1];
	    }
	}
	max = af == AF_INET ? MAX_GATEWAYS_4 : MAX_GATEWAYS_6;
#else
	max = MAX_GATEWAYS_4;
#endif

	if (num_gateways > max)
	    ex_error ("Too many gateways specified. No more than %d", max);


	gates = alloca (num_gateways * sizeof (*gates));

	for (i = 0; i < num_gateways; i++) {

	    if (getaddr (gateways[i], &gates[i]) < 0)
		    exit (2);      /*  already reported   */
	    if (gates[i].sa.sa_family != af)
		    ex_error ("IP versions mismatch in gateway addresses");

	    free (gateways[i]);
	}

	free (gateways);
	gateways = NULL;

	if (af == AF_INET) {
	    struct in_addr *in;

	    rtbuf_len = 4 + (num_gateways + 1) * sizeof (*in);
	    rtbuf = xmalloc (rtbuf_len);

	    in = (struct in_addr *) &rtbuf[4];
	    for (i = 0; i < num_gateways; i++)
		    memcpy (&in[i], &gates[i].sin.sin_addr, sizeof (*in));
	    /*  final hop   */
	    memcpy (&in[i], &dst_addr.sin.sin_addr, sizeof (*in));
	    i++;

	    rtbuf[0] = IPOPT_NOP;
	    rtbuf[1] = IPOPT_LSRR;
	    rtbuf[2] = (i * sizeof (*in)) + 3;
	    rtbuf[3] = IPOPT_MINOFF;

	}
#if ENABLE_FEATURE_IPV6
	else if (af == AF_INET6) {
	    struct in6_addr *in6;
	    struct ip6_rthdr *rth;

	    /*  IPV6_RTHDR_TYPE_0 length is 8   */
	    rtbuf_len = 8 + num_gateways * sizeof (*in6);
	    rtbuf = xmalloc (rtbuf_len);

	    rth = (struct ip6_rthdr *) rtbuf;
	    rth->ip6r_nxt = 0;
	    rth->ip6r_len = 2 * num_gateways;
	    rth->ip6r_type = ipv6_rthdr_type;
	    rth->ip6r_segleft = num_gateways;

	    *((uint32_t *) (rth + 1)) = 0;

	    in6 = (struct in6_addr *) (rtbuf + 8);
	    for (i = 0; i < num_gateways; i++)
		    memcpy (&in6[i], &gates[i].sin6.sin6_addr, sizeof (*in6));
	}
#endif
	/* no opt if !CONFIG_FEATURE_TRACEROUTE_SPEC_GATEWAYS */
#endif
}

/*      Command line stuff          */

#ifdef CONFIG_FEATURE_TRACEROUTE_SPEC_GATEWAYS
static int add_gateway (char *arg) {

	if (num_gateways >= MAX_GATEWAYS_6) {   /*  127 > 8 ... :)   */
		fprintf (stderr, "Too many gateways specified.");
		return -1;
	}

	gateways = xrealloc (gateways, (num_gateways + 1) * sizeof (*gateways));
	gateways[num_gateways++] = xstrdup (arg);

	return 0;
}
#endif

static void set_port (unsigned int *up, char *arg, char opt) {
	char *q;

	if(arg == NULL) return;
	*up = strtoul (arg, &q, 0);
	if (q == arg) {
	    struct servent *s = getservbyname (arg, NULL);

	    if (!s)  ex_error("cannot find service -%c '%s'", opt, arg);
	    *up = ntohs (s->s_port);
	}
}

static void set_raw (char *arg)
{
	module = "raw";
	O = llist_add_to(O, xasprintf ("protocol=%s", arg));
}

static void set_wait_specs (char *arg) {
	char *p, *q;

	if(arg == NULL) return;
	here_factor = near_factor = 0;

	wait_secs = strtod (p = arg, &q);
	if (q == p || wait_secs < 0) ex_error ("invalid -w wait seconds");
	if (!*q++)  return;

	here_factor = strtod (p = q, &q);
	if (q == p || here_factor < 0)  ex_error ("invalid here factor in option -w waitseconds,here");
	if (!*q++)  return;

	near_factor = strtod (p = q, &q);
	if (q == p || *q || near_factor < 0)  ex_error ("invalid near_factor in option -w waitseconds,here,near");
}


/*      PRINT  STUFF        */

static void print_header (void) {

	/*  Note, without ending new-line!  */
	char addr2str_buf[INET_4_6_ADDRSTRLEN];
	printf ("traceroute to %s (%s), %u hops max, %zu byte packets",
				dst_name, addr2str (&dst_addr, addr2str_buf), max_hops,
				g_header_len + data_len);
	fflush (stdout);
}

static void print_addr (sockaddr_any *res) {
	const char *str;
	char addr2str_buf[INET_4_6_ADDRSTRLEN];

	if (!res->sa.sa_family)
		return;

	str = addr2str (res, addr2str_buf);

	if ((bss_flags & OPT_n))
		printf (" %s", str);
	else {
	    char buf_print_addr[1024];

	    buf_print_addr[0] = '\0';
	    getnameinfo (&res->sa, sizeof (*res), buf_print_addr, sizeof (buf_print_addr),
							    0, 0, NI_IDN);
	    printf (" %s (%s)", buf_print_addr[0] ? buf_print_addr : str, str);
	}
}

static void print_probe (probe *pb) {
	unsigned int idx = (pb - probes);
	unsigned int ttl = idx / probes_per_hop + 1;
	unsigned int np = idx % probes_per_hop;

	if (np == 0)
		printf ("\n%2u ", ttl);

	if (!pb->res.sa.sa_family)
		printf (" *");
	else {
	    int prn = !np;      /*  print if the first...  */

	    if (np) {       /*  ...and if differs with previous   */
		probe *p;

		/*  skip expired   */
		for (p = pb - 1; np && !p->res.sa.sa_family; p--, np--) ;

		if (!np ||
		    !equal_addr (&p->res, &pb->res) ||
		    (p->ext != pb->ext &&
			!(p->ext && pb->ext && !strcmp (p->ext, pb->ext))) ||
		    (backward && p->recv_ttl != pb->recv_ttl)
		)  prn = 1;
	    }

	    if (prn) {
		print_addr (&pb->res);

		if (pb->ext)  printf (" <%s>", pb->ext);

		if (backward && pb->recv_ttl) {
		    int hops = ttl2hops (pb->recv_ttl);
		    if (hops != (int)ttl)  printf (" '-%d'", hops);
		}
	    }
	}

	if (pb->recv_time) {
	    double diff = pb->recv_time - pb->send_time;

	    printf ("  %.3f ms", diff * 1000);
	}

	if (pb->err_str[0])
		printf (" %s", pb->err_str);


	fflush (stdout);
}

/*      Compute  timeout  stuff         */

static double get_timeout (probe *pb) {
	double value;

	if (here_factor) {
	    /*  check for already replied from the same hop   */
	    unsigned int i;
	    int idx = (pb - probes);
	    probe *p = &probes[idx - (idx % probes_per_hop)];

	    for (i = 0; i < probes_per_hop; i++, p++) {
		/*   `p == pb' skipped since  !pb->done   */

		if (p->done && (value = p->recv_time - p->send_time) > 0) {
		    value += DEF_WAIT_PREC;
		    value *= here_factor;
		    return  value < wait_secs ? value : wait_secs;
		}
	    }
	}

	if (near_factor) {
	    /*  check forward for already replied   */
	    probe *p, *endp = probes + num_probes;

	    for (p = pb + 1; p < endp && p->send_time; p++) {

		if (p->done && (value = p->recv_time - p->send_time) > 0) {
		    value += DEF_WAIT_PREC;
		    value *= near_factor;
		    return  value < wait_secs ? value : wait_secs;
		}
	    }
	}

	return wait_secs;
}

/*      Check  expiration  stuff        */

static void check_expired (probe *pb) {
	int idx = (pb - probes);
	probe *p, *endp = probes + num_probes;
	probe *fp = NULL, *pfp = NULL;

	if (!pb->done)      /*  an ops method still not release it  */
	    return;

	/*  check all the previous in the same hop   */
	for (p = &probes[idx - (idx % probes_per_hop)]; p < pb; p++) {

	    if (!p->done ||     /*  too early to decide something  */
		!p->final       /*  already ttl-exceeded in the same hop  */
	    )  return;

	    pfp = p;    /*  some of the previous probes is final   */
	}

	/*  check forward all the sent probes   */
	for (p = pb + 1; p < endp && p->send_time; p++) {

	    if (p->done) {      /*  some next probe already done...  */
		if (!p->final)  /*  ...was ttl-exceeded. OK, we are expired.  */
		    return;
		else {
		    fp = p;
		    break;
		}
	    }
	}

	if (!fp)    /*  no any final probe found. Assume expired.   */
	    return;

	/*  Well. There is a situation "*(this) * * * * ... * * final"
	   We cannot guarantee that "final" is in its right place.
	   We've sent "sim_probes" simultaneously, and the final hop
	   can drop some of them and answer only for latest ones.
	   If we can detect/assume that it so, then just put "final"
	   to the (pseudo-expired) "this" place.
	*/

	/*  It seems that the case of "answers for latest ones only"
	   occurs mostly with icmp_unreach error answers ("!H" etc.).
	   Icmp_echoreply, tcp_reset and even icmp_port_unreach looks
	   like going in the right order.
	*/
	if (!fp->err_str[0])    /*  not an icmp_unreach error report...  */
		return;

	if (pfp || (idx % probes_per_hop) + (fp - pb) < probes_per_hop) {
	    /*  Either some previous (pfp) or some next probe
		in this hop is final. It means that the whole hop is final.
		Do the replace (it also causes further "final"s to be shifted
		here too).
	    */
	    goto  replace_by_final;
	}

	/*  If the final probe is an icmp_unreachable report
	    (either in a case of some error, like "!H", or just port_unreach),
	    it could follow the "time-exceed" report from the *same* hop.
	*/
	for (p = pb - 1; p >= probes; p--) {
	    if (equal_addr (&p->res, &fp->res)) {
		/*  ...Yes. Put "final" to the "this" place.  */
		goto  replace_by_final;
	    }
	}

	if (fp->recv_ttl) {
	    /*  Consider the ttl value of the report packet and guess where
		the "final" should be. If it seems that it should be
		in the same hop as "this", then do replace.
	    */
	    int back_hops, ttl;

	    /*  We assume that the reporting one has an initial ttl value
		of either 64, or 128, or 255. It is most widely used
		in the modern routers and computers.
		The idea comes from tracepath(1) routine.
	    */
	    back_hops = ttl2hops (fp->recv_ttl);

	    /*  It is possible that the back path differs from the forward
		and therefore has different number of hops. To minimize
		such an influence, get the nearest previous time-exceeded
		probe and compare with it.
	    */
	    for (p = pb - 1; p >= probes; p--) {
		if (p->done && !p->final && p->recv_ttl) {
		    int hops = ttl2hops (p->recv_ttl);

		    if (hops < back_hops) {
			ttl = (p - probes) / probes_per_hop + 1;
			back_hops = (back_hops - hops) + ttl;
			break;
		    }
		}
	    }

	    ttl = idx / probes_per_hop + 1;
	    if (back_hops == ttl)
		/*  Yes! It seems that "final" should be at "this" place   */
		goto  replace_by_final;
	    else if (back_hops < ttl)
		/*  Hmmm... Assume better to replace here too...  */
		goto  replace_by_final;

	}

	/*  No idea what to do. Assume expired.  */

	return;

replace_by_final:

	*pb = *fp;

	memset (fp, 0, sizeof (*fp));
	/*  block extra re-send  */
	fp->send_time = 1.;
}

static void poll_callback (int fd, int revents) {

	ops->recv_probe (fd, revents);
}

static void do_it (void) {
	unsigned int start = (first_hop - 1) * probes_per_hop;
	unsigned int end = num_probes;
	double last_send = 0;

	print_header ();

	while (start < end) {
	    unsigned int n, num = 0;
	    double next_time = 0;
	    double now_time = get_time ();

	    for (n = start; n < end; n++) {
		probe *pb = &probes[n];

		if (n == start &&               /*  probably time to print...  */
		    !pb->done && pb->send_time  /*  ...but yet not replied   */
		) {
		    double expire_time = pb->send_time + get_timeout (pb);

		    if (expire_time > now_time)
			    next_time = expire_time;
		    else {
			probe_done (pb);
			check_expired (pb);
		    }
		}

		if (pb->done) {

		    if (n == start) {   /*  can print it now   */
			print_probe (pb);
			start++;
		    }

		    if (pb->final)
			end = (n / probes_per_hop + 1) * probes_per_hop;

		    continue;
		}


		if (!pb->send_time) {
		    int ttl;
		    double next;

		    if (send_secs && (next = last_send + send_secs) > now_time) {
			next_time = next;
			break;
		    }

		    ttl = n / probes_per_hop + 1;

		    ops->send_probe (pb, ttl);

		    if (!pb->send_time) {
			if (next_time)  break;  /*  have chances later   */
			else  bb_perror_msg_and_die ("send probe");
		    }

		    last_send = pb->send_time;
		}

		if (!next_time)
		    next_time = pb->send_time + get_timeout (pb);

		num++;
		if (num >= sim_probes)  break;
	    }

	    if (next_time) {
		double timeout = next_time - get_time ();

		if (timeout < 0)  timeout = 0;

		do_poll (timeout, poll_callback);
	    }

	}

	printf ("\n");
}

static const tr_module default_ops = {
	.name = bb_default,
	.init = udp_default_init,
	.send_probe = udp_send_probe,
	.recv_probe = udp_recv_probe,
	.header_len = sizeof (struct udphdr),
};

static const tr_module raw_ops = {
	.name = "raw",
	.init = raw_init,
	.send_probe = raw_send_probe,
	.recv_probe = raw_recv_probe,
	.one_per_time = 1,
};

static const tr_module * const base[] =
{
	&default_ops, &raw_ops,
#ifdef CONFIG_FEATURE_TRACEROUTE_USE_DCCP
	&dccp_ops,
#endif /* CONFIG_FEATURE_TRACEROUTE_USE_DCCP */
#ifdef CONFIG_FEATURE_TRACEROUTE_USE_ICMP
	&icmp_ops,
#endif /* CONFIG_FEATURE_TRACEROUTE_USE_ICMP */
#ifdef CONFIG_FEATURE_TRACEROUTE_USE_TCP
	&tcp_ops,
#endif /* CONFIG_FEATURE_TRACEROUTE_USE_TCP */
#ifdef CONFIG_FEATURE_TRACEROUTE_USE_TCPCONN
	&tcp_ops_tcpconn,
#endif /* CONFIG_FEATURE_TRACEROUTE_USE_TCPCONN */
#ifdef CONFIG_FEATURE_TRACEROUTE_USE_UDP
	&udp_ops,
#endif /* CONFIG_FEATURE_TRACEROUTE_USE_UDP */
#ifdef CONFIG_FEATURE_TRACEROUTE_USE_UDPL
	&udplite_ops,
#endif /* CONFIG_FEATURE_TRACEROUTE_USE_UDPL */
	NULL
};

static const tr_module *tr_get_module (const char *name) {
	size_t i;

	if (!name)  return NULL;

	for (i = 0; base[i] != NULL; i++) {
	    if (!strcasecmp (name, base[i]->name))
		    return base[i];
	}

	return NULL;
}


int traceroute_main(void)
{
    unsigned long o;
    char *s_first_hop = NULL, *s_max_hops = NULL, *s_sim_probes = NULL;
    char *s_port = NULL, *s_tos = NULL, *s_w = NULL, *s_probes_per_hop = NULL;
    char *s_arg = NULL, *s_z = NULL, *s_S = NULL, *s_P = NULL;
#if ENABLE_FEATURE_IPV6
    char *s_flow_label = NULL;
#endif
#if defined(SO_MARK) && defined(CONFIG_FEATURE_TRACEROUTE_FWMARK)
    char *s_a = NULL;
#endif
#if ENABLE_FEATURE_LONG_OPTIONS
    static const struct option longopts[] =
    {
	USE_FEATURE_TRACEROUTE_USE_ICMP({"icmp", 0, NULL, 'I'},)
	USE_FEATURE_TRACEROUTE_USE_TCP({"tcp", 0, NULL, 'T'},)
	USE_FEATURE_TRACEROUTE_USE_UDP({"udp", 0, NULL, 'U'},)
	USE_FEATURE_TRACEROUTE_USE_DCCP({"dccp", 0, NULL, 'D'},)
	USE_FEATURE_TRACEROUTE_DEBUG({"debug", 0, NULL, 'd'},)
	{"dont-fragment", 0, NULL, 'F'},
	{"first", 1, NULL, 'f'},
	{"gateway", 1, NULL, 'g'},
	{"interface", 1, NULL, 'i'},
	{"max-hops", 1, NULL, 'm'},
	{"sim-queries", 1, NULL, 'N'},
	{"port", 1, NULL, 'p'},
	{"tos", 1, NULL, 't'},
	{"flowlabel", 1, NULL, 'l'},
	{"wait", 1, NULL, 'w'},
	{"queries", 1, NULL, 'q'},
	{"source", 1, NULL, 's'},
	{"sendwait", 1, NULL, 'z'},
	{"extensions", 0, NULL, 'e'},
	{"version", 0, NULL, 'V'},
	{"sport", 1, NULL, 'S'},
	{"fwmark", 1, NULL, 'a'},
	{"module", 1, NULL, 'M'},
	{"options", 1, NULL, 'O'},
	{"protocol", 1, NULL, 'P'},
	{"mtu", 0, NULL, 'c'},
	{"back", 0, NULL, 'b'},
	{NULL, 0, NULL, 0}
    };
	bb_applet_long_options = longopts;
#endif
	setlocale (LC_ALL, "");
	setlocale (LC_NUMERIC, "C");    /*  avoid commas in msec printed  */
#if defined(CONFIG_FEATURE_TRACEROUTE_USE_TCP) || defined(CONFIG_FEATURE_TRACEROUTE_USE_DCCP)
	init_random_seq();
#endif

	check_progname ();
	module = bb_default;
	first_hop = 1;
	max_hops = DEF_HOPS;
	sim_probes = DEF_SIM_PROBES;
	probes_per_hop = DEF_NUM_PROBES;
	can_connect = -1;
	wait_secs = DEF_WAIT_SECS;
	here_factor = DEF_HERE_FACTOR;
	near_factor = DEF_NEAR_FACTOR;
	send_secs = DEF_SEND_SECS;
#ifdef CONFIG_FEATURE_TRACEROUTE_USE_ICMP
	seq_icmp = 1;
#endif
	packet_len = -1;
	protocol_raw = DEF_RAW_PROT;
	raw_sk = -1;
	protocol_udp = IPPROTO_UDP;
#if defined(CONFIG_FEATURE_TRACEROUTE_USE_TCPCONN) || defined(CONFIG_FEATURE_TRACEROUTE_USE_ICMP)
	icmp_sk = -1;
#endif
#if ENABLE_FEATURE_IPV6
	ipv6_rthdr_type = 2;        /*  IPV6_RTHDR_TYPE_2   */
#endif

	o = bb_xgetopt_ulflags(opt_str, &s_first_hop,
#ifdef CONFIG_FEATURE_TRACEROUTE_SPEC_GATEWAYS
				add_gateway,
#endif
				&device, &s_max_hops, &s_sim_probes, &s_port, &s_tos,
				USE_FEATURE_IPV6(&s_flow_label,) &s_w,
				&s_probes_per_hop, &s_arg, &s_z, &module, &s_S, &O, &s_P
#if defined(SO_MARK) && defined(CONFIG_FEATURE_TRACEROUTE_FWMARK)
				,&s_a
#endif
				);


	if((o & OPT_4)) af = AF_INET;
	USE_FEATURE_IPV6(if((o & OPT_6)) af = AF_INET6;)
	USE_FEATURE_TRACEROUTE_DEBUG(debug = o & OPT_DEBUG;)
	dontfrag = o & OPT_DONTFRAG;
	if(s_max_hops != NULL) {
		max_hops = bb_xparse_ui_number(BB_PN_SET_LOWER1|BB_PN_SET_UPPER|BB_PN_ERR_STR1,
			s_max_hops, MAX_HOPS, "max hops cannot be more than " _TEXT(MAX_HOPS));
	}
	if(s_first_hop != NULL) {
		first_hop = bb_xparse_i_number(BB_PN_SET_LOWER1|BB_PN_SET_UPPER|BB_PN_ERR_STR1,
			s_first_hop, max_hops, "first hop out of range");
	}
	USE_FEATURE_TRACEROUTE_USE_ICMP(if((o & OPT_ICMP)) module = "icmp";)
	USE_FEATURE_TRACEROUTE_USE_TCP(if((o & OPT_TCP)) module = "tcp";)
	USE_FEATURE_TRACEROUTE_USE_UDP(if((o & OPT_UDP)) module = "udp";)
	USE_FEATURE_TRACEROUTE_USE_UDPL(if((o & OPT_UDPL)) module = "udplite";)
	USE_FEATURE_TRACEROUTE_USE_DCCP(if((o & OPT_DCCP)) module = "dccp";)
	if(s_sim_probes != NULL)
		sim_probes = bb_xparse_ui_number(BB_PN_ERR_STR1, s_sim_probes, "invalid value -m '%s'");
	set_port(&dst_port_seq, s_port, 'p');
	if(s_tos != NULL)
		tos = bb_xparse_ui_number(BB_PN_ERR_STR1, s_tos, "invalid value -t '%s'");
	set_wait_specs(s_w);
	if(s_probes_per_hop != NULL) {
		probes_per_hop = bb_xparse_ui_number(BB_PN_SET_LOWER1|BB_PN_SET_UPPER|BB_PN_ERR_STR1,
		  s_probes_per_hop, MAX_PROBES, "invalid value -q '%s', need 1<>" _TEXT(MAX_PROBES));
	}
	noroute = o & OPT_r;
	if(s_arg != NULL && getaddr (s_arg, &src_addr))
		return 2;
	set_send_secs(s_z);
	set_port(&src_port, s_S, 'S');
#if defined(SO_MARK) && defined(CONFIG_FEATURE_TRACEROUTE_FWMARK)
	if(s_a != NULL)
		fwmark = bb_xparse_ui_number(BB_PN_ERR_STR1, s_a, "invalid value fwmark '%s'");
#endif
	mtudisc = o & OPT_c;
	backward = o & OPT_b;
	if(s_P != NULL) set_raw(s_P);

	if((o & OPT_V)) {
		fprintf(stderr, "%s\n", version_string);
		return 0;
	}
	if((o & OPT_h) || Gargc == 0 || Gargc > 2) bb_show_usage();

	if (getaddr (*Gargv, &dst_addr) < 0)
		return 1;
	dst_name = *Gargv;
	/*  i.e., guess it by the addr in cmdline...  */
	if (!af)  af = dst_addr.sa.sa_family;

	if(Gargc == 2)
		packet_len = bb_xparse_i_number(BB_PN_SET_UPPER|BB_PN_ERR_STR1,
				Gargv[1], MAX_PACKET_LEN, "invalid packet_len value '%s'");

	ops = tr_get_module (module);
	if (!ops)  ex_error ("Unknown traceroute module %s", module);

	if (src_addr.sa.sa_family && src_addr.sa.sa_family != af)
		ex_error ("IP version mismatch in addresses specified");
	if (send_secs >= 10)    /*  it is milliseconds   */
		send_secs /= 1000;

#if ENABLE_FEATURE_IPV6
	if(s_flow_label != NULL)
		flow_label = bb_xparse_ui_number(BB_PN_SET_UPPER|BB_PN_ERR_STR1, s_flow_label, 0xfffff, "invalid value -t '%s'");
	if (af == AF_INET6 && (tos || flow_label))
		dst_addr.sin6.sin6_flowinfo =
		    htonl (((tos & 0xff) << 20) | (flow_label & 0x000fffff));
#endif

	if (src_port) {
	    src_addr.sin.sin_port = htons ((uint16_t) src_port);
	    src_addr.sa.sa_family = af;
	}

	if (src_port || ops->one_per_time) {
		sim_probes = 1;
		here_factor = near_factor = 0;
	}

	/*  make sure we don't std{in,out,err} to open sockets  */
	make_fd_used (0);
	make_fd_used (1);
	make_fd_used (2);

	init_ip_options ();

#if ENABLE_FEATURE_IPV6
	g_header_len = (af == AF_INET ? sizeof (struct iphdr)
				    : sizeof (struct ip6_hdr)) +
			rtbuf_len + ops->header_len;
#else
	g_header_len = sizeof (struct iphdr) + rtbuf_len + ops->header_len;
#endif

	if (mtudisc) {
	    dontfrag = 1;
	    sim_probes = 1;
	    if (packet_len < 0)
		    packet_len = MAX_PACKET_LEN;
	}

	if (packet_len < 0) {
	    if (DEF_DATA_LEN >= ops->header_len)
		    data_len = DEF_DATA_LEN - ops->header_len;
	} else {
	    if ((size_t)packet_len >= g_header_len)
		    data_len = packet_len - g_header_len;
	}

	num_probes = max_hops * probes_per_hop;
	probes = zmalloc (num_probes * sizeof (*probes));

	if (ops->init (&dst_addr, dst_port_seq, &data_len) < 0)
		ex_error ("trace method's init failed");

	do_it ();

	return 0;
}
#endif /* APPLET_USAGE */
