/*
 * pidof implementation for busybox
 *
 * Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 *
 * Licensed under the GPL v2, see the file LICENSE in this tarball.
 */
#ifdef APPLET_USAGE
#if (ENABLE_FEATURE_PIDOF_SINGLE || ENABLE_FEATURE_PIDOF_OMIT)
#define USAGE_PIDOF "Options:"
#else
#define USAGE_PIDOF "\n\tThis version of pidof accepts no options."
#endif

#define pidof_trivial_usage \
	"process-name [OPTION] [process-name ...]"
#define pidof_full_usage \
	"Lists the PIDs of all processes with names that match the\n" \
	"names on the command line.\n" \
	USAGE_PIDOF \
	USE_FEATURE_PIDOF_SINGLE("\n\t-s\t\tdisplay only a single PID") \
	USE_FEATURE_PIDOF_OMIT("\n\t-o\t\tomit given pid.") \
	USE_FEATURE_PIDOF_OMIT("\n\t\t\tUse %PPID to omit the parent pid of pidof itself")
#else

#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include "busybox.h"

#if ENABLE_FEATURE_PIDOF_SINGLE
#define _SINGLE_COMPL(a) a
#define SINGLE (1<<0)
#else
#define _SINGLE_COMPL(a)
#define SINGLE (0)
#endif

#if ENABLE_FEATURE_PIDOF_OMIT
#define OMIT (SINGLE + 1)
#else
#define OMIT 0
#define omitted 0
#endif

static long *pidlist_reverse(long *pidList)
{
	int i=0;
	while (pidList[i] > 0 && ++i);
	if ( i-- > 0) {
		long k;
		int j;
		for (j = 0; i > j; i--, j++) {
			k = pidList[i];
			pidList[i] = pidList[j];
			pidList[j] = k;
		}
	}
	return pidList;
}


extern int pidof_main(void)
{
	char **argv;
	unsigned n = 0;
	unsigned fail = 1;
	unsigned long int opt;

	/* do unconditional option parsing */
#if ENABLE_FEATURE_PIDOF_OMIT
	llist_t *omits = NULL; /* list of pids to omit */
	llist_t * omits_p;
	char getppid_str[32];

	opt = bb_xgetopt_ulflags(_SINGLE_COMPL("s") "o:\0o::", &omits);
#else
	opt = bb_getopt_ulflags(_SINGLE_COMPL("s"));
#endif
	argv = Gargv;
#if ENABLE_FEATURE_PIDOF_OMIT
	/* fill omit list.  */
	for (omits_p = omits; omits_p != NULL; omits_p = omits_p->link) {
		/* are we asked to exclude the parent's process ID? */
		if (!strncmp(omits_p->data, "%PPID", 5)) {
			snprintf(getppid_str, sizeof(getppid_str), "%d", getppid());
			omits_p->data = getppid_str;
#if 0
		} else {
			bb_error_msg_and_die("illegal omit pid value (%s)!\n",
						omits_p->data);
#endif
		}
	}
#endif
	/* Looks like everything is set to go.  */
	while(*argv != NULL) {
		long *pidList;
		long *pl;

		/* reverse the pidlist like GNU pidof does.  */
		pidList = pidlist_reverse(find_pid_by_name(*argv));
		argv++;
		for(pl = pidList; *pl > 0; pl++) {
#if ENABLE_FEATURE_PIDOF_OMIT
			unsigned omitted = 0;
			if (opt & OMIT) {
				omits_p = omits;
				while (omits_p)
					if (bb_xparse_l_number(10 | BB_PN_SILENT_RE, omits_p->data) == *pl) {
						omitted = 1; break;
					} else
						omits_p = omits_p->link;
			}
#endif
			if (!omitted) {
				if (n) {
					putchar(' ');
				} else {
					n = 1;
				}
				printf("%ld", *pl);
			}
			fail = (!ENABLE_FEATURE_PIDOF_OMIT && omitted);

			if (ENABLE_FEATURE_PIDOF_SINGLE && (opt & SINGLE))
				break;
		}
		free(pidList);
	}
	putchar('\n');

	return fail ? EXIT_FAILURE : EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
