/*
 * Mini ps implementation(s) for busybox
 *
 * Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 *
 * Licensed under the GPL v2, see the file LICENSE in this tarball.
 */
#ifdef APPLET_USAGE
#if !ENABLE_FEATURE_PS_WIDE
#define USAGE_PS "\n\tThis version of ps accepts no options."
#else
#define USAGE_PS "\nOptions:"
#endif

#define ps_trivial_usage \
	""
#define ps_full_usage \
	"Report process status\n" \
	USAGE_PS \
	USE_FEATURE_PS_WIDE("\n\tw\twide output")
#else

#include "busybox.h"
#include <unistd.h>
#include <dirent.h>
#include <ctype.h>
#include <string.h>
#include <termios.h>
#include <sys/ioctl.h>
#include "procps.h"

extern int ps_main(void)
{
	procps_status_t *p, procps_status;
	int i, len;
	char *d = NULL;

#if ENABLE_FEATURE_PS_WIDE
	int terminal_width;
	int w_count = 0;

#else
# define terminal_width 79
#endif

#if ENABLE_FEATURE_PS_WIDE
	/* handle arguments */
	bb_xgetopt_ulflags("w\0-:ww", &w_count);
#endif

#if ENABLE_FEATURE_PS_WIDE
	/* if w is given once, GNU ps sets the width to 132,
	 * if w is given more than once, it is "unlimited"
	 */
	if(w_count) {
		terminal_width = (w_count==1) ? 132 : INT_MAX;
	} else {
		get_terminal_width_height(&terminal_width, NULL);
		/* Go one less... */
		terminal_width--;
	}
#endif

	printf("  PID  Uid     VmSize Stat Command\n");

	while ((p = procps_scan(1, &procps_status, &d)) != 0)  {
		char *namecmd = p->cmd;

		if(p->rss == 0)
		    len = printf("%5d %-8s        %s ", p->pid, p->user, p->state);
		  else
		    len = printf("%5d %-8s %6ld %s ", p->pid, p->user, p->rss, p->state);

		i = terminal_width-len;

		if(namecmd && namecmd[0]) {
			if(i < 0)
				i = 0;
			if(strlen(namecmd) > (size_t)i)
				namecmd[i] = 0;
			printf("%s\n", namecmd);
		} else {
			namecmd = p->short_cmd;
			if(i < 2)
				i = 2;
			if(strlen(namecmd) > ((size_t)i-2))
				namecmd[i-2] = 0;
			printf("[%s]\n", namecmd);
		}
	}
	return EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
