#!/bin/sh

# Copyright (C) 2013-2018 by Vladimir Oleynik <dzo@simtreas.ru>
# may safely be consumed by a GPL license.
#
# autoconfigure:
# detect native 64 or 32 target (cross!) bit system
# detect (cross!) the off_t size bits
# detect (cross!) the nlink_t_MAX for "du" applet
# set the printf format modificator for int-64
# set the printf format modificator for off_t
# set the printf format modificator for st_size field of struct stat
# define macros: BB_fseeko and BB_ftello BB_strtoll
# tst crypt() dprintf() strchrnul() shm*() getpass() stpcpy() mempcpy()
#  tcdrain() ether_aton() ether_hostton() sysinfo() mntent()
#  sethostname() clearenv() killpg() (set/get/end)usershell() w*(ctype)()
#  strlcpy() strlcat() isnan() login() ftw() nftw() ffs() posix_memalign()
# tst MAXHOSTNAMELEN from <sys/param.h> or <netdb.h> or set 64
# tst <arpa/telnet.h> and <arpa/ftp.h>
# tst usleep() return value
# tst have timezone and daylight variables in libc
# tst regex in libc, tst re_match(GNU) for mc.c
# tst zlib
# detect MUSL and __fseterr
# tst getifaddrs() freeifaddrs()
# tst have the 'gecos' field in 'struct passwd'
# tst have #include <linux/ip_vs.h>, #include <linux/ppp-comp.h>,
#  #include <sys/personality.h>
#  #include <langinfo.h> and nl_langinfo()
# tst have setns() adjtimex() getrandom() explicit_bzero()
# tst int==long and intmax_t==long

PWD0=`pwd`
cd $1 || exit 1
shift
OUTCFG=$1
shift
C="$@"

DEBUG_AC=0 # set to 1 if need debugging bb_ac.sh

OUTCFG_NEW=$OUTCFG-new

. ../.config || exit 1

if [ x"$CONFIG_STATIC" = xy ]; then
       C="$C -static"
fi

cct() {
	local __make_name=$1
	local is_ok=1 is_fail=0
	local out a1 err __ret_val exit_error r

	shift
	if [ x"$__make_name" = x-t ]; then
		__make_name=out
	elif [ x"$1" != x-1 ]; then
		is_ok=$1
		shift
		is_fail=$1
		shift
	else
		shift
	fi
	if [ x"$1" = x-e -o x"$1" = x-w ]; then
		exit_error=x$1
		shift
		err=$1
		shift
	fi
	a1=$1
	shift

	out=$PWD0/bb_ac_a.out

	if [ 1 -eq "$DEBUG_AC" ]; then
		echo "$C -o $out $a1.c $@"
		$C -o "$out" "$a1".c "$@"
	else
		$C -o "$out" "$a1".c "$@" > /dev/null 2> /dev/null
	fi

	r=$?
	rm -f "$out"
	if [ $r -eq 0 ]; then
		__ret_val=$is_ok
	else
		__ret_val=$is_fail

	fi
	if [ "$__ret_val" = 0 -a -n "$exit_error" ]; then
		[ $exit_error = x-w ] && echo -n 'WARNING: ' 1>&2
		echo "$err" 1>&2
		[ $exit_error = x-e ] && exit 1
	fi
	eval $__make_name=\$__ret_val
	return $r
}


# Test target C-int bits
cct SYSTEM_IS 0 int_64 tst_64_32 -DTST=int -DTST_B=8
if [ $SYSTEM_IS = 0 ]; then
  cct SYSTEM_IS 0 long_64 tst_64_32 -DTST=long -DTST_B=8
  if [ $SYSTEM_IS = 0 ]; then
    cct SYSTEM_IS 0 int_32 -e 'Can not detect 32 or 64 bit system' tst_64_32 -DTST=int -DTST_B=4
    cct SYSTEM_IS 0 $SYSTEM_IS -e 'Target compiller unsupport 64-bit "long long" type' tst_64_32 "-DTST=long long" -DTST_B=8
  fi
fi

cct -t -e 'Target compiller unsupport "long long" type' tst_long_long

if [ $SYSTEM_IS = int_64 ]; then
   cct TYPE_32T 0 short -e 'Can not detect 32 bit integer' tst_64_32 -DTST=short -DTST_B=4
fi
if [ $SYSTEM_IS = long_64 ]; then
   cct TYPE_32T 0 int tst_64_32 -DTST=int -DTST_B=4
   if [ $TYPE_32T = 0 ]; then
	cct TYPE_32T 0 short -e 'Can not detect 32 bit integer' tst_64_32 -DTST=short -DTST_B=4
   fi
fi

echo The target system is $SYSTEM_IS

cct INT_IS_LONG 0 1 tst_intmax_t -DTST_1=int -DTST_2=long
cct INTMAX_IS_LONG 0 1 tst_intmax_t -DTST_1=intmax_t -DTST_2=long
cct LL_IS_LONG 0 1 tst_intmax_t -DTST_1='long long' -DTST_2=long

# test nlink_t size for XXX_MAX (need "du" applet)
cct NLINK_MAX 0 LONG_MAX tst_nlink_t -DNLINK_T_LIKE=long
if [ $NLINK_MAX = 0 ]; then
  cct NLINK_MAX 0 INT_MAX tst_nlink_t -DNLINK_T_LIKE=int
  if [ $NLINK_MAX = 0 ]; then
    cct NLINK_MAX 0 SHRT_MAX tst_nlink_t -DNLINK_T_LIKE=short
    if [ $NLINK_MAX = 0 ]; then
      cct NLINK_MAX 0 CHAR_MAX -e 'Can not detect nlink_t size' tst_nlink_t -DNLINK_T_LIKE=char
    fi
  fi
fi


# test off_t
cct OFF_T 64 0 -e "Can not find 'off_t' in '#include <sys/types.h>'" tst_off_t

# test off_t is 64-bit in 32-bit system
if [ $SYSTEM_IS = int_32 ]; then
	cct OFF_T 64 32 tst_off_t64
fi

# test struct stat st_size member, may be not off_t
cct ST_SIZE_T 64 32 tst_st_size_t64


if [ "x$CONFIG_FEATURE_UTF8_SUPPORT" = xy ]; then
  cct -t -e 'Target system can not use multibyte function for UTF8 support. You need switch off CONFIG_FEATURE_UTF8_SUPPORT' tst_utf8
fi

cct HAVE_CRYPT -1 -w 'Target system have not crypt. Make all "password" is "true"!' tst_crypt -lcrypt -lm


MAXHOSTNAMELEN_NETDB=0
MAXHOSTNAMELEN_SET_CONST=0
cct MAXHOSTNAMELEN_SYS_PARAM_H -1 tst_mhnl_sp
if [ $MAXHOSTNAMELEN_SYS_PARAM_H -eq 0 ]; then
  cct MAXHOSTNAMELEN_SET_CONST 0 64 tst_mhnl_netdb
  if [ $MAXHOSTNAMELEN_SET_CONST -ne 0 ]; then
    echo Cannot find declare MAXHOSTNAMELEN, set $MAXHOSTNAMELEN_SET_CONST
  else
    MAXHOSTNAMELEN_NETDB=1
  fi
fi


HAVE_FDPRINTF=0
cct HAVE_DPRINTF -1 tst_dprintf -DTST=dprintf
if [ $HAVE_DPRINTF -eq 0 ]; then
  cct HAVE_FDPRINTF -1 -w 'Cannot find (f)dprintf' tst_dprintf -DTST=fdprintf
fi


cct HAVE_ISNAN -1 tst_isnan -lm
cct HAVE_STRCHRNUL -1 tst_strchrnul
cct HAVE_SHM -1 tst_shm
cct HAVE_ARPA_TELNET -1 tst_arpa_telnet
cct HAVE_ARPA_FTP -1 tst_arpa_ftp
cct HAVE_ARPA_TFTP -1 tst_arpa_tftp
cct HAVE_GETPASS -1 tst_getpass
cct HAVE_STRLCAT -1 tst_strlcat
cct HAVE_STRLCPY -1 tst_strlcpy
cct HAVE_STPCPY -1 tst_stpcpy
cct HAVE_MEMPCPY -1 tst_mempcpy
cct HAVE_TCDRAIN -1 tst_tcdrain
cct HAVE_ETHER_ATON -1 tst_ether_aton
cct HAVE_ETHER_HOSTTON -1 tst_ether_hostton
cct HAVE_SYSINFO -1 tst_sysinfo

cct HAVE_MNTENT -1 tst_mntent
HAVE_ADDMNTENT=$HAVE_MNTENT
if [ $HAVE_MNTENT -eq 1 ]; then
	cct HAVE_ADDMNTENT -1 tst_addmntent
fi

cct HAVE_SETHOSTNAME -1 tst_sethostname
cct HAVE_SETFSUID -1 tst_setfsuid
cct HAVE_GETPWENT -1 tst_getpwent
cct HAVE_GETGRENT -1 tst_getgrent
cct HAVE_USLEEP_RETURN_VALUE -1 tst_usleep
cct HAVE_TIMEZONE_DAYLIGHT -1 tst_tzd

cct HAVE_REGEX -1 tst_regex
if [ $HAVE_REGEX -eq 1 -a "x$CONFIG_MC" = xy ]; then
	cct HAVE_REGEX -1 tst_regex -DTST_RE_MATCH
fi

cct HAVE_CLEARENV -1 tst_clearenv
cct HAVE_KILLPG -1 tst_killpg
cct HAVE_USERSHELL -1 tst_usershell

if [ "x$CONFIG_FEATURE_DROPBEAR_ZLIB" = xy ]; then
  cct -t -e 'Target system have not zlib. You need switch off CONFIG_FEATURE_DROPBEAR_ZLIB' tst_zlib -lz
fi

cct __MUSL__ -1 tst_musl -DTST_MUSL_ONLY
cct HAVE___FSETERR -1 tst_musl

cct HAVE_LOGIN -1 tst_login
cct HAVE_GET_FREE_IFADDRS -1 tst_ifaddrs
cct HAVE_GECOS -1 tst_gecos
cct HAVE_LINUX_IP_VS_H -1 tst_ip_vs
cct HAVE_LINUX_PPP_COMP_H -1 tst_ppp_comp
cct HAVE_SYS_PERSONALITY_H -1 tst_sys_personality
cct HAVE_SYS_SWAP_H -1 tst_sys_swap

HAVE_FTW=1
cct HAVE_NFTW -1 tst_nftw
if [ $HAVE_NFTW -eq 0 ]; then
	cct HAVE_FTW -1 tst_ftw
fi

cct HAVE_FFS -1 tst_ffs

cct HAVE_POSIX_MEMALIGN -1 tst_posix_memalign
cct HAVE_LANGINFO -1 tst_langinfo

cct HAVE_SETNS -1 tst_setns
cct HAVE_ADJTIMEX -1 tst_adjtimex
cct HAVE_GETRANDOM -1 tst_getrandom
cct HAVE_EXPLICIT_BZERO -1 tst_explicit_bzero

# out config defines

cd "$PWD0"

out_cfg() {
	for i do
		echo "$i"
	done >> $OUTCFG_NEW
}

echo '/* autoconfigured, do not edit */' > $OUTCFG_NEW
out_cfg '#ifndef __BB_AC' '#define __BB_AC'

if [ $SYSTEM_IS = int_64 ]; then
	out_cfg '#define NATIVE_64 int' 'typedef NATIVE_64 bb64_t;'
	out_cfg 'typedef unsigned NATIVE_64 ubb64_t;'
	out_cfg "typedef $TYPE_32T bb32_t;"
	out_cfg 'typedef unsigned $TYPE_32T ubb32_t;'
	out_cfg '#define P_BB64T ""' '#define P_OFFT "l"'
	out_cfg '#define BB_ftello(f) ftell((f))'
	out_cfg '#define BB_fseeko(f, o, t) fseek((f), (o), (t))'
	out_cfg '#define BB_strtoll(s, e, b) strtol((s), (e), (b))'
fi
if [ $SYSTEM_IS = long_64 ]; then
	out_cfg "#define NATIVE_64 long" "typedef NATIVE_64 bb64_t;"
	out_cfg 'typedef unsigned NATIVE_64 ubb64_t;'
	out_cfg "typedef $TYPE_32T bb32_t;"
	out_cfg "typedef unsigned $TYPE_32T ubb32_t;"
	out_cfg '#define P_BB64T "l"' '#define P_OFFT "l"'
	out_cfg '#define BB_ftello(f) ftell((f))'
	out_cfg '#define BB_fseeko(f, o, t) fseek((f), (o), (t))'
	if [ $LL_IS_LONG -eq 1 ]; then
		out_cfg '#define BB_strtoll(s, e, b) strtol((s), (e), (b))'
	else
		out_cfg '#define BB_strtoll(s, e, b) strtoll((s), (e), (b))'
	fi
fi
if [ $SYSTEM_IS = int_32 ]; then
	out_cfg "typedef long long bb64_t;"
	out_cfg 'typedef unsigned long long ubb64_t;'
	out_cfg "typedef int bb32_t;"
	out_cfg 'typedef unsigned int ubb32_t;'
	out_cfg '#define P_BB64T "ll"'
	if [ $OFF_T -eq 32 ]; then
		out_cfg '#define P_OFFT "l"'
		out_cfg '#define BB_ftello(f) ftell((f))'
		out_cfg '#define BB_fseeko(f, o, t) fseek((f), (o), (t))'
	else
		out_cfg '#define P_OFFT "ll"'
		out_cfg '#define BB_ftello(f) ftello((f))'
		out_cfg '#define BB_fseeko(f, o, t) fseeko((f), (o), (t))'
	fi
	out_cfg '#define BB_strtoll(s, e, b) strtoll((s), (e), (b))'
fi

out_cfg "#define NLINK_MAX $NLINK_MAX"


if [ $OFF_T -ne $ST_SIZE_T ]; then
	if [ $ST_SIZE_T -eq 32 ]; then
		out_cfg "typedef long bbst_size_t;"
		out_cfg '#define P_ST_SIZE "l"'
		out_cfg '#define BB_st_size_strtol(s, e, b) strtol((s), (e), (b))'
	else
		out_cfg "typedef bb64_t bbst_size_t;"
		out_cfg '#define P_ST_SIZE "ll"'
		out_cfg '#define BB_st_size_strtol(s, e, b) BB_strtoll((s), (e), (b))'
	fi
else
	out_cfg "typedef off_t bbst_size_t;"
	out_cfg "#define P_ST_SIZE P_OFFT"
	if [ $ST_SIZE_T -eq 32 ]; then
		out_cfg '#define BB_st_size_strtol(s, e, b) strtol((s), (e), (b))'
	else
		out_cfg '#define BB_st_size_strtol(s, e, b) BB_strtoll((s), (e), (b))'
	fi
fi

out_have() {
	local __make_name_have=$1
	local __val
	eval __val=\$$__make_name_have
	if [ "x$__val" = x1 ]; then
		out_cfg "#define $__make_name_have 1"
	else
		if [ x"$2" != x ]; then
			out_cfg "$2"
		fi
	fi
}

if [ $INT_IS_LONG -eq 1 ]; then
	out_cfg '#define bb_xparse_i_number   bb_xparse_l_number'
	out_cfg '#define bb_xparse_ui_number  bb_xparse_ul_number'
fi
if [ $INTMAX_IS_LONG -eq 1 ]; then
	out_cfg '#define bb_xparse_im_number  bb_xparse_l_number'
	out_cfg '#define bb_xparse_uim_number bb_xparse_ul_number'
fi

out_have HAVE_CRYPT
out_have MAXHOSTNAMELEN_SYS_PARAM_H
out_have MAXHOSTNAMELEN_NETDB

if [ $MAXHOSTNAMELEN_SET_CONST -ne 0 ]; then
	out_cfg '#ifndef MAXHOSTNAMELEN'
	out_cfg "#define MAXHOSTNAMELEN $MAXHOSTNAMELEN_SET_CONST"
	out_cfg '#endif'
fi

out_have HAVE_DPRINTF 'int dprintf(int fd, const char *fmt, ...);'
out_have HAVE_FDPRINTF
out_have HAVE_ISNAN
out_have HAVE_STRCHRNUL 'char *strchrnul(const char *s, int c);'
out_have HAVE_SHM
out_have HAVE_ARPA_TELNET
out_have HAVE_ARPA_FTP
out_have HAVE_ARPA_TFTP
out_have HAVE_GETPASS 'char *getpass(const char *prompt);'
out_have HAVE_STRLCAT
out_have HAVE_STRLCPY
out_have HAVE_STPCPY 'char *stpcpy(char * s1, const char * s2);'
out_have HAVE_MEMPCPY 'void *mempcpy(void * s1, const void * s2, size_t n);'
out_have HAVE_TCDRAIN
out_have HAVE_ETHER_ATON
out_have HAVE_ETHER_HOSTTON
out_have HAVE_SYSINFO

if [ "x$HAVE_MNTENT" = x1 ]; then
	out_cfg '#define HAVE_MNTENT 1'
	out_have HAVE_ADDMNTENT
else
	out_cfg '#ifdef NEED_MNTENT'
	out_cfg 'FILE *setmntent(const char *name, const char *mode);'
	out_cfg 'struct mntent *getmntent(FILE * filep);'
	out_cfg 'struct mntent *getmntent_r (FILE *filep, struct mntent *mnt, char *buff, int bufsize);'
	out_cfg 'int addmntent(FILE * filep, const struct mntent *mnt);'
	out_cfg 'int endmntent(FILE * filep);' '#endif'
fi

out_have HAVE_SETHOSTNAME
out_have HAVE_SETFSUID
out_have HAVE_GETPWENT
out_have HAVE_GETGRENT
out_have HAVE_USLEEP_RETURN_VALUE
out_have HAVE_TIMEZONE_DAYLIGHT
out_have HAVE_REGEX
out_have HAVE_CLEARENV 'int clearenv(void);'
out_have HAVE_KILLPG
out_have HAVE_USERSHELL
out_have __MUSL__
out_have HAVE___FSETERR
out_have HAVE_LOGIN
out_have HAVE_GET_FREE_IFADDRS
out_have HAVE_GECOS
out_have HAVE_LINUX_IP_VS_H
out_have HAVE_LINUX_PPP_COMP_H
out_have HAVE_SYS_PERSONALITY_H
out_have HAVE_SYS_SWAP_H
out_have HAVE_FTW
out_have HAVE_NFTW
out_have HAVE_FFS
out_have HAVE_POSIX_MEMALIGN
out_have HAVE_LANGINFO
out_have HAVE_SETNS
out_have HAVE_ADJTIMEX
out_have HAVE_GETRANDOM
out_have HAVE_EXPLICIT_BZERO

out_cfg '#endif /* __BB_AC */'

mv $OUTCFG_NEW $OUTCFG
