/* dmesg.c -- Print out the contents of the kernel ring buffer
 * Created: Sat Oct  9 16:19:47 1993
 * Revised: Thu Oct 28 21:52:17 1993 by faith@cs.unc.edu
 * Copyright 1993 Theodore Ts'o (tytso@athena.mit.edu)
 * This program comes with ABSOLUTELY NO WARRANTY.
 * Modifications by Rick Sladkey (jrs@world.std.com)
 * Larger buffersize 3 June 1998 by Nicolai Langfeldt, based on a patch
 * by Peeter Joot.  This was also suggested by John Hudson.
 * 1999-02-22 Arkadiusz Mikiewicz <misiek@misiek.eu.org>
 * - added Native Language Support
 *
 * from util-linux -- adapted for busybox by
 * Erik Andersen <andersen@codepoet.org>. I ripped out Native Language
 * Support, replaced getopt, added some gotos for redundant stuff.
 *
 * Audited and cleaned up on 7 March 2003 to reduce size of
 * check error handling by Erik Andersen <andersen@codepoet.org>
 */
#ifdef APPLET_USAGE
#define dmesg_trivial_usage \
	"[-c] [-n LEVEL] [-s SIZE]"
#define dmesg_full_usage \
	"Prints or controls the kernel ring buffer\n\n" \
	"Options:\n" \
	"\t-c\t\tClears the ring buffer's contents after printing\n" \
	"\t-n LEVEL\tSets console logging level\n" \
	"\t-s SIZE\t\tUse a buffer of size SIZE"
#else

#include "busybox.h"
#include <sys/klog.h>

int dmesg_main(void)
{
	char *buf;
	int i, n;
	int bufsize = 8196;
	int level = 0;
	char *s_level = NULL, *s_bs = NULL;
	int lastc;
	int cmd = 3;

	i = (int)bb_xgetopt_ulflags("n:s:c\0?0", &s_level, &s_bs);
	if(i & 4)
		cmd = 4;
	if(s_level) {
		cmd = 8;
		level = bb_xparse_i_number(10 | BB_PN_SET_UPPER | BB_PN_ERR_STR1,
						s_level, 10, "invalid level `%s'");
	}
	if(s_bs) {
		/* I think a 512k max kernel ring buffer is big enough for
		 * anybody, as the default is 16k...  Could be wrong though.
		 * If so I'm sure I'll hear about it by the enraged masses*/
		bufsize = bb_xparse_i_number(10 | BB_PN_SET_LOWER | BB_PN_SET_UPPER | BB_PN_ERR_STR1,
					s_bs, 4096, 512*1024, "failed to parse buffer size: '%s'");
	}

	if (cmd == 8) {
		if (klogctl(cmd, NULL, level) < 0)
			goto die_the_death;
		return EXIT_SUCCESS;
	}

	buf = xmalloc(bufsize);
	if ((n = klogctl(cmd, buf, bufsize)) < 0)
		goto die_the_death;

	lastc = '\n';
	for (i = 0; i < n; i++) {
		if (lastc == '\n' && buf[i] == '<') {
			i++;
			while (buf[i] >= '0' && buf[i] <= '9')
				i++;
			if (buf[i] == '>')
				i++;
		}
		lastc = buf[i];
		putchar(lastc);
	}
	if (lastc != '\n')
		putchar('\n');
	return EXIT_SUCCESS;
die_the_death:
	bb_perror_nomsg_and_die();
}
#endif /* APPLET_USAGE */
