/* vi: set sw=4 ts=4: */
/*
 * lspci implementation for busybox
 *
 * Copyright (C) 2009  Malek Degachi <malek-degachi@laposte.net>
 *
 * Licensed under GPLv2 or later, see file LICENSE in this source tree.
 */
#ifdef APPLET_USAGE
#define lspci_trivial_usage \
       "[-mk]"
#define lspci_full_usage \
       "List all PCI devices\n\n" \
       "Options:\n" \
	"\t-m\tParsable output\n" \
	"\t-k\tShow driver"
#else

#include "opt_gen.h"
OPT_GEN(
opt_simple(m),
opt_simple(k), /*non-compat:*/
opt_simple(n), /*ignored:*/
opt_simple(v)  /*ignored:*/
)
#include "opt_gen.h"

#include "busybox.h"
#include <string.h>
#include "config_read.h"

/* PCI_SLOT_NAME PCI_CLASS: PCI_VID:PCI_DID [PCI_SUBSYS_VID:PCI_SUBSYS_DID] [DRIVER] */
static int fileAction(const char *fileName,
	struct stat ATTRIBUTE_UNUSED *statbuf, void ATTRIBUTE_UNUSED *userData)
{
	char pci_slot_name[128], driver[128];
	int pci_class = 0, pci_vid = 0, pci_did = 0;
	int pci_subsys_vid = 0, pci_subsys_did = 0;
	char *uevent_filename = concat_path_file(fileName, "/uevent");
	config_read_t *crt = config_read_prepare(uevent_filename, 's');

	free(uevent_filename);
	if(crt == NULL)
		return TRUE;
	pci_slot_name[0] = driver[0] = '\0';
	while( (crt = config_read(crt, "\0\1\2\0\2=")) != NULL) {
		config_read_t *tparse;

		if((crt->flags[0] & CR_FLAGS_FOUND_SPEC_DELIM) == 0)
			break;
		tparse = config_read(config_read_prepare(crt->tokens[1], 'm'), "\0\1\2\3\0\1:");
		if (strcmp(crt->tokens[0], "DRIVER") == 0) {
			strcpy(driver, tparse->tokens[0]);
		} else if (strcmp(crt->tokens[0], "PCI_CLASS") == 0) {
			pci_class = bb_xparse_ui_number(16 | BB_PN_SILENT_RE, tparse->tokens[0])>>8;
		} else if (strcmp(crt->tokens[0], "PCI_ID") == 0) {
			pci_vid = bb_xparse_ui_number(16 | BB_PN_SILENT_RE, tparse->tokens[0]);
			pci_did = bb_xparse_ui_number(16 | BB_PN_SILENT_RE, tparse->tokens[1]);
		} else if (strcmp(crt->tokens[0], "PCI_SUBSYS_ID") == 0) {
			pci_subsys_vid = bb_xparse_ui_number(16 | BB_PN_SILENT_RE, tparse->tokens[0]);
			pci_subsys_did = bb_xparse_ui_number(16 | BB_PN_SILENT_RE, tparse->tokens[1]);
		} else if (strcmp(crt->tokens[0], "PCI_SLOT_NAME") == 0) {
			sprintf(pci_slot_name, "%s:%s", tparse->tokens[1], tparse->tokens[2]);
		}
		config_close(tparse);
	}
	config_close(crt);

	if (bss_flags & OPT_m) {
		printf("%s \"Class %04x\" \"%04x\" \"%04x\" \"%04x\" \"%04x\"",
		       pci_slot_name, pci_class, pci_vid, pci_did,
		       pci_subsys_vid, pci_subsys_did);
	} else {
		printf("%s Class %04x: %04x:%04x",
		       pci_slot_name, pci_class, pci_vid, pci_did);
	}

	if ((bss_flags & OPT_k) && driver[0] != '\0') {
		if (bss_flags & OPT_m) {
			printf(" \"%s\"", driver);
		} else {
			printf(" %s", driver);
		}
	}
	putchar('\n');

	return TRUE;
}

int lspci_main(void)
{
	bb_getopt_ulflags(opt_str);

	recursive_action("/sys/bus/pci/devices", TRUE, FALSE, FALSE,
			fileAction,
			NULL, /* dirAction */
			NULL  /* userData */
			);

	return EXIT_SUCCESS;
}
#endif /* APPLET_USAGE */
