/* vi: set sw=4 ts=4: */
/*
 * lsscsi implementation for busybox
 *
 * Copyright (C) 2017 Markus Gothe <nietzsche@lysator.liu.se>
 *
 * Licensed under GPLv2 or later, see file LICENSE in this source tree.
 */

#include "busybox.h"
#include <string.h>
#include <dirent.h>
#include <unistd.h>
#define BUF_LINE_SIZE 80

static char *get_line_80(const char *filename, char *buf)
{
	FILE *file = fopen(filename, "r");
	int ch, idx = 0;

	if(file == NULL)
		return NULL;

	while (1) {
		ch = getc(file);
		if (idx > (BUF_LINE_SIZE - 1) || ch == '\n' || ch == EOF)
			ch = 0;
		buf[idx++] = (char)ch;
		if (!ch) break;
	}
	fclose(file);
	return idx ? buf : NULL;
}

#define scsi_device_types \
	"disk\0"    "tape\0"    "printer\0" "process\0" \
	"worm\0"    "\0"        "scanner\0" "optical\0" \
	"mediumx\0" "comms\0"   "\0"        "\0"        \
	"storage\0" "enclosu\0" "sim dsk\0" "opti rd\0" \
	"bridge\0"  "osd\0"     "adi\0"     "\0"        \
	"\0"        "\0"        "\0"        "\0"        \
	"\0"        "\0"        "\0"        "\0"        \
	"\0"        "\0"        "wlun\0"    "no dev"

static const char *scsi_device_types_string(int t)
{
	const char *p = scsi_device_types;
	while(t) {
		p += strlen(p) + 1;
		t--;
	}
	return p;
}

int lsscsi_main(void)
{
	struct dirent *de;
	DIR *dir;
	int dfd;

	xchdir("/sys/bus/scsi/devices");

	if((dir = opendir(".")) == NULL)
		bb_perror_msg_and_die("%s", "/sys/bus/scsi/devices");
	dfd = dirfd(dir);
	while ((de = readdir(dir)) != NULL) {
		char vendor[80];
		char type_str[80];
		char model[80];
		char rev[80];
		int type;
		const char *type_name;

		if (de->d_name[0] < '0' || de->d_name[0] > '9')
			continue;
		if (!strchr(de->d_name, ':'))
			continue;
		if (chdir(de->d_name) != 0)
			continue;

		if(get_line_80("vendor", vendor) == NULL)
			continue;
		if(get_line_80("type", type_str) == NULL)
			continue;
		if(get_line_80("model", model) == NULL)
			continue;
		if(get_line_80("rev", rev) == NULL)
			continue;

		printf("[%s]\t", de->d_name);

		if ((type = bb_xparse_i_number(BB_PN_SET_UPPER | BB_PN_SILENT_RE,
						type_str, 0x20)) < 0
			|| (type_name = scsi_device_types_string(type))[0] == '\0'
		) {
			printf("(%s)\t", type_str);
		} else {
			printf("%s\t", type_name);
		}

		printf("%s\t""%s\t""%s\n", vendor, model, rev);
		/* TODO: also output device column, e.g. "/dev/sdX" */

		fchdir(dfd);
	}

	return EXIT_SUCCESS;
}
