/* vi: set sw=4 ts=4: */
/*
 * lsusb implementation for busybox
 *
 * Copyright (C) 2009  Malek Degachi <malek-degachi@laposte.net>
 *
 * Licensed under GPLv2 or later, see file LICENSE in this source tree.
 */

#include "busybox.h"
#include <string.h>
#include "config_read.h"

static int fileAction(const char *fileName,
	struct stat ATTRIBUTE_UNUSED *statbuf, void ATTRIBUTE_UNUSED *userData)
{
	char busnum[128], devnum[128];
	int product_vid = 0, product_did = 0, found = 0;
	char *uevent_filename = concat_path_file(fileName, "uevent");
	config_read_t *crt = config_read_prepare(uevent_filename, 's');

	free(uevent_filename);
	if(crt == NULL)
		return TRUE;
	while( (crt = config_read(crt, "\0\1\2\0\2=")) != NULL) {
		config_read_t *tparse;

		if((crt->flags[0] & CR_FLAGS_FOUND_SPEC_DELIM) == 0)
			break;
		if(crt->num_line == 1 && strcmp(crt->tokens[0], "DEVTYPE") == 0)
			break;
		tparse = config_read(config_read_prepare(crt->tokens[1], 'm'), "\0\1\2\0\1/");
		if (strcmp(crt->tokens[0], "PRODUCT") == 0) {
			product_vid = bb_xparse_ui_number(16 | BB_PN_SILENT_RE, tparse->tokens[0]);
			product_did = bb_xparse_ui_number(16 | BB_PN_SILENT_RE, tparse->tokens[1]);
		} else if (strcmp(crt->tokens[0], "BUSNUM") == 0) {
			strcpy(busnum, tparse->tokens[0]);
			found++;
		} else if (strcmp(crt->tokens[0], "DEVNUM") == 0) {
			strcpy(devnum, tparse->tokens[0]);
			found++;
		}
		config_close(tparse);
		if (found == 2) {
			printf("Bus %s Device %s: ID %04x:%04x\n", busnum, devnum, product_vid, product_did);
			break;
		}
	}
	config_close(crt);
	return TRUE;
}

int lsusb_main(void)
{
	/* no options, no getopt */

	recursive_action("/sys/bus/usb/devices",
			TRUE, FALSE, FALSE,
			fileAction,
			NULL, /* dirAction */
			NULL  /* userData */
			);
	return EXIT_SUCCESS;
}
