/*
 * Linux32/linux64 allows for changing uname emulation.
 *
 * Copyright 2002 Andi Kleen, SuSE Labs.
 * This file is subject to the GNU General Public License v2
 *
 * Licensed under GPL v2 or later, see file License for details.
*/
#ifdef APPLET_USAGE
#define setarch_trivial_usage \
	"<personality> <program> [args ...]"
#define setarch_full_usage \
	"Personality may be:\n" \
	"\tlinux32\tSet 32bit uname emulation\n" \
	"\tlinux64\tSet 64bit uname emulation"
#else

#include <unistd.h>
#include <string.h>
#include "busybox.h"
#ifdef HAVE_SYS_PERSONALITY_H
#include <sys/personality.h>
#else
#include <linux/personality.h>
#endif

int setarch_main(void)
{
	int pers = -1;
	char **argv = Gargv;

	/* Figure out what personality we are supposed to switch to ...
	 * we can be invoked as either:
	 * argv[0],argv[1] -> "setarch","personality"
	 * argv[0]         -> "personality"
	 */
retry:
	if (!strcmp(argv[0], "linux64"))
		pers = PER_LINUX;
	else if (!strcmp(argv[0], "linux32"))
		pers = PER_LINUX32;
	else if (pers == -1 && argv[1] != NULL) {
		pers = PER_LINUX32;
		++argv;
		goto retry;
	}

	/* make user actually gave us something to do */
	++argv;
	if (argv[0] == NULL)
		bb_show_usage();

	/* Try to set personality */
	if (personality(pers) < 0)
		goto failure;

	/* Try to execute the program */
	execvp(argv[0], argv);

failure:
	bb_perror_msg_and_die(argv[0]);
}
#endif /* APPLET_USAGE */
