/*
   Copyright (C) 1991-1998 Free Software Foundation, Inc.

   Copyright (C) 1994 Janne Kukonlehto
   Copyright (C) 1994 Miguel de Icaza
   Copyright (C) 1994 Radek Doulik
   Copyright (C) 1994 Fred Leeflang
   Copyright (C) 1995 Jakub Jelinek
   Copyright (C) 1995 Andrej Borsenkow
   Copyright (C) 1995 Ching Hui (mr854307@cs.nthu.edu.tw)
   Copyright (C) 1996 Dugan Porter
   Copyright (C) 1996 Mauricio Plaza
   Copyright (C) 1997 Norbert Warmuth
   Copyright (C) 1997 Paul Sheer
   Copyright (C) 1997 Timur Bakeyev
   Copyright (C) 1998 Pavel Machek

   SLANG: Copyright (c) 1992, 1995 John E. Davis

   special version for busybox or cygwin
   by Vladimir Oleynik <dzo@simtreas.ru> (C) 2001-2015

   Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
*/

#ifdef APPLET_USAGE
#define mc_trivial_usage \
	"[flags] [this_dir] [other_panel_dir]"
#define mc_full_usage \
	"Midnight Commander\n" \
	"Flags:\n" \
	USE_FEATURE_LONG_OPTIONS( \
	"-a, --stickchars\tForce use of +, -, | for line drawing.\n" \
	"-b, --nocolor\tForce black and white display.\n" \
	"-c, --color\tForce color mode.\n" \
	"-C, --colors\tSpecify colors (use --help-colors to get a list).\n" \
	"-d, --nomouse\tDisable mouse support.\n" \
	"-f, --libdir\tPrint configured paths.\n" \
	"-k, --resetsoft\tReset softkeys to their terminfo/termcap default.\n" \
	"-P, --printwd\tAt exit, print the last working directory.\n" \
	"-s, --slow\tDisables verbose operation (for slow terminals).\n" \
	"-V, --version\tReport version and configuration options.\n" \
	"-x, --xterm\tForce xterm mouse support and screen save/restore.") \
	SKIP_FEATURE_LONG_OPTIONS( \
	"-a\tForce use of +, -, | for line drawing.\n" \
	"-b\tForce black and white display.\n" \
	"-c\tForce color mode.\n" \
	"-C\tSpecify colors (use --help-colors to get a list).\n" \
	"-d\tDisable mouse support.\n" \
	"-f\tPrint configured paths.\n" \
	"-k\tReset softkeys to their terminfo/termcap default.\n" \
	"-P\tAt exit, print the last working directory.\n" \
	"-s\tDisables verbose operation (for slow terminals).\n" \
	"-V\tReport version and configuration options.\n" \
	"-x\tForce xterm mouse support and screen save/restore.")
#else

#define VERSION "4.1.36"


#define LIBDIR "/usr/lib/mc/"
#define BINDIR "/usr/bin/"

#ifdef __CYGWIN__
# define NO_INFOMOUNT   /* not realise statfs/statvfs */
#endif

#ifdef I386
#define stdcall2 __attribute__ ((stdcall))
#else
#define stdcall2
#endif

/* Define umode_t if your system does not provide it */
#define umode_t int

/* Does the file command accepts the -L option */
#define FILE_L 1

/* Support the Midnight Commander Virtual File System? */
//#define USE_VFS 1
#undef USE_VFS

/* If you have socket and the rest of the net functions use this */
//#define USE_NETCODE 1
#undef USE_NETCODE

#ifndef USE_VFS
# ifdef USE_NETCODE
#  undef USE_NETCODE
# endif
#endif

#define MC_GLOBAL_MENU "mc.menu"
#define MC_LOCAL_MENU  ".mc.menu"
#define MC_HOME_MENU   ".mc/menu"
#define MC_HINT        "mc.hint"

#include "busybox.h"
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h> /* For sockaddr_in needed by get_remote_port */
#include <sys/time.h>    /* alex: this redefines struct timeval */
#include <sys/mman.h>
#include <sys/wait.h>
#include <time.h>
#include <sys/select.h>
#include <netdb.h>
#include <fcntl.h>
#include <utime.h>

#include <dirent.h>
#include <sys/mount.h>
#include <sys/vfs.h>

#if !defined(NO_INFOMOUNT)
#include <sys/statfs.h>
#endif

#include <mntent.h>
#include <termios.h>
#include <sys/ioctl.h>

#include <locale.h>
#include <paths.h>

#include "bb_sig.h"
#include "bb_inet.h"

#ifdef HAVE_ARPA_FTP
#include <arpa/ftp.h>
#else
#include "ftp.h"
#endif
#ifdef HAVE_ARPA_TELNET
#include <arpa/telnet.h>
#else
#include "telnet.h"
#endif

#include "xregex.h"
#include "speed_table.h"

/* SLANG 9938 */

/* linklist.h */

struct linklist {
    void *data;
    struct linklist *next;
    struct linklist *prev;
};

struct list_iterator {
    struct linklist *linklist;
    struct linklist *current_pos;
};

#ifdef USE_VFS
struct extfs_inode;

struct extfs_entry {
    int has_changed;
    struct extfs_entry *next_in_dir;
    struct extfs_entry *dir;
    char *name;
    struct extfs_inode *inode;
};

struct extfs_archive;

struct extfs_inode {
    int has_changed;
    nlink_t nlink;
    struct extfs_entry *first_in_subdir; /* only used if this is a directory */
    struct extfs_entry *last_in_subdir;
    ino_t inode;        /* This is inode # */
    dev_t dev;          /* This is an internal identification of the extfs archive */
    struct extfs_archive *archive; /* And this is an archive structure */
    dev_t rdev;
    umode_t mode;
    uid_t uid;
    gid_t gid;
    int size;
    time_t mtime;
    char linkflag;
    char *linkname;
    time_t atime;
    time_t ctime;
    char *local_filename;
};

struct extfs_archive {
    int fstype;
    char *name;
    char *local_name;
    struct stat extfsstat;
    struct stat local_stat;
    dev_t rdev;
    int fd_usage;
    ino_t __inode_counter;
    struct extfs_entry *root_entry;
    struct extfs_entry *current_dir;
    struct extfs_archive *next;
};

static int extfs_prefix_to_type (const char *path);
static void extfs_run (char *path);
#endif

/*  ftpfs.h */

#ifdef USE_NETCODE
struct ftpentry
{
    char *name;
    int count;
    char *linkname;
    char *local_filename;
    int local_is_temp:1;
    int freshly_created:1;
    int tmp_reget;
    struct stat local_stat;
    char *remote_filename;
    struct stat s;
    struct stat *l_stat;
    struct ftpfs_connection *bucket;
};

struct ftpfs_dir
{
    int count;
    struct timeval timestamp;
    char *remote_path;
    struct linklist *file_list;
    int symlink_status;
};

/* valid values for ftpfs_dir->symlink_status */
#define FTPFS_NO_SYMLINKS          0
#define FTPFS_UNRESOLVED_SYMLINKS  1
#define FTPFS_RESOLVED_SYMLINKS    2

struct ftpfs_connection {
    char *host;
    char *user;
    char *current_directory;
    char *home;
    char *updir;
    char *password;
    int port;
    int sock;
    struct linklist *dcache;
    ino_t __inode_counter;
    int lock;
    int failed_on_login;        /* used to pass the failure reason to upper levels */
    int use_proxy;              /* use a proxy server */
    int result_pending;
    int use_source_route;
    int use_passive_connection;
    int isbinary;
    int cwd_defered;  /* current_directory was changed but CWD command hasn't
			 been sent yet */
    int strict_rfc959_list_cmd; /* ftp server doesn't understand
				   "LIST -la <path>", use "CWD <path>"/
				   "LIST" instead */
};

#define qhost(b) (b)->host
#define quser(b) (b)->user
#define qcdir(b) (b)->current_directory
#define qport(b) (b)->port
#define qsock(b) (b)->sock
#define qlock(b) (b)->lock
#define qdcache(b) (b)->dcache
#define qhome(b) (b)->home
#define qupdir(b) (b)->updir
#define qproxy(b) (b)->use_proxy

/* Increased since now we may use C-r to reread the contents */
#define FTPFS_DIRECTORY_TIMEOUT 30 * 60

#define FTPFS_DO_RESOLVE_SYMLINK 1
#define FTPFS_OPEN            2
#define FTPFS_FREE_RESOURCE   4

static char *ftpfs_anonymous_passwd;
static char *ftpfs_proxy_host;
static int ftpfs_directory_timeout;
/* wether we have to use proxy by default? */
static int ftpfs_always_use_proxy;
static void ftpfs_init_passwd (void);
#endif /* USE_NETCODE */

/* Declarations for the tarfs. */

/*
 * Header block on tape.
 *
 * I'm going to use traditional DP naming conventions here.
 * A "block" is a big chunk of stuff that we do I/O on.
 * A "record" is a piece of info that we care about.
 * Typically many "record"s fit into a "block".
 */
#define RECORDSIZE      512
#define NAMSIZ          100
#define TUNMLEN         32
#define TGNMLEN         32
#define SPARSE_EXT_HDR  21
#define SPARSE_IN_HDR   4

struct sparse {
    char offset[12];
    char numbytes[12];
};

struct sp_array {
    int offset;
    int numbytes;
};

union record {
    char charptr[RECORDSIZE];
    struct header {
	char arch_name[NAMSIZ];
	char mode[8];
	char uid[8];
	char gid[8];
	char size[12];
	char mtime[12];
	char chksum[8];
	char linkflag;
	char arch_linkname[NAMSIZ];
	char magic[8];
	char uname[TUNMLEN];
	char gname[TGNMLEN];
	char devmajor[8];
	char devminor[8];
	/* these following fields were added by JF for gnu */
	/* and are NOT standard */
	char atime[12];
	char ctime[12];
	char offset[12];
	char longnames[4];
	char pad;
	struct sparse sp[SPARSE_IN_HDR];
	char isextended;
	char realsize[12];      /* true size of the sparse file */
	/* char ending_blanks[12];*//* number of nulls at the
	   end of the file, if any */
    } header;
    struct extended_header {
	struct sparse sp[21];
	char isextended;
    } ext_hdr;
};

/* The checksum field is filled with this while the checksum is computed. */
#define CHKBLANKS       "        "      /* 8 blanks, no null */

/* The magic field is filled with this if uname and gname are valid. */
#define TMAGIC          "ustar  "       /* 7 chars and a null */

/* The linkflag defines the type of file */
#define LF_OLDNORMAL    '\0'    /* Normal disk file, Unix compat */
#define LF_NORMAL       '0'     /* Normal disk file */
#define LF_LINK         '1'     /* Link to previously dumped file */
#define LF_SYMLINK      '2'     /* Symbolic link */
#define LF_CHR          '3'     /* Character special file */
#define LF_BLK          '4'     /* Block special file */
#define LF_DIR          '5'     /* Directory */
#define LF_FIFO         '6'     /* FIFO special file */
#define LF_CONTIG       '7'     /* Contiguous file */
/* Further link types may be defined later. */

/* Note that the standards committee allows only capital A through
   capital Z for user-defined expansion.  This means that defining something
   as, say '8' is a *bad* idea. */
#define LF_DUMPDIR      'D'     /* This is a dir entry that contains
					   the names of files that were in
					   the dir at the time the dump
					   was made */
#define LF_LONGLINK     'K'     /* Identifies the NEXT file on the tape
					   as having a long linkname */
#define LF_LONGNAME     'L'     /* Identifies the NEXT file on the tape
					   as having a long name. */
#define LF_MULTIVOL     'M'     /* This is the continuation
					   of a file that began on another
					   volume */
#define LF_NAMES        'N'     /* For storing filenames that didn't
					   fit in 100 characters */
#define LF_SPARSE       'S'     /* This is for sparse files */
#define LF_VOLHDR       'V'     /* This file is a tape/volume header */
/* Ignore it on extraction */

/*
 * Exit codes from the "tar" program
 */
#define EX_SUCCESS      0       /* success! */
#define EX_ARGSBAD      1       /* invalid args */
#define EX_BADFILE      2       /* invalid filename */
#define EX_BADARCH      3       /* bad archive */
#define EX_SYSTEM       4       /* system gave unexpected error */
#define EX_BADVOL       5       /* Special error code means
				   Tape volume doesn't match the one
				   specified on the command line */

/*
 * We default to Unix Standard format rather than 4.2BSD tar format.
 * The code can actually produce all three:
 *      f_standard      ANSI standard
 *      f_oldarch       V7
 *      neither         4.2BSD
 * but we don't bother, since 4.2BSD can read ANSI standard format anyway.
 * The only advantage to the "neither" option is that we can cmp our
 * output to the output of 4.2BSD tar, for debugging.
 */
#define         f_standard              (!f_oldarch)

#ifndef TAR_NAMES

struct tarfs_inode;

struct tarfs_entry {
    int has_changed;
    off_t header_offset; /* -1 if not in stored in archive */
    int header_size;
    struct tarfs_entry *next_in_dir;
    struct tarfs_entry *dir;
    off_t extended_offset; /* -1 if not present */
    off_t extended_size;

    char *name;
    struct tarfs_inode *inode;
};

struct tarfs_archive;

struct tarfs_inode {
    int has_changed;
    int is_open;
    nlink_t nlink;
    struct tarfs_entry *first_in_subdir; /* only used if linkflag == L_DIR */
    struct tarfs_entry *last_in_subdir;
    off_t data_offset;
    char *local_filename;
    ino_t inode;        /* This is inode # */
    dev_t dev;          /* This is an internal identification of the tar archive */
    struct tarfs_archive *archive; /* And this is an archive structure */
    dev_t rdev;
    int std;            /* 0 if old Unix inode */

    umode_t mode;
    uid_t uid;
    gid_t gid;
    int size;
    time_t mtime;
    char linkflag;
    char *linkname;
    time_t atime;
    time_t ctime;
};

struct tarfs_archive {
    char *name;
    struct stat tarstat;
    int is_gzipped;
    dev_t rdev;
    ino_t __inode_counter;
    struct tarfs_entry *root_entry;
    struct tarfs_entry *current_dir;
    struct tarfs_archive *next;
    off_t current_tar_position;
    int fd;
    int fd_usage; /* Zero means fd is invalid, otherwise is number of
		     pseudofiles that have the archive open */
    char *tmpname; /* For case of is_gzipped == 2 */
/* This part is only for gzipped archives: (growable buffers) */
    union record *block_first;
    int count_first; /* In records */
    union record **block_ptr;
    int count_blocks; /* In blocks */
    union record *current_record; /* position when reading */
};

/* This constant determines how many RECORDSIZE blocks are in one growing
   chunk */
#define TAR_GROWING_CHUNK_SIZE 64

enum {
    tar_normal,
    targz_growing,
    tar_uncompressed_local
};

#endif /* ! TAR_NAMES */

#if defined(USE_NETCODE)
static char *get_host_and_username (char *path, char **host, char **user, int *port,
			     int default_port, int default_to_anon, char **pass);

/* Use the ~/.netrc */
static int use_netrc = 1;
#endif

#ifdef USE_VFS

#ifdef USE_NETCODE
    static int got_sigpipe;
#endif
    /* Our virtual file system layer */

    typedef void * vfsid;

    struct vfs_stamping;

    typedef struct {
	void  *(*open)(char *fname, int flags, int mode);
	int   (*close)(void *vfs_info);
	int   (*read)(void *vfs_info, char *buffer, int count);
	int   (*write)(void *vfs_info, char *buf, int count);

	void  *(*opendir)(char *dirnm);
	void  *(*readdir)(void *vfs_info);
	int   (*closedir)(void *vfs_info);

	int  (*stat)(const char *path, struct stat *buf);
	int  (*lstat)(const char *path, struct stat *buf);
	int  (*fstat)(void *vfs_info, struct stat *buf);

	int  (*chmod)(char *path, int mode);
	int  (*chown)(char *path, int owner, int group);
	int  (*utime)(char *path, struct utimbuf *times);

	int  (*readlink)(char *path, char *buf, int size);
	int  (*symlink)(char *n1, char *n2);
	int  (*link)(char *p1, char *p2);
	int  (*unlink)(char *path);
	int  (*rename)(char *p1, char *p2);
	int  (*chdir)(char *path);
	int  (*ferrno)(void);
	int  (*lseek)(void *vfs_info, off_t offset, int whence);
	int  (*mknod)(char *path, int mode, int dev);

	vfsid (*getid)(char *path, struct vfs_stamping **parent);
	int  (*nothingisopen)(vfsid id);
	void (*free)(vfsid id);

	char *(*getlocalcopy)(char *filename);
	void (*ungetlocalcopy)(char *filename, char *local, int has_changed);

	int  (*mkdir)(char *path, mode_t mode);
	int  (*rmdir)(char *path);

	int  (*ctl)(void *vfs_info, int ctlop, int arg);
	int  (*setctl)(char *path, int ctlop, char *arg);
	void (*forget_about)(char *path);
	caddr_t (*mmap)(caddr_t addr, size_t len, int prot, int flags, void *vfs_info, off_t offset);
	int (*munmap)(caddr_t addr, size_t len, void *vfs_info);
    } vfs;

    struct vfs_stamping {
	vfs *v;
	vfsid id;
	struct vfs_stamping *parent; /* At the moment applies to tarfs only */
	struct vfs_stamping *next;
	struct timeval time;
    };

    static vfs *vfs_type (const char *path);
    static vfsid vfs_ncs_getid (vfs *nvfs, char *dir, struct vfs_stamping **par);
    static void vfs_rm_parents (struct vfs_stamping *stamp);
    static char *vfs_path (char *path);
    static char *vfs_canon (const char *path);
    static char *mc_get_current_wd (char *buffer, int bufsize);
    static int vfs_current_is_local (void);
    static int vfs_current_is_extfs (void);
    static int vfs_current_is_tarfs (void);
    static int vfs_file_is_local (char *name);
    static char *vfs_get_current_dir (void);

    static void vfs_stamp (vfs *, vfsid);
    static void vfs_rmstamp (vfs *, vfsid, int);
    static void vfs_addstamp (vfs *, vfsid, struct vfs_stamping *);
    static void vfs_add_noncurrent_stamps (vfs *, vfsid, struct vfs_stamping *);
    static void vfs_add_current_stamps (void);
    static void vfs_timeout_handler (void);
    static int  vfs_timeouts (void);
    static void vfs_fill_names (void (*)(char *));

    /* Required for the vfs_canon routine */
    static char *tarfs_analysis (const char *inname, char **archive, int is_dir);

#ifdef USE_NETCODE
    static int vfs_file_is_ftp (char *filename);
    static void ftpfs_hint_reread(int reread);
    static void ftpfs_flushdir(void);
    static char *ftpfs_gethome (char *);
#else
#   define ftpfs_flushdir()
#   define ftpfs_hint_reread(x)
#endif
    /* They fill the file system names */
    static void tarfs_fill_names (void (*)(char *));

    /* Only the routines outside of the VFS module need the emulation macros */

    static int mc_open (char *file, int flags, ...);
    static int mc_close (int handle);
    static int mc_read (int handle, char *buffer, int count);
    static int mc_write (int hanlde, char *buffer, int count);
    static off_t mc_lseek (int fd, off_t offset, int whence);
    static int mc_chdir (char *);

    static DIR *mc_opendir (char *dirnm);
    static struct dirent *mc_readdir(DIR *dirp);
    static int mc_closedir (DIR *dir);

    static int mc_stat (const char *path, struct stat *buf);
    static int mc_lstat (const char *path, struct stat *buf);
    static int mc_fstat (int fd, struct stat *buf);

    static int mc_chmod (char *path, int mode);
    static int mc_chown (char *path, int owner, int group);
    static int mc_utime (char *path, struct utimbuf *times);
    static int mc_readlink(char *path, char *buf, int bufsiz);
    static int mc_unlink (char *path);
    static int mc_symlink (char *name1, char *name2);
    static int mc_link (char *name1, char *name2);
    static int mc_mknod (char *, int, int);
    static int mc_rename (char *original, char *target);
    static int mc_write (int fd, char *buf, int nbyte);
    static int mc_rmdir (char *path);
    static int mc_mkdir (char *path, mode_t mode);
    static char *mc_getlocalcopy (char *filename);
    static void mc_ungetlocalcopy (char *filename, char *local, int has_changed);
    static char *mc_def_getlocalcopy (char *filename);
    static void mc_def_ungetlocalcopy (char *filename, char *local, int has_changed);
    static int mc_ctl (int fd, int ctlop, int arg);
    static int mc_setctl (char *path, int ctlop, char *arg);
    static caddr_t mc_mmap (caddr_t, size_t, int, int, int, off_t);
    static int mc_munmap (caddr_t addr, size_t len);

#else

#ifdef USE_NETCODE
#    undef USE_NETCODE
#endif

#   define vfs_fill_names(x)
#   define vfs_add_current_stamps()
#   define vfs_current_is_local() 1
#   define vfs_file_is_local(x) 1
#   define vfs_file_is_ftp(x) 0
#   define vfs_current_is_tarfs() 0
#   define vfs_current_is_extfs() 0
#   define vfs_path(x) x
#   define mc_close close
#   define mc_read read
#   define mc_write write
#   define mc_lseek lseek
#   define mc_opendir opendir
#   define mc_readdir readdir
#   define mc_closedir closedir

#   define mc_get_current_wd(x,size) getcwd (x, size)
#   define mc_fstat fstat
#   define mc_lstat lstat

#   define mc_readlink readlink
#   define mc_symlink symlink
#   define mc_rename rename

#ifndef __os2__
#   define mc_open open
#   define mc_utime local_utime
#   define mc_chmod chmod
#   define mc_chown chown
#   define mc_chdir chdir
#   define mc_unlink unlink
#endif

#   define mc_mmap mmap
#   define mc_munmap munmap

#   define mc_ctl(a,b,c) 0
#   define mc_setctl(a,b,c)

#   define mc_stat stat
#   define mc_mknod mknod
#   define mc_link link
#   define mc_mkdir mkdir
#   define mc_rmdir rmdir
#   define is_special_prefix(x) 0
#   define vfs_type(x) (vfs *)(NULL)
#   define vfs_setup_wd()
#   define vfs_init()
#   define vfs_shut()
#   define vfs_canon(p) canonicalize_pathname(strdup(p))
#   define vfs_timeout_handler()
#   define vfs_timeouts() 0

    typedef int vfs;

#   define mc_getlocalcopy(x) NULL
#   define mc_ungetlocalcopy(x,y,z)

#   define ftpfs_hint_reread(x)
#   define ftpfs_flushdir()

#endif /* USE_VFS */

#ifndef USE_NETCODE
# ifdef USE_VFS
#  define vfs_file_is_ftp(x) 0
# endif
#endif

#define MCCTL_SETREMOTECOPY     0
#define MCCTL_ISREMOTECOPY      1
#define MCCTL_REMOTECOPYCHUNK   2
#define MCCTL_FINISHREMOTE      3
#define MCCTL_FLUSHDIR          4

/* Return codes from the ${fs}_ctl routine */

#define MCERR_TARGETOPEN        -1
    /* Can't open target file */
#define MCERR_READ              -2
    /* Read error on source file */
#define MCERR_WRITE             -3
    /* Write error on target file */
#define MCERR_FINISH            -4
    /* Finished transfer */
#define MCERR_DATA_ON_STDIN     -5
    /* Data waiting on stdin to be processed */


/* This file takes care of loading ncurses or slang */

static int got_interrupt (void);

/* __MYSLANG_H */

enum {
    KEY_BACKSPACE = 400,
    KEY_END, KEY_UP, KEY_DOWN, KEY_LEFT, KEY_RIGHT,
    KEY_HOME, KEY_A1, KEY_C1, KEY_NPAGE, KEY_PPAGE, KEY_IC,
    KEY_ENTER, KEY_DC, KEY_SCANCEL, KEY_BTAB
};

#define KEY_F(x) 1000+x

# define SLSMG_HLINE_CHAR       'q'
# define SLSMG_VLINE_CHAR       'x'
# define SLSMG_ULCORN_CHAR      'l'
# define SLSMG_URCORN_CHAR      'k'
# define SLSMG_LLCORN_CHAR      'm'
# define SLSMG_LRCORN_CHAR      'j'
# define SLSMG_CKBRD_CHAR       'a'
# define SLSMG_RTEE_CHAR        'u'
# define SLSMG_LTEE_CHAR        't'
# define SLSMG_UTEE_CHAR        'w'
# define SLSMG_DTEE_CHAR        'v'
# define SLSMG_PLUS_CHAR        'n'

# define SLSMG_COLOR_BLACK              0x000000
# define SLSMG_COLOR_RED                0x000001
# define SLSMG_COLOR_GREEN              0x000002
# define SLSMG_COLOR_BROWN              0x000003
# define SLSMG_COLOR_BLUE               0x000004
# define SLSMG_COLOR_MAGENTA            0x000005
# define SLSMG_COLOR_CYAN               0x000006
# define SLSMG_COLOR_LGRAY              0x000007
# define SLSMG_COLOR_GRAY               0x000008
# define SLSMG_COLOR_BRIGHT_RED         0x000009
# define SLSMG_COLOR_BRIGHT_GREEN       0x00000A
# define SLSMG_COLOR_BRIGHT_BROWN       0x00000B
# define SLSMG_COLOR_BRIGHT_BLUE        0x00000C
# define SLSMG_COLOR_BRIGHT_CYAN        0x00000D
# define SLSMG_COLOR_BRIGHT_MAGENTA     0x00000E
# define SLSMG_COLOR_BRIGHT_WHITE       0x00000F

#define SLTT_BOLD_MASK  0x01000000
#define SLTT_BLINK_MASK 0x02000000
#define SLTT_ULINE_MASK 0x04000000
#define SLTT_REV_MASK   0x08000000
#define SLTT_ALTC_MASK  0x10000000

#define ACS_VLINE SLSMG_VLINE_CHAR
#define ACS_HLINE SLSMG_HLINE_CHAR
#define ACS_ULCORNER SLSMG_ULCORN_CHAR
#define ACS_LLCORNER SLSMG_LLCORN_CHAR
#define ACS_URCORNER SLSMG_URCORN_CHAR
#define ACS_LRCORNER SLSMG_LRCORN_CHAR

#define ACS_LTEE 't'

enum {
    COLOR_BLACK, COLOR_RED, COLOR_GREEN, COLOR_YELLOW, COLOR_BLUE,
    COLOR_MAGENTA, COLOR_CYAN, COLOR_WHITE
};

/* When using Slang with color, we have all the indexes free but
 * those defined here (A_BOLD, A_UNDERLINE, A_REVERSE, A_BOLD_REVERSE)
 */
#define A_BOLD      0x40
#define A_UNDERLINE 0x40
#define A_REVERSE   0x20
#define A_BOLD_REVERSE 0x21

#ifndef A_NORMAL
#    define A_NORMAL    0x00
#endif

#define ERR -1

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

static void slang_set_raw_mode (void);

#define raw() slang_set_raw_mode()
#define nodelay(val) no_slang_delay = val
#define beep() SLtt_beep ()
#define keypad(value) slang_keypad (value)

#define touchwin() SLsmg_touch_lines0(LINES)
#define reset_shell_mode slang_shell_mode
#define reset_prog_mode slang_prog_mode

static stdcall2 void attrset (int color);
static void slang_prog_mode (void);
static stdcall2 void hline (int ch, int len);
static stdcall2 void vline (int ch, int len);
static int  getch (void);
static stdcall2 void slang_keypad (int set);
static void slang_shell_mode (void);
static int  has_colors (void);
/* Internal function prototypes */
static void load_terminfo_keys (void);

/* copied from slcurses.h (MC version 4.0.7) */
#define move SLsmg_gotorc
#define printw SLsmg_printf
#define COLS SLtt_Screen_Cols
#define LINES SLtt_Screen_Rows
#define standend  SLsmg_normal_video
#define addch SLsmg_write_char
#define addstr SLsmg_write_string
#define mc_refresh SLsmg_refresh
#define endwin SLsmg_reset_smg

#define SLsmg_draw_double_box(r,c,dr,dc) SLsmg_draw_box ((r), (c), (dr), (dc))

#define SLsmg_get_row() This_Row
#define SLsmg_get_column() This_Col

    /* This is slow, but works well :| */
#define one_vline() SLsmg_draw_object (SLsmg_get_row(), SLsmg_get_column(), slow_terminal ? ' ' : ACS_VLINE)
#define one_hline() SLsmg_draw_object (SLsmg_get_row(), SLsmg_get_column(), slow_terminal ? ' ' : ACS_HLINE)

static void enable_interrupt_key (void);
static void disable_interrupt_key (void);
/* end __MYSLANG_H */

#define KEY_KP_ADD      4001
#define KEY_KP_SUBTRACT 4002
#define KEY_KP_MULTIPLY 4003

#ifndef MAXPATHLEN
#   define MC_MAXPATHLEN 4096
#else
#   define MC_MAXPATHLEN MAXPATHLEN
#endif


/* String managing functions */

static stdcall2 int  is_printable (int c);
static stdcall2 int  msglen (const char *text, int *lines);
static char *name_quote (const char *c, int quote_percent);
static char *fake_name_quote (const char *c, int quote_percent);
static stdcall2 char *name_trunc (const char *txt, int trunc_len);
static stdcall2 char *size_trunc_sep (off_t size);
static stdcall2 int  is_exe (mode_t mode);
static stdcall2 char *strip_home_and_password(char *dir, char newdir [MC_MAXPATHLEN]);
static stdcall2 const char *extension (const char *);
static stdcall2 char *get_full_name (const char *dir, char *file);

static char *concat_mcpath_file (const char *file);
static char *unix_error_string (void);
static stdcall2 char *skip_separators (char *s);
static char *convert_controls (char *s);
static stdcall2 char *resolve_symlinks (char *path);
static stdcall2 char *diff_two_paths (char *first, char *second);

#define x_malloc(a,b) xmalloc(a)

static char *mc_mkstemp(int *pfd)
{
    char *r;
    static const char t[] = "/tmp/mc-XXXXXX";

    r = strdup(t);
    if(r == NULL)
	*pfd = -1;
    else
	*pfd = mkstemp(r);
    return r;
}

/* Profile managing functions */
static stdcall2 char *load_file (const char *filename);

#ifndef S_ISLNK
#   define S_ISLNK(x) (((x) & S_IFLNK) == S_IFLNK)
#endif

#ifndef S_ISSOCK
#   ifdef S_IFSOCK
#       define S_ISSOCK(x) (((x) & S_IFSOCK) == S_IFSOCK)
#   else
#       define S_ISSOCK(x) 0
#   endif
#endif

/* uid/gid managing */
typedef struct user_in_groups{
    struct user_in_groups *next;
    int gid;
} user_in_groups;

static void init_groups (void);

static stdcall2 char *get_group (int);
static stdcall2 char *get_owner (int);

static stdcall2 char *file_date (time_t);
static stdcall2 int exist_file (const char *name);

/* Matching */
enum { match_file, match_normal };
static char *convert_pattern (const char *pattern, int match_type, int do_group);
static int regexp_match (char *pattern, char *string, int match_type);

/* Error pipes */
static void open_error_pipe (void);
static void check_error_pipe (void);
static stdcall2 void close_error_pipe (int error, char *text);

/* Process spawning */
static void my_putenv (const char*, const char*);
#define EXECUTE_INTERNAL   1
#define EXECUTE_TEMPFILE   2
#define EXECUTE_AS_SHELL   4
static int my_system (int flags, const char *shell, const char *command);
/* Tilde expansion */
static stdcall2 char *tilde_expand (const char *);

/* Pathname canonicalization */
static stdcall2 char *canonicalize_pathname (char *);

/* Misc Unix functions */
static stdcall2 int my_mkdir (char *s, mode_t mode);
static stdcall2 int my_rmdir (char *s);

/* Filesystem status */
struct my_statfs {
    int type;
    char *typename;
    char *mpoint;
    char *device;
    int avail;
    int total;
    int nfree;
    int nodes;
};

#ifndef NO_INFOMOUNT
static void init_my_statfs (void);
#endif
static stdcall2 void my_statfs (const char *path);

/* Rotating dash routines */
static void rotate_dash (void);

static const char app_text [] = "Midnight-Commander";

enum {
	ISGUNZIPABLE_GUNZIP,
	ISGUNZIPABLE_BZIP,
	ISGUNZIPABLE_BZIP2
};

#ifdef USE_VFS
static void decompress_command_and_arg (int type, char **cmd, char **flags);
#endif

static int mc_doublepopen (int inhandle, int inlen, pid_t *tp, const char *command, ...);
static stdcall2 int mc_doublepclose (int pipehandle, pid_t pid);

/* Hook functions */

typedef struct hook {
    void (*hook_fn)(void *);
    void *hook_data;
    struct hook *next;
} Hook;

static void add_hook (Hook **hook_list, void (*hook_fn)(void *), void *data);
static stdcall2 void execute_hooks (Hook *hook_list);
static void delete_hook (Hook **hook_list, void (*hook_fn)(void *));
static int hook_present (Hook *hook_list, void (*hook_fn)(void *));

/* message stubs: used by those routines that may output something during the panel_operate process */
static void message_1s (int flags, char *title, char *str1);
static void message_2s (int flags, char *title, char *str1, char *str2);
static void message_3s (int flags, char *title, char *str1, char *str2, const char *str3);

#define get_default_editor() "vi"
#define OS_SORT_CASE_SENSITIVE_DEFAULT 1
#define MC_ARCH_FLAGS 0

/*
 * This describes a format item.  The parse_display_format routine parses
 * the user specified format and creates a linked list of format_e structures.
 *
 * parse_display_format computes the actual field allocations if
 * the COMPUTE_FORMAT_ALLOCATIONs define is set.  MC frontends that are
 * just interested in the parsed display format should not set this define.
 */
typedef struct {
    /* file attributes */
    int  fnamelen;
    char *fname;
    struct stat  buf;

    /* flags */
    struct {
	unsigned int marked:1;          /* file marked in pane window */
	unsigned int exists:1;          /* use for rereading file */
	unsigned int link_to_dir:1;     /* if this is a link, does it point to directory? */
	unsigned int stalled_link:1;    /* if this is a symlink and points to charon's land */
    } f;
    char *cache;
} file_entry;

typedef struct format_e {
    struct format_e *next;
    int    requested_field_len;
    int    field_len;
    int    just_mode;
    int    expand;
    char   *(*string_fn)(file_entry *, int len);
    const char   *title;
    const char   *id;

    /* first format_e has the number of items */
    int    items;
    int    use_in_gui;
} format_e;

typedef struct {
    file_entry *list;
    int         size;
} dir_list;

/* Xterm mouse support supports only GPM_DOWN and GPM_UP */
/* If you use others make sure your code also works without them */
enum Gpm_Etype {
  GPM_MOVE=1,
  GPM_DRAG=2,   /* exactly one in four is active at a time */
  GPM_DOWN=4,
  GPM_UP=  8,

#define GPM_BARE_EVENTS(ev) ((ev)&0xF)

  GPM_SINGLE=16,            /* at most one in three is set */
  GPM_DOUBLE=32,
  GPM_TRIPLE=64,

  GPM_MFLAG=128,            /* motion during click? */
  GPM_HARD=256             /* if set in the defaultMask, force an already
			      used event to pass over to another handler */
};

typedef struct Gpm_Event {
  int buttons, x, y;
  enum Gpm_Etype type;
} Gpm_Event;

/* General mouse support definitions */

typedef int (*mouse_h)(Gpm_Event *, void *);

typedef unsigned long widget_data;

typedef enum {
    XV_WLAY_DONTCARE, /* Place the widget wherever it is reasonable */

    XV_WLAY_RIGHTOF,  /* Place the widget to the right of the last widget
		       * created - note: add_widget creates widgets from
		       *  the last to the first one.
		       */

    XV_WLAY_BELOWOF,  /* Place it in a column like style */

    XV_WLAY_BELOWCLOSE,/* The same, but without any gap between them */

    XV_WLAY_NEXTROW,  /* Place it on the left margin with Y bellow all the
		       * previous widgets
		       */

    XV_WLAY_CENTERROW,/* The same as previous, but when the dialog is
		       * ready to show, tries to center that row of widgets
		       */

    XV_WLAY_NEXTCOLUMN, /* Place it on the top margin with X behind all the
		       * previous widgets
		       */

    XV_WLAY_RIGHTDOWN, /* Place the widget to the right of the last one with
			* y set so that both y + h and yold + hold are equal.
			  This is usefull if the previous widget was a radio,
			  which has multiple lines */
    XV_WLAY_EXTENDWIDTH  /* Like nextrow, but later on tries to extend the widget
			  * to fit in the frame (only for PANEL_LIST) */
} WLay;


typedef struct Dlg_head {
    int *color;                 /* color set */
    int count;                  /* number of widgets */
    int ret_value;

    /* mouse status */
    int mouse_status;           /* For the autorepeat status of the mouse */

    void *previous_dialog;      /* Pointer to the previously running Dlg_head */
    int  refresh_pushed;        /* Did the dialog actually run? */

    /* position */
    int x, y;                   /* Position relative to screen origin */

    /* Flags */
    int running;
    int direction;
    int send_idle_msg;

    const char *help_ctx;

    /* Internal variables */
    struct Widget_Item *current, *first, *last;
    int (*callback) (struct Dlg_head *, int, int);

    struct Widget_Item *initfocus;

    /* Hacks */
    char *title;

    int cols;
    int lines;
    const void *data;

    int  has_menubar;   /* GrossHack: Send events on row 1 to a menubar? */
    int  raw;           /* Should the tab key be sent to the dialog? */

    widget_data wdata;
    int  grided;        /* Does it use the automatic layout? */
} Dlg_head;

typedef struct Widget_Item {
    int dlg_id;
    struct Widget_Item *next;   /* next in circle buffer */
    struct Widget_Item *prev;   /* previous in circle buffer */
    struct Widget *widget;      /* point to the component */
} Widget_Item;

/* Every Widget must have this as it's first element */
typedef struct Widget {
    int x, y;
    int cols, lines;
    int color;                  /* If the widget uses it, the color */
    int options;
    int focused;                /* Tells if the widget is focused */
    int (*callback)(Dlg_head *, void *, int, int);  /* The callback function */
    void (*destroy)(void *);
    mouse_h mouse;
    struct Dlg_head *parent;
    widget_data wdata;
    widget_data wcontainer;   /* For children of midnight_dlg */
    enum {
	AREA_TOP,
	AREA_LEFT,
	AREA_RIGHT,
	AREA_BOTTOM
    } area; /* Used by X platforms, should stay here always because the size
	       of this structure has to be same everywhere :) */
    WLay  layout;
} Widget;

typedef struct hist_entry {
    struct hist_entry *prev;
    struct hist_entry *next;
    char   *text;
} Hist;

typedef int sortfn (const file_entry *a, const file_entry *b);

#define LIST_TYPES      5

typedef struct {
    Widget   widget;
    dir_list dir;               /* Directory contents */

    int      list_type;         /* listing type (was view_type) */
    int      active;            /* If panel is currently selected */
    char     cwd [MC_MAXPATHLEN];/* Current Working Directory */
    char     lwd [MC_MAXPATHLEN];/* Last Working Directory */
    Hist     *dir_history;      /* directory history */
    char     *hist_name;        /* directory history name for history file */
    int      count;             /* Number of files in dir structure */
    int      marked;            /* Count of marked files */
    int      dirs_marked;       /* Count of marked directories */
    off_t    total;             /* Bytes in marked files */
    int      top_file;          /* The file showed on the top of the panel */
    int      selected;          /* Index to the selected file */
    int      reverse;           /* Show listing in reverse? */
    int      case_sensitive;    /* Listing is case sensitive? */
    int      split;             /* Split panel to allow two columns */
    int      is_panelized;      /* Flag: special filelisting, can't reload */
    int      frame_size;        /* half or full frame */
    int      icons_per_row;     /* Icon view; how many icons displayed per row */
    sortfn   *sort_type;        /* Sort type */
    char     *filter;           /* File name filter */

    int      dirty;             /* Should we redisplay the panel? */

    int      user_mini_status;          /* Is user_status_format used */
    char     *user_format;              /* User format */
    char     *user_status_format[LIST_TYPES];/* User format for status line */

    format_e *format;           /* Display format */
    format_e *status_format;    /* Mini status format */

    int      format_modified;   /* If the format was changed this is set */

    char     *panel_name;       /* The panel name */
    struct   stat dir_stat;     /* Stat of current dir: used by execute () */

    char     *gc;
    void     *font;
    int      item_height;
    int      total_width;
    int      ascent;
    int      descent;

    int      searching;
    char     search_buffer [256];

    int      has_dir_sizes;     /* Set if directories have sizes = to du -s */

} WPanel;

typedef void (*callfn) (void);
/* FIXME: We have to leave this type ambiguous, because `callfn' is
   used both for functions that take an argument and ones that don't.
   That ought to be cleared up. */

typedef struct {
    char first_letter;
    const char *text;
    int  hot_key;
    callfn call_back;
} menu_entry;

typedef struct {
    const char *name;
    int    count;
    int    max_entry_len;
    int    selected;
    const menu_entry *entries;
    int    start_x;             /* position relative to menubar start */
} sMenu;
typedef sMenu *Menu;

static Menu create_menu (const char *name, const menu_entry *entries, int count);


/* Initialization-File Functions. From the Wine project */

#define STRSIZE 4096
#define overflow (next == &CharBuffer [STRSIZE-1])

enum { FirstBrace, OnSecHeader, IgnoreToEOL, KeyDef, KeyDefOnKey, KeyValue };

typedef struct TKeys {
    char *KeyName;
    char *Value;
    struct TKeys *link;
} TKeys;

typedef struct TSecHeader {
    char *AppName;
    TKeys *Keys;
    struct TSecHeader *link;
} TSecHeader;

typedef struct TProfile {
    char *FileName;
    TSecHeader *Section;
    struct TProfile *link;
} TProfile;


/* It is crucial that JMAX_COLORS must be less than 128 since the high bit
 * is used to indicate a character from the ACS (alt char set). */
#define JMAX_COLORS 256
#define JNORMAL_COLOR 0


#ifndef MC_MOUSE_H
#define MC_MOUSE_H

/* Equivalent definitions for non-GPM mouse support */
/* These lines are modified version from the lines appearing in the */
/* gpm.h include file of the Linux General Purpose Mouse server */

#define GPM_B_LEFT      4
#define GPM_B_MIDDLE    2
#define GPM_B_RIGHT     1


/* Type of mouse */
#define NO_MOUSE 0
#define GPM_MOUSE 1
#define XTERM_MOUSE 2

static void init_mouse (void);
static void shut_mouse (void);

/* Constants returned from mouse handlers */

#define MOU_NORMAL    0x00
#define MOU_REPEAT    0x01
#define MOU_ENDLOOP   0x02
#define MOU_LOCK      0x04

#endif /* MC_MOUSE_H */


/* The button bar menu */
typedef struct {
    Widget widget;

    int    active;              /* If the menubar is in use */
    int    dropped;             /* If the menubar has dropped */
    Menu   *menu;               /* The actual menus */
    int    items;
    int    selected;            /* Selected menu on the top bar */
    int    subsel;              /* Selected entry on the submenu */
    int    max_entry_len;       /* Cache value for the columns in a box */
    int    previous_selection;  /* Selected widget before activating menu */
} WMenu;

static WMenu *menubar_new (int y, int x, int cols, Menu menu [], int items);

typedef int (*qsortfn) (const void *, const void *);

#ifndef __PANEL_H
#define __PANEL_H


enum list_types {
    list_full,                  /* Name, size, perm/date */
    list_brief,                 /* Name */
    list_long,                  /* Like ls -l */
    list_user,                  /* User defined */
    list_icons                  /* iconic display */
};

enum view_modes {
    view_listing,               /* Directory listing */
    view_info,                  /* Information panel */
    view_tree,                  /* Tree view */
    view_quick,                 /* Quick view */
    view_nothing                /* Undefined */
};

enum panel_display_enum {
    frame_full,                 /* full screen frame */
    frame_half                  /* half screen frame */
};

#define is_view_special(x) ((x == view_info) || (x == view_quick))

#define J_LEFT  0
#define J_RIGHT 1

#define NORMAL          0
#define SELECTED        1
#define MARKED          2
#define MARKED_SELECTED 3
#define STATUS          5


static WPanel *panel_new (const char *panel_name);
static stdcall2 void Xtry_to_select (WPanel *panel, char *name);

#define selection(p) (&(p->dir.list [p->selected]))


#define PANEL_ISVIEW(p) (p->view_type == view_brief || \
			 p->view_type == view_full  || \
			 p->view_type == view_long  || \
			 p->view_type == view_user  || \
			 p->view_type == view_tree)

static void set_colors (WPanel *panel);
static void paint_panel (WPanel *panel);
static void format_file (char *dest, WPanel *panel, int file_index, int panel_width, int attr, int isstatus);
static void repaint_file (WPanel *panel, int file_index, int move, int attr, int isstatus);
static void display_mini_info (WPanel *panel);
static void panel_reload (WPanel *panel);
static void paint_dir (WPanel *panel);
static void show_dir (WPanel *panel);

static void paint_frame (WPanel *panel);
static void panel_update_contents (WPanel *panel);
static void panel_update_cols (Widget *widget, int frame_size);
static char *panel_format (WPanel *panel);
static char *mini_status_format (WPanel *panel);
static int set_panel_formats (WPanel *p);

static WPanel *get_other_panel (void);

#define other_panel get_other_panel()


static void try_to_select (WPanel *panel, char *name);

#define DEFAULT_USER_FORMAT "half type,name,|,size,|,perm"

/* This were in main: */
static void unmark_files (WPanel *panel);
static void select_item (WPanel *panel);
static int ITEMS (WPanel *p);
static void unselect_item (WPanel *panel);

static char *string_file_type (file_entry *fe, int len);
static char *string_file_size_brief (file_entry *fe, int len);
static char *string_file_permission (file_entry *fe, int len);
static char *string_file_nlinks (file_entry *fe, int len);
static char *string_file_owner (file_entry *fe, int len);
static char *string_file_group (file_entry *fe, int len);
static char *string_file_size (file_entry *fe, int len);
static char *string_file_mtime (file_entry *fe, int len);
static char *string_file_atime (file_entry *fe, int len);
static char *string_file_ctime (file_entry *fe, int len);
static char *string_file_name (file_entry *fe, int len);
static char *string_space (file_entry *fe, int len);
static char *string_dot (file_entry *fe, int len);
static char *string_marked (file_entry *fe, int len);
static char *string_file_perm_octal (file_entry *fe, int len);
static char *string_inode (file_entry *fe, int len);
static char *string_file_ngid (file_entry *fe, int len);
static char *string_file_nuid (file_entry *fe, int len);

static void file_mark (WPanel *panel, int Index, int val);
static void do_file_mark (WPanel *panel, int Index, int val);
static int  file_compute_color (int attr, file_entry *fe);
static int  file_entry_color (file_entry *fe);

static void update_one_panel_widget (WPanel *panel, int force_update, char *current_file);

#endif  /* __PANEL_H */

static void reread_cmd (void);

#ifndef __VIEW_H
#define __VIEW_H

/* The growing buffers data types */
typedef struct {
    unsigned char *data;
    int  present;               /* Unused, for DOS port maybe */
} block_ptr_t;

enum ViewSide {
    view_side_left,
    view_side_right
};

typedef struct {
    Widget widget;

    char *filename;             /* Name of the file */
    char *command;              /* Command used to pipe data in */
    char *localcopy;
    int view_active;
    int have_frame;

    unsigned char *data;        /* Memory area for the file to be viewed */

    /* File information */
    int file;                   /* File descriptor (for mmap and munmap) */
    FILE *stdfile;              /* Stdio struct for reading file in parts */
    int reading_pipe;           /* Flag: Reading from pipe(use popen/pclose) */
    off_t bytes_read;           /* How much of file is read */
    int mmapping;               /* Did we use mmap on the file? */

    /* Display information */
    off_t last;                 /* Last byte shown */
    off_t last_byte;            /* Last byte of file */
    off_t first;                /* First byte in file */
    off_t bottom_first;  /* First byte shown when very last page is displayed */
				/* For the case of WINCH we should reset it to -1 */
    off_t start_display;        /* First char displayed */
    int  start_col;             /* First displayed column, negative */
    int  edit_cursor;           /* HexEdit cursor position in file */
    char hexedit_mode;          /* Hexidecimal editing mode flag */
    char nib_shift;             /* A flag for inserting nibbles into bytes */
    enum ViewSide view_side;    /* A flag for the active editing panel */
    int  file_dirty;            /* Number of changes */
    int  start_save;            /* Line start shift between Ascii and Hex */
    int  cursor_col;            /* Cursor column */
    int  cursor_row;            /* Cursor row */
    struct hexedit_change_node *change_list;   /* Linked list of changes */

    int dirty;                  /* Number of skipped updates */
    int wrap_mode;              /* wrap_mode */

    /* Mode variables */
    int hex_mode;               /* Hexadecimal mode flag */
    int bytes_per_line;         /* Number of bytes per line in hex mode */
    int viewer_magic_flag;      /* Selected viewer */
    int viewer_nroff_flag;      /* Do we do nroff style highlighting? */

    /* Growing buffers information */
    int growing_buffer;         /* Use the growing buffers? */
    block_ptr_t *block_ptr;     /* Pointer to the block pointers */
    int          blocks;        /* The number of blocks in *block_ptr */


    /* Search variables */
    int search_start;           /* First character to start searching from */
    int found_len;              /* Length of found string or 0 if none was found */
    char *search_exp;           /* The search expression */
    int  direction;             /* 1= forward; -1 backward */
    void (*last_search)(void *, char *);
				/* Pointer to the last search command */
    int view_quit;              /* Quit flag */

    int monitor;                /* Monitor file growth (like tail -f) */
    /* Markers */
    int marker;                 /* mark to use */
    int marks [10];             /* 10 marks: 0..9 */

    int  move_dir;              /* return value from widget:
				 * 0 do nothing
				 * -1 view previous file
				 * 1 view next file
				 */
    struct stat s;              /* stat for file */
} WView;

#define vwidth (view->widget.cols - (view->have_frame ? 2 : 0))
#define vheight (view->widget.lines - (view->have_frame ? 2 : 0))

/* Creation/initialization of a new view widget */
static WView *view_new (int y, int x, int cols, int lines, int is_panel);
static int view_init (WView *view, char *_command, char *_file, int start_line);

/* Internal view routines */
static void view_status        (WView *);
static void view_percent       (WView *, int, int);
static void view_update        (WView *view);
static void view_labels        (WView *view);
static int  view_event         (WView *view, Gpm_Event *event,int *result);
static void toggle_wrap_mode   (WView *);
static void toggle_hex_mode    (WView *);
static void goto_line          (WView *);
static void regexp_search_cmd  (WView *);
static void normal_search_cmd  (WView *);
static void continue_search    (WView *);
static void change_nroff       (WView *view);
static void set_monitor        (WView *view, int set_on);
static void view_move_forward  (WView *view, int i);
static void view_move_backward (WView *view, int i);

/* Command: view a file, if _command != NULL we use popen on _command */
/* move direction should be apointer that will hold the direction in which the user */
/* wants to move (-1 previous file, 1 next file, 0 do nothing) */
static int View (char *_command, char *_file, int *move_direction, int start_line);

/* Maxlimit for skipping updates */
#define max_dirt_limit 10

static void view_adjust_size (Dlg_head *h);

/* A node for building a change list on change_list */
struct hexedit_change_node {
   struct hexedit_change_node *next;
   off_t                      offset;
   unsigned char              value;
};

#endif  /* __VIEW_H */

/* taken from regex.c: */

#define ISASCII(c) 1

typedef void (*movefn)(void *, int);

#ifndef __WIN_H
#define __WIN_H


/* Labels at the screen bottom */

/* Keys managing */
static int check_movement_keys (int c, int additional, int page_size, void *,
    movefn backfn, movefn forfn, movefn topfn, movefn bottomfn);


static void do_enter_ca_mode (void);
static void do_exit_ca_mode (void);
#define wclr(w) wclrn(w, 0)

#endif  /* __WIN_H */

#ifndef __LAYOUT_H
#define __LAYOUT_H

static void done_screen (void);
static void setup_panels (void);
static void change_screen_size (void);
static void layout_change (void);
static void set_display_type (int num, int type);
static void swap_panels (void);
static int get_display_type (int Index);
static int get_current_type (void);
static int get_other_type (void);
static int get_current_index (void);
static int get_other_index (void);
static char *get_nth_panel_name (int num, char buffer[20]);

static Widget *get_panel_widget (int Index);

static void set_hintbar (char *str);

#endif /* __LAYOUT_H */

#ifndef __EXT_H
#define __EXT_H

static char *regex_command (char *filename, char *action, char **drops, int *move_dir);

/* Call it after the user has edited the mc.ext file,
 * to flush the cached mc.ext file
 */
static void flush_extension_file (void);

#define MC_USER_EXT ".mc/ext"
#define MC_LIB_EXT  "mc.ext"
#endif  /* __EXT_H */

#ifndef __COLOR_H
#define __COLOR_H

#define PORT_COLOR(co,bw) (use_colors?co:bw)

#define NORMAL_COLOR          (PORT_COLOR ((1), 0))
#define SELECTED_COLOR        (PORT_COLOR ((2),A_REVERSE))
#define MARKED_COLOR          (PORT_COLOR ((3),A_BOLD))
#define MARKED_SELECTED_COLOR (PORT_COLOR ((4),(SLtt_Use_Ansi_Colors ? A_BOLD_REVERSE : A_REVERSE | A_BOLD)))
#define ERROR_COLOR           (PORT_COLOR ((5),0))
#define MENU_ENTRY_COLOR      (PORT_COLOR ((6),A_REVERSE))
#define REVERSE_COLOR         (PORT_COLOR ((7),A_REVERSE))
#define Q_SELECTED_COLOR      (PORT_COLOR (SELECTED_COLOR, 0))
#define Q_UNSELECTED_COLOR    REVERSE_COLOR
#define VIEW_UNDERLINED_COLOR (PORT_COLOR ((12),A_UNDERLINE))
#define MENU_SELECTED_COLOR   (PORT_COLOR ((13),A_BOLD))
#define MENU_HOT_COLOR        (PORT_COLOR ((14),0))
#define MENU_HOTSEL_COLOR     (PORT_COLOR ((15),0))

/*
 * This should be selectable independently. Default has to be black background
 * foreground does not matter at all.
 */
#define GAUGE_COLOR        (PORT_COLOR ((21),0))
#define INPUT_COLOR        (PORT_COLOR ((22),0))

#define DEFAULT_COLOR  (PORT_COLOR ((31),0))
#define HELP_NORMAL_COLOR  (PORT_COLOR ((16),A_REVERSE))
#define HELP_ITALIC_COLOR  (PORT_COLOR ((17),A_REVERSE))
#define HELP_BOLD_COLOR    (PORT_COLOR ((18),A_REVERSE))
#define HELP_LINK_COLOR    (PORT_COLOR ((19),0))
#define HELP_SLINK_COLOR   (PORT_COLOR ((20),A_BOLD))

#define COLOR_NORMAL       (PORT_COLOR ((8),A_REVERSE))
#define COLOR_FOCUS        (PORT_COLOR ((9),A_BOLD))
#define COLOR_HOT_NORMAL   (PORT_COLOR ((10),0))
#define COLOR_HOT_FOCUS    (PORT_COLOR ((11),0))

/* Add this to color panel, on BW all pairs are normal */
#define STALLED_COLOR      (PORT_COLOR ((12),0))

#define DIRECTORY_COLOR    (PORT_COLOR ((23),0))
#define EXECUTABLE_COLOR   (PORT_COLOR ((24),0))
#define LINK_COLOR         (PORT_COLOR ((25),0))
#define DEVICE_COLOR       (PORT_COLOR ((26),0))
#define SPECIAL_COLOR      (PORT_COLOR ((27),0))
#define CORE_COLOR         (PORT_COLOR ((28),0))

#endif /* __COLOR_H */

#ifndef MC_DLG_H
#define MC_DLG_H


/* Color constants */
#define FOCUSC           h->color[1]
#define NORMALC          h->color[0]
#define HOT_NORMALC      h->color[2]
#define HOT_FOCUSC       h->color[3]

/* Possible directions */
#define DIR_FORWARD     1
#define DIR_BACKWARD    0

/* Common return values */
#define B_EXIT          0
#define B_CANCEL        1
#define B_ENTER         2
#define B_HELP          3
#define B_USER          100

/* Widget messages */
enum {
    WIDGET_INIT,                /* Initialize widget */
    WIDGET_FOCUS,               /* Draw widget in focused state */
    WIDGET_UNFOCUS,             /* Draw widget in unfocused state */
    WIDGET_DRAW,                /* Sent to widget to draw themselves */
    WIDGET_KEY,                 /* Sent to widgets on key press */
    WIDGET_HOTKEY,              /* Sent to widget to catch preprocess key */
    WIDGET_DESTROY,             /* Sent to widget at destruction time */
    WIDGET_CURSOR,              /* Sent to widget to position the cursor */
    WIDGET_IDLE,                /* Send to widgets with options & W_WANT_IDLE*/
    WIDGET_USER  = 0x100000

} /* Widget_Messages */;

enum {
    MSG_NOT_HANDLED,
    MSG_HANDLED
} /* WRET */;

/* Widgets are expected to answer to the following messages:

   WIDGET_FOCUS:   1 if the accept the focus, 0 if they do not.
   WIDGET_UNFOCUS: 1 if they accept to release the focus, 0 if they don't.
   WIDGET_KEY:     1 if they actually used the key, 0 if not.
   WIDGET_HOTKEY:  1 if they actually used the key, 0 if not.
*/

/* Dialog messages */
enum {
    DLG_KEY,                    /* Sent on keypress before sending to widget */
    DLG_INIT,                   /* Sent on init */
    DLG_END,                    /* Sent on shutdown */
    DLG_ACTION,
    DLG_DRAW,                   /* Sent for updating dialog managed area */
    DLG_FOCUS,                  /* Sent on give focus to a widget */
    DLG_UNFOCUS,                /* Sent on remove focus from widget */
    DLG_ONE_UP,                 /* Sent on selecting next */
    DLG_ONE_DOWN,               /* Sent on selecting prev */
    DLG_POST_KEY,               /* Sent after key has been sent */
    DLG_IDLE,                   /* Sent if idle is active */
    DLG_UNHANDLED_KEY,          /* Send if no widget wanted the key */
    DLG_HOTKEY_HANDLED,         /* Send if a child got the hotkey */
    DLG_PRE_EVENT               /* Send before calling get_event */
} /* Dialog_Messages */;

/* XView widget layout */


/* The options for the widgets */
#define  W_WANT_POST_KEY     1
#define  W_WANT_HOTKEY       2
#define  W_WANT_CURSOR       4
#define  W_WANT_IDLE         8
#define  W_IS_INPUT         16
#define  W_PANEL_HIDDEN     32


/* draw box in window */
static void draw_box (Dlg_head *h, int y, int x, int ys, int xs);

/* doubled line if possible */
static void draw_double_box (Dlg_head *h, int y, int x, int ys, int xs);

/* Creates a dialog head  */
static Dlg_head *create_dlg (int y1, int x1, int lines, int cols,
		      int *col,
		      int (*callback) (struct Dlg_head *, int, int),
		      const char *help_ctx, int flags);

/* The flags: */
#define DLG_NO_TOPLEVEL 32      /* GNOME only: Do not create a toplevel window, user provides it */
#define DLG_GNOME_APP   16      /* GNOME only: use a gnome-app for the toplevel window */
#define DLG_NO_TED       8      /* GNOME only: do not manage layout with a GNOME GtkTed widget */
#define DLG_GRID         4      /* Widgets should be created under .widgets */
#define DLG_TRYUP        2      /* Try to move two lines up the dialog */
#define DLG_CENTER       1      /* Center the dialog */
#define DLG_NONE         0      /* No options */
static int  add_widget           (Dlg_head *dest, void *pWidget);
static int  add_widgetl          (Dlg_head *dest, void *pWidget, WLay layout);

/* Runs dialog d */
static void run_dlg              (Dlg_head *d);

static void dlg_run_done         (Dlg_head *h);
static void dlg_process_event    (Dlg_head *h, int key, Gpm_Event *event);
static void init_dlg             (Dlg_head *h);

/* To activate/deactivate the idle message generation */
static void set_idle_proc        (Dlg_head *d, int state);

static void dlg_redraw           (Dlg_head *h);
static void dlg_refresh          (void *parameter);
static void destroy_dlg          (Dlg_head *h);

static void widget_set_size      (Widget *widget, int x1, int y1, int x2, int y2);

static void dlg_broadcast_msg_to (Dlg_head *h, int message, int reverse, int flags);
static void dlg_broadcast_msg    (Dlg_head *h, int message, int reverse);

typedef void  (*destroy_fn)(void *);
typedef int   (*callback_fn)(Dlg_head *, void *, int, int);

static void init_widget (Widget *w, int y, int x, int lines, int cols,
		  callback_fn callback, destroy_fn destroy,
		  mouse_h mouse_handler);

/* Various default service provision */
static int default_dlg_callback  (Dlg_head *h, int id, int msg);
static int default_proc          (Dlg_head *h, int Msg, int Par);

#define real_widget_move(w, _y, _x) move((w)->y + _y, (w)->x + _x)
#define dlg_move(h, _y, _x) move(((Dlg_head *) h)->y + _y, \
			     ((Dlg_head *) h)->x + _x)

#define widget_move(w,y,x) real_widget_move((Widget*)w,y,x)

static int  send_message         (Dlg_head *h, Widget *w, int msg, int par);
static void dlg_erase            (Dlg_head *h);
static void dlg_stop             (Dlg_head *h);

/* Widget selection */
static int  dlg_select_widget     (Dlg_head *h, void *widget);
static void dlg_one_up            (Dlg_head *h);
static void dlg_one_down          (Dlg_head *h);
static int  dlg_focus             (Dlg_head *h);

/* Sets/clear the specified flag in the options field */
#define widget_option(w,f,i) \
    w.options = ((i) ? (w.options | (f)) : (w.options & (~(f))))

#define widget_want_cursor(w,i) widget_option(w, W_WANT_CURSOR, i)
#define widget_want_hotkey(w,i) widget_option(w, W_WANT_HOTKEY, i)


/* Layout definitions */

static void x_set_dialog_title (Dlg_head *h, const char *title);

#endif /* MC_DLG_H */

typedef struct {
    Widget widget;
    int ready;
} WInfo;

#ifndef __WIDGET_H
#define __WIDGET_H

#define C_BOOL          1
#define C_CHANGE        2

/* Please note that the first element in all the widgets is a     */
/* widget variable of type Widget.  We abuse this fact everywhere */
/* Widget_Items */

#define HIDDEN_BUTTON           0
#define NARROW_BUTTON           1
#define NORMAL_BUTTON           2
#define DEFPUSH_BUTTON          3

typedef struct WButton {
    Widget widget;
    int action;                 /* what to do when pressed */
    int selected;               /* button state */
    unsigned int flags;         /* button flags */
    char *text;                 /* text of button */
    int  hotkey;                /* hot KEY */
    int  hotpos;                /* offset hot KEY char in text */
    int  (*callback)(int, void*); /* Callback function */
    void *callback_data;
} WButton;

typedef struct WRadio {
    Widget widget;
    unsigned int state;         /* radio button state */
    int pos, sel;
    int count;                  /* number of members */
    const char * const *texts;  /* texts of labels */
    int upper_letter_is_hotkey; /* If true, then the capital letter is a hk */
} WRadio;

typedef struct WCheck {
    Widget widget;
    unsigned int state;         /* check button state */
    char *text;                 /* text of check button */
    int hotkey;                 /* hot KEY */
    int hotpos;                 /* offset hot KEY char in text */
} WCheck;

typedef struct WGauge {
    Widget widget;
    int shown;
    off_t max;
    off_t current;
} WGauge;


static Hist *history_get (const char *input_name);
static char *show_hist (Hist *history, int widget_y, int widget_x);

typedef struct {
    Widget widget;
    int  point;                 /* cursor position in the input line */
    int  mark;                  /* The mark position */
    int  first_shown;           /* Index of the first shown character */
    int  current_max_len;       /* Maximum length of input line */
    int  field_len;             /* Length of the editing field */
    int  color;                 /* color used */
    int  first;                 /* Is first keystroke? */
    int  disable_update;        /* Do we want to skip updates? */
    int  is_password;           /* Is this a password input line? */
    char *buffer;               /* pointer to editing buffer */
    Hist *history;              /* The history */
    int  need_push;             /* need to push the current Input on hist? */
    char **completions;         /* Possible completions array */
    int  completion_flags;      /* INPUT_COMPLETE* bitwise flags(complete.h) */
    int  inserted_one;          /* TK: just one char inserted, nothing fancy */
    char *history_name;         /* name of history for loading and saving */
} WInput;

/* For history load-save functions */
#define INPUT_LAST_TEXT ((char *) 2)
#define HISTORY_FILE_NAME ".mc/history"

typedef struct {
    Widget widget;
    int    auto_adjust_cols;    /* compute widget.cols from strlen(text)? */
    char   *text;
    int    transparent;         /* Paint in the default color fg/bg */
} WLabel;

typedef struct WLEntry {
    char *text;                 /* Text to display */
    int  hotkey;
    void *data;                 /* Client information */
    struct WLEntry *next;
    struct WLEntry *prev;
} WLEntry;

enum {
    listbox_begin, listbox_end
} /* listbox_insert */;

/* Listbox actions when selecting an option: */
enum {
    listbox_nothing,
    listbox_finish,             /* finish dialog */
    listbox_cback               /* call the callback routine */
} /* listbox_action */;

typedef int (*lcback) (void *);

typedef struct {
    Widget widget;
    WLEntry *list;              /* Pointer to the circular double linked list. */
    WLEntry *top;               /* The first element displayed */
    WLEntry *current;           /* The current element displayed */
    int pos;                    /* Cur. pos, must be kept in sync with current */
    int count;                  /* Number of items in the listbox */
    int width;
    int height;                 /* Size of the widget */
    int action;                 /* Action type */
    int allow_duplicates;       /* Do we allow duplicates on the list? */
    int scrollbar;              /* Draw a scrollbar? */
    lcback cback;               /* The callback function */
    int cursor_x, cursor_y;     /* Cache the values */
} WListbox;

typedef struct {
    Widget widget;
    int    visible;             /* Is it visible? */
    struct {
	char   *text;
	void   (*function)(void *data);
	void   *data;
    } labels [10];
} WButtonBar;

/* Constructors */
static WButton *button_new   (int y, int x, int action, int flags, const char *text,
			int (*callback)(int, void *), void *extra);
static WRadio  *radio_new    (int y, int x, int count, const char * const *texts, int use_hotkey);
static WCheck  *check_new    (int y, int x, int state, const char *text);
static WInput  *input_new    (int y, int x, int color, int len, const char *text);
static WLabel  *label_new    (int y, int x, const char *text);
static WGauge  *gauge_new    (int y, int x, int shown, int max, int current);
static WListbox *listbox_new (int x, int y, int width, int height, int action, lcback);

/* Input lines */
static void winput_set_origin (WInput *i, int x, int field_len);
static int handle_char (WInput *in, int c_code);
static int is_in_input_map (WInput *in, int c_code);
static void update_input (WInput *in, int clear_first);
static void new_input (WInput *in);
static int push_history (WInput *in, char *text);
static void stuff (WInput *in, char *text, int insert_extra_space);
static void input_enable_update (WInput *in);
static void input_set_point (WInput *in, int pos);
static void assign_text (WInput *in, char *text);

/* Labels */
static void label_set_text (WLabel *label, const char *text);

/* Gauges */
static void gauge_set_value (WGauge *g, off_t max, off_t current);
static void gauge_show (WGauge *g, int shown);

/* Buttons */
static void button_set_text (WButton *b, const char *text);

/* search text int listbox entries */
static WLEntry *listbox_search_text (WListbox *l, const char *text);
static void listbox_select_entry (WListbox *l, WLEntry *dest);
static void listbox_select_by_number (WListbox *l, int n);
static void listbox_remove_current (WListbox *l, int force);
static void listbox_remove_list (WListbox *l);
static void listbox_get_current (WListbox *l, char **string, char **extra);

enum append_pos {
    LISTBOX_APPEND_AT_END,      /* append at the end */
    LISTBOX_APPEND_BEFORE,      /* insert before current */
    LISTBOX_APPEND_AFTER        /* insert after current */
};

static char *listbox_add_item (WListbox *l, enum append_pos pos, int
			hotkey, const char *text, void *data);

/* Buttonbar routines */
static WButtonBar *buttonbar_new (int visible);
typedef void (*buttonbarfn )(void *);
typedef void (*voidfn)(void);
static void define_label (Dlg_head *, Widget *paneletc, int Index, const char *text, voidfn);
static void define_label_data (Dlg_head *h, Widget *paneletc, int idx, const char *text,
			buttonbarfn cback, void *data);
static void redraw_labels (Dlg_head *h, Widget *paneletc);
static WButtonBar *find_buttonbar (Dlg_head *h, Widget *paneletc);

#endif  /* __WIDGET_H */

#ifndef __MAIN_H
#define __MAIN_H

enum {
    RP_NOCLEAR,
    RP_CLEAR
};

#define UP_OPTIMIZE 0
#define UP_RELOAD   1
#define UP_ONLY_CURRENT  2

#define UP_KEEPSEL (char *) -1

/* Execute functions: the base and the routines that use it */
static void do_execute (const char *shell, const char *command, int internal_command);
#define execute_internal(command,args) do_execute (command, args, 1)

/* Execute functions that use the shell to execute */
static void shell_execute (char *command, int flags);
static void execute (char *command);

/* If true, after executing a command, wait for a keystroke */
enum { pause_never, pause_on_dumb_terminals, pause_always };

/* The menu panels */
/* If the selected menu was the right */
#define MENU_PANEL (is_right ? right_panel : left_panel)
#define SELECTED_IS_PANEL (get_display_type (is_right ? 1 : 0) == view_listing)

/* Useful macros to avoid too much typing */
#define cpanel current_panel
#define opanel get_other_panel()

typedef void (*key_callback) ();
/* FIXME: We have to leave this type ambiguous, because `key_callback'
   is used both for functions that take an argument and ones that don't.
   That ought to be cleared up. */

/* The keymaps are of this type */
typedef struct {
    int   key_code;
    key_callback fn;
} key_map;

static void update_panels (int force_update, char *current_file);
static void repaint_screen (void);
static void suspend_cmd (void);

static void clr_scr (void);
static void restore_console (void);

enum cd_enum {
    cd_parse_command,
    cd_exact
};

static int do_cd           (char *new_dir, enum cd_enum cd_type);      /* For find.c */
static void change_panel   (void);
static void save_cwds_stat (void);
static void load_hint      (void);

#ifdef USE_NETCODE
static void print_vfs_message(char *msg, ...);
#endif

static int maybe_cd (int char_code, int move_up_dir);

/* Back hack to define the following routines only if the client code
 * has included panel.h
 */
static void directory_history_add   (WPanel *panel, char *s);
static int  do_panel_cd             (WPanel *panel, char *new_dir, enum cd_enum cd_type);

static void edition_post_exec (void);

#define MC_BASE "/.mc/"

#endif  /* __MAIN_H */

#ifndef __COMMAND_H
#define __COMMAND_H

typedef struct {
    WInput input;
    callback_fn old_callback;
} WCommand;

static WCommand *command_new (int y, int x, int len);

/* Return the Input * from a WCommand */
#define input_w(x) (&(x->input))

static void do_cd_command (char *cmd);

#endif /* __COMMAND_H */

#ifndef __COMPLETE_H
#define __COMPLETE_H

#define INPUT_COMPLETE_FILENAMES         1
#define INPUT_COMPLETE_HOSTNAMES         2
#define INPUT_COMPLETE_COMMANDS          4
#define INPUT_COMPLETE_VARIABLES         8
#define INPUT_COMPLETE_USERNAMES        16
#define INPUT_COMPLETE_CD               32

typedef char *CompletionFunction (const char *, int);

static void free_completions (WInput *);
static void complete (WInput *);

#endif  /* __COMPLETE_H */

#ifndef __DIALOG_H
#define __DIALOG_H

#define MSG_ERROR ((char *) -1)
static Dlg_head *MC_message (int error, const char *header, char *text, ...);

static int query_dialog (const char *header, const char *text, int flags, int count, ...);

enum {
   D_NORMAL = 0,
   D_ERROR  = 1,
   D_INSERT = 2
} /* dialog options */;

/* The refresh stack */
typedef struct Refresh {
    void (*refresh_fn)(void *);
    void *parameter;
    int  flags;
    struct Refresh *next;
} Refresh;

/* We search under the stack until we find a refresh function that covers */
/* the complete screen, and from this point we go up refreshing the */
/* individual regions */

enum {
    REFRESH_COVERS_PART,        /* If the refresh fn convers only a part */
    REFRESH_COVERS_ALL          /* If the refresh fn convers all the screen */
};

static void do_refresh (void);


#endif  /* __DIALOG_H */

#ifndef __FILE_H
#define __FILE_H

enum { OP_COPY, OP_MOVE, OP_DELETE };
enum { FILE_CONT, FILE_RETRY, FILE_SKIP, FILE_ABORT };


struct link;

static int copy_file_file (char *s, char *d, int ask_overwrite);
static int move_file_file (char *s, char *d);
static int erase_dir (char *s);
static int erase_dir_iff_empty (char *s);
static int move_dir_dir (char *s, char *d);
static int copy_dir_dir (char *s, char *d, int toplevel, int move_over, int delete, struct link *parent_dirs);

static void create_op_win (int op, int with_eta);
static void destroy_op_win (void);
static int panel_operate (void *source_panel, int op, char *thedefault);
static void file_mask_defaults (void);

/* Error reporting routines */
/* Skip/Retry/Abort routine */
/* This one just displays buf */
static int do_file_error (char *error);

/* Report error with one file */
static int file_error (const char *format, const char *file);

/* Report error with two files */
static int files_error (const char *format, const char *file1, const char *file2);

/* Query routines */
/* Replace existing file */
static int query_replace (char *destname, struct stat *_s_stat, struct stat *_d_stat);

/* Query recursive delete */
static int query_recursive (char *s);

#endif  /* __FILE_H */

#ifndef __BACKGROUND_H
#define __BACKGROUND_H

/* Used for parent/child communication.  These are numbers that
 * could not possible be a routine address.
 */

/* First argument passed to real functions */
enum OperationMode {
    Foreground,
    Background
};

#endif  /* __BACKGROUND_H */

static int stat_file (char *, struct stat *);

#define MAX_FIND_MENU      4   /* Maximum Menu */
#define MAX_FIND_FLINES   11   /* Size of Window Where files are stored */
#define MAX_FIND_TLINES    4   /* Size of Window Where Menu and text .. */
#define MAX_FIND_COLS     48   /* Length of Windows */
#define FIND_DIALOG_SIZE  MAX_FIND_FLINES+MAX_FIND_TLINES
			  /* Total size of the whole dialog */

typedef struct find_list {
    int               selected;  /* Selection field */
    int               ypos;      /* For Scrolling */
    int               isdir;     /* For adding a '\t' on FALSE */
    char              *fname;    /* Name of the file */
    char              *path;     /* For changing panel */
    struct find_list  *up;
    struct find_list  *down;
} find_list;

typedef struct FStack {         /* The Stack will be used to store */
    char *dir_name;             /* the directories to search */
    struct FStack *next;        /*  single-linked */
} FStack;

#define LIST_VFSLIST    0x01
#define LIST_HOTLIST    0x02
#define LIST_MOVELIST   0x04

static void add2hotlist_cmd (void);
static char *hotlist_cmd (int list_vfs);
static void load_hotlist (void);
static int  save_hotlist (void);
static void done_hotlist (void);

typedef struct tree_entry {
    char *name;                 /* The full path of directory */
    int sublevel;               /* Number of parent directories (slashes) */
    long submask;               /* Bitmask of existing sublevels after this entry */
    char *subname;              /* The last part of name (the actual name) */
    int mark;                   /* Flag: Is this entry marked (e. g. for delete)? */
    struct tree_entry *next;    /* Next item in the list */
    struct tree_entry *prev;    /* Previous item in the list */
} tree_entry;

typedef struct {
    Widget     widget;
    tree_entry *tree_first;             /* First entry in the list */
    tree_entry *tree_last;              /* Last entry in the list */
    tree_entry *selected_ptr;           /* The selected directory */
    char       search_buffer [256];     /* Current search string */
    int        done;                    /* Flag: exit tree */
    char       check_name [MC_MAXPATHLEN];/* Directory which is been checked */
    tree_entry *check_start;            /* Start of checked subdirectories */
    tree_entry **tree_shown;            /* Entries currently on screen */
    int        is_panel;                /* panel or plain widget flag */
    int        active;                  /* if it's currently selected */
    int        searching;               /* Are we on searching mode? */
    int topdiff;                        /* The difference between the topmost shown and the selected */
} WTree;

#define tlines(t) (t->is_panel ? t->widget.lines-2 - (show_mini_info ? 2 : 0) : t->widget.lines)

static void load_tree (WTree *tree);
static void save_tree (WTree *tree);
static void show_tree (WTree *tree);
static void tree_chdir (WTree *tree, char *dir);
static void tree_rescan_cmd (WTree *tree);
static int tree_forget_cmd (WTree *tree);
static void tree_copy (WTree *tree, char *default_dest);
static void tree_move (WTree *tree, char *default_dest);
static void tree_event (WTree *tree, int y);
static char *Tree (char *current_dir);

static int search_tree (WTree *ptree, char *text);

static tree_entry *tree_add_entry (WTree *ptree, char *name);
static void tree_remove_entry (WTree *ptree, char *name);
static void tree_destroy (WTree *ptree);
static void tree_check (const char *subname);
static void start_tree_check (WTree *ptree);
static void do_tree_check (WTree *ptree, const char *subname);
static void end_tree_check (WTree *ptree);

static void tree_move_backward (WTree *ptree, int i);
static void tree_move_forward (WTree *ptree, int i);
static int tree_move_to_parent (WTree *ptree);
static void tree_move_to_child (WTree *ptree);
static void tree_move_to_top (WTree *ptree);
static void tree_move_to_bottom (WTree *ptree);

static WTree *tree_new (int is_panel, int y, int x, int lines, int cols);

#define MC_TREE ".mc/tree"

#ifndef __SUBSHELL_H
#define __SUBSHELL_H

/* Used to distinguish between a normal MC termination and */
/* one caused by typing `exit' or `logout' in the subshell */
#define SUBSHELL_EXIT 128

#define use_subshell 0

#endif /* __SUBSHELL_H */

enum {
    CONSOLE_INIT = '1',
    CONSOLE_DONE,
    CONSOLE_SAVE,
    CONSOLE_RESTORE,
    CONSOLE_CONTENTS
};

static void show_console_contents (int starty, unsigned char begin_line, unsigned char end_line);
static void handle_console (unsigned char action);

static void show_rxvt_contents (int starty, unsigned char y1, unsigned char y2);

#ifdef min
#undef min
#endif

#ifdef max
#undef max
#endif

#define min(x,y) (x>y ? y: x)
#define max(x,y) (x>y ? x: y)

/* AIX compiler doesn't understand '\e' */
#define ESC_CHAR '\033'
#define ESC_STR  "\033"

#ifdef USE_BSD_CURSES
#   define xgetch x_getch
#else
#   define xgetch getch
#endif

#ifndef __WTOOLS_H
#define __WTOOLS_H

/* Dialog default background repaint routines */
static void dialog_repaint (struct Dlg_head *h, int back, int title_fore);
static void common_dialog_repaint (struct Dlg_head *h);

/* For common dialogs, just repaint background */
static int  common_dialog_callback (struct Dlg_head *h, int id, int msg);

/* Listbox utility functions */
typedef struct {
    Dlg_head *dlg;
    WListbox *list;
} Listbox;

static Listbox *
create_listbox_window (int cols, int lines, const char *title, char *help);
#define LISTBOX_APPEND_TEXT(l,h,t,d) \
    listbox_add_item (l->list, 0, h, t, d);

static int run_listbox (Listbox *l);

/* Quick Widgets */
enum {
    quick_end, quick_checkbox,
    quick_button, quick_input,
    quick_label, quick_radio
} /* quick_t */;

/* The widget is placed on relative_?/divisions_? of the parent widget */
/* Please note that the contents of the fields in the union are just */
/* used for setting up the dialog.  They are a convenient place to put */
/* the values for a widget */

typedef struct {
    int widget_type;
    int relative_x;
    int x_divisions;
    int relative_y;
    int y_divisions;

    const char *text;           /* Text */
    int  hotkey_pos;            /* the hotkey position */
    int  value;                 /* Buttons only: value of button */
    int  *result;               /* Checkbutton: where to store result */
    char **str_result;          /* Input lines: destination  */
    WLay layout;                /* XView Layouting stuff */
    void *the_widget;           /* For the quick quick dialog manager */
} QuickWidget;

typedef struct {
    int  xlen, ylen;
    int  xpos, ypos; /* if -1, then center the dialog */
    const char *title;
    const char *help;
    QuickWidget *widgets;
} QuickDialog;

static int quick_dialog (const QuickDialog *qd);
static int quick_dialog_skip (const QuickDialog *qd, int nskip);

/* Choosers */

#define CHOOSE_EDITABLE  1
#define CHOOSE_BROWSE    0

/* Chooser dialog boxes */
typedef struct {
    Dlg_head *dialog;
    WListbox *listbox;
} Chooser;

/* The input dialogs */
static char *input_dialog (char *header, char *text, char *def_text);
static char *input_expand_dialog (char *header, char *text, char *def_text);
static char *real_input_dialog_help (char *header, char *text, char *help, char *def_text);

static void query_set_sel (int new_sel);
#endif  /* __WTOOLS_H */
#ifndef __KEY_H
#define __KEY_H

static int get_event (Gpm_Event *event, int redo_event, int block);
static int is_idle (void);
static int ctrl_pressed (void);

static int mi_getch (void);
/* Possible return values from get_event: */
#define EV_MOUSE   -2
#define EV_NONE    -1

/* Used to get the modifier information          */
/* Currently, it just works on the Linux console */
#define SHIFT_PRESSED 1
#define ALTR_PRESSED 2
#define CONTROL_PRESSED 4
#define ALTL_PRESSED 8

/* While waiting for input, the program can select on more than one file */

typedef int (*select_fn)(int fd, void *info);

/* Abort/Quit chars */
static int is_abort_char (int c);

#define XCTRL(x) ((x) & 31)
#define ALT(x) (0x200 | (unsigned int)(x))

/* To define sequences and return codes */
#define MCKEY_NOACTION  0
#define MCKEY_ESCAPE    1

/* Return code for the mouse sequence */
#define MCKEY_MOUSE     -2

static void define_sequence (int code, const char *seq, int action);

/* Learn a single key */
static char *learn_key (void);

/* Returns a key code (interpreted) */
static int get_key_code (int nodelay);

typedef struct {
    int code;
    const char *name;
    const char *longname;
} key_code_name_t;

static const key_code_name_t key_name_conv_tab [] = {
/* KEY_F(0) is not here, since we are mapping it to f10, so there is no reason
   to define f0 as well. Also, it makes Learn keys a bunch of problems :( */
    { KEY_F(1),      "f1",         "Function key 1" },
    { KEY_F(2),      "f2",         "Function key 2" },
    { KEY_F(3),      "f3",         "Function key 3" },
    { KEY_F(4),      "f4",         "Function key 4" },
    { KEY_F(5),      "f5",         "Function key 5" },
    { KEY_F(6),      "f6",         "Function key 6" },
    { KEY_F(7),      "f7",         "Function key 7" },
    { KEY_F(8),      "f8",         "Function key 8" },
    { KEY_F(9),      "f9",         "Function key 9" },
    { KEY_F(10),     "f10",        "Function key 10" },
    { KEY_F(11),     "f11",        "Function key 11" },
    { KEY_F(12),     "f12",        "Function key 12" },
    { KEY_F(13),     "f13",        "Function key 13" },
    { KEY_F(14),     "f14",        "Function key 14" },
    { KEY_F(15),     "f15",        "Function key 15" },
    { KEY_F(16),     "f16",        "Function key 16" },
    { KEY_F(17),     "f17",        "Function key 17" },
    { KEY_F(18),     "f18",        "Function key 18" },
    { KEY_F(19),     "f19",        "Function key 19" },
    { KEY_F(20),     "f20",        "Function key 20" },
    { KEY_BACKSPACE, "bs",         "Backspace key" },
    { KEY_END,       "end",        "End key" },
    { KEY_UP,        "up",         "Up arrow key" },
    { KEY_DOWN,      "down",       "Down arrow key" },
    { KEY_LEFT,      "left",       "Left arrow key" },
    { KEY_RIGHT,     "right",      "Right arrow key" },
    { KEY_HOME,      "home",       "Home key" },
    { KEY_NPAGE,     "pgdn",       "Page Down key" },
    { KEY_PPAGE,     "pgup",       "Page Up key" },
    { KEY_IC,        "insert",     "Insert key" },
    { KEY_DC,        "delete",     "Delete key" },
    { ALT('\t'),     "complete",   "Completion/M-tab" },
    { KEY_KP_ADD,    "kpplus",     "+ on keypad" },
    { KEY_KP_SUBTRACT,"kpminus",   "- on keypad" },
    { KEY_KP_MULTIPLY,"kpasterix", "* on keypad" },
/* From here on, these won't be shown in Learn keys (no space) */
    { KEY_LEFT,      "kpleft",     "Left arrow keypad" },
    { KEY_RIGHT,     "kpright",    "Right arrow keypad" },
    { KEY_UP,        "kpup",       "Up arrow keypad" },
    { KEY_DOWN,      "kpdown",     "Down arrow keypad" },
    { KEY_HOME,      "kphome",     "Home on keypad" },
    { KEY_END,       "kpend",      "End on keypad" },
    { KEY_NPAGE,     "kpnpage",    "Page Down keypad" },
    { KEY_PPAGE,     "kpppage",    "Page Up keypad" },
    { KEY_IC,        "kpinsert",   "Insert on keypad" },
    { KEY_DC,        "kpdelete",   "Delete on keypad" },
    { (int) '\n',    "kpenter",    "Enter on keypad" },
    { (int) '/',     "kpslash",    "Slash on keypad" },
    { (int) '#',     "kpnumlock",  "NumLock on keypad" },
    { 0, 0 }
    };


/* Set keypad mode (xterm and linux console only) */
static void numeric_keypad_mode (void);
static void application_keypad_mode (void);

#endif  /* __KEY_H */

#ifndef __DIR_H
#define __DIR_H

#define MIN_FILES 128
#define RESIZE_STEPS 128

/* Sorting functions */
static int unsorted   (const file_entry *a, const file_entry *b);
static int sort_name  (const file_entry *a, const file_entry *b);
static int sort_ext   (const file_entry *a, const file_entry *b);
static int sort_time  (const file_entry *a, const file_entry *b);
static int sort_atime (const file_entry *a, const file_entry *b);
static int sort_ctime (const file_entry *a, const file_entry *b);
static int sort_size  (const file_entry *a, const file_entry *b);
static int sort_inode (const file_entry *a, const file_entry *b);
static int sort_type  (const file_entry *a, const file_entry *b);
static int sort_links (const file_entry *a, const file_entry *b);
static int sort_nuid  (const file_entry *a, const file_entry *b);
static int sort_ngid  (const file_entry *a, const file_entry *b);
static int sort_owner (const file_entry *a, const file_entry *b);
static int sort_group (const file_entry *a, const file_entry *b);

/* SORT_TYPES is used to build the nice dialog box entries */
#define SORT_TYPES 8

/* This is the number of sort types not available in that dialog box */
#define SORT_TYPES_EXTRA 6

/* The total, used by Tk version */
#define SORT_TYPES_TOTAL (SORT_TYPES + SORT_TYPES_EXTRA)

typedef struct {
    const char *sort_name;
    sortfn *sort_fn;
} sort_orders_t;

#define MY_ISDIR(x) ( (S_ISDIR (x->buf.st_mode) || x->f.link_to_dir) ? 1 : 0)

static const sort_orders_t sort_orders [SORT_TYPES_TOTAL] = {
    { "&Unsorted",    unsorted },
    { "&Name",        sort_name },
    { "&Extension",   sort_ext },
    { "&Modify time", sort_time },
    { "&Access time", sort_atime },
    { "&Change time", sort_ctime },
    { "&Size",        sort_size },
    { "&Inode",       sort_inode },

    /* New sort orders */
    { "&Type",        sort_type },
    { "&Links",       sort_links },
    { "N&GID",        sort_ngid },
    { "N&UID",        sort_nuid },
    { "&Owner",       sort_owner },
    { "&Group",       sort_group }
};

#define string_sortcomp(a,b) (sort_case_sensitive ? strcmp (a,b) : strcasecmp (a,b))
#endif  /* __DIR_H */

/* This file is included by help.c and man2hlp.c */

/* Some useful constants */
#define CHAR_NODE_END           '\04'
#define CHAR_LINK_START         '\01'
#define CHAR_LINK_POINTER       '\02'
#define CHAR_LINK_END           '\03'
#define CHAR_ALTERNATE          '\05'
#define CHAR_NORMAL             '\06'
#define CHAR_VERSION            '\07'
#define CHAR_BOLD_ON            '\010'
#define CHAR_BOLD_OFF           '\013'
#define CHAR_MCLOGO             '\014'
#define CHAR_TEXTONLY_START     '\016'
#define CHAR_TEXTONLY_END       '\017'
#define CHAR_XONLY_START        '\020'
#define CHAR_XONLY_END          '\021'
#define CHAR_TITLE_ON           '\022'
#define CHAR_TITLE_OFF          '\023'
#define CHAR_ITALIC_ON          '\024'
#define CHAR_RESERVED           '\025'
#define STRING_LINK_START       "\01"
#define STRING_LINK_POINTER     "\02"
#define STRING_LINK_END         "\03"
#define STRING_NODE_END         "\04"

static void interactive_display (const char *filename, const char *node);

static void user_menu_cmd (void);
static char *expand_format (char, int);
static int check_format_view (char *);
static int check_format_var (char *p, char **v);
static int check_format_cd (char *);

static void save_layout (void);
static void save_configure (void);
static void save_setup (void);
static void done_setup (void);
static void panel_save_setup (WPanel *panel, char *section);
static void panel_load_setup (WPanel *panel, char *section);

#define load_int(a,b,c) GetPrivateProfileInt(a,b,c,profile_name)
#define load_string(a,b,c,d,e) GetPrivateProfileString(a,b,c,d,e,profile_name)
#define save_string WritePrivateProfileString

#define PROFILE_NAME     ".mc/ini"
#define HOTLIST_FILENAME ".mc/hotlist"

/* Prototypes for the profile management functions */

static int GetPrivateProfileInt (const char * AppName, const char * KeyName, int Default,
			   const char * File);

static int WritePrivateProfileString (const char * AppName, const char * KeyName, const char * String,
				const char * FileName);

static void sync_profiles (void);

static void free_profiles (void);

/* New profile functions */

/* Returns a pointer for iterating on appname section, on profile file */
static void *profile_init_iterator (const char *appname, const char *file);

/* Returns both the key and the value of the current section. */
/* You pass the current iterating pointer and it returns the new pointer */
static void *profile_iterator_next (void *s, char **key, char **value);

/* Removes all the definitions from section appname on file */
static void profile_clean_section (const char *appname, const char *file);
static int profile_has_section (const char *section_name, const char *profile);

/* Forgets about a .ini file, to disable updating of it */
static void profile_forget_profile (const char *file);

#ifndef MAP_FILE
#   define MAP_FILE 0
#endif

#ifndef __FSUSAGE_H
#define __FSUSAGE_H
/* Space usage statistics for a filesystem.  Blocks are 512-byte. */
struct fs_usage
{
  long fsu_blocks;              /* Total blocks. */
  long fsu_bfree;               /* Free blocks available to superuser. */
  long fsu_bavail;              /* Free blocks available to non-superuser. */
  long fsu_files;               /* Total file nodes. */
  long fsu_ffree;               /* Free file nodes. */
};
#endif

#if !defined(NO_INFOMOUNT)
/* A mount table entry. */
struct mount_entry
{
  char *me_devname;             /* Device node pathname, including "/dev/". */
  char *me_mountdir;            /* Mount point directory pathname. */
  char *me_type;                /* "nfs", "4.2", etc. */
  dev_t me_dev;                 /* Device number of me_mountdir. */
  struct mount_entry *me_next;
};
#endif  /* !defined(NO_INFOMOUNT) */

#define NLENGTH(dirent) (strlen ((dirent)->d_name))

#define MAX_INPUT_BUFFER_LEN 1024

typedef unsigned long SLtt_Char_Type;


/* API */
static int   SLang_init_tty (void);
static void  SLang_reset_tty (void);
static void  SLang_ungetkey_string (unsigned char *s, unsigned int n);
static void  SLsmg_draw_box (int r, int c, int dr, int dc);
static void  SLsmg_draw_hline (int n);
static void  SLsmg_draw_object (int r, int c, unsigned char object);
static void  SLsmg_draw_vline (int n);
static void  SLsmg_gotorc (int r, int c);
static int   SLsmg_init_smg (void);
static void  SLsmg_normal_video (void);
static void  SLsmg_printf (const char *fmt, ...);
static void  SLsmg_refresh (void);
static void  SLsmg_reset_smg (void);
static void  SLsmg_set_char_set (int i);
static void  SLsmg_set_color (int color);
static void  SLsmg_touch_lines0 (int n);
static void  SLsmg_write_char (char ch);
static void  SLsmg_write_string (const char *str);
static void  SLtt_beep (void);
static void  SLtt_get_terminfo (void);
static void  SLtt_set_color (int obj, const char *fg, const char *bg);
static void  SLtt_set_color_esc (int obj, const char *esc);
static void  SLtt_set_mono (int obj, SLtt_Char_Type mask);
static const char *SLtt_tgetstr (const char *s);
static void  SLtt_write_string (const char *str);

typedef struct Screen_Type
{
     int n;                    /* number of chars written last time */
     int flags;                /* line untouched, etc... */
     unsigned short *old, *neew;
#ifndef pc_system
     unsigned long old_hash, new_hash;
#endif
} Screen_Type;

#define TOUCHED 0x1
#define TRASHED 0x2

#ifndef pc_system
#define MAX_SCREEN_SIZE 120
#else
#define MAX_SCREEN_SIZE 75
#endif

typedef struct
{
   SLtt_Char_Type fgbg;
   SLtt_Char_Type mono;
   char *custom_esc;
} Ansi_Color_Type;

struct FBUTS {
	const char *text;
	int len;        /* length including space and brackets */
	int x;
};

/* This keeps track of the directory stack */
typedef struct dir_stack {
    char *name;
    struct dir_stack *prev;
} dir_stack;


struct LA_CHECK_OPTIONS {
    const char   *text;
    int    *variable;
    WCheck *widget;
};

struct O_CHECK_OPTIONS {
    const char   *text;
    int    *variable;
    void   (*toggle_function)(void);
    WCheck *widget;
};

struct CHECK_PERM {
    mode_t mode;
    const char *text;
    int selected;
    WCheck *check;
};

#define LABELS_CHOWN 5

struct CHOWN_LABEL {
    int y, x;
    WLabel *l;
};

struct colorpair {
    const char *name;               /* Name of the entry */
    const char * fg;                /* foreground color */
    const char * bg;                /* background color */
};

struct _hotlist_but {
    int ret_cmd, flags, y, x;
    const char *text;
    int   type;
};


#define UX_HOTLIST              5
#define UY_HOTLIST              2

#define BX_HOTLIST              UX_HOTLIST
#define BY_HOTLIST              LINES-6

#define LABELS_HOTLIST          3
#define B_ADD_CURRENT           B_USER
#define B_REMOVE_HOTLIST        (B_USER + 1)
#define B_NEW_GROUP             (B_USER + 2)
#define B_NEW_ENTRY             (B_USER + 3)
#define B_UP_GROUP              (B_USER + 4)
#define B_INSERT                (B_USER + 5)
#define B_APPEND                (B_USER + 6)
#define B_MOVE                  (B_USER + 7)

static const struct _hotlist_but hotlist_but_st[] = {
    { B_MOVE, NORMAL_BUTTON,           1, 42, "&Move",       LIST_HOTLIST},
    { B_REMOVE_HOTLIST, NORMAL_BUTTON, 1, 30, "&Remove",     LIST_HOTLIST},
    { B_APPEND, NORMAL_BUTTON,         1, 15, "&Append",     LIST_MOVELIST},
    { B_INSERT, NORMAL_BUTTON,         1,  0, "&Insert",     LIST_MOVELIST},
    { B_NEW_ENTRY, NORMAL_BUTTON,      1, 15, "New &Entry",  LIST_HOTLIST},
    { B_NEW_GROUP, NORMAL_BUTTON,      1,  0, "New &Group",  LIST_HOTLIST},
    { B_CANCEL, NORMAL_BUTTON,         0, 53, "&Cancel",     LIST_HOTLIST|LIST_VFSLIST|LIST_MOVELIST},
    { B_UP_GROUP, NORMAL_BUTTON,       0, 42, "&Up",         LIST_HOTLIST|LIST_MOVELIST},
    { B_ADD_CURRENT, NORMAL_BUTTON,    0, 20, "&Add current",LIST_HOTLIST},
    { B_ENTER, DEFPUSH_BUTTON,         0,  0, "Change &To",  LIST_HOTLIST|LIST_VFSLIST|LIST_MOVELIST}
};
#define ELEMENTS(arr) ( sizeof(arr) / sizeof((arr)[0]) )
#define BUTTONS_HOTLIST         (ELEMENTS(hotlist_but_st))

typedef struct termios TTY_Termio_Type;

/* termcap stuff */

typedef struct
{
   unsigned int name_section_size;
   const char *terminal_names;

   unsigned int boolean_section_size;
   const unsigned char *boolean_flags;

   unsigned int num_numbers;
   const unsigned char *numbers;

   unsigned int num_string_offsets;
   const unsigned char *string_offsets;

   unsigned int string_table_size;
   const char *string_table;

} Terminfo_Type;



enum HotListType {
    HL_TYPE_GROUP,
    HL_TYPE_ENTRY,
    HL_TYPE_COMMENT
};


/* Directory hotlist */
struct hotlist {
    enum  HotListType type;
    char *directory;
    char *label;
    struct hotlist *head;
    struct hotlist *up;
    struct hotlist *next;
};

#define MAXLINKNAME 80
#define HISTORY_SIZE 20
#define HELP_WINDOW_WIDTH 62

struct HISTORY {
    char *page;                 /* Pointer to the selected page */
    char *link;                 /* Pointer to the selected link */
};

/* Link areas for the mouse */
typedef struct Link_Area {
    int x1, y1, x2, y2;
    char *link_name;
    struct Link_Area *next;
} Link_Area;

#define MAX_ENTRIES 40

/* This is a hard link cache */
struct link {
    struct link *next;
    vfs *vfs;
    dev_t dev;
    ino_t ino;
    short linkcount;
    umode_t st_mode;
    char name[1];
};

typedef struct key_def {
    char ch;                    /* Holds the matching char code */
    int code;                   /* The code returned, valid if child == NULL */
    struct key_def *next;
    struct key_def *child;      /* sequence continuation */
    int action;                 /* optional action to be done. Now used only
				   to mark that we are just after the first
				   Escape */
} key_def;

#define UID_CACHE_SIZE 200
#define GID_CACHE_SIZE 30

typedef struct {
    int  Index;
    char *string;
} int_cache;

struct HOTLIST_STATE {
    /*
     * these parameters are intended to be user configurable
     */
    int         expanded;       /* expanded view of all groups at startup */

    /*
     * these reflect run time state
     */

    int         loaded;         /* hotlist is loaded */
    int         readonly;       /* hotlist readonly */
    int         file_error;     /* parse error while reading file */
    int         running;        /* we are running dlg (and have to
				   update listbox */
    int         moving;         /* we are in moving hotlist currently */
    int         modified;       /* hotlist was modified */
    int         type;           /* LIST_HOTLIST || LIST_VFSLIST */
};

typedef struct {
    Widget *button;
    Widget *label;
    int ok;
    char *sequence;
} learnkey;

/* Directory panelize */
struct panelize {
    char *command;
    char *label;
    struct panelize *next;
};

/* The maximum number of views managed by the set_display_type routine */
/* Must be at least two (for current and other).  Please note that until */
/* Janne gets around this, we will only manage two of them :-) */
#define MAX_VIEWS 2

struct DIM_PANELS {
    int    type;
    Widget *widget;
};

struct text_struct {
    const char *text;
    const char *header;
};

CALC_BSS struct BB_COMMON_BSS
{
	const char *SLtt_Graphics_Char_Pairs;        /* ac termcap string -- def is vt100 */
/* 1 if terminal lacks the ability to do into insert mode or into delete
   mode. Currently controlled by S-Lang but later perhaps termcap. */
	const char *UnderLine_Vid_Str;
	const char *Blink_Vid_Str;
	const char *Bold_Vid_Str;
	const char *Eins_Mode_Str; /* = "\033[4l"; */  /* end ins mode (ei) */
	const char *Scroll_R_Str; /* = "\033[%d;%dr"; */ /* scroll region */
	const char *Cls_Str; /* = "\033[2J\033[H"; */  /* cl termcap STR  for ansi terminals */
	const char *Rev_Vid_Str; /* = "\033[7m"; */    /* mr,so termcap string */
	const char *Norm_Vid_Str; /* = "\033[m"; */   /* me,se termcap string */
	const char *Del_Eol_Str; /* = "\033[K"; */            /* ce */
	const char *Del_N_Lines_Str; /* = "\033[%dM"; */  /* DL */
	const char *Add_N_Lines_Str; /* = "\033[%dL"; */  /* AL */
	const char *Rev_Scroll_Str;
	const char *Curs_Up_Str;
	const char *Curs_F_Str;    /* RI termcap string */
	const char *Cursor_Visible_Str;    /* ve termcap string */
	const char *Cursor_Invisible_Str;    /* vi termcap string */
	const char *Start_Alt_Chars_Str;  /* as */
	const char *End_Alt_Chars_Str;   /* ae */
	const char *Enable_Alt_Char_Set;  /* eA */
	const char *Term_Init_Str;
	const char *Keypad_Init_Str;
	const char *Term_Reset_Str;
	const char *Keypad_Reset_Str;
	const char *Curs_Pos_Str; /* = "\033[%i%d;%dH";*/   /* cm termcap string */

	TProfile *Current;
	TProfile *Base;
	unsigned long Blank_Hash;
	int Cls_Flag;
	unsigned char FgBg_Stats[JMAX_COLORS];
	int Color_0_Modified;
	const char *Reset_Color_String;
	int Linux_Console;
	unsigned char Alt_Char_Set[129]; /* 129th is used as a flag */
	int Automatic_Margins;
	int Worthless_Highlight;
    /* This glitch is exclusive to HP term.  Basically it means that to clear
     * attributes, one has to erase to the end of the line.
     */
	int Has_HP_Glitch;
    /* 1 if cursor position known, 0 if not.  -1 if only row is known */
	int Cursor_Set;
	int Scroll_r1;
	int Cursor_r, Cursor_c ;         /* 0 based */
	int last_percentage [3];

    /* Replace dialog: color set, descriptor and filename */
	int replace_colors [4];
	Dlg_head *replace_dlg;
	char *replace_filename;
	int replace_result;

	struct stat *s_stat, *d_stat;
    /* Pointer to the operate dialog */
	Dlg_head *op_dlg;
	int      showing_eta;
	int      showing_bps;
	off_t bps, bps_time;
/* Describe the components in the panel operations window */
	WLabel *FileLabel [2];
	WLabel *FileString [2];
	WLabel *ProgressLabel [3];
	WGauge *ProgressGauge [3];
	WLabel *eta_label;
	WLabel *bps_label;
	WLabel *stalled_label;
	char debug_out_msg [256];
	char unix_error_string_buffer [256];
	char cmd_buf [512];
	gid_t my_gid;
	uid_t my_uid;
	uid_t my_euid;
	/* If use_mouse_p is XTERM_MOUSE: is mouse currently active? */
	int xmouse_flag;

	unsigned int SLang_Input_Buffer_Len;
	unsigned char SLang_Input_Buffer [MAX_INPUT_BUFFER_LEN];
	int   SLang_TT_Baud_Rate;
	int   SLtt_Has_Alt_Charset;
	int   SLtt_Screen_Cols;
	int   SLtt_Screen_Rows;
	int   SLtt_Use_Ansi_Colors;

	/* widget colors for the midnight commander */
	int midnight_colors [4];
	/* Force colors, only used by Slang */
	int force_colors;
/* colors specified on the command line: they override any other setting */
	char *command_line_colors;
	/* File name to view if argument was supplied */
	char *edit_one_file;
	/* Used so that widgets know if they are being destroyed or shut down */
	int midnight_shutdown;
	/* to show nice prompts */
	int last_paused;
	char *profile_name;
	char setup_color_string [4096];
	char term_color_string [4096];
	char color_terminal_string [512];
	/* Asks for confirmation before executing a program by pressing enter */
	int confirm_execute;
	/* Asks for confirmation when using F3 to view a directory and there
	    are tagged files */
	int confirm_view_dir;

	/* This flag indicates if the pull down menus by default drop down */
	int drop_menus;
	/* The home directory */
	char *home_dir;
	/* The value of the other directory, only used when loading the setup */
	char *other_dir;
	char *this_dir;
/* If set and you don't have subshell support,then C-o will give you a shell */
	int output_starts_shell;
	const char *TERM_env;
	int permission_mode;    /* If 1, we use permission hilighting */
	/* To avoid excessive calls to ncurses' has_colors () */
	int hascolors;
	/* Set to force black and white display at program startup */
	int disable_colors;
	/* Set if we are actually using colors */
	int use_colors;
	int dialog_colors [4];
#define CONTROL_FILE "/tmp/mc.%d.control"
	char control_file [sizeof (CONTROL_FILE) + 8];
	const char *MC_shell;   /* The user's shell */
	char *mc_home;       /* mc_home: The home of MC */
	WMenu *the_menubar;  /* The Menubar */
	WButton *b_att[3];      /* permission */
	WButton *b_user, *b_group;      /* owner */

	int files_on_begin;      /* Number of files at startup */
	int flag_pos;
	int x_toggle;
	char ch_flags[11];
	umode_t ch_cmode;
	struct stat *sf_stat;
	int a_need_update;
	int a_end_chown;
	int a_current_file;
	int a_single_set;
	char *g_fname;

	Dlg_head *dd;
	WRadio *my_radio;
	WInput *status;
	WCheck *check_status;
	int current_mode;

	char **displays_status;
	/* Controls whether the array strings have been translated */
	int i18n_displays_flag;

	const char *sort_orders_names [SORT_TYPES];
	int tree_colors [4];

	int dash_number;     /* used to rotate the dash */
	char *ret_timeout;
	char *ret_limit;
	int chm_single_set;

	int mode_change, chm_need_update;
	int c_file, end_chmod;
	umode_t and_mask, or_mask;
	WLabel *statl;
	int normal_color;
	int title_color;
	int selection_color;

	int cho_need_update, end_chown;
	int cho_current_file;
	int cho_single_set;
	WListbox *l_user, *l_group;

	char *editor;
	off_t compare_src_size;
	/* Recursive operation on subdirectories */
	int dive_into_subdirs;
	/* Preserve the original files' owner, group, permissions, and
	   timestamps (owner, group only as root). */
	int preserve;

	/* If running as root, preserve the original uid/gid
	   (we don't want to try chwon for non root)
	   preserve_uidgid = preserve && uid == 0 */
	int preserve_uidgid;

	/* If on, it gets a little scrict with dangerous operations */
	int know_not_what_am_i_doing;

	int stable_symlinks;

	/* With ETA on we have extra screen space */
	int eta_extra;

	/* result from the recursive query */
	int recursive_result;

	/* The estimated time of arrival in seconds */
	double eta_secs;

	char string_inode_buffer [16];

	/* Set if the window has changed it's size */
	int winch_flag;

	/* The prompt */
	char *MC_prompt;

	/* Ugly hack in order to distinguish between left and right panel in menubar */
	/* If the selected menu was the right */
	int is_right;

	int pipefd1 [2], pipefd2 [2];

	Screen_Type SL_Screen[MAX_SCREEN_SIZE];
#define MAX_OUTPUT_BUFFER_SIZE 4096

	unsigned char Output_Buffer[MAX_OUTPUT_BUFFER_SIZE];
	unsigned char *Output_Bufferp;
	int rxvt_extensions;
	int linux_terminfo_offset;
	int SLang_TT_Read_FD;
	int DISPLAY_X;
	int DISPLAY_Y;
	/* scrolling region */
	int Scroll_r2;
	int FIND2_X;
	/* Are the files sorted case sensitively? */
	int sort_case_sensitive;
	int SORT_X;
	/* If true, assume we are running on an xterm terminal */
	int force_xterm;
	/* This flag is set by xterm detection routine in function main() */
	/* It is used by function view_other_cmd() */
	int xterm_flag;
	/* For slow terminals */
	int slow_terminal;
	/* Have we shown the fast-reload warning in the past? */
	int fast_reload_w;
	/* If true: l&r arrows are used to chdir if the input line is empty */
	int navigate_with_arrows;
	/* If true use +, -, | for line drawing */
	int force_ugly_line_drawing;
	/* If true program softkeys (HP terminals only) on startup and after every
	   command ran in the subshell to the description found in the termcap/terminfo
	   database */
	int reset_hp_softkeys;
	int old_esc_mode;
	/* Set when we want use advanced chmod command instead of chmod and/or chown */
	int advanced_chfns;
	/* The number of output lines shown (if available) */
	int output_lines;
	/* Set if you want the possible completions dialog for the first time */
	int show_all_if_ambiguous;
	/* If set then dialogs just clean the screen when refreshing, else */
	/* they do a complete refresh, refreshing all the parts of the program */
	int fast_refresh;
	/* If true, at startup the user-menu is invoked */
	int auto_menu;
	/* If true, use + and \ keys normally and select/unselect do
	   if M-+ / M-\ and M-- and keypad + / - */
	int alternate_plus_minus;
	/* This variable points to a copy of the mc.ext file in memory
	 * With this we avoid loading/parsing the file each time we
	 * need it */
	char *ext_data;
	/* If true, then use stat() on the cwd to determine directory changes */
	int fast_reload;
	/* If true, use some usability hacks by Torben */
	int torben_fj_mode;
	/* if set, show a ruler */
	int ruler;
	/* Used to compute the bottom first variable */
	int have_fast_cpu;
	int default_hex_mode;
	int altered_hex_mode;
	int altered_magic_flag;
	int altered_nroff_flag;
	/* If set, then we have to call the layout_change routine from main */
	int layout_do_change;
	/* Set if the panels are split horizontally */
	int horizontal_split;
	/* First panel size if the panel are not split equally */
	int first_panel_size;
	/* Set if you want the message bar shown in xterm title bar to save space */
	int xterm_hintbar;
	/* The starting line for the output of the subprogram */
	int output_start_y;
	/* Ugh, ugly hack */
	/* Controls appending to files, shared with filequery.c */
	int do_append;
	/* Reget flag */
	int do_reget;
	/* Used to save the hint line */
	int last_hint_line;
	/* Specifies the display mode: 1d or 2d */
	int tree_navigation_flag;
	/* If this is true, then when browsing the tree the other window will
	 * automatically reload it's directory with the contents of the currently
	 * selected directory. */
	int xtree_mode;
	signed char console_flag;
	/* If true show files ending in ~ */
	int show_backups;
	/* If false then directories are shown separately from files */
	int mix_all_files;
	int This_Row, This_Col;
	int error_pipe[2];       /* File descriptors of error pipe */
	int old_error;           /* File descriptor of old standard error */
	/* The structures for the panels */
	WPanel *left_panel;
	WPanel *right_panel;
	/* Pointers to the selected and unselected panel */
	WPanel *current_panel;
	/* The hook list for the select file function */
	Hook *select_file_hook;
	/* Primitive way to check if the the current dialog is our dialog */
	/* This is needed by async routines like load_prompt */
	Dlg_head *current_dlg;
	/* A hook list for idle events */
	Hook *idle_hook;
	/* The dialog handle for the main program */
	Dlg_head *midnight_dlg;
	int quote;
	/* The widget where we draw the prompt */
	WLabel *the_prompt;
	/* The hint bar */
	WLabel *the_hint;
	/* The button bar */
	WButtonBar *the_bar;
	/* The pointer to the tree */
	WTree *the_tree;
	/* This holds the command line */
	WCommand *cmdline;
	struct Dlg_head *ch_dlg;
	umode_t c_stat;
	char *c_fname, *c_fown, *c_fgrp;
	int startup_left_mode;
	int startup_right_mode;
	WInput *WIuser;
	int my_delete;
	int my_overwrite;
	int my_execute;
	int my_exit;
	int box_new_mode;
	int new_meta;
#ifdef USE_VFS
#define MAXEXTFS 32
	char *extfs_prefixes [MAXEXTFS];
	char extfs_need_archive [MAXEXTFS];
	int extfs_no;
	int extfs_errloop;
	int extfs_notadir;
#endif
	Ansi_Color_Type Ansi_Color_Map[JMAX_COLORS];
	SLtt_Char_Type Current_Fgbg;
	unsigned long SLtt_Num_Chars_Output;
	int MC_verbose;
	int OX, opt_first_width, opt_second_width;
	int Video_Initialized;
	int SLtt_Blink_Mode;
	int SLsmg_Display_Eight_Bit;

	struct FBUTS fbuts[8];
	dir_stack *dir_stack_base;
	struct LA_CHECK_OPTIONS la_check_options[10];
	WRadio *radio_widget;
	/* To symlinks the difference between `follow Links' checked and not
	   checked is the stat call used (mc_stat resp. mc_lstat) */
	int (*mc_xstat)(const char *, struct stat *);
	struct O_CHECK_OPTIONS o_check_options[19];
	/* The value of the "preserve Attributes" checkbox in the copy file dialog.
	   We can't use the value of "preserve" because it can change in order to
	   preserve file attributs when moving files across filesystem boundaries
	   (we want to keep the value of the checkbox between copy operations). */
	int op_preserve;
	/* The bits to preserve in created files' modes on file copy */
	int umask_kill;
	/* use mouse? */
	int use_mouse_p;
	/* Controls screen clearing before an exec */
	int clear_before_exec;
	/* Asks for confirmation before deleting a file */
	int confirm_delete;
	/* Asks for confirmation before overwriting a file */
	int confirm_overwrite;
	/* Asks for confirmation before leaving the program */
	int confirm_exit;
#ifdef USE_NETCODE
	/* Source routing destination */
	int source_route;
#endif
	/* If set, use the builtin editor */
	int use_internal_edit;
	/* It true saves the setup when quitting */
	int auto_save_setup;
	/* If true, be eight bit clean */
	int eight_bit_clean;
	/* If true, then display chars 0-255, else iso-8859-1, requires eight_bit_clean */
	int full_eight_bits;
	/* If true use the internal viewer */
	int use_internal_view;
	/* Move page/item? When clicking on the top or bottom of a panel */
	int mouse_move_pages;
	int mou_auto_repeat;
	int double_click_speed;

	int use_8th_bit_as_meta;
	/* Set when cd symlink following is desirable (bash mode) */
	int cd_symlinks;
	/* If true, marking a files moves the cursor down */
	int mark_moves_down;
	/* If true, then the +, - and \ keys have their special meaning only if the
	 * command line is emtpy, otherwise they behave like regular letters */
	int only_leading_plus_minus;
	/* Only used at program boot */
	int boot_current_is_left;
	/* If set, we execute the file command to check the file type */
	int use_file_to_check_type;
#ifdef USE_VFS
	int vfs_use_limit;
#endif
#define PERMISSIONS          12
	struct CHECK_PERM check_perm[PERMISSIONS];
	struct CHOWN_LABEL chown_label[LABELS_CHOWN];
#define COLOR_MAP_SIZE 28
	struct colorpair color_map[COLOR_MAP_SIZE];

	/* Controls the display of the rotating dash on the verbose mode */
	int nice_rotating_dash;
	/* Set if the split is the same */
	int equal_split;
	/* Set if the command prompt is to be displayed */
	int command_prompt;
	/* Set if the nice and usefull keybar is visible */
	int keybar_visible;
	/* Set if the nice message (hint) bar is visible */
	int message_visible;
	/* Used only in the principal program */
	int cons_saver_pid;
	/* Scrolling is done in pages or line increments */
	int mouse_move_pages_viewer;
	/* wrap mode default */
	int global_wrap_mode;
	int default_magic_flag;
	int default_nroff_flag;

	int dirt_limit;
	int get_event_dirty;
	int easy_patterns;

	struct timeval esctime;
	int lastnodelay;
	/* If 1 - then add per file type hilighting */
	int filetype_mode;
	/* If true, show the mini-info on the panel */
	int show_mini_info;
	/* If true, up/down keys scroll the pane listing by pages */
	int panel_scroll_pages;

	int fmd_xlen;
	int help_lines;             /* Lines in help viewer */
	struct _hotlist_but hotlist_but[10];
	int menubar_visible;
	int w_message_flag;
	int num_history_items_recorded;
	int ok_to_refresh;
	int pause_after_run;
	/* If true show files starting with a dot */
	int show_dot_files;
	/* Reverse flag */
	int reverse_sort_flag;

	int user_hotkey;

	sigset_t Old_Signal_Mask;
	volatile unsigned int Blocked_Depth;
	/* A list of directories to be ignores, separated with ':' */
	char *find_ignore_dirs;

	Dlg_head *find_dlg;      /* The dialog */
	WInput *in_start;        /* Start path */
	WInput *in_name;         /* Pattern to search */
	WInput *in_with;         /* text inside filename */
	WListbox *Find_list;     /* Listbox with the file list */
	int bb_running;          /* nice flag */
	WButton *stop_button;    /* pointer to the stop button */
	WLabel *status_label;    /* Finished, Searching etc. */
	char *find_pattern;      /* Pattern to search */
	char *content_pattern;   /* pattern to search inside files */
	int bb_count;            /* Number of files displayed */
	int bb_matches;          /* Number of matches */
	int is_start;            /* Status of the start/stop toggle button */
	char *old_dir;
	char *if_Dirname;
	char *fc_Dirname;
	/* Set when main loop should be terminated */
	volatile int MC_quit;
	Menu MenuBar [5];
	Menu MenuBarEmpty [5];
	TTY_Termio_Type Old_TTY;
	int TTY_Open;
	int SLtt_Term_Cannot_Scroll;
	int Start_Col, Start_Row;
	int Screen_Cols, Screen_Rows;
	/* only the first 8 bits of this are used.  The highest bit is used
	   to indicate an alternate character set.  This leaves 127 userdefineable
	   color combination. */
	int This_Color;
	int Smg_Inited;
	Terminfo_Type Tbuf;
	int This_Alt_Char;
	/* These variables are used to avoid updating the information unless */
	/* we need it */
	int old_first_panel_size;
	int old_horizontal_split;
	int old_output_lines;
	/* Internal variables */
	int _horizontal_split;
	int _equal_split;
	int _first_panel_size;
	int _output_lines;
	int la_first_width, la_second_width;
	WButton *bleft_widget, *bright_widget;
	/* Various saved termios settings that we control here */
	struct termios boot_mode;
	struct termios new_mode;
	/* Set if we get an interrupt */
	volatile int slinterrupt;
	/* Controls whether we should wait for input in getch */
	int no_slang_delay;
	Dlg_head *conf_dlg;
	WRadio *pause_radio;
	int ctl_x_map_enabled;
	WListbox *l_hotlist;
	WListbox *l_movelist;
	Dlg_head *hotlist_dlg;
	Dlg_head *movelist_dlg;
	WLabel *ho_pname, *pname_group, *movelist_group;
	struct hotlist *current_group;
	uid_t current_user_uid;
	user_in_groups *current_user_gid;
	char *help_data;        /* Pointer to the loaded help_data file */
	int  history_ptr;       /* For the history queue */
	char *last_shown;       /* Last byte shown in a screen */
	int end_of_node;        /* Flag: the last character of the node shown? */
	char *currentpoint, *startpoint;
	char *selected_item;
	Link_Area *link_area;
	int inside_link_area;
	struct HISTORY bb_history [HISTORY_SIZE];
	char *um_entries [MAX_ENTRIES];
	int debug_flag;
	int debug_error;
	char bb_mode_string_buf[12];

	struct link *linklist;  /* the hard link cache */
	struct link *erase_list;        /* the files-to-be-erased list */
	/* In copy_dir_dir we use two additional single linked lists: The first -
	   variable name `parent_dirs' - holds information about already copied
	   directories and is used to detect cyclic symbolic links.
	   The second (`dest_dirs' below) holds information about just created
	   target directories and is used to detect when an directory is copied
	   into itself (we don't want to copy infinitly).
	   Both lists don't use the linkcount and name structure members of struct
	   link. */
	struct link *dest_dirs;
	struct re_pattern_buffer rx;
	struct re_registers regs;
	char *dest_mask;
	int op_follow_links;
	dir_list dir_copy;

	key_def *keys;   /* This holds all the key definitions */
	fd_set select_set;
	int text_start, text_end, min_end;
	int_cache uid_cache [UID_CACHE_SIZE];
	int_cache gid_cache [GID_CACHE_SIZE];
	char **hosts;
	char **hosts_p;
	int hosts_alloclen;
	struct hotlist *g_hotlist;
	/* This flag is used in filename_completion_function */
	int ignore_filenames;
	/* This flag is used by command_completion_function */
	/* to hint the filename_completion_function */
	int look_for_executables;
	/* Pointer to killed data */
	char *kill_buffer;
	WListbox *l_panelize;
	Dlg_head *panelize_dlg;
	WInput *pname;
	int last_listitem;
	Dlg_head *last_query_dlg;
	int sel_pos;
	int la_height;
	int query_height, query_width;
	WInput *query_input;
	char *tkn_buf;
	char *hotlist_file_name;
	FILE *hotlist_file;
	time_t hotlist_file_mtime;
	struct HOTLIST_STATE hotlist_state;
	/* If true, then print on stdout the last directory we were at */
	int print_last_wd;
	char *last_wd_string;
	int print_last_revert;
	/* Used for keeping track of the original stdout */
	int stdout_fd;
	/* if on, it displays the information that files have been moved to ~/.mc */
	int show_change_notice;
	Dlg_head *layout_dlg;
	Dlg_head *learn_dlg;
	learnkey *learnkeys;
	int learn_total;
	int learnok;
	int learnchanged;
	int list_level;
#ifndef NO_INFOMOUNT
	struct mount_entry *mount_list;
#endif
	int mouse_marking;
	struct my_statfs myfs_stats;
	struct panelize *d_panelize;
	struct DIM_PANELS panels [MAX_VIEWS];
	int *pending_keys;
	Refresh *refresh_list;
	char *regex_command_title;
	struct sigaction save_intr, save_quit, save_stop;
	int saving_setup;
	/* The maximum sequence length (32 + null terminator) */
#define SEQ_BUFFER_SZ 33
	int seq_buffer [SEQ_BUFFER_SZ];
	int *seq_append;
	struct sigaction startup_handler;
	int transform_error;
	/* The number of bytes between percent increments */
	int update_steps;
	/* Last point where we updated the status */
	long update_activate;
	/* Only the text mode edition uses this */
	Dlg_head  *view_dlg;
	int query_colors [4];
	struct text_struct query_dialog_pass;
	char timebuf [40];
	int tkn_buf_length;
	int check_pos, button_pos;
	struct timeval tv1;  /* Force first click as single */
	int clicks;
	int isabsolute;
	int phase;
	int text_len;
	const char *const *words;
	char *cur_path;
	char *cur_word;
	int init_state;
	const char *egrep_path;
	char **env_p;
	int varlen, isbrace;
	char **hc_host_p;
	int textstart, textlen;
	char fds_directory [MC_MAXPATHLEN+2];
	DIR *fcf_directory;
	char fntarget [MC_MAXPATHLEN];
	int uid_last;
	int  gid_last;
	char *users_dirname;
	int subdirs_left;
	int filename_len;
	char *in_contents;
	char *in_start_dir;
	char *in_start_name;
	int last_i;
	int last_sec;
	int userlen;
	char *qq;
	char *old_pattern;
	char *m_old_pattern;
	int m_old_type;
	int old_type;
	key_def *k_this, *k_parent;
	int l_b1, l_b2, l_b3;
	int c_b1, c_b2, c_b3;
	char *fill_listbox_buf;
	int  fill_listbox_buflen;
	char string_file_type_buffer[2];
	char string_file_nlinks_buffer[20];
	char string_file_size_buffer[16];
	char string_file_perm_octal_buffer [10];
	char string_file_ngid_buffer[10];
	char string_file_nuid_buffer[10];
	struct dirent *find_do_search_dp;
	DIR  *find_do_search_dirp;
	int find_do_search_pos;
	char *fc_filename;
	char get_group_gbuf[8];
	char get_owner_ibuf[8];
	char *regexp_search_old;
	char *normal_search_old;
	char *ccf_path;
	int rotate_dash_pos;
	regex_t regexp_match_r;
	regex_t rvs_r;
	struct link *cdd_tail;
	char *viewer;
	char nt_x[MC_MAXPATHLEN+MC_MAXPATHLEN];
	char st_x[30];
	char sts_x[60];
	char *load_setup_buffer;
} bb_common_bss;
#define SLtt_Graphics_Char_Pairs bb_common_bss.SLtt_Graphics_Char_Pairs
#define UnderLine_Vid_Str bb_common_bss.UnderLine_Vid_Str
#define Blink_Vid_Str bb_common_bss.Blink_Vid_Str
#define Bold_Vid_Str bb_common_bss.Bold_Vid_Str
#define Eins_Mode_Str bb_common_bss.Eins_Mode_Str
#define Scroll_R_Str bb_common_bss.Scroll_R_Str
#define Cls_Str bb_common_bss.Cls_Str
#define Rev_Vid_Str bb_common_bss.Rev_Vid_Str
#define Norm_Vid_Str bb_common_bss.Norm_Vid_Str
#define Del_Eol_Str bb_common_bss.Del_Eol_Str
#define Del_N_Lines_Str bb_common_bss.Del_N_Lines_Str
#define Add_N_Lines_Str bb_common_bss.Add_N_Lines_Str
#define Rev_Scroll_Str bb_common_bss.Rev_Scroll_Str
#define Curs_Up_Str bb_common_bss.Curs_Up_Str
#define Curs_F_Str bb_common_bss.Curs_F_Str
#define Cursor_Visible_Str bb_common_bss.Cursor_Visible_Str
#define Cursor_Invisible_Str bb_common_bss.Cursor_Invisible_Str
#define Start_Alt_Chars_Str bb_common_bss.Start_Alt_Chars_Str
#define End_Alt_Chars_Str bb_common_bss.End_Alt_Chars_Str
#define Enable_Alt_Char_Set bb_common_bss.Enable_Alt_Char_Set
#define Term_Init_Str bb_common_bss.Term_Init_Str
#define Keypad_Init_Str bb_common_bss.Keypad_Init_Str
#define Term_Reset_Str bb_common_bss.Term_Reset_Str
#define Keypad_Reset_Str bb_common_bss.Keypad_Reset_Str
#define Curs_Pos_Str bb_common_bss.Curs_Pos_Str
#define Current bb_common_bss.Current
#define Base bb_common_bss.Base
#define Blank_Hash bb_common_bss.Blank_Hash
#define Cls_Flag bb_common_bss.Cls_Flag
#define FgBg_Stats bb_common_bss.FgBg_Stats
#define Color_0_Modified bb_common_bss.Color_0_Modified
#define Reset_Color_String bb_common_bss.Reset_Color_String
#define Linux_Console bb_common_bss.Linux_Console
#define Alt_Char_Set bb_common_bss.Alt_Char_Set
#define Automatic_Margins bb_common_bss.Automatic_Margins
#define Worthless_Highlight bb_common_bss.Worthless_Highlight
#define Has_HP_Glitch bb_common_bss.Has_HP_Glitch
#define Cursor_Set bb_common_bss.Cursor_Set
#define Scroll_r1 bb_common_bss.Scroll_r1
#define Cursor_r bb_common_bss.Cursor_r
#define Cursor_c bb_common_bss.Cursor_c
#define last_percentage bb_common_bss.last_percentage
#define replace_colors bb_common_bss.replace_colors
#define replace_dlg bb_common_bss.replace_dlg
#define replace_filename bb_common_bss.replace_filename
#define replace_result bb_common_bss.replace_result
#define s_stat bb_common_bss.s_stat
#define d_stat bb_common_bss.d_stat
#define op_dlg bb_common_bss.op_dlg
#define showing_eta bb_common_bss.showing_eta
#define showing_bps bb_common_bss.showing_bps
#define bps bb_common_bss.bps
#define bps_time bb_common_bss.bps_time
#define FileLabel bb_common_bss.FileLabel
#define FileString bb_common_bss.FileString
#define ProgressLabel bb_common_bss.ProgressLabel
#define ProgressGauge bb_common_bss.ProgressGauge
#define eta_label bb_common_bss.eta_label
#define bps_label bb_common_bss.bps_label
#define stalled_label bb_common_bss.stalled_label
#define debug_out_msg bb_common_bss.debug_out_msg
#define unix_error_string_buffer bb_common_bss.unix_error_string_buffer
#define cmd_buf bb_common_bss.cmd_buf
#define my_gid bb_common_bss.my_gid
#define my_uid bb_common_bss.my_uid
#define my_euid bb_common_bss.my_euid
#define xmouse_flag bb_common_bss.xmouse_flag
#define SLang_Input_Buffer_Len bb_common_bss.SLang_Input_Buffer_Len
#define SLang_Input_Buffer bb_common_bss.SLang_Input_Buffer
#define SLang_TT_Baud_Rate bb_common_bss.SLang_TT_Baud_Rate
#define SLtt_Has_Alt_Charset bb_common_bss.SLtt_Has_Alt_Charset
#define SLtt_Screen_Cols bb_common_bss.SLtt_Screen_Cols
#define SLtt_Screen_Rows bb_common_bss.SLtt_Screen_Rows
#define SLtt_Use_Ansi_Colors bb_common_bss.SLtt_Use_Ansi_Colors
#define midnight_colors bb_common_bss.midnight_colors
#define force_colors bb_common_bss.force_colors
#define command_line_colors bb_common_bss.command_line_colors
#define edit_one_file bb_common_bss.edit_one_file
#define midnight_shutdown bb_common_bss.midnight_shutdown
#define last_paused bb_common_bss.last_paused
#define profile_name bb_common_bss.profile_name
#define setup_color_string bb_common_bss.setup_color_string
#define term_color_string bb_common_bss.term_color_string
#define color_terminal_string bb_common_bss.color_terminal_string
#define confirm_execute bb_common_bss.confirm_execute
#define confirm_view_dir bb_common_bss.confirm_view_dir
#define drop_menus bb_common_bss.drop_menus
#define home_dir bb_common_bss.home_dir
#define other_dir bb_common_bss.other_dir
#define this_dir bb_common_bss.this_dir
#define output_starts_shell bb_common_bss.output_starts_shell
#define TERM_env bb_common_bss.TERM_env
#define permission_mode bb_common_bss.permission_mode
#define hascolors bb_common_bss.hascolors
#define disable_colors bb_common_bss.disable_colors
#define use_colors bb_common_bss.use_colors
#define dialog_colors bb_common_bss.dialog_colors
#define control_file bb_common_bss.control_file
#define MC_shell bb_common_bss.MC_shell
#define mc_home bb_common_bss.mc_home
#define the_menubar bb_common_bss.the_menubar
#define b_att bb_common_bss.b_att
#define b_user bb_common_bss.b_user
#define b_group bb_common_bss.b_group
#define files_on_begin bb_common_bss.files_on_begin
#define flag_pos bb_common_bss.flag_pos
#define x_toggle bb_common_bss.x_toggle
#define ch_flags bb_common_bss.ch_flags
#define ch_cmode bb_common_bss.ch_cmode
#define sf_stat bb_common_bss.sf_stat
#define a_need_update bb_common_bss.a_need_update
#define a_end_chown bb_common_bss.a_end_chown
#define a_current_file bb_common_bss.a_current_file
#define a_single_set bb_common_bss.a_single_set
#define g_fname bb_common_bss.g_fname
#define dd bb_common_bss.dd
#define my_radio bb_common_bss.my_radio
#define status bb_common_bss.status
#define check_status bb_common_bss.check_status
#define current_mode bb_common_bss.current_mode
#define displays_status bb_common_bss.displays_status
#define i18n_displays_flag bb_common_bss.i18n_displays_flag
#define sort_orders_names bb_common_bss.sort_orders_names
#define tree_colors bb_common_bss.tree_colors
#define dash_number bb_common_bss.dash_number
#define ret_timeout bb_common_bss.ret_timeout
#define ret_limit bb_common_bss.ret_limit
#define chm_single_set bb_common_bss.chm_single_set
#define mode_change bb_common_bss.mode_change
#define chm_need_update bb_common_bss.chm_need_update
#define c_file bb_common_bss.c_file
#define end_chmod bb_common_bss.end_chmod
#define and_mask bb_common_bss.and_mask
#define or_mask bb_common_bss.or_mask
#define statl bb_common_bss.statl
#define normal_color bb_common_bss.normal_color
#define title_color bb_common_bss.title_color
#define selection_color bb_common_bss.selection_color
#define cho_need_update bb_common_bss.cho_need_update
#define end_chown bb_common_bss.end_chown
#define cho_current_file bb_common_bss.cho_current_file
#define cho_single_set bb_common_bss.cho_single_set
#define l_user bb_common_bss.l_user
#define l_group bb_common_bss.l_group
#define editor bb_common_bss.editor
#define compare_src_size bb_common_bss.compare_src_size
#define dive_into_subdirs bb_common_bss.dive_into_subdirs
#define preserve bb_common_bss.preserve
#define preserve_uidgid bb_common_bss.preserve_uidgid
#define know_not_what_am_i_doing bb_common_bss.know_not_what_am_i_doing
#define stable_symlinks bb_common_bss.stable_symlinks
#define eta_extra bb_common_bss.eta_extra
#define recursive_result bb_common_bss.recursive_result
#define eta_secs bb_common_bss.eta_secs
#define string_inode_buffer bb_common_bss.string_inode_buffer
#define winch_flag bb_common_bss.winch_flag
#define MC_prompt bb_common_bss.MC_prompt
#define is_right bb_common_bss.is_right
#define pipefd1 bb_common_bss.pipefd1
#define pipefd2 bb_common_bss.pipefd2
#define SL_Screen bb_common_bss.SL_Screen
#define Output_Buffer bb_common_bss.Output_Buffer
#define Output_Bufferp bb_common_bss.Output_Bufferp
#define rxvt_extensions bb_common_bss.rxvt_extensions
#define linux_terminfo_offset bb_common_bss.linux_terminfo_offset
#define SLang_TT_Read_FD bb_common_bss.SLang_TT_Read_FD
#define DISPLAY_X bb_common_bss.DISPLAY_X
#define DISPLAY_Y bb_common_bss.DISPLAY_Y
#define Scroll_r2 bb_common_bss.Scroll_r2
#define FIND2_X bb_common_bss.FIND2_X
#define sort_case_sensitive bb_common_bss.sort_case_sensitive
#define SORT_X bb_common_bss.SORT_X
#define force_xterm bb_common_bss.force_xterm
#define slow_terminal bb_common_bss.slow_terminal
#define fast_reload_w bb_common_bss.fast_reload_w
#define navigate_with_arrows bb_common_bss.navigate_with_arrows
#define force_ugly_line_drawing bb_common_bss.force_ugly_line_drawing
#define reset_hp_softkeys bb_common_bss.reset_hp_softkeys
#define old_esc_mode bb_common_bss.old_esc_mode
#define advanced_chfns bb_common_bss.advanced_chfns
#define output_lines bb_common_bss.output_lines
#define show_all_if_ambiguous bb_common_bss.show_all_if_ambiguous
#define fast_refresh bb_common_bss.fast_refresh
#define auto_menu bb_common_bss.auto_menu
#define alternate_plus_minus bb_common_bss.alternate_plus_minus
#define ext_data bb_common_bss.ext_data
#define fast_reload bb_common_bss.fast_reload
#define torben_fj_mode bb_common_bss.torben_fj_mode
#define ruler bb_common_bss.ruler
#define have_fast_cpu bb_common_bss.have_fast_cpu
#define default_hex_mode bb_common_bss.default_hex_mode
#define altered_hex_mode bb_common_bss.altered_hex_mode
#define altered_magic_flag bb_common_bss.altered_magic_flag
#define altered_nroff_flag bb_common_bss.altered_nroff_flag
#define xterm_flag bb_common_bss.xterm_flag
#define layout_do_change bb_common_bss.layout_do_change
#define horizontal_split bb_common_bss.horizontal_split
#define first_panel_size bb_common_bss.first_panel_size
#define xterm_hintbar bb_common_bss.xterm_hintbar
#define output_start_y bb_common_bss.output_start_y
#define do_append bb_common_bss.do_append
#define do_reget bb_common_bss.do_reget
#define last_hint_line bb_common_bss.last_hint_line
#define tree_navigation_flag bb_common_bss.tree_navigation_flag
#define xtree_mode bb_common_bss.xtree_mode
#define console_flag bb_common_bss.console_flag
#define show_backups bb_common_bss.show_backups
#define mix_all_files bb_common_bss.mix_all_files
#define This_Row bb_common_bss.This_Row
#define This_Col bb_common_bss.This_Col
#define error_pipe bb_common_bss.error_pipe
#define old_error bb_common_bss.old_error
#define left_panel bb_common_bss.left_panel
#define right_panel bb_common_bss.right_panel
#define current_panel bb_common_bss.current_panel
#define select_file_hook bb_common_bss.select_file_hook
#define current_dlg bb_common_bss.current_dlg
#define idle_hook bb_common_bss.idle_hook
#define midnight_dlg bb_common_bss.midnight_dlg
#define quote bb_common_bss.quote
#define the_prompt bb_common_bss.the_prompt
#define the_hint bb_common_bss.the_hint
#define the_bar bb_common_bss.the_bar
#define cmdline bb_common_bss.cmdline
#define ch_dlg bb_common_bss.ch_dlg
#define c_stat bb_common_bss.c_stat
#define c_fname bb_common_bss.c_fname
#define c_fown bb_common_bss.c_fown
#define c_fgrp bb_common_bss.c_fgrp
#define the_tree bb_common_bss.the_tree
#define startup_left_mode bb_common_bss.startup_left_mode
#define startup_right_mode bb_common_bss.startup_right_mode
#define WIuser bb_common_bss.WIuser
#define my_delete bb_common_bss.my_delete
#define my_overwrite bb_common_bss.my_overwrite
#define my_execute bb_common_bss.my_execute
#define my_exit bb_common_bss.my_exit
#define box_new_mode bb_common_bss.box_new_mode
#define new_meta bb_common_bss.new_meta
#define extfs_prefixes bb_common_bss.extfs_prefixes
#define extfs_need_archive bb_common_bss.extfs_need_archive
#define extfs_no bb_common_bss.extfs_no
#define extfs_errloop bb_common_bss.extfs_errloop
#define extfs_notadir bb_common_bss.extfs_notadir
#define Ansi_Color_Map bb_common_bss.Ansi_Color_Map
#define Current_Fgbg bb_common_bss.Current_Fgbg
#define SLtt_Num_Chars_Output bb_common_bss.SLtt_Num_Chars_Output
#define MC_verbose bb_common_bss.MC_verbose
#define OX bb_common_bss.OX
#define opt_first_width bb_common_bss.opt_first_width
#define opt_second_width bb_common_bss.opt_second_width
#define Video_Initialized bb_common_bss.Video_Initialized
#define fbuts bb_common_bss.fbuts
#define dir_stack_base bb_common_bss.dir_stack_base
#define la_check_options bb_common_bss.la_check_options
#define radio_widget bb_common_bss.radio_widget
#define mc_xstat bb_common_bss.mc_xstat
#define o_check_options bb_common_bss.o_check_options
#define op_preserve bb_common_bss.op_preserve
#define umask_kill bb_common_bss.umask_kill
#define SLtt_Blink_Mode bb_common_bss.SLtt_Blink_Mode
#define SLsmg_Display_Eight_Bit bb_common_bss.SLsmg_Display_Eight_Bit
#define use_mouse_p bb_common_bss.use_mouse_p
#define clear_before_exec bb_common_bss.clear_before_exec
#define confirm_delete bb_common_bss.confirm_delete
#define confirm_overwrite bb_common_bss.confirm_overwrite
#define confirm_exit bb_common_bss.confirm_exit
#define use_internal_edit bb_common_bss.use_internal_edit
#define auto_save_setup bb_common_bss.auto_save_setup
#define eight_bit_clean bb_common_bss.eight_bit_clean
#define full_eight_bits bb_common_bss.full_eight_bits
#define use_internal_view bb_common_bss.use_internal_view
#define mouse_move_pages bb_common_bss.mouse_move_pages
#define mou_auto_repeat bb_common_bss.mou_auto_repeat
#define double_click_speed bb_common_bss.double_click_speed
#define use_8th_bit_as_meta bb_common_bss.use_8th_bit_as_meta
#define cd_symlinks bb_common_bss.cd_symlinks
#define mark_moves_down bb_common_bss.mark_moves_down
#define only_leading_plus_minus bb_common_bss.only_leading_plus_minus
#define boot_current_is_left bb_common_bss.boot_current_is_left
#define use_file_to_check_type bb_common_bss.use_file_to_check_type
#define vfs_use_limit bb_common_bss.vfs_use_limit
#define check_perm bb_common_bss.check_perm
#define chown_label bb_common_bss.chown_label
#define color_map bb_common_bss.color_map
#define nice_rotating_dash bb_common_bss.nice_rotating_dash
#define equal_split bb_common_bss.equal_split
#define command_prompt bb_common_bss.command_prompt
#define keybar_visible bb_common_bss.keybar_visible
#define message_visible bb_common_bss.message_visible
#define cons_saver_pid bb_common_bss.cons_saver_pid
#define mouse_move_pages_viewer bb_common_bss.mouse_move_pages_viewer
#define global_wrap_mode bb_common_bss.global_wrap_mode
#define default_magic_flag bb_common_bss.default_magic_flag
#define default_nroff_flag bb_common_bss.default_nroff_flag
#define dirt_limit bb_common_bss.dirt_limit
#define get_event_dirty bb_common_bss.get_event_dirty
#define easy_patterns bb_common_bss.easy_patterns
#define esctime bb_common_bss.esctime
#define lastnodelay bb_common_bss.lastnodelay
#define filetype_mode bb_common_bss.filetype_mode
#define show_mini_info bb_common_bss.show_mini_info
#define panel_scroll_pages bb_common_bss.panel_scroll_pages
#define fmd_xlen bb_common_bss.fmd_xlen
#define help_lines bb_common_bss.help_lines
#define hotlist_but bb_common_bss.hotlist_but
#define menubar_visible bb_common_bss.menubar_visible
#define w_message_flag bb_common_bss.w_message_flag
#define num_history_items_recorded bb_common_bss.num_history_items_recorded
#define ok_to_refresh bb_common_bss.ok_to_refresh
#define pause_after_run bb_common_bss.pause_after_run
#define show_dot_files bb_common_bss.show_dot_files
#define reverse_sort_flag bb_common_bss.reverse_sort_flag
#define user_hotkey bb_common_bss.user_hotkey
#define Old_Signal_Mask bb_common_bss.Old_Signal_Mask
#define Blocked_Depth bb_common_bss.Blocked_Depth
#define find_ignore_dirs bb_common_bss.find_ignore_dirs
#define find_dlg bb_common_bss.find_dlg
#define in_start bb_common_bss.in_start
#define in_name bb_common_bss.in_name
#define in_with bb_common_bss.in_with
#define Find_list bb_common_bss.Find_list
#define bb_running bb_common_bss.bb_running
#define stop_button bb_common_bss.stop_button
#define status_label bb_common_bss.status_label
#define find_pattern bb_common_bss.find_pattern
#define content_pattern bb_common_bss.content_pattern
#define bb_count bb_common_bss.bb_count
#define bb_matches bb_common_bss.bb_matches
#define is_start bb_common_bss.is_start
#define old_dir bb_common_bss.old_dir
#define if_Dirname bb_common_bss.if_Dirname
#define fc_Dirname bb_common_bss.fc_Dirname
#define MC_quit bb_common_bss.MC_quit
#define MenuBar bb_common_bss.MenuBar
#define MenuBarEmpty bb_common_bss.MenuBarEmpty
#define Old_TTY bb_common_bss.Old_TTY
#define TTY_Open bb_common_bss.TTY_Open
#define SLtt_Term_Cannot_Scroll bb_common_bss.SLtt_Term_Cannot_Scroll
#define Start_Col bb_common_bss.Start_Col
#define Start_Row bb_common_bss.Start_Row
#define Screen_Cols bb_common_bss.Screen_Cols
#define Screen_Rows bb_common_bss.Screen_Rows
#define This_Color bb_common_bss.This_Color
#define Smg_Inited bb_common_bss.Smg_Inited
#define Tbuf bb_common_bss.Tbuf
#define This_Alt_Char bb_common_bss.This_Alt_Char
#define old_first_panel_size bb_common_bss.old_first_panel_size
#define old_horizontal_split bb_common_bss.old_horizontal_split
#define old_output_lines bb_common_bss.old_output_lines
#define _horizontal_split bb_common_bss._horizontal_split
#define _equal_split bb_common_bss._equal_split
#define _first_panel_size bb_common_bss._first_panel_size
#define _output_lines bb_common_bss._output_lines
#define la_first_width bb_common_bss.la_first_width
#define la_second_width bb_common_bss.la_second_width
#define bleft_widget bb_common_bss.bleft_widget
#define bright_widget bb_common_bss.bright_widget
#define boot_mode bb_common_bss.boot_mode
#define new_mode bb_common_bss.new_mode
#define slinterrupt bb_common_bss.slinterrupt
#define no_slang_delay bb_common_bss.no_slang_delay
#define conf_dlg bb_common_bss.conf_dlg
#define pause_radio bb_common_bss.pause_radio
#define ctl_x_map_enabled bb_common_bss.ctl_x_map_enabled
#define l_hotlist bb_common_bss.l_hotlist
#define l_movelist bb_common_bss.l_movelist
#define hotlist_dlg bb_common_bss.hotlist_dlg
#define movelist_dlg bb_common_bss.movelist_dlg
#define ho_pname bb_common_bss.ho_pname
#define pname_group bb_common_bss.pname_group
#define movelist_group bb_common_bss.movelist_group
#define current_group bb_common_bss.current_group
#define current_user_uid bb_common_bss.current_user_uid
#define current_user_gid bb_common_bss.current_user_gid
#define help_data bb_common_bss.help_data
#define history_ptr bb_common_bss.history_ptr
#define last_shown bb_common_bss.last_shown
#define end_of_node bb_common_bss.end_of_node
#define currentpoint bb_common_bss.currentpoint
#define startpoint bb_common_bss.startpoint
#define selected_item bb_common_bss.selected_item
#define link_area bb_common_bss.link_area
#define inside_link_area bb_common_bss.inside_link_area
#define bb_history bb_common_bss.bb_history
#define um_entries bb_common_bss.um_entries
#define debug_flag bb_common_bss.debug_flag
#define debug_error bb_common_bss.debug_error
#define bb_mode_string_buf bb_common_bss.bb_mode_string_buf
#define linklist bb_common_bss.linklist
#define erase_list bb_common_bss.erase_list
#define dest_dirs bb_common_bss.dest_dirs
#define rx bb_common_bss.rx
#define regs bb_common_bss.regs
#define dest_mask bb_common_bss.dest_mask
#define op_follow_links bb_common_bss.op_follow_links
#define dir_copy bb_common_bss.dir_copy
#define keys bb_common_bss.keys
#define select_set bb_common_bss.select_set
#define text_start bb_common_bss.text_start
#define text_end bb_common_bss.text_end
#define min_end bb_common_bss.min_end
#define uid_cache bb_common_bss.uid_cache
#define gid_cache bb_common_bss.gid_cache
#define hosts bb_common_bss.hosts
#define hosts_p bb_common_bss.hosts_p
#define hosts_alloclen bb_common_bss.hosts_alloclen
#define g_hotlist bb_common_bss.g_hotlist
#define ignore_filenames bb_common_bss.ignore_filenames
#define look_for_executables bb_common_bss.look_for_executables
#define kill_buffer bb_common_bss.kill_buffer
#define l_panelize bb_common_bss.l_panelize
#define panelize_dlg bb_common_bss.panelize_dlg
#define pname bb_common_bss.pname
#define last_listitem bb_common_bss.last_listitem
#define last_query_dlg bb_common_bss.last_query_dlg
#define sel_pos bb_common_bss.sel_pos
#define la_height bb_common_bss.la_height
#define query_height bb_common_bss.query_height
#define query_width bb_common_bss.query_width
#define query_input bb_common_bss.query_input
#define tkn_buf bb_common_bss.tkn_buf
#define hotlist_file_name bb_common_bss.hotlist_file_name
#define hotlist_file bb_common_bss.hotlist_file
#define hotlist_file_mtime bb_common_bss.hotlist_file_mtime
#define hotlist_state bb_common_bss.hotlist_state
#define print_last_wd bb_common_bss.print_last_wd
#define last_wd_string bb_common_bss.last_wd_string
#define print_last_revert bb_common_bss.print_last_revert
#define stdout_fd bb_common_bss.stdout_fd
#define show_change_notice bb_common_bss.show_change_notice
#define layout_dlg bb_common_bss.layout_dlg
#define learn_dlg bb_common_bss.learn_dlg
#define learnkeys bb_common_bss.learnkeys
#define learn_total bb_common_bss.learn_total
#define learnok bb_common_bss.learnok
#define learnchanged bb_common_bss.learnchanged
#define list_level bb_common_bss.list_level
#define mount_list bb_common_bss.mount_list
#define mouse_marking bb_common_bss.mouse_marking
#define myfs_stats bb_common_bss.myfs_stats
#define d_panelize bb_common_bss.d_panelize
#define panels bb_common_bss.panels
#define pending_keys bb_common_bss.pending_keys
#define refresh_list bb_common_bss.refresh_list
#define regex_command_title bb_common_bss.regex_command_title
#define save_intr bb_common_bss.save_intr
#define save_quit bb_common_bss.save_quit
#define save_stop bb_common_bss.save_stop
#define saving_setup bb_common_bss.saving_setup
#define seq_buffer bb_common_bss.seq_buffer
#define seq_append bb_common_bss.seq_append
#define startup_handler bb_common_bss.startup_handler
#define transform_error bb_common_bss.transform_error
#define update_steps bb_common_bss.update_steps
#define update_activate bb_common_bss.update_activate
#define view_dlg bb_common_bss.view_dlg
#define query_colors bb_common_bss.query_colors
#define query_dialog_pass bb_common_bss.query_dialog_pass
#define timebuf bb_common_bss.timebuf
#define tkn_buf_length bb_common_bss.tkn_buf_length
#define check_pos bb_common_bss.check_pos
#define button_pos bb_common_bss.button_pos
#define tv1 bb_common_bss.tv1
#define clicks bb_common_bss.clicks
#define isabsolute bb_common_bss.isabsolute
#define phase bb_common_bss.phase
#define text_len bb_common_bss.text_len
#define words bb_common_bss.words
#define cur_path bb_common_bss.cur_path
#define cur_word bb_common_bss.cur_word
#define init_state bb_common_bss.init_state
#define egrep_path bb_common_bss.egrep_path
#define env_p bb_common_bss.env_p
#define varlen bb_common_bss.varlen
#define isbrace bb_common_bss.isbrace
#define hc_host_p bb_common_bss.hc_host_p
#define textstart bb_common_bss.textstart
#define textlen bb_common_bss.textlen
#define fds_directory bb_common_bss.fds_directory
#define fcf_directory bb_common_bss.fcf_directory
#define fntarget bb_common_bss.fntarget
#define uid_last bb_common_bss.uid_last
#define gid_last bb_common_bss.gid_last
#define users_dirname bb_common_bss.users_dirname
#define subdirs_left bb_common_bss.subdirs_left
#define filename_len bb_common_bss.filename_len
#define in_contents bb_common_bss.in_contents
#define in_start_dir bb_common_bss.in_start_dir
#define in_start_name bb_common_bss.in_start_name
#define last_i bb_common_bss.last_i
#define last_sec bb_common_bss.last_sec
#define userlen bb_common_bss.userlen
#define qq bb_common_bss.qq
#define old_pattern bb_common_bss.old_pattern
#define m_old_pattern bb_common_bss.m_old_pattern
#define m_old_type bb_common_bss.m_old_type
#define old_type bb_common_bss.old_type
#define k_this bb_common_bss.k_this
#define k_parent bb_common_bss.k_parent
#define l_b1 bb_common_bss.l_b1
#define l_b2 bb_common_bss.l_b2
#define l_b3 bb_common_bss.l_b3
#define c_b1 bb_common_bss.c_b1
#define c_b2 bb_common_bss.c_b2
#define c_b3 bb_common_bss.c_b3
#define fill_listbox_buf bb_common_bss.fill_listbox_buf
#define fill_listbox_buflen bb_common_bss.fill_listbox_buflen
#define string_file_type_buffer bb_common_bss.string_file_type_buffer
#define string_file_nlinks_buffer bb_common_bss.string_file_nlinks_buffer
#define string_file_size_buffer bb_common_bss.string_file_size_buffer
#define string_file_perm_octal_buffer bb_common_bss.string_file_perm_octal_buffer
#define string_file_ngid_buffer bb_common_bss.string_file_ngid_buffer
#define string_file_nuid_buffer bb_common_bss.string_file_nuid_buffer
#define find_do_search_dp bb_common_bss.find_do_search_dp
#define find_do_search_dirp bb_common_bss.find_do_search_dirp
#define find_do_search_pos bb_common_bss.find_do_search_pos
#define fc_filename bb_common_bss.fc_filename
#define get_group_gbuf bb_common_bss.get_group_gbuf
#define get_owner_ibuf bb_common_bss.get_owner_ibuf
#define regexp_search_old bb_common_bss.regexp_search_old
#define normal_search_old bb_common_bss.normal_search_old
#define ccf_path bb_common_bss.ccf_path
#define rotate_dash_pos bb_common_bss.rotate_dash_pos
#define regexp_match_r bb_common_bss.regexp_match_r
#define rvs_r bb_common_bss.rvs_r
#define cdd_tail bb_common_bss.cdd_tail
#define viewer bb_common_bss.viewer
#define nt_x bb_common_bss.nt_x
#define st_x bb_common_bss.st_x
#define sts_x bb_common_bss.sts_x
#define load_setup_buffer bb_common_bss.load_setup_buffer


/*
	Chown-advanced command -- for the Midnight Commander
*/

#define BX_ACHOWN              5
#define BY_ACHOWN              6

#define TX_ACHOWN              50
#define TY_ACHOWN              2

#define BUTTONS_ACHOWN         9

#define B_SETALL        B_USER
#define B_SKIP          B_USER + 1

#define B_OWN           B_USER + 3
#define B_GRP           B_USER + 4
#define B_OTH           B_USER + 5
#define B_OUSER         B_USER + 6
#define B_OGROUP        B_USER + 7

static const struct {
    int ret_cmd, flags, y, x;
    const char *text;
} chown_advanced_but [BUTTONS_ACHOWN] = {
    { B_CANCEL, NORMAL_BUTTON, 4, 55, "&Cancel" },
    { B_ENTER,  DEFPUSH_BUTTON,4, 45, "&Set" },
    { B_SKIP,   NORMAL_BUTTON, 4, 36, "S&kip" },
    { B_SETALL, NORMAL_BUTTON, 4, 24, "Set &all"},
    { B_ENTER,  NARROW_BUTTON, 0, 47, "               "},
    { B_ENTER,  NARROW_BUTTON, 0, 29, "               "},
    { B_ENTER,  NARROW_BUTTON, 0, 19, "   "},
    { B_ENTER,  NARROW_BUTTON, 0, 11, "   "},
    { B_ENTER,  NARROW_BUTTON, 0, 3, "   "},
};

static const char ch_perm[] = "rwx";

static void get_ownership (void)
{                               /* set buttons  - ownership */
    char *name_t;

    name_t = name_trunc (get_owner (sf_stat->st_uid), 15);
    memset (b_user->text, ' ', 15);
    strncpy (b_user->text, name_t, strlen (name_t));
    name_t = name_trunc (get_group (sf_stat->st_gid), 15);
    memset (b_group->text, ' ', 15);
    strncpy (b_group->text, name_t, strlen (name_t));
}


static int inc_flag_pos (int f_pos)
{
    if (flag_pos == 10) {
	flag_pos = 0;
	return 0;
    }
    flag_pos++;
    if (!(flag_pos % 3) || f_pos > 2)
	return 0;
    return 1;
}

static int dec_flag_pos (int f_pos)
{
    if (!flag_pos) {
	flag_pos = 10;
	return 0;
    }
    flag_pos--;
    if (!((flag_pos + 1) % 3) || f_pos > 2)
	return 0;
    return 1;
}

static void set_perm_by_flags (char *s, int f_p)
{
    int i;

    for (i = 0; i < 3; i++)
	if (ch_flags[f_p + i] == '+')
	    s[i] = ch_perm[i];
	else if (ch_flags[f_p + i] == '-')
	    s[i] = '-';
	else
	    s[i] = (ch_cmode & (1 << (8 - f_p - i))) ? ch_perm[i] : '-';
}

static void set_perm (char *s, int p)
{
    s[0] = (p & 4) ? 'r' : '-';
    s[1] = (p & 2) ? 'w' : '-';
    s[2] = (p & 1) ? 'x' : '-';
}

static umode_t get_perm (const char *s, int base)
{
    umode_t m;

    m = 0;
    m |= (s [0] == '-') ? 0 :
	((s[0] == '+') ? (1 << (base + 2)) : (1 << (base + 2)) & ch_cmode);

    m |= (s [1] == '-') ? 0 :
	((s[1] == '+') ? (1 << (base + 1)) : (1 << (base + 1)) & ch_cmode);

    m |= (s [2] == '-') ? 0 :
	((s[2] == '+') ? (1 << base) : (1 << base) & ch_cmode);

    return m;
}

static umode_t get_mode (void)
{
    umode_t m;

    m = ch_cmode ^ (ch_cmode & 0777);
    m |= get_perm (ch_flags, 6);
    m |= get_perm (ch_flags + 3, 3);
    m |= get_perm (ch_flags + 6, 0);

    return m;
}

static void print_flags (void)
{
    int i;

    attrset (COLOR_NORMAL);

    for (i = 0; i < 3; i++) {
	dlg_move (ch_dlg, BY_ACHOWN+1, 9+i);
	addch (ch_flags [i]);
    }

    for (i = 0; i < 3; i++) {
	dlg_move (ch_dlg, BY_ACHOWN + 1, 17 + i);
	addch (ch_flags [i+3]);
    }

    for (i = 0; i < 3; i++) {
	dlg_move (ch_dlg, BY_ACHOWN + 1, 25 + i);
	addch (ch_flags [i+6]);
    }

    set_perm_by_flags (b_att[0]->text, 0);
    set_perm_by_flags (b_att[1]->text, 3);
    set_perm_by_flags (b_att[2]->text, 6);

    for (i = 0; i < 15; i++) {
	dlg_move (ch_dlg, BY_ACHOWN+1, 35+i);
	addch (ch_flags[9]);
    }
    for (i = 0; i < 15; i++) {
	dlg_move (ch_dlg, BY_ACHOWN + 1, 53 + i);
	addch (ch_flags[10]);
    }
}

static void update_mode (Dlg_head * h)
{
    print_flags ();
    attrset (COLOR_NORMAL);
    dlg_move (h, BY_ACHOWN + 2, 9);
    printw ("%12o", get_mode ());
    send_message (h, h->current->widget, WIDGET_FOCUS, 0);
}

static int l_call_1 (void *data)
{
    return 1;
}

static int chl_callback (Dlg_head * h, int Par, int Msg)
{
    switch (Msg) {
    case DLG_DRAW:
	attrset (COLOR_NORMAL);
	dlg_erase (h);
	draw_box (h, 0, 0, 13, 17);
	break;

    case DLG_KEY:
	switch (Par) {
	case KEY_LEFT:
	case KEY_RIGHT:
	    h->ret_value = Par;
	    dlg_stop (h);
	}
    }
    return 0;
}

static void do_enter_key (Dlg_head *h, int f_pos)
{
    Dlg_head *chl_dlg;
    WListbox *chl_list;
    struct   passwd *chl_pass;
    struct   group *chl_grp;
    WLEntry  *fe;
    int      lxx, lyy, chl_end, b_pos;

    do {
	lxx = (COLS - 74) / 2 + ((f_pos == 3) ? 35 : 53);
	lyy = (LINES - 13) / 2;
	chl_end = 0;

	chl_dlg = create_dlg (lyy, lxx, 13, 17, dialog_colors, chl_callback,
			      "[Chown-advanced]", DLG_NONE);

	/* get new listboxes */
	chl_list = listbox_new (1, 1, 15, 11, 0, l_call_1);

	listbox_add_item (chl_list, 0, 0, "<Unknown>", NULL);

	if (f_pos == 3) {
	    /* get and put user names in the listbox */
	    setpwent ();
	    while ((chl_pass = getpwent ()))
		listbox_add_item (chl_list, 0, 0, chl_pass->pw_name, NULL);
	    endpwent ();
	    fe = listbox_search_text (chl_list, get_owner (sf_stat->st_uid));
	}
	else
	{
	    /* get and put group names in the listbox */
	    setgrent ();
	    while ((chl_grp = getgrent ())) {
		listbox_add_item (chl_list, 0, 0, chl_grp->gr_name, NULL);
	    }
	    endgrent ();
	    fe = listbox_search_text (chl_list, get_group (sf_stat->st_gid));
	}

	if (fe)
	    listbox_select_entry (chl_list, fe);

	b_pos = chl_list->pos;
	add_widget (chl_dlg, chl_list);

	run_dlg (chl_dlg);

	if (b_pos != chl_list->pos) {
	    int ok = 0;
	    if (f_pos == 3) {
		chl_pass = getpwnam (chl_list->current->text);
		if (chl_pass) {
		    ok = 1;
		    sf_stat->st_uid = chl_pass->pw_uid;
		}
	    } else {
		chl_grp = getgrnam (chl_list->current->text);
		if (chl_grp) {
		    sf_stat->st_gid = chl_grp->gr_gid;
		    ok = 1;
		}
	    }
	    if (ok) {
		ch_flags [f_pos + 6] = '+';
		get_ownership ();
	    }
	    dlg_focus (h);
	    if (ok)
		print_flags ();
	}
	if (chl_dlg->ret_value == KEY_LEFT) {
	    if (f_pos == 4)
		chl_end = 1;
	    dlg_one_up (ch_dlg);
	    f_pos--;
	} else if (chl_dlg->ret_value == KEY_RIGHT) {
	    if (f_pos == 3)
		chl_end = 1;
	    dlg_one_down (ch_dlg);
	    f_pos++;
	}
	/* Here we used to redraw the window */
	destroy_dlg (chl_dlg);
    } while (chl_end);
}

static void a_chown_refresh (void)
{
    attrset (COLOR_NORMAL);
    dlg_erase (ch_dlg);

    draw_box (ch_dlg, 1, 2, 11, 70);

    dlg_move (ch_dlg, BY_ACHOWN - 1, 8);
    addstr ("owner");
    dlg_move (ch_dlg, BY_ACHOWN - 1, 16);
    addstr ("group");
    dlg_move (ch_dlg, BY_ACHOWN - 1, 24);
    addstr ("other");

    dlg_move (ch_dlg, BY_ACHOWN - 1, 35);
    addstr ("owner");
    dlg_move (ch_dlg, BY_ACHOWN - 1, 53);
    addstr ("group");

    dlg_move (ch_dlg, 3, 4);
    addstr ("On");
    dlg_move (ch_dlg, BY_ACHOWN + 1, 4);
    addstr ("Flag");
    dlg_move (ch_dlg, BY_ACHOWN + 2, 4);
    addstr ("Mode");


    if (!a_single_set) {
	dlg_move (ch_dlg, 3, 54);
	printw ("%6d of %d", files_on_begin - (cpanel->marked) + 1,
		   files_on_begin);
    }

    print_flags ();

    attrset (COLOR_HOT_NORMAL);
    dlg_move (ch_dlg, 1, 24);
    addstr (" Chown advanced command ");
}

static void chown_info_update (void)
{
    /* display file info */
    attrset (COLOR_NORMAL);

    /* name && mode */
    dlg_move (ch_dlg, 3, 8);
    printw ("%s", name_trunc (g_fname, 45));
    dlg_move (ch_dlg, BY_ACHOWN + 2, 9);
    printw ("%12o", get_mode ());

    /* permissions */
    set_perm (b_att[0]->text, sf_stat->st_mode >> 6);
    set_perm (b_att[1]->text, sf_stat->st_mode >> 3);
    set_perm (b_att[2]->text, sf_stat->st_mode);
}

static void b_setpos (int f_pos) {
	b_att[0]->hotpos=-1;
	b_att[1]->hotpos=-1;
	b_att[2]->hotpos=-1;
	b_att[f_pos]->hotpos = (flag_pos % 3);
}

static int advanced_chown_callback (Dlg_head * h, int Par, int Msg)
{
    int i = 0, f_pos = BUTTONS_ACHOWN - h->current->dlg_id - a_single_set - 1;

    switch (Msg) {
    case DLG_DRAW:
	a_chown_refresh ();
	chown_info_update ();
	return 1;

    case DLG_POST_KEY:
	if (f_pos < 3)
		b_setpos (f_pos);
	break;

    case DLG_FOCUS:
	if (f_pos < 3) {
	    if ((flag_pos / 3) != f_pos)
		flag_pos = f_pos * 3;
	    b_setpos (f_pos);
	} else if (f_pos < 5)
	    flag_pos = f_pos + 6;
	break;

    case DLG_KEY:
	switch (Par) {

	case XCTRL('b'):
	case KEY_LEFT:
	    if (f_pos < 5)
		return (dec_flag_pos (f_pos));
	    break;

	case XCTRL('f'):
	case KEY_RIGHT:
	    if (f_pos < 5)
		return (inc_flag_pos (f_pos));
	    break;

	case ' ':
	    if (f_pos < 3)
		return 1;
	    break;

	case '\n':
	case KEY_ENTER:
	    if (f_pos <= 2 || f_pos >= 5)
		break;
	    do_enter_key (h, f_pos);
	    return 1;

	case ALT ('x'):
	    i++;

	case ALT ('w'):
	    i++;

	case ALT ('r'):
	    Par = i + 3;
	    for (i = 0; i < 3; i++)
		ch_flags[i * 3 + Par - 3] = (x_toggle & (1 << Par)) ? '-' : '+';
	    x_toggle ^= (1 << Par);
	    update_mode (h);
	    dlg_broadcast_msg (h, WIDGET_DRAW, 0);
	    send_message (h, h->current->widget, WIDGET_FOCUS, 0);
	    break;

	case XCTRL ('x'):
	    i++;

	case XCTRL ('w'):
	    i++;

	case XCTRL ('r'):
	    Par = i;
	    for (i = 0; i < 3; i++)
		ch_flags[i * 3 + Par] = (x_toggle & (1 << Par)) ? '-' : '+';
	    x_toggle ^= (1 << Par);
	    update_mode (h);
	    dlg_broadcast_msg (h, WIDGET_DRAW, 0);
	    send_message (h, h->current->widget, WIDGET_FOCUS, 0);
	    break;

	case 'x':
	    i++;

	case 'w':
	    i++;

	case 'r':
	    if (f_pos > 2)
		break;
	    flag_pos = f_pos * 3 + i;   /* (strchr(ch_perm,Par)-ch_perm); */
	    if (((WButton *) h->current->widget)->text[(flag_pos % 3)] == '-')
		ch_flags[flag_pos] = '+';
	    else
		ch_flags[flag_pos] = '-';
	    update_mode (h);
	    break;

	case '4':
	    i++;

	case '2':
	    i++;

	case '1':
	    if (f_pos > 2)
		break;
	    flag_pos = i + f_pos * 3;
	    ch_flags[flag_pos] = '=';
	    update_mode (h);
	    break;

	case '-':
	    if (f_pos > 2)
		break;

	case '*':
	    if (Par == '*')
		Par = '=';

	case '=':
	case '+':
	    if (f_pos > 4)
		break;
	    ch_flags[flag_pos] = Par;
	    update_mode (h);
	    advanced_chown_callback (h, KEY_RIGHT, DLG_KEY);
	    if (flag_pos>8 || !(flag_pos%3)) dlg_one_down (h);

	    break;
	}
	return 0;
    }
    return 0;
}

static void init_chown_advanced (void)
{
    int i;

    sf_stat = (struct stat *) xmalloc (sizeof (struct stat));
    do_refresh ();
    a_end_chown = a_need_update = a_current_file = 0;
    a_single_set = (cpanel->marked < 2) ? 2 : 0;
    memset (ch_flags, '=', 11);
    flag_pos = 0;
    x_toggle = 070;

    ch_dlg = create_dlg (0, 0, 13, 74, dialog_colors, advanced_chown_callback,
			 "[Chown-advanced]", DLG_CENTER);

#define XTRACT_ACHOWN(i) BY_ACHOWN+chown_advanced_but[i].y, BX_ACHOWN+chown_advanced_but[i].x, \
	chown_advanced_but[i].ret_cmd, chown_advanced_but[i].flags, chown_advanced_but[i].text, \
	0, 0

    for (i = 0; i < BUTTONS_ACHOWN - 5; i++)
	if (!a_single_set || i < 2)
	    add_widget (ch_dlg, button_new (XTRACT_ACHOWN (i)));

    b_att[0] = button_new (XTRACT_ACHOWN (8));
    b_att[1] = button_new (XTRACT_ACHOWN (7));
    b_att[2] = button_new (XTRACT_ACHOWN (6));
    b_user = button_new (XTRACT_ACHOWN (5));
    b_group = button_new (XTRACT_ACHOWN (4));

    add_widget (ch_dlg, b_group);
    add_widget (ch_dlg, b_user);
    add_widget (ch_dlg, b_att[2]);
    add_widget (ch_dlg, b_att[1]);
    add_widget (ch_dlg, b_att[0]);
}

static void chown_advanced_done (void)
{
    free (sf_stat);
    if (a_need_update)
	update_panels (UP_OPTIMIZE, UP_KEEPSEL);
    repaint_screen ();
}

static char *a_next_file (void)
{
    while (!cpanel->dir.list[a_current_file].f.marked)
	a_current_file++;

    return cpanel->dir.list[a_current_file].fname;
}

static void apply_advanced_chowns (struct stat *sf)
{
    char *Fname;
    gid_t a_gid = sf->st_gid;
    uid_t a_uid = sf->st_uid;

    Fname = cpanel->dir.list[a_current_file].fname;
    a_need_update = a_end_chown = 1;
    if (mc_chmod (Fname, get_mode ()) == -1)
	MC_message (1, MSG_ERROR, " Couldn't chmod \"%s\" \n %m ",
		 Fname);
    /* call mc_chown only, if mc_chmod didn't fail */
    else if (mc_chown (Fname, (ch_flags[9] == '+') ? sf->st_uid : -1,
		       (ch_flags[10] == '+') ? sf->st_gid : -1) == -1)
	MC_message (1, MSG_ERROR, " Couldn't chown \"%s\" \n %m ",
		 Fname);
    do_file_mark (cpanel, a_current_file, 0);

    do {
	Fname = a_next_file ();

	if (!stat_file (Fname, sf))
	    break;
	ch_cmode = sf->st_mode;
	if (mc_chmod (Fname, get_mode ()) == -1)
	    MC_message (1, MSG_ERROR, " Couldn't chmod \"%s\" \n %m ",
		     Fname);
	/* call mc_chown only, if mc_chmod didn't fail */
	else if (mc_chown (Fname, (ch_flags[9] == '+') ? a_uid : -1, (ch_flags[10] == '+') ? a_gid : -1) == -1)
	    MC_message (1, MSG_ERROR, " Couldn't chown \"%s\" \n %m ",
		     Fname);

	do_file_mark (cpanel, a_current_file, 0);
    } while (cpanel->marked);
}

static void chown_advanced_cmd (void)
{

    files_on_begin = cpanel->marked;

    if (!vfs_current_is_local ()) {
	if (vfs_current_is_extfs ()) {
	    MC_message (1, " Oops... ",
		     " I can't run the Advanced Chown command on an extfs ");
	    return;
	} else if (vfs_current_is_tarfs ()) {
	    MC_message (1, " Oops... ",
		     " I can't run the Advanced Chown command on a tarfs ");
	    return;
	}
    }

    do {                        /* do while any files remaining */
	init_chown_advanced ();

	if (cpanel->marked)
	    g_fname = a_next_file ();       /* next marked file */
	else
	    g_fname = selection (cpanel)->fname;  /* single file */

	if (!stat_file (g_fname, sf_stat)) {       /* get status of file */
	    destroy_dlg (ch_dlg);
	    break;
	}
	ch_cmode = sf_stat->st_mode;

	a_chown_refresh ();

	get_ownership ();

	/* game can begin */
	run_dlg (ch_dlg);

	switch (ch_dlg->ret_value) {
	case B_CANCEL:
	    a_end_chown = 1;
	    break;

	case B_ENTER:
	    a_need_update = 1;
	    if (mc_chmod (g_fname, get_mode ()) == -1)
		MC_message (1, MSG_ERROR, " Couldn't chmod \"%s\" \n %m ",
			 g_fname);
	    /* call mc_chown only, if mc_chmod didn't fail */
	    else if (mc_chown (g_fname, (ch_flags[9] == '+') ? sf_stat->st_uid : -1, (ch_flags[10] == '+') ? sf_stat->st_gid : -1) == -1)
		MC_message (1, MSG_ERROR, " Couldn't chown \"%s\" \n %m ",
			 g_fname);
	    break;
	case B_SETALL:
	    apply_advanced_chowns (sf_stat);
	    break;

	case B_SKIP:
	    break;

	}

	if (cpanel->marked && ch_dlg->ret_value != B_CANCEL) {
	    do_file_mark (cpanel, a_current_file, 0);
	    a_need_update = 1;
	}
	destroy_dlg (ch_dlg);
    } while (cpanel->marked && !a_end_chown);

    chown_advanced_done ();
}
/* Background support. */

/*
 * We currenlty only support one way of comunicating the background
 * and foreground process by using the socketpair system call
 */


/* {{{ Stubs if background code is not supported */
void
message_1s (int flags, char *title, char *str1)
{
    MC_message (flags, title, str1);
}

void
message_2s (int flags, char *title, char *str1, char *str2)
{
    MC_message (flags, title, str1, str2);
}

void
message_3s (int flags, char *title, char *str1, char *str2, const char *str3)
{
    MC_message (flags, title, str1, str2, str3);
}

static inline char *
input_dialog_help (char *header, char *text, char *help, char *def_text)
{
    return real_input_dialog_help (header, text, help, def_text);
}
/* }}} */

static const char display_title[] = " Listing mode ";

static const char * const displays [LIST_TYPES] = {
    "&Full file list",
    "&Brief file list",
    "&Long file list",
    "&User defined:",
    "&Icon view",
};

static int display_callback (struct Dlg_head *h, int id, int Msg)
{
    switch (Msg) {
    case DLG_DRAW:
	attrset (COLOR_NORMAL);
	dlg_erase (h);
	draw_box (h, 1, 2, h->lines - 2, h->cols - 4);

	attrset (COLOR_HOT_NORMAL);
	dlg_move (h, 1, (h->cols - strlen(display_title))/2);
	addstr (display_title);
	attrset (COLOR_NORMAL);
	break;

    case DLG_UNFOCUS:
	if((WRadio *) h->current->widget == my_radio) {
	    assign_text (status, displays_status [my_radio->sel]);
	    input_set_point (status, 0);
	}
	break;

    case DLG_KEY:
	if (id == '\n') {
	    if((WRadio *) h->current->widget == my_radio) {
		assign_text (status, displays_status [my_radio->sel]);
		dlg_stop (h);
		break;
	    }

	    if ((WInput *) h->current->widget == WIuser) {
		h->ret_value = B_USER + 6;
		dlg_stop (h);
		break;
	    }

	    if ((WInput *) h->current->widget == status) {
		h->ret_value = B_USER + 7;
		dlg_stop (h);
		break;
	    }
	}

	if (tolower(id) == user_hotkey && h->current->widget != (Widget *) WIuser
	    && h->current->widget != (Widget *) status) {
	    my_radio->sel = 3;
	    dlg_select_widget (h, my_radio); /* force redraw */
	    dlg_select_widget (h, WIuser);
	    return MSG_HANDLED;
	}
    }
    return MSG_NOT_HANDLED;
}

static const char ok_button[] = "&Ok";
static const char cancel_button[] = "&Cancel";

static void display_init (int radio_sel, char *init_text,
			  int _check_status, char ** _status)
{
    static const char user_mini_status[] = "user &Mini status";
    int button_start = i18n_displays_flag;

    displays_status = _status;

    if (!button_start) {
		int i, l, maxlen = 0;
		char* cp;

		for (i = 0; i < LIST_TYPES; i++) {
			if ((l = strlen(displays [i])) > maxlen)
				maxlen = l;
		}

		i = strlen (ok_button) + 5;
		l = strlen (cancel_button) + 3;
		l = max(i, l);

		i = maxlen + l + 16;
		if (i > DISPLAY_X)
			DISPLAY_X = i;

		i = strlen (user_mini_status) + 13;
		if (i > DISPLAY_X)
			DISPLAY_X = i;

		i = strlen (display_title) + 8;
		if (i > DISPLAY_X)
			DISPLAY_X = i;

		button_start = DISPLAY_X - l - 5;

		/* get hotkey of user-defined format string */
		cp = strchr(displays[LIST_TYPES-1],'&');
		if (cp != NULL && *++cp != '\0')
			user_hotkey = tolower(*cp);

	i18n_displays_flag = button_start;
    }
    dd = create_dlg (0, 0, DISPLAY_Y, DISPLAY_X, dialog_colors,
		     display_callback, "[Left and Right Menus]", DLG_CENTER | DLG_GRID);

    x_set_dialog_title (dd, "Listing mode");
    add_widgetl (dd, button_new (4, button_start, B_CANCEL, NORMAL_BUTTON, cancel_button, 0, 0), XV_WLAY_RIGHTOF);

    add_widgetl (dd, button_new (3, button_start, B_ENTER, DEFPUSH_BUTTON, ok_button, 0, 0), XV_WLAY_CENTERROW);

    status = input_new (10, 9, INPUT_COLOR, DISPLAY_X-14, _status [radio_sel]);
    add_widgetl (dd, status, XV_WLAY_RIGHTDOWN);
    input_set_point (status, 0);

    check_status = check_new (9, 5, _check_status, user_mini_status);
    add_widgetl (dd, check_status, XV_WLAY_NEXTROW);

    WIuser = input_new  (7, 9, INPUT_COLOR, DISPLAY_X-14, init_text);
    add_widgetl (dd, WIuser, XV_WLAY_RIGHTDOWN);
    input_set_point (WIuser, 0);

    my_radio = radio_new (3, 5, LIST_TYPES-1, displays, 1);
    my_radio->sel = my_radio->pos = current_mode;
    add_widgetl (dd, my_radio, XV_WLAY_BELOWCLOSE);
}

static int
display_box (WPanel *panel, char **userp, char **minip, int *use_msformat, int num)
{
    int result, i;
    char *section = NULL;
    char *p;

    if (!panel) {
	char buffer[20];

	p = get_nth_panel_name (num, buffer);
	panel = x_malloc (sizeof (WPanel), "temporary panel");
	panel->list_type = list_full;
	panel->user_format = strdup (DEFAULT_USER_FORMAT);
	panel->user_mini_status = 0;
	for (i = 0; i < LIST_TYPES; i++)
	    panel->user_status_format[i] = strdup (DEFAULT_USER_FORMAT);
	section = xasprintf ("Temporal:%s", p);
	if (!profile_has_section (section, profile_name)) {
	    free (section);
	    section = strdup (p);
	}
	panel_load_setup (panel, section);
	free (section);
    }

    current_mode = panel->list_type;
    display_init (current_mode, panel->user_format,
	panel->user_mini_status, panel->user_status_format);

    run_dlg (dd);

    result = -1;

    if (section) {
	free (panel->user_format);
	for (i = 0; i < LIST_TYPES; i++)
	    free (panel->user_status_format [i]);
	free (panel);
    }

    if (dd->ret_value != B_CANCEL) {
	result = my_radio->sel;
	*userp = strdup (WIuser->buffer);
	*minip = strdup (status->buffer);
	*use_msformat = check_status->state & C_BOOL;
    }
    destroy_dlg (dd);

    return result;
}

#define SORT_Y 14

static sortfn *sort_box (sortfn *sort_fn, int *revers, int *case_sensitiv)
{
    int i, r, l;
    sortfn *result;
    WCheck *c, *case_sense;

	static const char reverse_label[] = "&Reverse";
	static const char case_label[] = "case sensi&tive";
	static const char sort_title[] = "Sort order";

	if (!check_pos) {
		int maxlen = 0;
		for (i = SORT_TYPES-1; i >= 0; i--) {
			sort_orders_names[i] = sort_orders[i].sort_name;
			r = strlen (sort_orders_names [i]);
			if (r > maxlen)
				maxlen = r;
		}

		check_pos = maxlen + 9;

		r = strlen (reverse_label) + 4;
		i = strlen (case_label) + 4;
		if (i > r)
			r = i;

		l = strlen (ok_button) + 6;
		i = strlen (cancel_button) + 4;
		if (i > l)
			l = i;

		i = check_pos + max(r,l) + 2;
		SORT_X = i > 40 ? i : 40;

		i = strlen (sort_title) + 6;
		if (i > SORT_X)
			SORT_X = i;

		button_pos = SORT_X - l - 2;
    }

    result = 0;

    for (i = 0; i < SORT_TYPES; i++)
	if (sort_orders [i].sort_fn == sort_fn) {
	    current_mode = i;
	    break;
	}

    dd = create_dlg (0, 0, SORT_Y, SORT_X, dialog_colors, common_dialog_callback,
		     "[Left and Right Menus]", DLG_CENTER | DLG_GRID);

    x_set_dialog_title (dd, sort_title);

    add_widgetl (dd,
		button_new (10, button_pos, B_CANCEL, NORMAL_BUTTON, cancel_button,
		0, 0), XV_WLAY_CENTERROW);

    add_widgetl (dd,
		button_new (9, button_pos, B_ENTER, DEFPUSH_BUTTON, ok_button,
		0, 0), XV_WLAY_RIGHTDOWN);

    case_sense = check_new (4, check_pos, *case_sensitiv, case_label);
    add_widgetl (dd, case_sense, XV_WLAY_RIGHTDOWN);
    c = check_new (3, check_pos, *revers, reverse_label);
    add_widgetl (dd, c, XV_WLAY_RIGHTDOWN);

    my_radio = radio_new (3, 3, SORT_TYPES, sort_orders_names, 1);
    my_radio->sel = my_radio->pos = current_mode;

    add_widget (dd, my_radio);
    run_dlg (dd);

    r = dd->ret_value;
    if (r != B_CANCEL) {
	result = sort_orders [my_radio->sel].sort_fn;
	*revers = c->state & C_BOOL;
	*case_sensitiv = case_sense->state & C_BOOL;
    } else
	result = sort_fn;
    destroy_dlg (dd);

    return result;
}

#define CONFY 10
#define CONFX 46

static const QuickWidget conf_widgets_st [] = {
  { quick_button,   4, 6, 4, CONFY, "&Cancel",
      0, B_CANCEL, 0, 0, XV_WLAY_RIGHTOF, "c" },
  { quick_button,   4, 6, 3, CONFY, "&Ok",
      0, B_ENTER, 0, 0, XV_WLAY_CENTERROW, "o" },

  { quick_checkbox, 1, 13, 6, CONFY, " confirm &Exit ",
      9, 0, &my_exit, 0, XV_WLAY_BELOWCLOSE, "e" },
  { quick_checkbox, 1, 13, 5, CONFY, " confirm e&Xecute ",
      10, 0, &my_execute, 0, XV_WLAY_BELOWCLOSE, "x" },
  { quick_checkbox, 1, 13, 4, CONFY, " confirm o&Verwrite ",
      10, 0, &my_overwrite, 0, XV_WLAY_BELOWCLOSE, "ov" },
  { quick_checkbox, 1, 13, 3, CONFY, " confirm &Delete ",
      9, 0, &my_delete, 0, XV_WLAY_BELOWCLOSE, "de" },
  { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, XV_WLAY_DONTCARE }
};

static void confirm_box (void)
{
    QuickDialog confirmation;

    confirmation.xlen = CONFX;
    confirmation.ylen = CONFY;
    confirmation.xpos = -1;
    confirmation.ypos = -1;
    confirmation.title = " Confirmation ";
    confirmation.help = "[Confirmation]";
    confirmation.widgets = alloca(sizeof(conf_widgets_st));
    memcpy(confirmation.widgets, conf_widgets_st, sizeof(conf_widgets_st));
    my_delete    = confirm_delete;
    my_overwrite = confirm_overwrite;
    my_execute   = confirm_execute;
    my_exit      = confirm_exit;

    if (quick_dialog (&confirmation) != B_CANCEL) {
	confirm_delete    = my_delete;
	confirm_overwrite = my_overwrite;
	confirm_execute   = my_execute;
	confirm_exit      = my_exit;
    }
}

#define DISPY 11
#define DISPX 46

static const char * const display_bits_str [] =
{ "Full 8 bits output", "ISO 8859-1", "7 bits" };

static const QuickWidget display_widgets_st [] = {
  { quick_button,   4,  6,    4, DISPY, "&Cancel",
      0, B_CANCEL, 0, 0, XV_WLAY_CENTERROW, "c" },
  { quick_button,   4,  6,    3, DISPY, "&Ok",
      0, B_ENTER, 0, 0, XV_WLAY_CENTERROW, "o" },
  { quick_checkbox, 4, DISPX, 7, DISPY, "F&ull 8 bits input",
      0, 0, &new_meta, 0, XV_WLAY_BELOWCLOSE, "u" },
  { quick_radio,    4, DISPX, 3, DISPY, "", 3, 0,
      &box_new_mode, (char **)display_bits_str, XV_WLAY_BELOWCLOSE, "r" },
  { 0, 0, 0, 0, 0, 0,  0, 0, 0, 0, XV_WLAY_DONTCARE }
};

static void display_bits_box (void)
{
     QuickDialog display_bits;

     display_bits.xlen = DISPX;
     display_bits.ylen = DISPY;
     display_bits.xpos = -1;
     display_bits.ypos = -1;
     display_bits.title = " Display bits ";
     display_bits.help = "[Display bits]";
     display_bits.widgets = alloca(sizeof(display_widgets_st));
     memcpy(display_bits.widgets, display_widgets_st, sizeof(display_widgets_st));

    /* display_widgets [3].value = 0; */

    new_meta = !use_8th_bit_as_meta;
    if (quick_dialog (&display_bits) != B_ENTER)
	    return;

    eight_bit_clean = box_new_mode < 2;
    full_eight_bits = box_new_mode == 0;
    SLsmg_Display_Eight_Bit = full_eight_bits ? 128 : 160;
    use_8th_bit_as_meta = !new_meta;
}

#define TREE_Y 20
#define TREE_X 60

static int box_tree_callback (struct Dlg_head *h, int id, int msg)
{
    switch (msg) {

    case DLG_POST_KEY:
	/* The enter key will be processed by the tree widget */
	if (id == '\n' || ((WTree *)(h->current->widget))->done) {
	    h->ret_value = B_ENTER;
	    dlg_stop (h);
	}
	return MSG_HANDLED;

    case DLG_DRAW:
	common_dialog_repaint (h);
	break;
    }
    return MSG_NOT_HANDLED;
}

char *Tree (char *current_dir)
{
    WTree    *mytree;
    Dlg_head *dlg;
    char     *val;
    WButtonBar *bar;

    tree_colors [3] = dialog_colors [0];
    tree_colors [1] = dialog_colors [1];

    /* Create the components */
    dlg = create_dlg (0, 0, TREE_Y, TREE_X, tree_colors,
		      box_tree_callback, "[Directory Tree]", DLG_CENTER);
    mytree = tree_new (0, 2, 2, TREE_Y - 6, TREE_X - 5);
    add_widget (dlg, mytree);
    bar = buttonbar_new(1);
    add_widget (dlg, bar);
    bar->widget.x = 0;
    bar->widget.y = LINES - 1;

    run_dlg (dlg);
    if (dlg->ret_value == B_ENTER)
	val = strdup (mytree->selected_ptr->name);
    else
	val = 0;

    destroy_dlg (dlg);
    return val;
}

static int local_utime (char *path, struct utimbuf *times)
{
    return bb_utime (path, times->actime, times->modtime);
}

#ifdef USE_VFS

#if defined(USE_NETCODE)
#define VFSY 15
#else
#define VFSY 11
#endif

#define VFSX 56

static int vfs_timeout = 60; /* VFS timeout in seconds */

/* The limit for a compressed tar file to be loaded in core */
static int tar_gzipped_memlimit = 1*1024*1024;

#if defined(USE_NETCODE)
static char *ret_passwd;
static char *ret_directory_timeout;
static char *ret_ftp_proxy;
static int ret_use_netrc;
#endif

static char *confvfs_str [] =
{ "Always to memory", "If size less than:" };

static const QuickWidget confvfs_widgets_st [] = {
  { quick_button,   30,  VFSX,    VFSY - 3, VFSY, "&Cancel",
      0, B_CANCEL, 0, 0, XV_WLAY_RIGHTOF, "button-cancel" },
  { quick_button,   12, VFSX,    VFSY - 3, VFSY, "&Ok",
      0, B_ENTER, 0, 0, XV_WLAY_CENTERROW, "button-ok" },
#if defined(USE_NETCODE)
  { quick_input,    30, VFSX, 10, VFSY, "", 22, 0, 0, &ret_ftp_proxy,
      XV_WLAY_RIGHTDOWN, "input-ftp-proxy" },
  { quick_checkbox,    4, VFSX, 10, VFSY, "&Always use ftp proxy", 0, 0,
      &ftpfs_always_use_proxy, 0, XV_WLAY_RIGHTDOWN, "check-ftp-proxy" },
  { quick_label,    46, VFSX, 9, VFSY, "sec",
      0, 0, 0, 0, XV_WLAY_RIGHTOF, "label-sec" },
  { quick_input,    35, VFSX, 9, VFSY, "", 10, 0, 0, &ret_directory_timeout,
      XV_WLAY_RIGHTDOWN, "input-timeout" },
  { quick_label,     4, VFSX, 9, VFSY, "ftpfs directory cache timeout:",
      0, 0, 0, 0, XV_WLAY_NEXTROW, "label-cache"},
  { quick_input,    28, VFSX, 8, VFSY, "", 24, 0, 0, &ret_passwd,
      XV_WLAY_RIGHTDOWN, "input-passwd" },
  { quick_label,     4, VFSX, 8, VFSY, "ftp anonymous password:",
      0, 0, 0, 0, XV_WLAY_NEXTROW, "label-pass"},
#endif
  { quick_input,    26, VFSX, 6, VFSY, "", 10, 0, 0, &ret_limit,
      XV_WLAY_RIGHTDOWN, "input-limit" },
  { quick_radio,    4, VFSX, 5, VFSY, "", 2, 0,
      &vfs_use_limit, confvfs_str, XV_WLAY_BELOWCLOSE, "radio" },
  { quick_label,    4,  VFSX, 4, VFSY, "Gzipped tar archive extract:",
      0, 0, 0, 0, XV_WLAY_NEXTROW, "label-tar" },
  { quick_label,    46, VFSX, 3, VFSY, "sec",
      0, 0, 0, 0, XV_WLAY_RIGHTOF, "label-sec2" },
  { quick_input,    35, VFSX, 3, VFSY, "", 10, 0, 0, &ret_timeout,
      XV_WLAY_RIGHTOF, "input-timo-vfs" },
  { quick_label,    4,  VFSX, 3, VFSY, "Timeout for freeing VFSs:",
      0, 0, 0, 0, XV_WLAY_BELOWCLOSE, "label-vfs" },
  { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, XV_WLAY_DONTCARE, 0 }
};

#if defined(USE_NETCODE)
#define VFS_WIDGETBASE 7
#else
#define VFS_WIDGETBASE 0
#endif

static void configure_vfs (void)
{
    char buffer1 [15], buffer2 [15];
#if defined(USE_NETCODE)
    char buffer3[15];
#endif
     QuickDialog confvfs_dlg;

     confvfs_dlg.xlen = VFSX;
     confvfs_dlg.ylen = VFSY;
     confvfs_dlg.xpos = -1;
     confvfs_dlg.ypos = -1;
     confvfs_dlg.title = " Virtual File System Setting ";
     confvfs_dlg.help = "[Virtual FS]";
     confvfs_dlg.widgets = alloca(sizeof(confvfs_widgets_st));
     memcpy(confvfs_dlg.widgets, confvfs_widgets_st, sizeof(confvfs_widgets_st));

    if (tar_gzipped_memlimit > -1) {
	if (tar_gzipped_memlimit == 0)
	    strcpy (buffer1, "0 B");
	else if ((tar_gzipped_memlimit % (1024*1024)) == 0) /* I.e. in M */
	    sprintf (buffer1, "%i MB", (int)(((unsigned)tar_gzipped_memlimit) >> 20));
	else if ((tar_gzipped_memlimit % 1024) == 0) /* I.e. in K */
	    sprintf (buffer1, "%i KB", (int)(((unsigned)tar_gzipped_memlimit) >> 10));
	else if ((tar_gzipped_memlimit % 1000) == 0)
	    sprintf (buffer1, "%i kB", (int)(tar_gzipped_memlimit / 1000));
	else
	    sprintf (buffer1, "%i B", (int)tar_gzipped_memlimit);
	confvfs_dlg.widgets [2 + VFS_WIDGETBASE].text = buffer1;
    } else
	confvfs_dlg.widgets [2 + VFS_WIDGETBASE].text = "5 MB";
    sprintf (buffer2, "%i", vfs_timeout);
    confvfs_dlg.widgets [6 + VFS_WIDGETBASE].text = buffer2;
    confvfs_dlg.widgets [3 + VFS_WIDGETBASE].value = vfs_use_limit;
#if defined(USE_NETCODE)
    ret_use_netrc = use_netrc;
    sprintf(buffer3, "%i", ftpfs_directory_timeout);
    confvfs_dlg.widgets[5].text = buffer3;
    confvfs_dlg.widgets[7].text = ftpfs_anonymous_passwd;
    confvfs_dlg.widgets[2].text = ftpfs_proxy_host ? ftpfs_proxy_host : "";
#endif

    if (quick_dialog (&confvfs_dlg) != B_CANCEL) {
	char *p;

	vfs_timeout = atoi (ret_timeout);
	free (ret_timeout);
	if (vfs_timeout < 0 || vfs_timeout > 10000)
	    vfs_timeout = 10;
	if (!vfs_use_limit)
	    tar_gzipped_memlimit = -1;
	else {
	    tar_gzipped_memlimit = atoi (ret_limit);
	    if (tar_gzipped_memlimit < 0)
		tar_gzipped_memlimit = -1;
	    else {
		for (p = ret_limit; *p == ' ' || (*p >= '0' && *p <= '9'); p++);
		switch (*p) {
		case 'm':
		case 'M': tar_gzipped_memlimit <<= 20; break;
		case 'K': tar_gzipped_memlimit <<= 10; break;
		case 'k': tar_gzipped_memlimit *= 1000; break;
		}
	    }
	}
	free (ret_limit);
#if defined(USE_NETCODE)
	free(ftpfs_anonymous_passwd);
	ftpfs_anonymous_passwd = ret_passwd;
	if (ftpfs_proxy_host)
	    free(ftpfs_proxy_host);
	ftpfs_proxy_host = ret_ftp_proxy;
	ftpfs_directory_timeout = atoi(ret_directory_timeout);
	use_netrc = ret_use_netrc;
	free(ret_directory_timeout);
#endif
    }
}

#endif

static char *cd_dialog (void)
{
    QuickDialog Quick_input;
    static const QuickWidget quick_widgets_st [] = {
      { quick_input,  6, 57, 5, 0, "", 50, 0, 0, 0, XV_WLAY_RIGHTOF, "input" },
      { quick_label,  3, 57, 2, 0, "",  0, 0, 0, 0, XV_WLAY_DONTCARE, "label" },
      { 0 }
    };

    char *my_str;
    int len;

    QuickWidget *quick_widgets;

    quick_widgets = alloca(sizeof(quick_widgets_st));
    memcpy(quick_widgets, quick_widgets_st, sizeof(quick_widgets_st));
    Quick_input.xlen  = 57;
    Quick_input.title = "Quick cd";
    Quick_input.help  = "[Quick cd]";
    quick_widgets [0].text = "";
    quick_widgets [0].value = 2; /* want cd like completion */
    quick_widgets [0+1].text = "cd";
    quick_widgets [0+1].y_divisions = quick_widgets [0].y_divisions = Quick_input.ylen = 5;

    len = strlen (quick_widgets [0+1].text);

    quick_widgets [0+1].relative_x = 3;
    quick_widgets [0].relative_x = quick_widgets [0+1].relative_x + len + 1;

    Quick_input.xlen = len + quick_widgets [0].hotkey_pos + 7;
    quick_widgets [0].x_divisions = quick_widgets [0+1].x_divisions = Quick_input.xlen;

    Quick_input.xpos = 2;
    Quick_input.ypos = LINES - 2 - Quick_input.ylen;
    quick_widgets [0].relative_y = 2;
    quick_widgets [0].str_result = &my_str;

    Quick_input.widgets = quick_widgets;
    if (quick_dialog (&Quick_input) != B_CANCEL) {
	return *(quick_widgets [0].str_result);
    } else
	return 0;
}

static void symlink_dialog (char *existing, char *new, char **ret_existing,
    char **ret_new)
{
    QuickDialog Quick_input;
    static const QuickWidget quick_widgets_st [] = {
      { quick_input,  6, 80, 5, 8, "", 58, 0, 0, 0, XV_WLAY_BELOWCLOSE, "input-1" },
      { quick_label,  6, 80, 4, 8, "", 0, 0, 0, 0, XV_WLAY_BELOWOF, "label-1" },
      { quick_input,  6, 80, 3, 8, "", 58, 0, 0, 0, XV_WLAY_BELOWCLOSE, "input-2" },
      { quick_label,  6, 80, 2, 8, "", 0, 0, 0, 0, XV_WLAY_DONTCARE, "label-2" },
      { 0 }
    };

    QuickWidget *quick_widgets;

    quick_widgets = alloca(sizeof(quick_widgets_st));
    memcpy(quick_widgets, quick_widgets_st, sizeof(quick_widgets_st));
    Quick_input.xlen  = 64;
    Quick_input.ylen  = 8;
    Quick_input.title = "Symbolic link";
    Quick_input.help  = "[File Menu]";
    quick_widgets [0].text = new;
    quick_widgets [0+1].text = "Symbolic link filename:";
    quick_widgets [0+2].text = existing;
    quick_widgets [0+3].text = "Existing filename (filename symlink will point to):";
    Quick_input.xpos = -1;
    quick_widgets [0].str_result = ret_new;
    quick_widgets [0+2].str_result = ret_existing;

    Quick_input.widgets = quick_widgets;
    if (quick_dialog (&Quick_input) == B_CANCEL) {
	*ret_new = NULL;
	*ret_existing = NULL;
    }
}

/*
	Chmod command -- for the Midnight Commander
*/

#define PX_CHMOD              5
#define PY_CHMOD              2

#define FX_CHMOD              40
#define FY_CHMOD              2

#define BX_CHMOD              6
#define BY_CHMOD              17

#define TX_CHMOD              40
#define TY_CHMOD              12

#define BUTTONS_CHMOD         6

#define B_MARKED        B_USER
#define B_ALL           B_USER+1
#define B_SETMRK        B_USER+2
#define B_CLRMRK        B_USER+3

static const struct {
    int ret_cmd, flags, y, x;
    const char *text;
} chmod_but[BUTTONS_CHMOD] =
{
    { B_CANCEL, NORMAL_BUTTON,  2, 33, "&Cancel" },
    { B_ENTER,  DEFPUSH_BUTTON, 2, 17, "&Set" },
    { B_CLRMRK, NORMAL_BUTTON,  0, 42, "C&lear marked" },
    { B_SETMRK, NORMAL_BUTTON,  0, 27, "S&et marked" },
    { B_MARKED, NORMAL_BUTTON,  0, 12, "&Marked all" },
    { B_ALL,    NORMAL_BUTTON,  0, 0,  "Set &all" },
};

static void chmod_toggle_select (void)
{
    int Id = ch_dlg->current->dlg_id - BUTTONS_CHMOD + chm_single_set * 2;

    attrset (normal_color);
    check_perm[Id].selected ^= 1;

    dlg_move (ch_dlg, PY_CHMOD + PERMISSIONS - Id, PX_CHMOD + 1);
    addch ((check_perm[Id].selected) ? '*' : ' ');
    dlg_move (ch_dlg, PY_CHMOD + PERMISSIONS - Id, PX_CHMOD + 3);
}

static void chmod_refresh (void)
{
    attrset (COLOR_NORMAL);
    dlg_erase (ch_dlg);

    draw_box (ch_dlg, 1, 2, 20 - chm_single_set, 66);
    draw_box (ch_dlg, PY_CHMOD, PX_CHMOD, PERMISSIONS + 2, 33);
    draw_box (ch_dlg, FY_CHMOD, FX_CHMOD, 10, 25);

    dlg_move (ch_dlg, FY_CHMOD + 1, FX_CHMOD + 2);
    addstr ("Name");
    dlg_move (ch_dlg, FY_CHMOD + 3, FX_CHMOD + 2);
    addstr ("Permissions (Octal)");
    dlg_move (ch_dlg, FY_CHMOD + 5, FX_CHMOD + 2);
    addstr ("Owner name");
    dlg_move (ch_dlg, FY_CHMOD + 7, FX_CHMOD + 2);
    addstr ("Group name");

    attrset (title_color);
    dlg_move (ch_dlg, 1, 28);
    addstr (" Chmod command ");
    dlg_move (ch_dlg, PY_CHMOD, PX_CHMOD + 1);
    addstr (" Permission ");
    dlg_move (ch_dlg, FY_CHMOD, FX_CHMOD + 1);
    addstr (" File ");

    attrset (selection_color);

    dlg_move (ch_dlg, TY_CHMOD, TX_CHMOD);
    addstr ("Use SPACE to change");
    dlg_move (ch_dlg, TY_CHMOD + 1, TX_CHMOD);
    addstr ("an option, ARROW KEYS");
    dlg_move (ch_dlg, TY_CHMOD + 2, TX_CHMOD);
    addstr ("to move between options");
    dlg_move (ch_dlg, TY_CHMOD + 3, TX_CHMOD);
    addstr ("and T or INS to mark");
}

static int chmod_callback (Dlg_head *h, int Par, int Msg)
{
    char buffer [10];

    switch (Msg) {
    case DLG_ACTION:
	if (Par >= BUTTONS_CHMOD - chm_single_set * 2) {
	    c_stat ^= check_perm[Par - BUTTONS_CHMOD + chm_single_set * 2].mode;
	    sprintf (buffer, "%o", c_stat);
	    label_set_text (statl, buffer);
	    chmod_toggle_select ();
	    mode_change = 1;
	}
	break;

    case DLG_KEY:
	if ((Par == 'T' || Par == 't' || Par == KEY_IC) &&
	    ch_dlg->current->dlg_id >= BUTTONS_CHMOD - chm_single_set * 2) {
	    chmod_toggle_select ();
	    if (Par == KEY_IC)
		dlg_one_down (ch_dlg);
	    return 1;
	}
	break;
    case DLG_DRAW:
	chmod_refresh ();
	break;
    }
    return 0;
}

static void init_chmod (void)
{
    int i;

    check_perm[ 0].mode = S_IXOTH; check_perm[ 0].text = "execute/search by others";
    check_perm[ 1].mode = S_IWOTH; check_perm[ 1].text = "write by others";
    check_perm[ 2].mode = S_IROTH; check_perm[ 2].text = "read by others";
    check_perm[ 3].mode = S_IXGRP; check_perm[ 3].text = "execute/search by group";
    check_perm[ 4].mode = S_IWGRP; check_perm[ 4].text = "write by group";
    check_perm[ 5].mode = S_IRGRP; check_perm[ 5].text = "read by group";
    check_perm[ 6].mode = S_IXUSR; check_perm[ 6].text = "execute/search by owner";
    check_perm[ 7].mode = S_IWUSR; check_perm[ 7].text = "write by owner";
    check_perm[ 8].mode = S_IRUSR; check_perm[ 8].text = "read by owner";
    check_perm[ 9].mode = S_ISVTX; check_perm[ 9].text = "sticky bit";
    check_perm[10].mode = S_ISGID; check_perm[10].text = "set group ID on execution";
    check_perm[11].mode = S_ISUID; check_perm[11].text = "set user ID on execution";
    do_refresh ();
    end_chmod = c_file = chm_need_update = 0;
    chm_single_set = (cpanel->marked < 2) ? 2 : 0;

    if (use_colors) {
	normal_color = COLOR_NORMAL;
	title_color  = COLOR_HOT_NORMAL;
	selection_color = COLOR_NORMAL;
    } else {
	normal_color = NORMAL_COLOR;
	title_color  = SELECTED_COLOR;
	selection_color = SELECTED_COLOR;
    }

    ch_dlg = create_dlg (0, 0, 22 - chm_single_set, 70, dialog_colors,
			 chmod_callback, "[Chmod]", DLG_CENTER);

    x_set_dialog_title (ch_dlg, "Chmod command");

#define XTRACT_CHMOD(i) BY_CHMOD+chmod_but[i].y-chm_single_set, BX_CHMOD+chmod_but[i].x, \
     chmod_but[i].ret_cmd, chmod_but[i].flags, chmod_but[i].text, 0, 0

    for (i = 0; i < BUTTONS_CHMOD; i++) {
	if (i == 2 && chm_single_set)
	    break;
	else
	    add_widgetl (ch_dlg, button_new (XTRACT_CHMOD (i)), XV_WLAY_RIGHTOF);
    }


#define XTRACT2(i) 0, check_perm [i].text
    for (i = 0; i < PERMISSIONS; i++) {
	check_perm[i].check = check_new (PY_CHMOD + (PERMISSIONS - i), PX_CHMOD + 2,
					 XTRACT2 (i));
	add_widget (ch_dlg, check_perm[i].check);
    }
}

static int stat_file (char *filename, struct stat *st)
{
    if (mc_stat (filename, st))
	return 0;
    if (!(S_ISREG(st->st_mode) || S_ISDIR(st->st_mode) ||
	  S_ISLNK(st->st_mode)))
	return 0;

    return 1;
}

static void chmod_done (void)
{
    if (chm_need_update)
	update_panels (UP_OPTIMIZE, UP_KEEPSEL);
    repaint_screen ();
}

static char *chm_next_file (void)
{
    while (!cpanel->dir.list[c_file].f.marked)
	c_file++;

    return cpanel->dir.list[c_file].fname;
}

static void do_chmod (struct stat *sf)
{
    sf->st_mode &= and_mask;
    sf->st_mode |= or_mask;
    if (mc_chmod (cpanel->dir.list [c_file].fname, sf->st_mode) == -1)
	MC_message (1, MSG_ERROR, " Couldn't chmod \"%s\" \n %m ",
	     cpanel->dir.list [c_file].fname);

    do_file_mark (cpanel, c_file, 0);
}

static void apply_mask (struct stat *sf)
{
    char *Fname;

    chm_need_update = end_chmod = 1;
    do_chmod (sf);

    do {
	Fname = chm_next_file ();
	if (!stat_file (Fname, sf))
	    return;
	c_stat = sf->st_mode;

	do_chmod (sf);
    } while (cpanel->marked);
}

static void chmod_cmd (void)
{
    char buffer [10];
    char *Fname;
    int i;
    struct stat chmod_cmd_sf_stat;

    if (!vfs_current_is_local ()) {
	if (vfs_current_is_extfs ()) {
	    MC_message (1, " Oops... ",
		     " I can't run the Chmod command on an extfs ");
	    return;
	} else if (vfs_current_is_tarfs ()) {
	    MC_message (1, " Oops... ",
		     (" I can't run the Chmod command on a tarfs "));
	    return;
	}
    }

    do {                        /* do while any files remaining */
	init_chmod ();
	if (cpanel->marked)
	    Fname = chm_next_file ();       /* next marked file */
	else
	    Fname = selection (cpanel)->fname;  /* single file */

	if (!stat_file (Fname, &chmod_cmd_sf_stat)) {      /* get status of file */
	    destroy_dlg (ch_dlg);
	    break;
	}

	c_stat = chmod_cmd_sf_stat.st_mode;
	mode_change = 0;        /* clear changes flag */

	/* set check buttons */
	for (i = 0; i < PERMISSIONS; i++) {
	    check_perm[i].check->state = (c_stat & check_perm[i].mode) ? 1 : 0;
	    check_perm[i].selected = 0;
	}

	/* Set the labels */
	c_fname = name_trunc (Fname, 21);
	add_widget (ch_dlg, label_new (FY_CHMOD+2, FX_CHMOD+2, c_fname));
	c_fown = name_trunc (get_owner (chmod_cmd_sf_stat.st_uid), 21);
	add_widget (ch_dlg, label_new (FY_CHMOD+6, FX_CHMOD+2, c_fown));
	c_fgrp = name_trunc (get_group (chmod_cmd_sf_stat.st_gid), 21);
	add_widget (ch_dlg, label_new (FY_CHMOD+8, FX_CHMOD+2, c_fgrp));
	sprintf (buffer, "%o", c_stat);
	statl = label_new (FY_CHMOD+4, FX_CHMOD+2, buffer);
	add_widget (ch_dlg, statl);

	run_dlg (ch_dlg);       /* retrieve an action */

	/* do action */
	switch (ch_dlg->ret_value) {
	case B_ENTER:
	    if (mode_change)
		if (mc_chmod (Fname, c_stat) == -1)
		    MC_message (1, MSG_ERROR, " Couldn't chmod \"%s\" \n %m ",
			 Fname);
	    chm_need_update = 1;
	    break;

	case B_CANCEL:
	    end_chmod = 1;
	    break;

	case B_ALL:
	case B_MARKED:
	    and_mask = or_mask = 0;
	    and_mask = ~and_mask;

	    for (i = 0; i < PERMISSIONS; i++)
		if (check_perm[i].selected || ch_dlg->ret_value == B_ALL) {
		    if (check_perm[i].check->state & C_BOOL)
			or_mask |= check_perm[i].mode;
		    else
			and_mask &= ~check_perm[i].mode;
		}

	    apply_mask (&chmod_cmd_sf_stat);
	    break;

	case B_SETMRK:
	    and_mask = or_mask = 0;
	    and_mask = ~and_mask;

	    for (i = 0; i < PERMISSIONS; i++) {
		if (check_perm[i].selected)
		    or_mask |= check_perm[i].mode;
	    }

	    apply_mask (&chmod_cmd_sf_stat);
	    break;
	case B_CLRMRK:
	    and_mask = or_mask = 0;
	    and_mask = ~and_mask;

	    for (i = 0; i < PERMISSIONS; i++) {
		if (check_perm[i].selected)
		    and_mask &= ~check_perm[i].mode;
	    }

	    apply_mask (&chmod_cmd_sf_stat);
	    break;
	}

	if (cpanel->marked && ch_dlg->ret_value!=B_CANCEL) {
	    do_file_mark (cpanel, c_file, 0);
	    chm_need_update = 1;
	}
	destroy_dlg (ch_dlg);
    } while (cpanel->marked && !end_chmod);
    chmod_done ();
}

/*
	Chown command -- for the Midnight Commander
*/

#define UX_CHOWN              5
#define UY_CHOWN              2

#define GX_CHOWN              27
#define GY_CHOWN              2

#define BX_CHOWN              5
#define BY_CHOWN              15

#define TX_CHOWN              50
#define TY_CHOWN              2

#define BUTTONS_CHOWN         5

#define B_SETALL        B_USER
#define B_SETUSR        B_USER + 1
#define B_SETGRP        B_USER + 2


static const struct {
    int ret_cmd, flags, y, x;
    const char *text;
} chown_but[BUTTONS_CHOWN] = {
    { B_CANCEL, NORMAL_BUTTON,  0, 54, "&Cancel" },
    { B_ENTER,  DEFPUSH_BUTTON, 0, 44, "&Set" },
    { B_SETUSR, NORMAL_BUTTON,  0, 30, "Set &users" },
    { B_SETGRP, NORMAL_BUTTON,  0, 15, "Set &groups" },
    { B_SETALL, NORMAL_BUTTON,  0, 3,  "Set &all" },
};


static void chown_refresh (void)
{
    attrset (COLOR_NORMAL);
    dlg_erase (ch_dlg);

    draw_box (ch_dlg, 1, 2, 16, 70);
    draw_box (ch_dlg, UY_CHOWN, UX_CHOWN, 12, 21);
    draw_box (ch_dlg, GY_CHOWN, GX_CHOWN, 12, 21);
    draw_box (ch_dlg, TY_CHOWN, TX_CHOWN, 12, 19);

    attrset (COLOR_NORMAL);
    dlg_move (ch_dlg, TY_CHOWN + 1, TX_CHOWN + 1);
    addstr (" Name ");
    dlg_move (ch_dlg, TY_CHOWN + 3, TX_CHOWN + 1);
    addstr (" Owner name ");
    dlg_move (ch_dlg, TY_CHOWN + 5, TX_CHOWN + 1);
    addstr (" Group name ");
    dlg_move (ch_dlg, TY_CHOWN + 7, TX_CHOWN + 1);
    addstr (" Size ");
    dlg_move (ch_dlg, TY_CHOWN + 9, TX_CHOWN + 1);
    addstr (" Permission ");

    attrset (COLOR_HOT_NORMAL);
    dlg_move (ch_dlg, 1, 28);
    addstr (" Chown command ");
    dlg_move (ch_dlg, UY_CHOWN, UX_CHOWN + 1);
    addstr (" User name ");
    dlg_move (ch_dlg, GY_CHOWN, GX_CHOWN + 1);
    addstr (" Group name ");
    dlg_move (ch_dlg, TY_CHOWN, TX_CHOWN + 1);
    addstr (" File ");
}

static char *cho_next_file (void)
{
    while (!cpanel->dir.list[cho_current_file].f.marked)
	cho_current_file++;

    return cpanel->dir.list[cho_current_file].fname;
}

static int chown_callback (Dlg_head * h, int Par, int Msg)
{
    if (Msg == DLG_DRAW)
      chown_refresh ();
    return 0;
}


static void init_chown (void)
{
    int i;
    struct passwd *l_pass;
    struct group *l_grp;

    do_refresh ();
    end_chown = cho_need_update = cho_current_file = 0;
    cho_single_set = (cpanel->marked < 2) ? 3 : 0;

    ch_dlg = create_dlg (0, 0, 18, 74, dialog_colors, chown_callback,
			 "[Chown]", DLG_CENTER);

#define XTRACT_CHOWN(i) BY_CHOWN+chown_but[i].y, BX_CHOWN+chown_but[i].x, chown_but[i].ret_cmd, chown_but[i].flags, chown_but[i].text, 0, 0

    for (i = 0; i < BUTTONS_CHOWN-cho_single_set; i++)
	add_widget (ch_dlg, button_new (XTRACT_CHOWN (i)));

    /* Add the widgets for the file information */
    for (i = 0; i < LABELS_CHOWN; i++) {
	chown_label[i].y = TY_CHOWN + (i + 1) * 2;
	chown_label[i].x = TX_CHOWN + 2;
	chown_label[i].l = label_new (chown_label[i].y, chown_label[i].x, "");
	add_widget (ch_dlg, chown_label [i].l);
    }

    /* get new listboxes */
    l_user = listbox_new (UY_CHOWN + 1, UX_CHOWN + 1, 19, 10, 0, l_call_1);
    l_group = listbox_new (GY_CHOWN + 1, GX_CHOWN + 1, 19, 10, 0, l_call_1);

    listbox_add_item (l_user, 0, 0, "<Unknown user>", NULL); /* add fields for unknown names (numbers) */
    listbox_add_item (l_group, 0, 0, "<Unknown group>", NULL);

    setpwent ();                /* get and put user names in the listbox */
    while ((l_pass = getpwent ()))
	listbox_add_item (l_user, 0, 0, l_pass->pw_name, NULL);
    endpwent ();

    setgrent ();                /* get and put group names in the listbox */
    while ((l_grp = getgrent ())) {
	listbox_add_item (l_group, 0, 0, l_grp->gr_name, NULL);
    }
    endgrent ();

    add_widget (ch_dlg, l_group);
    add_widget (ch_dlg, l_user);        /* add listboxes to the dialogs */
}

static void chown_done (void)
{
    if (cho_need_update)
	update_panels (UP_OPTIMIZE, UP_KEEPSEL);
    repaint_screen ();
}

static inline void do_chown (uid_t u, gid_t g)
{
    if (mc_chown (cpanel->dir.list [cho_current_file].fname, u, g) == -1)
	MC_message (1, MSG_ERROR, " Couldn't chown \"%s\" \n %m ",
	     cpanel->dir.list [cho_current_file].fname);

    do_file_mark (cpanel, cho_current_file, 0);
}

static void apply_chowns (uid_t u, gid_t g)
{
    cho_need_update = end_chown = 1;
    do_chown (u,g);

    do {
	cho_next_file ();
	do_chown (u,g);
    } while (cpanel->marked);
}

static char *sprintf_pd(char *buffer, int val)
{
	sprintf (buffer, "%d", val);
	return buffer;
}

#define Xchown_label(n,txt) label_set_text (chown_label [n].l, txt)

static void chown_cmd (void)
{
    char *Fname;
    struct stat chown_cmd_sf_stat;
    WLEntry *fe;
    uid_t new_user;
    gid_t new_group;
    char  buffer [15];

    if (!vfs_current_is_local ()) {
	if (vfs_current_is_extfs ()) {
	    MC_message (1, " Oops... ",
		     " I can't run the Chown command on an extfs ");
	    return;
	} else if (vfs_current_is_tarfs ()) {
	    MC_message (1, " Oops... ",
		     " I can't run the Chown command on a tarfs ");
	    return;
	}
    }

    do {                        /* do while any files remaining */
	init_chown ();
	new_user = new_group = -1;

	if (cpanel->marked)
	    Fname = cho_next_file ();               /* next marked file */
	else
	    Fname = selection (cpanel)->fname;          /* single file */

	if (!stat_file (Fname, &chown_cmd_sf_stat)) {      /* get status of file */
	    destroy_dlg (ch_dlg);
	    break;
	}

	/* select in listboxes */
	fe = listbox_search_text (l_user, get_owner(chown_cmd_sf_stat.st_uid));
	if (fe)
	    listbox_select_entry (l_user, fe);

	fe = listbox_search_text (l_group, get_group(chown_cmd_sf_stat.st_gid));
	if (fe)
	    listbox_select_entry (l_group, fe);

	Xchown_label (0, name_trunc (Fname, 15));
	Xchown_label (1, name_trunc (get_owner (chown_cmd_sf_stat.st_uid), 15));
	Xchown_label (2, name_trunc (get_group (chown_cmd_sf_stat.st_gid), 15));
	buffer[0] = '0'; buffer[1] = '\0';
	Xchown_label (3, buffer);
	Xchown_label (4, bb_mode_string (chown_cmd_sf_stat.st_mode, bb_mode_string_buf));

	run_dlg (ch_dlg);

	switch (ch_dlg->ret_value) {
	case B_CANCEL:
	    end_chown = 1;
	    break;

	case B_SETUSR:
	{
	    struct passwd *User;

	    User = getpwnam (l_user->current->text);
	    if (User) {
		new_user = User->pw_uid;
		apply_chowns (new_user, new_group);
	    }
	    break;
	}
	case B_SETGRP:
	{
	    struct group *grp;

	    grp = getgrnam (l_group->current->text);
	    if (grp) {
		new_group = grp->gr_gid;
		apply_chowns (new_user, new_group);
	    }
	    break;
	}
	case B_SETALL:
	case B_ENTER:
	{
	    struct group *grp;
	    struct passwd *User;

	    grp = getgrnam (l_group->current->text);
	    if (grp)
		new_group = grp->gr_gid;
	    User = getpwnam (l_user->current->text);
	    if (User)
		new_user = User->pw_uid;
	    if (ch_dlg->ret_value==B_ENTER) {
		cho_need_update = 1;
		if (mc_chown (Fname, new_user, new_group) == -1)
		    MC_message (1, MSG_ERROR, " Couldn't chown \"%s\" \n %m ",
			 Fname);
	    } else
		apply_chowns (new_user, new_group);
	    break;
	}
	}

	if (cpanel->marked && ch_dlg->ret_value != B_CANCEL) {
	    do_file_mark (cpanel, cho_current_file, 0);
	    cho_need_update = 1;
	}
	destroy_dlg (ch_dlg);
    } while (cpanel->marked && !end_chown);

    chown_done ();
}
/* Routines invoked by a function key
   They normally operate on the current panel.
*/

#define MENU_PANEL_IDX  (is_right ? 1 : 0)


#ifndef PORT_HAS_FILTER_CHANGED
#    define x_filter_changed(p)
#endif

/* This is used since the parameter panel on some of the commands */
/* defined in this file may receive a 0 parameter if they are invoked */
/* The drop down menu */

static WPanel *get_a_panel (WPanel *panel)
{
    if (panel)
	return panel;
    if (get_current_type () == view_listing) {
	return cpanel;
    } else
	return other_panel;
}

/* view_file (filename, normal, internal)
 *
 * Inputs:
 *   filename:   The file name to view
 *   plain_view: If set does not do any fancy pre-processing (no filtering) and
 *               always invokes the internal viewer.
 *   internal:   If set uses the internal viewer, otherwise an external viewer.
 */
static int
view_file_at_line (char *filename, int plain_view, int internal, int start_line)
{
    int move_dir = 0;

    if (plain_view) {
	int changed_hex_mode = 0;
	int changed_nroff_flag = 0;
	int changed_magic_flag = 0;

	altered_hex_mode = 0;
	altered_nroff_flag = 0;
	altered_magic_flag = 0;
	if (default_hex_mode)
	    changed_hex_mode = 1;
	if (default_nroff_flag)
	    changed_nroff_flag = 1;
	if (default_magic_flag)
	    changed_magic_flag = 1;
	default_hex_mode = 0;
	default_nroff_flag = 0;
	default_magic_flag = 0;
	View (0, filename, &move_dir, start_line);
	if (changed_hex_mode && !altered_hex_mode)
	    default_hex_mode = 1;
	if (changed_nroff_flag && !altered_nroff_flag)
	    default_nroff_flag = 1;
	if (changed_magic_flag && !altered_magic_flag)
	    default_magic_flag = 1;
	repaint_screen ();
	return move_dir;
    }
    if (internal) {
	char view_entry [32];

	if (start_line != 0)
	    sprintf (view_entry, "View:%d", start_line);
	else
	    strcpy (view_entry, "View");

	if (!regex_command (filename, view_entry, NULL, &move_dir)) {
	    View (0, filename, &move_dir, start_line);
	    repaint_screen ();
	}
    } else {
	char *localcopy;

	/* The file may be a non local file, get a copy */
	if (!vfs_file_is_local (filename)) {
	    localcopy = mc_getlocalcopy (filename);
	    if (localcopy == NULL) {
		MC_message (1, MSG_ERROR, " Can not fetch a local copy of %s ", filename);
		return 0;
	    }
	    execute_internal (viewer, localcopy);
	    mc_ungetlocalcopy (filename, localcopy, 0);
	} else
	    execute_internal (viewer, filename);
    }
    return move_dir;
}

static int
view_file (char *filename, int plain_view, int internal)
{
    return view_file_at_line (filename, plain_view, internal, 0);
}

/* scan_for_file (panel, idx, direction)
 *
 * Inputs:
 *   panel:     pointer to the panel on which we operate
 *   idx:       starting file.
 *   direction: 1, or -1
 */
static int scan_for_file (WPanel *panel, int idx, int direction)
{
    int i = idx + direction;

    while (i != idx) {
	if (i < 0)
	    i = panel->count - 1;
	if (i == panel->count)
	    i = 0;
	if (!S_ISDIR (panel->dir.list [i].buf.st_mode))
	    return i;
	i += direction;
    }
    return i;
}

static int link_isdir (file_entry *file)
{
    struct stat b;

    if (S_ISLNK (file->buf.st_mode)) {
	mc_stat (file->fname, &b);
	if (S_ISDIR (b.st_mode))
	    return 1;
    }
    return 0;
}

/* do_view: Invoked as the F3/F13 key. */
static void do_view_cmd (WPanel *panel, int normal)
{
    int dir, file_idx;
    panel = get_a_panel (panel);

    /* Directories are viewed by changing to them */
    if (S_ISDIR (selection (panel)->buf.st_mode) ||
	link_isdir (selection (panel))) {
	if (confirm_view_dir && (panel->marked || panel->dirs_marked)) {
	    if (query_dialog (" CD ", "Files tagged, want to cd?",
			      0, 2, "&Yes", "&No") == 1) {
		return;
	    }
	}
	do_cd (selection (panel)->fname, cd_exact);
	return;

    }

    file_idx = panel->selected;
    while (1) {
	char *filename;

	filename = panel->dir.list [file_idx].fname;

	dir = view_file (filename, normal, use_internal_view);
	if (dir == 0)
	    break;
	file_idx = scan_for_file (panel, file_idx, dir);
    }
}

static void view_cmd (WPanel *panel)
{
    do_view_cmd (panel, 0);
}

static void view_simple_cmd (WPanel *panel)
{
    do_view_cmd (panel, 1);
}

static void filtered_view_cmd (WPanel *panel)
{
    char *command;

    panel = get_a_panel (panel);
    command = input_dialog (" Filtered view ", " Filter command and arguments:",
			    selection (panel)->fname);
    if (!command)
	return;

    View (command, "", 0, 0);

    free (command);
}

static void filtered_view_cmd_cpanel (void)
{
    filtered_view_cmd (cpanel);
}

static void do_edit_at_line (const char *what, int start_line)
{
    if (!editor) {
	editor = getenv ("EDITOR");
	if (!editor)
	    editor = get_default_editor ();
    }
    execute_internal (editor, what);
    reread_cmd ();
}

static void do_edit (const char *what)
{
    do_edit_at_line (what, 1);
}

static void edit_cmd (WPanel *panel)
{
    panel = get_a_panel(panel);
    if (!regex_command (selection (panel)->fname, "Edit", NULL, 0))
	do_edit (selection (panel)->fname);
}

static void edit_cmd_new (WPanel *panel)
{
    do_edit ("");
}

static void copy_cmd (void)
{
    save_cwds_stat ();
    if (panel_operate (cpanel, OP_COPY, NULL)) {
	update_panels (UP_OPTIMIZE, UP_KEEPSEL);
	repaint_screen ();
    }
}

static void ren_cmd (void)
{
    save_cwds_stat ();
    if (panel_operate (cpanel, OP_MOVE, NULL)) {
	update_panels (UP_OPTIMIZE, UP_KEEPSEL);
	repaint_screen ();
    }
}

static void mkdir_cmd (WPanel *panel)
{
    char *dir;

    panel = get_a_panel (panel);
    dir = input_expand_dialog (" Mkdir ", " Enter directory name:" , "");

    if (!dir)
	return;

    save_cwds_stat ();
    if (my_mkdir (dir, 0777) == 0) {
	update_panels (UP_OPTIMIZE, dir);
	repaint_screen ();
	select_item (cpanel);
	free (dir);
	return;
    }
    free (dir);
    MC_message (1, MSG_ERROR, "  %m  ");
}

static void delete_cmd (void)
{
    save_cwds_stat ();

    if (panel_operate (cpanel, OP_DELETE, NULL)) {
	update_panels (UP_OPTIMIZE, UP_KEEPSEL);
	repaint_screen ();
    }
}

static void
set_panel_filter_to (WPanel *p, char *allocated_filter_string)
{
    if (p->filter) {
	free (p->filter);
	p->filter = 0;
    }
    if (!(allocated_filter_string [0] == '*' && allocated_filter_string [1] == 0))
	p->filter = allocated_filter_string;
    else
	free (allocated_filter_string);
    reread_cmd ();
    x_filter_changed (p);
}

/* Set a given panel filter expression */
static void set_panel_filter (WPanel *p)
{
    char *reg_exp;
    char *x;

    x = p->filter ? p->filter : easy_patterns ? "*" : ".";

    reg_exp = input_dialog (" Filter ", " Set expression for filtering filenames", x);
    if (!reg_exp)
	return;
    set_panel_filter_to (p, reg_exp);
}

/* Invoked from the left/right menus */
static void filter_cmd (void)
{
    WPanel *p;

    if (!SELECTED_IS_PANEL)
	return;

    p = MENU_PANEL;
    set_panel_filter (p);
}

static void reread_cmd (void)
{
    int flag;

    if (get_current_type () == view_listing &&
	get_other_type () == view_listing)
	flag = strcmp (cpanel->cwd, opanel->cwd) ? UP_ONLY_CURRENT : 0;
    else
	flag = UP_ONLY_CURRENT;

    update_panels (UP_RELOAD|flag, UP_KEEPSEL);
    repaint_screen ();
}

static void
do_sort (dir_list *list, sortfn *sort, int top, int reverse_f, int case_sensitive_f)
{
    int i;
    file_entry tmp_fe;

    for (i = 0; i < top + 1; i++) {             /* put ".." first in list */
	if (!strcmp (list->list [i].fname, "..")) {
	    if (i > 0) {                        /* swap [i] and [0] */
		memcpy (&tmp_fe, &(list->list [0]), sizeof (file_entry));
		memcpy (&(list->list [0]), &(list->list [i]), sizeof (file_entry));
		memcpy (&(list->list [i]), &tmp_fe, sizeof (file_entry));
	    }
	    break;
	}
    }

    reverse_sort_flag = reverse_f ? -1 : 1;
    sort_case_sensitive = case_sensitive_f;
    qsort (&(list->list) [1], top, sizeof (file_entry), (qsortfn)sort);
}

/* Panel sorting related routines */
static void do_re_sort (WPanel *panel)
{
    char *filename;
    int  i;

    panel = get_a_panel (panel);
    filename = strdup (selection (cpanel)->fname);
    unselect_item (panel);
    do_sort (&panel->dir, panel->sort_type, panel->count-1, panel->reverse, panel->case_sensitive);
    panel->selected = -1;
    for (i = panel->count; i; i--) {
	if (!strcmp (panel->dir.list [i-1].fname, filename)) {
	    panel->selected = i-1;
	    break;
	}
    }
    free (filename);
    cpanel->top_file = cpanel->selected - ITEMS (cpanel)/2;
    if (cpanel->top_file < 0)
	cpanel->top_file = 0;
    select_item (panel);
    panel_update_contents (panel);
}

static void reverse_selection_cmd_panel (WPanel *panel)
{
    file_entry *file;
    int i;

    for (i = 0; i < panel->count; i++) {
	file = &panel->dir.list [i];
	if (S_ISDIR (file->buf.st_mode))
	    continue;
	do_file_mark (panel, i, !file->f.marked);
    }
    paint_panel (panel);
}

static void reverse_selection_cmd (void)
{
    reverse_selection_cmd_panel (cpanel);
}

static void select_cmd_panel (WPanel *panel)
{
    char *reg_exp, *reg_exp_t;
    int i;
    int c;
    int dirflag = 0;

    reg_exp = input_dialog (" Select ", "", easy_patterns ? "*" : ".");
    if (!reg_exp)
	return;

    reg_exp_t = reg_exp;

    /* Check if they specified a directory */
    if (*reg_exp_t == '/') {
	dirflag = 1;
	reg_exp_t++;
    }
    if (reg_exp_t [strlen(reg_exp_t) - 1] == '/') {
	dirflag = 1;
	reg_exp_t [strlen(reg_exp_t) - 1] = 0;
    }

    for (i = 0; i < panel->count; i++) {
	if (!strcmp (panel->dir.list [i].fname, ".."))
	    continue;
	if (S_ISDIR (panel->dir.list [i].buf.st_mode)) {
	    if (!dirflag)
		continue;
	} else {
	    if (dirflag)
		continue;
	}
	c = regexp_match (reg_exp_t, panel->dir.list [i].fname, match_file);
	if (c == -1) {
	    MC_message (1, MSG_ERROR, "  Malformed regular expression  ");
	    free (reg_exp);
	    return;
	}
	if (c) {
	    do_file_mark (panel, i, 1);
	}
    }
    paint_panel (panel);
    free (reg_exp);
}

static void select_cmd (void)
{
	select_cmd_panel (cpanel);
}

static void unselect_cmd_panel (WPanel *panel)
{
    char *reg_exp, *reg_exp_t;
    int i;
    int c;
    int dirflag = 0;

    reg_exp = input_dialog (" Unselect ","", easy_patterns ? "*" : ".");
    if (!reg_exp)
	return;

    reg_exp_t = reg_exp;

    /* Check if they specified directory matching */
    if (*reg_exp_t == '/') {
	dirflag = 1;
	reg_exp_t ++;
    }
    if (reg_exp_t [strlen(reg_exp_t) - 1] == '/') {
	dirflag = 1;
	reg_exp_t [strlen(reg_exp_t) - 1] = 0;
    }
    for (i = 0; i < panel->count; i++) {
	if (!strcmp (panel->dir.list [i].fname, ".."))
	    continue;
	if (S_ISDIR (panel->dir.list [i].buf.st_mode)) {
	    if (!dirflag)
		continue;
	} else {
	    if (dirflag)
		continue;
	}
	c = regexp_match (reg_exp_t, panel->dir.list [i].fname, match_file);
	if (c == -1) {
	    MC_message (1, MSG_ERROR, "  Malformed regular expression  ");
	    free (reg_exp);
	    return;
	}
	if (c) {
	    do_file_mark (panel, i, 0);
	}
    }
    paint_panel (panel);
    free (reg_exp);
}

static void unselect_cmd (void)
{
	unselect_cmd_panel (cpanel);
}

/* Check if the file exists */
/* If not copy the default */
static int check_for_default(char *default_file, char *file)
{
    struct stat s;
    if (mc_stat (file, &s)) {
	if (mc_stat (default_file, &s)) {
	    return -1;
	}
	create_op_win (OP_COPY, 0);
	file_mask_defaults ();
	copy_file_file (default_file, file, 1);
	destroy_op_win ();
    }
    return 0;
}

static void ext_cmd (void)
{
    char *buffer;
    char *extdir;
    int  dir;

    dir = 0;
    if (my_euid == 0) {
	dir = query_dialog ("Extension file edit",
			    " Which extension file you want to edit? ", 0, 2,
			    "&User", "&System Wide");
    }
    extdir = concat_mcpath_file (MC_LIB_EXT);

    if (dir == 0) {
	buffer = concat_path_file (home_dir, MC_USER_EXT);
	check_for_default (extdir, buffer);
	do_edit (buffer);
	free (buffer);
    } else if (dir == 1)
	do_edit (extdir);

   free (extdir);
   flush_extension_file ();
}

static void menu_edit_cmd (void)
{
    char *buffer;
    char *menufile;
    int dir;

    dir = query_dialog (
	"Menu file edit",
	" Which menu file will you edit? ",
	0, my_euid ? 2 : 3,
	"&Local", "&Home", "&System Wide"
    );

    menufile = concat_path_file(mc_home, MC_GLOBAL_MENU);

    switch (dir) {
	case 0:
	    buffer = strdup (MC_LOCAL_MENU);
	    check_for_default (menufile, buffer);
	    break;

	case 1:
	    buffer = concat_path_file (home_dir, MC_HOME_MENU);
	    check_for_default (menufile, buffer);
	    break;

	case 2:
	    buffer = concat_mcpath_file (MC_GLOBAL_MENU);
	    break;

	default:
	    free (menufile);
	    return;
    }
    do_edit (buffer);
	if (dir == 0)
		chmod(buffer, 0600);
    free (buffer);
    free (menufile);
}

static void quick_chdir_cmd (void)
{
    char *target;

    target = hotlist_cmd (LIST_HOTLIST);
    if (!target)
	return;

    if (get_current_type () == view_tree)
	tree_chdir (the_tree, target);
    else
	do_cd (target, cd_exact);
    free (target);
}

#ifdef USE_VFS
static void reselect_vfs (void)
{
    char *target;

    target = hotlist_cmd (LIST_VFSLIST);
    if (!target)
	return;

    do_cd (target, cd_exact);
    free (target);
}
#endif

static int prepare4compare(const char *name, char **data)
{
    int fd = open (name, O_RDONLY);
    if(fd >= 0) {
	*data = mmap (0, compare_src_size, PROT_READ,
				MAP_FILE | MAP_PRIVATE, fd, 0);
    }
    return fd;
}

static int compare_files (const char *name1, const char *name2)
{
    int file1, file2;
    char *data1, *data2;
    int result = -1;            /* Different by default */

    file1 = prepare4compare(name1, &data1);
    if (file1 >= 0) {
	if (data1 != (char*) -1) {
	    file2 = prepare4compare(name2, &data2);
	    if (file2 >= 0) {
		if (data2 != (char*) -1) {
		    rotate_dash ();
		    result = memcmp (data1, data2, compare_src_size);
		    munmap (data2, compare_src_size);
		}
		close (file2);
	    }
	    munmap (data1, compare_src_size);
	}
	close (file1);
    }
    return result;
}

enum CompareMode {
    compare_quick, compare_size_only, compare_thourough
};

static void
compare_dir (WPanel *panel, WPanel *other, enum CompareMode mode)
{
    int i, j;
    char *src_name, *dst_name;

    panel = get_a_panel (panel);

    /* No marks by default */
    panel->marked = 0;
    panel->total = 0;
    panel->dirs_marked = 0;

    /* Handle all files in the panel */
    for (i = 0; i < panel->count; i++) {
	file_entry *source = &panel->dir.list[i];

	/* Default: unmarked */
	file_mark (panel, i, 0);

	/* Skip directories */
	if (S_ISDIR (source->buf.st_mode))
	    continue;

	/* Search the corresponding entry from the other panel */
	for (j = 0; j < other->count; j++) {
	    if (strcmp (source->fname,
			other->dir.list[j].fname) == 0)
		break;
	}
	if (j >= other->count)
	    /* Not found -> mark */
	    do_file_mark (panel, i, 1);
	else {
	    /* Found */
	    file_entry *target = &other->dir.list[j];

	    if (mode != compare_size_only) {
		/* Older version is not marked */
		if (source->buf.st_mtime < target->buf.st_mtime)
		    continue;
	    }

	    /* Newer version with different size is marked */
	    if (source->buf.st_size != target->buf.st_size) {
		do_file_mark (panel, i, 1);
		continue;

	    }
	    if (mode == compare_size_only)
		continue;

	    if (mode == compare_quick) {
		/* Thorough compare off, compare only time stamps */
		/* Mark newer version, don't mark version with the same date */
		if (source->buf.st_mtime > target->buf.st_mtime) {
		    do_file_mark (panel, i, 1);
		}
		continue;
	    }

	    /* Thorough compare on, do byte-by-byte comparison */
	    src_name = get_full_name (panel->cwd, source->fname);
	    dst_name = get_full_name (other->cwd, target->fname);
	    compare_src_size = source->buf.st_size;
	    if (compare_files (src_name, dst_name))
		do_file_mark (panel, i, 1);
	    free (src_name);
	    free (dst_name);
	}
    } /* for (i ...) */
}

static void compare_dirs_cmd (void)
{
    enum CompareMode thorough_flag = compare_quick;

    thorough_flag = query_dialog (" Compare directories ", " Select compare method: ",
				  0, 3, "&Quick", "&Size only", "&Thorough", "&Cancel");
    if (thorough_flag < 0 || thorough_flag > 2)
	return;
    if (get_current_type () == view_listing &&
	get_other_type () == view_listing) {
	compare_dir (cpanel, opanel, thorough_flag);
	compare_dir (opanel, cpanel, thorough_flag);
	paint_panel (cpanel);
	paint_panel (opanel);
    } else {
	MC_message (1, MSG_ERROR, " Both panels should be on the listing view mode to use this command ");
    }
}

static void history_cmd (void)
{
    Listbox *listbox;
    Hist *current;

    if (input_w (cmdline)->need_push) {
	if (push_history (input_w (cmdline), input_w (cmdline)->buffer) == 2)
	    input_w (cmdline)->need_push = 0;
    }
    if (!input_w (cmdline)->history) {
	MC_message (1, MSG_ERROR, " The command history is empty ");
	return;
    }
    current = input_w (cmdline)->history;
    while (current->prev)
	current = current->prev;
    listbox = create_listbox_window (60, 10, " Command history ",
				     "[Command Menu]");
    while (current) {
	LISTBOX_APPEND_TEXT (listbox, 0, current->text,
			     current);
	current = current->next;
    }
    run_dlg (listbox->dlg);
    if (listbox->dlg->ret_value == B_CANCEL)
	current = NULL;
    else
	current = listbox->list->current->data;
    destroy_dlg (listbox->dlg);
    free (listbox);

    if (!current)
	return;
    input_w (cmdline)->history = current;
    assign_text (input_w (cmdline), input_w (cmdline)->history->text);
    update_input (input_w (cmdline), 1);
}

static void swap_cmd (void)
{
    swap_panels ();
    touchwin ();
    repaint_screen ();
}

static void
view_other_cmd (void)
{
    if (!xterm_flag && !console_flag && !use_subshell) {
	if (!w_message_flag)
	    MC_message (1, MSG_ERROR, " Not an xterm or Linux console; \n"
				     " the panels cannot be toggled. ");
	w_message_flag = 1;
    } else {
	shut_mouse ();
	if (clear_before_exec)
	    clr_scr ();
	if (alternate_plus_minus)
	    numeric_keypad_mode ();
	keypad(FALSE);
	endwin ();
	do_exit_ca_mode ();
	raw ();
	if (console_flag)
	    restore_console ();
	if (output_starts_shell) {
		fprintf (stderr,
		 "Type `exit' to return to the Midnight Commander\n\r\n\r");
		my_system (EXECUTE_AS_SHELL, MC_shell, NULL);
	} else
		get_key_code (0);
	if (console_flag)
	    handle_console (CONSOLE_SAVE);

	do_enter_ca_mode ();

	reset_prog_mode ();
	keypad(TRUE);
	init_mouse ();
	if (alternate_plus_minus)
	    application_keypad_mode ();

	touchwin ();

	/* prevent screen flash when user did 'exit' or 'logout' within
	   subshell */
	if (!MC_quit)
	    repaint_screen ();
    }
}

static void
do_link (int symbolic_link, char *Fname)
{
    struct stat st;
    char *dest, *src;
    int  stat_r;

    if (!symbolic_link) {
	stat_r = mc_stat (Fname, &st);
	if (stat_r != 0) {
	    MC_message (1, MSG_ERROR, " Couldn't stat %s \n %m ",
		     Fname);
	    return;
	}
	if (!S_ISREG (st.st_mode))
	    return;
    }

    if (!symbolic_link) {
	src = xasprintf (" Link %s to:", name_trunc (Fname, 46));
	dest = input_expand_dialog (" Link ", src, "");
	free (src);
	if (!dest)
	    return;
	if (!*dest) {
	    free (dest);
	    return;
	}
	save_cwds_stat ();
	if (-1 == mc_link (Fname, dest))
	    MC_message (1, MSG_ERROR, " link: %m ");
    } else {
#ifdef OLD_SYMLINK_VERSION
	symlink_dialog (Fname, "", &dest, &src);
#else
	/* suggest the full path for symlink */
	char s[MC_MAXPATHLEN];
	char d[MC_MAXPATHLEN];

	strcpy(s, cpanel->cwd);
	if ( ! ((s[0] == '/') && (s[1] == 0)))
	    strcat(s, "/");
	strcat(s, Fname);
	if (get_other_type () == view_listing)
	    strcpy(d, opanel->cwd);
	else
	    strcpy (d, "");

	if ( ! ((d[0] == '/') && (d[1] == 0)))
	    strcat(d, "/");
	symlink_dialog (s, d, &dest, &src);
#endif
	if (!dest || !*dest) {
	    if (src)
		free (src);
	    if (dest)
		free (dest);
	    return;
	}
	if (src) {
	    if (*src) {
		save_cwds_stat ();
		if (-1 == mc_symlink (dest, src))
		    MC_message (1, MSG_ERROR, " symlink: %m ");
	    }
	    free (src);
	}
    }
    free (dest);
    update_panels (UP_OPTIMIZE, UP_KEEPSEL);
    repaint_screen ();
}

static void link_cmd (void)
{
    do_link (0, selection (cpanel)->fname);
}

static void symlink_cmd (void)
{
    do_link (1, selection (cpanel)->fname);
}

static void edit_symlink_cmd (void)
{
    if (S_ISLNK (selection (cpanel)->buf.st_mode)) {
	char buffer [MC_MAXPATHLEN], *p = selection (cpanel)->fname;
	int i;
	char *dest, *q = xasprintf (" Symlink %s points to:",
							name_trunc (p, 32));

	i = readlink (p, buffer, MC_MAXPATHLEN);
	if (i > 0) {
	    buffer [i] = 0;
	    dest = input_expand_dialog (" Edit symlink ", q, buffer);
	    if (dest) {
		if (*dest && strcmp (buffer, dest)) {
		    save_cwds_stat ();
		    mc_unlink (p);
		    if (-1 == mc_symlink (dest, p))
			MC_message (1, MSG_ERROR, " edit symlink: %m ");
		    update_panels (UP_OPTIMIZE, UP_KEEPSEL);
		    repaint_screen ();
		}
		free (dest);
	    }
	}
	free (q);
    }
}

static void other_symlink_cmd (void)
{
    char *dest, *q, *p, *r, *s, *t;

    if (get_other_type () != view_listing)
	return;

    if (!strcmp (selection (opanel)->fname, ".."))
	return;
    p = concat_path_file (cpanel->cwd, selection (cpanel)->fname);
    r = concat_path_file (opanel->cwd, selection (cpanel)->fname);

    q = xasprintf (" Link symbolically %s to:", name_trunc (p, 32));
    dest = input_expand_dialog (" Relative symlink ", q, r);
    if (dest) {
	if (*dest) {
	    t = strrchr (dest, '/');
	    if (t) {
		t[1] = 0;
		s = diff_two_paths (dest, p);
		t[1] = '/';
		if (s) {
		    save_cwds_stat ();
		    if (-1 == mc_symlink (dest, s))
			MC_message (1, MSG_ERROR, " relative symlink: %m ");
		    update_panels (UP_OPTIMIZE, UP_KEEPSEL);
		    repaint_screen ();
		    free (s);
		}
	    }
	}
	free (dest);
    }
    free (q);
    free (p);
    free (r);
}

static void help_cmd (void)
{
   char *hlpfile = concat_mcpath_file ("mc.hlp");
   interactive_display (hlpfile, "[main]");
   free (hlpfile);
}

static void view_panel_cmd (void)
{
    view_cmd (cpanel);
}

static void edit_panel_cmd (void)
{
    edit_cmd (cpanel);
}

static void mkdir_panel_cmd (void)
{
    mkdir_cmd (cpanel);
}

/* Returns a random hint */
static char *get_random_hint (void)
{
    char *data, *result;
    char *hintfile;
    int  len;
    int start;

    /* Do not change hints more often than one minute */

    struct timeval tv;

    gettimeofday (&tv, NULL);
    if (!(tv.tv_sec> last_sec+60))
	return strdup ("");
    last_sec = tv.tv_sec;

    hintfile = concat_mcpath_file (MC_HINT);
    data = load_file (hintfile);
    free (hintfile);
    if (!data)
	return 0;

    srand (tv.tv_sec);
    /* get a random entry */
    len = strlen (data);
    start = rand () % len;

    for (; start; start--) {
	if (data [start] == '\n') {
	    start++;
	    break;
	}
    }
    strchr_set0 (&data [start], '\n');
    result = strdup (&data [start]);
    free (data);
    return result;
}

#ifdef USE_NETCODE

static char *machine_str = " Enter machine name (F1 for details): ";

static void nice_cd (char *text, char *xtext, char *help, char *prefix, int to_home)
{
    char *machine;
    char *cd_path;

    if (!SELECTED_IS_PANEL)
	return;

    machine = input_dialog_help (text,
				 xtext,
				 help, "");
    if (!machine)
	return;

    if (strncmp (prefix, machine, strlen (prefix)) == 0)
	cd_path = xasprintf ("%s%s", machine, to_home ? "/~/" : "");
    else
	cd_path = xasprintf ("%s%s%s", prefix, machine, to_home ? "/~/" : "");

    if (do_panel_cd (MENU_PANEL, cd_path, 0))
	directory_history_add (MENU_PANEL, (MENU_PANEL)->cwd);
    else
	MC_message (1, MSG_ERROR, " Could not chdir to %s ", cd_path);
    free (cd_path);
    free (machine);
}

static void netlink_cmd (void)
{
    nice_cd (" Link to a remote machine ", machine_str,
	     "[Network File System]", "mc:", 1);
}

static void ftplink_cmd (void)
{
    nice_cd (" FTP to machine ", machine_str,
	     "[FTP File System]", "ftp://", 1);
}

#endif /* USE_NETCODE */

static void quick_cd_cmd (void)
{
    char *p = cd_dialog ();

    if (p && *p) {
	char *q = xasprintf ("cd %s", p);

	do_cd_command (q);
	free (q);
    }
    if (p)
	free (p);
}

static void save_setup_cmd (void)
{
    char *str;

    save_setup ();
    sync_profiles ();
    str = xstrdup ( " Setup saved to ~/" PROFILE_NAME);

    MC_message (0, " Setup ", str);
    free (str);
}

static void
configure_panel_listing (WPanel *p, int view_type, int use_msformat, char *User, char *Status)
{
    int err;

    p->user_mini_status = use_msformat;
    p->list_type = view_type;

    if (view_type == list_user || use_msformat) {
	free (p->user_format);
	p->user_format = User;

	free (p->user_status_format [view_type]);
	p->user_status_format [view_type] = Status;

	err = set_panel_formats (p);

	if (err) {
	    if (err & 0x01) {
		free (p->user_format);
		p->user_format  = strdup (DEFAULT_USER_FORMAT);
	    }

	    if (err & 0x02) {
		free (p->user_status_format [view_type]);
		p->user_status_format [view_type]  = strdup (DEFAULT_USER_FORMAT);
	    }
	}
    }
    else {
	free (User);
	free (Status);
    }

    set_panel_formats (p);
    paint_panel (p);

    do_refresh ();
}

static void
info_cmd_no_menu (void)
{
    set_display_type (cpanel == left_panel ? 1 : 0, view_info);
}

static void
quick_cmd_no_menu (void)
{
    set_display_type (cpanel == left_panel ? 1 : 0, view_quick);
}

static void
switch_to_listing (int panel_index)
{
    if (get_display_type (panel_index) != view_listing)
	set_display_type (panel_index, view_listing);
}

static void
listing_cmd (void)
{
    int   view_type, use_msformat;
    char  *User, *Status;
    WPanel *p;
    int   display_type;

    display_type = get_display_type (MENU_PANEL_IDX);
    if (display_type == view_listing)
	p = MENU_PANEL_IDX == 0 ? left_panel : right_panel;
    else
	p = 0;

    view_type = display_box (p, &User, &Status, &use_msformat, MENU_PANEL_IDX);

    if (view_type == -1)
	return;

    switch_to_listing (MENU_PANEL_IDX);

    p = MENU_PANEL_IDX == 0 ? left_panel : right_panel;

    configure_panel_listing (p, view_type, use_msformat, User, Status);
}

static void
tree_cmd (void)
{
    set_display_type (MENU_PANEL_IDX, view_tree);
}

static void
info_cmd (void)
{
    set_display_type (MENU_PANEL_IDX, view_info);
}

static void
quick_view_cmd (void)
{
    set_display_type (MENU_PANEL_IDX, view_quick);
}

/* Handle the tree internal listing modes switching */
static int
set_basic_panel_listing_to (int panel_index, int listing_mode)
{
    WPanel *p = (WPanel *) get_panel_widget (panel_index);

    switch_to_listing (panel_index);
    p->list_type = listing_mode;
    if (set_panel_formats (p))
	return 0;

    paint_panel (p);
    do_refresh ();
    return 1;
}

static void
toggle_listing_cmd (void)
{
    int current = get_current_index ();
    WPanel *p = (WPanel *) get_panel_widget (current);
    int list_mode = p->list_type;
    int m;

    switch (list_mode) {
    case list_full:
    case list_brief:
	m = list_long;
	break;
    case list_long:
	m = list_user;
	break;
    default:
	m = list_full;
    }
    if (set_basic_panel_listing_to (current, m))
	return;
    set_basic_panel_listing_to (current, list_full);
}

/* Color setup */



#define color_map_fg(n) color_map[n].fg
#define color_map_bg(n) color_map[n].bg


struct color_table_s {
    const char *name;
    int  value;
};

static const struct color_table_s color_table [] = {
    { "black",         COLOR_BLACK   },
    { "gray",          COLOR_BLACK   | A_BOLD },
    { "red",           COLOR_RED     },
    { "brightred",     COLOR_RED     | A_BOLD },
    { "green",         COLOR_GREEN   },
    { "brightgreen",   COLOR_GREEN   | A_BOLD },
    { "brown",         COLOR_YELLOW  },
    { "yellow",        COLOR_YELLOW  | A_BOLD },
    { "blue",          COLOR_BLUE    },
    { "brightblue",    COLOR_BLUE    | A_BOLD },
    { "magenta",       COLOR_MAGENTA },
    { "brightmagenta", COLOR_MAGENTA | A_BOLD },
    { "cyan",          COLOR_CYAN    },
    { "brightcyan",    COLOR_CYAN    | A_BOLD },
    { "lightgray",     COLOR_WHITE },
    { "white",         COLOR_WHITE   | A_BOLD }
};

#define color_value(i) color_table [i].name
#define color_name(i)  color_table [i].name

static const char * get_color (char *cpp)
{
    int i;

    for (i = 0; i < ELEMENTS(color_table); i++) {
	if (strcmp (cpp, color_name (i)) == 0) {
	    return color_value (i);
	}
    }
    return color_value (0);
}

static void get_two_colors (char **cpp, struct colorpair *colorpairp)
{
    char *p = *cpp;
    int state;
    const char *c;

    state = 0;

    for (; *p; p++) {
	if (*p == ':') {
	    *p = 0;
	    c = get_color (*cpp);
	    if(state)
		colorpairp->bg = c;
	    else
		colorpairp->fg = c;
	    *p = ':';
	    *cpp = p + 1;
	    return;
	}

	if (*p == ',') {
	    state = 1;
	    *p = 0;
	    colorpairp->fg = get_color (*cpp);
	    *p = ',';
	    *cpp = p + 1;
	}
    }
    c = get_color (*cpp);
    if(state)
	colorpairp->bg = c;
    else
	colorpairp->fg = c;
}

static void configure_colors_string (const char *the_color_string)
{
    char *color_string, *p;
    int  i, found;

    if (!the_color_string)
	return;

    p = color_string = strdup (the_color_string);
    while (color_string && *color_string) {
	while (*color_string == ' ' || *color_string == '\t')
	    color_string++;

	found = 0;
	for (i = 0; i < COLOR_MAP_SIZE; i++) {
	    int klen = strlen (color_map [i].name);

	    if (strncmp (color_string, color_map [i].name, klen) == 0) {
		color_string += klen;
		get_two_colors (&color_string, &color_map [i]);
		found = 1;
	    }
	}
	if (!found) {
		while (*color_string && *color_string != ':')
			color_string++;
		if (*color_string)
			color_string++;
	}
    }
    free (p);
}

/*
 * Text edition support code
 *
 *
 */
static const char default_edition_colors[] =
"normal=lightgray,blue:"
"selected=black,cyan:"
"marked=yellow,blue:"
"markselect=yellow,cyan:"
"errors=white,red:"
"menu=white,cyan:"
"reverse=black,lightgray:"
"dnormal=black,lightgray:"
"dfocus=black,cyan:"
"dhotnormal=yellow,lightgray:"
"dhotfocus=yellow,cyan:"
"viewunderline=brightred,blue:"
"menuhot=yellow,cyan:"
"menusel=white,black:"
"menuhotsel=yellow,black:"
"helpnormal=black,lightgray:"
"helpitalic=red,lightgray:"
"helpbold=blue,lightgray:"
"helplink=black,cyan:"
"helpslink=yellow,blue:"
"gauge=white,black:"
"input=black,cyan:"
"directory=white,blue:"
"execute=brightgreen,blue:"
"link=lightgray,blue:"
"device=brightmagenta,blue:"
"core=red,blue:"
"special=black,blue";


static void configure_colors (void)
{
    configure_colors_string (default_edition_colors);
    configure_colors_string (setup_color_string);
    configure_colors_string (term_color_string);
    configure_colors_string (getenv ("MC_COLOR_TABLE"));
    configure_colors_string (command_line_colors);
}


static void
load_dialog_colors (void)
{
    dialog_colors [0] = COLOR_NORMAL;
    dialog_colors [1] = COLOR_FOCUS;
    dialog_colors [2] = COLOR_HOT_NORMAL;
    dialog_colors [3] = COLOR_HOT_FOCUS;
}

static void init_colors (void)
{
    int i;

    color_map[ 0].name = "normal=";     /* normal */
    color_map[ 1].name = "selected=";   /* selected */
    color_map[ 2].name = "marked=";     /* marked */
    color_map[ 3].name = "markselect="; /* marked/selected */
    color_map[ 4].name = "errors=";     /* errors */
    color_map[ 5].name = "menu=";       /* menu entry */
    color_map[ 6].name = "reverse=";    /* reverse */
    /* Dialog colors */
    color_map[ 7].name = "dnormal=";    /* Dialog normal */
    color_map[ 8].name = "dfocus=";     /* Dialog focused */
    color_map[ 9].name = "dhotnormal="; /* Dialog normal/hot */
    color_map[10].name = "dhotfocus=";  /* Dialog focused/hot */

    color_map[11].name = "viewunderline="; /* _\b? sequence in view */
    color_map[12].name = "menusel=";    /* Menu selected color */
    color_map[13].name = "menuhot=";    /* Color for menu hotkeys */
    color_map[14].name = "menuhotsel="; /* Menu hotkeys/selected entry */

    color_map[15].name = "helpnormal="; /* Help normal */
    color_map[16].name = "helpitalic="; /* Italic in help */
    color_map[17].name = "helpbold=";   /* Bold in help */
    color_map[18].name = "helplink=";   /* Not selected hyperlink */
    color_map[19].name = "helpslink=";  /* Selected hyperlink */

    color_map[20].name = "gauge=";      /* Color of the progress bar (percentage) */
    color_map[21].name = "input=";
    /* Per file types colors */
    color_map[22].name = "directory=";
    color_map[23].name = "execute=";
    color_map[24].name = "link=";
    color_map[25].name = "device=";
    color_map[26].name = "special=";
    color_map[27].name = "core=";

    hascolors = has_colors ();

    if (!disable_colors && hascolors)
	use_colors = 1;

    if (use_colors) {
	configure_colors ();

	if (use_colors) { /* Hack to make COLOR_PAIR(31) be the default fg/bg
			     of the terminal */
	    const char *Norm_Vid = SLtt_tgetstr ("me");

	    if (Norm_Vid == NULL)
		Norm_Vid = SLtt_tgetstr ("se");
	    if (Norm_Vid == NULL)
		Norm_Vid = "\033[0m";
	    SLtt_set_color_esc (31, Norm_Vid);
	}

	for (i = 0; i < COLOR_MAP_SIZE; i++)
	    SLtt_set_color (i+1, color_map_fg(i), color_map_bg(i));
    }
    load_dialog_colors ();
}

/* Command line widget. */


/* Tries variable substitution, and if a variable CDPATH
of the form e.g. CDPATH=".:~:/usr" exists, we try then all the paths which
are mentioned there. Also, we do not support such extraordinary things as
${var:-value}, etc. Use the eval cd 'path' command instead.
Bugs: No quoting occurrs here, so ${VAR} and $VAR will be always
substituted. I think we can encourage users to use in such extreme
cases instead of >cd path< command a >eval cd 'path'< command, which works
as they might expect :)
FIXME: Perhaps we should do wildcard matching as well? */
static int examine_cd (char *path)
{
    char *p;
    int result;
    char *q = x_malloc (MC_MAXPATHLEN + 10, "examine_cd"), *r, *s, *t, c;

    /* Variable expansion */
    for (p = path, r = q; *p && r < q + MC_MAXPATHLEN; ) {
	if (*p != '$' || (p [1] == '[' || p [1] == '('))
	    *(r++)=*(p++);
	else {
	    p++;
	    if (*p == '{') {
		p++;
		s = strchr (p, '}');
	    } else
		s = NULL;
	    if (s == NULL)
		s = strchr (p, '/');
	    if (s == NULL)
		s = strchr (p, 0);
	    c = *s;
	    *s = 0;
	    t = getenv (p);
	    *s = c;
	    if (t == NULL) {
		*(r++) = '$';
		if (*(p - 1) != '$')
		    *(r++) = '{';
	    } else {
		if (r + strlen (t) < q + MC_MAXPATHLEN) {
		    strcpy (r, t);
		    r = strchr (r, 0);
		}
		if (*s == '}')
		    p = s + 1;
		else
		    p = s;
	    }
	}
    }
    *r = 0;

    result = do_cd (q, cd_parse_command);

    /* CDPATH handling */
    if (*q != '/' && !result) {
	p = getenv ("CDPATH");
	if (p == NULL)
	    c = 0;
	else
	    c = ':';
	while (!result && c == ':') {
	    s = strchr (p, ':');
	    if (s == NULL)
		s = strchr (p, 0);
	    c = *s;
	    *s = 0;
	    if (*p) {
		r = concat_path_file (p, q);
		result = do_cd (r, cd_parse_command);
		free (r);
	    }
	    *s = c;
	    p = s + 1;
	}
    }
    free (q);
    return result;
}

/* Execute the cd command on the command line */
void do_cd_command (char *cmd)
{
    int len;

    /* Any final whitespace should be removed here
       (to see why, try "cd fred "). */
    /* NOTE: I think we should not remove the extra space,
       that way, we can cd into hidden directories */
    len = strlen (cmd) - 1;
    while (len >= 0 &&
	   (cmd [len] == ' ' || cmd [len] == '\t' || cmd [len] == '\n')) {
	cmd [len] = 0;
	len --;
    }

    if (cmd [2] == 0)
	cmd = "cd ";

    if (get_current_type () == view_tree) {
	if (cmd [0] == 0) {
	    tree_chdir (the_tree, home_dir);
	} else if (strcmp (cmd+3, "..") == 0) {
	    char *dir = cpanel->cwd;

	    len = strlen (dir);
	    while (len && dir [--len] != '/');
	    dir [len] = 0;
	    if (len)
		tree_chdir (the_tree, dir);
	    else
		tree_chdir (the_tree, "/");
	} else if (cmd [3] == '/') {
	    tree_chdir (the_tree, cmd+3);
	} else {
	    char *old = cpanel->cwd;
	    char *new;
	    new = concat_path_file (old, cmd+3);
	    tree_chdir (the_tree, new);
	    free (new);
	}
    } else
	if (!examine_cd (&cmd [3])) {
	    MC_message (1, MSG_ERROR, " Can't chdir to '%s' \n %m ",
		     &cmd [3]);
	    return;
	}
}

/* Returns 1 if the we could handle the enter, 0 if not */
static int enter (WCommand *Cmdline)
{
    Dlg_head *old_dlg;

    if (command_prompt && strlen (input_w (Cmdline)->buffer)) {
	char *cmd;

	/* Any initial whitespace should be removed at this point */
	cmd = input_w (Cmdline)->buffer;
	while (*cmd == ' ' || *cmd == '\t' || *cmd == '\n')
	    cmd++;

	if (strncmp (cmd, "cd ", 3) == 0 || strcmp (cmd, "cd") == 0) {
	    do_cd_command (cmd);
	    new_input (input_w (Cmdline));
	    return MSG_HANDLED;
	} else {
	    char *command, *s;
	    int i, j;

	    if (!vfs_current_is_local ())
		return MSG_NOT_HANDLED;
	    command = zmalloc (strlen (cmd) + 1);
	    for (i = j = 0; i < strlen (cmd); i ++) {
		if (cmd [i] == '%') {
		    i++;
		    s = expand_format (cmd [i], 1);
		    command = realloc (command, strlen (command) + strlen (s)
				       + strlen (cmd) - i + 1);
		    strcat (command, s);
		    free (s);
		    j = strlen (command);
		} else {
		    command [j] = cmd [i];
		    j ++;
		}
		command [j] = 0;
	    }
	    old_dlg = current_dlg;
	    current_dlg = 0;
	    new_input (input_w (Cmdline));
	    execute (command);
	    free (command);

	    current_dlg = old_dlg;
	}
    }
    return MSG_HANDLED;
}

static int command_callback (Dlg_head *h, WCommand *cmd, int msg, int par)
{
    switch (msg) {
    case WIDGET_FOCUS:
	/* We refuse the focus always: needed not to unselect the panel */
	return MSG_NOT_HANDLED;

    case WIDGET_KEY:
	/* Special case: we handle the enter key */
	if (par == '\n') {
	    return enter (cmd);
	}
    }
    return (*cmd->old_callback)(h, cmd, msg, par);
}

WCommand *command_new (int y, int x, int cols)
{
    WInput *in;
    WCommand *cmd = x_malloc (sizeof (WCommand), "command_new");

    in = input_new (y, x, DEFAULT_COLOR, cols, "");
    cmd->input = *in;
    free (in);

    /* Add our hooks */
    cmd->old_callback = (callback_fn) cmd->input.widget.callback;
    cmd->input.widget.callback = (int (*) (Dlg_head *, void *, int, int))
			command_callback;

    cmd->input.completion_flags |= INPUT_COMPLETE_COMMANDS;
    return cmd;
}


/* Input line filename/username/hostname/variable/command completion.
   (Let mc type for you...) */

static char *filename_completion_function (const char *text, int state)
{
    int isdir = 1, isexec = 0;
    struct dirent *entry = NULL;

    /* If we're starting the match process, initialize us a bit. */
    if (!state) {
	char *temp;

	free (fc_Dirname);
	free (fc_filename);
	free (users_dirname);

	fc_filename = strdup (text);
	if (!*text)
	    text = ".";
	fc_Dirname = strdup (text);

	temp = strrchr (fc_Dirname, '/');

	if (temp) {
	    strcpy (fc_filename, ++temp);
	    *temp = 0;
	}
	else
	    strcpy (fc_Dirname, ".");

	/* We aren't done yet.  We also support the "~user" syntax. */

	/* Save the version of the directory that the user typed. */
	users_dirname = strdup (fc_Dirname);
	{
	    char *temp_dirname;

	    temp_dirname = tilde_expand (fc_Dirname);
	    if (!temp_dirname) {
		free (fc_Dirname);
		free (users_dirname);
		free (fc_filename);
		fc_Dirname = users_dirname = fc_filename = NULL;
		return NULL;
	    }
	    free (fc_Dirname);
	    fc_Dirname = temp_dirname;
	    canonicalize_pathname (fc_Dirname);
	    /* Here we should do something with variable expansion
	       and `command`.
	       Maybe a dream - UNIMPLEMENTED yet. */
	}
	fcf_directory = opendir (fc_Dirname);
	filename_len = strlen (fc_filename);
    }

    /* Now that we have some state, we can read the directory. */

    while (fcf_directory && (entry = readdir (fcf_directory))) {
	/* Special case for no fc_filename.
	   All entries except "." and ".." match. */
	if (!filename_len) {
	    if (!strcmp (entry->d_name, ".") || !strcmp (entry->d_name, ".."))
		continue;
	} else {
	    /* Otherwise, if these match up to the length of fc_filename, then
	       it may be a match. */
	    if ((entry->d_name[0] != fc_filename[0]) ||
		((NLENGTH (entry)) < filename_len) ||
		strncmp (fc_filename, entry->d_name, filename_len))
		continue;
	}
	isdir = 1; isexec = 0;
	{
	    char *tmp;
	    struct stat tempstat;

	    tmp = concat_path_file(fc_Dirname, entry->d_name);
	    canonicalize_pathname (tmp);
	    /* Unix version */
	    if (!is_directory (tmp, 1, &tempstat)) {
		    isdir = 0;
		    if ((!my_uid && (tempstat.st_mode & 0111)) ||
			(my_uid == tempstat.st_uid && (tempstat.st_mode & 0100)) ||
			(my_gid == tempstat.st_gid && (tempstat.st_mode & 0010)) ||
			(tempstat.st_mode & 0001))
			isexec = 1;
	    }
	    free (tmp);
	}
	switch (look_for_executables)
	{
	    case 2: if (!isexec)
			continue;
		    break;
	    case 1: if (!isexec && !isdir)
			continue;
		    break;
	}
	if (ignore_filenames && !isdir)
	    continue;
	break;
    }

    if (!entry) {
	if (fcf_directory) {
	    closedir (fcf_directory);
	    fcf_directory = NULL;
	}
	free (fc_Dirname);
	fc_Dirname = NULL;
	free (fc_filename);
	fc_filename = NULL;
	free (users_dirname);
	users_dirname = NULL;
	return NULL;
    } else {
	char *temp;

	if (users_dirname && (users_dirname[0] != '.' || users_dirname[1])) {
	    int dirlen = strlen (users_dirname);
	    temp = x_malloc (3 + dirlen + NLENGTH (entry), "Filename completion");
	    strcpy (temp, users_dirname);
	    /* We need a `/' at the end. */
	    if (users_dirname[dirlen - 1] != '/') {
		temp[dirlen] = '/';
		temp[dirlen + 1] = 0;
	    }
	    strcat (temp, entry->d_name);
	} else {
	    temp = x_malloc (2 + NLENGTH (entry), "Filename completion");
	    strcpy (temp, entry->d_name);
	}
	if (isdir)
	    strcat (temp, "/");
	return temp;
    }
}

/* We assume here that text[0] == '~' , if you want to call it in another way,
   you have to change the code */
static char *username_completion_function (const char *text, int state)
{
    struct passwd *entry;

    if (!state) {       /* Initialization stuff */
	setpwent ();
	userlen = strlen (text + 1);
    }
    while ((entry = getpwent ()) != NULL) {
	/* Null usernames should result in all users as possible completions. */
	if (!userlen)
	    break;
	else if (text[1] == entry->pw_name[0] &&
		 !strncmp (text + 1, entry->pw_name, userlen))
	    break;
    }

    if (!entry) {
	endpwent ();
	return NULL;
    } else {
	return xasprintf("~%s/", entry->pw_name);
    }
}

extern char **environ;

/* We assume text [0] == '$' and want to have a look at text [1], if it is
   equal to '{', so that we should append '}' at the end */
static char *variable_completion_function (const char *text, int state)
{
    char *p = 0;

    if (!state) {                       /* Initialization stuff */
	isbrace = (text [1] == '{');
	varlen = strlen (text + 1 + isbrace);
	env_p = environ;
    }

    while (*env_p) {
	p = strchr (*env_p, '=');
	if (p && p - *env_p >= varlen && !strncmp (text + 1 + isbrace, *env_p, varlen))
	    break;
	env_p++;
    }

    if (!*env_p)
	return NULL;
    else {
	char *temp = x_malloc (2 + 2 * isbrace + p - *env_p, "Variable completion");

	*temp = '$';
	if (isbrace)
	    temp [1] = '{';
	strncpy (temp + 1 + isbrace, *env_p, p - *env_p);
	if (isbrace)
	    strcpy (temp + 2 + (p - *env_p), "}");
	else
	    temp [1 + p - *env_p] = 0;
	env_p++;
	return temp;
    }
}

#define whitespace(c) ((c) == ' ' || (c) == '\t')
#define cr_whitespace(c) (whitespace (c) || (c) == '\n' || (c) == '\r')

static void fetch_hosts (const char *filename)
{
    FILE *file = fopen (filename, "r");
    char *temp, buffer[256], *name;
    int i, start;

    if (!file)
	return;

    while ((temp = fgets (buffer, 255, file)) != NULL) {
	/* Skip to first character. */
	for (i = 0; buffer[i] && cr_whitespace (buffer[i]); i++);
	/* Ignore comments... */
	if (buffer[i] == '#')
	    continue;
	/* Handle $include. */
	if (!strncmp (buffer + i, "$include ", 9)) {
	    char *includefile = buffer + i + 9;
	    char *t;

	    /* Find start of filename. */
	    while (*includefile && whitespace (*includefile))
		includefile++;
	    t = includefile;

	    /* Find end of filename. */
	    while (*t && !cr_whitespace (*t))
		t++;
	    *t = '\0';

	    fetch_hosts (includefile);
	    continue;
	}

	/* Skip IP #s. */
	for (; buffer[i] && !cr_whitespace (buffer[i]); i++);

	/* Get the host names separated by white space. */
	while (buffer[i] && buffer[i] != '#') {
	    for (; i && cr_whitespace (buffer[i]); i++);
	    if (buffer[i] ==  '#')
		continue;
	    for (start = i; buffer[i] && !cr_whitespace (buffer[i]); i++);
	    if ((i - start) == 0)
		continue;
	    name = (char *) x_malloc (i - start + 1, "Hostname completion");
	    strncpy (name, buffer + start, i - start);
	    name [i - start] = 0;
	    {
		char **host_p;

		if (hosts_p - hosts >= hosts_alloclen) {
		    int j = hosts_p - hosts;

		    hosts = realloc ((void *)hosts, ((hosts_alloclen += 30) + 1) * sizeof (char *));
		    hosts_p = hosts + j;
		}
		for (host_p = hosts; host_p < hosts_p; host_p++)
		    if (!strcmp (name, *host_p))
			break; /* We do not want any duplicates */
		if (host_p == hosts_p) {
		    *(hosts_p++) = name;
		    *hosts_p = NULL;
		} else
		    free (name);
	    }
	}
    }
    fclose (file);
}

static char *hostname_completion_function (const char *text, int state)
{
    if (!state) {       /* Initialization stuff */
	char *p;

	if (hosts != NULL) {
	    for (hc_host_p = hosts; *hc_host_p; hc_host_p++)
		free (*hc_host_p);
	    free (hosts);
	}
	hosts = (char **) x_malloc (((hosts_alloclen = 30) + 1) * sizeof (char *), "Hostname completion");
	*hosts = NULL;
	hosts_p = hosts;
	fetch_hosts ((p = getenv ("HOSTFILE")) ? p : "/etc/hosts");
	hc_host_p = hosts;
	textstart = (*text == '@') ? 1 : 0;
	textlen = strlen (text + textstart);
    }

    while (*hc_host_p) {
	if (!textlen)
	    break; /* Match all of them */
	else if (!strncmp (text + textstart, *hc_host_p, textlen))
	    break;
	hc_host_p++;
    }

    if (!*hc_host_p) {
	for (hc_host_p = hosts; *hc_host_p; hc_host_p++)
	    free (*hc_host_p);
	free (hosts);
	hosts = NULL;
	return NULL;
    } else {
	char *temp = x_malloc (2 + strlen (*hc_host_p), "Hostname completion");

	if (textstart)
	    *temp = '@';
	strcpy (temp + textstart, *hc_host_p);
	hc_host_p++;
	return temp;
    }
}

/* This is the function to call when the word to complete is in a position
   where a command word can be found. It looks around $PATH, looking for
   commands that match. It also scans aliases, function names, and the
   table of shell built-ins. */
static char *command_completion_function (const char *text, int state)
{
    static const char * const bash_reserved [] = { "if", "then", "else", "elif", "fi",
				      "case", "esac", "for", "select", "while",
				      "until", "do", "done", "in", "function" , 0};
    static const char * const bash_builtins [] = { "alias", "bg", "bind", "break", "builtin",
				      "cd", "command", "continue", "declare",
				      "dirs", "echo", "enable", "eval", "exec",
				      "exit", "export", "fc", "fg", "getopts",
				      "hash", "help", "history", "jobs", "kill",
				      "let", "local", "logout", "popd", "pushd",
				      "pwd", "read", "readonly", "return", "set",
				      "shift", "source", "suspend", "test",
				      "times", "trap", "type", "typeset",
				      "ulimit", "umask", "unalias", "unset",
				      "wait" , 0};
    char *p, *found;

    if (!state) {       /* Initialize us a little bit */
	isabsolute = strchr (text, '/') != 0;
	look_for_executables = isabsolute ? 1 : 2;
	if (!isabsolute) {
	    words = bash_reserved;
	    phase = 0;
	    text_len = strlen (text);
	    p = getenv ("PATH");
	    if (!p)
		ccf_path = NULL;
	    else {
		ccf_path = x_malloc (strlen (p) + 2, "Command completion");
		strcpy (ccf_path, p);
		p = strchr_set0 (ccf_path, ':');
		while (p)
		    p = strchr_set0 (p + 1, ':');
	    }
	}
    }

    if (isabsolute) {
	p = filename_completion_function (text, state);
	if (!p)
	    look_for_executables = 0;
	return p;
    }

    found = NULL;
    switch (phase) {
	case 0: /* Reserved words */
	    while (*words) {
		if (!strncmp (*words, text, text_len))
		    return strdup (*(words++));
		words++;
	    }
	    phase++;
	    words = bash_builtins;
	case 1: /* Builtin commands */
	    while (*words) {
		if (!strncmp (*words, text, text_len))
		    return strdup (*(words++));
		words++;
	    }
	    phase++;
	    if (!ccf_path)
		break;
	    cur_path = ccf_path;
	    cur_word = NULL;
	case 2: /* And looking through the $PATH */
	    while (!found) {
		if (!cur_word) {
		    char *expanded;

		    if (!*cur_path)
			break;
		    expanded = tilde_expand (cur_path);
		    if (!expanded) {
			free (ccf_path);
			ccf_path = NULL;
			return NULL;
		    }
		    p = canonicalize_pathname (expanded);
		    cur_word = x_malloc (strlen (p) + 2 + text_len, "Command completion");
		    strcpy (cur_word, p);
		    if (cur_word [strlen (cur_word) - 1] != '/')
			strcat (cur_word, "/");
		    strcat (cur_word, text);
		    free (p);
		    cur_path = strchr (cur_path, 0) + 1;
		    init_state = state;
		}
		found = filename_completion_function (cur_word, state - init_state);
		if (!found) {
		    free (cur_word);
		    cur_word = NULL;
		}
	    }
    }

    if (!found) {
	look_for_executables = 0;
	if (ccf_path)
	    free (ccf_path);
	return NULL;
    }
    if ((p = (char *)bb_basename (found)) != found) {
	p = strdup (p);
	free (found);
	return p;
    }
    return found;
}

static int match_compare (const void *a, const void *b)
{
    return strcmp (*(char **)a, *(char **)b);
}

/* Returns an array of char * matches with the longest common denominator
   in the 1st entry. Then a NULL terminated list of different possible
   completions follows.
   You have to supply your own CompletionFunction with the word you
   want to complete as the first argument and an count of previous matches
   as the second.
   In case no matches were found we return NULL. */
static char **completion_matches (char *text, CompletionFunction entry_function)
{
    /* Number of slots in match_list. */
    int match_list_size;

    /* The list of matches. */
    char **match_list = x_malloc (((match_list_size = 30) + 1) * sizeof (char *), "completion match list");

    /* Number of matches actually found. */
    int matches = 0;

    /* Temporary string binder. */
    char *string;

    match_list[1] = NULL;

    while ((string = (*entry_function) (text, matches)) != NULL) {
	if (matches + 1 == match_list_size)
	    match_list = realloc (match_list, ((match_list_size += 30) + 1) * sizeof (char *));
	match_list[++matches] = string;
	match_list[matches + 1] = NULL;
    }

    /* If there were any matches, then look through them finding out the
       lowest common denominator.  That then becomes match_list[0]. */
    if (matches) {
	int i = 1;
	int low = 4096;         /* Count of max-matched characters. */

	/* If only one match, just use that. */
	if (matches == 1) {
	    match_list[0] = match_list[1];
	    match_list[1] = (char *)NULL;
	} else {
	    int j;

	    qsort (match_list + 1, matches, sizeof (char *), match_compare);

	    /* And compare each member of the list with
	       the next, finding out where they stop matching.
	       If we find two equal strings, we have to put one away... */

	    j = i + 1;
	    while (j < matches + 1)
	    {
		int c1, c2, si;

		for (si = 0;(c1 = match_list [i][si]) && (c2 = match_list [j][si]); si++)
		    if (c1 != c2) break;

		if (!c1 && !match_list [j][si]) { /* Two equal strings */
		    free (match_list [j]);
		    j++;
		    if (j > matches)
			break;
		} else
		    if (low > si) low = si;
		if (i + 1 != j) /* So there's some gap */
		    match_list [i + 1] = match_list [j];
		i++; j++;
	    }
	    matches = i;
	    match_list [matches + 1] = NULL;
	    match_list[0] = x_malloc (low + 1, "Completion matching list");
	    strncpy (match_list[0], match_list[1], low);
	    match_list[0][low] = 0;
	}
    } else {                            /* There were no matches. */
	free (match_list);
	match_list = NULL;
    }
    return match_list;
}

/* rewrote to parse "ls aa; cd a<TAB>" best now work in dir with files:
   a.out and aa/
   by vodz */

static const char command_separator_chars[] = ";|&{(`";

static int check_is_cd (const char *text, int start, int flags)
{
    const char *p = text;
    const char *q = text + start;

    const char *pcsc;

    if((flags & INPUT_COMPLETE_CD))
	return 1;
    for(pcsc=command_separator_chars; *pcsc; pcsc++) {
	const char *pp;

	pp = strrchr(p, *pcsc);
	if(pp!=0) {
		if(pp >= q) {
			pp = strchr(p, *pcsc);
			if(pp < q) {
				pcsc--;
				p = pp+1;
				}
			}
		 else
			p = pp+1;
		}
    }
    for (; *p && p < q && (*p == ' ' || *p == '\t'); p++);
    if ( (flags & INPUT_COMPLETE_COMMANDS) &&
	!strncmp (p, "cd", 2) && (p [2] == ' ' || p [2] == '\t') &&
	p + 2 < q )
	return 1;
    return 0;
}

/* Returns an array of matches, or NULL if none. */
static char **try_complete (char *text, int *start, int *end, int flags)
{
    int in_command_position = 0, i;
    char *word, c;
    char **matches = NULL;
    char *p = NULL, *q = NULL, *r = NULL;
    int is_cd = check_is_cd (text, *start, flags);

    ignore_filenames = 0;
    c = text [*end];
    text [*end] = 0;
    word = strdup (text + *start);
    text [*end] = c;

    /* Determine if this could be a command word. It is if it appears at
       the start of the line (ignoring preceding whitespace), or if it
       appears after a character that separates commands. And we have to
       be in a INPUT_COMPLETE_COMMANDS flagged Input line. */
    if (!is_cd && (flags & INPUT_COMPLETE_COMMANDS)) {
	i = *start - 1;
	while (i > -1 && (text[i] == ' ' || text[i] == '\t'))
	    i--;
	if (i < 0)
	    in_command_position++;
	else if (strchr (command_separator_chars, text[i])) {
	    int this_char, prev_char;

	    in_command_position++;

	    if (i) {
		/* Handle the two character tokens `>&', `<&', and `>|'.
		   We are not in a command position after one of these. */
		this_char = text[i];
		prev_char = text[i - 1];

		if ((this_char == '&' && (prev_char == '<' || prev_char == '>')) ||
		    (this_char == '|' && prev_char == '>'))
		    in_command_position = 0;
		else if (i > 0 && text [i-1] == '\\') /* Quoted */
		    in_command_position = 0;
	    }
	}
    }

    if (flags & INPUT_COMPLETE_COMMANDS)
	p = strrchr (word, '`');
    if (flags & (INPUT_COMPLETE_COMMANDS | INPUT_COMPLETE_VARIABLES))
	q = strrchr (word, '$');
    if (flags & INPUT_COMPLETE_HOSTNAMES)
	r = strrchr (word, '@');
    if (q && q [1] == '(' && INPUT_COMPLETE_COMMANDS) {
	if (q > p)
	    p = q + 1;
	q = NULL;
    }

    /* Command substitution? */
    if (p > q && p > r) {
	matches = completion_matches (p + 1, command_completion_function);
	if (matches)
	    *start += p + 1 - word;
    }

    /* Variable name? */
    else if  (q > p && q > r) {
	matches = completion_matches (q, variable_completion_function);
	if (matches)
	    *start += q - word;
    }

    /* Starts with '@', then look through the known hostnames for
       completion first. */
    else if (r > p && r > q) {
	matches = completion_matches (r, hostname_completion_function);
	if (matches)
	    *start += r - word;
    }

    /* Starts with `~' and there is no slash in the word, then
       try completing this word as a username. */
    if (!matches && *word == '~' && (flags & INPUT_COMPLETE_USERNAMES) && !strchr (word, '/'))
	matches = completion_matches (word, username_completion_function);


    /* And finally if this word is in a command position, then
       complete over possible command names, including aliases, functions,
       and command names. */
    if (!matches && in_command_position)
	matches = completion_matches (word, command_completion_function);

    else if (!matches && (flags & INPUT_COMPLETE_FILENAMES)) {
	if (is_cd)
	    ignore_filenames = 1;
	matches = completion_matches (word, filename_completion_function);
	ignore_filenames = 0;
	if (!matches && is_cd && *word != '/' && *word != '~') {
	    q = text + *start;

	    for (p = text; *p && p < q && (*p == ' ' || *p == '\t'); p++);
	    if (!strncmp (p, "cd", 2))
		for (p += 2; *p && p < q && (*p == ' ' || *p == '\t'); p++);
	    if (p == q) {
		char *cdpath = getenv ("CDPATH");
		char *s;

		if (cdpath == NULL)
		    c = 0;
		else
		    c = ':';
		while (!matches && c == ':') {
		    s = strchr (cdpath, ':');
		    if (s == NULL)
			s = strchr (cdpath, 0);
		    c = *s;
		    *s = 0;
		    if (*cdpath) {
			r = concat_path_file (cdpath, word);
			ignore_filenames = 1;
			matches = completion_matches (r, filename_completion_function);
			ignore_filenames = 0;
			free (r);
		    }
		    *s = c;
		    cdpath = s + 1;
		}
	    }
	}
    }

    if (word)
	free (word);

    return matches;
}

void free_completions (WInput *in)
{
    char **p;

    if (!in->completions)
	return;
    for (p=in->completions; *p; p++)
	free (*p);
    free (in->completions);
    in->completions = NULL;
}

static int insert_text (WInput *in, char *text, int len)
{
    len = min (len, strlen (text)) + text_start - text_end;
    if (strlen (in->buffer) + len >= in->current_max_len) {
    /* Expand the buffer */
	char *narea = realloc(in->buffer, in->current_max_len + len + in->field_len);
	if (narea) {
	    in->buffer = narea;
	    in->current_max_len += len + in->field_len;
	}
    }
    if (strlen (in->buffer)+1 < in->current_max_len) {
	if (len > 0) {
	    int i, l = strlen (&in->buffer [text_end]);
	    for (i = l + 1; i >= 0; i--)
		in->buffer [text_end + len + i] = in->buffer [text_end + i];
	} else if (len < 0) {
	    char *p = in->buffer + text_end + len, *q = in->buffer + text_end;
	    while (*q)
		*(p++) = *(q++);
	    *p = 0;
	}
	strncpy (in->buffer + text_start, text, len - text_start + text_end);
	in->point += len;
	update_input (in, 1);
	text_end += len;
    }
    return len != 0;
}

static int cmplt_query_callback (Dlg_head * h, int Par, int Msg)
{
    switch (Msg) {
	case DLG_DRAW:
	    attrset (COLOR_NORMAL);
	    dlg_erase (h);
	    draw_box (h, 0, 0, query_height, query_width);
	    break;

	case DLG_KEY:
	    switch (Par) {
		case KEY_LEFT:
		case KEY_RIGHT:
		    h->ret_value = 0;
		    dlg_stop (h);
		    return 1;

		case 0177:
		case KEY_BACKSPACE:
		case XCTRL('h'):
		    if (text_end == min_end) {
			h->ret_value = 0;
			dlg_stop (h);
			return 1;
		    } else {
			WLEntry *e, *e1;

			e1 = e = ((WListbox *)(h->current->widget))->list;
			do {
			    if (!strncmp (query_input->buffer + text_start, e1->text, text_end - text_start - 1)) {
				listbox_select_entry((WListbox *)(h->current->widget), e1);
				handle_char (query_input, Par);
				text_end--;
				send_message (h, h->current->widget,
				    WIDGET_DRAW, 0);
				break;
			    }
			    e1 = e1->next;
			} while (e != e1);
		    }
		    return 1;

		default:
		    if (Par > 0xff || !is_printable (Par)) {
			if (is_in_input_map (query_input, Par) == 2) {
			    if (text_end == min_end)
				return 1;
			    h->ret_value = B_USER; /* This means we want to refill the
						      list box and start again */
			    dlg_stop (h);
			    return 1;
			} else
			    return 0;
		    } else {
			WLEntry *e, *e1;
			int need_redraw = 0;
			int low = 4096;
			char *last_text = NULL;

			e1 = e = ((WListbox *)(h->current->widget))->list;
			do {
			    if (!strncmp (query_input->buffer + text_start, e1->text, text_end - text_start)) {
				if (e1->text [text_end - text_start] == Par) {
				    if (need_redraw) {
					int c1, c2, si;

					for (si = text_end - text_start + 1;
					     (c1 = last_text [si]) &&
					     (c2 = e1->text [si]); si++)
					    if (c1 != c2)
						break;
					if (low > si)
					    low = si;
					last_text = e1->text;
					need_redraw = 2;
				    } else {
					need_redraw = 1;
					listbox_select_entry((WListbox *)(h->current->widget), e1);
					last_text = e1->text;
				    }
				}
			    }
			    e1 = e1->next;
			} while (e != e1);
			if (need_redraw == 2) {
			    insert_text (query_input, last_text, low);
			    send_message (h, h->current->widget,WIDGET_DRAW,0);
			} else if (need_redraw == 1) {
			    h->ret_value = B_ENTER;
			    dlg_stop (h);
			}
		    }
		    return 1;
	    }
	    break;
    }
    return 0;
}

static int querylist_callback (void *data)
{
    return 1;
}

#define DO_INSERTION 1
#define DO_QUERY     2
/* Returns 1 if the user would like to see us again */
static int complete_engine (WInput *in, int what_to_do)
{
    if (in->completions && in->point != text_end)
	free_completions (in);
    if (!in->completions) {
	text_end = in->point;
	for (text_start = text_end ? text_end - 1 : 0; text_start > -1; text_start--)
	    if (strchr (" \t;|<>", in->buffer [text_start]))
		break;
	if (text_start < text_end)
	    text_start++;
	in->completions = try_complete (in->buffer, &text_start, &text_end, in->completion_flags);
    }
    if (in->completions) {
	if (what_to_do & DO_INSERTION) {
	    if (insert_text (in, in->completions [0], strlen (in->completions [0]))) {
		if (in->completions [1])
		    beep ();
	    } else
		beep ();
	}
	if ((what_to_do & DO_QUERY) && in->completions [1]) {
	    int maxlen = 0, i, count = 0;
	    int x, y, w, h;
	    int start_x, start_y;
	    char **p, *q;
	    Dlg_head *query_dlg;
	    WListbox *query_list;

	    for (p=in->completions + 1; *p; count++, p++)
		if ((i = strlen (*p)) > maxlen)
		    maxlen = i;
	    start_x = in->widget.x;
	    start_y = in->widget.y;
	    if (start_y - 2 >= count) {
		y = start_y - 2 - count;
		h = 2 + count;
	    } else {
		if (start_y >= LINES - start_y - 1) {
		    y = 0;
		    h = start_y;
		} else {
		    y = start_y + 1;
		    h = LINES - start_y - 1;
		}
	    }
	    x = text_start - in->first_shown - 2 + start_x;
	    w = maxlen + 4;
	    if (x + w > COLS)
		x = COLS - w;
	    if (x < 0)
		x = 0;
	    if (x + w > COLS)
		w = COLS;
	    query_input = in;
	    min_end = text_end;
	    query_height = h;
	    query_width  = w;
	    query_dlg = create_dlg (y, x, query_height, query_width,
				    dialog_colors, cmplt_query_callback,
				    "[Completion-query]", DLG_NONE);
	    query_list = listbox_new (1, 1, w - 2, h - 2, 0, querylist_callback);
	    add_widget (query_dlg, query_list);
	    for (p = in->completions + 1; *p; p++)
		listbox_add_item (query_list, 0, 0, *p, NULL);
	    run_dlg (query_dlg);
	    q = NULL;
	    if (query_dlg->ret_value == B_ENTER) {
		listbox_get_current (query_list, &q, NULL);
		if (q)
		    insert_text (in, q, strlen (q));
	    }
	    if (q || text_end != min_end)
		free_completions (in);
	    i = query_dlg->ret_value; /* B_USER if user wants to start over again */
	    destroy_dlg (query_dlg);
	    if (i == B_USER)
		return 1;
	}
    } else
	beep ();
    return 0;
}

void complete (WInput *in)
{
    if (in->completions)
	while (complete_engine (in, DO_QUERY));
    else if (show_all_if_ambiguous) {
	complete_engine (in, DO_INSERTION);
	while (complete_engine (in, DO_QUERY));
    } else
	complete_engine (in, DO_INSERTION);
}

/* Client interface for General purpose Linux console save/restore server */

/* rxvt.c - gives output lines on rxvt with a special rxvt patch */

#define look_for_rxvt_extensions() rxvt_extensions

static void init_look_for_rxvt_extensions (void)
{
	char *e = getenv ("RXVT_EXT");

	if (e && e[0] == '1' && e[1] == '.') {
		rxvt_extensions = 1;
		console_flag = 4;
	}
}

/* The cons saver can't have a pid of 1, used to prevent bunches of */

void show_console_contents (int starty, unsigned char begin_line, unsigned char end_line)
{
    unsigned char message = 0;
    unsigned short bytes = 0;
    int i;

    standend ();

    if (look_for_rxvt_extensions ()) {
	show_rxvt_contents (starty, begin_line, end_line);
	return;
    }

    /* Is tty console? */
    if (!console_flag)
	return;
    /* Paranoid: Is the cons.saver still running? */
    if (cons_saver_pid < 1 || kill (cons_saver_pid, SIGCONT)) {
	cons_saver_pid = 0;
	console_flag = 0;
	return;
    }

    /* Send command to the console handler */
    message = CONSOLE_CONTENTS;
    write (pipefd1[1], &message, 1);
    /* Check for outdated cons.saver */
    read (pipefd2[0], &message, 1);
    if (message != CONSOLE_CONTENTS)
	return;

    /* Send the range of lines that we want */
    write (pipefd1[1], &begin_line, 1);
    write (pipefd1[1], &end_line, 1);
    /* Read the corresponding number of bytes */
    read (pipefd2[0], &bytes, 2);

    /* Read the bytes and output them */
    for (i = 0; i < bytes; i++) {
	if ((i % COLS) == 0)
	    move (starty+(i/COLS), 0);
	read (pipefd2[0], &message, 1);
	addch (message);
    }

    /* Read the value of the console_flag */
    read (pipefd2[0], &message, 1);
}

void handle_console (unsigned char action)
{
    char *tty_name;
    char *mc_conssaver;

    switch (action) {
    case CONSOLE_INIT:
	if (look_for_rxvt_extensions ())
	    return;
	/* Close old pipe ends in case it is the 2nd time we run cons.saver */
	if(pipefd1[1] != -1)
		close (pipefd1[1]);
	if(pipefd2[0] != -1)
		close (pipefd2[0]);
	/* Create two pipes for communication */
	pipe (pipefd1);
	pipe (pipefd2);
	/* Get the console saver running */
	cons_saver_pid = fork ();
	if (cons_saver_pid < 0) {
	    /* Can't fork */
	    /* Delete pipes */
	    close (pipefd1[1]);
	    close (pipefd1[0]);
	    pipefd1[0] = pipefd1[1] = -1;
	    close (pipefd2[1]);
	    close (pipefd2[0]);
	    pipefd2[0] = pipefd2[1] = -1;
	    console_flag = 0;
	} else if (cons_saver_pid > 0) {
	    /* Parent */
	    /* Close the extra pipe ends */
	    close (pipefd1[0]);
	    close (pipefd2[1]);
	    pipefd1[0] = pipefd2[1] = -1;
	    /* Was the child successful? */
	    read (pipefd2[0], &console_flag, 1);
	    if (!console_flag) {
		close (pipefd1[1]);
		close (pipefd2[0]);
		pipefd1[1] = pipefd2[0] = -1;
		waitpid_no_int (cons_saver_pid, NULL, 0, 0);
	    }
	} else {
	    /* Child */
	    /* Close the extra pipe ends */
	    close (pipefd1[1]);
	    close (pipefd2[0]);
	    tty_name = ttyname (0);
	    /* Bind the pipe 0 to the standard input */
	    close (0);
	    dup (pipefd1[0]);
	    close (pipefd1[0]);
	    /* Bind the pipe 1 to the standard output */
	    close (1);
	    dup (pipefd2[1]);
	    close (pipefd2[1]);
	    /* Bind standard error to /dev/null */
	    close (2);
	    open (bb_dev_null, O_WRONLY);
	    /* Exec the console save/restore handler */
	    mc_conssaver = concat_mcpath_file ("bin/cons.saver");
	    execl (mc_conssaver, "cons.saver", tty_name, NULL);
	    /* Exec failed */
	    console_flag = 0;
	    write (1, &console_flag, 1);
	    close (1);
	    close (0);
	    exit (3);
	} /* if (cons_saver_pid ...) */
	break;

    case CONSOLE_DONE:
    case CONSOLE_SAVE:
    case CONSOLE_RESTORE:
	if (look_for_rxvt_extensions ())
	    return;
	/* Is tty console? */
	if (!console_flag)
	    return;
	/* Paranoid: Is the cons.saver still running? */
	if (cons_saver_pid < 1 || kill (cons_saver_pid, SIGCONT)) {
	    cons_saver_pid = 0;
	    console_flag = 0;
	    return;
	}
	/* Send command to the console handler */
	write (pipefd1[1], &action, 1);
	if (action != CONSOLE_DONE) {
	    /* Wait the console handler to do its job */
	    read (pipefd2[0], &console_flag, 1);
	}
	if (action == CONSOLE_DONE || !console_flag) {
	    /* We are done -> Let's clean up */
	    close (pipefd1 [1]);
	    close (pipefd2 [0]);
	    pipefd1[1] = pipefd2[0] = -1;
	    waitpid_no_int (cons_saver_pid, NULL, 0, 0);
	    console_flag = 0;
	}
	break;
    default:
	;
	/* Nothing */
    }
}

/* Dialog managing. */

static void push_refresh (void (*new_refresh)(void *), void *parameter, int flags)
{
    Refresh *new;

    new = x_malloc (sizeof (Refresh), "push_refresh");
    new->next = refresh_list;
    new->refresh_fn = new_refresh;
    new->parameter = parameter;
    new->flags     = flags;
    refresh_list = new;
}

static void pop_refresh (void)
{
    Refresh *old;

    if (!refresh_list)
	fprintf (stderr, "\n\n\nrefresh stack underflow!\n\n\n");
    else {
	old = refresh_list;
	refresh_list = refresh_list->next;
	free (old);
    }
}

static void do_complete_refresh (Refresh *Refresh_list)
{
    if (!Refresh_list)
	return;

    if (Refresh_list->flags != REFRESH_COVERS_ALL)
	do_complete_refresh (Refresh_list->next);

    (*(Refresh_list->refresh_fn))(Refresh_list->parameter);
}

void do_refresh (void)
{
    if (!refresh_list)
	return;
    else {
	if (fast_refresh)
	    (*(refresh_list->refresh_fn))(refresh_list->parameter);
	else {
	    do_complete_refresh (refresh_list);
	}
    }
}

/* Directory routines */

int
unsorted (const file_entry *a, const file_entry *b)
{
    return 0;
}

int
sort_name (const file_entry *a, const file_entry *b)
{
    int ad = MY_ISDIR (a);
    int bd = MY_ISDIR (b);

    if (ad == bd || mix_all_files)
	return string_sortcomp (a->fname, b->fname) * reverse_sort_flag;
    return bd-ad;
}

int
sort_ext (const file_entry *a, const file_entry *b)
{
    const char *exta;
    const char *extb;
    int r;
    int ad = MY_ISDIR (a);
    int bd = MY_ISDIR (b);

    if (ad == bd || mix_all_files) {
	exta = extension (a->fname);
	extb = extension (b->fname);
	r = string_sortcomp (exta, extb);
	if (r)
	    return r * reverse_sort_flag;
	else
	    return sort_name (a, b);
    } else
	return bd-ad;
}

int
sort_owner (const file_entry *a, const file_entry *b)
{
    int ad = MY_ISDIR (a);
    int bd = MY_ISDIR (b);

    if (ad == bd || mix_all_files)
	return string_sortcomp (get_owner (a->buf.st_uid), get_owner (a->buf.st_uid)) * reverse_sort_flag;
    return bd-ad;
}

int
sort_group (const file_entry *a, const file_entry *b)
{
    int ad = MY_ISDIR (a);
    int bd = MY_ISDIR (b);

    if (ad == bd || mix_all_files)
	return string_sortcomp (get_group (a->buf.st_gid), get_group (a->buf.st_gid)) * reverse_sort_flag;
    return bd-ad;
}

int
sort_time (const file_entry *a, const file_entry *b)
{
    int ad = MY_ISDIR (a);
    int bd = MY_ISDIR (b);

    if (ad == bd || mix_all_files)
	return (a->buf.st_mtime - b->buf.st_mtime) * reverse_sort_flag;
    else
	return bd-ad;
}

int
sort_ctime (const file_entry *a, const file_entry *b)
{
    int ad = MY_ISDIR (a);
    int bd = MY_ISDIR (b);

    if (ad == bd || mix_all_files)
	return (a->buf.st_ctime - b->buf.st_ctime) * reverse_sort_flag;
    else
	return bd-ad;
}

int
sort_atime (const file_entry *a, const file_entry *b)
{
    int ad = MY_ISDIR (a);
    int bd = MY_ISDIR (b);

    if (ad == bd || mix_all_files)
	return (a->buf.st_atime - b->buf.st_atime) * reverse_sort_flag;
    else
	return bd-ad;
}

int
sort_inode (const file_entry *a, const file_entry *b)
{
    int ad = MY_ISDIR (a);
    int bd = MY_ISDIR (b);

    if (ad == bd || mix_all_files) {
	ino_t ino_cmp;
	int rez = 0;    /* long long diff to int sign value */

	ino_cmp = a->buf.st_ino - b->buf.st_ino;
	if(ino_cmp < 0)
		rez = -1;
	else if(ino_cmp > 0)
		rez = 1;
	return rez * reverse_sort_flag;
    } else {
	return bd-ad;
    }
}

int
sort_size (const file_entry *a, const file_entry *b)
{
    int ad = MY_ISDIR (a);
    int bd = MY_ISDIR (b);

    if (ad == bd || mix_all_files) {
	off_t size_cmp;
	int rez = 0;    /* long long diff to int sign value */

	size_cmp = b->buf.st_size - a->buf.st_size;
	if(size_cmp < 0)
		rez = -1;
	else if(size_cmp > 0)
		rez = 1;
	return rez * reverse_sort_flag;
    } else {
	return bd-ad;
    }
}

int
sort_links (const file_entry *a, const file_entry *b)
{
    int ad = MY_ISDIR (a);
    int bd = MY_ISDIR (b);

    if (ad == bd || mix_all_files)
	return (b->buf.st_nlink - a->buf.st_nlink) * reverse_sort_flag;
    else
	return bd-ad;
}

int
sort_ngid (const file_entry *a, const file_entry *b)
{
    int ad = MY_ISDIR (a);
    int bd = MY_ISDIR (b);

    if (ad == bd || mix_all_files)
	return (b->buf.st_gid - a->buf.st_gid) * reverse_sort_flag;
    else
	return bd-ad;
}

int
sort_nuid (const file_entry *a, const file_entry *b)
{
    int ad = MY_ISDIR (a);
    int bd = MY_ISDIR (b);

    if (ad == bd || mix_all_files)
	return (b->buf.st_uid - a->buf.st_uid) * reverse_sort_flag;
    else
	return bd-ad;
}

static int
file_type_to_num (const file_entry *fe)
{
    const struct stat *s = &fe->buf;

    if (S_ISDIR (s->st_mode))
	return 0;
    if (S_ISLNK (s->st_mode)) {
	if (fe->f.link_to_dir)
	    return 1;
	if (fe->f.stalled_link)
	    return 2;
	else
	    return 3;
    }
    if (S_ISSOCK (s->st_mode))
	return 4;
    if (S_ISCHR (s->st_mode))
	return 5;
    if (S_ISBLK (s->st_mode))
	return 6;
    if (S_ISFIFO (s->st_mode))
	return 7;
    if (is_exe (s->st_mode))
	return 8;
    return 9;
}

int
sort_type (const file_entry *a, const file_entry *b)
{
    int aa  = file_type_to_num (a);
    int bb  = file_type_to_num (b);

    return bb-aa;
}



static void clean_dir (dir_list *list, int count)
{
    int i;

    for (i = 0; i < count; i++) {
	free (list->list [i].fname);
	list->list [i].fname = 0;
	if (list->list [i].cache != NULL) {
	    free (list->list [i].cache);
	    list->list [i].cache = NULL;
	}
    }
}

static int
add_dotdot_to_list (dir_list *list, int Index)
{
    char buffer [MC_MAXPATHLEN + MC_MAXPATHLEN];
    char *p, *s;

    /* Need to grow the *list? */
    if (Index == list->size) {
	list->list = realloc (list->list, sizeof (file_entry) *
			      (list->size + RESIZE_STEPS));
	if (!list->list)
	    return 0;
	list->size += RESIZE_STEPS;
    }

    (list->list) [Index].fnamelen = 2;
    (list->list) [Index].fname = strdup ("..");
    (list->list) [Index].cache = NULL;
    (list->list) [Index].f.link_to_dir = 0;
    (list->list) [Index].f.stalled_link = 0;

    /* FIXME: We need to get the panel definition! to use file_mark */
    (list->list) [Index].f.marked = 0;
    mc_get_current_wd (buffer, sizeof (buffer) - 1 );
    if (buffer [strlen (buffer) - 1] == '/')
	buffer [strlen (buffer) - 1] = 0;
    for (;;) {
	strcat (buffer, "/..");
	p = vfs_canon (buffer);
	if (mc_stat (p, &((list->list) [Index].buf)) != -1) {
	    free (p);
	    break;
	}
	if ((s = vfs_path (p)) && !strcmp (s, "/")) {
	    free (p);
	    return 1;
	}
	strcpy (buffer, p);
	free (p);
    }

    return 1;
}

/* Used to set up a directory list when there is no access to a directory */
static int set_zero_dir (dir_list *list)
{
    return (add_dotdot_to_list (list, 0));
}

/* If you change handle_dirent then check also handle_path. */
/* Return values: -1 = failure, 0 = don't add, 1 = add to the list */
static int handle_dirent (dir_list *list, char *filter, struct dirent *dp,
		   struct stat *buf1, int next_free, int *link_to_dir,
		   int *stalled_link)
{
    if (dp->d_name [0] == '.' && dp->d_name [1] == 0)
	return 0;
    if (!show_dot_files) {
	if (dp->d_name [0] == '.') {
	    if (!(dp->d_name [1] == 0))
		if (!(dp->d_name [1] == '.' && NLENGTH (dp) == 2))
		    return 0;
	}
    }
    if (!show_backups && dp->d_name [NLENGTH (dp)-1] == '~')
	return 0;
    if (mc_lstat (dp->d_name, buf1) == -1)
	return 0;

    if (S_ISDIR (buf1->st_mode))
	tree_check (dp->d_name);

    /* A link to a file or a directory? */
    *link_to_dir = 0;
    *stalled_link = 0;
    if (S_ISLNK(buf1->st_mode)) {
	struct stat buf2;
	if (!mc_stat (dp->d_name, &buf2))
	    *link_to_dir = S_ISDIR(buf2.st_mode) != 0;
	else
	    *stalled_link = 1;
    }
    if (!(S_ISDIR(buf1->st_mode) || *link_to_dir) && filter &&
	!regexp_match (filter, dp->d_name, match_file))
	return 0;

    /* Need to grow the *list? */
    if (next_free == list->size) {
	list->list = realloc (list->list, sizeof (file_entry) *
			      (list->size + RESIZE_STEPS));
	if (!list->list)
	    return -1;
	list->size += RESIZE_STEPS;
    }
    return 1;
}

/* handle_path is a simplified handle_dirent. The difference is that
   handle_path doesn't pay attention to show_dot_files and show_backups.
   Moreover handle_path can't be used with a filemask.
   If you change handle_path then check also handle_dirent. */
/* Return values: -1 = failure, 0 = don't add, 1 = add to the list */
static int handle_path (dir_list *list, char *path,
		   struct stat *buf1, int next_free, int *link_to_dir,
		   int *stalled_link)
{
    if (path [0] == '.' && path [1] == 0)
	return 0;
    if (mc_lstat (path, buf1) == -1)
	return 0;

    if (S_ISDIR (buf1->st_mode))
	tree_check (path);

    /* A link to a file or a directory? */
    *link_to_dir = 0;
    *stalled_link = 0;
    if (S_ISLNK(buf1->st_mode)) {
	struct stat buf2;
	if (!mc_stat (path, &buf2))
	    *link_to_dir = S_ISDIR(buf2.st_mode) != 0;
	else
	    *stalled_link = 1;
    }

    /* Need to grow the *list? */
    if (next_free == list->size) {
	list->list = realloc (list->list, sizeof (file_entry) *
			      (list->size + RESIZE_STEPS));
	if (!list->list)
	    return -1;
	list->size += RESIZE_STEPS;
    }
    return 1;
}

static int
do_load_dir(dir_list *list, sortfn *sort, int revers, int case_sensitiv, char *filter)
{
    DIR           *dirp;
    struct dirent *dp;
    int           Status, link_to_dir, stalled_link;
    int           next_free = 0;
    struct stat   buf;
    int dotdot_found = 0;

    start_tree_check (NULL);

    dirp = mc_opendir (".");
    if (!dirp) {
	return set_zero_dir (list);
    }
    for (dp = mc_readdir (dirp); dp; dp = mc_readdir (dirp)) {
	Status = handle_dirent (list, filter, dp, &buf, next_free, &link_to_dir,
	    &stalled_link);
	if (Status == 0)
	    continue;
	if (Status == -1)
	    return next_free;
	list->list [next_free].fnamelen = NLENGTH (dp);
	list->list [next_free].fname = strdup (dp->d_name);
	list->list [next_free].cache = NULL;
	list->list [next_free].f.marked = 0;
	list->list [next_free].f.link_to_dir = link_to_dir;
	list->list [next_free].f.stalled_link = stalled_link;
	list->list [next_free].buf = buf;
	if (strcmp (dp->d_name, ".." ) == 0)
	    dotdot_found = 1;
	next_free++;
	if (!(next_free % 32))
	    rotate_dash ();
    }

    if (next_free) {
	if (!dotdot_found)
	    add_dotdot_to_list (list, next_free++);
	do_sort (list, sort, next_free-1, revers, case_sensitiv);
    }
    else
	return set_zero_dir (list);

    mc_closedir (dirp);
    end_tree_check (NULL);
    return next_free;
}


static int if_link_is_exe (file_entry *file)
{
    struct stat b;

    if (S_ISLNK (file->buf.st_mode)) {
	mc_stat (file->fname, &b);
	return is_exe (b.st_mode);
    }
    return 1;
}

static void alloc_dir_copy (int size)
{
    int i;

    if (dir_copy.size < size) {
	if (dir_copy.list) {

	    for (i = 0; i < dir_copy.size; i++) {
		if (dir_copy.list [i].fname)
		    free (dir_copy.list [i].fname);
		if (dir_copy.list [i].cache)
		    free (dir_copy.list [i].cache);
	    }
	    free (dir_copy.list);
	    dir_copy.list = 0;
	}

	dir_copy.list = zmalloc (sizeof (file_entry) * size);
	dir_copy.size = size;
    }
}

/* If filter is null, then it is a match */
static int do_reload_dir (dir_list *list, sortfn *sort, int count, int rev,
		   int case_sensitiv, char *filter)
{
    DIR           *dirp;
    struct dirent *dp;
    int           next_free = 0;
    int           i, found, Status, link_to_dir, stalled_link;
    struct stat   buf;
    int           tmp_len;  /* For optimisation */
    int           dotdot_found = 0;

    start_tree_check (NULL);
    dirp = mc_opendir (".");
    if (!dirp) {
	clean_dir (list, count);
	return set_zero_dir (list);
    }

    alloc_dir_copy (list->size);
    for (i = 0; i < count; i++) {
	dir_copy.list [i].fnamelen = list->list [i].fnamelen;
	dir_copy.list [i].fname =    list->list [i].fname;
	dir_copy.list [i].cache =    list->list [i].cache;
	dir_copy.list [i].f.marked = list->list [i].f.marked;
	dir_copy.list [i].f.link_to_dir = list->list [i].f.link_to_dir;
	dir_copy.list [i].f.stalled_link = list->list [i].f.stalled_link;
    }

    for (dp = mc_readdir (dirp); dp; dp = mc_readdir (dirp)) {
	Status = handle_dirent (list, filter, dp, &buf, next_free, &link_to_dir,
	    &stalled_link);
	if (Status == 0)
	    continue;
	if (Status == -1) {
	    mc_closedir (dirp);
	    /* Norbert (Feb 12, 1997):
	     Just in case someone finds this memory leak:
	     -1 means big trouble (at the moment no memory left),
	     I don't bother with further cleanup because if one gets to
	     this point he will have more problems than a few memory
	     leaks and because one 'clean_dir' would not be enough (and
	     because I don't want to spent the time to make it working,
	     IMHO it's not worthwhile).
	    clean_dir (&dir_copy, count);
	     */
	    return next_free;
	}

	tmp_len = NLENGTH (dp);
	for (found = i = 0; i < count; i++)
	    if (tmp_len == dir_copy.list [i].fnamelen
		&& !strcmp (dp->d_name, dir_copy.list [i].fname)) {
		list->list [next_free].f.marked = dir_copy.list [i].f.marked;
		found = 1;
		break;
	    }

	if (!found)
	    list->list [next_free].f.marked = 0;

	list->list [next_free].fnamelen = tmp_len;
	list->list [next_free].fname = strdup (dp->d_name);
	list->list [next_free].cache = NULL;
	list->list [next_free].f.link_to_dir = link_to_dir;
	list->list [next_free].f.stalled_link = stalled_link;
	list->list [next_free].buf = buf;
	if (strcmp (dp->d_name, ".." ) == 0)
	    dotdot_found = 1;
	next_free++;
	if (!(next_free % 16))
	    rotate_dash ();
    }
    mc_closedir (dirp);
    end_tree_check (NULL);
    if (next_free) {
	if (!dotdot_found)
	    add_dotdot_to_list (list, next_free++);
	do_sort (list, sort, next_free-1, rev, case_sensitiv);
    }
    else
	next_free = set_zero_dir (list);
    clean_dir (&dir_copy, count);
    return next_free;
}

/* Dlg box features module for the Midnight Commander */

#define waddc(w,y1,x1,c) move (w->y+y1, w->x+x1); addch (c)

#ifndef PORT_HAS_SET_IDLE
#    define x_set_idle(d,x)
#endif

#ifndef PORT_HAS_DIALOG_STOP
#   define x_dialog_stop(d)
#endif

static void slow_box (Dlg_head *h, int y, int x, int ys, int xs)
{
    move (h->y+y, h->x+x);
    hline (' ', xs);
    vline (' ', ys);
    move (h->y+y, h->x+x+xs-1);
    vline (' ', ys);
    move (h->y+y+ys-1, h->x+x);
    hline (' ', xs);
}

/* draw box in window */
void draw_box (Dlg_head *h, int y, int x, int ys, int xs)
{
    if (slow_terminal) {
	slow_box (h, y, x, ys, xs);
	return;
    }

    SLsmg_draw_box (h->y+y, h->x+x, ys, xs);
}

/* draw box in window */
void draw_double_box (Dlg_head *h, int y, int x, int ys, int xs)
{
    SLsmg_draw_double_box (h->y+y, h->x+x, ys, xs);
}

static void widget_erase (Widget *w)
{
    int x, y;

    for (y = 0; y < w->lines; y++) {
	widget_move (w, y, 0);
	for (x = 0; x < w->cols; x++)
	    addch (' ');
    }
}

void dlg_erase (Dlg_head *h)
{
    int x, y;

    for (y = 0; y < h->lines; y++) {
	move (y+h->y, h->x);    /* FIXME: should test if ERR */
	for (x = 0; x < h->cols; x++) {
	    addch (' ');
	}
    }
}

void init_widget (Widget *w, int y, int x, int lines, int cols,
		  int (*callback)(Dlg_head *, void *, int, int),
		  destroy_fn destroy, mouse_h mouse_handler)
{
    w->x = x;
    w->y = y;
    w->cols = cols;
    w->lines = lines;
    w->color = -1;
    w->callback = callback;
    w->destroy  = destroy;
    w->mouse = mouse_handler;
    w->wdata = 0;
    w->wcontainer = 0;
    w->parent = 0;

    /* Almost all widgets want to put the cursor in a suitable place */
    w->options = W_WANT_CURSOR;
}

int default_proc (Dlg_head *h, int Msg, int Par)
{
    switch (Msg) {

    case WIDGET_HOTKEY:         /* Didn't use the key */
	return 0;

    case WIDGET_INIT:           /* We could tell if something went wrong */
	return 1;

    case WIDGET_KEY:
	return 0;               /* Didn't use the key */

    case WIDGET_FOCUS:          /* We accept FOCUSes */
	return 1;

    case WIDGET_UNFOCUS:        /* We accept loose FOCUSes */
	return 1;

    case WIDGET_DRAW:
	return 1;

    case WIDGET_DESTROY:
	return 1;

    case WIDGET_CURSOR:
	/* Move the cursor to the default widget position */
	return 1;

    case WIDGET_IDLE:
	return 1;
    }
    printf ("Internal error: unhandled message: %d\n", Msg);
    return 1;
}

int default_dlg_callback (Dlg_head *h, int id, int msg)
{
    if (msg == DLG_IDLE) {
	dlg_broadcast_msg_to (h, WIDGET_IDLE, 0, W_WANT_IDLE);
    }
    return 0;
}

static Dlg_head *create_dlg (int y1, int x1, int lines, int cols,
		      int *color_set,
		      int (*callback) (struct Dlg_head *, int, int),
		      const char *help_ctx, int flags)
{
    Dlg_head *new_d;

    if (flags & DLG_CENTER) {
	y1 = (LINES-lines)/2;
	x1 = (COLS-cols)/2;
    }
    if ((flags & DLG_TRYUP) && (y1 > 3))
	y1 -= 2;

    new_d = zmalloc (sizeof (Dlg_head));
    new_d->direction = DIR_FORWARD;
    new_d->color = color_set;
    new_d->help_ctx = help_ctx;
    new_d->callback = callback ? callback : default_dlg_callback;
    new_d->x = x1;
    new_d->y = y1;
    new_d->cols = cols;
    new_d->lines = lines;
    return (new_d);
}

void set_idle_proc (Dlg_head *d, int state)
{
    d->send_idle_msg = state;
    x_set_idle (d, state);
}

/* add component to dialog buffer */
int add_widgetl (Dlg_head *where, void *what, WLay layout)
{
    Widget_Item *back;
    Widget      *widget = (Widget *) what;

    widget->layout = layout;
    /* Don't accept 0 widgets, this could be from widgets that could not */
    /* initialize properly */
    if (!what)
	return 0;

    widget->x += where->x;
    widget->y += where->y;

    if (where->running) {
	    Widget_Item *point = where->current;

	    where->current = xmalloc (sizeof (Widget_Item));

	    if (point) {
		    where->current->next = point->next;
		    where->current->prev = point;
		    point->next->prev = where->current;
		    point->next = where->current;
	    } else {
		    where->current->next = where->current;
		    where->first = where->current;
		    where->current->prev = where->first;
		    where->last = where->current;
		    where->first->next = where->last;
	    }
    } else {
	    back = where->current;
	    where->current = xmalloc (sizeof (Widget_Item));
	    if (back) {
		    back->prev = where->current;
		    where->current->next = back;
	    } else {
		    where->current->next = where->current;
		    where->first = where->current;
	    }

	    where->current->prev = where->first;
	    where->last = where->current;
	    where->first->next = where->last;

    }
    where->current->dlg_id = where->count;
    where->current->widget = what;
    where->current->widget->parent = where;

    where->count++;

    /* If the widget is inserted in a running dialog */
    if (where->running) {
	send_message (where, widget, WIDGET_INIT, 0);
	send_message (where, widget, WIDGET_DRAW, 0);
    }
    return (where->count - 1);
}

int add_widget (Dlg_head *where, void *what)
{
    return add_widgetl (where, what, XV_WLAY_DONTCARE);
}

int send_message (Dlg_head *h, Widget *w, int msg, int par)
{
    return (*(w->callback))(h, w, msg, par);
}

/* broadcast a message to all the widgets in a dialog that have
 * the options set to flags.
 */
void dlg_broadcast_msg_to (Dlg_head *h, int message, int revers, int flags)
{
    Widget_Item *p, *first;

    if (!h->current)
	    return;

    if (revers)
	first = p = h->current->prev;
    else
	first = p = h->current->next;
    do {
/*      if (p->widget->options & flags) */
	    send_message (h, p->widget, message, 0);

	if (revers)
	    p = p->prev;
	else
	    p = p->next;
    } while (first != p);
}

/* broadcast a message to all the widgets in a dialog */
void dlg_broadcast_msg (Dlg_head *h, int message, int revers)
{
    dlg_broadcast_msg_to (h, message, revers, ~0);
}

int dlg_focus (Dlg_head *h)
{
    if (send_message (h, h->current->widget, WIDGET_FOCUS, 0)) {
	(*h->callback) (h, h->current->dlg_id, DLG_FOCUS);
	return 1;
    }
    return 0;
}

static int dlg_unfocus (Dlg_head *h)
{
    if (send_message (h, h->current->widget, WIDGET_UNFOCUS, 0)) {
	(*h->callback) (h, h->current->dlg_id, DLG_UNFOCUS);
	return 1;
    }
    return 0;
}

static void select_a_widget (Dlg_head *h, int down)
{
    int direction = h->direction;

    if (!down)
	direction = !direction;

    do {
	if (direction)
	    h->current = h->current->next;
	else
	    h->current = h->current->prev;

	(*h->callback) (h, h->current->dlg_id, DLG_ONE_DOWN);
    } while (!dlg_focus (h));
}

/* Return true if the windows overlap */
static int dlg_overlap (Widget *a, Widget *b)
{
    if ((b->x >= a->x + a->cols)
	|| (a->x >= b->x + b->cols)
	|| (b->y >= a->y + a->lines)
	|| (a->y >= b->y + b->lines))
	return 0;
    return 1;
}


/* Searches a widget, uses the callback as a signature in the dialog h */
static Widget *find_widget_type (Dlg_head *h, callback_fn signature)
{
    Widget *w;
    Widget_Item *item;
    int i;

    if (!h)
	return 0;

    w = 0;
    for (i = 0, item = h->current; i < h->count; i++, item = item->next) {
	if (item->widget->callback == signature) {
	    w = item->widget;
	    break;
	}
    }
    return w;
}

void dlg_one_up (Dlg_head *h)
{
    Widget_Item *old;

    old = h->current;
    /* If it accepts unFOCUSion */
    if (!dlg_unfocus(h))
	return;

    select_a_widget (h, 0);
    if (dlg_overlap (old->widget, h->current->widget)) {
	send_message (h, h->current->widget, WIDGET_DRAW, 0);
	send_message (h, h->current->widget, WIDGET_FOCUS, 0);
    }
}

void dlg_one_down (Dlg_head *h)
{
    Widget_Item *old;

    old = h->current;
    if (!dlg_unfocus (h))
	return;

    select_a_widget (h, 1);
    if (dlg_overlap (old->widget, h->current->widget)) {
	send_message (h, h->current->widget, WIDGET_DRAW, 0);
	send_message (h, h->current->widget, WIDGET_FOCUS, 0);
    }
}

int dlg_select_widget (Dlg_head *h, void *w)
{
    if (dlg_unfocus (h)) {
	while (h->current->widget != w)
	    h->current = h->current->next;
	while (!dlg_focus (h))
	    h->current = h->current->next;

	return 1;
    }
    return 0;
}

static int send_message_to (Dlg_head *h, Widget *w, int msg, int par)
{
    Widget_Item *p = h->current;
    int v, i;

    v = 0;
    for (i = 0; i < h->count; i++) {
	if (w == (void *) p->widget) {
	    v = send_message (h, p->widget, msg, par);
	    break;
	}
	p = p->next;
    }
    return v;
}

#define callback(h) (h->current->widget->callback)

static void update_cursor (Dlg_head *h)
{
    if (!h->current)
	 return;
    if (h->current->widget->options & W_WANT_CURSOR)
	send_message (h, h->current->widget, WIDGET_CURSOR, 0);
    else {
	Widget_Item *p = h->current;

	do {
	    if (p->widget->options & W_WANT_CURSOR)
		if ((*p->widget->callback)(h, p->widget, WIDGET_CURSOR, 0)) {
		    break;
		}
	    p = p->next;
	} while (h->current != p);
    }
}

/* Redraw the widgets in reverse order, leaving the current widget
 * as the last one
 */
void dlg_redraw (Dlg_head *h)
{
    (h->callback)(h, 0, DLG_DRAW);

    dlg_broadcast_msg (h, WIDGET_DRAW, 1);

    update_cursor (h);
}

void dlg_refresh (void *parameter)
{
    dlg_redraw ((Dlg_head *) parameter);
}

void dlg_stop (Dlg_head *h)
{
    h->running = 0;
    x_dialog_stop (h);
}

static inline void dialog_handle_key (Dlg_head *h, int d_key)
{
   char *hlpfile;

    switch (d_key) {
    case KEY_LEFT:
    case KEY_UP:
	dlg_one_up (h);
	break;

    case KEY_RIGHT:
    case KEY_DOWN:
	dlg_one_down (h);
	break;

    case KEY_F(1):
	hlpfile = concat_mcpath_file ("mc.hlp");
	interactive_display (hlpfile, h->help_ctx);
	free (hlpfile);
	do_refresh ();
	break;

    case XCTRL('z'):
	suspend_cmd ();
	/* Fall through */

    case XCTRL('l'):
	touchwin ();
	mc_refresh ();
	break;

    case '\n':
    case KEY_ENTER:
	h->ret_value = B_ENTER;
	h->running = 0;
	x_dialog_stop (h);
	break;

    case ESC_CHAR:
    case KEY_F (10):
    case XCTRL ('c'):
    case XCTRL ('g'):
	h->ret_value = B_CANCEL;
	dlg_stop (h);
	break;
    }
}

static int dlg_try_hotkey (Dlg_head *h, int d_key)
{
    Widget_Item *hot_cur;
    Widget_Item *previous;
    int    handled, c;

    /*
     * Explanation: we don't send letter hotkeys to other widgets if
     * the currently selected widget is an input line
     */

    if (h->current->widget->options & W_IS_INPUT) {
		if(d_key < 255 && isalpha(d_key))
			return 0;
    }

    /* If it's an alt key, send the message */
    c = d_key & ~ALT(0);
    if (d_key & ALT(0) && c < 255 && isalpha(c))
	d_key = tolower(c);

    handled = 0;
    if (h->current->widget->options & W_WANT_HOTKEY)
	handled = callback (h) (h, h->current->widget, WIDGET_HOTKEY, d_key);

    /* If not used, send hotkey to other widgets */
    if (handled)
	return handled;

    hot_cur = h->current;

    /* send it to all widgets */
    do {
	if (hot_cur->widget->options & W_WANT_HOTKEY)
	    handled |= (*hot_cur->widget->callback)
		(h, hot_cur->widget, WIDGET_HOTKEY, d_key);

	if (!handled)
	    hot_cur = hot_cur->next;
    } while (h->current != hot_cur && !handled);

    if (!handled)
	return 0;

    (*h->callback) (h, 0, DLG_HOTKEY_HANDLED);
    previous = h->current;
    if (!dlg_unfocus (h))
	return handled;

    h->current = hot_cur;
    if (!dlg_focus (h)) {
	h->current = previous;
	dlg_focus (h);
    }
    return handled;
}

static void dlg_key_event (Dlg_head *h, int d_key)
{
    int handled;

    /* TAB used to cycle */
    if (!h->raw && (d_key == '\t' || d_key == KEY_BTAB))
	if (d_key == '\t')
	    dlg_one_down (h);
	else
	    dlg_one_up (h);
    else {

	/* first can dlg_callback handle the key */
	handled = (*h->callback) (h, d_key, DLG_KEY);

	/* next try the hotkey */
	if (!handled)
	    handled = dlg_try_hotkey (h, d_key);

	/* not used - then try widget_callback */
	if (!handled)
	    handled |= callback (h)(h, h->current->widget, WIDGET_KEY, d_key);

	/* not used- try to use the unhandled case */
	if (!handled)
	    handled |= (*h->callback) (h, d_key, DLG_UNHANDLED_KEY);

	if (!handled)
	    dialog_handle_key (h, d_key);
	(*h->callback) (h, d_key, DLG_POST_KEY);
    }
}

static inline int dlg_mouse_event (Dlg_head *h, Gpm_Event *event)
{
    Widget_Item *item;
    Widget_Item *starting_widget = h->current;
    Gpm_Event   new_event;
    int x = event->x;
    int y = event->y;
    int ret_value;

    /* kludge for the menubar: start at h->first, not current  */
    /* Must be carefull in the insertion order to the dlg list */
    if (y == 1 && h->has_menubar)
	starting_widget = h->first;

    item = starting_widget;
    do {
	Widget *widget = item->widget;

	item = item->next;

	if (!((x > widget->x) && (x <= widget->x+widget->cols)
	    && (y > widget->y) && (y <= widget->y+widget->lines)))
	    continue;

	new_event = *event;
	new_event.x -= widget->x;
	new_event.y -= widget->y;

	ret_value = widget->mouse ? (*widget->mouse) (&new_event, widget) :
	    MOU_NORMAL;

	return ret_value;
    } while (item != starting_widget);
    return 0;
}

/* Run dialog routines */

/* Init the process */
void init_dlg (Dlg_head *h)
{
    int refresh_mode;

    /* Initialize dialog manager and widgets */
    (*h->callback) (h, 0, DLG_INIT);
    dlg_broadcast_msg (h, WIDGET_INIT, 0);

    if (h->x == 0 && h->y == 0 && h->cols == COLS && h->lines == LINES)
	refresh_mode = REFRESH_COVERS_ALL;
    else
	refresh_mode = REFRESH_COVERS_PART;
    push_refresh (dlg_refresh, h, refresh_mode);
    h->refresh_pushed = 1;

    /* Initialize direction */
    if (!h->direction)
	h->current =  h->first;

    if (h->initfocus != NULL)
	h->current = h->initfocus;

    h->previous_dialog = current_dlg;
    current_dlg = h;

    /* Initialize the mouse status */
    h->mouse_status = 0;

    /* Redraw the screen */
    dlg_redraw (h);

    while (!dlg_focus (h))
	h->current = h->current->next;

    h->ret_value = 0;
    h->running = 1;
}

/* Shutdown the run_dlg */
void dlg_run_done (Dlg_head *h)
{
    (*h->callback) (h, h->current->dlg_id, DLG_END);
    current_dlg = (Dlg_head *) h->previous_dialog;
}

void dlg_process_event (Dlg_head *h, int key, Gpm_Event *event)
{
    if (key == EV_NONE) {
	if (got_interrupt ())
	    key = XCTRL('g');
	else
	    return;
    }

    if (key == EV_MOUSE)
	h->mouse_status = dlg_mouse_event (h, event);
    else
	dlg_key_event (h, key);
}

#ifndef PORT_HAS_FRONTEND_RUN_DLG
static inline void
frontend_run_dlg (Dlg_head *h)
{
    int d_key;
    Gpm_Event event;

    event.x = -1;
    while (h->running) {
	/* It does not work with ncurses before 1.9.9g, it will break */
	if (winch_flag)
	    change_screen_size ();
	if (is_idle ()) {
	    if (idle_hook)
		execute_hooks (idle_hook);

	    while (h->send_idle_msg && is_idle ()) {
		(*h->callback) (h, 0, DLG_IDLE);
	    }
	}

	update_cursor (h);
	(*h->callback)(h, 0, DLG_PRE_EVENT);

	/* Clear interrupt flag */
	got_interrupt ();
	d_key = get_event (&event, h->mouse_status == MOU_REPEAT, 1);

	dlg_process_event (h, d_key, &event);
    }
}
#endif /* PORT_HAS_FRONTEND_RUN_DLG */

/* Standard run dialog routine
 * We have to keep this routine small so that we can duplicate it's
 * behavior on complex routines like the file routines, this way,
 * they can call the dlg_process_event without rewriting all the code
 */
void run_dlg (Dlg_head *h)
{
    init_dlg (h);
    frontend_run_dlg (h);
    dlg_run_done (h);
}

void
destroy_dlg (Dlg_head *h)
{
    int i;
    Widget_Item *c;

    if (h->refresh_pushed)
	pop_refresh ();

    dlg_broadcast_msg (h, WIDGET_DESTROY, 0);
    c = h->current;
    for (i = 0; i < h->count; i++) {
	if (c->widget->destroy)
	    c->widget->destroy (c->widget);
	c = c->next;
	free (h->current->widget);
	free (h->current);
	h->current = c;
    }
    if (h->title)
	free (h->title);
    free (h);
    if (refresh_list)
	do_refresh ();
}

void widget_set_size (Widget *widget, int y, int x, int lines, int cols)
{
    widget->x = x;
    widget->y = y;
    widget->cols = cols;
    widget->lines = lines;
}

/* Replace widget old for widget new in the h dialog */
static void dlg_replace_widget (Dlg_head *h, Widget *old, Widget *new)
{
    Widget_Item *p = h->current;
    int should_focus = 0;

    do {
	if (p->widget == old) {

	    if (old == h->current->widget)
		should_focus = 1;

	    /* We found the widget */
	    /* First kill the widget */
	    new->focused = old->focused;
	    new->parent  = h;
	    send_message_to (h, old, WIDGET_DESTROY, 0);
	    (*old->destroy) (old);

	    /* We insert the new widget */
	    p->widget = new;
	    send_message_to (h, new, WIDGET_INIT, 0);
	    if (should_focus) {
		if (dlg_focus (h) == 0)
		    select_a_widget (h, 1);
	    }
	    send_message_to (h, new, WIDGET_DRAW, 0);
	    break;
	}
	p = p->next;
    } while (p != h->current);
}

static void widget_redraw (Dlg_head *h, Widget_Item *w)
{
    Widget_Item *save = h->current;

    h->current = w;
    (*w->widget->callback)(h, h->current->widget, WIDGET_DRAW, 0);
    h->current = save;
}

/* Returns the Index of h->current from h->first */
static int dlg_item_number (Dlg_head *h)
{
    Widget_Item *p;
    int i = 0;

    p = h->first;

    do {
	if (p == h->current)
	    return i;
	i++;
	p = p->next;
    } while (p != h->first);
    fprintf (stderr, "Internal error: current not in dialog list\n\r");
    exit (1);
}

static int dlg_select_nth_widget (Dlg_head *h, int n)
{
    Widget_Item *w;
    int i;

    w = h->first;
    for (i = 0; i < n; i++)
	w = w->next;

    return dlg_select_widget (h, w->widget);
}

static void
x_set_dialog_title (Dlg_head *h, const char *title)
{
  h->title = strdup(title);
}

/* Extension dependent execution. */

void
flush_extension_file (void)
{
    if (ext_data) {
	free (ext_data);
	ext_data = NULL;
    }

}

typedef char *(*quote_func_t)(const char *name, int i);

static char *
quote_block (quote_func_t quote_func, char **quoting_block)
{
	char **p = quoting_block;
	char *result = 0;
	char *tail   = 0;
	int current_len = 0;

	for (p = quoting_block; *p; p++) {
		int  temp_len;
		char *temp = quote_func (*p, 0);

		temp_len = strlen (temp);
		current_len += temp_len + 2;
		result = realloc (result, current_len);
		if (!tail)
			tail = result;
		strcpy (tail, temp);
		strcat (tail, " ");
		tail += temp_len + 1;
		free (temp);
	}

	return result;
}

static void
exec_extension (char *filename, char *Ext_data, char **drops, int *move_dir, int start_line)
{
    char *file_name;
    int  cmd_file_fd;
    FILE *cmd_file;
    int  expand_prefix_found = 0;
    int  parameter_found = 0;
    char prompt [80];
    int  run_view = 0;
    int  def_hex_mode = default_hex_mode, changed_hex_mode = 0;
    int  def_nroff_flag = default_nroff_flag, changed_nroff_flag = 0;
    int  written_nonspace = 0;
    int  is_cd = 0;
    char buffer [1024];
    char *p = 0;
    char *localcopy = NULL;
    int  do_local_copy;
#ifdef USE_NETCODE
    time_t localmtime = 0;
#endif
    struct stat mystat;
    quote_func_t quote_func = name_quote;

    /* Avoid making a local copy if we are doing a cd */
    if (!vfs_file_is_local(filename))
	do_local_copy = 1;
    else
	do_local_copy = 0;

    /* Note: this has to be done after the getlocalcopy call,
     * since it uses tmpnam as well
     */
    file_name = mc_mkstemp(&cmd_file_fd);

    if (cmd_file_fd == -1) {
	MC_message (1, MSG_ERROR, " Can't create temporary command file \n %m ");
	return;
    }
    cmd_file = fdopen (cmd_file_fd, "w");
    fprintf (cmd_file, "#!%s\n", MC_shell);

    prompt [0] = 0;
    for (;*Ext_data && *Ext_data != '\n'; Ext_data++) {
	if (parameter_found) {
	    if (*Ext_data == '}') {
		char *parameter;
		parameter_found = 0;
		parameter = input_dialog (" Parameter ", prompt, "");
		if (!parameter) {
		    /* User canceled */
		    fclose (cmd_file);
		    unlink (file_name);
		    if (localcopy) {
			mc_ungetlocalcopy (filename, localcopy, 0);
		    }
		    free (file_name);
		    return;
		}
		fputs (parameter, cmd_file);
		written_nonspace = 1;
		free (parameter);
	    } else {
		int len = strlen (prompt);

		if (len < sizeof (prompt) - 1) {
		    prompt [len] = *Ext_data;
		    prompt [len+1] = 0;
		}
	    }
	} else if (expand_prefix_found) {
	    expand_prefix_found = 0;
	    if (*Ext_data == '{')
		parameter_found = 1;
	    else {
		int i = check_format_view (Ext_data);
		char *v;

		if (i) {
		    Ext_data += i - 1;
		    run_view = 1;
		} else if ((i = check_format_cd (Ext_data)) > 0) {
		    is_cd = 1;
		    quote_func = fake_name_quote;
		    do_local_copy = 0;
		    p = buffer;
		    Ext_data += i - 1;
		} else if ((i = check_format_var (Ext_data, &v)) > 0 && v) {
		    fputs (v, cmd_file);
		    free (v);
		    Ext_data += i;
		} else {
		    char *text;

		    if (*Ext_data == 'f') {
			if (do_local_copy) {
			    localcopy = mc_getlocalcopy (filename);
			    if (localcopy == NULL) {
				fclose(cmd_file);
				unlink(file_name);
				free(file_name);
				return;
			    }
			    mc_stat (localcopy, &mystat);
#ifdef USE_NETCODE
			    localmtime = mystat.st_mtime;
#endif
			    text = (*quote_func) (localcopy, 0);
			} else {
			    text = (*quote_func) (filename, 0);
			}
		    } else if (*Ext_data == 'q') {
			text = quote_block (quote_func, drops);
		    } else
			text = expand_format (*Ext_data, !is_cd);
		    if (!is_cd)
			fputs (text, cmd_file);
		    else {
			strcpy (p, text);
			p = strchr (p, 0);
		    }
		    free (text);
		    written_nonspace = 1;
		}
	    }
	} else {
	    if (*Ext_data == '%')
		expand_prefix_found = 1;
	    else {
		if (*Ext_data != ' ' && *Ext_data != '\t')
		    written_nonspace = 1;
		if (is_cd)
		    *(p++) = *Ext_data;
		else
		    fputc (*Ext_data, cmd_file);
	    }
	}
    } /* for */
    fputc ('\n', cmd_file);
    fclose (cmd_file);
    chmod (file_name, S_IRWXU);
    if (run_view) {
	altered_hex_mode = 0;
	altered_nroff_flag = 0;
	if (def_hex_mode != default_hex_mode)
	    changed_hex_mode = 1;
	if (def_nroff_flag != default_nroff_flag)
	    changed_nroff_flag = 1;

	/* If we've written whitespace only, then just load filename
	 * into view
	 */
	if (written_nonspace)
	    View (file_name, filename, move_dir, start_line);
	else
	    View (0, filename, move_dir, start_line);
	if (changed_hex_mode && !altered_hex_mode)
	    default_hex_mode = def_hex_mode;
	if (changed_nroff_flag && !altered_nroff_flag)
	    default_nroff_flag = def_nroff_flag;
	repaint_screen ();
    } else if (is_cd) {
	char *q;
	*p = 0;
	p = buffer;
	while (*p == ' ' && *p == '\t')
	    p++;

	/* Search last non-space character. Start search at the end in order
	   not to short filenames containing spaces. */
	q = p + strlen (p) - 1;
	while (q >= p && (*q == ' ' || *q == '\t'))
	    q--;
	q[1] = 0;
	do_cd (p, cd_parse_command);
    } else {
	shell_execute (file_name, EXECUTE_INTERNAL | EXECUTE_TEMPFILE);
	if (console_flag)
	{
	    handle_console (CONSOLE_SAVE);
	    if (output_lines && keybar_visible)
	    {
		show_console_contents (output_start_y,
				       LINES-keybar_visible-output_lines-1,
				       LINES-keybar_visible-1);

	    }
	}

#ifdef OLD_CODE
	if (vfs_current_is_local ())
	    shell_execute (file_name, EXECUTE_INTERNAL);
	else
	    MC_message (1, " Warning ", " Can't execute commands on a Virtual File System directory ");
#endif
    }
#ifndef PORT_DOES_BACKGROUND_EXEC
    unlink (file_name);
#endif
    if (localcopy) {
	mc_stat (localcopy, &mystat);
	mc_ungetlocalcopy (filename, localcopy, localmtime != mystat.st_mtime);
    }
    free (file_name);
}

#ifdef FILE_L
#   define FILE_CMD "file -L "
#else
#   define FILE_CMD "file "
#endif

/* The second argument is action, i.e. Open, View, Edit, Drop, or NULL if
 * we want regex_command to return a list of all user defined actions.
 * Third argument is space separated list of dropped files (for actions
 * other then Drop it should be NULL);
 *
 * This function returns:
 *
 * If action != NULL, then it returns "Success" (not allocated) if it ran
 * some command or NULL if not.
 *
 * If action == NULL, it returns NULL if there are no user defined commands
 * or an allocated space separated list of user defined Actions.
 *
 * If action == "Icon", we are doing again something special. We return
 * icon name and we set the variable regex_command_title to Title for
 * that icon.
 *
 * If action == "View" then a parameter is checked in the form of "View:%d",
 * if the value for %d exists, then the viewer is started up at that line number.
 */
static char *regex_command (char *filename, char *action, char **drops, int *move_dir)
{
    char *extension_file;
    char *p, *q, *r, c;
    char *buffer;
    int  file_len = strlen (filename);
    int found = 0;
    char content_string [2048];
    int content_shift = 0;
    char *to_return = NULL;
    int old_patterns;
    struct stat mystat;
    int asked_file;
    int view_at_line_number;
    char *include_target;
    int include_target_len=0;

    int file_supports_stdin = 1;

    /* Check for the special View:%d parameter */
    if (action && strncmp (action, "View:", 5) == 0) {
	view_at_line_number = atoi (action + 5);
	action [4] = 0;
    } else {
	view_at_line_number = 0;
    }
    /* Have we asked file for the file contents? */
    asked_file = 0;

    if (ext_data == NULL) {
	int home_error = 0;

	buffer = concat_path_file (home_dir, MC_USER_EXT);
	if (exist_file (buffer))
	    extension_file = buffer;
	else
check_stock_mc_ext:
	    extension_file = concat_mcpath_file (MC_LIB_EXT);
	if ((ext_data = load_file (extension_file)) == NULL) {
	    free (buffer);
	    return 0;
	}
	if (!strstr (ext_data, "default/")) {
	    if (!strstr (ext_data, "regex/") && !strstr (ext_data, "shell/") &&
		!strstr (ext_data, "type/")) {
		free (ext_data);
		ext_data = NULL;
		if (extension_file == buffer) {
		    home_error = 1;
		    goto check_stock_mc_ext;
		} else {
		    char *msg;
		    char *msg2;
		    msg = xasprintf(" %s%s file error", mc_home, MC_LIB_EXT);
		    msg2 = xasprintf("Format of the %s"
					"mc.ext file has changed\n\
with version 3.0. It seems that installation\n\
failed. Please fetch a fresh new copy from the\n\
Midnight Commander package or in case you don't\n\
have any, get it from ftp://ftp.nuclecu.unam.mx.", mc_home);
		    MC_message (1, msg, msg2);
		    free (msg);
		    free (msg2);
		    free (buffer);
		    return 0;
		}
	    }
	}
	if (home_error) {
	    char *msg;
	    char *msg2;
	    msg = xasprintf(" ~/%s file error ", MC_USER_EXT);
	    msg2 = xasprintf("Format of the ~/%s file has changed\n\
with version 3.0. You may want either to\n\
copy it from %smc.ext or use that\n\
file as an example of how to write it.\n\
%smc.ext will be used for this moment.", MC_USER_EXT, mc_home, mc_home);
	    MC_message (1, msg, msg2);
	    free (msg);
	    free (msg2);
	}
	free (buffer);
    }
    mc_stat (filename, &mystat);

    if (regex_command_title) {
	free (regex_command_title);
	regex_command_title = NULL;
    }
    old_patterns = easy_patterns;
    easy_patterns = 0; /* Real regular expressions are needed :) */
    include_target = NULL;
    for (p = ext_data; *p; p++) {
	for (q = p; *q == ' ' || *q == '\t'; q++)
		;
	if (*q == '\n' || !*q)
	    p = q; /* empty line */
	if (*p == '#') /* comment */
	    while (*p && *p != '\n')
		p++;
	if (*p == '\n')
	    continue;
	if (!*p)
	    break;
	if (p == q) { /* i.e. starts in the first column, should be
		       * keyword/descNL
		       */
	    if (found && action == NULL) /* We have already accumulated all
					  * the user actions
					  */
		break;
	    found = 0;
	    q = strchr (p, '\n');
	    if (q == NULL)
		q = strchr (p, 0);
	    c = *q;
	    *q = 0;
	    if (include_target) {
		if ((strncmp (p, "include/", 8) == 0) &&
		    (strncmp (p+8, include_target, include_target_len) == 0))
		    found = 1;
	    } else if (!strncmp (p, "regex/", 6)) {
		p += 6;
		/* Do not transform shell patterns, you can use shell/ for
		 * that
		 */
		if (regexp_match (p, filename, match_normal))
		    found = 1;
	    } else if (!strncmp (p, "directory/", 10)) {
		if (S_ISDIR (mystat.st_mode) && regexp_match (p+10, filename, match_normal))
		    found = 1;
	    } else if (!strncmp (p, "shell/", 6)) {
		p += 6;
		if (*p == '.') {
		    if (!strncmp (p, filename + file_len - (q - p),
			q - p))
			found = 1;
		} else {
		    if (q - p == file_len && !strncmp (p, filename, q - p))
			found = 1;
		}
	    } else if (!strncmp (p, "type/", 5)) {
		int islocal = vfs_file_is_local (filename);
		p += 5;

		if (islocal || file_supports_stdin) {
		    char *pp;
		    int hasread = use_file_to_check_type;

		    if (asked_file || !use_file_to_check_type)
			goto match_file_output;

		    hasread = 0;
		    if (islocal) {
			char *tmp = name_quote (filename, 0);
			char *command = xasprintf ("%s%s", FILE_CMD, tmp);
			FILE *f = popen (command, "r");

			free (tmp);
			free (command);
			if (f != NULL) {
			    hasread = (fgets (content_string, 2047, f) != NULL);
			    if (!hasread)
				content_string [0] = 0;
			    pclose (f);
			}
		    } else {
			int pipehandle, remotehandle;
			pid_t P;

			remotehandle = mc_open (filename, O_RDONLY);
			if (remotehandle != -1) {
			/* 8192 is HOWMANY hardcoded value in the file-3.14
			 * sources. Tell me if any other file uses larger
			 * chunk from beginning
			 */
			    pipehandle = mc_doublepopen(remotehandle, 8192, &P,"file", "file", "-", NULL);
			    if (pipehandle != -1) {
				int i;
				while ((i = read (pipehandle, content_string
				     + hasread, 2047 - hasread)) > 0)
				    hasread += i;
				mc_doublepclose (pipehandle, P);
				content_string [hasread] = 0;
			    }
			    mc_close (remotehandle);
			}
		    }
		    asked_file = 1;
match_file_output:
		    if (hasread) {
			if ((pp = strchr (content_string, '\n')) != 0)
			    *pp = 0;
			if (islocal && !strncmp (content_string,
			    filename, file_len)) {
			    content_shift = file_len;
			    if (content_string [content_shift] == ':')
				for (content_shift++;
				    content_string [content_shift] == ' ';
				    content_shift++);
			} else if (!islocal
				   && !strncmp (content_string,
						"standard input:", 15)) {
			    for (content_shift = 15;
				content_string [content_shift] == ' ';
				content_shift++);
			}
			if (regexp_match (p, content_string +
				content_shift, match_normal)) {
			    found = 1;
			}
		    }
		}
	    } else if (!strncmp (p, "default/", 8)) {
		p += 8;
		found = 1;
	    }
	    *q = c;
	    p = q;
	    if (!*p)
		break;
	} else { /* List of actions */
	    p = q;
	    q = strchr (p, '\n');
	    if (q == NULL)
		q = strchr (p, 0);
	    if (found) {
		r = strchr (p, '=');
		if (r != NULL) {
		    c = *r;
		    *r = 0;
		    if (strcmp (p, "Include") == 0) {
			char *t;

			include_target = p + 8;
			t = strchr (include_target, '\n');
			if (t) *t = 0;
			include_target_len = strlen (include_target);
			if (t) *t = '\n';

			*r = c;
			p = q;
			found = 0;

			if (!*p)
			    break;
			continue;
		    }
		    if (action == NULL) {
			if (strcmp (p, "Open") &&
			    strcmp (p, "View") &&
			    strcmp (p, "Edit") &&
			    strcmp (p, "Drop") &&
			    strcmp (p, "Icon") &&
			    strcmp (p, "Include") &&
			    strcmp (p, "Title")) {
			    /* I.e. this is a name of a user defined action */

				if (to_return == NULL) {
				    to_return = x_malloc (512, "Action list");
				    qq = to_return;
				} else
				    *(qq++) = '='; /* Mark separator */
				strcpy (qq, p);
				qq = strchr (qq, 0);
			}
			*r = c;
		    } else if (!strcmp (action, "Icon")) {
			if (!strcmp (p, "Icon") && to_return == NULL) {
			    *r = c;
			    c = *q;
			    *q = 0;
			    to_return = strdup (r + 1);
			} else if (!strcmp (p, "Title") && regex_command_title == NULL) {
			    *r = c;
			    c = *q;
			    *q = 0;
			    regex_command_title = strdup (r + 1);
			} else {
			    *r = c;
			    c = *q;
			}
			*q = c;
			if (to_return != NULL && regex_command_title != NULL)
			    break;
		    } else if (!strcmp (action, p)) {
			*r = c;
			for (p = r + 1; *p == ' ' || *p == '\t'; p++)
			    ;

			/* Empty commands just stop searching
			 * through, they don't do anything
			 *
			 * We need to copy the filename because exec_extension
			 * may end up invoking update_panels thus making the
			 * filename parameter invalid (ie, most of the time,
			 * we get filename as a pointer from cpanel->dir).
			 */
			if (p < q) {
			    char *filename_copy = strdup (filename);

			    exec_extension (filename_copy, r + 1, drops, move_dir, view_at_line_number);
			    free (filename_copy);

			    to_return = "Success";
			}
			break;
		    } else
			*r = c;
		}
	    }
	    p = q;
	    if (!*p)
		break;
	}
    }
    easy_patterns = old_patterns;
    return to_return;
}
/* File managing

   The copy code was based in GNU's cp, and was written by:
   Torbjorn Granlund, David MacKenzie, and Jim Meyering.
*/

/* File operate window sizes */
#define WX 62
#define WY 10
#define BY 10
#define WX_ETA_EXTRA  12

#define FCOPY_GAUGE_X 14
#define FCOPY_LABEL_X 5

/* Used for button result values */
enum {
    REPLACE_YES = B_USER,
    REPLACE_NO,
    REPLACE_APPEND,
    REPLACE_ALWAYS,
    REPLACE_UPDATE,
    REPLACE_NEVER,
    REPLACE_ABORT,
    REPLACE_SIZE,
    REPLACE_REGET
};

enum {
    RECURSIVE_YES,
    RECURSIVE_NO,
    RECURSIVE_ALWAYS,
    RECURSIVE_NEVER,
    RECURSIVE_ABORT
};


static const char * const op_names [] = { " Copy ", " Move ", " Delete " };

static int recursive_erase (char *s);
static int erase_file (char *s);


/* }}} */

/* {{{ File progress display routines */

static int
check_buttons (void)
{
    int c;
    Gpm_Event event;

    c = get_event (&event, 0, 0);
    if (c == EV_NONE)
      return FILE_CONT;
    dlg_process_event (op_dlg, c, &event);
    switch (op_dlg->ret_value) {
    case FILE_SKIP:
	return FILE_SKIP;
	break;
    case B_CANCEL:
    case FILE_ABORT:
	return FILE_ABORT;
	break;
    default:
	return FILE_CONT;
    }
}

static int
op_win_callback (struct Dlg_head *h, int id, int msg)
{
    switch (msg) {
    case DLG_DRAW:
	attrset (COLOR_NORMAL);
	dlg_erase (h);
	draw_box (h, 1, 2, h->lines-2, h->cols-4);
	return 1;
    }
    return 0;
}

void
create_op_win (int op, int with_eta)
{
    int i, x_size;
    int minus = MC_verbose ? 0 : 3;
    int eta_offset = with_eta ? (WX_ETA_EXTRA) / 2 : 0;

    char *sixty = "";
    char *fifteen = "";

    replace_result = 0;
    recursive_result = 0;
    showing_eta = with_eta;
    showing_bps = with_eta;
    eta_extra = with_eta ? WX_ETA_EXTRA : 0;
    x_size = (WX + 4) + eta_extra;

    op_dlg = create_dlg (0, 0, WY-minus+4, x_size, dialog_colors,
			 op_win_callback, "", DLG_CENTER);

    last_hint_line = the_hint->widget.y;
    if ((op_dlg->y + op_dlg->lines) > last_hint_line)
	the_hint->widget.y = op_dlg->y + op_dlg->lines+1;

    x_set_dialog_title (op_dlg, "");

    add_widgetl (op_dlg, button_new (BY-minus, WX - 19 + eta_offset, FILE_ABORT,
				     NORMAL_BUTTON, "&Abort", 0, 0),
	XV_WLAY_RIGHTOF);
    add_widgetl (op_dlg, button_new (BY-minus, 14 + eta_offset, FILE_SKIP,
				     NORMAL_BUTTON, "&Skip", 0, 0),
	XV_WLAY_CENTERROW);

    add_widgetl (op_dlg, ProgressGauge [2] = gauge_new (7, FCOPY_GAUGE_X, 0, 100, 0),
	XV_WLAY_RIGHTOF);
    add_widgetl (op_dlg, ProgressLabel [2] = label_new (7, FCOPY_LABEL_X, fifteen),
	XV_WLAY_NEXTROW);
    add_widgetl (op_dlg, bps_label = label_new (7, WX, ""), XV_WLAY_NEXTROW);

    add_widgetl (op_dlg, ProgressGauge [1] = gauge_new (8, FCOPY_GAUGE_X, 0, 100, 0),
	XV_WLAY_RIGHTOF);
    add_widgetl (op_dlg, ProgressLabel [1] = label_new (8, FCOPY_LABEL_X, fifteen),
	XV_WLAY_RIGHTOF);
    add_widgetl (op_dlg, stalled_label = label_new (8, WX, ""), XV_WLAY_NEXTROW);

    add_widgetl (op_dlg, ProgressGauge [0] = gauge_new (6, FCOPY_GAUGE_X, 0, 100, 0),
	XV_WLAY_RIGHTOF);
    add_widgetl (op_dlg, ProgressLabel [0] = label_new (6, FCOPY_LABEL_X, fifteen),
	XV_WLAY_RIGHTOF);
    add_widgetl (op_dlg, eta_label = label_new (6, WX, ""), XV_WLAY_NEXTROW);

    add_widgetl (op_dlg, FileString [1] = label_new (4, FCOPY_GAUGE_X, sixty),
	XV_WLAY_RIGHTOF);
    add_widgetl (op_dlg, FileLabel [1] = label_new (4, FCOPY_LABEL_X, fifteen),
	XV_WLAY_NEXTROW);
    add_widgetl (op_dlg, FileString [0] = label_new (3, FCOPY_GAUGE_X, sixty),
	XV_WLAY_RIGHTOF);
    add_widgetl (op_dlg, FileLabel [0] = label_new (3, FCOPY_LABEL_X, fifteen),
	XV_WLAY_NEXTROW);

    /* We will manage the dialog without any help, that's why
       we have to call init_dlg */
    init_dlg (op_dlg);
    op_dlg->running = 1;
    for (i = 0; i < 3; i++)
	last_percentage [i] = -99;
}

void
destroy_op_win (void)
{
    dlg_run_done (op_dlg);
    destroy_dlg (op_dlg);
    the_hint->widget.y = last_hint_line;
}

static int
show_no_bar (int n)
{
    if (n >= 0) {
	label_set_text (ProgressLabel [n], "");
	gauge_show (ProgressGauge [n], 0);
    }
    return check_buttons ();
}

#define truncFileString(s) name_trunc (s, eta_extra + 47)

static int
show_source (char *s)
{
    if (s != NULL) {

	int i = strlen (cpanel->cwd);

	/* We remove the full path we have added before */
	if (!strncmp (s, cpanel->cwd, i)) {
	    if (s[i] == '/')
		s += i + 1;
	}

	label_set_text (FileLabel [0], "Source");
	label_set_text (FileString [0], truncFileString (s));
	return check_buttons ();
    } else {
	label_set_text (FileLabel [0], "");
	label_set_text (FileString [0], "");
	return check_buttons ();
    }
}

static int
show_target (char *s)
{
    if (s != NULL) {
	label_set_text (FileLabel [1], "Target");
	label_set_text (FileString [1], truncFileString (s));
	return check_buttons ();
    } else {
	label_set_text (FileLabel [1], "");
	label_set_text (FileString [1], "");
	return check_buttons ();
    }
}

static int
show_deleting (char *s)
{
    label_set_text (FileLabel [0], "Deleting");
    label_set_text (FileString [0], truncFileString (s));
    return check_buttons ();
}

static int
show_bar (int n, off_t done, off_t total)
{
    gauge_set_value (ProgressGauge [n], total, done);
    gauge_show (ProgressGauge [n], 1);
    return check_buttons ();
}

static inline void
file_eta_show (void)
{
    int eta_hours, eta_mins, eta_s;
    char eta_buffer [30];

    if (!showing_eta)
	return;

    eta_hours = eta_secs / (60 * 60);
    eta_mins  = (eta_secs - (eta_hours * 60 * 60)) / 60;
    eta_s     = eta_secs - ((eta_hours * 60 * 60) + eta_mins * 60 );
    sprintf (eta_buffer, "ETA %d:%02d.%02d", eta_hours, eta_mins, eta_s);
    label_set_text (eta_label, eta_buffer);
}

static void
file_bps_show (void)
{
    char bps_buffer [30];

    file_eta_show();
    if (!showing_bps)
	return;
    if (bps > 1024) {
	    if (bps > 1024*1024) {
		    sprintf (bps_buffer, "%.2f MBS", (double)bps / (1024*1024.0));
	    } else
		    sprintf (bps_buffer, "%.2f KBS", (double)bps / 1024.0);
    } else
	    sprintf (bps_buffer, "%d BPS", (int)bps);
    label_set_text (bps_label, bps_buffer);
}

static int
show_file_progress (off_t done, off_t total)
{
    if (!MC_verbose)
	return check_buttons ();
    if (total > 0) {
	label_set_text (ProgressLabel [0], "File");
	file_bps_show ();
	return show_bar (0, done, total);
    } else
	return show_no_bar (0);
}

static int
show_count_progress (off_t done, off_t total)
{
    if (!MC_verbose)
	return check_buttons ();
    if (total > 0) {
	label_set_text (ProgressLabel [1], "Count");
	return show_bar (1, done, total);
    } else
	return show_no_bar (1);
}

static int
show_bytes_progress (off_t done, off_t total)
{
    if (!MC_verbose)
	return check_buttons ();
    if (total > 0) {
	label_set_text (ProgressLabel [2], "Bytes");
	return show_bar (2, done, total);
    } else
	return show_no_bar (2);
}

/* }}} */


/* {{{ Copy routines */

enum CaseConvs { NO_CONV=0, UP_CHAR=1, LOW_CHAR=2, UP_SECT=4, LOW_SECT=8 };

static int
convert_case (int c, enum CaseConvs *conversion)
{
    if (*conversion & UP_CHAR) {
	*conversion &= ~UP_CHAR;
	return toupper (c);
    } else if (*conversion & LOW_CHAR) {
	*conversion &= ~LOW_CHAR;
	return tolower (c);
    } else if (*conversion & UP_SECT) {
	return toupper (c);
    } else if (*conversion & LOW_SECT) {
	return tolower (c);
    } else
	return c;
}

static char *
do_transform_source (char *source)
{
    int j, k, l, len;
    const char *fnsource = bb_basename (source);
    int next_reg;
    enum CaseConvs case_conv = NO_CONV;

    len = strlen (fnsource);
    j = re_match (&rx, fnsource, len, 0, &regs);
    if (j != len) {
	transform_error = FILE_SKIP;
	return NULL;
    }
    for (next_reg = 1, j = 0, k = 0; j < strlen (dest_mask); j++) {
	switch (dest_mask [j]) {
	case '\\':
	    j++;
	    if (! isdigit (dest_mask [j])) {
		/* Backslash followed by non-digit */
		switch (dest_mask [j]) {
		case 'U':
		    case_conv |= UP_SECT;
		    case_conv &= ~LOW_SECT;
		    break;
		case 'u':
		    case_conv |= UP_CHAR;
		    break;
		case 'L':
		    case_conv |= LOW_SECT;
		    case_conv &= ~UP_SECT;
		    break;
		case 'l':
		    case_conv |= LOW_CHAR;
		    break;
		case 'E':
		    case_conv = NO_CONV;
		    break;
		default:
		    /* Backslash as quote mark */
		    fntarget [k++] = convert_case (dest_mask [j], &case_conv);
		}
		break;
	    } else {
		/* Backslash followed by digit */
		next_reg = dest_mask [j] - '0';
		/* Fall through */
	    }

	case '*':
	    if (next_reg < 0 || next_reg >= RE_NREGS
		|| regs.start [next_reg] < 0) {
		message_1s (1, MSG_ERROR, " Invalid target mask ");
		transform_error = FILE_ABORT;
		return NULL;
	    }
	    for (l = regs.start [next_reg]; l < regs.end [next_reg]; l++)
		fntarget [k++] = convert_case (fnsource [l], &case_conv);
	    next_reg ++;
	    break;

	default:
	    fntarget [k++] = convert_case (dest_mask [j], &case_conv);
	    break;
	}
    }
    fntarget [k] = 0;
    return fntarget;
}

static char *
transform_source (char *source)
{
    char *s = strdup (source);
    char *q;

    /* We remove \n from the filename since regex routines would use \n as an anchor */
    /* this is just to be allowed to maniupulate file names with \n on it */
    for (q = s; *q; q++) {
	if (*q == '\n')
	    *q = ' ';
    }
    q = do_transform_source (s);
    free (s);
    return q;
}

static void
free_linklist (struct link **plinklist)
{
    struct link *lp, *lp2;

    for (lp = *plinklist; lp != NULL; lp = lp2) {
	lp2 = lp -> next;
	free (lp);
    }
    *plinklist = NULL;
}

#ifdef USE_VFS
static int
is_in_linklist (struct link *lp, char *path, struct stat *sb)
{
   ino_t ino = sb->st_ino;
   dev_t dev = sb->st_dev;
   vfs *Vfs = vfs_type (path);

   while (lp) {
      if (lp->vfs == Vfs && lp->ino == ino && lp->dev == dev )
	  return 1;
      lp = lp->next;
   }
   return 0;
}
#else
static int
is_in_linklist (struct link *lp, char *path, struct stat *sb)
{
   ino_t ino = sb->st_ino;
   dev_t dev = sb->st_dev;

   while (lp) {
      if (lp->ino == ino && lp->dev == dev )
	  return 1;
      lp = lp->next;
   }
   return 0;
}
#endif

/* Returns 0 if the inode wasn't found in the cache and 1 if it was found
   and a hardlink was succesfully made */
static int
check_hardlinks (char *src_name, char *dst_name, struct stat *pstat)
{
    struct link *lp;
    vfs *my_vfs = vfs_type (src_name);
    ino_t ino = pstat->st_ino;
    dev_t dev = pstat->st_dev;
    struct stat link_stat;
    char *p;

    if (vfs_file_is_ftp (src_name))
	return 0;
    for (lp = linklist; lp != NULL; lp = lp -> next)
	if (lp->vfs == my_vfs && lp->ino == ino && lp->dev == dev) {
	    if (!mc_stat (lp->name, &link_stat) && link_stat.st_ino == ino &&
		link_stat.st_dev == dev && vfs_type (lp->name) == my_vfs) {
		p = strchr (lp->name, 0) + 1; /* i.e. where the `name' file
						 was copied to */
		if (vfs_type (dst_name) == vfs_type (p)) {
		    if (!mc_stat (p, &link_stat)) {
			if (!mc_link (p, dst_name))
			    return 1;
		    }
		}
	    }
	    /* FIXME: Announce we couldn't make the hardlink */
	    return 0;
	}
    lp = x_malloc (sizeof (struct link) + strlen (src_name)
				  + strlen (dst_name) + 1, "Hardlink cache");
    if (lp) {
	lp->vfs = my_vfs;
	lp->ino = ino;
	lp->dev = dev;
	strcpy (lp->name, src_name);
	p = strchr (lp->name, 0) + 1;
	strcpy (p, dst_name);
	lp->next = linklist;
	linklist = lp;
    }
    return 0;
}

/* Duplicate the contents of the symbolic link src_path in dst_path.
   Try to make a stable symlink if the option "stable symlink" was
   set in the file mask dialog.
   If dst_path is an existing symlink it will be deleted silently
   (upper levels take already care of existing files at dst_path).
  */
static int
make_symlink (char *src_path, char *dst_path)
{
    char link_target[MC_MAXPATHLEN];
    int len;
    int return_status;
    struct stat sb;
    int dst_is_symlink;

    if (mc_lstat (dst_path, &sb) == 0 && S_ISLNK (sb.st_mode))
	dst_is_symlink = 1;
    else
	dst_is_symlink = 0;

  retry_src_readlink:
    len = mc_readlink (src_path, link_target, MC_MAXPATHLEN);
    if (len < 0) {
	return_status = file_error
	    (" Cannot read source link \"%s\" \n %s ", src_path);
	if (return_status == FILE_RETRY)
	    goto retry_src_readlink;
	return return_status;
    }
    link_target[len] = 0;

    if (stable_symlinks && (!vfs_file_is_local (src_path) ||
			    !vfs_file_is_local (dst_path))) {
	message_1s (1, MSG_ERROR, " Cannot make stable symlinks across "
				    "non-local filesystems: \n\n"
				    " Option Stable Symlinks will be disabled ");
	stable_symlinks = 0;
    }

    if (stable_symlinks && *link_target != '/') {
	char *p, *q, *r, *s;

	p = strdup (src_path);
	r = strrchr (p, '/');
	if (r) {
	    r[1] = 0;
	    if (*dst_path == '/')
		q = strdup (dst_path);
	    else
		q = xasprintf ("%s%s", p, dst_path);
	    r = strrchr (q, '/');
	    if (r) {
		r[1] = 0;
		s = xasprintf ("%s%s", p, link_target);
		strcpy (link_target, s);
		free (s);
		s = diff_two_paths (q, link_target);
		if (s) {
		    strcpy (link_target, s);
		    free (s);
		}
	    }
	    free (q);
	}
	free (p);
    }
  retry_dst_symlink:
    if (mc_symlink (link_target, dst_path) == 0)
	/* Success */
	return FILE_CONT;
    /*
     * if dst_exists, it is obvious that this had failed.
     * We can delete the old symlink and try again...
     */
    if (dst_is_symlink) {
	if (!mc_unlink (dst_path))
	    if (mc_symlink (link_target, dst_path) == 0)
		/* Success */
		return FILE_CONT;
    }
    return_status = file_error
	(" Cannot create target symlink \"%s\" \n %s ", dst_path);
    if (return_status == FILE_RETRY)
	goto retry_dst_symlink;
    return return_status;
}


int
copy_file_file (char *src_path, char *dst_path, int ask_overwrite)
{
    uid_t src_uid = 0;
    gid_t src_gid = 0;
    char *buf = 0;
    int  buf_size = 8*1024;
    int  dest_desc = 0;
    int  source_desc;
    int  n_read;
    int  n_written;
    int  src_mode = 0;              /* The mode of the source file */
    struct stat sb, sb2;
    struct utimbuf utb;
    int  dst_exists = 0;
    off_t n_read_total = 0;
    off_t file_size;
    int  return_status, temp_status;
    int do_remote_copy = 0;
    int appending = 0;
    /* bitmask used to remember which resourses we should release on return
       A single goto label is much easier to handle than a bunch of gotos ;-). */
    unsigned resources = 0;

    memset(&utb, 0, sizeof(utb));
    return_status = FILE_RETRY;

    if (show_source (src_path) == FILE_ABORT
	|| show_target (dst_path) == FILE_ABORT)
	return FILE_ABORT;
    mc_refresh ();

 retry_dst_stat:
    if (mc_stat (dst_path, &sb2) == 0) {
	if (S_ISDIR (sb2.st_mode)) {
	    return_status = file_error (" Cannot overwrite directory \"%s\" \n %s ",
					dst_path);
	    if (return_status == FILE_RETRY)
		goto retry_dst_stat;
	    return return_status;
	}
	dst_exists = 1;
    }

 retry_src_xstat:
    if ((*mc_xstat)(src_path, &sb)) {
	return_status = file_error (" Cannot stat source file \"%s\" \n %s ",
				    src_path);
	if (return_status == FILE_RETRY)
	    goto retry_src_xstat;
	return return_status;
    }

    if (dst_exists) {
	    /* .ado: For OS/2 or NT: no st_ino exists, it is better to just try to
	     * overwrite the target file
	     */
	/* Destination already exists */
	if (sb.st_dev == sb2.st_dev && sb.st_ino == sb2.st_ino) {
	    message_3s (1, MSG_ERROR, " `%s' and `%s' are the same file. ",
		     src_path, dst_path);
	    do_refresh ();
	    return FILE_SKIP;
	}

	/* Should we replace destination? */
	if (ask_overwrite) {
	    if (vfs_file_is_ftp (src_path))
		do_reget = -1;
	    else
		do_reget = 0;

	    return_status = query_replace (dst_path, &sb, &sb2);
	    if (return_status != FILE_CONT)
		return return_status;
	}
    }

    if (!do_append) {
       /* .ado: OS2 and NT don't have hardlinks */
	/* Check the hardlinks */
	if (!op_follow_links && sb.st_nlink > 1 &&
	     check_hardlinks (src_path, dst_path, &sb) == 1) {
	    /* We have made a hardlink - no more processing is necessary */
	    return return_status;
	}

	if (S_ISLNK (sb.st_mode))
	    return make_symlink (src_path, dst_path);

	if (S_ISCHR (sb.st_mode) || S_ISBLK (sb.st_mode) || S_ISFIFO (sb.st_mode)
	    || S_ISSOCK (sb.st_mode)) {
	retry_mknod:
	    if (mc_mknod (dst_path, sb.st_mode & umask_kill, sb.st_rdev) < 0) {
		return_status = file_error
		    (" Cannot create special file \"%s\" \n %s ", dst_path);
		if (return_status == FILE_RETRY)
		    goto retry_mknod;
		return return_status;
	    }
	    /* Success */

	retry_mknod_uidgid:
	    if (preserve_uidgid && mc_chown (dst_path, sb.st_uid, sb.st_gid)) {
		temp_status = file_error
		    (" Cannot chown target file \"%s\" \n %s ", dst_path);
		if (temp_status == FILE_RETRY)
		    goto retry_mknod_uidgid;
		return temp_status;
	    }
	retry_mknod_chmod:
	    if (preserve && mc_chmod (dst_path, sb.st_mode & umask_kill) < 0) {
		temp_status = file_error (" Cannnot chmod target file \"%s\" \n %s ", dst_path);
		if (temp_status == FILE_RETRY)
		    goto retry_mknod_chmod;
		return temp_status;
	    }
	    return FILE_CONT;
	}
    }

    if (!do_append && !vfs_file_is_local (src_path) && vfs_file_is_local (dst_path)) {
	mc_setctl (src_path, MCCTL_SETREMOTECOPY, dst_path);
    }
 retry_src_open:
    if ((source_desc = mc_open (src_path, O_RDONLY)) < 0) {
	return_status = file_error
	    (" Cannot open source file \"%s\" \n %s ", src_path);
	if (return_status == FILE_RETRY)
	    goto retry_src_open;
	do_append = 0;
	return return_status;
    }

    resources |= 1;
    do_remote_copy = mc_ctl (source_desc, MCCTL_ISREMOTECOPY, 0);

    if (!do_remote_copy) {
 retry_src_fstat:
	if (mc_fstat (source_desc, &sb)) {
	    return_status = file_error
		(" Cannot fstat source file \"%s\" \n %s ", src_path);
	    if (return_status == FILE_RETRY)
		goto retry_src_fstat;
	    do_append = 0;
	    goto ret;
	}
    }
    src_mode = sb.st_mode;
    src_uid = sb.st_uid;
    src_gid = sb.st_gid;
    utb.actime = sb.st_atime;
    utb.modtime = sb.st_mtime;
    file_size = sb.st_size;

    /* Create the new regular file with small permissions initially,
       do not create a security hole.  */

    if (!do_remote_copy) {
 retry_dst_open:
	if ((do_append &&
	    (dest_desc = mc_open (dst_path, O_WRONLY | O_APPEND)) < 0) ||
	    (!do_append &&
	    (dest_desc = mc_open (dst_path, O_WRONLY | O_CREAT | O_TRUNC, 0600)) < 0)) {
	    return_status = file_error
		(" Cannot create target file \"%s\" \n %s ", dst_path);
	    if (return_status == FILE_RETRY)
		goto retry_dst_open;
	    do_append = 0;
	    goto ret;
	}
	resources |= 2; /* dst_path exists/dst_path opened */
	resources |= 4; /* remove short file */
    }
    appending = do_append;
    do_append = 0;

    if (!do_remote_copy) {
 retry_dst_fstat:
	/* Find out the optimal buffer size.  */
	if (mc_fstat (dest_desc, &sb)) {
	    return_status = file_error
		(" Cannot fstat target file \"%s\" \n %s ", dst_path);
	    if (return_status == FILE_RETRY)
		goto retry_dst_fstat;
	    goto ret;
	}
	buf_size = 8*1024;

	buf = (char *) x_malloc (buf_size, "copy_file_file");
    }

    return_status = show_file_progress (0, file_size);
    mc_refresh ();
    if (return_status != FILE_CONT)
	goto ret;

    if (!do_remote_copy) {
	for (;;) {
    retry_src_read:
	    n_read = mc_read (source_desc, buf, buf_size);
	    if (n_read < 0) {
		return_status = file_error
		    (" Cannot read source file \"%s\" \n %s ", src_path);
		if (return_status == FILE_RETRY)
		    goto retry_src_read;
		goto ret;
	    }
	    if (n_read == 0)
		break;

	    n_read_total += n_read;

    retry_dst_write:
	    n_written = mc_write (dest_desc, buf, n_read);
	    if (n_written < n_read) {
		return_status = file_error
		    (" Cannot write target file \"%s\" \n %s ", dst_path);
		if (return_status == FILE_RETRY)
		    goto retry_dst_write;
		goto ret;
	    }
	    return_status = show_file_progress (n_read_total, file_size);
	    mc_refresh ();
	    if (return_status != FILE_CONT)
		goto ret;
	}
    } else {
	struct timeval tv_current;
	struct timeval tv_transfer_start;
	struct timeval tv_last_update;
	struct timeval tv_last_input;
	int    i, size, secs, update_secs;
	long   dt;
	char   *stalled_msg;

	gettimeofday (&tv_transfer_start, (struct timezone *) NULL);
	tv_last_update = tv_transfer_start;
	eta_secs = 0.0;

	for (i = 1; i;) {
	    switch (size = mc_ctl (source_desc, MCCTL_REMOTECOPYCHUNK, 8192)) {
		case MCERR_TARGETOPEN:
		    message_1s (1, MSG_ERROR, " Can't open target file ");
		    goto ret;
		case MCERR_READ:
		    goto ret;
		case MCERR_WRITE:
		    message_1s (1, MSG_ERROR, " Can't write to local target file ");
		    goto ret;
		case MCERR_DATA_ON_STDIN:
		    break;
		case MCERR_FINISH:
		    resources |= 8;
		    i = 0;
		    break;
	    }

	    /* the first time we reach this line the target file has been created
	       or truncated and we actually have a short target file.
	       Do we really want to delete the target file when the ftp transfer
	       fails? If we don't delete it we would be able to use reget later.
	       (Norbert) */
	    resources |= 4; /* remove short file */

	    if (i && size != MCERR_DATA_ON_STDIN) {
		n_read_total += size;

		/* Windows NT ftp servers report that files have no
		 * permissions: -------, so if we happen to have actually
		 * read something, we should fix the permissions.
		 */
		if (!(src_mode &
		      ((S_IRUSR|S_IWUSR|S_IXUSR)    /* user */
		       |(S_IXOTH|S_IWOTH|S_IROTH)  /* other */
		       |(S_IXGRP|S_IWGRP|S_IRGRP)))) /* group */
		    src_mode = S_IRUSR|S_IWUSR|S_IROTH|S_IRGRP;

		gettimeofday (&tv_last_input, NULL);
	    }
	    /* Timed operations: */
	    gettimeofday (&tv_current, NULL);

	    /* 1. Update rotating dash after some time (hardcoded to 2 seconds) */
	    secs = (tv_current.tv_sec - tv_last_update.tv_sec);
	    if (secs > 2) {
		rotate_dash ();
		tv_last_update = tv_current;
	    }

	    /* 2. Check for a stalled condition */
	    update_secs = (tv_current.tv_sec - tv_last_input.tv_sec);
	    stalled_msg = "";
	    if (update_secs > 4) {
		stalled_msg = "(stalled)";
	    }

	    /* 3. Compute ETA */
	    if (secs > 2 || eta_secs == 0.0) {
		dt = (tv_current.tv_sec - tv_transfer_start.tv_sec);

		if (n_read_total) {
		    eta_secs = ((dt / (double) n_read_total) * file_size) - dt;
		    bps = n_read_total / ((dt < 1) ? 1 : dt);
		} else
		    eta_secs = 0.0;
	    }

	    /* 4. Compute BPS rate */
	    if (secs > 2) {
		bps_time = (tv_current.tv_sec - tv_transfer_start.tv_sec);
		if (bps_time < 1)
		    bps_time = 1;
		bps = n_read_total / bps_time;
	    }

	    label_set_text (stalled_label, stalled_msg);

	    return_status = show_file_progress (n_read_total, file_size);
	    mc_refresh ();
	    if (return_status != FILE_CONT)
		goto ret;
	}
    }
    resources &= ~4; /* copy successful, don't remove target file */

ret:
    if (buf)
	free (buf);

 retry_src_close:
    if ((resources & 1) && mc_close (source_desc) < 0) {
	temp_status = file_error
	    (" Cannot close source file \"%s\" \n %s ", src_path);
	if (temp_status == FILE_RETRY)
	    goto retry_src_close;
	if (temp_status == FILE_ABORT)
	    return_status = temp_status;
    }

 retry_dst_close:
    if ((resources & 2) && mc_close (dest_desc) < 0) {
	temp_status = file_error
	    (" Cannot close target file \"%s\" \n %s ", dst_path);
	if (temp_status == FILE_RETRY)
	    goto retry_dst_close;
	return_status = temp_status;
    }

    if (resources & 4) {
	/* Remove short file */
	mc_unlink (dst_path);
#ifdef USE_VFS
	if (do_remote_copy) {
	    mc_ctl (source_desc, MCCTL_FINISHREMOTE, -1);
	}
#endif
    } else if (resources & (2|8)) {
	/* no short file and destination file exists */
	if (!appending && preserve_uidgid) {
	 retry_dst_chown:
	    if (mc_chown (dst_path, src_uid, src_gid)) {
		temp_status = file_error
		    (" Cannot chown target file \"%s\" \n %s ", dst_path);
		if (temp_status == FILE_RETRY)
		    goto retry_dst_chown;
		return_status = temp_status;
	    }
	}

     /* .ado: according to the XPG4 standard, the file must be closed before
      * chmod can be invoked
      */
     retry_dst_chmod:
	if (!appending && mc_chmod (dst_path, src_mode & umask_kill)) {
	    temp_status = file_error
		(" Cannot chmod target file \"%s\" \n %s ", dst_path);
	    if (temp_status == FILE_RETRY)
		goto retry_dst_chmod;
	    return_status = temp_status;
	}

	if (!appending && preserve)
	    mc_utime (dst_path, &utb);
    }
    return return_status;
}

/*
 * I think these copy_*_* functions should have a return type.
 * anyway, this function *must* have two directories as arguments.
 */
/* FIXME: This function needs to check the return values of the
   function calls */
int
copy_dir_dir (char *s, char *d, int toplevel, int move_over, int delete,
	      struct link *parent_dirs)
{
#ifdef __os2__
    DIR    *next;
#else
    struct dirent *next;
#endif
    struct stat   buf, cbuf;
    DIR    *reading;
    char   *path, *mdpath, *dest_file, *dest_dir;
    int    return_status = FILE_CONT;
    struct utimbuf utb;
    struct link *lp;

    /* First get the mode of the source dir */
 retry_src_stat:
    if ((*mc_xstat) (s, &cbuf)) {
	return_status = file_error (" Cannot stat source directory \"%s\" \n %s ", s);
	if (return_status == FILE_RETRY)
	    goto retry_src_stat;
	return return_status;
    }

    if (is_in_linklist (dest_dirs, s, &cbuf)) {
	/* Don't copy a directory we created before (we don't want to copy
	   infinitely if a directory is copied into itself) */
	/* FIXME: should there be an error message and FILE_SKIP? - Norbert */
	return FILE_CONT;
    }

/* Hmm, hardlink to directory??? - Norbert */
/* FIXME: In this step we should do something
   in case the destination already exist */
    /* Check the hardlinks */
    if (preserve && cbuf.st_nlink > 1 && check_hardlinks (s, d, &cbuf) == 1) {
	/* We have made a hardlink - no more processing is necessary */
	return return_status;
    }

    if (!S_ISDIR (cbuf.st_mode)) {
	return_status = file_error (" Source directory \"%s\" is not a directory \n %s ", s);
	if (return_status == FILE_RETRY)
	    goto retry_src_stat;
	return return_status;
    }

    if (is_in_linklist (parent_dirs, s, &cbuf)) {
	/* we found a cyclic symbolic link */
	   message_2s (1, MSG_ERROR, " Cannot copy cyclic symbolic link \n `%s' ", s);
	return FILE_SKIP;
    }

    lp = x_malloc (sizeof (struct link), "parent_dirs");
    lp->vfs = vfs_type (s);
    lp->ino = cbuf.st_ino;
    lp->dev = cbuf.st_dev;
    lp->next = parent_dirs;
    parent_dirs = lp;

    /* Now, check if the dest dir exists, if not, create it. */
    if (mc_stat (d, &buf)) {
	/* Here the dir doesn't exist : make it !*/

	if (move_over) {
	    if (mc_rename (s, d) == 0) {
		free (parent_dirs);
		return FILE_CONT;
	    }
	}
	dest_dir = xstrdup (d);
    } else {
	/*
	 * If the destination directory exists, we want to copy the whole
	 * directory, but we only want this to happen once.
	 *
	 * Escape sequences added to the * to avoid compiler warnings.
	 * so, say /bla exists, if we copy /tmp/\* to /bla, we get /bla/tmp/\*
	 * or ( /bla doesn't exist )       /tmp/\* to /bla     ->  /bla/\*
	 */
#if 1
/* Again, I'm getting curious. Is not d already what we wanted, incl.
 *  masked source basename? Is not this just a relict of the past versions?
 *  I'm afraid this will lead into a two level deep dive :(
 *
 * I think this is indeed the problem.  I can not remember any case where
 * we actually would like that behaviour -miguel
 *
 * It's a documented feature (option `Dive into subdir if exists' in the
 * copy/move dialog). -Norbert
 */
	if (toplevel && dive_into_subdirs) {
	    dest_dir = concat_path_file (d, bb_basename (s));
	} else
#endif
	{
	    dest_dir = xstrdup (d);
	    goto dont_mkdir;
	}
    }
 retry_dst_mkdir:
    if (my_mkdir (dest_dir, (cbuf.st_mode & umask_kill) | S_IRWXU)) {
	return_status = file_error (" Cannot create target directory \"%s\" \n %s ", dest_dir);
	if (return_status == FILE_RETRY)
	    goto retry_dst_mkdir;
	goto ret;
    }

    lp = x_malloc (sizeof (struct link), "dest_dirs");
    mc_stat (dest_dir, &buf);
    lp->vfs = vfs_type (dest_dir);
    lp->ino = buf.st_ino;
    lp->dev = buf.st_dev;
    lp->next = dest_dirs;
    dest_dirs = lp;

    if (preserve_uidgid) {
     retry_dst_chown:
	if (mc_chown (dest_dir, cbuf.st_uid, cbuf.st_gid)) {
	    return_status = file_error
		(" Cannot chown target directory \"%s\" \n %s ", dest_dir);
	    if (return_status == FILE_RETRY)
		goto retry_dst_chown;
	    goto ret;
	}
    }

 dont_mkdir:
    /* open the source dir for reading */
    if ((reading = mc_opendir (s)) == 0) {
	goto ret;
    }

    while ((next = mc_readdir (reading)) && return_status != FILE_ABORT) {
	/* get the filename and add it to the src directory */
	path = concat_subpath_file (s, next->d_name);
	if(path == NULL)
		continue;

	(*mc_xstat)(path, &buf);
	if (S_ISDIR (buf.st_mode)) {
	    mdpath = concat_path_file (dest_dir, next->d_name);
	    /*
	     * From here, we just intend to recursively copy subdirs, not
	     * the double functionality of copying different when the target
	     * dir already exists. So, we give the recursive call the flag 0
	     * meaning no toplevel.
	     */
	    return_status = copy_dir_dir (path, mdpath, 0, 0, delete, parent_dirs);
	    free (mdpath);
	} else {
	    dest_file = concat_path_file (dest_dir, bb_basename (path));
	    return_status = copy_file_file (path, dest_file, 1);
	    free (dest_file);
	}
	if (delete && return_status == FILE_CONT) {
		lp = x_malloc (sizeof (struct link) + strlen (path), "erase_list");
		strcpy (lp->name, path);
		lp->st_mode = buf.st_mode;
		lp->next = 0;
		if (erase_list) {
		   cdd_tail->next = lp;
		   cdd_tail = lp;
		} else
		   erase_list = cdd_tail = lp;
	}

#ifdef __os2__
	/* The OS/2 mc_readdir returns a block of memory DIR
	 * next should be freed: .ado
	 */
	if (!next)
		free (next);
#endif
	free (path);
    }
    mc_closedir (reading);

    /* .ado: Directories can not have permission set in OS/2 */
#ifndef __os2__
    if (preserve) {
	mc_chmod (dest_dir, cbuf.st_mode & umask_kill);
	utb.actime = cbuf.st_atime;
	utb.modtime = cbuf.st_mtime;
	mc_utime(dest_dir, &utb);
    }

#endif
ret:
    free (dest_dir);
    free (parent_dirs);
    return return_status;
}

/* }}} */

/* {{{ Move routines */

int
move_file_file (char *s, char *d)
{
    struct stat src_stats, dst_stats;
    int return_status = FILE_CONT;

    if (show_source (s) == FILE_ABORT
	|| show_target (d) == FILE_ABORT)
	return FILE_ABORT;

    mc_refresh ();

 retry_src_lstat:
    if (mc_lstat (s, &src_stats) != 0) {
	/* Source doesn't exist */
	return_status = file_error (" Cannot stat file \"%s\" \n %s ", s);
	if (return_status == FILE_RETRY)
	    goto retry_src_lstat;
	return return_status;
    }

    if (mc_lstat (d, &dst_stats) == 0) {
	/* Destination already exists */
	/* .ado: for OS/2 and NT, no st_ino exists */
	if (src_stats.st_dev == dst_stats.st_dev
	    && src_stats.st_ino == dst_stats.st_ino) {
	    int msize = COLS - 36;
	    char st[MC_MAXPATHLEN];
	    char dt[MC_MAXPATHLEN];

	    if (msize < 0)
		msize = 40;
	    msize /= 2;

	    strcpy (st, name_trunc (s, msize));
	    strcpy (dt, name_trunc (d, msize));
	    message_3s (1, MSG_ERROR, " `%s' and `%s' are the same file ",
		     st, dt );
	    do_refresh ();
	    return FILE_SKIP;
	}
	if (S_ISDIR (dst_stats.st_mode)) {
	    message_2s (1, MSG_ERROR, " Cannot overwrite directory `%s' ", d);
	    do_refresh ();
	    return FILE_SKIP;
	}

	if (confirm_overwrite) {
	    if (vfs_file_is_ftp (s))
		do_reget = -1;
	    else
		do_reget = 0;

	    return_status = query_replace (d, &src_stats, &dst_stats);
	    if (return_status != FILE_CONT)
		return return_status;
	}
	/* Ok to overwrite */
    }
    if (!do_append) {
	if (S_ISLNK (src_stats.st_mode) && stable_symlinks) {
	    if ((return_status = make_symlink (s, d)) == FILE_CONT)
		goto retry_src_remove;
	    else
		return return_status;
	}

	if (mc_rename (s, d) == 0)
	    return FILE_CONT;
    }

    /* Failed because filesystem boundary -> copy the file instead */
    if ((return_status = copy_file_file (s, d, 0)) != FILE_CONT)
	return return_status;
    if ((return_status = show_source (NULL)) != FILE_CONT
	|| (return_status = show_file_progress (0, 0)) != FILE_CONT)
	return return_status;

    mc_refresh ();

 retry_src_remove:
    if (mc_unlink (s)) {
	return_status = file_error (" Cannot remove file \"%s\" \n %s ", s);
	if (return_status == FILE_RETRY)
	    goto retry_src_remove;
	return return_status;
    }

    return FILE_CONT;
}

int
move_dir_dir (char *s, char *d)
{
    struct stat sbuf, dbuf, destbuf;
    struct link *lp;
    char *destdir;
    int return_status;
    int move_over = 0;

    if (show_source (s) == FILE_ABORT
	|| show_target (d) == FILE_ABORT)
	return FILE_ABORT;

    mc_refresh ();

    mc_stat (s, &sbuf);
    if (mc_stat (d, &dbuf))
	destdir = xstrdup (d);       /* destination doesn't exist */
    else if (!dive_into_subdirs) {
	destdir = xstrdup (d);
	move_over = 1;
    } else
	destdir = concat_path_file (d, bb_basename (s));

    /* Check if the user inputted an existing dir */
 retry_dst_stat:
    if (!mc_stat (destdir, &destbuf)) {
	if (move_over) {
	    if ((return_status = copy_dir_dir (s, destdir, 0, 1, 1, 0)) != FILE_CONT)
		goto ret;
	    goto oktoret;
	} else {
	    if (S_ISDIR (destbuf.st_mode))
		return_status = file_error (" Cannot overwrite directory \"%s\" %s ", destdir);
	    else
		return_status = file_error (" Cannot overwrite file \"%s\" %s ", destdir);
	    if (return_status == FILE_RETRY)
		goto retry_dst_stat;
	}
	free (destdir);
	return return_status;
    }

 retry_rename:
    if (mc_rename (s, destdir) == 0) {
	return_status = FILE_CONT;
	goto ret;
    }
/* .ado: Drive, Do we need this anymore? */
#ifdef WIN32
	else {
	/* EXDEV: cross device; does not work everywhere */
		if (toupper(s[0]) != toupper(destdir[0]))
			goto w32try;
	}
#endif

    if (errno != EXDEV) {
	return_status = files_error (" Cannot move directory \"%s\" to \"%s\" \n %s ", s, d);
	if (return_status == FILE_RETRY)
	    goto retry_rename;
	goto ret;
    }

#ifdef WIN32
w32try:
#endif
    /* Failed because of filesystem boundary -> copy dir instead */
    if ((return_status = copy_dir_dir (s, destdir, 0, 0, 1, 0)) != FILE_CONT)
	goto ret;
oktoret:
    if ((return_status = show_source (NULL)) != FILE_CONT
	|| (return_status = show_file_progress (0, 0)) != FILE_CONT)
	goto ret;

    mc_refresh ();
    for ( ; erase_list && return_status != FILE_ABORT; ) {
	    if (S_ISDIR (erase_list->st_mode)) {
		return_status = erase_dir_iff_empty (erase_list->name);
	    } else
		return_status = erase_file (erase_list->name);
	    lp = erase_list;
	    erase_list = erase_list->next;
	    free (lp);
    }
    erase_dir_iff_empty (s);

 ret:
    free (destdir);
    for ( ; erase_list; ) {
	lp = erase_list;
	erase_list = erase_list->next;
	free (lp);
    }
    return return_status;
}

/* }}} */

/* {{{ Erase routines */

static int
erase_file (char *s)
{
    int return_status;

    if (show_deleting (s) == FILE_ABORT)
	return FILE_ABORT;

    mc_refresh ();

 retry_unlink:
    if (mc_unlink (s)) {
	return_status = file_error (" Cannot delete file \"%s\" \n %s ", s);
	if (return_status == FILE_RETRY)
	    goto retry_unlink;
	return return_status;
    }
    return FILE_CONT;
}

static int
recursive_erase (char *s)
{
    struct dirent *next;
    struct stat buf;
    DIR    *reading;
    char   *path;
    int    return_status = FILE_CONT;

    if (!strcmp (s, ".."))
	return 1;

    reading = mc_opendir (s);

    if (!reading)
	return 1;

    while ((next = mc_readdir (reading)) && return_status == FILE_CONT) {
	path = concat_subpath_file (s, next->d_name);
	if(path == NULL)
	    continue;
	if (mc_lstat (path, &buf)) {
	    free (path);
	    return 1;
	}
	if (S_ISDIR (buf.st_mode))
	    return_status = (recursive_erase (path) != FILE_CONT);
	else
	    return_status = erase_file (path);
	free (path);
	/* .ado: OS/2 returns a block of memory DIR to next and must be freed */
#ifdef __os2__
	if (!next)
	    free (next);
#endif
    }
    mc_closedir (reading);
    if (return_status != FILE_CONT)
	return return_status;
    if (show_deleting (s) == FILE_ABORT)
	return FILE_ABORT;
    mc_refresh ();
 retry_rmdir:
    if (my_rmdir (s)) {
	return_status = file_error (" Cannot remove directory \"%s\" \n %s ", s);
	if (return_status == FILE_RETRY)
	    goto retry_rmdir;
	return return_status;
    }
    return FILE_CONT;
}

/* Return -1 on error, 1 if there are no entries besides "." and ".."
   in the directory path points to, 0 else. */
static int
check_dir_is_empty(char *path)
{
    DIR *dir;
    struct dirent *d;
    int i;

    dir = mc_opendir (path);
    if (!dir)
	return -1;

    for (i = 1, d = mc_readdir (dir); d; d = mc_readdir (dir)) {
	if (d->d_name[0] == '.' && (d->d_name[1] == '\0' ||
	    (d->d_name[1] == '.' && d->d_name[2] == '\0')))
	    continue; /* "." or ".." */
	i = 0;
	break;
    }

    mc_closedir (dir);
    return i;
}

int
erase_dir (char *s)
{
    int error;

    if (s[0] == '.' && (s[1] == '\0' || (s[1] == '.' && s[2] == '\0')))
	return FILE_SKIP;
    if (show_deleting (s) == FILE_ABORT)
	return FILE_ABORT;
    mc_refresh ();

    /* The old way to detect a non empty directory was:
	    error = my_rmdir (s);
	    if (error && (errno == ENOTEMPTY || errno == EEXIST))) {
       For the linux user space nfs server (nfs-server-2.2beta29-2)
       we would have to check also for EIO. I hope the new way is
       fool proof. (Norbert)
     */
    error = check_dir_is_empty (s);
    if (error == 0) { /* not empty */
	error = query_recursive (s);
	if (error == FILE_CONT)
	    return recursive_erase (s);
	else
	    return error;
    }

 retry_rmdir:
    error = my_rmdir (s);
    if (error == -1) {
	error = file_error (" Cannot remove directory \"%s\" \n %s ", s);
	if (error == FILE_RETRY)
	    goto retry_rmdir;
	return error;
    }
    return FILE_CONT;
}

int
erase_dir_iff_empty (char *s)
{
    int error;

    if (s[0] == '.' && (s[1] == '\0' || (s[1] == '.' && s[2] == '\0')))
	return FILE_SKIP;

    if (show_deleting (s) == FILE_ABORT)
	return FILE_ABORT;
    mc_refresh ();

    if (1 != check_dir_is_empty (s)) /* not empty or error */
	return FILE_CONT;

 retry_rmdir:
    error = my_rmdir (s);
    if (error) {
	error = file_error (" Cannot remove directory \"%s\" \n %s ", s);
	if (error == FILE_RETRY)
	    goto retry_rmdir;
	return error;
    }
    return FILE_CONT;
}

/* }}} */

/* {{{ Panel operate routines */

/* Returns currently selected file or the first marked file if there is one */
static char *
get_file (WPanel *panel, struct stat *stat_buf)
{
    int i;

    /* No problem with Gnome, as get_current_type never returns view_tree there */
    if (get_current_type () == view_tree) {
	WTree *tree = (WTree *)get_panel_widget (get_current_index ());

	mc_stat (tree->selected_ptr->name, stat_buf);
	return tree->selected_ptr->name;
    }

    if (panel->marked) {
	for (i = 0; i < panel->count; i++)
	    if (panel->dir.list [i].f.marked) {
		*stat_buf = panel->dir.list [i].buf;
		return panel->dir.list [i].fname;
	    }
    } else {
	*stat_buf = panel->dir.list [panel->selected].buf;
	return panel->dir.list [panel->selected].fname;
    }
    fprintf (stderr, " Internal error: get_file \n");
    mi_getch ();
    return "";
}

static int
is_wildcarded (char *p)
{
    for (; *p; p++) {
	if (*p == '*')
	    return 1;
	else if (*p == '\\' && p [1] >= '1' && p [1] <= '9')
	    return 1;
    }
    return 0;
}

/* Sets all global variables used by copy_file_file/move_file_file to a
   resonable default
   (file_mask_dialog sets these global variables interactively)
 */
static void
file_mask_defaults (void)
{
    stable_symlinks = 0;
    op_follow_links = 0;
    dive_into_subdirs = 0;
    mc_xstat = mc_lstat;

    preserve = 1;
    umask_kill = 0777777;
    preserve_uidgid = (my_euid == 0) ? 1 : 0;
}

#define FMDY 13
#define FMD_XLEN 64

static const QuickWidget fmd_widgets_st [] = {

#define FMCB0  FMDC
#define FMCB12 0
#define FMCB11 1
	/* follow symlinks and preserve Attributes must be the first */
    { quick_checkbox, 3, 64, 8, FMDY, "preserve &Attributes", 9, 0,
      &op_preserve, 0, XV_WLAY_BELOWCLOSE, "preserve" },
    { quick_checkbox, 3, 64, 7, FMDY, "follow &Links", 7, 0,
      &op_follow_links, 0, XV_WLAY_BELOWCLOSE, "follow" },
    { quick_label, 3, 64, 5, FMDY, "to:", 0, 0, 0, 0, XV_WLAY_BELOWOF,"to"},
    { quick_checkbox, 37, 64, 4, FMDY, "&Using shell patterns", 0, 0,
      0/* &source_easy_patterns */, 0, XV_WLAY_BELOWCLOSE, "using-shell" },
    { quick_input, 3, 64, 3, FMDY, "", 58,
      0, 0, 0, XV_WLAY_BELOWCLOSE, "input-def" },
#define FMDI1 4
#define FMDI2 5
#define FMDC 3
    { quick_input, 3, 64, 6, FMDY, "", 58, 0,
      0, 0, XV_WLAY_BELOWCLOSE, "input2" },
#define FMDI0 6
    { quick_label, 3, 64, 2, FMDY, "", 0, 0, 0, 0, XV_WLAY_DONTCARE, "ql" },
#define FMBRGT 7
    { quick_button, 42, 64, 9, FMDY, "&Cancel", 0, B_CANCEL, 0, 0, XV_WLAY_DONTCARE,
	  "cancel" },
#undef SKIP
#define SKIP 4
#define FMCB21 10
#define FMCB22 9
#define FMBLFT 8
#undef  FMBMID
    { quick_button, 14, 64, 9, FMDY, "&Ok", 0, B_ENTER, 0, 0, XV_WLAY_NEXTROW, "ok" },
    { quick_checkbox, 42, 64, 8, FMDY, "&Stable Symlinks", 0, 0,
      &stable_symlinks, 0, XV_WLAY_BELOWCLOSE, "stab-sym" },
    { quick_checkbox, 31, 64, 7, FMDY, "&Dive into subdir if exists", 0, 0,
      &dive_into_subdirs, 0, XV_WLAY_BELOWOF, "dive" },
    { 0 }
};


static char *
file_mask_dialog (int operation, char *text, char *def_text, int only_one, int *do_background)
{
    int source_easy_patterns = easy_patterns;
    char *source_mask, *orig_mask, *dest_dir;
    const char *error;
    struct stat buf;
    int val;
    QuickWidget *fmd_widgets;
    QuickDialog Quick_input;

    fmd_widgets = alloca(sizeof(fmd_widgets_st));
    memcpy(fmd_widgets, fmd_widgets_st, sizeof(fmd_widgets_st));

    stable_symlinks = 0;
    fmd_widgets [FMDC].result = &source_easy_patterns;
    fmd_widgets [FMDI1].text = easy_patterns ? "*" : "^\\(.*\\)$";
    Quick_input.xlen  = fmd_xlen;
    Quick_input.xpos  = -1;
    Quick_input.title = op_names [operation];
    Quick_input.help  = "[Mask Copy/Rename]";
    Quick_input.ylen  = FMDY;

    if (operation == OP_COPY) {
	Quick_input.widgets = fmd_widgets;
    } else { /* operation == OP_MOVE */
	Quick_input.widgets = fmd_widgets + 2;
    }
    fmd_widgets [FMDI0].text = text;
    fmd_widgets [FMDI2].text = def_text;
    fmd_widgets [FMDI2].str_result = &dest_dir;
    fmd_widgets [FMDI1].str_result = &source_mask;

    *do_background = 0;
ask_file_mask:

    if ((val = quick_dialog_skip (&Quick_input, SKIP)) == B_CANCEL)
	return 0;

    if (op_follow_links && operation != OP_MOVE)
	mc_xstat = mc_stat;
    else
	mc_xstat = mc_lstat;

    if (op_preserve || operation == OP_MOVE) {
	preserve = 1;
	umask_kill = 0777777;
	preserve_uidgid = (my_euid == 0) ? 1 : 0;
    }
    else {
	int i;
	preserve = preserve_uidgid = 0;
	i = umask (0);
	umask (i);
	umask_kill = i ^ 0777777;
    }

    orig_mask = source_mask;
    if (!dest_dir || !*dest_dir) {
	if (source_mask)
	    free (source_mask);
	return dest_dir;
    }
    if (source_easy_patterns) {
	source_easy_patterns = easy_patterns;
	easy_patterns = 1;
	source_mask = convert_pattern (source_mask, match_file, 1);
	easy_patterns = source_easy_patterns;
	error = re_compile_pattern (source_mask, strlen (source_mask), &rx);
	free (source_mask);
    } else
	error = re_compile_pattern (source_mask, strlen (source_mask), &rx);

    if (error) {
	message_3s (1, MSG_ERROR, "Invalid source pattern `%s' \n %s ",
		 orig_mask, error);
	if (orig_mask)
	    free (orig_mask);
	goto ask_file_mask;
    }
    if (orig_mask)
	free (orig_mask);
    dest_mask = (char *)bb_basename (dest_dir);
    orig_mask = dest_mask;
    if (!*dest_mask || (!dive_into_subdirs && !is_wildcarded (dest_mask) &&
			(!only_one || (!mc_stat (dest_dir, &buf) && S_ISDIR (buf.st_mode)))) ||
	(dive_into_subdirs && ((!only_one && !is_wildcarded (dest_mask)) ||
			       (only_one && !mc_stat (dest_dir, &buf) && S_ISDIR (buf.st_mode)))))
	dest_mask = strdup ("*");
    else {
	dest_mask = strdup (dest_mask);
	*orig_mask = 0;
    }
    if (!*dest_dir) {
	free (dest_dir);
	dest_dir = strdup ("./");
    }
    if (val == B_USER)
	*do_background = 1;
    return dest_dir;
}

/*
 * This array introduced to avoid translation problems. The former (op_names)
 * is assumed to be nouns, suitable in dialog box titles; this one should
 * contain whatever is used in prompt itself (i.e. in russian, it's verb).
 * Notice first symbol - it is to fool gettext and force these strings to
 * be different for it. First symbol is skipped while building a prompt.
 * (I don't use spaces around the words, because someday they could be
 * dropped, when widgets get smarter)
 */
static const
char * const op_names1 [] = { "1Copy", "1Move", "1Delete" };

/*
 * These are formats for building a prompt. Parts encoded as follows:
 * %o - operation from op_names1
 * %f - file/files or files/directories, as appropriate
 * %m - "with source mask" or question mark for delete
 * %s - source name (truncated)
 * %d - number of marked files
 */
static const char one_format[]  = "%o %f \"%s\"%m";
static const char many_format[] = "%o %d %f%m";

static const char* const prompt_parts [] =
{
	"file", "files", "directory", "directories",
	"files/directories", " with source mask:"
};

static char*
generate_prompt(char* Cmd_buf, WPanel* panel, int operation, int only_one,
	struct stat* src_stat)
{
	const char *cp;
	const char *sp;
	int i;
	char format_string [200];
	char *dp = format_string;
	char* source = NULL;

	sp = only_one ? one_format : many_format;

	if (only_one)
		source = get_file (panel, src_stat);

	while (*sp) {
		switch (*sp) {
			case '%':
				cp = NULL;
				switch (sp[1]) {
					case 'o':
						cp = op_names1 [operation] + 1;
						break;
					case 'm':
						cp = operation == OP_DELETE ? "?" : prompt_parts [5];
						break;
					case 'f':
						if (only_one) {
							cp = S_ISDIR (src_stat->st_mode) ?
								prompt_parts [2] : prompt_parts [0];
						} else {
							cp = (panel->marked == panel->dirs_marked)
							? prompt_parts [3]
							: (panel->dirs_marked ? prompt_parts [4]
							: prompt_parts [1]);
						}
						break;
					default:
						*dp++ = *sp++;
				}
				if (cp) {
					sp += 2;
					while (*cp)
						*dp++ = *cp++;
				}
				break;
			default:
				*dp++ = *sp++;
		}
	}
	*dp = '\0';

	if (only_one) {
		i = fmd_xlen - strlen(format_string) - 4;
		sprintf (Cmd_buf, format_string, name_trunc (source, i));
	} else {
		sprintf (Cmd_buf, format_string, panel->marked);
		i = strlen (Cmd_buf) + 6 - fmd_xlen;
		if (i > 0)
			fmd_xlen += i;
	}

	return source;
}


/* Returns 1 if did change the directory structure,
   Returns 0 if user aborted */
int
panel_operate (void *source_panel, int operation, char *thedefault)
{
    WPanel *panel = source_panel;
    char *source_with_path = NULL;
    char *source = NULL;
    char *dest = NULL;
    char *temp = NULL;
    int only_one = (get_current_type () == view_tree) || (panel->marked <= 1);
    struct stat src_stat, dst_stat;
    int i, value;
    off_t marked, total;
    off_t count = 0, bytes = 0;
    int  dst_result;
    int  do_bg;                 /* do background operation? */

    do_bg = 0;
    rx.buffer = NULL;
    free_linklist (&linklist);
    free_linklist (&dest_dirs);
    if (get_current_type () == view_listing)
	if (!panel->marked && !strcmp (selection (panel)->fname, "..")) {
	    MC_message (1, MSG_ERROR, " Can't operate on \"..\"! ");
	    return 0;
	}

    if (operation < OP_COPY || operation > OP_DELETE)
	return 0;

    /* Generate confirmation prompt */
    source = generate_prompt(cmd_buf, panel, operation, only_one, &src_stat);

    /* Show confirmation dialog */
    if (operation == OP_DELETE && confirm_delete) {
	if (know_not_what_am_i_doing)
	    query_set_sel (1);
	i = query_dialog (op_names [operation], cmd_buf,
			  D_ERROR, 2, "&Yes", "&No");
	if (i != 0)
	    return 0;
    } else if (operation != OP_DELETE) {
	char *dest_dir;

	if (thedefault != NULL)
	    dest_dir = thedefault;
	else if (get_other_type () == view_listing)
	    dest_dir = opanel->cwd;
	else
	    dest_dir = panel->cwd;

	rx.buffer = (unsigned char *) x_malloc (MC_MAXPATHLEN, "mask copying");
	rx.allocated = MC_MAXPATHLEN;
	rx.translate = 0;
	dest = file_mask_dialog (operation, cmd_buf, dest_dir, only_one, &do_bg);
	if (!dest) {
	    free (rx.buffer);
	    return 0;
	}
	if (!*dest) {
	    free (rx.buffer);
	    free (dest);
	    return 0;
	}
    }

    /* Initialize things */
    /* We turn on ETA display if the source is an ftp file system */
    create_op_win (operation, vfs_file_is_ftp (panel->cwd));
    ftpfs_hint_reread (0);

    /* Now, let's do the job */
    /* This code is only called by the tree and panel code */
    if (only_one) {
	/* One file: FIXME mc_chdir will take user out of any vfs */
	if (operation != OP_COPY && get_current_type () == view_tree)
	    mc_chdir ("/");

	/* The source and src_stat variables have been initialized before */
	source_with_path = concat_path_file (panel->cwd, source);
	if (operation == OP_DELETE) {
	    /* Delete operation */
	    if (S_ISDIR (src_stat.st_mode))
		value = erase_dir (source_with_path);
	    else
		value = erase_file (source_with_path);
	} else {
	    /* Copy or move operation */
	    temp = transform_source (source_with_path);
	    if (temp == NULL) {
		value = transform_error;
	    } else {
		temp = get_full_name (dest, temp);
		free (dest);
		dest = temp;
		temp = 0;

		switch (operation) {
		case OP_COPY:
		    /* we use op_follow_links only with OP_COPY,
		      */
		    (*mc_xstat) (source_with_path, &src_stat);
		    if (S_ISDIR (src_stat.st_mode))
			value = copy_dir_dir (source_with_path, dest, 1, 0, 0, 0);
		    else
			value = copy_file_file (source_with_path, dest, 1);
		    break;
		case OP_MOVE:
		    if (S_ISDIR (src_stat.st_mode))
			value = move_dir_dir (source_with_path, dest);
		    else
			value = move_file_file (source_with_path, dest);
		    break;
		default:
		    value = FILE_CONT;
		    message_1s (1, " Internal failure ", " Unknown file operation ");
		}
	    }
	} /* Copy or move operation */

	if (value == FILE_CONT)
	    unmark_files (panel);

    } else {
	/* Many files */

	if (operation != OP_DELETE) {
	    /* Check destination for copy or move operation */
	retry_many_dst_stat:
	    dst_result = mc_stat (dest, &dst_stat);
	    if (dst_result == 0 && !S_ISDIR (dst_stat.st_mode)) {
		if (file_error (" Destination \"%s\" must be a directory \n %s ", dest) == FILE_RETRY)
		    goto retry_many_dst_stat;
		goto clean_up;
	    }
	}

	/* Initialize variables for progress bars */
	marked = panel->marked;
	total = panel->total;

	/* Loop for every file */
	for (i = 0; i < panel->count; i++) {
	    if (!panel->dir.list [i].f.marked)
		continue;       /* Skip the unmarked ones */
	    source = panel->dir.list [i].fname;
	    src_stat = panel->dir.list [i].buf; /* Inefficient, should we use pointers? */
	    if (source_with_path)
		free (source_with_path);
	    source_with_path = concat_path_file (panel->cwd, source);
	    if (operation == OP_DELETE) {
		/* Delete operation */
		if (S_ISDIR (src_stat.st_mode))
		    value = erase_dir (source_with_path);
		else
		    value = erase_file (source_with_path);
	    } else {
		/* Copy or move operation */
		if (temp)
		    free (temp);
		temp = transform_source (source_with_path);
		if (temp == NULL) {
		    value = transform_error;
		} else {
		    temp = get_full_name (dest, temp);
		    switch (operation) {
		    case OP_COPY:
		       /* we use op_follow_links only with OP_COPY,
			*/
			(*mc_xstat) (source_with_path, &src_stat);
			if (S_ISDIR (src_stat.st_mode))
			    value = copy_dir_dir (source_with_path, temp, 1, 0, 0, 0);
			else
			    value = copy_file_file (source_with_path, temp, 1);
			free_linklist (&dest_dirs);
			break;
		    case OP_MOVE:
			if (S_ISDIR (src_stat.st_mode))
			    value = move_dir_dir (source_with_path, temp);
			else
			    value = move_file_file (source_with_path, temp);
			break;
		    default:
			message_1s (1, " Internal failure ",
				 " Unknown file operation ");
			goto clean_up;
		    }
		}
	    } /* Copy or move operation */

	    if (value == FILE_ABORT)
		goto clean_up;
	    if (value == FILE_CONT) {
		do_file_mark (panel, i, 0);
	    }
	    count ++;
	    if (show_count_progress (count, marked) == FILE_ABORT)
		goto clean_up;
	    bytes += src_stat.st_size;
	    if (MC_verbose &&
		show_bytes_progress (bytes, total) == FILE_ABORT)
		goto clean_up;
	    if (operation != OP_DELETE && MC_verbose
		&& show_file_progress (0, 0) == FILE_ABORT)
		goto clean_up;
	    mc_refresh ();
	} /* Loop for every file */
    } /* Many files */

 clean_up:
    /* Clean up */
    destroy_op_win ();
    ftpfs_hint_reread (1);
    free_linklist (&linklist);
    free_linklist (&dest_dirs);
    if (source_with_path)
	free (source_with_path);
    if (dest)
	free (dest);
    if (temp)
	free (temp);
    if (rx.buffer) {
	free (rx.buffer);
	rx.buffer = NULL;
    }
    if (dest_mask) {
	free (dest_mask);
	dest_mask = NULL;
    }

    return 1;
}

/* }}} */

/* {{{ Query/status report routines */

static int
real_do_file_error (enum OperationMode mode, char *error)
{
    int result;
    char *msg;

    msg = mode == Foreground ? MSG_ERROR : " Background process error ";
    result = query_dialog (msg, error, D_ERROR, 3, "&Skip", "&Retry", "&Abort");

    switch (result) {
    case 0:
	do_refresh ();
	return FILE_SKIP;
    case 1:
	do_refresh ();
	return FILE_RETRY;
    case 2:
    default:
	return FILE_ABORT;
    }
}

/* Report error with one file */
static int
file_error (const char *format, const char *file)
{
    sprintf (cmd_buf, format, name_trunc (file, 30), unix_error_string());
    return do_file_error (cmd_buf);
}

/* Report error with two files */
int
files_error (const char *format, const char *file1, const char *file2)
{
    char nfile1 [16];
    char nfile2 [16];

    strcpy (nfile1, name_trunc (file1, 15));
    strcpy (nfile2, name_trunc (file2, 15));

    sprintf (cmd_buf, format, nfile1, nfile2, unix_error_string());
    return do_file_error (cmd_buf);
}

static int
replace_callback (struct Dlg_head *h, int Id, int Msg)
{
    switch (Msg) {
    case DLG_DRAW:
	dialog_repaint (h, ERROR_COLOR, ERROR_COLOR);
	break;
    }
    return 0;
}

#define X_TRUNC 52

/*
 * FIXME: probably it is better to replace this with quick dialog machinery,
 * but actually I'm not familiar with it and have not much time :(
 *   alex
 */
static const struct
{
	char* text;
	int   ypos, xpos;
	int   value;            /* 0 for labels */
	WLay  layout;
}
rd_widgets [] = {
	{"Target file \"%s\" already exists!", 3,      4,  0,  XV_WLAY_CENTERROW},
	{"&Abort",                  BY + 3, 25, REPLACE_ABORT, XV_WLAY_CENTERROW},
	{"if &Size differs",        BY + 1, 28, REPLACE_SIZE,  XV_WLAY_RIGHTOF},
	{"non&E",                   BY,     47, REPLACE_NEVER, XV_WLAY_RIGHTOF},
	{"&Update",                 BY,     36, REPLACE_UPDATE,XV_WLAY_RIGHTOF},
	{"al&L",                    BY,     28, REPLACE_ALWAYS,XV_WLAY_RIGHTOF},
	{"Overwrite all targets?",  BY,     4,  0,             XV_WLAY_CENTERROW},
	{"&Reget",                  BY - 1, 28, REPLACE_REGET, XV_WLAY_RIGHTOF},
	{"ap&Pend",                 BY - 2, 45, REPLACE_APPEND,XV_WLAY_RIGHTOF},
	{"&No",                     BY - 2, 37, REPLACE_NO,    XV_WLAY_RIGHTOF},
	{"&Yes",                    BY - 2, 28, REPLACE_YES,   XV_WLAY_RIGHTOF},
	{"Overwrite this target?",  BY - 2, 4,  0,             XV_WLAY_CENTERROW},
	{"Target date: %s, size %d",6,      4,  0,             XV_WLAY_CENTERROW},
	{"Source date: %s, size %d",5,      4,  0,             XV_WLAY_CENTERROW}
};

#define ADD_RD_BUTTON(i)\
	add_widgetl (replace_dlg,\
		button_new (rd_widgets [i].ypos, rd_widgets [i].xpos, rd_widgets [i].value,\
		NORMAL_BUTTON, rd_widgets [i].text, 0, 0), \
		rd_widgets [i].layout)

#define ADD_RD_LABEL(i,p1,p2)\
	sprintf (buffer, rd_widgets [i].text, p1, p2);\
	add_widgetl (replace_dlg,\
		label_new (rd_widgets [i].ypos, rd_widgets [i].xpos, buffer), rd_widgets [i].layout)

static void
init_replace (enum OperationMode mode)
{
    char buffer [128];

    replace_colors [0] = ERROR_COLOR;
    replace_colors [1] = COLOR_NORMAL;
    replace_colors [2] = ERROR_COLOR;
    replace_colors [3] = COLOR_NORMAL;

    replace_dlg = create_dlg (0, 0, 16, 60, replace_colors, replace_callback,
			      "[ Replace ]", DLG_CENTER);

    x_set_dialog_title (replace_dlg,
	mode == Foreground ? " File exists " : " Background process: File exists ");

    ADD_RD_LABEL(0, name_trunc (replace_filename, X_TRUNC - strlen (rd_widgets [0].text)), 0 );
    ADD_RD_BUTTON(1);

    ADD_RD_BUTTON(2);
    ADD_RD_BUTTON(3);
    ADD_RD_BUTTON(4);
    ADD_RD_BUTTON(5);
    ADD_RD_LABEL(6,0,0);

    /* "this target..." widgets */
    if (!S_ISDIR (d_stat->st_mode)) {
		if ((do_reget == -1 && d_stat->st_size && s_stat->st_size > d_stat->st_size))
			ADD_RD_BUTTON(7);

		ADD_RD_BUTTON(8);
    }
    ADD_RD_BUTTON(9);
    ADD_RD_BUTTON(10);
    ADD_RD_LABEL(11,0,0);

    ADD_RD_LABEL(12, file_date (d_stat->st_mtime), (int) d_stat->st_size);
    ADD_RD_LABEL(13, file_date (s_stat->st_mtime), (int) s_stat->st_size);
}

static int
real_query_replace (enum OperationMode mode, char *destname, struct stat *_s_stat,
		    struct stat *_d_stat)
{
    if (replace_result < REPLACE_ALWAYS) {
	replace_filename = destname;
	s_stat = _s_stat;
	d_stat = _d_stat;
	init_replace (mode);
	run_dlg (replace_dlg);
	replace_result = replace_dlg->ret_value;
	if (replace_result == B_CANCEL)
	    replace_result = REPLACE_ABORT;
	destroy_dlg (replace_dlg);
    }

    switch (replace_result) {
    case REPLACE_UPDATE:
	do_refresh ();
	if (_s_stat->st_mtime > _d_stat->st_mtime)
	    return FILE_CONT;
	else
	    return FILE_SKIP;

    case REPLACE_SIZE:
	do_refresh ();
	if (_s_stat->st_size == _d_stat->st_size)
	    return FILE_SKIP;
	else
	    return FILE_CONT;

    case REPLACE_REGET:
	do_reget = _d_stat->st_size;

    case REPLACE_APPEND:
	do_append = 1;

    case REPLACE_YES:
    case REPLACE_ALWAYS:
	do_refresh ();
	return FILE_CONT;
    case REPLACE_NO:
    case REPLACE_NEVER:
	do_refresh ();
	return FILE_SKIP;
    case REPLACE_ABORT:
    default:
	return FILE_ABORT;
    }
}

static int
real_query_recursive (enum OperationMode mode, char *s)
{
    char *confirm, *text;

    if (recursive_result < RECURSIVE_ALWAYS) {
	char *msg =
	    mode == Foreground ? "\n   Directory not empty.   \n   Delete it recursively? "
			       : "\n   Background process: Directory not empty \n   Delete it recursively? ";
	text = xasprintf (" Delete: %s ", name_trunc (s, 30));

	if (know_not_what_am_i_doing)
	    query_set_sel (1);
	recursive_result = query_dialog (text, msg, D_ERROR, 5,
				     "&Yes", "&No", "a&ll", "non&E", "&Abort");


	if (recursive_result != RECURSIVE_ABORT)
	    do_refresh ();
	free (text);
	if (know_not_what_am_i_doing && (recursive_result == RECURSIVE_YES
	    || recursive_result == RECURSIVE_ALWAYS)) {
	    text = xasprintf (" Type 'yes' if you REALLY want to delete %s ",
				 recursive_result == RECURSIVE_YES
				 ? name_trunc (s, 19) : "all the directories ");
	    confirm = input_dialog (mode == Foreground ? " Recursive Delete "
						       : " Background process: Recursive Delete ",
				    text, "no");
	    do_refresh ();
	    if (!confirm || strcmp (confirm, "yes"))
		recursive_result = RECURSIVE_NEVER;
	    free (confirm);
	    free (text);
	}
    }
    switch (recursive_result) {
    case RECURSIVE_YES:
    case RECURSIVE_ALWAYS:
	return FILE_CONT;
    case RECURSIVE_NO:
    case RECURSIVE_NEVER:
	return FILE_SKIP;
    case RECURSIVE_ABORT:
    default:
	return FILE_ABORT;
    }
}

static int
do_file_error (char *str)
{
    return real_do_file_error (Foreground, str);
}

static int
query_recursive (char *s)
{
    return real_query_recursive (Foreground, s);
}

static int
query_replace (char *destname, struct stat *_s_stat, struct stat *_d_stat)
{
    return real_query_replace (Foreground, destname, _s_stat, _d_stat);
}

/* Find file command for the Midnight Commander */


/* Size of the find window */
#define FIND2_Y LINES-4

#define FIND2_X_USE FIND2_X-20

/* A couple of extra messages we need */
enum {
    FIND_B_STOP = B_USER + 1,
    B_AGAIN,
    B_PANELIZE,
    B_TREE,
    B_VIEW
};

/* For nice updating */
static const char rotating_dash[] = "|/-\\";

/*
 * find_parameters: gets information from the user
 *
 * If the return value is true, then the following holds:
 *
 * START_DIR and PATTERN are pointers to char * and upon return they
 * contain the information provided by the user.
 *
 * CONTENT holds a strdup of the contents specified by the user if he
 * asked for them or 0 if not (note, this is different from the
 * behavior for the other two parameters.
 *
 */

static int
find_parameters (char **start_dir, char **pattern, char **content)
{
	int return_value;
	char *temp_dir;

	static const char * const Labs[] = {"Start at:", "Filename:", "Content: "};
	static const char * const buts[] = {"&Ok", "&Tree", "&Cancel"};
	static const int ilen = 30, istart = 14, b0 = 3, b1 = 16, b2 = 36;

find_par_start:
    if (!in_start_dir)
	in_start_dir = strdup (".");
    if (!in_start_name)
	in_start_name = strdup (easy_patterns ? "*" : ".");
    if (!in_contents)
	in_contents = strdup ("");

/* Size of the find parameters window */
#define FIND_Y 12
#define FIND_X 50
    find_dlg = create_dlg (0, 0, FIND_Y, FIND_X, dialog_colors,
			   common_dialog_callback, "[Find File]", DLG_CENTER | DLG_GRID);
    x_set_dialog_title (find_dlg, "Find File");

    add_widgetl (find_dlg, button_new (9, b2, B_CANCEL, NORMAL_BUTTON,
		buts[2], 0 ,0), XV_WLAY_RIGHTOF);
    add_widgetl (find_dlg, button_new (9, b1, B_TREE, NORMAL_BUTTON,
		buts[1], 0, 0), XV_WLAY_RIGHTOF);
    add_widgetl (find_dlg, button_new (9, b0, B_ENTER, DEFPUSH_BUTTON,
		buts[0], 0, 0), XV_WLAY_CENTERROW);

    in_with  = input_new (7, istart, INPUT_COLOR, ilen, in_contents);
    add_widgetl (find_dlg, in_with, XV_WLAY_BELOWOF);

    in_name  = input_new (5, istart, INPUT_COLOR, ilen, in_start_name);
    add_widgetl (find_dlg, in_name, XV_WLAY_BELOWOF);

    in_start = input_new (3, istart, INPUT_COLOR, ilen, in_start_dir);
    add_widgetl (find_dlg, in_start, XV_WLAY_NEXTCOLUMN);

    add_widgetl (find_dlg, label_new (7, 3, Labs[2]), XV_WLAY_BELOWOF);
    add_widgetl (find_dlg, label_new (5, 3, Labs[1]), XV_WLAY_BELOWOF);
    add_widgetl (find_dlg, label_new (3, 3, Labs[0]), XV_WLAY_NEXTCOLUMN);

    run_dlg (find_dlg);
    if (find_dlg->ret_value == B_CANCEL)
	return_value = 0;
    else if (find_dlg->ret_value == B_TREE) {
	temp_dir = strdup (in_start->buffer);
	destroy_dlg (find_dlg);
	free (in_start_dir);
	if (strcmp (temp_dir, ".") == 0) {
	    free (temp_dir);
	    temp_dir = strdup (cpanel->cwd);
	}
	in_start_dir = Tree (temp_dir);
	if (in_start_dir)
	    free (temp_dir);
	else
	    in_start_dir = temp_dir;
	/* Warning: Dreadful goto */
	goto find_par_start;
    } else {
	return_value = 1;
	*start_dir = strdup (in_start->buffer);
	*pattern   = strdup (in_name->buffer);

	free (in_contents);
	if (in_with->buffer [0]) {
	    *content    = strdup (in_with->buffer);
	    in_contents = strdup (*content);
	} else
	    *content = in_contents = NULL;

	free (in_start_dir);
	in_start_dir = strdup (*start_dir);
	free (in_start_name);
	in_start_name = strdup (*pattern);
    }

    destroy_dlg (find_dlg);

    return return_value;
}

static void
push_directory (char *dir)
{
    dir_stack *new;

    new = x_malloc (sizeof (dir_stack), "find: push_directory");
    new->name = strdup (dir);
    new->prev = dir_stack_base;
    dir_stack_base = new;
}

static char*
pop_directory (void)
{
    char *name;
    dir_stack *next;

    if (dir_stack_base) {
	name = dir_stack_base->name;
	next = dir_stack_base->prev;
	free (dir_stack_base);
	dir_stack_base = next;
	return name;
    } else
	return 0;
}

static void
insert_file (char *dir, char *file)
{
    char *tmp_name;
    int i;

    if (dir [0] == '/' && dir [1] == '/')
	dir++;
    i = strlen (dir);
    if (i) {
	if (dir [i - 1] != '/') {
	    dir [i] = '/';
	    dir [i + 1] = 0;
	}
    }

    if (old_dir) {
	if (strcmp (old_dir, dir)) {
	    free (old_dir);
	    old_dir = strdup (dir);
	    if_Dirname = listbox_add_item (Find_list, 0, 0, dir, 0);
	}
    } else {
	old_dir = strdup (dir);
	if_Dirname = listbox_add_item (Find_list, 0, 0, dir, 0);
    }

    tmp_name = xasprintf ("    %s", file);
    listbox_add_item (Find_list, 0, 0, tmp_name, if_Dirname);
    free (tmp_name);
}

/* Used only for display updating, for avoiding line at a time scroll */
static void
listbox_select_last (WListbox *l, int set_top)
{
    if (l->list) {
	l->current = l->list->prev;
	l->pos = l->count - 1;
	if (set_top)
	    l->top = l->list->prev;
    }
}

static void
find_add_match (Dlg_head *h, char *dir, char *file)
{
	int p = ++bb_matches & 7;

	insert_file (dir, file);

	/* Scroll nicely */
	listbox_select_last (Find_list, p == 0);

	/* Updates the current listing */
	send_message (h, &Find_list->widget, WIDGET_DRAW, 0);
	if (p == 7)
		mc_refresh ();
}

static const char *
locate_egrep (void)
{
    static const char * const paths [] = {
	"/bin/egrep",
	"/usr/bin/egrep",
	"/sbin/egrep",
	"/usr/sbin/egrep",
	NULL
    };
    struct stat s;
    const char * const *p;

    for (p = &paths [0]; *p; p++) {
	if (stat (*p, &s) == 0)
	    return *p;
    }
    return "egrep";
}

/*
 * search_content:
 *
 * Search with egrep the global (FIXME) content_pattern string in the
 * DIRECTORY/FILE.  It will add the found entries to the find listbox.
 */
static void
search_content (Dlg_head *h, char *directory, char *filename)
{
    struct stat s;
    char buffer [128];
    char *Fname, *p;
    int file_fd, Pipe, ignoring;
    char c;
    int i;
    pid_t pid;

    Fname = get_full_name (directory, filename);

    if (mc_stat (Fname, &s) != 0 && !S_ISREG (s.st_mode)) {
	free (Fname);
	return;
    }
    if (!S_ISREG (s.st_mode)) {
	free (Fname);
	return;
    }

    file_fd = mc_open (Fname, O_RDONLY);
    free (Fname);

    if (file_fd == -1)
	return;

    if (!egrep_path)
	egrep_path = locate_egrep ();

    Pipe = mc_doublepopen (file_fd, -1, &pid, egrep_path, egrep_path, "-n", content_pattern, "-", NULL);

    if (Pipe == -1) {
	mc_close (file_fd);
	return;
    }

    sprintf (buffer, "Grepping in %s", name_trunc (filename, FIND2_X_USE));

    label_set_text (status_label, buffer);
    mc_refresh ();
    p = buffer;
    ignoring = 0;

    enable_interrupt_key ();
    got_interrupt ();
    while (1) {
	i = read (Pipe, &c, 1);
	if (i != 1)
	    break;

	if (c == '\n') {
	    p = buffer;
	    ignoring = 0;
	}
	if (ignoring)
	    continue;

	if (c == ':') {
	    char *the_name;

	    *p = 0;
	    ignoring = 1;
	    the_name = xasprintf ("%s:%s", buffer, filename);
	    find_add_match (h, directory, the_name);
	    free (the_name);
	} else {
	    if (p - buffer < (sizeof (buffer)-1) && ISASCII (c) && isdigit (c))
		*p++ = c;
	    else
		*p = 0;
	}
    }
    disable_interrupt_key ();
    if (i == -1)
	MC_message (1, " Find/read ", " Problem reading from child ");

    mc_doublepclose (Pipe, pid);
    mc_close (file_fd);
}

static void
find_do_search (struct Dlg_head *h)
{
    struct stat tmp_stat;
    char *tmp_name;             /* For bulding file names */

    if (!h) { /* someone forces me to close find_do_search_dirp */
	if (find_do_search_dirp) {
	    mc_closedir (find_do_search_dirp);
	    find_do_search_dirp = 0;
	}
	find_do_search_dp = 0;
	return;
    }

 do_search_begin:
    while (!find_do_search_dp) {

	if (find_do_search_dirp) {
	    mc_closedir (find_do_search_dirp);
	    find_do_search_dirp = 0;
	}

	while (!find_do_search_dirp) {
	    char *tmp;

	    attrset (REVERSE_COLOR);
	    while (1) {
		tmp = pop_directory ();
		if (!tmp) {
		    bb_running = 0;
		    label_set_text (status_label, "Finished");
		    set_idle_proc (h, 0);
		    return;
		}
		if (find_ignore_dirs) {
		    char *temp_dir = xasprintf (":%s:", tmp);
		    if (strstr (find_ignore_dirs, temp_dir))
			free (tmp);
		    else
			break;
		} else
		    break;
	    }

	    strcpy (fds_directory, tmp);
	    free (tmp);

	    if (MC_verbose) {
		    char buffer [50];

		    sprintf (buffer, "Searching %s", name_trunc (fds_directory, FIND2_X_USE));
		    label_set_text (status_label, buffer);
	    }
	    find_do_search_dirp = mc_opendir (fds_directory);
	    mc_stat (fds_directory, &tmp_stat);
	    subdirs_left = tmp_stat.st_nlink - 2;
	    /* Commented out as unnecessary
	       if (subdirs_left < 0)
	       subdirs_left = MAXINT;
	    */
	}
	find_do_search_dp = mc_readdir (find_do_search_dirp);
    }

    if (find_do_search_dp->d_name[0] == '.' && (find_do_search_dp->d_name[1] == '\0' ||
			(find_do_search_dp->d_name[1] == '.' && find_do_search_dp->d_name[2] == '\0'))) {
	find_do_search_dp = mc_readdir (find_do_search_dirp);
	return;
    }

    tmp_name = get_full_name (fds_directory, find_do_search_dp->d_name);

    if (subdirs_left) {
	mc_lstat (tmp_name, &tmp_stat);
	if (S_ISDIR (tmp_stat.st_mode)) {
	    push_directory (tmp_name);
	    subdirs_left--;
	}
    }

    if (regexp_match (find_pattern, find_do_search_dp->d_name, match_file)) {
	if (content_pattern)
	    search_content (h, fds_directory, find_do_search_dp->d_name);
	else
	    find_add_match (h, fds_directory, find_do_search_dp->d_name);
    }

    free (tmp_name);
    find_do_search_dp = mc_readdir (find_do_search_dirp);

    /* Displays the nice dot */
    bb_count++;
    if (!(bb_count & 31)) {
	if (MC_verbose) {
	    find_do_search_pos = (find_do_search_pos + 1) % 4;
	    attrset (NORMALC);
	    dlg_move (h, FIND2_Y-6, FIND2_X - 4);
	    addch (rotating_dash [find_do_search_pos]);
	    mc_refresh ();
	}
    } else
	goto do_search_begin;
}

static int
view_edit_currently_selected_file (int unparsed_view, int edit)
{
    WLEntry *entry = Find_list->current;
    char *dir, *fullname, *filename;
    int line;

    if (!entry)
	return MSG_NOT_HANDLED;

    dir = entry->data;

    if (!entry->text || !dir)
	return MSG_NOT_HANDLED;

    if (content_pattern) {
	filename = strchr (entry->text + 4, ':') + 1;
	line = atoi (entry->text + 4);
    } else {
	 filename = entry->text + 4;
	 line = 0;
    }
    if (dir [0] == '.' && dir [1] == 0)
	 fullname = strdup (filename);
    else if (dir [0] == '.' && dir [1] == '/')
	 fullname = get_full_name (dir+2, filename);
    else
	 fullname = get_full_name (dir, filename);

    if (edit)
	do_edit_at_line (fullname, line);
    else
	view_file_at_line (fullname, unparsed_view, use_internal_view, line);
    free (fullname);
    return MSG_HANDLED;
}

static int
find_callback (struct Dlg_head *h, int id, int Msg)
{
    switch (Msg) {
    case DLG_DRAW:
	common_dialog_repaint (h);
	break;
    case DLG_KEY:
	if (id == KEY_F(3) || id == KEY_F(13)) {
	    int unparsed_view = (id == KEY_F(13));
	    return view_edit_currently_selected_file (unparsed_view, 0);
	}
	if (id == KEY_F(4)) {
	    return view_edit_currently_selected_file (0, 1);
	 }
	 return MSG_NOT_HANDLED;

     case DLG_IDLE:
	 find_do_search (h);
	 break;
     }
     return 0;
}

/* Handles the Stop/Start button in the find window */
static int
start_stop (int button, void *extra)
{
    bb_running = is_start;
    set_idle_proc (find_dlg, bb_running);
    is_start = !is_start;

    label_set_text (status_label, is_start ? "Stopped" : "Searching");
    button_set_text (stop_button, fbuts [is_start].text);

    return 0;
}

/* Handle view command, when invoked as a button */
static int
find_do_view_file (int button, void *extra)
{
    view_edit_currently_selected_file (0, 0);
    return 0;
}

/* Handle edit command, when invoked as a button */
static int
find_do_edit_file (int button, void *extra)
{
    view_edit_currently_selected_file (0, 1);
    return 0;
}

static void
init_find_vars (void)
{
    char *dir;

    if (old_dir) {
	free (old_dir);
	old_dir = 0;
    }
    bb_count = 0;
    bb_matches = 0;

    /* Remove all the items in the stack */
    while ((dir = pop_directory ()) != NULL)
	free (dir);
}

static int
find_file (char *start_dir, char *pattern, char *content, char **Dirname,  char **filename)
{
    int return_value = 0;
    char *dir;
    char *dir_tmp, *file_tmp;

	/* Dynamically place buttons centered within current window size */
	{
		fbuts[0].text = "&Suspend";  fbuts[0].len = 11;
		fbuts[1].text = "Con&tinue"; fbuts[1].len = 12;
		fbuts[2].text = "&Chdir";    fbuts[2].len = 11;
		fbuts[3].text = "&Again";    fbuts[3].len = 9;
		fbuts[4].text = "&Quit";     fbuts[4].len = 8;
		fbuts[5].text = "Pane&lize"; fbuts[5].len = 12;
		fbuts[6].text = "&View - F3";fbuts[6].len = 13;
		fbuts[7].text = "&Edit - F4";fbuts[7].len = 13;

		int l0 = max (fbuts[0].len, fbuts[1].len);
		int l1 = fbuts[2].len + fbuts[3].len + l0 + fbuts[4].len;
		int l2 = fbuts[5].len + fbuts[6].len + fbuts[7].len;
		int r1, r2;

		FIND2_X = COLS - 16;

		/* Check, if both button rows fit within FIND2_X */
		if (l1 + 9 > FIND2_X) FIND2_X = l1 + 9;
		if (l2 + 8 > FIND2_X) FIND2_X = l2 + 8;

		/* compute amount of space between buttons for each row */
		r1 = (FIND2_X - 4 - l1) % 5;
		l1 = (FIND2_X - 4 - l1) / 5;
		r2 = (FIND2_X - 4 - l2) % 4;
		l2 = (FIND2_X - 4 - l2) / 4;

		/* ...and finally, place buttons */
		fbuts [2].x = 2 + r1/2 + l1;
		fbuts [3].x = fbuts [2].x + fbuts [2].len + l1;
		fbuts [0].x = fbuts [3].x + fbuts [3].len + l1;
		fbuts [4].x = fbuts [0].x + l0 + l1;
		fbuts [5].x = 2 + r2/2 + l2;
		fbuts [6].x = fbuts [5].x + fbuts [5].len + l2;
		fbuts [7].x = fbuts [6].x + fbuts [6].len + l2;
	}

    find_dlg = create_dlg (0, 0, FIND2_Y, FIND2_X, dialog_colors,
			   find_callback, "[Find File]", DLG_CENTER | DLG_GRID);

    x_set_dialog_title (find_dlg, "Find file");

    add_widgetl (find_dlg,
		button_new (FIND2_Y-3, fbuts[7].x, B_VIEW, NORMAL_BUTTON,
			fbuts[7].text, find_do_edit_file, find_dlg), 0);
    add_widgetl (find_dlg,
		button_new (FIND2_Y-3, fbuts[6].x, B_VIEW, NORMAL_BUTTON,
			fbuts[6].text, find_do_view_file, find_dlg), 0);
    add_widgetl (find_dlg,
		button_new (FIND2_Y-3, fbuts[5].x, B_PANELIZE, NORMAL_BUTTON,
			fbuts[5].text, 0, 0), XV_WLAY_CENTERROW);

    add_widgetl (find_dlg,
		button_new (FIND2_Y-4, fbuts[4].x, B_CANCEL, NORMAL_BUTTON,
			fbuts[4].text, 0, 0), XV_WLAY_RIGHTOF);
    stop_button = button_new (FIND2_Y-4, fbuts[0].x, FIND_B_STOP, NORMAL_BUTTON,
		fbuts[0].text, start_stop, find_dlg);
    add_widgetl (find_dlg, stop_button, XV_WLAY_RIGHTOF);
    add_widgetl (find_dlg,
		button_new (FIND2_Y-4, fbuts[3].x, B_AGAIN, NORMAL_BUTTON,
			fbuts[3].text, 0, 0), XV_WLAY_RIGHTOF);
    add_widgetl (find_dlg,
		button_new (FIND2_Y-4, fbuts[2].x, B_ENTER, DEFPUSH_BUTTON,
			fbuts[2].text, 0, 0), XV_WLAY_CENTERROW);

    status_label = label_new (FIND2_Y-6, 4, "Searching");
    add_widgetl (find_dlg, status_label, XV_WLAY_BELOWOF);

    Find_list = listbox_new (2, 2, FIND2_X-4, FIND2_Y-9, listbox_finish, 0);
    add_widgetl (find_dlg, Find_list, XV_WLAY_EXTENDWIDTH);

    /* FIXME: Need to cleanup this, this ought to be passed non-globaly */
    find_pattern    = pattern;
    content_pattern = content;

    set_idle_proc (find_dlg, 1);
    init_find_vars ();
    push_directory (start_dir);

    run_dlg (find_dlg);

    return_value = find_dlg->ret_value;

    /* Remove all the items in the stack */
    while ((dir = pop_directory ()) != NULL)
	free (dir);

    listbox_get_current (Find_list, &file_tmp, &dir_tmp);

    if (dir_tmp)
	*Dirname  = strdup (dir_tmp);
    if (file_tmp)
	*filename = strdup (file_tmp);
    if (return_value == B_PANELIZE && *filename) {
	int Status, link_to_dir, stalled_link;
	int next_free = 0;
	int i;
	struct stat buf;
	WLEntry *entry = Find_list->list;
	dir_list *list = &cpanel->dir;
	char *name;

	for (i = 0; entry && i < Find_list->count; entry = entry->next, i++) {
	    char *Filename;

	    if (content_pattern)
		Filename = strchr (entry->text+4, ':')+1;
	    else
		Filename = entry->text+4;

	    if (!entry->text || !entry->data)
		continue;
	    dir = entry->data;
	    if (dir [0] == '.' && dir [1] == 0)
		name = strdup (Filename);
	    else if (dir [0] == '.' && dir [1] == '/')
		name = get_full_name (dir + 2, Filename);
	    else
		name = get_full_name (dir, Filename);
	    Status = handle_path (list, name, &buf, next_free, &link_to_dir,
		&stalled_link);
	    if (Status == 0) {
		free (name);
		continue;
	    }
	    if (Status == -1) {
		free (name);
		break;
	    }

	    /* don't add files more than once to the panel */
	    if (content_pattern && next_free > 0) {
		if (strcmp (list->list [next_free-1].fname, name) == 0) {
		    free (name);
		    continue;
		}
	    }

	    if (!next_free) /* first turn i.e clean old list */
		clean_dir (list, cpanel->count);
	    list->list [next_free].fnamelen = strlen (name);
	    list->list [next_free].fname = name;
	    list->list [next_free].cache = NULL;
	    file_mark (cpanel, next_free, 0);
	    list->list [next_free].f.link_to_dir = link_to_dir;
	    list->list [next_free].f.stalled_link = stalled_link;
	    list->list [next_free].buf = buf;
	    next_free++;
	   if (!(next_free & 15))
	       rotate_dash ();
	}
	if (next_free) {
	    cpanel->count = next_free;
	    cpanel->is_panelized = 1;
	    cpanel->dirs_marked = 0;
	    cpanel->has_dir_sizes = 0;
	    cpanel->marked = 0;
	    cpanel->total = 0;
	    cpanel->top_file = 0;
	    cpanel->selected = 0;

	    if (start_dir [0] == '/') {
		strcpy (cpanel->cwd, "/");
		chdir ("/");
	    }
	}
    }

    set_idle_proc (find_dlg, 0);
    destroy_dlg (find_dlg);
    find_do_search (0); /* force find_do_search to release resources */
    if (old_dir) {
	free (old_dir);
	old_dir = 0;
    }
    return return_value;
}

static void do_find (void)
{
    char *filename, *Dirname;
    int  v, dir_and_file_set;
    int done = 0;

    while (!done) {
	char *start_dir = NULL, *pattern = NULL, *content = NULL;
	if (!find_parameters (&start_dir, &pattern, &content))
	    break;

	Dirname = filename = NULL;
	is_start = 0;
	v = find_file (start_dir, pattern, content, &Dirname, &filename);
	free (start_dir);
	free (pattern);

	if (v == B_ENTER) {
	    if (Dirname || filename) {
		if (Dirname) {
		    do_cd (Dirname, cd_exact);
		    if (filename)
			try_to_select (cpanel, filename + (content ?
			   (strchr (filename + 4, ':') - filename + 1) : 4) );
		} else if (filename)
		    do_cd (filename, cd_exact);
		paint_panel (cpanel);
		select_item (cpanel);
	    }
	    if (Dirname)
		free (Dirname);
	    if (filename)
		free (filename);
	    break;
	}
	free (content);
	dir_and_file_set = Dirname && filename;
	if (Dirname)  free (Dirname);
	if (filename) free (filename);
	if (v == B_CANCEL)
	    break;

	if (v == B_PANELIZE) {
	    if (dir_and_file_set) {
		try_to_select (cpanel, NULL);
		paint_panel (cpanel);
	    }
	    break;
	}
    }
}

/* fsusage.c -- return space usage of mounted filesystems */

#if !defined(NO_INFOMOUNT)

/* Return the number of TOSIZE-byte blocks used by
   BLOCKS FROMSIZE-byte blocks, rounding away from zero.
   TOSIZE must be positive.  Return -1 if FROMSIZE is not positive.  */

static long fs_adjust_blocks (long blocks, int fromsize, int tosize)
{
    if (tosize <= 0)
	abort ();
    if (fromsize <= 0)
	return -1;

    if (fromsize == tosize)     /* E.g., from 512 to 512.  */
	return blocks;
    else if (fromsize > tosize) /* E.g., from 2048 to 512.  */
	return blocks * (fromsize / tosize);
    else                        /* E.g., from 256 to 512.  */
	return (blocks + (blocks < 0 ? -1 : 1)) / (tosize / fromsize);
}

/* Fill in the fields of FSP with information about space usage for
   the filesystem on which PATH resides.
   Return 0 if successful, -1 if not. */

static int get_fs_usage (char *path, struct fs_usage *fsp)
{
    struct statfs fsd;

    if (statfs (path, &fsd) < 0)
	return -1;
#define CONVERT_BLOCKS(b) fs_adjust_blocks ((b), fsd.f_frsize ? fsd.f_frsize : fsd.f_bsize, 512)

    fsp->fsu_blocks = CONVERT_BLOCKS (fsd.f_blocks);
    fsp->fsu_bfree = CONVERT_BLOCKS (fsd.f_bfree);
    fsp->fsu_bavail = CONVERT_BLOCKS (fsd.f_bavail);
    fsp->fsu_files = fsd.f_files;
    fsp->fsu_ffree = fsd.f_ffree;

    return 0;
}

#if defined(_AIX) && defined(_I386)
/* AIX PS/2 does not supply statfs.  */

int statfs (char *path, struct statfs *fsb)
{
    struct stat stats;
    struct dustat fsd;

    if (stat (path, &stats))
	return -1;
    if (dustat (stats.st_dev, 0, &fsd, sizeof (fsd)))
	return -1;
    fsb->f_type = 0;
    fsb->f_bsize = fsd.du_bsize;
    fsb->f_blocks = fsd.du_fsize - fsd.du_isize;
    fsb->f_bfree = fsd.du_tfree;
    fsb->f_bavail = fsd.du_tfree;
    fsb->f_files = (fsd.du_isize - 2) * fsd.du_inopb;
    fsb->f_ffree = fsd.du_tinode;
    fsb->f_fsid.val[0] = fsd.du_site;
    fsb->f_fsid.val[1] = fsd.du_pckno;
    return 0;
}
#endif                          /* _AIX && _I386 */
#endif /* NO_INFOMOUNT */
/* Hypertext file browser. */

static int help_callback (struct Dlg_head *h, int id, int msg);

/* returns the position where text was found in the start buffer */
/* or 0 if not found */
static char *search_string (char *start_b, char *text)
{
    char *d = text;
    char *e = start_b;

    /* fmt sometimes replaces a space with a newline in the help file */
    /* Replace the newlines in the link name with spaces to correct the situation */
    while (*d) {
	if (*d == '\n')
	    *d = ' ';
	d++;
    }
    /* Do search */
    for (d = text; *e; e++) {
	if (*d == *e)
	    d++;
	else
	    d = text;
	if (!*d)
	    return e+1;
    }
    return 0;
}

/* Searches text in the buffer pointed by start.  Search ends */
/* if the CHAR_NODE_END is found in the text.  Returns 0 on failure */
static char *search_string_node (char *start_b, char *text)
{
    char *d = text;
    char *e = start_b;

    if (!start_b)
	return 0;

    for (; *e && *e != CHAR_NODE_END; e++) {
	if (*d == *e)
	    d++;
	else
	    d = text;
	if (!*d)
	    return e+1;
    }
    return 0;
}

/* Searches the_char in the buffer pointer by start and searches */
/* it can search forward (direction = 1) or backward (direction = -1) */
static char *search_char_node (char *start_b, char the_char, int direction)
{
    char *e;

    e = start_b;

    for (; *e && (*e != CHAR_NODE_END); e += direction) {
	if (*e == the_char)
	    return e;
    }
    return 0;
}

/* Returns the new current pointer when moved lines lines */
static char *help_move_forward2 (char *c, int lines)
{
    char *p;
    int  line;

    currentpoint = c;
    for (line = 0, p = currentpoint; *p && *p != CHAR_NODE_END; p++) {
	if (line == lines)
	    return currentpoint = p;
	if (*p == '\n')
	    line++;
    }
    return currentpoint = c;
}

static char *help_move_backward2 (char *c, int lines)
{
    char *p;
    int line;

    currentpoint = c;
    for (line = 0, p = currentpoint; *p && p >= help_data; p--) {
	if (*p == CHAR_NODE_END)
	{
	    /* We reached the beginning of the node */
	    /* Skip the node headers */
	    while (*p != ']') p++;
	    return currentpoint = p + 2;
	}
	if (*(p - 1) == '\n')
	    line++;
	if (line == lines)
	    return currentpoint = p;
    }
    return currentpoint = c;
}

static void move_forward (int i)
{
    if (end_of_node)
	return;
    currentpoint = help_move_forward2 (currentpoint, i);
}

static void move_backward (int i)
{
    currentpoint = help_move_backward2 (currentpoint, ++i);
}

static void help_move_to_top (int dummy)
{
    while (currentpoint > help_data && *currentpoint != CHAR_NODE_END)
	currentpoint--;
    while (*currentpoint != ']')
	currentpoint++;
    currentpoint = currentpoint + 1;
    selected_item = NULL;
}

static void help_move_to_bottom (int dummy)
{
    while (*currentpoint && *currentpoint != CHAR_NODE_END)
	currentpoint++;
    currentpoint--;
    move_backward (help_lines - 1);
}

static char *help_follow_link (char *start_l, char *Selected_item)
{
    char link_name [MAXLINKNAME];
    char *p;
    int  i = 0;

    if (!Selected_item)
	return start_l;

    for (p = Selected_item; *p && *p != CHAR_NODE_END && *p != CHAR_LINK_POINTER; p++)
	;
    if (*p == CHAR_LINK_POINTER) {
	link_name [0] = '[';
	for (i = 1; *p != CHAR_LINK_END && *p && *p != CHAR_NODE_END && i < MAXLINKNAME-3; )
	    link_name [i++] = *++p;
	link_name [i-1] = ']';
	link_name [i] = 0;
	p = search_string (help_data, link_name);
	if (p)
	    return p;
    }
    return " Help file format error\n\x4";   /*  */
}

static char *select_next_link (char *start_l, char *current_link)
{
    char *p;

    if (!current_link)
	return 0;

    p = search_string_node (current_link, STRING_LINK_END);
    if (!p)
	return NULL;
    p = search_string_node (p, STRING_LINK_START);
    if (!p)
	return NULL;
    return p - 1;
}

static char *select_prev_link (char *start_l, char *current_link)
{
    char *p;

    if (!current_link)
	return 0;

    p = current_link - 1;
    if (p <= start_l)
	return 0;

    p = search_char_node (p, CHAR_LINK_START, -1);
    return p;
}

static void start_link_area (int x, int y, char *link_name)
{
    Link_Area *new;

    if (inside_link_area)
	MC_message (0, " Warning ", " Internal bug: Double start of link area ");

    /* Allocate memory for a new link area */
    new = x_malloc (sizeof (Link_Area), "Help, link_area");
    new->next = link_area;
    link_area = new;

    /* Save the beginning coordinates of the link area */
    link_area->x1 = x;
    link_area->y1 = y;

    /* Save the name of the destination anchor */
    link_area->link_name = link_name;

    inside_link_area = 1;
}

static void end_link_area (int x, int y)
{
    if (inside_link_area) {
	/* Save the end coordinates of the link area */
	link_area->x2 = x;
	link_area->y2 = y;

	inside_link_area = 0;
    }
}

static void clear_link_areas (void)
{
    Link_Area *current;

    while (link_area) {
	current = link_area;
	link_area = current -> next;
	free (current);
    }
    inside_link_area = 0;
}

static void show (Dlg_head *h, char *paint_start)
{
    char *p;
    int  col, line, c;
    int  painting = 1;
    int acs;                    /* Flag: Alternate character set active? */
    int repeat_paint;
    int active_col, active_line;/* Active link position */

    do {
	line = col = acs = active_col = active_line = repeat_paint = 0;

	clear_link_areas ();
	if (selected_item < paint_start)
	    selected_item = NULL;

	for (p = paint_start; *p != CHAR_NODE_END && line < help_lines; p++) {
	    c = *p;
	    switch (c) {
	    case CHAR_LINK_START:
		if (selected_item == NULL)
		    selected_item = p;
		if (p == selected_item) {
		    attrset (HELP_SLINK_COLOR);

		    /* Store the coordinates of the link */
		    active_col = col + 2;
		    active_line = line + 2;
		}
		else
		    attrset (HELP_LINK_COLOR);
		start_link_area (col, line, p);
		break;
	    case CHAR_LINK_POINTER:
		painting = 0;
		end_link_area (col - 1, line);
		break;
	    case CHAR_LINK_END:
		painting = 1;
		attrset (HELP_NORMAL_COLOR);
		break;
	    case CHAR_ALTERNATE:
		acs = 1;
		break;
	    case CHAR_NORMAL:
		acs = 0;
		break;
	    case CHAR_VERSION:
		dlg_move (h, line+2, col+2);
		addstr (VERSION);
		col += strlen (VERSION);
		break;
	    case CHAR_BOLD_ON:
		attrset (HELP_BOLD_COLOR);
		break;
	    case CHAR_ITALIC_ON:
		attrset (HELP_ITALIC_COLOR);
		break;
	    case CHAR_BOLD_OFF:
		attrset (HELP_NORMAL_COLOR);
		break;
	    case '\n':
		line++;
		col = 0;
		break;
	    case '\t':
		col = (col/8 + 1) * 8;
		break;
	    case CHAR_MCLOGO:
	    case CHAR_TEXTONLY_START:
	    case CHAR_TEXTONLY_END:
		break;
	    case CHAR_XONLY_START:
		while (*p && *p != CHAR_NODE_END && *p != CHAR_XONLY_END)
		    p++;
		if (*p == CHAR_NODE_END || !*p)
		    p--;
		break;
	    default:
		if (!painting)
		    continue;
		if (col > HELP_WINDOW_WIDTH-1)
		    continue;

		dlg_move (h, line+2, col+2);
		if (acs) {
		    if (c == ' ' || c == '.')
			addch (c);
		    else
			SLsmg_draw_object (h->y + line + 2, h->x + col + 2, c);
		} else
		    addch (c);
		col++;
		break;
	    }
	}
	last_shown = p;
	end_of_node = line < help_lines;
	attrset (HELP_NORMAL_COLOR);
	if (selected_item >= last_shown) {
	    if (link_area != NULL) {
		selected_item = link_area->link_name;
		repeat_paint = 1;
	    }
	    else
		selected_item = NULL;
	}
    } while (repeat_paint);

    /* Position the cursor over a nice link */
    if (active_col)
	dlg_move (h, active_line, active_col);
}

static int help_event (Gpm_Event *event, Widget *w)
{
    Link_Area *current_area;

    if (! (event->type & GPM_UP))
	return 0;

    /* The event is relative to the dialog window, adjust it: */
    event->y -= 2;

    if (event->buttons & GPM_B_RIGHT) {
	currentpoint = startpoint = bb_history [history_ptr].page;
	selected_item = bb_history [history_ptr].link;
	history_ptr--;
	if (history_ptr < 0)
	    history_ptr = HISTORY_SIZE-1;

	help_callback (w->parent, 0, DLG_DRAW);
	return 0;
    }

    /* Test whether the mouse click is inside one of the link areas */
    current_area = link_area;
    while (current_area)
    {
	/* Test one line link area */
	if (event->y == current_area->y1 && event->x >= current_area->x1 &&
	    event->y == current_area->y2 && event->x <= current_area->x2)
	    break;
	/* Test two line link area */
	if (current_area->y1 + 1 == current_area->y2) {
	    /* The first line */
	    if (event->y == current_area->y1 && event->x >= current_area->x1)
		break;
	    /* The second line */
	    if (event->y == current_area->y2 && event->x <= current_area->x2)
		break;
	}
	/* Mouse will not work with link areas of more than two lines */

	current_area = current_area -> next;
    }

    /* Test whether a link area was found */
    if (current_area) {
	/* The click was inside a link area -> follow the link */
	history_ptr = (history_ptr+1) % HISTORY_SIZE;
	bb_history [history_ptr].page = currentpoint;
	bb_history [history_ptr].link = current_area->link_name;
	currentpoint = startpoint = help_follow_link (currentpoint, current_area->link_name);
	selected_item = NULL;
    } else{
	if (event->y < 0)
	    move_backward (help_lines - 1);
	else if (event->y >= help_lines)
	    move_forward (help_lines - 1);
	else if (event->y < help_lines/2)
	    move_backward (1);
	else
	    move_forward (1);
    }

    /* Show the new node */
    help_callback (w->parent, 0, DLG_DRAW);

    return 0;
}

/* show help */
static void help_help_cmd (Dlg_head *h)
{
    history_ptr = (history_ptr+1) % HISTORY_SIZE;
    bb_history [history_ptr].page = currentpoint;
    bb_history [history_ptr].link = selected_item;
    currentpoint = startpoint = search_string (help_data, "[How to use help]") + 1;
    selected_item = NULL;
    help_callback (h, 0, DLG_DRAW);
}

static void help_index_cmd (Dlg_head *h)
{
    char *new_item;

    history_ptr = (history_ptr+1) % HISTORY_SIZE;
    bb_history [history_ptr].page = currentpoint;
    bb_history [history_ptr].link = selected_item;
    currentpoint = startpoint = search_string (help_data, "[Help]") + 1;

    if (!(new_item = search_string (help_data, "[Contents]")))
	MC_message (1, MSG_ERROR, " Can't find node [Contents] in help file ");
    else
	currentpoint = startpoint = new_item + 1;
    selected_item = NULL;
    help_callback (h, 0, DLG_DRAW);
}

static void help_quit_cmd (void *x)
{
    dlg_stop (x);
}

static void prev_node_cmd (Dlg_head *h)
{
    currentpoint = startpoint = bb_history [history_ptr].page;
    selected_item = bb_history [history_ptr].link;
    history_ptr--;
    if (history_ptr < 0)
	history_ptr = HISTORY_SIZE-1;
    help_callback (h, 0, DLG_DRAW);
}

static int md_callback (Dlg_head *h, Widget *w, int msg, int par)
{
    return default_proc (h, msg, par);
}

static Widget *mousedispatch_new (int y, int x, int yl, int xl)
{
    Widget *w = x_malloc (sizeof (Widget), "disp_new");

    init_widget (w, y, x, yl, xl, (callback_fn) md_callback, 0, (mouse_h)  help_event);

    return w;
}

static int help_handle_key (struct Dlg_head *h, int c)
{
    char *new_item;

    if (c != KEY_UP && c != KEY_DOWN &&
	check_movement_keys (c, 1, help_lines, currentpoint,
			     (movefn) help_move_backward2,
			     (movefn) help_move_forward2,
			     (movefn) help_move_to_top,
			     (movefn) help_move_to_bottom))
	/* Nothing */;
    else switch (c) {
    case 'l':
    case KEY_LEFT:
	prev_node_cmd (h);
	break;

    case '\n':
    case KEY_RIGHT:
	/* follow link */
	if (selected_item) {
	    history_ptr = (history_ptr+1) % HISTORY_SIZE;
	    bb_history [history_ptr].page = currentpoint;
	    bb_history [history_ptr].link = selected_item;
	    currentpoint = startpoint = help_follow_link (currentpoint, selected_item) + 1;
	}
	selected_item = NULL;
	break;

    case KEY_DOWN:
    case '\t':
	/* select next link */
	new_item = select_next_link (startpoint, selected_item);
	if (new_item) {
	    selected_item = new_item;
	    if (selected_item >= last_shown) {
		if (c == KEY_DOWN)
		    move_forward (1);
		else
		    selected_item = NULL;
	    }
	} else if (c == KEY_DOWN)
	    move_forward (1);
	else
	    selected_item = NULL;
	break;

    case KEY_UP:
    case ALT ('\t'):
	/* select previous link */
	new_item = select_prev_link (startpoint, selected_item);
	selected_item = new_item;
	if (selected_item < currentpoint || selected_item >= last_shown) {
	    if (c == KEY_UP)
		move_backward (1);
	    else{
		if (link_area != NULL)
		    selected_item = link_area->link_name;
		else
		    selected_item = NULL;
	    }
	}
	break;

    case 'n':
	/* Next node */
	new_item = currentpoint;
	while (*new_item && *new_item != CHAR_NODE_END)
	    new_item++;
	if (*++new_item == '[') {
	    while (*new_item != ']')
		new_item++;
	    currentpoint = new_item + 2;
	    selected_item = NULL;
	}
	break;

    case 'p':
	/* Previous node */
	new_item = currentpoint;
	while (new_item > help_data + 1 && *new_item != CHAR_NODE_END)
	    new_item--;
	new_item--;
	while (new_item > help_data && *new_item != CHAR_NODE_END)
	    new_item--;
	while (*new_item != ']')
	    new_item++;
	currentpoint = new_item + 2;
	selected_item = NULL;
	break;

    case 'c':
	help_index_cmd (h);
	break;

    case ESC_CHAR:
    case XCTRL('g'):
	dlg_stop (h);
	break;

    default:
	return 0;

    }
    help_callback (h, 0, DLG_DRAW);
    return 1;
}

static int help_callback (struct Dlg_head *h, int id, int msg)
{
    switch (msg) {
    case DLG_DRAW:
	attrset (HELP_NORMAL_COLOR);
	dlg_erase (h);
	draw_box (h, 1, 1, help_lines+2, HELP_WINDOW_WIDTH+2);
	attrset (COLOR_HOT_NORMAL);
	dlg_move (h, 1, (HELP_WINDOW_WIDTH - 1) / 2);
	addstr (" Help ");
	attrset (HELP_NORMAL_COLOR);
	show (h, currentpoint);
	break;

    case DLG_KEY:
	return help_handle_key (h, id);
    }
    return 0;
}

static void interactive_display_finish (void)
{
    clear_link_areas ();
    free (help_data);
}

void interactive_display (const char *filename, const char *node)
{
    WButtonBar *help_bar;
    Widget     *md;
    char *main_code;        /* The main_code node */
    Dlg_head *whelp;

    if ((help_data = load_file (filename)) == 0) {
	MC_message (1, MSG_ERROR, " Can't open file %s \n %m ",
		 filename);
	return;
    }
    if (!(main_code = search_string (help_data, (char *)node))) {
	MC_message (1, MSG_ERROR, " Can't find node %s in help file ", node);
	interactive_display_finish ();
	return;
    }

    if (help_lines > LINES - 4)
	help_lines = LINES - 4;

    whelp = create_dlg (0, 0, help_lines+4, HELP_WINDOW_WIDTH+4, dialog_colors,
			help_callback, "[Help]", DLG_TRYUP|DLG_CENTER);

    /* allow us to process the tab key */
    whelp->raw = 1;

    selected_item = search_string_node (main_code, STRING_LINK_START) - 1;
    currentpoint = startpoint = main_code + 1;

    for (history_ptr = HISTORY_SIZE; history_ptr;) {
	history_ptr--;
	bb_history [history_ptr].page = currentpoint;
	bb_history [history_ptr].link = selected_item;
    }

    help_bar = buttonbar_new (keybar_visible);
    help_bar->widget.y -= whelp->y;
    help_bar->widget.x -= whelp->x;

    md = mousedispatch_new (1, 1, help_lines, HELP_WINDOW_WIDTH-2);

    add_widget (whelp, help_bar);
    add_widget (whelp, md);

    define_label_data (whelp, (Widget *)NULL, 1, "Help",  (buttonbarfn) help_help_cmd, whelp);
    define_label_data (whelp, (Widget *)NULL, 2, "Index", (buttonbarfn) help_index_cmd,whelp);
    define_label_data (whelp, (Widget *)NULL, 3, "Prev",  (buttonbarfn) prev_node_cmd, whelp);
    define_label (whelp, (Widget *) NULL, 4, "", 0);
    define_label (whelp, (Widget *) NULL, 5, "", 0);
    define_label (whelp, (Widget *) NULL, 6, "", 0);
    define_label (whelp, (Widget *) NULL, 7, "", 0);
    define_label (whelp, (Widget *) NULL, 8, "", 0);
    define_label (whelp, (Widget *) NULL, 9, "", 0);
    define_label_data (whelp, (Widget *) NULL, 10, "Quit", help_quit_cmd, whelp);

    run_dlg (whelp);
    interactive_display_finish ();
    destroy_dlg (whelp);
}

/* Directory hotlist -- for the Midnight Commander */

static void remove_from_hotlist (struct hotlist *entry);
static void add_new_group_cmd (void);

static struct hotlist *new_hotlist (void)
{
    return zmalloc (sizeof (struct hotlist));
}

static void hotlist_refresh (Dlg_head *dlg)
{
    dialog_repaint (dlg, COLOR_NORMAL, COLOR_HOT_NORMAL);
    attrset (COLOR_NORMAL);
    draw_box (dlg, 2, 5,
		dlg->lines - (hotlist_state.moving ? 6 : 10),
		dlg->cols - (UX_HOTLIST*2));
    if (!hotlist_state.moving)
	draw_box (dlg, dlg->lines-8, 5, 3, dlg->cols - (UX_HOTLIST*2));
}

/* If current->data is 0, then we are dealing with a VFS pathname */
static inline void update_path_name (void)
{
    char *text, *p;
    WListbox  *list = hotlist_state.moving ? l_movelist : l_hotlist;
    Dlg_head  *dlg = hotlist_state.moving ? movelist_dlg : hotlist_dlg;

    if (list->current) {
	if (list->current->data != 0) {
	    struct hotlist *hlp = (struct hotlist *)list->current->data;

	    if (hlp->type == HL_TYPE_ENTRY)
		text = hlp->directory;
	    else
		text = "Subgroup - press ENTER to see list";

	    p = xasprintf (" %s ", current_group->label);
	    if (!hotlist_state.moving)
		label_set_text (pname_group, name_trunc (p, dlg->cols - (UX_HOTLIST*2+4)));
	    else
		label_set_text (movelist_group, name_trunc (p, dlg->cols - (UX_HOTLIST*2+4)));
	    free (p);
	} else {
	    text = list->current->text;
	}
    } else {
	text = "";
    }
    if (!hotlist_state.moving)
	label_set_text (ho_pname, name_trunc (text, dlg->cols - (UX_HOTLIST*2+4)));
    dlg_redraw (dlg);
}

#define CHECK_BUFFER  \
do { \
    int               i; \
\
    if ((i = strlen (current->label) + 3) > fill_listbox_buflen) { \
      free (fill_listbox_buf); \
      fill_listbox_buf = x_malloc (fill_listbox_buflen = 1024 * (i/1024 + 1), "fill_listbox"); \
    } \
    fill_listbox_buf[0] = '\0'; \
} while (0)

static void fill_listbox (void)
{
    struct hotlist *current = current_group->head;

    if (!fill_listbox_buf)
	fill_listbox_buf = x_malloc (fill_listbox_buflen = 1024, "fill_listbox");
    fill_listbox_buf[0] = '\0';

    while (current) {
	switch (current->type) {
	case HL_TYPE_GROUP:
	    {
		CHECK_BUFFER;
		strcat (strcat (fill_listbox_buf, "->"), current->label);
		if (hotlist_state.moving)
		    listbox_add_item (l_movelist, 0, 0, fill_listbox_buf, current);
		else
		    listbox_add_item (l_hotlist, 0, 0, fill_listbox_buf, current);
	    }
	    break;
	case HL_TYPE_ENTRY:
		if (hotlist_state.moving)
		    listbox_add_item (l_movelist, 0, 0, current->label, current);
		else
		    listbox_add_item (l_hotlist, 0, 0, current->label, current);
	    break;
	default:
		break;
	}
	current = current->next;
    }
}

#undef CHECK_BUFFER

static void
unlink_entry (struct hotlist *entry)
{
    struct hotlist *current = current_group->head;

    if (current == entry)
	current_group->head = entry->next;
    else
	while (current && current->next != entry)
	    current = current->next;
    if (current)
	    current->next = entry->next;
    entry->next = entry->up = 0;
}

static void add_new_entry_cmd (void);
static void init_movelist (int, struct hotlist *);

static int hotlist_button_callback (int action, void *data)
{
    switch (action) {
    case B_MOVE:
	{
	    struct hotlist  *saved = current_group;
	    struct hotlist  *item;
	    struct hotlist  *moveto_item = 0;
	    struct hotlist  *moveto_group = 0;
	    int             ret;

	    if (!l_hotlist->current)
		return 0;               /* empty group - nothing to do */
	    item = l_hotlist->current->data;
	    hotlist_state.moving = 1;
	    init_movelist (LIST_MOVELIST, item);
	    run_dlg (movelist_dlg);
	    ret = movelist_dlg->ret_value;
	    hotlist_state.moving = 0;
	    if (l_movelist->current)
		moveto_item = l_movelist->current->data;
	    moveto_group = current_group;
	    destroy_dlg (movelist_dlg);
	    current_group = saved;
	    if (ret == B_CANCEL)
		return 0;
	    if (moveto_item == item)
		return 0;               /* If we insert/append a before/after a
					   it hardly changes anything ;) */
	    unlink_entry (item);
	    listbox_remove_current (l_hotlist, 1);
	    item->up = moveto_group;
	    if (!moveto_group->head)
		moveto_group->head = item;
	    else if (!moveto_item) {    /* we have group with just comments */
		struct hotlist *p = moveto_group->head;

		/* skip comments */
		while (p->next)
		    p = p->next;
		p->next = item;
	    } else if (ret == B_ENTER || ret == B_APPEND)
		if (!moveto_item->next)
		    moveto_item->next = item;
		else {
		    item->next = moveto_item->next;
		    moveto_item->next = item;
		}
	    else if (moveto_group->head == moveto_item) {
		    moveto_group->head = item;
		    item->next = moveto_item;
		} else {
		    struct hotlist *p = moveto_group->head;

		    while (p->next != moveto_item)
			p = p->next;
		    item->next = p->next;
		    p->next = item;
		}
	    listbox_remove_list (l_hotlist);
	    fill_listbox ();
	    repaint_screen ();
	    hotlist_state.modified = 1;
	    return 0;
	    break;
	}
    case B_REMOVE_HOTLIST:
	if (l_hotlist->current)
	    remove_from_hotlist (l_hotlist->current->data);
	return 0;
	break;

    case B_NEW_GROUP:
	add_new_group_cmd ();
	return 0;
	break;

    case B_ADD_CURRENT:
	add2hotlist_cmd ();
	return 0;
	break;

    case B_NEW_ENTRY:
	add_new_entry_cmd ();
	return 0;
	break;

    case B_ENTER:
	{
	WListbox *list = hotlist_state.moving ? l_movelist : l_hotlist;
	if (list->current) {
	    if (list->current->data) {
		struct hotlist *hlp = (struct hotlist*) list->current->data;
		if (hlp->type == HL_TYPE_ENTRY)
		    return 1;
		else {
		    listbox_remove_list (list);
		    current_group = hlp;
		    fill_listbox ();
		    return 0;
		}
	    } else
		return 1;
	}
	}
	/* Fall through if list empty - just go up */

    case B_UP_GROUP:
	{
	WListbox *list = hotlist_state.moving ? l_movelist : l_hotlist;
	listbox_remove_list (list);
	current_group = current_group->up;
	fill_listbox ();
	return 0;
	break;
	}

    default:
	return 1;
	break;

    }
}

static int hotlist_callback (Dlg_head * h, int Par, int Msg)
{
    switch (Msg) {
    case DLG_DRAW:
	hotlist_refresh (h);
	break;

    case DLG_UNHANDLED_KEY:
	switch (Par) {
	case '\n':
	    if (ctrl_pressed())
		goto l1;
	case KEY_ENTER:
	case KEY_RIGHT:
	    if (hotlist_button_callback (B_ENTER, 0)) {
		h->ret_value = B_ENTER;
		dlg_stop (h);
	    };
	    return 1;
	    break;
	case KEY_LEFT:
	    if (hotlist_state.type != LIST_VFSLIST )
		return !hotlist_button_callback (B_UP_GROUP, 0);
	    else
		return 0;
	    break;
l1:
	case ALT('\n'):
	case ALT('\r'):
	    if (!hotlist_state.moving)
	    {
		if (l_hotlist->current) {
		    if (l_hotlist->current->data) {
			struct hotlist *hlp = (struct hotlist*) l_hotlist->current->data;
			if (hlp->type == HL_TYPE_ENTRY) {
			    char *tmp = xasprintf( "cd %s", hlp->directory);
			    stuff (input_w (cmdline), tmp, 0);
			    free (tmp);
			    dlg_stop (h);
			    h->ret_value = B_CANCEL;
			    return 1;
			}
		    }
		}
	    }
	    return 1; /* ignore key */
	default:
	    return 0;
	}
	break;

    case DLG_POST_KEY:
	if (hotlist_state.moving)
	    dlg_select_widget (movelist_dlg, l_movelist);
	else
	    dlg_select_widget (hotlist_dlg, l_hotlist);
	/* always stay on hotlist */
	/* fall through */

    case DLG_INIT:
	attrset (MENU_ENTRY_COLOR);
	update_path_name ();
	break;
    }
    return 0;
}

static int ho_l_call (void *l)
{
    WListbox *list = (WListbox *) l;
    Dlg_head *dlg = hotlist_state.moving ? movelist_dlg : hotlist_dlg;

    if (list->current) {
	if (list->current->data) {
	    struct hotlist *hlp = (struct hotlist*) list->current->data;
	    if (hlp->type == HL_TYPE_ENTRY) {
		dlg->ret_value = B_ENTER;
		dlg_stop (dlg);
		return listbox_finish;
	    } else {
		hotlist_button_callback (B_ENTER, (void *)0);
		hotlist_callback (dlg, '\n', DLG_POST_KEY);
		return listbox_nothing;
	    }
	} else {
	    dlg->ret_value = B_ENTER;
	    dlg_stop (dlg);
	    return listbox_finish;
	}
    }

    hotlist_button_callback (B_UP_GROUP, (void *)0);
    hotlist_callback (dlg, 'u', DLG_POST_KEY);
    return listbox_nothing;
}

#ifdef USE_VFS
static void add_name_to_list (char *path)
{
    listbox_add_item (l_hotlist, 0, 0, path, 0);
}
#endif
/*
 * Expands all button names (once) and recalculates button positions.
 * returns number of columns in the dialog box, which is 10 chars longer
 * then buttonbar.
 *
 * If common width of the window (i.e. in xterm) is less than returned
 * width - sorry :)  (anyway this did not handled in previous version too)
 */
static int
init_i18n_stuff(int list_type, int cols)
{
	int i;

	/* Dynamic resizing of buttonbars */
	{
		int len[2], count_row[2]; /* at most two lines of buttons */
		int cur_x[2], row;

		memcpy(hotlist_but, hotlist_but_st, sizeof(hotlist_but_st));
		len[0] = len[1] = count_row[0] = count_row[1] = 0;

		i = ELEMENTS(hotlist_but_st);
		/* Count len of buttonbars, assuming 2 extra space between buttons */
		while (i--) {
			if (! (hotlist_but[i].type & list_type))
				continue;

			row = hotlist_but [i].y;
			++count_row [row];
			len [row] += strlen (hotlist_but [i].text) + 5;
			if (hotlist_but [i].flags == DEFPUSH_BUTTON)
				len [row] += 2;
		}
		len[0] -= 2;
		len[1] -= 2;

		cols = max(cols, max(len[0], len[1]));

		/* arrange buttons */

		cur_x[0] = cur_x[1] = 0;
		i = ELEMENTS(hotlist_but_st);
		while (i--)
		{
			if (! (hotlist_but[i].type & list_type))
				continue;

			row = hotlist_but [i].y;

			if (hotlist_but [i].x != 0)
			{
				/* not first int the row */
				if (!strcmp (hotlist_but [i].text, cancel_button))
					hotlist_but [i].x =
						cols - strlen (hotlist_but [i].text) - 13;
				else
					hotlist_but [i].x = cur_x [row];
			}

			cur_x [row] += strlen (hotlist_but [i].text) + 2
				+ (hotlist_but [i].flags == DEFPUSH_BUTTON ? 5 : 3);
		}
	}

	return cols;
}

static void init_hotlist (int list_type)
{
    int i;
	int hotlist_cols = init_i18n_stuff (list_type, COLS - 6);

    do_refresh ();

    hotlist_state.expanded = GetPrivateProfileInt ("HotlistConfig",
			    "expanded_view_of_groups", 0, profile_name);

    hotlist_dlg = create_dlg (0, 0, LINES-2, hotlist_cols, dialog_colors,
			      hotlist_callback,
			      list_type == LIST_VFSLIST ? "[vfshot]" : "[Hotlist]",
			      DLG_CENTER|DLG_GRID);
    x_set_dialog_title (hotlist_dlg,
	list_type == LIST_VFSLIST ? "Active VFS directories" : "Directory hotlist");

#define XTRACT_HOTLIST(i) BY_HOTLIST+hotlist_but[i].y, BX_HOTLIST+hotlist_but[i].x, hotlist_but[i].ret_cmd, hotlist_but[i].flags, hotlist_but[i].text, hotlist_button_callback, 0

    for (i = 0; i < BUTTONS_HOTLIST; i++) {
	if (hotlist_but[i].type & list_type)
	    add_widgetl (hotlist_dlg, button_new (XTRACT_HOTLIST (i)), (i == BUTTONS_HOTLIST - 1) ?
		XV_WLAY_CENTERROW : XV_WLAY_RIGHTOF);
    }
#undef XTRACT_HOTLIST

    /* We add the labels.
     *    ho_pname       will hold entry's pathname;
     *    pname_group will hold name of current group
     */
    ho_pname = label_new (UY_HOTLIST-11+LINES, UX_HOTLIST+2, "");
    add_widget (hotlist_dlg, ho_pname);
    if (!hotlist_state.moving) {
	add_widget (hotlist_dlg, label_new (UY_HOTLIST-12+LINES, UX_HOTLIST+1, " Directory path "));

	/* This one holds the displayed pathname */
	pname_group = label_new (UY_HOTLIST, UX_HOTLIST+1, " Directory label ");
	add_widget (hotlist_dlg, pname_group);
    }
    /* get new listbox */
    l_hotlist = listbox_new (UY_HOTLIST + 1, UX_HOTLIST + 1, COLS-2*UX_HOTLIST-8, LINES-14, listbox_cback, ho_l_call);

    /* Fill the hotlist with the active VFS or the hotlist */
    if (list_type == LIST_VFSLIST) {
	listbox_add_item (l_hotlist, 0, 0, home_dir, 0);
	vfs_fill_names (add_name_to_list);
    } else
	fill_listbox ();

    add_widgetl (hotlist_dlg, l_hotlist, XV_WLAY_EXTENDWIDTH);
    /* add listbox to the dialogs */
}

static void init_movelist (int list_type, struct hotlist *item)
{
    int i;
    char *hdr = xasprintf ("Moving %s", item->label);
	int movelist_cols = init_i18n_stuff (list_type, COLS - 6);

    do_refresh ();

    movelist_dlg = create_dlg (0, 0, LINES-6, movelist_cols, dialog_colors,
			      hotlist_callback, "[Hotlist]", DLG_CENTER|DLG_GRID);
    x_set_dialog_title (movelist_dlg, hdr);
    free (hdr);

#define XTRACT_HOTLIST(i) BY_HOTLIST-4+hotlist_but[i].y, BX_HOTLIST+hotlist_but[i].x, hotlist_but[i].ret_cmd, hotlist_but[i].flags, hotlist_but[i].text, hotlist_button_callback, 0

    for (i = 0; i < BUTTONS_HOTLIST; i++) {
	if (hotlist_but[i].type & list_type)
	    add_widgetl (movelist_dlg, button_new (XTRACT_HOTLIST (i)), (i == BUTTONS_HOTLIST - 1) ?
		XV_WLAY_CENTERROW : XV_WLAY_RIGHTOF);
    }

    /* We add the labels.  We are interested in the last one,
     * that one will hold the path name label
     */
    movelist_group = label_new (UY_HOTLIST, UX_HOTLIST+1, " Directory label ");
    add_widget (movelist_dlg, movelist_group);
    /* get new listbox */
    l_movelist = listbox_new (UY_HOTLIST + 1, UX_HOTLIST + 1,
		movelist_dlg->cols - 2*UX_HOTLIST - 2, movelist_dlg->lines - 8,
		listbox_cback, ho_l_call);

    fill_listbox ();

    add_widgetl (movelist_dlg, l_movelist, XV_WLAY_EXTENDWIDTH);
    /* add listbox to the dialogs */
}

static void hotlist_done (void)
{
    destroy_dlg (hotlist_dlg);
    if (0)
	update_panels (UP_OPTIMIZE, UP_KEEPSEL);
    repaint_screen ();
}

static char *
find_group_section (struct hotlist *grp)
{
    return xasprintf ("%s.Group", grp->directory);

}


/* 1.11.96 bor: added pos parameter to control placement of new item.
   see widget.c, listbox_add_item()
		now hotlist is in unsorted mode
 */
static struct hotlist *
add2hotlist (char *label, char *directory, enum HotListType type, int pos)
{
    struct hotlist *current_hl = NULL;
    struct hotlist *new;

    if (l_hotlist && l_hotlist->current)
	current_hl = l_hotlist->current->data;

    new = new_hotlist ();

    new->type      = type;
    new->label     = label;
    new->directory = directory;
    new->up        = current_group;

    if (!current_group->head) { /* first element in group */
	current_group->head = new;
    } else if (pos == 2) { /* should be appended after current */
	new->next     = current_hl->next;
	current_hl->next = new;
    } else if (pos == 1 &&
	       current_hl == current_group->head) {
			   /* should be inserted before first item */
	new->next = current_hl;
	current_group->head = new;
    } else if (pos == 1) { /* befor current */
	struct hotlist  *p = current_group->head;

	while (p->next != current_hl)
	    p = p->next;

	new->next = current_hl;
	p->next   = new;
    } else {                /* append at the end */
	struct hotlist  *p = current_group->head;

	while (p->next)
	    p = p->next;

	p->next = new;
    }

    if (hotlist_state.running && type != HL_TYPE_COMMENT) {
	if (type == HL_TYPE_GROUP) {
	    char  *lbl = xasprintf ("->%s", new->label);

	    listbox_add_item (l_hotlist, pos, 0, lbl, new);
	    free (lbl);
	} else
	    listbox_add_item (l_hotlist, pos, 0, new->label, new);
	listbox_select_entry (l_hotlist, l_hotlist->current);
    }
    return new;

}

/*
 * Support routine for add_new_entry_input()/add_new_group_input()
 * Change positions of buttons (first three widgets).
 *
 * This is just a quick hack. Accurate procedure must take care of
 * internationalized label lengths and total buttonbar length...assume
 * 64 is longer anyway.
 */

static int add_new_entry_input (char *header, char *text1, char *text2, char *help, char **r1, char **r2)
{
    QuickDialog Quick_input;
    static const QuickWidget quick_widgets_st [] = {
      { quick_button, 55, 80, 4, 0, "&Cancel", 0, B_CANCEL, 0, 0,
	  XV_WLAY_DONTCARE, "button-cancel" },
      { quick_button, 30, 80, 4, 0, "&Insert", 0, B_INSERT, 0, 0,
	  XV_WLAY_DONTCARE, "button-insert" },
      { quick_button, 10, 80, 4, 0, "&Append", 0, B_APPEND, 0, 0,
	  XV_WLAY_DONTCARE, "button-append" },
      { quick_input,  4, 80, 4, 0, "",58, 0, 0, 0, XV_WLAY_BELOWCLOSE, "input-pth" },
      { quick_label,  3, 80, 3, 0, 0, 0, 0, 0, 0, XV_WLAY_DONTCARE, "label-pth" },
      { quick_input,  4, 80, 3, 0, "", 58, 0, 0, 0, XV_WLAY_BELOWCLOSE, "input-lbl" },
      { quick_label,  3, 80, 2, 0, 0, 0, 0, 0, 0, XV_WLAY_DONTCARE, "label-lbl" },
      { 0 }
    };

    int len;
    int i;
    int lines1, lines2;
    char *my_str1, *my_str2;
    QuickWidget *quick_widgets;

    quick_widgets = alloca(sizeof(quick_widgets_st));
    memcpy(quick_widgets, quick_widgets_st, sizeof(quick_widgets_st));

    len = max (strlen (header), msglen (text1, &lines1));
    len = max (len, msglen (text2, &lines2)) + 4;
    len = max (len, 64);

    Quick_input.xlen  = len;
    Quick_input.xpos  = -1;
    Quick_input.title = header;
    Quick_input.help  = help;
    quick_widgets [6].text = text1;
    quick_widgets [4].text = text2;
    quick_widgets [5].text = *r1;
    quick_widgets [3].text = *r2;

    for (i = 0; i < 7; i++)
	quick_widgets [i].y_divisions = lines1+lines2+7;
    Quick_input.ylen  = lines1 + lines2 + 7;

    quick_widgets [0].relative_y += (lines1 + lines2);
    quick_widgets [1].relative_y += (lines1 + lines2);
    quick_widgets [2].relative_y += (lines1 + lines2);
    quick_widgets [3].relative_y += (lines1);
    quick_widgets [4].relative_y += (lines1);

    quick_widgets [5].str_result = &my_str1;
    quick_widgets [3].str_result = &my_str2;

    Quick_input.widgets = quick_widgets;
    if ((i = quick_dialog (&Quick_input)) != B_CANCEL) {
	 *r1 = *(quick_widgets [5].str_result);
	 *r2 = *(quick_widgets [3].str_result);
	 return i;
    } else
	 return 0;
}

static void add_new_entry_cmd (void)
{
    char *title = 0, *url = 0;
    int ret;

    /* Take current directory as default value for input fields */
    title = url = cpanel->cwd;

    ret = add_new_entry_input ("New hotlist entry", "Directory label", "Directory path",
	 "[Hotlist]", &title, &url);

    if (!ret || !title || !*title || !url || !*url)
	return;

    if (ret == B_ENTER || ret == B_APPEND)
	add2hotlist (strdup (title), strdup (url), HL_TYPE_ENTRY, 2);
    else
	add2hotlist (strdup (title), strdup (url), HL_TYPE_ENTRY, 1);

    hotlist_state.modified = 1;
}

static int add_new_group_input (char *header, char *label, char **result)
{
    int         ret;
    QuickDialog Quick_input;
    static const QuickWidget quick_widgets_st [] = {
      { quick_button, 55, 80, 1, 0, "&Cancel", 0, B_CANCEL, 0, 0,
	  XV_WLAY_DONTCARE, "button-cancel" },
      { quick_button, 30, 80, 1, 0, "&Insert", 0, B_INSERT, 0, 0,
	  XV_WLAY_DONTCARE, "button-insert" },
      { quick_button, 10, 80, 1, 0, "&Append", 0, B_APPEND, 0, 0,
	  XV_WLAY_DONTCARE, "button-append" },
      { quick_input,  4, 80,  0, 0, "", 58, 0, 0, 0, XV_WLAY_BELOWCLOSE, "input" },
      { quick_label,  3, 80,  2, 0,  0,  0, 0, 0, 0, XV_WLAY_DONTCARE, "label" },
      { 0 }
    };

    int len;
    int i;
    int lines;
    char *my_str;
    QuickWidget *quick_widgets;

    quick_widgets = alloca(sizeof(quick_widgets_st));
    memcpy(quick_widgets, quick_widgets_st, sizeof(quick_widgets_st));

    len = max (strlen (header), msglen (label, &lines)) + 4;
    len = max (len, 64);

    Quick_input.xlen  = len;
    Quick_input.xpos  = -1;
    Quick_input.title = header;
    Quick_input.help  = "[Hotlist]";
    quick_widgets [4].text = label;

    for (i = 0; i < 5; i++)
	quick_widgets [i].y_divisions = lines+6;
    Quick_input.ylen  = lines + 6;

    for (i = 0; i < 4; i++)
	quick_widgets [i].relative_y += 2 + lines;

    quick_widgets [3].str_result = &my_str;
    quick_widgets [3].text       = "";

    Quick_input.widgets = quick_widgets;
    if ((ret = quick_dialog (&Quick_input)) != B_CANCEL) {
	*result = *(quick_widgets [3].str_result);
	return ret;
    } else
	return 0;
}

static void add_new_group_cmd (void)
{
    char   *label;
    int         ret;

    ret = add_new_group_input (" New hotlist group ", "Name of new group", &label);
    if (!ret || !label || !*label)
	return;

    if (ret == B_ENTER || ret == B_APPEND)
	add2hotlist (label, 0, HL_TYPE_GROUP, 2);
    else
	add2hotlist (label, 0, HL_TYPE_GROUP, 1);

    hotlist_state.modified = 1;
}

static void add2hotlist_cmd (void)
{
    char *prompt, *label;
	char* cp = "Label for \"%s\":";
	int l = strlen(cp);

    prompt = x_malloc (strlen (cpanel->cwd) + l, "add2hotlist_cmd");
    sprintf (prompt, cp, name_trunc (cpanel->cwd, COLS-2*UX_HOTLIST-(l+8)));
    label = input_dialog (" Add to hotlist ", prompt, cpanel->cwd);
    free (prompt);
    if (!label || !*label)
	return;

    add2hotlist (label, strdup (cpanel->cwd), HL_TYPE_ENTRY, 0);
    hotlist_state.modified = 1;
}

static void remove_group (struct hotlist *grp)
{
    struct hotlist *current = grp->head;

    while (current) {
	struct hotlist *next = current->next;

	if (current->type == HL_TYPE_GROUP)
	    remove_group (current);

	if (current->label)
	    free (current->label);
	if (current->directory)
	    free (current->directory);
	free (current);

	current = next;
    }

}

static void remove_from_hotlist (struct hotlist *entry)
{
    if (entry->type == HL_TYPE_GROUP) {
	if (entry->head) {
	    char *header;
	    int   result;

	    header = xasprintf(" Remove: %s ", name_trunc(entry->label, 30));
	    result = query_dialog (header, "\n Group not empty.\n Remove it?",
				   D_ERROR, 2,
				   "&No", "&Yes");
	    free (header);

	    if (!result)
		return;
	}

	remove_group (entry);
    }

    unlink_entry (entry);

    if (entry->label)
	  free (entry->label);
    if (entry->directory)
	free (entry->directory);
    free (entry);
    /* now remove list entry from screen */
    listbox_remove_current (l_hotlist, 1);
    hotlist_state.modified = 1;
}

char *hotlist_cmd (int vfs_or_hotlist)
{
    char *target = NULL;

    hotlist_state.type = vfs_or_hotlist;
    load_hotlist ();

    init_hotlist (vfs_or_hotlist);

    /* display file info */
    attrset (SELECTED_COLOR);

    hotlist_state.running = 1;
    run_dlg (hotlist_dlg);
    hotlist_state.running = 0;
    save_hotlist ();

    switch (hotlist_dlg->ret_value) {
    case B_CANCEL:
	break;

    case B_ENTER:
	if (l_hotlist->current->data) {
	    struct hotlist *hlp = (struct hotlist*) l_hotlist->current->data;
	    target = strdup (hlp->directory);
	} else
	    target = strdup (l_hotlist->current->text);
	break;
    }

    hotlist_done ();
    return target;
}

static void load_group (struct hotlist *grp)
{
    void *profile_keys;
    char *key, *value;
    char *group_section;
    struct hotlist *current = 0;

    group_section = find_group_section (grp);

    profile_keys = profile_init_iterator (group_section, profile_name);

    current_group = grp;

    while (profile_keys) {
	profile_keys = profile_iterator_next (profile_keys, &key, &value);
	add2hotlist (strdup (value), strdup (key), HL_TYPE_GROUP, 0);
    }
    free (group_section);

    profile_keys = profile_init_iterator (grp->directory, profile_name);

    while (profile_keys) {
	profile_keys = profile_iterator_next (profile_keys, &key, &value);
	add2hotlist (strdup (value), strdup (key), HL_TYPE_ENTRY, 0);
    }

    for (current = grp->head; current; current = current->next)
	load_group (current);
}

#define TKN_GROUP       0
#define TKN_ENTRY       1
#define TKN_STRING      2
#define TKN_URL         3
#define TKN_ENDGROUP    4
#define TKN_COMMENT     5
#define TKN_EOL         125
#define TKN_EOF         126
#define TKN_UNKNOWN     127

static int hot_skip_blanks (void)
{
    int c;

    while ((c = getc (hotlist_file)) != EOF && c != '\n' && isspace (c))
	;
    return c;

}

static int hot_next_token (void)
{
    int c;
    int tkn_length;

#define CHECK_BUF() \
do { \
    if (tkn_length == tkn_buf_length) \
	tkn_buf = xrealloc (tkn_buf, tkn_buf_length += 1024); \
} while (0)

    tkn_length = 0;

again:
    c = hot_skip_blanks ();
    switch (c) {
    case EOF:
	return TKN_EOF;
	break;
    case '\n':
	return TKN_EOL;
	break;
    case '#':
	while ((c = getc (hotlist_file)) != EOF && c != '\n') {
	    if (c == EOF)
		return TKN_EOF;
	    if (c != '\n') {
		CHECK_BUF();
		tkn_buf[tkn_length++] = c == '\n' ? ' ' : c;
	    }
	}
	CHECK_BUF();
	tkn_buf[tkn_length] = '\0';
	return TKN_COMMENT;
	break;
    case '"':
	while ((c = getc (hotlist_file)) != EOF && c != '"') {
	    if (c == '\\')
		if ((c = getc (hotlist_file)) == EOF)
		    return TKN_EOF;
	    CHECK_BUF();
	    tkn_buf[tkn_length++] = c == '\n' ? ' ' : c;
	}
	if (c == EOF)
	    return TKN_EOF;
	CHECK_BUF();
	tkn_buf[tkn_length] = '\0';
	return TKN_STRING;
	break;
    case '\\':
	if ((c = getc (hotlist_file)) == EOF)
	    return TKN_EOF;
	if (c == '\n')
	    goto again;

	/* fall through; it is taken as normal character */

    default:
	do {
	    CHECK_BUF();
	    tkn_buf[tkn_length++] = toupper(c);
	} while ((c = fgetc (hotlist_file)) != EOF && isalnum (c));
	if (c != EOF)
	    ungetc (c, hotlist_file);
	CHECK_BUF();
	tkn_buf[tkn_length] = '\0';
	if (strncmp (tkn_buf, "GROUP", tkn_length) == 0)
	    return TKN_GROUP;
	else if (strncmp (tkn_buf, "ENTRY", tkn_length) == 0)
	    return TKN_ENTRY;
	else if (strncmp (tkn_buf, "ENDGROUP", tkn_length) == 0)
	    return TKN_ENDGROUP;
	else if (strncmp (tkn_buf, "URL", tkn_length) == 0)
	    return TKN_URL;
	else
	    return TKN_UNKNOWN;
	break;
    }
}

#define SKIP_TO_EOL     { \
int _tkn; \
while ((_tkn = hot_next_token ()) != TKN_EOF && _tkn != TKN_EOL) ; \
}

#define CHECK_TOKEN(_TKN_) \
if ((tkn = hot_next_token ()) != _TKN_) { \
    hotlist_state.readonly = 1; \
    hotlist_state.file_error = 1; \
    while (tkn != TKN_EOL && tkn != TKN_EOF) \
	tkn = hot_next_token (); \
break; \
}

static void
hot_load_group (struct hotlist * grp)
{
    int         tkn;
    struct hotlist *new_grp;
    char        *label, *url;

    current_group = grp;

    while ((tkn = hot_next_token()) != TKN_ENDGROUP)
	switch (tkn) {
	case TKN_GROUP:
	    CHECK_TOKEN(TKN_STRING);
	    new_grp = add2hotlist (strdup (tkn_buf), 0, HL_TYPE_GROUP, 0);
	    SKIP_TO_EOL;
	    hot_load_group (new_grp);
	    current_group = grp;
	    break;
	case TKN_ENTRY:
	    CHECK_TOKEN(TKN_STRING);
	    label = strdup (tkn_buf);
	    CHECK_TOKEN(TKN_URL);
	    CHECK_TOKEN(TKN_STRING);
	    url = strdup (tkn_buf);
	    add2hotlist (label, url, HL_TYPE_ENTRY, 0);
	    SKIP_TO_EOL;
	    break;
	case TKN_COMMENT:
	    label = strdup (tkn_buf);
	    add2hotlist (label, 0, HL_TYPE_COMMENT, 0);
	    break;
	case TKN_EOF:
	    hotlist_state.readonly = 1;
	    hotlist_state.file_error = 1;
	    return;
	    break;
	case TKN_EOL:
	    /* skip empty lines */
	    break;
	default:
	    hotlist_state.readonly = 1;
	    hotlist_state.file_error = 1;
	    SKIP_TO_EOL;
	    break;
	}
    SKIP_TO_EOL;
}

static void
hot_load_file (struct hotlist * grp)
{
    int         tkn;
    struct hotlist *new_grp;
    char        *label, *url;

    current_group = grp;

    while ((tkn = hot_next_token())!= TKN_EOF)
	switch (tkn) {
	case TKN_GROUP:
	    CHECK_TOKEN(TKN_STRING);
	    new_grp = add2hotlist (strdup (tkn_buf), 0, HL_TYPE_GROUP, 0);
	    SKIP_TO_EOL;
	    hot_load_group (new_grp);
	    current_group = grp;
	    break;
	case TKN_ENTRY:
	    CHECK_TOKEN(TKN_STRING);
	    label = strdup (tkn_buf);
	    CHECK_TOKEN(TKN_URL);
	    CHECK_TOKEN(TKN_STRING);
	    url = strdup (tkn_buf);
	    add2hotlist (label, url, HL_TYPE_ENTRY, 0);
	    SKIP_TO_EOL;
	    break;
	case TKN_COMMENT:
	    label = strdup (tkn_buf);
	    add2hotlist (label, 0, HL_TYPE_COMMENT, 0);
	    break;
	case TKN_EOL:
	    /* skip empty lines */
	    break;
	default:
	    hotlist_state.readonly = 1;
	    hotlist_state.file_error = 1;
	    SKIP_TO_EOL;
	    break;
	}
}

static void
clean_up_hotlist_groups (const char *section)
{
    char        *grp_section;
    void        *profile_keys;
    char        *key, *value;

    grp_section = xasprintf ("%s.Group", section);
    if (profile_has_section (section, profile_name))
	profile_clean_section (section, profile_name);
    if (profile_has_section (grp_section, profile_name)) {
	profile_keys = profile_init_iterator (grp_section, profile_name);

	while (profile_keys) {
	    profile_keys = profile_iterator_next (profile_keys, &key, &value);
	    clean_up_hotlist_groups (key);
	}
	profile_clean_section (grp_section, profile_name);
    }
    free (grp_section);
}

void load_hotlist (void)
{
    char        *grp_section;
    int         has_old_list = 0;
    int         remove_old_list = 0;
    struct stat stat_buf;
    char *Msg;

    if (hotlist_state.loaded) {
	stat (hotlist_file_name, &stat_buf);
	if (hotlist_file_mtime < stat_buf.st_mtime)
	    done_hotlist ();
	else
	    return;
    }

    if (!hotlist_file_name)
	hotlist_file_name = concat_path_file (home_dir, HOTLIST_FILENAME);

    g_hotlist          = new_hotlist ();
    g_hotlist->type    = HL_TYPE_GROUP;
    g_hotlist->label   = strdup (" Top level group ");
    g_hotlist->up      = g_hotlist;
    /*
     * compatibility :-(
     */
    g_hotlist->directory = strdup ("Hotlist");

    grp_section = xstrdup ("Hotlist.Group");
    has_old_list = profile_has_section ("Hotlist", profile_name) ||
		   profile_has_section (grp_section, profile_name);
    free (grp_section);

    if ((hotlist_file = fopen (hotlist_file_name, "r")) == 0) {
	int     result;

	Msg = xstrdup ("Hotlist is now kept in file ~/"
			    HOTLIST_FILENAME
			    " MC will load hotlist from ~/"
			    PROFILE_NAME
			    " and then delete [Hotlist] section there");
	MC_message (0, " Hotlist Load ", Msg);
	free (Msg);

	load_group (g_hotlist);
	hotlist_state.loaded   = 1;
	/*
	 * just to be shure we got copy
	 */
	hotlist_state.modified = 1;
	result = save_hotlist ();
	hotlist_state.modified = 0;
	if (result) {
	    remove_old_list = 1;
	} else {
	    Msg = xstrdup ("MC was unable to write ~/" HOTLIST_FILENAME
			" file, your old hotlist entries were not deleted");

	    MC_message (D_ERROR, " Hotlist Load ", Msg);
	    free (Msg);
	}
    } else {
	hot_load_file (g_hotlist);
	fclose (hotlist_file);
	hotlist_state.loaded = 1;
	if (has_old_list) {
	    int         result;

	    Msg = xstrdup (
		    "You have ~/" HOTLIST_FILENAME " file and [Hotlist] section in ~/" PROFILE_NAME "\n"
		    "Your ~/" HOTLIST_FILENAME " most probably was created\n"
		    "by an earlier development version of MC\nand is more actual than ~/"
		    PROFILE_NAME " entries\n\n"
		    "You can choose between\n\n"
		      "  Remove - remove old hotlist entries from ~/" PROFILE_NAME "\n"
		      "  Keep   - keep your old entries; you will be asked\n"
		      "           the same question next time\n"
		      "  Merge  - add old entries to hotlist as group \"Entries from ~/"
		    PROFILE_NAME "\"\n\n");

	    result = query_dialog (" Hotlist Load ",
				   Msg, D_ERROR, 3, "&Remove", "&Keep", "&Merge");
	    if (result == 0)
		remove_old_list = 1;
	    else if (result == 2) {
		struct hotlist  *grp = g_hotlist->head;
		struct hotlist  *old;

		g_hotlist->head = 0;
		load_group (g_hotlist);

		old            = new_hotlist ();
		old->type      = HL_TYPE_GROUP;
		old->label     = xstrdup (" Entries from ~/" PROFILE_NAME);
		old->up        = g_hotlist;
		old->head      = g_hotlist->head;
		old->next      = grp;
		g_hotlist->head  = old;
		hotlist_state.modified = 1;
		if (!save_hotlist ()) {
		    char *str;

		    str = xstrdup ("MC was unable to write ~/" HOTLIST_FILENAME
					" file your old hotlist entries were not deleted");

		    MC_message (D_ERROR, " Hotlist Load ", str);
		    free (str);
		} else
		    remove_old_list = 1;
		hotlist_state.modified = 0;
	    }
	}
    }

    if (remove_old_list) {
	clean_up_hotlist_groups ("Hotlist");
	sync_profiles ();
    }

    stat (hotlist_file_name, &stat_buf);
    hotlist_file_mtime = stat_buf.st_mtime;
    current_group = g_hotlist;
}

static void hot_save_group (struct hotlist *grp)
{
    struct hotlist *current = grp->head;
    int             i;
    char           *s;

#define INDENT(n) \
do { \
    for (i = 0; i < n; i++) \
	putc (' ', hotlist_file); \
} while (0)

    for (;current; current = current->next)
	switch (current->type) {
	case HL_TYPE_GROUP:
	    INDENT (list_level);
	    fputs ("GROUP \"", hotlist_file);
	    for (s = current->label; *s; s++) {
		if (*s == '"')
		    putc ('\\', hotlist_file);
		else if (*s == '\\')
		    putc ('\\', hotlist_file);
		putc (*s, hotlist_file);
	    }
	    fputs ("\"\n", hotlist_file);
	    list_level += 2;
	    hot_save_group (current);
	    list_level -= 2;
	    INDENT (list_level);
	    fputs ("ENDGROUP\n", hotlist_file);
	    break;
	case HL_TYPE_ENTRY:
	    INDENT(list_level);
	    fputs ("ENTRY \"", hotlist_file);
	    for (s = current->label; *s; s++) {
		if (*s == '"')
		    putc ('\\', hotlist_file);
		else if (*s == '\\')
		    putc ('\\', hotlist_file);
		putc (*s, hotlist_file);
	    }
	    fputs ("\" URL \"", hotlist_file);
	    for (s = current->directory; *s; s++) {
		if (*s == '"')
		    putc ('\\', hotlist_file);
		else if (*s == '\\')
		    putc ('\\', hotlist_file);
		putc (*s, hotlist_file);
	    }
	    fputs ("\"\n", hotlist_file);
	    break;
	case HL_TYPE_COMMENT:
	    fprintf (hotlist_file, "#%s\n", current->label);
	    break;

	}
}

int save_hotlist (void)
{
    int         saved = 0;
    struct      stat stat_buf;

    if (!hotlist_state.readonly && hotlist_state.modified && hotlist_file_name) {
	char    *fbak = xasprintf ("%s.bak", hotlist_file_name);

	rename (hotlist_file_name, fbak);
	if ((hotlist_file = fopen (hotlist_file_name, "w")) != 0) {
	    if (stat (fbak, &stat_buf) == 0)
		chmod (hotlist_file_name, stat_buf.st_mode);
	    else
		chmod (hotlist_file_name, S_IRUSR | S_IWUSR);
	    hot_save_group (g_hotlist);
	    fflush (hotlist_file);
	    fclose (hotlist_file);
	    stat (hotlist_file_name, &stat_buf);
	    hotlist_file_mtime = stat_buf.st_mtime;
	    saved = 1;
	    hotlist_state.modified = 0;
	} else
	    rename (fbak, hotlist_file_name);
	free (fbak);
    }

    return saved;
}

void done_hotlist (void)
{
    remove_group (g_hotlist);
    hotlist_state.loaded = 0;
    if (g_hotlist->label)
	free (g_hotlist->label);
    if (g_hotlist->directory)
	free (g_hotlist->directory);
    free (g_hotlist);
    if (hotlist_file_name)
	free (hotlist_file_name);
    hotlist_file_name = 0;
    g_hotlist = current_group = 0;
    l_hotlist = 0;
    current_group = 0;
}

static void print_bytesize (off_t size, int scale)
{
    char buffer [10];
    char scales[] = " kMGT";

    if (size > 0) {
	while (size > 9999 && scale < sizeof (scales)) {
	    size = (size + 512) / 1024;
	    scale ++;
	}
    }
    if (scale > 0)
	sprintf (buffer, "%4d %cb", (int)size, scales[scale]);
    else
	sprintf (buffer, "%4d b ", (int)size);
    printw (buffer);
}


/* Panel managing. */

static int info_event (Gpm_Event *event, WInfo *info)
{
    return 0;
}

static void info_box (Dlg_head *h, WInfo *info)
{
    standend ();
    if (hascolors)
	attrset (NORMAL_COLOR);
    widget_erase (&info->widget);
    draw_double_box (h, info->widget.y,  info->widget.x,
		     info->widget.lines, info->widget.cols);
}

static void info_show_info (WInfo *info)
{
    struct stat buf;

    if (!is_idle ())
	return;

    info_box (info->widget.parent, info);
    attrset (MARKED_COLOR);
    widget_move (&info->widget, 1, 3);
    printw ("Midnight Commander %s", VERSION);
    attrset (NORMAL_COLOR);
    widget_move (&info->widget, 2, 1);
    /* .ado: info->widget.x has wrong value (==0) on NT and OS/2, why? */
    hline (ACS_HLINE|NORMAL_COLOR, info->widget.x-2);
    if (get_current_type () != view_listing)
	return;

    if (!info->ready)
	return;

    my_statfs (cpanel->cwd);
    buf = cpanel->dir.list [cpanel->selected].buf;
    /* Print only lines which fit */
    switch (info->widget.y-2) {
	/* Note: all cases are fall-throughs */

    default:

    case 16:
	widget_move (&info->widget, 16, 3);
	if (myfs_stats.nfree >0 || myfs_stats.nodes > 0)
	    printw ("Free nodes %d (%d%%) of %d",
		    myfs_stats.nfree,
		    myfs_stats.total
		    ? 100 * myfs_stats.nfree / myfs_stats.nodes : 0,
		    myfs_stats.nodes);
	else
	    addstr ("No node information");

    case 15:
	widget_move (&info->widget, 15, 3);
	if (myfs_stats.avail > 0 || myfs_stats.total > 0) {
	    addstr ("Free space ");
	    print_bytesize (myfs_stats.avail, 1);
	    printw (" (%d%%) of ", myfs_stats.total
		    ? (int)(100ll * myfs_stats.avail / myfs_stats.total) : 0);
	    print_bytesize (myfs_stats.total, 1);
	} else
	    addstr ("No space information");

    case 14:
	widget_move (&info->widget, 14, 3);
	printw ("Type:      %s ", myfs_stats.typename ? myfs_stats.typename : "non-local vfs");
	if (myfs_stats.type != 0xffff && myfs_stats.type != 0xffffffff)
	    printw (" (%Xh)", myfs_stats.type);

    case 13:
	widget_move (&info->widget, 13, 3);
	printw ("Device:    %s",
		name_trunc (myfs_stats.device, info->widget.cols - 15));
    case 12:
	widget_move (&info->widget, 12, 3);
	printw ("Filesystem: %s",
		name_trunc (myfs_stats.mpoint, info->widget.cols - 15));

    case 11:
	widget_move (&info->widget, 11, 3);
	printw ("Accessed:  %s", file_date (buf.st_atime));

    case 10:
	widget_move (&info->widget, 10, 3);
	printw ("Modified:  %s", file_date (buf.st_mtime));

    case 9:
	widget_move (&info->widget, 9, 3);
	printw ("Created:   %s", file_date (buf.st_ctime));

    case 8:
	widget_move (&info->widget, 8, 3);
	printw ("Size:      ");
	print_bytesize (buf.st_size, 0);
	printw (" (%d blocks)", buf.st_blocks);

    case 7:
	widget_move (&info->widget, 7, 3);
	printw ("Owner:     %s/%s", get_owner (buf.st_uid),
		get_group (buf.st_gid));

    case 6:
	widget_move (&info->widget, 6, 3);
	printw ("Links:     %d", buf.st_nlink);

    case 5:
	widget_move (&info->widget, 5, 3);
	printw ("Mode:      %s (%04o)",
		bb_mode_string (buf.st_mode, bb_mode_string_buf), buf.st_mode & 07777);

    case 4:
	widget_move (&info->widget, 4, 3);
	printw ("Location:  %Xh:%Xh", buf.st_dev, buf.st_ino);

    case 3:
	widget_move (&info->widget, 3, 2);
	/* .ado: fname is invalid if selected == 0 && ifno called from current panel */
	if (cpanel->selected) {
		printw ("File:       %s",
			name_trunc (cpanel->dir.list [cpanel->selected].fname,
				    info->widget.cols - 15));
	} else
		printw ("File:       None");

    case 2:
    case 1:
    case 0:
	;
    } /* switch */
}

static void info_hook (void *data)
{
    WInfo *info = (WInfo *) data;
    Widget *other_widget;

    other_widget = get_panel_widget (get_current_index ());
    if (!other_widget)
	return;
    if (dlg_overlap (&info->widget, other_widget))
	return;

    info->ready = 1;
    info_show_info (info);
}

static void info_destroy (WInfo *info)
{
    delete_hook (&select_file_hook, info_hook);
}

static int info_callback (Dlg_head *h, WInfo *info, int msg, int par)
{
    switch (msg) {

    case WIDGET_INIT:
	add_hook (&select_file_hook, info_hook, info);
	info->ready = 0;
	break;

    case WIDGET_DRAW:
	info_hook (info);
	info_show_info (info);
	return 1;

    case WIDGET_FOCUS:
	return 0;
    }
    return default_proc (h, msg, par);
}

static WInfo *info_new (void)
{
    WInfo *info = x_malloc (sizeof (WInfo), "panel_info");

    init_widget (&info->widget, 0, 0, 0, 0, (callback_fn)
		 info_callback, (destroy_fn) info_destroy,
		 (mouse_h) info_event);

    /* We do not want the cursor */
    widget_want_cursor (info->widget, 0);

#ifndef NO_INFOMOUNT
    init_my_statfs ();
#endif
    return info;
}

/* Keyboard support routines. */

/* This macros were stolen from gpm 0.15 */
#define GET_TIME(tv) (gettimeofday(&tv, (struct timezone *)NULL))
#define DIF_TIME(t1,t2) ((t2.tv_sec -t1.tv_sec) *1000+ \
			 (t2.tv_usec-t1.tv_usec)/1000)

/* timeout for old_esc_mode in usec */
#define ESCMODE_TIMEOUT 1000000

static int xgetch_second (void);

typedef struct {
    int code;
    char *seq;
    int action;
} key_define_t;

static const key_define_t mc_bindings [] = {
    { KEY_END,    ESC_STR ">", MCKEY_NOACTION },
    { KEY_HOME,   ESC_STR "<", MCKEY_NOACTION },

#ifdef linux
    /* Incredible, but many Linuxes still have old databases */
    { KEY_IC,     ESC_STR "[2~", MCKEY_NOACTION },
#endif
    { 0, 0, MCKEY_NOACTION },
};

/* Broken terminfo and termcap databases on xterminals */
static const key_define_t xterm_key_defines [] = {
    { KEY_F(1),   ESC_STR "OP",   MCKEY_NOACTION },
    { KEY_F(2),   ESC_STR "OQ",   MCKEY_NOACTION },
    { KEY_F(3),   ESC_STR "OR",   MCKEY_NOACTION },
    { KEY_F(4),   ESC_STR "OS",   MCKEY_NOACTION },
    { KEY_F(1),   ESC_STR "[11~", MCKEY_NOACTION },
    { KEY_F(2),   ESC_STR "[12~", MCKEY_NOACTION },
    { KEY_F(3),   ESC_STR "[13~", MCKEY_NOACTION },
    { KEY_F(4),   ESC_STR "[14~", MCKEY_NOACTION },
    { KEY_F(5),   ESC_STR "[15~", MCKEY_NOACTION },
    { KEY_F(6),   ESC_STR "[17~", MCKEY_NOACTION },
    { KEY_F(7),   ESC_STR "[18~", MCKEY_NOACTION },
    { KEY_F(8),   ESC_STR "[19~", MCKEY_NOACTION },
    { KEY_F(9),   ESC_STR "[20~", MCKEY_NOACTION },
    { KEY_F(10),  ESC_STR "[21~", MCKEY_NOACTION },
    { 0, 0, MCKEY_NOACTION },
};

static const key_define_t mc_default_keys [] = {
    { ESC_CHAR, ESC_STR, MCKEY_ESCAPE },
    { ESC_CHAR, ESC_STR ESC_STR, MCKEY_NOACTION },
    { 0, 0, MCKEY_NOACTION },
};

static void define_sequences (const key_define_t *kd)
{
    int i;

    for (i = 0; kd [i].code; i++)
	define_sequence(kd [i].code, kd [i].seq, kd [i].action);
}

/* This has to be called before slang_init or whatever routine
   calls any define_sequence */
static void init_key (void)
{
    /* This has to be the first define_sequence */
    /* So, we can assume that the first keys member has ESC */
    define_sequences (mc_default_keys);

    /* Terminfo on irix does not have some keys */
    if ((!strncmp (TERM_env, "iris-ansi", 9)) || (!strncmp (TERM_env, "xterm", 5)))
	define_sequences (xterm_key_defines);

    define_sequences (mc_bindings);

}

static void xmouse_get_event (Gpm_Event *ev)
{
    int btn;

    /* Decode Xterm mouse information to a GPM style event */

    /* Variable btn has following meaning: */
    /* 0 = btn1 dn, 1 = btn2 dn, 2 = btn3 dn, 3 = btn up */
    btn = xgetch () - 32;

    /* There seems to be no way of knowing which button was released */
    /* So we assume all the buttons were released */

    if (btn == 3) {
	ev->type = GPM_UP | (GPM_SINGLE << clicks);
	ev->buttons = 0;
	GET_TIME (tv1);
	clicks = 0;
    } else {
	struct timeval tv2;

	ev->type = GPM_DOWN;
	GET_TIME (tv2);
	if (tv1.tv_sec && (DIF_TIME (tv1,tv2) < double_click_speed)) {
	    clicks++;
	    clicks %= 3;
	} else
	    clicks = 0;

	switch (btn) {
	case 0:
	    ev->buttons = GPM_B_LEFT;
	    break;
	case 1:
	    ev->buttons = GPM_B_MIDDLE;
	    break;
	case 2:
	    ev->buttons = GPM_B_RIGHT;
	    break;
	default:
	    /* Nothing */
	    break;
	}
    }
    /* Coordinates are 33-based */
    /* Transform them to 1-based */
    ev->x = xgetch () - 32;
    ev->y = xgetch () - 32;
}

static key_def *create_sequence (const char *seq, int code, int action)
{
    key_def *base, *p, *attach;

    for (base = attach = NULL; *seq; seq++) {
	p = x_malloc (sizeof (key_def), "create_sequence");
	if (!base) base = p;
	if (attach) attach->child = p;

	p->ch   = *seq;
	p->code = code;
	p->child = p->next = NULL;
	if (!seq[1])
	    p->action = action;
	else
	    p->action = MCKEY_NOACTION;
	attach = p;
    }
    return base;
}

static int push_char (int c)
{
    if (!seq_append)
	seq_append = seq_buffer;

    if (seq_append == &(seq_buffer [SEQ_BUFFER_SZ-2]))
	return 0;
    *(seq_append++) = c;
    *seq_append = 0;
    return 1;
}

static void define_sequence (int code, const char *seq, int action)
{
    key_def *base;

    if (strlen (seq) > (SEQ_BUFFER_SZ-1))
	return;

    for (base = keys; (base != 0) && *seq; ) {
	if (*seq == base->ch) {
	    if (base->child == 0) {
		if (*(seq+1)) {
		    base->child = create_sequence (seq+1, code, action);
		    return;
		} else {
		    /* The sequence clashes */
		    return;
		}
	    } else {
		base = base->child;
		seq++;
	    }
	} else {
	    if (base->next)
		base = base->next;
	    else {
		base->next = create_sequence (seq, code, action);
		return;
	    }
	}
    }
    keys = create_sequence (seq, code, action);
}

static int correct_key_code (int c)
{
    /* This is needed on some OS that do not support ncurses and */
    /* do some magic after read()ing the data */
    if (c == '\r')
	return '\n';

#ifdef IS_AIX
    if (c == KEY_SCANCEL)
	return '\t';
#endif

    if (c == KEY_F(0))
	return KEY_F(10);

    if (!alternate_plus_minus)
	switch (c) {
	    case KEY_KP_ADD: c = '+'; break;
	    case KEY_KP_SUBTRACT: c = '-'; break;
	    case KEY_KP_MULTIPLY: c = '*'; break;
	}

    return c;
}

int get_key_code (int no_delay)
{
    int c;

    if (no_delay != lastnodelay) {
	k_this = NULL;
	lastnodelay = no_delay;
    }

 pend_send:
    if (pending_keys) {
	int d = *pending_keys++;
 check_pend:
	if (!*pending_keys) {
	    pending_keys = 0;
	    seq_append = 0;
	}
	if (d == ESC_CHAR && pending_keys) {
	    d = ALT(*pending_keys++);
	    goto check_pend;
	}
	if ((d & 0x80) && use_8th_bit_as_meta)
	    d = ALT(d & 0x7f);
	k_this = NULL;
	return correct_key_code (d);
    }

 nodelay_try_again:
    if (no_delay)
	nodelay (TRUE);
    c = xgetch ();
    if (no_delay) {
	nodelay (FALSE);
	if (c == ERR) {
	    if (k_this != NULL && k_parent != NULL &&
		k_parent->action == MCKEY_ESCAPE && old_esc_mode) {
		struct timeval current, timeout;

		if (esctime.tv_sec == -1)
		    return ERR;
		GET_TIME (current);
		timeout.tv_sec = ESCMODE_TIMEOUT / 1000000 + esctime.tv_sec;
		timeout.tv_usec = ESCMODE_TIMEOUT % 1000000 + esctime.tv_usec;
		if (timeout.tv_usec > 1000000) {
		    timeout.tv_usec -= 1000000;
		    timeout.tv_sec++;
		}
		if (current.tv_sec < timeout.tv_sec)
		    return ERR;
		if (current.tv_sec == timeout.tv_sec &&
		    current.tv_usec < timeout.tv_usec)
		    return ERR;
		k_this = NULL;
		pending_keys = seq_append = NULL;
		return ESC_CHAR;
	    }
	    return ERR;
	}
    } else if (c == ERR) {
	/* Maybe we got an incomplete match.
	   This we do only in delay mode, since otherwise
	   xgetch can return ERR at any time. */
	if (seq_append) {
	    pending_keys = seq_buffer;
	    goto pend_send;
	}
	k_this = NULL;
	return ERR;
    }

    /* Search the key on the root */
    if (!no_delay || k_this == NULL) {
	k_this = keys;
	k_parent = NULL;

	if ((c & 0x80) && use_8th_bit_as_meta) {
	    c &= ~0x7f;

	    /* The first sequence defined starts with esc */
	    k_parent = keys;
	    k_this = keys->child;
	}
    }
    while (k_this) {
	if (c == k_this->ch) {
	    if (k_this->child) {
		if (!push_char (c)) {
		    pending_keys = seq_buffer;
		    goto pend_send;
		}
		k_parent = k_this;
		k_this = k_this->child;
		if (k_parent->action == MCKEY_ESCAPE && old_esc_mode) {
		    if (no_delay) {
			GET_TIME (esctime);
			if (k_this == NULL) {
			    /* Shouldn't happen */
			    fprintf (stderr, "Internal error\n");
			    exit (1);
			}
			goto nodelay_try_again;
		    }
		    esctime.tv_sec = -1;
		    c = xgetch_second ();
		    if (c == ERR) {
			pending_keys = seq_append = NULL;
			k_this = NULL;
			return ESC_CHAR;
		    }
		} else {
		    if (no_delay)
			goto nodelay_try_again;
		    c = xgetch ();
		}
	    } else {
		/* We got a complete match, return and reset search */
		int code;

		pending_keys = seq_append = NULL;
		code = k_this->code;
		k_this = NULL;
		return correct_key_code (code);
	    }
	} else {
	    if (k_this->next)
		k_this = k_this->next;
	    else {
		if (k_parent != NULL && k_parent->action == MCKEY_ESCAPE) {
		    /* This is just to save a lot of define_sequences */
		    if (isalpha(c)
			|| (c == '\n') || (c == '\t') || (c == XCTRL('h'))
			|| (c == KEY_BACKSPACE) || (c == '!') || (c == '\r')
			|| c == 127 || c == '+' || c == '-' || c == '\\'
			|| c == '?')
			c = ALT(c);
		    else if (isdigit(c))
			c = KEY_F (c-'0');
		    else if (c == ' ')
			c = ESC_CHAR;
		    pending_keys = seq_append = NULL;
		    k_this = NULL;
		    return correct_key_code (c);
		}
		/* Did not find a match or {c} was changed in the if above,
		   so we have to return everything we had skipped
		 */
		push_char (c);
		pending_keys = seq_buffer;
		goto pend_send;
	    }
	}
    }
    k_this = NULL;
    return correct_key_code (c);
}

#ifndef PORT_HAS_FILE_HANDLERS
/* If set timeout is set, then we wait 0.1 seconds, else, we block */
static void try_channels (int set_timeout)
{
    struct timeval timeout;
    fd_set select_set_a;
    struct timeval *timeptr;
    int v;

    while (1) {
	FD_ZERO (&select_set_a);
	FD_SET  (SLang_TT_Read_FD, &select_set_a);        /* Add stdin */

	if (set_timeout) {
	    timeout.tv_sec = 0;
	    timeout.tv_usec = 100000;
	    timeptr = &timeout;
	} else
	    timeptr = 0;

	v = select (FD_SETSIZE, &select_set_a, NULL, NULL, timeptr);
	if (v > 0) {
	    if (FD_ISSET (SLang_TT_Read_FD, &select_set_a))
		return;
	}
    }
}

/* Workaround for System V Curses vt100 bug */
static int getch_with_delay (void)
{
    int c;

    /* This routine could be used on systems without mouse support,
       so we need to do the select check :-( */
    while (1) {
	if (!pending_keys)
	    try_channels (0);

	/* Try to get a character */
	c = get_key_code (0);
	if (c != ERR)
	    break;
	/* Failed -> wait 0.1 secs and try again */
	try_channels (1);
    }
    /* Success -> return the character */
    return c;
}

#endif /* !HAVE_FILE_HANDLERS */

/* Returns a character read from stdin with appropriate interpretation */
/* Also takes care of generated mouse events */
/* Returns EV_MOUSE if it is a mouse event */
/* Returns EV_NONE  if non-blocking or interrupt set and nothing was done */
static int get_event (Gpm_Event *event, int redo_event, int block)
{
    int c;
    int flag;                    /* Return value from select */
    struct timeval timeout;
    struct timeval *time_addr = NULL;

    if ((get_event_dirty == 3) || is_idle ()) {
	mc_refresh ();
	get_event_dirty = 1;
    } else
	get_event_dirty++;

    vfs_timeout_handler ();

    /* Ok, we use (event->x < 0) to signal that the event does not contain
       a suitable position for the mouse, so we can't use show_mouse_pointer
       on it.
    */
    if (event->x > 0) {
	if (!redo_event)
	    event->x = -1;
    }

    /* Repeat if using mouse */
    while ((xmouse_flag) && !pending_keys) {
	if (xmouse_flag) {
	    FD_ZERO (&select_set);
	    FD_SET  (SLang_TT_Read_FD, &select_set);

	    if (redo_event) {
		timeout.tv_usec = mou_auto_repeat * 1000;
		timeout.tv_sec = 0;

		time_addr = &timeout;
	    } else {
		int seconds;

		if ((seconds = vfs_timeouts ())) {
		    /* the timeout could be improved and actually be
		     * the number of seconds until the next vfs entry
		     * timeouts in the stamp list.
		     */

		    timeout.tv_sec = seconds;
		    timeout.tv_usec = 0;
		    time_addr = &timeout;
		} else
		    time_addr = NULL;
	    }

	    if (!block) {
		time_addr = &timeout;
		timeout.tv_sec = 0;
		timeout.tv_usec = 0;
	    }
	    enable_interrupt_key ();
	    flag = select (FD_SETSIZE, &select_set, NULL, NULL, time_addr);
	    disable_interrupt_key ();

	    /* select timed out: it could be for any of the following reasons:
	     * redo_event -> it was because of the MOU_REPEAT handler
	     * !block     -> we did not block in the select call
	     * else       -> 10 second timeout to check the vfs status.
	     */
	    if (flag == 0) {
		if (redo_event)
		    return EV_MOUSE;
		if (!block)
		    return EV_NONE;
		vfs_timeout_handler ();
	    }
	    if (flag == -1 && errno == EINTR)
		return EV_NONE;

	    if (FD_ISSET (SLang_TT_Read_FD, &select_set))
		break;
	}
    }
    c = block ? getch_with_delay () : get_key_code(1);

    if (c == MCKEY_MOUSE) { /* Mouse event */
	xmouse_get_event (event);
	return EV_MOUSE;
    }

    return c;
}

/* Returns a key press, mouse events are discarded */
static int mi_getch (void)
{
    Gpm_Event ev;
    int       key;

    ev.x = -1;
    while ((key = get_event (&ev, 0, 1)) == 0)
	;
    return key;
}

int xgetch_second (void)
{
    fd_set Read_FD_Set;
    int c;
    struct timeval timeout;

    timeout.tv_sec = ESCMODE_TIMEOUT / 1000000;
    timeout.tv_usec = ESCMODE_TIMEOUT % 1000000;
    nodelay (TRUE);
    FD_ZERO (&Read_FD_Set);
    FD_SET (SLang_TT_Read_FD, &Read_FD_Set);
    select (FD_SETSIZE, &Read_FD_Set, NULL, NULL, &timeout);
    c = xgetch ();
    nodelay (FALSE);
    return c;
}

static void learn_store_key (char *buffer, char **p, int c)
{
    if (*p - buffer > 253)
	return;
    if (c == ESC_CHAR) {
	*(*p)++ = '\\';
	*(*p)++ = 'e';
    } else if (c < ' ') {
	*(*p)++ = '^';
	*(*p)++ = c + 'a' - 1;
    } else if (c == '^') {
	*(*p)++ = '^';
	*(*p)++ = '^';
    } else
	*(*p)++ = (char) c;
}

static char *learn_key (void)
{
/* LEARN_TIMEOUT in usec */
#define LEARN_TIMEOUT 200000

    fd_set Read_FD_Set;
    struct timeval endtime;
    struct timeval timeout;
    int c = xgetch ();
    char buffer [256];
    char *p = buffer;

    while (c == ERR)
	c = xgetch (); /* Sanity check, should be unnecessary */
    learn_store_key (buffer, &p, c);
    GET_TIME (endtime);
    endtime.tv_usec += LEARN_TIMEOUT;
    if (endtime.tv_usec > 1000000) {
	endtime.tv_usec -= 1000000;
	endtime.tv_sec++;
    }
    nodelay (TRUE);
    for (;;) {
	while ((c = xgetch ()) == ERR) {
	    GET_TIME (timeout);
	    timeout.tv_usec = endtime.tv_usec - timeout.tv_usec;
	    if (timeout.tv_usec < 0)
		timeout.tv_sec++;
	    timeout.tv_sec = endtime.tv_sec - timeout.tv_sec;
	    if (timeout.tv_sec >= 0 && timeout.tv_usec > 0) {
		FD_ZERO (&Read_FD_Set);
		FD_SET (SLang_TT_Read_FD, &Read_FD_Set);
		select (FD_SETSIZE, &Read_FD_Set, NULL, NULL, &timeout);
	    } else
		break;
	}
	if (c == ERR)
	    break;
	learn_store_key (buffer, &p, c);
    }
    nodelay (FALSE);
    *p = 0;
    return strdup (buffer);
}

/* xterm and linux console only: set keypad to numeric or application
   mode. Only in application keypad mode it's possible to distinguish
   the '+' key and the '+' on the keypad ('*' and '-' ditto)*/
static void
numeric_keypad_mode (void)
{
    if (console_flag || xterm_flag) {
	fprintf (stdout, "\033>");
	fflush (stdout);
    }
}

static void
application_keypad_mode (void)
{
    if (console_flag || xterm_flag) {
	fprintf (stdout, "\033=");
	fflush (stdout);
    }
}

/* A function to check if we're idle.
   Currently checks only for key presses.
   We could also check the mouse. */
static int is_idle (void)
{
    /* Check for incoming key presses     *
     * If there are any we say we're busy */

    fd_set select_set_a;
    struct timeval timeout;
    FD_ZERO (&select_set_a);
    FD_SET (0, &select_set_a);
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;
    select (FD_SETSIZE, &select_set_a, 0, 0, &timeout);
    return ! FD_ISSET (0, &select_set_a);
}


static int get_modifier (void)
{
#ifdef __linux__
    unsigned char modifiers;

    modifiers = 6;

    if (ioctl (0, TIOCLINUX, &modifiers) < 0)
	return 0;

    return (int) modifiers;
#else
    return 0;
#endif
}

int ctrl_pressed ()
{
#ifdef __linux__
    if (get_modifier () & CONTROL_PRESSED)
	return 1;
#endif
    return 0;
}

/* Panel layout module for the Midnight Commander */



#define MINWIDTH 10
#define MINHEIGHT 5

#define BY_LAYOUT      12

#define B_2LEFT B_USER
#define B_2RIGHT B_USER + 1
#define B_PLUS_LAYOUT B_USER + 2
#define B_MINUS_LAYOUT B_USER + 3

static const char * const s_split_direction [2] = {
    "&Vertical",
    "&Horizontal"
};

static const char layout_title[] = " Layout ";
static const char la_title1[] = " Panel split ";
static const char la_title2[] = " Highlight... ";
static const char la_title3[] = " Other options ";
static const char output_lines_label[] = "output lines";

static void _check_split (void)
{
    if (_horizontal_split) {
	if (_equal_split)
	    _first_panel_size = la_height / 2;
	else if (_first_panel_size < MINHEIGHT)
	    _first_panel_size = MINHEIGHT;
	else if (_first_panel_size > la_height - MINHEIGHT)
	    _first_panel_size = la_height - MINHEIGHT;
    } else {
	if (_equal_split)
	    _first_panel_size = COLS / 2;
	else if (_first_panel_size < MINWIDTH)
	    _first_panel_size = MINWIDTH;
	else if (_first_panel_size > COLS - MINWIDTH)
	    _first_panel_size = COLS - MINWIDTH;
    }
}

static void update_split (void)
{
    /* Check split has to be done before testing if it changed, since
       it can change due to calling _check_split() as well*/
    _check_split ();

    /* To avoid setting the cursor to the wrong place */
    if ((old_first_panel_size == _first_panel_size) &&
	(old_horizontal_split == _horizontal_split)) {
	return;
    }

    old_first_panel_size = _first_panel_size;
    old_horizontal_split = _horizontal_split;

    attrset (COLOR_NORMAL);
    dlg_move (layout_dlg, 6, 6);
    printw ("%03d", _first_panel_size);
    dlg_move (layout_dlg, 6, 18);
    if (_horizontal_split)
	printw ("%03d", la_height - _first_panel_size);
    else
	printw ("%03d", COLS - _first_panel_size);
}

static int b2left_cback (int action, void *data)
{
    if (_equal_split) {
	/* Turn equal split off */
	_equal_split = 0;
	la_check_options [6].widget->state = la_check_options [6].widget->state & ~C_BOOL;
	dlg_select_widget (layout_dlg, la_check_options [6].widget);
	dlg_select_widget (layout_dlg, bleft_widget);
    }
    _first_panel_size++;
    return 0;
}

static int b2right_cback (int action, void *data)
{
    if (_equal_split) {
	/* Turn equal split off */
	_equal_split = 0;
	la_check_options [6].widget->state = la_check_options [6].widget->state & ~C_BOOL;
	dlg_select_widget (layout_dlg, la_check_options [6].widget);
	dlg_select_widget (layout_dlg, bright_widget);
    }
    _first_panel_size--;
    return 0;
}

static int l_bplus_cback (int action, void *data)
{
    if (_output_lines < 99)
	_output_lines++;
    return 0;
}

static int l_bminus_cback (int action, void *data)
{
    if (_output_lines > 0)
	_output_lines--;
    return 0;
}

static int layout_callback (struct Dlg_head *h, int Id, int Msg)
{
    int _command_prompt;
    int _menubar_visible;
    int _keybar_visible;
    int _message_visible;
    int _xterm_hintbar;

    switch (Msg) {
    case DLG_DRAW:
	/*When repainting the whole dialog (e.g. with C-l) we have to
	  update everything*/
	old_first_panel_size = -1;
	old_horizontal_split = -1;
	old_output_lines     = -1;
	attrset (COLOR_NORMAL);
	dlg_erase (h);
	draw_box (h, 1, 2, h->lines - 2, h->cols - 4);
	draw_box (h, 2, 4, 6, la_first_width);
	draw_box (h, 8, 4, 4, la_first_width);
	draw_box (h, 2, 5 + la_first_width, 10, la_second_width);

	attrset (COLOR_HOT_NORMAL);
	dlg_move (h, 1, (h->cols - strlen(layout_title))/2);
	addstr (layout_title);
	dlg_move (h, 2, 5);
	addstr (la_title1);
	dlg_move (h, 8, 5);
	addstr (la_title2);
	dlg_move (h, 2, 6 + la_first_width);
	addstr (la_title3);
	update_split ();
	dlg_move (h, 6, 13);
	addch ('=');
	if (console_flag) {
	    if (old_output_lines != _output_lines) {
		old_output_lines = _output_lines;
		attrset (COLOR_NORMAL);
		dlg_move (h, 9, 16 + la_first_width);
		addstr (output_lines_label);
		dlg_move (h, 9, 10 + la_first_width);
		printw ("%02d", _output_lines);
	    }
	}
	break;

    case DLG_POST_KEY:
	_equal_split = la_check_options [6].widget->state & C_BOOL;
	_menubar_visible = la_check_options [5].widget->state & C_BOOL;
	_command_prompt = la_check_options [4].widget->state & C_BOOL;
	_keybar_visible = la_check_options [2].widget->state & C_BOOL;
	_message_visible = la_check_options [1].widget->state & C_BOOL;
	_xterm_hintbar = la_check_options [0].widget->state & C_BOOL;
	if (console_flag) {
	    int minimum;
	    if (_output_lines < 0)
		_output_lines = 0;
	    la_height = LINES - _keybar_visible - _command_prompt -
		     _menubar_visible - _output_lines - _message_visible;
	    if (_message_visible && _xterm_hintbar && xterm_flag) la_height++;
	    minimum = MINHEIGHT * (1 + _horizontal_split);
	    if (la_height < minimum) {
		_output_lines -= minimum - la_height;
		la_height = minimum;
	    }
	} else {
	    la_height = LINES - _keybar_visible - _command_prompt -
		_menubar_visible - _output_lines - _message_visible;
	    if (_message_visible && _xterm_hintbar && xterm_flag) la_height++;
	}
	if (_horizontal_split != radio_widget->sel) {
	    _horizontal_split = radio_widget->sel;
	    if (_horizontal_split)
		_first_panel_size = la_height / 2;
	    else
		_first_panel_size = COLS / 2;
	}
	update_split ();
	if (console_flag) {
	    if (old_output_lines != _output_lines) {
		old_output_lines = _output_lines;
		attrset (COLOR_NORMAL);
		dlg_move (h, 9, 10 + la_first_width);
		printw ("%02d", _output_lines);
	    }
	}
	break;

    case DLG_END:
	break;
    }
    return 0;
}

static const char save_button[] = "&Save";

static void init_layout (void)
{
    int i = sizeof (s_split_direction) / sizeof(char*);

    if (!l_b1) {
		int l1;

		la_first_width = 19; /* length of line with '<' '>' buttons */
		la_check_options[0].text = "&Xterm hintbar";    la_check_options[0].variable = &xterm_hintbar;
		la_check_options[1].text = "h&Intbar visible";  la_check_options[1].variable = &message_visible;
		la_check_options[2].text = "&Keybar visible";   la_check_options[2].variable = &keybar_visible;
		la_check_options[3].text = "command &Prompt";   la_check_options[3].variable = &command_prompt;
		la_check_options[4].text = "show &Mini status"; la_check_options[4].variable = &show_mini_info;
		la_check_options[5].text = "menu&Bar visible";  la_check_options[5].variable = &menubar_visible;
		la_check_options[6].text = "&Equal split";      la_check_options[6].variable = &equal_split;
		la_check_options[7].text = "pe&Rmissions";      la_check_options[7].variable = &permission_mode;
		la_check_options[8].text = "&File types";       la_check_options[8].variable = &filetype_mode;

		while (i--) {
			l1 = strlen (s_split_direction [i]) + 7;
			if (l1 > la_first_width)
				la_first_width = l1;
		}

		for (i = 0; i <= 8; i++) {
			l1 = strlen (la_check_options[i].text) + 7;
			if (l1 > la_first_width)
				la_first_width = l1;
		}

		l1 = strlen (la_title1) + 1;
		if (l1 > la_first_width)
			la_first_width = l1;

		l1 = strlen (la_title2) + 1;
		if (l1 > la_first_width)
			la_first_width = l1;


		la_second_width = strlen (la_title3) + 1;
		for (i = 0; i < 6; i++) {
			l1 = strlen (la_check_options[i].text) + 7;
			if (l1 > la_second_width)
				la_second_width = l1;
		}
		if (console_flag) {
			l1 = strlen (output_lines_label) + 13;
			if (l1 > la_second_width)
				la_second_width = l1;
		}

		/*
		 * alex@bcs.zp.ua:
		 * To be completely correct, one need to check if layout_title
		 * does not exceed dialog length and total length of 3 buttons
		 * allows their placement in one row. But assuming this dialog
		 * is wide enough, I don't include such a tests.
		 *
		 * Now the last thing to do - properly space buttons...
		 */
		l1 = 11 + strlen (ok_button)   /* 14 - all brackets and inner space */
			+ strlen (save_button)     /* notice: it is 3 char less because */
			+ strlen (cancel_button);  /* of '&' char in button text */

		i = (la_first_width + la_second_width - l1) / 4;
		l_b1 = 5 + i;
		l_b2 = l_b1 + strlen(ok_button) + i + 6;
		l_b3 = l_b2 + strlen(save_button) + i + 4;
    }

    layout_dlg = create_dlg (0, 0, 15, la_first_width + la_second_width + 9,
		dialog_colors, layout_callback,
		"[Layout]", DLG_CENTER | DLG_GRID);

    x_set_dialog_title (layout_dlg, "Layout");

    add_widgetl (layout_dlg,
		button_new (BY_LAYOUT, l_b3, B_CANCEL, NORMAL_BUTTON, cancel_button, 0, 0),
		XV_WLAY_RIGHTOF);
    add_widgetl (layout_dlg,
		button_new (BY_LAYOUT, l_b2, B_EXIT, NORMAL_BUTTON, save_button, 0, 0),
		XV_WLAY_RIGHTOF);
    add_widgetl (layout_dlg,
		button_new (BY_LAYOUT, l_b1, B_ENTER, DEFPUSH_BUTTON, ok_button, 0, 0),
		XV_WLAY_CENTERROW);
    if (console_flag) {
	add_widget (layout_dlg,
		    button_new (9, 12 + la_first_width, B_MINUS_LAYOUT, NARROW_BUTTON, "&-",
			l_bminus_cback, 0));
	add_widget (layout_dlg,
		    button_new (9, 7 + la_first_width, B_PLUS_LAYOUT, NARROW_BUTTON, "&+",
			l_bplus_cback, 0));
    }

#define XTRACT_LAYOUT(i) *la_check_options[i].variable, la_check_options[i].text

    for (i = 0; i < 6; i++) {
	la_check_options [i].widget = check_new (8 - i, 7 + la_first_width, XTRACT_LAYOUT(i));
	add_widgetl (layout_dlg, la_check_options [i].widget, XV_WLAY_BELOWCLOSE);
    }
    la_check_options [8].widget = check_new (10, 6, XTRACT_LAYOUT(8));
    add_widgetl (layout_dlg, la_check_options [8].widget, XV_WLAY_BELOWCLOSE);
    la_check_options [7].widget = check_new (9, 6, XTRACT_LAYOUT(7));
    add_widgetl (layout_dlg, la_check_options [7].widget, XV_WLAY_BELOWCLOSE);

    _equal_split = equal_split;
    bright_widget = button_new(6, 15, B_2RIGHT, NARROW_BUTTON, "&>", b2right_cback, 0);
    add_widgetl (layout_dlg, bright_widget, XV_WLAY_RIGHTOF);
    bleft_widget = button_new (6, 9, B_2LEFT, NARROW_BUTTON, "&<", b2left_cback, 0);
    add_widgetl (layout_dlg, bleft_widget, XV_WLAY_RIGHTOF);
    la_check_options [6].widget = check_new (5, 6, XTRACT_LAYOUT(6));
    old_first_panel_size = -1;
    old_horizontal_split = -1;
    old_output_lines     = -1;

    _first_panel_size = first_panel_size;
    _output_lines = output_lines;
    add_widget (layout_dlg, la_check_options [6].widget);
    radio_widget = radio_new (3, 6, 2, s_split_direction, 1);
    add_widget (layout_dlg, radio_widget);
    radio_widget->sel = horizontal_split;
}

static void init_menu (void);
static void done_menu (void);
static void menubar_arrange(WMenu* menubar);


static void layout_change (void)
{
    setup_panels ();
    layout_do_change = 0;
    /* re-init the menu, because perhaps there was a change in the way
	how the panel are split (horizontal/vertical). */
    done_menu();
    init_menu();
    menubar_arrange(the_menubar);
}

static void layout_cmd (void)
{
    int result;
    int i;

    init_layout ();
    run_dlg (layout_dlg);
    result = layout_dlg->ret_value;

    if (result == B_ENTER || result == B_EXIT) {
	for (i = 0; la_check_options [i].text; i++)
	    if (la_check_options [i].widget)
		*la_check_options [i].variable = la_check_options [i].widget->state & C_BOOL;
	horizontal_split = radio_widget->sel;
	first_panel_size = _first_panel_size;
	output_lines = _output_lines;
	layout_do_change = 1;
    }
    if (result == B_EXIT) {
	save_layout ();
	sync_profiles ();
    }

    destroy_dlg (layout_dlg);
    if (layout_do_change)
	layout_change ();
}

static void check_split (void)
{
    if (horizontal_split) {
	if (equal_split)
	    first_panel_size = la_height / 2;
	else if (first_panel_size < MINHEIGHT)
	    first_panel_size = MINHEIGHT;
	else if (first_panel_size > la_height - MINHEIGHT)
	    first_panel_size = la_height - MINHEIGHT;
    } else {
	if (equal_split)
	    first_panel_size = COLS / 2;
	else if (first_panel_size < MINWIDTH)
	    first_panel_size = MINWIDTH;
	else if (first_panel_size > COLS - MINWIDTH)
	    first_panel_size = COLS - MINWIDTH;
    }
}

static int panel_event    (Gpm_Event *event, WPanel *panel);

static void init_curses (void)
{
    SLtt_get_terminfo ();
    if (force_ugly_line_drawing)
	SLtt_Has_Alt_Charset = 0;
    SLsmg_init_smg ();
    do_enter_ca_mode ();
    init_colors ();
    keypad (TRUE);
    nodelay (FALSE);
}

static void done_screen (void)
{
    if (!(MC_quit & SUBSHELL_EXIT))
	clr_scr ();
    reset_shell_mode ();
    shut_mouse ();
    keypad (FALSE);
}

static void panel_do_cols (int Index)
{
    if (get_display_type (Index) == view_listing)
	set_panel_formats ((WPanel *) panels [Index].widget);
    else {
	panel_update_cols (panels [Index].widget, frame_half);
    }
}

static void setup_panels (void)
{
    int start_y;
    int promptl;                /* the prompt len */

    if (console_flag) {
	int minimum;
	if (output_lines < 0)
	    output_lines = 0;
	la_height = LINES - keybar_visible - command_prompt - menubar_visible
		 - output_lines - message_visible;
	if (message_visible && xterm_hintbar && xterm_flag) la_height++;
	minimum = MINHEIGHT * (1 + horizontal_split);
	if (la_height < minimum) {
	    output_lines -= minimum - la_height;
	    la_height = minimum;
	}
    } else {
	la_height = LINES - menubar_visible - command_prompt -
	    keybar_visible - message_visible;
	if (message_visible && xterm_hintbar && xterm_flag) la_height++;
    }
    check_split ();
    start_y = menubar_visible;

    /* The column computing is defered until panel_do_cols */
    if (horizontal_split) {
	widget_set_size (panels [0].widget, start_y, 0,
			 first_panel_size, 0);

	widget_set_size (panels [1].widget, start_y+first_panel_size, 0,
			 la_height-first_panel_size, 0);
    } else {
	int first_x = first_panel_size;

	widget_set_size (panels [0].widget, start_y, 0,
			 la_height, 0);

	widget_set_size (panels [1].widget, start_y, first_x,
			 la_height, 0);

    }
    panel_do_cols (0);
    panel_do_cols (1);

    promptl = strlen (MC_prompt);

    widget_set_size (&the_menubar->widget, 0, 0, 1, COLS);

    if (command_prompt) {
	    widget_set_size (&cmdline->input.widget,
			     LINES-1-keybar_visible, promptl,
			     1, COLS-promptl-(keybar_visible ? 0 : 1));
	    winput_set_origin (&cmdline->input, promptl, COLS-promptl-(keybar_visible ? 0 : 1));
	    widget_set_size (&the_prompt->widget,
			     LINES-1-keybar_visible, 0,
			     1, promptl);
    } else {
	    widget_set_size (&cmdline->input.widget, 0, 0, 0, 0);
	    winput_set_origin (&cmdline->input, 0, 0);
	    widget_set_size (&the_prompt->widget, LINES, COLS, 0, 0);
    }

    widget_set_size (&the_bar->widget, LINES-1, 0, 1, COLS);
    the_bar->visible = keybar_visible;

    /* Output window */
    if (console_flag && output_lines) {
	output_start_y = LINES - command_prompt - keybar_visible - output_lines;
	show_console_contents (output_start_y,
			       LINES-output_lines-keybar_visible-1,
			       LINES-keybar_visible-1);
    }
    if (message_visible && (!xterm_hintbar || !xterm_flag))
	widget_set_size (&the_hint->widget, la_height+start_y, 0, 1,COLS);
    else
	widget_set_size (&the_hint->widget, 0, 0, 0, 0);

    load_hint ();
}


static void low_level_change_screen_size (void)
{
    get_terminal_width_height(&COLS, &LINES);
}

static void change_screen_size (void)
{
#if defined TIOCGWINSZ && !defined SCO_FLAVOR

    endwin ();
    low_level_change_screen_size ();
    check_split ();
    init_curses ();
    setup_panels ();
    if (current_dlg == view_dlg)
	view_adjust_size (view_dlg);

#ifdef RESIZABLE_MENUBAR
	menubar_arrange(the_menubar);
#endif

    /* Now, force the redraw */
    do_refresh ();
    touchwin ();
#endif /* TIOCGWINSZ && !SCO_FLAVOR */
    winch_flag = 0;
}

static void set_hintbar(char *str)
{
    if (xterm_flag && xterm_hintbar) {
	fprintf (stderr, "\33]0;mc - %s\7", str);
    } else {
	label_set_text (the_hint, str);
	if (ok_to_refresh > 0)
	    mc_refresh();
    }
}

#ifdef USE_NETCODE
static void print_vfs_message(char *msg, ...)
{
    va_list ap;
    char str[128];

    va_start(ap, msg);
    vsprintf(str, msg, ap);
    va_end(ap);
    if (midnight_shutdown || !the_hint || !the_hint->widget.parent)
	return;

    if (message_visible || (xterm_flag && xterm_hintbar)) {
	set_hintbar(str);
    }
}
#endif

static void rotate_dash (void)
{
    if (!nice_rotating_dash || (ok_to_refresh <= 0))
	return;

    if (rotate_dash_pos >= sizeof (rotating_dash)-1)
	rotate_dash_pos = 0;
    move (0, COLS-1);
    addch (rotating_dash [rotate_dash_pos]);
    mc_refresh ();
    rotate_dash_pos++;
}

static char *get_nth_panel_name (int num, char buffer[20])
{
    if (!num)
	return "New Left Panel";
    else if (num == 1)
	return "New Right Panel";
    else {
	sprintf (buffer, "%ith Panel", num);
	return buffer;
    }
}

/* I wonder if I should start to use the folding mode than Dugan uses */
/*                                                                     */
/* This is the centralized managing of the panel display types         */
/* This routine takes care of destroying and creating new widgets      */
/* Please note that it could manage MAX_VIEWS, not just left and right */
/* Currently nothing in the code takes advantage of this and has hard- */
/* coded values for two panels only                                    */

/* Set the num-th panel to the view type: type */
/* This routine also keeps at least one WPanel object in the screen */
/* since a lot of routines depend on the current_panel variable */
void set_display_type (int num, int type)
{
    int x, y, cols, lines;
    int    the_other;           /* Index to the other panel */
    char   *file_name = 0;      /* For Quick view */
    Widget *new_widget, *old_widget;
    WPanel  *the_other_panel;

    x =y = cols = lines = 0;
    old_widget = 0;
    if (num >= MAX_VIEWS) {
	fprintf (stderr, "Could not allocate more that %d views\n", MAX_VIEWS);
	abort ();
    }

    /* Check that we will have a WPanel * at least */
    the_other = 0;
    if (type != view_listing) {
	the_other = num == 0 ? 1 : 0;

	if (panels [the_other].type != view_listing)
	    return;

    }

    /* Get rid of it */
    if (panels [num].widget) {
	Widget *w = panels [num].widget;
	WPanel *panel = (WPanel *) panels [num].widget;

	x = w->x;
	y = w->y;
	cols  = w->cols;
	lines = w->lines;
	old_widget = panels [num].widget;

	if (panels [num].type == view_listing) {
	    if (panel->frame_size == frame_full && type != view_listing) {
		cols = COLS - first_panel_size;
		if (num == 1)
		    x = first_panel_size;
	    }
	}
    }

    new_widget = 0;

    switch (type) {
    case view_listing:
	{
		char buffer[20];
		new_widget = (Widget *) panel_new (get_nth_panel_name (num, buffer));
	}
	break;

    case view_info:
	new_widget = (Widget *) info_new ();

	break;

    case view_tree:
	new_widget = (Widget *) tree_new (1, 0, 0, 0, 0);
	break;

    case view_quick:
	new_widget = (Widget *) view_new (0, 0, 0, 0, 1);
	the_other_panel = (WPanel *) panels [the_other].widget;
	if (the_other_panel)
	    file_name =
		the_other_panel->dir.list[the_other_panel->selected].fname;
	else
	    file_name = "";

	view_init ((WView *) new_widget, 0, file_name, 0);
	break;
    }
    panels [num].type = type;
    panels [num].widget = (Widget *) new_widget;

    /* We set the same size the old widget had */
    widget_set_size ((Widget *) new_widget, y, x, lines, cols);

    /* We wanna the new widget at the same position */
    /* XView sets wcontainer to !0 <- Not XView, but we, when we create it */
    /* Ok, the XView support code does it */
    if (old_widget && old_widget->wcontainer) {
	new_widget->wcontainer = old_widget->wcontainer;
	new_widget->area = old_widget->area;
    }

    /* We use replace to keep the circular list of the dialog in the */
    /* same state.  Maybe we could just kill it and then replace it  */
    if (midnight_dlg && old_widget) {
	dlg_replace_widget (midnight_dlg, old_widget, panels [num].widget);
    }
    if (type == view_listing) {
	if (num == 0)
	    left_panel = (WPanel *) new_widget;
	else
	    right_panel = (WPanel *) new_widget;
    }

    if (type == view_tree)
	the_tree = (WTree *) new_widget;

    /* Prevent current_panel's value from becoming invalid.
     * It's just a quick hack to prevent segfaults. Comment out and
     * try following:
     * - select left panel
     * - invoke menue left/tree
     * - as long as you stay in the left panel almost everything that uses
     *   cpanel causes segfault, e.g. C-Enter, C-x c, ...
     */

    if (type != view_listing)
	if (cpanel == (WPanel *) old_widget)
	    current_panel = num == 0 ? right_panel : left_panel;
}

/* This routine is deeply sticked to the two panels idea.
   What should it do in more panels. ANSWER - don't use it
   in any multiple panels environment. */
static void swap_panels (void)
{
    Widget tmp;
    Widget *tmp_widget;
    WPanel panel;
    WPanel *panel1, *panel2;
    int tmp_type;

#define panelswap(x) panel.x = panel1->x; panel1->x = panel2->x; panel2->x = panel.x;

#define panelswapstr(e) strcpy (panel.e, panel1->e); \
			strcpy (panel1->e, panel2->e); \
			strcpy (panel2->e, panel.e);
    panel1 = (WPanel *) panels [0].widget;
    panel2 = (WPanel *) panels [1].widget;
    if (panels [0].type == view_listing && panels [1].type == view_listing) {
	/* Change everything except format/sort/panel_name etc. */
	panelswap (dir);
	panelswap (active);
	panelswapstr (cwd);
	panelswapstr (lwd);
	panelswap (count);
	panelswap (marked);
	panelswap (dirs_marked);
	panelswap (total);
	panelswap (top_file);
	panelswap (selected);
	panelswap (is_panelized);
	panelswap (dir_stat);

	panel1->searching = 0;
	panel2->searching = 0;
	if (cpanel == panel1)
	    current_panel = panel2;
	else
	    current_panel = panel1;
	if (midnight_dlg->current->widget == panels [0].widget)
	    dlg_select_widget (midnight_dlg, (void *) panels [1].widget);
	else if (midnight_dlg->current->widget == panels [1].widget)
	    dlg_select_widget (midnight_dlg, (void *) panels [0].widget);
    } else {
	WPanel *tmp_panel;
	char buffer[20];

	tmp_panel = right_panel;
	right_panel = left_panel;
	left_panel = tmp_panel;

	if (panels [0].type == view_listing) {
	    if (!strcmp (panel1->panel_name, get_nth_panel_name (0, buffer))) {
		free (panel1->panel_name);
		panel1->panel_name = strdup (get_nth_panel_name (1, buffer));
	    }
	}
	if (panels [1].type == view_listing) {
	    if (!strcmp (panel2->panel_name, get_nth_panel_name (1, buffer))) {
		free (panel2->panel_name);
		panel2->panel_name = strdup (get_nth_panel_name (0, buffer));
	    }
	}

	tmp.x = panels [0].widget->x;
	tmp.y = panels [0].widget->y;
	tmp.cols = panels [0].widget->cols;
	tmp.lines = panels [0].widget->lines;

	panels [0].widget->x = panels [1].widget->x;
	panels [0].widget->y = panels [1].widget->y;
	panels [0].widget->cols = panels [1].widget->cols;
	panels [0].widget->lines = panels [1].widget->lines;

	panels [1].widget->x = tmp.x;
	panels [1].widget->y = tmp.y;
	panels [1].widget->cols = tmp.cols;
	panels [1].widget->lines = tmp.lines;

	tmp_widget = panels [0].widget;
	panels [0].widget = panels [1].widget;
	panels [1].widget = tmp_widget;
	tmp_type = panels [0].type;
	panels [0].type = panels [1].type;
	panels [1].type = tmp_type;
    }
}

static int get_display_type (int Index)
{
    return panels [Index].type;
}

static Widget *get_panel_widget (int Index)
{
    return panels [Index].widget;
}

int get_current_index (void)
{
    if (panels [0].widget == ((Widget *) cpanel))
	return 0;
    else
	return 1;
}

int get_other_index (void)
{
    return !get_current_index ();
}

/* Returns the view type for the current panel/view */
int get_current_type (void)
{
    if (panels [0].widget == (Widget *) cpanel)
	return panels [0].type;
    else
	return panels [1].type;
}

/* Returns the view type of the unselected panel */
int get_other_type (void)
{
    if (panels [0].widget == (Widget *) cpanel)
	return panels [1].type;
    else
	return panels [0].type;
}

/* Learn keys */

#define UX_LEARN              4
#define UY_LEARN              3

#define BY_LEARN              UY_LEARN + 17

#define ROWS            13
#define COLSHIFT        23

#define BUTTONS_LEARN         2

static const struct {
    int ret_cmd, flags, y, x;
    unsigned int hotkey;
    const char *text;
} learn_but[BUTTONS_LEARN] = {
    { B_CANCEL, NORMAL_BUTTON, 0, 39, 'C', "&Cancel" },
    { B_ENTER, DEFPUSH_BUTTON, 0, 25, 'S', "&Save" }
};

static const char learn_title[] = " Learn keys ";

static void learn_refresh (void)
{
    attrset (COLOR_NORMAL);
    dlg_erase (learn_dlg);

    draw_box (learn_dlg, 1, 2, learn_dlg->lines - 2, learn_dlg->cols - 4);

    attrset (COLOR_HOT_NORMAL);
    dlg_move (learn_dlg, 1, (learn_dlg->cols - strlen (learn_title)) / 2);
    addstr (learn_title);
}

static int learn_button (int action, void *param)
{
    char *seq;
    Dlg_head *d = MC_message (D_INSERT | 1, " Teach me a key ",
"Please press the %s\n"
"and then wait until this message disappears.\n\n"
"Then, press it again to see if OK appears\n"
"next to its button.\n\n"
"If you want to escape, press a single Escape key\n"
"and wait as well.",
	key_name_conv_tab [action - B_USER].longname);
    mc_refresh ();
    if (learnkeys [action - B_USER].sequence != NULL) {
	free (learnkeys [action - B_USER].sequence);
	learnkeys [action - B_USER].sequence = NULL;
    }
    seq = learn_key ();

    if (seq) {
	/* Esc hides the dialog and do not allow definitions of
	 * regular characters
	 */
	if (*seq && strcmp (seq, "\\e") && strcmp (seq, "\\e\\e")
	    && strcmp (seq, "^m" )
	    && (seq [1]
	    || (((unsigned char)*seq) < ' ' || ((unsigned char)*seq) > '~'))) {

	    learnchanged = 1;
	    learnkeys [action - B_USER].sequence = seq;
	    seq = convert_controls (seq);
	    define_sequence (key_name_conv_tab [action - B_USER].code, seq,
			     MCKEY_NOACTION);
	} else {
	    MC_message (0, " Cannot accept this key ",
		" You have entered \"%s\"", seq);
	}

	free (seq);
    }

    dlg_run_done (d);
    destroy_dlg (d);
    dlg_select_widget (learn_dlg, learnkeys [action - B_USER].button);
    return 0; /* Do not kill learn_dlg */
}

static int learn_move (int right)
{
    int i, totalcols;

    totalcols = (learn_total - 1) / ROWS + 1;
    for (i = 0; i < learn_total; i++)
	if (learnkeys [i].button == learn_dlg->current->widget) {
	    if (right) {
		if (i < learn_total - ROWS)
		    i += ROWS;
		else
		    i %= ROWS;
	    } else {
		if (i / ROWS)
		    i -= ROWS;
		else if (i + (totalcols - 1) * ROWS >= learn_total)
		    i += (totalcols - 2) * ROWS;
		else
		    i += (totalcols - 1) * ROWS;
	    }
	    dlg_select_widget (learn_dlg, (void *) learnkeys [i].button);
	    return 1;
	}
    return 0;
}

static int learn_check_key (int c)
{
    int i;

    for (i = 0; i < learn_total; i++) {
	if (key_name_conv_tab [i].code == c) {
	    if (!learnkeys [i].ok) {
		dlg_select_widget (learn_dlg, learnkeys [i].button);
		label_set_text ((WLabel *) learnkeys [i].label,
		    "OK");
		learnkeys [i].ok = 1;
		learnok++;
		if (learnok >= learn_total) {
		    learn_dlg->ret_value = B_CANCEL;
		    if (learnchanged) {
			if (query_dialog (learn_title,
			    "It seems that all your keys already\n"
			      "work fine. That's great.",
			    1, 2, "&Save", "&Discard") == 0)
			    learn_dlg->ret_value = B_ENTER;
		    } else {
			MC_message (1, learn_title,
			"Great! You have a complete terminal database!\n"
			  "All your keys work well.");
		    }
		    dlg_stop (learn_dlg);
		}
		return 1;
	    }
	}
    }
    switch (c) {
	case KEY_LEFT:
	case 'h':
	    return learn_move (0);
	case KEY_RIGHT:
	case 'l':
	    return learn_move (1);
	case 'j':
	    dlg_one_down (learn_dlg);
	    return 1;
	case 'k':
	    dlg_one_up (learn_dlg);
	    return 1;
    }

    /* Prevent from disappearing if a non-defined sequence is pressed
       and contains s or c. Use ALT('s') or ALT('c'). */
	if (c < 255 && isalpha(c))
	{
		c = toupper(c);
		for (i = 0; i < BUTTONS_LEARN; i++)
			if (c == learn_but [i].hotkey)
				return 1;
	}

    return 0;
}

static int learn_callback (Dlg_head * h, int Par, int Msg)
{
    switch (Msg) {
    case DLG_DRAW:
	learn_refresh ();
	break;
    case DLG_KEY:
	return learn_check_key (Par);
    }
    return 0;
}

static void init_learn (void)
{
    int x, y, i, j;
    const key_code_name_t *key;
    char buffer [22];

    do_refresh ();

    learn_dlg = create_dlg (0, 0, 23, 78, dialog_colors,
			      learn_callback, "[Learn keys]",
			      DLG_CENTER);
    x_set_dialog_title (learn_dlg, "Learn keys");

#define XTRACT_LEARN(i) BY_LEARN+learn_but[i].y, learn_but[i].x, learn_but[i].ret_cmd, learn_but[i].flags, learn_but[i].text, 0, 0

    for (i = 0; i < BUTTONS_LEARN; i++)
	add_widget (learn_dlg, button_new (XTRACT_LEARN (i)));

    x = UX_LEARN;
    y = UY_LEARN;
    for (key = key_name_conv_tab, j = 0; key->name != NULL &&
	strcmp (key->name, "kpleft"); key++, j++);
    learnkeys = (learnkey *) x_malloc (sizeof (learnkey) * j, "Learn keys");
    x += ((j - 1) / ROWS) * COLSHIFT;
    y += (j - 1) % ROWS;
    learn_total = j;
    learnok = 0;
    learnchanged = 0;
    for (i = j - 1, key = key_name_conv_tab + j - 1; i >= 0; i--, key--) {
	learnkeys [i].ok = 0;
	learnkeys [i].sequence = NULL;
	sprintf (buffer, "%-16s", key->longname);
	add_widget (learn_dlg, learnkeys [i].button = (Widget *)
	    button_new (y, x, B_USER + i, NARROW_BUTTON, buffer, learn_button, 0));
	add_widget (learn_dlg, learnkeys [i].label = (Widget *)label_new (y, x + 19, ""));
	if (i % 13)
	    y--;
	else {
	    x -= COLSHIFT;
	    y = UY_LEARN + ROWS - 1;
	}
    }
    add_widget (learn_dlg,
		label_new (UY_LEARN+14, 5, "Press all the keys mentioned here. After you have done it, check"));
    add_widget (learn_dlg,
		label_new (UY_LEARN+15, 5, "which keys are not marked with OK.  Press space on the missing"));
    add_widget (learn_dlg,
		label_new (UY_LEARN+16, 5, "key, or click with the mouse to define it. Move around with Tab."));
}

static void learn_done (void)
{
    destroy_dlg (learn_dlg);
    repaint_screen ();
}

static void learn_save (void)
{
    int i;
    int profile_changed = 0;
    char *section = xasprintf ("terminal:%s", TERM_env);

    for (i = 0; i < learn_total; i++) {
	if (learnkeys [i].sequence != NULL) {
	    profile_changed = 1;
	    WritePrivateProfileString (section, key_name_conv_tab [i].name,
		learnkeys [i].sequence, profile_name);
	}
    }

    /* On the one hand no good idea to save the complete setup but
     * without 'Auto save setup' the new key-definitions will not be
     * saved unless the user does an 'Options/Save Setup'.
     * On the other hand a save-button that does not save anything to
     * disk is much worse.
     */
    if (profile_changed)
	sync_profiles ();
}

static void learn_keys (void)
{
    int save_old_esc_mode = old_esc_mode;
    int save_alternate_plus_minus = alternate_plus_minus;

    old_esc_mode = 0; /* old_esc_mode cannot work in learn keys dialog */
    alternate_plus_minus = 1; /* don't translate KP_ADD, KP_SUBTRACT and
				 KP_MULTIPLY to '+', '-' and '*' in
				 correct_key_code */
    application_keypad_mode ();
    init_learn ();

    run_dlg (learn_dlg);

    old_esc_mode = save_old_esc_mode;
    alternate_plus_minus = save_alternate_plus_minus;

    if (!alternate_plus_minus)
	numeric_keypad_mode ();

    switch (learn_dlg->ret_value) {
    case B_ENTER:
	learn_save ();
	break;
    }

    learn_done ();
}


/* Main program for the Midnight Commander */

static void configure_box (void);

/* When the modes are active, left_panel, right_panel and tree_panel */
/* Point to a proper data structure.  You should check with the functions */
/* get_current_type and get_other_type the types of the panels before using */
/* This pointer variables */


/* Forward declarations */
static void menu_cmd (void);

static WPanel *
get_other_panel (void)
{
	return (WPanel *) get_panel_widget (get_other_index ());
}

static void
try_to_select (WPanel *panel, char *name)
{
    Xtry_to_select (panel, name);
    select_item (panel);
    display_mini_info (panel);
}

#ifdef USE_VFS
#define extfs_get_prefix(idx) extfs_prefixes[idx]
#endif

/*
 * cd_try_to_select:
 *
 *  If we moved to the parent directory move the selection pointer to
 *  the old directory name
 */
static void
cd_try_to_select (WPanel *panel)
{
#ifdef USE_VFS
    char *p, *q;
    int i, j = 4;
#endif

    if (strlen (panel->lwd) > strlen (panel->cwd)
	&& strncmp (panel->cwd, panel->lwd, strlen (panel->cwd)) == 0
	&& strchr (panel->lwd + strlen (panel->cwd) + 1, '/') == 0)
	try_to_select (panel, panel->lwd);
    else
#ifdef USE_VFS
	if ((!strncmp (panel->lwd, "tar:", 4) &&
	     !strncmp (panel->lwd + 4, panel->cwd, strlen (panel->cwd))) ||
	     ((i = extfs_prefix_to_type (panel->lwd)) != -1 &&
	     !strncmp (panel->lwd + (j = strlen (extfs_get_prefix (i)) + 1),
	     panel->cwd, strlen (panel->cwd)))) {
	p = strdup (panel->lwd + j + strlen (panel->cwd));
	q = strchr (p, '/');
	if (q != NULL && (q != p || (q = strchr (q + 1, '/')) != NULL))
	    *q = 0;
	try_to_select (panel, p);
	free (p);
    } else
#endif
	try_to_select (panel, NULL);
}

static void
reload_panelized (WPanel *panel)
{
    int i, j;
    dir_list *list = &panel->dir;

    if (panel != cpanel)
	mc_chdir (panel->cwd);

    for (i = 0, j = 0; i < panel->count; i++) {
	if (list->list [i].f.marked) {
	    /* Unmark the file in advance. In case the following mc_lstat
	     * fails we are done, else we have to mark the file again
	     * (Note: do_file_mark depends on a valid "list->list [i].buf").
	     * IMO that's the best way to update the panel's summary status
	     * -- Norbert
	     */
	    do_file_mark (panel, i, 0);
	}
	if (mc_lstat (list->list [i].fname, &list->list [i].buf)) {
	    free (list->list [i].fname);
	    continue;
	}
	if (list->list [i].f.marked)
	    do_file_mark (panel, i, 1);
	if (j != i)
	    list->list [j] = list->list [i];
	j++;
    }
    if (j == 0)
	panel->count = set_zero_dir (list);
    else
	panel->count = j;

    if (panel != cpanel)
	mc_chdir (cpanel->cwd);
}

void
update_one_panel_widget (WPanel *panel, int force_update, char *current_file_u)
{
    int free_pointer;

    if (force_update & UP_RELOAD) {
	panel->is_panelized = 0;

	ftpfs_flushdir ();
	memset (&(panel->dir_stat), 0, sizeof (panel->dir_stat));
    }

    /* If current_file == -1 (an invalid pointer) then preserve selection */
    if (current_file_u == UP_KEEPSEL) {
	free_pointer = 1;
	current_file_u = strdup (panel->dir.list [panel->selected].fname);
    } else
	free_pointer = 0;

    if (panel->is_panelized)
	reload_panelized (panel);
    else
	panel_reload (panel);

    try_to_select (panel, current_file_u);
    panel->dirty = 1;

    if (free_pointer)
	free (current_file_u);
}

#ifndef PORT_HAS_UPDATE_PANELS
static void
update_one_panel (int which, int force_update, char *current_file_u)
{
    WPanel *panel;

    if (get_display_type (which) != view_listing)
	return;

    panel = (WPanel *) get_panel_widget (which);
    update_one_panel_widget (panel, force_update, current_file_u);
}

/* This routine reloads the directory in both panels. It tries to
 * select current_file in current_panel and other_file in other_panel.
 * If current_file == -1 then it automatically sets current_file and
 * other_file to the currently selected files in the panels.
 *
 * if force_update has the UP_ONLY_CURRENT bit toggled on, then it
 * will not reload the other panel.
*/
void
update_panels (int force_update, char *current_file_u)
{
    int reload_other = !(force_update & UP_ONLY_CURRENT);
    WPanel *panel;

    update_one_panel (get_current_index (), force_update, current_file_u);
    if (reload_other)
	update_one_panel (get_other_index (), force_update, UP_KEEPSEL);

    if (get_current_type () == view_listing)
	panel = (WPanel *) get_panel_widget (get_current_index ());
    else
	panel = (WPanel *) get_panel_widget (get_other_index ());

    mc_chdir (panel->cwd);
}
#endif

#ifdef WANT_PARSE
static void select_by_index (WPanel *panel, int i);

/* Called by parse_control_file */
static int index_by_name (file_entry *list, int count)
{
    char *name;
    int i;

    name = strtok (NULL, " \t\n");
    if (!name || !*name)
	return -1;
    for (i = 0; i < count; i++) {
	if (strcmp (name, list[i].fname) == 0)
	    return i;
    }
    return -1;
}

/* Called by parse_control_file */
static void select_by_index (WPanel *panel, int i)
{
    if (i >= panel->count)
	return;

    unselect_item (panel);
    panel->selected = i;

    while (panel->selected - panel->top_file >= ITEMS (panel)) {
	/* Scroll window half screen */
	panel->top_file += ITEMS (panel)/2;
	paint_dir (panel);
	select_item (panel);
    }
    while (panel->selected < panel->top_file) {
	/* Scroll window half screen */
	panel->top_file -= ITEMS (panel)/2;
	if (panel->top_file < 0) panel->top_file = 0;
	paint_dir (panel);
    }
    select_item (panel);
}

/* Called by my_system
   No error reporting, just exits on the first sign of trouble */
static void parse_control_file (void)
{
    char *data, *current;
    WPanel *panel;
    file_entry *list;
    int i;
    FILE *file;
    struct stat s;

    if ((file = fopen (control_file, "r")) == NULL) {
	return;
    }
    /* Use of fstat prevents race conditions */
    if (fstat (fileno (file), &s) != 0) {
	fclose (file);
	return;
    }
    /* Security: Check that the user owns the control file to prevent
       other users from playing tricks on him/her. */
    if (s.st_uid != my_uid) {
	fclose (file);
	return;
    }
    data = (char *) x_malloc (s.st_size+1, "main, parse_control_file");
    if (s.st_size != fread (data, 1, s.st_size, file)) {
	free (data);
	fclose (file);
	return;
    }
    data [s.st_size] = 0;
    fclose (file);

    /* The Control file has now been loaded to memory -> start parsing. */
    current = strtok (data, " \t\n");
    while (current && *current) {
	if (isupper (*current)) {
	    if (get_other_type () != view_listing)
		break;
	    else
		panel = other_panel;
	} else
	    panel = cpanel;

	list = panel->dir.list;
	*current = tolower (*current);

	if (strcmp (current, "clear_tags") == 0) {
	    unmark_files (panel);
	} else if (strcmp (current, "tag") == 0) {
	    i = index_by_name (list, panel->count);
	    if (i >= 0) {
		do_file_mark (panel, i, 1);
	    }
	} else if (strcmp (current, "untag") == 0) {
	    i = index_by_name (list, panel->count);
	    if (i >= 0) {
		do_file_mark (panel, i, 0);
	    }
	} else if (strcmp (current, "select") == 0) {
	    i = index_by_name (list, panel->count);
	    if (i >= 0) {
		select_by_index (panel, i);
	    }
	} else if (strcmp (current, "change_panel") == 0) {
	    change_panel ();
	} else if (strcmp (current, "cd") == 0) {
	    int change = 0;
	    current = strtok (NULL, " \t\n");
	    if (!current) break;
	    if (cpanel != panel) {
		change_panel ();
		change = 1;
	    }
	    do_cd (current, cd_parse_command);
	    if (change)
		change_panel ();
	} else {
	    /* Unknown command -> let's give up */
	    break;
	}
	current = strtok (NULL, " \t\n");
    }

    free (data);
    paint_panel (cpanel);
    paint_panel (opanel);
}
#else
#define parse_control_file()
#endif /* WANT_PARSE */

/* Sets up the terminal before executing a program */
static void
pre_exec (void)
{
    ok_to_refresh--;
    if (clear_before_exec)
	clr_scr ();
    else {
	if (!(console_flag || xterm_flag))
	    printf ("\n\n");
    }

    shut_mouse ();

    reset_shell_mode ();
    keypad (FALSE);
    endwin ();

    numeric_keypad_mode ();

    /* on xterms: maybe endwin did not leave the terminal on the shell
     * screen page: do it now.
     *
     * Do not move this before endwin: in some systems rmcup includes
     * a call to clear screen, so it will end up clearing the sheel screen.
     */
    do_exit_ca_mode ();
}

/* Save current stat of directories to avoid reloading the panels */
/* when no modifications have taken place */
void
save_cwds_stat (void)
{
    if (fast_reload) {
	mc_stat (cpanel->cwd, &(cpanel->dir_stat));
	if (get_other_type () == view_listing)
	    mc_stat (opanel->cwd, &(opanel->dir_stat));
    }
}

void
restore_console (void)
{
    handle_console (CONSOLE_RESTORE);
}

void
do_execute (const char *shell, const char *command, int flags)
{
#ifdef USE_VFS
    char *old_vfs_dir = 0;

    if (!vfs_current_is_local ())
	old_vfs_dir = strdup (vfs_get_current_dir ());
#endif

    save_cwds_stat ();
    pre_exec ();
    if (console_flag)
	restore_console ();

#ifndef __os2__
    unlink (control_file);
#endif
    if (!use_subshell && !(flags & EXECUTE_INTERNAL)) {
	printf ("%s%s%s\n", last_paused ? "\r\n":"", MC_prompt, command);
	last_paused = 0;
    }

	my_system (flags, shell, command);

    if (!(flags & EXECUTE_INTERNAL)) {
	if ((pause_after_run == pause_always ||
	    (pause_after_run == pause_on_dumb_terminals &&
	     !xterm_flag && !console_flag)) && !MC_quit) {
	    printf ("Press any key to continue...");
	    last_paused = 1;
	    fflush (stdout);
	    raw ();
	    xgetch ();
	}
	if (console_flag) {
	    if (output_lines && keybar_visible) {
		putchar('\n');
		fflush(stdout);
	    }
	}
    }

    if (console_flag)
	handle_console (CONSOLE_SAVE);
    edition_post_exec ();

#ifdef USE_VFS
	if (old_vfs_dir) {
	    mc_chdir (old_vfs_dir);
	    free (old_vfs_dir);
	}
#endif

    update_panels (UP_OPTIMIZE, UP_KEEPSEL);

    parse_control_file ();
#ifndef __os2__
    unlink (control_file);
#endif
    do_refresh ();
    ok_to_refresh++;
}

/* Executes a command */
static void
shell_execute (char *command, int flags)
{
	do_execute (MC_shell, command, flags | EXECUTE_AS_SHELL);
}

static void
execute (char *command)
{
    shell_execute (command, 0);
}

static void
change_panel (void)
{
    free_completions (input_w (cmdline));
    dlg_one_down (midnight_dlg);
}

static int
quit_cmd_internal (int quiet)
{
    int q = MC_quit;

    if (quiet || !confirm_exit) {
	q = 1;
    } else  {
	if (query_dialog (" The Midnight Commander ",
		     " Do you really want to quit the Midnight Commander? ",
			  0, 2, "&Yes", "&No") == 0)
	    q = 1;
    }
    if (q) {
		midnight_dlg->running = 0;
    }
    if (q)
	MC_quit |= 1;
    return MC_quit;
}

static void quit_cmd (void)
{
    quit_cmd_internal (0);
}

static void quiet_quit_cmd (void)
{
    print_last_revert = 1;
    quit_cmd_internal (1);
}

/*
 * Touch window and refresh window functions
 */

static void
repaint_screen (void)
{
    do_refresh ();
    mc_refresh ();
}

static void
directory_history_add (WPanel * panel, char *s)
{
    if (!panel->dir_history) {
	panel->dir_history = zmalloc (sizeof (Hist));
	panel->dir_history->text = strdup (s);
	return;
    }
    if (!strcmp (panel->dir_history->text, s))
	return;
    if (panel->dir_history->next) {
	if (panel->dir_history->next->text) {
	    free (panel->dir_history->next->text);
	    panel->dir_history->next->text = 0;
	}
    } else {
	panel->dir_history->next = zmalloc (sizeof (Hist));
	panel->dir_history->next->prev = panel->dir_history;
    }
    panel->dir_history = panel->dir_history->next;
    panel->dir_history->text = strdup (s);
}

/* Changes the current panel directory */
static int
_do_panel_cd (WPanel *panel, char *new_dir, enum cd_enum cd_type)
{
    char *directory, *olddir;
    char temp [MC_MAXPATHLEN];
#ifdef USE_VFS
    vfs *oldvfs;
    vfsid oldvfsid;
    struct vfs_stamping *parent;
#endif
    olddir = strdup (panel->cwd);

    /* Convert *new_path to a suitable pathname, handle ~user */

    if (cd_type == cd_parse_command) {
	while (*new_dir == ' ')
	    new_dir++;

	if (!strcmp (new_dir, "-")) {
	    strcpy (temp, panel->lwd);
	    new_dir = temp;
	}
    }
    directory = *new_dir ? new_dir : home_dir;

    if (mc_chdir (directory) == -1) {
	strcpy (panel->cwd, olddir);
	free (olddir);
	return 0;
    }

    /* Success: save previous directory, shutdown status of previous dir */
    strcpy (panel->lwd, olddir);
    free_completions (input_w (cmdline));

    mc_get_current_wd (panel->cwd, sizeof (panel->cwd) - 2);

#ifdef USE_VFS
    oldvfs = vfs_type (olddir);
    oldvfsid = vfs_ncs_getid (oldvfs, olddir, &parent);
    vfs_add_noncurrent_stamps (oldvfs, oldvfsid, parent);
    vfs_rm_parents (parent);
#endif
    free (olddir);

    /* Reload current panel */
    clean_dir (&panel->dir, panel->count);
    panel->count = do_load_dir (&panel->dir, panel->sort_type,
				 panel->reverse, panel->case_sensitive, panel->filter);
    panel->top_file = 0;
    panel->selected = 0;
    panel->marked = 0;
    panel->dirs_marked = 0;
    panel->total = 0;
    panel->searching = 0;
    cd_try_to_select (panel);
    load_hint ();
    panel_update_contents (panel);
    return 1;
}

static int
do_panel_cd (WPanel *panel, char *new_dir, enum cd_enum cd_type)
{
    int r;

    r = _do_panel_cd (panel, new_dir, cd_type);
    if (r)
	directory_history_add (cpanel, cpanel->cwd);
    return r;
}

static int
do_cd (char *new_dir, enum cd_enum exact)
{
    return (do_panel_cd (cpanel, new_dir, exact));
}

static void
directory_history_next (WPanel * panel)
{
    if (!panel->dir_history->next)
	return;
    if (_do_panel_cd (panel, panel->dir_history->next->text, cd_exact))
	panel->dir_history = panel->dir_history->next;
}

static void
directory_history_prev (WPanel * panel)
{
    if (!panel->dir_history->prev)
	return;
    if (_do_panel_cd (panel, panel->dir_history->prev->text, cd_exact))
	panel->dir_history = panel->dir_history->prev;
}

static void
directory_history_list (WPanel * panel)
{
    char *s;
/* must be at least two to show a history */
    if (panel->dir_history) {
	if (panel->dir_history->prev || panel->dir_history->next) {
	    s = show_hist (panel->dir_history, panel->widget.x, panel->widget.y);
	    if (s) {
		int r;
		r = _do_panel_cd (panel, s, cd_exact);
		if (r)
		    directory_history_add (panel, panel->cwd);
		free (s);
	    }
	}
    }
}

/* Used to emulate Lynx's entering leaving a directory with the arrow keys */
static int
maybe_cd (int char_code, int move_up_dir)
{
    if (navigate_with_arrows) {
	if (!input_w (cmdline)->buffer [0]) {
	    if (!move_up_dir) {
		do_cd ("..", cd_exact);
		return 1;
	    }
	    if (S_ISDIR (selection (cpanel)->buf.st_mode)
		|| link_isdir (selection (cpanel))) {
		do_cd (selection (cpanel)->fname, cd_exact);
		return 1;
	    }
	}
    }
    return 0;
}

static void
sort_cmd (void)
{
    WPanel  *p;
    sortfn *sort_order;

    if (!SELECTED_IS_PANEL)
	return;

    p = MENU_PANEL;
    sort_order = sort_box (p->sort_type, &p->reverse, &p->case_sensitive);

    if (sort_order == 0)
	return;

    p->sort_type = sort_order;

    /* The directory is already sorted, we have to load the unsorted stuff */
    if (sort_order == unsorted) {
	char *current_file_u;

	current_file_u = strdup (cpanel->dir.list [cpanel->selected].fname);
	panel_reload (cpanel);
	try_to_select (cpanel, current_file_u);
	free (current_file_u);
    }
    do_re_sort (p);
}

static void
tree_box (void)
{
    char *sel_dir;

    sel_dir = Tree (selection (cpanel)->fname);
    if (sel_dir) {
	do_cd(sel_dir, cd_exact);
	free (sel_dir);
    }
}

/* NOTICE: hotkeys specified here are overriden in menubar_paint_idx (alex) */
static const menu_entry PanelMenu [] = {
    { ' ', "&Listing mode...",          'l', listing_cmd },
    { ' ', "&Quick view     C-x q",     'q', quick_view_cmd },
    { ' ', "&Info           C-x i",     'i', info_cmd },
    { ' ', "&Tree",                     't', tree_cmd },
    { ' ', "", ' ', 0 },
    { ' ', "&Sort order...",            's', sort_cmd },
    { ' ', "", ' ', 0 },
    { ' ', "&Filter...",                'f', filter_cmd },
#ifdef USE_NETCODE
    { ' ', "", ' ', 0 },
    { ' ', "&Network link...",          'n', netlink_cmd },
    { ' ', "FT&P link...",              'p', ftplink_cmd },
#endif
    { ' ', "", ' ', 0 },
    { ' ', "&Rescan         C-r",       'r', reread_cmd }
};

static const menu_entry RightMenu [] = {
    { ' ', "&Listing mode...",       'l', listing_cmd },
    { ' ', "&Quick view     C-x q",  'q', quick_view_cmd },
    { ' ', "&Info           C-x i",  'i', info_cmd },
    { ' ', "&Tree",                  't', tree_cmd },
    { ' ', "", ' ', 0 },
    { ' ', "&Sort order...",         's', sort_cmd },
    { ' ', "", ' ', 0 },
    { ' ', "&Filter...",             'f', filter_cmd },
#ifdef USE_NETCODE
    { ' ', "", ' ', 0 },
    { ' ', "&Network link...",       'n', netlink_cmd },
    { ' ', "FT&P link...",           'p', ftplink_cmd },
#endif
    { ' ', "", ' ', 0 },
    { ' ', "&Rescan         C-r",    'r', reread_cmd }
};

static const menu_entry FileMenu [] = {
    { ' ', "&User menu          F2", 'u', user_menu_cmd },
    { ' ', "&View               F3", 'v', view_panel_cmd },
    { ' ', "&Filtered view     M-!", 'f', filtered_view_cmd_cpanel },
    { ' ', "&Edit               F4", 'e', edit_panel_cmd },
    { ' ', "&Copy               F5", 'c', copy_cmd },
    { ' ', "c&Hmod           C-x c", 'h', chmod_cmd },
    { ' ', "&Link            C-x l", 'l', link_cmd },
    { ' ', "&SymLink         C-x s", 's', symlink_cmd },
    { ' ', "edit s&Ymlink  C-x C-s", 'y', edit_symlink_cmd },
    { ' ', "ch&Own           C-x o", 'o', chown_cmd },
    { ' ', "&Advanced chown       ", 'a', chown_advanced_cmd },
    { ' ', "&Rename/Move        F6", 'r', ren_cmd },
    { ' ', "&Mkdir              F7", 'm', mkdir_panel_cmd },
    { ' ', "&Delete             F8", 'd', delete_cmd },
    { ' ', "&Quick cd          M-c", 'q', quick_cd_cmd },
    { ' ', "", ' ', 0 },
    { ' ', "select &Group      M-+", 'g', select_cmd },
    { ' ', "u&Nselect group    M-\\",'n', unselect_cmd },
    { ' ', "reverse selec&Tion M-*", 't', reverse_selection_cmd },
    { ' ', "", ' ', 0 },
    { ' ', "e&Xit              F10", 'x', quit_cmd }
};

static void external_panelize (void);

static const menu_entry CmdMenu [] = {
    /* I know, I'm lazy, but the tree widget when it's not running
     * as a panel still has some problems, I have not yet finished
     * the WTree widget port, sorry.
     */
    { ' ', "&Directory tree",               'd', tree_box },
    { ' ', "&Find file            M-?",     'f', do_find },
    { ' ', "s&Wap panels          C-u",     'w', swap_cmd },
    { ' ', "switch &Panels on/off C-o",     'p', view_other_cmd },
    { ' ', "&Compare directories  C-x d",   'c', compare_dirs_cmd },
    { ' ', "e&Xternal panelize    C-x !",   'x', external_panelize },
    { ' ', "", ' ', 0 },
    { ' ', "command &History",              'h', history_cmd },
    { ' ', "di&Rectory hotlist    C-\\",    'r', quick_chdir_cmd },
#ifdef USE_VFS
    { ' ', "&Active VFS list      C-x a",   'a', reselect_vfs },
#endif
    { ' ', "", ' ', 0 },
    { ' ', "&Extension file edit",          'e', ext_cmd },
    { ' ', "&Menu file edit",               'm', menu_edit_cmd }
};

/* Must keep in sync with the constants in menu_cmd */
static const menu_entry OptMenu [] = {
    { ' ', "&Configuration...",    'c', configure_box },
    { ' ', "&Layout...",           'l', layout_cmd },
    { ' ', "c&Onfirmation...",     'o', confirm_box },
    { ' ', "&Display bits...",     'd', display_bits_box },
    { ' ', "learn &Keys...",       'k', learn_keys },
#ifdef USE_VFS
    { ' ', "&Virtual FS...",       'v', configure_vfs },
#endif
    { ' ', "", ' ', 0 },
    { ' ', "&Save setup",          's', save_setup_cmd }
};

#define menu_entries(x) sizeof(x)/sizeof(menu_entry)

static void
init_menu (void)
{
    int i;

    MenuBar [0] = create_menu ( horizontal_split ? " Above " : " Left ",
				PanelMenu, menu_entries (PanelMenu));
    MenuBar [1] = create_menu (" File ", FileMenu, menu_entries (FileMenu));
    MenuBar [2] = create_menu (" Command ", CmdMenu, menu_entries (CmdMenu));
    MenuBar [3] = create_menu (" Options ", OptMenu, menu_entries (OptMenu));
    MenuBar [4] = create_menu (horizontal_split ? " Below " : " Right ",
			       RightMenu, menu_entries (PanelMenu));
    for (i = 0; i < 5; i++)
	MenuBarEmpty [i] = create_menu (MenuBar [i]->name, 0, 0);
}


static inline void
destroy_menu (Menu menu)
{
    free (menu);
}

static void
done_menu (void)
{
    int i;

    for (i = 0; i < 5; i++) {
	destroy_menu (MenuBar [i]);
	destroy_menu (MenuBarEmpty [i]);
    }
}

static void
menu_last_selected_cmd (void)
{
    the_menubar->active = 1;
    the_menubar->dropped = drop_menus;
    the_menubar->previous_selection = dlg_item_number (midnight_dlg);
    dlg_select_widget (midnight_dlg, the_menubar);
}

static void
menu_cmd (void)
{
    if (the_menubar->active)
	return;

    if (get_current_index () == 0)
	the_menubar->selected = 0;
    else
	the_menubar->selected = 4;
    menu_last_selected_cmd ();
}

/* Flag toggling functions */

static void toggle_fast_reload (void)
{
    fast_reload = !fast_reload;
    if (fast_reload_w == 0 && fast_reload) {
	MC_message (0, " Information ",
		 " Using the fast reload option may not reflect the exact \n"
		   " directory contents. In this cases you'll need to do a  \n"
		   " manual reload of the directory. See the man page for   \n"
		   " the details.                                           ");
	fast_reload_w = 1;
    }
}

static void toggle_mix_all_files (void)
{
    mix_all_files = !mix_all_files;
    update_panels (UP_RELOAD, UP_KEEPSEL);
}

static void toggle_show_backup (void)
{
    show_backups = !show_backups;
    update_panels (UP_RELOAD, UP_KEEPSEL);
}

static void toggle_show_hidden (void)
{
    show_dot_files = !show_dot_files;
    update_panels (UP_RELOAD, UP_KEEPSEL);
}

static void create_panels (void)
{
    int current_index;
    int other_index;
    int current_mode_c;
    int other_mode;
    char original_dir [1024];

    original_dir [0] = 0;

    if (boot_current_is_left) {
	current_index = 0;
	other_index = 1;
	current_mode_c = startup_left_mode;
	other_mode = startup_right_mode;
    } else {
	current_index = 1;
	other_index = 0;
	current_mode_c = startup_right_mode;
	other_mode = startup_left_mode;
    }
    /* Creates the left panel */
    if (this_dir) {
	if (other_dir) {
	    /* Ok, user has specified two dirs, save the original one,
	     * since we may not be able to chdir to the proper
	     * second directory later
	     */
	    mc_get_current_wd (original_dir, sizeof (original_dir)-2);
	}
	mc_chdir (this_dir);
    }
    set_display_type (current_index, current_mode_c);

    /* The other panel */
    if (other_dir) {
	if (original_dir [0])
	    mc_chdir (original_dir);
	mc_chdir (other_dir);
    }
    set_display_type (other_index, other_mode);

    if (startup_left_mode == view_listing) {
	current_panel = left_panel;
    } else {
	if (right_panel)
	    current_panel = right_panel;
	else
	    current_panel = left_panel;
    }

    /* Create the nice widgets */
    cmdline     = command_new (0, 0, 0);
    the_prompt  = label_new (0, 0, MC_prompt);
    the_prompt->transparent = 1;
    the_bar     = buttonbar_new (keybar_visible);

    the_hint    = label_new (0, 0, 0);
    the_hint->transparent = 1;
    the_hint->auto_adjust_cols = 0;
    the_hint->widget.cols = COLS;

    the_menubar = menubar_new (0, 0, COLS, MenuBar, 5);
}

static void copy_current_pathname (void)
{
    if (!command_prompt)
	return;

    stuff (input_w (cmdline), cpanel->cwd, 0);
    if (cpanel->cwd [strlen (cpanel->cwd) - 1] != '/')
	stuff (input_w (cmdline), "/", 0);
}

static void copy_other_pathname (void)
{
    if (get_other_type () != view_listing)
	return;

    if (!command_prompt)
	return;

    stuff (input_w (cmdline), opanel->cwd, 0);
    if (cpanel->cwd [strlen (opanel->cwd) - 1] != '/')
	stuff (input_w (cmdline), "/", 0);
}

static void copy_readlink (WPanel *panel)
{
    if (!command_prompt)
	return;
    if (S_ISLNK (selection (panel)->buf.st_mode)) {
	char buffer [MC_MAXPATHLEN];
	char *p = concat_path_file (panel->cwd, selection (panel)->fname);
	int i;

	i = mc_readlink (p, buffer, MC_MAXPATHLEN);
	free (p);
	if (i > 0) {
	    buffer [i] = 0;
	    stuff (input_w (cmdline), buffer, 0);
	}
    }
}

static void copy_current_readlink (void)
{
    copy_readlink (cpanel);
}

static void copy_other_readlink (void)
{
    if (get_other_type () != view_listing)
	return;
    copy_readlink (opanel);
}

/* Inserts the selected file name into the input line */
/* Exported so that the command modules uses it */
static void copy_prog_name (void)
{
    char *tmp;
    if (!command_prompt)
	return;

    if (get_current_type () == view_tree) {
	WTree *tree = (WTree *) get_panel_widget (get_current_index ());
	tmp = name_quote (tree->selected_ptr->name, 1);
    } else
	tmp = name_quote (selection (cpanel)->fname, 1);
    stuff (input_w (cmdline), tmp, 1);
    free (tmp);
}

static inline void
input_disable_update (WInput *in)
{
    in->disable_update++;
}


static void copy_tagged (WPanel *panel)
{
    int i;

    if (!command_prompt)
	return;
    input_disable_update (input_w (cmdline));
    if (panel->marked) {
	for (i = 0; i < panel->count; i++)
	    if (panel->dir.list [i].f.marked) {
		char *tmp = name_quote (panel->dir.list [i].fname, 1);
		stuff (input_w (cmdline), tmp, 1);
		free (tmp);
	    }
    } else {
	char *tmp = name_quote (panel->dir.list [panel->selected].fname, 1);
	stuff (input_w (cmdline), tmp, 1);
	free (tmp);
    }
    input_enable_update (input_w (cmdline));
}

static void copy_current_tagged (void)
{
    copy_tagged (cpanel);
}

static void copy_other_tagged (void)
{
    if (get_other_type () != view_listing)
	return;
    copy_tagged (opanel);
}

void suspend_cmd (void)
{
    save_cwds_stat ();
    pre_exec ();

    if (console_flag && !use_subshell)
	restore_console ();

    {
	struct sigaction sigtstp_action;

	/* Make sure that the SIGTSTP below will suspend us directly,
	   without calling ncurses' SIGTSTP handler; we *don't* want
	   ncurses to redraw the screen immediately after the SIGCONT */
	sigaction (SIGTSTP, &startup_handler, &sigtstp_action);

	kill (getpid (), SIGTSTP);

	/* Restore previous SIGTSTP action */
	bb_sigaction2 (SIGTSTP, &sigtstp_action);
    }

    if (console_flag && !use_subshell)
	handle_console (CONSOLE_SAVE);

    edition_post_exec ();
    update_panels (UP_OPTIMIZE, UP_KEEPSEL);
    do_refresh ();
}

static void init_labels (Widget *paneletc)
{
    define_label (midnight_dlg, paneletc, 1, "Help", help_cmd);
    define_label (midnight_dlg, paneletc, 2, "Menu", user_menu_cmd);
    define_label (midnight_dlg, paneletc, 9, "PullDn", menu_cmd);
    define_label (midnight_dlg, paneletc, 10, "Quit", quit_cmd);
}

static const key_map ctl_x_map [] = {
    { XCTRL('c'),   quit_cmd },
#ifdef USE_VFS
    { 'a',          reselect_vfs },
#endif
    { 'd',          compare_dirs_cmd },
    { 'p',          copy_current_pathname },
    { XCTRL('p'),   copy_other_pathname },
    { 't',          copy_current_tagged },
    { XCTRL('t'),   copy_other_tagged },
    { 'c',          chmod_cmd },
    { 'o',          chown_cmd },
    { 'l',          link_cmd },
    { XCTRL('l'),   other_symlink_cmd },
    { 's',          symlink_cmd },
    { XCTRL('s'),   edit_symlink_cmd },
    { 'r',          copy_current_readlink },
    { XCTRL('r'),   copy_other_readlink },
    { 'i',          info_cmd_no_menu },
    { 'q',          quick_cmd_no_menu },
    { 'h',          add2hotlist_cmd },
    { '!',          external_panelize },
    { 0,  0 }
};

static void ctl_x_cmd (void)
{
	ctl_x_map_enabled = 1;
}

static void Nothing (void)
{
}

static const key_map default_map [] = {
    { KEY_F(19),  menu_last_selected_cmd },
    { KEY_F(20),  quiet_quit_cmd },

    /* Copy useful information to the command line */
    { ALT('\n'),  copy_prog_name },
    { ALT('\r'),  copy_prog_name },
    { ALT('a'),   copy_current_pathname },
    { ALT('A'),   copy_other_pathname },

    { ALT('c'),   quick_cd_cmd },

    /* To access the directory hotlist */
    { XCTRL('\\'), quick_chdir_cmd },

    /* Suspend */
    { XCTRL('z'), suspend_cmd },
    /* The filtered view command */
    { ALT('!'),   filtered_view_cmd_cpanel },

    /* Find file */
    { ALT('?'),   do_find },

    /* Panel refresh */
    { XCTRL('r'), reread_cmd },

    { ALT('t'),   toggle_listing_cmd },

    /* Swap panels */
    { XCTRL('u'), swap_cmd },

    /* View output */
    { XCTRL('o'), view_other_cmd },

    /* Control-X keybindings */
    { XCTRL('x'), ctl_x_cmd },

    /* Trap dlg's exit commands */
    { ESC_CHAR,   Nothing },
    { XCTRL('c'), Nothing },
    { XCTRL('g'), Nothing },
    { 0, 0 },
};

#ifdef SIGWINCH
static void flag_winch (int dummy)
{
	winch_flag = 1;
}
#endif

static void
setup_post (void)
{
    int i;

#ifdef SIGWINCH
    bb_signal((1 << SIGWINCH), flag_winch);
#endif

    for (i = 0; i < UID_CACHE_SIZE; i++)
	uid_cache [i].string = 0;

    for (i = 0; i < GID_CACHE_SIZE; i++)
	 gid_cache [i].string = 0;

    if (SLang_TT_Baud_Rate < 9600 || slow_terminal)
	MC_verbose = 0;
    init_mouse ();

    midnight_colors [0] = 0;
    midnight_colors [1] = REVERSE_COLOR;     /* FOCUSC */
    midnight_colors [2] = INPUT_COLOR;       /* HOT_NORMALC */
    midnight_colors [3] = NORMAL_COLOR;      /* HOT_FOCUSC */
}

static void setup_mc (void)
{
    /* Call all the inits */
    SLsmg_Display_Eight_Bit = full_eight_bits ? 128 : 160;
    init_menu ();
    create_panels ();
    setup_panels ();
    setup_post ();
}

static void done_mc (void)
{
    done_menu ();

    /* Setup shutdown
     *
     * We sync the profiles since the hotlist may have changed, while
     * we only change the setup data if we have the auto save feature set
     */
    if (auto_save_setup)
	save_setup ();   /* does also call save_hotlist */
    else
	save_hotlist();
    done_screen ();
    vfs_add_current_stamps ();
    if (xterm_flag && xterm_hintbar)
	set_hintbar("Thank you for using GNU Midnight Commander");
}

/* This should be called after destroy_dlg since panel widgets
 *  save their state on the profiles
 */
static inline void done_mc_profile (void)
{
    if (!auto_save_setup)
	profile_forget_profile (profile_name);
    sync_profiles ();
    done_setup ();
    free_profiles ();
}

/* This routine only handles cpanel, and opanel, it is easy to
 * change to use npanels, just loop over the number of panels
 * and use get_panel_widget (i) and make the test done below
 */
static void make_panels_dirty (void)
{
    if (cpanel->dirty)
	panel_update_contents (cpanel);

    if ((get_other_type () == view_listing) && opanel->dirty)
	panel_update_contents (opanel);
}

#define check_key_backslash(x) ((x) == '\\')

static int midnight_callback (struct Dlg_head *h, int id, int msg)
{
    int i;

    switch (msg) {

	/* Speed up routine: now, we just set the  */
    case DLG_PRE_EVENT:
	make_panels_dirty ();
	return MSG_HANDLED;

    case DLG_KEY:
	if (ctl_x_map_enabled) {
		ctl_x_map_enabled = 0;
		for (i = 0; ctl_x_map [i].key_code; i++)
			if (id == ctl_x_map [i].key_code) {
				(*ctl_x_map [i].fn)(id);
				return MSG_HANDLED;
			}
	}

	if (id == KEY_F(10) && !the_menubar->active) {
	    quit_cmd ();
	    return MSG_HANDLED;
	}

	if (id == '\t')
	    free_completions (input_w (cmdline));

	/* On Linux, we can tell the difference */
	if (id == '\n' && ctrl_pressed ()) {
	    copy_prog_name ();
	    return MSG_HANDLED;
	}

	if (id == '\n' && input_w (cmdline)->buffer [0]) {
	    send_message_to (h, (Widget *) cmdline, WIDGET_KEY, id);
	    return MSG_HANDLED;
	}

	if ((!alternate_plus_minus || !(console_flag || xterm_flag)) &&
	     !quote && !cpanel->searching) {
	    if(!only_leading_plus_minus) {
		/* Special treatement, since the input line will eat them */
		if (id == '+' ) {
		    select_cmd ();
		    return MSG_HANDLED;
		}

		if (check_key_backslash (id) || id == '-') {
		    unselect_cmd ();
		    return MSG_HANDLED;
		}

		if (id == '*') {
		    reverse_selection_cmd ();
		    return MSG_HANDLED;
		}
	    } else if (command_prompt && !strlen (input_w (cmdline)->buffer)) {
		/* Special treatement '+', '-', '\', '*' only when this is
		 * first char on input line
		 */

		if (id == '+') {
		    select_cmd ();
		    return MSG_HANDLED;
		}

		if (check_key_backslash (id) || id == '-') {
		    unselect_cmd ();
		    return MSG_HANDLED;
		}

		if (id == '*') {
		    reverse_selection_cmd ();
		    return MSG_HANDLED;
		}
	    }
	}
	break;

    case DLG_HOTKEY_HANDLED:
	if (get_current_type () == view_listing)
	    cpanel->searching = 0;
	break;

    case DLG_UNHANDLED_KEY:
	if (command_prompt) {
	    int v;

	    v = send_message_to (h, (Widget *) cmdline, WIDGET_KEY, id);
	    if (v)
		return v;
	}
	if (ctl_x_map_enabled) {
		ctl_x_map_enabled = 0;
		for (i = 0; ctl_x_map [i].key_code; i++)
			if (id == ctl_x_map [i].key_code) {
				(*ctl_x_map [i].fn)(id);
				return MSG_HANDLED;
			}
	} else {
		for (i = 0; default_map [i].key_code; i++) {
			if (id == default_map [i].key_code) {
				(*default_map [i].fn)(id);
				return MSG_HANDLED;
			}
		}
	}
	return MSG_NOT_HANDLED;

	/* We handle the special case of the output lines */
    case DLG_DRAW:
	attrset (SELECTED_COLOR);
	if (console_flag && output_lines)
	    show_console_contents (output_start_y,
				   LINES-output_lines-keybar_visible-1,
				   LINES-keybar_visible-1);
	break;

    }
    return default_dlg_callback (h, id, msg);
}

#ifndef PORT_HAS_LOAD_HINT
void load_hint (void)
{
    char *hint;

    if (!the_hint->widget.parent)
	return;

    if (!message_visible && (!xterm_flag || !xterm_hintbar)) {
	label_set_text (the_hint, 0);
	return;
    }

    if ((hint = get_random_hint ())) {
	if (*hint)
	    set_hintbar (hint);
	free (hint);
    } else {
	set_hintbar ("The Midnight Commander " VERSION
			" (C) 1995-1997 the Free Software Foundation");
    }
}
#endif

static void
setup_panels_and_run_mc (void)
{
    int first, second;

    add_widget (midnight_dlg, the_hint);
    load_hint ();
    add_widgetl (midnight_dlg, cmdline, XV_WLAY_RIGHTOF);
    add_widgetl (midnight_dlg, the_prompt, XV_WLAY_DONTCARE);
    add_widget (midnight_dlg, the_bar);
    if (boot_current_is_left) {
	first = 1;
	second = 0;
    } else {
	first = 0;
	second = 1;
    }
    add_widget (midnight_dlg, get_panel_widget (first));
    add_widget (midnight_dlg, get_panel_widget (second));
    add_widget (midnight_dlg, the_menubar);

    init_labels (get_panel_widget (0));
    init_labels (get_panel_widget (1));

    /* Run the Midnight Commander if no file was specified in the command line */
    run_dlg (midnight_dlg);
}

static void
do_nc (void)
{
    midnight_dlg = create_dlg (0, 0, LINES, COLS, midnight_colors, midnight_callback, "[main]", 0);
    midnight_dlg->has_menubar = 1;

    setup_mc ();

    setup_panels_and_run_mc ();

    /* Program end */
    midnight_shutdown = 1;

    /* destroy_dlg destroys even cpanel->cwd, so we have to save a copy :) */
    if (print_last_wd) {
	if (!vfs_current_is_local ())
	    last_wd_string = strdup (".");
	else
	    last_wd_string = strdup (cpanel->cwd);
    }
    done_mc ();

    destroy_dlg (midnight_dlg);
    current_panel = 0;
    done_mc_profile ();
}

/* This just computes a nice value for the features variable */

#ifndef VERSION
#    define VERSION "undefined"
#endif

static const char features[] =
    "Edition: text mode\n"

#ifdef USE_VFS
    "Virtual File System: tarfs, extfs"
#ifdef USE_NETCODE
    ", ftpfs"
    ", mcfs"
#endif
    ".\n"
#endif

    "Using S-lang library with harcoded-terminfo database"
    "\n"
;

static void version (void)
{
     fprintf (stderr, "The Midnight Commander %s\n"
		"with mouse support on xterm.\n%s", VERSION, features);
    if (print_last_wd)
	write (stdout_fd, ".", 1);
}

static void
OS_Setup (void)
{
    const char *termvalue;
    char   *mc_libdir;

    TERM_env = getenv ("TERM");
    termvalue = TERM_env;
    if (!termvalue) {
	bb_error_msg ("The TERM environment variable is unset!");
	termvalue = "";
    }
    if (force_xterm || (strncmp (termvalue, "xterm", 5) == 0 || strcmp (termvalue, "dtterm") == 0)) {
	use_mouse_p = XTERM_MOUSE;
	xterm_flag = 1;
#ifdef SET_TITLE
	printf ("\33]0;GNU Midnight Commander\7");
#endif
    }
    MC_shell = getenv ("SHELL");
    if (!MC_shell || !*MC_shell)
	MC_shell = strdup (getpwuid (my_euid)->pw_shell);
    if (!MC_shell || !*MC_shell)
	MC_shell = DEFAULT_SHELL;

    sprintf (control_file, CONTROL_FILE, getpid ());
    my_putenv ("MC_CONTROL_FILE", control_file);

    /* This is the directory, where MC was installed, on Unix this is LIBDIR */
    /* and can be overriden by the MCHOME environment variable */
    if ((mc_libdir = getenv ("MCHOME")) != NULL) {
	mc_home = strdup (mc_libdir);
    } else {
	mc_home = strdup (LIBDIR);
    }
}

static void
sigchld_handler_no_subshell (int sig)
{
    int pid, Status;

    if (!console_flag)
	return;

    /* COMMENT: if it were true that after the call to handle_console(..INIT)
       the value of console_flag never changed, we could simply not install
       this handler at all if (!console_flag && !use_subshell). */

    /* That comment is no longer true.  We need to wait() on a sigchld
       handler (that's at least what the tarfs code expects currently). */

    pid = waitpid_no_int (cons_saver_pid, &Status, WUNTRACED | WNOHANG, 0);

    if (pid == cons_saver_pid) {
	/* {{{ Someone has stopped or killed cons.saver; restart it */

	if (WIFSTOPPED (Status))
	    kill (pid, SIGCONT);
	else
	{
	    handle_console (CONSOLE_DONE);
	    handle_console (CONSOLE_INIT);
	}
	/* }}} */
    }

    /* If we get here, some other child exited; ignore it */
}

static void init_sigchld (void)
{
    bb_signal((1 << SIGCHLD), sigchld_handler_no_subshell);
}


/*
 * The compatibility_move_mc_files routine is intended to
 * move all of the hidden .mc files into a private ~/.mc
 * directory in the home directory, to avoid cluttering the users.
 *
 * Previous versions of the program had all of their files in
 * the $HOME, we are now putting them in $HOME/.mc
 */
static int
do_mc_filename_rename (const char *mc_dir, char *o_name, char *n_name)
{
	char *full_o_name = concat_path_file (home_dir, o_name);
	char *full_n_name = xasprintf ("%s%s%s", home_dir, MC_BASE, n_name);
	int flg_move;

	flg_move = 0 == rename (full_o_name, full_n_name);
	free (full_o_name);
	free (full_n_name);
	return flg_move;
}

static void
do_compatibility_move (const char *mc_dir)
{
	struct stat s;
	int flg_move;

	if (stat (mc_dir, &s) == 0)
		return;
	if (errno != ENOENT)
		return;

	if (mkdir (mc_dir, 0777) == -1)
		return;

	flg_move = do_mc_filename_rename (mc_dir, ".mc.ini", "ini");
	flg_move += do_mc_filename_rename (mc_dir, ".mc.hot", "hotlist");
	flg_move += do_mc_filename_rename (mc_dir, ".mc.ext", "ext");
	flg_move += do_mc_filename_rename (mc_dir, MC_LOCAL_MENU, "menu");
	flg_move += do_mc_filename_rename (mc_dir, ".mc.bashrc", "bashrc");
	flg_move += do_mc_filename_rename (mc_dir, ".mc.inputrc", "inputrc");
	flg_move += do_mc_filename_rename (mc_dir, ".mc.tcshrc", "tcshrc");
	flg_move += do_mc_filename_rename (mc_dir, ".mc.tree", "tree");

	if (!flg_move)
		return;

	show_change_notice = 1;
}

static void
compatibility_move_mc_files (void)
{
	char *mc_dir = concat_path_file (home_dir, ".mc");

	do_compatibility_move (mc_dir);
	free (mc_dir);
}

/* Pulldown menu code. */

static Menu create_menu (const char *name, const menu_entry *entries, int count_m)
{
    Menu menu;

    menu = x_malloc (sizeof (*menu), "create_menu");
    menu->count = count_m;
    menu->max_entry_len = 0;
    menu->entries = entries;

    menu->name = name;
    menu->start_x = 0;
    return menu;
}

static void menubar_drop_compute (WMenu *menubar)
{
    const Menu menu = menubar->menu [menubar->selected];
    int   max_entry_len = 0;
    int i;

    for (i = 0; i < menu->count; i++)
	max_entry_len = max (max_entry_len, strlen (menu->entries [i].text));
    menubar->max_entry_len = max_entry_len = max (max_entry_len, 20);
}

static void menubar_paint_idx (WMenu *menubar, int idx, int color)
{
    const Menu menu = menubar->menu [menubar->selected];
    const int y = 2 + idx;
    int x = menubar-> menu[menubar->selected]->start_x;

    if (x + menubar->max_entry_len + 3 > menubar->widget.cols)
	x = menubar->widget.cols - menubar->max_entry_len - 3;

    widget_move (&menubar->widget, y, x);
    attrset (color);
    hline (' ', menubar->max_entry_len+2);
    if (!*menu->entries [idx].text) {
	attrset (SELECTED_COLOR);
	widget_move (&menubar->widget, y, x + 1);
	hline (slow_terminal ? ' ' : ACS_HLINE, menubar->max_entry_len);
    } else {
	const unsigned char *text;

	addch((unsigned char)menu->entries [idx].first_letter);
	for (text = (unsigned char *)menu->entries [idx].text; *text; text++)
	{
		if (*text == '&')
		{
			++text;
		     /*   menu->entries [idx].hot_key = tolower(*text); */
			attrset (color == MENU_SELECTED_COLOR ?
				MENU_HOTSEL_COLOR : MENU_HOT_COLOR);
			addch(*text);
			attrset(color);
			continue;
		}
		addch(*text);
	}
    }
    widget_move (&menubar->widget, y, x + 1);
}

static inline void menubar_draw_drop (WMenu *menubar)
{
    const int count_m = (menubar->menu [menubar->selected])->count;
    int   i;
    int   sel = menubar->subsel;
    int   column = menubar-> menu[menubar->selected]->start_x - 1;

    if (column + menubar->max_entry_len + 4 > menubar->widget.cols)
	column = menubar->widget.cols - menubar->max_entry_len - 4;

    attrset (SELECTED_COLOR);
    draw_box (menubar->widget.parent,
	      menubar->widget.y+1, menubar->widget.x + column,
	      count_m+2, menubar->max_entry_len + 4);

    column++;
    for (i = 0; i < count_m; i++) {
	if (i == sel)
	    continue;
	menubar_paint_idx (menubar, i, MENU_ENTRY_COLOR);
    }
    menubar_paint_idx (menubar, sel, MENU_SELECTED_COLOR);
}

static void menubar_draw (WMenu *menubar)
{
    const int items = menubar->items;
    int   i;

    /* First draw the complete menubar */
    attrset (SELECTED_COLOR);
    widget_move (&menubar->widget, 0, 0);

    /* ncurses bug: it should work with hline but it does not */
    for (i =  menubar->widget.cols; i; i--)
	addch (' ');

    attrset (SELECTED_COLOR);
    /* Now each one of the entries */
    for (i = 0; i < items; i++) {
	if (menubar->active)
	    attrset(i == menubar->selected?MENU_SELECTED_COLOR:SELECTED_COLOR);
	widget_move (&menubar->widget, 0, menubar->menu [i]->start_x);
	printw ("%s", menubar->menu [i]->name);
    }

    if (menubar->dropped)
	menubar_draw_drop (menubar);
    else
	widget_move (&menubar->widget, 0,
		menubar-> menu[menubar->selected]->start_x);
}

static void menubar_remove (WMenu *menubar)
{
    menubar->subsel = 0;
    if (menubar->dropped) {
	menubar->dropped = 0;
	do_refresh ();
	menubar->dropped = 1;
    }
}

static void menubar_left (WMenu *menu)
{
    menubar_remove (menu);
    menu->selected = (menu->selected - 1) % menu->items;
    if (menu->selected < 0)
	menu->selected = menu->items -1;
    menubar_drop_compute (menu);
    menubar_draw (menu);
}

static void menubar_right (WMenu *menu)
{
    menubar_remove (menu);
    menu->selected = (menu->selected + 1) % menu->items;
    menubar_drop_compute (menu);
    menubar_draw (menu);
}

static void menubar_finish (WMenu *menubar)
{
    menubar->dropped = 0;
    menubar->active = 0;
    menubar->widget.lines = 1;
    widget_want_hotkey (menubar->widget, 0);
    dlg_select_nth_widget (menubar->widget.parent,
			   menubar->previous_selection);
    do_refresh ();
}

static void menubar_drop (WMenu *menubar, int selected)
{
    menubar->dropped = 1;
    menubar->selected = selected;
    menubar->subsel = 0;
    menubar_drop_compute (menubar);
    menubar_draw (menubar);
}

static void menubar_execute (WMenu *menubar, int entry)
{
    const Menu menu = menubar->menu [menubar->selected];

    is_right = menubar->selected != 0;
    (*menu->entries [entry].call_back)();
    menubar_finish (menubar);
}

static void menubar_move (WMenu *menubar, int step)
{
    const Menu menu = menubar->menu [menubar->selected];

    menubar_paint_idx (menubar, menubar->subsel, MENU_ENTRY_COLOR);
    do {
	menubar->subsel += step;
	if (menubar->subsel < 0)
	    menubar->subsel = menu->count - 1;

	menubar->subsel %= menu->count;
    } while  (!menu->entries [menubar->subsel].call_back);
    menubar_paint_idx (menubar, menubar->subsel, MENU_SELECTED_COLOR);
}

static int menubar_handle_key (WMenu *menubar, int key)
{
    int   i;

    /* Lowercase */
    if (key < 256 && isalpha (key)) /* Linux libc.so.5.x.x bug fix */
	key = tolower (key);

    if (is_abort_char (key)) {
	menubar_finish (menubar);
	return 1;
    }

    if (key == KEY_LEFT || key == XCTRL('b')) {
	menubar_left (menubar);
	return 1;
    } else if (key == KEY_RIGHT || key == XCTRL ('f')) {
	menubar_right (menubar);
	return 1;
    }

    /* .ado: NT Alpha can not allow CTRL in Menubar */
#if defined(_OS_NT)
       if (!key)
	       return 0;
#endif

    if (!menubar->dropped) {
	const int items = menubar->items;
	for (i = 0; i < items; i++) {
	    const Menu menu = menubar->menu [i];

	    /* Hack, we should check for the upper case letter */
	    if (tolower (menu->name [1]) == key) {
		menubar_drop (menubar, i);
		return 1;
	    }
	}
	if (key == KEY_ENTER || key == XCTRL ('n') || key == KEY_DOWN
	    || key == '\n') {
	    menubar_drop (menubar, menubar->selected);
	    return 1;
	}
	return 1;
    } else {
	const int selected = menubar->selected;
	const Menu menu = menubar->menu [selected];
	const int items = menu->count;

	for (i = 0; i < items; i++) {
	    if (!menu->entries [i].call_back)
		continue;

	    if (key != menu->entries [i].hot_key)
		continue;

	    menubar_execute (menubar, i);
	    return 1;
	}

	if (key == KEY_ENTER || key == '\n') {
	    menubar_execute (menubar, menubar->subsel);
	    return 1;
	}


	if (key == KEY_DOWN || key == XCTRL ('n'))
	    menubar_move (menubar, 1);

	if (key == KEY_UP || key == XCTRL ('p'))
	    menubar_move (menubar, -1);
    }
    return 0;
}

static int menubar_callback (Dlg_head *h, WMenu *menubar, int msg, int par)
{
    switch (msg) {
	/* We do not want the focus unless we have been activated */
    case WIDGET_FOCUS:
	if (menubar->active) {
	    widget_want_cursor (menubar->widget, 1);

	    /* Trick to get all the mouse events */
	    menubar->widget.lines = LINES;

	    /* Trick to get all of the hotkeys */
	    widget_want_hotkey (menubar->widget, 1);
	    menubar->subsel = 0;
	    menubar_drop_compute (menubar);
	    menubar_draw (menubar);
	    return 1;
	} else
	    return 0;

	/* We don't want the buttonbar to activate while using the menubar */
    case WIDGET_HOTKEY:
    case WIDGET_KEY:
	if (menubar->active) {
	    menubar_handle_key (menubar, par);
	    return 1;
	} else
	    return 0;

    case WIDGET_CURSOR:
	/* Put the cursor in a suitable place */
	return 0;

    case WIDGET_UNFOCUS:
	if (menubar->active)
	    return 0;
	else {
	    widget_want_cursor (menubar->widget, 0);
	    return 1;
	}

    case WIDGET_DRAW:
	if (menubar_visible)
	    menubar_draw (menubar);
    }
    return default_proc (h, msg, par);
}

static int
menubar_event    (Gpm_Event *event, WMenu *menubar)
{
    int was_active;
    int new_selection;
    int left_x, right_x, bottom_y;

    if (!(event->type & (GPM_UP|GPM_DOWN|GPM_DRAG)))
	return MOU_NORMAL;

    if (!menubar->dropped) {
	menubar->previous_selection = dlg_item_number(menubar->widget.parent);
	menubar->active = 1;
	menubar->dropped = 1;
	was_active = 0;
    } else
	was_active = 1;

    /* Mouse operations on the menubar */
    if (event->y == 1 || !was_active) {
	if (event->type & GPM_UP)
	    return MOU_NORMAL;

	new_selection = 0;
	while (new_selection < menubar->items
		&& event->x > menubar->menu[new_selection]->start_x
	)
		new_selection++;

	if (new_selection) /* Don't set the invalid value -1 */
		--new_selection;

	if (!was_active) {
	    menubar->selected = new_selection;
	    dlg_select_widget (menubar->widget.parent, menubar);
	    menubar_drop_compute (menubar);
	    menubar_draw (menubar);
	    return MOU_NORMAL;
	}

	menubar_remove (menubar);

	menubar->selected = new_selection;

	menubar_drop_compute (menubar);
	menubar_draw (menubar);
	return MOU_NORMAL;
    }

    if (!menubar->dropped)
	return MOU_NORMAL;

    /* Ignore the events on anything below the third line */
    if (event->y <= 2)
	return MOU_NORMAL;

    /* Else, the mouse operation is on the menus or it is not */
	left_x = menubar->menu[menubar->selected]->start_x;
	right_x = left_x + menubar->max_entry_len + 4;
	if (right_x > menubar->widget.cols)
	{
		left_x = menubar->widget.cols - menubar->max_entry_len - 3;
		right_x = menubar->widget.cols - 1;
	}

    bottom_y = (menubar->menu [menubar->selected])->count + 3;

    if ((event->x > left_x) && (event->x < right_x) && (event->y < bottom_y)) {
	int pos = event->y - 3;

	if (!menubar->menu [menubar->selected]->entries [pos].call_back)
	    return MOU_NORMAL;

	menubar_paint_idx (menubar, menubar->subsel, MENU_ENTRY_COLOR);
	menubar->subsel = pos;
	menubar_paint_idx (menubar, menubar->subsel, MENU_SELECTED_COLOR);

	if (event->type & GPM_UP)
	    menubar_execute (menubar, pos);
    } else
	if (event->type & GPM_DOWN)
	    menubar_finish (menubar);

    return MOU_NORMAL;
}

static void menubar_destroy (WMenu *menubar)
{
}

/*
 * Properly space menubar items. Should be called when menubar is created
 * and also when widget width is changed (i.e. upon xterm resize).
 */
void
menubar_arrange(WMenu* menubar)
{
	int i, start_x = 1;
	int items = menubar->items;

#ifndef RESIZABLE_MENUBAR
	int gap = 3;

	for (i = 0; i < items; i++)
	{
		int len = strlen(menubar->menu[i]->name);
		menubar->menu[i]->start_x = start_x;
		start_x += len + gap;
	}

#else /* RESIZABLE_MENUBAR */

	int gap = menubar->widget.cols - 2;

	/* First, calculate gap between items... */
	for (i = 0; i < items; i++)
	{
		/* preserve length here, to be used below */
		gap -= (menubar->menu[i]->start_x = strlen(menubar->menu[i]->name));
	}

	gap /= (items - 1);

	if (gap <= 0)
	{
		/* We are out of luck - window is too narrow... */
		gap = 1;
	}

	/* ...and now fix start positions of menubar items */
	for (i = 0; i < items; i++)
	{
		int len = menubar->menu[i]->start_x;
		menubar->menu[i]->start_x = start_x;
		start_x += len + gap;
	}
#endif /* RESIZABLE_MENUBAR */
}


WMenu *menubar_new (int y, int x, int cols, Menu menu [], int items)
{
    WMenu *menubar = zmalloc (sizeof (WMenu));

    /* FIXME: subsel used w/o being set */
    init_widget (&menubar->widget, y, x, 1, cols,
		 (callback_fn) menubar_callback,
		 (destroy_fn)  menubar_destroy,
		 (mouse_h)     menubar_event);
    menubar->menu = menu;
    menubar->items = items;
    widget_want_cursor (menubar->widget, 0);
    menubar_arrange(menubar);

    return menubar;
}
/* mountlist.c -- return a list of mounted filesystems */

#ifndef NO_INFOMOUNT

/* Return the value of the hexadecimal number represented by CP.
   No prefix (like '0x') or suffix (like 'h') is expected to be
   part of CP. */

static int xatoi (char *cp)
{
    int val;

    val = 0;
    while (*cp) {
	if (*cp >= 'a' && *cp <= 'f')
	    val = val * 16 + *cp - 'a' + 10;
	else if (*cp >= 'A' && *cp <= 'F')
	    val = val * 16 + *cp - 'A' + 10;
	else if (*cp >= '0' && *cp <= '9')
	    val = val * 16 + *cp - '0';
	else
	    break;
	cp++;
    }
    return val;
}

#if defined (MOUNTED_GETMNTINFO) && !defined (xBSD)
static char *fstype_to_string (short t)
{
    switch (t) {
    case MOUNT_UFS:
	return "ufs";
    case MOUNT_NFS:
	return "nfs";
#ifdef MOUNT_PC
    case MOUNT_PC:
	return "pc";
#endif
#ifdef MOUNT_MFS
    case MOUNT_MFS:
	return "mfs";
#endif
#ifdef MOUNT_LO
    case MOUNT_LO:
	return "lo";
#endif
#ifdef MOUNT_TFS
    case MOUNT_TFS:
	return "tfs";
#endif
#ifdef MOUNT_TMP
    case MOUNT_TMP:
	return "tmp";
#endif
    default:
	return "?";
    }
}
#endif                          /* MOUNTED_GETMNTINFO */

#ifdef MOUNTED_VMOUNT           /* AIX.  */
static char *fstype_to_string (int t)
{
    struct vfs_ent *e;

    e = getvfsbytype (t);
    if (!e || !e->vfsent_name)
	return "none";
    else
	return e->vfsent_name;
}
#endif                          /* MOUNTED_VMOUNT */

/* Return a list of the currently mounted filesystems, or NULL on error.
   Add each entry to the tail of the list so that they stay in order.
   If NEED_FS_TYPE is nonzero, ensure that the filesystem type fields in
   the returned list are valid.  Otherwise, they might not be.
   If ALL_FS is zero, do not return entries for filesystems that
   are automounter (dummy) entries.  */

static void read_filesystem_list (void)
{
    struct mount_entry *mount_list0;
    struct mount_entry *me;
    struct mount_entry *mtail;

    /* Start the list off with a dummy entry. */
    me = zmalloc (sizeof (struct mount_entry));
    mount_list0 = mtail = me;

    {
	struct mntent *mnt;
#if !defined(MOUNTED) && defined(_PATH_MOUNTED)
#define MOUNTED _PATH_MOUNTED
#endif
	char *table = MOUNTED;
	FILE *fp;
	char *devopt;

	fp = setmntent (table, "r");
	if (fp == NULL)
	    return;

	while ((mnt = getmntent (fp))) {
	    me = xmalloc (sizeof (struct mount_entry));
	    me->me_devname = strdup (mnt->mnt_fsname);
	    me->me_mountdir = strdup (mnt->mnt_dir);
	    me->me_type = strdup (mnt->mnt_type);
	    devopt = strstr (mnt->mnt_opts, "dev=");
	    if (devopt) {
		if (devopt[4] == '0' && (devopt[5] == 'x' || devopt[5] == 'X'))
		    me->me_dev = xatoi (devopt + 6);
		else
		    me->me_dev = xatoi (devopt + 4);
	    } else
		me->me_dev = -1;        /* Magic; means not known yet. */
	    me->me_next = NULL;

	    /* Add to the linked list. */
	    mtail->me_next = me;
	    mtail = me;
	}

	if (endmntent (fp) == 0)
	    return;
    }

#ifdef MOUNTED_GETMNTINFO       /* 4.4BSD.  */
    {
	struct statfs *fsp;
	int entries;

	entries = getmntinfo (&fsp, MNT_NOWAIT);
	if (entries < 0)
	    return;
	while (entries-- > 0) {
	    me = xmalloc (sizeof (struct mount_entry));
	    me->me_devname = strdup (fsp->f_mntfromname);
	    me->me_mountdir = strdup (fsp->f_mntonname);
#ifdef xBSD
	    me->me_type = strdup (fsp->f_fstypename);
#else
	    me->me_type = fstype_to_string (fsp->f_type);
#endif
	    me->me_dev = -1;    /* Magic; means not known yet. */
	    me->me_next = NULL;

	    /* Add to the linked list. */
	    mtail->me_next = me;
	    mtail = me;
	    fsp++;
	}
    }
#endif                          /* MOUNTED_GETMNTINFO */

#ifdef MOUNTED_GETMNT           /* Ultrix.  */
    {
	int offset = 0;
	int val;
	struct fs_data fsd;

	while ((val = getmnt (&offset, &fsd, sizeof (fsd), NOSTAT_MANY, (char *) 0)) > 0) {
	    me = zmalloc (sizeof (struct mount_entry));
	    me->me_devname = strdup (fsd.fd_req.devname);
	    me->me_mountdir = strdup (fsd.fd_req.path);
	    me->me_type = gt_names[fsd.fd_req.fstype];
	    me->me_dev = fsd.fd_req.dev;
	    /* Add to the linked list. */
	    mtail->me_next = me;
	    mtail = me;
	}
	if (val < 0)
	    return;
    }
#endif                          /* MOUNTED_GETMNT. */

#if defined (MOUNTED_GETFSSTAT) /* __alpha running OSF_1 */
    {
	int numsys, counter, bufsize;
	struct statfs *stats;

	numsys = getfsstat ((struct statfs *) 0, 0L, MNT_WAIT);
	if (numsys < 0)
	    return;

	bufsize = (1 + numsys) * sizeof (struct statfs);
	stats = (struct statfs *) xmalloc (bufsize);
	numsys = getfsstat (stats, bufsize, MNT_WAIT);

	if (numsys < 0) {
	    free (stats);
	    return;
	}
	for (counter = 0; counter < numsys; counter++) {
	    me = zmalloc (sizeof (struct mount_entry));
	    me->me_devname = strdup (stats[counter].f_mntfromname);
	    me->me_mountdir = strdup (stats[counter].f_mntonname);
	    me->me_type = mnt_names[stats[counter].f_type];
	    me->me_dev = -1;    /* Magic; means not known yet. */
	    /* Add to the linked list. */
	    mtail->me_next = me;
	    mtail = me;
	}

	free (stats);
    }
#endif                          /* MOUNTED_GETFSSTAT */

#if defined (MOUNTED_FREAD) || defined (MOUNTED_FREAD_FSTYP)    /* SVR[23].  */
    {
	struct mnttab mnt;
	char *table = "/etc/mnttab";
	FILE *fp;

	fp = fopen (table, "r");
	if (fp == NULL)
	    return;

	while (fread (&mnt, sizeof mnt, 1, fp) > 0) {
	    me = xmalloc (sizeof (struct mount_entry));
#ifdef GETFSTYP                 /* SVR3.  */
	    me->me_devname = strdup (mnt.mt_dev);
#else
	    me->me_devname = xasprintf ("/dev/%s", mnt.mt_dev);
#endif
	    me->me_mountdir = strdup (mnt.mt_filsys);
	    me->me_dev = -1;    /* Magic; means not known yet. */
	    me->me_type = "";
#ifdef GETFSTYP                 /* SVR3.  */
	    if (1) {
		struct statfs fsd;
		char typebuf[FSTYPSZ];

		if (statfs (me->me_mountdir, &fsd, sizeof fsd, 0) != -1
		    && sysfs (GETFSTYP, fsd.f_fstyp, typebuf) != -1)
		    me->me_type = strdup (typebuf);
	    }
#endif
	    me->me_next = NULL;

	    /* Add to the linked list. */
	    mtail->me_next = me;
	    mtail = me;
	}

	if (fclose (fp) == EOF)
	    return;
    }
#endif                          /* MOUNTED_FREAD || MOUNTED_FREAD_FSTYP.  */

#ifdef MOUNTED_GETMNTTBL        /* DolphinOS goes it's own way */
    {
	struct mntent **mnttbl = getmnttbl (), **ent;
	for (ent = mnttbl; *ent; ent++) {
	    me = zmalloc (sizeof (struct mount_entry));
	    me->me_devname = strdup ((*ent)->mt_resource);
	    me->me_mountdir = strdup ((*ent)->mt_directory);
	    me->me_type = strdup ((*ent)->mt_fstype);
	    me->me_dev = -1;    /* Magic; means not known yet. */
	    /* Add to the linked list. */
	    mtail->me_next = me;
	    mtail = me;
	}
	endmnttbl ();
    }
#endif

#ifdef MOUNTED_GETMNTENT2       /* SVR4.  */
    {
	struct mnttab mnt;
	char *table = MNTTAB;
	FILE *fp;
	int ret;

	fp = fopen (table, "r");
	if (fp == NULL)
	    return;

	while ((ret = getmntent (fp, &mnt)) == 0) {
	    me = zmalloc (sizeof (struct mount_entry));
	    me->me_devname = strdup (mnt.mnt_special);
	    me->me_mountdir = strdup (mnt.mnt_mountp);
	    me->me_type = strdup (mnt.mnt_fstype);
	    me->me_dev = -1;    /* Magic; means not known yet. */
	    /* Add to the linked list. */
	    mtail->me_next = me;
	    mtail = me;
	}

	if (ret > 0)
	    return;
	if (fclose (fp) == EOF)
	    return;
    }
#endif                          /* MOUNTED_GETMNTENT2.  */

#ifdef MOUNTED_VMOUNT           /* AIX.  */
    {
	int bufsize;
	char *entries, *thisent;
	struct vmount *vmp;

	/* Ask how many bytes to allocate for the mounted filesystem info.  */
	mntctl (MCTL_QUERY, sizeof bufsize, (struct vmount *) &bufsize);
	entries = xmalloc (bufsize);

	/* Get the list of mounted filesystems.  */
	mntctl (MCTL_QUERY, bufsize, (struct vmount *) entries);

	for (thisent = entries; thisent < entries + bufsize;
	     thisent += vmp->vmt_length) {
	    vmp = (struct vmount *) thisent;
	    me = xmalloc (sizeof (struct mount_entry));
	    if (vmp->vmt_flags & MNT_REMOTE) {
		char *host, *path;

		/* Prepend the remote pathname.  */
		host = thisent + vmp->vmt_data[VMT_HOSTNAME].vmt_off;
		path = thisent + vmp->vmt_data[VMT_OBJECT].vmt_off;
		me->me_devname = xasprintf ("%s:%s", host, path);
	    } else {
		me->me_devname = strdup (thisent +
				      vmp->vmt_data[VMT_OBJECT].vmt_off);
	    }
	    me->me_mountdir = strdup (thisent + vmp->vmt_data[VMT_STUB].vmt_off);
	    me->me_type = strdup (fstype_to_string (vmp->vmt_gfstype));
	    me->me_dev = -1;    /* vmt_fsid might be the info we want.  */
	    me->me_next = NULL;

	    /* Add to the linked list. */
	    mtail->me_next = me;
	    mtail = me;
	}
	free (entries);
    }
#endif                          /* MOUNTED_VMOUNT. */

    /* Free the dummy head. */
    me = mount_list0;
    mount_list = mount_list0->me_next;
    free (me);
}
#endif /* NO_INFOMOUNT */

/* Mouse managing */

/* Events received by clients of this library have their coordinates 0 based */

void init_mouse (void)
{
	if (use_mouse_p == XTERM_MOUSE && !xmouse_flag) {
		/* save old highlight mouse tracking and enable mouse tracking */
		printf("\033[?1001s\033[?1000h");
		fflush (stdout);
		/* turn on */
		xmouse_flag = 1;
		define_sequence (MCKEY_MOUSE, ESC_STR "[M", MCKEY_NOACTION);
	}
}

void shut_mouse (void)
{
	if (use_mouse_p == XTERM_MOUSE && xmouse_flag) {
		/* disable mouse tracking */
		/* Changed the 1 for an 'l' below: */
		/* restore old highlight mouse tracking */
		printf("\033[?1000l\033[?1001r");
		fflush (stdout);
		/* off */
		xmouse_flag = 0;
	}
}

/* Configure box module for the Midnight Commander */

#define PX_OPTION      4
#define PY_OPTION      2

#define RX_OPTION      4
#define RY_OPTION      11

#define BY_OPTION      16
#define OY_OPTION      2

static const char configure_title[] = " Configure options ";
static const char opt_title1[] = " Panel options ";
static const char opt_title2[] = " Pause after run... ";
static const char opt_title3[] = " Other options ";
static const char * const pause_options [3] = {
    "&Never",
    "on dumb &Terminals",
    "alwa&Ys" };

static int configure_callback (struct Dlg_head *h, int Id, int Msg)
{
    switch (Msg) {
    case DLG_DRAW:
	attrset (COLOR_NORMAL);
	dlg_erase (h);
	draw_box (h, 1, 2, h->lines - 2, h->cols - 4);
	draw_box (h, PY_OPTION, PX_OPTION, 8, opt_first_width);
	draw_box (h, RY_OPTION, RX_OPTION, 5, opt_first_width);
	draw_box (h, OY_OPTION, OX, 14, opt_second_width);

	attrset (COLOR_HOT_NORMAL);
	dlg_move (h, 1, (h->cols - strlen(configure_title))/2);
	addstr (configure_title);
	dlg_move (h, OY_OPTION, OX+1);
	addstr (opt_title3);
	dlg_move (h, RY_OPTION, RX_OPTION+1);
	addstr (opt_title2);
	dlg_move (h, PY_OPTION, PX_OPTION+1);
	addstr (opt_title1);
	break;

    case DLG_END:
	break;
    }
    return 0;
}

static void init_configure (void)
{
    int i;

    if (!c_b1) {
	int l1;

	o_check_options[ 0].text = "safe de&Lete";       o_check_options[ 0].variable = &know_not_what_am_i_doing;
	o_check_options[ 1].text = "cd follows lin&Ks";  o_check_options[ 1].variable = &cd_symlinks;
	o_check_options[ 2].text = "advanced cho&Wn";    o_check_options[ 2].variable = &advanced_chfns;
	o_check_options[ 3].text = "l&Ynx-like motion";  o_check_options[ 3].variable = &navigate_with_arrows;
	o_check_options[ 4].text = "ro&Tating dash";     o_check_options[ 4].variable = &nice_rotating_dash;
	o_check_options[ 5].text = "co&Mplete: show all";o_check_options[ 5].variable = &show_all_if_ambiguous;
	o_check_options[ 6].text = "&Use internal view"; o_check_options[ 6].variable = &use_internal_view;
	o_check_options[ 7].text = "use internal ed&It"; o_check_options[ 7].variable = &use_internal_edit;
	o_check_options[ 8].text = "auto m&Enus";        o_check_options[ 8].variable = &auto_menu;
	o_check_options[ 9].text = "&Auto save setup";   o_check_options[ 9].variable = &auto_save_setup;
	o_check_options[10].text = "shell &Patterns";    o_check_options[10].variable = &easy_patterns;
	o_check_options[11].text = "&Verbose operation"; o_check_options[11].variable = &MC_verbose;
	o_check_options[12].text = "&Fast dir reload";   o_check_options[12].variable = &fast_reload;     o_check_options[12].toggle_function = toggle_fast_reload;
	o_check_options[13].text = "mi&X all files";     o_check_options[13].variable = &mix_all_files;   o_check_options[13].toggle_function = toggle_mix_all_files;
	o_check_options[14].text = "&Drop down menus";   o_check_options[14].variable = &drop_menus;
	o_check_options[15].text = "ma&Rk moves down";   o_check_options[15].variable = &mark_moves_down;
	o_check_options[16].text = "show &Hidden files"; o_check_options[16].variable = &show_dot_files;  o_check_options[16].toggle_function = toggle_show_hidden;
	o_check_options[17].text = "show &Backup files"; o_check_options[17].variable = &show_backups;    o_check_options[17].toggle_function = toggle_show_backup;

	/* Similar code is in layout.c (init_layout())  */

	opt_first_width = strlen (opt_title1) + 1;
	for (i = 12; i < 18; i++) {
		l1 = strlen (o_check_options[i].text) + 7;
		if (l1 > opt_first_width)
			opt_first_width = l1;
	}

	i = sizeof(pause_options)/sizeof(char*);
	while (i--) {
		l1 = strlen (pause_options [i]) + 7;
		if (l1 > opt_first_width)
			opt_first_width = l1;
	}

	l1 = strlen (opt_title2) + 1;
	if (l1 > opt_first_width)
		opt_first_width = l1;

	OX = opt_first_width + 5;

	opt_second_width = strlen (opt_title3) + 1;
	for (i = 0; i < 12; i++) {
		l1 = strlen (o_check_options[i].text) + 7;
		if (l1 > opt_second_width)
			opt_second_width = l1;
	}

	l1 = 11 + strlen (ok_button) + strlen (save_button) +
					    strlen (cancel_button);

	i = (opt_first_width + opt_second_width - l1) / 4;
	c_b1 = 5 + i;
	c_b2 = c_b1 + strlen(ok_button) + i + 6;
	c_b3 = c_b2 + strlen(save_button) + i + 4;
    }

    conf_dlg = create_dlg (0, 0, 19, opt_first_width + opt_second_width + 9,
		dialog_colors, configure_callback, "[Options Menu]", DLG_CENTER | DLG_GRID);

    x_set_dialog_title (conf_dlg, "Configure options");

    add_widgetl (conf_dlg,
	button_new (BY_OPTION, c_b3, B_CANCEL, NORMAL_BUTTON, cancel_button, 0, 0),
	XV_WLAY_RIGHTOF);

    add_widgetl (conf_dlg,
	button_new (BY_OPTION, c_b2, B_EXIT, NORMAL_BUTTON, save_button, 0, 0),
	XV_WLAY_RIGHTOF);

    add_widgetl (conf_dlg,
	button_new (BY_OPTION, c_b1, B_ENTER, DEFPUSH_BUTTON, ok_button, 0, 0),
	XV_WLAY_CENTERROW);

#define XTRACT_OPTION(i) *o_check_options[i].variable, o_check_options[i].text

    /* Add all the checkboxes */
    for (i = 0; i < 12; i++) {
	o_check_options [i].widget = check_new (OY_OPTION + (12-i), OX+2, XTRACT_OPTION(i));
	add_widgetl (conf_dlg, o_check_options [i].widget, XV_WLAY_BELOWCLOSE);
    }

    pause_radio = radio_new (RY_OPTION+1, RX_OPTION+2, 3, pause_options, 1);
    pause_radio->sel = pause_after_run;
    add_widgetl (conf_dlg, pause_radio, XV_WLAY_BELOWCLOSE);
    for (i = 0; i < 6; i++) {
	o_check_options [i+12].widget = check_new (PY_OPTION + (6-i), PX_OPTION+2,
						  XTRACT_OPTION(i+12));
	add_widgetl (conf_dlg, o_check_options [i+12].widget, XV_WLAY_BELOWCLOSE);
    }
}


static void configure_box (void)
{
    int result, i;

    init_configure ();
    run_dlg (conf_dlg);

    result = conf_dlg->ret_value;
    if (result == B_ENTER || result == B_EXIT) {
	for (i = 0; o_check_options [i].text; i++)
	    if (o_check_options [i].widget->state & C_CHANGE) {
		if (o_check_options [i].toggle_function)
		    (*o_check_options [i].toggle_function)();
		else
		    *o_check_options [i].variable = !(*o_check_options [i].variable);
	    }
	pause_after_run = pause_radio->sel;
    }

    /* If they pressed the save button */
    if (result == B_EXIT) {
	save_configure ();
	sync_profiles ();
    }

    destroy_dlg (conf_dlg);
}
/* External panelize */

static void do_external_panelize (char *command);

#define UX_PANELIZE              5
#define UY_PANELIZE              2

#define BX_PANELIZE              5
#define BY_PANELIZE              18

#define BUTTONS_PANELIZE         4
#define LABELS_PANELIZE          3
#define B_ADD_PANELIZE           B_USER
#define B_REMOVE_PANELIZE        B_USER + 1


static const char panelize_section[] = "Panelize";

static const struct {
    int ret_cmd, flags, y, x;
    const char *text;
} panelize_but[BUTTONS_PANELIZE] = {
    { B_CANCEL, NORMAL_BUTTON, 0, 53, "&Cancel"},
    { B_ADD_PANELIZE, NORMAL_BUTTON,    0, 28, "&Add new"},
    { B_REMOVE_PANELIZE, NORMAL_BUTTON, 0, 16, "&Remove"},
    { B_ENTER, DEFPUSH_BUTTON, 0,  0, "Pane&lize"},
};

static const char panelize_title[] = " External panelize ";

static void
panelize_refresh (void)
{
    attrset (COLOR_NORMAL);
    dlg_erase (panelize_dlg);

    draw_box (panelize_dlg, 1, 2, panelize_dlg->lines-2, panelize_dlg->cols-4);
    draw_box (panelize_dlg, UY_PANELIZE, UX_PANELIZE, panelize_dlg->lines-10, panelize_dlg->cols-10);

    attrset (COLOR_HOT_NORMAL);
    dlg_move (panelize_dlg, 1, (panelize_dlg->cols - strlen(panelize_title)) / 2);
    addstr (panelize_title);
}

static void
update_command (void)
{
    if (l_panelize->pos != last_listitem) {
	last_listitem = l_panelize->pos;
	assign_text (pname,
	    ((struct panelize *) l_panelize->current->data)->command);
	pname->point = 0;
	update_input (pname, 1);
    }
}

static int
panelize_callback (Dlg_head * h, int Par, int Msg)
{
    switch (Msg) {
    case DLG_DRAW:
	panelize_refresh ();
	break;

    case DLG_POST_KEY:
	/* fall */
    case DLG_INIT:
	attrset (MENU_ENTRY_COLOR);
	update_command ();
	break;
    }
    return 0;
}

static int pa_l_call (void *data)
{
	return listbox_nothing;
}

static void init_panelize (void)
{
    int i, panelize_cols = COLS - 6;
    struct panelize *current = d_panelize;

    last_listitem = 0;

    do_refresh ();

    panelize_dlg = create_dlg (0, 0, 22, panelize_cols, dialog_colors,
			      panelize_callback, "[External panelize]",
			      DLG_CENTER|DLG_GRID);
    x_set_dialog_title (panelize_dlg, "External panelize");

#define XTRACT_PANELIZE(i) BY_PANELIZE+panelize_but[i].y, BX_PANELIZE+panelize_but[i].x, panelize_but[i].ret_cmd, panelize_but[i].flags, panelize_but[i].text, 0, 0

    for (i = 0; i < BUTTONS_PANELIZE; i++)
	add_widgetl (panelize_dlg, button_new (XTRACT_PANELIZE (i)), (i == BUTTONS_PANELIZE - 1) ?
	    XV_WLAY_CENTERROW : XV_WLAY_RIGHTOF);

    pname = input_new (UY_PANELIZE+14, UX_PANELIZE, INPUT_COLOR, panelize_dlg->cols-10, "");
    add_widgetl (panelize_dlg, pname, XV_WLAY_RIGHTOF);

    add_widgetl (panelize_dlg, label_new (UY_PANELIZE+13, UX_PANELIZE, "Command"), XV_WLAY_NEXTROW);

    /* get new listbox */
    l_panelize = listbox_new (UY_PANELIZE + 1, UX_PANELIZE + 1, panelize_dlg->cols-12, 10, 0, pa_l_call);

    while (current) {
	listbox_add_item (l_panelize, 0, 0, current->label, current);
	current = current->next;
    }

    /* add listbox to the dialogs */
    add_widgetl (panelize_dlg, l_panelize, XV_WLAY_EXTENDWIDTH);

    listbox_select_entry (l_panelize,
	listbox_search_text (l_panelize, "Other command"));
}

static void panelize_done (void)
{
    destroy_dlg (panelize_dlg);
    repaint_screen ();
}

static void add2panelize (char *label, char *command)
{
    struct panelize *current, *old;

    old = NULL;
    current = d_panelize;
    while (current && strcmp (current->label, label) <= 0) {
	old = current;
	current = current->next;
    }

    if (old == NULL) {
	d_panelize = xmalloc (sizeof (struct panelize));
	d_panelize->label = label;
	d_panelize->command = command;
	d_panelize->next = current;
    } else {
	struct panelize *new;
	new = xmalloc (sizeof (struct panelize));
	new->label = label;
	new->command = command;
	old->next = new;
	new->next = current;
    }
}

static void add2panelize_cmd (void)
{
    char *label;

    if (pname->buffer && (*pname->buffer)) {
	label = input_dialog (" Add to external panelize ",
				" Enter command label: ", "");
	if (!label)
	    return;
	if (!*label) {
	    free (label);
	    return;
	}

	add2panelize (label, strdup(pname->buffer));
    }
}

static void remove_from_panelize (struct panelize *entry)
{
    if (strcmp (entry->label, "Other command") != 0) {
	if (entry == d_panelize) {
	    d_panelize = d_panelize->next;
	} else {
	    struct panelize *current = d_panelize;
	    while (current && current->next != entry)
		current = current->next;
	    if (current) {
		current->next = entry->next;
	    }
	}

	free (entry->label);
	free (entry->command);
	free (entry);
    }
}

static void external_panelize (void)
{
    char *target = NULL;

    if (!vfs_current_is_local ()) {
	MC_message (1, " Oops... ",
		 " I can't run external panelize while logged on a non local directory ");
	return;
    }

    init_panelize ();

    /* display file info */
    attrset (SELECTED_COLOR);

    run_dlg (panelize_dlg);

    switch (panelize_dlg->ret_value) {
    case B_CANCEL:
	break;

    case B_ADD_PANELIZE:
	add2panelize_cmd ();
	break;

    case B_REMOVE_PANELIZE:
	remove_from_panelize (l_panelize->current->data);
	break;

    case B_ENTER:
	target = pname->buffer;
	if (target != NULL && *target) {
	    char *cmd = strdup (target);
	    destroy_dlg (panelize_dlg);
	    do_external_panelize (cmd);
	    free (cmd);
	    repaint_screen ();
	    return;
	}
	break;
    }

    panelize_done ();
}

static void load_panelize (void)
{
    void *profile_keys;
    char *key, *value;

    profile_keys = profile_init_iterator (panelize_section, profile_name);

    add2panelize (strdup ("Other command"), strdup (""));

    if (!profile_keys) {
	add2panelize (strdup ("Find rejects after patching"), strdup ("find . -name \\*.rej -print"));
	add2panelize (strdup ("Find *.orig after patching"), strdup ("find . -name \\*.orig -print"));
	add2panelize (strdup ("Find SUID and SGID programs"), strdup ("find . \\( \\( -perm -04000 -a -perm +011 \\) -o \\( -perm -02000 -a -perm +01 \\) \\) -print"));
	return;
    }

    while (profile_keys) {
	profile_keys = profile_iterator_next (profile_keys, &key, &value);
	add2panelize (strdup (key), strdup (value));
    }
}

static void save_panelize (void)
{
    struct panelize *current = d_panelize;

    profile_clean_section (panelize_section, profile_name);
    for (; current; current = current->next) {
	if (strcmp (current->label, "Other command"))
	    WritePrivateProfileString (panelize_section,
				       current->label,
				       current->command,
				       profile_name);
    }
    sync_profiles ();
}

static void done_panelize (void)
{
    struct panelize *current = d_panelize;
    struct panelize *next;

    for (; current; current = next) {
	next = current->next;
	free (current->label);
	free (current->command);
	free (current);
    }
}

static void do_external_panelize (char *command)
{
    int Status, link_to_dir, stalled_link;
    int next_free = 0;
    struct stat buf;
    dir_list *list = &cpanel->dir;
    char line [MC_MAXPATHLEN];
    char *name;
    FILE *external;

    open_error_pipe ();
    external = popen (command, "r");
    if (!external) {
	close_error_pipe (1, "Cannot invoke command.");
	return;
    }
    clean_dir (list, cpanel->count);

    /* Clear the counters */
    cpanel->total = cpanel->dirs_marked = cpanel->marked = 0;
    cpanel->has_dir_sizes = 0;
    while (1) {
	clearerr(external);
	if (fgets (line, MC_MAXPATHLEN, external) == NULL) {
	    if (ferror(external) && errno == EINTR)
		continue;
	    else
		break;
	}
	if (line[strlen(line)-1] == '\n')
	    line[strlen(line)-1] = 0;
	if (strlen(line) < 1)
	    continue;
	if (line [0] == '.' && line[1] == '/')
	    name = line + 2;
	else
	    name = line;
	Status = handle_path (list, name, &buf, next_free, &link_to_dir,
	    &stalled_link);
	if (Status == 0)
	    continue;
	if (Status == -1)
	    break;
	list->list [next_free].fnamelen = strlen (name);
	list->list [next_free].fname = strdup (name);
	list->list [next_free].cache = NULL;
	file_mark (cpanel, next_free, 0);
	list->list [next_free].f.link_to_dir = link_to_dir;
	list->list [next_free].f.stalled_link = stalled_link;
	list->list [next_free].buf = buf;
	next_free++;
	if (!(next_free & 32))
	    rotate_dash ();
    }
    if (next_free) {
	cpanel->count = next_free;
	cpanel->is_panelized = 1;
	if (list->list [0].fname [0] == '/') {
	    strcpy (cpanel->cwd, "/");
	    chdir ("/");
	}
    } else {
	cpanel->count = set_zero_dir (list);
    }
    if (pclose (external) < 0)
	MC_message (0, "External panelize", "Pipe close failed");
    close_error_pipe (0, 0);
    try_to_select (cpanel, NULL);
    paint_panel (cpanel);
}

static int is_loaded (const char *FileName, TSecHeader **section)
{
    TProfile *p = Base;

    while (p) {
	if (!strcasecmp (FileName, p->FileName)) {
	    Current = p;
	    *section = p->Section;
	    return 1;
	}
	p = p->link;
    }
    return 0;
}

#define TRANSLATION_CHAR '\200'

static char *str_untranslate_newline_dup (const char *s)
{
    int l = 0;
    char *p, *q;
    const char *ss = s;
    while (*ss) {
	l++;
	if(*ss == '\n' || *ss == TRANSLATION_CHAR)
		l++;
	ss++;
    }
    q = p = malloc (l + 1);
    if (!q)
	return 0;
    for (;;) {
	switch (*s) {
	case '\n':
	    *p++ = TRANSLATION_CHAR;
	    *p++ = 'n';
	    break;
	case TRANSLATION_CHAR:
	    if (s[1] == 'n' || s[1] == TRANSLATION_CHAR)
		*p++ = TRANSLATION_CHAR;
	    *p++ = TRANSLATION_CHAR;
	    break;
	case '\0':
	    *p = '\0';
	    return q;
	    break;
	default:
	    *p++ = *s;
	}
	s++;
    }
    return 0;                   /* not reached */
}

static char *str_translate_newline_dup (const char *s)
{
    char *p, *q;
    q = p = strdup (s);
    if (!q)
	return 0;
    while (*s) {
	if (*s == TRANSLATION_CHAR) {
	    switch (*(++s)) {
	    case 'n':
		*p++ = '\n';
		break;
	    case TRANSLATION_CHAR:
		*p++ = TRANSLATION_CHAR;
		break;
	    case '\0':
		*p++ = TRANSLATION_CHAR;
		*p++ = '\0';
		return q;
	    default:
		*p++ = TRANSLATION_CHAR;
		*p++ = *s;
	    }
	} else {
	    *p++ = *s;
	}
	s++;
    }
    *p = '\0';
    return q;                   /* not reached */
}

static TSecHeader *load (const char *file)
{
    FILE *f;
    int state;
    TSecHeader *SecHeader = 0;
    char CharBuffer [STRSIZE];
    char *next = "";            /* Not needed */
    int c;

    if ((f = fopen (file, "r"))==NULL)
	return NULL;

    state = FirstBrace;
    while ((c = getc (f)) != EOF) {
	if (c == '\r')          /* Ignore Carriage Return */
	    continue;

	switch (state) {

	case OnSecHeader:
	    if (c == ']' || overflow) {
		*next = '\0';
		next = CharBuffer;
		SecHeader->AppName = strdup (CharBuffer);
		state = IgnoreToEOL;
	    } else
		*next++ = c;
	    break;

	case IgnoreToEOL:
	    if (c == '\n') {
		state = KeyDef;
		next = CharBuffer;
	    }
	    break;

	case FirstBrace:
	case KeyDef:
	case KeyDefOnKey:
	    if (c == '[') {
		TSecHeader *temp;

		temp = SecHeader;
		SecHeader = x_malloc (sizeof (TSecHeader), "KeyDef");
		SecHeader->link = temp;
		SecHeader->Keys = 0;
		state = OnSecHeader;
		next = CharBuffer;
		break;
	    }
	    if (state == FirstBrace) /* On first pass, don't allow dangling keys */
		break;

	    if ((c == ' ' && state != KeyDefOnKey) || c == '\t')
		break;

	    if (c == '\n' || overflow) /* Abort Definition */
		next = CharBuffer;

	    if (c == '=' || overflow) {
		TKeys *temp;

		temp = SecHeader->Keys;
		*next = '\0';
		SecHeader->Keys = x_malloc (sizeof (TKeys), "KD2");
		SecHeader->Keys->link = temp;
		SecHeader->Keys->KeyName = strdup (CharBuffer);
		state = KeyValue;
		next = CharBuffer;
	    } else {
		*next++ = c;
		state = KeyDefOnKey;
	    }
	    break;

	case KeyValue:
	    if (overflow || c == '\n') {
		*next = '\0';
		SecHeader->Keys->Value = str_translate_newline_dup (CharBuffer);
		state = c == '\n' ? KeyDef : IgnoreToEOL;
		next = CharBuffer;
#if 0
		printf ("[%s] (%s)=%s\n", SecHeader->AppName,
			SecHeader->Keys->KeyName, SecHeader->Keys->Value);
#endif
	    } else
		*next++ = c;
	    break;

	} /* switch */

    } /* while ((c = getc (f)) != EOF) */
    if (c == EOF && state == KeyValue) {
	*next = '\0';
	SecHeader->Keys->Value = str_translate_newline_dup (CharBuffer);
    }
    fclose (f);
    return SecHeader;
}

static void new_key (TSecHeader *section, const char *KeyName, const char *Value)
{
    TKeys *key;

    key = x_malloc (sizeof (TKeys), "new_key");
    key->KeyName = strdup (KeyName);
    key->Value   = strdup (Value);
    key->link = section->Keys;
    section->Keys = key;
}

static const char *GetSetProfileChar (int set, const char *AppName, const char *KeyName,
			 const char *Default, const char *FileName)
{

    TProfile   *New;
    TSecHeader *section;
    TKeys      *key;

    if (!is_loaded (FileName, &section)) {
	New = (TProfile *) x_malloc (sizeof (TProfile), "GetSetProfile");
	New->link = Base;
	New->FileName = strdup (FileName);
	New->Section = load (FileName);
	Base = New;
	section = New->Section;
	Current = New;
    }

    /* Start search */
    for (; section; section = section->link) {
	if (section->AppName == 0 || strcasecmp (section->AppName, AppName))
	    continue;
	for (key = section->Keys; key; key = key->link) {
	    if (strcasecmp (key->KeyName, KeyName))
		continue;
	    if (set) {
		free (key->Value);
		key->Value = strdup (Default);
	    }
	    return key->Value;
	}
	/* If getting the information, then don't write the information
	   to the INI file, need to run a couple of tests with windog */
	/* No key found */
	if (set) {
	    new_key (section, KeyName, Default);
	    return 0;
	}
    }

    /* Non existent section */
    if (set && Default) {
	section = (TSecHeader *) x_malloc (sizeof (TSecHeader), "GSP3");
	section->AppName = strdup (AppName);
	section->Keys = 0;
	new_key (section, KeyName, Default);
	section->link = Current->Section;
	Current->Section = section;
    }
    return Default;
}

static short GetSetProfile (int set, const char * AppName, const char * KeyName, const char * Default,
		     char * ReturnedString, short Size, const char * FileName)

{
    const char *s;

    s = GetSetProfileChar (set, AppName, KeyName, Default, FileName);
    if (!set) {
	ReturnedString [Size-1] = 0;
	strncpy (ReturnedString, s, Size-1);
   }
    return 1;
}

static short GetPrivateProfileString (const char * AppName, const char * KeyName,
			       char * Default, char * ReturnedString,
			       short Size, const char * FileName)
{
    return (GetSetProfile (0, AppName, KeyName, Default, ReturnedString, Size, FileName));
}

static const char *get_profile_string (const char *AppName, const char *KeyName, const char *Default,
			  const char *FileName)
{
    return GetSetProfileChar (0, AppName, KeyName, Default, FileName);
}

static int GetPrivateProfileInt (const char * AppName, const char * KeyName, int Default,
			   const char * File)
{
    char IntBuf [15];
    char buf [15];

    /* Check the exact semantic with the SDK */
    GetPrivateProfileString (AppName, KeyName, sprintf_pd (buf, Default), IntBuf, 15, File);
    if (!strcasecmp (IntBuf, "true"))
	return 1;
    if (!strcasecmp (IntBuf, "yes"))
	return 1;
    return atoi (IntBuf);
}

static int WritePrivateProfileString (const char * AppName, const char * KeyName, const char * String,
				const char * FileName)
{
    return GetSetProfile (1, AppName, KeyName, String, "", 0, FileName);
}

static void dump_keys (FILE * profile, const TKeys * p)
{
    char *t;
    if (!p)
	return;
    dump_keys (profile, p->link);
    t = str_untranslate_newline_dup (p->Value);
    fprintf (profile, "%s=%s\n", p->KeyName, t);
    free (t);
}

static void dump_sections (FILE *profile, const TSecHeader *p)
{
    if (!p)
	return;
    dump_sections (profile, p->link);
    if (p->AppName [0]) {
	fprintf (profile, "\n[%s]\n", p->AppName);
	dump_keys (profile, p->Keys);
    }
}

static void dump_profile (const TProfile *p)
{
    FILE *profile;

    if (!p)
	return;
    dump_profile (p->link);
    /* .ado: p->FileName can be empty, it's better to jump over */
    if (p->FileName[0] != (char) 0)
      if ((profile = fopen (p->FileName, "w")) != NULL) {
	dump_sections (profile, p->Section);
	fclose (profile);
      }
}

/*
 * Must be called at the end of wine run
*/

static void sync_profiles (void)
{
    dump_profile (Base);
}

static void free_keys (TKeys *p)
{
    if (!p)
	return;
    free_keys (p->link);
    free (p->KeyName);
    free (p->Value);
    free (p);
}

static void free_sections (TSecHeader *p)
{
    if (!p)
	return;
    free_sections (p->link);
    free_keys (p->Keys);
    free (p->AppName);
    p->link = 0;
    p->Keys = 0;
    free (p);
}

static void free_profile (TProfile *p)
{
    if (!p)
	return;
    free_profile (p->link);
    free_sections (p->Section);
    free (p->FileName);
    free (p);
}

static void free_profile_name (const char *s)
{
    TProfile *p;

    if (!s)
	return;

    for (p = Base; p; p = p->link) {
	if (strcmp (s, p->FileName) == 0) {
	    free_sections (p->Section);
	    p->Section = 0;
	    p->FileName [0] = 0;
	    return;
	}
    }
}

static void free_profiles (void)
{
    free_profile (Base);
}

static void *profile_init_iterator (const char *appname, const char *file)
{
    TProfile   *New;
    TSecHeader *section;

    if (!is_loaded (file, &section)) {
	New = x_malloc (sizeof (TProfile), "GetSetProfile");
	New->link = Base;
	New->FileName = strdup (file);
	New->Section = load (file);
	Base = New;
	section = New->Section;
	Current = New;
    }
    for (; section; section = section->link) {
	if (strcasecmp (section->AppName, appname))
	    continue;
	return section->Keys;
    }
    return 0;
}

static void *profile_iterator_next (void *s, char **key, char **value)
{
    TKeys *Keys = (TKeys *) s;

    if (Keys) {
	*key   = Keys->KeyName;
	*value = Keys->Value;
	Keys   = Keys->link;
    }
    return Keys;
}

static void profile_clean_section (const char *appname, const char *file)
{
    TSecHeader *section;

    /* We assume the user has called one of the other initialization funcs */
    if (!is_loaded (file, &section)) {
	fprintf (stderr,"Warning: profile_clean_section called before init\n");
	return;
    }
    /* We only disable the section, so it will still be freed, but it */
    /* won't be find by further walks of the structure */

    for (; section; section = section->link) {
	if (strcasecmp (section->AppName, appname))
	    continue;
	section->AppName [0] = 0;
    }
}

static int profile_has_section (const char *section_name, const char *profile)
{
    TSecHeader *section;

    /* We assume the user has called one of the other initialization funcs */
    if (!is_loaded (profile, &section)) {
	return 0;
    }
    for (; section; section = section->link) {
	if (strcasecmp (section->AppName, section_name))
	    continue;
	return 1;
    }
    return 0;
}

static void profile_forget_profile (const char *file)
{
    TProfile *p;

    for (p = Base; p; p = p->link) {
	if (strcasecmp (file, p->FileName))
	    continue;
	p->FileName [0] = 0;
    }
}



/* my own wierd protocol base 16 - paul */
static int rxvt_getc (void)
{
    int r;
    unsigned char c;
    while (read (0, &c, 1) != 1);
    if (c == '\n')
	return -1;
    r = (c - 'A') * 16;
    while (read (0, &c, 1) != 1);
    r += (c - 'A');
    return r;
}

static int anything_ready (void)
{
    fd_set fds;
    struct timeval tv;

    FD_ZERO (&fds);
    FD_SET (0, &fds);
    tv.tv_sec = 0;
    tv.tv_usec = 0;
    return select (1, &fds, 0, 0, &tv);
}

static void show_rxvt_contents (int starty, unsigned char y1, unsigned char y2)
{
    unsigned char *k;
    int bytes, i, j, cols = 0;
    y1 += (keybar_visible != 0);        /* i don't knwo why we need this - paul */
    y2 += (keybar_visible != 0);
    while (anything_ready ())
	getch ();

/* my own wierd protocol base 26 - paul */
    printf ("\033CL%c%c%c%c\n",
	    (y1 / 26) + 'A', (y1 % 26) + 'A',
	    (y2 / 26) + 'A', (y2 % 26) + 'A');

    bytes = (y2 - y1) * (COLS + 1) + 1;         /* *should* be the number of bytes read */
    j = 0;
    k = xmalloc (bytes);
    for (;;) {
	int c;
	c = rxvt_getc ();
	if (c < 0)
	    break;
	if (j < bytes)
	    k[j++] = c;
	for (cols = 1;;cols++) {
	    c = rxvt_getc ();
	    if (c < 0)
		break;
	    if (j < bytes)
		k[j++] = c;
	}
    }
    for (i = 0; i < j; i++) {
	if ((i % cols) == 0)
	    move (starty + (i / cols), 0);
	addch (is_printable (k[i]) ? k[i] : ' ');
    }
    free (k);
}

/* Panel managing. */

static int panel_event (Gpm_Event *event, WPanel *panel);

#ifndef PORT_HAS_PANEL_ADJUST_TOP_FILE
#   define x_adjust_top_file(p)
#endif

#ifndef PORT_HAS_PANEL_RESET_SORT_LABELS
#   define x_reset_sort_labels(x)
#endif

/* This macro extracts the number of available lines in a panel */
#ifndef PORT_HAS_LLINES
#define llines(p) (p->widget.lines-3 - (show_mini_info ? 2 : 0))
#else
#define llines(p) (p->widget.lines)
#endif

#ifdef PORT_NOT_FOCUS_SELECT_ITEM
#   define focus_select_item(x)
#else
#   define focus_select_item(x) select_item(x)
#endif

#ifdef PORT_NOT_UNFOCUS_UNSELECT_ITEM
#    define unfocus_unselect_item(x)
#else
#    define unfocus_unselect_item(x) unselect_item(x)
#endif

static void
set_colors (WPanel *panel)
{
    standend ();
    if (hascolors)
	attrset (NORMAL_COLOR);
}

#ifndef ICONS_PER_ROW
#    define ICONS_PER_ROW(x) 1
#endif

/* Delete format string, it is a linked list */
static void
delete_format (format_e *format)
{
    format_e *next;

    while (format) {
	next = format->next;
	free (format);
	format = next;
     }
}

/* Return the Index of permission triplet */
/* This gives abilitiy to determine colored user priveleges */
static inline int
get_user_rights (const struct stat *buf)
{
    const user_in_groups *cug;

    if (buf->st_uid == current_user_uid || current_user_uid == 0)
       return 0;

    for (cug = current_user_gid; cug; cug = cug->next)
       if (cug->gid == buf->st_gid) return 1;

    return 2;
}

/* This code relies on the default justification!!! */
static void
add_permission_string (char *dest, int width, file_entry *fe, int attr, int color, int flg_is_octal)
{
    int i, r, l;

    l = get_user_rights (&fe->buf);

    if (flg_is_octal) {
	/* Place of the access bit in octal mode */
	l = width + l - 3;
	r = l + 1;
    } else {
	/* The same to the triplet in string mode */
	l = l * 3 + 1;
	r = l + 3;
    }

    for(i = 0; i < width; i++) {
	if (i >= l && i < r) {
	    if (attr == SELECTED || attr == MARKED_SELECTED)
		attrset (MARKED_SELECTED_COLOR);
	    else
		attrset (MARKED_COLOR);
	} else
	    attrset (color);

	addch (dest[i]);
    }
}

static int
file_entry_color (file_entry *fe)
{
    if (filetype_mode) {
	if (S_ISDIR (fe->buf.st_mode))
	    return (DIRECTORY_COLOR);
	else if (S_ISLNK (fe->buf.st_mode)) {
	    if (fe->f.link_to_dir)
		return (DIRECTORY_COLOR);
	    else if (fe->f.stalled_link)
		return (STALLED_COLOR);
	    else
		return (LINK_COLOR);
	} else if (S_ISSOCK (fe->buf.st_mode))
	    return (SPECIAL_COLOR);
	else if (S_ISCHR (fe->buf.st_mode))
	    return (DEVICE_COLOR);
	else if (S_ISBLK (fe->buf.st_mode))
	    return (DEVICE_COLOR);
	else if (S_ISFIFO (fe->buf.st_mode))
	    return (SPECIAL_COLOR);
	else if (is_exe (fe->buf.st_mode))
	    return (EXECUTABLE_COLOR);
	else if (fe->fname && (strcmp (fe->fname, "core") == 0))
	    return (CORE_COLOR);
    }
    return (NORMAL_COLOR);
}

/* This functions return a string representation of a file entry */
static char *
string_file_type (file_entry *fe, int len)
{
    if (S_ISDIR (fe->buf.st_mode))
	string_file_type_buffer [0] = '/';
    else if (S_ISLNK (fe->buf.st_mode)) {
	if (fe->f.link_to_dir)
	    string_file_type_buffer [0] = '~';
	else if (fe->f.stalled_link)
	    string_file_type_buffer [0] = '!';
	else
	    string_file_type_buffer [0] = '@';
    } else if (S_ISSOCK (fe->buf.st_mode))
	string_file_type_buffer [0] = '=';
    else if (S_ISCHR (fe->buf.st_mode))
	string_file_type_buffer [0] = '-';
    else if (S_ISBLK (fe->buf.st_mode))
	string_file_type_buffer [0] = '+';
    else if (S_ISFIFO (fe->buf.st_mode))
	string_file_type_buffer [0] = '|';
    else if (is_exe (fe->buf.st_mode))
	string_file_type_buffer [0] = '*';
    else
	string_file_type_buffer [0] = ' ';
    return string_file_type_buffer;
}

static char *
string_file_size_brief (file_entry *fe, int len)
{
    if (S_ISDIR (fe->buf.st_mode)) {
       return strcmp (fe->fname, "..") ? "SUB-DIR" : "UP--DIR";
    }

    return string_file_size (fe, len);
}

static char *
string_file_permission (file_entry *fe, int len)
{
    return (char *)bb_mode_string (fe->buf.st_mode, bb_mode_string_buf);
}

static char *
string_file_nlinks (file_entry *fe, int len)
{
    sprintf (string_file_nlinks_buffer, "%16u", (unsigned int)fe->buf.st_nlink);
    return string_file_nlinks_buffer;
}

static char *
string_file_owner (file_entry *fe, int len)
{
    return get_owner (fe->buf.st_uid);
}

static char *
string_file_group (file_entry *fe, int len)
{
    return get_group (fe->buf.st_gid);
}

static char *
string_file_size (file_entry *fe, int len)
{
    int i;

    if (S_ISBLK (fe->buf.st_mode) || S_ISCHR (fe->buf.st_mode))
	sprintf (string_file_size_buffer, "%3d,%3d", (int) (fe->buf.st_rdev >> 8),
	    (int) (fe->buf.st_rdev & 0xff));
    else
    {
	sprintf (string_file_size_buffer, "%" P_ST_SIZE "d", fe->buf.st_size);
	if (len && (i = strlen (string_file_size_buffer)) > len) {
	    if (i - 2 > len) {
		if (i - 5 > len)
		    sprintf (string_file_size_buffer, "%dG", (int) ((fe->buf.st_size) >> 30));
		else
		    sprintf (string_file_size_buffer, "%dM", (int) ((fe->buf.st_size) >> 20));
	    } else
		sprintf (string_file_size_buffer, "%luK", (unsigned long) ((fe->buf.st_size) >> 10));
	}
    }
    return string_file_size_buffer;
}

static char *
string_file_mtime (file_entry *fe, int len)
{
    return file_date (fe->buf.st_mtime);
}

static char *
string_file_atime (file_entry *fe, int len)
{
    return file_date (fe->buf.st_atime);
}

static char *
string_file_ctime (file_entry *fe, int len)
{
    return file_date (fe->buf.st_ctime);
}

static char *
string_file_name (file_entry *fe, int len)
{
    if (len)
	return name_trunc (fe->fname, len);
    else
	return fe->fname;
}

static char *
string_space (file_entry *fe, int len)
{
    return " ";
}

static char *
string_dot (file_entry *fe, int len)
{
    return ".";
}

static char *
string_marked (file_entry *fe, int len)
{
    return fe->f.marked ? "*" : " ";
}


static char *
string_file_perm_octal (file_entry *fe, int len)
{
    sprintf (string_file_perm_octal_buffer, "0%06o", fe->buf.st_mode);
    return string_file_perm_octal_buffer;
}

static char *
string_inode (file_entry *fe, int len)
{
    sprintf (string_inode_buffer, "%" P_BB64T "d", (bb64_t) fe->buf.st_ino);
    return string_inode_buffer;
}

static char *
string_file_ngid (file_entry *fe, int len)
{
    return sprintf_pd (string_file_ngid_buffer, fe->buf.st_gid);
}

static char *
string_file_nuid (file_entry *fe, int len)
{
    return sprintf_pd (string_file_nuid_buffer, fe->buf.st_uid);
}

#define GT 1

static const struct {
    const char *id;
    int  min_size;
    int  expands;
    int  default_just;
    const char *title;
    int  use_in_gui;
    char *(*string_fn)(file_entry *, int);
    sortfn *sort_routine;
} formats [] = {
{ "name",  12, 1, J_LEFT,  "Name",       1, string_file_name,       sort_name },
{ "size",  7,  0, J_RIGHT, "Size",       1, string_file_size,       sort_size },
{ "type",  GT, 0, J_LEFT,  "",           1, string_file_type,       sort_type },
{ "mtime", 12, 0, J_RIGHT, "MTime",      1, string_file_mtime,      sort_time },
{ "bsize", 7,  0, J_RIGHT, "Size",       1, string_file_size_brief, sort_size },
{ "perm",  10, 0, J_LEFT,  "Permission", 1, string_file_permission, NULL },
{ "mode",  6,  0, J_RIGHT, "Perm",       1, string_file_perm_octal, NULL },
{ "|",     1,  0, J_RIGHT, "|",          0, 0,                      NULL },
{ "nlink", 2,  0, J_RIGHT, "Nl",         1, string_file_nlinks,     sort_links },
{ "ngid",  5,  0, J_RIGHT, "GID",        1, string_file_ngid,       sort_ngid },
{ "nuid",  5,  0, J_RIGHT, "UID",        1, string_file_nuid,       sort_nuid },
{ "owner", 8,  0, J_LEFT,  "Owner",      1, string_file_owner,      sort_owner },
{ "group", 8,  0, J_LEFT,  "Group",      1, string_file_group,      sort_group },
{ "atime", 12, 0, J_RIGHT, "ATime",      1, string_file_atime,      sort_atime },
{ "ctime", 12, 0, J_RIGHT, "CTime",      1, string_file_ctime,      sort_ctime },
{ "space", 1,  0, J_RIGHT, " ",          0, string_space,           NULL },
{ "dot",   1,  0, J_RIGHT, " ",          0, string_dot,             NULL },
{ "mark",  1,  0, J_RIGHT, " ",          1, string_marked,          NULL },
{ "inode", 5,  0, J_RIGHT, "Inode",      1, string_inode,           sort_inode },
};

static char *
to_buffer (char *dest, int just_mode, int len, const char *txt)
{
    int txtlen = strlen (txt);
    int still;

    if (txtlen > len) {
	if (just_mode != J_LEFT)
	    txt += txtlen - len;
	txtlen = len;
    }
    still = len - txtlen;
    if (just_mode == J_LEFT) {
	strcpy (dest, txt);
	dest += txtlen;
	while (still--)
	    *dest++ = ' ';
	*dest = 0;
    } else {
	while (still--)
	    *dest++ = ' ';
	strcpy (dest, txt);
	dest += txtlen;
    }
    return dest;
}

int
file_compute_color (int attr, file_entry *fe)
{
    int color;

    switch (attr) {
    case SELECTED:
	color = SELECTED_COLOR;
	break;
    case MARKED:
	color = MARKED_COLOR;
	break;
    case MARKED_SELECTED:
	color = MARKED_SELECTED_COLOR;
	break;
    case STATUS:
	color = NORMAL_COLOR;
	break;
    case NORMAL:
    default:
	color = file_entry_color(fe);
    }
    return color;
}

/* Formats the file number file_index of panel in the buffer dest */
void
format_file (char *dest, WPanel *panel, int file_index, int width, int attr, int isstatus)
{
    int      color, length, empty_line;
    char     *txt;
    char     *old_pos;
    char     *cdest = dest;
    format_e *format, *home;
    file_entry *fe;

    length     = 0;
    empty_line = (file_index >= panel->count);
    home       = (isstatus) ? panel->status_format : panel->format;
    fe         = &panel->dir.list [file_index];

    if (!empty_line)
	color = file_compute_color (attr, fe);
    else
	color = NORMAL_COLOR;
    for (format = home; format; format = format->next) {
	if (format->string_fn) {
	    int len;

	    if (empty_line)
		txt = " ";
	    else
		txt = (*format->string_fn)(fe, format->field_len);

	    old_pos = cdest;

	    len = format->field_len;
	    if (len + length > width)
		len = width - length;
	    cdest = to_buffer (cdest, format->just_mode, len, txt);
	    length += len;

	    /* What shall we do? Will we color each line according to
	     * the file type? Any suggestions to mc@timur.kazan.su
	     */

	    attrset (color);

	    if (permission_mode && !strcmp(format->id, "perm"))
		add_permission_string (old_pos, format->field_len, fe, attr, color, 0);
	    else if (permission_mode && !strcmp(format->id, "mode"))
		add_permission_string (old_pos, format->field_len, fe, attr, color, 1);
	    else
		addstr (old_pos);
	} else {
	    /* I'm preffer the view without this 3 lines, try to kill it :-) */
	    if (attr == SELECTED || attr == MARKED_SELECTED)
		attrset (SELECTED_COLOR);
	    else
		attrset (NORMAL_COLOR);
	    one_vline ();
	    length++;
	}
    }

    if (length < width) {
	int still = width - length;
	while (still--)
	    addch (' ');
    }
}

static void
repaint_file (WPanel *panel, int file_index, int mv, int attr, int isstatus)
{
    int    second_column = 0;
    int    width, offset;
    char   buffer [255];

    offset = 0;
    if (!isstatus && panel->split) {

	second_column = (file_index - panel->top_file) / llines (panel);
	width = (panel->widget.cols - 2)/2 - 1;

	if (second_column) {
	    offset = 1 + width;
	    width = (panel->widget.cols-2) - (panel->widget.cols-2)/2 - 1;
	}
    } else
	width = (panel->widget.cols - 2);

    if (mv) {
	if (!isstatus && panel->split) {
	    widget_move (&panel->widget,
			 (file_index - panel->top_file) %
			 llines (panel) + 2,
			 (offset + 1));
	} else
	    widget_move (&panel->widget, file_index - panel->top_file + 2, 1);
    }

    format_file (buffer, panel, file_index, width, attr, isstatus);

    if (!isstatus && panel->split) {
	if (second_column)
	    addch (' ');
	else {
	    attrset (NORMAL_COLOR);
	    one_vline ();
	}
    }
}

#ifndef PORT_HAS_DISPLAY_MINI_INFO
void
display_mini_info (WPanel *panel)
{
    if (!show_mini_info)
	return;

    widget_move (&panel->widget, llines (panel)+3, 1);

    if (panel->searching) {
	attrset (INPUT_COLOR);
	printw ("/%-*s", panel->widget.cols-3, panel->search_buffer);
	attrset (NORMAL_COLOR);
	return;
    }

    /* Status displays total marked size */
    if (panel->marked) {
	char buffer [100];
	char *p;

	attrset (MARKED_COLOR);
	printw  ("%*s", panel->widget.cols-2, " ");
	widget_move (&panel->widget, llines (panel)+3, 1);
	sprintf (buffer, "  %s bytes in %d file%s",
		 size_trunc_sep (panel->total), panel->marked,
		 panel->marked == 1 ? "" : "s");
	p = buffer;
	if (strlen (buffer) > panel->widget.cols-4) {
	    buffer [panel->widget.cols-4] = 0;
	    p += 2;
	}
	printw ("%-*s", panel->widget.cols-2, p);
	return;
    }

    /* Status resolves links and show them */
    set_colors (panel);
    if (S_ISLNK (panel->dir.list [panel->selected].buf.st_mode)) {
	char *link_c, link_target [MC_MAXPATHLEN];
	int  len;

	link_c = concat_path_file (panel->cwd, panel->dir.list [panel->selected].fname);
	len = mc_readlink (link_c, link_target, MC_MAXPATHLEN);
	free (link_c);
	if (len > 0) {
	    link_target[len] = 0;
	    printw ("-> %-*s", panel->widget.cols - 5,
		     name_trunc (link_target, panel->widget.cols - 5));
	} else
	    addstr ("<readlink failed>");
	return;
    }
    /* Default behaviour */
    repaint_file (panel, panel->selected, 0, STATUS, 1);
    return;
}
#endif

static void
paint_dir (WPanel *panel)
{
    int i;
    int color;                  /* Color value of the line */
    int items;                  /* Number of items */

    items = llines (panel) * (panel->split ? 2 : 1);

    for (i = 0; i < items; i++) {
	if (i+panel->top_file >= panel->count)
	    color = 0;
	else {
	    color = 2 * (panel->dir.list [i+panel->top_file].f.marked);
	    color += (panel->selected==i+panel->top_file && panel->active);
	}
	repaint_file (panel, i+panel->top_file, 1, color, 0);
    }
    standend ();
    panel->dirty = 0;
}

static void
mini_info_separator (WPanel *panel)
{
    if (!show_mini_info)
	return;

    standend ();
    widget_move (&panel->widget, llines (panel)+2, 1);
    attrset (NORMAL_COLOR);
    hline (ACS_HLINE, panel->widget.cols-2);
}

static void trim_to_dots (char *s, int len)
{
    int l = strlen (s);

    if(len < 0)
	len = 0;
    if (l > len) {
	memmove(s, s + l - len, len + 1);
	for(l = 0; l < 3 && len > 0; l++) {
		s[l] = '.';
		len--;
	}
    }
}

static void
show_dir (WPanel *panel)
{
    char newdir [MC_MAXPATHLEN];

    set_colors (panel);
    draw_double_box (panel->widget.parent,
		    panel->widget.y,    panel->widget.x,
		    panel->widget.lines, panel->widget.cols);

    if (show_mini_info) {
	    SLsmg_draw_object (panel->widget.y + llines (panel) + 2,
		panel->widget.x, SLSMG_LTEE_CHAR);
	    SLsmg_draw_object (panel->widget.y + llines (panel) + 2,
		panel->widget.x + panel->widget.cols - 1, SLSMG_RTEE_CHAR);
    }

    if (panel->active)
	attrset (REVERSE_COLOR);

    widget_move (&panel->widget, 0, 3);

    trim_to_dots (strip_home_and_password (panel->cwd, newdir), panel->widget.cols-7);
    addstr (newdir);
    widget_move (&panel->widget, 0, 1);
    addstr ("<");
    widget_move (&panel->widget, 0, panel->widget.cols-2);
    addstr (">");
    widget_move (&panel->widget, 0, panel->widget.cols-3);
    addstr ("v");

    if (panel->active)
	standend ();
}

/* To be used only by long_frame and full_frame to adjust top_file */
static void
adjust_top_file (WPanel *panel)
{
    int old_top = panel->top_file;

    if (panel->selected - old_top > llines (panel))
	panel->top_file = panel->selected;
    if (old_top - panel->count > llines (panel))
	panel->top_file = panel->count - llines (panel);

}

/* Repaints the information that changes after a command */
static void
panel_update_contents (WPanel *panel)
{
    show_dir (panel);
    paint_dir (panel);
    display_mini_info (panel);
}

static void
paint_panel (WPanel *panel)
{
    paint_frame (panel);
    panel_update_contents (panel);
    mini_info_separator (panel);
}

static stdcall2 void
Xtry_to_select (WPanel *panel, char *name)
{
    int i;
    char *subdir;

    if (!name) {
	panel->selected = 0;
	panel->top_file = 0;
	x_adjust_top_file (panel);
	return;
    }

    /* We only want the last component of the directory */
    for (subdir = name + strlen (name) - 1; subdir >= name; subdir--) {
	if (*subdir == '/') {
	    subdir++;
	    break;
	}
    }
    if (subdir < name)
	subdir = name;

    /* Search that subdirectory, if found select it */
    for (i = 0; i < panel->count; i++) {
	if (strcmp (subdir, panel->dir.list [i].fname))
	    continue;

	if (i != panel->selected) {
	    panel->selected = i;
	    panel->top_file = panel->selected - (panel->widget.lines-2)/2;
	    if (panel->top_file < 0)
		panel->top_file = 0;
	    x_adjust_top_file (panel);
	}
	return;
    }

    /* Try to select a file near the file that is missing */
    if (panel->selected >= panel->count) {
	panel->selected = panel->count-1;
	panel->top_file = panel->selected - (panel->widget.lines)/2;
	if (panel->top_file < 0)
	    panel->top_file = 0;
	x_adjust_top_file (panel);
    } else
	return;
}

static void
panel_update_cols (Widget *widget, int frame_size)
{
    int cols, origin;

    if (horizontal_split) {
	widget->cols = COLS;
	return;
    }

    if (frame_size == frame_full) {
	cols = COLS;
	origin = 0;
    } else {
	if (widget == get_panel_widget (0)) {
	    cols   = first_panel_size;
	    origin = 0;
	} else {
	    cols   = COLS-first_panel_size;
	    origin = first_panel_size;
	}
    }

    widget->cols = cols;
    widget->x = origin;
}

/* Ugly hack to allow panel_save_setup to work as a place holder for */
/* default panel values */

static char *
panel_save_name (WPanel *panel)
{
    /* If the program is shuting down */
    if ((midnight_shutdown && auto_save_setup) || saving_setup)
	return xasprintf ("%s", panel->panel_name);
    else
	return xasprintf ("Temporal:%s", panel->panel_name);
}

static void
panel_destroy (WPanel *p)
{
    int i;

    char *name = panel_save_name (p);

    panel_save_setup (p, name);
    clean_dir (&p->dir, p->count);

/* save and clean history */
    if (p->dir_history) {
	Hist *current, *old;
	current = p->dir_history;
	while (current->next)
	    current = current->next;
	while (current) {
	    old = current;
	    current = current->prev;
	    free (old->text);
	    free (old);
	}
    }
    free (p->hist_name);

    delete_format (p->format);
    delete_format (p->status_format);

    free (p->user_format);
    for (i = 0; i < LIST_TYPES; i++)
	free (p->user_status_format [i]);
    free (p->dir.list);
    free (p->panel_name);
    free (name);
}

static void
panel_format_modified (WPanel *panel)
{
    panel->format_modified = 1;
    x_reset_sort_labels (panel);
}

static int
panel_callback (Dlg_head *h, WPanel *panel, int msg, int par);

/* Panel creation */
/* The parameter specifies the name of the panel for setup retieving */
WPanel *panel_new (const char *panel_name)
{
    WPanel *panel;
    char *section;
    int i, err;

    panel = zmalloc (sizeof (WPanel));

    /* No know sizes of the panel at startup */
    init_widget (&panel->widget, 0, 0, 0, 0, (callback_fn)
		 panel_callback, (destroy_fn) panel_destroy,
		 (mouse_h) panel_event);

    /* We do not want the cursor */
    widget_want_cursor (panel->widget, 0);

    mc_get_current_wd (panel->cwd, sizeof (panel->cwd)-2);
    strcpy (panel->lwd, ".");

    panel->hist_name = xasprintf ("Dir Hist %s", panel_name);
    panel->dir_history = history_get (panel->hist_name);
    directory_history_add (panel, panel->cwd);

    panel->dir.list         = zmalloc (MIN_FILES * sizeof (file_entry));
    panel->dir.size         = MIN_FILES;
    panel->dirty            = 1;
    panel->format_modified  = 1;

    panel->panel_name = strdup (panel_name);
    panel->user_format = strdup (DEFAULT_USER_FORMAT);

    for(i = 0; i < LIST_TYPES; i++)
	panel->user_status_format [i] = strdup (DEFAULT_USER_FORMAT);

    // panel->search_buffer [0] = 0;
    panel->frame_size = frame_half;
    section = xasprintf ("Temporal:%s", panel->panel_name);
    if (!profile_has_section (section, profile_name)) {
	free (section);
	section = strdup (panel->panel_name);
    }
    panel_load_setup (panel, section);
    free (section);

    /* Load format strings */
    err = set_panel_formats (panel);
    if (err) {
	if (err & 0x01) {
	    free (panel->user_format);
	    panel->user_format = strdup (DEFAULT_USER_FORMAT);
	}
	if (err & 0x02) {
	    free (panel->user_status_format [panel->list_type]);
	    panel->user_status_format [panel->list_type] = strdup (DEFAULT_USER_FORMAT);
	}
	set_panel_formats (panel);
    }

    /* Load the default format */
    panel->count = do_load_dir (&panel->dir, panel->sort_type,
				panel->reverse, panel->case_sensitive, panel->filter);
    return panel;
}

void
panel_reload (WPanel *panel)
{
    int i;
    struct stat current_stat;

    if (fast_reload
	&& !stat (panel->cwd, &current_stat)
	&& current_stat.st_ctime == panel->dir_stat.st_ctime
	&& current_stat.st_mtime == panel->dir_stat.st_mtime)
	return;

    while (mc_chdir (panel->cwd) == -1) {
	char *last_slash;

	if (panel->cwd [0] == '/' && panel->cwd [1] == 0) {
	    clean_dir (&panel->dir, panel->count);
	    panel->count = set_zero_dir (&panel->dir);
	    return;
	}
	last_slash = strrchr (panel->cwd, '/');
	if (!last_slash || last_slash == panel->cwd)
	    strcpy (panel->cwd, "/");
	else
	    *last_slash = 0;
	memset (&(panel->dir_stat), 0, sizeof (panel->dir_stat));
	show_dir (panel);
    }

    panel->count = do_reload_dir (&panel->dir, panel->sort_type, panel->count,
				  panel->reverse, panel->case_sensitive, panel->filter);
    panel->marked = 0;
    panel->dirs_marked = 0;
    panel->total  = 0;
    panel->has_dir_sizes = 0;

    for (i = 0; i < panel->count; i++)
	if (panel->dir.list [i].f.marked) {
	    /* do_file_mark will return immediately if newmark == oldmark.
	       So we have to first unmark it to get panel's summary information
	       updated. (Norbert) */
	    panel->dir.list [i].f.marked = 0;
	    do_file_mark (panel, i, 1);
	}
}

void
paint_frame (WPanel *panel)
{
    int  header_len;
    int  spaces, extra;
    int  side, width;

    if (!panel->split)
	adjust_top_file (panel);

    widget_erase (&panel->widget);
    show_dir (panel);

    widget_move (&panel->widget, 1, 1);

    for (side = 0; side <= panel->split; side++) {
	format_e *format;

	if (side) {
	    attrset (NORMAL_COLOR);
	    one_vline ();
	    width = panel->widget.cols - panel->widget.cols/2 - 1;
	} else if (panel->split)
	    width = panel->widget.cols/2 - 3;
	else
	    width = panel->widget.cols - 2;

	for (format = panel->format; format; format = format->next) {
	    if (format->string_fn) {
		char buffer[30]; /*Hope that this is enough ;-) */
		const char *txt = format->title;

		header_len = strlen (txt);
		if (header_len > format->field_len) {
		    txt = strcpy (buffer, txt);
		    buffer [format->field_len] = 0;
		    header_len = strlen (txt);
		}

		attrset (MARKED_COLOR);
		spaces = (format->field_len - header_len) / 2;
		extra  = (format->field_len - header_len) % 2;
		printw ("%*s%-s%*s", spaces, "",
			 txt, spaces+extra, "");
		width -= 2 * spaces + extra + header_len;
	    } else {
		attrset (NORMAL_COLOR);
		one_vline ();
		width --;
		continue;
	    }
	}

	if (width > 0)
	    printw ("%*s", width, "");
    }
}

static char *
parse_panel_size (WPanel *panel, char *format, int isstatus)
{
    int frame = frame_half;
    format = skip_separators (format);

    if (!strncmp (format, "full", 4)) {
	frame = frame_full;
	format += 4;
    } else if (!strncmp (format, "half", 4)) {
	frame = frame_half;
	format += 4;
    }

    if (!isstatus) {
	panel->frame_size = frame;
	panel->split = 0;
    }

    /* Now, the optional column specifier */
    format = skip_separators (format);

    if (*format == '1' || *format == '2') {
	if (!isstatus)
	    panel->split = *format == '2';
	format++;
    }

    if (!isstatus)
	panel_update_cols (&(panel->widget), panel->frame_size);

    return skip_separators (format);
}

/* Format is:

   all              := panel_format? format
   panel_format     := [full|half] [1|2]
   format           := one_format_e
		     | format , one_format_e

   one_format_e     := just format.id [opt_size]
   just             := [<|>]
   opt_size         := : size [opt_expand]
   size             := [0-9]+
   opt_expand       := +

*/

static inline char *skip_numbers (char *s)
{
    for (;*s; s++)
	if (!isdigit (*s))
	    break;
    return s;
}


static format_e *
parse_display_format (WPanel *panel, char *format, char **error, int isstatus, int *res_total_cols)
{
    format_e *darr, *old = 0, *home = 0;    /* The formats we return */
    int  total_cols = 0;                /* Used columns by the format */
    int  set_justify;                   /* flag: set justification mode? */
    int  justify = 0;                   /* Which mode. */
    int  items = 0;                     /* Number of items in the format */
    int  i;

    *error = 0;

    /*
     * This makes sure that the panel and mini status full/half mode
     * setting is equal
     */
    format = parse_panel_size (panel, format, isstatus);

    while (*format) {   /* format can be an empty string */
	int found = 0;

	darr = x_malloc (sizeof (format_e), "parse_display_format");

	/* I'm so ugly, don't look at me :-) */
	if (!home)
	    home = old = darr;

	old->next = darr;
	darr->next = 0;
	old = darr;

	format = skip_separators (format);

	if (*format == '<' || *format == '>') {
	    set_justify = 1;
	    justify = *format == '<' ? J_LEFT : J_RIGHT;
	    format = skip_separators (format+1);
	} else
	    set_justify = 0;

	for (i = 0; i < ELEMENTS(formats); i++) {
	    int klen = strlen (formats [i].id);

	    if (strncmp (format, formats [i].id, klen) != 0)
		continue;

	    format += klen;

	    if (formats [i].use_in_gui)
		items++;

	    darr->use_in_gui          = formats [i].use_in_gui;
	    darr->requested_field_len = formats [i].min_size;
	    darr->string_fn           = formats [i].string_fn;
	    if (formats [i].title [0])
		    darr->title = formats [i].title;
	    else
		    darr->title = "";
	    darr->id                  = formats [i].id;
	    darr->expand              = formats [i].expands;

	    if (set_justify)
		darr->just_mode = justify;
	    else
		darr->just_mode = formats [i].default_just;

	    found = 1;

	    format = skip_separators (format);

	    /* If we have a size specifier */
	    if (*format == ':') {
		int req_length;

		/* If the size was specified, we don't want
		 * auto-expansion by default
		 */
		darr->expand = 0;
		format++;
		req_length = atoi (format);
		darr->requested_field_len = req_length;

		format = skip_numbers (format);

		/* Now, if they insist on expansion */
		if (*format == '+') {
		    darr->expand = 1;
		    format++;
		}

	    }

	    break;
	}
	if (!found) {
	    char old_char;

	    int pos = min (8, strlen (format));
	    delete_format (home);
	    old_char = format [pos];
	    format [pos] = 0;
	    *error = xasprintf("Unknow tag on display format: %s", format);
	    format [pos] = old_char;
	    return 0;
	}
	total_cols += darr->requested_field_len;
    }

    *res_total_cols = total_cols;
    if (home)
       home->items = items;
    return home;
}

static format_e *
use_display_format (WPanel *panel, char *format, char **error, int isstatus)
{
#define MAX_EXPAND 4
    int  expand_top = 0;                /* Max used element in expand */
    int  usable_columns;                /* Usable columns in the panel */
    int  total_cols = 0;                /* make gcc happend */
#if 0
    char *expand_list [MAX_EXPAND];    /* Expand at most 4 fields. */
#endif
    int  i;
    format_e *darr, *home;

    if (!format)
	format = DEFAULT_USER_FORMAT;

    home = parse_display_format (panel, format, error, isstatus, &total_cols);

    if (*error)
	return 0;

    /* Status needn't to be split */
    usable_columns = ((panel->widget.cols-2)/((isstatus)
					      ? 1
					      : (panel->split+1))) - (!isstatus && panel->split);

#if 0
    /* Clean expand list */
    for (i = 0; i < MAX_EXPAND; i++)
	expand_list [i] = '\0';
#endif

    /* Look for the expandable fields and set field_len based on the requested field len */
    for (darr = home; darr && expand_top < MAX_EXPAND; darr = darr->next) {
	darr->field_len = darr->requested_field_len;
#if 0
	if (darr->expand)
	    expand_list [expand_top++] = darr->id;
#else
	if (darr->expand)
		expand_top++;
#endif
    }

    /* If we used more columns than the available columns, adjust that */
    if (total_cols > usable_columns) {
	int pdif, dif = total_cols - usable_columns;

	while (dif) {
	    pdif = dif;
	    for (darr = home; darr; darr = darr->next) {
		if (dif && darr->field_len - 1) {
		    darr->field_len--;
		    dif--;
		}
	    }

	    /* avoid endless loop if num fields > 40 */
	    if (pdif == dif)
		break;
	}
	total_cols  = usable_columns; /* give up, the rest should be truncated */
    }

    /* Expand the available space */
    if ((usable_columns > total_cols) && expand_top) {
	int spaces = (usable_columns - total_cols) / expand_top;
	int extra  = (usable_columns - total_cols) % expand_top;

	for (i = 0, darr = home; darr && (i < expand_top); darr = darr->next)
		if (darr->expand) {
			darr->field_len += (spaces + ((i == 0) ? extra : 0));
			i++;
		}
    }
    return home;
}

/* Switches the panel to the mode specified in the format           */
/* Seting up both format and status string. Return: 0 - on success; */
/* 1 - format error; 2 - status error; 3 - errors in both formats.  */
int
set_panel_formats (WPanel *p)
{
    format_e *form;
    char *err;
    int retcode = 0;

    form = use_display_format (p, panel_format (p), &err, 0);

    if (err) {
	free (err);
	retcode = 1;
    } else {
	if (p->format)
	    delete_format (p->format);

	p->format = form;
    }

    if (show_mini_info) {

	form = use_display_format (p, mini_status_format (p), &err, 1);

	if (err) {
	    free (err);
	    retcode += 2;
	}
	else {
	    if (p->status_format)
		delete_format (p->status_format);

	    p->status_format = form;
	}
    }

    panel_format_modified (p);
    panel_update_cols (&(p->widget), p->frame_size);

    return retcode;
}

/* Given the panel->view_type returns the format string to be parsed */
char *
panel_format (WPanel *panel)
{
    switch (panel->list_type) {

    case list_long:
	return "full perm,space,nlink,space,owner,space,group,space,size,space,mtime,space,name";

    case list_brief:
	return "half 2,type,name";

    case list_user:
	return panel->user_format;

    default:
    case list_full:
	return "half type,name,|,size,|,mtime";
    }
}

char *
mini_status_format (WPanel *panel)
{
    if (panel->user_mini_status)
       return panel->user_status_format [panel->list_type];

    switch (panel->list_type) {

    case list_long:
       return "full perm,space,nlink,space,owner,space,group,space,size,space,mtime,space,name";

    case list_brief:
       return "half type,name,space,bsize,space,perm,space";

    case list_full:
       return "half type,name";

    default:
    case list_user:
       return panel->user_format;
    }
}

/*                          */
/* Panel operation commands */
/*                          */

/* Returns the number of items in the given panel */
static int
ITEMS (WPanel *p)
{
    if (p->split)
	return llines (p) * 2;
    else
	return llines (p);
}

/* This function sets redisplays the selection */
static void
select_item (WPanel *panel)
{
    int repaint = 0;
    int items = ITEMS (panel);

    /* Although currently all over the code we set the selection and
       top file to decent values before calling select_item, I could
       forget it someday, so it's better to do the actual fitting here */

    if (panel->top_file < 0) {
	repaint = 1;
	panel->top_file = 0;
    }

    if (panel->selected < 0)
	panel->selected = 0;

    if (panel->selected > panel->count-1)
	panel->selected = panel->count - 1;

    if (panel->top_file > panel->count-1) {
	repaint = 1;
	panel->top_file = panel->count-1;
    }

    if ((panel->count - panel->top_file) < items) {
	repaint = 1;
	panel->top_file = panel->count - items;
	if (panel->top_file < 0)
	    panel->top_file = 0;
    }

    if (panel->selected < panel->top_file) {
	repaint = 1;
	panel->top_file = panel->selected;
    }

    if ((panel->selected - panel->top_file) >= items) {
	repaint = 1;
	panel->top_file = panel->selected - items + 1;
    }

    if (repaint)
	paint_panel (panel);
    else
	repaint_file (panel, panel->selected, 1, 2*selection (panel)->f.marked+1, 0);

    display_mini_info (panel);

    execute_hooks (select_file_hook);
}

/* Clears all files in the panel, used only when one file was marked */
void
unmark_files (WPanel *panel)
{
    int i;

    if (!panel->marked)
	return;
    for (i = 0; i < panel->count; i++)
	file_mark (panel, i, 0);

    panel->dirs_marked = 0;
    panel->marked = 0;
    panel->total = 0;
}

static void
unselect_item (WPanel *panel)
{
    repaint_file (panel, panel->selected, 1, 2*selection (panel)->f.marked, 0);
}

static void
do_move_down (WPanel *panel)
{
    if (panel->selected+1 == panel->count)
	return;

    unselect_item (panel);
    panel->selected++;

    if (panel->selected - panel->top_file == ITEMS (panel) &&
	panel_scroll_pages) {
	/* Scroll window half screen */
	panel->top_file += ITEMS (panel)/2;
	if (panel->top_file > panel->count - ITEMS (panel))
		panel->top_file = panel->count - ITEMS (panel);
	paint_dir (panel);
	select_item (panel);
    }
    select_item (panel);
}

static void
do_move_up (WPanel *panel)
{
    if (panel->selected == 0)
	return;

    unselect_item (panel);
    panel->selected--;
    if (panel->selected < panel->top_file && panel_scroll_pages) {
	/* Scroll window half screen */
	panel->top_file -= ITEMS (panel)/2;
	if (panel->top_file < 0) panel->top_file = 0;
	paint_dir (panel);
    }
    select_item (panel);
}

static void
move_rel (WPanel *panel, int rel)
{
    unselect_item (panel);

    if (rel < 0) {
	if (panel->selected + rel < 0)
	    panel->selected = 0;
	else
	    panel->selected = panel->selected + rel;
    } else {
	if (panel->selected + rel >= panel->count)
	    panel->selected = panel->count - 1;
	else
	    panel->selected = panel->selected + rel;
    }
    select_item (panel);
}

/*                           */
/* Panel key binded commands */
/*                           */
static void
scr_move_up (WPanel *panel)
{
    if (panel->list_type == list_icons) {
	move_rel (panel, -ICONS_PER_ROW (panel));
    } else
	do_move_up (panel);
}

static void
scr_move_down (WPanel *panel)
{
    if (panel->list_type == list_icons) {
	move_rel (panel, ICONS_PER_ROW (panel));
    } else
	do_move_down (panel);
}

/* Changes the selection by lines (may be negative) */
static void
move_selection (WPanel *panel, int lines)
{
    int new_pos;
    int adjust = 0;

    new_pos = panel->selected + lines;
    if (new_pos >= panel->count)
	new_pos = panel->count-1;

    if (new_pos < 0)
	new_pos = 0;

    unselect_item (panel);
    panel->selected = new_pos;

    if (panel->selected - panel->top_file >= ITEMS (panel)) {
	panel->top_file += lines;
	adjust = 1;
    }

    if (panel->selected - panel->top_file < 0) {
	panel->top_file += lines;
	adjust = 1;
    }

    if (adjust) {
	if (panel->top_file > panel->selected)
	    panel->top_file = panel->selected;
	if (panel->top_file < 0)
	    panel->top_file = 0;
	paint_dir (panel);
    }
    select_item (panel);
}

static int
scr_move_left (WPanel *panel, int c_code)
{
    if (panel->list_type == list_icons) {
	do_move_up (panel);
	return 1;
    } else {
	if (panel->split) {
	    move_selection (panel, -llines (panel));
	    return 1;
	} else
	    return maybe_cd (c_code, 0);
    }
}

static int
scr_move_right (WPanel *panel, int c_code)
{
    if (panel->list_type == list_icons) {
	do_move_down (panel);
	return 1;
    } else {
	if (panel->split) {
	    move_selection (panel, llines (panel));
	    return 1;
	} else
	    return maybe_cd (c_code, 1);
    }
}

static void
prev_page (WPanel *panel)
{
    int items;

    if (!panel->selected && !panel->top_file)
	return;
    unselect_item (panel);
    items = ITEMS (panel);
    if (panel->top_file < items)
	items = panel->top_file;
    if (!items)
	panel->selected = 0;
    else
	panel->selected -= items;
    panel->top_file -= items;

    /* This keeps the selection in a reasonable place */
    if (panel->selected < 0)
	panel->selected = 0;
    if (panel->top_file < 0)
	panel->top_file = 0;
    x_adjust_top_file (panel);
    select_item (panel);
    paint_dir (panel);
}

static void
prev_page_key (WPanel *panel)
{
    if (console_flag && ctrl_pressed ()) {
	do_cd ("..", cd_exact);
    } else
	prev_page (panel);
}

static void
next_page (WPanel *panel)
{
    int items;

    if (panel->selected == panel->count - 1)
	return;
    unselect_item (panel);
    items = ITEMS (panel);
    if (panel->top_file > panel->count - 2 * items)
	items = panel->count - items - panel->top_file;
    if (panel->top_file + items < 0)
	items = - panel->top_file;
    if (!items)
	panel->selected = panel->count - 1;
    else
	panel->selected += items;
    panel->top_file += items;

    /* This keeps the selection in it's relative position */
    if (panel->selected >= panel->count)
	panel->selected = panel->count - 1;
    if (panel->top_file >= panel->count)
	panel->top_file = panel->count - 1;
    x_adjust_top_file (panel);
    select_item (panel);
    paint_dir (panel);
}

static void next_page_key (WPanel *panel)
{
    if (console_flag&&ctrl_pressed()&&
	(S_ISDIR(selection (panel)->buf.st_mode) ||
	 link_isdir (selection (panel))))
	do_cd (selection (panel)->fname, cd_exact);
    else
	next_page (panel);
}

static void
goto_top_file (WPanel *panel)
{
    unselect_item (panel);
    panel->selected = panel->top_file;
    select_item (panel);
}

static void
goto_middle_file (WPanel *panel)
{
    unselect_item (panel);
    panel->selected = panel->top_file + (ITEMS (panel)/2);
    if (panel->selected >= panel->count)
	panel->selected = panel->count - 1;
    select_item (panel);
}

static void
goto_bottom_file (WPanel *panel)
{
    unselect_item (panel);
    panel->selected = panel->top_file + ITEMS (panel)-1;
    if (panel->selected >= panel->count)
	panel->selected = panel->count - 1;
    select_item (panel);
}

static void
scr_move_home (WPanel *panel)
{
    if (panel->selected == 0)
	return;
    unselect_item (panel);

    if (torben_fj_mode) {
	int middle_pos = panel->top_file + (ITEMS (panel)/2);

	if (panel->selected > middle_pos) {
	    goto_middle_file (panel);
	    return;
	}
	if (panel->selected != panel->top_file) {
	    goto_top_file (panel);
	    return;
	}
    }

    panel->top_file = 0;
    panel->selected = 0;

    paint_dir (panel);
    select_item (panel);
}

static void
scr_move_end (WPanel *panel)
{
    if (panel->selected == panel->count-1)
	return;
    unselect_item (panel);
    if (torben_fj_mode) {
	int middle_pos = panel->top_file + (ITEMS (panel)/2);

	if (panel->selected < middle_pos) {
	    goto_middle_file (panel);
	    return;
	}
	if (panel->selected != (panel->top_file + ITEMS(panel)-1)) {
	    goto_bottom_file (panel);
	    return;
	}
    }

    panel->selected = panel->count-1;
    paint_dir (panel);
    select_item (panel);
}

/* This routine marks a file or a directory */
void
do_file_mark (WPanel *panel, int idx, int mark)
{
    if (panel->dir.list [idx].f.marked == mark)
	return;
    /*
     * Only '..' can't be marked, '.' isn't visible.
     */
    if (strcmp (panel->dir.list [idx].fname, "..")) {
	file_mark (panel, idx, mark);
	if (panel->dir.list [idx].f.marked) {
	    panel->marked++;
	    if (S_ISDIR (panel->dir.list [idx].buf.st_mode)) {
		if (panel->has_dir_sizes)
		    panel->total += panel->dir.list [idx].buf.st_size;
		panel->dirs_marked++;
	    } else
		panel->total += panel->dir.list [idx].buf.st_size;
	    set_colors (panel);
	} else {
	    if (S_ISDIR(panel->dir.list [idx].buf.st_mode)) {
		if (panel->has_dir_sizes)
		    panel->total -= panel->dir.list [idx].buf.st_size;
		panel->dirs_marked--;
	    } else
		panel->total -= panel->dir.list [idx].buf.st_size;
	    panel->marked--;
	}
    }
}

static void
do_mark_file (WPanel *panel, int do_move)
{
    int idx = panel->selected;
    do_file_mark (panel, idx, selection (panel)->f.marked ? 0 : 1);
    repaint_file (panel, idx, 1, 2*panel->dir.list [idx].f.marked+1, 0);

    if (mark_moves_down && do_move)
	scr_move_down (panel);
    display_mini_info (panel);
}

static void
mark_file (WPanel *panel)
{
    do_mark_file (panel, 1);
}

/* Incremental search of a file name in the panel */
static void
scr_do_search (WPanel *panel, int c_code)
{
    int l, i;
    int wrapped = 0;
    int found;

    l = strlen (panel->search_buffer);
    if (l && (c_code == 8 || c_code == 0177 || c_code == KEY_BACKSPACE))
	panel->search_buffer [--l] = 0;
    else {
	if (c_code && l < sizeof (panel->search_buffer)) {
	    panel->search_buffer [l] = c_code;
	    panel->search_buffer [l+1] = 0;
	    l++;
	}
    }

    found = 0;
    for (i = panel->selected; !wrapped || i != panel->selected; i++) {
	if (i >= panel->count) {
	    i = 0;
	    if (wrapped)
		break;
	    wrapped = 1;
	}
	if (strncmp (panel->dir.list [i].fname, panel->search_buffer, l) == 0) {
	    unselect_item (panel);
	    panel->selected = i;
	    select_item (panel);
	    found = 1;
	    break;
	}
    }
    if (!found)
	panel->search_buffer [--l] = 0;
    paint_panel (panel);
}

static void
scr_start_search (WPanel *panel)
{
    if (panel->searching) {
	if (panel->selected+1 == panel->count)
	    panel->selected = 0;
	else
	    scr_move_down (panel);
	scr_do_search (panel, 0);
    } else {
	panel->searching = 1;
	panel->search_buffer [0] = 0;
	display_mini_info (panel);
	mc_refresh ();
    }
}

static void
do_enter (WPanel *panel)
{
    if (S_ISDIR (selection (panel)->buf.st_mode)
	|| link_isdir (selection (panel))) {
	do_cd (selection (panel)->fname, cd_exact);
	return;
    } else {
	if (is_exe (selection (panel)->buf.st_mode) &&
	    if_link_is_exe (selection (panel))) {
#ifdef USE_VFS
	    if (vfs_current_is_local ())
#endif
	    {
		char *tmp = name_quote (selection (panel)->fname, 0);
		char *cmd = xasprintf ("./%s", tmp);
		if (!confirm_execute || (query_dialog (" The Midnight Commander ",
						       " Do you really want to execute? ",
						       0, 2, "&Yes", "&No") == 0))
		    execute (cmd);
		free (tmp);
		free (cmd);
	    }
#ifdef USE_VFS
	    else if (vfs_current_is_extfs ()) {
		char *tmp = vfs_get_current_dir();
		char *tmp2;

		tmp2 = concat_path_file (tmp, selection (panel)->fname);
		extfs_run(tmp2);
		free (tmp2);
	    }
#endif /* USE_VFS */
	    return;
	} else {
		regex_command (selection (panel)->fname, "Open", NULL, 0);
		return;
	}
    }
}

static void
chdir_other_panel (WPanel *panel)
{
    char *new_dir;

    if (get_other_type () != view_listing)
	return;

    if (!S_ISDIR (panel->dir.list [panel->selected].buf.st_mode))
	new_dir = concat_path_file (panel->cwd, "..");
    else
	new_dir = concat_path_file (panel->cwd, panel->dir.list [panel->selected].fname);

    change_panel ();
    do_cd (new_dir, cd_exact);
    change_panel ();

    scr_move_down (panel);

    free (new_dir);
}

static void
chdir_to_readlink (WPanel *panel)
{
    char *new_dir;

    if (get_other_type () != view_listing)
	return;

    if (S_ISLNK (panel->dir.list [panel->selected].buf.st_mode)) {
	char buffer [MC_MAXPATHLEN], *p;
	int i;
	struct stat mybuf;

	i = readlink (selection (panel)->fname, buffer, MC_MAXPATHLEN);
	if (i < 0)
	    return;
	if (mc_stat (selection (panel)->fname, &mybuf) < 0)
	    return;
	buffer [i] = 0;
	if (!S_ISDIR (mybuf.st_mode)) {
	    p = strrchr (buffer, '/');
	    if (p && !p[1]) {
		*p = 0;
		p = strrchr (buffer, '/');
	    }
	    if (!p)
		return;
	    p[1] = 0;
	}
	if (*buffer == '/')
	    new_dir = strdup (buffer);
	else
	    new_dir = concat_path_file (panel->cwd, buffer);

	change_panel ();
	do_cd (new_dir, cd_exact);
	change_panel ();

	scr_move_down (panel);

	free (new_dir);
    }
}

static const key_map panel_keymap [] = {
    { KEY_DOWN,   scr_move_down },
    { KEY_UP,     scr_move_up },

    /* The action button :-) */
    { '\n',       do_enter },
    { KEY_ENTER,  do_enter },

    { KEY_IC,     mark_file },
    { KEY_HOME,   scr_move_home },
    { KEY_C1,     scr_move_end },
    { KEY_END,    scr_move_end },
    { KEY_A1,     scr_move_home },
    { KEY_NPAGE,  next_page_key },
    { KEY_PPAGE,  prev_page_key },

    /* To quickly move in the panel */
    { ALT('g'),   goto_top_file },
    { ALT('r'),   goto_middle_file }, /* M-r like emacs */
    { ALT('j'),   goto_bottom_file },

    /* Emacs-like bindings */
    { XCTRL('v'), next_page },          /* C-v like emacs */
    { ALT('v'),   prev_page },          /* M-v like emacs */
    { XCTRL('p'), scr_move_up },            /* C-p like emacs */
    { XCTRL('n'), scr_move_down },          /* C-n like emacs */
    { XCTRL('s'), scr_start_search },       /* C-s like emacs */
    { ALT('s'),   scr_start_search },       /* M-s not like emacs */
    { XCTRL('t'), mark_file },
    { ALT('o'),   chdir_other_panel },
    { ALT('l'),   chdir_to_readlink },
    { KEY_F(13),  view_simple_cmd },
    { KEY_F(14),  edit_cmd_new },
    { ALT('y'),   directory_history_prev },
    { ALT('u'),   directory_history_next },
    { ALT('H'),   directory_history_list },
    { ALT('+'),   select_cmd_panel },
    { KEY_KP_ADD, select_cmd_panel },
    { ALT('\\'),  unselect_cmd_panel },
    { ALT('-'),   unselect_cmd_panel },
    { KEY_KP_SUBTRACT, unselect_cmd_panel },
    { ALT('*'),   reverse_selection_cmd_panel },
    { KEY_KP_MULTIPLY, reverse_selection_cmd_panel },

    { 0, 0 }
};

static inline int
panel_key (WPanel *panel, int key)
{
    int i;

    for (i = 0; panel_keymap [i].key_code; i++) {
	if (key == panel_keymap [i].key_code) {
	    if (panel_keymap [i].fn != scr_start_search)
	       panel->searching = 0;
	    (*panel_keymap [i].fn)(panel);
	    return 1;
	}
    }
    if (torben_fj_mode && key == ALT('h')) {
	goto_middle_file (panel);
	return 1;
    }

    /* We do not want to take a key press if nothing can be done with it */
    /* The command line widget may do something more usefull */
    if (key == KEY_LEFT)
	return scr_move_left (panel, key);

    if (key == KEY_RIGHT)
	return scr_move_right (panel, key);

    if (is_abort_char (key)) {
	panel->searching = 0;
	display_mini_info (panel);
	return 1;
    }

    /* Do not eat characters not meant for the panel below ' ' (e.g. C-l). */
    if ((key >= ' '&& key <= 255) || key == 8 || key == KEY_BACKSPACE) {
	if (panel->searching) {
	    scr_do_search (panel, key);
	    return 1;
	}

	if (!command_prompt) {
	    scr_start_search (panel);
	    scr_do_search (panel, key);
	    return 1;
	}
    }

    return 0;
}

static int
panel_callback (Dlg_head *h, WPanel *panel, int msg, int par)
{
    switch (msg) {
    case WIDGET_INIT:
	return 1;

    case WIDGET_DRAW:
	paint_panel (panel);
	break;

    case WIDGET_FOCUS:
	current_panel = panel;
	panel->active = 1;
	if (mc_chdir (panel->cwd) != 0)
	    MC_message (1, " Error ", " Can't chdir to %s \n %m ", panel->cwd);

	show_dir (panel);
	focus_select_item (panel);
	define_label (h, (Widget *)panel, 1, "Help", help_cmd);
	define_label (h, (Widget *)panel, 2, "Menu", user_menu_cmd);
	define_label (h, (Widget *)panel, 3, "View", view_panel_cmd);
	define_label (h, (Widget *)panel, 4, "Edit", edit_panel_cmd);
	define_label (h, (Widget *)panel, 5, "Copy", copy_cmd);
	define_label (h, (Widget *)panel, 6, "RenMov", ren_cmd);
	define_label (h, (Widget *)panel, 7, "Mkdir", mkdir_panel_cmd);
	define_label (h, (Widget *)panel, 8, "Delete", delete_cmd);
	redraw_labels (h, (Widget *)panel);
	/* Chain behaviour */
	default_proc (h, WIDGET_FOCUS, par);
	return 1;

    case WIDGET_UNFOCUS:
	/* Janne: look at this for the multiple panel options */
	if (panel->searching) {
	    panel->searching = 0;
	    display_mini_info (panel);
	}
	panel->active = 0;
	show_dir (panel);
	unselect_item (panel);
	return 1;

    case WIDGET_KEY:
	return panel_key (panel, par);
	break;
    }
    return default_proc (h, msg, par);
}

/*                                     */
/* Panel mouse events support routines */
/*                                     */
static void
mouse_toggle_mark (WPanel *panel)
{
    do_mark_file (panel, 0);
    mouse_marking = selection (panel)->f.marked;
}

static void
mouse_set_mark (WPanel *panel)
{
    if (mouse_marking && !(selection (panel)->f.marked))
	do_mark_file (panel, 0);
    else if (!mouse_marking && (selection (panel)->f.marked))
	do_mark_file (panel, 0);
}

static int
mark_if_marking (WPanel *panel, Gpm_Event *event)
{
    if (event->buttons & GPM_B_RIGHT) {
	if (event->type & GPM_DOWN)
	    mouse_toggle_mark (panel);
	else
	    mouse_set_mark (panel);
	return 1;
    }
    return 0;
}

void
file_mark (WPanel *panel, int Index, int val)
{
    panel->dir.list [Index].f.marked = val;
}

int
panel_event (Gpm_Event *event, WPanel *panel)
{
    const int lines = llines (panel);

    int my_index;

    if (event->type & GPM_DOWN && event->x == 1 + 1 && event->y == 0 + 1) {
	directory_history_prev (panel);
	return MOU_NORMAL;
    }

    if (event->type & GPM_DOWN && event->x == panel->widget.cols - 2 + 1 && event->y == 0 + 1) {
	directory_history_next (panel);
	return MOU_NORMAL;
    }

    if (event->type & GPM_DOWN && event->x == panel->widget.cols - 3 + 1 && event->y == 0 + 1) {
	directory_history_list (panel);
	return MOU_NORMAL;
    }

    event->y -= 2;
    if ((event->type & (GPM_DOWN|GPM_DRAG))) {

	if (panel != (WPanel *) current_dlg->current->widget)
	    change_panel ();

	if (event->y <= 0) {
	    mark_if_marking (panel, event);
	    if (mouse_move_pages)
		prev_page (panel);
	    else
		scr_move_up (panel);
	    return MOU_REPEAT;
	}

	if (!((panel->top_file + event->y <= panel->count) &&
	      event->y <= lines)) {
	    mark_if_marking (panel, event);
	    if (mouse_move_pages)
		next_page (panel);
	    else
		scr_move_down (panel);
	    return MOU_REPEAT;
	}
	my_index = panel->top_file + event->y - 1;
	if (panel->split) {
	    if (event->x > ((panel->widget.cols-2)/2))
		my_index += llines (panel);
	}

	if (my_index >= panel->count)
	    my_index = panel->count - 1;

	if (my_index != panel->selected) {
	    unselect_item (panel);
	    panel->selected = my_index;
	    select_item (panel);
	}

	/* This one is new */
	mark_if_marking (panel, event);

    } else if ((event->type & (GPM_UP|GPM_DOUBLE)) == (GPM_UP|GPM_DOUBLE)) {
	    if (event->y > 0 && event->y <= lines)
		do_enter (panel);
    }
    return MOU_NORMAL;
}

/* Setup loading/saving. */

#ifndef PORT_LIST_MODE_NAME
#     define PORT_LIST_MODE_NAME "list_mode"
#endif

#ifndef PORT_LIST_MODE_DEFAULT
#     define PORT_LIST_MODE_DEFAULT "full"
#endif

#ifdef USE_NETCODE
/* Delay to retry a connection */
static int ftpfs_retry_seconds = 30;

/* Method to use to connect to ftp sites */
static int ftpfs_use_passive_connections = 1;

/* Method used to get directory listings:
   1: try 'LIST -la <path>', if it fails
      fall back to CWD <path>; LIST
   0: always use CWD <path>; LIST
 */
static int ftpfs_use_unix_list_options = 1;
#endif

static const struct {
    const char *key;
    sortfn *sort_type;
} sort_names [] = {
    { "name",      sort_name },
    { "extension", sort_ext },
    { "time",      sort_time },
    { "atime",     sort_atime },
    { "ctime",     sort_ctime },
    { "size",      sort_size },
    { "inode",     sort_inode },
    { "unsorted",  unsorted },
    { 0, 0 }
};

static const struct {
    const char *key;
    int  list_type;
} list_types [] = {
    { "full",  list_full  },
    { "brief", list_brief },
    { "long",  list_long  },
    { "user",  list_user  },
    { "icons", list_icons },
    { 0, 0 }
};

static const struct {
    const char *opt_name;
    int  opt_type;
} panel_types [] = {
    { "listing",   view_listing },
    { "quickview", view_quick   },
    { "info",      view_info },
    { "tree",      view_tree },
    { 0, 0 }
};

static const struct {
    const char *opt_name;
    int *opt_addr;
} layout [] = {
    { "equal_split", &equal_split },
    { "first_panel_size", &first_panel_size },
    { "message_visible", &message_visible },
    { "keybar_visible", &keybar_visible },
    { "xterm_hintbar", &xterm_hintbar },
    { "output_lines", &output_lines },
    { "menubar_visible", &menubar_visible },
    { "command_prompt", &command_prompt },
    { "show_mini_info", &show_mini_info },
    { "permission_mode", &permission_mode },
    { "filetype_mode", &filetype_mode },
    { 0, 0 }
};


#undef KEEP_UNUSE_OPTIONS
#ifdef KEEP_UNUSE_OPTIONS
static int align_extensions = 1;
static int iconify_on_exec = 1;
#endif

static const struct {
    const char *opt_name;
    int  *opt_addr;
} options [] = {
#ifdef KEEP_UNUSE_OPTIONS
    { "align_extensions", &align_extensions },
    { "iconify_on_exec", &iconify_on_exec },
#endif
    { "show_backups", &show_backups },
    { "show_dot_files", &show_dot_files },
    { "verbose", &MC_verbose },
    { "mark_moves_down", &mark_moves_down },
    { "pause_after_run", &pause_after_run },
    { "shell_patterns", &easy_patterns },
    { "auto_save_setup", &auto_save_setup },
    { "auto_menu", &auto_menu },
    { "use_internal_view", &use_internal_view },
    { "use_internal_edit", &use_internal_edit },
    { "clear_before_exec", &clear_before_exec },
    { "mix_all_files", &mix_all_files },
    { "fast_reload", &fast_reload },
    { "fast_reload_msg_shown", &fast_reload_w },
    { "confirm_delete", &confirm_delete },
    { "confirm_overwrite", &confirm_overwrite },
    { "confirm_execute", &confirm_execute },
    { "confirm_exit", &confirm_exit },
    { "safe_delete", &know_not_what_am_i_doing },
    { "mouse_repeat_rate", &mou_auto_repeat },
    { "double_click_speed", &double_click_speed },
    { "eight_bit_clean", &eight_bit_clean },
    { "full_eight_bits", &full_eight_bits },
    { "use_8th_bit_as_meta", &use_8th_bit_as_meta },
    { "confirm_view_dir", &confirm_view_dir },
    { "mouse_move_pages", &mouse_move_pages },
    { "mouse_move_pages_viewer", &mouse_move_pages_viewer },
    { "fast_refresh", &fast_refresh },
    { "navigate_with_arrows", &navigate_with_arrows },
    { "advanced_chown", &advanced_chfns },
    { "drop_menus", &drop_menus },
    { "wrap_mode",  &global_wrap_mode},
    { "old_esc_mode", &old_esc_mode },
    { "cd_symlinks", &cd_symlinks },
    { "show_all_if_ambiguous", &show_all_if_ambiguous },
    { "have_fast_cpu", &have_fast_cpu },
    { "torben_fj_mode", &torben_fj_mode },
    { "use_file_to_guess_type", &use_file_to_check_type },
    { "alternate_plus_minus", &alternate_plus_minus },
    { "only_leading_plus_minus", &only_leading_plus_minus },
    { "show_output_starts_shell", &output_starts_shell },
    { "panel_scroll_pages", &panel_scroll_pages },
    { "xtree_mode", &xtree_mode },
    { "num_history_items_recorded", &num_history_items_recorded },
#ifdef USE_VFS
    { "tar_gzipped_memlimit", &tar_gzipped_memlimit },
    { "vfs_timeout", &vfs_timeout },
    { "vfs_use_targz_memlimit", &vfs_use_limit },
#ifdef USE_NETCODE
    { "ftpfs_directory_timeout", &ftpfs_directory_timeout },
    { "use_netrc", &use_netrc },
    { "ftpfs_retry_seconds", &ftpfs_retry_seconds },
    { "ftpfs_always_use_proxy", &ftpfs_always_use_proxy },
    { "ftpfs_use_passive_connections", &ftpfs_use_passive_connections },
    { "ftpfs_use_unix_list_options", &ftpfs_use_passive_connections },
#endif
#endif
    { "nice_rotating_dash", &nice_rotating_dash },
    { "horizontal_split", &horizontal_split },
    { 0, 0 }
};

static void panel_save_setup (WPanel *panel, char *section)
{
    char buffer [40];
    int  i;

    save_string (section, "reverse", sprintf_pd (buffer, panel->reverse), profile_name);
    save_string (section, "case_sensitive", sprintf_pd (buffer, panel->case_sensitive), profile_name);
    for (i = 0; sort_names [i].key; i++)
	if (sort_names [i].sort_type == panel->sort_type) {
	    save_string (section, "sort_order", sort_names [i].key, profile_name);
	    break;
	}

    for (i = 0; list_types [i].key; i++)
	if (list_types [i].list_type == panel->list_type) {
	    save_string (section, PORT_LIST_MODE_NAME, list_types [i].key, profile_name);
	    break;
	}

    save_string (section, "user_format", panel->user_format, profile_name);

    for (i = 0; i < LIST_TYPES; i++) {
	sprintf (buffer, "user_status%d", i);
	save_string (section, buffer, panel->user_status_format [i], profile_name);
    }

    save_string (section, "user_mini_status", sprintf_pd (buffer, panel->user_mini_status), profile_name);
}

static void save_layout (void)
{
    char *profile;
    int  i;
    char buffer [6];

    profile = concat_path_file (home_dir, PROFILE_NAME);

    /* Save integer options */
    for (i = 0; layout [i].opt_name; i++)
	save_string ("Layout", layout [i].opt_name, sprintf_pd (buffer, *layout [i].opt_addr), profile);

    free (profile);
}

static inline int set_int (const char *file, const char *key, int value)
{
    char buffer [30];

    return WritePrivateProfileString (app_text, key, sprintf_pd (buffer, value), file);
}

static void save_configure (void)
{
    char *profile;
    int  i;

    profile = concat_path_file (home_dir, PROFILE_NAME);

    /* Save integer options */
    for (i = 0; options [i].opt_name; i++)
	set_int (profile, options [i].opt_name, *options [i].opt_addr);

    free (profile);
}

static void panel_save_type (char *section, int type)
{
    int i;

    for (i = 0; panel_types [i].opt_name; i++)
	if (panel_types [i].opt_type == type) {
	    save_string (section, "display", panel_types [i].opt_name,
			 profile_name);
	    break;
	}
}

static void save_panel_types (void)
{
    int type;

    type = get_display_type (0);
    panel_save_type ("New Left Panel", type);
    if (type == view_listing)
	panel_save_setup (left_panel, left_panel->panel_name);
    type = get_display_type (1);
    panel_save_type ("New Right Panel", type);
    if (type == view_listing)
	panel_save_setup (right_panel, right_panel->panel_name);
}

static void save_setup (void)
{
    char *profile;

    saving_setup = 1;
    profile = concat_path_file (home_dir, PROFILE_NAME);

    save_layout ();
    save_configure ();
    save_string ("Dirs", "other_dir",
			       get_other_type () == view_listing
			       ? opanel->cwd : ".", profile);
    WritePrivateProfileString ("Dirs", "current_is_left",
			       get_current_index () == 0 ? "1" : "0", profile);
    save_hotlist ();
    save_panelize ();
    save_panel_types ();
/*     directory_history_save (); */

#ifdef USE_VFS
#ifdef USE_NETCODE
    WritePrivateProfileString ("Misc", "ftpfs_password",
			       ftpfs_anonymous_passwd, profile);
    if (ftpfs_proxy_host)
	WritePrivateProfileString ("Misc", "ftp_proxy_host",
				   ftpfs_proxy_host, profile);
#endif
#endif
    free (profile);
    saving_setup = 0;
}

static void panel_load_setup (WPanel *panel, char *section)
{
    int i;
    char buffer [40];

    panel->reverse = load_int (section, "reverse", 0);
    panel->case_sensitive = load_int (section, "case_sensitive", OS_SORT_CASE_SENSITIVE_DEFAULT);

    /* Load sort order */
    load_string (section, "sort_order", "name", buffer, sizeof (buffer));
    panel->sort_type = sort_name;
    for (i = 0; sort_names [i].key; i++)
	if (strcasecmp (sort_names [i].key, buffer) == 0) {
	    panel->sort_type = sort_names [i].sort_type;
	    break;
	}

    /* Load the listing mode */
    load_string (section, PORT_LIST_MODE_NAME, PORT_LIST_MODE_DEFAULT, buffer, sizeof (buffer));
    panel->list_type = list_full;
    for (i = 0; list_types [i].key; i++)
	if (strcasecmp (list_types [i].key, buffer) == 0) {
	    panel->list_type = list_types [i].list_type;
	    break;
	}
    if (panel->list_type == list_icons)
	panel->list_type = list_full;
    /* User formats */
    if (panel->user_format) {
	free (panel->user_format);
	panel->user_format = 0;
    }
    panel->user_format = strdup (get_profile_string (section, "user_format",
						     DEFAULT_USER_FORMAT,
						     profile_name));
    for (i = 0; i < LIST_TYPES; i++) {
	if (panel->user_status_format [i])
	    free (panel->user_status_format [i]);
	sprintf (buffer, "user_status%d", i);
	panel->user_status_format [i] =
	    strdup (get_profile_string (section, buffer,
			DEFAULT_USER_FORMAT, profile_name));
    }

    panel->user_mini_status =
	load_int (section, "user_mini_status", 0);

}

static void load_layout (void)
{
    int i;

    for (i = 0; layout [i].opt_name; i++)
	*layout [i].opt_addr =
	    load_int ("Layout", layout [i].opt_name,
				  *layout [i].opt_addr);
}

static int load_mode (char *section)
{
    char buffer [20];
    int  i;

    int mode = view_listing;

    /* Load the display mode */
    load_string (section, "display", "listing", buffer, sizeof (buffer));

    for (i = 0; panel_types [i].opt_name; i++)
	if (strcasecmp (panel_types [i].opt_name, buffer) == 0) {
	    mode = panel_types [i].opt_type;
	    break;
	}

    return mode;
}

#ifdef USE_NETCODE
static char *do_load_string (char *s, char *ss, char *def)
{
    char *buffer = x_malloc (128, "dls");
    char *p;

    load_string (s, ss, def, buffer, 128);

    p = strdup (buffer);
    free (buffer);
    return p;
}
#endif

static inline int get_int (const char *file, const char *key, int def)
{
    return GetPrivateProfileInt (app_text, key, def, file);
}

static void load_setup (void)
{
    char   *profile;
    char   *inifile;
    int    i;

    load_setup_buffer = concat_path_file (home_dir, PROFILE_NAME);
    inifile = concat_mcpath_file ("mc.ini");
    if (exist_file (load_setup_buffer)) {
	profile = load_setup_buffer;
    } else if (exist_file (inifile)) {
	profile = strdup (inifile);
	free (load_setup_buffer);
    } else {
	profile = load_setup_buffer;
    }
    free (inifile);

    profile_name = profile;

    /* Load integer boolean options */
    for (i = 0; options [i].opt_name; i++)
	*options [i].opt_addr =
	    get_int (profile, options [i].opt_name, *options [i].opt_addr);

    load_layout ();

    load_panelize ();

    startup_left_mode = load_mode ("New Left Panel");
    startup_right_mode = load_mode ("New Right Panel");

    /* At least one of the panels is a listing panel */
    if (startup_left_mode != view_listing && startup_right_mode!=view_listing)
	startup_left_mode = view_listing;

    if (!other_dir) {
	load_setup_buffer = malloc (MC_MAXPATHLEN);
	load_string ("Dirs", "other_dir", ".", load_setup_buffer,
			     MC_MAXPATHLEN);
	if (vfs_file_is_local (load_setup_buffer))
	    other_dir = load_setup_buffer;
	else
	    free (load_setup_buffer);
    }
#ifdef USE_NETCODE
    ftpfs_proxy_host = do_load_string ("Misc", "ftp_proxy_host", "gate");
#endif
    boot_current_is_left =
	GetPrivateProfileInt ("Dirs", "current_is_left", 1, profile);

    load_string ("Misc", "find_ignore_dirs", "", setup_color_string,
		 sizeof (setup_color_string));
    if (setup_color_string [0])
	find_ignore_dirs = xasprintf (":%s:", setup_color_string);

    /* The default color and the terminal dependent color */
    load_string ("Colors", "base_color", "", setup_color_string,
			     sizeof (setup_color_string));
    load_string ("Colors", TERM_env, "",
			     term_color_string, sizeof (term_color_string));
    load_string ("Colors", "color_terminals", "",
			     color_terminal_string, sizeof (color_terminal_string));

    /* Load the directory history */
/*    directory_history_load (); */
    /* Remove the temporal entries */
    profile_clean_section ("Temporal:New Left Panel", profile_name);
    profile_clean_section ("Temporal:New Right Panel", profile_name);
#ifdef USE_VFS
#ifdef USE_NETCODE
    ftpfs_init_passwd ();
#endif
#endif
}

#ifdef USE_VFS
#ifdef USE_NETCODE
static char *load_anon_passwd (void)
{
    char buffer [255];

    load_string ("Misc", "ftpfs_password", "", buffer, sizeof (buffer));
    if (buffer [0])
	return strdup (buffer);
    else
	return 0;
}
#endif
#endif

static void done_setup (void)
{
    free (profile_name);
    done_hotlist ();
    done_panelize ();
/*    directory_history_free (); */
}

/* This table is a mapping between names and the constants we use
 * We use this to allow users to define alternate definitions for
 * certain keys that may be missing from the terminal database
 */
/* Return the code associated with the symbolic name keyname */
static int lookup_key (const char *keyname)
{
    int i;

    for (i = 0; key_name_conv_tab [i].code; i++) {
	if (strcasecmp (key_name_conv_tab [i].name, keyname))
	    continue;
	return key_name_conv_tab [i].code;
    }
    return 0;
}


static void load_keys_from_section (const char *terminal, char *Profile_name)
{
    char *section_name;
    void *profile_keys;
    char *key, *value, *valcopy;
    int  key_code;

    if (!terminal)
	return;

    section_name = xasprintf ("terminal:%s", terminal);
    profile_keys = profile_init_iterator (section_name, Profile_name);
    if (!profile_keys) {
	free (section_name);
	return;
    }

    while (profile_keys) {
	profile_keys = profile_iterator_next (profile_keys, &key, &value);
	key_code = lookup_key (key);
	valcopy = convert_controls (value);
	if (key_code)
	    define_sequence (key_code, valcopy, MCKEY_NOACTION);
	free (valcopy);
    }
    free (section_name);
    return;
}

static void load_key_defs (void)
{
    char *libfile = concat_mcpath_file ("mc.lib");

    load_keys_from_section (TERM_env, profile_name);
    load_keys_from_section ("general", profile_name);

    load_keys_from_section (TERM_env, libfile);
    load_keys_from_section ("general", libfile);

    /* We don't want a huge database loaded in core */
    free_profile_name (libfile);
    free (libfile);
}
/* Slang interface to the Midnight Commander
   This emulates some features of ncurses on top of slang */

/* Taken from S-Lang's slutty.c */

# ifdef _POSIX_VDISABLE
#  define NULL_VALUE _POSIX_VDISABLE
# else
#  define NULL_VALUE 255
# endif

/* Taken from S-Lang's sldisply.c file */

static int SLsys_input_pending(int tsecs)
{
    fd_set Read_FD_Set;
    struct timeval t_wait;

    t_wait.tv_sec = tsecs / 10;
    t_wait.tv_usec = (tsecs % 10) * 100000;

    FD_ZERO(&Read_FD_Set);
    FD_SET(SLang_TT_Read_FD, &Read_FD_Set);

    return select(SLang_TT_Read_FD + 1, &Read_FD_Set, NULL, NULL, &t_wait);
}

static void SLang_exit_error (char *s)
{
    if (s != NULL)
     fprintf (stderr, "%s\n", s);
    exit (-1);
}

#define SLANG_GETKEY_ERROR 0xFFFF

static unsigned int SLsys_getkey (void)
{
    unsigned char c;

    while (1) {
	int ret;

	if (0 == (ret = SLsys_input_pending (100)))
	  continue;

	if (ret != -1)
	  break;

	if (errno == EINTR)
	     continue;

	break;                         /* let read handle it */
    }

    while (-1 == read(SLang_TT_Read_FD, (char *) &c, 1)) {
	if (errno == EINTR)
	     continue;
#ifdef EAGAIN
	if (errno == EAGAIN) {
	     sleep (1);
	     continue;
	}
#endif
#ifdef EWOULDBLOCK
	if (errno == EWOULDBLOCK) {
	     sleep (1);
	     continue;
	}
#endif
#ifdef EIO
	if (errno == EIO)
	     SLang_exit_error ("SLsys_getkey: EIO error.");
#endif
	return SLANG_GETKEY_ERROR;
    }

    return((unsigned int) c);
}

static unsigned int SLang_getkey2 (void)
{
   unsigned int imax;
   unsigned int ch;

   if (SLang_Input_Buffer_Len)
     {
	ch = (unsigned int) *SLang_Input_Buffer;
	SLang_Input_Buffer_Len--;
	imax = SLang_Input_Buffer_Len;

	memcpy ((char *) SLang_Input_Buffer,
		(char *) (SLang_Input_Buffer + 1), imax);
	return(ch);
     }
   else return(SLsys_getkey());
}

static int SLang_input_pending2 (int tsecs)
{
   int n;
   unsigned char c;
   if (SLang_Input_Buffer_Len) return (int) SLang_Input_Buffer_Len;
   n = SLsys_input_pending (tsecs);
   if (n <= 0) return 0;

   c = (unsigned char) SLang_getkey2 ();
   SLang_ungetkey_string (&c, 1);

   return n;
}
/* }}} */

static void
slang_intr (int signo)
{
    slinterrupt = 1;
}

static void
enable_interrupt_key(void)
{
    bb_signal((1 << SIGINT), slang_intr);
    slinterrupt = 0;
}

void
disable_interrupt_key(void)
{
    bb_signal((1 << SIGINT), SIG_IGN);
}

static int
got_interrupt (void)
{
    int t;

    t = slinterrupt;
    slinterrupt = 0;
    return t;
}


#define RGB1(r, g, b)   ((r) | ((g) << 1) | ((b) << 2))
#define RGB(r, g, b, br, bg, bb)  ((RGB1(r, g, b) << 8) | (RGB1(br, bg, bb) << 16))


static void init_Ansi_Color_Map(void)
{
   Ansi_Color_Map[ 0].fgbg = RGB(1, 1, 1, 0, 0, 0); Ansi_Color_Map[ 0].mono = 0x00000000;    /* white/black */
   Ansi_Color_Map[ 1].fgbg = RGB(0, 1, 0, 0, 0, 0); Ansi_Color_Map[ 1].mono = SLTT_REV_MASK; /* green/black */
   Ansi_Color_Map[ 2].fgbg = RGB(1, 0, 1, 0, 0, 0); Ansi_Color_Map[ 2].mono = SLTT_REV_MASK; /* magenta/black */
   Ansi_Color_Map[ 3].fgbg = RGB(0, 1, 1, 0, 0, 0); Ansi_Color_Map[ 3].mono = SLTT_REV_MASK; /* cyan/black */
   Ansi_Color_Map[ 4].fgbg = RGB(1, 0, 0, 0, 0, 0); Ansi_Color_Map[ 4].mono = SLTT_REV_MASK;
   Ansi_Color_Map[ 5].fgbg = RGB(0, 1, 0, 0, 0, 1); Ansi_Color_Map[ 5].mono = SLTT_REV_MASK;
   Ansi_Color_Map[ 6].fgbg = RGB(1, 0, 0, 0, 0, 1); Ansi_Color_Map[ 6].mono = SLTT_REV_MASK;
   Ansi_Color_Map[ 7].fgbg = RGB(1, 0, 0, 0, 1, 0); Ansi_Color_Map[ 7].mono = SLTT_REV_MASK;
   Ansi_Color_Map[ 8].fgbg = RGB(0, 0, 1, 1, 0, 0); Ansi_Color_Map[ 8].mono = SLTT_REV_MASK;
   Ansi_Color_Map[ 9].fgbg = RGB(0, 1, 0, 1, 0, 0); Ansi_Color_Map[ 9].mono = SLTT_REV_MASK;
   Ansi_Color_Map[10].fgbg = RGB(0, 1, 1, 1, 1, 1); Ansi_Color_Map[10].mono = SLTT_REV_MASK;
   Ansi_Color_Map[11].fgbg = RGB(1, 1, 0, 1, 1, 1); Ansi_Color_Map[11].mono = SLTT_REV_MASK;
   Ansi_Color_Map[12].fgbg = RGB(1, 0, 1, 1, 1, 1); Ansi_Color_Map[12].mono = SLTT_REV_MASK;
   Ansi_Color_Map[13].fgbg = RGB(0, 0, 0, 0, 1, 1); Ansi_Color_Map[13].mono = SLTT_REV_MASK;
   Ansi_Color_Map[14].fgbg = RGB(0, 1, 0, 1, 1, 1); Ansi_Color_Map[14].mono = SLTT_REV_MASK;
   Ansi_Color_Map[15].fgbg = RGB(0, 1, 0, 1, 1, 1); Ansi_Color_Map[15].mono = SLTT_REV_MASK;
   Ansi_Color_Map[16].fgbg = RGB(0, 1, 0, 1, 1, 1); Ansi_Color_Map[16].mono = SLTT_REV_MASK;
   Ansi_Color_Map[17].fgbg = RGB(0, 1, 0, 1, 1, 1); Ansi_Color_Map[17].mono = SLTT_REV_MASK;
}

/* Only done the first time */
static void
slang_init (void)
{
    Output_Bufferp = Output_Buffer;
    init_Ansi_Color_Map();
    SLtt_Blink_Mode = 1;
    SLsmg_Display_Eight_Bit = 160;

    /* current attributes --- initialized to impossible value */
    Current_Fgbg = 0xFFFFFFFFU;

    SLtt_get_terminfo ();
    tcgetattr (fileno (stdin), &boot_mode);
    SLang_init_tty ();

    /* If SLang uses fileno(stderr) for terminal input MC will hang
       if we call SLang_getkey between calls to open_error_pipe and
       close_error_pipe, e.g. when we do a growing view of an gzipped
       file. */
    if (SLang_TT_Read_FD == fileno (stderr))
	SLang_TT_Read_FD = fileno (stdin);

    if (force_ugly_line_drawing)
	SLtt_Has_Alt_Charset = 0;
    if (tcgetattr (SLang_TT_Read_FD, &new_mode) == 0) {
#ifdef VDSUSP
	new_mode.c_cc[VDSUSP] = NULL_VALUE;   /* to ignore ^Y */
#endif
#ifdef VLNEXT
	new_mode.c_cc[VLNEXT] = NULL_VALUE;   /* to ignore ^V */
#endif
	tcsetattr (SLang_TT_Read_FD, TCSADRAIN, &new_mode);
    }
    slang_prog_mode ();
    load_terminfo_keys ();
    bb_signal((1 << SIGINT), slang_intr);
    SLtt_Blink_Mode = 0;
}

static void
slang_set_raw_mode (void)
{
    tcsetattr (SLang_TT_Read_FD, TCSANOW, &new_mode);
}

/* Done each time we come back from done mode */
static void
slang_prog_mode (void)
{
    tcsetattr (SLang_TT_Read_FD, TCSANOW, &new_mode);
    SLsmg_init_smg ();
    SLsmg_touch_lines0 (LINES);
}

/* Called each time we want to shutdown slang screen manager */
static void
slang_shell_mode (void)
{
    tcsetattr (SLang_TT_Read_FD, TCSANOW, &boot_mode);
}

static void
slang_shutdown (void)
{
    const char *op_cap;

    slang_shell_mode ();
    do_exit_ca_mode ();
    SLang_reset_tty ();

    /* Load the op capability to reset the colors to those that were
     * active when the program was started up
     */
    op_cap = SLtt_tgetstr ("op");
    if (op_cap) {
	fprintf (stdout, "%s", op_cap);
	fflush (stdout);
    }
}

/* HP Terminals have capabilities (pfkey, pfloc, pfx) to program function keys.
   elm 2.4pl15 invoked with the -K option utilizes these softkeys and the
   consequence is that function keys don't work in MC sometimes.
   Unfortunately I don't now the one and only escape sequence to turn off
   softkeys (elm uses three different capabilities to turn on softkeys and two
   capabilities to turn them off).
   Among other things elm uses the pair we already use in slang_keypad. That's
   the reason why I call slang_reset_softkeys from slang_keypad. In lack of
   something better the softkeys are programmed to their defaults from the
   termcap/terminfo database.
   The escape sequence to program the softkeys is taken from elm and it is
   hardcoded because neither slang nor ncurses 4.1 know how to 'printf' this
   sequence. -- Norbert
 */

static void
slang_reset_softkeys (void)
{
    int key;
    const char *Send;
    char *display = "                ";
    char tmp[100];

    for ( key = 1; key < 9; key++ ) {
	sprintf ( tmp, "k%d", key);
	Send = SLtt_tgetstr (tmp);
	if (Send) {
	    sprintf(tmp, "\033&f%dk%dd%dL%s%s", key,
		    (int)strlen(display), (int)strlen(Send), display, Send);
	    SLtt_write_string (tmp);
	}
    }
}

/* keypad routines */
static stdcall2 void
slang_keypad (int set)
{
    const char *keypad_string;

    keypad_string = SLtt_tgetstr (set ? "ks" : "ke");
    if (keypad_string)
	SLtt_write_string (keypad_string);
    if (set && reset_hp_softkeys)
	slang_reset_softkeys ();
}

static stdcall2 void
hline (int ch, int len)
{
    int last_x, last_y;

    last_x = SLsmg_get_column ();
    last_y = SLsmg_get_row ();

    if (ch == 0)
	ch = ACS_HLINE;

    if (ch == ACS_HLINE) {
	SLsmg_draw_hline (len);
    } else {
	while (len--)
	    addch (ch);
    }
    move (last_y, last_x);
}

static stdcall2 void
vline (int character, int len)
{
    if (!slow_terminal) {
	SLsmg_draw_vline (len);
    } else {
	int last_x, last_y, pos = 0;

	last_x = SLsmg_get_column ();
	last_y = SLsmg_get_row ();

	while (len--) {
	    move (last_y + pos++, last_x);
	    addch (' ');
	}
	move (last_x, last_y);
    }
}

static const char * const color_terminals [] = {
    "console",
    "linux",
    "xterm-color",
    "color-xterm",
    "dtterm",
    "xterm",
    "ansi",
/* 'qnx*' terminals have non-ANSI-compatible color sequences... */
    "qansi",
    "qansi-g",
    "qansi-m",
    "qansi-t",
    "qansi-w",
    0
};

static int has_colors (void)
{
    const char *terminal = TERM_env;
    char *cts = color_terminal_string, *s;
    int  i;

    SLtt_Use_Ansi_Colors = force_colors;
    if (NULL != getenv ("COLORTERM"))
	SLtt_Use_Ansi_Colors = 1;

    /* We want to allow overwriding */
    if (!disable_colors) {
      if (!*cts) {
      /* check hard-coded terminals */
	for (i = 0; color_terminals [i]; i++)
	    if (strcmp (color_terminals [i], terminal) == 0)
		SLtt_Use_Ansi_Colors = 1;
      } else {
	/* check color_terminal_string */
	while (*cts) {
	  while (*cts == ' ' || *cts == '\t') cts++;

	  s = cts;
	  i = 0;

	  while (*cts && *cts != ',') {
	    cts++;
	    i++;
	  }
	  if (i && i == strlen(terminal) && strncmp(s, terminal, i) == 0)
	    SLtt_Use_Ansi_Colors = 1;

	  if (*cts == ',') cts++;
	}
      }
    }

    /* Setup emulated colors */
    if (SLtt_Use_Ansi_Colors) {
	if (use_colors) {
	    SLtt_set_color (A_REVERSE, "black", "white");
	    SLtt_set_color (A_BOLD, "white", "black");
	} else {
	    SLtt_set_color (A_REVERSE, "black", "lightgray");
	    SLtt_set_color (A_BOLD, "white", "black");
	    SLtt_set_color (A_BOLD_REVERSE, "white", "lightgray");
	}
    } else {
	SLtt_set_mono (A_BOLD, SLTT_BOLD_MASK);
	SLtt_set_mono (A_REVERSE, SLTT_REV_MASK);
	SLtt_set_mono (A_BOLD|A_REVERSE, SLTT_BOLD_MASK | SLTT_REV_MASK);
    }
    return SLtt_Use_Ansi_Colors;
}

static stdcall2 void
attrset (int color)
{
    if (!SLtt_Use_Ansi_Colors) {
	SLsmg_set_color (color);
	return;
    }

    if (color & A_BOLD) {
	if (color == A_BOLD)
	    SLsmg_set_color (A_BOLD);
	else
	    SLsmg_set_color ((color & (~A_BOLD)) + 8);
	return;
    }

    if (color == A_REVERSE)
	SLsmg_set_color (A_REVERSE);
    else
	SLsmg_set_color (color);
}

/* This table describes which capabilities we want and which values we
 * assign to them.
 */
static const struct {
    int  key_code;
    const char *key_name;
} key_table [] = {
    { KEY_F(0),  "k0" },
    { KEY_F(1),  "k1" },
    { KEY_F(2),  "k2" },
    { KEY_F(3),  "k3" },
    { KEY_F(4),  "k4" },
    { KEY_F(5),  "k5" },
    { KEY_F(6),  "k6" },
    { KEY_F(7),  "k7" },
    { KEY_F(8),  "k8" },
    { KEY_F(9),  "k9" },
    { KEY_F(10), "k;" },
    { KEY_F(11), "F1" },
    { KEY_F(12), "F2" },
    { KEY_F(13), "F3" },
    { KEY_F(14), "F4" },
    { KEY_F(15), "F5" },
    { KEY_F(16), "F6" },
    { KEY_F(17), "F7" },
    { KEY_F(18), "F8" },
    { KEY_F(19), "F9" },
    { KEY_F(20), "FA" },
    { KEY_IC,    "kI" },
    { KEY_NPAGE, "kN" },
    { KEY_PPAGE, "kP" },
    { KEY_LEFT,  "kl" },
    { KEY_RIGHT, "kr" },
    { KEY_UP,    "ku" },
    { KEY_DOWN,  "kd" },
    { KEY_DC,    "kD" },
    { KEY_BACKSPACE, "kb" },
    { KEY_HOME,  "kh" },
    { KEY_END,   "@7" },
    { 0, 0}
};

static void
do_define_key (int code, const char *strcap)
{
    char    *seq;

    seq = (char *) SLtt_tgetstr (strcap);
    if (seq)
	define_sequence (code, seq, MCKEY_NOACTION);
}

static void
load_terminfo_keys (void)
{
    int i;

    for (i = 0; key_table [i].key_code; i++)
	do_define_key (key_table [i].key_code, key_table [i].key_name);
}

static int getch (void)
{
    if (no_slang_delay)
	if (SLang_input_pending2 (0) == 0)
	    return -1;

    return (SLang_getkey2 ());
}

/* Term name pollution avoiding. */

/* The following routines only work on xterm terminals */

static void do_enter_ca_mode (void)
{
    if (!xterm_flag)
	return;
    fprintf (stdout, /* ESC_STR ")0" */ ESC_STR "7" ESC_STR "[?47h");
    fflush (stdout);
}

static void do_exit_ca_mode (void)
{
    if (!xterm_flag)
	return;
    fprintf (stdout, ESC_STR "[?47l" ESC_STR "8" ESC_STR "[m");
    fflush (stdout);
}

static void
edition_post_exec (void)
{
    do_enter_ca_mode ();

    /* FIXME: Missing on slang endwin? */
    reset_prog_mode ();

    keypad (TRUE);
    raw ();
    init_mouse ();
    if (alternate_plus_minus)
	application_keypad_mode ();
}

static void
clr_scr (void)
{
    standend ();
    dlg_erase (midnight_dlg);
    mc_refresh ();
}

/* Directory tree browser for the Midnight Commander */

/*
   This module has been converted to be a widget.

   The program load and saves the tree each time the tree widget is
   created and destroyed.  This is required for the future vfs layer,
   it will be possible to have tree views over virtual file systems.
   */

#define TREE_NORMALC HOT_FOCUSC

/* Forwards */
static int tree_callback (Dlg_head *h, WTree *tree, int msg, int par);
#define tcallback (callback_fn) tree_callback

/* Returns number of common characters */
static inline int str_common (char *s1, char *s2)
{
    int result = 0;

    while (*s1++ == *s2++)
	result++;
    return result;
}

static tree_entry *back_ptr (tree_entry *ptr, int *Count)
{
    int i = 0;

    while (ptr && ptr->prev && i < *Count) {
	ptr = ptr->prev;
	i ++;
    }
    *Count = i;
    return ptr;
}

static tree_entry *forw_ptr (tree_entry *ptr, int *Count)
{
    int i = 0;

    while (ptr && ptr->next && i < *Count) {
	ptr = ptr->next;
	i ++;
    }
    *Count = i;
    return ptr;
}

/* The directory names are arranged in a single linked list in the same
   order as they are displayed. When the tree is displayed the expected
   order is like this:
	/
	/bin
	/etc
	/etc/X11
	/etc/rc.d
	/etc.old/X11
	/etc.old/rc.d
	/usr

   i.e. the required collating sequence when comparing two directory names is
	'\0' < '/' < all-other-characters-in-encoding-order

    Since strcmp doesn't fulfil this requirement we use pathcmp when
    inserting directory names into the list. The meaning of the return value
    of pathcmp and strcmp are the same (an integer less than, equal to, or
    greater than zero if p1 is found to be less than, to match, or be greater
    than p2.
 */
static int
pathcmp (const char *p1, const char *p2)
{
    for ( ; *p1 == *p2; p1++, p2++)
	if (*p1 == '\0' )
	    return 0;

    if (*p1 == '\0')
	return -1;
    if (*p2 == '\0')
	return 1;
    if (*p1 == '/')
	return -1;
    if (*p2 == '/')
	return 1;
    return (*p1 - *p2);
}

/* Searches for specified directory */
static tree_entry *whereis (WTree *tree, char *name)
{
    tree_entry *current = tree->tree_first;
    int flag = -1;

    while (current && (flag = pathcmp (current->name, name)) < 0)
	current = current->next;

    if (flag == 0)
	return current;
    else
	return NULL;
}

/* Add a directory to the list of directories */
tree_entry *tree_add_entry (WTree *tree, char *name)
{
    int flag = -1;
    tree_entry *current = tree->tree_first;
    tree_entry *old = NULL;
    tree_entry *new;
    int i, len;
    int submask = 0;

    if (!tree)
	return 0;

    if (tree->tree_last && tree->tree_last->next)
	abort ();
    /* Search for the correct place */
    while (current && (flag = pathcmp (current->name, name)) < 0) {
	old = current;
	current = current->next;
    }

    if (flag == 0)
	return current; /* Already in the list */

    /* Not in the list -> add it */
    new = x_malloc (sizeof (tree_entry), "tree, tree_entry");
    if (!current) {
	/* Append to the end of the list */
	if (!tree->tree_first) {
	    /* Empty list */
	    tree->tree_first = new;
	    new->prev = NULL;
	} else {
	    old->next = new;
	    new->prev = old;
	}
	new->next = NULL;
	tree->tree_last = new;
    } else {
	/* Insert in to the middle of the list */
	new->prev = old;
	if (old) {
	    /* Yes, in the middle */
	    new->next = old->next;
	    old->next = new;
	} else {
	    /* Nope, in the beginning of the list */
	    new->next = tree->tree_first;
	    tree->tree_first = new;
	}
	new->next->prev = new;
    }
    /* tree_count++; */

    /* Calculate attributes */
    new->name = strdup (name);
    len = strlen (new->name);
    new->sublevel = 0;
    for (i = 0; i < len; i++)
	if (new->name [i] == '/') {
	    new->sublevel++;
	    new->subname = new->name + i + 1;
	}
    if (new->next)
	submask = new->next->submask;
    else
	submask = 0;
    submask |= 1 << new->sublevel;
    submask &= (2 << new->sublevel) - 1;
    new->submask = submask;
    new->mark = 0;

    /* Correct the submasks of the previous entries */
    current = new->prev;
    while (current && current->sublevel > new->sublevel) {
	current->submask |= 1 << new->sublevel;
	current = current->prev;
    }

    /* The entry has now been added */

    if (new->sublevel > 1) {
	/* Let's check if the parent directory is in the tree */
	char *parent = strdup (new->name);

	for (i = strlen (parent) - 1; i > 1; i--) {
	    if (parent [i] == '/') {
		parent [i] = 0;
		tree_add_entry (tree, parent);
		break;
	    }
	}
	free (parent);
    }

    return new;
}

static void remove_entry (WTree *tree, tree_entry *entry)
{
    tree_entry *current = entry->prev;
    long submask = 0;

    if (tree->selected_ptr == entry) {
	if (tree->selected_ptr->next)
	    tree->selected_ptr = tree->selected_ptr->next;
	else
	    tree->selected_ptr = tree->selected_ptr->prev;
    }

    /* Correct the submasks of the previous entries */
    if (entry->next)
	submask = entry->next->submask;
    while (current && current->sublevel > entry->sublevel) {
	submask |= 1 << current->sublevel;
	submask &= (2 << current->sublevel) - 1;
	current->submask = submask;
	current = current->prev;
    }

    /* Unlink the entry from the list */
    if (entry->prev)
	entry->prev->next = entry->next;
    else
	tree->tree_first = entry->next;
    if (entry->next)
	entry->next->prev = entry->prev;
    else
	tree->tree_last = entry->prev;
    /* tree_count--; */

    /* Free the memory used by the entry */
    free (entry->name);
    free (entry);
}

void tree_remove_entry (WTree *tree, char *name)
{
    tree_entry *current, *base, *old;
    int len;
#if 0
    int base_sublevel;
#endif

    /* Miguel Ugly hack */
    if (name [0] == '/' && name [1] == 0)
	return;
    /* Miguel Ugly hack end */

    base = whereis (tree, name);
    if (!base)
	return; /* Doesn't exist */
#if 0
    if (tree->check_name [0] == '/' && tree->check_name [1] == 0)
	base_sublevel = base->sublevel;
    else
	base_sublevel = base->sublevel + 1;
#endif
    len = strlen (base->name);
    current = base->next;
    while (current
	   && strncmp (current->name, base->name, len) == 0
	   && (current->name[len] == '\0' || current->name[len] == '/')) {
	old = current;
	current = current->next;
	remove_entry (tree, old);
    }
    remove_entry (tree, base);
}

void tree_destroy (WTree *tree)
{
    tree_entry *current, *old;

    save_tree (tree);
    current = tree->tree_first;
    while (current) {
	old = current;
	current = current->next;
	free (old->name);
	free (old);
    }
    if (tree->tree_shown) {
	free (tree->tree_shown);
	tree->tree_shown = 0;
    }
    tree->selected_ptr = tree->tree_first = tree->tree_last = NULL;
}

/* Mark the subdirectories of the current directory for delete */
void start_tree_check (WTree *tree)
{
    tree_entry *current;
    int len;

    if (!tree)
	tree = (WTree *) find_widget_type (current_dlg, tcallback);
    if (!tree)
	return;

    /* Search for the start of subdirectories */
    mc_get_current_wd (tree->check_name, MC_MAXPATHLEN);
    tree->check_start = NULL;
    current = whereis (tree, tree->check_name);
    if (!current) {
	/* Cwd doesn't exist -> add it */
	current = tree_add_entry (tree, tree->check_name);
	return;
    }

    /* Mark old subdirectories for delete */
    tree->check_start = current->next;
    len = strlen (tree->check_name);

    current = tree->check_start;
    while (current
	   && strncmp (current->name, tree->check_name, len) == 0
	   && (current->name[len] == '\0' || current->name[len] == '/' || len == 1)) {
	current->mark = 1;
	current = current->next;
    }
}

/* This subdirectory exists -> clear deletion mark */
void do_tree_check (WTree *tree, const char *subname)
{
    char *name;
    tree_entry *current, *base;
    int flag = 1, len;

    /* Calculate the full name of the subdirectory */
    if (subname [0] == '.' &&
	(subname [1] == 0 || (subname [1] == '.' && subname [2] == 0)))
	return;
    if (tree->check_name [0] == '/' && tree->check_name [1] == 0)
	name = xasprintf ("/%s", subname);
    else
	name = concat_path_file (tree->check_name, subname);

    /* Search for the subdirectory */
    current = tree->check_start;
    while (current && (flag = pathcmp (current->name, name)) < 0)
	current = current->next;

    if (flag != 0)
	/* Doesn't exist -> add it */
	current = tree_add_entry (tree, name);
    free (name);

    /* Clear the deletion mark from the subdirectory and its children */
    base = current;
    if (base) {
	len = strlen (base->name);
	base->mark = 0;
	current = base->next;
	while (current
	       && strncmp (current->name, base->name, len) == 0
	       && (current->name[len] == '\0' || current->name[len] == '/' || len == 1)) {
	    current->mark = 0;
	    current = current->next;
	}
    }
}

/* Tree check searchs a tree widget in the current dialog and
 * if it finds it, it calls do_tree_check on the subname
 */
void tree_check (const char *subname)
{
    WTree *tree;

    tree = (WTree *) find_widget_type (current_dlg, tcallback);
    if (!tree)
	return;
    do_tree_check (tree, subname);
}


/* Delete subdirectories which still have the deletion mark */
void end_tree_check (WTree *tree)
{
    tree_entry *current, *old;
    int len;

    if (!tree)
	tree = (WTree *) find_widget_type (current_dlg, tcallback);
    if (!tree)
	return;

    /* Check delete marks and delete if found */
    len = strlen (tree->check_name);

    current = tree->check_start;
    while (current
	   && strncmp (current->name, tree->check_name, len) == 0
	   && (current->name[len] == '\0' || current->name[len] == '/' || len == 1)) {
	old = current;
	current = current->next;
	if (old->mark)
	    remove_entry (tree, old);
    }
}

/* Loads the .mc.tree file */
void load_tree (WTree *tree)
{
    char *filename;
    FILE *file;
    char name [MC_MAXPATHLEN], oldname[MC_MAXPATHLEN];
    char *different;
    int len, common;

    filename = concat_path_file (home_dir, MC_TREE);
    file = fopen (filename, "r");
    free (filename);
    if (!file) {
	/* No new tree file -> let's try the old file */
	filename = concat_path_file (home_dir, MC_TREE);
	file = fopen (filename, "r");
	free (filename);
    }

    if (file) {
	/* File open -> read contents */
	oldname [0] = 0;
	while (fgets (name, MC_MAXPATHLEN, file)) {
	    len = strlen (name);
	    if (name [len - 1] == '\n') {
		name [--len] = 0;
	    }
#if defined(__CYGWIN__)
	    /* .ado: Drives for NT and OS/2 */
	    if ((len > 2)         &&
		 isalpha(name[0]) &&
		 (name[1] == ':') &&
		 (name[2] == '\\')) {
			tree_add_entry (tree, name);
			strcpy (oldname, name);
	    } else
#endif
	    /* UNIX Version */
	    if (name [0] != '/') {
		/* Clear-text decompression */
		char *s = strtok (name, " ");

		if (s) {
		    common = atoi (s);
		    different = strtok (NULL, "");
		    if (different) {
			strcpy (oldname + common, different);
			tree_add_entry (tree, oldname);
		    }
		}
	    } else {
		tree_add_entry (tree, name);
		strcpy (oldname, name);
	    }
	}
	fclose (file);
    }
    if (!tree->tree_first) {
	/* Nothing loaded -> let's add some standard directories */
	tree_add_entry (tree, "/");
	tree->selected_ptr = tree->tree_first;
	tree_rescan_cmd (tree);
	tree_add_entry (tree, home_dir);
	tree_chdir (tree, home_dir);
	tree_rescan_cmd (tree);
    }
}

/* Save the .mc.tree file */
void save_tree (WTree *tree)
{
    tree_entry *current;
    char *filename;
    FILE *file;
    int i, common;

    filename = concat_path_file (home_dir, MC_TREE);
    file = bb_xfopen (filename, "Ww");
    free (filename);
    if (!file)
	return;

    current = tree->tree_first;
    while (current) {
	if (current->prev && (common = str_common (current->prev->name, current->name)) > 2)
	    /* Clear-text compression */
	    i = fprintf (file, "%d %s\n", common, current->name + common);
	else
	    i = fprintf (file, "%s\n", current->name);
	if (i == EOF) {
	    fprintf (stderr, "Can't write to the %s file:\n%m\n", MC_TREE);
	    break;
	}
	current = current->next;
    }
    fclose (file);
}

static void tree_show_mini_info (WTree *tree, int tree_lines, int tree_cols)
{
    Dlg_head *h = tree->widget.parent;
    int      line;

    /* Show mini info */
    if (tree->is_panel) {
	if (!show_mini_info)
	    return;
	line = tree_lines+2;
    } else
	line = tree_lines+1;

    widget_move (&tree->widget, line, 1);
    hline (' ', tree_cols);
    widget_move (&tree->widget, line, 1);

    if (tree->searching) {
	/* Show search string */
	attrset (TREE_NORMALC);
	attrset (FOCUSC);
	addch ('/');

	addstr (name_trunc (tree->search_buffer, tree_cols-2));
	addch (' ');
	attrset (FOCUSC);
    } else {
	/* Show full name of selected directory */
	addstr (name_trunc (tree->selected_ptr->name, tree_cols));
    }
}

void show_tree (WTree *tree)
{
    Dlg_head *h = tree->widget.parent;
    tree_entry *current;
    int i, j, topsublevel;
    int x, y;
    int tree_lines, tree_cols;

    /* Initialize */
    x = y = 0;
    tree_lines = tlines (tree);
    tree_cols  = tree->widget.cols;

    attrset (TREE_NORMALC);
    widget_move ((Widget*)tree, y, x);
    if (tree->is_panel) {
	tree_cols  -= 2;
	x = y = 1;
    }

    if (tree->tree_shown)
	free (tree->tree_shown);
    tree->tree_shown = (tree_entry**)x_malloc (sizeof (tree_entry*)*tree_lines,
					      "tree, show_tree");
    for (i = 0; i < tree_lines; i++)
	tree->tree_shown [i] = NULL;
    if (tree->tree_first)
	topsublevel = tree->tree_first->sublevel;
    else
	topsublevel = 0;
    if (!tree->selected_ptr) {
	tree->selected_ptr = tree->tree_first;
	tree->topdiff = 0;
    }
    current = tree->selected_ptr;

    /* Calculate the directory which is to be shown on the topmost line */
    if (tree_navigation_flag) {
	i = 0;
	while (current->prev && i < tree->topdiff) {
	    current = current->prev;
	    if (current->sublevel < tree->selected_ptr->sublevel) {
		if (strncmp (current->name, tree->selected_ptr->name,
			     strlen (current->name)) == 0)
		    i++;
	    } else if (current->sublevel == tree->selected_ptr->sublevel) {
		for (j = strlen (current->name) - 1; current->name [j] != '/'; j--);
		if (strncmp (current->name, tree->selected_ptr->name, j) == 0)
		    i++;
	    } else if (current->sublevel == tree->selected_ptr->sublevel + 1
		       && strlen (tree->selected_ptr->name) > 1) {
		if (strncmp (current->name, tree->selected_ptr->name,
			     strlen (tree->selected_ptr->name)) == 0)
		    i++;
	    }
	}
	tree->topdiff = i;
    } else
	current = back_ptr (current, &tree->topdiff);

    /* Loop for every line */
    for (i = 0; i < tree_lines; i++) {
	/* Move to the beginning of the line */
	widget_move (&tree->widget, y+i, x);

	hline (' ', tree_cols);
	widget_move (&tree->widget, y+i, x);

	if (!current)
	    continue;

	tree->tree_shown [i] = current;
	if (current->sublevel == topsublevel) {

	    /* Top level directory */
	    if (tree->active && current == tree->selected_ptr) {
		if (!use_colors && !tree->is_panel)
			attrset (MARKED_COLOR);
		else
			attrset (SELECTED_COLOR);
	    }
	    /* Show full name */
	    addstr (name_trunc (current->name, tree_cols - 6));
	} else{
	    /* Sub level directory */

	    SLsmg_set_char_set(1);
	    /* Output branch parts */
	    for (j = 0; j < current->sublevel - topsublevel - 1; j++) {
		if (tree_cols - 8 - 3 * j < 9)
		    break;
		addch (' ');
		if (current->submask & (1 << (j + topsublevel + 1)))
		    addch (ACS_VLINE);
		else
		    addch (' ');
		addch (' ');
	    }
	    addch (' '); j++;
	    if (!current->next || !(current->next->submask & (1 << current->sublevel)))
		addch (ACS_LLCORNER);
	    else
		addch (ACS_LTEE);
	    addch (ACS_HLINE);
	    SLsmg_set_char_set (0);

	    if (tree->active && current == tree->selected_ptr) {
		/* Selected directory -> change color */
		if (!use_colors && !tree->is_panel)
		    attrset (MARKED_COLOR);
		else
		    attrset (SELECTED_COLOR);
	    }
	    /* Show sub-name */
	    addch (' ');
	    addstr (name_trunc (current->subname,
				tree_cols - 2 - 4 - 3 * j));
	}
	addch (' ');

	/* Return to normal color */
	attrset (TREE_NORMALC);

	/* Calculate the next value for current */
	if (tree_navigation_flag) {
	    current = current->next;
	    while (current) {
		if (current->sublevel < tree->selected_ptr->sublevel) {
		    if (strncmp (current->name, tree->selected_ptr->name,
				 strlen (current->name)) == 0)
			break;
		} else if (current->sublevel == tree->selected_ptr->sublevel) {
		    for (j = strlen (current->name) - 1; current->name [j] != '/'; j--);
		    if (strncmp (current->name,tree->selected_ptr->name,j)== 0)
			break;
		} else if (current->sublevel == tree->selected_ptr->sublevel+1
			   && strlen (tree->selected_ptr->name) > 1) {
		    if (strncmp (current->name, tree->selected_ptr->name,
				 strlen (tree->selected_ptr->name)) == 0)
			break;
		}
		current = current->next;
	    }
	} else
	    current = current->next;
    }
    tree_show_mini_info (tree, tree_lines, tree_cols);
}

static void check_focus (WTree *tree)
{
    if (tree->topdiff < 3)
	tree->topdiff = 3;
    else if (tree->topdiff >= tlines (tree) - 3)
	tree->topdiff = tlines (tree) - 3 - 1;
}

void tree_move_backward (WTree *tree, int i)
{
    tree_entry *current;
    int j = 0;

    if (tree_navigation_flag) {
	current = tree->selected_ptr;
	while (j < i && current->prev
	       && current->prev->sublevel >= tree->selected_ptr->sublevel) {
	    current = current->prev;
	    if (current->sublevel == tree->selected_ptr->sublevel) {
		tree->selected_ptr = current;
		j ++;
	    }
	}
	i = j;
    } else
	tree->selected_ptr = back_ptr (tree->selected_ptr, &i);
    tree->topdiff -= i;
    check_focus (tree);
}

void tree_move_forward (WTree *tree, int i)
{
    tree_entry *current;
    int j = 0;

    if (tree_navigation_flag) {
	current = tree->selected_ptr;
	while (j < i && current->next
	       && current->next->sublevel >= tree->selected_ptr->sublevel) {
	    current = current->next;
	    if (current->sublevel == tree->selected_ptr->sublevel) {
		tree->selected_ptr = current;
		j ++;
	    }
	}
	i = j;
    } else
	tree->selected_ptr = forw_ptr (tree->selected_ptr, &i);
    tree->topdiff += i;
    check_focus (tree);
}

void tree_move_to_child (WTree *tree)
{
    tree_entry *current;

    /* Do we have a starting point? */
    if (!tree->selected_ptr)
	return;
    /* Take the next entry */
    current = tree->selected_ptr->next;
    /* Is it the child of the selected entry */
    if (current && current->sublevel > tree->selected_ptr->sublevel) {
	/* Yes -> select this entry */
	tree->selected_ptr = current;
	tree->topdiff++;
	check_focus (tree);
    } else {
	/* No -> rescan and try again */
	tree_rescan_cmd (tree);
	current = tree->selected_ptr->next;
	if (current && current->sublevel > tree->selected_ptr->sublevel) {
	    tree->selected_ptr = current;
	    tree->topdiff++;
	    check_focus (tree);
	}
    }
}

int tree_move_to_parent (WTree *tree)
{
    tree_entry *current;
    tree_entry *old;

    if (!tree->selected_ptr)
	return 0;
    old = tree->selected_ptr;
    current = tree->selected_ptr->prev;
    while (current && current->sublevel >= tree->selected_ptr->sublevel) {
	current = current->prev;
	tree->topdiff--;
    }
    if (!current)
	current = tree->tree_first;
    tree->selected_ptr = current;
    check_focus (tree);
    return tree->selected_ptr != old;
}

void tree_move_to_top (WTree *tree)
{
    tree->selected_ptr = tree->tree_first;
    tree->topdiff = 0;
}

void tree_move_to_bottom (WTree *tree)
{
    tree->selected_ptr = tree->tree_last;
    tree->topdiff = tlines (tree) - 3 - 1;
}

void tree_chdir (WTree *tree, char *dir)
{
    tree_entry *current;

    current = whereis (tree, dir);
    if (current) {
	tree->selected_ptr = current;
	check_focus (tree);
    }
}

/* Handle mouse click */
void tree_event (WTree *tree, int y)
{
    if (tree->tree_shown [y]) {
	tree->selected_ptr = tree->tree_shown [y];
	tree->topdiff = y;
    }
    show_tree (tree);
}

static void chdir_sel (WTree *tree);

static void maybe_chdir (WTree *tree)
{
    if (!(xtree_mode && tree->is_panel))
	return;
    if (is_idle ())
	chdir_sel (tree);
}

/* Mouse callback */
static int event_callback (Gpm_Event *event, WTree *tree)
{
    if (!(event->type & GPM_UP))
	return MOU_ENDLOOP;

    if (tree->is_panel)
	event->y--;

    event->y--;

    if (!tree->active)
	change_panel ();

    if (event->y < 0) {
	tree_move_backward (tree, tlines (tree) - 1);
	show_tree (tree);
    }
    else if (event->y >= tlines (tree)) {
	tree_move_forward (tree, tlines (tree) - 1);
	show_tree (tree);
    } else {
	tree_event (tree, event->y);
	if ((event->type & (GPM_UP|GPM_DOUBLE)) == (GPM_UP|GPM_DOUBLE)) {
	    chdir_sel (tree);
	}
    }
    return MOU_ENDLOOP;
}

/* Search tree for text */
int search_tree (WTree *tree, char *text)
{
    tree_entry *current;
    int len;
    int wrapped = 0;
    int found = 0;

    len = strlen (text);
    current = tree->selected_ptr;
    found = 0;
    while (!wrapped || current != tree->selected_ptr) {
	if (strncmp (current->subname, text, len) == 0) {
	    tree->selected_ptr = current;
	    found = 1;
	    break;
	}
	current = current->next;
	if (!current) {
	    current = tree->tree_first;
	    wrapped = 1;
	}
	tree->topdiff++;
    }
    check_focus (tree);
    return found;
}

static void tree_do_search (WTree *tree, int key)
{
    int l;

    l = strlen (tree->search_buffer);
    if (l && (key == 8 || key == 0177 || key == KEY_BACKSPACE))
	tree->search_buffer [--l] = 0;
    else {
	if (key && l < sizeof (tree->search_buffer)) {
	    tree->search_buffer [l] = key;
	    tree->search_buffer [l+1] = 0;
	    l++;
	}
    }

    if (!search_tree (tree, tree->search_buffer))
	tree->search_buffer [--l] = 0;

    show_tree (tree);
    maybe_chdir (tree);
}

void tree_rescan_cmd (WTree *tree)
{
    DIR *dirp;
    struct dirent *dp;
    char Old_dir [MC_MAXPATHLEN];

    if (!tree->selected_ptr || !mc_get_current_wd (Old_dir, MC_MAXPATHLEN) ||
	mc_chdir (tree->selected_ptr->name))
	return;

    start_tree_check (tree);
    dirp = opendir (".");
    if (dirp) {
	for (dp = readdir (dirp); dp; dp = readdir (dirp)) {
	    if (is_directory (dp->d_name, 0, NULL) > 0)
		do_tree_check (tree, dp->d_name);
	}
	closedir (dirp);
    }
    end_tree_check (tree);
    mc_chdir (Old_dir);
}

int tree_forget_cmd (WTree *tree)
{
    if (tree->selected_ptr)
	tree_remove_entry (tree, tree->selected_ptr->name);
    return 1;
}

void tree_copy (WTree *tree, char *default_dest)
{
    char *dest;

    if (!tree->selected_ptr)
	return;
    sprintf (cmd_buf, "Copy \"%s\" directory to:",
	     name_trunc (tree->selected_ptr->name, 50));
    dest = input_expand_dialog (" Copy ", cmd_buf, default_dest);
    if (!dest || !*dest) {
	return;
    }
    create_op_win (OP_COPY, 0);
    file_mask_defaults ();
    copy_dir_dir (tree->selected_ptr->name, dest, 1, 0, 0, 0);
    destroy_op_win ();
    free (dest);
}

static void tree_help_cmd (void)
{
    char *hlpfile = concat_mcpath_file ("mc.hlp");
    interactive_display (hlpfile,  "[Directory Tree]");
    free (hlpfile);
}

static int tree_copy_cmd (WTree *tree)
{
    tree_copy (tree, "");
    return 1;
}

void tree_move (WTree *tree, char *default_dest)
{
    char *dest;

    if (!tree->selected_ptr)
	return;
    sprintf (cmd_buf, "Move \"%s\" directory to:",
	     name_trunc (tree->selected_ptr->name, 50));
    dest = input_expand_dialog (" Move ", cmd_buf, default_dest);
    if (!dest || !*dest) {
	return;
    }
    if (is_directory(dest, 1, NULL) <= 0) {
	MC_message (1, " Error ", " The destination \n %m ");
	free (dest);
	return;
    }
    create_op_win (OP_MOVE, 0);
    file_mask_defaults ();
    move_dir_dir (tree->selected_ptr->name, dest);
    destroy_op_win ();
    free (dest);
}

static int tree_move_cmd (WTree *tree)
{
    tree_move (tree, "");
    return 1;
}

static void tree_rmdir_cmd (WTree *tree)
{
    char Old_dir [MC_MAXPATHLEN];

    if (tree->selected_ptr) {
	if (!mc_get_current_wd (Old_dir, MC_MAXPATHLEN))
	    return;
	if (mc_chdir ("/"))
	    return;
	if (confirm_delete) {
	    char *Cmd_buf;
	    int result;

	    Cmd_buf = xasprintf("  Delete %s?  ", tree->selected_ptr->name);
	    result = query_dialog (" Delete ", Cmd_buf, 3, 2, "&Yes", "&No");
	    free (Cmd_buf);
	    if (result != 0)
		return;
	}
	create_op_win (OP_DELETE, 0);
	if (erase_dir (tree->selected_ptr->name) == FILE_CONT)
	    tree_forget_cmd (tree);
	destroy_op_win ();
	mc_chdir (Old_dir);
	return;
    } else
	return;
}

static void set_navig_label (Dlg_head *h);
static void tree_toggle_navig (Dlg_head *h)
{
    tree_navigation_flag = 1 - tree_navigation_flag;
    set_navig_label (h);
}

void set_navig_label (Dlg_head *h)
{
    define_label_data (h, (Widget *)Tree,
		       4, tree_navigation_flag ? "Static" : "Dynamc",
		       (void (*)(void *))tree_toggle_navig, h);
}

static void move_down (WTree *tree)
{
    tree_move_forward (tree, 1);
    show_tree (tree);
    maybe_chdir (tree);
}

static void move_up (WTree *tree)
{
    tree_move_backward (tree, 1);
    show_tree (tree);
    maybe_chdir (tree);
}

static void move_home (WTree *tree)
{
    tree_move_to_top (tree);
    show_tree (tree);
    maybe_chdir (tree);
}

static void move_end (WTree *tree)
{
    tree_move_to_bottom (tree);
    show_tree (tree);
    maybe_chdir (tree);
}

static int tree_move_left (WTree *tree)
{
    int v;

    if (tree_navigation_flag) {
	v = tree_move_to_parent (tree);
	show_tree (tree);
	maybe_chdir (tree);
	return v;
    }
    return 0;
}

static int tree_move_right (WTree *tree)
{
    if (tree_navigation_flag) {
	tree_move_to_child (tree);
	show_tree (tree);
	maybe_chdir (tree);
	return 1;
    }
    return 0;
}

static void move_prevp (WTree *tree)
{
    tree_move_backward (tree, tlines (tree) - 1);
    show_tree (tree);
    maybe_chdir (tree);
}

static void move_nextp (WTree *tree)
{
    tree_move_forward (tree, tlines (tree) - 1);
    show_tree (tree);
    maybe_chdir (tree);
}

static void chdir_sel (WTree *tree)
{
    if (!tree->is_panel) {
	tree->done = 1;
	return;
    }
    change_panel ();
    if (do_cd (tree->selected_ptr->name, cd_exact)) {
	paint_panel (cpanel);
	select_item (cpanel);
    } else {
	MC_message (1, MSG_ERROR, " Can't chdir to \"%s\" \n %m ",
		 tree->selected_ptr->name);
    }
    change_panel ();
    show_tree (tree);
    return;
}

static void start_search (WTree *tree)
{
    int i;

    if (tree->searching) {

	if (tree->selected_ptr == tree->tree_last)
	    tree_move_to_top(tree);
	else {
	/* set navigation mode temporarily to 'Static' because in
	 * dynamic navigation mode tree_move_forward will not move
	 * to a lower sublevel if necessary (sequent searches must
	 * start with the directory followed the last found directory)
	 */
	    i = tree_navigation_flag;
	    tree_navigation_flag = 0;
	    tree_move_forward (tree, 1);
	    tree_navigation_flag = i;
	}
	tree_do_search (tree, 0);
    }
    else {
	tree->searching = 1;
	tree->search_buffer[0] = 0;
    }
}

static const key_map tree_keymap [] = {
    { XCTRL('n'), move_down    },
    { XCTRL('p'), move_up      },
    { KEY_DOWN,   move_down    },
    { KEY_UP,     move_up      },
    { '\n',       chdir_sel    },
    { KEY_ENTER,  chdir_sel    },
    { KEY_HOME,   move_home    },
    { KEY_C1,     move_end     },
    { KEY_END,    move_end     },
    { KEY_A1,     move_home    },
    { KEY_NPAGE,  move_nextp   },
    { KEY_PPAGE,  move_prevp   },
    { XCTRL('v'), move_nextp   },
    { ALT('v'),   move_prevp   },
    { XCTRL('p'), move_up      },
    { XCTRL('p'), move_down    },
    { XCTRL('s'), start_search },
    { ALT('s'),   start_search },
    { XCTRL('r'), tree_rescan_cmd },
    { KEY_DC,     tree_rmdir_cmd },
    { 0, 0 }
    };

static inline int tree_key (WTree *tree, int key)
{
    int i;

    for (i = 0; tree_keymap [i].key_code; i++) {
	if (key == tree_keymap [i].key_code) {
	    if (tree_keymap [i].fn != start_search)
		tree->searching = 0;
	    (*tree_keymap [i].fn)(tree);
	    show_tree (tree);
	    return 1;
	}
    }

    /* We do not want to use them if we do not need to */
    /* Input line may want to take the motion key event */
    if (key == KEY_LEFT)
	return tree_move_left (tree);

    if (key == KEY_RIGHT)
	return tree_move_right (tree);

    if (is_abort_char (key)) {
	if (tree->is_panel) {
	    tree->searching = 0;
	    show_tree (tree);
	    return 1;  /* eat abort char */
	}
	return 0;  /* modal tree dialog: let upper layer see the
		      abort character and close the dialog */
    }

    /* Do not eat characters not meant for the tree below ' ' (e.g. C-l). */
    if ((key >= ' '&& key <= 255) || key == 8 || key == KEY_BACKSPACE) {
	if (tree->searching) {
	    tree_do_search (tree, key);
	    show_tree (tree);
	    return 1;
	}

	if (!command_prompt) {
	    start_search (tree);
	    tree_do_search (tree, key);
	    return 1;
	}
	return tree->is_panel;
    }

    return 0;
}

static void tree_frame (Dlg_head *h, WTree *tree)
{
    attrset (NORMAL_COLOR);
    widget_erase ((Widget*) tree);
    if (tree->is_panel)
	draw_double_box (h, tree->widget.y, tree->widget.x, tree->widget.lines,
			 tree->widget.cols);

    if (show_mini_info && tree->is_panel) {
	widget_move (tree, tlines (tree) + 1, 1);
	hline (ACS_HLINE, tree->widget.cols - 2);
    }
}


static int tree_callback (Dlg_head *h, WTree *tree, int msg, int par)
{
    switch (msg) {
    case WIDGET_DRAW:
	tree_frame (h, tree);
	show_tree (tree);
	return 1;

    case WIDGET_KEY:
	return tree_key (tree, par);

    case WIDGET_FOCUS:
	tree->active = 1;
	define_label (h, (Widget *)tree, 1, "Help", tree_help_cmd);
	define_label_data (h, (Widget *)tree,
	    2, "Rescan", (buttonbarfn)tree_rescan_cmd, tree);
	define_label_data (h, (Widget *)tree,
	    3, "Forget", (buttonbarfn)tree_forget_cmd, tree);
	define_label_data (h, (Widget *)tree,
	    5, "Copy",   (buttonbarfn) tree_copy_cmd, tree);
	define_label_data (h, (Widget *)tree,
	    6, "RenMov", (buttonbarfn) tree_move_cmd, tree);
	define_label (h, (Widget *)tree, 7, "", 0);
	define_label_data (h, (Widget *)tree,
	    8, "Rmdir",  (buttonbarfn) tree_rmdir_cmd, tree);
	set_navig_label (h);
	redraw_labels (h, (Widget *)tree);


	/* FIXME: Should find a better way of only displaying the
	   currently selected item */
	show_tree (tree);
	return 1;

	/* FIXME: Should find a better way of changing the color of the
	   selected item */
    case WIDGET_UNFOCUS:
	tree->active = 0;
	show_tree (tree);
	return 1;
    }
    return default_proc (h, msg, par);
}

WTree *tree_new (int is_panel, int y, int x, int lines, int cols)
{
    WTree *tree = zmalloc (sizeof (WTree));

    init_widget (&tree->widget, y, x, lines, cols, tcallback,
		 (destroy_fn) tree_destroy, (mouse_h) event_callback);
    tree->is_panel = is_panel;
    tree->topdiff = tree->widget.lines / 2;

    /* We do not want to keep the cursor */
    widget_want_cursor (tree->widget, 0);
    load_tree (tree);
    return tree;
}

static char *get_absolute_name (char *file)
{
    char dir [MC_MAXPATHLEN];

    if (file [0] == '/')
	return strdup (file);
    mc_get_current_wd (dir, MC_MAXPATHLEN);
    return get_full_name (dir, file);
}

static int my_mkdir_rec (char *s, mode_t mode)
{
    char *p, *q;
    int result;

    if (!mc_mkdir (s, mode))
	return 0;

    /* FIXME: should check instead if s is at the root of that filesystem */
    if (!vfs_file_is_local (s))
	return -1;
    if (!strcmp (vfs_path(s), "/"))
	return ENOTDIR;
    p = concat_path_file (s, "..");
    q = vfs_canon (p);
    free (p);
    if (!(result = my_mkdir_rec (q, mode))) {
	result = mc_mkdir (s, mode);
    }
    free (q);
    return result;
}

static stdcall2 int my_mkdir (char *s, mode_t mode)
{
    int result;
#if FIXME
    WTree *tree = 0;
#endif

    result = mc_mkdir (s, mode);
    if (result) {
	char *p = vfs_canon (s);

	result = my_mkdir_rec (p, mode);
	free (p);
    }
    if (result == 0) {
	s = get_absolute_name (s);
#if FIXME
	/* FIXME: Should receive a Wtree! */

	tree_add_entry (tree, s);
#endif
	free (s);
    }
    return result;
}

static stdcall2 int my_rmdir (char *s)
{
    int result;
#if FIXME
    WTree *tree = 0;
#endif

    /* FIXME: Should receive a Wtree! */
    result = mc_rmdir (s);
    if (result == 0) {
	s = get_absolute_name (s);
#if FIXME
	tree_remove_entry (tree, s);
#endif
	free (s);
    }
    return result;
}


/* User Menu implementation */

/* Formats defined:
   %%  The % character
   %f  The current file (if non-local vfs, file will be copied locally and
       %f will be full path to it).
   %p  The current file
   %d  The current working directory
   %s  "Selected files"; the tagged files if any, otherwise the current file
   %t  Tagged files
   %u  Tagged files (and they are untagged on return from expand_format)
   %view Runs the commands and pipes standard output to the view command
       if %view is directly followed by '{', a list of keywords
       ascii, hex, nroff, unformated and

   If the format letter is in uppercase, it refers to the other panel.

   With a number followed the % character you can turn quoting on (default)
   and off. For example:
       %f    quote expanded macro
       %1f   ditto
       %0f   don't quote expanded macro

   expand_format returns a memory block that must be free()d.
*/

/* Returns how many characters we should advance if %view was found */
int check_format_view (char *p)
{
    char *q = p;
    if (!strncmp (p, "view", 4)) {
	q += 4;
	if (*q == '{') {
	    for (q++;*q && *q != '}';q++) {
		if (!strncmp (q, "ascii", 5)) {
		    default_hex_mode = 0;
		    q += 4;
		} else if (!strncmp (q, "hex", 3)) {
		    default_hex_mode = 1;
		    q += 2;
		} else if (!strncmp (q, "nroff", 5)) {
		    default_nroff_flag = 1;
		    q += 4;
		} else if (!strncmp (q, "unformated", 10)) {
		    default_nroff_flag = 0;
		    q += 9;
		}
	    }
	    if (*q == '}')
		q++;
	}
	return q - p;
    } else
	return 0;
}

int check_format_cd (char *p)
{
    if (!strncmp (p, "cd", 2))
	return 3;
    else
	return 0;
}

/* Check if p has a "^var\{var-name\}" */
/* Returns the number of skipped characters (zero on not found) */
/* V will be set to the expanded variable name */
int check_format_var (char *p, char **v)
{
    char *q = p;
    char *var_name;
    char *value;
    char *dots;

    *v = 0;
    dots = 0;
    if (!strncmp (p, "var{", 4)) {
	for (q += 4; *q && *q != '}'; q++) {
	    if (*q == ':')
		dots = q+1;
	    ;
	}
	if (!*q)
	    return 0;

	if (!dots || dots == q+5) {
	    MC_message (1,
		     " Format error on file Extensions File ",
		     !dots ? " The %%var macro does not have a default "
		     :       " The %%var macros does not have a variable " );
	    return 0;
	}

	/* Copy the variable name */
	var_name = x_malloc (dots - p, "check_format_var");
	strncpy (var_name, p+4, dots-2 - (p+3));
	var_name [dots-2 - (p+3)] = 0;

	value = getenv (var_name);
	if (value) {
	    *v = strdup (value);
	    return q-p;
	}
	free (var_name);
	var_name = xstrndup (dots, q - dots);
	*v = var_name;
	return q-p;
    }
    return 0;
}

/* strip file's extension */
static char *strip_ext(char *ss)
{
    char *s = ss;
    char *e = NULL;
    while(*s) {
	if(*s == '.') e = s;
	if(*s == '/' && e) e=NULL; /* '.' in *directory* name */
	s++;
    }
    if(e) *e = 0;
    return ss;
}

char *expand_format (char c, int quote_e)
{
    WPanel *panel;
    char *(*quote_func)(const char *, int);

    if (quote_e)
	quote_func = name_quote;
    else
	quote_func = fake_name_quote;

    if (c == '%')
	return strdup ("%");

    if (islower (c))
	panel = cpanel;
    else {
	if (get_other_type () == view_listing) {
	    panel = other_panel;
	} else
	    return strdup ("");
    }
    if (!panel)
	panel = cpanel;

    c = tolower (c);

    switch (c) {
    case 'f':
    case 'p': return (*quote_func) (panel->dir.list [panel->selected].fname, 0);
    case 'b':
	return strip_ext((*quote_func) (panel->dir.list [panel->selected].fname, 0));
    case 'd': return (*quote_func) (panel->cwd, 0);
    case 's':
	if (!panel->marked)
	    return (*quote_func) (panel->dir.list [panel->selected].fname, 0);

	/* Fall through */

    case 't':
    case 'u':
    {
	int length = 2, i;
	char *block, *tmp;

	for (i = 0; i < panel->count; i++)
	    if (panel->dir.list [i].f.marked)
		length += strlen (panel->dir.list [i].fname) + 1;

	block = x_malloc (length*2+1, "expand_format");
	*block = 0;
	for (i = 0; i < panel->count; i++)
	    if (panel->dir.list [i].f.marked) {
		strcat (block, tmp = (*quote_func) (panel->dir.list [i].fname, 0));
		free (tmp);
		strcat (block, " ");
		if (c == 'u')
		    do_file_mark (panel, i, 0);
	    }
	return block;
    } /* sub case block */
    } /* switch */
    return strdup ("");
}

/* Checks for shell patterns defination */
static char *check_patterns (char *p)
{
    const char *def_name = "shell_patterns=";
    int value;
    size_t l = strlen(def_name);

    if (strncmp (p, def_name, l) == 0) {
	p += l;
	value = *p++ - '0';
	if (value == 0 || value == 1)
	    easy_patterns = value;
	else
	    MC_message (1, MSG_ERROR, " Invalid shell pattern defination \"%c\". ", value + '0');
    }
    while (*p == '\n' || *p == '\t' || *p == ' ') p++;
    return p;
}

/* Copies a whitespace separated argument from p to arg. Returns the
   point after argument. */
static char *extract_arg (char *p, char *arg)
{
    while (*p && (*p == ' ' || *p == '\t' || *p == '\n'))
	p++;
    while (*p && *p != ' ' && *p != '\t' && *p != '\n')
	*arg++ = *p++;
    *arg = 0;
    if (!*p || *p == '\n')
	p --;
    return p;
}

/* Tests whether the selected file in the panel is of any of the types
   specified in argument. */
static int test_type (WPanel *panel, char *arg)
{
    int result = 0; /* False by default */
    int st_mode = panel->dir.list [panel->selected].buf.st_mode;

    for (;*arg != 0; arg++) {
	switch (*arg) {
	case 'n':       /* Not a directory */
	    result |= !S_ISDIR (st_mode);
	    break;
	case 'r':       /* Regular file */
	    result |= S_ISREG (st_mode);
	    break;
	case 'd':       /* Directory */
	    result |= S_ISDIR (st_mode);
	    break;
	case 'l':       /* Link */
	    result |= S_ISLNK (st_mode);
	    break;
	case 'c':       /* Character special */
	    result |= S_ISCHR (st_mode);
	    break;
	case 'b':       /* Block special */
	    result |= S_ISBLK (st_mode);
	    break;
	case 'f':       /* Fifo (named pipe) */
	    result |= S_ISFIFO (st_mode);
	    break;
	case 's':       /* Socket */
	    result |= S_ISSOCK (st_mode);
	    break;
	case 'x':       /* Executable */
	    result |= (st_mode & 0111) ? 1 : 0;
	    break;
	case 't':
	    result |= panel->marked ? 1 : 0;
	    break;
	default:
	    debug_error = 1;
	    break;
	}
    }
    return result;
}

/* Calculates the truth value of the next condition starting from
   p. Returns the point after condition. */
static char *test_condition (char *p, int *condition)
{
    WPanel *panel;
    char arg [256];

    /* Handle one condition */
    for (;*p != '\n' && *p != '&' && *p != '|'; p++) {
	if (*p == ' ' || *p == '\t')
	    continue;
	if (*p >= 'a')
	    panel = cpanel;
	else {
	    if (get_other_type () == view_listing)
		panel = other_panel;
	    else
		panel = NULL;
	}
	*p |= 0x20;

	switch (*p++) {
	case '!':
	    p = test_condition (p, condition);
	    *condition = ! *condition;
	    p--;
	    break;
	case 'f':
	    p = extract_arg (p, arg);
	    *condition = panel && regexp_match (arg, panel->dir.list [panel->selected].fname, match_file);
	    break;
	case 'd':
	    p = extract_arg (p, arg);
	    *condition = panel && regexp_match (arg, panel->cwd, match_file);
	    break;
	case 't':
	    p = extract_arg (p, arg);
	    *condition = panel && test_type (panel, arg);
	    break;
	default:
	    debug_error = 1;
	    break;
	} /* switch */

    } /* while */
    return p;
}

/* General purpose condition debug output handler */
static void debug_out (char *start_d, char *end_d, int cond)
{
    int len;

    if (start_d == NULL && end_d == NULL) {
	if (cond == 0) {
	    /* Init */
	    debug_out_msg [0] = 0;
	} else {
	    /* Show output */
	    if (!debug_flag)
		return;
	    len = strlen (debug_out_msg);
	    if (len)
		debug_out_msg [len - 1] = 0;
	    MC_message (0, " Debug ", debug_out_msg);
	    debug_flag = 0;
	}
    } else {
	/* Save debug info for later output */
	if (!debug_flag)
	    return;
	/* Save the result of the condition */
	if (debug_error) {
	    strcat (debug_out_msg, " ERROR: ");
	    debug_error = 0;
	}
	else if (cond)
	    strcat (debug_out_msg, " True:  ");
	else
	    strcat (debug_out_msg, " False: ");
	/* Copy condition statement */
	len = strlen (debug_out_msg);
	if (end_d == NULL) {
	    /* Copy one character */
	    debug_out_msg [len] = *start_d;
	    debug_out_msg [len + 1] = 0;
	} else {
	    /* Copy many characters */
	    while (start_d < end_d) {
		debug_out_msg [len++] = *start_d++;
	    }
	    debug_out_msg [len] = 0;
	}
	strcat (debug_out_msg, " \n");
    }
}

/* Calculates the truth value of one lineful of conditions. Returns
   the point just before the end of line. */
static char *test_line (char *p, int *result)
{
    int condition;
    char operator;
    char *debug_start, *debug_end;

    /* Init debugger */
    debug_out (NULL, NULL, 0);
    /* Repeat till end of line */
    while (*p && *p != '\n') {
	while (*p == ' ' || *p == '\t')
	    p++;
	if (!*p || *p == '\n')
	    break;
	operator = *p++;
	if (*p == '?') {
	    debug_flag = 1;
	    p++;
	}
	while (*p == ' ' || *p == '\t')
	    p++;
	if (!*p || *p == '\n')
	    break;
	condition = 1;  /* True by default */

	debug_start = p;
	p = test_condition (p, &condition);
	debug_end = p;
	/* Add one debug statement */
	debug_out (debug_start, debug_end, condition);

	switch (operator) {
	case '+':
	case '=':
	    /* Assignment */
	    *result = condition;
	    break;
	case '&':       /* Logical and */
	    *result &= condition;
	    break;
	case '|':       /* Logical or */
	    *result |= condition;
	    break;
	default:
	    debug_error = 1;
	    break;
	} /* switch */
	/* Add one debug statement */
	debug_out (&operator, NULL, *result);

    } /* while (*p != '\n') */
    /* Report debug message */
    debug_out (NULL, NULL, 1);

    if (!*p || *p == '\n')
	p --;
    return p;
}

/* FIXME: recode this routine on version 3.0, it could be cleaner */
static void execute_menu_command (char *s)
{
    char *commands;
    FILE *cmd_file;
    int  cmd_file_fd;
    int  expand_prefix_found = 0;
    int parameter_found = 0;
    int do_quote = 0;
    char prompt [80] = "";
    int  col;
    char *file_name;

    file_name = mc_mkstemp(&cmd_file_fd);
    if (cmd_file_fd == -1) {
	MC_message (1, MSG_ERROR, " Can't create temporary command file \n %m ");
	return;
    }
    cmd_file = fdopen (cmd_file_fd, "w");
    commands = strchr (s, '\n');
    if (!commands) {
end:
	free(file_name);
	fclose (cmd_file);
	unlink (file_name);
	return;
    }
    commands++;

    for (col = 0; *commands; commands++) {
	if (col == 0 && (*commands != ' ' && *commands != '\t'))
	    break;
	else if (col == 0)
	    while (*commands == ' ' || *commands == '\t')
		commands++;
	col++;
	if (*commands == '\n')
	    col = 0;
	if (parameter_found) {
	    if (*commands == '}') {
		char *parameter;
		char *tmp;
		parameter_found = 0;
		parameter = input_dialog (" Parameter ", prompt, "");
		if (!parameter || !*parameter) {
		    /* User canceled */
		    goto end;
		}
		if (do_quote) {
		    fputs (tmp = name_quote (parameter, 0), cmd_file);
		    free (tmp);
		} else
		    fputs (parameter, cmd_file);
		free (parameter);
	    } else {
		int len = strlen (prompt);

		if (len+1 < sizeof (prompt)) {
		    prompt [len] = *commands;
		    prompt [len+1] = 0;
		} else
		    prompt [sizeof (prompt)-1] = 0;
	    }
	} else if (expand_prefix_found) {
	    expand_prefix_found = 0;
	    if (isdigit (*commands)) {
		do_quote = atoi (commands);
		for ( ; isdigit (*commands); commands++)
		    ;
	    }
	    if (*commands == '{')
		parameter_found = 1;
	    else{
		char *text = expand_format (*commands, do_quote);
		fputs (text, cmd_file);
		free (text);
	    }
	} else {
	    if (*commands == '%') {
		do_quote = 1; /* Default: Quote expanded macro */
		expand_prefix_found = 1;
	    } else
		fputc (*commands, cmd_file);
	}
    }
    fclose (cmd_file);
    chmod (file_name, S_IRWXU);
    execute (file_name);
    unlink (file_name);
    free(file_name);
}

/* Check owner of the menu file. Using menu file is allowed, if
 * owner of the menu is root or the actual user. In either case
 * file should not be group and word-writable.
 *
 * Q. Should we apply this routine to system and home menu (and .ext files)? */
static int
menu_file_own(const char *path)
{
	struct stat st;

	if (stat (path, &st) == 0
			&& (!st.st_uid || (st.st_uid == my_euid))
			&& ((st.st_mode & (S_IWGRP | S_IWOTH)) == 0)) {
		return 1;
	} else {
		if (MC_verbose) {
			MC_message (0, " Warning -- ignoring file ",
				 "File %s is not owned by root or you or is world writable.\n"
				   "Using it may compromise your security", path);
		}
		return 0;
	}
}

#define TMP_LINE_SIZE 500
static char *extract_line (char *s, char *top, char tmp_line [TMP_LINE_SIZE])
{
    char *t = tmp_line;

    while (*s && *s != '\n' && (t - tmp_line) < (TMP_LINE_SIZE-1) && s < top)
	*t++ = *s++;
    *t = 0;
    return tmp_line;
}

void user_menu_cmd (void)
{
    char *menu, *p;
    int  col, i, accept_entry = 1;
    int  selected, old_patterns;
    Listbox *listbox;
    char *user_data;
    int max_cols;
    int menu_lines;
    char tmp_line [TMP_LINE_SIZE];

    if (!vfs_current_is_local ()) {
	MC_message (1, " Oops... ",
		 " I can't run programs while logged on a non local directory ");
	return;
    }

    menu = strdup (MC_LOCAL_MENU);
    if (!exist_file (menu) || !menu_file_own (menu)) {
	free (menu);
	menu = concat_path_file (home_dir, MC_HOME_MENU);
	if (!exist_file (menu)) {
	    free (menu);
	    menu = concat_mcpath_file (MC_GLOBAL_MENU);
	}
    }

    if ((user_data = load_file (menu)) == NULL) {
	MC_message (1, MSG_ERROR, " Can't open file %s \n %m ", menu);
	free (menu);
	return;
    }
    free (menu);

    max_cols = 0;
    for (i = 0; i < MAX_ENTRIES; i++)
	um_entries [i] = 0;
    selected = 0;

    /* Parse the menu file */
    old_patterns = easy_patterns;
    p = check_patterns (user_data);
    for (menu_lines = col = 0; *p; p++) {
	if (col == 0 && !um_entries [menu_lines]) {
	    if (*p == '#') {
		/* A commented menu entry */
		accept_entry = 1;
	    } else if (*p == '+') {
		if (*(p+1) == '=') {
		    /* Combined adding and default */
		    char *q = p++;

		    p = test_line (q, &accept_entry);
		    if (selected == 0 && accept_entry)
			selected = menu_lines;
		} else {
		    /* A condition for adding the entry */
		    p = test_line (p, &accept_entry);
		}
	    } else if (*p == '=') {
		if (*(p+1) == '+') {
		    char *q = p++;
		    /* Combined adding and default */
		    p = test_line (q, &accept_entry);
		    if (selected == 0 && accept_entry)
			selected = menu_lines;
		} else {
		    /* A condition for making the entry default */
		    i = 1;
		    p = test_line (p, &i);
		    if (selected == 0 && i)
			selected = menu_lines;
		}
	    }
	    else if (*p > ' ' && *p < 127) {
		/* A menu entry title line */
		if (accept_entry)
		    um_entries [menu_lines] = p;
		else
		    accept_entry = 1;
	    }
	}
	if (menu_lines == MAX_ENTRIES)
	    break;
	if (*p == '\t')
	    *p = ' ';
	col++;
	if (*p == '\n') {
	    if (um_entries [menu_lines]) {
		menu_lines++;
		accept_entry = 1;
	    }
	    max_cols = max (max_cols, col);
	    col = 0;
	}
    }
#define MAX_ENTRY_LEN 60
    max_cols = min (max (max_cols, col), MAX_ENTRY_LEN);

    /* Create listbox */
    listbox = create_listbox_window (max_cols+2, menu_lines, " User menu ",
				     "[Menu File Edit]");

    /* insert all the items found */
    for (i = 0; i < menu_lines; i++)
	LISTBOX_APPEND_TEXT (listbox, um_entries [i][0],
			     extract_line (um_entries [i],
					   um_entries [i]+MAX_ENTRY_LEN,
					   tmp_line),
			     um_entries [i]);

    /* Select the default entry */
    listbox_select_by_number (listbox->list, selected);

    selected = run_listbox (listbox);
    if (selected >= 0)
	execute_menu_command (um_entries [selected]);

    easy_patterns = old_patterns;
    do_refresh ();
    free (user_data);
}
/* Various utilities */


static stdcall2 int is_printable (int c)
{
    static const unsigned char xterm_printable[] = {
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,
	1,1,1,1,0,0,1,1,0,1,1,1,1,0,0,0,0,1,1,1,1,1,0,0,0,1,0,0,0,0,0,0,
	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,
	1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1
    };

    c &= 0xff;
    return xterm_printable [c];
}

/* Returns the message dimensions (lines and columns) */
static stdcall2 int msglen (const char *text, int *lines)
{
    int max = 0;
    int line_len = 0;

    for (*lines = 1;*text; text++) {
	if (*text == '\n') {
	    line_len = 0;
	    (*lines)++;
	} else {
	    line_len++;
	    if (line_len > max)
		max = line_len;
	}
    }
    return max;
}

char *
name_quote (const char *s, int quote_percent)
{
    char *ret, *d;

    d = ret = x_malloc (strlen (s)*2 + 2 + 1, "quote_name");
    if (*s == '-') {
	*d++ = '.';
	*d++ = '/';
    }

    for (; *s; s++, d++) {
	switch (*s)
	{
	    case '%':
		if (quote_percent)
		    *d++ = '%';
		break;
	    case '\'':
	    case '\\':
	    case '\r':
	    case '\n':
	    case '\t':
	    case '$':
	    case '?':
	    case '*':
	    case '(':
	    case ')':
	    case '[':
	    case ']':
	    case '"':
	    case '!':
	    case '&':
	    case '#':
	    case '`':
	    case ' ':
		*d++ = '\\';
	}
	*d = *s;
    }
    *d = '\0';
    return ret;
}

char *
fake_name_quote (const char *s, int quote_percent)
{
    return strdup (s);
}

/* If passed an empty txt (this usually means that there is an error)
 * in the upper layers, we return "/"
 */
static stdcall2 char *
name_trunc (const char *txt, int trunc_len)
{
    int    txt_len;
    char *p;

    if (!txt)
	txt = "/";

    if (trunc_len > sizeof (nt_x)-1) {
	fprintf (stderr, "name_trunc: too big");
	trunc_len = sizeof (nt_x)-1;
    }
    txt_len = strlen (txt);
    if (txt_len <= trunc_len)
	strcpy (nt_x, txt);
    else {
	int y = trunc_len % 2;
	strncpy (nt_x, txt, (trunc_len/2)+y);
	strncpy (nt_x+(trunc_len/2)+y, txt+txt_len-(trunc_len/2), trunc_len/2);
	nt_x [(trunc_len/2)+y] = '~';
    }
    nt_x [trunc_len] = 0;
    for (p = nt_x; *p; p++)
	if (!is_printable (*p))
	    *p = '?';
    return nt_x;
}

static stdcall2 char *size_trunc (off_t size)
{
    long int divisor = 1;
    char *xtra = "";

    if (size > 999999999L) {
	divisor = 1024;
	xtra = "kb";
	if (size/divisor > 999999999L) {
	    divisor = 1024*1024;
	    xtra = "Mb";
	}
    }
    sprintf (st_x, "%ld%s", (long)(size/divisor), xtra);
    return st_x;
}

static stdcall2 char *size_trunc_sep (off_t size)
{
    int  count_s;
    char *p, *d, *y;

    p = y = size_trunc (size);
    p += strlen (p) - 1;
    d = sts_x + sizeof (sts_x) - 1;
    *d-- = 0;
    while (p >= y && isalpha (*p))
	*d-- = *p--;
    for (count_s = 0; p >= y; count_s++) {
	if (count_s == 3) {
	    *d-- = ',';
	    count_s = 0;
	}
	*d-- = *p--;
    }
    d++;
    if (*d == ',')
	d++;
    return d;
}

static stdcall2 int is_exe (mode_t mode)
{
    if ((S_IXUSR & mode) || (S_IXGRP & mode) || (S_IXOTH & mode))
	return 1;
    return 0;
}

#ifdef USE_NETCODE
static char *
strip_password (char *path)
{
    char *at, *inner_colon, *dir;

    dir = strchr_set0 (path, '/');
    /* search for any possible user */
    at = strchr_set0 (path, '@');

    /* We have a username */
    if (at) {
	inner_colon = strchr (path, ':');
	*at = '@';
	if (inner_colon)
	    strcpy (inner_colon, at);
    }
    if (dir)
	*dir = '/';
    return (path);
}
#endif

static stdcall2 char *strip_home_and_password(char *dir, char newdir [MC_MAXPATHLEN])
{
    if (home_dir && !strncmp (dir, home_dir, strlen (home_dir))) {
	newdir [0] = '~';
	strcpy (&newdir [1], &dir [strlen (home_dir)]);
	return newdir;
    }
#ifdef USE_NETCODE
    else if (!strncmp (dir, "ftp://", 6)) {
	char *p;

	strip_password (strcpy (newdir, dir) + 6);
	if ((p = strchr_set0 (newdir + 6, '/')) != NULL) {
	    char *q = ftpfs_gethome (newdir);

	    *p = '/';
	    if (q != 0 && strcmp (q, "/") && !strncmp (p, q, strlen (q) - 1)) {
		strcpy (p, "/~");
		strcat (newdir, p + strlen (q) - 1);
	    }
	}
	return newdir;
    }
#endif
    return strcpy(newdir, dir);
}

static char *maybe_start_group (char *d, int do_group, int *was_wildcard)
{
    if (!do_group)
	return d;
    if (*was_wildcard)
	return d;
    *was_wildcard = 1;
    *d++ = '\\';
    *d++ = '(';
    return d;
}

static char *maybe_end_group (char *d, int do_group, int *was_wildcard)
{
    if (!do_group)
	return d;
    if (!*was_wildcard)
	return d;
    *was_wildcard = 0;
    *d++ = '\\';
    *d++ = ')';
    return d;
}

/* If shell patterns are on converts a shell pattern to a regular
   expression. Called by regexp_match and mask_rename. */
/* Shouldn't we support [a-fw] type wildcards as well ?? */
char *convert_pattern (const char *pattern, int match_type, int do_group)
{
    if (easy_patterns) {
	char *new_pattern = malloc (MC_MAXPATHLEN);
	char *d = new_pattern;
	int was_wildcard = 0;
	const char *s;

	if (match_type == match_file)
	    *d++ = '^';
	for (s = pattern; *s; s++, d++) {
	    switch (*s) {
	    case '*':
		d = maybe_start_group (d, do_group, &was_wildcard);
		*d++ = '.';
		*d   = '*';
		break;

	    case '?':
		d = maybe_start_group (d, do_group, &was_wildcard);
		*d = '.';
		break;

	    case '.':
		d = maybe_end_group (d, do_group, &was_wildcard);
		*d++ = '\\';
		*d   = '.';
		break;

	    default:
		d = maybe_end_group (d, do_group, &was_wildcard);
		*d = *s;
		break;
	    }
	}
	d = maybe_end_group (d, do_group, &was_wildcard);
	if (match_type == match_file)
	    *d++ = '$';
	*d = 0;
	return new_pattern;
    } else
	return strdup (pattern);
}

int regexp_match (char *pattern, char *string, int match_type)
{
    int    rval;

    if (!m_old_pattern || strcmp(m_old_pattern, pattern) || m_old_type != match_type) {
	if (m_old_pattern) {
	    regfree (&regexp_match_r);
	    free (m_old_pattern);
	}
	pattern = convert_pattern (pattern, match_type, 0);
	if (regcomp (&regexp_match_r, pattern, REG_EXTENDED|REG_NOSUB|MC_ARCH_FLAGS)) {
	    free (pattern);
	    return -1;
	}
	m_old_pattern = pattern;
	m_old_type = match_type;
    }
    rval = !regexec (&regexp_match_r, string, 0, NULL, 0);
    return rval;
}

static stdcall2 const char *extension (const char *filename)
{
    const char *d;

    if (!strlen (filename))
	return "";

    d = filename + strlen (filename) - 1;
    for (;d >= filename; d--) {
	if (*d == '.')
	    return d+1;
    }
    return "";
}



static stdcall2 int exist_file (const char *name)
{
    return access (name, R_OK) == 0;
}

static stdcall2 char *load_file (const char *filename)
{
    FILE *data_file;
    struct stat s;
    char *data;
    size_t read_size;

    if (stat (filename, &s) != 0) {
	return 0;
    }
    if ((data_file = fopen (filename, "r")) == NULL) {
	return 0;
    }
    data = (char *) x_malloc (s.st_size+1, "util, load_file");
    read_size = fread (data, 1, s.st_size, data_file);
    data [read_size] = 0;
    fclose (data_file);

    if (read_size > 0)
	return data;
    else {
	free (data);
	return 0;
    }
}

static stdcall2 char *file_date (time_t when)
{
    time_t current_time = time (NULL);

    strcpy (timebuf, ctime (&when));
    if (current_time > when + 6L * 30L * 24L * 60L * 60L /* Old. */
	|| current_time < when - 60L * 60L) /* In the future. */
    {
	/* The file is fairly old or in the future.
	   POSIX says the cutoff is 6 months old;
	   approximate this by 6*30 days.
	   Allow a 1 hour slop factor for what is considered "the future",
	   to allow for NFS server/client clock disagreement.
	   Show the year instead of the time of day.  */
	strcpy (timebuf + 11, timebuf + 19);
    }
    timebuf[16] = 0;
    return &timebuf [4];
}

/* Returns a copy of *s until a \n is found and is below top */
/* FIXME: I should write a faster version of this (Aho-Corasick stuff) */
static stdcall2 char * _icase_search (const char *text, char *data, int *lng)
{
    const char *d = text;
    char *e = data;
    int dlng = 0;

    if (lng)
	*lng = 0;
    for (;*e; e++) {
	while (*(e+1) == '\b' && *(e+2)) {
	    e += 2;
	    dlng += 2;
	}
	if (toupper((unsigned char) *d) == toupper((unsigned char) *e))
	    d++;
	else {
	    e -= d - text;
	    d = text;
	    dlng = 0;
	}
	if (!*d) {
	    if (lng)
		*lng = strlen (text) + dlng;
	    return e+1;
	}
    }
    return 0;
}

/* The basename routine */
static stdcall2 char *get_full_name (const char *dir, char *file)
{
    return concat_path_file(dir, bb_basename (file));
}

static void my_putenv (const char *name, const char *data)
{
    char *full;

    full = xasprintf ("%s=%s", name, data);
    putenv (full);
    /* WARNING: NEVER FREE THE full VARIABLE!!!!!!!!!!!!!!!!!!!!!!!! */
    /* It is used by putenv. Freeing it will corrupt the environment */
}

static char *unix_error_string (void)
{
    sprintf (unix_error_string_buffer, "%m (%d)", errno);
    return unix_error_string_buffer;
}

#ifndef NO_INFOMOUNT
void init_my_statfs (void)
{
    if(mount_list == NULL)
	read_filesystem_list();
}
#endif

static stdcall2 char *skip_separators (char *s)
{
    for (;*s; s++)
	if (*s != ' ' && *s != '\t' && *s != ',')
	    break;
    return s;
}


#define CHECK(x) if (x == -1) return 0;

static long get_small_endian_long (int fd)
{
    unsigned char a, b, c, d;

    /* It needs to be read one byte at the time to avoid endianess
       portability problems */
    CHECK (mc_read (fd, (char *)&a, 1));
    CHECK (mc_read (fd, (char *)&b, 1));
    CHECK (mc_read (fd, (char *)&c, 1));
    CHECK (mc_read (fd, (char *)&d, 1));
    return (d << 24) | (c << 16) | (b << 8) | a;
}

/* This function returns 0 if the file is not in gunzip format  */
/* or how much memory must be allocated to load the gziped file */
/* Warning: this function moves the current file pointer */
static stdcall2 long int is_gunzipable (int fd, int *type)
{
    unsigned char magic [4];

    *type = ISGUNZIPABLE_GUNZIP;

    /* Read the magic signature */
    CHECK (mc_read (fd, (char *)&magic [0], 1));
    CHECK (mc_read (fd, (char *)&magic [1], 1));
    CHECK (mc_read (fd, (char *)&magic [2], 1));
    CHECK (mc_read (fd, (char *)&magic [3], 1));

    /* GZIP_MAGIC and OLD_GZIP_MAGIC */
    if (magic [0] == 037 && (magic [1] == 0213 || magic [1] == 0236)) {
	/* Read the uncompressed size of the file */
	mc_lseek (fd, -4, SEEK_END);
	return get_small_endian_long (fd);
    }

    /* PKZIP_MAGIC */
    if (magic [0] == 0120 && magic [1] == 0113 && magic [2] == 003 && magic [3] == 004) {
	/* Read compression type */
	mc_lseek (fd, 8, SEEK_SET);
	CHECK (mc_read (fd, &magic [0], 1));
	CHECK (mc_read (fd, &magic [1], 1));

	/* Gzip can handle only deflated (8) or stored (0) files */
	if ((magic [0] != 8 && magic [0] != 0) || magic [1] != 0)
	     return 0;
	/* Read the uncompressed size of the first file in the archive */
	mc_lseek (fd, 22, SEEK_SET);
	return get_small_endian_long (fd);
    }

    /* PACK_MAGIC and LZH_MAGIC and compress magic */
    if (magic [0] == 037 && (magic [1] ==  036 || magic [1] == 0240 || magic [1] == 0235)) {
	 /* In case the file is packed, sco lzhed or compress_magic, the */
	 /* program guesses that the uncompressed size is (at most) four */
	 /* times the length of the compressed size, if the compression  */
	 /* ratio is more than 4:1 the end of the file is not displayed  */
	 return 4*mc_lseek (fd, 0, SEEK_END);
    }

    /* BZIP and BZIP2 files */
    if ((magic[0] == 'B') && (magic[1] == 'Z') &&
	(magic [3] >= '1') && (magic [3] <= '9')) {
	    switch (magic[2]) {
		case '0':
		    *type = ISGUNZIPABLE_BZIP;
		    return 5*mc_lseek (fd, 0, SEEK_END);
		case 'h':
		    *type = ISGUNZIPABLE_BZIP2;
		    return 5*mc_lseek (fd, 0, SEEK_END);
	    }
    }
    return 0;
}

static stdcall2 char *
decompress_command (int type)
{
	switch (type) {
	case ISGUNZIPABLE_GUNZIP:
		return "gzip -cdf";

	case ISGUNZIPABLE_BZIP:
		return "bzip -d";

	case ISGUNZIPABLE_BZIP2:
		return "bzip2 -d";
	}
	/* Should never reach this place */
	fprintf (stderr, "Fatal: decompress_command called with an unknown argument\n");
	return 0;
}

#ifdef USE_VFS
void
decompress_command_and_arg (int type, char **cmd, char **flags)
{
	switch (type) {
	case ISGUNZIPABLE_GUNZIP:
		*cmd   = "gzip";
		*flags = "-cdf";
		return;

	case ISGUNZIPABLE_BZIP:
		*cmd   = "bzip";
		*flags = "-d";
		return;


	case ISGUNZIPABLE_BZIP2:
		*cmd   = "bzip2";
		*flags = "-d";
		return;
	}
	*cmd   = 0;
	*flags = 0;

	/* Should never reach this place */
	fprintf (stderr, "Fatal: decompress_command called with an unknown argument\n");
}
#endif

/* Hooks */
void add_hook (Hook **hook_list, void (*hook_fn)(void *), void *data)
{
    Hook *new_hook = x_malloc (sizeof (Hook), "add_hook");

    new_hook->hook_fn = hook_fn;
    new_hook->next    = *hook_list;
    new_hook->hook_data = data;

    *hook_list = new_hook;
}

static stdcall2 void execute_hooks (Hook *hook_list)
{
    Hook *new_hook = 0;
    Hook *p;

    /* We copy the hook list first so tahat we let the hook
     * function call delete_hook
     */

    while (hook_list) {
	add_hook (&new_hook, hook_list->hook_fn, hook_list->hook_data);
	hook_list = hook_list->next;
    }
    p = new_hook;

    while (new_hook) {
	(*new_hook->hook_fn)(new_hook->hook_data);
	new_hook = new_hook->next;
    }

    for (hook_list = p; hook_list;) {
	p = hook_list;
	hook_list = hook_list->next;
	free (p);
    }
}

void delete_hook (Hook **hook_list, void (*hook_fn)(void *))
{
    Hook *current, *new_list, *next;

    new_list = 0;

    for (current = *hook_list; current; current = next) {
	next = current->next;
	if (current->hook_fn == hook_fn)
	    free (current);
	else
	    add_hook (&new_list, current->hook_fn, current->hook_data);
    }
    *hook_list = new_list;
}

int hook_present (Hook *hook_list, void (*hook_fn)(void *))
{
    Hook *p;

    for (p = hook_list; p; p = p->next)
	if (p->hook_fn == hook_fn)
	    return 1;
    return 0;
}

#ifdef USE_VFS
void wipe_password (char *passwd)
{
    char *p = passwd;

    for ( ; *p; p++)
	*p = 0;
    free (passwd);
}
#endif

/* Convert "\E" -> esc character and ^x to control-x key and ^^ to ^ key */
/* Returns a newly allocated string */
char *convert_controls (char *s)
{
    char *valcopy = strdup (s);
    char *p, *q;

    /* Parse the escape special character */
    for (p = s, q = valcopy; *p;) {
	if (*p == '\\') {
	    p++;
	    if ((*p == 'e') || (*p == 'E')) {
		p++;
		*q++ = ESC_CHAR;
	    }
	} else {
	    if (*p == '^') {
		p++;
		if (*p == '^')
		    *q++ = *p++;
		else {
		    *p = (*p | 0x20);
		    if (*p >= 'a' && *p <= 'z') {
			*q++ = *p++ - 'a' + 1;
		    } else
			p++;
		}
	    } else
		*q++ = *p++;
	}
    }
    *q++ = 0;
    return valcopy;
}

/* Reverse the string */
static stdcall2 char *reverse_string (char *string)
{
    int len = strlen (string);
    int i;
    const int steps = len/2;

    for (i = 0; i < steps; i++) {
	char c = string [i];

	string [i] = string [len-i-1];
	string [len-i-1] = c;
    }
    return string;
}

static stdcall2 char *resolve_symlinks (char *path)
{
    char *buf, *buf2, *p, *q, *r, c;
    int len;
    struct stat mybuf;

    if (*path != '/')
	return NULL;
    r = buf = x_malloc (MC_MAXPATHLEN, "resolve symlinks");
    buf2 = x_malloc (MC_MAXPATHLEN, "resolve symlinks");
    *r++ = '/';
    *r = 0;
    p = path;
    for (;;) {
	q = strchr (p + 1, '/');
	if (!q) {
	    q = strchr (p + 1, 0);
	    if (q == p + 1)
		break;
	}
	c = *q;
	*q = 0;
	if (mc_lstat (path, &mybuf) < 0) {
	    free (buf);
	    free (buf2);
	    *q = c;
	    return NULL;
	}
	if (!S_ISLNK (mybuf.st_mode))
	    strcpy (r, p + 1);
	else {
	    len = mc_readlink (path, buf2, MC_MAXPATHLEN);
	    if (len < 0) {
		free (buf);
		free (buf2);
		*q = c;
		return NULL;
	    }
	    buf2 [len] = 0;
	    if (*buf2 == '/')
		strcpy (buf, buf2);
	    else
		strcpy (r, buf2);
	}
	canonicalize_pathname (buf);
	r = strchr (buf, 0);
	if (!*r || *(r - 1) != '/') {
	    *r++ = '/';
	    *r = 0;
	}
	*q = c;
	p = q;
	if (!c)
	    break;
    }
    if (!*buf)
	strcpy (buf, "/");
    else if (*(r - 1) == '/' && r != buf + 1)
	*(r - 1) = 0;
    free (buf2);
    return buf;
}

/* Finds out a relative path from first to second, i.e. goes as many ..
 * as needed up in first and then goes down using second */
static stdcall2 char *diff_two_paths (char *first, char *second)
{
    char *p, *q, *r, *s, *buf = 0;
    int i, j, prevlen = -1, currlen;

    first = resolve_symlinks (first);
    if (first == NULL)
	return NULL;
    for (j = 0; j < 2; j++) {
	p = first;
	if (j) {
	    second = resolve_symlinks (second);
	    if (second == NULL) {
		free (first);
		return buf;
	    }
	}
	q = second;
	for (;;) {
	    r = strchr (p, '/');
	    s = strchr (q, '/');
	    if (!r || !s)
	      break;
	    *r = 0; *s = 0;
	    if (strcmp (p, q)) {
		*r = '/'; *s = '/';
		break;
	    } else {
		*r = '/'; *s = '/';
	    }
	    p = r + 1;
	    q = s + 1;
	}
	p--;
	for (i = 0; (p = strchr (p + 1, '/')) != NULL; i++);
	currlen = (i + 1) * 3 + strlen (q) + 1;
	if (j) {
	    if (currlen < prevlen)
		free (buf);
	    else {
		free (first);
		free (second);
		return buf;
	    }
	}
	p = buf = x_malloc (currlen, "diff 2 paths");
	prevlen = currlen;
	for (; i >= 0; i--, p += 3)
	  strcpy (p, "../");
	strcpy (p, q);
    }
    free (first);
    free (second);
    return buf;
}

static char *concat_mcpath_file (const char *file)
{
    return concat_path_file(mc_home, file);
}


/* Various utilities - Unix variants */


static void init_groups (void)
{
    int i;
    struct passwd *pwd = getpwuid (current_user_uid=my_uid);
    struct group *grp;
    user_in_groups *cug, *pug;

    current_user_gid = (pug = x_malloc (sizeof (user_in_groups), "init_groups"));
    current_user_gid->gid = my_gid; current_user_gid->next = 0;

    if (pwd == 0)
       return;

    setgrent ();
    while ((grp = getgrent ()))
       for (i = 0; grp->gr_mem[i]; i++)
	   if (!strcmp (pwd->pw_name, grp->gr_mem[i])) {
	       cug = x_malloc (sizeof (user_in_groups), "init_groups");
	       cug->gid  = grp->gr_gid;
	       pug->next = cug;
	       cug->next = 0;
	       pug = cug;
	       break;
	   }
    endgrent ();
}


static void delete_groups (void)
{
    user_in_groups *pug, *cug = current_user_gid;

    while (cug) {
       pug = cug->next;
       free (cug);
       cug = pug;
    }
}


static char *i_cache_match (int id, int_cache *cache, int size)
{
    int i;

    for (i = 0; i < size; i++)
	if (cache [i].Index == id)
	    return cache [i].string;
    return 0;
}

static void i_cache_add (int id, int_cache *cache, int size, char *text,
			 int *last)
{
    if (cache [*last].string)
	free (cache [*last].string);
    cache [*last].string = strdup (text);
    cache [*last].Index = id;
    *last = ((*last)+1) % size;
}

static stdcall2 char *get_owner (int uid)
{
    struct passwd *pwd;
    char   *name;

    if ((name = i_cache_match (uid, uid_cache, UID_CACHE_SIZE)) != NULL)
	return name;

    pwd = getpwuid (uid);
    if (pwd) {
	i_cache_add (uid, uid_cache, UID_CACHE_SIZE, pwd->pw_name, &uid_last);
	return pwd->pw_name;
    }
    else {
	return sprintf_pd (get_owner_ibuf, uid);
    }
}

static stdcall2 char *get_group (int gid)
{
    struct group *grp;
    char *name;

    if ((name = i_cache_match (gid, gid_cache, GID_CACHE_SIZE)) != NULL)
	return name;

    grp = getgrgid (gid);
    if (grp) {
	i_cache_add (gid, gid_cache, GID_CACHE_SIZE, grp->gr_name, &gid_last);
	return grp->gr_name;
    } else {
	return sprintf_pd (get_group_gbuf, gid);
    }
}

int my_system (int flags, const char *shell, const char *command)
{
    struct sigaction ignore, l_save_intr, l_save_quit, l_save_stop;
    pid_t pid;
    int Status = 0;
    int as_shell_command = flags & EXECUTE_AS_SHELL;

    ignore.sa_handler = SIG_IGN;
    sigemptyset (&ignore.sa_mask);
    ignore.sa_flags = 0;

    sigaction (SIGINT, &ignore, &l_save_intr);
    sigaction (SIGQUIT, &ignore, &l_save_quit);

    /* Restore the original SIGTSTP handler, we don't want ncurses' */
    /* handler messing the screen after the SIGCONT */
    sigaction (SIGTSTP, &startup_handler, &l_save_stop);

    if ((pid = fork ()) < 0) {
	fprintf (stderr, "\n\nfork () = -1\n");
	return -1;
    }
    if (pid == 0) {
	bb_sigaction2 (SIGINT,  &l_save_intr);
	bb_sigaction2 (SIGQUIT, &l_save_quit);

	if (as_shell_command)
	    execl (shell, shell, "-c", command, (char *) 0);
	else
	    execlp (shell, shell, command, (char *) 0);

	_exit (127);            /* Exec error */
    } else {
	waitpid_no_int (pid, &Status, 0, 0);
    }
    bb_sigaction2 (SIGINT,  &l_save_intr);
    bb_sigaction2 (SIGQUIT, &l_save_quit);
    bb_sigaction2 (SIGTSTP, &l_save_stop);

    return WEXITSTATUS(Status);
}

/* Returns a newly allocated string, if directory does not exist, return 0 */
static stdcall2 char *tilde_expand (const char *directory)
{
    struct passwd *passwd;
    const char *p;

    if (*directory != '~')
	return strdup (directory);

    directory++;

    p = strchr (directory, '/');

    /* d = "~" or d = "~/" */
    if (!(*directory) || (*directory == '/')) {
	passwd = getpwuid (my_euid);
	p = (*directory == '/') ? directory+1 : "";
    } else {
	if (!p) {
	    p = "";
	    passwd = getpwnam (directory);
	} else {
	    char *name = xstrndup (directory, p - directory);
	    passwd = getpwnam (name);
	    free (name);
	}
    }

    /* If we can't figure the user name, return NULL */
    if (!passwd)
	return 0;

    return concat_path_file(passwd->pw_dir, p);
}

/* Pipes are guaranteed to be able to hold at least 4096 bytes */
/* More than that would be unportable */
#define MAX_PIPE_SIZE 4096

/* Creates a pipe to hold standard error for a later analysis. */
/* The pipe can hold 4096 bytes. Make sure no more is written */
/* or a deadlock might occur. */
void open_error_pipe (void)
{
    if (pipe (error_pipe) < 0)
	MC_message (0, " Warning ", " Pipe failed ");

    old_error = dup (2);
    if(old_error < 0 || close(2) || dup (error_pipe[1]) != 2) {
	MC_message (0, " Warning ", " Dup failed ");
	close (error_pipe[0]);
	close (error_pipe[1]);
    }
    close (error_pipe[1]);
}

static stdcall2 void close_error_pipe (int error, char *text)
{
    char *title;
    char msg[MAX_PIPE_SIZE];
    int len = 0;

    if (error)
	title = MSG_ERROR;
    else
	title = " Warning ";
    if (old_error >= 0) {
	close (2);
	dup (old_error);
	close (old_error);
	len = read (error_pipe[0], msg, MAX_PIPE_SIZE);

	if (len >= 0)
	    msg[len] = 0;
	close (error_pipe[0]);
    }
    if (error < 0)
	return;         /* Just ignore error message */
    if (text == NULL) {
	if (len == 0) return;   /* Nothing to show */

	/* Show message from pipe */
	MC_message (error, title, msg);
    } else {
	/* Show given text and possible message from pipe */
	MC_message (error, title, " %s \n %s ", text, msg);
    }
}

/* Checks for messages in the error pipe,
 * closes the pipe and displays an error box if needed
 */
void check_error_pipe (void)
{
    char error[MAX_PIPE_SIZE];
    int len = 0;

    if (old_error >= 0) {
	while (len < MAX_PIPE_SIZE) {
	    fd_set select_set_a;
	    struct timeval timeout;
	    FD_ZERO (&select_set_a);
	    FD_SET (error_pipe[0], &select_set_a);
	    timeout.tv_sec = 0;
	    timeout.tv_usec = 0;
	    select (FD_SETSIZE, &select_set_a, 0, 0, &timeout);
	    if (!FD_ISSET (0, &select_set_a))
		break;
	    read (error_pipe[0], error + len, 1);
	    len ++;
	}
	error[len] = 0;
	close (error_pipe[0]);
    }
    if (len > 0)
	MC_message (0, " Warning ", error);
}

/* INHANDLE is a result of some mc_open call to any vfs, this function
   returns a normal handle (to be used with read) of a pipe for reading
   of the output of COMMAND with arguments ... (must include argv[0] as
   well) which gets as its input at most INLEN bytes from the INHANDLE
   using mc_read. You have to call mc_doublepclose to close the returned
   handle afterwards. If INLEN is -1, we read as much as we can :) */
int mc_doublepopen (int inhandle, int inlen, pid_t *the_pid, const char *command, ...)
{
    struct sigaction ignore;
    int pipe0 [2], pipe1 [2];
    pid_t pid;

#define closepipes() close(pipe0[0]);close(pipe0[1]);close(pipe1[0]);close(pipe1[1])
#define is_a_pipe_fd(f) ((pipe0[0] == f) || (pipe0[1] == f) || (pipe1[0] == f) || (pipe1[1] == f))

    pipe (pipe0); pipe (pipe1);

    memset(&ignore, 0, sizeof(ignore));
    ignore.sa_handler = SIG_IGN;

    sigaction (SIGINT, &ignore, &save_intr);
    sigaction (SIGQUIT, &ignore, &save_quit);
    sigaction (SIGTSTP, &startup_handler, &save_stop);

    switch (pid = fork ()) {
    case -1:
	    closepipes ();
	    return -1;
    case 0: {
	    bb_sigaction2 (SIGINT, &save_intr);
	    bb_sigaction2 (SIGQUIT, &save_quit);

	    switch (pid = fork ()) {
		case -1:
		    closepipes ();
		    exit (1);
		case 0: {
#define MAXARGS 16
		     int argno;
		     char *args[MAXARGS];
		     va_list ap;
		     int nulldevice;

		     nulldevice = open (bb_dev_null, O_WRONLY);
		     close (0);
		     dup (pipe0 [0]);
		     close (1);
		     dup (pipe1 [1]);
		     close (2);
		     dup (nulldevice);
		     close (nulldevice);
		     closepipes ();
		     va_start (ap, command);
		     argno = 0;
		     while ((args[argno++] = va_arg(ap, char *)) != NULL)
			 if (argno == (MAXARGS - 1)) {
			     args[argno] = NULL;
			     break;
			 }
		     va_end (ap);
		     execvp (command, args);
		     exit (0);
		}
		default:
		    {
			char buffer [8192];
			int i;

			close (pipe0 [0]);
			close (pipe1 [0]);
			close (pipe1 [1]);
			while ((i = mc_read (inhandle, buffer,
					     (inlen == -1 || inlen > 8192)
					     ? 8192 : inlen)) > 0) {
			    write (pipe0 [1], buffer, i);
			    if (inlen != -1) {
				inlen -= i;
				if (!inlen)
				    break;
			    }
			}
			close (pipe0 [1]);

			waitpid_no_int (pid, &i, 0, 0);
			exit (i);
		    }
	    }
    }
    default:
	*the_pid = pid;
	break;
    }
    close (pipe0 [0]);
    close (pipe0 [1]);
    close (pipe1 [1]);
    return pipe1 [0];
}

static stdcall2 int mc_doublepclose (int Pipe, pid_t pid)
{
    int Status = 0;

    close (Pipe);
    waitpid_no_int (pid, &Status, 0, 0);
    bb_sigaction2 (SIGINT, &save_intr);
    bb_sigaction2 (SIGQUIT, &save_quit);
    bb_sigaction2 (SIGTSTP, &save_stop);

    return Status;
}

/* Canonicalize path, and return a new path. Do everything in situ.
   The new path differs from path in:
	Multiple `/'s are collapsed to a single `/'.
	Leading `./'s and trailing `/.'s are removed.
	Trailing `/'s are removed.
	Non-leading `../'s and trailing `..'s are handled by removing
	portions of the path. */
static stdcall2 char *canonicalize_pathname (char *path)
{
    int i, start_c;
    char stub_char;

    stub_char = (*path == '/') ? '/' : '.';

    /* Walk along path looking for things to compact. */
    i = 0;
    for (;;) {
	if (!path[i])
	    break;

	while (path[i] && path[i] != '/')
	    i++;

	start_c = i++;

	/* If we didn't find any slashes, then there is nothing left to do. */
	if (!path[start_c])
	    break;

	/* Handle multiple `/'s in a row. */
	while (path[i] == '/')
	    i++;

	if ((start_c + 1) != i) {
	    memmove (path + start_c + 1, path + i, strlen(path + i) + 1);
	    i = start_c + 1;
	}

	/* Handle backquoted `/'. */
	if (start_c > 0 && path[start_c - 1] == '\\')
	    continue;

	/* Check for trailing `/'. */
	if (start_c && !path[i]) {
	zero_last:
	    path[--i] = '\0';
	    break;
	}

	/* Check for `../', `./' or trailing `.' by itself. */
	if (path[i] == '.') {
	    /* Handle trailing `.' by itself. */
	    if (!path[i + 1])
		goto zero_last;

	    /* Handle `./'. */
	    if (path[i + 1] == '/') {
		memmove (path + i, path + i + 1, strlen(path + i));
		i = start_c;
		continue;
	    }

	    /* Handle `../' or trailing `..' by itself.
	       Remove the previous ?/ part with the exception of
	       ../, which we should leave intact. */
	    if (path[i + 1] == '.' && (path[i + 2] == '/' || !path[i + 2])) {
		while (--start_c > -1 && path[start_c] != '/');
		if (!strncmp (path + start_c + 1, "../", 3))
		    continue;
		memmove (path + start_c + 1, path + i + 2, strlen(path + i + 1));
		i = start_c;
		continue;
	    }
	}
    }

    if (!*path) {
	*path = stub_char;
	path[1] = '\0';
    }
    return path;
}

static stdcall2 void my_statfs (const char *path)
{
#ifndef NO_INFOMOUNT
    int i, len = 0;

    struct mount_entry *entry = NULL;
    struct mount_entry *temp = mount_list;
    struct fs_usage fs_use;

    while (temp) {
	i = strlen (temp->me_mountdir);
	if (i > len && (strncmp (path, temp->me_mountdir, i) == 0))
	    if (!entry || (path [i] == '/' || path [i] == 0)) {
		len = i;
		entry = temp;
	    }
	temp = temp->me_next;
    }

    if (entry && get_fs_usage (entry->me_mountdir, &fs_use) == 0) {
	myfs_stats.type = entry->me_dev;
	myfs_stats.typename = entry->me_type;
	myfs_stats.mpoint = entry->me_mountdir;
	myfs_stats.device = entry->me_devname;
	myfs_stats.avail = my_uid ? fs_use.fsu_bavail/2 : fs_use.fsu_bfree/2;
	myfs_stats.total = fs_use.fsu_blocks/2;
	myfs_stats.nfree = fs_use.fsu_ffree;
	myfs_stats.nodes = fs_use.fsu_files;
    } else
#endif
    {
	myfs_stats.mpoint = "unknown";
	myfs_stats.device = "unknown";
    }
}

/* {{{ Declarations */

/* Block size for reading files in parts */
#define READ_BLOCK 8192
#define VIEW_PAGE_SIZE 8192

#ifdef IS_AIX
#   define IFAIX(x) case (x):
#else
#   define IFAIX(x)
#endif


/* Our callback */
static int view_callback (Dlg_head *h, WView *view, int msg, int par);

/* }}} */

static const char hex_char[] = "0123456789ABCDEF";

/* }}} */
/* {{{ Clean-up functions */

static void
close_view_file (WView *view)
{
    if (view->file != -1) {
	mc_close (view->file);
	view->file = -1;
    }
}

static void
free_file (WView *view)
{
    int i;

    if (view->mmapping) {
	mc_munmap ((char *)view->data, view->s.st_size);
	close_view_file (view);
    } else
    {
	if (view->reading_pipe) {
	    /* Check error messages */
	    if (!view->have_frame)
		check_error_pipe ();

	    /* Close pipe */
	    pclose (view->stdfile);
	    view->stdfile = NULL;

	    /* Ignore errors because we don't want to hear about broken pipe */
	    close_error_pipe (-1, NULL);
	} else
	    close_view_file (view);
    }
    /* Block_ptr may be zero if the file was a file with 0 bytes */
    if (view->growing_buffer && view->block_ptr) {
	for (i = 0; i < view->blocks; i++) {
	    free (view->block_ptr [i].data);
	}
	free (view->block_ptr);
    }
}

/* Valid parameters for second parameter to set_monitor */
enum { off, on };

/* Both views */
static void
view_done (WView *view)
{
	set_monitor (view, off);
    if (view->view_active) {
	if (view->localcopy)
	    mc_ungetlocalcopy (view->filename, view->localcopy, 0);
	free_file (view);
	free (view->filename);
	if (view->command)
	    free (view->command);
    }
    view->view_active = 0;
    default_hex_mode = view->hex_mode;
    default_nroff_flag = view->viewer_nroff_flag;
    default_magic_flag = view->viewer_magic_flag;
}

static void view_hook (void *);

static void
view_destroy (WView *view)
{
    view_done (view);
    if (view->have_frame)
	delete_hook (&select_file_hook, view_hook);
}

static int
get_byte (WView *view, int byte_index)
{
    int page   = byte_index / VIEW_PAGE_SIZE + 1;
    int offset = byte_index % VIEW_PAGE_SIZE;
    int i, n;
    block_ptr_t *tmp;

    if (view->growing_buffer) {
	if (page > view->blocks) {
	    tmp = x_malloc (sizeof (block_ptr_t) * page, "get_byte");
	    if (view->block_ptr) {
		memcpy (tmp, view->block_ptr, sizeof (block_ptr_t) *
		       view->blocks);
		free (view->block_ptr);
	    }
	    view->block_ptr = tmp;
	    for (i = view->blocks; i < page; i++) {
		char *p = malloc (VIEW_PAGE_SIZE);
		view->block_ptr [i].data = (unsigned char *)p;
		if (!p)
		    return '\n';
		if (view->stdfile != NULL)
		    n = fread (p, 1, VIEW_PAGE_SIZE, view->stdfile);
		else
		    n = mc_read (view->file, p, VIEW_PAGE_SIZE);
		if (n != -1)
		    view->bytes_read += n;
		if (view->s.st_size < view->bytes_read) {
		    view->bottom_first = -1; /* Invalidate cache */
		    view->s.st_size = view->bytes_read;
		    view->last_byte = view->bytes_read;
		    if (view->reading_pipe)
			view->last_byte = view->first + view->bytes_read;
		}
		/* To force loading the next page */
		if (n == VIEW_PAGE_SIZE && view->reading_pipe) {
		    view->last_byte++;
		}
	    }
	    view->blocks = page;
	}
	if (byte_index > view->bytes_read) {
	    return -1;
	} else
	    return view->block_ptr [page-1].data [offset];
    } else {
	if (byte_index >= view->last_byte)
	    return -1;
	else
	    return view->data [byte_index];
    }
}

static void
enqueue_change (struct hexedit_change_node **head, struct hexedit_change_node *node)
{
    struct hexedit_change_node *curr = *head;

    while (curr) {
	if (node->offset < curr->offset) {
	    *head = node;
	    node->next = curr;
	    return;
	}
	head = (struct hexedit_change_node **) curr;
	curr = curr->next;
    }
    *head = node;
    node->next = curr;
}

static void move_right (WView *);

static void
put_editkey (WView *view, unsigned char key)
{
    struct hexedit_change_node *node;
    unsigned char byte_val;

    if (!view->hexedit_mode || view->growing_buffer != 0)
	return;

    /* Has there been a change at this position ? */
    node = view->change_list;
    while (node) {
	if (node->offset != view->edit_cursor)
	    node = node->next;
	else
	    break;
    }

    if (view->view_side == view_side_left) {
	/* Hex editing */

	if (key >= '0' && key <= '9')
	    key -= '0';
	else if (key >= 'A' && key <= 'F')
	    key -= '7';
	else if (key >= 'a' && key <= 'f')
	    key -= 'W';
	else
	    return;

	if (node)
	    byte_val = node->value;
	else
	    byte_val = get_byte(view, view->edit_cursor);

	if (view->nib_shift == 0) {
	    byte_val = (byte_val & 0x0f) | (key << 4);
	} else {
	    byte_val = (byte_val & 0xf0) | (key);
	}
    } else {
	/* Text editing */
	byte_val = key;
    }
    if (!node) {
	node = (struct hexedit_change_node *)
		    x_malloc(sizeof(struct hexedit_change_node), "HexEdit");

	if (node) {
	    node->offset = view->edit_cursor;
	    node->value = byte_val;
	    enqueue_change (&view->change_list, node);
	}
    } else {
	node->value = byte_val;
    }
    view->dirty++;
    view_update (view);
    move_right (view);
}

static void
free_change_list (WView *view)
{
    struct hexedit_change_node *n = view->change_list;

    while (n) {
	view->change_list = n->next;
	free (n);
	n = view->change_list;
    }
    view->file_dirty = 0;
    view->dirty++;
}

static void
save_edit_changes (WView *view)
{
    struct hexedit_change_node *node = view->change_list;
    int fp;

    fp = open (view->filename, O_WRONLY);
    if (fp >= 0) {
	while (node) {
	    lseek (fp, node->offset, SEEK_SET);
	    write (fp, &node->value, 1);
	    node = node->next;
	}
	close (fp);
    }
    free_change_list (view);
}

static int
view_ok_to_quit (WView *view)
{
    int r;
    char *text;

    if (!view->change_list)
	return 1;

    query_set_sel (1);
    text = xasprintf ("File: \n\n    %s"
			 "\n\nhas been modified, do you want to save the changes?\n"
			 , view->filename);

    r = query_dialog (" Save changes ", text, 2, 3, "&Yes", "&No", "&Cancel");
    free (text);

    switch (r) {
    case 0:
	save_edit_changes (view);
	return 1;
    case 1:
	free_change_list (view);
	return 1;
    default:
	return 0;
    }
}

static char *
set_view_init_error (WView *view, char *msg)
{
    view->growing_buffer = 0;
    view->reading_pipe   = 0;
    view->first = 0;
    view->last_byte = 0;
    if (msg) {
	view->bytes_read = strlen (msg);
	return strdup (msg);
    }
    return 0;
}

/* return values: 0 for success, else points to error message */
static char *
init_growing_view (WView *view, char *name, char *filename)
{
    view->growing_buffer = 1;

    if (name) {
	view->reading_pipe = 1;
	view->s.st_size = 0;

	open_error_pipe ();
	if ((view->stdfile = popen (name, "r")) == NULL) {
	    close_error_pipe (view->have_frame?-1:1, (char *)view->data);
	    return set_view_init_error (view, " Can't spawn child program ");
	}

	/* First, check if filter produced any output */
	get_byte (view, 0);
	if (view->bytes_read <= 0) {
	    pclose (view->stdfile);
	    view->stdfile = NULL;
	    close_error_pipe (view->have_frame?-1:1, (char *)view->data);
	    return set_view_init_error (view, " Empty output from child filter ");
	}
    } else {
	view->stdfile = NULL;
	if ((view->file = mc_open (filename, O_RDONLY)) == -1)
	    return set_view_init_error (view, " Could not open file ");
    }
    return 0;
}

/* Load filename into core */
/* returns:
   -1 on failure.
   if (have_frame), we return success, but data points to a
   error message instead of the file buffer (quick_view feature).
*/
static char *load_view_file (WView *view, char *filename)
{
    char *cmd;
    int  type;

    if ((view->file = mc_open (filename, O_RDONLY)) < 0) {
	set_view_init_error (view, 0);
	return (xasprintf (" Can't open file \"%s\"\n %s ",
			      filename, unix_error_string()));

    }
    if (mc_fstat (view->file, &view->s) < 0) {
	set_view_init_error (view, 0);
	close_view_file (view);
	return xasprintf (" Can't stat file\n %s ", unix_error_string());
    }
    if (S_ISDIR (view->s.st_mode) || S_ISSOCK (view->s.st_mode)
	|| S_ISFIFO (view->s.st_mode)) {
	close_view_file (view);
	return set_view_init_error (view, " Can't view: not a regular file ");
    }

    if (view->s.st_size == 0) {
	/* Must be one of those nice files that grow (/proc) */
	close_view_file (view);
	return init_growing_view (view, 0, filename);
    }

    /* First, try to open a compressed file */
    if (view->viewer_magic_flag && (is_gunzipable (view->file, &type)) != 0) {
	close_view_file (view);
	if (vfs_file_is_local (filename)) {
	    char *tmp = name_quote (filename, 0);
	    cmd = xasprintf ("%s %s", decompress_command (type), tmp);
	    free (tmp);
	}
	else {
	    char *tmp;
	    if ((view->localcopy = mc_getlocalcopy (filename)) == 0)
		return set_view_init_error (view, " Can not fetch local copy " );
	    tmp = name_quote (view->localcopy, 0);
	    cmd = xasprintf ("%s %s", decompress_command (type), tmp);
	    free (tmp);
	}
	return init_growing_view (view, cmd, filename);
    }

    /* Otherwise, the file wasn't compressed */
    view->data = mc_mmap (0, view->s.st_size, PROT_READ, MAP_FILE | MAP_SHARED,
			  view->file, 0);
    if ((caddr_t) view->data == (caddr_t) -1) {
	close_view_file (view);
/*      set_view_init_error (view, 0);
	return xasprintf (" Can't mmap file\n %s ", unix_error_string());*/
	return init_growing_view (view, 0, filename);

    }

    view->first = 0;
    view->bytes_read = view->s.st_size;
    view->mmapping = 1;
    return 0;
}

/* Return zero on success, -1 on failure */
static int
do_view_init (WView *view, char *_command, char *_file, int start_line)
{
    char *error = 0;
    int i;

    if (view->view_active)
	view_done (view);

    /* Set up the state */
    view->block_ptr = 0;
    view->data = NULL;
    view->growing_buffer = 0;
    view->reading_pipe = 0;
    view->mmapping = 0;
    view->blocks = 0;
    view->block_ptr = 0;
    view->first = view->bytes_read = 0;
    view->last_byte = 0;
    view->filename = 0;
    view->localcopy = 0;
    view->command = 0;
    view->last = view->first + ((LINES-2) * view->bytes_per_line);

    /* Clear the markers */
    view->marker = 0;
    for (i = 0; i < 10; i++)
	view->marks [i] = 0;


    if (!view->have_frame) {
	view->start_col = 0;
    }
    if (_command && (view->viewer_magic_flag || _file[0] == '\0'))
	error = init_growing_view (view, _command, _file);
    else
	error = load_view_file (view, _file);

    if (error) {
	if (!view->have_frame) {
	    MC_message (1, MSG_ERROR, error);
	    free (error);
	    return -1;
	}
    }

    view->view_active = 1;
    view->filename = strdup (_file);
    if (_command)
	view->command = strdup (_command);
    else
	view->command = 0;
    view->search_start = view->start_display = view->start_save = view->first;
    view->found_len = 0;
    view->start_col = 0;
    view->last_search = 0;            /* Start a new search */

    /* Special case: The data points to the error message */
    if (error) {
	view->data = (unsigned char *)error;
	view->s.st_size = view->bytes_read = strlen ((char *)view->data);
    }
    view->last_byte = view->first + view->s.st_size;

    if (start_line > 1 && !error) {
	int saved_wrap_mode = view->wrap_mode;

	view->wrap_mode = 0;
	get_byte (view, 0);
	view_move_forward (view, start_line - 1);
	view->wrap_mode = saved_wrap_mode;
    }
    view->edit_cursor = view->first;
    view->file_dirty =  0;
    view->nib_shift = 0;
    view->view_side = view_side_left;
    view->change_list = NULL;

    return 0;
}


/* Both views */
/* Return zero on success, -1 on failure */
int
view_init (WView *view, char *_command, char *_file, int start_line)
{
    int cols;

    if (view->have_frame)
	cols = view->widget.cols - 2;
    else
	cols = view->widget.cols;

    view->bottom_first = -1;
    view->bytes_per_line = 2 * (cols - 7) / 9;
    view->bytes_per_line &= 0xfffc;
    view->dirty = max_dirt_limit + 1;   /* To force refresh */
    if (!view->view_active || strcmp (_file, view->filename) || altered_magic_flag)
	return do_view_init (view, _command, _file, start_line);
    else
	return 0;
}

/* }}} */

/* {{{ Screen update functions */

static void
view_percent (WView *view, int p, int w)
{
    int percent;

    percent = (view->s.st_size == 0 || view->last_byte == view->last) ? 100 :
	(p > (INT_MAX/100) ?
	 p / (view->s.st_size / 100) :
	 p * 100 / view->s.st_size);

    widget_move (view, view->have_frame, w - 5);
    printw ("%3d%% ", percent);
}

static void
view_status (WView *view)
{
    int w = view->widget.cols - (view->have_frame * 2);
    int i;

    attrset (SELECTED_COLOR);
    widget_move (view, view->have_frame, view->have_frame);
    hline (' ', w);
    if (w > 6) {
	i = w > 24 ? 18 : w - 6;
	printw ("File: %s", name_trunc (view->filename ? view->filename:
					view->command ? view->command:"", i));
	if (w > 30) {
	    widget_move (view, view->have_frame, 24);
	    if (view->hex_mode)
		printw ("Offset 0x%08x", view->edit_cursor);
	    else
		printw ("Col %d", -view->start_col);
	}
	if (w > 60) {
	    widget_move (view, view->have_frame, 42);
	    printw ("%s bytes", size_trunc (view->s.st_size));
	}
	if (w > 70) {
	    printw (" ");
	    if (view->growing_buffer)
		addstr ("  [grow]");
	}
	if (w - i > 4) {
	    if (view->hex_mode)
		view_percent (view, view->edit_cursor - view->first, w);
	    else
		view_percent (view, view->start_display - view->first, w);
	}
    }
    attrset (SELECTED_COLOR);
}

#define view_set_color(view,font) attrset (font)

static inline void
view_display_clean (WView *view, int height_d, int width)
{
    /* FIXME: Should I use widget_erase only and repaint the box? */
    if (view->have_frame) {
	int i;

	draw_double_box (view->widget.parent, view->widget.y, view->widget.x,
			 view->widget.lines, view->widget.cols);
	for (i = 1; i < height_d; i++) {
	    widget_move (view, i, 1);
	    printw ("%*s", width-1, "");
	}
    } else
	widget_erase ((Widget *) view);
}

#define view_add_character(view,c) addch (c)
#define view_add_string(view,s)    addstr (s)
#define view_gotoyx(v,r,c)    widget_move (v,r,c)

/* Both the text mode and gnome editions use this */
#define BOLD_COLOR        MARKED_COLOR
#define UNDERLINE_COLOR   VIEW_UNDERLINED_COLOR
#define MARK_COLOR        SELECTED_COLOR
#define DEF_COLOR         NORMAL_COLOR

#ifndef PORT_HAS_VIEW_FREEZE
#    define view_freeze(view)
#    define view_thaw(view)
#endif

#define PICK_COLOR(a,b) a : b

/* Shows the file pointed to by *start_display on view_win */
static long
display (WView *view)
{
    const int frame_shift = view->have_frame;
#define STATUS_LINES 1
    int col = 0 + frame_shift;
    int row = STATUS_LINES + frame_shift;
    int height_d, width;
    long from;
    int c;
    int boldflag = 0;
    struct hexedit_change_node *curr = view->change_list;

    height_d = view->widget.lines - frame_shift;
    width = view->widget.cols - frame_shift;
    from = view->start_display;
    view_set_color (view, DEF_COLOR);

    view_freeze (view);
    view_display_clean (view, height_d, width);

    /* Optionally, display a ruler */
    if ((!view->hex_mode) && (ruler)) {
	char r_buff[4];
	int cl;

	view_set_color (view, BOLD_COLOR);
	for (c = frame_shift; c < width; c++) {
	    cl = c-view->start_col;
	    if (ruler == 1)
		view_gotoyx (view, row, c);
	    else
		view_gotoyx (view, row+height_d-2, c);
	    r_buff[0] = '-';
	    if ((cl % 10) == 0)
		r_buff[0] = '|';
	    else
		if ((cl % 5) == 0)
		    r_buff[0] = '*';
	    view_add_character (view, r_buff[0]);
	    if ((cl != 0) && (cl % 10) == 0) {
		sprintf(r_buff, "%03d", cl);
		if (ruler == 1)
		    widget_move (view, row + 1, c - 1);
		else
		    widget_move (view, row + height_d - 3, c - 1);
		view_add_string (view, r_buff);
	    }
	}
	view_set_color (view, DEF_COLOR);
	if (ruler == 1)
	    row += 2;
	else
	    height_d -= 2;
    }

    /* Find the first displayable changed byte */
    while (curr) {
	if (curr->offset < from)
	    curr = curr->next;
	else
	    break;
    }
    if (view->hex_mode) {
	char hex_buff[10];   /* A temporary buffer for sprintf and mvwaddstr */
	int bytes;           /* Number of bytes already printed on the line */

	/* Start of text column */
	int text_start_col = width - view->bytes_per_line - 1 + frame_shift;

	for (;row < height_d && from < view->last_byte; row++) {
	    /* Print the hex offset */
	    sprintf (hex_buff, "%05X", (int) (from - view->first));
	    widget_move (view, row, frame_shift);
	    view_add_string (view, hex_buff);

	    /* Hex dump starts from column seven */
	    col = 7;

	    /* Each hex number is two digits */
	    hex_buff[2] = 0;
	    for (bytes = 0; bytes < view->bytes_per_line
		 && from < view->last_byte; bytes++, from++) {
		/* Display and mark changed bytes */
		if (curr && from == curr->offset) {
		    c = curr->value;
		    curr = curr->next;
		    boldflag = 3;
		    view_set_color (view, 7);
		} else
		c = (unsigned char) get_byte (view, from);

		if (view->found_len && from >= view->search_start
		    && from < view->search_start + view->found_len) {
		    boldflag = 1;
		    view_set_color (view, BOLD_COLOR);
		}
		/* Display the navigation cursor */
		if (from == view->edit_cursor) {
		    if (view->view_side == view_side_left) {
			view->cursor_row = row;
			view->cursor_col = col;
		    }
		    boldflag = 2;
		    view_set_color (view, view->view_side == view_side_left ? PICK_COLOR (15, 31));
		}

		/* Print a hex number (sprintf is too slow) */
		hex_buff [0] = hex_char [(c >> 4)];
		hex_buff [1] = hex_char [c & 15];
		view_gotoyx (view, row, col);
		view_add_string (view, hex_buff);
		col += 3;
		/* Turn off the cursor or changed byte highlighting here */
		if (boldflag > 1)
		    view_set_color (view, DEF_COLOR);
		if ((bytes & 3) == 3 && bytes + 1 < view->bytes_per_line) {
		    /* Turn off the search highlighting */
		    if (boldflag == 1 &&
			    from == view->search_start + view->found_len - 1)
			view_set_color (view, DEF_COLOR);

		    /* Hex numbers are printed in the groups of four */
		    /* Groups are separated by a vline */

		    view_add_character (view, ' ');
		    one_vline ();
		    view_gotoyx (view, row, col + 1);
		    col += 2;

		    if (boldflag && from==view->search_start+view->found_len-1)
			view_set_color (view, BOLD_COLOR);

		}
		if (boldflag && from < view->search_start + view->found_len - 1
		    && bytes != view->bytes_per_line - 1)
		    view_add_character (view, ' ');

		/* Print the corresponding ascii character */
		view_gotoyx (view, row, text_start_col + bytes);

		if (!is_printable (c))
		    c = '.';
		switch (boldflag) {
		    default:
			break;
		    case 1:
			view_set_color (view, BOLD_COLOR);
			goto setcursor;
		    case 2:
			view_set_color (view, view->view_side == view_side_left ? PICK_COLOR (31, 15));
			goto setcursor;
		    case 3:
			view_set_color (view, 7);

		    setcursor:
		    if (view->view_side == view_side_right) {
			view->cursor_col = text_start_col + bytes;
			view->cursor_row = row;
		    }
		}
		view_add_character (view, c);

		if (boldflag) {
		    boldflag = 0;
		    view_set_color (view, DEF_COLOR);
		}
	    }
	}
    } else {
	if (view->growing_buffer && from == view->last_byte)
	    get_byte (view, from);
	for (; row < height_d && from < view->last_byte; from++) {
	    c = get_byte (view, from);
	    if ((c == '\n') || (col == width && view->wrap_mode)) {
		col = frame_shift;
		row++;
		if (c == '\n' || row >= height_d)
		    continue;
	    }
	    if (c == '\r')
		continue;
	    if (c == '\t') {
		col = ((col - frame_shift)/8)*8 + 8 + frame_shift;
		continue;
	    }
	    if (view->viewer_nroff_flag && c == '\b') {
		if (from + 1 < view->last_byte
		    && is_printable (get_byte (view, from + 1)) &&
		    from > view->first
		    && is_printable (get_byte (view, from - 1)))
		{
		    if (col <= frame_shift) {
			/* So it has to be wrap_mode - do not need to check for it */
			if (row == 1 + frame_shift) {
			    from++;
			    continue; /* There had to be a bold character on the rightmost position
					 of the previous undisplayed line */
			}
			row--;
			col = width;
		    }
		    col--;
		    boldflag = 1;
		    if (get_byte (view, from - 1) == '_' && get_byte (view, from + 1) != '_')
			view_set_color (view, UNDERLINE_COLOR);
		    else
			view_set_color (view, BOLD_COLOR);
		    continue;
		}
	    }
	    if (view->found_len && from >= view->search_start
		&& from < view->search_start + view->found_len) {
		boldflag = 1;
		view_set_color (view, MARK_COLOR);
	    }
	    if (col >= frame_shift-view->start_col
		&& col < width-view->start_col)
	    {
		view_gotoyx (view, row, col+view->start_col);
		if (!is_printable (c))
		    c = '.';

		view_add_character (view, c);
	    }
	    col++;
	    if (boldflag) {
		boldflag = 0;
		view_set_color (view, DEF_COLOR);
	    }

	    /* Very last thing */
	    if (view->growing_buffer && from+1 == view->last_byte)
		 get_byte (view, from+1);
	}
    }
    view->last = from;
    view_thaw (view);
    return from;
}

static void
view_place_cursor (WView *view)
{
    int shift;

    if (view->view_side == view_side_left)
	shift = view->nib_shift;
    else
	shift = 0;

    widget_move (&view->widget, view->cursor_row, view->cursor_col + shift);
}

void
view_update (WView *view)
{


    if (view->dirty > dirt_limit) {
	/* Too many updates skipped -> force a update */
	display (view);
	view_status (view);
	view->dirty = 0;
	/* Raise the update skipping limit */
	dirt_limit++;
	if (dirt_limit > max_dirt_limit)
	    dirt_limit = max_dirt_limit;
    }
    if (view->dirty) {
	if (is_idle ()) {
	    /* We have time to update the screen properly */
	    display (view);
	    view_status (view);
	    view->dirty = 0;
	    if (dirt_limit > 1)
		dirt_limit--;
	} else {
	    /* We are busy -> skipping full update,
	       only the status line is updated */
	    view_status (view);
	}
	/* Here we had a refresh, if fast scrolling does not work
	   restore the refresh, although this should not happen */
    }
}

static inline void
my_define (Dlg_head *h, int idx, char *text,
			 void (*fn)(WView *), WView *view)
{
    define_label_data (h, (Widget *) view, idx, text, (buttonbarfn) fn, view);
}

/* }}} */
/* {{{ Movement functions */
/* If the last parameter is nonzero, it means we want get the count of lines
   from current up to the the upto position inclusive */
static long
move_forward2 (WView *view, long current, int lines, long upto)
{
    long p, q;
    int  line;
    int  col = 0;

    if (view->hex_mode) {
	p = current + lines * view->bytes_per_line;
	p = (p >= view->last_byte) ? current : p;
	if (lines == 1) {
	    q = view->edit_cursor + view->bytes_per_line;
	    line = q / view->bytes_per_line;
	    col = (view->last_byte-1) / view->bytes_per_line;
	    view->edit_cursor = (line > col) ? view->edit_cursor : q;
	    view->edit_cursor = (view->edit_cursor < view->last_byte) ?
		view->edit_cursor : view->last_byte-1;
	    q = current + ((LINES-2) * view->bytes_per_line);
	    p = (view->edit_cursor < q) ? current : p;
	} else {
	    view->edit_cursor = (view->edit_cursor < p) ?
		p : view->edit_cursor;
	}
	return p;
    } else {
	if (upto) {
	    lines = -1;
	    q = upto;
	} else
	    q = view->last_byte;
	if (get_byte (view, q) != '\n')
	    q++;
	for (line = col = 0, p = current; p < q; p++) {
	    int c;

	    if (lines != -1 && line >= lines)
		return p;

	    c = get_byte (view, p);

	    if (view->wrap_mode) {
		if (c == '\r')
		    continue; /* This characters is never displayed */
		else if (c == '\t')
		    col = ((col - view->have_frame)/8)*8 +8+ view->have_frame;
		else
		    col++;
		if (view->viewer_nroff_flag && c == '\b') {
		    if (p + 1 < view->last_byte
			&& is_printable (get_byte (view, p + 1))
			&& p > view->first
			&& is_printable (get_byte (view, p - 1)))
			col -= 2;
		} else if (col == vwidth) {
		    /* FIXME: the c in is_printable was a p, that is a bug,
		       I suspect I got that fix from Jakub, same applies
		       for d. */
		    int d = get_byte (view, p+2);

		    if (p + 2 >= view->last_byte || !is_printable (c) ||
			!view->viewer_nroff_flag || get_byte (view, p + 1) != '\b' ||
			!is_printable (d)) {
			col = 0;

			if (c == '\n' || get_byte (view, p+1) != '\n')
			    line++;
		    }
		} else if (c == '\n') {
		    line++;
		    col = 0;
		}
	    } else if (c == '\n')
		line++;
	}
	if (upto)
	    return line;
    }
    return current;
}

/* returns the new current pointer */
/* Cause even the forward routine became very complex, we in the wrap_mode
   just find the nearest '\n', use move_forward2(p, 0, q) to get the count
   of lines up to there and then use move_forward2(p, something, 0), which we
   return */
static long
move_backward2 (WView *view, long current, int lines)
{
    long p, q, pm;
    int line;

    if (!view->hex_mode && current == view->first)
	return current;

    if (view->hex_mode) {
	p = current - lines * view->bytes_per_line;
	p = (p < view->first) ? view->first : p;
	if (lines == 1) {
	    q = view->edit_cursor - view->bytes_per_line;
	    view->edit_cursor = (q < view->first) ? view->edit_cursor : q;
	    p = (view->edit_cursor >= current) ? current : p;
	} else {
	    q = p + ((LINES-2) * view->bytes_per_line);
	    view->edit_cursor = (view->edit_cursor >= q) ?
		p : view->edit_cursor;
	}
	return p;
    } else {
	if (current == view->last_byte
	    && get_byte (view, current - 1) != '\n')
	    /* There is one virtual '\n' at the end,
	       so that the last line is shown */
	    line = 1;
	else
	    line = 0;
	for (q = p = current - 1; p >= view->first; p--)
	    if (get_byte (view, p) == '\n' || p == view->first) {
		pm = p > view->first ? p + 1 : view->first;
		if (!view->wrap_mode) {
		    if (line == lines)
			return pm;
		    line++;
		} else {
		    line += move_forward2 (view, pm, 0, q);
		    if (line >= lines) {
			if (line == lines)
			    return pm;
			else
			    return move_forward2 (view, pm, line - lines, 0);
		    }
		    q = p + 1;
		}
	    }
    }
    return p > view->first ? p : view->first;
}

void
view_move_backward (WView *view, int i)
{
    view->search_start = view->start_display =
	move_backward2 (view, view->start_display, i);
    view->found_len = 0;
    view->last = view->first + ((LINES-2) * view->bytes_per_line);
    view->dirty++;
}

static long
get_bottom_first (WView *view, int do_not_cache, int really)
{
    int bottom_first;

    if (!have_fast_cpu && !really)
	return INT_MAX;

    if (!do_not_cache && view->bottom_first != -1)
	return view->bottom_first;

    /* Force loading */
    if (view->growing_buffer) {
	int old_last_byte;

	old_last_byte = -1;
	while (old_last_byte != view->last_byte) {
	    old_last_byte = view->last_byte;
	    get_byte (view, view->last_byte+VIEW_PAGE_SIZE);
	}
    }

    bottom_first = move_backward2 (view, view->last_byte, vheight - 1);

    if (view->hex_mode)
	bottom_first = (bottom_first + view->bytes_per_line - 1)
	    / view->bytes_per_line * view->bytes_per_line;
    view->bottom_first = bottom_first;

    return view->bottom_first;
}

void
view_move_forward (WView *view, int i)
{
    view->start_display = move_forward2 (view, view->start_display, i, 0);
    if (!view->reading_pipe && view->start_display > get_bottom_first (view, 0, 0))
	view->start_display = view->bottom_first;
    view->search_start = view->start_display;
    view->found_len = 0;
    view->last = view->first + ((LINES-2) * view->bytes_per_line);
    view->dirty++;
}


static void
move_to_top (WView *view)
{
    view->search_start = view->start_display = view->first;
    view->found_len = 0;
    view->last = view->first + ((LINES-2) * view->bytes_per_line);
    view->nib_shift = 0;
    view->edit_cursor = view->start_display;
    view->dirty++;
}

static void
move_to_bottom (WView *view)
{
    view->search_start = view->start_display = get_bottom_first (view, 0, 1);
    view->found_len = 0;
    view->last = view->first + ((LINES-2) * view->bytes_per_line);
    view->edit_cursor = (view->edit_cursor < view->start_display) ?
	view->start_display : view->edit_cursor;
    view->dirty++;
}

/* Scroll left/right the view panel functions */
static void
move_right (WView *view)
{
    if (view->wrap_mode && !view->hex_mode)
	return;
    if (view->hex_mode) {
	view->last = view->first + ((LINES-2) * view->bytes_per_line);

	if (view->hex_mode && view->view_side == view_side_left) {
	    view->nib_shift = 1 - view->nib_shift;
	    if (view->nib_shift == 1)
		return;
	}
	view->edit_cursor = (++view->edit_cursor <  view->last_byte) ?
	    view->edit_cursor : view->last_byte - 1;
	if (view->edit_cursor >= view->last) {
	   view->edit_cursor -= view->bytes_per_line;
	   view_move_forward(view, 1);
	}
    } else
    if (--view->start_col > 0)
	view->start_col = 0;
    view->dirty++;
}

static void
move_left (WView *view)
{
    if (view->wrap_mode && !view->hex_mode)
	return;
    if (view->hex_mode) {
	if (view->hex_mode && view->view_side == view_side_left) {
	    view->nib_shift = 1 - view->nib_shift;
	    if (view->nib_shift == 0)
		return;
	}
	view->edit_cursor = (--view->edit_cursor < view->first) ?
	    view->first : view->edit_cursor;
	if (view->edit_cursor < view->start_display) {
	   view->edit_cursor += view->bytes_per_line;
	   view_move_backward(view, 1);
	}
    } else
    if (++view->start_col > 0)
	view->start_col = 0;
    view->dirty++;
}

/* }}} */
/* {{{ Search routines */

/* Case insensitive search of text in data */
static int
icase_search_p (WView *view, char *text, char *data, int nothing)
{
    int p = 0, lng;
    char *q;

    p = (q = _icase_search (text, data, &lng)) != 0;
    if (p) {
	view->found_len = lng;
	view->search_start = q - data - view->found_len;
    }
    return p;
}

static char *
grow_string_buffer (char *text, int *size)
{
    char *new;
    int  old_size = *size;

    /* The grow steps */
    *size += 160;
    new = x_malloc (*size, "grow_string_buffer");
    if (text) {
	strncpy (new, text, old_size);
	free (text);
    } else {
	*new = 0;
    }
    return new;
}

static char *
get_line_at (WView *view, long *p)
{
    char *buffer = 0;
    int  buffer_size, usable_size;
    int  ch;
    int  direction;
    long pos = *p;
    long i;

    direction = view->direction;
    buffer_size = usable_size = 0;

    i = ch = 0;
    for (;pos >= 0 && (ch = get_byte (view, pos))!= -1; pos += direction, i++) {

	/* skip over all the possible zeros in the file */
	if (ch == 0 && i == 0) {
	    while (pos >= 0 && ((ch = get_byte (view, pos)) != -1) && ch == 0)
		pos+= direction;
	    if (ch == -1)
		break;
	}
	if (i == usable_size) {
	    buffer = grow_string_buffer (buffer, &buffer_size);
	    usable_size = buffer_size - 2;
	    buffer [0] = ' '; /* This makes possible strcpy of buffer */
	}
	buffer [i+1] = ch;
	if (ch == '\n' || !ch || ch == -1) {
	    pos += direction; i++;
	    break;
	}
    }
    if (buffer) {
	i--;
	buffer [0] = ' ';
	buffer [i+1] = 0;

	/* If we are searching backwards, reverse the string */
	if (view->direction < 0)
	    reverse_string (buffer);
    }

    *p = pos;
    return buffer;
}

/** Search status optmizations **/

static void
search_update_steps (WView *view)
{
    if (view->s.st_size)
	update_steps = 40000;
    else
	update_steps = view->last_byte / 100;

    /* Do not update the percent display but every 20 ks */
    if (update_steps < 20000)
	update_steps = 20000;
}

static void
search (WView *view, char *text, int (*Fsearch)(WView *, char *, char *, int))
{
    int w = view->widget.cols - (view->have_frame * 2);
    char *s = NULL;             /*  The line we read from the view buffer */
    long p, beginning;
    int isatbeg; /* Nonzero means we start search at beginning of some line */
    int found_len, search_start;
    Dlg_head *d = 0;
    int search_status;

    /* Used to keep track of where the line starts, when looking forward */
    /* is the Index before transfering the line; the reverse case uses   */
    /* the position returned after the line has been read */
    long forward_line_start;
    long reverse_line_start;
    long t;
    /* Clear interrupt status */
    got_interrupt ();

    if (MC_verbose) {
	d = MC_message (D_INSERT, " Search ", "Searching %s", text);
	mc_refresh ();
    }
    if (view->direction == 1) {
	p = view->found_len ? view->search_start + 1 : view->search_start;
    } else {
	p = (view->found_len ? view->search_start : view->last) - 1;
    }
    beginning = p;

    isatbeg = view->found_len == 0;
    found_len = view->found_len;
    search_start = view->search_start;

    /* Compute the percent steps */
    search_update_steps (view);
    update_activate = 0;

    for ( ; ; isatbeg = 1, free (s)) {
	if (p >= update_activate) {
	    update_activate += update_steps;
	    if (MC_verbose) {
		view_percent (view, p, w);
		mc_refresh ();
	    }
	    if (got_interrupt ())
		break;
	}
	forward_line_start = p;
	disable_interrupt_key ();
	s = get_line_at (view, &p);
	reverse_line_start = p;
	enable_interrupt_key ();
	if (!s)
	    break;

	search_status = (*Fsearch) (view, text, s + 1, match_normal);
	if (search_status < 0)
	    break;

	if (search_status == 0)
	    continue;

	/* We found the string */

	if (!isatbeg && !view->search_start) {

	    /* We do not want to match a
	     * ^ regexp when not at the real
	     * beginning of some line
	     */
	    view->found_len = found_len;
	    view->search_start = search_start;
	    if ((*Fsearch) (view, text, s, match_normal) <= 0)
		continue;
	    (*Fsearch) (view, text, s + 1, match_normal);
	}
	/* Record the position used to continue the search */
	if (view->direction == 1)
	    t = forward_line_start;
	else
	    t = reverse_line_start ? reverse_line_start + 3 : 0;
	view->search_start += t;

	if (t != beginning) {
	    if (t > get_bottom_first (view, 0, 0))
		view->start_display = view->bottom_first;
	    else
		view->start_display = t;
	}

	free (s);
	break;
    }
    disable_interrupt_key ();
    if (MC_verbose) {
	dlg_run_done (d);
	destroy_dlg (d);
    }

    if (!s) {
	MC_message (0, " Search ", " Search string not found ");
	view->found_len = 0;
    }
}

/* Search buffer (it's size is len) in the complete buffer */
/* returns the position where the block was found or -1 if not found */
static long
block_search (WView *view, char *buffer, int len)
{
    int w = view->widget.cols - (view->have_frame * 2);
    char *d = buffer, b;
    long e;

    /* clear interrupt status */
    got_interrupt ();
    enable_interrupt_key ();
    e = view->found_len ? view->search_start + 1 : view->search_start;

    search_update_steps (view);
    update_activate = 0;

    for (; e < view->last_byte; e++) {
	if (e >= update_activate) {
	    update_activate += update_steps;
	    if (MC_verbose) {
		view_percent (view, e, w);
		mc_refresh ();
	    }
	    if (got_interrupt ())
		break;
	}
	b = get_byte (view, e);

	if (*d == b) {
	    d++;
	} else {
	    e -= d - buffer;
	    d = buffer;
	}
	if (d - buffer == len) {
	    disable_interrupt_key ();
	    return e - len;
	}
    }
    disable_interrupt_key ();
    return -1;
}

/* States of our funny recognizer */
enum {
    normal,
    inside_quotes,
    zero,
    hex1,
    hex2,
    oct1
};

/* This routine doesn't report all the user mistakes, it just ignores them */
static void
hex_search (WView *view, char *text)
{
    char buffer [120];          /* Where we hold the information */
    int  i, block_len;
    int  v = 0;
    long pos;                   /* Where did we found the string */
    char *p;                    /* Temporary */
    int  state = normal;        /* Initial state of the micro-scanner */

    /* First convert the string to a stream of bytes */
    for (i = block_len = 0; text [i] && block_len < sizeof (buffer); i++) {
	switch (state) {
	case inside_quotes:
	    if (text [i] == '"')
		state = normal;
	    else
		buffer [block_len++] = text [i];
	    break;

	case normal:
	    if (text [i] == '"') {
		state = inside_quotes;
		break;
	    }
	    if (text [i] == '0') {
		state = zero;
		break;
	    }
	    if (text [i] == 'x') {
		state = hex1;
		break;
	    }
	    break;

	case zero:
	    if (text [i] == 'x')
		state = hex1;
	    break;

	case hex1:
	    v = 0;
	    text [i] = toupper (text [i]);
	    if ((p = strchr (hex_char, text [i])) != 0) {
		v = (p - hex_char) << 4;
		state = hex2;
	    }
	    break;

	case hex2:
	    text [i] = toupper (text [i]);
	    if ((p = strchr (hex_char, text [i])) != 0) {
		v |= (p - hex_char);
		state = normal;
	    }
	    buffer [block_len++] = v;
	    break;
	}
    }
    /* Then start the search */
    pos = block_search (view, buffer, block_len);
    if (pos == -1) {
	MC_message (0, " Search ", " Search string not found ");
	view->found_len = 0;
	return;
    }

    view->search_start = pos + 1;
    view->found_len = block_len;
    /* Set the edit cursor to the search position, left nibble */
    view->edit_cursor = view->search_start;
    view->nib_shift = 0;

    /* Adjust the file offset */
    view->start_display = (pos & (~(view->bytes_per_line-1)));
    if (view->start_display > get_bottom_first (view, 0, 0))
	view->start_display = view->bottom_first;
}

static int regexp_view_search (WView *view, char *pattern, char *string, int match_type)
{
    regmatch_t pmatch[1];
    int i, flags = REG_ICASE;

    if (!old_pattern || strcmp (old_pattern, pattern) || old_type != match_type) {
	if (old_pattern) {
	    regfree (&rvs_r);
	    free (old_pattern);
	    old_pattern = 0;
	}
	for (i = 0; pattern[i] != 0; i++) {
	    if (isupper ((unsigned char) pattern[i])) {
		flags = 0;
		break;
	    }
	}
	flags |= REG_EXTENDED;
	if (regcomp (&rvs_r, pattern, flags)) {
	    MC_message (1, MSG_ERROR, " Invalid regular expression ");
	    return -1;
	}
	old_pattern = strdup (pattern);
	old_type = match_type;
    }
    if (regexec (&rvs_r, string, 1, pmatch, 0) != 0)
	return 0;
    view->found_len = pmatch[0].rm_eo - pmatch[0].rm_so;
    view->search_start = pmatch[0].rm_so;
    return 1;
}

static void do_regexp_search (void *xview, char *regexp)
{
    WView *view = (WView *) xview;

    view->search_exp = regexp;
    search (view, regexp, regexp_view_search);
    /* Had a refresh here */
    view->dirty++;
    view_update (view);
}

static void do_normal_search (void *xview, char *text)
{
    WView *view = (WView *) xview;

    view->search_exp = text;
    if (view->hex_mode)
	hex_search (view, text);
    else
	search (view, text, icase_search_p);
    /* Had a refresh here */
    view->dirty++;
    view_update (view);
}

/* }}} */
/* {{{ Mouse and keyboard handling */

/* Real view only */
static void v_help_cmd (void)
{
    char *hlpfile = concat_mcpath_file ("mc.hlp");
    interactive_display (hlpfile, "[Internal File Viewer]");
    free (hlpfile);
    /*
    view_refresh (0);
    */
}

/* Both views */
void toggle_wrap_mode (WView *view)
{
    if (view->hex_mode) {
	if (view->growing_buffer != 0) {
	    return;
	}
	get_bottom_first (view, 1, 1);
	if (view->hexedit_mode) {
	    view->view_side = 1 - view->view_side;
	} else {
	    view->hexedit_mode = 1 - view->hexedit_mode;
	}
	view_labels (view);
	view->dirty++;
	view_update (view);
		return;
    }
    view->wrap_mode = 1 - view->wrap_mode;
    get_bottom_first (view, 1, 1);
    if (view->wrap_mode)
	view->start_col = 0;
    else {
	if (have_fast_cpu) {
	    if (view->bottom_first < view->start_display)
		view->search_start = view->start_display = view->bottom_first;
	    view->found_len = 0;
	}
    }
    view_labels (view);
    view->dirty++;
    view_update (view);
}

/* Both views */
void
toggle_hex_mode (WView *view)
{
    view->hex_mode = 1 - view->hex_mode;

    if (view->hex_mode) {
	/* Shift the line start to 0x____0 on entry, restore it for Ascii */
	view->start_save = view->start_display;
	view->start_display -= view->start_display % view->bytes_per_line;
	view->edit_cursor = view->start_display;
	view->widget.options |= W_WANT_CURSOR;
	view->widget.parent->raw = 1;
    } else {
	view->start_display = view->start_save;
	view->widget.parent->raw = 0;
	view->widget.options &= ~W_WANT_CURSOR;
    }
    altered_hex_mode = 1;
    get_bottom_first (view, 1, 1);
    view_labels (view);
    view->dirty++;
    view_update (view);
}

/* Both views */
static void
goto_line (WView *view)
{
    char *line, prompt [100];
    int i, oldline = 1;
    int saved_wrap_mode = view->wrap_mode;

    view->wrap_mode = 0;
    for (i = view->first; i < view->start_display; i++)
	if (get_byte (view, i) == '\n')
	    oldline ++;
    sprintf (prompt, " The current line number is %d.\n"
		       " Enter the new line number:", oldline);
    line = input_dialog (" Goto line ", prompt, "");
    if (line) {
	if (*line) {
	    move_to_top (view);
	    view_move_forward (view, atoi (line) - 1);
	}
	free (line);
    }
    view->dirty++;
    view->wrap_mode = saved_wrap_mode;
    view_update (view);
}

/* Both views */
static void
regexp_search (WView *view, int direction)
{
    char *regexp = "";
    /* This is really an F6 key handler */
    if (view->hex_mode) {
	/* Save it without a confirmation prompt */
	save_edit_changes(view);
	return;
    }

    regexp = regexp_search_old ? regexp_search_old : regexp;
    regexp = input_dialog (" Search ", " Enter regexp:", regexp);
    if ((!regexp) || (!*regexp)) {
	return;
    }
    if (regexp_search_old)
	free (regexp_search_old);
    regexp_search_old = regexp;
    view->direction = direction;
    do_regexp_search (view, regexp);

    view->last_search = do_regexp_search;
}

static void
regexp_search_cmd (WView *view)
{
    regexp_search (view, 1);
}

/* Both views */
static void
normal_search (WView *view, int direction)
{
    char *e = "";

    e = normal_search_old ? normal_search_old : e;
    e = input_dialog (" Search ", " Enter search string:", e);
    if ((!e) || (!*e)) {
	return;
    }
    if (normal_search_old)
	free (normal_search_old);
    normal_search_old = e;

    view->direction = direction;
    do_normal_search (view, e);
    view->last_search = do_normal_search;
}

static void
normal_search_cmd (WView *view)
{
    normal_search (view, 1);
}

static void
change_viewer (WView *view)
{
    char *s;
    char *t;

    if (*view->filename) {
	altered_magic_flag = 1;
	view->viewer_magic_flag = !view->viewer_magic_flag;
	s = strdup (view->filename);
	if (view->command)
	    t = strdup (view->command);
	else
	    t = 0;

	view_done (view);
	view_init (view, t, s, 0);
	free (s);
	if (t)
	    free (t);
	view_labels (view);
	view->dirty++;
	view_update (view);
    }
}

void
change_nroff (WView *view)
{
    view->viewer_nroff_flag = !view->viewer_nroff_flag;
    altered_nroff_flag = 1;
    view_labels (view);
    view->dirty++;
    view_update (view);
}

/* Real view only */
static void
view_quit_cmd (WView *view)
{
    if (view_ok_to_quit (view))
	dlg_stop (view->widget.parent);
}

/* Both views */
void
view_labels (WView *view)
{
    Dlg_head *h = view->widget.parent;

    define_label (h, (Widget *) view, 1, "Help", v_help_cmd);

    my_define (h, 10, "Quit", view_quit_cmd, view);
    my_define (h, 4, view->hex_mode ? "Ascii": "Hex", toggle_hex_mode, view);
    my_define (h, 5, "Line", goto_line, view);
    my_define (h, 6, view->hex_mode ? "Save" : "RxSrch", regexp_search_cmd, view);

    my_define (h, 2, view->hex_mode ? view->hexedit_mode ?
		     view->view_side == view_side_left ? "EdText" : "EdHex" :
		     view->growing_buffer ? "" : "Edit" :
		     view->wrap_mode ? "UnWrap" : "Wrap",
		     toggle_wrap_mode, view);

    my_define (h, 7, view->hex_mode ? "HxSrch" : "Search",
	       normal_search_cmd, view);

    my_define (h, 8, view->viewer_magic_flag ? "Raw" : "Parse",
	       change_viewer, view);

    if (!view->have_frame) {
	my_define (h, 9, view->viewer_nroff_flag ? "Unform" : "Format",
		   change_nroff, view);
	my_define (h, 3, "Quit", view_quit_cmd, view);
    }

    redraw_labels (h, (Widget *) view);
}

/* Both views */
static int
check_left_right_keys (WView *view, int c)
{
    if (c == KEY_LEFT)
	move_left (view);
    else if (c == KEY_RIGHT)
	move_right (view);
    else return 0;

    return 1;
}

void
set_monitor (WView *view, int set_on)
{
    int old = view->monitor;

    view->monitor = set_on;

    if (view->monitor) {
	move_to_bottom (view);
	view->bottom_first = -1;
	set_idle_proc (view->widget.parent, 1);
    } else {
	if (old)
	    set_idle_proc (view->widget.parent, 0);
    }
}

void
continue_search (WView *view)
{
    if (view->last_search) {
	(*view->last_search)(view, view->search_exp);
    } else {
	/* if not... then ask for an expression */
	normal_search (view, 1);
    }
}

/* Both views */
static int
view_handle_key (WView *view, int c)
{
    int prev_monitor = view->monitor;

    set_monitor (view, off);

    if (view->hex_mode) {
	switch (c) {
	case 0x09:              /* Tab key */
	    view->view_side = 1 - view->view_side;
	    view->dirty++;
	    return 1;

	case XCTRL('a'):        /* Beginning of line */
	    view->edit_cursor -= view->edit_cursor % view->bytes_per_line;
	    view->dirty++;
	    return 1;

	case XCTRL('b'):        /* Character back */
	    move_left(view);
	    return 1;

	case XCTRL('e'):        /* End of line */
	    view->edit_cursor -= view->edit_cursor % view->bytes_per_line;
	    view->edit_cursor += view->bytes_per_line - 1;
	    view->dirty++;
	    return 1;

	case XCTRL('f'):        /* Character forward */
	    move_right(view);
	    return 1;
	}

	/* Trap 0-9,A-F,a-f for left side data entry (hex editing) */
	if (view->view_side == view_side_left) {
	    if ((c >= '0' && c <= '9') ||
		(c >= 'A' && c <= 'F') ||
		(c >= 'a' && c <= 'f')) {

		put_editkey (view, c);
		return 1;
	    }
	}

	/* Trap all printable characters for right side data entry */
	/* Also enter the value of the Enter key */
	if (view->view_side == view_side_right) {
	    if (c < 256 && (is_printable (c) || (c == '\n'))) {
		put_editkey(view, c);
		return 1;
	    }
	}
    }

    if (check_left_right_keys (view, c))
	return 1;

    if (check_movement_keys (c, 1, vheight, view, (movefn) view_move_backward, (movefn) view_move_forward,
			     (movefn) move_to_top, (movefn) move_to_bottom)) {
	return 1;
    }
    switch (c) {

    case '?':
	regexp_search (view, -1);
	return 1;

    case '/':
	regexp_search (view, 1);
	return 1;

	/* Continue search */
    case XCTRL('s'):
    case 'n':
    case KEY_F(17):
	continue_search (view);
	return 1;

    case XCTRL('r'):
	if (view->last_search) {
	    (*view->last_search)(view, view->search_exp);
	} else {
	    normal_search (view, -1);
	}
	return 1;

	/* toggle ruler */
    case ALT('r'):
	switch (ruler) {
	case 0:
	    ruler = 1; break;
	case 1:
	    ruler = 2; break;
	default:
	    ruler = 0; break;
	}
	view->dirty++;
	return 1;

    case 'h':
	move_left (view);
	return 1;

    case 'j':
    case '\n':
    case 'e':
	view_move_forward (view, 1);
	return 1;

    case 'd':
	view_move_forward (view, vheight / 2);
	return 1;

    case 'u':
	view_move_backward (view, vheight / 2);
	return 1;

    case 'k':
    case 'y':
	view_move_backward (view, 1);
	return 1;

    case 'l':
	move_right (view);
	return 1;

    case ' ':
    case 'f':
	view_move_forward (view, vheight - 1);
	return 1;

    case '!':
	do_execute (MC_shell, 0, 0);
	return 1;

    case 'F':
	set_monitor (view, on);
	return 1;

    case 'b':
	view_move_backward (view, vheight - 1);
	return 1;

    case KEY_IC:
	view_move_backward (view, 2);
	return 1;

    case KEY_DC:
	view_move_forward (view, 2);
	return 1;

    case 'm':
	view->marks [view->marker] = view->start_display;
	return 1;

    case 'r':
	view->start_display = view->marks [view->marker];
	view->dirty++;
	return 1;

	/*  Use to indicate parent that we want to see the next/previous file */
	/* Only works on full screen mode */
    case XCTRL('f'):
    case XCTRL('b'):
	if (!view->have_frame)
	    view->move_dir = c == XCTRL('f') ? 1 : -1;
	/* fall */

    case 'q':
    case XCTRL('g'):
    case ESC_CHAR:
	if (view_ok_to_quit (view))
	    view->view_quit = 1;
	return 1;

    }
    if (c >= '0' && c <= '9')
	view->marker = c - '0';

    /* Restore the monitor status */
    set_monitor (view, prev_monitor);

    /* Key not used */
    return 0;
}

/* Both views */
int
view_event (WView *view, Gpm_Event *event, int *result)
{
    *result = MOU_NORMAL;
    if (event->type & (GPM_DOWN|GPM_DRAG)) {
	if (!view->wrap_mode) {
	    if (event->x < view->widget.cols / 4) {
		move_left (view);
		*result = MOU_REPEAT;
		return 1;
	    }
	    if (event->x > 3 * vwidth / 4) {
		move_right (view);
		*result = MOU_REPEAT;
		return 1;
	    }
	}
	if (event->y < view->widget.lines / 3) {
	    if (mouse_move_pages_viewer)
		view_move_backward (view, view->widget.lines / 2 - 1);
	    else
		view_move_backward (view, 1);
	    *result = MOU_REPEAT;
	    return 1;
	}
	else if (event->y > 2 * vheight /3) {
	    if (mouse_move_pages_viewer)
		view_move_forward (view, vheight / 2 - 1);
	    else
		view_move_forward (view, 1);
	    *result = MOU_REPEAT;
	    return 1;
	}
    }
    return 0;
}

/* Real view only */
static int
real_view_event (Gpm_Event *event, void *x)
{
    int result;

    if (view_event ((WView *) x, event, &result))
	view_update ((WView *) x);
    return result;
}

/* }}} */
/* {{{ Window creation, destruction and a driver stub for real view */

static int
view_mode_callback (struct Dlg_head *h, int id, int msg)
{
    return default_dlg_callback (h, id, msg);
}

void
view_adjust_size (Dlg_head *h)
{
    WView      *view;
    WButtonBar *bar;

    /* Look up the viewer and the buttonbar, we assume only two widgets here */
    view = (WView *) find_widget_type (h, (callback_fn) view_callback);
    bar  = (WButtonBar *) view->widget.parent->current->next->widget;
    widget_set_size (&view->widget, 0, 0, LINES-1, COLS);
    widget_set_size (&bar->widget, LINES-1, 0, 1, COLS);
}


/* Real view only */
int
View (char *_command, char *_file, int *move_dir_p, int start_line)
{
    int error;
    WView *wview;
    WButtonBar *bar;
    Dlg_head *our_dlg;

    /* Create dialog and widgets, put them on the dialog */
    our_dlg = create_dlg (0, 0, LINES, COLS, midnight_colors,
			   view_mode_callback, "[Internal File Viewer]", DLG_NONE);

    view_dlg = our_dlg;
    wview = view_new (0, 0, COLS, LINES-1, 0);

    bar  = buttonbar_new (1);

    add_widget (our_dlg, wview);
    add_widget (our_dlg, bar);

    error = view_init (wview, _command, _file, start_line);
    if (move_dir_p)
	*move_dir_p = 0;

    /* Please note that if you add another widget,
     * you have to modify view_adjust_size to
     * be aware of it
     */
    if (!error) {
	run_dlg (our_dlg);
	if (move_dir_p)
	    *move_dir_p = wview->move_dir;
    }
    destroy_dlg (our_dlg);

    return !error;
}

static void
view_hook (void *v)
{
    WView *view = (WView *) v;
    WPanel *panel;
    struct stat s;

    /* If the user is busy typing, wait until he finishes to update the
       screen */
    if (!is_idle ()) {
	if (!hook_present (idle_hook, view_hook))
	    add_hook (&idle_hook, view_hook, v);
	return;
    }

    delete_hook (&idle_hook, view_hook);

    if (get_current_type () == view_listing)
	panel = cpanel;
    else if (get_other_type () == view_listing)
	panel = other_panel;
    else
	return;

    if (S_ISLNK (panel->dir.list [panel->selected].buf.st_mode)) {
	if (mc_stat (panel->dir.list [panel->selected].fname, &s) != 0)
	    return;
	if (!S_ISREG (s.st_mode))
	    return;
    } else {
	if (!S_ISREG (panel->dir.list [panel->selected].buf.st_mode))
	    return;
    }

    view_init (view, 0, panel->dir.list [panel->selected].fname, 0);
    display (view);
    view_status (view);
}

static int
view_callback (Dlg_head *h, WView *v, int msg, int par)
{
    WView *view = (WView *) v;
    int i;

    switch (msg) {
    case WIDGET_INIT:
	if (view->have_frame)
	    add_hook (&select_file_hook, view_hook, view);
	else
	    view_labels (view);
	break;

    case WIDGET_DRAW:
	display (view);
	view_status (view);
	break;

    case WIDGET_CURSOR:
	if (view->hex_mode)
	    view_place_cursor (view);
	break;

    case WIDGET_KEY:
	i = view_handle_key ((WView *)view, par);
	if (view->view_quit)
	    dlg_stop (h);
	else {
	    view_update (view);
	}
	return i;

    case WIDGET_IDLE:
	/* This event is generated when the user is using the 'F' flag */
	view->bottom_first = -1;
	move_to_bottom (view);
	display (view);
	view_status (view);
	sleep (1);
	return 1;

    case WIDGET_FOCUS:
	view_labels (view);
	return 1;

    }
    return default_proc (h, msg, par);
}

WView *
view_new (int y, int x, int cols, int lines, int is_panel)
{
    WView *view = zmalloc (sizeof (WView));

    init_widget (&view->widget, y, x, lines, cols,
		 (callback_fn) view_callback,
		 (destroy_fn) view_destroy,
		 (mouse_h) real_view_event);

    view->hex_mode = default_hex_mode;
    view->hexedit_mode = 0;
    view->viewer_magic_flag = default_magic_flag;
    view->viewer_nroff_flag = default_nroff_flag;
    view->have_frame = is_panel;
    view->last_byte = -1;
    view->wrap_mode = global_wrap_mode;

    widget_want_cursor (view->widget, 0);

    return view;
}

/* }}} */
/* {{{ Emacs local variables */
/*
   Cause emacs to enter folding mode for this file:
   Local variables:
   end:
   */
/* }}} */
/* Widgets for the Midnight Commander */


#ifndef PORT_HAS_DESTROY_CMD
#   define x_destroy_cmd(w)
#endif

#ifndef PORT_HAS_RADIO_FOCUS_ITEM
#   define x_radio_focus_item(r)
#endif

#ifndef PORT_HAS_RADIO_TOGGLE
#   define x_radio_toggle(x)
#endif

static int button_event (Gpm_Event *event, WButton *b);

static int
button_callback (Dlg_head *h, WButton *b, int Msg, int Par)
{
    char *txt, buf[256];
    int stop = 0;
    int Off = 0;

    switch (Msg) {
    case WIDGET_INIT:
	return 1;
    case WIDGET_HOTKEY:
	if (b->hotkey == Par ||  toupper(b->hotkey) == Par) {
	    button_callback (h, b, WIDGET_KEY, ' '); /* to make action */
	    return 1;
	} else
	    return 0;

    case WIDGET_KEY:
	if (Par != ' ' && Par != '\n')
	    break;

	if (b->callback)
	    stop = (*b->callback)(b->action, b->callback_data);
	if (!b->callback || stop) {
	    h->ret_value = b->action;
	    dlg_stop (h);
	}
	return 1;

    case WIDGET_CURSOR:
	switch (b->flags) {
	    case DEFPUSH_BUTTON:
		Off = 3;
		break;
	    case NORMAL_BUTTON:
		Off = 2;
		break;
	    case NARROW_BUTTON:
		Off = 1;
		break;
	    case HIDDEN_BUTTON:
	    default:
		Off = 0;
		break;
	}
	widget_move (&b->widget, 0, b->hotpos + Off);
	return 1;

    case WIDGET_UNFOCUS:
    case WIDGET_FOCUS:
    case WIDGET_DRAW:
	if (Msg==WIDGET_UNFOCUS)
	    b->selected = 0;
	else if (Msg==WIDGET_FOCUS)
	    b->selected = 1;

	switch (b->flags) {
	    case DEFPUSH_BUTTON:
		sprintf (buf, "[< %s >]", b->text);
		Off = 3;
		break;
	    case NORMAL_BUTTON:
		sprintf (buf, "[ %s ]", b->text);
		Off = 2;
		break;
	    case NARROW_BUTTON:
		sprintf (buf, "[%s]", b->text);
		Off = 1;
		break;
	    case HIDDEN_BUTTON:
	    default:
		buf[0] = '\0';
		Off = 0;
		break;
	}
	txt = buf;

	attrset ((b->selected) ? FOCUSC : NORMALC);
	widget_move (&b->widget, 0, 0);

	addstr (txt);

	if (b->hotpos >= 0) {
	    attrset ((b->selected) ? HOT_FOCUSC : HOT_NORMALC);
	    widget_move (&b->widget, 0, b->hotpos+Off);
	    addch ((unsigned char)b->text [b->hotpos]);
	}
	if (Msg == WIDGET_FOCUS)
	    break;
	else
	    return 1;
	break;
    }
    return default_proc (h, Msg, Par);
}

static int
button_event (Gpm_Event *event, WButton *b)
{
    if (event->type & (GPM_DOWN|GPM_UP)) {
	Dlg_head *h=b->widget.parent;
	dlg_select_widget (h, b);
	if (event->type & GPM_UP) {
	    button_callback (h, b, WIDGET_KEY, ' ');
	    (*h->callback) (h, ' ', DLG_POST_KEY);
	    return MOU_NORMAL;
	}
    }
    return MOU_NORMAL;
}

static void
button_destroy (WButton *b)
{
    x_destroy_cmd (b);
    free (b->text);
}

static int
button_len (const char *text, unsigned int flags)
{
    int ret = strlen (text);
    switch (flags) {
	case DEFPUSH_BUTTON:
	    ret += 6;
	    break;
	case NORMAL_BUTTON:
	    ret += 4;
	    break;
	case NARROW_BUTTON:
	    ret += 2;
	    break;
	case HIDDEN_BUTTON:
	default:
	    return 0;
    }
    return ret;
}

/*
 * Assuming that button text is malloc'ed, we may safely change it
 * (as opposed to statically allocated); from other hand, excluding &
 * and shifting data past it to the left results to one unused byte.
 * This does not harm though :)
 */
static void
button_scan_hotkey(WButton* b)
{
    char* cp = strchr (b->text, '&');

    if (cp != NULL && cp[1] != '\0') {
	memmove (cp, cp+1, strlen(cp));
	b->hotkey = tolower (*cp);
	b->hotpos = cp - b->text;
    }
}

static WButton *
button_new (int y, int x, int action, int flags, const char *text,
	    int (*callback)(int, void *), void *callback_data)
{
    WButton *b = x_malloc (sizeof (WButton), "new_button");

    init_widget (&b->widget, y, x, 1, button_len (text, flags),
		 (callback_fn) button_callback,
		 (destroy_fn) button_destroy, (mouse_h)button_event);

    b->action = action;
    b->flags  = flags;
    b->selected = 0;
    b->text   = strdup (text);
    b->callback = callback;
    b->callback_data = callback_data;
    widget_want_hotkey (b->widget, 1);
    b->hotkey = 0;
    b->hotpos = -1;

    button_scan_hotkey(b);
    return b;
}

static void
button_set_text (WButton *b, const char *text)
{
    free (b->text);
    b->text = strdup (text);
    b->widget.cols = button_len (text, b->flags);
    button_scan_hotkey(b);
    dlg_redraw (b->widget.parent);
}

/* Radio button widget */
static int radio_event (Gpm_Event *event, WRadio *r);

static int
radio_callback (Dlg_head *h, WRadio *r, int Msg, int Par)
{
    int i;

    switch (Msg) {
    case WIDGET_INIT:
	return 1;

    case WIDGET_HOTKEY:
	{
	    int lp = tolower(Par);
	    char *cp;

	    for (i = 0; i < r->count; i++) {
		cp = strchr (r->texts [i], '&');
		if (cp != NULL && cp[1] != '\0') {
		    int c = tolower (cp [1]);

		    if (c != lp)
			continue;
		    r->pos = i;
		    radio_callback (h, r, WIDGET_KEY, ' '); /* Take action */
		    return 1;
		}
	    }
	}
	return 0;

    case WIDGET_KEY:
	switch (Par) {
	case ' ':
	    r->sel = r->pos;
	    (*h->callback) (h, h->current->dlg_id, DLG_ACTION);
	    radio_callback (h, r, WIDGET_FOCUS, ' ');
	    x_radio_toggle(r);
	    return 1;

	case KEY_UP:
	case KEY_LEFT:
	    if (r->pos > 0) {
		r->pos--;
		x_radio_focus_item (r);
		return 1;
	    }
	    return 0;

	case KEY_DOWN:
	case KEY_RIGHT:
	    if (r->count - 1 > r->pos) {
		r->pos++;
		x_radio_focus_item (r);
		return 1;
	    }
	}
	return 0;

    case WIDGET_CURSOR:
	(*h->callback) (h, h->current->dlg_id, DLG_ACTION);
	radio_callback (h, r, WIDGET_FOCUS, ' ');
	widget_move (&r->widget, r->pos, 1);
	break;

    case WIDGET_UNFOCUS:
    case WIDGET_FOCUS:
    case WIDGET_DRAW:
	for (i = 0; i < r->count; i++) {
		const unsigned char* cp;
	    attrset ((i==r->pos && Msg==WIDGET_FOCUS) ? FOCUSC :NORMALC);
	    widget_move (&r->widget, i, 0);

		printw("(%c) ", (r->sel == i) ? '*' : ' ');
		for (cp = (unsigned char*)r->texts[i]; *cp; cp++)
		{
			if (*cp == '&')
			{
				attrset ((i==r->pos && Msg==WIDGET_FOCUS)
					? HOT_FOCUSC : HOT_NORMALC);
				addch(*++cp);
				attrset ((i==r->pos && Msg==WIDGET_FOCUS) ? FOCUSC : NORMALC);
			}
			else
				addch(*cp);
		}
	}
	return 1;
	break;
    }
    return default_proc (h, Msg, Par);
}

static int
radio_event (Gpm_Event *event, WRadio *r)
{
    if (event->type & (GPM_DOWN|GPM_UP)) {
	Dlg_head *h = r->widget.parent;

	r->pos = event->y - 1;
	dlg_select_widget (h, r);
	if (event->type & GPM_UP) {
	    radio_callback (h, r, WIDGET_KEY, ' ');
	    radio_callback (h, r, WIDGET_FOCUS, 0);
	    (*h->callback) (h, ' ', DLG_POST_KEY);
	    return MOU_NORMAL;
	}
    }
    return MOU_NORMAL;
}

WRadio *
radio_new (int y, int x, int count_r, const char * const *texts, int use_hotkey)
{
    WRadio *r = zmalloc (sizeof (WRadio));
    int i, max, m;

    /* Compute the longest string */
    max = 0;
    for (i = 0; i < count_r; i++) {
	m = strlen (texts [i]);
	if (m > max)
	    max = m;
    }

    init_widget (&r->widget, y, x, count_r, max, (callback_fn) radio_callback,
		 0, (mouse_h) radio_event);
    r->state = 1;
    r->count = count_r;
    r->texts = texts;
    r->upper_letter_is_hotkey = use_hotkey;
    widget_want_hotkey (r->widget, 1);

    return r;
}

/* Checkbutton widget */

static int check_event (Gpm_Event *event, WCheck *b);

static int
check_callback (Dlg_head *h, WCheck *c, int Msg, int Par)
{
    switch (Msg) {
    case WIDGET_INIT:
	return 1;

    case WIDGET_HOTKEY:
	if (c->hotkey==Par ||
	    (c->hotkey>='a' && c->hotkey<='z' && c->hotkey-32==Par)) {
	    check_callback (h, c, WIDGET_KEY, ' ');        /* make action */
	    return 1;
	} else
	    return 0;

    case WIDGET_KEY:
	if (Par != ' ')
	    break;
	c->state ^= C_BOOL;
	c->state ^= C_CHANGE;
	(*h->callback) (h, h->current->dlg_id, DLG_ACTION);
	check_callback (h, c, WIDGET_FOCUS, ' ');
	return 1;

    case WIDGET_CURSOR:
	widget_move (&c->widget, 0, 1);
	break;

    case WIDGET_FOCUS:
    case WIDGET_UNFOCUS:
    case WIDGET_DRAW:
	attrset ((Msg == WIDGET_FOCUS) ? FOCUSC : NORMALC);
	widget_move (&c->widget, 0, 0);
	printw ("[%c] %s", (c->state & C_BOOL) ? 'x' : ' ', c->text);

	if (c->hotpos >= 0) {
	    attrset ((Msg == WIDGET_FOCUS) ? HOT_FOCUSC : HOT_NORMALC);
	    widget_move (&c->widget, 0, + c->hotpos+4);
	    addch ((unsigned char)c->text [c->hotpos]);
	}
	return 1;
    }
    return default_proc (h, Msg, Par);
}

static int
check_event (Gpm_Event *event, WCheck *c)
{
    if (event->type & (GPM_DOWN|GPM_UP)) {
	Dlg_head *h = c->widget.parent;

	dlg_select_widget (h, c);
	if (event->type & GPM_UP) {
	    check_callback (h, c, WIDGET_KEY, ' ');
	    check_callback (h, c, WIDGET_FOCUS, 0);
	    (*h->callback) (h, ' ', DLG_POST_KEY);
	    return MOU_NORMAL;
	}
    }
    return MOU_NORMAL;
}

static void
check_destroy (WCheck *c)
{
	x_destroy_cmd (c);
	free (c->text);
}

WCheck *
check_new (int y, int x, int state, const char *text)
{
    WCheck *c =  x_malloc (sizeof (WCheck), "check_new");
    char *t;
    const char *s;

    init_widget (&c->widget, y, x, 1, strlen (text),
		 (callback_fn)check_callback,
		 (destroy_fn)check_destroy, (mouse_h) check_event);
    c->state = state ? C_BOOL : 0;
    c->text = strdup (text);
    c->hotkey = 0;
    c->hotpos = -1;
    widget_want_hotkey (c->widget, 1);

    /* Scan for the hotkey */
    for (s = text, t = c->text; *s; s++, t++) {
	if (*s != '&') {
	    *t = *s;
	    continue;
	}
	s++;
	if (*s) {
	    c->hotkey = tolower (*s);
	    c->hotpos = t - c->text;
	}
	*t = *s;
    }
    *t = 0;
    return c;
}

/* Label widget */

static int
label_callback (Dlg_head *h, WLabel *l, int Msg, int Par)
{
    if (Msg == WIDGET_INIT)
	return 1;

    /* We don't want to get the focus */
    if (Msg == WIDGET_FOCUS)
	return 0;
    if (Msg == WIDGET_DRAW && l->text) {
	char *p = l->text, *q, c = 0;
	int y = 0;
	if (l->transparent)
	    attrset (DEFAULT_COLOR);
	else
	    attrset (NORMALC);
	for (;;) {
	    int xlen;

	    q = strchr (p, '\n');
	    if (q) {
		c = *q;
		*q = 0;
	    }
	    widget_move (&l->widget, y, 0);
	    printw ("%s", p);
	    xlen = l->widget.cols - strlen (p);
	    if (xlen > 0)
		printw ("%*s", xlen, " ");
	    if (!q)
		break;
	    *q = c;
	    p = q + 1;
	    y++;
	}
	return 1;
    }
    return default_proc (h, Msg, Par);
}

void
label_set_text (WLabel *label, const char *text)
{
    int newcols = label->widget.cols;

    if (label->text && text && !strcmp (label->text, text))
	return; /* Flickering is not nice */

    if (label->text) {
	free (label->text);
    }
    if (text) {
	label->text = strdup (text);
	if (label->auto_adjust_cols) {
	    newcols = strlen (text);
	    if (newcols > label->widget.cols)
	    label->widget.cols = newcols;
	}
    } else
	label->text = 0;

    if (label->widget.parent)
	label_callback (label->widget.parent, label, WIDGET_DRAW, 0);
    if (newcols < label->widget.cols)
	label->widget.cols = newcols;
}

static void
label_destroy (WLabel *l)
{
    x_destroy_cmd (l);
    if (l->text)
	free (l->text);
}

WLabel *
label_new (int y, int x, const char *text)
{
    WLabel *l = x_malloc (sizeof (WLabel), "label_new");

    init_widget (&l->widget, y, x, 1, 1,
		 (callback_fn) label_callback,
		 (destroy_fn) label_destroy, NULL);
    l->text = text ? strdup (text) : 0;
    l->auto_adjust_cols = 1;
    l->transparent = 0;
    widget_want_cursor (l->widget, 0);
    return l;
}

/* Gauge widget (progress indicator) */
/* Currently width is hardcoded here for text mode */
#define gauge_len 47

static int
gauge_callback (Dlg_head *h, WGauge *g, int Msg, int Par)
{

    if (Msg == WIDGET_INIT)
	return 1;

    /* We don't want to get the focus */
    if (Msg == WIDGET_FOCUS)
	return 0;

    if (Msg == WIDGET_DRAW) {
	widget_move (&g->widget, 0, 0);
	attrset (NORMALC);
	if (!g->shown)
	    printw ("%*s", gauge_len, "");
	else {
	    int percentage, columns;
	    off_t total = g->max, done = g->current;

	    if (total <= 0 || done < 0) {
		done = 0;
		total = 100;
	    }
	    if (done > total)
		done = total;
	    while (total > 65535) {
		total /= 256;
		done /= 256;
	    }
	    percentage = (done * 200 / total + 1) / 2;
	    columns = (2 * (gauge_len - 7) * (done / total) + 1) / 2;
	    addch ('[');
	    attrset (GAUGE_COLOR);
	    printw ("%*s", columns, "");
	    attrset (NORMALC);
	    printw ("%*s] %3d%%", gauge_len - 7 - columns, "", percentage);
	}
	return 1;
    }
    return default_proc (h, Msg, Par);
}

static void
gauge_set_value (WGauge *g, off_t max, off_t current)
{
    if (g->current == current && g->max == max)
	return; /* Do not flicker */
    if (max == 0)
	max = 1; /* I do not like division by zero :) */
    g->current = current;
    g->max = max;
    gauge_callback (g->widget.parent, g, WIDGET_DRAW, 0);
}

static void
gauge_show (WGauge *g, int shown)
{
    if (g->shown == shown)
	return;
    g->shown = shown;
    gauge_callback (g->widget.parent, g, WIDGET_DRAW, 0);
}

static void
gauge_destroy (WGauge *g)
{
    /* nothing */
}

WGauge *
gauge_new (int y, int x, int shown, int max, int current)
{
    WGauge *g = x_malloc (sizeof (WGauge), "gauge_new");

    init_widget (&g->widget, y, x, 1, gauge_len,
		 (callback_fn) gauge_callback,
		 (destroy_fn) gauge_destroy, NULL);
    g->shown = shown;
    if (max == 0)
	max = 1; /* I do not like division by zero :) */
    g->max = max;
    g->current = current;
    widget_want_cursor (g->widget, 0);
    return g;
}

/* Input widget */

/* {{{ history button */

#define LARGE_HISTORY_BUTTON 1

#ifdef LARGE_HISTORY_BUTTON
#  define HISTORY_BUTTON_WIDTH 3
#else
#  define HISTORY_BUTTON_WIDTH 1
#endif

#define should_show_history_button(in) \
	    (in->history && in->field_len > HISTORY_BUTTON_WIDTH * 2 + 1 && in->widget.parent)

static void draw_history_button (WInput * in)
{
    char c;
    c = in->history->next ? (in->history->prev ? '|' : 'v') : '^';
    widget_move (&in->widget, 0, in->field_len - HISTORY_BUTTON_WIDTH);
#ifdef LARGE_HISTORY_BUTTON
    {
	attrset (NORMAL_COLOR);
	addstr ("[ ]");
	/* Too distracting: attrset (MARKED_COLOR); */
	widget_move (&in->widget, 0, in->field_len - HISTORY_BUTTON_WIDTH + 1);
	addch (c);
    }
#else
    attrset (MARKED_COLOR);
    addch (c);
#endif
}

/* }}} history button */


/* Input widgets now have a global kill ring */

static void
update_input (WInput *in, int clear_first)
{
    int has_history = 0;
    int    i, j;
    char   c;
    int    buf_len = strlen (in->buffer);

    if (should_show_history_button (in))
	has_history = HISTORY_BUTTON_WIDTH;

    if (in->disable_update)
	return;

    /* Make the point visible */
    if ((in->point < in->first_shown) ||
	(in->point >= in->first_shown+in->field_len - has_history)) {
	in->first_shown = in->point - (in->field_len / 3);
	if (in->first_shown < 0)
	    in->first_shown = 0;
    }

    /* Adjust the mark */
    if (in->mark > buf_len)
	in->mark = buf_len;

    if (has_history)
	draw_history_button (in);

    attrset (in->color);

    widget_move (&in->widget, 0, 0);
    for (i = 0; i < in->field_len - has_history; i++)
	addch (' ');
    widget_move (&in->widget, 0, 0);

    for (i = 0, j = in->first_shown; i < in->field_len - has_history && in->buffer [j]; i++) {
	c = in->buffer [j++];
	c = is_printable (c) ? c : '.';
	if (in->is_password)
	    c = '*';
	addch (c);
    }
    widget_move (&in->widget, 0, in->point - in->first_shown);

    if (clear_first)
	    in->first = 0;
}

static void
winput_set_origin (WInput *in, int x, int field_len)
{
    in->widget.x    = x;
    in->field_len = in->widget.cols = field_len;
    update_input (in, 0);
}

/* {{{ history saving and loading */

/*
   This loads and saves the history of an input line to and from the
   widget. It is called with the widgets tk name on creation of the
   widget, and returns the Hist list. It stores histories in the file
   ~/.mc/history in using the profile code.

   If def_text is passed as INPUT_LAST_TEXT (to the input_new()
   function) then input_new assigns the default text to be the last text
   entered, or "" if not found.
 */

static Hist *history_get (const char *input_name)
{
    int i;
    Hist *old = 0, *new = 0;
    char *profile;

    if (!num_history_items_recorded)    /* this is how to disable */
	return 0;
    if (!input_name)
	return 0;
    if (!*input_name)
	return 0;
    profile = concat_path_file (home_dir, HISTORY_FILE_NAME);
    for (i = 0;; i++) {
	char key_name[32];
	char this_entry[1024];

	GetPrivateProfileString (input_name, sprintf_pd (key_name, i), "", this_entry, sizeof (this_entry), profile);
	if (!*this_entry)
	    break;
	new = zmalloc (sizeof (Hist));
	new->text = strdup (this_entry);
	new->prev = old;        /* set up list pointers */
	if (old)
	    old->next = new;
	old = new;
    }
    free (profile);
    return new;                 /* return pointer to last entry in list */
}

/* {{{ history display */

static const char history_title[] = " History ";

static int history_callback (Dlg_head * h, int Par, int Msg)
{
    switch (Msg) {
    case DLG_DRAW:
	attrset (COLOR_NORMAL);
	dlg_erase (h);
	draw_box (h, 0, 0, h->lines, h->cols);
	attrset (COLOR_HOT_NORMAL);
	dlg_move (h, 0, (h->cols - strlen (history_title)) / 2);
	printw (history_title);
	break;
    }
    return 0;
}

static int listbox_fwd (WListbox *l);

char *show_hist (Hist *history_s, int widget_x, int widget_y)
{
    Hist *hi, *z;
    int maxlen = strlen(history_title), i, count_s = 0;
    int x, y, w, h;
    char *q, *r = 0;
    Dlg_head *query_dlg;
    WListbox *query_list;

    z = history_s;
    if (!z)
		return 0;

    while (z->prev)             /* goto first */
		z = z->prev;
    hi = z;
    while (hi) {
		if ((i = strlen (hi->text)) > maxlen)
		    maxlen = i;
		count_s++;
		hi = hi->next;
    }

    y = widget_y;
    h = count_s + 2;
    if (h <= y || y > LINES - 6) {
	h = min(h, y - 1);
	y -= h;
    } else {
	y++;
	h = min(h, LINES - y);
    }

    x = widget_x - 2;
    if ((w = maxlen + 4) + x > COLS) {
	w = min(w,COLS);
	x = COLS - w;
    }

    query_dlg = create_dlg (y, x, h, w, dialog_colors, history_callback,
			    "[History-query]", DLG_NONE);
    query_list = listbox_new (1, 1, w - 2, h - 2, listbox_finish, 0);
    add_widget (query_dlg, query_list);
    hi = z;
    if (y < widget_y) {
		while (hi) {            /* traverse */
		    listbox_add_item (query_list, 0, 0, hi->text, NULL);
		hi = hi->next;
		}
		while (listbox_fwd (query_list));
    } else {
		while (hi->next)
		    hi = hi->next;
		while (hi) {            /* traverse backwards */
		    listbox_add_item (query_list, 0, 0, hi->text, NULL);
		hi = hi->prev;
		}
    }
    run_dlg (query_dlg);
    q = NULL;
    if (query_dlg->ret_value != B_CANCEL) {
		listbox_get_current (query_list, &q, NULL);
		if (q)
		    r = strdup (q);
    }
    destroy_dlg (query_dlg);
    return r;
}

static void do_show_hist (WInput * in)
{
    char *r;
    r = show_hist (in->history, in->widget.x, in->widget.y);
    if (r) {
	assign_text (in, r);
	free (r);
    }
}

/* }}} history display */

static void
input_destroy (WInput *in)
{
    if (!in) {
	fprintf (stderr, "Internal error: null Input *\n");
	exit (1);
    }

    new_input (in);
    if (in->history) {
	Hist *current, *old;

	current = in->history;
	while (current->next)
	    current = current->next;
	while (current) {
	    old = current;
	    current = current->prev;
	    free (old->text);
	    free (old);
	}
    }
    x_destroy_cmd (in);
    free (in->buffer);
    free_completions (in);
    if (in->history_name)
	free (in->history_name);
}

void
input_enable_update (WInput *in)
{
    in->disable_update--;
    update_input (in, 0);
}

int
push_history (WInput *in, char *text)
{
    Hist *new;
    char *p;

    for (p = text; *p == ' ' || *p == '\t'; p++);
    if (!*p)
	return 0;
    if (in->history) {
	while (in->history->next)
	    in->history = in->history->next;
	if (!strcmp (in->history->text, text))
	    return 1;
	new = x_malloc (sizeof (Hist), "push_history");
	in->history->next = new;
    } else
	new = x_malloc (sizeof (Hist), "push_history");
    in->need_push = 0;
    new->next = 0;
    new->prev = in->history;
    new->text = strdup (text);
    in->history = new;
    return 2;
}

/* Cleans the input line and adds the current text to the history */
void
new_input (WInput *in)
{
    if (in->buffer)
	push_history (in, in->buffer);
    in->need_push = 1;
    in->buffer [0] = 0;
    in->point = 0;
    in->mark = 0;
    free_completions (in);
    update_input (in, 0);
}

static int
insert_char (WInput *in, int c_code)
{
    int i;

    if (c_code == -1)
	return 0;

    in->need_push = 1;
    if (strlen (in->buffer)+1 == in->current_max_len) {
	/* Expand the buffer */
	char *narea = x_malloc(in->current_max_len + in->field_len, "string expansion");
	if (narea) {
	    char *p = in->buffer;

	    strcpy (narea, in->buffer);
	    in->buffer = narea;
	    in->current_max_len += in->field_len;
	    free (p);
	}
    }
    if (strlen (in->buffer)+1 < in->current_max_len) {
	int l = strlen (&in->buffer [in->point]);
	for (i = l+1; i > 0; i--)
	    in->buffer [in->point+i] = in->buffer [in->point+i-1];
	in->buffer [in->point] = c_code;
	in->point++;
    }
    return 1;
}

static void
beginning_of_line (WInput *in)
{
    in->point = 0;
}

static void
end_of_line (WInput *in)
{
    in->point = strlen (in->buffer);
}

static void
backward_char (WInput *in)
{
    if (in->point)
	in->point--;
}

static void
forward_char (WInput *in)
{
    if (in->buffer [in->point])
	in->point++;
}

static void
forward_word (WInput *in)
{
    char *p = in->buffer+in->point;

    while ((*p && isspace (*p)) || ispunct (*p))
	p++;
    while (*p && isalnum (*p))
	p++;
    in->point = p - in->buffer;
}

static void
backward_word (WInput *in)
{
    char *p = in->buffer+in->point;

    while (p-1 > in->buffer-1 && (isspace (*(p-1)) || ispunct (*(p-1))))
	p--;
    while (p-1 > in->buffer-1 && isalnum (*(p-1)))
	p--;
    in->point = p - in->buffer;
}

#ifdef __linux__
static void
key_left (WInput *in)
{
    if (ctrl_pressed ())
	backward_word (in);
    else
	backward_char (in);
}

static void
key_right (WInput *in)
{
    if (ctrl_pressed ())
	forward_word (in);
    else
	forward_char (in);
}
#else
#define key_left  backward_char
#define key_right forward_char
#endif

static void
backward_delete (WInput *in)
{
    int i;

    if (!in->point)
	return;
    for (i = in->point; in->buffer [i-1]; i++)
	in->buffer [i-1] = in->buffer [i];
    in->need_push = 1;
    in->point--;
}

static void
delete_char (WInput *in)
{
    int i;

    for (i = in->point; in->buffer [i]; i++)
	in->buffer [i] = in->buffer [i+1];
    in->need_push = 1;
}

static void
copy_region (WInput *in, int x_first, int x_last)
{
    int first = min (x_first, x_last);
    int last  = max (x_first, x_last);

    if (last == first)
	return;

    free (kill_buffer);
    kill_buffer = xstrndup (in->buffer+first, last-first);
}

static void
delete_region (WInput *in, int x_first, int x_last)
{
   int first = min (x_first, x_last);
   int last  = max (x_first, x_last);

   in->point = first;
   in->mark  = first;
   memmove (&in->buffer [first], &in->buffer [last], strlen(&in->buffer [last]) + 1);
   in->need_push = 1;
}

static void
kill_word (WInput *in)
{
    int old_point = in->point;
    int new_point;

    forward_word (in);
    new_point = in->point;
    in->point = old_point;

    copy_region (in, old_point, new_point);
    delete_region (in, old_point, new_point);
    in->need_push = 1;
}

static void
back_kill_word (WInput *in)
{
    int old_point = in->point;
    int new_point;

    backward_word (in);
    new_point = in->point;
    in->point = old_point;

    copy_region (in, old_point, new_point);
    delete_region (in, old_point, new_point);
    in->need_push = 1;
}

static void
set_mark (WInput *in)
{
    in->mark = in->point;
}

static void
kill_save (WInput *in)
{
    copy_region (in, in->mark, in->point);
}

static void
kill_region (WInput *in)
{
    kill_save (in);
    delete_region (in, in->point, in->mark);
}

static void
yank (WInput *in)
{
    char *p;

    if (!kill_buffer)
	return;
    for (p = kill_buffer; *p; p++)
	insert_char (in, *p);
}

static void
kill_line (WInput *in)
{
    free (kill_buffer);
    kill_buffer = strdup (&in->buffer [in->point]);
    in->buffer [in->point] = 0;
}

void
assign_text (WInput *in, char *text)
{
    free_completions (in);
    free (in->buffer);
    in->buffer = strdup (text); /* was in->buffer->text */
    in->current_max_len = strlen (in->buffer) + 1;
    in->point = strlen (in->buffer);
    in->mark = 0;
    in->need_push = 1;
}

static void
hist_prev (WInput *in)
{
    if (!in->history)
	return;

    if (in->need_push) {
	switch (push_history (in, in->buffer)) {
	 case 2: in->history = in->history->prev; break;
	 case 1: if (in->history->prev) in->history = in->history->prev; break;
	 case 0: break;
	}
    } else if (in->history->prev)
	in->history = in->history->prev;
    else
	return;
    assign_text (in, in->history->text);
    in->need_push = 0;
}

static void
hist_next (WInput *in)
{
    if (in->need_push) {
	switch (push_history (in, in->buffer)) {
	 case 2:
	    assign_text (in, "");
	    return;
	 case 0:
	    return;
	}
    }

    if (!in->history)
	return;

    if (!in->history->next) {
	assign_text (in, "");
	return;
    }

    in->history = in->history->next;
    assign_text (in, in->history->text);
    in->need_push = 0;
}

static const struct {
    int key_code;
    void (*fn)(WInput *in);
} input_map [] = {
    /* Motion */
    { XCTRL('a'),         beginning_of_line },
    { KEY_HOME,           beginning_of_line },
    { KEY_A1,             beginning_of_line },
    { XCTRL('e'),         end_of_line },
    { KEY_END,            end_of_line },
    { KEY_C1,             end_of_line },
    { KEY_LEFT,           key_left },
    { XCTRL('b'),         backward_char },
    { ALT('b'),           backward_word },
    { KEY_RIGHT,          key_right },
    { XCTRL('f'),         forward_char },
    { ALT('f'),           forward_word },

    /* Editing */
    { 0177,               backward_delete },
    { KEY_BACKSPACE,      backward_delete },
    { XCTRL('h'),         backward_delete },
    { KEY_DC,             delete_char },
    { XCTRL('d'),         delete_char },
    { ALT('d'),           kill_word },
    { ALT(KEY_BACKSPACE), back_kill_word },
    { ALT(XCTRL('h')),    back_kill_word },
    { ALT(127),           back_kill_word },

    /* Region manipulation */
    { 0,                  set_mark },
    { XCTRL('w'),         kill_region },
    { ALT('w'),           kill_save },
    { XCTRL('y'),         yank },
    { XCTRL('k'),         kill_line },

    /* History */
    { ALT('p'),           hist_prev },
    { ALT('n'),           hist_next },
    { ALT('h'),           do_show_hist },

    /* Completion */
    { ALT('\t'),          complete },

    { 0,            0 }
};

/* This function is a test for a special input key used in complete.c */
/* Returns 0 if it is not a special key, 1 if it is a non-complete key
   and 2 if it is a complete key */
int
is_in_input_map (WInput *in, int c_code)
{
    int i;

    for (i = 0; input_map [i].fn; i++)
	if (c_code == input_map [i].key_code) {
	    if (input_map [i].fn == complete)
		return 2;
	    else
		return 1;
	}
    return 0;
}

#ifdef PORT_WINPUT_DELETES_MARKED
static void
port_region_marked_for_delete (WInput *in)
{
	kill_region (in);
}
#else
static void
port_region_marked_for_delete (WInput *in)
{
    *in->buffer = 0;
    in->point = 0;
    in->first = 0;
}
#endif

int
handle_char (WInput *in, int c_code)
{
    int    i;
    int    v;

    v = 0;

    if (quote) {
	free_completions (in);
	v = insert_char (in, c_code);
	update_input (in, 1);
	quote = 0;
	return v;
    }

    for (i = 0; input_map [i].fn; i++) {
	if (c_code == input_map [i].key_code) {
	    if (input_map [i].fn != complete)
		free_completions (in);
	    (*input_map [i].fn)(in);
	    v = 1;
	    break;
	}
    }
    if (!input_map [i].fn) {
	if (c_code > 255 || !is_printable (c_code))
	    return 0;
	if (in->first) {
	    port_region_marked_for_delete (in);
	}
	free_completions (in);
	v = insert_char (in, c_code);
	in->inserted_one = c_code;
    }
    update_input (in, 1);
    return v;
}

/* Inserts text in input line */
void
stuff (WInput *in, char *text, int insert_extra_space)
{
    input_disable_update (in);
    while (*text)
	handle_char (in, *text++);
    if (insert_extra_space)
	handle_char (in, ' ');
    input_enable_update (in);
    update_input (in, 1);
}

void
input_set_point (WInput *in, int pos)
{
    if (pos > in->current_max_len)
	pos = in->current_max_len;
    if (pos != in->point)
	free_completions (in);
    in->point = pos;
    update_input (in, 1);
}

static int
input_callback (Dlg_head *h, WInput *in, int Msg, int Par)
{
    switch (Msg) {
    case WIDGET_INIT:
	return 1;

    case WIDGET_KEY:
	if (Par == XCTRL('q')) {
	    int v;

	    quote = 1;
	    v = handle_char (in, mi_getch ());
	    quote = 0;
	    return v;
	}
	if (Par == KEY_UP || Par == KEY_DOWN ||
	    Par == ESC_CHAR || Par == KEY_F(10) ||
	    Par == XCTRL('g'))
	    return 0;           /* We don't handle up/down */

	if (Par == '\n') {
	    dlg_one_down (h);
	    return 1;
	}
	return handle_char (in, Par);

    case WIDGET_FOCUS:
    case WIDGET_UNFOCUS:
    case WIDGET_DRAW:
	update_input (in, 0);
	break;
    case WIDGET_CURSOR:
	widget_move (&in->widget, 0, in->point - in->first_shown);
	return 1;

    }
    return default_proc (h, Msg, Par);
}

/* Not declared static, since we check against this value in dlg.c */
/* FIXME: Declare static again and provide an identification mechanism */
static int
input_event (Gpm_Event *event, WInput *in)
{
    if (event->type & (GPM_DOWN|GPM_DRAG)) {
	dlg_select_widget (in->widget.parent, in);

	if (event->x >= in->field_len - HISTORY_BUTTON_WIDTH + 1 && should_show_history_button (in)) {
	    do_show_hist (in);
	    update_input (in, 1);
	} else {
	    in->point = strlen (in->buffer);
	    if (event->x - in->first_shown - 1 < in->point)
		in->point = event->x - in->first_shown - 1;
	    if (in->point < 0)
		in->point = 0;

	    update_input (in, 1);
	}
    }
    return MOU_NORMAL;
}

static WInput *
input_new (int y, int x, int color, int len, const char *def_text)
{
    WInput *in = x_malloc (sizeof (WInput), "input_new");
    int initial_buffer_len;

    init_widget (&in->widget, y, x, 1, len,
		 (callback_fn) input_callback,
	      (destroy_fn) input_destroy, (mouse_h) input_event);

    /* history setup */
    in->history = NULL;
    in->history_name = 0;
    if (def_text == INPUT_LAST_TEXT) {
	def_text = "";
	if (in->history)
	    if (in->history->text)
		def_text = in->history->text;
    }
    initial_buffer_len = 1 + max (len, strlen (def_text));
    in->widget.options |= W_IS_INPUT;
    in->completions = NULL;
    in->completion_flags =
	INPUT_COMPLETE_FILENAMES | INPUT_COMPLETE_HOSTNAMES |
	INPUT_COMPLETE_VARIABLES | INPUT_COMPLETE_USERNAMES;
    in->current_max_len = initial_buffer_len;
    in->buffer = xstrndup (def_text, initial_buffer_len);
    in->color = color;
    in->field_len = len;
    in->first = 1;
    in->first_shown = 0;
    in->disable_update = 0;
    in->mark = 0;
    in->need_push = 1;
    in->is_password = 0;

    in->point = strlen (in->buffer);
    in->first = 1;
    return in;
}

/* Listbox widget */

/* Should draw the scrollbar, but currently draws only
 * indications that there is more information
 */
static int listbox_cdiff (WLEntry *s, WLEntry *e);

static void
listbox_drawscroll (WListbox *l)
{
    int line;
    int i, top;
    int max_line = l->height-1;

    /* Are we at the top? */
    widget_move (&l->widget, 0, l->width);
    if (l->list == l->top)
	one_vline ();
    else
	addch ('^');

    /* Are we at the bottom? */
    widget_move (&l->widget, max_line, l->width);
    top = listbox_cdiff (l->list, l->top);
    if ((top + l->height == l->count) || l->height >= l->count)
	one_vline ();
    else
	addch ('v');

    /* Now draw the nice relative pointer */
    if (l->count)
	line = 1+ ((l->pos * (l->height-2)) / l->count);
    else
	line = 0;

    for (i = 1; i < max_line; i++) {
	widget_move (&l->widget, i, l->width);
	if (i != line)
	    one_vline ();
	else
	    addch ('*');
    }
}

static void
listbox_draw (WListbox *l, Dlg_head *h, int focused)
{
    WLEntry *e;
    int i;
    int sel_line;
    int normalc, selc;
    char *text;

    if (focused) {
	normalc = NORMALC;
	selc    = FOCUSC;
    } else {
	normalc = NORMALC;
	selc    = HOT_FOCUSC;
    }
    sel_line = -1;

    for (e = l->top, i = 0; (i < l->height); i++) {

	/* Display the entry */
	if (e == l->current && sel_line == -1) {
	    sel_line = i;
	    attrset (selc);
	} else
	    attrset (normalc);

	widget_move (&l->widget, i, 0);

	if ((i > 0 && e == l->list) || !l->list)
	    text = "";
	else {
	    text = e->text;
	    e = e->next;
	}
	printw (" %-*s ", l->width-2, name_trunc (text, l->width-2));
    }
    l->cursor_y = sel_line;
    if (!l->scrollbar)
	return;
    attrset (normalc);
    listbox_drawscroll (l);
}

/* Returns the number of items between s and e,
   must be on the same linked list */
static int
listbox_cdiff (WLEntry *s, WLEntry *e)
{
    int c;

    for (c = 0; s != e; c++)
	s = s->next;
    return c;
}

static WLEntry *
listbox_check_hotkey (WListbox *l, int key)
{
    int i;
    WLEntry *e;

    i = 0;
    e = l->list;
    if (!e)
	return 0;

    while (1) {

	/* If we didn't find anything, return */
	if (i && e == l->list)
	    return 0;

	if (e->hotkey == key)
	    return e;

	i++;
	e = e->next;
    }
}

static void
listbox_remove_list (WListbox *l)
{
    WLEntry *p, *q;

    if (!l->count)
	return;

    p = l->list;

    while (l->count--) {
	q = p->next;
	free (p->text);
	free (p);
	p = q;
    }
    l->pos = l->count = 0;
    l->list = l->top = l->current = 0;
}

/*
 * bor 30.10.96: added force flag to remove *last* entry as well
 * bor 30.10.96: corrected selection bug if last entry was removed
 */

void
listbox_remove_current (WListbox *l, int force)
{
    WLEntry *p;

    /* Ok, note: this won't allow for emtpy lists */
    if (!force && (!l->count || l->count == 1))
	return;

    l->count--;
    p = l->current;

    if (l->count) {
	l->current->next->prev = l->current->prev;
	l->current->prev->next = l->current->next;
	if (p->next == l->list) {
	    l->current = p->prev;
	    l->pos--;
	}
	else
	    l->current = p->next;

	if (p == l->list)
	    l->list = l->top = p->next;
    } else {
	l->pos = 0;
	l->list = l->top = l->current = 0;
    }

    free (p->text);
    free (p);
}

/* Makes *e the selected entry (sets current and pos) */
void
listbox_select_entry (WListbox *l, WLEntry *dest)
{
    WLEntry *e;
    int pos;
    int top_seen;

    top_seen = 0;

    /* Special case */
    for (pos = 0, e = l->list; pos < l->count; e = e->next, pos++) {

	if (e == l->top)
	    top_seen = 1;

	if (e == dest) {
	    l->current = e;
	    if (top_seen) {
		while (listbox_cdiff (l->top, l->current) >= l->height)
		    l->top = l->top->next;
	    } else {
		l->top = l->current;
	    }
	    l->pos = pos;
	    return;
	}
    }
    /* If we are unable to find it, set decent values */
    l->current = l->top = l->list;
    l->pos = 0;
}

/* Selects from base the pos element */
static WLEntry *
listbox_select_pos (WListbox *l, WLEntry *base, int pos)
{
    WLEntry *last = l->list->prev;

    if (base == last)
	return last;
    while (pos--) {
	base = base->next;
	if (base == last)
	    break;
    }
    return base;
}

static inline int
listbox_back (WListbox *l)
{
    if (l->pos) {
	listbox_select_entry (l, listbox_select_pos (l, l->list, l->pos-1));
	return 1;
    }
    return 0;
}

static int
listbox_fwd (WListbox *l)
{
    if (l->current != l->list->prev) {
	listbox_select_entry (l, listbox_select_pos (l, l->list, l->pos+1));
	return 1;
    }
    return 0;
}

/* Returns 1 if we want a redraw */
static int
listbox_key (WListbox *l, int key)
{
    int i;
    int j = 0;

    if (!l->list)
	return 0;

    switch (key) {
    case KEY_HOME:
    case KEY_A1:
	l->current = l->top = l->list;
	l->pos = 0;
	return 1;

    case KEY_END:
    case KEY_C1:
	l->current = l->top = l->list->prev;
	for (i = min (l->height - 1, l->count - 1); i; i--)
	    l->top = l->top->prev;
	l->pos = l->count - 1;
	return 1;

    case XCTRL('p'):
    case KEY_UP:
	listbox_back (l);
	return 1;

    case XCTRL('n'):
    case KEY_DOWN:
	listbox_fwd (l);
	return 1;

    case KEY_NPAGE:
    case XCTRL('v'):
	for (i = 0; i < l->height-1; i++)
	    j |= listbox_fwd (l);
	return j > 0;

    case KEY_PPAGE:
    case ALT('v'):
	for (i = 0; i < l->height-1; i++)
	    j |= listbox_back (l);
	return j > 0;
    }
    return 0;
}

static int listbox_event (Gpm_Event *event, WListbox *l);
static int
wi_listbox_callback (Dlg_head *h, WListbox *l, int msg, int par)
{
    WLEntry  *e;
    /* int selected_color; Never used */
    int ret_code;

    switch (msg) {
    case WIDGET_INIT:
	return 1;

    case WIDGET_HOTKEY:
	if ((e = listbox_check_hotkey (l, par)) != NULL) {
	    listbox_select_entry (l, e);

	    /* Take the appropriate action */
	    if (l->action == listbox_finish) {
		l->widget.parent->running   = 0;
		l->widget.parent->ret_value = B_ENTER;
	    } else if (l->action == listbox_cback) {
		if ((*l->cback)(l) == listbox_finish) {
		    l->widget.parent->running = 0;
		    l->widget.parent->ret_value = B_ENTER;
		}
	    }
	    return 1;
	} else
	    return 0;

    case WIDGET_KEY:
	if ((ret_code = listbox_key (l, par)))
	    listbox_draw (l, h, 1);
	return ret_code;

    case WIDGET_CURSOR:
	widget_move (&l->widget, l->cursor_y, 0);
	return 1;

    case WIDGET_FOCUS:
    case WIDGET_UNFOCUS:
    case WIDGET_DRAW:
	listbox_draw (l, h, msg != WIDGET_UNFOCUS);
	return 1;
    }
    return default_proc (h, msg, par);
}

static int
listbox_event (Gpm_Event *event, WListbox *l)
{
    int i;

    Dlg_head *h = l->widget.parent;

    /* Single click */
    if (event->type & GPM_DOWN)
	dlg_select_widget (l->widget.parent, l);
    if (!l->list)
	return MOU_NORMAL;
    if (event->type & (GPM_DOWN|GPM_DRAG)) {
	if (event->x < 0 || event->x >= l->width)
	    return MOU_REPEAT;
	if (event->y < 1)
	    for (i = -event->y; i >= 0; i--)
		listbox_back (l);
	else if (event->y > l->height)
	    for (i = event->y - l->height; i > 0; i--)
		listbox_fwd (l);
	else
	    listbox_select_entry (l, listbox_select_pos (l, l->top,
							 event->y - 1));

	/* We need to refresh ourselves since the dialog manager doesn't */
	/* know about this event */
	wi_listbox_callback (h, l, WIDGET_DRAW, 0);
	mc_refresh ();
	return MOU_REPEAT;
    }

    /* Double click */
    if ((event->type & (GPM_DOUBLE|GPM_UP)) == (GPM_UP|GPM_DOUBLE)) {
	if (event->x < 0 || event->x >= l->width)
	    return MOU_NORMAL;
	if (event->y < 1 || event->y > l->height)
	    return MOU_NORMAL;

	dlg_select_widget (l->widget.parent, l);
	listbox_select_entry (l, listbox_select_pos (l, l->top, event->y - 1));

	switch (l->action) {
	case listbox_nothing:
	    break;

	case listbox_finish:
	    h->ret_value = B_ENTER;
	    dlg_stop (h);
	    return MOU_ENDLOOP;

	case listbox_cback:
	    if ((*l->cback)(l) == listbox_finish)
		return MOU_ENDLOOP;
	}
    }
    return MOU_NORMAL;
}

static void
listbox_destroy (WListbox *l)
{
    WLEntry *n, *p = l->list;
    int i;

    x_destroy_cmd (l);
    for (i = 0; i < l->count; i++) {
	n = p->next;
	free (p->text);
	free (p);
	p = n;
    }
}

WListbox *
listbox_new (int y, int x, int width, int height_l,
	     int action, lcback callback)
{
    WListbox *l = x_malloc (sizeof (WListbox), "listbox_new");

    init_widget (&l->widget, y, x, height_l, width,
		 (callback_fn)wi_listbox_callback,
		 (destroy_fn) listbox_destroy, (mouse_h)listbox_event);

    l->list   = l->top = l->current = 0;
    l->pos    = 0;
    l->width  = width;
    l->height = height_l;
    l->count  = 0;
    l->top    = 0;
    l->current= 0;
    l->cback  = callback;
    l->action = action;
    l->allow_duplicates = 1;
    l->scrollbar = slow_terminal ? 0 : 1;
    widget_want_hotkey (l->widget, 1);

    return l;
}

/* Listbox item adding function.  They still lack a lot of functionality */
/* any takers? */
/* 1.11.96 bor: added pos argument to control placement of new entry */
static void
listbox_append_item (WListbox *l, WLEntry *e, enum append_pos pos)
{
    if (!l->list) {
	l->list = e;
	l->top = e;
	l->current = e;
	e->next = l->list;
	e->prev = l->list;
    } else if (pos == LISTBOX_APPEND_AT_END) {
	e->next = l->list;
	e->prev = l->list->prev;
	l->list->prev->next = e;
	l->list->prev = e;
    } else if (pos == LISTBOX_APPEND_BEFORE) {
	e->next = l->current;
	e->prev = l->current->prev;
	l->current->prev->next = e;
	l->current->prev = e;
	if (l->list == l->current) {    /* move list one position down */
	    l->list = e;
	    l->top =  e;
	}
    } else if (pos == LISTBOX_APPEND_AFTER) {
	e->prev = l->current;
	e->next = l->current->next;
	l->current->next->prev = e;
	l->current->next = e;
    }
    l->count++;
}

static char *
listbox_add_item (WListbox *l, enum append_pos pos, int hotkey, const char *text,
			void *data)
{
    WLEntry *entry;

    if (!l)
	return 0;

    if (!l->allow_duplicates)
	if (listbox_search_text (l, text))
	    return 0;

    entry = x_malloc (sizeof (WLEntry), "listbox_add_item");
    entry->text = strdup (text);
    entry->data = data;
    entry->hotkey = hotkey;

    listbox_append_item (l, entry, pos);

    return entry->text;
}

/* Selects the nth entry in the listbox */
void
listbox_select_by_number (WListbox *l, int n)
{
    listbox_select_entry (l, listbox_select_pos (l, l->list, n));
}

WLEntry *
listbox_search_text (WListbox *l, const char *text)
{
    WLEntry *e;

    e = l->list;
    if (!e)
	return NULL;

    do {
	if(!strcmp (e->text, text))
	    return e;
	e = e->next;
    } while (e!=l->list);

    return NULL;
}

/* Returns the current string text as well as the associated extra data */
void
listbox_get_current (WListbox *l, char **string, char **extra)
{
    if (!l->current) {
	*string = 0;
	*extra  = 0;
    }
    if (string && l->current)
	*string = l->current->text;
    if (extra && l->current)
	*extra = l->current->data;
}

static int
buttonbar_callback (Dlg_head *h, WButtonBar *bb, int msg, int par)
{
    int i;

    switch (msg) {
    case WIDGET_INIT:
	return 1;

    case WIDGET_FOCUS:
	return 0;

    case WIDGET_HOTKEY:
	for (i = 0; i < 10; i++) {
	    if (par == KEY_F(i+1) && bb->labels [i].function) {
		(*bb->labels [i].function)(bb->labels [i].data);
		return 1;
	    }
	}
	return 0;

    case WIDGET_DRAW:
	if (!bb->visible)
	    return 1;
	widget_move (&bb->widget, 0, 0);
	attrset (DEFAULT_COLOR);
	printw ("%-*s", bb->widget.cols - 1, "");
	for (i = 0; i < COLS/8 && i < 10; i++) {
	    widget_move (&bb->widget, 0, i*8);
	    attrset (DEFAULT_COLOR);
	    printw ("%d", i+1);
	    attrset (SELECTED_COLOR);
	    printw ("%-*s", ((i+1) * 8 == COLS ? 5 : 6),
		    bb->labels [i].text ? bb->labels [i].text : "");
	    attrset (DEFAULT_COLOR);
	}
	attrset (SELECTED_COLOR);
	return 1;
    }
    return default_proc (h, msg, par);
}

static void
buttonbar_destroy (WButtonBar *bb)
{
    int i;

    for (i = 0; i < 10; i++) {
	if (bb->labels [i].text)
	    free (bb->labels [i].text);
    }
}

static int
buttonbar_event (Gpm_Event *event, WButtonBar *bb)
{
    int button;

    if (!(event->type & GPM_UP))
	return MOU_NORMAL;
    if (event->y == 2)
	return MOU_NORMAL;
    button = event->x / 8;
    if (button < 10 && bb->labels [button].function)
	(*bb->labels [button].function)(bb->labels [button].data);
    return MOU_NORMAL;
}

WButtonBar *
buttonbar_new (int visible)
{
    WButtonBar *bb = zmalloc (sizeof (WButtonBar));

    init_widget (&bb->widget, LINES-1, 0, 1, COLS,
		 (callback_fn) buttonbar_callback,
		 (destroy_fn) buttonbar_destroy, (mouse_h) buttonbar_event);

    bb->visible = visible;
    widget_want_hotkey (bb->widget, 1);
    widget_want_cursor (bb->widget, 0);

    return bb;
}

static void set_label_text (WButtonBar *bb, int Index, const char *text)
{
    if (bb->labels [Index-1].text)
	free (bb->labels [Index-1].text);

    bb->labels [Index-1].text = strdup (text);
}

/* paneletc is either the panel widget, or info or view or tree widget */
WButtonBar *
find_buttonbar (Dlg_head *h, Widget *paneletc)
{
    WButtonBar *bb;
    Widget_Item *item;
    int i;

    bb = 0;
    for (i = 0, item = h->current; i < h->count; i++, item = item->next) {
	if (item->widget->callback == (callback_fn) buttonbar_callback) {
	    bb = (WButtonBar *) item->widget;
	    break;
	}
    }
    return bb;
}

static void
define_label_data (Dlg_head *h, Widget *paneletc, int idx, const char *text,
		   buttonbarfn cback, void *data)
{
    WButtonBar *bb = find_buttonbar (h, paneletc);
    if (!bb)
	return;

    set_label_text (bb, idx, text);
    bb->labels [idx-1].function = (void (*)(void *)) cback;
    bb->labels [idx-1].data = data;
}

static void
define_label (Dlg_head *h, Widget *paneletc, int idx, const char *text, void (*cback)(void))
{
    define_label_data (h, paneletc, idx, text, (void (*)(void *)) cback, 0);
}

static void
redraw_labels (Dlg_head *h, Widget *paneletc)
{
    Widget_Item *item;
    int i;

    for (i = 0, item = h->current; i < h->count; i++, item = item->next) {
	if (item->widget->callback == (callback_fn) buttonbar_callback) {
	    widget_redraw (h, item);
	    return;
	}
    }
}
/* Curses utilities */

typedef void (*fnptr)(void);

typedef struct Fkey_Table_List {
    fnptr  actions[11];
    struct Fkey_Table_List *next;
    int    has_labels;
} Fkey_Table_List;



/* Return values: 0 = not a movement key, 1 = was a movement key */
static int check_movement_keys (int c, int additional, int page_size, void *data,
			 movefn backfn, movefn forfn, movefn topfn,
			 movefn bottomfn)
{
    switch (c) {
    case KEY_UP:
    case XCTRL ('p'):
	(*backfn)(data, 1);
	return 1;

    case KEY_DOWN:
    case XCTRL ('n'):
	(*forfn)(data, 1);
	return 1;

    case KEY_PPAGE:
    case ALT('v'):
	(*backfn)(data, page_size-1);
	return 1;

    case KEY_NPAGE:
    case XCTRL('v'):
	(*forfn)(data, page_size-1);
	return 1;

    case KEY_HOME:
    case KEY_A1:
	(*topfn)(data, 0);
	return 1;
    case KEY_END:
    case KEY_C1:
	(*bottomfn)(data, 0);
	return 1;
    }
    if (additional)
	switch (c) {
	case 'b':
	case XCTRL('h'):
	case KEY_BACKSPACE:
	case 0177:
	    (*backfn)(data, page_size-1);
	    return 1;
	case ' ':
	    (*forfn)(data, page_size-1);
	    return 1;
	case 'u':
	    (*backfn)(data, page_size / 2);
	    return 1;
	case 'd':
	    (*forfn)(data, page_size / 2);
	    return 1;
	case 'g':
	    (*topfn)(data, 0);
	    return 1;
	case 'G':
	    (*bottomfn)(data, 0);
	    return 1;
	}
    return 0;
}

/* Classification routines */
int is_abort_char (int c)
{
    return (c == XCTRL('c') || c == XCTRL('g') || c == ESC_CHAR ||
	    c == KEY_F(10));
}

/* Widget based utility functions. */

/* {{{ Common dialog callback */
static void
dialog_repaint (struct Dlg_head *h, int back, int title_fore)
{
    attrset (back);
    dlg_erase (h);
    draw_box (h, 1, 1, h->lines - 2, h->cols - 2);
    attrset (title_fore);
    if (h->title) {
	dlg_move (h, 1, (h->cols-strlen (h->title))/2);
	addstr (h->title);
    }
}

static void
common_dialog_repaint (struct Dlg_head *h)
{
    dialog_repaint (h, COLOR_NORMAL, COLOR_HOT_NORMAL);
}

static int
common_dialog_callback (struct Dlg_head *h, int id, int msg)
{
    if (msg == DLG_DRAW)
	common_dialog_repaint (h);
    return 0;
}

/* }}} */
/* {{{ Listbox utility functions */

static void
listbox_refresh (Dlg_head *h)
{
    dialog_repaint (h, COLOR_NORMAL, COLOR_HOT_NORMAL);
}

static int wt_listbox_callback (Dlg_head *h, int id, int msg)
{
    switch (msg) {
    case DLG_DRAW:
	listbox_refresh(h);
	return 1;
    }
    return 0;
}

static Listbox *
create_listbox_window (int cols, int lines, const char *title, char *help)
{
    int xpos, ypos, len;
    Listbox  *listbox = x_malloc (sizeof (Listbox), "create_listbox_window");
    char* cancel_string = "&Cancel";

    /* Adjust sizes */
    lines = (lines > LINES-6) ? LINES - 6 : lines;

	if (title && (cols < (len = strlen(title) + 2)))
		cols = len;

	/* no &, but 4 spaces around button for brackets and such */
	if (cols < (len = strlen(cancel_string) + 3))
		cols = len;

    cols = cols > COLS-6 ? COLS-6 : cols;

    /* I'm not sure if this -2 is safe, should test it */
    xpos = (COLS-cols)/2;
    ypos = (LINES-lines)/2 - 2;

    /* Create components */
    listbox->dlg = create_dlg (ypos, xpos, lines+6, cols+4, dialog_colors,
			       wt_listbox_callback, help, DLG_CENTER|DLG_GRID);
    x_set_dialog_title (listbox->dlg, title);

    listbox->list = listbox_new (2, 2, cols, lines, listbox_finish, 0);

    add_widget (listbox->dlg,
		button_new (lines+3, (cols/2 + 2) - len/2,
		B_CANCEL, NORMAL_BUTTON, cancel_string, 0, 0));
    add_widget (listbox->dlg, listbox->list);
    listbox_refresh(listbox->dlg);
    return listbox;
}

/* Returns the number of the item selected */
int run_listbox (Listbox *l)
{
    int val;

    run_dlg (l->dlg);
    if (l->dlg->ret_value == B_CANCEL)
	val = -1;
    else
	val = l->list->pos;
    destroy_dlg (l->dlg);
    free (l);
    return val;
}

/* }}} */


/* {{{ Query Dialog functions */
static int wt_query_callback (struct Dlg_head *h, int Id, int Msg)
{
    struct text_struct *info;

    info = (struct text_struct *) h->data;

    switch (Msg) {
    case DLG_DRAW:
	/* designate window */
	attrset (NORMALC);
	dlg_erase (h);
	draw_box (h, 1, 1, h->lines-2, h->cols-2);
	attrset (HOT_NORMALC);
	dlg_move (h, 1, (h->cols-strlen (info->header))/2);
	addstr (info->header);
	break;
    }
    return 0;
}


/* Used to ask questions to the user */
static int query_dialog (const char *header, const char *text, int flags, int count_q, ...)
{
    va_list ap;
    Dlg_head *query_dlg;
    int win_len = 0;
    int i;
    int result = -1;
    int xpos, ypos;
    int cols, lines;
    char *cur_name;

    /* set dialog colors */
    query_colors [0] = (flags & D_ERROR) ? ERROR_COLOR :  Q_UNSELECTED_COLOR;
    query_colors [1] = (flags & D_ERROR) ? REVERSE_COLOR : Q_SELECTED_COLOR;
    query_colors [2] = (flags & D_ERROR) ? ERROR_COLOR : COLOR_HOT_NORMAL;
    query_colors [3] = (flags & D_ERROR) ? COLOR_HOT_NORMAL :  COLOR_HOT_FOCUS;

    if (header == MSG_ERROR)
	    header = " Error ";

    if (count_q > 0) {
	va_start (ap, count_q);
	for (i = 0; i < count_q; i++) {
		char *cp = va_arg (ap, char *);

		win_len += strlen (cp) + 6;
		if (strchr (cp, '&') != NULL)
			win_len--;
	}
	va_end (ap);
    }

    /* count coordinates */
    cols = 6 + max (win_len, max (strlen (header), msglen (text, &lines)));
    lines += 4 + (count_q > 0 ? 2 : 0);
    xpos = COLS/2 - cols/2;
    ypos = LINES/3 - (lines-3)/2;
    query_dialog_pass.header = header;
    query_dialog_pass.text   = text;

    /* prepare dialog */
    query_dlg = create_dlg (ypos, xpos, lines, cols, query_colors,
			    wt_query_callback, "[QueryBox]", DLG_NONE);
    x_set_dialog_title (query_dlg, header);

    /* The data we need to query_dialog_pass to the callback */
    query_dlg->cols = cols;
    query_dlg->lines = lines;
    query_dlg->data  = &query_dialog_pass;

    query_dlg->direction = DIR_BACKWARD;

    if (count_q > 0) {

	cols = (cols-win_len-2)/2+2;
	va_start (ap, count_q);
	for (i = 0; i < count_q; i++) {
	    cur_name = va_arg (ap, char *);
	    xpos = strlen (cur_name)+6;
		if (strchr(cur_name, '&') != NULL)
			xpos--;
	    add_widget (query_dlg, button_new
			(lines-3, cols, B_USER+i, NORMAL_BUTTON, cur_name,
			 0, 0));
	    cols += xpos;
	    if (i == sel_pos)
		query_dlg->initfocus = query_dlg->current;
	}
	va_end (ap);

	add_widget (query_dlg, label_new (2, 3, text));

	/* run dialog and make result */
	run_dlg (query_dlg);
	switch (query_dlg->ret_value) {
	case B_CANCEL:
	    break;
	default:
	    result = query_dlg->ret_value-B_USER;
	}

	/* free used memory */
	destroy_dlg (query_dlg);
    } else {
	add_widget (query_dlg, label_new (2, 3, text));
	add_widget (query_dlg, button_new(0, 0, 0, HIDDEN_BUTTON, "-", 0, 0));
	last_query_dlg = query_dlg;
    }
    sel_pos = 0;
    return result;
}

static void query_set_sel (int new_sel)
{
    sel_pos = new_sel;
}

/* }}} */

/* {{{ The message function */

/* To show nice messages to the users */
static Dlg_head *MC_message (int error, const char *header, char *text, ...)
{
    va_list  args;
    char     buffer [4096];
    Dlg_head *d;

    /* Setup the display information */
    strcpy (buffer, "\n");
    va_start (args, text);
    vsprintf (&buffer [1], text, args);
    strcat (buffer, "\n");
    va_end (args);

    query_dialog (header, buffer, error, 0);
    d = last_query_dlg;
    init_dlg (d);
    if (!(error & D_INSERT)) {
	mi_getch ();
	dlg_run_done (d);
	destroy_dlg (d);
    } else
	return d;
    return 0;
}
/* }}} */

/* {{{ Quick dialog routines */

static int quick_callback (struct Dlg_head *h, int id, int Msg)
{
    switch (Msg) {
    case DLG_DRAW:
	attrset (COLOR_NORMAL);
	dlg_erase (h);
	draw_box (h, 1, 1, h->lines-2, h->cols-2);

	attrset (COLOR_HOT_NORMAL);
	dlg_move (h, 1,((h->cols-strlen (h->data))/2));
	addstr (h->data);
	break;
    case DLG_KEY:
	if (id == '\n') {
	    h->ret_value = B_ENTER;
	    dlg_stop (h);
	    break;
	}
    }
    return 0;
}


static int quick_dialog_skip (const QuickDialog *qd, int nskip)
{
    Dlg_head *Dd;
    void     *widget;
    WRadio   *r;
    int      xpos;
    int      ypos;
    int      return_val;
    WInput   *Input;
    QuickWidget *qw;

    if (qd->xpos == -1)
	Dd = create_dlg (0, 0, qd->ylen, qd->xlen, dialog_colors, quick_callback,
			 qd->help, DLG_CENTER | DLG_TRYUP | DLG_GRID);
    else
	Dd = create_dlg (qd->ypos, qd->xpos, qd->ylen, qd->xlen, dialog_colors,
			 quick_callback, qd->help, DLG_GRID);

    x_set_dialog_title (Dd, qd->title);

    /* We pass this to the callback */
    Dd->cols  = qd->xlen;
    Dd->lines = qd->ylen;
    Dd->data  = qd->title;

    for (qw = qd->widgets; qw->widget_type; qw++) {
	xpos = (qd->xlen * qw->relative_x)/qw->x_divisions;
	ypos = (qd->ylen * qw->relative_y)/qw->y_divisions;

	switch (qw->widget_type) {
	case quick_checkbox:
	    widget = check_new (ypos, xpos, *qw->result, qw->text);
	    break;

	case quick_radio:
	    r = radio_new (ypos, xpos, qw->hotkey_pos, (const char * const *)qw->str_result, 1);
	    r->pos = r->sel = qw->value;
	    widget = r;
	    break;

	case quick_button:
	    widget = button_new (ypos, xpos, qw->value, (qw->value==B_ENTER) ? DEFPUSH_BUTTON : NORMAL_BUTTON,
				    qw->text, 0, 0);
	    break;

	    /* We use the hotkey pos as the field length */
	case quick_input:
	    Input = input_new (ypos, xpos, INPUT_COLOR,
			       qw->hotkey_pos, qw->text);
	    Input->is_password = qw->value == 1;
	    Input->point = 0;
	    if (qw->value & 2)
		Input->completion_flags |= INPUT_COMPLETE_CD;
	    widget = Input;
	    break;

	case quick_label:
	    widget = label_new (ypos, xpos, qw->text);
	    break;

	default:
	    widget = 0;
	    break;
	}
	qw->the_widget = widget;
	add_widgetl (Dd, widget, qw->layout);
    }

    while (nskip--)
	Dd->current = Dd->current->next;

    run_dlg (Dd);

    /* Get the data if we found something interesting */
    if (Dd->ret_value != B_CANCEL) {
	for (qw = qd->widgets; qw->widget_type; qw++) {
	    switch (qw->widget_type) {
	    case quick_checkbox:
		*qw->result = ((WCheck *) qw->the_widget)->state & C_BOOL;
		break;

	    case quick_radio:
		*qw->result = ((WRadio *) qw->the_widget)->sel;
		break;

	    case quick_input:
		*qw->str_result = strdup (((WInput *) qw->the_widget)->buffer);
		break;
	    }
	}
    }
    return_val = Dd->ret_value;
    destroy_dlg (Dd);

    return return_val;
}

static int quick_dialog (const QuickDialog *qd)
{
    return quick_dialog_skip (qd, 0);
}

/* }}} */

/* {{{ Input routines */
char *real_input_dialog_help (char *header, char *text, char *help, char *def_text)
{
    QuickDialog Quick_input;
    static const QuickWidget quick_widgets_st [] = {
      { quick_button, 6, 10, 1, 0, "&Cancel", 0, B_CANCEL, 0, 0, XV_WLAY_RIGHTOF},
      { quick_button, 3, 10, 1, 0, "&Ok", 0, B_ENTER, 0, 0, XV_WLAY_CENTERROW },
      { quick_input,  4, 80, 0, 0, "", 58, 0, 0, 0, XV_WLAY_NEXTROW },
      { quick_label,  3, 80, 2, 0, "", 0, 0, 0, 0, XV_WLAY_NEXTROW },
      { 0 }
    };

    int len;
    int i;
    int lines;
    char *my_str;
    QuickWidget *quick_widgets;

    quick_widgets = alloca(sizeof(quick_widgets_st));
    memcpy(quick_widgets, quick_widgets_st, sizeof(quick_widgets_st));
    len = max (strlen (header), msglen (text, &lines)) + 4;
    len = max (len, 64);

    if (strncmp (text, "Password:", strlen ("Password")) == 0) {
	quick_widgets [2].value = 1;
    } else {
	quick_widgets [2].value = 0;
    }

    Quick_input.xlen  = len;
    Quick_input.xpos  = -1;
    Quick_input.title = header;
    Quick_input.help  = help;
    quick_widgets [2+1].text = text;
    quick_widgets [2].text = def_text;

    for (i = 0; i < 4; i++)
	quick_widgets [i].y_divisions = lines+6;
    Quick_input.ylen  = lines + 6;

    for (i = 0; i < 3; i++)
	quick_widgets [i].relative_y += 2 + lines;

    quick_widgets [2].str_result = &my_str;

    Quick_input.widgets = quick_widgets;
    if (quick_dialog (&Quick_input) != B_CANCEL) {
	return *(quick_widgets [2].str_result);
    } else
	return 0;
}

static char *input_dialog (char *header, char *text, char *def_text)
{
    return input_dialog_help (header, text, "[Input Line Keys]", def_text);
}

static char *input_expand_dialog (char *header, char *text, char *def_text)
{
    char *result;
    char *expanded;

    result = input_dialog (header, text, def_text);
    if (result) {
	expanded = tilde_expand (result);
	if (expanded) {
	    free (result);
	    return expanded;
	} else
	    return result;
    }
    return result;
}

/* }}} */

/* }}} */
/*
  Cause emacs to enter folding mode for this file:
  Local variables:
  end:
*/
/* Virtual File System */
#ifdef USE_VFS

static int parse_ls_lga (char *p, struct stat *s, char **filename, char **linkname);

#ifdef USE_NETCODE
static struct linklist *
linklist_init(void)
{
    struct linklist *head;

    head = x_malloc(sizeof(struct linklist), "struct linklist");
    if (head) {
	head->prev = head->next = head;
	head->data = NULL;
    }
    return head;
}

static void
linklist_destroy(struct linklist *head, void (*destructor) (void *))
{
    struct linklist *p, *q;

    for (p = head->next; p != head; p = q) {
	if (p->data && destructor)
	    (*destructor) (p->data);
	q = p->next;
	free(p);
    }
    free(head);
}

static int
linklist_insert(struct linklist *head, void *data)
{
    struct linklist *p;

    p = x_malloc(sizeof(struct linklist), "struct linklist");
    if (p == NULL)
	return 0;
    p->data = data;
    p->prev = head->prev;
    p->next = head;
    head->prev->next = p;
    head->prev = p;
    return 1;
}

static void
linklist_delete_all(struct linklist *head, void (*destructor) (void *))
{
    struct linklist *p, *q;

    for (p = head->next; p != head; p = q) {
	destructor(p->data);
	q = p->next;
	free(p);
    }
    head->next = head->prev = head;
    head->data = NULL;
}

static int
linklist_delete(struct linklist *head, void *data)
{
    struct linklist *h = head->next;

    while (h != head) {
	if (h->data == data) {
	    h->prev->next = h->next;
	    h->next->prev = h->prev;
	    free(h);
	    return 1;
	}
	h = h->next;
    }
    return 0;
}
#endif

static struct extfs_entry *
extfs_find_entry (struct extfs_entry *dir, char *name, int make_dirs, int make_file);

static struct extfs_archive *exrfs_first_archive = NULL;;
static int extfserrno;
static struct stat extfs_hstat;               /* Stat struct corresponding */
static char *extfs_current_file_name, *extfs_current_link_name;
static char *extfs_current_dir;

enum {
    EFS_ARG_EMPTY,
    EFS_NEED_FILE,
    EFS_NEED_ARG
};


struct extfs_ext {
    char *ext;
    struct extfs_ext *next;
};

static struct extfs_ext *extfs_extensions [MAXEXTFS];

static void extfs_fill_names (void (*func)(char *))
{
    struct extfs_archive *a = exrfs_first_archive;
    char *name;

    while (a) {
	name = xasprintf ("%s:%s/%s", extfs_prefixes [a->fstype],
			    (a->name ? a->name : ""),
			     a->current_dir->name);
	(*func)(name);
	free (name);
	a = a->next;
    }
}

static void extfs_make_dot_doubledot (struct extfs_entry *ent)
{
    struct extfs_entry *entry = (struct extfs_entry *)
	x_malloc (sizeof (struct extfs_entry), "Extfs: extfs_entry");
    struct extfs_entry *parentry = ent->dir;
    struct extfs_inode *inode = ent->inode, *parent;

    parent = (parentry != NULL) ? parentry->inode : NULL;
    entry->name = strdup (".");
    entry->has_changed = 0;
    entry->inode = inode;
    entry->dir = ent;
    inode->local_filename = NULL;
    inode->first_in_subdir = entry;
    inode->last_in_subdir = entry;
    inode->nlink++;
    entry->next_in_dir = (struct extfs_entry *)
	x_malloc (sizeof (struct extfs_entry), "Extfs: extfs_entry");
    entry=entry->next_in_dir;
    entry->name = strdup ("..");
    entry->has_changed = 0;
    inode->last_in_subdir = entry;
    entry->next_in_dir = NULL;
    if (parent != NULL) {
	entry->inode = parent;
	entry->dir = parentry;
	parent->nlink++;
    } else {
	entry->inode = inode;
	entry->dir = ent;
	inode->nlink++;
    }
}

static struct extfs_entry *extfs_generate_entry (struct extfs_archive *archive,
    char *name, struct extfs_entry *parentry, mode_t mode)
{
    mode_t myumask;
    struct extfs_inode *inode, *parent;
    struct extfs_entry *entry;

    parent = (parentry != NULL) ? parentry->inode : NULL;
    entry = (struct extfs_entry *)
	x_malloc (sizeof (struct extfs_entry), "Extfs: extfs_entry");

    entry->name = strdup (name);
    entry->has_changed = 0;
    entry->next_in_dir = NULL;
    entry->dir = parentry;
    if (parent != NULL) {
	parent->last_in_subdir->next_in_dir = entry;
	parent->last_in_subdir = entry;
    }
    inode = (struct extfs_inode *)
	x_malloc (sizeof (struct extfs_inode), "Extfs: extfs_inode");
    entry->inode = inode;
    inode->has_changed = 0;
    inode->local_filename = NULL;
    inode->linkname = 0;
    inode->inode = (archive->__inode_counter)++;
    inode->dev = archive->rdev;
    inode->archive = archive;
    myumask = umask (022);
    umask (myumask);
    inode->mode = mode & ~myumask;
    mode = inode->mode;
    inode->rdev = 0;
    inode->uid = my_uid;
    inode->gid = my_gid;
    inode->size = 0;
    inode->mtime = time (NULL);
    inode->atime = inode->mtime;
    inode->ctime = inode->mtime;
    inode->nlink = 1;
    if (S_ISDIR (mode))
	extfs_make_dot_doubledot (entry);
    return entry;
}

static void extfs_free_entries (struct extfs_entry *entry)
{
    return;
}

static void extfs_free_archive (struct extfs_archive *archive)
{
    extfs_free_entries (archive->root_entry);
    if (archive->local_name != NULL) {
	struct stat my;

	mc_stat (archive->local_name, &my);
	mc_ungetlocalcopy (archive->name, archive->local_name,
	    archive->local_stat.st_mtime != my.st_mtime);
	/* ungetlocalcopy frees local_name for us */
    }
    if (archive->name)
	free (archive->name);
    free (archive);
}

static FILE *open_extfs_archive (int fstype, char *name, struct extfs_archive **pparc)
{
    static dev_t __extfs_no = 0;
    FILE *result;
    mode_t mode;
    char *cmd;
    char *mc_extfsdir;
    struct stat mystat;
    struct extfs_archive *current_archive;
    struct extfs_entry *root_entry;
    char *local_name = NULL, *tmp = 0;
    const int uses_archive = extfs_need_archive [fstype];

    if (uses_archive == EFS_NEED_FILE) {
	if (mc_stat (name, &mystat) == -1)
	    return NULL;
	if (!vfs_file_is_local (name)) {
	    local_name = mc_getlocalcopy (name);
	    if (local_name == NULL)
		return NULL;
	}
	tmp = name_quote (name, 0);
    }

    if (uses_archive == EFS_NEED_ARG) {
	tmp = name_quote (name, 0);
    }

    mc_extfsdir = concat_mcpath_file ("extfs/");
    cmd = xasprintf ("%s%s list %s", mc_extfsdir, extfs_prefixes [fstype],
			local_name ? local_name : tmp);
    if (tmp)
	free (tmp);
    result = popen (cmd, "r");
    free (cmd);
    free (mc_extfsdir);
    if (result == NULL) {
	if (local_name != NULL && uses_archive)
	    mc_ungetlocalcopy (name, local_name, 0);
	return NULL;
    }

    current_archive = (struct extfs_archive *)
		      x_malloc (sizeof (struct extfs_archive), "Tar archive");
    current_archive->fstype = fstype;
    current_archive->name = name ? strdup (name): name;
    current_archive->local_name = local_name;

    if (local_name != NULL)
	mc_stat (local_name, &current_archive->local_stat);
    current_archive->__inode_counter = 0;
    current_archive->fd_usage = 0;
    current_archive->extfsstat = mystat;
    current_archive->rdev = __extfs_no++;
    current_archive->next = exrfs_first_archive;
    exrfs_first_archive = current_archive;
    mode = current_archive->extfsstat.st_mode & 07777;
    if (mode & 0400)
	mode |= 0100;
    if (mode & 0040)
	mode |= 0010;
    if (mode & 0004)
	mode |= 0001;
    mode |= S_IFDIR;
    root_entry = extfs_generate_entry (current_archive, "/", NULL, mode);
    root_entry->inode->uid = current_archive->extfsstat.st_uid;
    root_entry->inode->gid = current_archive->extfsstat.st_gid;
    root_entry->inode->atime = current_archive->extfsstat.st_atime;
    root_entry->inode->ctime = current_archive->extfsstat.st_ctime;
    root_entry->inode->mtime = current_archive->extfsstat.st_mtime;
    current_archive->root_entry = root_entry;
    current_archive->current_dir = root_entry;

    *pparc = current_archive;

    return result;
}

/*
 * Main loop for reading an archive.
 * Returns 0 on success, -1 on error.
 */
static int
read_extfs_archive (int fstype, char *name, struct extfs_archive **pparc)
{
    FILE *extfsd;
    char *buffer;
    struct extfs_archive *current_archive;

    if ((extfsd = open_extfs_archive (fstype, name, &current_archive)) == NULL) {
	message_3s (1, " Error ", "Couldn't open %s archive\n%s",
	    extfs_prefixes [fstype], name);
	return -1;
    }

    buffer = x_malloc (4096, "Extfs: buffer");
    while (fgets (buffer, 4096, extfsd) != NULL) {
	extfs_current_link_name = NULL;
	if (parse_ls_lga (buffer, &extfs_hstat, &extfs_current_file_name, &extfs_current_link_name)) {
	    struct extfs_entry *entry, *pent;
	    struct extfs_inode *inode;
	    char *p, *q, *cfn = extfs_current_file_name;

	    if (*cfn) {
		if (*cfn == '/')
		    cfn++;
		p = strchr (cfn, 0);
		if (p != NULL && p != cfn && *(p - 1) == '/')
		    *(p - 1) = 0;
		p = strrchr (cfn, '/');
		if (p == NULL) {
		    p = cfn;
		    q = strchr (cfn, 0);
		} else {
		    *(p++) = 0;
		    q = cfn;
		}
		if (S_ISDIR (extfs_hstat.st_mode) &&
		    (!strcmp (p, ".") || !strcmp (p, "..")))
		    goto read_extfs_continue;
		pent = extfs_find_entry (current_archive->root_entry, q, 1, 0);
		if (pent == NULL) {
		    message_1s (1, " Error ", "Inconsistent extfs archive");
		    /* FIXME: Should clean everything one day */
		    free (buffer);
		    pclose (extfsd);
		    return -1;
		}
		entry = (struct extfs_entry *) x_malloc (sizeof (struct extfs_entry), "Extfs: extfs_entry");
		entry->name = strdup (p);
		entry->has_changed = 0;
		entry->next_in_dir = NULL;
		entry->dir = pent;
		if (pent != NULL) {
		    if (pent->inode->last_in_subdir) {
			pent->inode->last_in_subdir->next_in_dir = entry;
			pent->inode->last_in_subdir = entry;
		    }
		}
		if (!S_ISLNK (extfs_hstat.st_mode) && extfs_current_link_name != NULL) {
		    pent = extfs_find_entry (current_archive->root_entry, extfs_current_link_name, 0, 0);
		    if (pent == NULL) {
			message_1s (1, " Error ", "Inconsistent extfs archive");
		    /* FIXME: Should clean everything one day */
			free (buffer);
			pclose (extfsd);
			return -1;
		    } else {
			entry->inode = pent->inode;
			pent->inode->nlink++;
		    }
		} else {
		    inode = (struct extfs_inode *) x_malloc (sizeof (struct extfs_inode), "Extfs: extfs_inode");
		    entry->inode = inode;
		    inode->local_filename = NULL;
		    inode->has_changed = 0;
		    inode->inode = (current_archive->__inode_counter)++;
		    inode->nlink = 1;
		    inode->dev = current_archive->rdev;
		    inode->archive = current_archive;
		    inode->mode = extfs_hstat.st_mode;
		    inode->rdev = extfs_hstat.st_rdev;
		    inode->uid = extfs_hstat.st_uid;
		    inode->gid = extfs_hstat.st_gid;
		    inode->size = extfs_hstat.st_size;
		    inode->mtime = extfs_hstat.st_mtime;
		    inode->atime = extfs_hstat.st_atime;
		    inode->ctime = extfs_hstat.st_ctime;
		    if (extfs_current_link_name != NULL && S_ISLNK (extfs_hstat.st_mode)) {
			inode->linkname = extfs_current_link_name;
			extfs_current_link_name = NULL;
		    } else {
			inode->mode &= ~S_IFLNK;
			inode->linkname = NULL;
		    }
		    if (S_ISDIR (extfs_hstat.st_mode))
			extfs_make_dot_doubledot (entry);
		}
	    }
 read_extfs_continue:
	    free (extfs_current_file_name);
	    if (extfs_current_link_name != NULL)
		free (extfs_current_link_name);
	}
    }
    pclose (extfsd);
    *pparc = current_archive;
    free (buffer);
    return 0;
}

static char *extfs_analysis (const char *name, char **archive, int *fstype, int is_dir)
{
    char *local;
    const char *p;
    int j;
    int prefname;
    struct extfs_ext *ext;
					       /*  |  this is len of prefix plus some minimum
						*  v  space needed for the extension */
    *fstype = extfs_prefix_to_type (name);
    if (*fstype == -1) {
	extfserrno = ENOENT;
	return NULL;
    }

    prefname = strlen (extfs_prefixes [*fstype]) + 1;
    if (extfs_need_archive [*fstype] == EFS_ARG_EMPTY) {
	*archive = 0;
	return strdup (name +  prefname);
    }

    if (extfs_need_archive [*fstype] == EFS_NEED_ARG) {
	name += prefname;
	j = strlen(name);
	p = strchr (name, '/');
	if (p)
	    j = (p - name);

	*archive = xstrndup (name, j);
	return p ? strdup (p+1) : NULL;
    }
    for (p = name + strlen (name); p > name; p--)
	if (*p == '/' || (is_dir && !*p))
	    for (ext = extfs_extensions [*fstype]; ext; ext = ext->next) {
		j = strlen (ext->ext);
		if (p - j < name + 1)
		    continue;
		if (!strncmp (p - j, ext->ext, j)) {
		    char c = *p;

		    local = strdup (p);
		    local[(p - name)] = 0;
		    *archive = vfs_canon (name + prefname);
		    local[(p - name)] = c;
		    return local;
		}
	    }
    extfserrno = ENOENT;
    return NULL;
}

struct extfs_pseudofile {
    struct extfs_archive *archive;
    unsigned int has_changed:1;
    int local_handle;
    struct extfs_entry *entry;
};


static caddr_t extfs_mmap (caddr_t addr, size_t len, int prot, int flags, void *data, off_t offset)
{
    return (caddr_t)-1;
}

static int extfs_munmap (caddr_t addr, size_t len, void *data)
{
    return -1;
}

static void *extfs_open (char *file, int flags, int mode);
static int extfs_close (void *data);

static int extfs_read (void *data, char *buffer, int count_r)
{
    struct extfs_pseudofile *file = (struct extfs_pseudofile *)data;

    return read (file->local_handle, buffer, count_r);
}


static int extfs_write (void *data, char *buf, int nbyte)
{
    struct extfs_pseudofile *file = (struct extfs_pseudofile *)data;

    file->has_changed = 1;
    return write (file->local_handle, buf, nbyte);
}

static void *extfs_opendir (char *Dirname);

static void *extfs_readdir (void *data)
{
    static struct {
	struct dirent dir;
    } dir;

    struct extfs_entry **extfs_info = (struct extfs_entry **) data;

    if (*extfs_info == NULL)
	return NULL;

    strcpy (&(dir.dir.d_name [0]), (*extfs_info)->name);

    *extfs_info = (*extfs_info)->next_in_dir;

    return (void *)&dir;
}

static int extfs_closedir (void *data)
{
    free (data);
    return 0;
}

static int _extfs_stat (const char *path, struct stat *buf, int resolve);

static int extfs_stat (const char *path, struct stat *buf)
{
    return _extfs_stat (path, buf, 1);
}

static int extfs_lstat (const char *path, struct stat *buf)
{
    return _extfs_stat (path, buf, 0);
}

static int extfs_fstat (void *data, struct stat *buf);

static int extfs_chmod (char *path, int mode)
{
    return 0;
}

static int extfs_chown (char *path, int owner, int group)
{
    return -1;
}

static char *
extfs_get_path (const char *inname, struct extfs_archive **archive, int is_dir,
    int do_not_open);

static int extfs_readlink (char *path, char *buf, int size)
{
    struct extfs_archive *archive;
    char *p, *q;
    int i;
    struct extfs_entry *entry;

    if ((p = extfs_get_path (path, &archive, 0, 0)) == NULL)
	return -1;
    q = (*p == '/') ? p + 1 : p;
    entry = extfs_find_entry (archive->root_entry, q, 0, 0);
    free (p);
    if (entry == NULL)
	return -1;
    if (!S_ISLNK (entry->inode->mode)) {
	extfserrno = EINVAL;
	return -1;
    }
    if (size > (i = strlen (entry->inode->linkname))) {
	size = i;
    }
    strncpy (buf, entry->inode->linkname, i);
    return i;
}

static int extfs_unlink (char *path)
{
    return -1;
}

static int extfs_symlink (char *n1, char *n2)
{
    return -1;
}

static int extfs_rename (char *a, char *b)
{
    return -1;
}

static int extfs_mknod (char *path, int mode, int dev)
{
    return -1;
}

static int extfs_link (char *p1, char *p2)
{
    return -1;
}

static int extfs_mkdir (char *path, mode_t mode)
{
    return -1;
}

static int extfs_rmdir (char *path)
{
    return -1;
}

static int extfs_chdir (char *path);

static int extfs_errno (void)
{
    return extfserrno;
}

static int extfs_lseek (void *data, off_t offset, int whence)
{
    struct extfs_pseudofile *file = (struct extfs_pseudofile *) data;

    return lseek (file->local_handle, offset, whence);
}

static vfsid extfs_getid (char *path, struct vfs_stamping **parent);

static int extfs_nothingisopen (vfsid id)
{
    if (((struct extfs_archive *)id)->fd_usage <= 0)
	return 1;
    else
	return 0;
}

static void extfs_free (vfsid id);

static char *extfs_getlocalcopy (char *path)
{
    struct extfs_pseudofile *fp =
	(struct extfs_pseudofile *) extfs_open (path, O_RDONLY, 0);
    char *p;

    if (fp == NULL)
	return NULL;
    if (fp->entry->inode->local_filename == NULL) {
	extfs_close ((void *) fp);
	return NULL;
    }
    p = strdup (fp->entry->inode->local_filename);
    fp->archive->fd_usage++;
    extfs_close ((void *) fp);
    return p;
}

static void extfs_ungetlocalcopy (char *path, char *local, int has_changed)
{
    struct extfs_pseudofile *fp =
	(struct extfs_pseudofile *) extfs_open (path, O_WRONLY, 0);

    if (fp == NULL)
	return;
    if (!strcmp (fp->entry->inode->local_filename, local)) {
	fp->entry->inode->has_changed = has_changed;
	fp->archive->fd_usage--;
	extfs_close ((void *) fp);
	return;
    } else {
	/* Should not happen */
	extfs_close ((void *) fp);
	mc_def_ungetlocalcopy (path, local, has_changed);
    }
}


static vfs extfs_vfs_ops =
{
    extfs_open,
    extfs_close,
    extfs_read,
    extfs_write,

    extfs_opendir,
    extfs_readdir,
    extfs_closedir,

    extfs_stat,
    extfs_lstat,
    extfs_fstat,

    extfs_chmod,                /* unimplemented */
    extfs_chown,                /* unimplemented */
    NULL,

    extfs_readlink,

    extfs_symlink,              /* unimplemented */
    extfs_link,                 /* unimplemented */
    extfs_unlink,               /* unimplemented */

    extfs_rename,               /* unimplemented */
    extfs_chdir,
    extfs_errno,
    extfs_lseek,
    extfs_mknod,                /* unimplemented */

    extfs_getid,
    extfs_nothingisopen,
    extfs_free,

    extfs_getlocalcopy,
    extfs_ungetlocalcopy,

    extfs_mkdir,                /* unimplemented */
    extfs_rmdir,                /* unimplemented */
    NULL,
    NULL,
    NULL,
    extfs_mmap,
    extfs_munmap
};

static void *local_open (char *file, int flags, int mode)
{
    int *local_info;
    int fd;

    fd = open (file, flags, mode);
    if (fd == -1)
	return 0;

    local_info = (int *) x_malloc (sizeof (int), "Local fs");
    *local_info = fd;

    return local_info;
}

static int local_read (void *data, char *buffer, int count_r)
{
    int n;

    if (!data)
	return -1;

    while ((n = read (*((int *) data), buffer, count_r)) == -1) {
#ifdef EAGAIN
	if (errno == EAGAIN) continue;
#endif
#ifdef EINTR
	if (errno == EINTR) continue;
#endif
	return -1;
    }
    return n;
}

static int local_close (void *data)
{
    int fd;

    if (!data)
	return -1;

    fd =  *(int *) data;
    free (data);
    return close (fd);
}

static int local_errno (void)
{
    return errno;
}

static void *local_opendir (char *Dirname)
{
    DIR **local_info;
    DIR *dir;

    dir = opendir (Dirname);
    if (!dir)
	return 0;

    local_info = (DIR **) x_malloc (sizeof (DIR *), "Local fs");
    *local_info = dir;

    return local_info;
}

static void *local_readdir (void *data)
{
    return readdir (*(DIR **) data);
}

static int local_closedir (void *data)
{
    int i;

    i = closedir (* (DIR **) data);
    if (data)
	free (data);
    return i;
}

static int local_stat (const char *path, struct stat *buf)
{
    return stat (path, buf);
}

static int local_lstat (const char *path, struct stat *buf)
{
    return lstat (path,buf);
}

static int local_fstat (void *data, struct stat *buf)
{
    return fstat (*((int *) data), buf);
}

static int local_chmod (char *path, int mode)
{
    return chmod (path, mode);
}

static int local_chown (char *path, int owner, int group)
{
    return chown (path, owner, group);
}


static int local_readlink (char *path, char *buf, int size)
{
    return readlink (path, buf, size);
}

static int local_unlink (char *path)
{
    return unlink (path);
}

static int local_symlink (char *n1, char *n2)
{
    return symlink (n1, n2);
}

static int local_write (void *data, char *buf, int nbyte)
{
    int fd;
    int n;

    if (!data)
	return -1;

    fd = * (int *) data;
    while ((n = write (fd, buf, nbyte)) == -1) {
#ifdef EAGAIN
	if (errno == EAGAIN) continue;
#endif
#ifdef EINTR
	if (errno == EINTR) continue;
#endif
	break;
    }
    return n;
}

static int local_rename (char *a, char *b)
{
    return rename (a, b);
}

static int local_chdir (char *path)
{
    return chdir (path);
}

static int local_lseek (void *data, off_t offset, int whence)
{
    int fd = * (int *) data;

    return lseek (fd, offset, whence);
}

static int local_mknod (char *path, int mode, int dev)
{
    return mknod (path, mode, dev);
}

static int local_link (char *p1, char *p2)
{
    return link (p1, p2);
}

static int local_mkdir (char *path, mode_t mode)
{
    return mkdir (path, mode);
}

static int local_rmdir (char *path)
{
    return rmdir (path);
}

static vfsid local_getid (char *path, struct vfs_stamping **parent)
{
    *parent = NULL;
    return (vfsid) -1; /* We do not free local fs stuff at all */
}

static int local_nothingisopen (vfsid id)
{
    return 0;
}

static void local_free (vfsid id)
{
}

static char *local_getlocalcopy (char *path)
{
    return strdup (path);
}

static void local_ungetlocalcopy (char *path, char *local, int has_changed)
{
}

static caddr_t local_mmap (caddr_t addr, size_t len, int prot, int flags, void *data, off_t offset)
{
    int fd = * (int *)data;

    return mmap (addr, len, prot, flags, fd, offset);
}

static int local_munmap (caddr_t addr, size_t len, void *data)
{
    return munmap (addr, len);
}

static vfs local_vfs_ops = {
    local_open,
    local_close,
    local_read,
    local_write,

    local_opendir,
    local_readdir,
    local_closedir,

    local_stat,
    local_lstat,
    local_fstat,

    local_chmod,
    local_chown,
    local_utime,

    local_readlink,
    local_symlink,
    local_link,
    local_unlink,

    local_rename,
    local_chdir,
    local_errno,
    local_lseek,
    local_mknod,

    local_getid,
    local_nothingisopen,
    local_free,

    local_getlocalcopy,
    local_ungetlocalcopy,

    local_mkdir,
    local_rmdir,

    NULL,
    NULL,
    NULL,
    local_mmap,
    local_munmap
};


/* Returns allocated path inside the archive or NULL */
static char *extfs_get_path (const char *inname, struct extfs_archive **archive, int is_dir,
    int do_not_open)
{
    char *local, *archive_name;
    int result = -1;
    struct extfs_archive *parc;
    struct vfs_stamping *parent;
    vfs *v;
    int fstype;

    local = extfs_analysis (inname, &archive_name, &fstype, is_dir);
    if (local == NULL) {
	extfserrno = ENOENT;
	return NULL;
    }

    /* In the case of the file systems that do not require a local
     * archive, we compare the fstype instead of the archive name,
     * of course, this has the drawback that we can't have more than
     * one external file system for each prefix.
     *
     * Actually, we should implement an alias mechanism that would
     * translate: "a:" to "dos:a.
     *
     */
    for (parc = exrfs_first_archive; parc != NULL; parc = parc->next)
	if (archive_name) {
	    if (parc->name) {
		if (!strcmp (parc->name, archive_name)) {
		    vfs_stamp (&extfs_vfs_ops, (vfsid) parc);
		    goto return_success;
		}
	    }
	} else {
	    if (parc->fstype == fstype) {
		vfs_stamp (&extfs_vfs_ops, (vfsid) parc);
		goto return_success;
	    }
	}

    if (do_not_open)
	result = -1;
    else
	result = read_extfs_archive (fstype, archive_name, &parc);
    if (result == -1) {
	extfserrno = EIO;
	free(local);
	if (archive_name)
	    free(archive_name);
	return NULL;
    }
    if (archive_name) {
	v = vfs_type (archive_name);
	if (v == &local_vfs_ops) {
	    parent = NULL;
	} else {
	    parent = x_malloc (sizeof (struct vfs_stamping), "vfs stamping");
	    parent->v = v;
	    parent->next = 0;
	    parent->id = (*v->getid) (archive_name, &(parent->parent));
	}
	vfs_add_noncurrent_stamps (&extfs_vfs_ops, (vfsid) parc, parent);
	vfs_rm_parents (parent);
    }
 return_success:
    *archive = parc;
    if (archive_name)
	free (archive_name);
    return local;
}

/* Returns allocated path (without leading slash) inside the archive  */
static char *extfs_get_path_from_entry (struct extfs_entry *entry)
{
    struct list {
	struct list *next;
	char *name;
    } *head, *p;
    char *localpath;
    size_t len;

    for (len = 0, head = 0; entry->dir; entry = entry->dir) {
	p = x_malloc (sizeof  (struct list), "Extfs: list");
	p->next = head;
	p->name = entry->name;
	head = p;
	len += strlen (entry->name) + 1;
    }

    localpath = x_malloc (len, "Extfs: localpath");
    *localpath = '\0';
    for ( ; head; ) {
	strcat (localpath, head->name);
	if (head->next)
	    strcat (localpath, "/");
	p = head;
	head = head->next;
	free (p);
    }
    return (localpath);
}


struct extfs_loop_protect {
    struct extfs_entry *entry;
    struct extfs_loop_protect *next;
};

static struct extfs_entry *
__extfs_find_entry (struct extfs_entry *dir, char *name,
		    struct extfs_loop_protect *list, int make_dirs, int make_file);

static struct extfs_entry *
__extfs_resolve_symlinks (struct extfs_entry *entry,
			  struct extfs_loop_protect *list)
{
    struct extfs_entry *pent;
    struct extfs_loop_protect *looping;

    if (!S_ISLNK (entry->inode->mode))
	return entry;
    for (looping = list; looping != NULL; looping = looping->next)
	if (entry == looping->entry) { /* Here we protect us against symlink looping */
	    extfs_errloop = 1;
	    return NULL;
	}
    looping = (struct extfs_loop_protect *)
	x_malloc (sizeof (struct extfs_loop_protect),
		 "Extfs: symlink looping protection");
    looping->entry = entry;
    looping->next = list;
    pent = __extfs_find_entry (entry->dir, entry->inode->linkname, looping, 0, 0);
    free (looping);
    if (pent == NULL)
	extfserrno = ENOENT;
    return pent;
}

static struct extfs_entry *extfs_resolve_symlinks (struct extfs_entry *entry)
{
    struct extfs_entry *res;

    extfs_errloop = 0;
    extfs_notadir = 0;
    res = __extfs_resolve_symlinks (entry, NULL);
    if (res == NULL) {
	if (extfs_errloop)
	    extfserrno = ELOOP;
	else if (extfs_notadir)
	    extfserrno = ENOTDIR;
    }
    return res;
}

static struct extfs_entry*
__extfs_find_entry (struct extfs_entry *dir, char *name,
		     struct extfs_loop_protect *list, int make_dirs, int make_file)
{
    struct extfs_entry *pent, *pdir;
    char *p, *q, *name_end;
    char c;

    if (*name == '/') { /* Handle absolute paths */
	name++;
	dir = dir->inode->archive->root_entry;
    }

    pent = dir;
    p = name;
    name_end = name + strlen (name);
    q = strchr (p, '/');
    c = '/';
    if (!q)
	q = strchr (p, 0);

    for (; pent != NULL && c && *p; ) {
	c = *q;
	*q = 0;

	if (strcmp (p, ".")) {
	    if (!strcmp (p, ".."))
		pent = pent->dir;
	    else {
		if ((pent = __extfs_resolve_symlinks (pent, list))==NULL) {
		    *q = c;
		    return NULL;
		}
		if (c == '/' && !S_ISDIR (pent->inode->mode)) {
		    *q = c;
		    extfs_notadir = 1;
		    return NULL;
		}
		pdir = pent;
		for (pent = pent->inode->first_in_subdir; pent; pent = pent->next_in_dir)
		    /* Hack: I keep the original semanthic unless
		       q+1 would break in the strchr */
		    if (!strcmp (pent->name, p)) {
			if (q + 1 > name_end) {
			    *q = c;
			    extfs_notadir = !S_ISDIR (pent->inode->mode);
			    return pent;
			}
			break;
		    }

		/* When we load archive, we create automagically
		 * non-existant directories
		 */
		if (pent == NULL && make_dirs) {
		    pent = extfs_generate_entry (dir->inode->archive, p, pdir, S_IFDIR | 0777);
		}
		if (pent == NULL && make_file) {
		    pent = extfs_generate_entry (dir->inode->archive, p, pdir, 0777);
		}
	    }
	}
	/* Next iteration */
	*q = c;
	p = q + 1;
	q = strchr (p, '/');
	if (!q)
	    q = strchr (p, 0);
    }
    if (pent == NULL)
	extfserrno = ENOENT;
    return pent;
}

static struct extfs_entry *extfs_find_entry (struct extfs_entry *dir, char *name, int make_dirs, int make_file)
{
    struct extfs_entry *res;

    extfs_errloop = 0;
    extfs_notadir = 0;
    res = __extfs_find_entry (dir, name, NULL, make_dirs, make_file);
    if (res == NULL) {
	if (extfs_errloop)
	    extfserrno = ELOOP;
	else if (extfs_notadir)
	    extfserrno = ENOTDIR;
    }
    return res;
}

static char *get_archive_name (struct extfs_archive *archive)
{
    char *archive_name;

    if (archive->local_name)
	archive_name = archive->local_name;
    else
	archive_name = archive->name;

    if (!archive_name || !*archive_name)
	return "no_archive_name";
    else
	return archive_name;
}

static void extfs_run (char *file)
{
    struct extfs_archive *archive;
    char *p, *q, *cmd, *archive_name, *mc_extfsdir;

    if ((p = extfs_get_path (file, &archive, 0, 0)) == NULL)
	return;
    q = name_quote (((*p == '/') ? p + 1 : p), 0);

    archive_name = name_quote (get_archive_name(archive), 0);

    mc_extfsdir = concat_mcpath_file ("extfs/");
    cmd = xasprintf ("%s%s run %s %s", mc_extfsdir,
			extfs_prefixes [archive->fstype],
			archive_name, q);
    free (archive_name);
    free (q);
    shell_execute(cmd, 0);
    free(cmd);
    free(mc_extfsdir);
    free(p);
}

static void *extfs_open (char *file, int flags, int mode)
{
    struct extfs_pseudofile *extfs_info;
    struct extfs_archive *archive;
    char *p, *q;
    char *mc_extfsdir;
    struct extfs_entry *entry;
    int local_handle;
    const int do_create = (flags & O_ACCMODE) != O_RDONLY;

    if ((p = extfs_get_path (file, &archive, 0, 0)) == NULL)
	return NULL;
    q = (*p == '/') ? p + 1 : p;
    entry = extfs_find_entry (archive->root_entry, q, 0, do_create);
    free (p);
    if (entry == NULL) {
	return NULL;
    }
    if ((entry = extfs_resolve_symlinks (entry)) == NULL) {
	return NULL;
    }
    if (S_ISDIR (entry->inode->mode)) {
	extfserrno = EISDIR;
	return NULL;
    }
    if (entry->inode->local_filename == NULL) {
	char *cmd, *archive_name;
	int fd_unuse;

	entry->inode->local_filename = mc_mkstemp(&fd_unuse);
	if(fd_unuse >= 0)
	    close(fd_unuse);
	p = extfs_get_path_from_entry (entry);
	q = name_quote (p, 0);
	free (p);
	archive_name = name_quote (get_archive_name (archive), 0);

	mc_extfsdir = concat_mcpath_file ("extfs/");
	cmd = xasprintf ("%s%s copyout %s %s %s", mc_extfsdir,
				extfs_prefixes [archive->fstype],
				archive_name,
				q, entry->inode->local_filename);
	free (q);
	free (archive_name);
	if (my_system (EXECUTE_AS_SHELL, MC_shell, cmd) && !do_create) {
	    free (entry->inode->local_filename);
	    entry->inode->local_filename = NULL;
	    free (cmd);
	    free (mc_extfsdir);
	    extfserrno = EIO;
	    return NULL;
	}
	free (cmd);
	free (mc_extfsdir);
    }

    local_handle = open (entry->inode->local_filename, flags, mode);
    if (local_handle == -1) {
	extfserrno = EIO;
	return NULL;
    }

    extfs_info = (struct extfs_pseudofile *) x_malloc (sizeof (struct extfs_pseudofile), "Extfs: extfs_open");
    extfs_info->archive = archive;
    extfs_info->entry = entry;
    extfs_info->has_changed = 0;
    extfs_info->local_handle = local_handle;

    /* i.e. we had no open files and now we have one */
    vfs_rmstamp (&extfs_vfs_ops, (vfsid) archive, 1);
    archive->fd_usage++;
    return extfs_info;
}

static int extfs_close (void *data)
{
    struct extfs_pseudofile *file;
    int    errno_code = 0;
    file = (struct extfs_pseudofile *)data;

    close (file->local_handle);

    /* Commit the file if it has changed */
    if (file->has_changed) {
	struct extfs_archive *archive;
	char   *archive_name, *file_name;
	char   *cmd;
	char   *mc_extfsdir;
	char   *p;

	archive = file->archive;
	archive_name = name_quote (get_archive_name (archive), 0);
	p = extfs_get_path_from_entry (file->entry);
	file_name = name_quote (p, 0);
	free (p);

	mc_extfsdir = concat_mcpath_file ("extfs/");
	cmd = xasprintf ("%s%s copyin %s %s %s", mc_extfsdir,
			    extfs_prefixes [archive->fstype],
			    archive_name,
			    file_name,
			    file->entry->inode->local_filename);
	free (archive_name);
	free (file_name);
	if (my_system (EXECUTE_AS_SHELL, MC_shell, cmd))
	    errno_code = EIO;
	free (cmd);
	free (mc_extfsdir);
	{
	  struct stat file_status;
	  if( stat(file->entry->inode->local_filename,&file_status) != 0 )
	    errno_code = EIO;
	  else  file->entry->inode->size = file_status.st_size;
       }

      file->entry->inode->mtime = time (NULL);
    }

    file->archive->fd_usage--;
    if (!file->archive->fd_usage) {
	struct vfs_stamping *parent;
	vfs *v;

	if (!file->archive->name || !*file->archive->name || (v = vfs_type (file->archive->name)) == &local_vfs_ops) {
	    parent = NULL;
	} else {
	    parent = x_malloc (sizeof (struct vfs_stamping), "vfs stamping");
	    parent->v = v;
	    parent->next = 0;
	    parent->id = (*v->getid) (file->archive->name, &(parent->parent));
	}
	vfs_add_noncurrent_stamps (&extfs_vfs_ops, (vfsid) (file->archive), parent);
	vfs_rm_parents (parent);
    }

    free (data);
    if (errno_code) {
	extfserrno = EIO;       /* Non standard, but who cares? */
	return -1;
    } else
	return 0;
}

static void *extfs_opendir (char *Dirname)
{
    struct extfs_archive *archive;
    char *p, *q;
    struct extfs_entry *entry;
    struct extfs_entry **extfs_info;

    if ((p = extfs_get_path (Dirname, &archive, 1, 0)) == NULL)
	return NULL;
    q = (*p == '/') ? p + 1 : p;
    entry = extfs_find_entry (archive->root_entry, q, 0, 0);
    free (p);
    if (entry == NULL)
	return NULL;
    if ((entry = extfs_resolve_symlinks (entry)) == NULL)
	return NULL;
    if (!S_ISDIR (entry->inode->mode)) {
	extfserrno = ENOTDIR;
	return NULL;
    }

    extfs_info = (struct extfs_entry **) x_malloc (sizeof (struct extfs_entry *), "Extfs: extfs_opendir");
    *extfs_info = entry->inode->first_in_subdir;

    return extfs_info;
}


static int _extfs_stat (const char *path, struct stat *buf, int resolve)
{
    struct extfs_archive *archive;
    char *p, *q;
    struct extfs_entry *entry;
    struct extfs_inode *inode;

    if ((p = extfs_get_path (path, &archive, 0, 0)) == NULL)
	return -1;
    q = (*p == '/') ? p + 1 : p;
    entry = extfs_find_entry (archive->root_entry, q, 0, 0);
    free (p);
    if (entry == NULL)
	return -1;
    if (resolve && (entry = extfs_resolve_symlinks (entry)) == NULL)
	return -1;
    inode = entry->inode;
    buf->st_dev = inode->dev;
    buf->st_ino = inode->inode;
    buf->st_mode = inode->mode;
    buf->st_nlink = inode->nlink;
    buf->st_uid = inode->uid;
    buf->st_gid = inode->gid;
    buf->st_rdev = inode->rdev;
    buf->st_size = inode->size;
    buf->st_blksize = 512;
    buf->st_blocks = (inode->size + 512 - 1) / 512;
    buf->st_atime = inode->atime;
    buf->st_mtime = inode->mtime;
    buf->st_ctime = inode->ctime;
    return 0;
}

static int extfs_fstat (void *data, struct stat *buf)
{
    struct extfs_pseudofile *file = (struct extfs_pseudofile *)data;
    struct extfs_inode *inode;

    inode = file->entry->inode;
    buf->st_dev = inode->dev;
    buf->st_ino = inode->inode;
    buf->st_mode = inode->mode;
    buf->st_nlink = inode->nlink;
    buf->st_uid = inode->uid;
    buf->st_gid = inode->gid;
    buf->st_rdev = inode->rdev;
    buf->st_size = inode->size;
    buf->st_blksize = 512;
    buf->st_blocks = (inode->size + 512 - 1) / 512;
    buf->st_atime = inode->atime;
    buf->st_mtime = inode->mtime;
    buf->st_ctime = inode->ctime;
    return 0;
}


static int extfs_chdir (char *path)
{
    struct extfs_archive *archive;
    char *p, *q, *res;
    struct extfs_entry *entry;

    extfserrno = ENOTDIR;
    if ((p = extfs_get_path (path, &archive, 1, 0)) == NULL)
	return -1;
    q = (*p == '/') ? p + 1 : p;
    entry = extfs_find_entry (archive->root_entry, q, 0, 0);
    if (entry == NULL) {
	free (p);
	return -1;
    }
    entry = extfs_resolve_symlinks (entry);
    if (entry == NULL) {
	free (p);
	return -1;
    }
    if (!S_ISDIR (entry->inode->mode)) {
	free (p);
	return -1;
    }
    entry->inode->archive->current_dir = entry;
    res = xasprintf ("%s:%s%s",
	extfs_prefixes [entry->inode->archive->fstype],
	entry->inode->archive->name, p);
    free (p);
    extfserrno = 0;
    if (extfs_current_dir)
	free (extfs_current_dir);
    extfs_current_dir = res;
    return 0;
}

static vfsid extfs_getid (char *path, struct vfs_stamping **parent)
{
    struct extfs_archive *archive;
    vfs *v;
    vfsid id;
    char *p;
    struct vfs_stamping *par;

    *parent = NULL;
    if ((p = extfs_get_path (path, &archive, 1, 1)) == NULL) {
	return (vfsid) -1;
    }
    free (p);
    if (archive->name) {
	v = vfs_type (archive->name);
	id = (*v->getid) (archive->name, &par);
	if (id != (vfsid)-1) {
	    *parent = x_malloc (sizeof (struct vfs_stamping), "vfs stamping");
	    (*parent)->v = v;
	    (*parent)->id = id;
	    (*parent)->parent = par;
	    (*parent)->next = NULL;
	}
    }
    return (vfsid) archive;
}

static void extfs_free_entry (struct extfs_entry *e)
{
    int i = --(e->inode->nlink);
    if (S_ISDIR (e->inode->mode) && e->inode->first_in_subdir != NULL) {
	struct extfs_entry *f = e->inode->first_in_subdir;

	e->inode->first_in_subdir = NULL;
	extfs_free_entry (f);
    }
    if (i <= 0) {
	if (e->inode->local_filename != NULL) {
	    unlink (e->inode->local_filename);
	    free (e->inode->local_filename);
	}
	if (e->inode->linkname != NULL)
	    free (e->inode->linkname);
	free (e->inode);
    }
    if (e->next_in_dir != NULL)
	extfs_free_entry (e->next_in_dir);
    free (e->name);
    free (e);
}

static void extfs_free (vfsid id)
{
    struct extfs_archive *parc;
    struct extfs_archive *archive = (struct extfs_archive *)id;

    extfs_free_entry (archive->root_entry);
    if (archive == exrfs_first_archive) {
	exrfs_first_archive = archive->next;
    } else {
	for (parc = exrfs_first_archive; parc != NULL; parc = parc->next)
	    if (parc->next == archive)
		break;
	if (parc != NULL)
	    parc->next = archive->next;
    }
    extfs_free_archive (archive);
}


static void extfs_init (void)
{
    void *Keys;
    char *key, *value;
    char *p, *q, c;
    char *mc_extfsini;

    mc_extfsini = concat_mcpath_file ("extfs/extfs.ini");
    Keys = profile_init_iterator ("extfs", mc_extfsini);
    extfs_no = 0;
    for (extfs_no = 0; Keys != NULL && extfs_no < MAXEXTFS; extfs_no++) {
	Keys = profile_iterator_next (Keys, &key, &value);

	/* Handle those with a trailing ':', those flag that the
	 * file system does not require an archive to work
	 */

	c = key [strlen (key)-1];
	if (c == ':') {
	    key [strlen (key)-1] = 0;
	    extfs_need_archive [extfs_no] = EFS_ARG_EMPTY;
	} else if (c == '*') {
	    key [strlen (key)-1] = 0;
	    extfs_need_archive [extfs_no] = EFS_NEED_ARG;
	} else
	    extfs_need_archive [extfs_no] = EFS_NEED_FILE;

	extfs_prefixes [extfs_no] = strdup (key);

	value = strdup (value);
	for (q = value; (p = strtok (q, bb_spacestok)) != NULL; ) {
	    struct extfs_ext *newn;

	    q = NULL;
	    newn = x_malloc (sizeof (struct extfs_ext), "prefixes");
	    newn->next = extfs_extensions [extfs_no];
	    newn->ext = strdup (p);
	    extfs_extensions [extfs_no] = newn;
	}
	free (value);
    }
    free_profile_name (mc_extfsini);
    free (mc_extfsini);
}

static int extfs_prefix_to_type (const char *path)
{
    int i, j;

    for (i = 0; i < extfs_no; i++) {
	j = strlen (extfs_prefixes [i]);
	if (!strncmp (path, extfs_prefixes [i], j) && path [j] == ':')
	    return i;
    }
    return -1;
}

static char *extfs_get_prefix (int idx)
{
    return extfs_prefixes [idx];
}

static void extfs_done (void)
{
    struct extfs_ext *current, *old;
    int i;

    for (i = 0; i < extfs_no; i++ ) {
	free (extfs_prefixes [i]);
	current = extfs_extensions [i];
	while (current) {
	    old = current;
	    current = current->next;
	    free (old->ext);
	    free (old);
	}
    }
    extfs_no = 0;
    if (extfs_current_dir)
	free (extfs_current_dir);
    extfs_current_dir = 0;
}

static char *last_current_dir = NULL;

#ifdef USE_NETCODE

#ifndef MAXHOSTNAMELEN
#    define MAXHOSTNAMELEN 64
#endif

#define UPLOAD_ZERO_LENGTH_FILE

static int ftpfserrno;
static int code;

/* If set, then log dialog between server and client */
static int ftpfs_debug_server_dialog = 0;

/* Where we store the transactions */
static FILE *ftpfs_logfile;

/* If true, the directory cache is forced to reload */
static int force_expiration;

static struct linklist *ftpfs_connections_list;



/* command wait_flag: */
#define NONE        0x00
#define WAIT_REPLY  0x01
#define WANT_STRING 0x02
static char reply_str [80];

static char *ftpfs_get_current_directory(struct ftpfs_connection *bucket);
static int __ftpfs_chdir (struct ftpfs_connection *bucket ,char *remote_path);
static struct ftpentry *_get_file_entry(struct ftpfs_connection *bucket,
		 char *file_name, int op, int flags);

/* Extract the hostname and username from the path */
/* path is in the form: [user@]hostname:port/remote-dir, e.g.:
 * ftp://sunsite.unc.edu/pub/linux
 * ftp://miguel@sphinx.nuclecu.unam.mx/c/nc
 * ftp://tsx-11.mit.edu:8192/
 * ftp://joe@foo.edu:11321/private
 * If the user is empty, e.g. ftp://@roxanne/private, then your login name
 * is supplied.
 * */

static char *ftpfs_get_host_and_username (char *path, char **host, char **user, int *port, char **pass)
{
    return get_host_and_username (path, host, user, port, 21, 1, pass);
}

static int
select_on_two (int fd1, int fd2)
{
    fd_set set;
    struct timeval timeout;
    int v;
    int maxfd = (fd1 > fd2 ? fd1 : fd2) + 1;

    timeout.tv_sec  = 1;
    timeout.tv_usec = 0;
    FD_ZERO(&set);
    FD_SET(fd1, &set);
    FD_SET(fd2, &set);
    v = select (maxfd, &set, 0, 0, &timeout);
    if (v <= 0)
	return v;
    if (FD_ISSET (fd1, &set))
	return 1;
    if (FD_ISSET (fd2, &set))
	return 2;
    return -1;
}

static int
get_line (int sock, char *buf, int buf_len)
{
    int i, Status;
    char c;

    for (i = 0; i < buf_len; i++, buf++) {
	if (read(sock, buf, sizeof(char)) <= 0)
	    return 0;
	if (ftpfs_debug_server_dialog) {
	    fwrite (buf, 1, 1, ftpfs_logfile);
	    fflush (ftpfs_logfile);
	}
	if (*buf == '\n') {
	    *buf = 0;
	    return 1;
	}
    }
    *buf = 0;
    while ((Status = read(sock, &c, sizeof(c))) > 0) {
	if (ftpfs_debug_server_dialog) {
	    fwrite (&c, 1, 1, ftpfs_logfile);
	    fflush (ftpfs_logfile);
	}
	if (c == '\n')
	    return 1;
    }
    return 0;
}

/* Returns a reply code, check /usr/include/arpa/ftp.h for possible values */
static int getreply (int sock, char *string_buf, int string_len)
{
    char answer[1024];
    int i;

    for (;;) {
	if (!get_line(sock, answer, sizeof(answer))) {
	    if (string_buf)
		*string_buf = 0;
	    code = 421;
	    return 4;
	}
	switch(sscanf(answer, "%d", &code)) {
	    case 0:
		if (string_buf) {
		    strncpy(string_buf, answer, string_len - 1);
		    *(string_buf + string_len - 1) = 0;
		}
		code = 500;
		return 5;
	    case 1:
		if (answer[3] == '-') {
		    while (1) {
			if (!get_line(sock, answer, sizeof(answer))) {
			    if (string_buf)
				*string_buf = 0;
			    code = 421;
			    return 4;
			}
			if ((sscanf(answer, "%d", &i) > 0) &&
			    (code == i) && (answer[3] == ' '))
			    break;
		    }
		}
		if (string_buf) {
		    strncpy(string_buf, answer, string_len - 1);
		    *(string_buf + string_len - 1) = 0;
		}
		return code / 100;
	}
    }
}

static void
ftpentry_destructor (void *data)
{
    struct ftpentry *fe = data;

    fe->count--;
    if ((fe->tmp_reget == 1 && fe->local_filename)) {
	unlink (fe->local_filename);
	fe->tmp_reget = 0;
    }

    if (fe->count > 0)
	return;
    free(fe->name);
    if (fe->linkname)
	free(fe->linkname);
    if (fe->local_filename) {
	if (fe->local_is_temp) {
	    if (!fe->local_stat.st_mtime)
		unlink(fe->local_filename);
	    else {
		struct stat sb;

		if (stat (fe->local_filename, &sb) >=0 &&
		    fe->local_stat.st_mtime == sb.st_mtime)
		    unlink (fe->local_filename); /* Delete only if it hasn't changed */
	    }
	}
	free(fe->local_filename);
	fe->local_filename = 0;
    }
    if (fe->remote_filename)
	free(fe->remote_filename);
    if (fe->l_stat)
	free(fe->l_stat);
    free(fe);
}

static void
ftpfs_dir_destructor(void *data)
{
    struct ftpfs_dir *fd = data;

    fd->count--;
    if (fd->count > 0)
	return;
    free(fd->remote_path);
    linklist_destroy(fd->file_list, ftpentry_destructor);
    free(fd);
}

static int command (struct ftpfs_connection *bucket, int wait_reply,
		    char *fmt, ...)
{
    va_list ap;
    char buf[2048]; /* FIXME: buffer exceed ?? */
    int n, Status;
    int sock = qsock (bucket);

    va_start (ap, fmt);
    vsprintf (buf, fmt, ap);
    va_end (ap);
    n = strlen(buf);
    buf[n++] = '\r';
    buf[n++] = '\n';
    buf[n] = 0;

    if (ftpfs_debug_server_dialog) {
	if (strncmp (buf, "PASS ", 5) == 0) {
	    char *tmp = "PASS <Password not logged>\r\n";
	    fwrite (tmp, strlen (tmp), 1, ftpfs_logfile);
	} else
	    fwrite (buf, strlen (buf), 1, ftpfs_logfile);
	fflush (ftpfs_logfile);
    }
    got_sigpipe = 0;
    enable_interrupt_key();
    Status = write(sock, buf, strlen(buf));
    if (Status < 0) {
	code = 421;
	if (errno == EPIPE) {
	    got_sigpipe = 1;
	}
	disable_interrupt_key();
	return TRANSIENT;
    }
    disable_interrupt_key();

    if (wait_reply)
	return getreply (sock, (wait_reply & WANT_STRING) ? reply_str : NULL, sizeof (reply_str)-1);
    return COMPLETE;
}

static void
ftpfs_connection_close (void *data)
{
    struct ftpfs_connection *bucket = data;

    if (qsock (bucket) != -1) {
	print_vfs_message ("ftpfs: Disconnecting from %s", qhost(bucket));
	command(bucket, NONE, "QUIT");
	close(qsock(bucket));
    }
}

static void
ftpfs_free_bucket (void *data)
{
    struct ftpfs_connection *bucket = data;

    free(qhost(bucket));
    free(quser(bucket));
    if (qcdir(bucket))
	free(qcdir(bucket));
    if (qhome(bucket))
	free(qhome(bucket));
    if (qupdir(bucket))
	free(qupdir(bucket));
    if (bucket->password)
	wipe_password (bucket->password);
    linklist_destroy(qdcache(bucket), ftpfs_dir_destructor);
    free(bucket);
}

static void
ftpfs_connection_destructor(void *data)
{
    ftpfs_connection_close (data);
    ftpfs_free_bucket (data);
}

/* some defines only used by changetype */
/* These two are valid values for the second parameter */
#define TYPE_ASCII    0
#define TYPE_BINARY   1

/* This one is only used to initialize bucket->isbinary, don't use it as
   second parameter to changetype. */
#define TYPE_UNKNOWN -1

static int
changetype (struct ftpfs_connection *bucket, int binary)
{
    if (binary != bucket->isbinary) {
	if (command (bucket, WAIT_REPLY, "TYPE %c", binary ? 'I' : 'A') != COMPLETE) {
	    ftpfserrno = EIO;
	    return -1;
	}
	bucket->isbinary = binary;
    }
    return binary;
}

static inline void
flush_all_directory(struct ftpfs_connection *bucket)
{
    linklist_delete_all(qdcache(bucket), ftpfs_dir_destructor);

}

/* This routine logs the user in */
static int
login_server (struct ftpfs_connection *bucket, char *netrcpass)
{
    char *pass;
    char *op;
    char *name;                 /* login user name */
    int  anon = 0;

    bucket->isbinary = TYPE_UNKNOWN;
    if (netrcpass)
	op = strdup (netrcpass);
    else {
	if (!strcmp (quser(bucket), "anonymous") ||
	    !strcmp (quser(bucket), "ftp")) {
	    op = strdup(ftpfs_anonymous_passwd);
	    anon = 1;
	 } else {
	    char *p;

	    if (!bucket->password) {
		p = xasprintf (" FTP: Password required for %s ",
					quser(bucket));
		op = input_dialog (p, "Password:", "");
		free (p);
		if (op == NULL) {
		    ftpfserrno = EPERM;
		    return 0;
		}
		bucket->password = strdup (op);
	    } else
		op = strdup (bucket->password);
	}
    }

    if (!anon || ftpfs_debug_server_dialog)
	pass = strdup (op);
    else
	pass = xasprintf ("-%s", op);
    wipe_password (op);


    /* Proxy server accepts: username@host-we-want-to-connect*/
    if (qproxy (bucket)) {
	name = xasprintf ("%s@%s", quser(bucket),
		qhost(bucket)[0] == '!' ? qhost(bucket)+1 : qhost(bucket));
    } else
	name = strdup (quser (bucket));

    if (getreply (qsock(bucket), NULL, 0) == COMPLETE) {
	print_vfs_message("ftpfs: sending login name");
	code = command (bucket, WAIT_REPLY, "USER %s", name);

	switch (code) {
	case CONTINUE:
	    print_vfs_message("ftpfs: sending user password");
	    if (command (bucket, WAIT_REPLY, "PASS %s", pass) != COMPLETE)
		break;

	case COMPLETE:
	    print_vfs_message("ftpfs: logged in");
	    wipe_password (pass);
	    free (name);
	    return 1;

	default:
	    bucket->failed_on_login = 1;
	    /* ftpfserrno = E; */
	    if (bucket->password)
		wipe_password (bucket->password);
	    bucket->password = 0;

	    /* This matches the end of the code below, just to make it
	     * obvious to the optimizer
	     */
	    wipe_password (pass);
	    free (name);
	    return 0;
	}
    }
    print_vfs_message ("ftpfs: Login incorrect for user %s ", quser(bucket));
    ftpfserrno = EPERM;
    wipe_password (pass);
    free (name);
    return 0;
}

static struct no_proxy_entry {
    char  *domain;
    void  *next;
} *no_proxy;

static void
load_no_proxy_list ()
{
    /* FixMe: shouldn't be hardcoded!!! */
    char                s[258]; /* provide for 256 characters and nl */
    struct no_proxy_entry *np, *current;
    FILE        *npf;
    int         c;
    char        *p;
    char        *mc_file;
    static int  loaded;

    if (loaded)
	return;

    mc_file = concat_mcpath_file ("mc.no_proxy");
    if (exist_file (mc_file) &&
	(npf = fopen (mc_file, "r"))) {
	while (fgets (s, 258, npf) || !(feof (npf) || ferror (npf))) {
	    if (!(p = strchr (s, '\n'))) {      /* skip bogus entries */
		while ((c = getc (npf)) != EOF && c != '\n')
		    ;
		continue;
	    }

	    if (p == s)
		continue;

	    *p = '\0';
	    p = x_malloc (strlen (s), "load_no_proxy_list:1");
	    np = x_malloc (sizeof (*np), "load_no_proxy_list:2");
	    strcpy (p, s);
	    np->domain = p;
	    np->next   = 0;
	    if (no_proxy)
		current->next = np;
	    else
		no_proxy = np;
	    current = np;
	}

	fclose (npf);
	loaded = 1;
    }
    free (mc_file);
}

static int
ftpfs_check_proxy (char *host)
{
    struct no_proxy_entry       *npe;

    if (!ftpfs_proxy_host || !*ftpfs_proxy_host || !host || !*host)
	return 0;               /* sanity check */

    if (*host == '!')
	return 1;

    if (!ftpfs_always_use_proxy)
	return 0;

    if (!strchr (host, '.'))
	return 0;

    load_no_proxy_list ();
    for (npe = no_proxy; npe; npe=npe->next) {
	char    *domain = npe->domain;

	if (domain[0] == '.') {
	    int         ld = strlen (domain);
	    int         lh = strlen (host);

	    while (ld && lh && host[lh - 1] == domain[ld - 1]) {
		ld--;
		lh--;
	    }

	    if (!ld)
		return 0;
	} else
	    if (!strcasecmp (host, domain))
		return 0;
    }

    return 1;
}

static void
ftpfs_get_proxy_host_and_port (char *proxy, char **host, int *port)
{
    char *user, *pass, *dir;

    dir = get_host_and_username(proxy, host, &user, port, 21, 1,
				 &pass);
    free(user);
    if (pass)
	wipe_password (pass);
    if (dir)
	free(dir);
}

static int
ftpfs_open_socket(struct ftpfs_connection *bucket)
{
    struct   sockaddr_in server_address;
    int      my_socket;
    char     *host;
    int      port;
    int      free_host = 0;

    /* Use a proxy host? */
    host = qhost(bucket);

    if (!host || !*host) {
	print_vfs_message ("ftpfs: Invalid host name.");
	ftpfserrno = EINVAL;
	return -1;
    }

    /* Hosts to connect to that start with a ! should use proxy */
    if (qproxy(bucket)) {
	ftpfs_get_proxy_host_and_port (ftpfs_proxy_host, &host, &port);
	free_host = 1;
    }

    /* Get host address */
    if(bb_gethostbyname4(BB_GHBN_SOCKADDR_IN | BB_GHBN_SILENT, host, &server_address) == NULL) {
	    print_vfs_message("ftpfs: Invalid host address.");
	    ftpfserrno = EINVAL;
	    if (free_host)
		free (host);
	    return -1;
	}
    }

#define THE_PORT qproxy(bucket) ? port : qport (bucket)

    server_address.sin_port = htons (THE_PORT);

    /* Connect */
    if ((my_socket = bb_xsocket (BB_XSOCKET_FLG_NO_WARN)) < 0) {
	ftpfserrno = errno;
	if (free_host)
	    free (host);
	return -1;
    }
    print_vfs_message("ftpfs: making connection to %s", host);
    if (free_host)
	free (host);

    enable_interrupt_key(); /* clear the interrupt flag */

    if (connect (my_socket, (struct sockaddr *) &server_address,
	     sizeof (server_address)) < 0) {
	ftpfserrno = errno;
	if (errno == EINTR && got_interrupt())
	    print_vfs_message("ftpfs: connection interrupted by user");
	else
	    print_vfs_message("ftpfs: connection to server failed: %m");
	disable_interrupt_key();
	close (my_socket);
	return -1;
    }
    disable_interrupt_key();
    return my_socket;
}

static struct ftpfs_connection *
open_command_connection (char *host, char *user, int port, char *netrcpass)
{
    struct ftpfs_connection *bucket;
    int retry_seconds, count_down;

    bucket = x_malloc(sizeof(struct ftpfs_connection),
		     "struct ftpfs_connection");

    if (bucket == NULL) {
	ftpfserrno = ENOMEM;
	return NULL;
    }
    qhost(bucket) = strdup (host);
    quser(bucket) = strdup (user);
    qcdir(bucket) = NULL;
    qport(bucket) = port;
    qlock(bucket) = 0;
    qhome(bucket) = NULL;
    qproxy(bucket)= 0;
    qupdir(bucket)= 0;
    qdcache(bucket)=0;
    bucket->__inode_counter = 0;
    bucket->lock = 0;
    bucket->use_proxy = ftpfs_check_proxy (host);
    bucket->password = 0;
    bucket->use_passive_connection = ftpfs_use_passive_connections | source_route;
    bucket->use_source_route = source_route;
    bucket->strict_rfc959_list_cmd = !ftpfs_use_unix_list_options;
    bucket->isbinary = TYPE_UNKNOWN;

    /* We do not want to use the passive if we are using proxies */
    if (bucket->use_proxy)
	bucket->use_passive_connection = 0;

    if ((qdcache(bucket) = linklist_init()) == NULL) {
	ftpfserrno = ENOMEM;
	free (qhost(bucket));
	free (quser(bucket));
	free (bucket);
	return NULL;
    }

    retry_seconds = 0;
    do {
	bucket->failed_on_login = 0;

	qsock(bucket) = ftpfs_open_socket(bucket);
	if (qsock(bucket) == -1)  {
	    ftpfs_free_bucket (bucket);
	    return NULL;
	}

	if (login_server(bucket, netrcpass)) {
	    /* Logged in, no need to retry the connection */
	    break;
	} else {
	    if (bucket->failed_on_login) {
		/* Close only the socket descriptor */
		close (qsock (bucket));
	    } else {
		ftpfs_connection_destructor (bucket);
		return NULL;
	    }
	    if (ftpfs_retry_seconds) {
		retry_seconds = ftpfs_retry_seconds;
		enable_interrupt_key ();
		for (count_down = retry_seconds; count_down; count_down--) {
		    print_vfs_message ("Waiting to retry... %d (Control-C to cancel)\n", count_down);
		    sleep (1);
		    if (got_interrupt ()) {
			/* ftpfserrno = E; */
			disable_interrupt_key ();
			ftpfs_free_bucket (bucket);
			return NULL;
		    }
		}
		disable_interrupt_key ();
	    }
	}
    } while (retry_seconds);

    qhome(bucket) = ftpfs_get_current_directory (bucket);
    if (!qhome(bucket))
	qhome(bucket) = strdup ("/");
    if (last_current_dir) {
	if (last_current_dir [strlen (last_current_dir) - 1] == '/')
	    qupdir(bucket) = strdup (last_current_dir);
	else
	    qupdir(bucket) = xasprintf ("%s/", last_current_dir);
    } else
	qupdir(bucket) = strdup ("/");
    return bucket;
}

static int
is_connection_closed(struct ftpfs_connection *bucket)
{
    fd_set rset;
    struct timeval t;

    if (got_sigpipe) {
	return 1;
    }
    t.tv_sec = 0;
    t.tv_usec = 0;
    FD_ZERO(&rset);
    FD_SET(qsock(bucket), &rset);
    while (1) {
	if (select(qsock(bucket) + 1, &rset, NULL, NULL, &t) < 0)
	    if (errno != EINTR)
		return 1;
	return 0;
    }
}

/* This routine keeps track of open connections */
/* Returns a connected socket to host */
static struct ftpfs_connection *
ftpfs_open_link (char *host, char *user, int port, char *netrcpass)
{
    int sock;
    struct ftpfs_connection *bucket;
    struct linklist *lptr;

    for (lptr = ftpfs_connections_list->next;
	 lptr != ftpfs_connections_list; lptr = lptr->next) {
	bucket = lptr->data;
	if ((strcmp (host, qhost(bucket)) == 0) &&
	    (strcmp (user, quser(bucket)) == 0) &&
	    (port == qport(bucket))) {

	    /* check the connection is closed or not, just hack */
	    if (is_connection_closed(bucket)) {
		flush_all_directory(bucket);
		sock = ftpfs_open_socket(bucket);
		if (sock != -1) {
		    close(qsock(bucket));
		    qsock(bucket) = sock;
		    if (login_server(bucket, netrcpass))
			return bucket;
		}

		/* connection refused */
		lptr->prev->next = lptr->next;
		lptr->next->prev = lptr->prev;
		ftpfs_connection_destructor(bucket);
		return NULL;
	    }
	    return bucket;
	}
    }
    bucket = open_command_connection(host, user, port, netrcpass);
    if (bucket == NULL)
	return NULL;
    if (!linklist_insert(ftpfs_connections_list, bucket)) {
	ftpfserrno = ENOMEM;
	ftpfs_connection_destructor(bucket);
	return NULL;
    }
    return bucket;
}

/* The returned directory should always contain a trailing slash */
static char *ftpfs_get_current_directory(struct ftpfs_connection *bucket)
{
    char buf[4096], *bufp, *bufq;

    if (command(bucket, NONE, "PWD") == COMPLETE &&
	getreply(qsock(bucket), buf, sizeof(buf)) == COMPLETE) {
	bufp = NULL;
	for (bufq = buf; *bufq; bufq++)
	    if (*bufq == '"') {
		if (!bufp) {
		    bufp = bufq + 1;
		} else {
		    *bufq = 0;
		    if (*bufp) {
			if (*(bufq - 1) != '/') {
			    *bufq++ = '/';
			    *bufq = 0;
			}
			return strdup (bufp);
		    } else {
			ftpfserrno = EIO;
			return NULL;
		    }
		}
	    }
    }
    ftpfserrno = EIO;
    return NULL;
}

static void ftpfs_fill_names (void (*func)(char *))
{
    struct linklist *lptr;
    char   *path_name;
    struct ftpfs_connection *bucket;

    if (!ftpfs_connections_list)
	return;
    lptr = ftpfs_connections_list;
    do {
	if ((bucket = lptr->data) != 0) {

	    path_name = xasprintf ("ftp://%s@%s%s", quser (bucket),
				      qhost (bucket), qcdir(bucket));
	    (*func)(path_name);
	    free (path_name);
	}
	lptr = lptr->next;
    } while (lptr != ftpfs_connections_list);
}


/* Setup Passive ftp connection, we use it for source routed connections */
static int
setup_passive (int my_socket, struct ftpfs_connection *bucket, struct sockaddr_in *sa)
{
    int xa, xb, xc, xd, xe, xf;
    char n [6];
    char *c = reply_str;

    if (command (bucket, WAIT_REPLY | WANT_STRING, "PASV") != COMPLETE)
	return 0;

    /* Parse remote parameters */
    for (c = reply_str + 4; (*c) && (!isdigit (*c)); c++)
	;
    if (!*c)
	return 0;
    if (!isdigit (*c))
	return 0;
    if (sscanf (c, "%d,%d,%d,%d,%d,%d", &xa, &xb, &xc, &xd, &xe, &xf) != 6)
	return 0;
    n [0] = (unsigned char) xa;
    n [1] = (unsigned char) xb;
    n [2] = (unsigned char) xc;
    n [3] = (unsigned char) xd;
    n [4] = (unsigned char) xe;
    n [5] = (unsigned char) xf;

    memcpy (&(sa->sin_addr.s_addr), n, 4);
    memcpy (&(sa->sin_port), &n[4], 2);
    if (connect (my_socket, (struct sockaddr *) sa, sizeof (struct sockaddr_in)) < 0)
	return 0;
    return 1;
}

static int
initconn (struct ftpfs_connection *bucket)
{
    struct sockaddr_in data_addr;
    int data, len = sizeof(data_addr);
    struct protoent *pe;

    getsockname(qsock(bucket), (struct sockaddr *) &data_addr, &len);
    data_addr.sin_port = 0;

    pe = getprotobyname("tcp");
    if (pe == NULL) {
	ftpfserrno = EIO;
	return -1;
    }
    data = bb_xsocket (BB_XSOCKET_FLG_NO_WARN);
    if (data < 0) {
	ftpfserrno = EIO;
	return -1;
    }

    if (bucket->use_passive_connection) {
	if ((bucket->use_passive_connection = setup_passive (data, bucket, &data_addr)))
	    return data;

	bucket->use_source_route = 0;
	bucket->use_passive_connection = 0;
	print_vfs_message ("ftpfs: could not setup passive mode");
    }
    /* If passive setup fails, fallback to active connections */
    /* Active FTP connection */
    if (bind (data, (struct sockaddr *)&data_addr, len) < 0)
	goto error_return;
    getsockname(data, (struct sockaddr *) &data_addr, &len);
    if (listen (data, 1) < 0)
	goto error_return;
    {
	unsigned char *a = (unsigned char *)&data_addr.sin_addr;
	unsigned char *p = (unsigned char *)&data_addr.sin_port;

	if (command (bucket, WAIT_REPLY, "PORT %d,%d,%d,%d,%d,%d", a[0], a[1],
		     a[2], a[3], p[0], p[1]) != COMPLETE)
	    goto error_return;
    }
    return data;
error_return:
    close(data);
    ftpfserrno = EIO;
    return -1;
}

static int
open_data_connection (struct ftpfs_connection *bucket, char *cmd, char *remote,
		int isbinary)
{
    struct sockaddr_in from;
    int s, j, data, fromlen = sizeof(from);

    if ((s = initconn (bucket)) == -1)
	return -1;
    if (changetype (bucket, isbinary) == -1)
	return -1;
    if (do_reget > 0) {
	j = command (bucket, WAIT_REPLY, "REST %d", do_reget);
	do_reget = 0;
	if (j != CONTINUE)
	    return -1;
    }
    if (remote)
	j = command (bucket, WAIT_REPLY, "%s %s", cmd, remote);
    else
	j = command (bucket, WAIT_REPLY, "%s", cmd);
    if (j != PRELIM) {
	ftpfserrno = EPERM;
	return -1;
    }
    enable_interrupt_key();
    if (bucket->use_passive_connection)
	data = s;
    else {
	data = accept (s, (struct sockaddr *)&from, &fromlen);
	if (data < 0) {
	    ftpfserrno = errno;
	    close(s);
	    return -1;
	}
	close(s);
    }
    disable_interrupt_key();
    return data;
}

/* ftpfs_get_path:
 * makes BUCKET point to the connection bucket descriptor for PATH
 * returns a malloced string with the pathname relative to BUCKET.
 */
static char*
ftpfs_get_path (struct ftpfs_connection **bucket, char *path)
{
    char *user, *host, *remote_path, *pass;
    int port;

    /* An absolute path name, try to determine connection socket */
    if (strncmp (path, "ftp://", 6) == 0) {
	path += 6;

	if (!(remote_path = ftpfs_get_host_and_username (path, &host, &user,
	    &port, &pass))) {
	    ftpfserrno = ENOENT;
	    free (host);
	    free (user);
	    if (pass)
		wipe_password (pass);
	    return NULL;
	}
	if ((*bucket = ftpfs_open_link (host, user, port, pass)) == NULL) {
	    free (remote_path);
	    free (host);
	    free (user);
	    if (pass)
		wipe_password (pass);
	    return NULL;
	}
	free (host);
	free (user);
	if (pass)
	    wipe_password (pass);
	return remote_path;
    }
    /* never get here !!! */
    message_1s(1, " Error ", " Oops, you just hit a bug in the code ");
    return NULL;
}

static void
ftpfs_abort (struct ftpfs_connection *bucket, int dsock)
{
    static unsigned char ipbuf[3] = { IAC, IP, IAC };
    fd_set mask;
    char buf[1024];

    print_vfs_message("ftpfs: aborting transfer.");
    if (send(qsock(bucket), ipbuf, sizeof(ipbuf), MSG_OOB) != sizeof(ipbuf)) {
	print_vfs_message("ftpfs: abort error: %m");
	return;
    }

    if (command(bucket, NONE, "%cABOR", DM) != COMPLETE) {
	print_vfs_message ("ftpfs: abort failed");
	return;
    }
    if (dsock != -1) {
	FD_ZERO(&mask);
	FD_SET(dsock, &mask);
	if (select(dsock + 1, &mask, NULL, NULL, NULL) > 0)
	    while (read(dsock, buf, sizeof(buf)) > 0);
    }
    if ((getreply(qsock(bucket), NULL, 0) == TRANSIENT) && (code == 426))
	getreply(qsock(bucket), NULL, 0);
}

static void
resolve_symlink(struct ftpfs_connection *bucket, struct ftpfs_dir *dir)
{
    struct linklist *flist;
    struct ftpentry *fe, *fel;
    char tmp[MC_MAXPATHLEN];

    print_vfs_message("Resolving symlink...");

    dir->symlink_status = FTPFS_RESOLVED_SYMLINKS;
    for (flist = dir->file_list->next; flist != dir->file_list; flist = flist->next) {
	/* flist->data->l_stat is alread initialized with 0 */
	fel = flist->data;
	if (S_ISLNK(fel->s.st_mode)) {
	    if (fel->linkname[0] == '/') {
		if (strlen (fel->linkname) >= MC_MAXPATHLEN)
		    continue;
		strcpy (tmp, fel->linkname);
	    } else {
		if ((strlen (dir->remote_path) + strlen (fel->linkname)) >= MC_MAXPATHLEN)
		    continue;
		strcpy (tmp, dir->remote_path);
		if (tmp[1] != '\0')
		   strcat (tmp, "/");
		strcat (tmp + 1, fel->linkname);
	    }
	    for ( ;; ) {
		canonicalize_pathname (tmp);
		fe = _get_file_entry(bucket, tmp, 0, 0);
		if (fe) {
		    if (S_ISLNK (fe->s.st_mode) && fe->l_stat == 0) {
			/* Symlink points to link which isn't resolved, yet. */
			if (fe->linkname[0] == '/') {
			    if (strlen (fe->linkname) >= MC_MAXPATHLEN)
				break;
			    strcpy (tmp, fe->linkname);
			} else {
			    /* at this point tmp looks always like this
			       /directory/filename, i.e. no need to check
				strrchr's return value */
			    *(strrchr (tmp, '/') + 1) = '\0'; /* dirname */
			    if ((strlen (tmp) + strlen (fe->linkname)) >= MC_MAXPATHLEN)
				break;
			    strcat (tmp, fe->linkname);
			}
			continue;
		    } else {
			fel->l_stat = x_malloc(sizeof(struct stat),
					 "resolve_symlink: struct stat");
			if ( S_ISLNK (fe->s.st_mode))
			    *fel->l_stat = *fe->l_stat;
			else
			    *fel->l_stat = fe->s;
			(*fel->l_stat).st_ino = bucket->__inode_counter++;
		    }
		}
		break;
	    }
	}
    }
}

static int
get_line_interruptible (char *buffer, int size, int fd)
{
    int n;
    int i = 0;

    for (i = 0; i < size-1; i++) {
	n = read (fd, buffer+i, 1);
	if (n == -1 && errno == EINTR) {
	    buffer [i] = 0;
	    return EINTR;
	}
	if (n == 0) {
	    buffer [i] = 0;
	    return 0;
	}
	if (buffer [i] == '\n') {
	    buffer [i] = 0;
	    return 1;
	}
    }
    buffer [size-1] = 0;
    return 0;
}

/* Return true if path is the same directoy as the one we are on now */
static int
ftpfs_same_dir (char *path, struct ftpfs_connection *bucket)
{
    if (!qcdir (bucket))
	return 0;
    if (strcmp (path, qcdir (bucket)) == 0)
	return 1;
    return 0;
}

static struct ftpfs_dir *
retrieve_dir(struct ftpfs_connection *bucket, char *remote_path, int resolve_symlinks)
{
    int sock, has_symlinks;
    struct linklist *file_list, *p;
    struct ftpentry *fe;
    char buffer[8192];
    struct ftpfs_dir *dcache;
    int got_intr = 0;
    int has_spaces = (strchr (remote_path, ' ') != NULL);

    canonicalize_pathname (remote_path);
    for (p = qdcache(bucket)->next;p != qdcache(bucket);
	 p = p->next) {
	dcache = p->data;
	if (strcmp(dcache->remote_path, remote_path) == 0) {
	    struct timeval tim;

	    gettimeofday(&tim, NULL);
	    if ((tim.tv_sec < dcache->timestamp.tv_sec) && !force_expiration) {
		if (resolve_symlinks && dcache->symlink_status == FTPFS_UNRESOLVED_SYMLINKS)
		    resolve_symlink(bucket, dcache);
		return dcache;
	    } else {
		force_expiration = 0;
		p->next->prev = p->prev;
		p->prev->next = p->next;
		ftpfs_dir_destructor(dcache);
		free (p);
		break;
	    }
	}
    }

    has_symlinks = 0;
    if (bucket->strict_rfc959_list_cmd)
	print_vfs_message("ftpfs: Reading FTP directory... (don't use UNIX ls options)");
    else
	print_vfs_message("ftpfs: Reading FTP directory...");
    if (has_spaces || bucket->strict_rfc959_list_cmd)
	if (__ftpfs_chdir(bucket, remote_path) != COMPLETE) {
	    ftpfserrno = ENOENT;
	    print_vfs_message("ftpfs: CWD failed.");
	    return NULL;
	}

    file_list = linklist_init();
    if (file_list == NULL) {
	ftpfserrno = ENOMEM;
	print_vfs_message("ftpfs: couldn't get a file listing");
	return NULL;
    }
    dcache = x_malloc(sizeof(struct ftpfs_dir),
		     "struct ftpfs_dir");
    if (dcache == NULL) {
	ftpfserrno = ENOMEM;
	linklist_destroy(file_list, NULL);
	print_vfs_message("ftpfs: FAIL");
	return NULL;
    }
    gettimeofday(&dcache->timestamp, NULL);
    dcache->timestamp.tv_sec += ftpfs_directory_timeout;
    dcache->file_list = file_list;
    dcache->remote_path = strdup(remote_path);
    dcache->count = 1;
    dcache->symlink_status = FTPFS_NO_SYMLINKS;

    if (bucket->strict_rfc959_list_cmd == 1)
	sock = open_data_connection (bucket, "LIST", 0, TYPE_ASCII);
    else if (has_spaces)
	sock = open_data_connection (bucket, "LIST -la", ".", TYPE_ASCII);
    else {
	char *path = xasprintf ("%s/.", remote_path);

	sock = open_data_connection (bucket, "LIST -la", path, TYPE_ASCII);
	free (path);
    }

    if (sock == -1)
	goto error_3;

#define close_this_sock(x,y) close (y)

    /* Clear the interrupt flag */
    enable_interrupt_key ();

    errno = 0;
    while ((got_intr = get_line_interruptible (buffer, sizeof (buffer), sock)) != EINTR) {
	int eof = got_intr == 0;

	if (ftpfs_debug_server_dialog) {
	    fputs (buffer, ftpfs_logfile);
	    fputs ("\n", ftpfs_logfile);
	    fflush (ftpfs_logfile);
	}
	if (buffer [0] == 0 && eof)
	    break;
	fe = x_malloc(sizeof(struct ftpentry), "struct ftpentry");
	fe->freshly_created = 0;
	fe->tmp_reget = 0;
	if (fe == NULL) {
	    ftpfserrno = ENOMEM;
	    goto error_1;
	}
	if (parse_ls_lga (buffer, &fe->s, &fe->name, &fe->linkname)) {
	    fe->count = 1;
	    fe->local_filename = fe->remote_filename = NULL;
	    fe->l_stat = NULL;
	    fe->bucket = bucket;
	    (fe->s).st_ino = bucket->__inode_counter++;
	    if (S_ISLNK(fe->s.st_mode))
		has_symlinks = 1;

	    if (!linklist_insert(file_list, fe)) {
		free(fe);
		ftpfserrno = ENOMEM;
		goto error_1;
	    }
	}
	else
	    free(fe);
	if (eof)
	    break;
    }
    if (got_intr) {
	disable_interrupt_key();
	print_vfs_message("ftpfs: reading FTP directory interrupt by user");
	ftpfs_abort(bucket, sock);
	close_this_sock(fp, sock);
	ftpfserrno = EINTR;
	goto error_3;
    }
    close_this_sock(fp, sock);
    disable_interrupt_key();
    if (getreply (qsock (bucket), NULL, 0) != COMPLETE) {
	ftpfserrno = EIO;
	goto error_3;
    }

    if (file_list->next == file_list) {
	if (bucket->__inode_counter == 0 && !bucket->strict_rfc959_list_cmd) {
	    /* It's our first attempt to get a directory listing from this
	       server (UNIX style LIST command) */
	    bucket->strict_rfc959_list_cmd = 1;
	    free(dcache->remote_path);
	    free(dcache);
	    linklist_destroy(file_list, ftpentry_destructor);
	    return retrieve_dir (bucket, remote_path, resolve_symlinks);
	}
	ftpfserrno = EACCES;
	goto error_3;
    }

    if (!linklist_insert(qdcache(bucket), dcache)) {
	ftpfserrno = ENOMEM;
	goto error_3;
    }

    if (has_symlinks) {
	if (resolve_symlinks)
	    resolve_symlink(bucket, dcache);
	else
	   dcache->symlink_status = FTPFS_UNRESOLVED_SYMLINKS;
    }

    print_vfs_message("ftpfs: got listing");
    return dcache;
error_1:
    disable_interrupt_key();
    close_this_sock(fp, sock);
    getreply(qsock(bucket), NULL, 0);
error_3:
    free(dcache->remote_path);
    free(dcache);
    linklist_destroy(file_list, ftpentry_destructor);
    print_vfs_message("ftpfs: failed");
    return NULL;
}

static int
store_file(struct ftpentry *fe)
{
    int local_handle, sock, n, total;
    struct linger li;
    char buffer[8192];
    struct stat s;

    local_handle = open(fe->local_filename, O_RDONLY);
    unlink (fe->local_filename);
    if (local_handle == -1) {
	ftpfserrno = EIO;
	return 0;
    }
    fstat(local_handle, &s);
    sock = open_data_connection(fe->bucket, "STOR", fe->remote_filename, TYPE_BINARY);
    if (sock < 0) {
	close(local_handle);
	return 0;
    }
    li.l_onoff = 1;
    li.l_linger = 120;
    setsockopt(sock, SOL_SOCKET, SO_LINGER, (char *) &li, sizeof(li));
    total = 0;

    enable_interrupt_key();
    while (1) {
	while ((n = read(local_handle, buffer, sizeof(buffer))) < 0) {
	    if (errno == EINTR) {
		if (got_interrupt()) {
		    ftpfserrno = EINTR;
		    goto error_return;
		}
		else
		    continue;
	    }
	    ftpfserrno = errno;
	    goto error_return;
	}
	if (n == 0)
	    break;
	while (write(sock, buffer, n) < 0) {
	    if (errno == EINTR) {
		if (got_interrupt()) {
		    ftpfserrno = EINTR;
		    goto error_return;
		}
		else
		    continue;
	    }
	    ftpfserrno = errno;
	    goto error_return;
	}
	total += n;
	print_vfs_message("ftpfs: storing file %d (%d)",
			  total, s.st_size);
    }
    disable_interrupt_key();
    close(sock);
    close(local_handle);
    if (getreply (qsock (fe->bucket), NULL, 0) != COMPLETE) {
	ftpfserrno = EIO;
	return 0;
    }
    return 1;
error_return:
    disable_interrupt_key();
    close(sock);
    close(local_handle);
    getreply(qsock(fe->bucket), NULL, 0);
    return 0;
}

/* These variables are for the _ctl routine */
static char *localname;
static struct ftpentry *remoteent;
static int remotetotal;
static int transfer_started;
static char *remotebuffer;
static int isremotecopy;
static int remotelocal_handle, remotesock, remoten, remotestat_size;

static int retrieve_file_start(struct ftpentry *fe)
{
    if (fe->local_filename == NULL) {
	ftpfserrno = ENOMEM;
	return 0;
    }
    remotesock = open_data_connection(fe->bucket, "RETR", fe->remote_filename, TYPE_BINARY);
    if (remotesock == -1) {
	ftpfserrno = EACCES;
	free (fe->local_filename);
	fe->local_filename = NULL;
	return 0;
    }
    remotetotal = 0;
    remoteent = fe;
    return 1;
}

static int retrieve_file_start2(struct ftpentry *fe)
{
    remotelocal_handle = open(fe->local_filename, O_RDWR | O_CREAT | O_TRUNC, 0600);
    if (remotelocal_handle == -1) {
	ftpfserrno = EIO;
	free(fe->local_filename);
	fe->local_filename = NULL;
	fe->local_is_temp = 1;
	close(remotesock);
	return 0;
    }
    remotestat_size = fe->s.st_size;
    remotebuffer = x_malloc (8192, "");
    return 1;
}

void ftpfs_flushdir ()
{
	    force_expiration = 1;
}

static int ftpfs_ctl (void *data, int ctlop, int arg)
{
    int n = 0;

    switch (ctlop) {
	case MCCTL_ISREMOTECOPY:
	    return isremotecopy;

	case MCCTL_REMOTECOPYCHUNK:
	    if (!transfer_started) {
		if (!retrieve_file_start2 (remoteent))
		    return MCERR_TARGETOPEN;
		 else
		    transfer_started = 1;
	    }
	    enable_interrupt_key ();
	    if (!remoten) {
		int v = select_on_two (remotesock, 0);

		if (((v < 0) && (errno == EINTR)) || v == 0) {
		    disable_interrupt_key ();
		    return MCERR_DATA_ON_STDIN;
		}

		if ((n = read(remotesock, remotebuffer, 8192)) < 0) {
		    disable_interrupt_key ();
		    if (errno == EINTR)
			return MCERR_DATA_ON_STDIN;
		    else
			return MCERR_READ;
		}
		if (n == 0) {
		    if (getreply (qsock (remoteent->bucket), NULL, 0) != COMPLETE) {
			ftpfserrno = EIO;
		    }
		    close(remotelocal_handle);
		    close(remotesock);
		    if (localname) {
			free (localname);
			localname = NULL;
		    }
		    disable_interrupt_key ();
		    transfer_started = 0;
		    return MCERR_FINISH;
		}
		disable_interrupt_key ();
		remotetotal += n;
		remoten = n;
	    } else
		n = remoten;
	    if (write(remotelocal_handle, remotebuffer, remoten) < 0)
		return MCERR_WRITE;
	    remoten = 0;
	    return n;

	    /* We get this message if the transfer was aborted */
	case MCCTL_FINISHREMOTE:
	    if (localname) {
		free (localname);
		localname = NULL;
	    }
	    if (!arg) { /* OK */
		if (stat (remoteent->local_filename, &remoteent->local_stat) < 0)
		    remoteent->local_stat.st_mtime = 0;
	    } else
		remoteent->local_stat.st_mtime = 0;
	    transfer_started = 0;
	    ftpfs_abort (remoteent->bucket, remotesock);
	    ftpfserrno = EINTR;
	    return 0;
    }
    return 0;
}

static int ftpfs_setctl (char *path, int ctlop, char *arg)
{
    switch (ctlop) {
	case MCCTL_SETREMOTECOPY: if (localname) free (localname);
	    localname = strdup (vfs_canon (arg));
	    return 1;
	default:
	    return 0;
    }
}

static int retrieve_file(struct ftpentry *fe)
{
    int total, tmp_reget = do_reget;
    char buffer[8192];
    int local_handle, sock, n;

    if (fe->local_filename)
	return 1;
    fe->local_stat.st_mtime = 0;
    fe->local_filename = mc_mkstemp(&local_handle);
    fe->local_is_temp = 1;
    if (fe->local_filename == NULL) {
	ftpfserrno = ENOMEM;
	return 0;
    }
    if (local_handle == -1) {
	ftpfserrno = EIO;
	free(fe->local_filename);
	fe->local_filename = NULL;
	return 0;
    }
    sock = open_data_connection(fe->bucket, "RETR", fe->remote_filename, TYPE_BINARY);
    if (sock == -1) {
	ftpfserrno = EACCES;
	goto error_3;
    }

    /* Clear the interrupt status */
    enable_interrupt_key ();
    total = 0;
    if (tmp_reget > 0)
	total = tmp_reget;
    else
	total = 0;

    while (1) {
	int stat_size = fe->s.st_size;
	while ((n = read(sock, buffer, sizeof(buffer))) < 0) {
	    if (errno == EINTR) {
		if (got_interrupt ()) {
		    disable_interrupt_key();
		    ftpfs_abort(fe->bucket, sock);
		    ftpfserrno = EINTR;
		    goto error_2;
		}
		else
		    continue;
	    }
	    ftpfserrno = errno;
	    disable_interrupt_key();
	    goto error_1;
	}
	if (n == 0)
	    break;
	total += n;
	if (stat_size == 0)
	    print_vfs_message ("ftpfs: Getting file: %ld bytes transfered",
			       total);
	else
	    print_vfs_message ("ftpfs: Getting file: %3d%% (%ld bytes transfered)",
			       total*100/stat_size, total);
	while (write(local_handle, buffer, n) < 0) {
	    if (errno == EINTR) {
		if (got_interrupt()) {
		    ftpfs_abort(fe->bucket, sock);
		    ftpfserrno = EINTR;
		    goto error_2;
		}
		else
		    continue;
	    }
	    ftpfserrno = errno;
	    goto error_1;
	}
    }
    close(local_handle);
    close(sock);
    if (getreply (qsock (fe->bucket), NULL, 0) != COMPLETE) {
	ftpfserrno = EIO;
	goto error_2;
    }

    if (stat (fe->local_filename, &fe->local_stat) < 0)
	fe->local_stat.st_mtime = 0;

    if (tmp_reget > 0)
	fe->tmp_reget = 1;

    return 1;
error_1:
    getreply(qsock(fe->bucket), NULL, 0);
error_2:
    close(sock);
error_3:
    disable_interrupt_key ();
    close(local_handle);
    unlink(fe->local_filename);
    free(fe->local_filename);
    fe->local_filename = NULL;
    return 0;
}

static struct ftpentry *
_get_file_entry(struct ftpfs_connection *bucket, char *file_name,
		int op, int flags)
{
    char *p, q;
    struct ftpentry *ent;
    struct linklist *file_list, *lptr;
    struct ftpfs_dir *dcache;
    struct stat sb;

    p = strrchr(file_name, '/');
    q = *p;
    *p = '\0';
    dcache = retrieve_dir(bucket, *file_name ? file_name : "/", op & FTPFS_DO_RESOLVE_SYMLINK);
    if (dcache == NULL)
	return NULL;
    file_list = dcache->file_list;
    *p++ = q;
    if (!*p)
	p = ".";
    for (lptr = file_list->next; lptr != file_list; lptr = lptr->next) {
	ent = lptr->data;
	if (strcmp(p, ent->name) == 0) {
	    if (S_ISLNK(ent->s.st_mode) && (op & FTPFS_DO_RESOLVE_SYMLINK)) {
		if (ent->l_stat == NULL) {
		    ftpfserrno = ENOENT;
		    return NULL;
		}
		if (S_ISLNK(ent->l_stat->st_mode)) {
		    ftpfserrno = ELOOP;
		    return NULL;
		}
	    }
	    if (ent && (op & FTPFS_OPEN)) {
		mode_t fmode;

		fmode = S_ISLNK(ent->s.st_mode)
		    ? ent->l_stat->st_mode
		    : ent->s.st_mode;
		if (S_ISDIR(fmode)) {
		    ftpfserrno = EISDIR;
		    return NULL;
		}
		if (!S_ISREG(fmode)) {
		    ftpfserrno = EPERM;
		    return NULL;
		}
		if ((flags & O_EXCL) && (flags & O_CREAT)) {
		    ftpfserrno = EEXIST;
		    return NULL;
		}
		if (ent->remote_filename == NULL) {
		    ent->remote_filename = strdup(file_name);
		    if (ent->remote_filename == NULL) {
			ftpfserrno = ENOMEM;
			return NULL;
		    }
		}
		if (ent->local_filename == NULL ||
		    !ent->local_stat.st_mtime ||
		    stat (ent->local_filename, &sb) < 0 ||
		    sb.st_mtime != ent->local_stat.st_mtime) {
		    int handle;

		    if (ent->local_filename) {
			free (ent->local_filename);
			ent->local_filename = 0;
		    }
		    if (flags & O_TRUNC) {
			ent->local_filename = mc_mkstemp(&handle);
			if (ent->local_filename == NULL) {
			    ftpfserrno = ENOMEM;
			    return NULL;
			}
			if (handle < 0) {
			    ftpfserrno = EIO;
			    return NULL;
			}
			close(handle);
			if (stat (ent->local_filename, &ent->local_stat) < 0)
			    ent->local_stat.st_mtime = 0;
		    }
		    else {
			if (localname != NULL) {
			    isremotecopy = 1;
			    ent->local_is_temp = 0;
			    ent->local_stat.st_mtime = 0;
			    ent->local_filename = strdup (localname);
			    if (!retrieve_file_start (ent)) {
				isremotecopy = 0;
				return NULL;
			    }
			    return ent;
			}
			if (!retrieve_file(ent))
			    return NULL;
		    }
		}
		else if (flags & O_TRUNC) {
			truncate(ent->local_filename, 0);
		}
	    }
	    return ent;
	}
    }
    if ((op & FTPFS_OPEN) && (flags & O_CREAT)) {
	int handle;

	ent = x_malloc(sizeof(struct ftpentry), "struct ftpentry");
	ent->freshly_created = 0;
	ent->tmp_reget = 0;
	if (ent == NULL) {
	    ftpfserrno = ENOMEM;
	    return NULL;
	}
	ent->count = 1;
	ent->linkname = NULL;
	ent->l_stat = NULL;
	ent->bucket = bucket;
	ent->name = strdup(p);
	ent->remote_filename = strdup(file_name);
	ent->local_filename = mc_mkstemp(&handle);
	if (!ent->name && !ent->remote_filename && !ent->local_filename) {
	    ftpentry_destructor(ent);
	    ftpfserrno = ENOMEM;
	    return NULL;
	}
	if (handle == -1) {
	    ftpfserrno = EIO;
	    ftpentry_destructor(ent);
	    return NULL;
	}
	fstat(handle, &ent->s);
	close(handle);
	if (!linklist_insert(file_list, ent)) {
	    ftpfserrno = ENOMEM;
	    ftpentry_destructor(ent);
	    return NULL;
	}
	ent->freshly_created = 1;
	return ent;
    }
    else {
	ftpfserrno = ENOENT;
	return NULL;
    }
}

static struct ftpentry *
get_file_entry(char *path, int op, int flags)
{
    struct ftpfs_connection *bucket;
    struct ftpentry *fe;
    char *remote_path;

    if (!(remote_path = ftpfs_get_path (&bucket, path)))
	return NULL;
    isremotecopy = 0;
    fe = _get_file_entry(bucket, remote_path, op,
			 flags);
    free(remote_path);
    return fe;
}

#define OPT_FLUSH        1
#define OPT_IGNORE_ERROR 2

struct ftpfs_filp {
    unsigned int has_changed:1;
    struct ftpentry *fe;
    int local_handle;
};

/* Explanation:
 * On some operating systems (Slowaris 2 for example)
 * the d_name member is just a char long (Nice trick that break everything,
 * so we need to set up some space for the filename.
 */
struct ftpfs_dirent {
    struct dirent dent;
    struct linklist *pos;
    struct ftpfs_dir *dcache;
};


static int ftpfs_normal_flush = 1;

void ftpfs_hint_reread(int reread)
{
    if (reread)
	ftpfs_normal_flush++;
    else
	ftpfs_normal_flush--;
}

static void *ftpfs_open (char *file, int flags, int mode);
static int ftpfs_close (void *data);

static int ftpfs_read (void *data, char *buffer, int count_r)
{
    struct ftpfs_filp *fp;
    int n;

    fp = data;
    n = read(fp->local_handle, buffer, count_r);
    if (n < 0)
      ftpfserrno = errno;
    return n;
}

static int ftpfs_write (void *data, char *buf, int nbyte)
{
    struct ftpfs_filp *fp;
    int n;

    fp = data;
    n = write(fp->local_handle, buf, nbyte);
    if (n < 0)
      ftpfserrno = errno;
    fp->has_changed = 1;
    return n;
}
static void *ftpfs_opendir (char *Dirname);

static void *ftpfs_readdir (void *data)
{
    struct ftpentry *fe;
    struct ftpfs_dirent *dirp = data;

    if (dirp->pos == dirp->dcache->file_list)
	return NULL;
    fe = dirp->pos->data;
    strcpy (&(dirp->dent.d_name [0]), fe->name);
    dirp->pos = dirp->pos->next;
    return (void *) &dirp->dent;
}

static int ftpfs_closedir (void *info)
{
    struct ftpfs_dirent *dirp = info;
    ftpfs_dir_destructor(dirp->dcache);
    free(dirp);
    return 0;
}


static int ftpfs_lstat (char *path, struct stat *buf)
{
    struct ftpentry *fe;

    fe = get_file_entry(path, FTPFS_FREE_RESOURCE, 0);
    if (fe) {
	*buf = fe->s;
	return 0;
    }
    else
	return -1;
}

static int ftpfs_stat (char *path, struct stat *buf)
{
    struct ftpentry *fe;

    fe = get_file_entry(path, FTPFS_DO_RESOLVE_SYMLINK | FTPFS_FREE_RESOURCE, 0);
    if (fe) {
	if (!S_ISLNK(fe->s.st_mode))
	    *buf = fe->s;
	else
	    *buf = *fe->l_stat;
	return 0;
    }
    else
	return -1;
}

static int ftpfs_fstat (void *data, struct stat *buf)
{
    struct ftpfs_filp *fp = data;

    if (!S_ISLNK(fp->fe->s.st_mode))
	*buf = fp->fe->s;
    else
	*buf = *fp->fe->l_stat;
    return 0;
}

static int
send_ftp_command(char *filename, char *cmd, int flags);


static int ftpfs_chmod (char *path, int mode)
{
    char buf[40];

    sprintf(buf, "SITE CHMOD %4.4o %%s", mode & 07777);
    return send_ftp_command(path, buf, OPT_IGNORE_ERROR | OPT_FLUSH);
}

static int ftpfs_chown (char *path, int owner, int group)
{
/* Everyone knows it is not possible to chown remotely, so why bother them.
   If someone's root, then copy/move will always try to chown it... */
    return 0;
}

static int ftpfs_readlink (char *path, char *buf, int size)
{
    struct ftpentry *fe;

    fe = get_file_entry(path, FTPFS_FREE_RESOURCE, 0);
    if (!fe)
	return -1;
    if (!S_ISLNK(fe->s.st_mode)) {
	ftpfserrno = EINVAL;
	return -1;
    }
    if (fe->linkname == NULL) {
	ftpfserrno = EACCES;
	return -1;
    }
    if (strlen(fe->linkname) >= size) {
	ftpfserrno = ERANGE;
	return -1;
    }
    strncpy(buf, fe->linkname, size);
    return strlen(fe->linkname);
}

static int ftpfs_unlink (char *path)
{
    return send_ftp_command(path, "DELE %s", 1);
}

static int ftpfs_symlink (char *n1, char *n2)
{
    ftpfserrno = EPERM;
    return -1;
}

static int ftpfs_rename (char *path1, char *path2)
{
    ftpfserrno = EPERM;
    return -1;
}

static int ftpfs_lseek (void *data, off_t offset, int whence)
{
    struct ftpfs_filp *fp = data;

    return lseek(fp->local_handle, offset, whence);
}

static int ftpfs_mknod (char *path, int mode, int dev)
{
    ftpfserrno = EPERM;
    return -1;
}

static int ftpfs_mkdir (char *path, mode_t mode)
{
    return send_ftp_command(path, "MKD %s", 1);
}

static int ftpfs_rmdir (char *path)
{
    return send_ftp_command(path, "RMD %s", 1);
}

static int ftpfs_link (char *p1, char *p2)
{
    ftpfserrno = EPERM;
    return -1;
}

static vfsid ftpfs_getid (char *p, struct vfs_stamping **parent)
{
    struct ftpfs_connection *bucket;
    char *remote_path;

    *parent = NULL; /* We are not enclosed in any other fs */

    if (!(remote_path = ftpfs_get_path (&bucket, p)))
	return (vfsid) -1;
    else {
	free(remote_path);
	return (vfsid) bucket;
    }
}

static int ftpfs_nothingisopen (vfsid id)
{
    return qlock((struct ftpfs_connection *)id) == 0;
}

static void ftpfs_free (vfsid id)
{
    struct ftpfs_connection *bucket = (struct ftpfs_connection *) id;

    ftpfs_connection_destructor(bucket);
    linklist_delete(ftpfs_connections_list, bucket);
}

static char *ftpfs_getlocalcopy (char *path)
{
    struct ftpfs_filp *fp = (struct ftpfs_filp *) ftpfs_open (path, O_RDONLY, 0);
    char *p;

    if (fp == NULL)
	return NULL;
    if (fp->fe->local_filename == NULL) {
	ftpfs_close ((void *) fp);
	return NULL;
    }
    p = strdup (fp->fe->local_filename);
    qlock(fp->fe->bucket)++;
    fp->fe->count++;
    ftpfs_close ((void *) fp);
    return p;
}

static void ftpfs_ungetlocalcopy (char *path, char *local, int has_changed)
{
    struct ftpfs_filp *fp = (struct ftpfs_filp *) ftpfs_open (path, O_WRONLY, 0);

    if (fp == NULL)
	return;
    if (!strcmp (fp->fe->local_filename, local)) {
	fp->has_changed = has_changed;
	qlock(fp->fe->bucket)--;
	ftpentry_destructor(fp->fe);
	ftpfs_close ((void *) fp);
    } else {
	/* Should not happen */
	ftpfs_close ((void *) fp);
	mc_def_ungetlocalcopy (path, local, has_changed);
    }
}

static void ftpfs_forget (char *file)
{
    struct linklist *l;
    char *host, *user, *pass, *rp;
    int port;

    if (strncmp (file, "ftp://", 6) != 0)
	return;

    file += 6;
    if (!(rp = ftpfs_get_host_and_username (file, &host, &user, &port, &pass))) {
	free (host);
	free (user);
	if (pass)
	    wipe_password (pass);
	return;
    }

    /* we do not care about the path actually */
    free (rp);

    for (l = ftpfs_connections_list->next; l != ftpfs_connections_list; l = l->next) {
	struct ftpfs_connection *bucket = l->data;

	if ((strcmp (host, qhost (bucket)) == 0) &&
	    (strcmp (user, quser (bucket)) == 0) &&
	    (port == qport (bucket))) {

	    /* close socket: the child owns it now */
	    close (bucket->sock);
	    bucket->sock = -1;

	    /* reopen the connection */
	    bucket->sock = ftpfs_open_socket (bucket);
	    if (bucket->sock != -1)
		login_server (bucket, pass);
	    break;
	}
    }
    free (host);
    free (user);
    if (pass)
	wipe_password (pass);
}

static caddr_t ftpfs_mmap (caddr_t addr, size_t len, int prot, int flags, void *data, off_t offset)
{
    return (caddr_t)-1; /* We do not mmap to far away */
}

static int ftpfs_munmap (caddr_t addr, size_t len, void *data)
{
    return -1;
}

static int ftpfs_chdir (char *path);


static int ftpfs_errno (void)
{
    return ftpfserrno;
}


static vfs ftpfs_vfs_ops = {
    ftpfs_open,
    ftpfs_close,
    ftpfs_read,
    ftpfs_write,

    ftpfs_opendir,
    ftpfs_readdir,
    ftpfs_closedir,

    ftpfs_stat,
    ftpfs_lstat,
    ftpfs_fstat,

    ftpfs_chmod,
    ftpfs_chown,
    NULL,

    ftpfs_readlink,
    ftpfs_symlink,
    ftpfs_link,
    ftpfs_unlink,

    ftpfs_rename,
    ftpfs_chdir,
    ftpfs_errno,
    ftpfs_lseek,
    ftpfs_mknod,

    ftpfs_getid,
    ftpfs_nothingisopen,
    ftpfs_free,

    ftpfs_getlocalcopy,
    ftpfs_ungetlocalcopy,

    ftpfs_mkdir,
    ftpfs_rmdir,
    ftpfs_ctl,
    ftpfs_setctl,
    ftpfs_forget,
    ftpfs_mmap,
    ftpfs_munmap
};


static int ftpfs_chdir (char *path)
{
    char *remote_path;
    struct ftpfs_connection *bucket;

    if (!(remote_path = ftpfs_get_path(&bucket, path)))
	return -1;
    if (qcdir(bucket))
	free(qcdir(bucket));
    qcdir(bucket) = remote_path;
    bucket->cwd_defered = 1;

    vfs_add_noncurrent_stamps (&ftpfs_vfs_ops, (vfsid) bucket, NULL);
    return 0;
}

static int
send_ftp_command(char *filename, char *cmd, int flags)
{
    char *remote_path;
    struct ftpfs_connection *bucket;
    int r;
    int flush_directory_cache = (flags & OPT_FLUSH) && (ftpfs_normal_flush > 0);

    if (!(remote_path = ftpfs_get_path(&bucket, filename)))
	return -1;
    r = command (bucket, WAIT_REPLY, cmd, remote_path);
    free(remote_path);
    vfs_add_noncurrent_stamps (&ftpfs_vfs_ops, (vfsid) bucket, NULL);
    if (flags & OPT_IGNORE_ERROR)
	r = COMPLETE;
    if (r != COMPLETE) {
	ftpfserrno = EPERM;
	return -1;
    }
    if (flush_directory_cache)
	flush_all_directory(bucket);
    return 0;
}

/* This routine is called as the last step in load_setup */
static void
ftpfs_init_passwd(void)
{
    char *p, hostname[MAXHOSTNAMELEN], *h;

    ftpfs_anonymous_passwd = load_anon_passwd ();
    if (ftpfs_anonymous_passwd)
	return;

    p = bb_getpwuid ("guest", my_euid, 0);
    gethostname(hostname, sizeof(hostname));
    if(bb_gethostbyname4(BB_GHBN_SILENT | BB_GHBN_CANON_NAME, hostname, &h) != NULL) {
	ftpfs_anonymous_passwd = xasprintf ("%s@%s", p, h);
	free(h);
    } else
	ftpfs_anonymous_passwd = xasprintf ("%s@%s", p, hostname);
}

static void ftpfs_init (void)
{
    ftpfs_connections_list = linklist_init();
    ftpfs_directory_timeout = FTPFS_DIRECTORY_TIMEOUT;
}

static void ftpfs_done(void)
{
    linklist_destroy(ftpfs_connections_list, ftpfs_connection_destructor);
    if (ftpfs_debug_server_dialog)
	fclose (ftpfs_logfile);
}

/* The callbacks */

static void *ftpfs_open (char *file, int flags, int mode)
{
    struct ftpfs_filp *fp;
    struct ftpentry *fe;

    fp = x_malloc(sizeof(struct ftpfs_filp), "struct ftpfs_filp");
    if (fp == NULL) {
	ftpfserrno = ENOMEM;
	return NULL;
    }
    fe = get_file_entry(file, FTPFS_OPEN | FTPFS_DO_RESOLVE_SYMLINK, flags);
    if (fe == NULL) {
	free(fp);
	return NULL;
    }
    if (!isremotecopy) {
	fp->local_handle = open(fe->local_filename, flags, mode);
	if (fp->local_handle < 0) {
	    ftpfserrno = errno;
	    free(fp);
	    return NULL;
	}
    } else
	fp->local_handle = -1;
#ifdef UPLOAD_ZERO_LENGTH_FILE
    fp->has_changed = fe->freshly_created;
#else
    fp->has_changed = 0;
#endif
    fp->fe = fe;
    qlock(fe->bucket)++;
    fe->count++;
    return fp;
}


static int ftpfs_close (void *data)
{
    struct ftpfs_filp *fp = data;
    int result = 0;

    if (fp->has_changed) {
	if (!store_file(fp->fe))
	    result = -1;
	if (ftpfs_normal_flush)
	    flush_all_directory(fp->fe->bucket);
    }
    if (fp->local_handle >= 0)
	close(fp->local_handle);
    qlock(fp->fe->bucket)--;
    ftpentry_destructor(fp->fe);
    free(fp);
    return result;
}


char *ftpfs_gethome (char *servername)
{
    struct ftpfs_connection *bucket;
    char *remote_path;

    if (!(remote_path = ftpfs_get_path (&bucket, servername)))
	return NULL;
    free (remote_path);
    return qhome(bucket);
}

static char *ftpfs_getupdir (char *servername)
{
    struct ftpfs_connection *bucket;
    char *remote_path;

    if (!(remote_path = ftpfs_get_path (&bucket, servername)))
	return NULL;
    free (remote_path);
    return qupdir(bucket);
}

static void *ftpfs_opendir (char *Dirname)
{
    struct ftpfs_connection *bucket;
    char *remote_path;
    struct ftpfs_dirent *dirp;

    if (!(remote_path = ftpfs_get_path (&bucket, Dirname)))
	return NULL;
    dirp = x_malloc(sizeof(struct ftpfs_dirent), "struct ftpfs_dirent");
    if (dirp == NULL) {
	ftpfserrno = ENOMEM;
	goto error_return;
    }
    dirp->dcache = retrieve_dir(bucket, remote_path, 1);
    if (dirp->dcache == NULL)
	goto error_return;
    dirp->pos = dirp->dcache->file_list->next;
    free(remote_path);
    dirp->dcache->count++;
    return (void *)dirp;
error_return:
    vfs_add_noncurrent_stamps (&ftpfs_vfs_ops, (vfsid) bucket, NULL);
    free(remote_path);
    free(dirp);
    return NULL;
}


static int
__ftpfs_chdir (struct ftpfs_connection *bucket ,char *remote_path)
{
    int r;

    if (!bucket->cwd_defered && ftpfs_same_dir (remote_path, bucket))
	return COMPLETE;

    r = command (bucket, WAIT_REPLY, "CWD %s", remote_path);
    if (r != COMPLETE) {
	ftpfserrno = EIO;
    } else {
	if (qcdir(bucket))
	    free(qcdir(bucket));
	qcdir(bucket) = strdup (remote_path);
	bucket->cwd_defered = 0;
    }
    return r;
}

static void ftpfs_set_debug (const char *file)
{
    if ((ftpfs_logfile = fopen (file, "w+")) != NULL)
	ftpfs_debug_server_dialog = 1;
}
#endif /* USE_NETCODE */

/*
 * Look up user and/or group names.
 *
 * This file should be modified for non-unix systems to do something
 * reasonable.
 */

static int saveuid = -993;
static char saveuname[TUNMLEN];

static int savegid = -993;
static char savegname[TGNMLEN];

/*
 * Look up a user or group name from a uid/gid, maintaining a cache.
 * FIXME, for now it's a one-entry cache.
 * FIXME2, the "-993" is to reduce the chance of a hit on the first lookup.
 *
 * This is ifdef'd because on Suns, it drags in about 38K of "yellow
 * pages" code, roughly doubling the program size.  Thanks guys.
 */

static int finduid (const char *uname)
{
    struct passwd *pw;

    if (uname[0] != saveuname[0]/* Quick test w/o proc call */
	||0 != strncmp (uname, saveuname, TUNMLEN)) {
	strncpy (saveuname, uname, TUNMLEN);
	pw = getpwnam (uname);
	if (pw) {
	    saveuid = pw->pw_uid;
	} else {
	    saveuid = my_uid;
	}
    }
    return saveuid;
}

static int findgid (const char *gname)
{
    struct group *gr;

    if (gname[0] != savegname[0]/* Quick test w/o proc call */
	||0 != strncmp (gname, savegname, TUNMLEN)) {
	strncpy (savegname, gname, TUNMLEN);
	gr = getgrnam (gname);
	if (gr) {
	    savegid = gr->gr_gid;
	} else {
	    savegid = my_gid;
	}
    }
    return savegid;
}

#define isodigit(c)     ( ((c) >= '0') && ((c) <= '7') )
/*
 * Quick and dirty octal conversion.
 *
 * Result is -1 if the field is invalid (all blank, or nonoctal).
 */
static long from_oct (int digs, char *where)
{
    long value;

    while (isspace (*where)) {  /* Skip spaces */
	where++;
	if (--digs <= 0)
	    return -1;          /* All blank field */
    }
    value = 0;
    while (digs > 0 && isodigit (*where)) {     /* Scan till nonoctal */
	value = (value << 3) | (*where++ - '0');
	--digs;
    }

    if (digs > 0 && *where && !isspace (*where))
	return -1;              /* Ended on non-space/nul */

    return value;
}

static struct tarfs_archive *tarfs_first_archive = NULL;
static int tarerrno = 0;
static struct stat tarfs_hstat;               /* Stat struct corresponding */
static char *tarfs_current_file_name, *tarfs_current_link_name;
static struct tarfs_entry *tarfs_find_entry (struct tarfs_entry *dir, char *name, int make_dirs);

void tarfs_fill_names (void (*func)(char *))
{
    struct tarfs_archive *a = tarfs_first_archive;
    char *name;

    while (a) {
	name = xasprintf ("tar:%s/%s", a->name, a->current_dir->name);
	(*func)(name);
	free (name);
	a = a->next;
    }
}

static void tarfs_make_dot_doubledot (struct tarfs_entry *ent)
{
    struct tarfs_entry *entry = (struct tarfs_entry *)
	x_malloc (sizeof (struct tarfs_entry), "Tar: tarfs_entry");
    struct tarfs_entry *parentry = ent->dir;
    struct tarfs_inode *inode = ent->inode, *parent;

    parent = (parentry != NULL) ? parentry->inode : NULL;
    entry->name = strdup (".");
    entry->has_changed = 0;
    entry->header_offset = -1;
    entry->header_size = 0;
    entry->extended_offset = -1;
    entry->extended_size = 0;
    entry->inode = inode;
    entry->dir = ent;
    inode->first_in_subdir = entry;
    inode->last_in_subdir = entry;
    inode->nlink++;
    entry->next_in_dir = (struct tarfs_entry *)
	x_malloc (sizeof (struct tarfs_entry), "Tar: tarfs_entry");
    entry=entry->next_in_dir;
    entry->name = strdup ("..");
    entry->has_changed = 0;
    entry->header_offset = -1;
    entry->header_size = 0;
    entry->extended_offset = -1;
    entry->extended_size = 0;
    inode->last_in_subdir = entry;
    entry->next_in_dir = NULL;
    if (parent != NULL) {
	entry->inode = parent;
	entry->dir = parentry;
	parent->nlink++;
    } else {
	entry->inode = inode;
	entry->dir = ent;
	inode->nlink++;
    }
}

static struct tarfs_entry *tarfs_generate_entry (struct tarfs_archive *archive,
    char *name, struct tarfs_entry *parentry, mode_t mode)
{
    mode_t myumask;
    struct tarfs_inode *inode, *parent;
    struct tarfs_entry *entry;

    parent = (parentry != NULL) ? parentry->inode : NULL;
    entry = (struct tarfs_entry *)
	x_malloc (sizeof (struct tarfs_entry), "Tar: tarfs_entry");

    entry->name = strdup (name);
    entry->has_changed = 0;
    entry->header_offset = -1;
    entry->header_size = 0;
    entry->extended_offset = -1;
    entry->extended_size = 0;
    entry->next_in_dir = NULL;
    entry->dir = parentry;
    if (parent != NULL) {
	parent->last_in_subdir->next_in_dir = entry;
	parent->last_in_subdir = entry;
    }
    inode = (struct tarfs_inode *)
	x_malloc (sizeof (struct tarfs_inode), "Tar: tarfs_inode");
    entry->inode = inode;
    inode->local_filename = NULL;
    inode->has_changed = 0;
    inode->is_open = 0;
    inode->linkname = 0;
    inode->inode = (archive->__inode_counter)++;
    inode->dev = archive->rdev;
    inode->archive = archive;
    inode->data_offset = -1;
    myumask = umask (022);
    umask (myumask);
    inode->mode = mode & ~myumask;
    mode = inode->mode;
    inode->rdev = 0;
    inode->uid = my_uid;
    inode->gid = my_gid;
    inode->std = 1;
    inode->size = 0;
    inode->mtime = time (NULL);
    inode->atime = inode->mtime;
    inode->ctime = inode->mtime;
    inode->nlink = 1;
    if (S_ISDIR (mode)) {
	inode->linkflag = LF_DIR;
	tarfs_make_dot_doubledot (entry);
    } else if (S_ISLNK (mode)) {
	inode->linkflag = LF_SYMLINK;
    } else if (S_ISCHR (mode)) {
	inode->linkflag = LF_CHR;
    } else if (S_ISBLK (mode)) {
	inode->linkflag = LF_BLK;
    } else if (S_ISFIFO (mode) || S_ISSOCK (mode)) {
	inode->linkflag = LF_FIFO;
    } else {
	inode->linkflag = LF_NORMAL;
    }
    return entry;
}

static void tarfs_free_entries (struct tarfs_entry *entry)
{
    return;
}

static void tarfs_free_archive (struct tarfs_archive *archive)
{
    long l;

    if (archive->is_gzipped == targz_growing) {
	free (archive->block_first);
	if (archive->block_ptr != NULL) {
	    for (l = 0; l < archive->count_blocks; l++)
		free (archive->block_ptr [l]);
	    free (archive->block_ptr);
	}
    } else {
	if (archive->is_gzipped == tar_uncompressed_local) {
	    mc_unlink (archive->tmpname);
	    free (archive->tmpname);
	}
	if (archive->fd != -1)
	    mc_close(archive->fd);
    }
    tarfs_free_entries (archive->root_entry);

    free (archive->name);
    free (archive);
}

static inline int gzip_limit_ok (int size)
{
    return (size <= tar_gzipped_memlimit || tar_gzipped_memlimit < 0);
}

/* So we have to decompress it...
 * It is not that easy, because we would like to handle all the files
 * from all the vfs's. So, we do this like this:
 * we run a pipe:
 *    for (;;) mc_read | gzip -cdf | store into growing buffer
 *
 * Returns: 0 on failure
 */
static inline int load_compressed_tar (struct tarfs_archive *current_archive,
					int size, int fd, int type)
{
    int pipehandle, i;
    long l, l2;
    union record *ur;
    pid_t p;
    char  *cmd, *cmd_flags;
    void *tmp;

    l2 = 0;
    current_archive->is_gzipped = targz_growing;
    size = (size + RECORDSIZE - 1) / RECORDSIZE * RECORDSIZE;
    current_archive->count_first = size / RECORDSIZE;
    current_archive->count_blocks = 0;
    current_archive->block_ptr = NULL;
    current_archive->root_entry = NULL;

    decompress_command_and_arg (type, &cmd, &cmd_flags);
    pipehandle = mc_doublepopen (fd, -1, &p, cmd, cmd, cmd_flags, NULL);
    if (pipehandle == -1)
    {
	free (current_archive);
	mc_close (fd);
	return 0;
    }

    /* On some systems it's better to allocate the memory for the tar-file
       this way:
       - allocate some memory we don't need (size 100k).
       - allocate the memory for the uncompressed tar-file (the size of
	 this memory block can be really big - five times the size of the
	 compressed tar-file).
       - free the 100k from the first step

       Without the extra steps a few malloc/free implementations can't give
       back memory to the operating system when the memory for the tar-file is
       freed after the vfs-timeout.
     */
    tmp = malloc (100*1024);
    current_archive->block_first = (union record *) malloc (size);
    if (tmp)
       free (tmp); /* make a hole in size of 100k */
    if (0 == current_archive->block_first) {
	mc_doublepclose (pipehandle, p);
	free (current_archive);
	mc_close (fd);
	return 0;
    }

    ur = current_archive->block_first;
    l = 0;
    while ((i = read (pipehandle, (char *) ur, RECORDSIZE)) == RECORDSIZE) {
	l++;
	if (l >= current_archive->count_first) {
	    l2 = l - current_archive->count_first;
	    if (l2 % TAR_GROWING_CHUNK_SIZE)
		ur++;
	    else {
		tmp = x_malloc ((++current_archive->count_blocks) *
			     sizeof (union record **), "Tar: Growing buffers");

		if (current_archive->block_ptr != NULL) {
		    memcpy (tmp, current_archive->block_ptr,
			   (current_archive->count_blocks - 1) *
			   sizeof (union record **));
		    free (current_archive->block_ptr);
		}
		current_archive->block_ptr = (union record **)tmp;
		ur = (union record *)
		    x_malloc (TAR_GROWING_CHUNK_SIZE * RECORDSIZE,
			     "Tar: Growing buffers");
		current_archive->block_ptr [current_archive->count_blocks - 1] = ur;
	    }
	} else
	    ur++;
	if ((dash_number++ % 64) == 0)
	    rotate_dash ();
    }
    i = mc_doublepclose (pipehandle, p);
    mc_close (fd);
    if (i == -1) {
	tarfs_free_archive (current_archive);
	return 0;
    }
    current_archive->current_record = current_archive->block_first;
    return 1;
}

/* Returns a file handle of the opened local tar file or -1 on error */
static inline int uncompress_tar_file  (struct tarfs_archive *current_archive,
					int size, int fd, int type)
{
    FILE *f;
    char *command;
    int i, result;
    char buffer [8192]; /* Changed to 8K: better transfer size */
    int fd_unuse;

    dash_number = 0;

    current_archive->is_gzipped = tar_uncompressed_local;
    current_archive->tmpname = mc_mkstemp(&fd_unuse);
    if(fd_unuse >= 0)
	close(fd_unuse);

    /* Some security is sometimes neccessary :) */
    command = xasprintf ("touch %s ; chmod 0600 %s ; %s 2>/dev/null >%s",
				current_archive->tmpname,
				current_archive->tmpname,
				decompress_command (type),
				current_archive->tmpname);

    if ((f = popen (command, "w")) == NULL) {
	mc_close (fd);
	tarfs_free_archive (current_archive);
	free (command);
	return -1;
    }
    free (command);

    while ((i = mc_read (fd, buffer, sizeof (buffer))) > 0) {
	if ((dash_number++ % 64) == 0)
	    rotate_dash ();
	fwrite (buffer, 1, i, f);
	if (ferror (f)) {
	    pclose (f);
	    mc_close (fd);
	    tarfs_free_archive (current_archive);
	    return -1;
	}
    }

    pclose (f);
    mc_close (fd);
    result = mc_open (current_archive->tmpname, O_RDONLY);
    if (result == -1) {
	tarfs_free_archive (current_archive);
	return -1;
    }
    return result;
}

/* Returns fd of the open tar file */
static int open_tar_archive (char *name, struct tarfs_archive **pparc)
{
    static dev_t __tar_no = 0;
    int result, type;
    long size;
    mode_t mode;
    struct tarfs_archive *current_archive;
    static struct tarfs_entry *root_entry;

    result = mc_open (name, O_RDONLY);
    if (result == -1)
	return -1;

    current_archive = (struct tarfs_archive *)
		      x_malloc (sizeof (struct tarfs_archive), "Tar archive");
    current_archive->current_tar_position = 0;
    current_archive->name = strdup (name);
    current_archive->__inode_counter = 0;
    mc_stat (name, &(current_archive->tarstat));
    current_archive->rdev = __tar_no++;
    current_archive->next = tarfs_first_archive;
    current_archive->fd_usage = 0;
    current_archive->fd = -1;
    size = is_gunzipable (result, &type);
    mc_lseek (result, 0, SEEK_SET);

    /* Find out the method to handle this tar file */
    if (size > 0 && gzip_limit_ok (size)) {
	if (load_compressed_tar (current_archive, size, result, type) == 0)
	    return -1;
	result = 0;
    } else if (size > 0) {
	result = uncompress_tar_file (current_archive, size, result, type);
	if (result == -1)
	    return -1;
    } else {
	current_archive->is_gzipped = tar_normal;
    }

    current_archive->fd = result;
    tarfs_first_archive = current_archive;
    mode = current_archive->tarstat.st_mode & 07777;
    if (mode & 0400)
	mode |= 0100;
    if (mode & 0040)
	mode |= 0010;
    if (mode & 0004)
	mode |= 0001;
    mode |= S_IFDIR;
    root_entry = tarfs_generate_entry (current_archive, "/", NULL, mode);
    root_entry->inode->uid = current_archive->tarstat.st_uid;
    root_entry->inode->gid = current_archive->tarstat.st_gid;
    root_entry->inode->atime = current_archive->tarstat.st_atime;
    root_entry->inode->ctime = current_archive->tarstat.st_ctime;
    root_entry->inode->mtime = current_archive->tarstat.st_mtime;
    current_archive->root_entry = root_entry;
    current_archive->current_dir = root_entry;

    *pparc = current_archive;

    return result;
}

static off_t get_current_position (struct tarfs_archive *archive, int tard)
{
    return archive->current_tar_position;
}

static union record *find_current_record (struct tarfs_archive *archive, long pos)
{
    long l, l2;
    static union record *ur;

    l = pos / RECORDSIZE;
    if (l >= archive->count_first) {
	l2 = l - archive->count_first;
	ur = archive->block_ptr [l2 / TAR_GROWING_CHUNK_SIZE]
	     + (l2 % TAR_GROWING_CHUNK_SIZE);
    } else
	ur = archive->block_first + l;
    return ur;
}

static union record rec_buf;

static union record *get_next_record (struct tarfs_archive *archive, int tard)
{
    if (archive->is_gzipped == targz_growing) {
	memcpy (rec_buf.charptr, archive->current_record, RECORDSIZE);
	archive->current_record = find_current_record (archive, archive->current_tar_position + RECORDSIZE);
    } else if (mc_read (tard, rec_buf.charptr, RECORDSIZE) != RECORDSIZE)
	return NULL;            /* An error has occurred */
    archive->current_tar_position += RECORDSIZE;
    return &rec_buf;
}

static void skip_n_records (struct tarfs_archive *archive, int tard, int n)
{
    if (archive->is_gzipped == targz_growing)
	archive->current_record = find_current_record (archive, archive->current_tar_position + n * RECORDSIZE);
    else
	mc_lseek (tard, n * RECORDSIZE, SEEK_CUR);
    archive->current_tar_position += n * RECORDSIZE;
}

/*
 * Return 1 for success, 0 if the checksum is bad, EOF on eof,
 * 2 for a record full of zeros (EOF marker).
 *
 */
static int read_header (struct tarfs_archive *archive, int tard)
{
    int i;
    long sum, signed_sum, recsum;
    char *p;
    union record *header;
    char **longp;
    char *bp, *data;
    int size, written;
    static char *next_long_name = NULL, *next_long_link = NULL;
    off_t header_position = get_current_position (archive, tard);

  recurse:

    header = get_next_record (archive, tard);
    if (NULL == header)
	return EOF;

    recsum = from_oct (8, header->header.chksum);

    sum = 0; signed_sum = 0;
    p = header->charptr;
    for (i = sizeof (*header); --i >= 0;) {
		/*
		 * We can't use unsigned char here because of old compilers,
		 * e.g. V7.
		 */
	signed_sum += *p;
	sum += 0xFF & *p++;
    }

    /* Adjust checksum to count the "chksum" field as blanks. */
    for (i = sizeof (header->header.chksum); --i >= 0;) {
	sum -= 0xFF & header->header.chksum[i];
	signed_sum -= (char) header->header.chksum[i];
    }
    sum += ' ' * sizeof header->header.chksum;
    signed_sum += ' ' * sizeof header->header.chksum;

    if (sum == 8 * ' ') {
	/*
	 * This is a zeroed record...whole record is 0's except
	 * for the 8 blanks we faked for the checksum field.
	 */
	    return 2;
    }
    if (sum != recsum && signed_sum != recsum)
	return 0;

    /*
     * linkflag on BSDI tar (pax) always '\000'
     */
    if(header->header.linkflag == '\000' &&
       strlen(header->header.arch_name) &&
       header->header.arch_name[strlen(header->header.arch_name) - 1] == '/')
	header->header.linkflag = LF_DIR;

    /*
     * Good record.  Decode file size and return.
     */
    if (header->header.linkflag == LF_LINK || header->header.linkflag == LF_DIR)
	tarfs_hstat.st_size = 0;      /* Links 0 size on tape */
    else
	tarfs_hstat.st_size = from_oct (1 + 12, header->header.size);

    header->header.arch_name[NAMSIZ - 1] = '\0';
    if (header->header.linkflag == LF_LONGNAME
	|| header->header.linkflag == LF_LONGLINK) {
	longp = ((header->header.linkflag == LF_LONGNAME)
		 ? &next_long_name
		 : &next_long_link);

	if (*longp)
	    free (*longp);
	bp = *longp = (char *) x_malloc (tarfs_hstat.st_size, "Tar: Long name");

	for (size = tarfs_hstat.st_size;
	     size > 0;
	     size -= written) {
	    data = get_next_record (archive, tard)->charptr;
	    if (data == NULL) {
		message_1s (1, " Error ", "Unexpected EOF on archive file");
		return 0;
	    }
	    written = RECORDSIZE;
	    if (written > size)
		written = size;

	    memcpy (bp, data, written);
	    bp += written;
	}
	goto recurse;
    } else {
	struct tarfs_entry *entry, *pent;
	struct tarfs_inode *inode;
	off_t data_position;
	char *q;
	int len;
	int isdir = 0;

	tarfs_current_file_name = (next_long_name
			     ? next_long_name
			     : strdup (header->header.arch_name));
	len = strlen (tarfs_current_file_name);
	if (tarfs_current_file_name[len - 1] == '/') {
	    tarfs_current_file_name[len - 1] = 0;
	    isdir = 1;
	}

	tarfs_current_link_name = (next_long_link
			     ? next_long_link
			     : strdup (header->header.arch_linkname));
	len = strlen (tarfs_current_link_name);
	if (len && tarfs_current_link_name [len - 1] == '/')
	    tarfs_current_link_name[len - 1] = 0;

	next_long_link = next_long_name = NULL;

	data_position = get_current_position (archive, tard);

	p = strrchr (tarfs_current_file_name, '/');
	if (p == NULL) {
	    p = tarfs_current_file_name;
	    q = tarfs_current_file_name + strlen (tarfs_current_file_name); /* "" */
	} else {
	    *(p++) = 0;
	    q = tarfs_current_file_name;
	}

	pent = tarfs_find_entry (archive->root_entry, q, 1);
	if (pent == NULL) {
	    message_1s (1, " Error ", "Inconsistent tar archive");
	}

	entry = (struct tarfs_entry *) x_malloc (sizeof (struct tarfs_entry), "Tar: tarfs_entry");
	entry->name = strdup (p);
	entry->has_changed = 0;
	entry->header_offset = header_position;
	entry->header_size = data_position - header_position;
	entry->extended_offset = -1;
	entry->extended_size = 0;
	entry->next_in_dir = NULL;
	entry->dir = pent;
	if (pent != NULL) {
	    pent->inode->last_in_subdir->next_in_dir = entry;
	    pent->inode->last_in_subdir = entry;
	}
	free (tarfs_current_file_name);

	if (header->header.linkflag == LF_LINK) {
	    pent = tarfs_find_entry (archive->root_entry, tarfs_current_link_name, 0);
	    if (pent == NULL) {
		message_1s (1, " Error ", "Inconsistent tar archive");
	    } else {
		entry->inode = pent->inode;
		pent->inode->nlink++;
		free (tarfs_current_link_name);
		if (header->header.isextended) {
		    entry->extended_offset = data_position;
		    while (get_next_record (archive, tard)->ext_hdr.isextended);
		    data_position = get_current_position (archive, tard);
		    entry->extended_size = data_position - entry->extended_offset;
		}
		return 1;
	    }
	}
	inode = (struct tarfs_inode *) x_malloc (sizeof (struct tarfs_inode), "Tar: tarfs_inode");
	entry->inode = inode;
	inode->local_filename = NULL;
	inode->has_changed = 0;
	inode->is_open = 0;
	inode->inode = (archive->__inode_counter)++;
	inode->nlink = 1;
	inode->dev = archive->rdev;
	inode->archive = archive;
	inode->data_offset = data_position;
	inode->mode = from_oct (8, header->header.mode);

	/* Adjust inode->mode because there are tar-files with
	 * linkflag==LF_SYMLINK and S_ISLNK(mod)==0. I don't
	 * know about the other modes but I think I cause no new
	 * problem when I adjust them, too. -- Norbert.
	 */
	if (header->header.linkflag == LF_DIR) {
	    inode->mode |= S_IFDIR;
	} else if (header->header.linkflag == LF_SYMLINK) {
	    inode->mode |= S_IFLNK;
	} else if (header->header.linkflag == LF_CHR) {
	    inode->mode |= S_IFCHR;
	} else if (header->header.linkflag == LF_BLK) {
	    inode->mode |= S_IFBLK;
	} else if (header->header.linkflag == LF_FIFO) {
	    inode->mode |= S_IFIFO;
	} else
	    inode->mode |= S_IFREG;

	inode->rdev = 0;
	if (!strcmp (header->header.magic, TMAGIC)) {
	    inode->uid = *header->header.uname ? finduid (header->header.uname) :
						 from_oct (8, header->header.uid);
	    inode->gid = *header->header.gname ? findgid (header->header.gname) :
						 from_oct (8, header->header.gid);
	    switch (header->header.linkflag) {
		case LF_BLK:
		case LF_CHR:
		    inode->rdev = (from_oct (8, header->header.devmajor) << 8) |
				   from_oct (8, header->header.devminor);
	    }
	    inode->std = 1;
	} else { /* Old Unix tar */
	    inode->uid = from_oct (8, header->header.uid);
	    inode->gid = from_oct (8, header->header.gid);
	    inode->std = 0;
	}
	inode->size = tarfs_hstat.st_size;
	inode->mtime = from_oct (1 + 12, header->header.mtime);
	inode->atime = from_oct (1 + 12, header->header.atime);
	inode->ctime = from_oct (1 + 12, header->header.ctime);
	inode->linkflag = header->header.linkflag;
	if (*tarfs_current_link_name) {
	    inode->linkname = tarfs_current_link_name;
	} else {
	    free (tarfs_current_link_name);
	    inode->linkname = NULL;
	}
	if (inode->linkflag == LF_DIR || isdir) {
	    inode->mode |= S_IFDIR;
	    tarfs_make_dot_doubledot (entry);
	}

	if (header->header.isextended) {
	    entry->extended_offset = data_position;
	    while (get_next_record (archive, tard)->ext_hdr.isextended);
	    inode->data_offset = get_current_position (archive, tard);
	    entry->extended_size = inode->data_offset - entry->extended_offset;
	}
	return 1;
    }
}

/*
 * Main loop for reading an archive.
 * Returns 0 on success, -1 on error.
 */
static int read_tar_archive (char *name, struct tarfs_archive **pparc)
{
    int Status = 3;             /* Initial status at start of archive */
    int prev_status;
    int tard;
    struct tarfs_archive *archive;

    if ((tard = open_tar_archive (name, &archive)) == -1) {     /* Open for reading */
	message_2s (1, " Error ", "Couldn't open tar archive\n%s", name);
	return -1;
    }

    for (;;) {
	prev_status = Status;
	Status = read_header (archive, tard);
	switch (Status) {

	case 1:         /* Valid header */
		skip_n_records (archive, tard, (tarfs_hstat.st_size + RECORDSIZE - 1) / RECORDSIZE);
		continue;
			/*
			 * If the previous header was good, tell them
			 * that we are skipping bad ones.
			 */
	case 0:         /* Invalid header */
	    switch (prev_status) {
	    case 3:             /* Error on first record */
		message_2s (1, " Error ", "Hmm,...\n%s\ndoesn't look like a tar archive.", name);
		/* FALL THRU */
	    case 2:             /* Error after record of zeroes */
	    case 1:             /* Error after header rec */
	    case 0:             /* Error after error */
		return -1;
	    }

	case 2:         /* Record of zeroes */
	    Status = prev_status;       /* If error after 0's */
	    /* FALL THRU */
	case EOF:               /* End of archive */
	    break;
	}
	break;
    };

    *pparc = archive;
    return 0;
}

static char *tarfs_analysis (const char *name, char **archive, int is_dir)
{
    static struct {
	int len; /* strlen (ext)  */
	char *ext;
    } tarext[] = {{4, ".tar"},
		  {4, ".tgz"},
		  {7, ".tar.gz"},
		  {4, ".taz"},
		  {4, ".tpz"},
		  {6, ".tar.z"},
		  {7, ".tar.bz"},
		  {8, ".tar.bz2"},
		  {6, ".tar.Z"} };
    char *local;
    const char *p;
    unsigned int i;
    char *archive_name = NULL;

					       /*  |  this is len of "tar:" plus some minimum
						*  v  space needed for the extension */
    for (p = name + strlen (name); p >= name + 8; p--)
	if (*p == '/' || (is_dir && !*p))
	    for (i = 0; i < ELEMENTS(tarext); i++)
		if (!strncmp (p - tarext [i].len, tarext [i].ext, tarext [i].len)) {
		    char c = *p;
		    char *ploc;

		    local = strdup (p);
		    ploc = local + (p - name);
		    *ploc = 0;
		    archive_name = vfs_canon (local + 4);
		    *archive = archive_name;
		    *ploc = c;
		    return local;
		}
    tarerrno = ENOENT;
    return NULL;
}

static caddr_t
tar_mmap (caddr_t addr, size_t len, int prot, int flags, void *data, off_t offset)
{
    return (caddr_t)-1;
}

static int tar_munmap (caddr_t addr, size_t len, void *data)
{
    return -1;
}

static void *tar_open (char *file, int flags, int mode);
static int tar_close (void *data);
static int tar_read (void *data, char *buffer, int count_r);

static int tar_unlink (char *path)
{
    return -1;
}

static int tar_symlink (char *n1, char *n2)
{
    return -1;
}

static int tar_write (void *data, char *buf, int nbyte)
{
    return -1;
}

static int tar_rename (char *a, char *b)
{
    return -1;
}

static void *tar_opendir (char *Dirname);
static void *tar_readdir (void *data);

static int tar_closedir (void *data)
{
    free (data);
    return 0;
}
static int tar_stat (const char *path, struct stat *buf);
static int tar_lstat (const char *path, struct stat *buf);
static int tar_fstat (void *data, struct stat *buf);

static int tar_chmod (char *path, int mode)
{
    return chmod (path, mode);
}

static int tar_chown (char *path, int owner, int group)
{
    return chown (path, owner, group);
}

static int tar_readlink (char *path, char *buf, int size);
static int tar_chdir (char *path);

static int tar_errno (void)
{
    return tarerrno;
}

static int tar_lseek (void *data, off_t offset, int whence);

static int tar_mknod (char *path, int mode, int dev)
{
    return -1;
}

static int tar_link (char *p1, char *p2)
{
    return -1;
}

static int tar_mkdir (char *path, mode_t mode)
{
    return -1;
}

static int tar_rmdir (char *path)
{
    return -1;
}

static vfsid tar_getid (char *path, struct vfs_stamping **parent);

static int tar_nothingisopen (vfsid id)
{
    if (((struct tarfs_archive *)id)->fd_usage <= 0)
	return 1;
    else
	return 0;
}

static void tar_free (vfsid id);

static char *tar_getlocalcopy (char *path);

static void tar_ungetlocalcopy (char *path, char *local, int has_changed)
{
/* We do just nothing. (We are read only and do not need to free local,
   since it will be freed when tar archive will be freed */
}


static vfs tarfs_vfs_ops =
{
    tar_open,
    tar_close,
    tar_read,
    tar_write,

    tar_opendir,
    tar_readdir,
    tar_closedir,

    tar_stat,
    tar_lstat,
    tar_fstat,

    tar_chmod,
    tar_chown,
    NULL,

    tar_readlink,
    tar_symlink,
    tar_link,
    tar_unlink,

    tar_rename,
    tar_chdir,
    tar_errno,
    tar_lseek,
    tar_mknod,

    tar_getid,
    tar_nothingisopen,
    tar_free,

    tar_getlocalcopy,
    tar_ungetlocalcopy,

    tar_mkdir,
    tar_rmdir,
    NULL,
    NULL,
    NULL,
    tar_mmap,
    tar_munmap
};


/* Returns allocated path inside the archive or NULL */
static char *tarfs_get_path (const char *inname, struct tarfs_archive **archive, int is_dir,
    int do_not_open)
{
    char *local, *archive_name;
    int result = -1;
    struct tarfs_archive *parc;
    struct vfs_stamping *parent;
    vfs *v;
    struct stat stat_buf;

    local = tarfs_analysis (inname, &archive_name, is_dir);
    if (local == NULL) {
	tarerrno = ENOENT;
	return NULL;
    }

    mc_stat (archive_name, &stat_buf);

    for (parc = tarfs_first_archive; parc != NULL; parc = parc->next)
	if (!strcmp (parc->name, archive_name)) {
/* Has the cached archive been changed on the disk? */
	    if (parc->tarstat.st_mtime < stat_buf.st_mtime) { /* Yes, reload! */
		(*tarfs_vfs_ops.free) ((vfsid) parc);
		vfs_rmstamp (&tarfs_vfs_ops, (vfsid) parc, 0);
		break;
	    }
		    /* Hasn't been modified, give it a new timeout */
	    vfs_stamp (&tarfs_vfs_ops, (vfsid) parc);
	    goto return_success;
	}
    if (do_not_open)
	result = -1;
    else
	result = read_tar_archive (archive_name, &parc);
    if (result == -1) {
	tarerrno = EIO;
	free(local);
	free(archive_name);
	return NULL;
    }
    v = vfs_type (archive_name);
    if (v == &local_vfs_ops) {
	parent = NULL;
    } else {
	parent = x_malloc (sizeof (struct vfs_stamping), "vfs stamping");
	parent->v = v;
	parent->next = 0;
	parent->id = (*v->getid) (archive_name, &(parent->parent));
    }
    vfs_add_noncurrent_stamps (&tarfs_vfs_ops, (vfsid) parc, parent);
    vfs_rm_parents (parent);
return_success:
    *archive = parc;
    free (archive_name);
    return local;
}

struct tarfs_loop_protect {
    struct tarfs_entry *entry;
    struct tarfs_loop_protect *next;
};
static int tarfs_errloop;
static int tarfs_notadir;

static struct tarfs_entry *
__tarfs_find_entry (struct tarfs_entry *dir, char *name,
		    struct tarfs_loop_protect *list, int make_dirs);

static struct tarfs_entry *
__tarfs_resolve_symlinks (struct tarfs_entry *entry,
			  struct tarfs_loop_protect *list)
{
    struct tarfs_entry *pent;
    struct tarfs_loop_protect *looping;

    if (!S_ISLNK (entry->inode->mode))
	return entry;
    for (looping = list; looping != NULL; looping = looping->next)
	if (entry == looping->entry) { /* Here we protect us against symlink looping */
	    tarfs_errloop = 1;
	    return NULL;
	}
    looping = (struct tarfs_loop_protect *)
	x_malloc (sizeof (struct tarfs_loop_protect),
		 "Tar: symlink looping protection");
    looping->entry = entry;
    looping->next = list;
    pent = __tarfs_find_entry (entry->dir, entry->inode->linkname, looping, 0);
    free (looping);
    if (pent == NULL)
	tarerrno = ENOENT;
    return pent;
}

static struct tarfs_entry *tarfs_resolve_symlinks (struct tarfs_entry *entry)
{
    struct tarfs_entry *res;

    tarfs_errloop = 0;
    tarfs_notadir = 0;
    res = __tarfs_resolve_symlinks (entry, NULL);
    if (res == NULL) {
	if (tarfs_errloop)
	    tarerrno = ELOOP;
	else if (tarfs_notadir)
	    tarerrno = ENOTDIR;
    }
    return res;
}

static struct tarfs_entry*
__tarfs_find_entry (struct tarfs_entry *dir, char *name,
		     struct tarfs_loop_protect *list, int make_dirs)
{
    struct tarfs_entry *pent, *pdir;
    char *p, *q, *name_end;
    char c;

    if (*name == '/') { /* Handle absolute paths */
	name++;
	dir = dir->inode->archive->root_entry;
    }

    pent = dir;
    p = name;
    name_end = name + strlen (name);
    q = strchr (p, '/');
    c = '/';
    if (!q)
	q = strchr (p, 0);

    for (; pent != NULL && c && *p; ) {
	c = *q;
	*q = 0;

	if (strcmp (p, ".")) {
	    if (!strcmp (p, ".."))
		pent = pent->dir;
	    else {
		if ((pent = __tarfs_resolve_symlinks (pent, list))==NULL) {
		    *q = c;
		    return NULL;
		}
		if (c == '/' && !S_ISDIR (pent->inode->mode)) {
		    *q = c;
		    tarfs_notadir = 1;
		    return NULL;
		}
		pdir = pent;
		for (pent = pent->inode->first_in_subdir; pent; pent = pent->next_in_dir)
		    /* Hack: I keep the original semanthic unless
		       q+1 would break in the strchr */
		    if (!strcmp (pent->name, p)) {
			if (q + 1 > name_end) {
			    *q = c;
			    tarfs_notadir = !S_ISDIR (pent->inode->mode);
			    return pent;
			}
			break;
		    }

		/* When we load archive, we create automagically
		 * non-existant directories
		 */
		if (pent == NULL && make_dirs) {
		    pent = tarfs_generate_entry (dir->inode->archive, p, pdir, S_IFDIR | 0777);
		}
	    }
	}
	/* Next iteration */
	*q = c;
	p = q + 1;
	q = strchr (p, '/');
	if (!q)
	    q = strchr (p, 0);
    }
    if (pent == NULL)
	tarerrno = ENOENT;
    return pent;
}

static struct tarfs_entry *tarfs_find_entry (struct tarfs_entry *dir, char *name, int make_dirs)
{
    struct tarfs_entry *res;

    tarfs_errloop = 0;
    tarfs_notadir = 0;
    res = __tarfs_find_entry (dir, name, NULL, make_dirs);
    if (res == NULL) {
	if (tarfs_errloop)
	    tarerrno = ELOOP;
	else if (tarfs_notadir)
	    tarerrno = ENOTDIR;
    }
    return res;
}

struct tar_pseudofile {
    struct tarfs_archive *archive;
    off_t pos;
    off_t begin;
    off_t end;
    struct tarfs_entry *entry;
};

static void *tar_open (char *file, int flags, int mode)
{
    struct tar_pseudofile *tar_info;
    struct tarfs_archive *archive;
    char *p, *q;
    struct tarfs_entry *entry;

    if ((p = tarfs_get_path (file, &archive, 0, 0)) == NULL)
	return NULL;
    q = (*p == '/') ? p + 1 : p;
    entry = tarfs_find_entry (archive->root_entry, q, 0);
    free (p);
    if (entry == NULL)
	return NULL;
    if ((entry = tarfs_resolve_symlinks (entry)) == NULL)
	return NULL;
    if (S_ISDIR (entry->inode->mode)) {
	tarerrno = EISDIR;
	return NULL;
    }
    if ((flags & O_ACCMODE) != O_RDONLY) {
	tarerrno = EROFS; /* At the moment we are RO */
	return NULL;
    }

    tar_info = (struct tar_pseudofile *) x_malloc (sizeof (struct tar_pseudofile), "Tar: tar_open");
    tar_info->archive = archive;
    tar_info->pos = 0;
    tar_info->begin = entry->inode->data_offset;
    tar_info->end = tar_info->begin + entry->inode->size;
    tar_info->entry = entry;
    entry->inode->is_open++;

     /* i.e. we had no open files and now we have one */
    vfs_rmstamp (&tarfs_vfs_ops, (vfsid) archive, 1);
    archive->fd_usage++;
    return tar_info;
}

static int tar_read (void *data, char *buffer, int Count)
{
    struct tar_pseudofile *file = (struct tar_pseudofile *)data;

    if (file->archive->is_gzipped != targz_growing &&
	mc_lseek (file->archive->fd, file->begin + file->pos, SEEK_SET) !=
	file->begin + file->pos) {
	tarerrno = EIO;
	return -1;
    }

    if (Count > file->end - file->begin - file->pos)
	Count = file->end - file->begin - file->pos;
    if (file->archive->is_gzipped == targz_growing) {
	char *p = buffer;
	int cnt = Count;
	int i = file->begin + file->pos, j;

	if (i % RECORDSIZE) {
	    j = RECORDSIZE - (i % RECORDSIZE);
	    if (cnt < j)
		j = cnt;
	    memcpy (p,
		   ((char *) find_current_record (file->archive, i / RECORDSIZE * RECORDSIZE)) + (i % RECORDSIZE),
		   j);
	    cnt -= j;
	    p += j;
	    i += j;
	}
	while (cnt) {
	    if (cnt > RECORDSIZE)
		j = RECORDSIZE;
	    else
		j = cnt;
	    memcpy (p, find_current_record (file->archive, i), j);
	    i += j;
	    p += j;
	    cnt -= j;
	}
    }
    else if ((Count = mc_read (file->archive->fd, buffer, Count)) == -1) {
	tarerrno = errno;
	return -1;
    }
    file->pos += Count;
    return Count;
}

static int tar_close (void *data)
{
    struct tar_pseudofile *file;

    file = (struct tar_pseudofile *)data;

    file->archive->fd_usage--;
    if (!file->archive->fd_usage) {
	struct vfs_stamping *parent;
	vfs *v;

	v = vfs_type (file->archive->name);
	if (v == &local_vfs_ops) {
	    parent = NULL;
	} else {
	    parent = x_malloc (sizeof (struct vfs_stamping), "vfs stamping");
	    parent->v = v;
	    parent->next = 0;
	    parent->id = (*v->getid) (file->archive->name, &(parent->parent));
	}
	vfs_add_noncurrent_stamps (&tarfs_vfs_ops, (vfsid) (file->archive), parent);
	vfs_rm_parents (parent);
    }
    (file->entry->inode->is_open)--;

    free (data);
    return 0;
}


static void *tar_opendir (char *Dirname)
{
    struct tarfs_archive *archive;
    char *p, *q;
    struct tarfs_entry *entry;
    struct tarfs_entry **tar_info;

    if ((p = tarfs_get_path (Dirname, &archive, 1, 0)) == NULL)
	return NULL;
    q = (*p == '/') ? p + 1 : p;
    entry = tarfs_find_entry (archive->root_entry, q, 0);
    free (p);
    if (entry == NULL)
	return NULL;
    if ((entry = tarfs_resolve_symlinks (entry)) == NULL)
	return NULL;
    if (!S_ISDIR (entry->inode->mode)) {
	tarerrno = ENOTDIR;
	return NULL;
    }

    tar_info = (struct tarfs_entry **) x_malloc (sizeof (struct tarfs_entry *), "Tar: tar_opendir");
    *tar_info = entry->inode->first_in_subdir;

    return tar_info;
}

static void *tar_readdir (void *data)
{
    static struct {
	struct dirent dir;
    } dir;

    struct tarfs_entry **tar_info = (struct tarfs_entry **) data;

    if (*tar_info == NULL)
	return NULL;

    strcpy (&(dir.dir.d_name [0]), (*tar_info)->name);

    *tar_info = (*tar_info)->next_in_dir;

    return (void *)&dir;
}


static int _tar_stat (const char *path, struct stat *buf, int resolve)
{
    struct tarfs_archive *archive;
    char *p, *q;
    struct tarfs_entry *entry;
    struct tarfs_inode *inode;

    if ((p = tarfs_get_path (path, &archive, 0, 0)) == NULL)
	return -1;
    q = (*p == '/') ? p + 1 : p;
    entry = tarfs_find_entry (archive->root_entry, q, 0);
    free (p);
    if (entry == NULL)
	return -1;
    if (resolve && (entry = tarfs_resolve_symlinks (entry)) == NULL)
	return -1;
    inode = entry->inode;
    buf->st_dev = inode->dev;
    buf->st_ino = inode->inode;
    buf->st_mode = inode->mode;
    buf->st_nlink = inode->nlink;
    buf->st_uid = inode->uid;
    buf->st_gid = inode->gid;
    buf->st_rdev = inode->rdev;
    buf->st_size = inode->size;
    buf->st_blksize = RECORDSIZE;
    buf->st_blocks = (inode->size + RECORDSIZE - 1) / RECORDSIZE;
    buf->st_atime = inode->atime;
    buf->st_mtime = inode->mtime;
    buf->st_ctime = inode->ctime;
    return 0;
}

static int tar_stat (const char *path, struct stat *buf)
{
    return _tar_stat (path, buf, 1);
}

static int tar_lstat (const char *path, struct stat *buf)
{
    return _tar_stat (path, buf, 0);
}

static int tar_fstat (void *data, struct stat *buf)
{
    struct tar_pseudofile *file = (struct tar_pseudofile *)data;
    struct tarfs_inode *inode;

    inode = file->entry->inode;
    buf->st_dev = inode->dev;
    buf->st_ino = inode->inode;
    buf->st_mode = inode->mode;
    buf->st_nlink = inode->nlink;
    buf->st_uid = inode->uid;
    buf->st_gid = inode->gid;
    buf->st_rdev = inode->rdev;
    buf->st_size = inode->size;
    buf->st_blksize = RECORDSIZE;
    buf->st_blocks = (inode->size + RECORDSIZE - 1) / RECORDSIZE;
    buf->st_atime = inode->atime;
    buf->st_mtime = inode->mtime;
    buf->st_ctime = inode->ctime;
    return 0;
}


static int tar_readlink (char *path, char *buf, int size)
{
    struct tarfs_archive *archive;
    char *p, *q;
    int i;
    struct tarfs_entry *entry;

    if ((p = tarfs_get_path (path, &archive, 0, 0)) == NULL)
	return -1;
    q = (*p == '/') ? p + 1 : p;
    entry = tarfs_find_entry (archive->root_entry, q, 0);
    free (p);
    if (entry == NULL)
	return -1;
    if (!S_ISLNK (entry->inode->mode)) {
	tarerrno = EINVAL;
	return -1;
    }
    if (size > (i = strlen (entry->inode->linkname))) {
	size = i;
    }
    strncpy (buf, entry->inode->linkname, i);
    return i;
}

static int tar_chdir (char *path)
{
    struct tarfs_archive *archive;
    char *p, *q;
    struct tarfs_entry *entry;

    tarerrno = ENOTDIR;
    if ((p = tarfs_get_path (path, &archive, 1, 0)) == NULL)
	return -1;
    q = (*p == '/') ? p + 1 : p;
    entry = tarfs_find_entry (archive->root_entry, q, 0);
    if (entry == NULL) {
	free (p);
	return -1;
    }
    entry = tarfs_resolve_symlinks (entry);
    if (entry == NULL) {
	free (p);
	return -1;
    }
    if (!S_ISDIR (entry->inode->mode)) {
	free (p);
	return -1;
    }
    entry->inode->archive->current_dir = entry;
    free (p);
    tarerrno = 0;
    return 0;
}

static int tar_lseek (void *data, off_t offset, int whence)
{
    struct tar_pseudofile *file = (struct tar_pseudofile *) data;

    switch (whence) {
	case SEEK_CUR:
	    offset += file->pos; break;
	case SEEK_END:
	    offset += file->end - file->begin; break;
    }
    if (offset < 0)
	file->pos = 0;
    else if (offset < file->end - file->begin)
	file->pos = offset;
    else
	file->pos = file->end;
    return file->pos;
}


static vfsid tar_getid (char *path, struct vfs_stamping **parent)
{
    struct tarfs_archive *archive;
    vfs *v;
    vfsid id;
    char *p;
    struct vfs_stamping *par;

    *parent = NULL;
    if ((p = tarfs_get_path (path, &archive, 0, 1)) == NULL) {
	return (vfsid) -1;
    }
    free (p);
    v = vfs_type (archive->name);
    id = (*v->getid) (archive->name, &par);
    if (id != (vfsid)-1) {
	*parent = x_malloc (sizeof (struct vfs_stamping), "vfs stamping");
	(*parent)->v = v;
	(*parent)->id = id;
	(*parent)->parent = par;
	(*parent)->next = NULL;
    }
    return (vfsid) archive;
}

static void tarfs_free_entry (struct tarfs_entry *e)
{
    int i = --(e->inode->nlink);
    if (S_ISDIR (e->inode->mode) && e->inode->first_in_subdir != NULL) {
	struct tarfs_entry *f = e->inode->first_in_subdir;

	e->inode->first_in_subdir = NULL;
	tarfs_free_entry (f);
    }
    if (i <= 0) {
	if (e->inode->linkname != NULL)
	    free (e->inode->linkname);
	if (e->inode->local_filename != NULL) {
	    unlink (e->inode->local_filename);
	    free (e->inode->local_filename);
	}
	free (e->inode);
    }
    if (e->next_in_dir != NULL)
	tarfs_free_entry (e->next_in_dir);
    free (e->name);
    free (e);
}

static void tar_free (vfsid id)
{
    struct tarfs_archive *parc;
    struct tarfs_archive *archive = (struct tarfs_archive *)id;

    tarfs_free_entry (archive->root_entry);
    if (archive == tarfs_first_archive) {
	tarfs_first_archive = archive->next;
    } else {
	for (parc = tarfs_first_archive; parc != NULL; parc = parc->next)
	    if (parc->next == archive)
		break;
	if (parc != NULL)
	    parc->next = archive->next;
    }
    tarfs_free_archive (archive);
}

static char *tar_getlocalcopy (char *path)
{
    struct tarfs_archive *archive;
    char *p, *q;
    struct tarfs_entry *entry;

    if ((p = tarfs_get_path (path, &archive, 1, 0)) == NULL)
	return NULL;
    q = (*p == '/') ? p + 1 : p;
    entry = tarfs_find_entry (archive->root_entry, q, 0);
    free (p);
    if (entry == NULL)
	return NULL;
    if ((entry = tarfs_resolve_symlinks (entry)) == NULL)
	return NULL;

    if (entry->inode->local_filename != NULL)
	return entry->inode->local_filename;
    p = mc_def_getlocalcopy (path);
    if (p != NULL) {
	entry->inode->local_filename = p;
    }
    return p;
}


#ifdef USE_NETCODE
#if defined(IS_AIX) || defined(linux) || defined(SCO_FLAVOR)
static void sig_pipe (int unused)
#else
static void sig_pipe (void)
#endif
{
    got_sigpipe = 1;
}

static void tcp_init (void)
{
    bb_signal((1 << SIGPIPE), sig_pipe);
    got_sigpipe = 0;
}
#endif

#ifdef USE_NETCODE
/* Extract the hostname and username from the path */
/* path is in the form: [user@]hostname:port/remote-dir, e.g.:
 *
 * ftp://sunsite.unc.edu/pub/linux
 * ftp://miguel@sphinx.nuclecu.unam.mx/c/nc
 * ftp://tsx-11.mit.edu:8192/
 * ftp://joe@foo.edu:11321/private
 * ftp://joe:password@foo.se
 *
 * If the user is empty, e.g. ftp://@roxanne/private, then your login name
 * is supplied.
 *
 * returns malloced host, user.
 * returns a malloced strings with the pathname relative to the host.
 * */

static char *get_host_and_username (char *path, char **host, char **user,
	int *port, int default_port, int default_is_anon, char **pass)
{
    struct passwd *passwd_info;
    char *dir, *colon, *inner_colon, *at, *rest;
    char *retval;
    char *pcopy = strdup (path);
    char *pend   = pcopy + strlen (pcopy);

    *pass = NULL;
    *port = default_port;
    *user = NULL;

    /* locate path component */
    for (dir = pcopy; *dir != '/' && *dir; dir++)
	;
    if (*dir) {
	retval = strdup (dir);
	*dir = 0;
    } else
	retval = strdup ("/");

    /* search for any possible user */
    at = strchr_set0 (pcopy, '@');

    /* We have a username */
    if (at) {
	inner_colon = strchr (pcopy, ':');
	if (inner_colon) {
	    *inner_colon = 0;
	    inner_colon++;
	    if (*inner_colon == '@')
		*pass = NULL;
	    else
		*pass = strdup (inner_colon);
	}
	if (*pcopy != 0)
	    *user = strdup (pcopy);
	else
	    default_is_anon = 0;

	if (pend == at+1)
	    rest = at;
	else
	    rest = at + 1;
    } else
	rest = pcopy;

    if (!*user) {
	if (default_is_anon)
	    *user = "anonymous";
	else
	    *user = bb_getpwuid ("anonymous", my_euid, 0);
	*user = strdup (*user);
    }
    /* Check if the host comes with a port spec, if so, chop it */
    colon = strchr (rest, ':');
    if (colon) {
	*port = atoi (colon+1);

	if (*port <= 0 || *port >= 65536)
	    *port = 21;
	*colon = 0;
    }
    *host = strdup (rest);

    free (pcopy);
    return retval;
}
#endif

/* They keep track of the current directory */
static vfs *current_vfs = &local_vfs_ops;
static char *current_dir;
static int current_mon;
static int current_year;

/* Open files managed by the vfs layer */
#define MAX_VFS_FILES 100
static struct {
    void *fs_info;
    vfs  *operations;
} vfs_file_table [MAX_VFS_FILES];

static int get_bucket (void)
{
    int i;

    /* 0, 1, 2 are reserved file descriptors, while (DIR *) 0 means error */
    for (i = 3; i < MAX_VFS_FILES; i++) {
	if (!vfs_file_table [i].fs_info)
	    return i;
    }
    fprintf (stderr, "No more virtual file handles\n");
    exit (1);
}

/* This flag is set on each vfs_type call to say if the path was
   absolute (1) or relative (0) to current dir. */

static vfs *vfs_type (const char *path)
{
    vfs *Vfs;

    Vfs = current_vfs;

    if (*path == '/')
	Vfs = &local_vfs_ops;

    if (strncmp (path, "local:", 6) == 0)
	Vfs = &local_vfs_ops;

#ifdef USE_NETCODE
    if (strncmp (path, "ftp://", 6) == 0)
	Vfs = &ftpfs_vfs_ops;
#endif

    if (strncmp (path, "tar:", 4) == 0)
	Vfs = &tarfs_vfs_ops;

    if (extfs_prefix_to_type (path) != -1)
	Vfs = &extfs_vfs_ops;
    return Vfs;
}

static char *vfs_path (char *path)
{
    int j;

    if (strncmp (path, "local:", 6) == 0)
	return path + 6;

#ifdef USE_NETCODE
    if (strncmp (path, "ftp://", 6) == 0)
	return strchr(path + 6, '/');
#endif

    if (strncmp (path, "tar:", 4) == 0)
	return path + 4;

    j = extfs_prefix_to_type (path);
    if (j != -1)
	return path + strlen (extfs_get_prefix (j)) + 1;

    return path;
}

static struct vfs_stamping *stamps;

/* Returns the number of seconds remaining to the vfs timeout
 *
 * FIXME: currently this is set to 10 seconds.  We should compute this.
 */
static int vfs_timeouts (void)
{
    return stamps ? 10 : 0;
}

static void vfs_addstamp (vfs *v, vfsid id, struct vfs_stamping *parent)
{
    if (v != &local_vfs_ops && id != (vfsid)-1) {
	struct vfs_stamping *stamp = stamps, *st1 = stamp;

	for (stamp = stamps; stamp != NULL; st1 = stamp, stamp = stamp->next)
	    if (stamp->v == v && stamp->id == id) {
		gettimeofday(&(stamp->time), NULL);
		return;
	    }
	stamp = x_malloc (sizeof (struct vfs_stamping), "vfs stamping");
	stamp->v = v;
	stamp->id = id;
	if (parent) {
	    struct vfs_stamping *st = stamp;
	    for ( ; parent; ) {
		st->parent = x_malloc (sizeof (struct vfs_stamping), "vfs stamping");
		*st->parent = *parent;
		parent = parent->parent;
		st = st->parent;
	    }
	    st->parent = 0;
	}
	else
	    stamp->parent = 0;
	gettimeofday (&(stamp->time), NULL);
	stamp->next = 0;
	if (stamps)
	    st1->next = stamp;
	else
	    stamps = stamp;
    }
}

static void vfs_stamp (vfs *v, vfsid id)
{
    struct vfs_stamping *stamp;

    for (stamp = stamps; stamp != NULL; stamp = stamp->next)
	if (stamp->v == v && stamp->id == id) {
	    gettimeofday (&(stamp->time), NULL);
	    if (stamp->parent != NULL)
		vfs_stamp (stamp->parent->v, stamp->parent->id);
	    return;
	}
}

static void vfs_rm_parents (struct vfs_stamping *stamp)
{
    struct vfs_stamping *st2, *st3;
    if (stamp) {
	for (st2 = stamp, st3 = st2->parent; st3 != NULL; st2 = st3, st3 = st3->parent)
	    free (st2);
	free (st2);
    }
}

static void vfs_rmstamp (vfs *v, vfsid id, int removeparents)
{
    struct vfs_stamping *stamp, *st1;

    for (stamp = stamps, st1 = NULL; stamp != NULL; st1 = stamp, stamp = stamp->next)
	if (stamp->v == v && stamp->id == id) {
	    if (stamp->parent != NULL) {
		if (removeparents)
		    vfs_rmstamp (stamp->parent->v, stamp->parent->id, 1);
		vfs_rm_parents (stamp->parent);
	    }
	    if (st1 == NULL) {
		stamps = stamp->next;
	    } else {
		st1->next = stamp->next;
	    }
	    free (stamp);
	    return;
	}
}

static int mc_open (char *file, int flags, ...)
{
    int  handle;
    int  mode;
    void *info;
    vfs  *Vfs;
    va_list ap;

    file = vfs_canon (file);
    Vfs = vfs_type (file);

    /* Get the mode flag */
    va_start (ap, flags);
    mode = va_arg (ap, int);
    va_end (ap);

    info = (*Vfs->open) (file, flags, mode);
    if (!info) {
	errno = (*Vfs->ferrno)();
	free (file);
	return -1;
    }
    handle = get_bucket ();
    vfs_file_table [handle].fs_info = info;
    vfs_file_table [handle].operations = Vfs;

    free (file);
    return handle;
}

#define vfs_op(handle) vfs_file_table [handle].operations
#define vfs_info(handle) vfs_file_table [handle].fs_info
#define vfs_free_bucket(handle) vfs_info(handle) = 0;

static int mc_read (int handle, char *buffer, int Count)
{
    vfs *Vfs;
    int result;

    if (handle == -1)
	return -1;

    Vfs = vfs_op (handle);
    result = (*Vfs->read)(vfs_info (handle), buffer, Count);
    if (result == -1) {
	errno = (*Vfs->ferrno)();
	return -1;
    }
    return result;
}

static int mc_ctl (int handle, int ctlop, int arg)
{
    vfs *Vfs;
    int result;

    Vfs = vfs_op (handle);
    if (Vfs->ctl == NULL)
	return 0;
    result = (*Vfs->ctl)(vfs_info (handle), ctlop, arg);
    return result;
}

int mc_setctl (char *path, int ctlop, char *arg)
{
    vfs *Vfs;
    int result;

    path = vfs_canon (path);
    Vfs = vfs_type (path);
    if (Vfs->setctl == NULL) {
	free (path);
	return 0;
    }
    result = (*Vfs->setctl)(path, ctlop, arg);
    free (path);
    return result;
}

static int mc_close (int handle)
{
    vfs *Vfs;
    int result;

    if (handle == -1 || !vfs_info (handle))
	return -1;

    Vfs = vfs_op (handle);
    if (handle < 3)
	return close (handle);

    result = (*Vfs->close)(vfs_info (handle));
    vfs_free_bucket (handle);
    if (result == -1)
	errno = (*Vfs->ferrno)();

    return result;
}

static DIR *mc_opendir (char *Dirname)
{
    int  handle, *handlep;
    void *info;
    vfs  *Vfs;
    char *p = NULL;
    int i = strlen (Dirname);

    if (Dirname [i - 1] != '/') {
	/* We should make possible reading of the root directory in a tar file */
	p = xasprintf ("%s/", Dirname);
	Dirname = p;
    }
    Dirname = vfs_canon (Dirname);
    Vfs = vfs_type (Dirname);

    info = (*Vfs->opendir)(Dirname);
    if (!info) {
	errno = (*Vfs->ferrno)();
	free (Dirname);
	if (p)
	    free (p);
	return NULL;
    }
    handle = get_bucket ();
    vfs_file_table [handle].fs_info = info;
    vfs_file_table [handle].operations = Vfs;

    free (Dirname);
    if (p)
	free (p);

    handlep = (int *) x_malloc (sizeof (int), "opendir handle");
    *handlep = handle;
    return (DIR *) handlep;
}

/* This should strip the non needed part of a path name */
#define vfs_name(x) x

static struct dirent *mc_readdir(DIR *dirp)
{
    int handle;
    vfs *Vfs;

    if (!dirp) {
#ifdef EBADF
	errno = EBADF;
#else
	errno = 1;
#endif
	return NULL;
    }
    handle = *(int *) dirp;
    Vfs = vfs_op (handle);
    return (*Vfs->readdir)(vfs_info (handle));
}

static int mc_closedir (DIR *dirp)
{
    int handle = *(int *) dirp;
    vfs *Vfs = vfs_op (handle);
    int result;

    result = (*Vfs->closedir)(vfs_info (handle));
    vfs_free_bucket (handle);
    free (dirp);
    return result;
}

static int mc_stat (const char *path, struct stat *buf)
{
    vfs *Vfs;
    int result;
    char *p;

    p = vfs_canon (path);
    Vfs = vfs_type (p);
    result = (*Vfs->stat)(vfs_name (p), buf);
    free (p);
    if (result == -1) {
	errno = (*Vfs->ferrno)();
	return -1;
    }
    return result;
}

static int mc_lstat (const char *path, struct stat *buf)
{
    vfs *Vfs;
    int result;
    char *p;

    p = vfs_canon (path);
    Vfs = vfs_type (p);
    result = (*Vfs->lstat)(vfs_name (p), buf);
    free (p);
    if (result == -1) {
	errno = (*Vfs->ferrno)();
	return -1;
    }
    return result;
}

static int mc_fstat (int handle, struct stat *buf)
{
    vfs *Vfs;
    int result;

    if (handle == -1)
	return -1;

    Vfs = vfs_op (handle);
    result = (*Vfs->fstat)(vfs_info (handle), buf);
    if (result == -1) {
	errno = (*Vfs->ferrno)();
	return -1;
    }
    return result;
}

/* FIXME: We should return the buffer, not set it, since it could
   overflow */
static char *mc_get_current_wd (char *buffer, int size)
{
    char *p;
    struct stat my_stat, my_stat2;

    if (current_vfs == &local_vfs_ops) {
	p = getcwd (buffer, size);
	if (!p) { /* One of the directories in the path is not readable */
	    strcpy (buffer, current_dir);
	    return buffer;
	}

	/* Otherwise check if it is O.K. to use the current_dir */
	mc_stat (buffer, &my_stat);
	mc_stat (current_dir, &my_stat2);
	if (my_stat.st_ino != my_stat2.st_ino ||
	    my_stat.st_dev != my_stat2.st_dev ||
	    !cd_symlinks) {
	    if (last_current_dir != current_dir)
		free (last_current_dir);
	    free (current_dir);
	    current_dir = strdup (p);
	    last_current_dir = current_dir;
	    return p;
	} /* Otherwise we return current_dir below */
    }
    strcpy (buffer, current_dir);
    return buffer;
}

static int mc_chmod (char *path, int mode)
{
    vfs *Vfs;
    int result;

    path = vfs_canon (path);
    Vfs = vfs_type (path);
    result = (*Vfs->chmod)(vfs_name (path), mode);
    free (path);
    if (result == -1) {
	errno = (*Vfs->ferrno)();
	return -1;
    }
    return result;
}

static int mc_chown (char *path, int owner, int group)
{
    vfs *Vfs;
    int result;

    path = vfs_canon (path);
    Vfs = vfs_type (path);
    result = (*Vfs->chown)(vfs_name (path), owner, group);
    free (path);
    if (result == -1) {
	errno = (*Vfs->ferrno)();
	return -1;
    }
    return result;
}

static int mc_utime (char *path, struct utimbuf *times)
{
    vfs *Vfs;
    int result;

    path = vfs_canon (path);
    Vfs = vfs_type (path);
    if (!Vfs->utime) {
	free (path);
	return 0;
    }
    result = (*Vfs->utime)(vfs_name (path), times);
    free (path);
    if (result == -1) {
	errno = (*Vfs->ferrno)();
	return -1;
    }
    return result;
}

static int mc_readlink(char *path, char *buf, int bufsiz)
{
    vfs *Vfs;
    int result;

    path = vfs_canon (path);
    Vfs = vfs_type (path);
    result = (*Vfs->readlink)(vfs_name (path), buf, bufsiz);
    free (path);
    if (result == -1) {
	errno = (*Vfs->ferrno)();
	return -1;
    }
    return result;
}

static int mc_unlink (char *path)
{
    vfs *Vfs;
    int result;

    path = vfs_canon (path);
    Vfs = vfs_type (path);
    result = (*Vfs->unlink)(vfs_name (path));
    free (path);
    if (result == -1) {
	errno = (*Vfs->ferrno)();
	return -1;
    }
    return result;
}

static int mc_symlink (char *name1, char *name2)
{
    vfs *vfs2 = 0;
    int result;

    name2 = vfs_canon (name2);
    vfs2 = vfs_type (name2);

    result = (*vfs2->symlink)(vfs_name (name1), vfs_name (name2));
    free (name2);
    if (result == -1) {
	errno = (*vfs2->ferrno)();
	return -1;
    }
    return result;
}

static int mc_link (char *name1, char *name2)
{
    vfs *vfs1;
    vfs *vfs2;
    int result;

    name1 = vfs_canon (name1);
    vfs1 = vfs_type (name1);
    name2 = vfs_canon (name2);
    vfs2 = vfs_type (name2);

    if (vfs1 != vfs2) {
	errno = EXDEV;
	free (name1);
	free (name2);
	return -1;
    }

    result = (*vfs1->link)(vfs_name (name1), vfs_name (name2));
    free (name1);
    free (name2);
    if (result == -1) {
	errno = (*vfs1->ferrno)();
	return -1;
    }
    return result;
}

static int mc_write (int fd, char *buf, int nbyte)
{
    vfs *Vfs;
    int result;

    if (fd == -1)
	return -1;

    Vfs = vfs_op (fd);
    result = (*Vfs->write)(vfs_info (fd), buf, nbyte);
    if (result == -1) {
	errno = (*Vfs->ferrno)();
	return -1;
    }
    return result;
}

static int mc_rename (char *path1, char *path2)
{
    vfs *vfs1;
    vfs *vfs2;
    int result;

    path1 = vfs_canon (path1);
    vfs1 = vfs_type (path1);
    path2 = vfs_canon (path2);
    vfs2 = vfs_type (path2);

    if (vfs1 != vfs2) {
	errno = EXDEV;
	free (path1);
	free (path2);
	return -1;
    }

    result = (*vfs1->rename)(vfs_name (path1), vfs_name (path2));
    free (path1);
    free (path2);
    if (result == -1) {
	errno = (*vfs1->ferrno)();
	return -1;
    }
    return result;
}

static off_t mc_lseek (int fd, off_t offset, int whence)
{
    vfs *Vfs;

    if (fd == -1)
	return -1;

    Vfs = vfs_op (fd);
    return (*Vfs->lseek)(vfs_info (fd), offset, whence);
}

static char *vfs_canon (const char *path)
{
    vfs *Vfs;

    Vfs = current_vfs;

    if (strncmp (path, "local:", 6) == 0) {
	return vfs_canon (path+6);

#ifdef USE_NETCODE
    } else if (strncmp (path, "ftp://", 6) == 0) {
	Vfs = &ftpfs_vfs_ops;
#endif
    } else if (strncmp (path, "tar:", 4) == 0) {
	Vfs = &tarfs_vfs_ops;
    } else if (extfs_prefix_to_type (path) != -1) {
	Vfs = &extfs_vfs_ops;
    } else if (*path == '/') { /* Absolute local */
	Vfs = &local_vfs_ops;
    } else if (*path == '~') { /* Tilde expansion */
	char *local, *result;

	local = tilde_expand (path);
	if (local) {
	    result = vfs_canon (local);
	    free (local);
	    return result;
	} else
	    return strdup (path);
    } else { /* Relative to current directory */
	char *result;
	char *local = concat_path_file(current_dir, path);
	result = vfs_canon (local);
	free (local);
	return result;
    }

    if (Vfs == &local_vfs_ops)
/* NOTE: No local:/ is necessary at the moment, we are handling absolute paths
	 at the moment, even it is impossible, since it doesn't work then at all. */
	return canonicalize_pathname (strdup (path));
#ifdef USE_NETCODE
    if (Vfs == &ftpfs_vfs_ops) {
	char *q, *p, *r, *s;
	int prefixsh, isftp;

	if (Vfs == &ftpfs_vfs_ops) {
	    prefixsh = 6;
	    isftp = 1;
	}

	q = strdup (path);
	p = strchr (q + prefixsh, '/');

	if (p != NULL) {
	    if (p [0] && p [1] == '~') { /* Tilde expansion */
		*p = 0;
		if (isftp) {
		    s = ftpfs_gethome (q);
		    if (s == NULL)
			s = "/";
		}
		if (p [2] == '/')
		    p += 3;
		else
		    p += 2;
		r = xasprintf ("%s%s%s", q, s, p);
		free (q);
		q = r;
		p = strchr (q + prefixsh, '/');
	    }
	    p++;
	    if (p [0])
		canonicalize_pathname (p);
	    else
		p = ".";

	    if (!strncmp (p, "..", 2) && (p [2] == '/' || !p [2])) {
		*p = 0;
		r = ftpfs_getupdir (q);
		if (r == NULL)
		    r = "/";
		p += 2;
		if (*p == '/')
		    p++;
		r = xasprintf ("%s%s", r, p);
		free (q);
		q = vfs_canon (r);
		free (r);
	    }
	}
	return q;
    }
#endif
    if (Vfs == &tarfs_vfs_ops || Vfs == &extfs_vfs_ops) {
	char *local_path, *arc_name, *p, *q, *result, *extfs_prefix = NULL;
	int istar, extfs_type;

	/* There has to be a leading tar: or extfs prefix in the path */

	if (Vfs == &tarfs_vfs_ops)
	    istar = 1;
	else
	    istar = 0;

	/* The original code does not compile right in NeXTStep:
	 * istar = (Vfs == &tarfs_vfs_ops);
	 */
	if (istar)
	    local_path = tarfs_analysis (path, &arc_name, 0);
	else {
	    local_path = extfs_analysis (path, &arc_name, &extfs_type, 0);
	    extfs_prefix = extfs_get_prefix (extfs_type);
	}
	if (local_path == NULL) {
	    return strdup ("/"); /* For error cases cd to the root dir */
	}
	p = (*local_path == '/') ? local_path + 1 : local_path;
	if (*p)
	    canonicalize_pathname (p);
	if (!strcmp (p, ".")) /* This is an incorrect result of canonicalization,
				 if we have p = somedir/.. */
	    *p = 0;
	if (p [0] != '.' || p [1] != '.' || (p [2] && p [2] != '/')) {
	    /* I.e. p doesn't start with the .. directory */
	    if (istar)
		result = xasprintf ("tar:%s/%s", arc_name, p);
	    else {
		/* Two chances: the extfs requires an archive to
		 * use or it does not need it:
		 */

		/* If it does need an archive... */
		if (arc_name && *arc_name)
		    result = xasprintf("%s:/%s", extfs_prefix, arc_name, p);
		else
		    result = canonicalize_pathname (strdup(path));
	    }
	    free (local_path);
	    if (arc_name)
		free (arc_name);
	    return result;
	} else {
	    /* We want to move outside of the archive */
	    if (!istar && (!arc_name || !*arc_name)) {
		free (local_path);
		if (arc_name) free (arc_name);
		return strdup ("/");
	    }
	    q = bb_basename (arc_name);
	    *q = 0;
	    if (p [2] == '/')
		p += 3;
	    else
		p += 2;
	    result = xasprintf ("%s%s", arc_name, p);
	    free (local_path);
	    free (arc_name);
	    local_path = vfs_canon (result);
	    free (result);
	    return local_path;
	}
    }

    fprintf (stderr, "Could not happend\n");
    return 0;
}

vfsid vfs_ncs_getid (vfs *nvfs, char *dir, struct vfs_stamping **par)
{
    vfsid nvfsid;
    int freeit = 0;

    if (dir [strlen (dir) - 1] != '/') {
	dir = xasprintf ("%s/", dir);
	freeit = 1;
    }
    nvfsid = (*nvfs->getid)(dir, par);
    if (freeit)
	free (dir);
    return nvfsid;
}

static int is_parent (vfs * nvfs, vfsid nvfsid, struct vfs_stamping *parent)
{
    struct vfs_stamping *stamp;
    for (stamp = parent; stamp; stamp = stamp->parent)
	if (stamp->v == nvfs && stamp->id == nvfsid)
	    break;
    return (stamp ? 1 : 0);
}

void vfs_add_noncurrent_stamps (vfs * oldvfs, vfsid oldvfsid, struct vfs_stamping *parent)
{
    vfs *nvfs, *n2vfs, *n3vfs;
    vfsid nvfsid, n2vfsid, n3vfsid;
    struct vfs_stamping *par, *stamp;
    int f;

    /* FIXME: As soon as we convert to multiple panels, this stuff
       has to change. It works like this: We do not time out the
       vfs's which are current in any panel and on the other
       side we add the old directory with all its parents which
       are not in any panel (if we find such one, we stop adding
       parents to the time-outing structure. */

    /* There are three directories we have to take care of: current_dir,
       cpanel->cwd and opanel->cwd. Athough most of the time either
       current_dir and cpanel->cwd or current_dir and opanel->cwd are the
       same, it's possible that all three are different -- Norbert */

    if (!cpanel)
	return;

    nvfs = vfs_type (current_dir);
    nvfsid = vfs_ncs_getid (nvfs, current_dir, &par);
    vfs_rmstamp (nvfs, nvfsid, 1);

    f = is_parent (oldvfs, oldvfsid, par);
    vfs_rm_parents (par);
    if ((nvfs == oldvfs && nvfs == oldvfsid) || oldvfsid == (vfsid *)-1 || f) {
	return;
    }

    if (get_current_type () == view_listing) {
	n2vfs = vfs_type (cpanel->cwd);
	n2vfsid = vfs_ncs_getid (n2vfs, cpanel->cwd, &par);
	f = is_parent (oldvfs, oldvfsid, par);
	vfs_rm_parents (par);
	if ((n2vfs == oldvfs && n2vfsid == oldvfsid) || f)
	    return;
    } else {
	n2vfs = (vfs *) -1;
	n2vfsid = (vfs *) -1;
    }

    if (get_other_type () == view_listing) {
	n3vfs = vfs_type (opanel->cwd);
	n3vfsid = vfs_ncs_getid (n3vfs, opanel->cwd, &par);
	f = is_parent (oldvfs, oldvfsid, par);
	vfs_rm_parents (par);
	if ((n3vfs == oldvfs && n3vfsid == oldvfsid) || f)
	    return;
    } else {
	n3vfs = (vfs *)-1;
	n3vfsid = (vfs *)-1;
    }

    if ((*oldvfs->nothingisopen) (oldvfsid)) {
	if (oldvfs == &extfs_vfs_ops && ((struct extfs_archive *) oldvfsid)->name == 0) {
	    /* Free the resources immediatly when we leave a mtools fs
	       ('cd a:') instead of waiting for the vfs-timeout */
	    (oldvfs->free) (oldvfsid);
	} else
	    vfs_addstamp (oldvfs, oldvfsid, parent);
	for (stamp = parent; stamp != NULL; stamp = stamp->parent) {
	    if ((stamp->v == nvfs && stamp->id == nvfsid) ||
		(stamp->v == n2vfs && stamp->id == n2vfsid) ||
		(stamp->v == n3vfs && stamp->id == n3vfsid) ||
		stamp->id == (vfsid) - 1 ||
		!(*stamp->v->nothingisopen) (stamp->id))
		break;
	    if (stamp->v == &extfs_vfs_ops && ((struct extfs_archive *) stamp->id)->name == 0) {
		(stamp->v->free) (stamp->id);
		vfs_rmstamp (stamp->v, stamp->id, 0);
	    } else
		vfs_addstamp (stamp->v, stamp->id, stamp->parent);
	}
    }
}

static void vfs_stamp_path (char *path)
{
    vfs *Vfs;
    vfsid id;
    struct vfs_stamping *par, *stamp;

    Vfs = vfs_type (path);
    id  = vfs_ncs_getid (Vfs, path, &par);
    vfs_addstamp (Vfs, id, par);

    for (stamp = par; stamp != NULL; stamp = stamp->parent)
	vfs_addstamp (stamp->v, stamp->id, stamp->parent);
    vfs_rm_parents (par);
}

void vfs_add_current_stamps (void)
{
    vfs_stamp_path (current_dir);
    if (get_current_type () == view_listing)
	vfs_stamp_path (cpanel->cwd);
    if (get_other_type () == view_listing)
	vfs_stamp_path (opanel->cwd);
}

/* This function is really broken */
int mc_chdir (char *path)
{
    char *a;
    int result;
    char *p = NULL;
    int i = strlen (path);
    vfs *oldvfs;
    vfsid oldvfsid;
    struct vfs_stamping *parent;

    if (path [i - 1] != '/') {
    /* We should make possible reading of the root directory in a tar archive */
	p = x_malloc (i + 2, "slash");
	strcpy (p, path);
	strcpy (p + i, "/");
	path = p;
    }

    if (!current_dir) {
	current_dir = strdup ("/"); /* Note: we should set current_dir
				       on startup to a reasonable pwd.
				       It cannot be relative, i.e. no
				       . or anything else except
				       /, tar:, mcfs:, local:, ftp:,
				       $extfs_prefix .
				       This is done by the vfs_setup_wd call
				       in main.c */
	last_current_dir = current_dir;
    }

    a = current_dir; /* Save a copy for case of failure */
    if (last_current_dir != current_dir)
	free (last_current_dir);
    last_current_dir = a;
    current_dir = vfs_canon (path);
    current_vfs = vfs_type (current_dir);
    result = (*current_vfs->chdir)(vfs_name (current_dir));
    if (result == -1) {
	errno = (*current_vfs->ferrno)();
	free (current_dir);
	current_vfs = vfs_type (a);
	current_dir = a;
    } else {
	last_current_dir = current_dir;
	oldvfs = vfs_type (a);
	oldvfsid = vfs_ncs_getid (oldvfs, a, &parent);
	free (a);
	vfs_add_noncurrent_stamps (oldvfs, oldvfsid, parent);
	vfs_rm_parents (parent);
    }
    if (p) {
	free (p);
    }
    if (*current_dir) {
	p = strchr (current_dir, 0) - 1;
	if (*p == '/' && p > current_dir)
	    *p = 0; /* Sometimes we assume no trailing slash on cwd */
    }
    return result;
}

int vfs_current_is_local (void)
{
    return current_vfs == &local_vfs_ops;
}

int vfs_current_is_extfs (void)
{
    return current_vfs == &extfs_vfs_ops;
}

int vfs_current_is_tarfs (void)
{
    return current_vfs == &tarfs_vfs_ops;
}

int vfs_file_is_local (char *filename)
{
    vfs *Vfs;

    filename = vfs_canon (filename);
    Vfs = vfs_type (filename);
    free (filename);
    return Vfs == &local_vfs_ops;
}

#ifdef USE_NETCODE
int vfs_file_is_ftp (char *filename)
{

    vfs *Vfs;

    filename = vfs_canon (filename);
    Vfs = vfs_type (filename);
    free (filename);
    return Vfs == &ftpfs_vfs_ops;
}
#endif

char *vfs_get_current_dir (void)
{
    return current_dir;
}

static void vfs_setup_wd (void)
{
    current_dir = x_malloc (MC_MAXPATHLEN, "Current directory");
    if (!getcwd (current_dir, MC_MAXPATHLEN))
	if (errno != 0)
	     current_dir [0] = 0;
    current_dir [MC_MAXPATHLEN - 1] = 0;
    last_current_dir = current_dir;
    return;
}

int mc_mkdir (char *path, mode_t mode)
{
    vfs *Vfs;
    int result;

    path = vfs_canon (path);
    Vfs = vfs_type (path);
    result = (*Vfs->mkdir)(vfs_name (path), mode);
    free (path);
    if (result == -1) {
	errno = (*Vfs->ferrno)();
	return -1;
    }
    return result;
}

int mc_rmdir (char *path)
{
    vfs *Vfs;
    int result;

    path = vfs_canon (path);
    Vfs = vfs_type (path);
    result = (*Vfs->rmdir)(vfs_name (path));
    free (path);
    if (result == -1) {
	errno = (*Vfs->ferrno)();
	return -1;
    }
    return result;
}

int mc_mknod (char *path, int mode, int dev)
{
    vfs *Vfs;
    int result;

    path = vfs_canon (path);
    Vfs = vfs_type (path);
    result = (*Vfs->mknod)(vfs_name (path), mode, dev);
    free (path);
    if (result == -1) {
	errno = (*Vfs->ferrno)();
	return -1;
    }
    return result;
}

static struct mc_mmapping {
    caddr_t addr;
    void *vfs_info;
    vfs *vfs;
    struct mc_mmapping *next;
} *mc_mmaparray = NULL;

static caddr_t
mc_mmap (caddr_t addr, size_t len, int prot, int flags, int fd, off_t offset)
{
    vfs *Vfs;
    caddr_t result;
    struct mc_mmapping *mcm;

    if (fd == -1)
	return (caddr_t) -1;

    Vfs = vfs_op (fd);
    result = (*Vfs->mmap)(addr, len, prot, flags, vfs_info (fd), offset);
    if (result == (caddr_t)-1) {
	errno = (*Vfs->ferrno)();
	return (caddr_t)-1;
    }
    mcm = (struct mc_mmapping *) x_malloc (sizeof (struct mc_mmapping), "vfs: mmap handling");
    mcm->addr = result;
    mcm->vfs_info = vfs_info (fd);
    mcm->vfs = Vfs;
    mcm->next = mc_mmaparray;
    mc_mmaparray = mcm;
    return result;
}

int mc_munmap (caddr_t addr, size_t len)
{
    struct mc_mmapping *mcm, *mcm2 = NULL;

    for (mcm = mc_mmaparray; mcm != NULL; mcm2 = mcm, mcm = mcm->next) {
	if (mcm->addr == addr) {
	    if (mcm2 == NULL)
		mc_mmaparray = mcm->next;
	    else
		mcm2->next = mcm->next;
	    (*mcm->vfs->munmap)(addr, len, mcm->vfs_info);
	    free (mcm);
	    return 0;
	}
    }
    return -1;
}

char *mc_def_getlocalcopy (char *filename)
{
    char *tmp;
    int fdin, fdout, i;
    char buffer[8192];
    struct stat mystat;

    fdin = mc_open (filename, O_RDONLY);
    if (fdin == -1)
	return NULL;
    tmp = mc_mkstemp(&fdout);
    if (fdout == -1) {
	mc_close (fdin);
	return NULL;
    }
    while ((i = mc_read (fdin, buffer, sizeof (buffer))) == sizeof (buffer)) {
	write (fdout, buffer, i);
    }
    if (i != -1)
	write (fdout, buffer, i);
    mc_close (fdin);
    close (fdout);
    if (mc_stat (filename, &mystat) != -1) {
	chmod (tmp, mystat.st_mode);
    }
    return tmp;
}

char *mc_getlocalcopy (char *path)
{
    vfs *Vfs;
    char *result;

    path = vfs_canon (path);
    Vfs = vfs_type (path);
    result = (*Vfs->getlocalcopy)(vfs_name (path));
    free (path);
    if (result == NULL) {
	errno = (*Vfs->ferrno)();
	return NULL;
    }
    return result;
}

void mc_def_ungetlocalcopy (char *filename, char *local, int has_changed)
{
    if (has_changed) {
	int fdin, fdout, i;
	char buffer[8192];

	fdin = open (local, O_RDONLY);
	if (fdin == -1) {
	    unlink (local);
	    free (local);
	    return;
	}
	fdout = mc_open (filename, O_WRONLY | O_TRUNC);
	if (fdout == -1) {
	    close (fdin);
	    unlink (local);
	    free (local);
	    return;
	}
	while ((i = read (fdin, buffer, sizeof (buffer))) == sizeof (buffer)) {
	    mc_write (fdout, buffer, i);
	}
	if (i != -1)
	    mc_write (fdout, buffer, i);
	close (fdin);
	mc_close (fdout);
    }
    unlink (local);
    free (local);
}

static void mc_ungetlocalcopy (char *path, char *local, int has_changed)
{
    vfs *Vfs;

    path = vfs_canon (path);
    Vfs = vfs_type (path);
    (*Vfs->ungetlocalcopy)(vfs_name (path), local, has_changed);
    free (path);
}

static inline int timeoutcmp (struct timeval *t1, struct timeval *t2)
{
    return ((t1->tv_sec < t2->tv_sec)
	    || ((t1->tv_sec == t2->tv_sec) && (t1->tv_usec <= t2->tv_usec)));
}

static void vfs_timeout_handler (void)
{
    struct timeval Time;
    struct vfs_stamping *stamp, *st;

    gettimeofday (&Time, NULL);
    Time.tv_sec -= vfs_timeout;
    for (stamp = stamps; stamp != NULL;) {
	if (timeoutcmp (&stamp->time, &Time)) {
	    st = stamp->next;
	    (*stamp->v->free) (stamp->id);
	    vfs_rmstamp (stamp->v, stamp->id, 0);
	    stamp = st;
	} else
	    stamp = stamp->next;
    }
}

static void vfs_init (void)
{
    time_t current_time;
    struct tm *t;

    current_time = time (NULL);
    t = localtime (&current_time);
    current_mon = t->tm_mon;
    current_year = t->tm_year;

#ifdef USE_NETCODE
    tcp_init();
    ftpfs_init();
#endif
    extfs_init ();
    vfs_setup_wd ();
}

static void vfs_shut (void)
{
    struct vfs_stamping *stamp, *st;

    for ( stamp = stamps, stamps = 0; stamp != NULL; ) {
	(*stamp->v->free)(stamp->id);
	st = stamp->next;
	free (stamp);
	stamp = st;
    }

    if (stamps)
	vfs_rmstamp (stamps->v, stamps->id, 1);

    if (last_current_dir && last_current_dir != current_dir)
	free (last_current_dir);
    if (current_dir)
	free (current_dir);

    extfs_done ();

#ifdef USE_NETCODE
    ftpfs_done();
#endif

}

/* These ones grab information from the VFS
 *  and handles them to an upper layer
 */
void vfs_fill_names (void (*func)(char *))
{
#ifdef USE_NETCODE
    ftpfs_fill_names (func);
#endif
    tarfs_fill_names (func);
    extfs_fill_names (func);
}

/* Following stuff (parse_ls_lga) is used by ftpfs and extfs */
#define MAXCOLS 30

static char *columns [MAXCOLS]; /* Points to the string in column n */
static int   column_ptr [MAXCOLS]; /* Index from 0 to the starting positions of the columns */

static int split_text (char *p)
{
    char *original = p;
    int  numcols;


    for (numcols = 0; *p && numcols < MAXCOLS; numcols++) {
	while (*p == ' ' || *p == '\r' || *p == '\n') {
	    *p = 0;
	    p++;
	}
	columns [numcols] = p;
	column_ptr [numcols] = p - original;
	while (*p && *p != ' ' && *p != '\r' && *p != '\n')
	    p++;
    }
    return numcols;
}

static int is_num (int idx)
{
    if (columns [idx][0] < '0' || columns [idx][0] > '9')
	return 0;
    return 1;
}

#define Free_and_Return(x) {free (p_copy); return (x); }
static int parse_ls_lga (char *p, struct stat *s, char **filename, char **linkname)
{
    int idx, idx2, num_cols, isconc = 0;
    int i;
    long l;
    struct tm tim;
    int extfs_format_date = 0;
    int year_supplied = 0;
    char *p_copy;

    s->st_mode = 0;
    if (strncmp (p, "total", 5) == 0) {
	return 0;
    }
    switch (*(p++)) {
	case 'd': s->st_mode |= S_IFDIR; break;
	case 'b': s->st_mode |= S_IFBLK; break;
	case 'c': s->st_mode |= S_IFCHR; break;
	case 'm': s->st_mode |= S_IFREG; break; /* Don't know what it is :-) */
	case 'n': s->st_mode |= S_IFREG; break; /* and this as well */
	case 'l': s->st_mode |= S_IFLNK; break;
#ifdef IS_IFSOCK
	case 's': s->st_mode |= S_IFSOCK; break;
#endif
	case 'p': s->st_mode |= S_IFIFO; break;
	case '-': s->st_mode |= S_IFREG; break;
	case '?': s->st_mode |= S_IFREG; break;
	default: return 0;
    }
    if (*p == '[') {
	if (strlen (p) <= 8 || p [8] != ']')
	    return 0;
	/* Should parse here the Notwell permissions :) */
	if (S_ISDIR (s->st_mode))
	    s->st_mode |= (S_IRUSR | S_IRGRP | S_IROTH | S_IWUSR | S_IXUSR | S_IXGRP | S_IXOTH);
	else
	    s->st_mode |= (S_IRUSR | S_IRGRP | S_IROTH | S_IWUSR);
	p += 9;
    } else {
	switch (*(p++)) {
	 case 'r': s->st_mode |= 0400; break;
	 case '-': break;
	 default: return 0;
	}
	switch (*(p++)) {
	 case 'w': s->st_mode |= 0200; break;
	 case '-': break;
	 default: return 0;
	}
	switch (*(p++)) {
	 case 'x': s->st_mode |= 0100; break;
	 case 's': s->st_mode |= 0100 | S_ISUID; break;
	 case 'S': s->st_mode |= S_ISUID; break;
	 case '-': break;
	 default: return 0;
	}
	switch (*(p++)) {
	 case 'r': s->st_mode |= 0040; break;
	 case '-': break;
	 default: return 0;
	}
	switch (*(p++)) {
	 case 'w': s->st_mode |= 0020; break;
	 case '-': break;
	 default: return 0;
	}
	switch (*(p++)) {
	 case 'x': s->st_mode |= 0010; break;
	 case 's': s->st_mode |= 0010 | S_ISGID; break;
	 case 'S': s->st_mode |= S_ISGID; break;
	 case '-': break;
	 default: return 0;
	}
	switch (*(p++)) {
	 case 'r': s->st_mode |= 0004; break;
	 case '-': break;
	 default: return 0;
	}
	switch (*(p++)) {
	 case 'w': s->st_mode |= 0002; break;
	 case '-': break;
	 default: return 0;
	}
	switch (*(p++)) {
	 case 'x': s->st_mode |= 0001; break;
	 case 't': s->st_mode |= 0001 | S_ISVTX; break;
	 case 'T': s->st_mode |= S_ISVTX; break;
	 case '-': break;
	 default: return 0;
	}
    }

    p_copy = strdup (p);
    num_cols = split_text (p);

    s->st_nlink = atol (columns [0]);
    if (s->st_nlink <= 0)
	Free_and_Return (0);

    if (!is_num (1))
	s->st_uid = finduid (columns [1]);
    else
	s->st_uid = (uid_t) atol (columns [1]);

    /* Mhm, the ls -lg did not produce a group field */
    for (idx = 3; idx <= 5; idx++)
	if ((*columns [idx] >= 'A' && *columns [idx] <= 'S' &&
	    strlen (columns[idx]) == 3) || (strlen (columns[idx])==8 &&
	    columns [idx][2] == '-' && columns [idx][5] == '-'))
	    break;
    if (idx == 6 || (idx == 5 && !S_ISCHR (s->st_mode) && !S_ISBLK (s->st_mode)))
	Free_and_Return (0);
    if (idx < 5) {
	p = strchr(columns [idx - 1], ',');
	if (p && p[1] >= '0' && p[1] <= '9')
	    isconc = 1;
    }
    if (idx == 3 || (idx == 4 && !isconc && (S_ISCHR(s->st_mode) || S_ISBLK (s->st_mode))))
	idx2 = 2;
    else {
	if (is_num (2))
	    s->st_gid = (gid_t) atol (columns [2]);
	else
	    s->st_gid = findgid (columns [2]);
	idx2 = 3;
    }

    if (S_ISCHR (s->st_mode) || S_ISBLK (s->st_mode)) {
	if (!is_num (idx2))
	    Free_and_Return (0);
	s->st_rdev = (atol (columns [idx2]) & 0xff) << 8;
	if (isconc) {
	    p = strchr (columns [idx2], ',');
	    if (!p || p [1] < '0' || p [1] > '9')
		Free_and_Return (0);
	    p++;
	} else {
	    p = columns [idx2 + 1];
	    if (!is_num (idx2+1))
		Free_and_Return (0);
	}
	s->st_rdev |= (atol (p) & 0xff);
	s->st_size = 0;
    } else {
	if (!is_num (idx2))
	    Free_and_Return (0);

	s->st_size = (size_t) atol (columns [idx2]);
	s->st_rdev = 0;
    }

    p = columns [idx++];

    if (!strcmp (p, "Jan"))
	tim.tm_mon = 0;
    else if (!strcmp (p, "Feb"))
	tim.tm_mon = 1;
    else if (!strcmp (p, "Mar"))
	tim.tm_mon = 2;
    else if (!strcmp (p, "Apr"))
	tim.tm_mon = 3;
    else if (!strcmp (p, "May"))
	tim.tm_mon = 4;
    else if (!strcmp (p, "Jun"))
	tim.tm_mon = 5;
    else if (!strcmp (p, "Jul"))
	tim.tm_mon = 6;
    else if (!strcmp (p, "Aug"))
	tim.tm_mon = 7;
    else if (!strcmp (p, "Sep"))
	tim.tm_mon = 8;
    else if (!strcmp (p, "Oct"))
	tim.tm_mon = 9;
    else if (!strcmp (p, "Nov"))
	tim.tm_mon = 10;
    else if (!strcmp (p, "Dec"))
	tim.tm_mon = 11;
    else {
	/* This case should not normaly happen, but in extfs we allow these
	   date formats:
	   Mon DD hh:mm
	   Mon DD YYYY
	   Mon DD YYYY hh:mm
	   MM-DD-YY hh:mm
	   where Mon is Jan-Dec, DD, MM, YY two digit day, month, year,
	   YYYY four digit year, hh, mm two digit hour and minute. */
	if (strlen (p) == 8 && p [2] == '-' && p [5] == '-') {
	    p [2] = 0;
	    p [5] = 0;
	    tim.tm_mon = (int) atol (p);
	    if (!tim.tm_mon)
		Free_and_Return (0)
	    else
		tim.tm_mon--;
	    tim.tm_mday = (int) atol (p + 3);
	    tim.tm_year = (int) atol (p + 6);
	    if (tim.tm_year < 70)
		tim.tm_year += 70;
	    extfs_format_date = 1;
	} else
	    Free_and_Return (0);
    }

    if (!extfs_format_date) {
	if (!is_num (idx))
	    Free_and_Return (0);
	tim.tm_mday = (int)atol (columns [idx++]);
    }

    /* Microsoft ftp server may send a non padded digit */
    if (columns [idx][2] != ':' && columns [idx][1] != ':') {
	/* There is a year */
	l = atol (columns [idx++]);
	if (l < 1900 || l > 3000)
	    Free_and_Return (0);
	tim.tm_year = (int) (l - 1900);
	tim.tm_hour = 0;
	tim.tm_min = 0;
	tim.tm_sec = 0;
	year_supplied = 1;
    }
    if (columns [idx][2] == ':' || columns [idx][1] == ':') {
	if (sscanf (columns [idx],
		    "%2d:%2d", &tim.tm_hour, &tim.tm_min) != 2) {
	    if (!year_supplied)
		Free_and_Return (0);
	    tim.tm_hour = 0;
	    tim.tm_min = 0;
	    tim.tm_sec = 0;
	} else {
	    idx++;
	    tim.tm_sec = 0;
	    if (!extfs_format_date && !year_supplied) {
		tim.tm_year = current_year;
		if (tim.tm_mon > current_mon)
		    tim.tm_year--;
	    }
	}
    } else if (!year_supplied)
	Free_and_Return (0);
    tim.tm_isdst = 0;
    s->st_mtime = mktime (&tim);
    if (s->st_mtime == -1)
	s->st_mtime = 0;
    s->st_atime = s->st_mtime;
    s->st_ctime = s->st_mtime;
    s->st_dev = 0;
    s->st_ino = 0;
    s->st_blksize = 512;
    s->st_blocks = (s->st_size + 511) / 512;

    for (i = idx + 1, idx2 = 0; i < num_cols; i++ )
	if (strcmp (columns [i], "->") == 0) {
	    idx2 = i;
	    break;
	}

    if (((S_ISLNK (s->st_mode) ||
	(num_cols == idx + 3 && s->st_nlink > 1))) /* Maybe a hardlink? (in extfs) */
	&& idx2) {
	int pp;
	char *ss;

	if (filename) {
	    pp = column_ptr [idx2] - column_ptr [idx];
	    ss = x_malloc (pp, "filename");
	    strncpy (ss, p_copy + column_ptr [idx], pp - 1);
	    ss[pp - 1] = '\0';
	    *filename = ss;
	}
	if (linkname) {
	    ss = strdup (p_copy + column_ptr [idx2+1]);
	    pp = strlen (ss);
	    if (ss [pp-1] == '\r' || ss [pp-1] == '\n')
		ss [pp-1] = 0;
	    if (ss [pp-2] == '\r' || ss [pp-2] == '\n')
		ss [pp-2] = 0;

	    *linkname = ss;
	}
    } else {
	/* Extract the filename from the string copy, not from the columns
	 * this way we have a chance of entering hidden directories like ". ."
	 */
	if (filename) {
	    /*
	    *filename = strdup (columns [idx++]);
	    */
	    int pp;
	    char *ss;

	    ss = strdup (p_copy + column_ptr [idx++]);
	    pp = strlen (ss);
	    if (ss [pp-1] == '\r' || ss [pp-1] == '\n')
		ss [pp-1] = 0;
	    if (ss [pp-2] == '\r' || ss [pp-2] == '\n')
		ss [pp-2] = 0;

	    *filename = ss;
	}
	if (linkname)
	    *linkname = NULL;
    }
    Free_and_Return (1);
}
#endif /* VFS */


/*{{{ Signal Routines */

typedef void SLSig_Fun_Type (int);

/* Colors:  These definitions are used for the display.  However, the
 * application only uses object handles which get mapped to this
 * internal representation.  The mapping is performed by the Color_Map
 * structure below. */

#define FG_MASK         0x0000FF00
#define BG_MASK         0x00FF0000
#define ATTR_MASK       0x1F000000
#define BGALL_MASK      0x0FFF0000

/* The 0x10000000 bit represents the alternate character set.  BGALL_MASK does
 * not include this attribute.
 */

#define GET_FG(color) ((color & FG_MASK) >> 8)
#define GET_BG(color) ((color & BG_MASK) >> 16)
#define MAKE_COLOR(fg, bg) (((fg) | ((bg) << 8)) << 8)

/* This is the string to use to use when outputting color information.
 */
#ifdef M_UNIX
/* work around for sco console bug that can't handle combined sequences */
  static const char Color_Escape_Sequence[] = "\033[3%dm\033[4%dm";
#else
/* Believe it or not, this is what is in the linux terminfo database.  It
 * produces the same escape sequence but it is much more CPU intensive.
 * Why not just encode it as "\033[3%p1%dm\033[4%p2%dm" ???
 */
 /* static char *Color_Escape_Sequence = "\033[%p1%{30}%+%dm\033[%p2%{40}%+%dm"; */
static const char Color_Escape_Sequence[] = "\033[3%d;4%dm";
#endif

static int SLtt_flush_output (void)
{
   int nwrite = 0;
   unsigned int total;
   int n = (int) (Output_Bufferp - Output_Buffer);

   SLtt_Num_Chars_Output += n;

   total = 0;
   while (n > 0) {
	nwrite = write (fileno(stdout), (char *) Output_Buffer + total, n);
	if (nwrite == -1) {
	     nwrite = 0;
#ifdef EAGAIN
	     if (errno == EAGAIN) {
		  xusleep (100000);   /* 1/10 sec */
		  continue;
	     }
#endif
#ifdef EWOULDBLOCK
	     if (errno == EWOULDBLOCK) {
		  xusleep (100000);
		  continue;
	     }
#endif
#ifdef EINTR
	     if (errno == EINTR) continue;
#endif
	     break;
	}
	n -= nwrite;
	total += nwrite;
   }
   Output_Bufferp = Output_Buffer;
   return n;
}


static void tt_write(const char *str, unsigned int n)
{
   unsigned int ndiff;

   if ((str == NULL) || (n == 0)) return;

   while (1) {
	ndiff = MAX_OUTPUT_BUFFER_SIZE - (int) (Output_Bufferp - Output_Buffer);
	if (ndiff < n) {
	     memcpy ((char *) Output_Bufferp, str, ndiff);
	     Output_Bufferp += ndiff;
	     SLtt_flush_output ();
	     n -= ndiff;
	     str += ndiff;
	} else {
	     memcpy ((char *) Output_Bufferp, str, n);
	     Output_Bufferp += n;
	     break;
	}
   }
}


static void SLtt_write_string (const char *str)
{
   if (str != NULL) tt_write(str, strlen(str));
}



/* this is supposed to be fast--- also handles
   termcap: %d, &i, %., %+, %r strings as well as terminfo stuff */
static unsigned int tt_sprintf(char *buf, const char *fmt, int x, int y)
{
   unsigned char *b, ch;
   const unsigned char *f = (const unsigned char *) fmt;
   int offset = 0, tinfo = 0;
   int stack[10];
   int i = 0, z;
   stack[0] = y; stack[1] = x; i = 2;

   b = (unsigned char *) buf;
   if (fmt != NULL) while ((ch = *f++) != 0) {
	if (ch != '%') *b++ = ch;
	else {
	     ch = *f++;
	     if (tinfo) {
		  if ((ch <= '3') && (ch >= '0')) {
		       /* map it to termcap.  Since this is terminfo,
			* it must be one of:
			*   %2d, %3d, %02d, %03d
			*
			* I am assuming that a terminal that understands
			* %2d form will also understand the %02d form.  These
			* only differ by a space padding the field.
			*/

		       /* skip the 'd'-- hope it is there */
		       if (ch == '0') {
			    ch = *f;
			    f += 2;
		       } else
			    f++;
		  }
	     }

	     switch (ch) {
		case 'p':
		  tinfo = 1;
		  ch = *f++;
		  if (ch == '1') stack[i++] = x; else stack[i++] = y;
		  break;

		case '\'':   /* 'x' */
		  stack[i++] = *f++;
		  f++;
		  break;

		case '{':              /* literal constant, e.g. {30} */
		  z = 0;
		  while (((ch = *f) <= '9') && (ch >= '0')) {
		       z = z * 10 + (ch - '0');
		       f++;
		  }
		  stack[i++] = z;
		  if (ch == '}') f++;
		  break;

		case 'd':
		case '2':
		case '3':
		  z = stack[--i];
		  z += offset;
		  if (z >= 100) {
			*b++ = z / 100 + '0';
			z = z % 100;
			goto ten;
		  } else if (ch == 3) {
			*b++ = '0';
			ch = '2';
		  }

		  if (z >= 10) {
		       ten:
			*b++ = z / 10 + '0';
			z = z % 10;
		  } else if (ch == 2)
			*b++ = '0';

		  *b++ = z + '0';
		  break;

		case 'i':
		  offset = 1;
		  break;

		case '+':
		  if (tinfo) {
		       z = stack[--i];
		       stack[i-1] += z;
		  } else {
		       ch = *f++;
		       if ((unsigned char) ch == 128) ch = 0;
		       ch = ch + (unsigned char) stack[--i];
		       if (ch == '\n') ch++;
		       *b++ = ch;
		  }
		  break;

		case 'r':
		  stack[0] = x;
		  stack[1] = y;
		  break;

		case '.':
		case 'c':
		  ch = (unsigned char) stack[--i];
		  if (ch == '\n') ch++;
		  *b++ = ch;
		  break;

		default:
		  *b++ = ch;
	     }
	}
   }
   *b = 0;
   return((unsigned int) (b - (unsigned char *) buf));
}

static void tt_printf(const char *fmt, int x, int y)
{
   char buf[256];
   unsigned int n;

   if (fmt == NULL) return;
   n = tt_sprintf(buf, fmt, x, y);
   tt_write(buf, n);
}


static void SLtt_set_scroll_region (int r1, int r2)
{
   Scroll_r1 = r1;
   Scroll_r2 = r2;
   tt_printf (Scroll_R_Str, Scroll_r1, Scroll_r2);
   Cursor_Set = 0;
}

static void SLtt_reset_scroll_region (void)
{
    SLtt_set_scroll_region(0, SLtt_Screen_Rows - 1);
}

/* the goto_rc function moves to row relative to scrolling region */
static void SLtt_goto_rc(int r, int c)
{
   const char *sc = NULL;
   char *s = NULL;
   int n;
   char buf[6];

   if (c < 0) {
	c = -c - 1;
	Cursor_Set = 0;
   }

   r += Scroll_r1;

   if ((Cursor_Set > 0) || ((Cursor_Set < 0) && !Automatic_Margins)) {
	n = r - Cursor_r;
	if ((n == -1) && (Cursor_Set > 0) && (Cursor_c == c) && (Curs_Up_Str != NULL)) {
	     sc = Curs_Up_Str;
	} else if ((n >= 0) && (n <= 4)) {
	     if ((n == 0) && (Cursor_Set == 1)
		 && ((c > 1) || (c == Cursor_c))) {
		  if (Cursor_c == c) return;
		  if (Cursor_c == c + 1) {
		       s = buf;
		       *s++ = '\b'; *s = 0;
		       s = buf;
		  }
	     } else if (c == 0) {
		  s = buf;
		  if ((Cursor_Set != 1) || (Cursor_c != 0)) *s++ = '\r';
		  while (n--) *s++ = '\n';
		  *s = 0;
		  s = buf;
	     }
	}
   }
   if (sc != NULL) SLtt_write_string(sc);
   else if (s != NULL) SLtt_write_string(s);
   else tt_printf(Curs_Pos_Str, r, c);
   Cursor_c = c; Cursor_r = r;
   Cursor_Set = 1;
}

static void SLtt_end_insert (void)
{
   SLtt_write_string(Eins_Mode_Str);
}

static void SLtt_reverse_video (int color);

static void SLtt_normal_video (void)
{
   SLtt_reverse_video(JNORMAL_COLOR);
}

static void SLtt_delete_nlines (int n)
{
   int r1, curs;
   char buf[132];

   if (n <= 0) return;
   SLtt_normal_video ();
   if (Del_N_Lines_Str != NULL) tt_printf(Del_N_Lines_Str, n, 0);
   else {
	/* get a new terminal */
	r1 = Scroll_r1;
	curs = Cursor_r;
	SLtt_set_scroll_region(curs, Scroll_r2);
	SLtt_goto_rc(Scroll_r2 - Scroll_r1, 0);
	memset(buf, '\n', (unsigned int) n);
	tt_write(buf, (unsigned int) n);
	SLtt_set_scroll_region(r1, Scroll_r2);
	SLtt_goto_rc(curs, 0);
   }
}

static void SLtt_cls (void)
{
   SLtt_normal_video();
   SLtt_reset_scroll_region ();
   SLtt_write_string(Cls_Str);
}

static void SLtt_reverse_index (int n)
{
   if (!n) return;

   SLtt_normal_video();
   if (Add_N_Lines_Str != NULL) tt_printf(Add_N_Lines_Str,n, 0);
   else {
	while(n--) SLtt_write_string(Rev_Scroll_Str);
   }
}

static void SLtt_putchar (char ch)
{
   SLtt_normal_video ();
   if (Cursor_Set == 1) {
	if (ch >= ' ') Cursor_c++;
	else if (ch == '\b') Cursor_c--;
	else if (ch == '\r') Cursor_c = 0;
	else Cursor_Set = 0;

	if ((Cursor_c + 1 == SLtt_Screen_Cols)
	    && Automatic_Margins) Cursor_Set = 0;
   }

   if (Output_Bufferp < Output_Buffer + MAX_OUTPUT_BUFFER_SIZE)
	*Output_Bufferp++ = (unsigned char) ch;
    else
	tt_write (&ch, 1);
}

static void SLtt_beep (void)
{
   SLtt_putchar('\007');
   SLtt_flush_output ();
}

static void SLtt_del_eol (void)
{
   if (Current_Fgbg != 0xFFFFFFFFU) SLtt_normal_video ();
   SLtt_write_string(Del_Eol_Str);
}

static void SLtt_erase_line (void)
{
   SLtt_write_string("\r");
   Cursor_Set = 1; Cursor_c = 0;
   SLtt_del_eol();
}


typedef struct
{
   const char *name;
   SLtt_Char_Type color;
} Color_Def_Type;

#define MAX_COLOR_NAMES 17
static const Color_Def_Type Color_Defs [MAX_COLOR_NAMES] =
{
   {"black",            SLSMG_COLOR_BLACK},
   {"red",              SLSMG_COLOR_RED},
   {"green",            SLSMG_COLOR_GREEN},
   {"brown",            SLSMG_COLOR_BROWN},
   {"blue",             SLSMG_COLOR_BLUE},
   {"magenta",          SLSMG_COLOR_MAGENTA},
   {"cyan",             SLSMG_COLOR_CYAN},
   {"lightgray",        SLSMG_COLOR_LGRAY},
   {"gray",             SLSMG_COLOR_GRAY},
   {"brightred",        SLSMG_COLOR_BRIGHT_RED},
   {"brightgreen",      SLSMG_COLOR_BRIGHT_GREEN},
   {"yellow",           SLSMG_COLOR_BRIGHT_BROWN},
   {"brightblue",       SLSMG_COLOR_BRIGHT_BLUE},
   {"brightmagenta",    SLSMG_COLOR_BRIGHT_CYAN},
   {"brightcyan",       SLSMG_COLOR_BRIGHT_MAGENTA},
   {"white",            SLSMG_COLOR_BRIGHT_WHITE},
   {bb_default,         25}
};


static void SLtt_set_mono (int obj, SLtt_Char_Type mask)
{
   if ((obj < 0) || (obj >= JMAX_COLORS))
	return;
   Ansi_Color_Map[obj].mono = mask & ATTR_MASK;
}

static void SLtt_set_color_object (int obj, SLtt_Char_Type attr)
{
   char *cust_esc;

   if ((obj < 0) || (obj >= JMAX_COLORS)) return;

   cust_esc = Ansi_Color_Map[obj].custom_esc;
   if (cust_esc != NULL) {
	free (cust_esc);
	FgBg_Stats[(Ansi_Color_Map[obj].fgbg >> 8) & 0x7F] -= 1;
	Ansi_Color_Map[obj].custom_esc = NULL;
   }

   Ansi_Color_Map[obj].fgbg = attr;
   if (obj == 0) Color_0_Modified = 1;
}

static SLtt_Char_Type fb_to_fgbg (SLtt_Char_Type f, SLtt_Char_Type b)
{
   SLtt_Char_Type attr = 0;

   if ((f & 0xF0) == 0) {
	if (f & 0x8) attr = SLTT_BOLD_MASK;
	f &= 0x7;
   } else
	f = 9;

   if ((b & 0xF0) == 0) {
	if (b & 0x8) attr |= SLTT_BLINK_MASK;
	b &= 0x7;
   } else
	b = 9;

   return ((f << 8) | (b << 16) | attr);
}

static int make_color_fgbg (const char *fg, const char *bg, SLtt_Char_Type *fgbg)
{
   SLtt_Char_Type f = 0xFFFFFFFFU, b = 0xFFFFFFFFU;
   unsigned int i;

   if ((fg != NULL) && (*fg == 0)) fg = NULL;
   if ((bg != NULL) && (*bg == 0)) bg = NULL;

   if ((fg == NULL) || (bg == NULL))
	return -1;

   for (i = 0; i < MAX_COLOR_NAMES; i++) {
	if (strcmp(fg, Color_Defs[i].name)) continue;
	f = Color_Defs[i].color;
	break;
   }

   for (i = 0; i < MAX_COLOR_NAMES; i++) {
	if (strcmp(bg, Color_Defs[i].name)) continue;
	b = Color_Defs[i].color;
	break;
   }

   if ((f == 0xFFFFFFFFU) || (b == 0xFFFFFFFFU))
     return -1;

   *fgbg = fb_to_fgbg (f, b);
   return 0;
}

static void SLtt_set_color (int obj, const char *fg, const char *bg)
{
   SLtt_Char_Type fgbg;

   if ((obj < 0) || (obj >= JMAX_COLORS))
     return;

   if (-1 != make_color_fgbg (fg, bg, &fgbg))
     SLtt_set_color_object (obj, fgbg);
}

static void SLtt_set_color_esc (int obj, const char *esc)
{
   char *cust_esc;
   SLtt_Char_Type fgbg = 0;
   int i;

   if ((obj < 0) || (obj >= JMAX_COLORS))
	return;

   cust_esc = Ansi_Color_Map[obj].custom_esc;
   if (cust_esc != NULL) {
	free (cust_esc);
	FgBg_Stats[(Ansi_Color_Map[obj].fgbg >> 8) & 0x7F] -= 1;
   }

   cust_esc = strdup (esc);

   Ansi_Color_Map[obj].custom_esc = cust_esc;
   if (cust_esc == NULL) fgbg = 0;
   else {
	/* The whole point of this is to generate a unique fgbg */
	for (i = 0; i < JMAX_COLORS; i++) {
	     if (FgBg_Stats[i] == 0) fgbg = i;

	     if (obj == i) continue;
	     if ((Ansi_Color_Map[i].custom_esc) == NULL) continue;
	     if (!strcmp (Ansi_Color_Map[i].custom_esc, cust_esc)) {
		  fgbg = (Ansi_Color_Map[i].fgbg >> 8) & 0x7F;
		  break;
	     }
	}
	FgBg_Stats[fgbg] += 1;
   }

   fgbg |= 0x80;
   Ansi_Color_Map[obj].fgbg = (fgbg | (fgbg << 8)) << 8;
   if (obj == 0) Color_0_Modified = 1;
}

static void SLtt_set_alt_char_set (int i)
{
   if (SLtt_Has_Alt_Charset == 0) return;
   if (i == last_i) return;
   SLtt_write_string (i ? Start_Alt_Chars_Str : End_Alt_Chars_Str );
   /* if (i) Current_Fgbg |= SLTT_ALTC_MASK;
    else Current_Fgbg &= ~SLTT_ALTC_MASK; */
   last_i = i;
}

static void write_attributes (SLtt_Char_Type fgbg)
{
   int bg0, fg0;

   if (Worthless_Highlight) return;
   if (fgbg == Current_Fgbg) return;

   /* Before spitting out colors, fix attributes */
   if ((fgbg & ATTR_MASK) != (Current_Fgbg & ATTR_MASK)) {
	if (Current_Fgbg & ATTR_MASK) {
	     SLtt_write_string(Norm_Vid_Str);
	     /* In case normal video turns off ALL attributes: */
	     if (fgbg & SLTT_ALTC_MASK)
	       Current_Fgbg &= ~SLTT_ALTC_MASK;
	     SLtt_set_alt_char_set (0);
	}

	if ((fgbg & SLTT_ALTC_MASK) != (Current_Fgbg & SLTT_ALTC_MASK))
	     SLtt_set_alt_char_set ((int) (fgbg & SLTT_ALTC_MASK));

	if (fgbg & SLTT_ULINE_MASK) SLtt_write_string (UnderLine_Vid_Str);
	if (fgbg & SLTT_BOLD_MASK) SLtt_write_string (Bold_Vid_Str);
	if (fgbg & SLTT_REV_MASK) SLtt_write_string (Rev_Vid_Str);
	if (fgbg & SLTT_BLINK_MASK) {
	     /* Someday Linux will have a blink mode that set high intensity
	      * background.  Lets be prepared.
	      */
	     if (SLtt_Blink_Mode) SLtt_write_string (Blink_Vid_Str);
	}
   }

   if (SLtt_Use_Ansi_Colors) {
	fg0 = (int) GET_FG(fgbg);
	bg0 = (int) GET_BG(fgbg);
	tt_printf(Color_Escape_Sequence, fg0, bg0);
   }
   Current_Fgbg = fgbg;
}


static void SLtt_reverse_video (int color)
{
   SLtt_Char_Type fgbg;
   char *esc;

   if (Worthless_Highlight) return;
   if ((color < 0) || (color >= JMAX_COLORS)) return;

   if (Video_Initialized == 0) {
	if (color == JNORMAL_COLOR)
		SLtt_write_string (Norm_Vid_Str);
	else
		SLtt_write_string (Rev_Vid_Str);
	Current_Fgbg = 0xFFFFFFFFU;
	return;
   }

   if (SLtt_Use_Ansi_Colors) {
	fgbg = Ansi_Color_Map[color].fgbg;
	if ((esc = Ansi_Color_Map[color].custom_esc) != NULL) {
	     if (fgbg != Current_Fgbg) {
		  Current_Fgbg = fgbg;
		  SLtt_write_string (esc);
		  return;
	     }
	}
   } else
	fgbg = Ansi_Color_Map[color].mono;

   if (fgbg == Current_Fgbg) return;
   write_attributes (fgbg);
}



/* Highest bit represents the character set. */
#define COLOR_MASK 0x7F00

#define COLOR_OF(x) (((unsigned int)(x) & COLOR_MASK) >> 8)

#define COLOR_EQS(a, b) \
   (SLtt_Use_Ansi_Colors \
    ? (Ansi_Color_Map[COLOR_OF(a)].fgbg == Ansi_Color_Map[COLOR_OF(b)].fgbg)\
    :  (Ansi_Color_Map[COLOR_OF(a)].mono == Ansi_Color_Map[COLOR_OF(b)].mono))


#define CHAR_EQS(a, b) (((a) == (b))\
			|| ((((a) & ~COLOR_MASK) == ((b) & ~COLOR_MASK))\
			    && COLOR_EQS((a), (b))))


/* The whole point of this routine is to prevent writing to the last column
 * and last row on terminals with automatic margins.
 */
static void write_string_with_care (char *str)
{
   unsigned int len;

   if (str == NULL) return;

   len = strlen (str);
   if (Automatic_Margins && (Cursor_r + 1 == SLtt_Screen_Rows)) {
	if (len + (unsigned int) Cursor_c >= (unsigned int) SLtt_Screen_Cols) {
	     /* For now, just do not write there.  Later, something more
	      * sophisticated will be implemented.
	      */
	     if (SLtt_Screen_Cols > Cursor_c)
	       len = SLtt_Screen_Cols - Cursor_c - 1;
	     else len = 0;
	}
   }
   tt_write (str, len);
}

static void send_attr_str (unsigned short *s)
{
   unsigned char out[256], ch, *p;
   SLtt_Char_Type attr;
   unsigned short sh;
   int color, last_color = -1;

   p = out;
   while (0 != (sh = *s++)) {
	ch = sh & 0xFF;
	color = ((int) sh & 0xFF00) >> 8;
	if (color != last_color) {
	     if (SLtt_Use_Ansi_Colors) attr = Ansi_Color_Map[color & 0x7F].fgbg;
	     else attr = Ansi_Color_Map[color & 0x7F].mono;

	     /* sh => color */
	     if (color & 0x80) /* alternate char set */
		  attr |= SLTT_ALTC_MASK;


	     if (attr != Current_Fgbg) {
		  if ((ch != ' ') ||
		      /* it is a space so only consider it different if it
		       * has different attributes.
		       */
		      (attr & BGALL_MASK) != (Current_Fgbg & BGALL_MASK)) {
		       if (p != out) {
			    *p = 0;
			    write_string_with_care ((char *) out);
			    Cursor_c += (int) (p - out);
			    p = out;
		       }

		       if (SLtt_Use_Ansi_Colors && (NULL != Ansi_Color_Map[color & 0x7F].custom_esc)) {
			    SLtt_write_string (Ansi_Color_Map[color & 0x7F].custom_esc);
			    /* Just in case the custom escape sequence screwed up
			     * the alt character set state...
			     */
			    if ((attr & SLTT_ALTC_MASK) != (Current_Fgbg & SLTT_ALTC_MASK))
			      SLtt_set_alt_char_set ((int) (attr & SLTT_ALTC_MASK));
			    Current_Fgbg = attr;
		       } else
			    write_attributes (attr);

		       last_color = color;
		  }
	     }
	}
	*p++ = ch;
   }
   *p = 0;
   if (p != out) write_string_with_care ((char *) out);
   Cursor_c += (int) (p - out);
}

static void forward_cursor (unsigned int n, int row)
{
   char buf[30];

   if (n <= 4) {
	SLtt_normal_video ();
	memset (buf, ' ', n);
	buf[n] = 0;
	write_string_with_care (buf);
	Cursor_c += n;
   } else if (Curs_F_Str != NULL) {
	Cursor_c += n;
	n = tt_sprintf(buf, Curs_F_Str, (int) n, 0);
	tt_write(buf, n);
   } else
	SLtt_goto_rc (row, (int) (Cursor_c + n));
}


#define SPACE_CHAR (0x20 | (JNORMAL_COLOR << 8))

static void SLtt_smart_puts(unsigned short *neww, unsigned short *oldd, int len, int row)
{
   unsigned short *p, *q, *qmax, *pmax, *buf;
   unsigned short buffer[256];
   unsigned int n_spaces;
   unsigned short *space_match, *last_buffered_match;
   int handle_hp_glitch = 0;

   q = oldd; p = neww;
   qmax = oldd + len;
   pmax = p + len;

   /* Find out where to begin --- while they match, we are ok */
   while (1) {
	if (q == qmax) return;
	if (!CHAR_EQS(*q, *p)) break;
	q++; p++;
   }

   /*position the cursor */
   SLtt_goto_rc (row, (int) (p - neww));

   if (Has_HP_Glitch) {
	unsigned short *qq2 = q;
	while (qq2 < qmax) {
	     if (*qq2 & 0xFF00) {
		  SLtt_normal_video ();
		  SLtt_del_eol ();
		  qmax = q;
		  handle_hp_glitch = 1;
		  break;
	     }
	     qq2++;
	}
   }
   /* Find where the last non-blank character on old/new screen is */

   while (qmax > q) {
	qmax--;
	if (!CHAR_EQS(*qmax, SPACE_CHAR)) {
	     qmax++;
	     break;
	}
   }

   while (pmax > p) {
	pmax--;
	if (!CHAR_EQS(*pmax, SPACE_CHAR)) {
	     pmax++;
	     break;
	}
   }

   last_buffered_match = buf = buffer;                 /* buffer is empty */

   if (handle_hp_glitch)
	while (p < pmax)
	     *buf++ = *p++;

   /* loop using overwrite then skip algorithm until done */
   while (1) {
	/* while they do not match and we do not hit a space, buffer them up */
	n_spaces = 0;
	while (p < pmax) {
	     if (CHAR_EQS(*q,SPACE_CHAR) && CHAR_EQS(*p, SPACE_CHAR)) {
		  /* If *q is not a space, we would have to overwrite it.
		   * However, if *q is a space, then while *p is also one,
		   * we only need to skip over the blank field.
		   */
		  space_match = p;
		  p++; q++;
		  while ((p < pmax) && CHAR_EQS(*q,SPACE_CHAR) && CHAR_EQS(*p, SPACE_CHAR)) {
		       p++;
		       q++;
		  }
		  n_spaces = (unsigned int) (p - space_match);
		  break;
	     }
	     if (CHAR_EQS(*q, *p)) break;
	     *buf++ = *p++;
	     q++;
	}
	*buf = 0;

	if (buf != buffer) send_attr_str (buffer);
	buf = buffer;

	if (n_spaces && (p < pmax))
	     forward_cursor (n_spaces, row);

	/* Now we overwrote what we could and cursor is placed at position
	 * of a possible match of new and old.  If this is the case, skip
	 * some more.
	 */
	while ((p < pmax) && CHAR_EQS(*p, *q)) {
	     *buf++ = *p++;
	     q++;
	}
	last_buffered_match = buf;
	if (p >= pmax) break;

	/* jump to new position is it is greater than 5 otherwise
	 * let it sit in the buffer and output it later.
	 */
	if ((int) (buf - buffer) >= 5) {
	     forward_cursor ((unsigned int) (buf - buffer), row);
	     last_buffered_match = buf = buffer;
	}
   }

   if (buf != buffer) {
	if (q < qmax) {
	     if ((buf == last_buffered_match) && ((int) (buf - buffer) >= 5)) {
		  forward_cursor ((unsigned int) (buf - buffer), row);
	     } else {
		  *buf = 0;
		  send_attr_str (buffer);
	     }
	}
   }
   if (q < qmax) SLtt_del_eol ();
   if (Automatic_Margins && (Cursor_c + 1 >= SLtt_Screen_Cols)) Cursor_Set = 0;
}


static const char *SLtt_tigetstr (const char *cap);
static int SLtt_tigetflag (const char *cap);

/* This file contains enough terminfo reading capabilities sufficient for
 * the slang SLtt interface.
 */

/*
 * The majority of the comments found in the file were taken from the
 * term(4) man page on an SGI.
 */

/* Short integers are stored in two 8-bit bytes.  The first byte contains
 * the least significant 8 bits of the value, and the second byte contains
 * the most significant 8 bits.  (Thus, the value represented is
 * 256*second+first.)  The value -1 is represented by 0377,0377, and the
 * value -2 is represented by 0376,0377; other negative values are illegal.
 * The -1 generally means that a capability is missing from this terminal.
 * The -2 means that the capability has been cancelled in the terminfo
 * source and also is to be considered missing.
 */

static int make_integer (const unsigned char *buf)
{
   int lo, hi;
   lo = (int) *buf++; hi = (int) *buf;
   if (hi == 0377) {
	if (lo == 0377) return -1;
	if (lo == 0376) return -2;
   }
   return lo + 256 * hi;
}

/*
 * The compiled file is created from the source file descriptions of the
 * terminals (see the -I option of infocmp) by using the terminfo compiler,
 * tic, and read by the routine setupterm [see curses(3X).]  The file is
 * divided into six parts in the following order:  the header, terminal
 * names, boolean flags, numbers, strings, and string table.
 *
 * The header section begins the file.  This section contains six short
 * integers in the format described below.  These integers are (1) the magic
 * number (octal 0432); (2) the size, in bytes, of the names section; (3)
 * the number of bytes in the boolean section; (4) the number of short
 * integers in the numbers section; (5) the number of offsets (short
 * integers) in the strings section; (6) the size, in bytes, of the string
 * table.
 */

#define MAGIC 0432


static const unsigned char linux_terminfo[]=
"\032\0013\000\035\000\020\000}\001G\003linux-koi8|linux with koi8 alterna\
te character set\000\000\001\000\000\001\001\000\000\000\000\000\000\000\001\
\001\000\000\000\000\000\001\000\000\000\000\000\000\001\001\377\377\010\000\
\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\
\377\377\377\010\000@\000\022\000\377\377\000\000\002\000\004\000\025\000\032\
\000!\000%\000)\000\377\3774\000E\000G\000K\000W\000\377\377Y\000e\000\377\
\377i\000m\000y\000}\000\377\377\377\377\201\000\207\000\214\000\377\377\377\
\377\221\000\226\000\377\377\377\377\233\000\240\000\245\000\252\000\263\000\
\271\000\377\377\377\377\301\000\306\000\314\000\322\000\377\377\377\377\377\
\377\377\377\377\377\377\377\344\000\350\000\377\377\354\000\377\377\377\377\
\377\377\356\000\377\377\363\000\377\377\377\377\377\377\377\377\367\000\374\
\000\002\001\007\001\014\001\021\001\026\001\034\001\"\001(\001.\0013\001\377\
\3778\001\377\377<\001A\001F\001\377\377\377\377\377\377J\001\377\377\377\377\
\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\
\377\377\377\377\377\377\377N\001\377\377Q\001Z\001c\001l\001\377\377u\001\
~\001\207\001\377\377\220\001\377\377\377\377\377\377\377\377\377\377\377\377\
\377\377\231\001\377\377\377\377\377\377\237\001\242\001\255\001\260\001\262\
\001\265\001\013\002\377\377\016\002\377\377\377\377\377\377\377\377\377\377\
\377\377\020\002\377\377\377\377\377\377\377\377\024\002\377\377S\002\377\377\
\377\377W\002]\002\377\377\377\377\377\377\377\377\377\377\377\377\377\377\
\377\377\377\377\377\377\377\377c\002\377\377\377\377\377\377\377\377\377\377\
\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\
\377\377\377\377\377\377\377\377\377h\002\377\377\377\377\377\377\377\377\377\
\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\
\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\
\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377j\002p\002v\002\
|\002\202\002\210\002\216\002\224\002\232\002\240\002\377\377\377\377\377\377\
\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\
\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\
\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\
\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\
\377\377\377\377\246\002\377\377\377\377\377\377\377\377\377\377\377\377\377\
\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\
\377\377\377\377\377\377\377\377\377\377\377\377\377\377\253\002\266\002\273\
\002\301\002\305\002\316\002\322\002\377\377\377\377\377\377\377\377\377\377\
\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\
\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\
\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\
\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\
\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\
\377\377\377\377\377#\003\377\377\377\377\377\377'\0031\003\377\377\377\377\
\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\377\
\377\377\377\377\377\377\377\377\377\377\377\377\377;\003A\003\007\000\015\
\000\033[%i%p1%d;%p2%dr\000\033[3g\000\033[H\033[J\000\033[K\000\033[J\000\
\033[%i%p1%dG\000\033[%i%p1%d;%p2%dH\000\012\000\033[H\000\033[?25l\033[?1\
c\000\010\000\033[?25h\033[?0c\000\033[C\000\033[A\000\033[?25h\033[?8c\000\
\033[P\000\033[M\000\033[11m\000\033[5m\000\033[1m\000\033[2m\000\033[4h\000\
\033[7m\000\033[7m\000\033[4m\000\033[%p1%dX\000\033[10m\000\033[0;10m\000\
\033[4l\000\033[27m\000\033[24m\000\033[?5h\033[?5l$<200/>\000\033[@\000\033\
[L\000\177\000\033[3~\000\033[B\000\033[[A\000\033[21~\000\033[[B\000\033[\
[C\000\033[[D\000\033[[E\000\033[17~\000\033[18~\000\033[19~\000\033[20~\000\
\033[1~\000\033[2~\000\033[D\000\033[6~\000\033[5~\000\033[C\000\033[A\000\
\015\012\000\033[%p1%dP\000\033[%p1%dM\000\033[%p1%dB\000\033[%p1%d@\000\033\
[%p1%dL\000\033[%p1%dD\000\033[%p1%dC\000\033[%p1%dA\000\033c\033]R\000\033\
8\000\033[%i%p1%dd\000\0337\000\012\000\033M\000\033[0;10%?%p1%t;7%;%?%p2%\
t;4%;%?%p3%t;7%;%?%p4%t;5%;%?%p5%t;2%;%?%p6%t;1%;%?%p9%t;11%;m\000\033H\000\
\011\000\033[G\000+\020,\021-\030.\0310\215`\004a\221f\234g\237h\220i\276j\
\205k\203l\202m\204n\212o~p\200q\200r\200s_t\206u\207v\211w\210x\201y\230z\
\231{\267|\274~\224\000\033[Z\000\033[?7h\000\033[?7l\000\033[4~\000\032\000\
\033[23~\000\033[24~\000\033[25~\000\033[26~\000\033[28~\000\033[29~\000\033\
[31~\000\033[32~\000\033[33~\000\033[34~\000\033[1K\000\033[%i%d;%dR\000\033\
[6n\000\033[?6c\000\033[c\000\033[39;49m\000\033]R\000\033]P%p1%x%p2%{255}\
%*%{1000}%/%02x%p3%{255}%*%{1000}%/%02x%p4%{255}%*%{1000}%/%02x\000\033[M\000\
\033[3%p1%dm\000\033[4%p1%dm\000\033[11m\000\033[10m\000\000\003\000\001\000\
\007\000\022\000%\000\001\000\000\000\001\000\377\377\377\377\377\377\377\377\
\377\377\377\377\377\377\000\000\003\000\006\000\011\000\014\000\017\000\022\
\000\025\000\031\000\035\000!\000AX\000G0\000XT\000U8\000E0\000E3\000S0\000\
ka2\000kb1\000kb3\000kc2\000";

static void open_terminfo (Terminfo_Type *h)
{
   const unsigned char *buf = linux_terminfo;
   linux_terminfo_offset = 12;

   if ((MAGIC == make_integer (buf))) {
	h->name_section_size = make_integer (buf + 2);
	h->boolean_section_size = make_integer (buf + 4);
	h->num_numbers = make_integer (buf + 6);
	h->num_string_offsets = make_integer (buf + 8);
	h->string_table_size = make_integer (buf + 10);
   }
}

/* The terminal names section comes next.  It contains the first line of the
 * terminfo description, listing the various names for the terminal,
 * separated by the bar ( | ) character (see term(5)).  The section is
 * terminated with an ASCII NUL character.
 */

/* returns pointer to malloced space */
static const unsigned char *read_terminfo_section (unsigned int size)
{
   const unsigned char *s = linux_terminfo + linux_terminfo_offset;
   linux_terminfo_offset += size;
   return s;
}

static inline void read_terminal_names (Terminfo_Type *t)
{
   t->terminal_names = (const char *) read_terminfo_section (t->name_section_size);
}

/* The boolean flags have one byte for each flag.  This byte is either 0 or
 * 1 as the flag is present or absent.  The value of 2 means that the flag
 * has been cancelled.  The capabilities are in the same order as the file
 * <term.h>.
 */

static inline void read_boolean_flags (Terminfo_Type *t)
{
   /* Between the boolean section and the number section, a null byte is
    * inserted, if necessary, to ensure that the number section begins on an
    * even byte offset. All short integers are aligned on a short word
    * boundary.
    */

   unsigned int size = (t->name_section_size + t->boolean_section_size) % 2;
   size += t->boolean_section_size;

   t->boolean_flags = read_terminfo_section (size);
}


/* The numbers section is similar to the boolean flags section.  Each
 * capability takes up two bytes, and is stored as a short integer.  If the
 * value represented is -1 or -2, the capability is taken to be missing.
 */

static inline void read_numbers (Terminfo_Type *t)
{
   t->numbers = read_terminfo_section (2 * t->num_numbers);
}


/* The strings section is also similar.  Each capability is stored as a
 * short integer, in the format above.  A value of -1 or -2 means the
 * capability is missing.  Otherwise, the value is taken as an offset from
 * the beginning of the string table.  Special characters in ^X or \c
 * notation are stored in their interpreted form, not the printing
 * representation.  Padding information ($<nn>) and parameter information
 * (%x) are stored intact in uninterpreted form.
 */

static inline void read_string_offsets (Terminfo_Type *t)
{
   t->string_offsets = (const unsigned char *) read_terminfo_section (2 * t->num_string_offsets);
}


/* The final section is the string table.  It contains all the values of
 * string capabilities referenced in the string section.  Each string is
 * null terminated.
 */

static inline void read_string_table (Terminfo_Type *t)
{
   t->string_table = (const char *) read_terminfo_section (t->string_table_size);
}


static void SLtt_tigetent (void)
{
   open_terminfo(&Tbuf);
   read_terminal_names (&Tbuf);
   read_boolean_flags (&Tbuf);
   read_numbers (&Tbuf);
   read_string_offsets (&Tbuf);
   read_string_table (&Tbuf);
}

static const char *SLtt_tgetstr (const char *s)
{
#if 0
   const char *w, *w1, *wsave;
#endif
   const char *what;

   what = SLtt_tigetstr(s);
   if (what != NULL) {
	/* Check for AIX brain-damage */
	if (*what == '@')
	  return NULL;

#if 0
	/* lose pad info --- with today's technology, term is a loser if
	   it is really needed */
	while ((*what == '.') ||
	       ((*what >= '0') && (*what <= '9'))) what++;
	if (*what == '*') what++;

	/* lose terminfo padding--- looks like $<...> */
	w = what;
	while (*w) if ((*w++ == '$') && (*w == '<')) {
	     w1 = w - 1;
	     while (*w && (*w != '>')) w++;
	     if (*w == 0) break;
	     w++;
	     wsave = w1;
	     while ((*w1++ = *w++) != 0);
	     w = wsave;
	}
#endif
	if (*what == 0) what = NULL;
   }
   return(what);
}

static void SLtt_get_screen_size (void)
{
    int r = 0, c = 0;

#if defined(TIOCGWINSZ)
    struct winsize wind_struct;

    do {
	if ((ioctl(1,TIOCGWINSZ,&wind_struct) == 0)
	    || (ioctl(0, TIOCGWINSZ, &wind_struct) == 0)
	    || (ioctl(2, TIOCGWINSZ, &wind_struct) == 0))
	  {
	     c = (int) wind_struct.ws_col;
	     r = (int) wind_struct.ws_row;
	     break;
	  }
    } while (errno == EINTR);

#endif

    if (r <= 0) {
	char *s = getenv ("LINES");
	if (s != NULL) r = atoi (s);
    }

    if (c <= 0) {
	char *s = getenv ("COLUMNS");
	if (s != NULL) c = atoi (s);
    }

    if ((r <= 0) || (r > 200)) r = 24;
    if ((c <= 0) || (c > 250)) c = 80;
    SLtt_Screen_Rows = r;
    SLtt_Screen_Cols = c;
}

static int SLtt_tgetnum (const char *s);

/* This sets term for vt102 terminals it parameter vt100 is 0.  If vt100
 * is non-zero, set terminal appropriate for a only vt100
 * (no add line capability). */

static void SLtt_set_term_vtxxx(int *vt100)
{
   Norm_Vid_Str = "\033[m";

   Scroll_R_Str = "\033[%i%d;%dr";
   Cls_Str = "\033[2J\033[H";
   Rev_Vid_Str = "\033[7m";
   Bold_Vid_Str = "\033[1m";
   Blink_Vid_Str = "\033[5m";
   UnderLine_Vid_Str = "\033[4m";
   Del_Eol_Str = "\033[K";
   Rev_Scroll_Str = "\033M";
   Curs_F_Str = "\033[%dC";
   /* Len_Curs_F_Str = 5; */
   Curs_Pos_Str = "\033[%i%d;%dH";
   if ((vt100 == NULL) || (*vt100 == 0)) {
	Eins_Mode_Str = "\033[4l";
	Del_N_Lines_Str = "\033[%dM";
	Add_N_Lines_Str = "\033[%dL";
   } else {
	Del_N_Lines_Str = NULL;
	Add_N_Lines_Str = NULL;
   }
   SLtt_Term_Cannot_Scroll = 0;
}

static void SLtt_get_terminfo (void)
{
   const char *term, *t;
   char ch;
   int is_xterm;
   int almost_vtxxx;
   int Vt100_Like = 0;

   SLtt_Use_Ansi_Colors = (NULL != getenv ("COLORTERM"));

   if (NULL == (term = TERM_env))
	SLang_exit_error("TERM environment variable needs set.");

   Linux_Console = (!strncmp (term, "linux", 5)
#ifdef linux
		    || !strncmp(term, "con", 3)
#endif
		    );

   t = term;

   if (strcmp(t, "vt52") && (*t++ == 'v') && (*t++ == 't')
       && (ch = *t, (ch >= '1') && (ch <= '9'))) Vt100_Like = 1;

   is_xterm = !strncmp (term, "xterm", 5);
   almost_vtxxx = (Vt100_Like
		   || Linux_Console
		   || is_xterm
		   || !strcmp (term, "screen"));

   SLtt_tigetent();
   Cls_Str = SLtt_tgetstr("cl");
   Curs_Pos_Str = SLtt_tgetstr("cm");

   Eins_Mode_Str = SLtt_tgetstr("ei");

   Curs_Up_Str = SLtt_tgetstr ("up");
   Rev_Scroll_Str = SLtt_tgetstr("sr");
   Del_N_Lines_Str = SLtt_tgetstr("DL");
   Add_N_Lines_Str = SLtt_tgetstr("AL");

   /* Actually these are used to initialize terminals that use cursor
    * addressing.  Hard to believe.
    */
   Term_Init_Str = SLtt_tgetstr ("ti");
   Term_Reset_Str = SLtt_tgetstr ("te");

   /* If I do this for vtxxx terminals, arrow keys start sending ESC O A,
    * which I do not want.  This is mainly for HP terminals.
    */
   if (almost_vtxxx == 0) {
	Keypad_Init_Str = SLtt_tgetstr ("ks");
	Keypad_Reset_Str = SLtt_tgetstr ("ke");
   }

   /* Make up for defective termcap/terminfo databases */
   if ((Vt100_Like && (term[2] != '1')) || Linux_Console || is_xterm) {
	if (Del_N_Lines_Str == NULL) Del_N_Lines_Str = "\033[%dM";
	if (Add_N_Lines_Str == NULL) Add_N_Lines_Str = "\033[%dL";
   }

   Scroll_R_Str = SLtt_tgetstr("cs");

   SLtt_get_screen_size ();

   if ((Scroll_R_Str == NULL)
       || (((NULL == Del_N_Lines_Str) || (NULL == Add_N_Lines_Str))
					&& (NULL == Rev_Scroll_Str))) {
	if (is_xterm || Linux_Console) {
	     /* Defective termcap mode!!!! */
	    SLtt_set_term_vtxxx (NULL);
	} else
	    SLtt_Term_Cannot_Scroll = 1;
   }

   Del_Eol_Str = SLtt_tgetstr("ce");

   Rev_Vid_Str = SLtt_tgetstr("mr");
   if (Rev_Vid_Str == NULL) Rev_Vid_Str = SLtt_tgetstr("so");

   Bold_Vid_Str = SLtt_tgetstr("md");

   /* Although xterm cannot blink, it does display the blinking characters
    * as bold ones.  Some Rxvt will display the background as high intensity.
    */
   if ((NULL == (Blink_Vid_Str = SLtt_tgetstr("mb"))) && is_xterm)
     Blink_Vid_Str = "\033[5m";

   UnderLine_Vid_Str = SLtt_tgetstr("us");

   Start_Alt_Chars_Str = SLtt_tgetstr ("as");   /* smacs */
   End_Alt_Chars_Str = SLtt_tgetstr ("ae");   /* rmacs */
   Enable_Alt_Char_Set = SLtt_tgetstr ("eA");   /* enacs */
   SLtt_Graphics_Char_Pairs = SLtt_tgetstr ("ac");

   if (NULL == SLtt_Graphics_Char_Pairs) {
	/* make up for defective termcap/terminfo */
	if (Vt100_Like) {
	     Start_Alt_Chars_Str = "\016";
	     End_Alt_Chars_Str = "\017";
	     Enable_Alt_Char_Set = "\033)0";
	}
   }

    /* aixterm added by willi */
   if (is_xterm || !strncmp (term, "aixterm", 7)) {
	Start_Alt_Chars_Str = "\016";
	End_Alt_Chars_Str = "\017";
	Enable_Alt_Char_Set = "\033(B\033)0";
   }

   if (SLtt_Graphics_Char_Pairs == NULL &&
	    (Start_Alt_Chars_Str == NULL || End_Alt_Chars_Str == NULL)) {
	SLtt_Has_Alt_Charset = 0;
	Enable_Alt_Char_Set = NULL;
   } else
	SLtt_Has_Alt_Charset = 1;

   if (NULL == (Norm_Vid_Str = SLtt_tgetstr("me")))
	Norm_Vid_Str = SLtt_tgetstr("se");

   Cursor_Invisible_Str = SLtt_tgetstr("vi");
   Cursor_Visible_Str = SLtt_tgetstr("ve");
   Curs_F_Str = SLtt_tgetstr("RI");

   Automatic_Margins = SLtt_tigetflag ("am");
   Has_HP_Glitch = SLtt_tigetflag ("xs");

   /* Magic cookie glitch */
   Worthless_Highlight = (SLtt_tgetnum ("sg") > 0);

   if (Worthless_Highlight)
     SLtt_Has_Alt_Charset = 0;

   /* Check for color information in the termcap.  A program should not
    * rely on this information being accurate.
    */
   if (SLtt_Use_Ansi_Colors == 0) {
	Reset_Color_String = SLtt_tgetstr ("op");

	SLtt_Use_Ansi_Colors = ((NULL != Reset_Color_String)
				|| (NULL != SLtt_tgetstr ("Sf"))
				|| (NULL != SLtt_tgetstr ("Sb"))
				|| (NULL != SLtt_tgetstr ("AF"))
				|| (NULL != SLtt_tgetstr ("AB"))
				|| (-1 != SLtt_tgetnum ("Co"))
				|| (-1 != SLtt_tgetnum ("pa")));

   }
}


static void SLtt_reset_video (void)
{
   SLtt_goto_rc (SLtt_Screen_Rows - 1, 0);
   Cursor_Set = 0;
   SLtt_normal_video ();               /* MSKermit requires this  */
   SLtt_write_string(Norm_Vid_Str);

   Current_Fgbg = 0xFFFFFFFFU;
   SLtt_set_alt_char_set (0);
   if (SLtt_Use_Ansi_Colors) {
	if (Reset_Color_String == NULL) {
	     SLtt_Char_Type attr;
	     if (-1 != make_color_fgbg (NULL, NULL, &attr))
		write_attributes (attr);
	     else
		SLtt_write_string ("\033[0m\033[m");
	} else
	    SLtt_write_string (Reset_Color_String);
	Current_Fgbg = 0xFFFFFFFFU;
   }
   SLtt_erase_line ();
   SLtt_write_string (Keypad_Reset_Str);
   SLtt_write_string (Term_Reset_Str);
   SLtt_flush_output ();
   Video_Initialized = 0;
}

/* This has the effect of mapping all characters in the range 128-169 to
 * ESC [ something */
#define DEC_8BIT_HACK 64

static void SLang_ungetkey_string (unsigned char *s, unsigned int n)
{
   unsigned char *bmax, *b, *b1;
   if (SLang_Input_Buffer_Len + n + 3 > MAX_INPUT_BUFFER_LEN) return;

   b = SLang_Input_Buffer;
   bmax = (b - 1) + SLang_Input_Buffer_Len;
   b1 = bmax + n;
   while (bmax >= b) *b1-- = *bmax--;
   bmax = b + n;
   while (b < bmax) *b++ = *s++;
   SLang_Input_Buffer_Len += n;
}


/* We are primarily interested in blocking signals that would cause the
 * application to reset the tty.  These include suspend signals and
 * possibly interrupt signals. */
static void SLsig_block_signals(void)
{
   Blocked_Depth++;
   if (Blocked_Depth == 1)

       Old_Signal_Mask = bb_sigblock( (1 << SIGQUIT) | (1 << SIGINT)
# ifdef SIGTSTP
	 | (1 << SIGTSTP)
# endif
# ifdef SIGTTIN
	| (1 << SIGTTIN)
# endif
# ifdef SIGTTOU
	| (1 << SIGTTOU)
# endif
			);
}

static void SLsig_unblock_signals(void)
{
   if (Blocked_Depth == 0)
     return;

   Blocked_Depth--;

   if (Blocked_Depth != 0)
     return;

   bb_sigunblock(Old_Signal_Mask);
}


#ifndef pc_system
#define ALT_CHAR_FLAG 0x80
#else
#define ALT_CHAR_FLAG 0x00
#endif

static void blank_line (unsigned short *p, int n, unsigned char ch)
{
   unsigned short *pmax = p + n;
   unsigned short color_ch;

   color_ch = (This_Color << 8) | (unsigned short) ch;

   while (p < pmax)
	*p++ = color_ch;
}

static void SLsmg_gotorc (int r, int c)
{
   This_Row = r;
   This_Col = c;
}

static void SLsmg_set_char_set (int i)
{
   if (i) This_Alt_Char = ALT_CHAR_FLAG;
   else This_Alt_Char = 0;

   This_Color &= 0x7F;
   This_Color |= This_Alt_Char;
}

static void SLsmg_set_color (int color)
{
   if (color < 0) return;
   This_Color = color | This_Alt_Char;
}


static void SLsmg_normal_video (void)
{
   This_Color = This_Alt_Char;         /* reset video but NOT char set. */
}


static int point_visible (int col_too)
{
   return ((This_Row >= Start_Row) && (This_Row < Start_Row + Screen_Rows)
	   && ((col_too == 0)
	       || ((This_Col >= Start_Col)
		   && (This_Col < Start_Col + Screen_Cols))));
}

static void SLsmg_printf (const char *fmt, ...)
{
   char p[1000];
   va_list ap;

   va_start(ap, fmt);
   (void) vsprintf(p, fmt, ap);
   va_end(ap);

   SLsmg_write_string (p);
}

#define SLsmg_Tab_Width  8

/* Minimum value for which eight bit char is displayed as is. */


static void SLsmg_write_nchars (const char *str, int n)
{
   unsigned short *p, old, neew, color;
   unsigned char ch;
   unsigned int flags;
   int len, start_len, max_len;
   const char *str_max;
#ifndef pc_system
   int alt_char_set_flag;

   alt_char_set_flag = (This_Color & ALT_CHAR_FLAG);
#endif

   str_max = str + n;
   color = This_Color << 8;

   start_len = Start_Col;

   if (point_visible (0) == 0) return;

   len = This_Col;
   max_len = start_len + Screen_Cols;

   p = SL_Screen[This_Row].neew;
   if (len > start_len) p += (len - start_len);

   flags = SL_Screen[This_Row].flags;
   while ((len < max_len) && (str < str_max)) {
	ch = (unsigned char) *str++;

#ifndef pc_system
	if (alt_char_set_flag)
	  ch = Alt_Char_Set [ch & 0x7F];
#endif
	if (((ch >= ' ') && (ch < 127))
	    || (ch >= (unsigned char) SLsmg_Display_Eight_Bit)
#ifndef pc_system
	    || alt_char_set_flag
#endif
				) {
	     len += 1;
	     if (len > start_len) {
		  old = *p;
		  neew = color | (unsigned short) ch;
		  if (old != neew) {
		       flags |= TOUCHED;
		       *p = neew;
		  }
		  p++;
	     }
	} else if ((ch == '\t') && (SLsmg_Tab_Width > 0)) {
	     n = len;
	     n += SLsmg_Tab_Width;
	     n = SLsmg_Tab_Width - (n % SLsmg_Tab_Width);
	     if (len + n > max_len) n = max_len - len;
	     neew = color | (unsigned short) ' ';
	     while (n--) {
		  len += 1;
		  if (len > start_len) {
		       if (*p != neew) {
			    flags |= TOUCHED;
			    *p = neew;
		       }
		       p++;
		  }
	     }
	} else if (ch == '\n') {
	     break;
	} else {
	     if (ch & 0x80) {
		  neew = color | (unsigned short) '~';
		  len += 1;
		  if (len > start_len) {
		       if (*p != neew) {
			    *p = neew;
			    flags |= TOUCHED;
		       }
		       p++;
		       if (len == max_len) break;
		       ch &= 0x7F;
		  }
	     }

	     len += 1;
	     if (len > start_len) {
		  neew = color | (unsigned short) '^';
		  if (*p != neew) {
		       *p = neew;
		       flags |= TOUCHED;
		  }
		  p++;
		  if (len == max_len) break;
	     }

	     if (ch == 127) ch = '?'; else ch = ch + '@';
	     len++;
	     if (len > start_len) {
		  neew = color | (unsigned short) ch;
		  if (*p != neew) {
		       *p = neew;
		       flags |= TOUCHED;
		  }
		  p++;
	     }
	}
   }

   SL_Screen[This_Row].flags = flags;
   This_Col = len;
}


static void SLsmg_write_char (char ch)
{
   SLsmg_write_nchars (&ch, 1);
}

static void SLsmg_write_string (const char *str)
{
   SLsmg_write_nchars (str, strlen (str));
}


static unsigned long compute_hash (unsigned short *s, int n)
{
   unsigned long h = 0, g;
   unsigned long sum = 0;
   unsigned short *smax, ch;
   int is_blank = 2;

   smax = s + n;
   while (s < smax) {
	ch = *s++;
	if (is_blank && ((ch & 0xFF) != 32)) is_blank--;

	sum += ch;

	h = sum + (h << 3);
	if ((g = h & 0xE0000000UL) != 0) {
	     h = h ^ (g >> 24);
	     h = h ^ g;
	}
   }
   if (is_blank) return 0;
   return h;
}


static void try_scroll (void)
{
   int i, j, di, r1, r2, rmin, rmax;
   unsigned long hash;
   int color, did_scroll = 0;
   unsigned short *tmp;
   int Ignore;

   /* find region limits. */

   for (rmax = Screen_Rows - 1; rmax > 0; rmax--)
     {
	if (SL_Screen[rmax].new_hash != SL_Screen[rmax].old_hash)
	  {
	     r1 = rmax - 1;
	     if ((r1 == 0)
		 || (SL_Screen[r1].new_hash != SL_Screen[r1].old_hash))
	       break;

	     rmax = r1;
	  }
     }

   for (rmin = 0; rmin < rmax; rmin++)
     {
	if (SL_Screen[rmin].new_hash != SL_Screen[rmin].old_hash)
	  {
	     r1 = rmin + 1;
	     if ((r1 == rmax)
		 || (SL_Screen[r1].new_hash != SL_Screen[r1].old_hash))
	       break;

	     rmin = r1;
	  }
     }


   for (i = rmax; i > rmin; i--)
     {
	hash = SL_Screen[i].new_hash;
	if (hash == Blank_Hash) continue;

	if ((hash == SL_Screen[i].old_hash)
	    || ((i + 1 < Screen_Rows) && (hash == SL_Screen[i + 1].old_hash))
	    || ((i - 1 > rmin) && (SL_Screen[i].old_hash == SL_Screen[i - 1].new_hash)))
	  continue;

	for (j = i - 1; j >= rmin; j--)
	  {
	     if (hash == SL_Screen[j].old_hash) break;
	  }
	if (j < rmin) continue;

	r2 = i;                        /* end scroll region */

	di = i - j;
	j--;
	Ignore = 0;
	while ((j >= rmin) && (SL_Screen[j].old_hash == SL_Screen[j + di].new_hash))
	  {
	     if (SL_Screen[j].old_hash == Blank_Hash) Ignore++;
	     j--;
	  }
	r1 = j + 1;

	/* If this scroll only scrolls this line into place, don't do it.
	 */
	if ((di > 1) && (r1 + di + Ignore == r2)) continue;

	/* If there is anything in the scrolling region that is ok, abort the
	 * scroll.
	 */

	for (j = r1; j <= r2; j++)
	  {
	     if ((SL_Screen[j].old_hash != Blank_Hash)
		 && (SL_Screen[j].old_hash == SL_Screen[j].new_hash))
	       {
		  /* See if the scroll is happens to scroll this one into place. */
		  if ((j + di > r2) || (SL_Screen[j].old_hash != SL_Screen[j + di].new_hash))
		    break;
	       }
	  }
	if (j <= r2) continue;

	color = This_Color;  This_Color = 0;
	did_scroll = 1;
	SLtt_normal_video ();
	SLtt_set_scroll_region (r1, r2);
	SLtt_goto_rc (0, 0);
	SLtt_reverse_index (di);
	SLtt_reset_scroll_region ();
	/* Now we have a hole in the screen.  Make the virtual screen look
	 * like it.
	 */
	for (j = r1; j <= r2; j++) SL_Screen[j].flags = TOUCHED;

	while (di--)
	  {
	     tmp = SL_Screen[r2].old;
	     for (j = r2; j > r1; j--)
	       {
		  SL_Screen[j].old = SL_Screen[j - 1].old;
		  SL_Screen[j].old_hash = SL_Screen[j - 1].old_hash;
	       }
	     SL_Screen[r1].old = tmp;
	     blank_line (SL_Screen[r1].old, Screen_Cols, ' ');
	     SL_Screen[r1].old_hash = Blank_Hash;
	     r1++;
	  }
	This_Color = color;
     }
   if (did_scroll) return;

   /* Try other direction */

   for (i = rmin; i < rmax; i++)
     {
	hash = SL_Screen[i].new_hash;
	if (hash == Blank_Hash) continue;
	if (hash == SL_Screen[i].old_hash) continue;

	/* find a match further down screen */
	for (j = i + 1; j <= rmax; j++)
	  {
	     if (hash == SL_Screen[j].old_hash) break;
	  }
	if (j > rmax) continue;

	r1 = i;                        /* beg scroll region */
	di = j - i;                    /* number of lines to scroll */
	j++;                           /* since we know this is a match */

	/* find end of scroll region */
	Ignore = 0;
	while ((j <= rmax) && (SL_Screen[j].old_hash == SL_Screen[j - di].new_hash))
	  {
	     if (SL_Screen[j].old_hash == Blank_Hash) Ignore++;
	     j++;
	  }
	r2 = j - 1;                    /* end of scroll region */

	/* If this scroll only scrolls this line into place, don't do it.
	 */
	if ((di > 1) && (r1 + di + Ignore == r2)) continue;

	/* If there is anything in the scrolling region that is ok, abort the
	 * scroll.
	 */

	for (j = r1; j <= r2; j++)
	  {
	     if ((SL_Screen[j].old_hash != Blank_Hash)
		 && (SL_Screen[j].old_hash == SL_Screen[j].new_hash))
	       {
		  if ((j - di < r1) || (SL_Screen[j].old_hash != SL_Screen[j - di].new_hash))
		    break;
	       }

	  }
	if (j <= r2) continue;

	color = This_Color;  This_Color = 0;
	SLtt_normal_video ();
	SLtt_set_scroll_region (r1, r2);
	SLtt_goto_rc (0, 0);           /* relative to scroll region */
	SLtt_delete_nlines (di);
	SLtt_reset_scroll_region ();
	/* Now we have a hole in the screen.  Make the virtual screen look
	 * like it.
	 */
	for (j = r1; j <= r2; j++) SL_Screen[j].flags = TOUCHED;

	while (di--)
	  {
	     tmp = SL_Screen[r1].old;
	     for (j = r1; j < r2; j++)
	       {
		  SL_Screen[j].old = SL_Screen[j + 1].old;
		  SL_Screen[j].old_hash = SL_Screen[j + 1].old_hash;
	       }
	     SL_Screen[r2].old = tmp;
	     blank_line (SL_Screen[r2].old, Screen_Cols, ' ');
	     SL_Screen[r2].old_hash = Blank_Hash;
	     r2--;
	  }
	This_Color = color;
     }
}


static void SLsmg_refresh (void)
{
   int i;

   if (Smg_Inited == 0) return;
#ifndef pc_system
   for (i = 0; i < Screen_Rows; i++) {
	if (SL_Screen[i].flags == 0) continue;
	SL_Screen[i].new_hash = compute_hash (SL_Screen[i].neew, Screen_Cols);
   }
#endif

   if (Cls_Flag) {
	SLtt_normal_video ();  SLtt_cls ();
   }
#ifndef pc_system
   else if (SLtt_Term_Cannot_Scroll == 0) try_scroll ();
#endif

   for (i = 0; i < Screen_Rows; i++) {
	int trashed;

	if (SL_Screen[i].flags == 0) continue;

	if (SL_Screen[i].flags & TRASHED) {
	     SLtt_goto_rc (i, -1); /* Force cursor to move */
	     SLtt_goto_rc (i, 0);
	     if (Cls_Flag == 0) SLtt_del_eol ();
	     trashed = 1;
	} else
	     trashed = 0;

	if (Cls_Flag || trashed) {
	     int color = This_Color;
	     This_Color = 0;
	     blank_line (SL_Screen[i].old, Screen_Cols, ' ');
	     This_Color = color;
	}

	SL_Screen[i].old[Screen_Cols] = 0;
	SL_Screen[i].neew[Screen_Cols] = 0;

	SLtt_smart_puts (SL_Screen[i].neew, SL_Screen[i].old, Screen_Cols, i);

	memcpy ((char *) SL_Screen[i].old, (char *) SL_Screen[i].neew,
		  Screen_Cols * sizeof (short));

	SL_Screen[i].flags = 0;
#ifndef pc_system
	SL_Screen[i].old_hash = SL_Screen[i].new_hash;
#endif
   }

   if (point_visible (1)) SLtt_goto_rc (This_Row - Start_Row, This_Col - Start_Col);
   SLtt_flush_output ();
   Cls_Flag = 0;
}

static int compute_clip (int row, int n, int box_start, int box_end,
			 int *rmin, int *rmax)
{
   int row_max;

   if (n < 0) return 0;
   if (row >= box_end) return 0;
   row_max = row + n;
   if (row_max <= box_start) return 0;

   if (row < box_start) row = box_start;
   if (row_max >= box_end) row_max = box_end;
   *rmin = row;
   *rmax = row_max;
   return 1;
}

static void SLsmg_touch_lines0 (int n)
{
   int i;
   int r1, r2;

   if (0 == compute_clip (0, n, Start_Row, Start_Row + Screen_Rows, &r1, &r2))
     return;

   r1 -= Start_Row;
   r2 -= Start_Row;
   for (i = r1; i < r2; i++)
	SL_Screen[i].flags |= TRASHED;
}

static const char Fake_Alt_Char_Pairs [] = "a:j+k+l+m+q-t+u+v+w+x|";

static void init_alt_char_set (void)
{
   int i;
   const unsigned char *p, *pmax;
   unsigned char ch;

   if (Alt_Char_Set[128] == 128) return;

   i = 32;
   memset (Alt_Char_Set, ' ', i);
   while (i <= 128) {
	Alt_Char_Set [i] = i;
	i++;
   }

   /* Map to VT100 */
   if (SLtt_Has_Alt_Charset) {
	p = (unsigned char *) SLtt_Graphics_Char_Pairs;
	if (p == NULL) return;
   }
   else p = (unsigned char *) Fake_Alt_Char_Pairs;
   pmax = p + strlen ((char *) p);

   /* Some systems have messed up entries for this */
   while (p < pmax) {
	ch = *p++;
	ch &= 0x7F;                    /* should be unnecessary */
	Alt_Char_Set [ch] = *p;
	p++;
   }
}

static int SLsmg_init_smg (void)
{
   int i, len;
   unsigned short *old, *neew;

   SLsig_block_signals();

   if (Smg_Inited) SLsmg_reset_smg ();
   /*   send_string_to_term("\033[?6h"); */
   /* relative origin mode */
   SLtt_write_string (Term_Init_Str);
   SLtt_write_string (Keypad_Init_Str);
   SLtt_reset_scroll_region();
   SLtt_end_insert();
   SLtt_write_string (Enable_Alt_Char_Set);
   Video_Initialized = 1;
   Screen_Cols = SLtt_Screen_Cols;
   Screen_Rows = SLtt_Screen_Rows;
   This_Col = This_Row = Start_Col = Start_Row = 0;

   This_Color = 0;
   This_Alt_Char = 0;
   Cls_Flag = 1;
#ifndef pc_system
   init_alt_char_set ();
#endif
   len = Screen_Cols + 3;
   for (i = 0; i < Screen_Rows; i++) {
	if ((NULL == (old = (unsigned short *) malloc (sizeof(short) * len)))
	    || ((NULL == (neew = (unsigned short *) malloc (sizeof(short) * len)))))
	  {
	     SLsig_unblock_signals();
	     return 0;
	  }
	blank_line (old, len, ' ');
	blank_line (neew, len, ' ');
	SL_Screen[i].old = old;
	SL_Screen[i].neew = neew;
	SL_Screen[i].flags = 0;
#ifndef pc_system
	Blank_Hash = compute_hash (old, Screen_Cols);
	SL_Screen[i].new_hash = SL_Screen[i].old_hash =  Blank_Hash;
#endif
   }
   Smg_Inited = 1;
   SLsig_unblock_signals();
   return 1;
}


static void SLsmg_reset_smg (void)
{
   int i;

   SLsig_block_signals();

   if (Smg_Inited == 0) {
	SLsig_unblock_signals();
	return;
   }
   for (i = 0; i < Screen_Rows; i++) {
	if (SL_Screen[i].old != NULL) free (SL_Screen[i].old);
	if (SL_Screen[i].neew != NULL) free (SL_Screen[i].neew);
	SL_Screen[i].old = SL_Screen[i].neew = NULL;
   }
   SLtt_reset_video ();
   This_Alt_Char = This_Color = 0;
   Smg_Inited = 0;

   SLsig_unblock_signals();
}


static void SLsmg_draw_object (int r, int c, unsigned char object)
{
   This_Row = r;  This_Col = c;

   if (point_visible (1)) {
	int color = This_Color;
	This_Color |= ALT_CHAR_FLAG;
	SLsmg_write_char (object);
	This_Color = color;
   }

   This_Col = c + 1;
}

static void SLsmg_draw_hline (int n)
{
   char hbuf[16];
   int Count;
   int cmin, cmax;
   int final_col = This_Col + n;
   int save_color;

   if ((This_Row < Start_Row) || (This_Row >= Start_Row + Screen_Rows)
       || (0 == compute_clip (This_Col, n, Start_Col, Start_Col + Screen_Cols,
			      &cmin, &cmax)))
     {
	This_Col = final_col;
	return;
     }

   memset (hbuf, SLSMG_HLINE_CHAR, 16);

   n = cmax - cmin;
   Count = n / 16;

   save_color = This_Color;
   This_Color |= ALT_CHAR_FLAG;
   This_Col = cmin;

   SLsmg_write_nchars (hbuf, n % 16);
   while (Count-- > 0)
	SLsmg_write_nchars (hbuf, 16);

   This_Color = save_color;
   This_Col = final_col;
}

static void SLsmg_draw_vline (int n)
{
   unsigned char ch = SLSMG_VLINE_CHAR;
   int c = This_Col, rmin, rmax;
   int final_row = This_Row + n;
   int save_color;

   if (((c < Start_Col) || (c >= Start_Col + Screen_Cols)) ||
       (0 == compute_clip (This_Row, n, Start_Row, Start_Row + Screen_Rows,
			  &rmin, &rmax)))
     {
	This_Row = final_row;
	return;
     }

   save_color = This_Color;
   This_Color |= ALT_CHAR_FLAG;

   for (This_Row = rmin; This_Row < rmax; This_Row++) {
	This_Col = c;
	SLsmg_write_nchars ((char *) &ch, 1);
   }

   This_Col = c;  This_Row = final_row;
   This_Color = save_color;
}

static void SLsmg_draw_box (int r, int c, int dr, int dc)
{
   if (!dr || !dc) return;
   This_Row = r;  This_Col = c;
   dr--; dc--;
   SLsmg_draw_hline (dc);
   SLsmg_draw_vline (dr);
   This_Row = r;  This_Col = c;
   SLsmg_draw_vline (dr);
   SLsmg_draw_hline (dc);
   SLsmg_draw_object (r, c, SLSMG_ULCORN_CHAR);
   SLsmg_draw_object (r, c + dc, SLSMG_URCORN_CHAR);
   SLsmg_draw_object (r + dr, c, SLSMG_LLCORN_CHAR);
   SLsmg_draw_object (r + dr, c + dc, SLSMG_LRCORN_CHAR);
   This_Row = r; This_Col = c;
}


typedef struct
{
   const char name[3];
   int offset;
} Tgetstr_Map_Type;

/* I need to add: K1-5, %0-5(not important), @8, &8... */
static const Tgetstr_Map_Type Tgetstr_Map [] =
{
   {"@7", 164 },
   {"AB", 360 },
   {"AF", 359 },
   {"AL", 110 },
   {"DL", 106 },
   {"RI", 112 },
   {"Sf", 302 },
   {"Sb", 303 },
   {"ac", 146 },
   {"ae", 38  },
   {"as", 25  },
   {"ce", 6   },
   {"cl", 5   },
   {"cm", 10  },
   {"cs", 3   },
   {"dc", 21  },
   {"ds", 23  },
   {"eA", 155 },
   {"ei", 42  },
   {"fs", 47  },
   {"im", 31  },
   {"k0", 65  },
   {"k1", 66  },
   {"k2", 68  },
   {"k3", 69  },
   {"k4", 70  },
   {"k5", 71  },
   {"k6", 72  },
   {"k7", 73  },
   {"k8", 74  },
   {"k9", 75  },
   {"kA", 78  },
   {"kC", 57  },
   {"kD", 59  },
   {"kE", 63  },
   {"kF", 84  },
   {"kH", 80  },
   {"kI", 77  },
   {"kL", 60  },
   {"kM", 62  },
   {"kN", 81  },
   {"kP", 82  },
   {"kR", 85  },
   {"kS", 64  },
   {"kT", 86  },
   {"ka", 56  },
   {"k;", 67  },
   {"kb", 55  },
   {"kd", 61  },
   {"ke", 88  },
   {"kh", 76  },
   {"kl", 79  },
   {"kr", 83  },
   {"ks", 89  },
   {"kt", 58  },
   {"ku", 87  },
   {"mb", 26  },
   {"md", 27  },
   {"me", 39  },
   {"mr", 34  },
   {"op", 297 },
   {"pf", 119 },
   {"po", 120 },
   {"se", 43  },
   {"so", 35  },
   {"sr", 130 },
   {"te", 40  },
   {"ti", 28  },
   {"ts", 135 },
   {"up", 19  },
   {"us", 36  },
   {"vb", 45  },
   {"ve", 16  },
   {"vi", 13  },
   {"vs", 20  },
   {"", 0     }
};

static int compute_cap_offset (const char *cap, const Tgetstr_Map_Type *map, unsigned int max_ofs)
{
   char cha, chb;

   cha = *cap++; chb = *cap;

   while (*map->name != 0) {
	if ((cha == *map->name) && (chb == *(map->name + 1))) {
	     if (map->offset >= (int) max_ofs) return -1;
	     return map->offset;
	}
	map++;
   }
   return -1;
}


static const char *SLtt_tigetstr (const char *cap)
{
   int offset;
   Terminfo_Type *t = &Tbuf;

   offset = compute_cap_offset (cap, Tgetstr_Map, t->num_string_offsets);
   if (offset < 0) return NULL;
   offset = make_integer (t->string_offsets + 2 * offset);
   if (offset < 0) return NULL;
   return t->string_table + offset;
}

static const Tgetstr_Map_Type Tgetnum_Map[] =
{
   {"co", 0   },
   {"li", 2   },
   {"Co", 13  },
   {"pa", 14  },
   {"sg", 4   },
   {"ws", 7   },
   {"", -1    }
};

static int SLtt_tgetnum (const char *cap)
{
   int offset;
   Terminfo_Type *t = &Tbuf;

   offset = compute_cap_offset (cap, Tgetnum_Map, t->num_numbers);
   if (offset < 0) return -1;
   return make_integer (t->numbers + 2 * offset);
}

static const Tgetstr_Map_Type Tgetflag_Map[] =
{
   {"am", 1  },
   {"hs", 9  },
   {"ms", 14 },
   {"xs", 3  },
   {"xn", 4  },
   {"es", 16 },
   {"", -1   }
};

static int SLtt_tigetflag (const char *cap)
{
   int offset;
   Terminfo_Type *t = &Tbuf;

   offset = compute_cap_offset (cap, Tgetflag_Map, t->boolean_section_size);
   if (offset < 0) return -1;
   return (int) *(t->boolean_flags + offset);
}


/* slutty.c --- Unix Low level terminal (tty) functions for S-Lang */
# define GET_TERMIOS(fd, x) tcgetattr(fd, x)
# define SET_TERMIOS(fd, x) tcsetattr(fd, TCSADRAIN, x)


#ifdef _POSIX_VDISABLE
#  define NULL_VALUE _POSIX_VDISABLE
# else
#  define NULL_VALUE 255
#endif

static int SLang_init_tty (void)
{
   TTY_Termio_Type newtty;

   SLsig_block_signals();

   TTY_Open = 0;

   if ((SLang_TT_Read_FD == -1) || (1 != isatty (SLang_TT_Read_FD))) {
	if (TTY_Open == 0) {
	    SLang_TT_Read_FD = fileno (stdin);
	    if (1 != isatty (SLang_TT_Read_FD)) {
		bb_error_msg ("Failed to open terminal.");
		return -1;
	    }
	}
   }

   /* Some systems may not permit signals to be blocked.  As a result, the
    * return code must be checked.
    */
   while (-1 == GET_TERMIOS(SLang_TT_Read_FD, &Old_TTY)) {
	if (errno != EINTR) {
	     SLsig_unblock_signals();
	     return -1;
	}
   }

   while (-1 == GET_TERMIOS(SLang_TT_Read_FD, &newtty)) {
	if (errno != EINTR) {
	     SLsig_unblock_signals();
	     return -1;
	}
   }

   /* get baud rate */

   /* [not only QNX related !?!]
    * ECHO(0x08) is a c_lflag bit, it means here PARMRK(0x08) in c_iflag!!!
    */
   /*newtty.c_iflag &= ~(ECHO | INLCR | ICRNL);*/
   newtty.c_iflag &= ~(INLCR | ICRNL);
   /* newtty.c_iflag &= ~ISTRIP; */
   newtty.c_oflag &= ~OPOST;

   if (SLang_TT_Baud_Rate == 0)
	SLang_TT_Baud_Rate = bb_baud_to_value (cfgetospeed (&newtty));
   newtty.c_iflag &= ~IXON;

   newtty.c_cc[VMIN] = 1;
   newtty.c_cc[VTIME] = 0;
   newtty.c_cc[VEOF] = 1;
   newtty.c_lflag = ISIG | NOFLSH;
   newtty.c_cc[VINTR] = XCTRL('c');
   newtty.c_cc[VQUIT] = NULL_VALUE;
   newtty.c_cc[VSUSP] = NULL_VALUE;   /* to ignore ^Z */
#ifdef VSWTCH
   newtty.c_cc[VSWTCH] = NULL_VALUE;   /* to ignore who knows what */
#endif

   while (-1 == SET_TERMIOS(SLang_TT_Read_FD, &newtty)) {
	if (errno != EINTR) {
	     SLsig_unblock_signals();
	     return -1;
	}
   }

   SLsig_unblock_signals();
   return 0;
}

static void SLang_reset_tty (void)
{
   SLsig_block_signals();

   while ((-1 == SET_TERMIOS(SLang_TT_Read_FD, &Old_TTY))
	  && (errno == EINTR))
     ;

   if (TTY_Open) {
	while ((-1 == close (SLang_TT_Read_FD))
	       && (errno == EINTR))
	  ;

	TTY_Open = 0;
	SLang_TT_Read_FD = -1;
   }

   SLsig_unblock_signals();
}


#if ENABLE_FEATURE_LONG_OPTIONS
static const struct option mc_long_options[] = {
	{"color", 0, 0, 'c'},
	{"colors", 1, 0, 'C'},
	{"edit", 1, 0, 'e'},
	{"help-colors", 0, 0, 'H'},
#ifdef USE_NETCODE
	{"ftplog", 1, 0, 'l'},
#endif
	{"libdir", 0, 0, 'f'},
	{"nocolor", 0, 0, 'b'},
	{"nomouse", 0, 0, 'd'},
	{"printwd", 0, 0, 'P'},
	{"resetsoft", 0, 0, 'k'},
	{"slow", 0, 0, 's'},
	{"stickchars", 0, 0, 'a'},
	{"version", 0, 0, 'V'},
	{"xterm", 0, 0, 'x'},
	{0, 0, 0, 0}
};
#endif /* long option */

static void
print_color_usage (void)
{
     fprintf (stderr,
	     "--colors KEYWORD={FORE},{BACK}\n\n"
	     "{FORE} and {BACK} can be ommited, and the default will be used\n"
	     "\n"
	     "Keywords:\n"
	     "   Global:       errors, reverse, gauge, input\n"
	     "   File display: normal, selected, marked, markselect\n"
	     "   Dialog boxes: dnormal, dfocus, dhotnormal, dhotfocus\n"
	     "   Menus:        menu, menuhot, menusel, menuhotsel\n"
	     "   Help:         helpnormal, helpitalic, helplink, helpslink\n"
	     "   File types:   directory, execute, link, device, special, core\n"
	     "\n"
	     "Colors:\n"
	     "   black, gray, red, brightred, green, brightgreen, brown,\n"
	     "   yellow, blue, brightblue, magenta, brightmagenta, cyan,\n"
	     "   brightcyan, lightgray and white\n\n");
}

/* int main (int argc, char *argv []) */
int mc_main(void)
{
    int    c;
#ifdef USE_NETCODE
    char *s_ftpfs_set_debug = NULL;
#endif
    int finish_program = 0;

    my_uid = getuid ();
    my_euid = geteuid ();
    my_gid = getgid ();

    MC_verbose = 1;
    mc_xstat = mc_lstat;
    op_preserve = 1;
    umask_kill = 0777777;
    use_mouse_p = GPM_MOUSE;
    clear_before_exec = 1;
    confirm_delete = 1;
    confirm_overwrite = 1;
    confirm_exit = 1;
    use_internal_edit = 1;
    auto_save_setup = 1;
    eight_bit_clean = 1;
    full_eight_bits = 1;
    use_internal_view = 1;
    mouse_move_pages = 1;
    mou_auto_repeat = 100;
    double_click_speed = 250;
    use_8th_bit_as_meta = 1;
    cd_symlinks = 1;
    mark_moves_down = 1;
    only_leading_plus_minus = 1;
    boot_current_is_left = 1;
    use_file_to_check_type = 1;
#ifdef USE_VFS
    vfs_use_limit = 1;
#endif
    nice_rotating_dash = 1;
    equal_split = 1;
    command_prompt = 1;
    keybar_visible = 1;
    message_visible = 1;
    cons_saver_pid = 1;
    mouse_move_pages_viewer = 1;
    global_wrap_mode = 1;
    default_magic_flag = 1;
    default_nroff_flag = 1;
    dirt_limit = 1;
    get_event_dirty = 3;
    easy_patterns = 1;
    esctime.tv_sec = esctime.tv_usec = -1;
    lastnodelay = -1;
    filetype_mode = 1;
    show_mini_info = 1;
    panel_scroll_pages = 1;
    fmd_xlen = FMD_XLEN;
    help_lines = 18;
    menubar_visible = 1;
    num_history_items_recorded = 60;
    ok_to_refresh = 1;
    pause_after_run = pause_on_dumb_terminals;
    /* If true show files starting with a dot */
    show_dot_files = 1;
    /* Reverse flag */
    reverse_sort_flag = 1;
    user_hotkey = 'u';

    viewer = getenv ("PAGER");
    if (!viewer)
	viewer = "view";

    /* Initialize list of all user group for timur_clr_mode */
    init_groups ();

    OS_Setup ();

    vfs_init ();

    /* NOTE: This has to be called before slang_init or whatever routine
       calls any define_sequence */
    init_key ();

#if ENABLE_FEATURE_LONG_OPTIONS
    bb_applet_long_options = mc_long_options;
#endif

    c = bb_getopt_ulflags("cHfbdPksaxVmC:e:"
#ifdef USE_NETCODE
	"l:"
#endif
		, &command_line_colors, &edit_one_file
#ifdef USE_NETCODE
		, s_ftpfs_set_debug
#endif
			);
    if((c & 1)) {
	/* 'c' */
	disable_colors = 0;
	force_colors = 1;
    }
    if((c & 2)) {
	/* 'H' */
	print_color_usage ();
	finish_program = 1;
    }
    if((c & 4)) {
	/* 'f' */
	bb_error_msg ("Library directory for the Midnight Commander: %s", mc_home);
	finish_program = 1;
    }
    disable_colors = c & 8;     /* 'b' */

    if((c & 16)) {
	/* 'd' */
	use_mouse_p = NO_MOUSE;
    }

    print_last_wd = c & 32;     /* 'P' */
    reset_hp_softkeys = c & 64;     /* 'k' */
    slow_terminal = c & 128;     /* 's' */
    force_ugly_line_drawing = c & 256;     /* 'a' */
    force_xterm = c & 512;     /* 'x' */

    if((c & 1024)) {
	/* 'V' */
	version ();
	finish_program = 1;
    }
    if((c & 2048)) {
	/* 'm' */
	bb_error_msg ("Option -m is obsolete. Please look at Display Bits... in the Option's menu");
	finish_program = 1;
    }
#ifdef USE_NETCODE
    if(s_ftpfs_set_debug != NULL)
	ftpfs_set_debug (option_arg);
#endif

    if(finish_program) {
	if (print_last_wd)
	    printf (".");
	return 1;
    }

    /* Check for special invocation names mcedit and mcview,
     * if none apply then set the current directory and the other
     * directory from the command line arguments
     */
    if(*Gargv != NULL)
	*Gargv = xstrdup(*Gargv);

    if (!strncmp (bb_applet_name, "mce", 3) || !strcmp(bb_applet_name, "vi")) {
	edit_one_file = *Gargv == NULL ? "" : *Gargv;
    } else {
	/* sets the current dir and the other dir */
	if (*Gargv != NULL) {
	    this_dir = *Gargv;
	    Gargv++;
	    if (*Gargv != NULL)
		other_dir = xstrdup(*Gargv);
	}
    }

    /* Used to report the last working directory at program end */
    if (print_last_wd) {
	stdout_fd = dup (1);
	close (1);
	if (open (ttyname (0), O_RDWR) < 0)
	    if (open (bb_dev_tty, O_RDWR) < 0) {
	    /* Try if stderr is not redirected as the last chance */
		char *p = strdup (ttyname (0));

		if (!strcmp (p, ttyname (2)))
		    dup2 (2, 1);
		else {
		    bb_error_msg_and_die (
			     "Couldn't open tty line. You have to run mc without the -P flag.\n"
			       "On some systems you may want to run # `which mc`");
		}
		free (p);
	    }
    }

    sort_case_sensitive = OS_SORT_CASE_SENSITIVE_DEFAULT;

    DISPLAY_X = 45;
    DISPLAY_Y = 14;
    Scroll_r2 = 23;
    init_look_for_rxvt_extensions();
    /* Must be done before installing the SIGCHLD handler [[FIXME]] */
    pipefd1 [0] = -1;
    pipefd1 [1] = -1;
    pipefd2 [0] = -1;
    pipefd2 [1] = -1;
    SLang_TT_Read_FD = -1;
    handle_console (CONSOLE_INIT);

    /* Install the SIGCHLD handler; must be done before init_subshell() */
    init_sigchld ();

    /* This variable is used by the subshell */
    home_dir = getenv ("HOME");
    if (!home_dir)
	home_dir = "/";

    compatibility_move_mc_files ();

    /* Must be done before init_subshell, to set up the terminal size: */
    /* FIXME: Should be removed and LINES and COLS computed on subshell */
    slang_init ();

    load_setup ();

    init_curses ();

    /* Removing this from the X code let's us type C-c */
    load_key_defs ();

    /* Also done after init_subshell, to save any shell init file messages */
    if (console_flag)
	handle_console (CONSOLE_SAVE);

    if (alternate_plus_minus)
	application_keypad_mode ();

    /* The directory hot list */
    load_hotlist ();

    if (show_change_notice) {
	MC_message (1, " Notice ",
		 " The Midnight Commander configuration files \n"
		   " are now stored in the ~/.mc directory, the \n"
		   " files have been moved now\n");
    }

    MC_prompt = (my_euid == 0) ? "# " : "$ ";

    /* Program main loop */
    do_nc ();

    /* Virtual File System shutdown */
    vfs_shut ();

    /* Delete list of all user groups */
    delete_groups ();

    flush_extension_file (); /* does only free memory */

    /* Miguel, maybe the fix in slang is not required and
     * it could be done by removing the slang_done_screen.
     * Do I need to call slang_reset_tty then?
     */
    endwin ();
    slang_shutdown ();

    if (console_flag && !(MC_quit & SUBSHELL_EXIT))
	restore_console ();
    if (alternate_plus_minus)
	numeric_keypad_mode ();

    bb_signal((1 << SIGCHLD), SIG_DFL);  /* Disable the SIGCHLD handler */

    if (console_flag)
	handle_console (CONSOLE_DONE);
    putchar ('\r');  /* Hack to make shell's prompt start at left of screen */

    if (print_last_wd) {
	if (print_last_revert || edit_one_file)
	    last_wd_string = ".";
	write (stdout_fd, last_wd_string, strlen (last_wd_string));
    }

    return 0;
}
#endif /* APPLET_USAGE */
